/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.config;

import java.io.File;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.runtime.model.DefinitionAndValue;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.model.type.xml.XMLElementType;

/**
 * Bean to read / Write config file
 */
public final class Config
{
    // Logger for traces
    private static Logger __logger = LoggerFactory.getLogger(Config.class);
    
    // shared instance for optimization
    private static Config __config;
    
    // Initialization status
    private static boolean __initialized;
    
    // config file
    private static String __filename;
    
    // true if the config file does exist
    private static boolean __fileExists;

    // The last modification date
    private static long __lastModified = -1;
    
    // The configuration model
    private static Model __model;
    
    // Typed value (filled after read)
    private Map<String, DefinitionAndValue> _definitionAndValues;
    
    private Config()
    {
    }

    /**
     * Get the instance of Config using the config file.
     * @return An instance of Config containing config file values and definition, or null if config file cannot be read.
     */
    public static Config getInstance()
    {
        if (!__initialized)
        {
            return null;
        }
        
        if (__config == null)
        {
            try
            {
                __config = new Config();
                __config._load();
            }
            catch (Exception e)
            {
                __logger.warn("Exception creating Config, it won't be accessible.", e);
                return null;
            }
        }
        else
        {
            __config._reloadIfNeeded();
        }
        
        return __config;
    }
    
    /**
     * Gets the configuration model
     * @return the model or null if the configuration is not initialized
     */
    public static Model getModel()
    {
        return __model;
    }
    
    /**
     * Sets the configuration model
     * @param model the model to set
     */
    public static void setModel(Model model)
    {
        __model = model;
    }
    
    private void _load() throws Exception
    {
        __logger.info("Loading configuration values from file {}.", __filename);
        _definitionAndValues = __read();
    }
    
    private void _reloadIfNeeded()
    {
        if (__fileExists && new File(__filename).exists() && __lastModified < new File(__filename).lastModified())
        {
            try
            {
                __logger.info("The config file has changed. Let's reload."); 
                _load();
            }
            catch (Exception e)
            {
                // __lastModified was changed, so we will not fail several times
                // _values was not modified
                // __fileExists is still true
                __logger.error("The config file '" + __filename + "' was modified but could not be reloaded due to an exception", e);
            }
        }
    }
    
    /**
     * Read configuration file
     * @return the configuration definition and values
     * @throws Exception if a problem occurs reading values
     */
    static Map<String, DefinitionAndValue> __read() throws Exception
    {
        Map<String, DefinitionAndValue> definitionAndValues = new HashMap<>();

        File configFile = new File(__filename);
        __fileExists = configFile.exists();

        if (__fileExists)
        {
            __lastModified  = configFile.lastModified();

            Configuration configuration = new DefaultConfigurationBuilder().buildFromFile(configFile);
            for (Configuration parameter : configuration.getChildren())
            {
                String parameterName = parameter.getName();
                if (__model.hasModelItem(parameterName))
                {
                    ModelItem modelItem = __model.getModelItem(parameterName);
    
                    if (modelItem instanceof ElementDefinition)
                    {
                        ElementDefinition definition = (ElementDefinition) modelItem;
                        ElementType type = definition.getType();
                        
                        if (type instanceof XMLElementType)
                        {
                            Object value = ((XMLElementType) type).read(configuration, parameterName);
                            DefinitionAndValue definitionAndValue = new DefinitionAndValue<>(null, definition, value);
                            definitionAndValues.put(parameterName, definitionAndValue);
                        }
                    }
                }
                else
                {
                    __logger.warn("The parameter {} is not defined in configuration. This parameter is ignored.", parameterName);
                }
            }
        }
        else
        {
            __lastModified = -1;
        }

        return definitionAndValues;
    }
    
    /**
     * Set the initialization status of the configuration
     * @param initialized the initialization status of the configuration
     */
    public static void setInitialized(boolean initialized)
    {
        __initialized = initialized;
    }

    /**
     * Set the configuration filename
     * @param filename Name with path of the configuration file
     */
    public static void setFilename(String filename)
    {
        __filename = filename;
        __fileExists = new File(__filename).exists();
    }
    
    /**
     * Returns true if the configuration filename exists.
     * @return true if the configuration filename exists.
     */
    public static boolean fileExists()
    {
        return __fileExists;
    }
    
    /**
     * Retrieves all configuration parameter values
     * @return the configuration values
     */
    public Map<String, Object> getValues()
    {
        return __extractValues(_definitionAndValues);
    }
    
    static Map<String, Object> __extractValues(Map<String, DefinitionAndValue> definitionAndValues)
    {
        return DefinitionAndValue.extractValues(definitionAndValues);
    }
    
    /**
     * Retrieves all configuration parameter values as JSON objects
     * @return the configuration values as JSON objects
     */
    public static Map<String, Object> getValuesAsJSONForClient()
    {
        Map<String, Object> jsonValues = new HashMap<>();
        
        Set<Entry<String, DefinitionAndValue>> values;
        
        try
        {
            values = __read().entrySet();
        }
        catch (Exception e)
        {
            __logger.warn("Config values are unreadable. Using default values", e);
            return jsonValues;
        }
        
        for (Map.Entry<String, DefinitionAndValue> entry : values)
        {
            DefinitionAndValue definitionAndValue = entry.getValue();
            
            Object value = definitionAndValue.getValue();
            ElementDefinition definition = (ElementDefinition) definitionAndValue.getDefinition();
            
            ElementType<Object> type = definition.getType();
            Object jsonValue = type.valueToJSONForClient(value, DataContext.newInstance());
            
            jsonValues.put(entry.getKey(), jsonValue);
        }
        
        return jsonValues;
        
    }
    
    /**
     * Retrieves the typed parameter value
     * @param <T> type of the value to retrieve
     * @param name name of the configuration parameter
     * @return the parameter value
     */
    @SuppressWarnings("unchecked")
    public <T extends Object> T getValue(String name)
    {
        DefinitionAndValue definitionAndValue = _definitionAndValues.get(name);
        if (definitionAndValue == null)
        {
            __logger.warn("The configuration parameter {} doesn't exist. A null value is retrieved.", name);
            return null;
        }
        
        return (T) definitionAndValue.getValue();
    }
    
    /**
     * Retrieves the typed parameter value, or the default value
     * @param name name of the configuration parameter
     * @param useDefaultFromModel true to use the default value from the model, false to use the give default value
     * @param defaultValue default value used if value is null and useDefaultFromModel is false, or if there is no default value on model
     * @param <T> type of the value to retrieve
     * @return the parameter value
     */
    public <T> T getValue(String name, boolean useDefaultFromModel, T defaultValue)
    {
        DefinitionAndValue definitionAndValue = _definitionAndValues.get(name);
        if (definitionAndValue == null)
        {
            __logger.warn("The configuration parameter {} doesn't exist. The given default value is retrieved.", name);
            return defaultValue;
        }
        
        @SuppressWarnings("unchecked")
        T value = (T) definitionAndValue.getValue();
        if (value != null)
        {
            return value;
        }
        
        if (useDefaultFromModel)
        {
            ElementDefinition definition = (ElementDefinition) definitionAndValue.getDefinition();
            @SuppressWarnings("unchecked")
            T defaultFromModel = (T) definition.getDefaultValue();
            if (defaultFromModel != null)
            {
                return defaultFromModel;
            }
            __logger.debug("There is no default value in model for the configuration parameter {}. The given default value is retrieved.", name);
        }
        
        return defaultValue;
    }
    
    /**
     * Dispose this {@link Config} instance
     */
    public static void dispose()
    {
        __config = null;
    }
}
