/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.config;

import java.util.Collection;
import java.util.Comparator;

import org.ametys.runtime.model.CategorizedElementDefinitionHelper;
import org.ametys.runtime.model.CategorizedElementDefinitionWrapper;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemGroup;

/**
 * Comparator for the element definitions. Uses the position given by the {@link CategorizedElementDefinitionWrapper}s
 * If no position is set, compares the definitions themselves
 */
public class ConfigParameterDefinitionComparator implements Comparator<ModelItem>
{
    private final Collection<ConfigParameterDefinitionWrapper> _configParameterDefinitionWrappers;
    
    /**
     * Constructor
     * @param configParameterDefinitionWrappers the wrappers containing the position of the config parameter definitions
     */
    public ConfigParameterDefinitionComparator(Collection<ConfigParameterDefinitionWrapper> configParameterDefinitionWrappers)
    {
        _configParameterDefinitionWrappers = configParameterDefinitionWrappers;
    }
    
    public int compare(ModelItem item1, ModelItem item2)
    {
        if (!(item1 instanceof ElementDefinition) || !(item2 instanceof ElementDefinition))
        {
            throw new IllegalArgumentException("This comparator can only be called with element definitions");
        }
        
        ElementDefinition definition1 = (ElementDefinition) item1;
        ConfigParameterDefinitionWrapper wrapper1 = getWrapperFromDefinition((ElementDefinition) item1);
        ElementDefinition definition2 = (ElementDefinition) item2;
        ConfigParameterDefinitionWrapper wrapper2 = getWrapperFromDefinition((ElementDefinition) item2);
        
        int positionComparison = CategorizedElementDefinitionHelper.compareWrapperPositions(wrapper1, wrapper2);
        if (positionComparison != 0)
        {
            return positionComparison;
        }
        else
        {
            return definition1.compareTo(definition2);
        }
    }
    
    /**
     * Retrieves the {@link ConfigParameterDefinitionWrapper} corresponding to the given definition
     * @param definition the definition
     * @return the corresponding {@link ConfigParameterDefinitionWrapper}
     */
    protected ConfigParameterDefinitionWrapper getWrapperFromDefinition(ElementDefinition definition)
    {
        ModelItemGroup category = null;
        ModelItemGroup group = definition.getParent();
        if (group != null)
        {
            category = group.getParent();
        }
        
        for (ConfigParameterDefinitionWrapper wrapper : _configParameterDefinitionWrappers)
        {
            if ((category == null || category.getLabel().equals(wrapper.getDisplayCategory()))
                 && (group == null || group.getLabel().equals(wrapper.getDisplayGroup()))
                 && definition.getName().equals(wrapper.getDefinition().getName()))
            {
                return wrapper;
            }
        }
        
        // no wrapper has been found
        return null;
    }
}
