/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.config;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.CategorizedElementDefinitionHelper;
import org.ametys.runtime.model.CategorizedElementDefinitionWrapper;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Model;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.model.ViewItemGroup;

/**
 * Helper for {@link ConfigParameterDefinitionWrapper}
 */
public final class ConfigParameterDefinitionHelper
{
    private ConfigParameterDefinitionHelper()
    {
        //Nothing
    }
    
    /**
     * Get definitions of all elements parsed in this {@link CategorizedElementDefinitionWrapper} {@link Collection}, as a flat map.
     * @param elements a {@link Collection} of {@link CategorizedElementDefinitionWrapper}
     * @return a map with the definition name as key, containing all definitions
     */
    public static Map<String, ElementDefinition> getFlatDefinitions(Collection<ConfigParameterDefinitionWrapper> elements)
    {
        Map<String, ElementDefinition> flatDefinitions = new HashMap<>();
        for (ConfigParameterDefinitionWrapper element : elements)
        {
            ElementDefinition definition = element.getDefinition();
            flatDefinitions.put(definition.getName(), definition);
        }
        return flatDefinitions;
    }
    
    /**
     * Categorize the ModlItems based on the informations in the {@link CategorizedElementDefinitionWrapper} collection
     * Category and groups are not sorted, only definitions, based on the position.
     * Sorting of category and groups have to be done when the view is created (each usage can be different)
     * @param elements collection of categorized elements
     * @return a list of {@link ModelItem} with the correct tree (category/group/definitions)
     */
    public static List<ModelItem> categorizeConfigParameters(Collection<ConfigParameterDefinitionWrapper> elements)
    {
        Model model = ConfigManager.getInstance();
        Map<I18nizableText, Map<I18nizableText, List<ConfigParameterDefinitionWrapper>>> categories = CategorizedElementDefinitionHelper.categorizeElementDefinitionWrappers(elements);
        //This list is not sorted, only definition are sorted, not groups.
        //Groups should be sorted elsewhere (when creating the view for example)
        List<ModelItem> modelItems = new ArrayList<>();
        for (Entry<I18nizableText, Map<I18nizableText, List<ConfigParameterDefinitionWrapper>>> categoryEntry : categories.entrySet())
        {
            ModelItemGroup category = new ModelItemGroup();
            category.setModel(model);
            category.setLabel(categoryEntry.getKey());
            modelItems.add(category);
            Map<I18nizableText, List<ConfigParameterDefinitionWrapper>> values = categoryEntry.getValue();
            for (Entry<I18nizableText, List<ConfigParameterDefinitionWrapper>> groupEntry : values.entrySet())
            {
                ModelItemGroup group = new ModelItemGroup();
                group.setModel(model);
                group.setLabel(groupEntry.getKey());
                category.addChild(group);
                
                for (ConfigParameterDefinitionWrapper orderedDefinition : groupEntry.getValue())
                {
                    ElementDefinition definition = orderedDefinition.getDefinition();
                    group.addChild(definition, orderedDefinition.isGroupSwitch());
                }
            }
        }
        return modelItems;
    }
    
    /**
     * Generate the view for a list of {@link ModelItem}, using comparators to sort the categories and the groups
     * This works only on the very specific categories/group/fieldset hierarchy and an {@link IllegalArgumentException} will be thrown if the hierarchy is incorrect.
     * @param categories list of {@link ModelItem}
     * @param categoriesComparator {@link Comparator} used for the categories (can be null to keep order)
     * @param groupsComparator {@link Comparator} for the groups (can be null to keep order)
     * @param elementsComparator {@link Comparator} for the elements (can be null to keep order)
     * @return A {@link View} ordered from the {@link ModelItem} using the {@link Comparator}
     * @throws IllegalArgumentException the hierarchy of the categories/groups/elements is incorrect
     */
    public static View buildConfigParametersView(Collection<? extends ModelItem> categories, Comparator<? super ModelItem> categoriesComparator, Comparator<? super ModelItem> groupsComparator, Comparator<? super ModelItem> elementsComparator) throws IllegalArgumentException
    {
        View view = new View();
        
        Collection<? extends ModelItem> sorted = _sort(categories, categoriesComparator);
        
        for (ModelItem modelItem : sorted)
        {
            view.addViewItem(_buildCategoryViewItem(modelItem, groupsComparator, elementsComparator));
        }
        return view;
    }
    
    private static ModelViewItemGroup _buildCategoryViewItem(ModelItem modelItem, Comparator<? super ModelItem> groupsComparator, Comparator<? super ModelItem> elementsComparator) throws IllegalArgumentException
    {
        if (modelItem instanceof ModelItemGroup)
        {
            ModelItemGroup category = (ModelItemGroup) modelItem;
            ModelViewItemGroup<ModelItemGroup> categoryViewItem = new ModelViewItemGroup<>();
            categoryViewItem.setRole(ViewItemGroup.TAB_ROLE);
            categoryViewItem.setDefinition(category);
            
            Collection<? extends ModelItem> groups = _sort(category.getChildren(), groupsComparator);
            
            for (ModelItem group : groups)
            {
                categoryViewItem.addViewItem(_buildGroupViewItem(group, elementsComparator));
            }
            return categoryViewItem;
        }
        else
        {
            throw new IllegalArgumentException("Category " + modelItem.getPath() + " should be an instance of ModelItemGroup");
        }
    }
    
    private static ModelViewItemGroup _buildGroupViewItem(ModelItem modelItem, Comparator<? super ModelItem> elementsComparator) throws IllegalArgumentException
    {
        if (modelItem instanceof ModelItemGroup)
        {
            ModelItemGroup group = (ModelItemGroup) modelItem;
            ModelViewItemGroup<ModelItemGroup> groupViewItem = new ModelViewItemGroup<>();
            groupViewItem.setRole(ViewItemGroup.FIELDSET_ROLE);
            groupViewItem.setDefinition(group);
            
            Collection<? extends ModelItem> items = _sort(group.getChildren(), elementsComparator);
            
            for (ModelItem item : items)
            {
                groupViewItem.addViewItem(_buildElementViewItem(item));
            }
            return groupViewItem;
        }
        else
        {
            throw new IllegalArgumentException("Group " + modelItem.getPath() + " should be an instance of ModelItemGroup");
        }
    }
    
    private static ModelViewItem _buildElementViewItem(ModelItem modelItem) throws IllegalArgumentException
    {
        if (modelItem instanceof ElementDefinition)
        {
            ElementDefinition definition = (ElementDefinition) modelItem;
            
            ViewElement parameterViewItem = new ViewElement();
            parameterViewItem.setDefinition(definition);
            return parameterViewItem;
        }
        else
        {
            throw new IllegalArgumentException("Item " + modelItem.getPath() + " should be an instance of ElementDefinition");
        }
    }
    
    /**
     * sort a model item list using comparator
     * @param items list of items to sort
     * @param comparator a comparator, can be null to avoid sort
     * @return a sorted list (or a copy of the list if the comparator is null)
     */
    private static Collection<? extends ModelItem> _sort(Collection<? extends ModelItem> items, Comparator<? super ModelItem> comparator)
    {
        if (comparator == null)
        {
            return items;
        }
        else
        {
            List<ModelItem> sorted = new ArrayList<>(items);
            sorted.sort(comparator);
            return sorted;
        }
    }
}
