/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.i18n;

import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.FormatStyle;
import java.util.Locale;

import org.apache.cocoon.transformation.I18nTransformer;
import org.apache.cocoon.xml.AttributesImpl;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.core.util.DateUtils;

/**
 * Allow to use {@link ZonedDateTime} as a parameter for {@link I18nizableText}.
 */
public class I18nizableDateTime implements FormatableI18nizable, I18nizableTextParameter
{
    private ZonedDateTime _dateTime;
    private DateTimeFormatter _formatter;
    private String _pattern;
    
    /**
     * Build an {@link I18nizableDateTime} from {@link ZonedDateTime} and style.
     * @param dateTime the value 
     * @param zoneId the time zone to use for formatting
     * @param style the style to use for formatting
     */
    public I18nizableDateTime(ZonedDateTime dateTime, ZoneId zoneId, FormatStyle style)
    {
        _dateTime = dateTime.withZoneSameInstant(zoneId);
        _formatter = DateTimeFormatter.ofLocalizedDateTime(style);
        _pattern = style.name();
    }
    
    /**
     * Build an {@link I18nizableDateTime} from {@link ZonedDateTime} and pattern.
     * @param dateTime the value 
     * @param zoneId the time zone to use for formatting
     * @param pattern the pattern to use for formatting
     */
    public I18nizableDateTime(ZonedDateTime dateTime, ZoneId zoneId, String pattern)
    {
        _dateTime = dateTime.withZoneSameInstant(zoneId);
        _formatter = DateTimeFormatter.ofPattern(pattern);
        _pattern = pattern;
    }
    
    public void toSAX(ContentHandler handler) throws SAXException
    {            
        AttributesImpl atts = new AttributesImpl();
        atts.addCDATAAttribute(I18nTransformer.I18N_NAMESPACE_URI, "pattern", "pattern", _pattern);
        atts.addCDATAAttribute(I18nTransformer.I18N_NAMESPACE_URI, "src-pattern", "src-pattern", "yyyy-MM-dd'T'HH:mm:ss.SSSXXX");
        atts.addCDATAAttribute(I18nTransformer.I18N_NAMESPACE_URI, "value", "value", DateUtils.zonedDateTimeToString(_dateTime));

        handler.startElement(I18nTransformer.I18N_NAMESPACE_URI, "date-time", "i18n:date-time", atts);
        handler.endElement(I18nTransformer.I18N_NAMESPACE_URI, "date-time", "i18n:date-time");
    }
    
    public void toSAXAsParam(ContentHandler handler) throws SAXException
    {
        toSAX(handler);
    }

    public String format(Locale locale)
    {
        return _formatter.withLocale(locale).format(_dateTime);
    }
}
