/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.log;

import org.apache.log4j.Level;
import org.apache.log4j.spi.LoggingEvent;
import org.apache.log4j.spi.ThrowableInformation;

/**
 *  Log record for the Memory Appender storage needs.
 *  
 * @param category the category of this record
 * @param millis the time of this record in milliseconds from 1970.
 * @param level the Level of this record.
 * @param message the message of this record.
 * @param location the string containing the location information.
 * @param user the user of this record.
 * @param requestURI the request URI of this record.
 * @param threadDescription the thread description of this record.
 * @param thrownStackTrace the Stack Trace for the associated Throwable of this record.
 */
public record MemoryLogRecord(
    String category,
    long millis,
    Level level,
    String message,
    String location,
    String user,
    String requestURI,
    String threadDescription,
    String thrownStackTrace
) {

    MemoryLogRecord(LoggingEvent event)
    {
        this(
            event.getLoggerName(),
            event.timeStamp,
            event.getLevel(),
            event.getRenderedMessage(),
            event.getLocationInformation().fullInfo,
            (String) event.getMDC("user"),
            (String) event.getMDC("requestURI"),
            Thread.currentThread().toString(),
            setThrownStackTrace(event.getThrowableInformation())
        );
    }

    /**
     * Set stack trace information associated with this MemoryLogRecord.
     * When this method is called, the stack trace in a
     * String-based format is made
     * available via the getThrownStackTrace() method.
     *
     * @param throwableInfo An org.apache.log4j.spi.ThrowableInformation to
     * associate with this MemoryLogRecord.
     * @return The Stack Trace for the associated Throwable of this LogRecord.
     */
    private static String setThrownStackTrace(ThrowableInformation throwableInfo)
    {
        if (throwableInfo != null)
        {
            String[] stackTraceArray = throwableInfo.getThrowableStrRep();
            
            StringBuffer stackTrace = new StringBuffer();
            String nextLine;
            
            for (int i = 0; i < stackTraceArray.length; i++)
            {
                nextLine = stackTraceArray[i] + "\n";
                stackTrace.append(nextLine);
            }
            
            return stackTrace.toString();
        }
        return null;
    }
}
