/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.log;

import java.io.IOException;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Appender;
import org.apache.log4j.FileAppender;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.rolling.RollingFileAppender;
import org.apache.log4j.rolling.TimeBasedRollingPolicy;

import org.ametys.core.util.AmetysExceptionFilter;
import org.ametys.runtime.plugin.component.PluginAware;
import org.ametys.runtime.util.AmetysHomeHelper;

/**
 * Default static implementation of {@link ExternalLog}
 * For implementing the {@link ExternalLog} interface (while being {@link Configurable}), extends this class and implements the {@link #getAppender()} and/or {@link #getCategory()} methods.
 * 
 * Example of XML
 * <pre>
 * {@code
 * &lt;extension id="runtime.core.externalLogger.myTest"
 *         class="org.ametys.runtime.log.StaticExternalLog"
 *         point="org.ametys.runtime.log.ExternalLogExtensionPoint">
 *     &lt;appender-name>my-logger&lt;/appender-name>
 *     &lt;category-name>com.example.mylog&lt;/category-name>
 *     &lt;file-prefix>my-file-name&lt;/file-prefix>
 *     &lt;additive>false&lt;/additive>
 *     &lt;rolling-file>true&lt;/rolling-file>
 *     &lt;log-level>INFO&lt;/log-level>
 * &lt;/extension>
 * }
 * </pre>
 */
public class StaticExternalLog implements ExternalLog, Configurable, PluginAware
{
    /** default layout pattern */
    protected static final String PATTERN_LAYOUT = "%d %-5p [%c] (%t;%X{requestURI}) %m%n";
    
    /** Default log level */
    protected static final Level DEFAULT_LEVEL = Level.WARN;
    private String _appenderName;
    private String _categoryName;
    private String _filePrefix;
    private boolean _additive;
    private Level _level;
    private boolean _rollingFileAppender;
    private Appender _appender;

    public Appender getAppender() throws IOException
    {
        if (_appender == null)
        {
            if (_rollingFileAppender)
            {
                _appender = _createRollingFileAppender();
            }
            else
            {
                _appender = _createFileAppender();
            }
        }
        return _appender;
    }

    private RollingFileAppender _createRollingFileAppender()
    {
        RollingFileAppender fileAppender = new RollingFileAppender();
        fileAppender.setName(_appenderName);
        fileAppender.setEncoding("UTF-8");
        TimeBasedRollingPolicy rollingPolicy = new TimeBasedRollingPolicy();
        String filePattern = AmetysHomeHelper.getAmetysHome() + "/logs/" + _filePrefix + "-%d.log";
        rollingPolicy.setFileNamePattern(filePattern);
        fileAppender.setRollingPolicy(rollingPolicy);

        PatternLayout layout = new PatternLayout(PATTERN_LAYOUT);
        fileAppender.setLayout(layout);

        AmetysExceptionFilter filter = new AmetysExceptionFilter();
        fileAppender.addFilter(filter);
        fileAppender.activateOptions();
        return fileAppender;
    }

    private FileAppender _createFileAppender() throws IOException
    {
        String filename = AmetysHomeHelper.getAmetysHome() + "/logs/" + _filePrefix + ".log";
        PatternLayout layout = new PatternLayout(PATTERN_LAYOUT);
        FileAppender fileAppender = new FileAppender(layout, filename);
        AmetysExceptionFilter filter = new AmetysExceptionFilter();
        fileAppender.addFilter(filter);
        fileAppender.activateOptions();
        return fileAppender;
    }

    public Logger getCategory()
    {
        org.apache.log4j.Logger logger = LogManager.getLogger(_categoryName);
        logger.setAdditivity(_additive);
        logger.setLevel(_level);
        return logger;
    }

    public void configure(Configuration configuration) throws ConfigurationException
    {
        _categoryName = configuration.getChild("category-name").getValue("").trim();
        _filePrefix = configuration.getChild("file-prefix").getValue("").trim();
        _additive = configuration.getChild("additive").getValueAsBoolean(true);
        _rollingFileAppender = configuration.getChild("rolling-file").getValueAsBoolean(true);
        String levelAsString = configuration.getChild("log-level").getValue(DEFAULT_LEVEL.toString());
        _level = Level.toLevel(levelAsString);

        if (StringUtils.isBlank(_filePrefix))
        {
            throw new ConfigurationException("<file-prefix> is mandatory and should not be blank", configuration);
        }
        if (StringUtils.isBlank(_categoryName))
        {
            throw new ConfigurationException("<category-name> is mandatory and should not be blank", configuration);
        }
    }

    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        _appenderName = id;
    }

}
