/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.runtime.maintenance;

import java.util.Arrays;
import java.util.Collection;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;

import org.ametys.core.captcha.CaptchaHelper;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.runtime.exception.ServiceUnavailableException;
import org.ametys.runtime.servlet.RuntimeServlet;
import org.ametys.runtime.servlet.RuntimeServlet.MaintenanceStatus;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * This action will stop user during the maintenance mode (except administrators)
 */
public class MaintenanceAction extends ServiceableAction
{
    /** url requires for authentication */
    protected static final Collection<Pattern> ACCEPTED_URL_PATTERNS = Arrays.asList(new Pattern[] {
        Pattern.compile("^plugins/core-ui/current-user/image_[0-9]+$"), 
        Pattern.compile("^plugins/core-ui/user/[^/]+/[^/]+/image_[0-9]+$")
    });

    private CurrentUserProvider _currentUserProvider;
    private RightManager _rightManager;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _currentUserProvider = (CurrentUserProvider) smanager.lookup(CurrentUserProvider.ROLE);
        _rightManager = (RightManager) smanager.lookup(RightManager.ROLE);
    }
    
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        if (RuntimeServlet.getMaintenanceStatus() != MaintenanceStatus.NONE && !acceptedUrl(ObjectModelHelper.getRequest(objectModel)))
        {
            UserIdentity user = _currentUserProvider.getUser();
             
            if ((user == null
                    || user.getPopulationId() != UserPopulationDAO.ADMIN_POPULATION_ID
                        && _rightManager.hasRight(user, "Runtime_Rights_Admin_Maintenance_Access", "/admin") != RightResult.RIGHT_ALLOW)
                  && !_isACaptchaUrl((String) ObjectModelHelper.getRequest(objectModel).getAttribute(WorkspaceMatcher.IN_WORKSPACE_URL)))
            {
                throw new ServiceUnavailableException("User " + (user != null ? UserIdentity.userIdentityToString(user) : "[anonymous]") + " cannot access during maintenance");
            }
        }
        
        return EMPTY_MAP;
    }
    
    private boolean _isACaptchaUrl(String url)
    {
        for (Pattern pattern : CaptchaHelper.getUsedUrlPatterns())
        {
            if (pattern.matcher(url).matches())
            {
                return true;
            }
        }
        
        return false;
    }

    /**
     * Determine if the request is one of the maintenance process
     * @param request The request
     * @return true to allow even in maintenance
     */
    public static boolean acceptedUrl(Request request)
    {
        // URL without server context and leading slash.
        String url = (String) request.getAttribute(WorkspaceMatcher.IN_WORKSPACE_URL);
        for (Pattern pattern : ACCEPTED_URL_PATTERNS)
        {
            if (pattern.matcher(url).matches())
            {
                return true;
            }
        }
        
        return false;
    }
}
