/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.plugins.core.ui.util.ConfigurationHelper;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.disableconditions.DisableCondition;
import org.ametys.runtime.model.disableconditions.DisableConditions;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.model.type.ModelItemType;
import org.ametys.runtime.model.type.ModelItemTypeExtensionPoint;

/**
 * {@link ModelItem} parser from an XML configuration.
 */
public abstract class AbstractModelItemParser
{
    private static final Logger __LOGGER = LoggerFactory.getLogger(AbstractModelItemParser.class);

    /** The extension point to use to get available model items types */
    protected ModelItemTypeExtensionPoint _modelItemTypeExtensionPoint;
    
    /**
     * Creates a model item parser.
     * @param modelItemTypeExtensionPoint the extension point to use to get available model items types
     */
    public AbstractModelItemParser(ModelItemTypeExtensionPoint modelItemTypeExtensionPoint)
    {
        _modelItemTypeExtensionPoint = modelItemTypeExtensionPoint;
    }
    
    /**
     * Parses an element definition from a XML configuration.
     * @param <T> type of the parsed item
     * @param serviceManager the service manager
     * @param pluginName the plugin name declaring this item.
     * @param itemConfig the XML configuration of the model item.
     * @param model the model which defines the model item
     * @param parent the parent of the model item to create. Can be null if the model item to parse has no parent
     * @return the parsed model item.
     * @throws ConfigurationException if the configuration is not valid.
     */
    @SuppressWarnings("unchecked")
    public <T extends ModelItem> T parse(ServiceManager serviceManager, String pluginName, Configuration itemConfig, Model model, ModelItemGroup parent) throws ConfigurationException
    {
        return (T) parse(serviceManager, pluginName, "plugin." + pluginName, itemConfig, model, parent);
    }
    
    /**
     * Parses an element definition from a XML configuration.
     * @param <T> type of the parsed item
     * @param serviceManager the service manager
     * @param pluginName the plugin name declaring this item.
     * @param catalog the catalog
     * @param itemConfig the XML configuration of the model item.
     * @param model the model which defines the model item
     * @param parent the parent of the model item to create. Can be null if the model item to parse has no parent
     * @return the parsed model item.
     * @throws ConfigurationException if the configuration is not valid.
     */
    @SuppressWarnings("unchecked")
    public <T extends ModelItem> T parse(ServiceManager serviceManager, String pluginName, String catalog, Configuration itemConfig, Model model, ModelItemGroup parent) throws ConfigurationException
    {
        ModelItem modelItem = _createModelItem(itemConfig);
        
        modelItem.setModel(model);
        if (parent != null)
        {
            parent.addChild(modelItem);
        }
        
        modelItem.setName(_parseName(itemConfig));
        modelItem.setPluginName(pluginName);
        modelItem.setLabel(_parseI18nizableText(itemConfig, catalog, "label"));
        modelItem.setDescription(_parseI18nizableText(itemConfig, catalog, "description"));
        modelItem.setType(_parseType(itemConfig));
        
        modelItem.setWidget(_parseWidget(itemConfig));
        modelItem.setWidgetParameters(_parseWidgetParameters(itemConfig, pluginName));
        modelItem.setDisableConditions(_parseDisableConditions(itemConfig));
        
        return (T) modelItem;
    }
    
    /**
     * Create the model item to populate it.
     * @param itemConfig the model item configuration to use.
     * @return the item instantiated.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected abstract ModelItem _createModelItem(Configuration itemConfig) throws ConfigurationException;

    /**
     * Parses the name of the model item.
     * @param itemConfig the model item configuration to use.
     * @return the name of the model item.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected String _parseName(Configuration itemConfig) throws ConfigurationException
    {
        String itemName = itemConfig.getAttribute(_getNameConfigurationAttribute());
        
        if (!itemName.matches("^[a-zA-Z]((?!__)[a-zA-Z0-9_\\.-])*$"))
        {
            throw new ConfigurationException("Invalid model item name: " + itemName + ". The item name must start with a letter and must contain only letters, digits, underscore, dot or dash characters.", itemConfig);
        }
        
        return itemName;
    }
    
    /**
     * Retrieves the name of the configuration attribute that contains the name of the model item
     * @return the name of the configuration attribute that contains the name of the model item
     */
    protected String _getNameConfigurationAttribute()
    {
        return "name";
    }
    
    /**
     * Parses a mandatory i18n text configuration, throwing an exception if empty.
     * @param config the configuration to use.
     * @param catalog the catalog
     * @param name the child name.
     * @return the i18n text.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected I18nizableText _parseI18nizableText(Configuration config, String catalog, String name) throws ConfigurationException
    {
        return I18nizableText.parseI18nizableText(config.getChild(name), catalog);
    }
    
    /**
     * Parse an optional i18n text configuration, with a default i18n text value.
     * @param config the configuration to use.
     * @param catalog the catalog
     * @param name the child name.
     * @param defaultValueCatalog the catalog of the default i18n text value
     * @param defaultValue the default i18n text value.
     * @return the i18n text.
     */
    protected I18nizableText _parseI18nizableText(Configuration config, String catalog, String name, String defaultValueCatalog, String defaultValue)
    {
        return I18nizableText.parseI18nizableText(config.getChild(name), catalog, new I18nizableText(defaultValueCatalog, defaultValue));
    }
    
    /**
     * Parses the type.
     * @param config the model item configuration to use.
     * @return the type.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected ModelItemType _parseType(Configuration config) throws ConfigurationException
    {
        String typeId = config.getAttribute("type");
        
        if (!_modelItemTypeExtensionPoint.hasExtension(typeId))
        {
            String modelItemName = _parseName(config);
            String availableTypes = StringUtils.join(_modelItemTypeExtensionPoint.getExtensionsIds(), ", ");
            UnknownTypeException ute = new UnknownTypeException("The type '" + typeId + "' is not available for the extension point '" + _modelItemTypeExtensionPoint + "'. Available types are: '" + availableTypes + "'.");
            throw new ConfigurationException("Unable to find the type '" + typeId + "' defined on the item '" + modelItemName + "'.", ute);
        }
        
        return _modelItemTypeExtensionPoint.getExtension(typeId);
    }
    
    /**
     * Parses the disable condition.
     * @param definitionConfiguration the configuration of the model item
     * @return result the parsed disable condition to be converted in JSON
     * @throws ConfigurationException if an error occurred
     */
    protected DisableConditions _parseDisableConditions(Configuration definitionConfiguration) throws ConfigurationException
    {
        Configuration disableConditionsConfiguration = definitionConfiguration.getChild("disable-conditions", false);
        return _parseConditions(disableConditionsConfiguration);
    }
    
    /**
     * Parses the disable condition.
     * @param disableConditionsConfiguration the configuration of the disable condition
     * @return result the parsed disable condition to be converted in JSON
     * @throws ConfigurationException if an error occurred
     */
    protected DisableConditions _parseConditions(Configuration disableConditionsConfiguration) throws ConfigurationException
    {
        DisableConditions conditions = null;
        
        if (disableConditionsConfiguration != null)
        {
            conditions = new DisableConditions();

            Configuration[] conditionsConfiguration = disableConditionsConfiguration.getChildren();
            for (Configuration conditionConfiguration : conditionsConfiguration)
            {
                String tagName = conditionConfiguration.getName();
                
                // Recursive case
                if (tagName.equals("conditions"))
                {
                    conditions.getSubConditions().add(_parseConditions(conditionConfiguration));
                }
                else if (tagName.equals("condition"))
                {
                    String id = conditionConfiguration.getAttribute("id");
                    DisableCondition.OPERATOR operator = DisableCondition.OPERATOR.valueOf(conditionConfiguration.getAttribute("operator", "eq").toUpperCase());
                    String value = conditionConfiguration.getValue("");
                    
                    
                    DisableCondition condition = new DisableCondition(id, operator, value);
                    conditions.getConditions().add(condition);
                }
            }
            
            conditions.setAssociation(DisableConditions.ASSOCIATION_TYPE.valueOf(disableConditionsConfiguration.getAttribute("type", "and").toUpperCase()));
        }
        
        return conditions;
    }

    /**
     * Parses the widget.
     * @param definitionConfig the element definition configuration to use.
     * @return the widget or <code>null</code> if none defined.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected String _parseWidget(Configuration definitionConfig) throws ConfigurationException
    {
        return definitionConfig.getChild("widget").getValue(null);
    }
    
    /**
     * Parses the widget's parameters
     * @param definitionConfig the parameter ele;ent definition to use.
     * @param pluginName the current plugin name.
     * @return the widget's parameters in a Map
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected Map<String, I18nizableText> _parseWidgetParameters(Configuration definitionConfig, String pluginName) throws ConfigurationException
    {
        Map<String, I18nizableText> widgetParams = new HashMap<>();
        
        Configuration widgetParamsConfig = definitionConfig.getChild("widget-params", false);
        if (widgetParamsConfig != null)
        {
            Map<String, Object> parsedParams = ConfigurationHelper.parsePluginParameters(widgetParamsConfig, pluginName, __LOGGER);
            
            for (Entry<String, Object> param : parsedParams.entrySet())
            {
                String paramName = param.getKey();
                Object value = param.getValue();
                if (value instanceof I18nizableText)
                {
                    widgetParams.put(paramName, (I18nizableText) value);
                }
                else if (value instanceof String)
                {
                    widgetParams.put(paramName, new I18nizableText((String) value));
                }
                else
                {
                    __LOGGER.warn("Widget parameter '{}' at location {} is of type [{}] which is not supported. It will be ignored.", paramName, definitionConfig.getLocation(), value.getClass());
                }
            }
        }
        
        return widgetParams;
    }
}
