/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * Abstract class for group of view items 
 */
public abstract class AbstractViewItemGroup implements ViewItemGroup
{
    /** The role of the view item group */
    protected String _role;
    
    /** The label of the view item group */
    protected I18nizableText _label;
    
    /** The description of the view item group */
    protected I18nizableText _description;
    
    /** The view item group's children */
    protected List<ViewItem> _children = new ArrayList<>();
    
    public I18nizableText getLabel()
    {
        return _label;
    }

    public void setLabel(I18nizableText label)
    {
        _label = label;
    }
    
    public I18nizableText getDescription()
    {
        return _description;
    }
    
    public void setDescription(I18nizableText description)
    {
        _description = description;
    }
    
    public String getRole()
    {
        return _role;
    }
    
    public void setRole(String role)
    {
        _role = role;
    }
    
    public List<ViewItem> getViewItems()
    {
        return Collections.unmodifiableList(_children);
    }
    
    public void addViewItem(ViewItem item)
    {
        _children.add(item);
        item.setParent(this);
    }
    
    public void insertViewItem(ViewItem item, int index)
    {
        if (index >= 0 && index <= _children.size())
        {
            _children.add(index, item);
            item.setParent(this);
        }
        else
        {
            throw new IllegalArgumentException("Unable to insert an item at index " + index + ". This group contains " + _children.size() + " items.");
        }
    }
    
    public boolean removeViewItem(ViewItem item)
    {
        return _children.remove(item);
    }
    
    public void clear()
    {
        _children.clear();
    }
    
    public void copyTo(ViewItem item)
    {
        assert item instanceof AbstractViewItemGroup;
        AbstractViewItemGroup viewItemGroup = (AbstractViewItemGroup) item;
        
        viewItemGroup.setRole(getRole());
        viewItemGroup.setLabel(getLabel());
        viewItemGroup.setDescription(getDescription());
    }

    @Override
    public int hashCode()
    {
        return Objects.hash(_children);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        if (obj == null)
        {
            return false;
        }
        if (getClass() != obj.getClass())
        {
            return false;
        }
        AbstractViewItemGroup other = (AbstractViewItemGroup) obj;
        return Objects.equals(_children, other._children);
    }
    
    public boolean equals(Object obj, boolean checkDetails)
    {
        if (!equals(obj))
        {
            return false;
        }
        else if (checkDetails)
        {
            AbstractViewItemGroup other = (AbstractViewItemGroup) obj;
            
            for (int i = 0; i < _children.size(); i++)
            {
                ViewItem child = _children.get(i);
                ViewItem otherChild = other._children.get(i);
                
                if (!child.equals(otherChild, checkDetails))
                {
                    return false;
                }
            }
            
            return Objects.equals(_label, other._label) && Objects.equals(_description, other._description) && Objects.equals(_role, other._role);
        }
        else
        {
            return true;
        }
    }
}
