/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.slf4j.Logger;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * Helper for {@link CategorizedElementDefinitionWrapper}
 */
public final class CategorizedElementDefinitionHelper
{
    private CategorizedElementDefinitionHelper()
    {
        //Nothing
    }
    
    /**
     * Generate the view for categorized element definition wrappers, using comparators to sort the categories, groups and elements
     * @param <T> type of the wrappers
     * @param wrappers the wrappers
     * @param categoriesComparator {@link Comparator} used for the categories (can be null to keep order)
     * @param groupsComparator {@link Comparator} for the groups (can be null to keep order)
     * @param elementsComparator {@link Comparator} for the elements (can be null to keep order)
     * @return A {@link View} ordered from the {@link ModelItem} using the {@link Comparator}
     */
    public static <T extends CategorizedElementDefinitionWrapper> View buildViewFromCategories(Collection<T> wrappers, Comparator<I18nizableText> categoriesComparator, Comparator<I18nizableText> groupsComparator, Comparator<T> elementsComparator)
    {
        View view = new View();
        
        Map<I18nizableText, Map<I18nizableText, List<T>>> categories = categorizeElementDefinitionWrappers(wrappers);
        
        Collection<I18nizableText> sortedCategories = _sort(categories.keySet(), categoriesComparator);
        for (I18nizableText categoryLabel : sortedCategories)
        {
            ViewItem categoryViewItem = _buildCategoryViewItem(categoryLabel, categories.get(categoryLabel), groupsComparator, elementsComparator);
            view.addViewItem(categoryViewItem);
        }
        return view;
    }
    
    /**
     * Organize a collection of categorized element definition wrappers by categories and groups.
     * @param <T> type of the wrappers
     * @param wrappers a collection of element definition wrappers.
     * @return a Map with the same element definition wrappers, sorted first by category then by group.
     */
    public static <T extends CategorizedElementDefinitionWrapper> Map<I18nizableText, Map<I18nizableText, List<T>>> categorizeElementDefinitionWrappers(Collection<T> wrappers)
    {
        Map<I18nizableText, Map<I18nizableText, List<T>>> categories = new HashMap<>();

        // Classify parameters by groups and categories
        for (T parameter : wrappers)
        {
            I18nizableText displayCategory = parameter.getDisplayCategory();
            I18nizableText displayGroup = parameter.getDisplayGroup();

            // Get the map of groups of the category
            Map<I18nizableText, List<T>> category = categories.get(displayCategory);
            if (category == null)
            {
                category = new HashMap<>();
                categories.put(displayCategory, category);
            }

            // Get the map of parameters of the group
            List<T> group = category.get(displayGroup);
            if (group == null)
            {
                group = new ArrayList<>();
                category.put(displayGroup, group);
            }

            group.add(parameter);
        }
        
        return categories;
    }
    
    private static <T extends CategorizedElementDefinitionWrapper> ViewItem _buildCategoryViewItem(I18nizableText categoryLabel, Map<I18nizableText, List<T>> groups, Comparator<I18nizableText> groupsComparator, Comparator<T> elementsComparator) throws IllegalArgumentException
    {
        SimpleViewItemGroup categoryViewItem = new SimpleViewItemGroup();
        categoryViewItem.setRole(ViewItemGroup.TAB_ROLE);
        categoryViewItem.setLabel(categoryLabel);

        Collection<I18nizableText> sortedGroups = _sort(groups.keySet(), groupsComparator);
        for (I18nizableText groupLabel : sortedGroups)
        {
            ViewItem groupViewItem = _buildGroupViewItem(groupLabel, groups.get(groupLabel), elementsComparator);
            categoryViewItem.addViewItem(groupViewItem);
        }
        return categoryViewItem;
    }
    
    private static <T extends CategorizedElementDefinitionWrapper> ViewItem _buildGroupViewItem(I18nizableText groupLabel, List<T> wrappers, Comparator<T> elementsComparator) throws IllegalArgumentException
    {
        SimpleViewItemGroup groupViewItem = new SimpleViewItemGroup();
        groupViewItem.setRole(ViewItemGroup.FIELDSET_ROLE);
        groupViewItem.setLabel(groupLabel);

        Collection<T> items = _sort(wrappers, elementsComparator);
        for (T item : items)
        {
            ViewElement parameterViewItem = new ViewElement();
            parameterViewItem.setDefinition(item.getDefinition());
            groupViewItem.addViewItem(parameterViewItem);
        }
        
        return groupViewItem;
    }
    
    /**
     * Sort a model item list using comparator
     * @param <T> The type of item to compare
     * @param items list of items to sort
     * @param comparator a comparator, can be null to avoid sort
     * @return a sorted list (or a copy of the list if the comparator is null)
     */
    private static <T> Collection<T> _sort(Collection<T> items, Comparator<T> comparator)
    {
        if (comparator == null)
        {
            return items;
        }
        else
        {
            List<T> sorted = new ArrayList<>(items);
            sorted.sort(comparator);
            return sorted;
        }
    }
    
    /**
     * Validate parameters before writing
     * @param definitionAndValues a map of all parameters and their values
     * @param flatDefinitions flat definition of {@link ElementDefinition} from the model file
     * @param logger a logger
     * @return a map containing the potential errors
     * @deprecated TODO NEWATTRIBUTEAPI_CONFIG RUNTIME-2851 remove this method when it is not necessary anymore to have the flat definitions
     */
    @Deprecated
    public static Map<String, List<I18nizableText>> validateValuesForWriting (Map<String, DefinitionAndValue> definitionAndValues, Map<String, ElementDefinition> flatDefinitions, Logger logger)
    {
        Map<String, List<I18nizableText>> errorFields = new HashMap<>();
        
        for (ElementDefinition definition : flatDefinitions.values())
        {
            boolean isGroupSwitchOn = ModelHelper.isGroupSwitchOn(definition, DefinitionAndValue.extractValues(definitionAndValues));
            boolean isDisabled = ModelHelper.evaluateDisableConditions(definition.getDisableConditions(), definitionAndValues, logger);
            
            if (isGroupSwitchOn && !isDisabled)
            {
                Object value = null;
                DefinitionAndValue definitionAndValue = definitionAndValues.get(definition.getName());
                if (definitionAndValue != null)
                {
                    value = definitionAndValue.getValue();
                }
                
                List<I18nizableText> errors = ModelHelper.validateValue(definition, value);
                
                if (!errors.isEmpty())
                {
                    logger.warn("The configuration parameter '{}' is not valid", definition.getName());
                    errorFields.put(definition.getName(), errors);
                }
            }
        }
        
        if (!errorFields.isEmpty())
        {
            logger.debug("Failed to save configuration because of invalid parameter values");
        }
        return errorFields;
    }
    
    /**
     * Validate parameters before writing
     * @param definitionAndValues a map of all parameters and their values
     * @param logger a logger
     * @return a map containing the potential errors
     */
    public static Map<String, List<I18nizableText>> validateValuesForWriting (Map<String, DefinitionAndValue> definitionAndValues, Logger logger)
    {
        Map<String, List<I18nizableText>> errorFields = new HashMap<>();
        for (Entry<String, DefinitionAndValue> entry : definitionAndValues.entrySet())
        {
            DefinitionAndValue definitionAndValue = entry.getValue();
            ModelItem definition = definitionAndValue.getDefinition();
            if (definition instanceof ElementDefinition)
            {
                ElementDefinition elementDefinition = (ElementDefinition) definition;
                boolean isGroupSwitchOn = ModelHelper.isGroupSwitchOn(elementDefinition, DefinitionAndValue.extractValues(definitionAndValues));
                boolean isDisabled = ModelHelper.evaluateDisableConditions(elementDefinition.getDisableConditions(), definitionAndValues, logger);
                
                if (isGroupSwitchOn && !isDisabled)
                {
                    Object value = definitionAndValue.getValue();
                    
                    List<I18nizableText> errors = ModelHelper.validateValue(elementDefinition, value);
                    
                    if (!errors.isEmpty())
                    {
                        logger.warn("The configuration parameter '{}' is not valid", elementDefinition.getName());
                        errorFields.put(elementDefinition.getName(), errors);
                    }
                }
            }
        }
        
        if (!errorFields.isEmpty())
        {
            logger.debug("Failed to save configuration because of invalid parameter values");
        }
        return errorFields;
    }
    
    /**
     * Compares the two given wrappers' position.
     * Returns a negative integer, zero, or a positive integer as the position of the first wrapper
     * is less than, equal to, or greater than the second.
     * @param <T> type of the wrappers to compare
     * @param wrapper1 the first wrapper
     * @param wrapper2 the second wrapper
     * @return a negative integer, zero, or a positive integer as the position of the first wrapper
     * is less than, equal to, or greater than the second.
     */
    public static <T extends CategorizedElementDefinitionWrapper> int compareWrapperPositions(T wrapper1, T wrapper2)
    {
        int positionComparison = ((Long) wrapper1.getPosition()).compareTo(wrapper2.getPosition());
        if (wrapper1.getPosition() < 0 && wrapper2.getPosition() < 0)
        {
            positionComparison = 0;
        }
        else if (wrapper1.getPosition() < 0 && wrapper2.getPosition() >= 0)
        {
            positionComparison = 1;
        }
        else if (wrapper2.getPosition() < 0)
        {
            positionComparison = -1;
        }
        
        return positionComparison;
    }
}
