/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * Parser for categorized {@link ElementDefinition}
 */
public class CategorizedElementDefinitionParser
{
    private ElementDefinitionParser _parser;
    
    /**
     * Creates a categorized element definition parser.
     * @param parser the element definition parser
     */
    public CategorizedElementDefinitionParser(ElementDefinitionParser parser)
    {
        _parser = parser;
    }

    /**
     * Parses an element definition from a XML configuration.
     * @param serviceManager the service manager
     * @param pluginName the plugin name declaring this item.
     * @param definitionConfig the XML configuration of the model item.
     * @param model the model which defines the model item
     * @param parent the parent of the model item to create. Can be null if the model item to parse has no parent
     * @return the parsed model item.
     * @throws ConfigurationException if the configuration is not valid.
     */
    @SuppressWarnings("unchecked")
    public CategorizedElementDefinitionWrapper parse(ServiceManager serviceManager, String pluginName, Configuration definitionConfig, Model model, ModelItemGroup parent) throws ConfigurationException
    {
        ElementDefinition definition = _parser.parse(serviceManager, pluginName, definitionConfig, model, parent);
        CategorizedElementDefinitionWrapper elementDefinitionWrapper = _createElementDefinitionWrapper();
        
        elementDefinitionWrapper.setDefinition(definition);
        elementDefinitionWrapper.setDisplayCategory(_parseI18nizableText(definitionConfig, pluginName, "category"));
        elementDefinitionWrapper.setDisplayGroup(_parseI18nizableText(definitionConfig, pluginName, "group"));
        elementDefinitionWrapper.setPosition(definitionConfig.getChild("order").getValueAsLong(-1));
        
        return elementDefinitionWrapper;
    }
    
    /**
     * Create the element definition wrapper to populate it.
     * @return the wrapper instantiated.
     */
    protected CategorizedElementDefinitionWrapper _createElementDefinitionWrapper()
    {
        return new CategorizedElementDefinitionWrapper<>();
    }
    
    /**
     * Parses an i18n text.
     * @param config the configuration to use.
     * @param pluginName the current plugin name.
     * @param name the child name.
     * @return the i18n text.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected I18nizableText _parseI18nizableText(Configuration config, String pluginName, String name) throws ConfigurationException
    {
        return I18nizableText.parseI18nizableText(config.getChild(name), "plugin." + pluginName);
    }
}
