/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * Object containing the definition of an element and informations about its categories / groups
 * @param <T> Type of the element value
 */
public class CategorizedElementDefinitionWrapper<T>
{
    /** The element definition */
    protected ElementDefinition<T> _definition;
    /** the category in which to display the element */
    protected I18nizableText _displayCategory;
    /** the group in which to display the element */
    protected I18nizableText _displayGroup;
    /** the position in the group where the element has to be displayed */
    protected long _position;
    
    /**
     * Default constructor.
     */
    public CategorizedElementDefinitionWrapper()
    {
        // Do nothing
    }
    
    /**
     * Constructor by copying an existing {@link CategorizedElementDefinitionWrapper}.
     * @param wrapperToCopy The {@link CategorizedElementDefinitionWrapper} to copy
     */
    public CategorizedElementDefinitionWrapper(CategorizedElementDefinitionWrapper<T> wrapperToCopy)
    {
        // ElementDefinition
        ElementDefinition<T> copiedDefinition = new DefaultElementDefinition<>(wrapperToCopy.getDefinition());
        setDefinition(copiedDefinition);
        
        // Categories
        setDisplayCategory(wrapperToCopy.getDisplayCategory());
        setDisplayGroup(wrapperToCopy.getDisplayGroup());
        setPosition(wrapperToCopy.getPosition());
    }
    
    /**
     * Retrieves the element definition
     * @return the element definition
     */
    public ElementDefinition<T> getDefinition()
    {
        return _definition;
    }
    
    /**
     * Sets the element definition
     * @param definition the element definition to set
     */
    public void setDefinition(ElementDefinition<T> definition)
    {
        _definition = definition;
    }
    
    /**
     * Retrieves the category in which to display the element
     * @return the category
     */
    public I18nizableText getDisplayCategory()
    {
        return _displayCategory;
    }
    
    /**
     * Sets the category in which to display the element
     * @param displayCategory the category to set
     */
    public void setDisplayCategory(I18nizableText displayCategory)
    {
        this._displayCategory = displayCategory;
    }
    
    /**
     * Retrieves the group in which to display the element
     * @return the group
     */
    public I18nizableText getDisplayGroup()
    {
        return _displayGroup;
    }
    
    /**
     * Sets the group in which to display the element
     * @param displayGroup the group to set
     */
    public void setDisplayGroup(I18nizableText displayGroup)
    {
        this._displayGroup = displayGroup;
    }
    
    /**
     * Retrieves the position in the group where the element has to be displayed 
     * @return the position in the group
     */
    public long getPosition()
    {
        return _position;
    }
    
    /**
     * Sets the position of the element in the group
     * @param position the position to set
     */
    public void setPosition(long position)
    {
        this._position = position;
    }

    @Override
    public boolean equals(Object obj)
    {
        return getDefinition().equals(obj);
    }

    @Override
    public int hashCode()
    {
        return getDefinition().hashCode();
    }

    @Override
    public String toString()
    {
        return getDefinition().toString();
    }
    
    
}
