/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.HashMap;
import java.util.Map;

/**
 * Pair of a value and its definition
 * @param <T> Type of the root object
 */
public class DefinitionAndValue<T>
{
    private T _root;
    private ModelItem _definition;
    private Object _value;
    private DefinitionAndValue<T> _parent;
    
    /**
     * Creates a definition and value pair
     * @param root the root object of this definition and value pair
     * @param definition the definition
     * @param value the value
     */
    public DefinitionAndValue(T root, ModelItem definition, Object value)
    {
        this(root, definition, value, null);
    }
    
    /**
     * Creates a definition and value pair
     * @param root the root object of this definition and value pair
     * @param definition the definition
     * @param value the value
     * @param parent the parent of the definition and value pair
     */
    public DefinitionAndValue(T root, ModelItem definition, Object value, DefinitionAndValue<T> parent)
    {
        this._root = root;
        this._definition = definition;
        this._value = value;
        this._parent = parent;
    }

    /**
     * Retrieves the root object of this definition and value pair
     * @return the root
     */
    public T getRoot()
    {
        return _root;
    }
    
    /**
     * Retrieves the definition
     * @return the definition
     */
    public ModelItem getDefinition()
    {
        return _definition;
    }
    
    /**
     * Retrieves the value
     * @return the value
     */
    public Object getValue()
    {
        return _value;
    }
    
    /**
     * Retrieves the parent of the definition and value pair
     * @return the parent
     */
    public DefinitionAndValue<T> getParent()
    {
        return _parent;
    }
    
    /**
     * extract a map of values from a map of DefinitionAndValue
     * @param definitionAndValues a map containing definition and values
     * @return a map containing only values
     */
    public static Map<String, Object> extractValues(Map<String, DefinitionAndValue> definitionAndValues)
    {
        Map<String, Object> values = new HashMap<>();
        for (Map.Entry<String, DefinitionAndValue> entry : definitionAndValues.entrySet())
        {
            Object value = entry.getValue().getValue();
            values.put(entry.getKey(), value);
        }
        return values;
    }
}
