/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.Optional;

/**
 * Object that gives some context for definitions manipulation
 */
public final class DefinitionContext
{
    private Optional<Object> _object = Optional.empty();
    private boolean _isEdition;

    private DefinitionContext()
    {
        // Empty private constructor
    }
    
    /**
     * Creates a new instance of a {@link DefinitionContext}
     * @return the created instance
     */
    public static DefinitionContext newInstance()
    {
        return new DefinitionContext();
    }
    
    /**
     * Retrieves the context object for the definition
     * @return the context object
     */
    public Optional<Object> getObject()
    {
        return _object;
    }
    
    /**
     * Set the context object
     * @param object the context object to set
     * @return the current {@link DefinitionContext}
     */
    public DefinitionContext withObject(Object object)
    {
        _object = Optional.ofNullable(object);
        return this;
    }
    
    /**
     * Checks if the context is in edition
     * @return <code>true</code> if the context is in edition, <code>false</code> otherwise
     */
    public boolean isEdition()
    {
        return _isEdition;
    }
    
    /**
     * Set to <code>true</code> if the context is in edition (default to <code>false</code>) 
     * @param isEdition <code>true</code> if the context is in edition, <code>false</code> otherwise
     * @return the current {@link DefinitionContext}
     */
    public DefinitionContext withEdition(boolean isEdition)
    {
        _isEdition = isEdition;
        return this;
    }
}
