/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.List;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.lang3.tuple.Pair;

import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.parameter.Validator;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * Interface of the definition of a single model item (parameter, attribute)
 * @param <T> Type of the element value
 */
public interface ElementDefinition<T> extends ModelItem
{
    /** config type for default values */
    public static final String CONFIG_DEFAULT_VALUE_TYPE = "config";
    
    @Override
    public ElementType<T> getType();
    
    /**
     * Retrieves the enumerator.
     * @return the enumerator or <code>null</code> if none is defined.
     */
    public Enumerator<T> getEnumerator();
    
    /**
     * Set the enumerator.
     * @param enumerator the enumerator.
     */
    public void setEnumerator(Enumerator<T> enumerator);
    
    /**
     * Retrieves the custom enumerator's class name
     * @return the custom enumerator's class name
     */
    public String getCustomEnumerator();
    
    /**
     * Set the custom enumerator's class name
     * @param customEnumerator the custom enumerator's class name
     */
    public void setCustomEnumerator(String customEnumerator);
    
    /**
     * Retrieves the custom enumerator's configuration
     * @return the custom enumerator's configuration
     */
    public Configuration getEnumeratorConfiguration();
    
    /**
     * Set the custom enumerator's configuration
     * @param enumeratorConfiguration the custom enumerator's configuration
     */
    public void setEnumeratorConfiguration(Configuration enumeratorConfiguration);
    
    /**
     * Get the enumerator to use when rendering this element as a criterion
     * @param configuration The enumerator configuration.
     * @param enumeratorManager ComponentManager for the criterion's {@link Enumerator}
     * @return The enumerator or null if the element'critérion is not enumerated.
     * @throws ConfigurationException If an error occurs while initializing the enumerator
     */
    public default Enumerator getCriterionEnumerator(Configuration configuration, ThreadSafeComponentManager<Enumerator> enumeratorManager) throws ConfigurationException
    {
        return getEnumerator();
    }
    
    /**
     * Retrieves the validator.
     * @return the validator or <code>null</code> if none is defined.
     */
    public Validator getValidator();
    
    /**
     * Set the validator.
     * @param validator the validator.
     */
    public void setValidator(Validator validator);
    
    /**
     * Retrieves the custom validator's class name
     * @return the custom validator's class name
     */
    public String getCustomValidator();
    
    /**
     * Set the custom validator's class name
     * @param customValidator the custom validator's class name
     */
    public void setCustomValidator(String customValidator);
    
    /**
     * Retrieves the custom validator's configuraiton
     * @return the custom validator's configuration
     */
    public Configuration getValidatorConfiguration();
    
    /**
     * Set the custom validator's configuration
     * @param validatorConfiguration the custom validator's configuration
     */
    public void setValidatorConfiguration(Configuration validatorConfiguration);
    
    /**
     * Retrieves the default value, as an object corresponding to the definition's type and cardinality
     * Retrieves <code>null</code> if no default value is defined for this definition
     * @param <X> The type of the default value
     * @return the default value.
     */
    public <X> X getDefaultValue();
    
    /**
     * Set the parsed default values.
     * If the definition is not multiple, the list should contain only one element
     * A parsed default value is described by its type and an object depending on the type 
     * @param parsedDefaultValues the parsed default values.
     */
    public void setParsedDefaultValues(List<Pair<String, Object>> parsedDefaultValues);
    
    /**
     * Set a default value to the definition
     * The default value is single, classic default value
     * @param defaultValue the default value to set
     */
    public void setDefaultValue(T defaultValue);
    
    /**
     * Retrieves the parsed default values
     * @return the parsed default values
     */
    public List<Pair<String, Object>> getParsedDefaultValues();
    
    /**
     * Test if the element is multiple.
     * @return <code>true</code> if the metadata is multiple.
     */
    public boolean isMultiple();
    
    /**
     * Set the element multiple status.
     * @param isMultiple the element multiple status.
     */
    public void setMultiple(boolean isMultiple);
    
    /**
     * Determines if the definition is editable, i.e. if it can be part of a view used for edition
     * @return <code>true</code> if the definition is editable, <code>false</code> otherwise
     */
    public default boolean isEditable()
    {
        // A definition is modifable by default
        return true;
    }
}
