/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;

import org.ametys.runtime.config.ConfigManager;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.model.type.ModelItemTypeExtensionPoint;
import org.ametys.runtime.parameter.DefaultValidator;
import org.ametys.runtime.parameter.Validator;
import org.ametys.runtime.plugin.component.ThreadSafeComponentManager;

/**
 * {@link ElementDefinition} parser from an XML configuration.
 */
public class ElementDefinitionParser extends AbstractModelItemParser
{
    private ThreadSafeComponentManager<Enumerator> _enumeratorManager;
    private ThreadSafeComponentManager<Validator> _validatorManager;
    private final Map<ElementDefinition, String> _validatorsToLookup = new HashMap<>();
    private final Map<ElementDefinition, String> _enumeratorsToLookup = new HashMap<>();
    
    /**
     * Creates an element definition parser.
     * @param modelItemTypeExtensionPoint the extension point to use to get available element types
     * @param enumeratorManager the enumerator component manager.
     * @param validatorManager the validator component manager.
     */
    public ElementDefinitionParser(ModelItemTypeExtensionPoint modelItemTypeExtensionPoint, ThreadSafeComponentManager<Enumerator> enumeratorManager, ThreadSafeComponentManager<Validator> validatorManager)
    {
        super(modelItemTypeExtensionPoint);
        _enumeratorManager = enumeratorManager;
        _validatorManager = validatorManager;
    }
    
    @Override
    @SuppressWarnings("unchecked")
    public <T extends ModelItem> T parse(ServiceManager serviceManager, String pluginName, String catalog, Configuration definitionConfig, Model model, ModelItemGroup parent) throws ConfigurationException
    {
        ElementDefinition definition = (ElementDefinition) super.parse(serviceManager, pluginName, catalog, definitionConfig, model, parent);
        
        definition.setParsedDefaultValues(_parseDefaultValues(definitionConfig, definition));
        definition.setMultiple(_parseMultiple(definitionConfig));
        
        _parseAndSetEnumerator(pluginName, catalog, definition, definitionConfig);
        _parseAndSetValidator(pluginName, definition, definitionConfig);
        
        return (T) definition;
    }
    
    @Override
    protected ElementDefinition _createModelItem(Configuration definitionConfig) throws ConfigurationException
    {
        return new DefaultElementDefinition();
    }

    /**
     * Parses the default values.
     * @param definitionConfig the element definition configuration.
     * @param definition the element definition.
     * @return the default values and their types or <code>null</code> if none default value is defined.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected List<Pair<String, Object>> _parseDefaultValues(Configuration definitionConfig, ElementDefinition definition) throws ConfigurationException
    {
        Configuration[] defaultValueConfigs = definitionConfig.getChildren("default-value");
        if (defaultValueConfigs.length > 0)
        {
            List<Pair<String, Object>> defaultValues = new ArrayList<>();
            
            for (Configuration defaultValueConfig : defaultValueConfigs)
            {
                String defaultValueType = defaultValueConfig.getAttribute("type", null);
                Object defaultValue = _parseDefaultValue(defaultValueConfig, definition, defaultValueType);
                defaultValues.add(new ImmutablePair<>(defaultValueType, defaultValue));
            }
            
            return defaultValues;
        }
        else
        {
            return null;
        }
    }
    
    /**
     * Parses the default value.
     * @param defaultValueConfig the default value configuration.
     * @param definition the element definition.
     * @param defaultValueType the type of the default value
     * @return the default value or <code>null</code> if none default value is defined.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected Object _parseDefaultValue(Configuration defaultValueConfig, ElementDefinition definition, String defaultValueType) throws ConfigurationException
    {
        ElementType type = definition.getType();
        if (defaultValueType != null)
        {
            if (ElementDefinition.CONFIG_DEFAULT_VALUE_TYPE.equals(defaultValueType))
            {
                String configParamName = defaultValueConfig.getValue();
                if (ConfigManager.getInstance().hasModelItem(configParamName))
                {
                    ModelItem configParamDefinition = ConfigManager.getInstance().getModelItem(configParamName);
                    String configParamTypeId = configParamDefinition.getType().getId();
                    if (configParamTypeId.equals(type.getId()))
                    {
                        return configParamName;
                    }
                    else
                    {
                        throw new ConfigurationException("The configuration parameter '" + configParamName + " (" + configParamTypeId + ")' cannot be used as default value for item '" + definition.getPath() + " (" + type.getId() + ")': types are not the same.", defaultValueConfig);
                    }
                }
                else
                {
                    throw new ConfigurationException("The configuration parameter '" + configParamName + "' does not exist, it cannot be used as default value for item '" + definition.getPath() + " (" + type.getId() + ")'.", defaultValueConfig);
                }
            }
            else
            {
                throw new ConfigurationException("The type '" + defaultValueType + "' is not available for the default value of item '" + definition.getPath() + " (" + type.getId() + ")'.", defaultValueConfig);
            }
        }
        else
        {
            return type.parseConfiguration(defaultValueConfig);
        }
    }
    
    /**
     * Parses the multiple attribute.
     * @param definitionConfig the element definition configuration to use.
     * @return the true if the element is multiple, false otherwise.
     * @throws ConfigurationException if the configuration is not valid.
     */
    protected Boolean _parseMultiple(Configuration definitionConfig) throws ConfigurationException
    {
        return definitionConfig.getAttributeAsBoolean("multiple", false);
    }
    
    /**
     * Parses the enumerator.
     * @param pluginName the plugin name.
     * @param catalog the catalog
     * @param definition the element definition.
     * @param definitionConfig the element definition configuration.
     * @throws ConfigurationException if the configuration is not valid.
     */
    @SuppressWarnings("unchecked")
    protected void _parseAndSetEnumerator(String pluginName, String catalog, ElementDefinition definition, Configuration definitionConfig) throws ConfigurationException
    {
        Configuration enumeratorConfig = definitionConfig.getChild("enumeration", false);
        
        if (enumeratorConfig != null)
        {
            Configuration customEnumerator = enumeratorConfig.getChild("custom-enumerator", false);
            
            if (customEnumerator != null)
            {
                final String enumeratorClassName = customEnumerator.getAttribute("class");
                final String enumeratorRole = definition.getPath();
                
                try
                {
                    Class enumeratorClass = Class.forName(enumeratorClassName);
                    _enumeratorManager.addComponent(pluginName, null, enumeratorRole, enumeratorClass, definitionConfig);
                }
                catch (Exception e)
                {
                    throw new ConfigurationException("Unable to instantiate enumerator for class: " + enumeratorClassName, e);
                }

                // This enumerator will be affected later when enumeratorManager
                // will be initialized in lookupComponents() call
                _enumeratorsToLookup.put(definition, enumeratorRole);
                    
                // Add the custom enumerator information to the element definition
                definition.setCustomEnumerator(enumeratorClassName);
                definition.setEnumeratorConfiguration(customEnumerator);
            }
            else
            {
                StaticEnumerator staticEnumerator = new StaticEnumerator();
                
                for (Configuration entryConfig : enumeratorConfig.getChildren("entry"))
                {
                    Configuration valueConfiguration = entryConfig.getChild("value");
                    Object value = definition.getType().parseConfiguration(valueConfiguration);
                    I18nizableText label = null;
                    
                    if (entryConfig.getChild("label", false) != null)
                    {
                        label = _parseI18nizableText(entryConfig, catalog, "label");
                    }
                    
                    staticEnumerator.add(label, value);
                }
                
                definition.setEnumerator(staticEnumerator);
            }
        }
    }

    /**
     * Parses the validator.
     * @param pluginName the plugin name.
     * @param definition the element definition.
     * @param definitionConfig the element definition configuration.
     * @throws ConfigurationException if the configuration is not valid.
     */
    @SuppressWarnings("unchecked")
    protected void _parseAndSetValidator(String pluginName, ElementDefinition definition, Configuration definitionConfig) throws ConfigurationException
    {
        Configuration validatorConfig = definitionConfig.getChild("validation", false);
        
        if (validatorConfig != null)
        {
            Configuration customValidator = validatorConfig.getChild("custom-validator", false);
            String validatorClassName;

            if (customValidator != null)
            {
                validatorClassName = customValidator.getAttribute("class");
                
                // Add the custom validator information to the element definition
                definition.setCustomValidator(validatorClassName);
                definition.setValidatorConfiguration(customValidator);
            }
            else
            {
                validatorClassName = DefaultValidator.class.getName();
            }
            
            final String validatorRole = definition.getPath();
            
            try
            {
                Class validatorClass = Class.forName(validatorClassName);
                _validatorManager.addComponent(pluginName, null, validatorRole, validatorClass, definitionConfig);
            }
            catch (Exception e)
            {
                throw new ConfigurationException("Unable to instantiate validator for class: " + validatorClassName, e);
            }

            // Will be affected later when validatorManager will be initialized
            // in lookupComponents() call
            _validatorsToLookup.put(definition, validatorRole);
        }
    }
    
    /**
     * Retrieves local validators and enumerators components and set them into
     * previous parsed element definition.
     * @throws Exception if an error occurs.
     */
    @SuppressWarnings("unchecked")
    public void lookupComponents() throws Exception
    {
        _validatorManager.initialize();
        _enumeratorManager.initialize();
        
        for (Map.Entry<ElementDefinition, String> entry : _validatorsToLookup.entrySet())
        {
            ElementDefinition definition = entry.getKey();
            String validatorRole = entry.getValue();
            
            try
            {
                definition.setValidator(_validatorManager.lookup(validatorRole));
            }
            catch (ComponentException e)
            {
                throw new Exception("Unable to lookup validator role: '" + validatorRole + "' for parameter: " + definition, e);
            }
        }
        
        for (Map.Entry<ElementDefinition, String> entry : _enumeratorsToLookup.entrySet())
        {
            ElementDefinition definition = entry.getKey();
            String enumeratorRole = entry.getValue();
            
            try
            {
                definition.setEnumerator(_enumeratorManager.lookup(enumeratorRole));
            }
            catch (ComponentException e)
            {
                throw new Exception("Unable to lookup enumerator role: '" + enumeratorRole + "' for parameter: " + definition, e);
            }
        }
    }
}
