/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.cocoon.ProcessingException;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.checker.ItemCheckerDescriptor;
import org.ametys.runtime.model.disableconditions.DisableConditions;
import org.ametys.runtime.model.type.ModelItemType;
import org.ametys.runtime.util.ModifiableLabelable;

/**
 * Interface for model items
 */
public interface ModelItem extends ModifiableLabelable, Comparable<ModelItem>
{
    /** Separator used for item paths in definitions */
    public static final String ITEM_PATH_SEPARATOR = "/";
    
    /**
     * Returns the {@link ItemCheckerDescriptor}s associated with this group.
     * @return the {@link ItemCheckerDescriptor}s associated with this group.
     */
    public Set<ItemCheckerDescriptor> getItemCheckers();
    
    /**
     * Add an item checker to the group
     * @param itemChecker the item checker to add
     */
    public void addItemChecker(ItemCheckerDescriptor itemChecker);

    /**
     * Retrieves the path of the model item
     * @return the item path
     */
    public String getPath();

    /**
     * Retrieves the model of the item
     * @return the model
     */
    public Model getModel();
    
    /**
     * Sets the model of the element
     * @param model the model to set
     */
    public void setModel(Model model);
    
    /**
     * Retrieves the parent of the item
     * @return the parent group
     */
    public ModelItemGroup getParent();

    /**
     * Sets the parent of the item
     * @param parent the parent to set
     */
    public void setParent(ModelItemGroup parent);
    
    /**
     * Converts the model item in a JSON map
     * @param context the context of the definition
     * @return The model item as a JSON map, or an empty map
     * @throws ProcessingException If an error occurs when converting the model item
     */
    public Map<String, Object> toJSON(DefinitionContext context) throws ProcessingException;
    
    /**
     * Generates SAX events for the model item
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events 
     * @param context the context of the definition
     * @throws SAXException if an error occurs during the SAX events generation
     */
    public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException;
    
    /**
     * Retrieves the type.
     * @return the type.
     */
    public ModelItemType getType();
    
    /**
     * Set the type.
     * @param type the type.
     */
    public void setType(ModelItemType type);
    
    /**
     * Retrieves the name of the plugin declaring this element.
     * @return the plugin name.
     */
    public String getPluginName();

    /**
     * Set the name of the plugin declaring this element.
     * @param pluginName the plugin name.
     */
    public void setPluginName(String pluginName);
    

    /**
     * Retrieves the widget to use for rendering.
     * @return the widget or <code>null</code> if none is defined.
     */
    public String getWidget();

    /**
     * Set the widget.
     * @param widget the widget.
     */
    public void setWidget(String widget);
    
    /**
     * Get the default widget to use when rendering this item as a criterion.
     * @return The default widget to use, or <code>null</code> if no specific widget is necessary.
     */
    public default String getCriterionWidget()
    {
        return getWidget();
    }
    
    /**
     * Get the widget's parameters
     * @return the widget's parameters
     */
    public Map<String, I18nizableText> getWidgetParameters();
    
    /**
     * Set the widget's parameters
     * @param params the parameters to set
     */
    public void setWidgetParameters (Map<String, I18nizableText> params);
    
    /**
     * Get the widget parameters to use when rendering this item as a criterion.
     * @param configuration The widget parameters configuration
     * @return The default widget parameters to use, or an empty Map if no specific widget parameters are necessary.
     */
    public default Map<String, I18nizableText> getCriterionWidgetParameters(Configuration configuration)
    {
        return getWidgetParameters();
    }
    
    /**
     * Retrieves the disable condition.
     * @return the disable condition or <code>null</code> if none is defined.
     */
    public DisableConditions getDisableConditions();

    /**
     * Sets the disable condition.
     * @param disableConditions the disable condition.
     */
    public void setDisableConditions(DisableConditions disableConditions);    
}
