/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.Collection;

import org.apache.commons.lang3.StringUtils;

import org.ametys.runtime.model.exception.UndefinedItemPathException;

/**
 * Interface for objects that can access to some model items
 */
public interface ModelItemAccessor
{
    /**
     * Retrieves the child model item with the given name
     * @param childName name of the model item child to retrieve
     * @return the child with the given name, or <code>null</code> if no child is found
     */
    public default ModelItem getChild(String childName)
    {
        for (ModelItem child : getModelItems())
        {
            if (StringUtils.equals(child.getName(), childName))
            {
                return child;
            }
            
            // If the accessor has no name, check its children
            if (child.getName() == null && child instanceof ModelItemAccessor)
            {
                ModelItem item = ((ModelItemAccessor) child).getChild(childName);
                if (item != null)
                {
                    return item;
                }
            }
        }

        return null;
    }
    
    /**
     * Retrieves the model item at given path
     * @param itemPath the item path
     * @return the model item.
     * @throws UndefinedItemPathException if there is no item defined at the given path 
     */
    public default ModelItem getModelItem(String itemPath) throws UndefinedItemPathException
    {
        String[] pathSegments = StringUtils.split(itemPath, ModelItem.ITEM_PATH_SEPARATOR);
        
        if (pathSegments.length > 0)
        {
            ModelItemAccessor currentAccessor = this;
            ModelItem child = null;
            for (int i = 0; i < pathSegments.length; i++)
            {
                child = currentAccessor.getChild(pathSegments[i]);

                if (child == null)
                {
                    throw new UndefinedItemPathException("The item at path '" + itemPath + "' is not defined in the model item accessor '" + toString() + "'.");
                }
                else if (i < pathSegments.length - 1 && !(child instanceof ModelItemAccessor))
                {
                    String currentAccessorPath = StringUtils.join(pathSegments, ModelItem.ITEM_PATH_SEPARATOR, 0, i + 1);
                    String subItemPath = StringUtils.join(pathSegments, ModelItem.ITEM_PATH_SEPARATOR, i + 1, pathSegments.length);
                    throw new UndefinedItemPathException("In the model item accessor '" + toString() + "', the item at path '" + currentAccessorPath + "' is not a model item accessor. So it can not access to a sub item with path '" + subItemPath + "'.");
                }
                else if (child instanceof ModelItemAccessor)
                {
                    currentAccessor = (ModelItemAccessor) child;
                }
            }
            
            return child;
        }
        else
        {
            throw new UndefinedItemPathException("The item at path '" + itemPath + "' is not defined in the model item accessor '" + toString() + "'.");
        }
    }
    
    /**
     * Checks if there is an item defined with the given path
     * @param itemPath path of the item
     * @return <code>true</code> if there is an item, <code>false</code> otherwise
     */
    public default boolean hasModelItem(String itemPath)
    {
        try
        {
            getModelItem(itemPath);
            return true;
        }
        catch (UndefinedItemPathException e)
        {
            // The getModelItem method throws an UndefinedItemPathException if there is no model item defined with the given path 
            return false;
        }
    }
    
    /**
     * Retrieves all the model items of this accessor
     * @return the model items
     */
    public Collection<? extends ModelItem> getModelItems();
}
