/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.xml.AttributesImpl;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.core.util.XMLUtils;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.model.type.ModelItemType;
import org.ametys.runtime.model.type.ModelItemTypeConstants;

/**
 * Class for group of model items
 */
public class ModelItemGroup extends AbstractModelItem implements ModelItemContainer
{
    /** Id for model item group types */
    public static final String DEFAULT_TYPE_ID = "composite";
    
    private List<ModelItem> _children = new ArrayList<>();
    private ElementDefinition<Boolean> _switcher;
    private ModelItemType _type;

    /**
     * Default constructor.
     */
    public ModelItemGroup()
    {
        super();
    }
    
    /**
     * Constructor used to create simple models and items 
     * @param name the name of the definition
     * @param children the group's children
     */
    public ModelItemGroup(String name, ModelItem... children)
    {
        super(name);
        Arrays.stream(children)
            .forEach(item -> addChild(item));
    }
    
    public Collection<ModelItem> getModelItems()
    {
        return getChildren();
    }
    
    /**
     * Retrieves the list of children model items
     * @return the children
     */
    public List<ModelItem> getChildren()
    {
        return getChildren(true);
    }
    
    /**
     * Retrieves the list of children model items, with or without the switcher
     * @param withSwitch true to retrieve the switcher with the other children
     * @return the children with or without the switcher
     */
    public List<ModelItem> getChildren(boolean withSwitch)
    {
        if (withSwitch || _switcher == null)
        {
            return Collections.unmodifiableList(_children);
        }
        
        List<ModelItem> childrenWithoutSwitcher = new ArrayList<>(_children);
        childrenWithoutSwitcher.remove(_switcher);
        return childrenWithoutSwitcher;
    }
    
    /**
     * Add a child in the group
     * @param child the item to add
     */
    public void addChild(ModelItem child)
    {
        addChild(child, false);
    }
    
    /**
     * Add a child in the group
     * @param child the child to add
     * @param isGroupSwitch true if the child to add is the group switch, false otherwise
     */
    @SuppressWarnings("unchecked")
    public void addChild(ModelItem child, boolean isGroupSwitch)
    {
        _children.add(child);
        child.setParent(this);
        
        if (isGroupSwitch)
        {
            if (_switcher == null)
            {
                if (child instanceof ElementDefinition)
                {
                    ElementType type = ((ElementDefinition) child).getType();
                    if (ModelItemTypeConstants.BOOLEAN_TYPE_ID.equals(type.getId()))
                    {
                        _switcher = (ElementDefinition<Boolean>) child;
                    }
                }
                
                if (_switcher == null)
                {
                    // If switcher is still null, the type of child is not compatible
                    throw new RuntimeException("The group '" + getLabel() + "' has a switch '" + child + "' that is not valid because it is not a boolean.");
                }
            }
            else
            {
                throw new RuntimeException("At least two group-switches have been defined for the configuration group '" + getLabel() + "'. These parameters are '" + _switcher + "' and '" + child + "'.");
            }
        }
    }
    
    /**
     * Retrieves the switcher element definition
     * @return the switcher
     */
    public ElementDefinition<Boolean> getSwitcher()
    {
        return _switcher;
    }
    
    @Override
    public void setModel(Model model)
    {
        super.setModel(model);
        for (ModelItem modelItem : _children)
        {
            modelItem.setModel(model);
        }
    }
    
    @Override
    public ModelItemType getType()
    {
        return _type;
    }

    public void setType(ModelItemType type)
    {
        _type = type;
    }
    
    @Override
    protected Map<String, Object> _toJSON(DefinitionContext context) throws ProcessingException
    {
        return _toJSON(context, true);
    }
    
    /**
     * Converts the model item group in a JSON map
     * @param context the context of the definition
     * @param includeChildren true to iterate and add children as elements in the returned JSON map, false otherwise
     * @return The model item as a JSON map
     * @throws ProcessingException If an error occurs when converting the model item group
     */
    public Map<String, Object> toJSON(DefinitionContext context, boolean includeChildren) throws ProcessingException
    {
        if (_shouldJSONBeEmpty(context))
        {
            return Map.of();
        }
        else
        {
            return _toJSON(context, includeChildren);
        }
    }
    
    /**
     * Converts the model item group in a JSON map
     * @param context the context of the definition
     * @param includeChildren true to iterate and add children as elements in the returned JSON map, false otherwise
     * @return The model item as a JSON map
     * @throws ProcessingException If an error occurs when converting the model item group
     */
    protected Map<String, Object> _toJSON(DefinitionContext context, boolean includeChildren) throws ProcessingException
    {
        Map<String, Object> result = super._toJSON(context);
        
        ModelItemType type = getType();
        result.put("type", type != null ? type.getId() : DEFAULT_TYPE_ID);
        
        ElementDefinition<Boolean> switcher = getSwitcher();
        if (switcher != null)
        {
            Map<String, Object> switcherToJSON = new HashMap<>();
            
            switcherToJSON.put("id", switcher.getName());
            switcherToJSON.put("label", switcher.getLabel());
            
            if (switcher.getType() != null)
            {
                switcherToJSON.put("default-value", switcher.getType().valueToJSONForClient(switcher.getDefaultValue(), DataContext.newInstance()));
            }
            
            result.put("switcher", switcherToJSON);
        }
        
        // Do not include switcher in children because it is already included above
        List<ModelItem> children = getChildren(false);
        if (includeChildren && !children.isEmpty())
        {
            Map<String, Object> elements = new LinkedHashMap<>();
            for (ModelItem child : children)
            {
                String name = child.getName();
                if (name != null)
                {
                    elements.put(name, child.toJSON(context));
                }
            }
            
            result.put("elements", elements);
        }
        
        return result;
    }
    
    /**
     * {@inheritDoc}
     * Do not generate SAX events for children.
     * This method has been created to generate SAX events for views, the children are taken into account via view items children 
     */
    @SuppressWarnings("static-access")
    @Override
    public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException
    {
        super.toSAX(contentHandler, context);
        
        ElementDefinition<Boolean> switcher = getSwitcher();
        if (switcher != null)
        {
            AttributesImpl switcherAttributes = new AttributesImpl();
            switcherAttributes.addCDATAAttribute("name", switcher.getName());
            
            XMLUtils.startElement(contentHandler, "switcher", switcherAttributes);
            XMLUtils.createI18nElementIfNotNull(contentHandler, "label", switcher.getLabel());
            
            if (switcher.getType() != null)
            {
                switcher.getType().valueToSAX(contentHandler, "default-value", switcher.getDefaultValue(), DataContext.newInstance());
            }
            
            XMLUtils.endElement(contentHandler, "switcher");
        }
    }
    
    /**
     * Creates a {@link ModelItemGroup}
     * @param name the group's name
     * @param children the group's children
     * @return the created {@link ModelItemGroup}
     */
    public static ModelItemGroup of(String name, ModelItem... children)
    {
        return new ModelItemGroup(name, children);
    }
}
