/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.core.util.XMLUtils;
import org.ametys.core.util.SizeUtils.ExcludeFromSizeCalculation;

/**
 * Simple group of view items
 */
public class SimpleViewItemGroup extends AbstractViewItemGroup
{
    @ExcludeFromSizeCalculation
    private ViewItemAccessor _parent;
    
    private String _name;
    
    public String getName()
    {
        return _name;
    }

    public void setName(String name)
    {
        _name = name;
    }
    
    public ViewItemAccessor getParent()
    {
        return _parent;
    }
    
    public void setParent(ViewItemAccessor parent)
    {
        _parent = parent;
    }
    
    public Map<String, Object> toJSON(DefinitionContext context) throws ProcessingException
    {
        Map<String, Object> result = new HashMap<>();
        
        result.put("name", getName());
        result.put("label", getLabel());
        result.put("description", getDescription());
        result.put("role", getRole());

        List<ViewItem> viewItems = getViewItems();
        if (!viewItems.isEmpty())
        {
            result.put("elements", ViewHelper.viewItemsToJSON(viewItems, context));
        }
        
        return result;
    }
    
    @SuppressWarnings("static-access")
    public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException
    {
        AttributesImpl attributes = new AttributesImpl();
        attributes.addCDATAAttribute("name", getName());
        attributes.addCDATAAttribute("role", getRole());
        
        XMLUtils.startElement(contentHandler, "group", attributes);
        
        XMLUtils.createI18nElementIfNotNull(contentHandler, "label", getLabel());
        XMLUtils.createI18nElementIfNotNull(contentHandler, "description", getDescription());
        
        for (ViewItem viewItem : getViewItems())
        {
            viewItem.toSAX(contentHandler, context);
        }
        
        XMLUtils.endElement(contentHandler, "group");
    }
    
    /**
     * Include the given view to the group.
     * Add the items of the view to include if they are not already present in the group or the reference view
     * @param viewToInclude the view to include
     * @param referenceView the reference view
     */
    public void includeView(View viewToInclude, View referenceView)
    {
        ViewHelper.addViewAccessorItems(this, viewToInclude, referenceView, StringUtils.EMPTY);
    }
    
    @Override
    public void copyTo(ViewItem item)
    {
        super.copyTo(item);
        
        assert item instanceof SimpleViewItemGroup;
        ((SimpleViewItemGroup) item).setName(getName());
    }
    
    public ViewItem createInstance()
    {
        return new SimpleViewItemGroup();
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = super.hashCode();
        result = prime * result + Objects.hash(_name);
        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        if (!super.equals(obj))
        {
            return false;
        }
        if (getClass() != obj.getClass())
        {
            return false;
        }
        SimpleViewItemGroup other = (SimpleViewItemGroup) obj;
        return Objects.equals(_name, other._name);
    }
    
    @Override
    public String toString()
    {
        return _name + ": " + _children.toString();
    }
}
