/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * View element that can access to other view items
 */
public class ViewElementAccessor extends ViewElement implements ViewItemAccessor
{
    private List<ViewItem> _viewItems = new ArrayList<>();

    public List<ViewItem> getViewItems()
    {
        return _viewItems;
    }

    public void addViewItem(ViewItem item)
    {
        _viewItems.add(item);
        item.setParent(this);
    }

    public void insertViewItem(ViewItem item, int index)
    {
        if (index >= 0 && index <= _viewItems.size())
        {
            _viewItems.add(index, item);
            item.setParent(this);
        }
        else
        {
            throw new IllegalArgumentException("Unable to insert an item at index " + index + ". This group contains " + _viewItems.size() + " items.");
        }
    }
    
    public boolean removeViewItem(ViewItem item)
    {
        return _viewItems.remove(item);
    }
    
    public void clear()
    {
        _viewItems.clear();
    }
    
    /**
     * Creates a {@link ViewElementAccessor} with the items of the given {@link ModelItemAccessor}
     * @param modelItemAccessor the model item accessor
     * @return the created {@link ViewElementAccessor}
     * @throws IllegalArgumentException if the model item accessor is <code>null</code>
     */
    public static ViewElementAccessor of(ModelItemAccessor modelItemAccessor) throws IllegalArgumentException
    {
        if (modelItemAccessor == null)
        {
            throw new IllegalArgumentException("Unable to create the view from a null model");
        }
        else
        {
            return ViewHelper.createViewItemAccessor(List.of(modelItemAccessor));
        }
    }
    
    /**
     * Creates a {@link ViewElementAccessor} with the given items
     * @param modelItemAccessor the model item accessing the items definitions
     * @param itemPaths the paths of the items to put in the view item
     * @return the created {@link ViewElementAccessor}
     * @throws IllegalArgumentException if the model item accessor is <code>null</code> or if an item path is <code>null</code>, empty, or is not defined in the given model items
     * @throws BadItemTypeException if a segment in a path (but not the last) does not represent a group item
     */
    public static ViewElementAccessor of(ModelItemGroup modelItemAccessor, String... itemPaths) throws IllegalArgumentException, BadItemTypeException
    {
        if (modelItemAccessor == null)
        {
            throw new IllegalArgumentException("Unable to create the view from a null model");
        }
        else
        {
            return ViewHelper.createViewItemAccessor(List.of(modelItemAccessor), itemPaths);
        }
    }
    
    @Override
    public void copyTo(ViewItem item, View referenceView, String itemPath)
    {
        super.copyTo(item, referenceView, itemPath);
        
        assert item instanceof ViewElementAccessor;
        ViewHelper.addViewAccessorItems((ViewElementAccessor) item, this, referenceView, itemPath);
    }
    
    @Override
    public ViewElementAccessor createInstance()
    {
        return new ViewElementAccessor();
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = super.hashCode();
        result = prime * result + Objects.hash(_viewItems);
        return result;
    }

    @Override
    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }
        if (!super.equals(obj))
        {
            return false;
        }
        if (getClass() != obj.getClass())
        {
            return false;
        }
        ViewElementAccessor other = (ViewElementAccessor) obj;
        return Objects.equals(_viewItems, other._viewItems);
    }
    
    @Override
    public String toString()
    {
        return super.toString() + ": " + _viewItems.toString();
    }
}
