/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.Map;

import org.apache.cocoon.ProcessingException;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.runtime.util.ModifiableLabelable;

/**
 * Interface for view items
 */
public interface ViewItem extends ModifiableLabelable
{
    /**
     * Create an instance of {@link ViewItem}
     * @return the created instance
     */
    public ViewItem createInstance();
    
    /**
     * Retrieves the parent of the view item
     * @return the parent group
     */
    public ViewItemAccessor getParent();

    /**
     * Sets the parent of the view item
     * @param parent the parent to set
     */
    public void setParent(ViewItemAccessor parent);

    /**
     * Converts the view item in a JSON map
     * @param context the context of the definitions referenced in this view item and/or its children
     * @return The view item as a JSON map
     * @throws ProcessingException If an error occurs when converting the view item
     */
    public Map<String, Object> toJSON(DefinitionContext context) throws ProcessingException;
    
    /**
     * Generates SAX events for the view item
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events 
     * @param context the context of the definitions included in the view item
     * @throws SAXException if an error occurs during the SAX events generation
     */
    public void toSAX(ContentHandler contentHandler, DefinitionContext context) throws SAXException;
    
    /**
     * Copy the current view item in the given one.
     * If the view item is an accessor, its view items are copied only if they are not already present in the reference view
     * @param item the copy
     * @param referenceView the reference view
     * @param itemPath the path of the view item in the reference view
     */
    public default void copyTo(ViewItem item, View referenceView, String itemPath)
    {
        copyTo(item);
    }
    
    /**
     * Copy the current view item in the given one.
     * If the view item is an accessor, its view items are not copied
     * @param item the copy
     */
    public void copyTo(ViewItem item);
    
    /**
     * Indicates whether some other object is "equal to" this one.
     * @param obj the reference object with which to compare.
     * @param checkDetails <code>true</code> to check the view item's details during comparison (label, description, ...)
     * @return <code>true</code> if this object is the same as the given obj, <code>false</code> otherwise.
     */
    public boolean equals(Object obj, boolean checkDetails);
}
