/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.List;

import org.apache.commons.lang3.StringUtils;

/**
 * Interface for objects that can access to some view items
 */
public interface ViewItemAccessor
{
    /**
     * Retrieves all the view items of this accessor
     * @return the view items
     */
    public List<ViewItem> getViewItems();
    
    /**
     * Add a view item in the accessor
     * @param item the item to add
     */
    public void addViewItem(ViewItem item);
    
    /**
     * Insert a view item in the accessor at the given index
     * @param item the item to insert
     * @param index index at which the item has to be inserted
     */
    public void insertViewItem(ViewItem item, int index);
    
    /**
     * Removes the given items of this accessor
     * @param item the item to remove
     * @return <code>true</code> if this accessor contained the given item, <code>false</code> otherwise
     */
    public boolean removeViewItem(ViewItem item);
    
    /**
     * Removes all items of this accessor
     */
    public void clear();
    
    /**
     * Add view items in the accessor
     * @param items the items to add
     */
    public default void addViewItems(List<ViewItem> items)
    {
        for (ViewItem item : items)
        {
            addViewItem(item);
        }
    }
    
    /**
     * Checks if the current {@link ViewItemAccessor} contains the given model view item
     * Consider that the accessor contains the given item if it contains an item representing the same definition
     * @param item the item
     * @return <code>true</code> if the accessor contains the given item, <code>false</code> otherwise 
     */
    public default boolean hasModelViewItem(ModelViewItem item)
    {
        return hasModelViewItem(item, StringUtils.EMPTY, item.getDefinition().getPath());
    }
    
    /**
     * Checks if the current {@link ViewItemAccessor} contains the given model view item
     * Consider that the accessor contains the given item if it contains an item representing the same definition AND with the same children if it is a group AND at the same position as the given path
     * @param item the item
     * @param currentAccessorPath The path of the current item accessor
     * @param itemPath the path of the item in the accessor
     * @return <code>true</code> if the accessor contains the given item, <code>false</code> otherwise 
     */
    public default boolean hasModelViewItem(ModelViewItem item, String currentAccessorPath, String itemPath)
    {
        for (ViewItem currentItem : getViewItems())
        {
            String currentItemPath = currentAccessorPath;
            if (currentItem instanceof ModelViewItem)
            {
                currentItemPath += StringUtils.isNotEmpty(currentAccessorPath) ? ModelItem.ITEM_PATH_SEPARATOR + currentItem.getName() : currentItem.getName();
                if (currentItemPath.equals(itemPath) && currentItem.equals(item))
                {
                    return true;
                }
            }
            
            if (currentItem instanceof ViewItemAccessor && ((ViewItemAccessor) currentItem).hasModelViewItem(item, currentItemPath, itemPath))
            {
                return true;
            }
        }
        
        // No item found
        return false;
    }
    
    /**
     * Checks if there is a model view item with the given name
     * @param itemName the name of the item
     * @return <code>true</code> if the accessor can access to an item with the given name, <code>false</code> otherwise 
     */
    public default boolean hasModelViewItem(String itemName)
    {
        return getModelViewItem(itemName) != null;
    }
    
    /**
     * Retrieves the model view item with the given name
     * If there are more than one corresponding items, the first one is retrieved
     * @param itemName the name of the item
     * @return the model view item, or <code>null</code> if none was found
     */
    public default ModelViewItem getModelViewItem(String itemName)
    {
        if (StringUtils.isEmpty(itemName))
        {
            return null;
        }
        else
        {
            for (ViewItem item : getViewItems())
            {
                // The item is a reference to the model
                if (item instanceof ModelViewItem modelViewItem)
                {
                    if (itemName.equals(item.getName()))
                    {
                        return modelViewItem;
                    }
                }
                // The item is not a reference to the model and is a group => check its children
                else if (item instanceof ViewItemAccessor viewItemAccessor)
                {
                    ModelViewItem modelViewItem = viewItemAccessor.getModelViewItem(itemName);
                    if (modelViewItem != null)
                    {
                        return modelViewItem;
                    }
                }
            }
        
            // No item found
            return null;
        }
    }
    
    /**
     * Retrieves the view item with the given name
     * If there are more than one corresponding items, the first one is retrieved
     * @param itemName the name of the item
     * @return the view item, or <code>null</code> if none was found
     */
    public default ViewItem getViewItem(String itemName)
    {
        return getViewItems().stream()
                             .filter(viewItem -> itemName.equals(viewItem.getName()))
                             .findFirst()
                             .orElse(null);
    }
}
