/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model.disableconditions;

import java.util.Arrays;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.component.Component;
import org.apache.commons.lang3.StringUtils;

import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.disableconditions.DisableCondition.OPERATOR;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Abstract evaluator for {@link DisableConditions}
 * @param <T> Type of object holding the data to evaluate
 */
public abstract class AbstractDisableConditionsEvaluator<T extends Object> extends AbstractLogEnabled implements Component, DisableConditionsEvaluator<T>
{
    public boolean evaluateDisableConditions(ModelItem definition, String dataPath, Optional<String> oldDataPath, Map<String, Object> values, T object, Map<String, Object> contextualParameters)
    {
        return evaluateDisableConditions(definition, definition.getDisableConditions(), dataPath, oldDataPath, values, object, contextualParameters);
    }
    
    /**
     * Recursively evaluate the given {@link DisableConditions} against the given values
     * @param definition the definition of the evaluated data
     * @param disableConditions the conditions to evaluate
     * @param dataPath the path of the evaluated data. Needed to get the value to compare as condition ids are relative to this one
     * @param oldDataPath the old path of the evaluated data. Needed to get stored value if the data has been moved
     * @param values values to check conditions on 
     * @param object the object holding the data to evaluate and the condition value
     * @param contextualParameters the contextual parameters
     * @return <code>true</code> if the disable conditions are <code>true</code>, <code>false</code> otherwise
     */
    protected boolean evaluateDisableConditions(ModelItem definition, DisableConditions disableConditions, String dataPath, Optional<String> oldDataPath, Map<String, Object> values, T object, Map<String, Object> contextualParameters)
    {
        if (!ModelHelper.hasDisableConditions(disableConditions))
        {
            return false;
        }
        
        boolean andOperator = disableConditions.getAssociationType() == DisableConditions.ASSOCIATION_TYPE.AND;
        
        // initial value depends on OR or AND associations
        boolean disabled = andOperator;
        
        for (DisableConditions subConditions : disableConditions.getSubConditions())
        {
            boolean result = evaluateDisableConditions(definition, subConditions, dataPath, oldDataPath, values, object, contextualParameters);
            if (_resultIsSufficient(andOperator, result))
            {
                return result;
            }
            else
            {
                disabled = andOperator ? disabled && result : disabled || result;
            }
        }
        
        for (DisableCondition condition : disableConditions.getConditions())
        {
            boolean result = evaluateDisableCondition(definition, condition, dataPath, oldDataPath, values, object, contextualParameters);
            if (_resultIsSufficient(andOperator, result))
            {
                return result;
            }
            else
            {
                disabled = andOperator ? disabled && result : disabled || result;
            }
        }
                
        return disabled;
    }
    
    /**
     * Check if the given result is sufficient to evaluate the current disable condition 
     * @param andOperator the conditions operator
     * @param result the result of the current condition
     * @return <code>true</code> if the result of evaluation is sufficient, <code>false</code> otherwise
     */
    protected boolean _resultIsSufficient(boolean andOperator, boolean result)
    {
        return andOperator && !result || !andOperator && result;
    }

    /**
     * Evaluate the given {@link DisableCondition} against the given values
     * @param definition the definition of the evaluated data
     * @param condition the condition to evaluate
     * @param dataPath the path of the evaluated data. Needed to get the value to compare as condition ids are relative to this one
     * @param oldDataPath the old path of the evaluated data. Needed to get stored value if the data has been moved
     * @param values values to check conditions on
     * @param object the object holding the data to evaluate and the condition value
     * @param contextualParameters the contextual parameters
     * @return <code>true</code> if the disable condition is <code>true</code>, <code>false</code> otherwise
     */
    protected boolean evaluateDisableCondition(ModelItem definition, DisableCondition condition, String dataPath, Optional<String> oldDataPath, Map<String, Object> values, T object, Map<String, Object> contextualParameters)
    {
        Optional<DefinitionAndValue> definitionAndValue = _getConditionDefinitionAndValue(definition, condition, dataPath, oldDataPath, values, object, contextualParameters);
        if (definitionAndValue.isEmpty())
        {
            getLogger().debug("Cannot evaluate the disable condition on the undefined element {}.\nReturning false.", condition.getId());
            return false;
        }

        ElementDefinition conditionDefinition = definitionAndValue.get().definition();
        ElementType type = conditionDefinition.getType();
        String valueFromCondition = condition.getValue();
        try
        {
            Object typedvalueFromCondition = type.castValue(valueFromCondition);
            Object value = definitionAndValue.get().value();
            
            DisableCondition.OPERATOR operator = condition.getOperator();
            if (value != null && value.getClass().isArray())
            {
                if (operator == OPERATOR.EQ)
                {
                    return Arrays.stream((Object[]) value)
                                 .anyMatch(v -> evaluateDisableConditionValue(typedvalueFromCondition, operator, v));
                }
                else
                {
                    return Arrays.stream((Object[]) value)
                                 .allMatch(v -> evaluateDisableConditionValue(typedvalueFromCondition, operator, v));
                }
            }
            else
            {
                return evaluateDisableConditionValue(typedvalueFromCondition, operator, value);
            }
        }
        catch (BadItemTypeException e)
        {
            throw new IllegalStateException("Cannot convert the condition value '" + valueFromCondition + "' to a '" + type.getId() + "' for model item '" + condition.getId() + "'", e);
        }
        catch (Exception e)
        {
            throw new IllegalStateException("An error occurred while comparing values in type '" + type.getId() + "' for model item '" + condition.getId() + "'.", e);
        }
    }
    
    /**
     * Retrieve the value corresponding to the identifier of the condition, and its type
     * @param object the object holding the data to evaluate and the condition value
     * @param definition the definition containing the condition
     * @param condition the condition 
     * @param dataPath the path of the evaluated data
     * @param oldDataPath the old path of the evaluated data. Needed to get stored value if the data has been moved
     * @param values values to check conditions on
     * @param contextualParameters the contextual parameters
     * @return the {@link DefinitionAndValue} corresponding to the condition identifier
     */
    protected Optional<DefinitionAndValue> _getConditionDefinitionAndValue(ModelItem definition, DisableCondition condition, String dataPath, Optional<String> oldDataPath, Map<String, Object> values, T object, Map<String, Object> contextualParameters)
    {
        String conditionAbsoluteDataPath = ModelHelper.getDisableConditionAbsolutePath(condition, dataPath);
        Optional<String> oldConditionAbsoluteDataPath = oldDataPath.map(path -> ModelHelper.getDisableConditionAbsolutePath(condition, path));
        Optional<ElementDefinition> conditionDefinition = getConditionDefinition(conditionAbsoluteDataPath, object);
        if (conditionDefinition.isEmpty())
        {
            return Optional.empty();
        }
        
        Object conditionValue = containsValue(conditionAbsoluteDataPath, oldConditionAbsoluteDataPath, values, object, contextualParameters)
                ? getValueFromMap(conditionAbsoluteDataPath, oldConditionAbsoluteDataPath, values, object, contextualParameters)
                : getStoredValue(conditionAbsoluteDataPath, oldConditionAbsoluteDataPath, object, contextualParameters);
        
        return Optional.of(new DefinitionAndValue(conditionDefinition.get(), conditionValue));
    }
    
    /**
     * Retrieves the condition definition
     * @param conditionDataPath the absolute data path of the condition
     * @param object the object holding the data to evaluate and the condition value
     * @return the condition definition
     */
    protected abstract Optional<ElementDefinition> getConditionDefinition(String conditionDataPath, T object);
    
    /**
     * Check if the values {@link Map} contains a value for the condition
     * @param conditionDataPath the absolute data path of the condition
     * @param oldConditionDataPath the absolute old data path of the condition
     * @param values the values {@link Map}
     * @param object the object holding the data to evaluate
     * @param contextualParameters the contextual parameters
     * @return <code>true</code> if there is a value (even empty) for the condition in the values {@link Map}, <code>false</code> otherwise
     */
    protected abstract boolean containsValue(String conditionDataPath, Optional<String> oldConditionDataPath, Map<String, Object> values, T object, Map<String, Object> contextualParameters);
    
    /**
     * Retrieves the condition value from the values {@link Map}
     * @param conditionDataPath the absolute data path of the condition
     * @param oldConditionDataPath the absolute old data path of the condition
     * @param values the values {@link Map}
     * @param object the object holding the data to evaluate
     * @param contextualParameters the contextual parameters
     * @return the condition value found in the values {@link Map}
     */
    protected abstract Object getValueFromMap(String conditionDataPath, Optional<String> oldConditionDataPath, Map<String, Object> values, T object, Map<String, Object> contextualParameters);
    
    /**
     * Retrieves the condition value from data stored in the given object
     * @param conditionDataPath the absolute data path of the condition
     * @param oldConditionDataPath the absolute old data path of the condition
     * @param object the object holding the data to evaluate and the condition value
     * @param contextualParameters the contextual parameters
     * @return the condition value from data stored in the object
     */
    protected abstract Object getStoredValue(String conditionDataPath, Optional<String> oldConditionDataPath, T object, Map<String, Object> contextualParameters);
    
    /**
     * Evaluate the given condition value against the given value
     * @param conditionValue the condition value
     * @param conditionOperator the condition operator
     * @param value the value to check
     * @return <code>true</code> if the condition is <code>true</code>, <code>false</code> otherwise
     */
    public static boolean evaluateDisableConditionValue(Object conditionValue, DisableCondition.OPERATOR conditionOperator, Object value)
    {
        if ((value == null || value instanceof Comparable) && (conditionValue == null || conditionValue instanceof Comparable))
        {
            @SuppressWarnings("unchecked")
            Comparable<Object> comparableParameterValue = (Comparable<Object>) _emptyStringToNull(value);
            @SuppressWarnings("unchecked")
            Comparable<Object> comparableCompareValue = (Comparable<Object>) _emptyStringToNull(conditionValue);
            
            int comparison;
            if (comparableParameterValue != null && comparableCompareValue != null)
            {
                comparison = comparableParameterValue.compareTo(comparableCompareValue);
            }
            else if (comparableCompareValue != null)
            {
                comparison = -1; // null comparableParameterValue is considered less than non-null comparableCompareValue
            }
            else if (comparableParameterValue != null)
            {
                comparison = 1; // non-null comparableParameterValue is considered greater than null comparableCompareValue
            }
            else
            {
                comparison = 0; // both are null
            }
            
            switch (conditionOperator)
            {
                case NEQ:
                    return comparison != 0;
                case GEQ:
                    return comparison >= 0;
                case GT:
                    return comparison > 0;
                case LT:
                    return comparison < 0;
                case LEQ:
                    return comparison <= 0;
                case EQ:
                default:
                    return comparison == 0;
            }
        }
        else
        {
            throw new IllegalStateException("Values '" + value + "' and '" + conditionValue + "' are not comparable");
        }
    }
    
    /**
     * Convert empty string to null object, or retrieve the given value
     * @param value the value
     * @return null if the given value is an empty string, the value itself otherwise
     */
    protected static Object _emptyStringToNull(Object value)
    {
        if (StringUtils.EMPTY.equals(value))
        {
            return null;
        }
        else
        {
            return  value;
        }
    }
    
    /**
     * Definition and value corresponding to a {@link DisableCondition}
     * @param definition {@link ElementDefinition} corresponding to the value
     * @param value the value
     */
    protected record DefinitionAndValue(ElementDefinition definition,  Object value) { /* empty */ }
}
