/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model.disableconditions;

import java.util.Map;
import java.util.Optional;

import org.ametys.runtime.model.ModelItem;

/**
 * Evaluator for {@link DisableConditions}
 * @param <T> Type of object holding the data to evaluate
 */
public interface DisableConditionsEvaluator<T extends Object>
{
    /**
     * Recursively evaluate the {@link DisableConditions} of the given model item, against the values stored in the given object
     * @param definition the definition of the evaluated data
     * @param dataPath the path of the evaluated data. Needed to get the value to compare as condition ids are relative to this one
     * @param object the object holding the data to evaluate and the condition value
     * @return <code>true</code> if the disable conditions are <code>true</code>, <code>false</code> otherwise
     */
    public default boolean evaluateDisableConditions(ModelItem definition, String dataPath, T object)
    {
        return evaluateDisableConditions(definition, dataPath, Optional.empty(), Map.of(), object, Map.of());
    }
    
    /**
     * Recursively evaluate the {@link DisableConditions} of the given model item, against the given values.
     * If the value of the condition is not present in the given {@link Map}, search in the stored values in the given object  
     * @param definition the definition of the evaluated data
     * @param dataPath the path of the evaluated data. Needed to get the value to compare as condition ids are relative to this one
     * @param oldDataPath the old path of the evaluated data. Needed to get stored value if the data has been moved
     * @param values values to check conditions on 
     * @param object the object holding the data to evaluate and the condition value
     * @param contextualParameters the contextual parameters
     * @return <code>true</code> if the disable conditions are <code>true</code>, <code>false</code> otherwise
     */
    public boolean evaluateDisableConditions(ModelItem definition, String dataPath, Optional<String> oldDataPath, Map<String, Object> values, T object, Map<String, Object> contextualParameters);
}
