/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model.type;

import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;

import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ViewItem;

/**
 * Object that gives some context for data manipulation
 */
public class DataContext
{
    private List<String> _fullDataPath = new ArrayList<>();
    private List<String> _dataPath = new ArrayList<>();
    
    private Optional<ModelItem> _modelItem = Optional.empty();
    private Optional<ViewItem> _viewItem = Optional.empty();

    private Locale _locale;
    private boolean _renderEmptyValues = true;

    /**
     * Creates a new instance of a {@link DataContext}
     */
    protected DataContext()
    {
        // Empty private constructor
    }
    
    /**
     * Creates a new instance of a {@link DataContext} from another {@link DataContext}
     * @param context the data context to copy
     */
    protected DataContext(DataContext context)
    {
        withDataPath(context.getDataPath());
        withFullDataPath(context.getFullDataPath());
        
        withModelItem(context.getModelItem().orElse(null));
        withViewItem(context.getViewItem().orElse(null));

        withLocale(context.getLocale());
        withEmptyValues(context.renderEmptyValues());
    }
    
    /**
     * Creates a new instance of a {@link DataContext}
     * @return the created instance
     */
    public static DataContext newInstance()
    {
        return new DataContext();
    }
    
    /**
     * Creates a new instance of a {@link DataContext} from another {@link DataContext}.
     * It can be the same implementation or another one, but it will be casted to the current implementation.
     * @param context the data context to copy
     * @return the created instance
     */
    public static DataContext newInstance(DataContext context)
    {
        return new DataContext(context);
    }
    
    /**
     * Creates a new instance of a {@link DataContext}, with the current context values
     * @return the created instance
     */
    public DataContext cloneContext()
    {
        return newInstance(this);
    }
    
    /**
     * Retrieves the data path
     * @return the dataPath
     */
    public String getDataPath()
    {
        return StringUtils.join(_dataPath, ModelItem.ITEM_PATH_SEPARATOR);
    }
    
    /**
     * Retrieves the last segment of the data path
     * @return the last segment of the data path
     */
    public String getDataPathLastSegment()
    {
        return _dataPath.get(_dataPath.size() - 1);
    }
    
    /**
     * Sets the data path
     * @param <T> the type of the retrieved {@link DataContext}
     * @param dataPath the data path to set
     * @return the current {@link DataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends DataContext> T withDataPath(String dataPath)
    {
        _dataPath = new ArrayList<>();
        if (StringUtils.isNotEmpty(dataPath))
        {
            _dataPath.add(dataPath);
            _fullDataPath.add(dataPath);
        }
        return (T) this;
    }
    
    /**
     * Retrieves the full data path
     * @return the fullDataPath
     */
    public String getFullDataPath()
    {
        return StringUtils.join(_fullDataPath, ModelItem.ITEM_PATH_SEPARATOR);
    }
    
    /**
     * Sets the full data path
     * @param <T> the type of the retrieved {@link DataContext}
     * @param fullDataPath the full data path to set
     * @return the current {@link DataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends DataContext> T withFullDataPath(String fullDataPath)
    {
        _fullDataPath = new ArrayList<>();
        if (StringUtils.isNotEmpty(fullDataPath))
        {
            _fullDataPath.add(fullDataPath);
        }
        return (T) this;
    }
    
    /**
     * Add a segment to the data path
     * @param <T> the type of the retrieved {@link DataContext}
     * @param segment the segment to add to the data path
     * @return the current {@link DataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends DataContext> T addSegmentToDataPath(String segment)
    {
        _dataPath.add(segment);
        _fullDataPath.add(segment);
        return (T) this;
    }
    
    /**
     * Add a suffix to the last segment of the data path
     * @param <T> the type of the retrieved {@link DataContext}
     * @param suffix the suffix to add to the last segment
     * @return the current {@link DataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends DataContext> T addSuffixToLastSegment(String suffix)
    {
        int segmentIndex = _dataPath.size() - 1;
        _dataPath.set(segmentIndex, _dataPath.get(segmentIndex) + suffix);
        
        segmentIndex = _fullDataPath.size() - 1;
        _fullDataPath.set(segmentIndex, _fullDataPath.get(segmentIndex) + suffix);
        
        return (T) this;
    }
    
    /**
     * Retrieves the model item
     * @return the model item
     */
    public Optional<ModelItem> getModelItem()
    {
        return _modelItem;
    }
    
    /**
     * Sets the model item
     * @param <T> the type of the retrieved {@link DataContext}
     * @param modelItem the model item to set
     * @return the current {@link DataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends DataContext> T withModelItem(ModelItem modelItem)
    {
        _modelItem = Optional.ofNullable(modelItem);
        return (T) this;
    }
    
    /**
     * Retrieves the view item
     * @return the view item
     */
    public Optional<ViewItem> getViewItem()
    {
        return _viewItem;
    }
    
    /**
     * Sets the view item
     * @param <T> the type of the retrieved {@link DataContext}
     * @param viewItem the view item to set
     * @return the current {@link DataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends DataContext> T withViewItem(ViewItem viewItem)
    {
        _viewItem = Optional.ofNullable(viewItem);
        return (T) this;
    }
    
    /**
     * Retrieves the locale to use for localized data
     * @return the locale, or null if no local has been set
     */
    public Locale getLocale()
    {
        return _locale;
    }
    
    /**
     * Sets the locale to use for localized data.
     * Do not set {@link Locale} to use all existing ones.
     * @param <T> the type of the retrieved {@link DataContext}
     * @param locale the locale to set
     * @return the current {@link DataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends DataContext> T withLocale(Locale locale)
    {
        _locale = locale;
        return (T) this;
    }
    
    /**
     * Determines if the empty values have to be rendered
     * @return <code>true</code> if empty values must be rendered (default), <code>false</code> otherwise
     */
    public boolean renderEmptyValues()
    {
        return _renderEmptyValues;
    }
    
    /**
     * Set to <code>false</code> to not render empty values (default to <code>true</code>)
     * @param <T> the type of the retrieved {@link DataContext}
     * @param renderEmptyValues <code>true</code> to render the empty values, <code>false</code> otherwise
     * @return the current {@link DataContext}
     */
    @SuppressWarnings("unchecked")
    public <T extends DataContext> T withEmptyValues(boolean renderEmptyValues)
    {
        _renderEmptyValues = renderEmptyValues;
        return (T) this;
    }
}
