/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model.type;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.plugin.component.PluginAware;

/**
 * Interface for model item types
 */
public interface ModelItemType extends PluginAware
{
    /**
     * Retrieves the identifier of the parameter type
     * @return the identifier of the parameter type
     */
    public String getId();

    /**
     * Generates SAX events for the given value
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param value the value to SAX
     * @param context The context of the data to SAX
     * @throws SAXException if an error occurs during the SAX events generation
     * @throws BadItemTypeException if the given value is not compatible with the current type
     */
    public void valueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context) throws SAXException, BadItemTypeException;
    
    /**
     * Generates SAX events for the given value in edition mode.
     * This method exist for legacy purposes because some types have to be saxed as JSON
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param value the value to SAX
     * @param context The context of the data to SAX
     * @throws SAXException if an error occurs during the SAX events generation
     * @throws BadItemTypeException if the given value is not compatible with the current type
     */
    public void valueToSAXForEdition(ContentHandler contentHandler, String tagName, Object value, DataContext context) throws SAXException, BadItemTypeException;
    
    /**
     * Convert the value into a JSON object to use client side
     * @param value the value to convert
     * @param context The context of the data to convert. Can not be null.
     * @return The value as JSON
     * @throws BadItemTypeException if the given value is not compatible is the current type
     */
    public Object valueToJSONForClient(Object value, DataContext context) throws BadItemTypeException;
    
    /**
     * Convert the value into a JSON object to use in edition mode
     * @param value the value to convert
     * @param context The context of the data to convert. Can not be null.
     * @return The value as JSON
     * @throws BadItemTypeException if the given value is not compatible is the current type
     */
    public Object valueToJSONForEdition(Object value, DataContext context) throws BadItemTypeException;
}
