/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model.type.xml;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.List;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.ElementType;

/**
 * Interfaces for XML parameter types
 * @param <T> Type of the parameter value
 */
public interface XMLElementType<T> extends ElementType<T>
{
    /**
     * Read the value in the given XML configuration
     * @param parentConfiguration XML configuration containing the value
     * @param name the name of the element to read
     * @return the value
     * @throws ConfigurationException if an error occurs while reading the given configuration
     */
    public default Object read(Configuration parentConfiguration, String name) throws ConfigurationException
    {
        Configuration elementConfiguration = parentConfiguration.getChild(name);
        
        return readValueFromNode(elementConfiguration);
    }
    
    /**
     * Read the value in the given XML configuration (the node directly, not the parent configuration)
     * @param elementConfiguration XML configuration node containing the value
     * @return the value
     * @throws ConfigurationException if an error occurs while reading the given configuration
     */
    public default Object readValueFromNode(Configuration elementConfiguration) throws ConfigurationException
    {
        if (elementConfiguration == null)
        {
            return null;
        }
        
        // Multiple element management
        Configuration[] valuesConfiguration = elementConfiguration.getChildren("value");
        if (valuesConfiguration.length > 0)
        {
            List<T> values = new ArrayList<>();
            for (Configuration valueConfiguration : valuesConfiguration)
            {
                T value = parseConfiguration(valueConfiguration);
                if (value != null)
                {
                    values.add(value);
                }
            }
            return Array.newInstance(getManagedClass(), values.size());
        }
        else
        {
            // Simple element management
            return parseConfiguration(elementConfiguration);
        }
    }
    
    /**
     * Write the value into the given content handler
     * @param contentHandler the content handler where to SAX into.
     * @param name the name of the element to write
     * @param value the value to write
     * @throws SAXException if an errors occurs during the value writing
     * @throws BadItemTypeException If the given value doesn't match this element type
     */
    @SuppressWarnings("unchecked")
    public default void write(ContentHandler contentHandler, String name, Object value) throws SAXException, BadItemTypeException
    {
        if (value == null)
        {
            XMLUtils.createElement(contentHandler, name);
        }
        else if (getManagedClass().isInstance(value))
        {
            XMLUtils.createElement(contentHandler, name, StringUtils.defaultString(toString((T) value)));
        }
        else if (getManagedClassArray().isInstance(value))
        {
            XMLUtils.startElement(contentHandler, name);
            for (T singleValue : (T[]) value)
            {
                XMLUtils.createElement(contentHandler, "value", StringUtils.defaultString(toString(singleValue)));
            }
            XMLUtils.endElement(contentHandler, name);
        }
        else
        {
            throw new BadItemTypeException("Try to write the non " + getManagedClass().getName() + " value '" + value + "' on element named '" + name + "'");
        }
    }
}
