/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugin;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.collections4.CollectionUtils;
import org.slf4j.Logger;

import org.ametys.runtime.plugin.IncomingDeactivation.Type;

class CorrectedDependencies
{
    private Map<String, Feature> _allFeatures;
    private Map<Feature, Collection<String>> _correctedDependencies = new ConcurrentHashMap<>();
    private Map<String, Collection<IncomingDeactivation>> _incomingDeactivations;
    private Logger _logger;
    
    CorrectedDependencies(Map<String, Feature> allFeatures, Map<String, Collection<IncomingDeactivation>> incomingDeactivations, Logger logger)
    {
        _allFeatures = allFeatures;
        _incomingDeactivations = incomingDeactivations;
        _logger = logger;
        _computeCorrectedDependencies();
    }
    
    private void _computeCorrectedDependencies()
    {
        for (Feature feature : _allFeatures.values())
        {
            _correctedDependencies.put(feature, _computeCorrectedDependencies(feature));
        }
    }
    
    Collection<String> getCorrectedDependencies(Feature feature)
    {
        return _correctedDependencies.get(feature);
    }
    
    // getDependencies taking account of deactivations
    private Collection<String> _computeCorrectedDependencies(Feature feature)
    {
        Collection<String> dependencies = feature.getDependencies();
        // If we have for instance feature 'overrider' which overrides feature 'overridden' 
        // _incomingDeactivations={'overridden'=[IncomingDeactivation('overrider', OVERRIDDEN)], ...}
        // 
        // And if dependencies contains 'overridden', then it must be replaced with 'overrider'
        List<String> correctedDependencies = dependencies
                .stream()
                .map(dependency -> _maybeReplaceDependencyByDeactivators(dependency, feature))
                .flatMap(Collection::stream)
                .collect(Collectors.toList());
        return correctedDependencies;
    }
    
    private Collection<String> _maybeReplaceDependencyByDeactivators(String dependency, Feature originalFeature)
    {
        Collection<IncomingDeactivation> deactivators = _incomingDeactivations.get(dependency);
        if (CollectionUtils.isEmpty(deactivators))
        {
            // nobody deactivates or overrides it => keep the dependency
            return Collections.singleton(dependency);
        }
        else if (_hasDeactivatorsAndNoOverride(deactivators))
        {
            _logger.debug("Dependency '{}' (of feature '{}') is deactivated by one or more feature(s) ({}). Nothing special made.", 
                    dependency, originalFeature.getFeatureId(), deactivators);
            return Collections.singleton(dependency);
        }
        else
        {
            // AbstractFeatureActivator._checkNoMultipleOverriders ensures that deactivators contains at most 1 overrider
            // Thus here (with previous if statements), deactivators contains ONE AND ONLY ONE overrider
            
            // replace the dependency by its unique feature which overrides it
            String overrider = _getOverride(deactivators);
            _logger.info("Dependency '{}' (of feature '{}') is overridden by one feature, it is replaced by it: '{}'.", 
                    dependency, originalFeature.getFeatureId(), overrider);
            Collection<String> recursiveOverride = _maybeReplaceDependencyByDeactivators(overrider, _allFeatures.get(overrider));
            if (CollectionUtils.isNotEmpty(recursiveOverride))
            {
                overrider = recursiveOverride.iterator().next();
            }
            return Collections.singleton(overrider);
        }
    }
    
    private String _getOverride(Collection<IncomingDeactivation> deactivators)
    {
        return _onlyOverriden(deactivators)
                .findAny()
                .map(IncomingDeactivation::getFeatureId)
                .orElseThrow();
    }
    
    private Stream<IncomingDeactivation> _onlyOverriden(Collection<IncomingDeactivation> deactivators)
    {
        return deactivators.stream()
                .filter(deactivator -> deactivator.getType() == Type.OVERRIDDEN);
    }
    
    private boolean _hasDeactivatorsAndNoOverride(Collection<IncomingDeactivation> deactivators)
    {
        return _onlyOverriden(deactivators)
                .findAny()
                .isEmpty();
    }
    
    @Override
    public String toString()
    {
        return _correctedDependencies.entrySet()
                .stream()
                .collect(Collectors.toMap(
                    e -> e.getKey().getFeatureId(), 
                    e -> e.getValue()))
                .toString();
    }
}
