/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugin;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.ametys.runtime.plugin.PluginsManager.InactivityCause;

/**
 * Default {@link FeatureActivator} which activates all non-passive and non-excluded features.
 */
public final class ExcludePolicyFeatureActivator extends AbstractFeatureActivator
{
    private Collection<String> _excludedPlugins;
    private Collection<String> _excludedFeatures;
    /**
     * Constructs a new feature activator with an 'exclude' policy
     * @param allPlugins all plugins
     * @param excludedPlugins The plugins to exclude
     * @param excludedFeatures The features to exclude
     */
    public ExcludePolicyFeatureActivator(Map<String, Plugin> allPlugins, Collection<String> excludedPlugins, Collection<String> excludedFeatures)
    {
        super(allPlugins);
        _excludedPlugins = excludedPlugins;
        _excludedFeatures = excludedFeatures;
    }

    @Override
    public PluginsInformation computeActiveFeatures(Map<String, String> componentsConfig, boolean safeMode)
    {
        _safeMode = safeMode;
        
        Map<String, Feature> initialFeatures = new HashMap<>();
        Map<String, ExtensionPointDefinition> extensionPoints = new HashMap<>();
        Map<String, InactivityCause> inactiveFeatures = new HashMap<>();
        
        Collection<PluginIssue> errors = new ArrayList<>();

        // Get actual plugin list, corresponding extension points and initial feature list
        Map<String, Plugin> plugins = computeActivePlugins(_excludedPlugins, initialFeatures, inactiveFeatures, extensionPoints, errors);

        // Compute incoming deactivations
        Map<String, Collection<IncomingDeactivation>> incomingDeactivations = computeIncomingDeactivations(initialFeatures);
        _correctedDependencies = new CorrectedDependencies(initialFeatures, incomingDeactivations, _logger);
        
        // First remove user-excluded features
        _removeUserExcludedFeatures(initialFeatures, inactiveFeatures);
        
        // Then remove deactivated features
        // Also remove feature containing inactive components
        removeInactiveFeatures(initialFeatures, inactiveFeatures, incomingDeactivations, componentsConfig);
        
        removeWrongPointReferences(initialFeatures, inactiveFeatures, extensionPoints, errors);
        
        // Process outgoing dependencies
        Map<String, Feature> features = processOutgoingDependencies(initialFeatures, inactiveFeatures, errors);

        // Compute incoming dependencies
        Map<String, Collection<String>> incomingDependencies = computeIncomingDependencies(features);
        
        // Finally remove unused passive features
        removeUnusedPassiveFeatures(features, inactiveFeatures, incomingDependencies);
        
        // Check uniqueness of extensions and components
        Map<String, Map<String, ExtensionDefinition>> extensions = computeExtensions(features, errors);
        Map<String, ComponentDefinition> components = computeComponents(features, componentsConfig, errors);
        
        return new PluginsInformation(plugins, features, inactiveFeatures, extensionPoints, extensions, components, errors);
    }
    
    private void _removeUserExcludedFeatures(Map<String, Feature> initialFeatures, Map<String, InactivityCause> inactiveFeatures)
    {
        Set<String> ids = initialFeatures.keySet();
        Iterator<String> it = ids.iterator();
        while (it.hasNext())
        {
            String id = it.next();
            
            if (_excludedFeatures.contains(id))
            {
                _logger.debug("Remove excluded feature '{}'", id);
                it.remove();
                inactiveFeatures.put(id, InactivityCause.EXCLUDED);
            }
        }
    }
    
    @Override
    public String shortDump(PluginsInformation pluginInfo)
    {
        // Do nothing in INFO log level
        return "";
    }
}
