/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugin;

import java.util.Collection;
import java.util.Map;

import org.ametys.runtime.plugin.PluginsManager.InactivityCause;

/**
 * An activator of {@link Feature features}, which will be loaded by the {@link PluginsManager}.
 * <br>The implemententations must define the strategy in order to determine which extension points, components, etc. need to be loaded.
 */
public interface FeatureActivator
{
    /**
     * Computes the actual plugins and features to load, based on values selected by the administrator.<br>
     * This method does not actually load nor execute any Java code. It reads plugins definitions, selects active features and gets components and extensions definitions.
     * @param componentsConfig chosen components, among those with the same role.
     * @param safeMode <code>true</code> if application is in safe mode
     * @return all information gathered during plugins reading.
     */
    PluginsInformation computeActiveFeatures(Map<String, String> componentsConfig, boolean safeMode);
    
    /**
     * Dumps the given {@link PluginsInformation}, in order to log in debug level.
     * @param pluginInfo The information about plugins
     * @return The string to display in debug level
     */
    String fullDump(PluginsInformation pluginInfo);
    
    /**
     * Dumps the given {@link PluginsInformation}, in order to log in info level.
     * @param pluginInfo The information about plugins
     * @return The string to display in info level
     */
    String shortDump(PluginsInformation pluginInfo);
    
    /**
     * Helper class containing all relevant information after features list computation.
     */
    static class PluginsInformation
    {
        private Map<String, Plugin> _plugins;
        private Map<String, Feature> _features;
        private Map<String, InactivityCause> _inactiveFeatures;
        private Map<String, ExtensionPointDefinition> _extensionPoints;
        private Map<String, Map<String, ExtensionDefinition>> _extensions;
        private Map<String, ComponentDefinition> _components;
        private Collection<PluginIssue> _errors;
        
        PluginsInformation(Map<String, Plugin> plugins, Map<String, Feature> features, Map<String, InactivityCause> inactiveFeatures, Map<String, ExtensionPointDefinition> extensionPoints, Map<String, Map<String, ExtensionDefinition>> extensions, Map<String, ComponentDefinition> components, Collection<PluginIssue> errors)
        {
            _plugins = plugins;
            _features = features;
            _inactiveFeatures = inactiveFeatures;
            _extensionPoints = extensionPoints;
            _extensions = extensions;
            _components = components;
            _errors = errors;
        }
        
        Map<String, Plugin> getPlugins()
        {
            return _plugins;
        }
        
        Map<String, Feature> getFeatures()
        {
            return _features;
        }
        
        Map<String, InactivityCause> getInactiveFeatures()
        {
            return _inactiveFeatures;
        }
        
        Map<String, ExtensionPointDefinition> getExtensionPoints()
        {
            return _extensionPoints;
        }
        
        Map<String, Map<String, ExtensionDefinition>> getExtensions()
        {
            return _extensions;
        }
        
        Map<String, ComponentDefinition> getComponents()
        {
            return _components;
        }
        
        /**
         * Returns all errors collected during initialization phase.
         * @return all errors collected during initialization phase.
         */
        Collection<PluginIssue> getErrors()
        {
            return _errors;
        }
    }
}
