/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugin;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import org.ametys.runtime.plugin.IncludePolicyFeatureActivator.IncludedFeature;

class IncludedFeatureConsistencyChecker
{
    private Collection<IncludedFeature> _includedFeatures;
    private Map<String, Feature> _initialFeatures;
    private Map<String, Collection<IncomingDeactivation>> _incomingDeactivations;

    IncludedFeatureConsistencyChecker(Collection<IncludedFeature> includedFeatures, Map<String, Feature> initialFeatures, Map<String, Collection<IncomingDeactivation>> incomingDeactivations)
    {
        _includedFeatures = includedFeatures;
        _initialFeatures = initialFeatures;
        _incomingDeactivations = incomingDeactivations;
    }
    
    void checkConsistency()
    {
        List<String> includedFeatures = _includedFeatures.stream()
                .map(IncludedFeature::featureId)
                .collect(Collectors.toList());
        for (String includedFeatureId : includedFeatures)
        {
            _checkFeature(includedFeatureId);
        }
    }
    
    private void _checkFeature(String featureId)
    {
        Feature feature = Objects.requireNonNull(_initialFeatures.get(featureId), String.format("Feature '%s' cannot be found.", featureId));
        Collection<String> dependencies = feature.getDependencies();
        for (String dependency : dependencies)
        {
            if (_incomingDeactivations.containsKey(dependency) && _allDeactivationsAreDeactivatedType(dependency))
            {
                throw new IllegalStateException(String.format(
                        "Problem of consistency with included features %s. At least one of them is deactivated by another. The application will not be able to boot.", 
                        _includedFeatures));
            }
            _checkFeature(dependency);
        }
    }
    
    private boolean _allDeactivationsAreDeactivatedType(String dependency)
    {
        Collection<IncomingDeactivation> incomingDeactivations = _incomingDeactivations.get(dependency);
        return incomingDeactivations.stream()
                .map(IncomingDeactivation::getType)
                .allMatch(IncomingDeactivation.Type.DEACTIVATED::equals);
    }
}
