/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugin;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.ArrayUtils;

import org.ametys.runtime.plugin.FeatureActivator.PluginsInformation;
import org.ametys.runtime.plugin.IncludePolicyFeatureActivator.IncludedFeature;
import org.ametys.runtime.plugin.PluginsManager.InactivityCause;
import org.ametys.runtime.servlet.RuntimeConfig;

import com.google.common.base.Predicates;

class LoadedFeaturesDump
{
    private AbstractFeatureActivator _featureActivator;

    LoadedFeaturesDump(AbstractFeatureActivator featureActivator)
    {
        _featureActivator = featureActivator;
    }
    
    String fullDump(PluginsInformation pluginInfo)
    {
        Map<String, InactivityCause> inactiveFeatures = pluginInfo.getInactiveFeatures();
        Collection<PluginIssue> errors = pluginInfo.getErrors();
        
        Collection<String> excludedPlugins = RuntimeConfig.getInstance().getExcludedPlugins();
        StringBuilder sb = new StringBuilder();
        
        for (String pluginName : _featureActivator._allPlugins.keySet())
        {
            Plugin plugin = _featureActivator._allPlugins.get(pluginName);
            _fullDumpPlugin(sb, plugin, excludedPlugins, inactiveFeatures);
        }
        
        if (!errors.isEmpty())
        {
            sb.append("\nErrors :\n");
            errors.forEach(issue -> sb.append(issue.toString()).append('\n'));
        }
        
        return sb.toString();
    }
    
    private void _fullDumpPlugin(StringBuilder sb, Plugin plugin, Collection<String> excludedPlugins, Map<String, InactivityCause> inactiveFeatures)
    {
        String pluginName = plugin.getName();
        sb.append("Plugin ").append(pluginName);
        
        if (excludedPlugins.contains(pluginName))
        {
            sb.append("   *** excluded ***");
        }
        
        sb.append('\n');
        
        Collection<String> configParameters = plugin.getConfigParameters().keySet();
        if (!CollectionUtils.isEmpty(configParameters))
        {
            sb.append("  Config parameters : \n");
            configParameters.forEach(param -> sb.append("    ").append(param).append('\n'));
        }
        
        Collection<String> paramCheckers = plugin.getParameterCheckers().keySet();
        if (!CollectionUtils.isEmpty(paramCheckers))
        {
            sb.append("  Parameters checkers : \n");
            paramCheckers.forEach(param -> sb.append("    ").append(param).append('\n'));
        }
        
        Collection<String> extensionPoints = plugin.getExtensionPoints();
        if (!CollectionUtils.isEmpty(extensionPoints))
        {
            sb.append("  Extension points : \n");
            extensionPoints.forEach(point -> sb.append("    ").append(point).append('\n'));
        }
        
        Map<String, Feature> features = plugin.getFeatures();
        for (String featureId : features.keySet())
        {
            Feature feature = features.get(featureId);
            _fullDumpFeature(sb, feature, inactiveFeatures);
        }
        
        sb.append('\n');
    }
    
    private void _fullDumpFeature(StringBuilder sb, Feature feature, Map<String, InactivityCause> inactiveFeatures)
    {
        String featureId = feature.getFeatureId();
        
        sb.append("  Feature ").append(featureId);
        if (feature.isPassive())
        {
            sb.append(" (passive)");
        }

        if (feature.isSafe())
        {
            sb.append(" (safe)");
        }
        
        if (inactiveFeatures != null && inactiveFeatures.containsKey(featureId))
        {
            sb.append("   *** inactive (").append(inactiveFeatures.get(featureId)).append(") ***");
        }
        
        sb.append('\n');
        
        Collection<String> featureConfigParameters = feature.getConfigParameters().keySet();
        if (!CollectionUtils.isEmpty(featureConfigParameters))
        {
            sb.append("    Config parameters : \n");
            featureConfigParameters.forEach(param -> sb.append("      ").append(param).append('\n'));
        }
        
        Collection<String> configParametersReferences = feature.getConfigParametersReferences();
        if (!CollectionUtils.isEmpty(configParametersReferences))
        {
            sb.append("    Config parameters references : \n");
            configParametersReferences.forEach(param -> sb.append("      ").append(param).append('\n'));
        }
        
        Collection<String> featureParamCheckers = feature.getParameterCheckers().keySet();
        if (!CollectionUtils.isEmpty(featureParamCheckers))
        {
            sb.append("    Parameters checkers : \n");
            featureParamCheckers.forEach(param -> sb.append("    ").append(param).append('\n'));
        }
        
        Map<String, String> componentsIds = feature.getComponentsIds();
        if (!componentsIds.isEmpty())
        {
            sb.append("    Components : \n");
            
            for (String role : componentsIds.keySet())
            {
                String id = componentsIds.get(role);
                sb.append("      ").append(role).append(" : ").append(id).append('\n');
            }
            
            sb.append('\n');
        }

        Map<String, Collection<String>> extensionsIds = feature.getExtensionsIds();
        if (!extensionsIds.isEmpty())
        {
            sb.append("    Extensions : \n");
            
            for (Entry<String, Collection<String>> extensionEntry : extensionsIds.entrySet())
            {
                String point = extensionEntry.getKey();
                Collection<String> ids = extensionEntry.getValue();
                
                sb.append("      ").append(point).append(" :\n");
                ids.forEach(id -> sb.append("        ").append(id).append('\n'));
            }
            
            sb.append('\n');
        }
    }
    
    String shortDump(PluginsInformation pluginInfo, Collection<IncludedFeature> includedFeatures)
    {
        Collection<PluginIssue> errors = pluginInfo.getErrors();
        Map<String, InactivityCause> inactiveFeatures = pluginInfo.getInactiveFeatures();
        Map<String, Collection<String>> dependencies = _featureActivator.computeIncomingDependencies(pluginInfo.getFeatures());
        StringBuilder sb = new StringBuilder();
        
        String separatorLine = new String(new char[66]).replace('\0', '-') + "\n";
        sb.append(separatorLine);
        _dumpLoaded(sb, separatorLine, pluginInfo);
        sb.append(separatorLine);
        
        List<String> pluginNames = _featureActivator._allPlugins.keySet()
                   .stream()
                   .sorted()
                   .collect(Collectors.toList());
        for (String pluginName : pluginNames)
        {
            Plugin plugin = _featureActivator._allPlugins.get(pluginName);
            _shortDumpPlugin(sb, plugin, dependencies, includedFeatures, inactiveFeatures);
        }
        
        if (!errors.isEmpty())
        {
            sb.append("\nErrors :\n");
            errors.forEach(
                issue -> sb.append(issue.toString()).append('\n')
            );
        }
        
        sb.append(separatorLine);
        
        return sb.toString();
    }
    
    private void _dumpLoaded(StringBuilder sb, String separatorLine, PluginsInformation pluginInfo)
    {
        Map<String, Feature> loadedFeatures = pluginInfo.getFeatures();
        int nbLoadedFeatures = loadedFeatures.size();
        long nbLoadedComponents = loadedFeatures.values()
                .parallelStream()
                .map(Feature::getComponents)
                .map(Map::values)
                .flatMap(Collection::parallelStream)
                .count();
        long nbLoadedExtensions = loadedFeatures.values()
                .parallelStream()
                .map(Feature::getExtensions)
                .map(Map::values)
                .flatMap(Collection::parallelStream)
                .map(Map::values)
                .flatMap(Collection::parallelStream)
                .count();
        
        int nbInactiveFeatures = pluginInfo.getInactiveFeatures().size();
        Set<String> inactiveFeatureIds = pluginInfo.getInactiveFeatures().keySet();
        Supplier<Stream<Feature>> inactiveFeatures = () -> _featureActivator._allPlugins
                .values()
                .stream()
                .map(Plugin::getFeatures)
                .map(Map::values)
                .flatMap(Collection::parallelStream)
                .filter(Predicates.compose(inactiveFeatureIds::contains, Feature::getFeatureId));
        long nbInactiveComponents = inactiveFeatures.get()
                .map(Feature::getComponentsIds)
                .mapToInt(Map::size)
                .sum();
        long nbInactiveExtensions = inactiveFeatures.get()
                .map(Feature::getExtensionsIds)
                .map(Map::values)
                .flatMap(Collection::parallelStream)
                .flatMap(Collection::parallelStream)
                .count();
        
        _addCell(sb, "\\");
        _addCell(sb, "loaded");
        _addCell(sb, "inactive");
        _addCell(sb, "total");
        _addCell(sb, "load factor");
        sb.append("|\n");
        sb.append(separatorLine);
        _addCell(sb, "features");
        _addCell(sb, String.valueOf(nbLoadedFeatures));
        _addCell(sb, String.valueOf(nbInactiveFeatures));
        int nbTotalFeatures = nbLoadedFeatures + nbInactiveFeatures;
        _addCell(sb, String.valueOf(nbTotalFeatures));
        _addCell(sb, new BigDecimal((double) nbLoadedFeatures / (double) nbTotalFeatures).setScale(3, RoundingMode.HALF_UP));
        sb.append("|\n");
        sb.append(separatorLine);
        _addCell(sb, "components");
        _addCell(sb, String.valueOf(nbLoadedComponents));
        _addCell(sb, String.valueOf(nbInactiveComponents));
        long nbTotalComponents = nbLoadedComponents + nbInactiveComponents;
        _addCell(sb, String.valueOf(nbTotalComponents));
        _addCell(sb, new BigDecimal((double) nbLoadedComponents / (double) nbTotalComponents).setScale(3, RoundingMode.HALF_UP));
        sb.append("|\n");
        sb.append(separatorLine);
        _addCell(sb, "extensions");
        _addCell(sb, String.valueOf(nbLoadedExtensions));
        _addCell(sb, String.valueOf(nbInactiveExtensions));
        long nbTotalExtensions = nbLoadedExtensions + nbInactiveExtensions;
        _addCell(sb, String.valueOf(nbTotalExtensions));
        _addCell(sb, new BigDecimal((double) nbLoadedExtensions / (double) nbTotalExtensions).setScale(3, RoundingMode.HALF_UP));
        sb.append("|\n");
    }
    
    private void _addCell(StringBuilder sb, Object content)
    {
        final int limit = 12;
        String limitedContent = String.valueOf(content);
        if (limitedContent.length() >= limit)
        {
            limitedContent = limitedContent.substring(0, limit);
        }
        else
        {
            int missing = limit - limitedContent.length();
            int half = missing / 2;
            char[] characters = new char[limit];
            Arrays.fill(characters, ' ');
            int startPosTarget = missing - half;
            System.arraycopy(limitedContent.toCharArray(), 0, characters, startPosTarget, limitedContent.length());
            limitedContent = new String(characters);
        }
        sb.append("|");
        sb.append(limitedContent);
    }
    
    private void _shortDumpPlugin(
            StringBuilder sb, 
            Plugin plugin, 
            Map<String, Collection<String>> dependencies, 
            Collection<IncludedFeature> includedFeatures, 
            Map<String, InactivityCause> inactiveFeatures)
    {
        String pluginName = plugin.getName();
        sb.append("Plugin ").append(pluginName);
        
        sb.append("\n\n");
        
        Map<String, Feature> activeFeatures = plugin.getFeatures();
        List<String> featureIds = activeFeatures.keySet()
                .stream()
                .sorted()
                .collect(Collectors.toList());
        for (String featureId : featureIds)
        {
            Feature feature = activeFeatures.get(featureId);
            _shortDumpFeature(sb, feature, dependencies, includedFeatures, inactiveFeatures);
        }
        
        sb.append('\n');
    }
    
    private void _shortDumpFeature(
            StringBuilder sb, 
            Feature feature, 
            Map<String, Collection<String>> dependencies, 
            Collection<IncludedFeature> includedFeatures, 
            Map<String, InactivityCause> inactiveFeatures)
    {
        String featureId = feature.getFeatureId();
        
        if (!inactiveFeatures.containsKey(featureId))
        {
            sb.append("  Feature ").append(featureId);
            List<String> dependenciesOfFeature = new ArrayList<>();
            if (dependencies.containsKey(featureId))
            {
                dependenciesOfFeature.addAll(dependencies.get(featureId));
            }
            Optional<String> causes = _findIncludedFeatureCauses(featureId, includedFeatures);
            if (causes.isPresent())
            {
                dependenciesOfFeature.add(0, "INCLUDED_FEATURE BY " + causes.get());
            }
            if (ArrayUtils.contains(IncludePolicyFeatureActivator.__KERNEL_DEPENDENCIES_NEEDED, featureId))
            {
                dependenciesOfFeature.add(0, "KERNEL");
            }
            
            if (!dependenciesOfFeature.isEmpty())
            {
                sb.append("\n    --> brought by ").append(_dependenciesToString(dependenciesOfFeature));
            }
            
            sb.append("\n\n");
        }
    }
    
    private String _dependenciesToString(List<String> dependenciesOfFeature)
    {
        switch (dependenciesOfFeature.size())
        {
            case 0:
            case 1:
                // display it on one line
                return dependenciesOfFeature.toString();
            default:
                // one item of the list perl ine to be readable (+ handle indentation)
                final String indent1 = "                   ";
                final String indent2 = indent1 + "    ";
                return "[\n" + indent2 + String.join(",\n" + indent2, dependenciesOfFeature) + "\n" + indent1 + "]";
        }
    }
    
    private Optional<String> _findIncludedFeatureCauses(String featureId, Collection<IncludedFeature> includedFeatures)
    {
        List<String> causes = includedFeatures.stream()
                .filter(Predicates.compose(Predicates.equalTo(featureId), IncludedFeature::featureId))
                .map(IncludedFeature::cause)
                .collect(Collectors.toList());
        switch (causes.size())
        {
            case 0:
                return Optional.empty();
            case 1:
                return Optional.of(causes.get(0));
            default:
                return Optional.of("{" + String.join(" & ", causes) + "}");
        }
    }
}
