/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.runtime.plugin.component;

import java.util.Comparator;
import java.util.Set;
import java.util.TreeSet;
import java.util.function.Function;

/**
 * Avalon based implementation of an {@link Prioritizable} ExtensionPoint.<br>
 * @param <T> the type of the managed extensions, must implement {@link Prioritizable}
 */
public abstract class AbstractThreadSafeComponentPrioritizableExtensionPoint<T extends Prioritizable> extends AbstractThreadSafeComponentExtensionPoint<T>
{
    private Set<String> _sortedIds;
    
    @Override
    public void initializeExtensions() throws Exception
    {
        super.initializeExtensions();
        _sortedIds = new TreeSet<>(getComparator());
        _sortedIds.addAll(super.getExtensionsIds());
    }
    
    @Override
    public Set<String> getExtensionsIds()
    {
        return _sortedIds;
    }
    
    /**
     * Define the comparator used to order components.
     * Default is based on priority (could be ascending or descending depending of {@link #sortPriorityAscending()} value, then on the extension ID.
     * @return the comparator used to order components
     */
    protected Comparator<String> getComparator()
    {
        return Comparator
            // Comparing on priority (ascending or descending)
            .<String, Integer>comparing(
                id -> getExtension(id).getPriority(),
                sortPriorityAscending()
                    ? Comparator.naturalOrder()
                    : Comparator.reverseOrder()
            )
            // Then comparing on extension id
            .thenComparing(Function.identity());
    }
    
    /**
     * Define if the priority is ascending sorted or reversed, default implementation is <code>true</code>.
     * @return <code>true</code> if the priority should be ascending.
     */
    protected boolean sortPriorityAscending()
    {
        return true;
    }
}
