/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.request;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

import org.apache.cocoon.servlet.multipart.MultipartException;
import org.apache.cocoon.servlet.multipart.MultipartHttpServletRequest;
import org.apache.cocoon.servlet.multipart.Part;
import org.apache.cocoon.servlet.multipart.PartOnDisk;
import org.apache.cocoon.servlet.multipart.RejectedPart;

import org.ametys.runtime.config.ConfigManager;
import org.ametys.runtime.servlet.AnalyseFileForVirusHelper;

/**
 * Extension of the cocoon request factory to allow antivirus scanning of uploaded file
 */
public class RequestFactory extends org.apache.cocoon.servlet.multipart.RequestFactory
{
    // check for antivirus activation at instantiation
    private boolean _antivirusEnabled;

    /**
     * Default constructor see {@link org.apache.cocoon.servlet.multipart.RequestFactory} for documentation
     * @param saveUploadedFilesToDisk to save files to disk
     * @param uploadDirectory where to save files on disk
     * @param allowOverwrite to allow overwrite
     * @param silentlyRename to rename silently
     * @param maxUploadSize the maximum size of an uploaded file
     * @param defaultCharEncoding the default encoding
     */
    public RequestFactory(boolean saveUploadedFilesToDisk,
            File uploadDirectory,
            boolean allowOverwrite,
            boolean silentlyRename,
            int maxUploadSize,
            String defaultCharEncoding)
    {
        super(saveUploadedFilesToDisk, uploadDirectory, allowOverwrite, silentlyRename, maxUploadSize, defaultCharEncoding);
        
        // Check antivirus activation
        if (ConfigManager.getInstance().isComplete())
        {
            _antivirusEnabled = AnalyseFileForVirusHelper.isAntivirusEnabled();
        }
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public HttpServletRequest getServletRequest(HttpServletRequest request) throws IOException, MultipartException
    {
        HttpServletRequest servletRequest = super.getServletRequest(request);
        // Check the activation of the antivirus to ensure minimal impact if antivirus is disabled
        if (servletRequest instanceof MultipartHttpServletRequest multipart && _antivirusEnabled)
        {
            Map<String, Object> rejectedParam = new HashMap<>();
            Enumeration<String> parameterNames = multipart.getParameterNames();
            while (parameterNames.hasMoreElements())
            {
                String name = parameterNames.nextElement();
                Object value = multipart.get(name);
                if (value instanceof PartOnDisk diskPart)
                {
                    if (_analyseAndRejectFile(diskPart, name))
                    {
                        // Record the name of param to remove
                        rejectedParam.put(name, new RejectedPart(diskPart.getHeaders(), 0, 0, 0));
                    }
                }
                else if (value instanceof List valueList && !valueList.isEmpty() && valueList.get(0) instanceof PartOnDisk)
                {
                    List<Part> replaceFiles = new ArrayList<>();
                    for (PartOnDisk diskPart : (List<PartOnDisk>) valueList)
                    {
                        // Analyse the file and reject and delete it if it has a virus
                        if (_analyseAndRejectFile(diskPart, name))
                        {
                            replaceFiles.add(new RejectedPart(diskPart.getHeaders(), 0, 0, 0));
                        }
                        else
                        {
                            replaceFiles.add(diskPart);
                        }
                    }
                    
                    // Record the name of param to remove
                    rejectedParam.put(name, replaceFiles);
                }
            }
            
            if (!rejectedParam.isEmpty())
            {
                Hashtable<String, Object> params = new Hashtable<>();
                // some part were rejected. 
                // We have to copy the request parameters
                parameterNames = multipart.getParameterNames();
                while (parameterNames.hasMoreElements())
                {
                    String name = parameterNames.nextElement();
                    if (rejectedParam.containsKey(name))
                    {
                        // We represent a contaminated part with a rejected part with max size = 0byte
                        params.put(name, rejectedParam.get(name));
                    }
                    else
                    {
                        params.put(name, multipart.get(name));
                    }
                }
                
                // and wrap the request with the new parameters including the rejected parts
                servletRequest = new MultipartHttpServletRequest(servletRequest, params);
            }
        }
        return servletRequest;
    }
    
    /**
     * Analyse the file for viruses, and delete it if a virus has been detected
     * @param diskPart The part to analyse
     * @param name The name of the param containing the file
     * @return a {@link Boolean}, true if a virus has been detected, false otherwise
     */
    private boolean _analyseAndRejectFile(PartOnDisk diskPart, String name)
    {
        File file = diskPart.getFile();
        // Check file here
        if (!AnalyseFileForVirusHelper.analysefile(file.getAbsolutePath()))
        {
            // Delete dangerous file instantly
            file.delete();
            
            return true;
        }
        
        return false;
    }
}
