/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.runtime.servlet;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.stream.Collectors;

import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.runtime.config.Config;

/**
 * Helper that analyzes a file for viruses
 */
public final class AnalyseFileForVirusHelper
{
    /** The result of the antivirus when no viruses is found */
    public static final Integer ANTIVIRUS_RESULT_OK = 0;
    
    private static final Logger __LOGGER = LoggerFactory.getLogger(AnalyseFileForVirusHelper.class);
    
    private AnalyseFileForVirusHelper()
    {
        // utility class
    }
    
    /**k
     * Checks if the antivirus is enabled
     * @return <code>true</code> if the antivirus is enabled, <code>false</code> otherwise 
     */
    public static boolean isAntivirusEnabled()
    {
        return Config.getInstance().getValue("runtime.upload.antivirus.activated", false, false);
    }
    
    /**
     * Antivirus analysis. Based on clamscan results.
     * 
     * @param absolutePath the absolute path of the file to analyse
     * @return true if the file is correct, false if a malware was discovered in
     *         the file
     */
    public static boolean analysefile(String absolutePath)
    {
        if (!isAntivirusEnabled())
        {
            return true;
        }
        
        try
        {
            String command = Config.getInstance().getValue("runtime.upload.antivirus.command");
            // Split before replacing with file path to avoid issue with space in file name
            String[] cmdArray = StringUtils.split(command);
            cmdArray = Arrays.stream(cmdArray).map(str -> StringUtils.replace(str, "%f", absolutePath)).toArray(String[]::new);
            if (__LOGGER.isDebugEnabled())
            {
                __LOGGER.debug("Executing antivirus analysis : {}", StringUtils.join(cmdArray, " "));
            }
            
            // Execute command
            // use the String[] constructor to unsure correct handling of filename with special character
            Process child = new ProcessBuilder(cmdArray)
                .redirectErrorStream(true)
                .start();
            child.waitFor();
            
            // Get the input stream and read from it
            if (__LOGGER.isDebugEnabled() || child.exitValue() == 2)
            {
                StringBuilder builder = new StringBuilder("Result of the command : (").append(child.exitValue()).append(")\n");
                try (InputStream in = child.getInputStream())
                {
                    String cmdOutput = new BufferedReader(new InputStreamReader(in, StandardCharsets.UTF_8))
                        .lines()
                        .collect(Collectors.joining("\n"));
                    builder.append(cmdOutput);
                }
                if (child.exitValue() == 2)
                {
                    __LOGGER.error(builder.toString());
                }
                else
                {
                    __LOGGER.debug(builder.toString());
                }
            }
            return ANTIVIRUS_RESULT_OK.equals(child.exitValue());
        }
        catch (IOException | InterruptedException e)
        {
            __LOGGER.error("Unable to get output from the command", e);
            return false;
        }
    }
}
