/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.runtime.workspace;

import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.matching.WildcardURIMatcher;
import org.apache.cocoon.sitemap.PatternException;

import org.ametys.runtime.servlet.RuntimeConfig;

/**
 * Workspace aware Cocoon matcher.<br>
 * Retrieve workspaces URI from names through the WorkspaceManager.
 */
public class WorkspaceMatcher extends WildcardURIMatcher
{
    /** The request attribute name where the current workspace is saved */
    public static final String WORKSPACE_NAME = "workspaceName";
    /** The request attribute name where the current workspace uri is saved */
    public static final String WORKSPACE_URI = "workspaceURI";
    /** The request attribute name where the current url inside the workspace is saved */
    public static final String IN_WORKSPACE_URL = "inWorkspaceURL";
    /** The request attribute name where the current workspace theme name is saved */
    public static final String WORKSPACE_THEME = "workspaceTheme";
    /** The request attribute name where the current workspace theme url is saved */
    public static final String WORKSPACE_THEME_URL = "workspaceThemeURL";
    
    /**
     * In the wildcard expression where to find the first relevant information
     * @return An integer &gt; 0
     */
    protected int getWildcardStartIndex()
    {
        return 1;
    }
    
    /**
     * Get the selected workspace
     * @param workspaceName The workspace name to get
     * @return The workspace if it exists and is supported
     */
    protected Workspace getWorkspace(String workspaceName)
    {
        WorkspaceManager wm = WorkspaceManager.getInstance();
        return wm.getWorkspaces().get(workspaceName);
    }
    
    @Override
    public Map match(String pattern, Map objectModel, Parameters parameters) throws PatternException
    {
        // Get the url wildcard part
        Map<String, String> result = super.match(pattern, objectModel, parameters);
        if (result == null)
        {
            return null;
        }
        
        int wildcardStartIndex = getWildcardStartIndex();
        
        // Get the workspace name
        String workspaceName;
        boolean defaultWorkspace = "true".equals(parameters.getParameter("default", null)); 
        if (defaultWorkspace)
        {
            workspaceName = RuntimeConfig.getInstance().getDefaultWorkspace();
        }
        else
        {
            workspaceName = result.get(Integer.toString(wildcardStartIndex)); 
        }
        
        // Check and get workspace name
        Workspace workspace = getWorkspace(workspaceName);
        if (workspace == null)
        {
            return null;
        }
        
        // Set response
        Request request = ObjectModelHelper.getRequest(objectModel);
        request.setAttribute(WORKSPACE_NAME, workspaceName);
        request.setAttribute(WORKSPACE_URI, defaultWorkspace ? "" : "/_" + workspaceName);
        request.setAttribute(IN_WORKSPACE_URL, result.get(Integer.toString(wildcardStartIndex + (defaultWorkspace ? 0 : 1))));
        request.setAttribute(WORKSPACE_THEME, workspace.getThemeName());
        request.setAttribute(WORKSPACE_THEME_URL, workspace.getThemeURL());
        
        result.put("workspaceName", workspaceName);
        
        return result;
    }
}
