/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugins.admin.configuration;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;

import org.ametys.core.cocoon.ActionResultGenerator;
import org.ametys.core.util.JSONUtils;
import org.ametys.runtime.config.ConfigManager;
import org.ametys.runtime.config.ConfigParameterCheckerDescriptor;
import org.ametys.runtime.model.DefinitionAndValue;
import org.ametys.runtime.model.checker.ItemChecker;

/**
 * Fetch the test values and dispatch the requests to the appropriate parameter checkers.
 */
public class CheckConfigAction extends ServiceableAction
{
    /** Helper component gathering utility methods for the management of JSON entities */
    private JSONUtils _jsonUtils;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        _jsonUtils = (JSONUtils) serviceManager.lookup(JSONUtils.ROLE);
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Map<String, String> result = new HashMap<> ();
        
        // Handle the 'configuration not initialized' case
        Map<String, DefinitionAndValue> oldDefinitionAndValues = ConfigManager.getInstance().getOldDefinitionAndValues();
        
        // Fetch the request parameters
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        // Prepare the values for the test
        String fieldCheckersInfoJSON = request.getParameter("fieldCheckersInfo");
        Map<String, Object> fieldCheckersInfo = _jsonUtils.convertJsonToMap(fieldCheckersInfoJSON);
        Map<String, Object> valuesByParameterChecker = _getValuesByParamCheckerId(fieldCheckersInfo, oldDefinitionAndValues);

        // Dispatch the requests
        for (String fieldCheckerId : valuesByParameterChecker.keySet())
        {
            ItemChecker fieldChecker = (ItemChecker) ((Map<String, Object>) valuesByParameterChecker.get(fieldCheckerId)).get("checker");
            List<String> values = (List<String>) ((Map<String, Object>) valuesByParameterChecker.get(fieldCheckerId)).get("values");
            try 
            {
                fieldChecker.check(values); 
            }
            catch (Throwable t)
            {
                getLogger().error("The test '" + fieldCheckerId + "' failed : \n" + t.getMessage(), t);
                String msg = t.getMessage() != null ? t.getMessage() : "Unknown error";
                result.put(fieldCheckerId, msg);
            }
        }
        
        request.setAttribute(ActionResultGenerator.MAP_REQUEST_ATTR, result);
        return result;
    }

    /**
     * Compute the proper values and {@link ItemChecker} implementations to use for the test and order them by parameter checker id
     * @param paramCheckersInfo the information concerning the parameter checkers
     * @param oldDefinitionAndValues the map of old definition and values of the configuration
     * @return the map of values lists ordered by parameter checker
     */
    @SuppressWarnings("unchecked")
    private Map<String, Object> _getValuesByParamCheckerId(Map<String, Object> paramCheckersInfo, Map<String, DefinitionAndValue> oldDefinitionAndValues)
    {
        Map<String, Object> result = new HashMap<> ();
        
        ConfigManager configManager = ConfigManager.getInstance();
        for (String paramCheckerId : paramCheckersInfo.keySet())
        {
            Map<String, Object> valuesByParamCheckerId = new HashMap<> (); 
            
            // Check the ids of the parameter checkers and build the parameter checkers' list
            ConfigParameterCheckerDescriptor parameterCheckerDescriptor = configManager.getParameterChecker(paramCheckerId);
            if (parameterCheckerDescriptor == null)
            {
                throw new IllegalArgumentException("The parameter checker '" + paramCheckerId + "' was not found.");
            }
            
            ItemChecker parameterChecker  = parameterCheckerDescriptor.getParameterChecker();
            
            valuesByParamCheckerId.put("checker", parameterChecker);
            
            List<String> paramNames = (List<String>) ((Map<String, Object>) paramCheckersInfo.get(paramCheckerId)).get("testParamsNames");
            List<Object> paramRawValues = (List<Object>) ((Map<String, Object>) paramCheckersInfo.get(paramCheckerId)).get("rawTestValues");
            
            List<String> values = new ArrayList<> ();
            
            // Compute the proper values for the test
            for (int i = 0; i < paramNames.size(); i++)
            {
                String parameterName = paramNames.get(i);
                Object value = paramRawValues.get(i);
                
                // Resolve the value
                String resolvedValue = configManager.resolveValueForParameterChecker(parameterName, value, oldDefinitionAndValues);
                values.add(resolvedValue);
            }
            
            valuesByParamCheckerId.put("values", values);
            
            result.put(paramCheckerId, valuesByParamCheckerId);
        }
        
        return result;
    }
}
