/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugins.admin.configuration;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.thread.ThreadSafe;
import org.apache.cocoon.acting.AbstractAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;

import org.ametys.core.cocoon.ActionResultGenerator;
import org.ametys.runtime.config.ConfigManager;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.servlet.RuntimeServlet;
import org.ametys.runtime.util.AmetysHomeHelper;


/**
 * This action is in charge to get and save the config values entered by the user.<br>
 * The backup is delegated to <code>Config</code>
 */
public class SaveConfigAction extends AbstractAction implements ThreadSafe
{   
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {

        if (getLogger().isDebugEnabled())
        {
            getLogger().debug("Starting SaveConfigAction");
        }

        Request request = ObjectModelHelper.getRequest(objectModel);
        Map<String, Object> result = new HashMap<> ();
        
        try
        {
            Map<String, Object> values = new HashMap<>();
            
            ConfigManager configManager = ConfigManager.getInstance();
            Collection<ElementDefinition> definitions = configManager.getConfigurationParameters();
            Map<String, List<I18nizableText>> errorFields = new HashMap<>();
            
            for (ElementDefinition definition : definitions)
            {
                final String parameterId = definition.getName();
                try
                {
                    Object value = _getTypedValueFromRequest(request, definition);
                    values.put(parameterId, value);
                }
                catch (Exception e)
                {
                    if (getLogger().isWarnEnabled())
                    {
                        getLogger().warn("The configuration parameter '" + parameterId + "' is not valid.", e);
                    }
                    I18nizableText errorLabel = new I18nizableText("plugin.core", "PLUGINS_CORE_SAVE_PARAMETER_VALUE_INVALID", Collections.singletonList(e.getMessage()));
                    errorFields.put(parameterId, Collections.singletonList(errorLabel));
                }
            }
            
            if (errorFields.isEmpty())
            {
                // No type error, let's try to save configuration
                String configFileName = new File(AmetysHomeHelper.getAmetysHomeConfig(), RuntimeServlet.CONFIG_FILE_NAME).getCanonicalPath();
                errorFields.putAll(configManager.save(values, configFileName));
            }
            
            if (!errorFields.isEmpty())
            {
                for (String paramId : errorFields.keySet())
                {
                    List<I18nizableText> errors = errorFields.get(paramId);
                    result.put(paramId, errors);
                }
                
                request.setAttribute(ActionResultGenerator.MAP_REQUEST_ATTR, result);
                return EMPTY_MAP;
            }
        }
        catch (Exception e)
        {
            getLogger().error("An error occured while saving config modifications", e);
            
            result.put("error", e.getMessage());
            request.setAttribute(ActionResultGenerator.MAP_REQUEST_ATTR, result);
            return EMPTY_MAP;
        }

        // Set the request attribute for Cocoon reloading
        if (getLogger().isDebugEnabled())
        {
            getLogger().debug("Positionning org.ametys.runtime.reload=true for Cocoon reloading");
        }
        request.setAttribute("org.ametys.runtime.reload", true);

        return EMPTY_MAP;
    }

    private Object _getTypedValueFromRequest(Request request, ElementDefinition definition)
    {
        final String parameterId = definition.getName();
        final ElementType type = definition.getType();
        
        Object value = null;
        if (definition.isMultiple())
        {
            String[] valuesAsString = request.getParameterValues(parameterId);
            if (valuesAsString != null)
            {
                List<Object> valuesAsList = new ArrayList<>();
                for (String valueAsString : valuesAsString)
                {
                    valuesAsList.add(type.castValue(valueAsString));
                }
                value = valuesAsList.toArray(new Object[valuesAsList.size()]);
            }
        }
        else
        {
            final String valueAsString = request.getParameter(parameterId);
            value = definition.getType().castValue(valueAsString);
        }

        return value;
    }
}
