/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugins.admin.jvmstatus;

import java.time.Instant;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpSessionEvent;
import javax.servlet.http.HttpSessionListener;

import org.ametys.core.authentication.AuthenticateAction;
import org.ametys.core.trace.ForensicLogger;
import org.ametys.core.user.UserIdentity;

/**
 * Simple HttpSessionListener resgistering active sessions. 
 */
public class ActiveSessionListener implements HttpSessionListener
{
    private static Set<HttpSession> _activeSessions = new HashSet<>();
    
    public void sessionCreated(HttpSessionEvent se)
    {
        HttpSession session = se.getSession();
        _activeSessions.add(session);
        _logSessionEvent("session.created", session);
    }

    public void sessionDestroyed(HttpSessionEvent se)
    {
        HttpSession session = se.getSession();
        _activeSessions.remove(session);
        _logSessionEvent("session.destroyed", session);
    }
    
    private void _logSessionEvent(String eventId, HttpSession session)
    {
        ForensicLogger.info(eventId, _transformSessionToJson(session), null);
    }
    
    /**
     * Get the active HTTP sessions with ID, created time, last accessed time and user if available.
     * @return all the active HTTP sessions
     */
    public static List<Map<String, Object>> getActiveSessions()
    {
        return _activeSessions.stream()
            .map(ActiveSessionListener::_transformSessionToJson)
            .toList();
    }
    
    private static Map<String, Object> _transformSessionToJson(HttpSession session)
    {
        Map<String, Object> map = new HashMap<>();
        map.put("id", session.getId());
        map.put("creation", _convertMillisToReadableDate(session.getCreationTime()));
        map.put("lastAccess", _convertMillisToReadableDate(session.getLastAccessedTime()));
        UserIdentity user = (UserIdentity) session.getAttribute(AuthenticateAction.SESSION_USERIDENTITY);
        if (user != null)
        {
            map.put("user", UserIdentity.userIdentityToString(user));
        }
        return map;
    }
    
    private static String _convertMillisToReadableDate(long time)
    {
        DateTimeFormatter formatter = DateTimeFormatter.ofPattern("dd/MM/yyyy HH:mm:ss").withZone(ZoneId.systemDefault());
        Instant instant = Instant.ofEpochMilli(time);
        return formatter.format(instant);
    }
}
