/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugins.admin.migration;

import java.time.Instant;
import java.util.Map.Entry;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.core.migration.MigrationEngine;
import org.ametys.core.migration.MigrationEngine.ActionData;
import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationEngine.VersionList;
import org.ametys.core.migration.MigrationEngine.Versions;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.MigrationExtensionPoint;
import org.ametys.core.migration.action.ActionConfiguration;
import org.ametys.core.migration.action.ActionExtensionPoint;
import org.ametys.core.migration.version.Version;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;

/**
 * Client side element with callables for migrations actions
 */
public class MigrationsActionsClientSideElement extends StaticClientSideElement
{
    private MigrationEngine _migrationEngine;
    private MigrationExtensionPoint _migrationExtensionPoint;
    private MigrationExtensionPoint _migrationDataExtensionPoint;
    private ActionExtensionPoint _upgradeEP;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        
        _migrationEngine = (MigrationEngine) manager.lookup(MigrationEngine.ROLE);
        _migrationExtensionPoint = (MigrationExtensionPoint) manager.lookup(MigrationExtensionPoint.ROLE);
        _migrationDataExtensionPoint = (MigrationExtensionPoint) manager.lookup(MigrationExtensionPoint.ROLE + "/internal");
        _upgradeEP = (ActionExtensionPoint) manager.lookup(ActionExtensionPoint.ROLE_UPGRADE);
    }
    
    /**
     * Marks a version as done
     * @param componentId the main component id
     * @param internal true if the component is from internel migration
     * @param versionListId the id of the {@link VersionList}
     * @param versionNumber the number of the version to be deleted
     * @param comment the comment associated to this operation
     * @throws MigrationException if an error occurs while storing version
     */
    @Callable (rights = "Runtime_Rights_Admin_Access", context = "/admin")
    public void acknowledge(String componentId, boolean internal, String versionListId, String versionNumber, String comment) throws MigrationException
    {
        MigrationComponent component = internal ? _migrationDataExtensionPoint.getExtension(componentId) : _migrationExtensionPoint.getExtension(componentId);
        
        VersionList versionList = _getVersionList(component, versionListId);
        
        Version version = new Version(component, versionList.componentId(), component.versionStorage(), versionList.storageConfiguration(), versionNumber, Instant.now(), comment);
        component.versionStorage().storeVersion(version);
    }
    
    /**
     * Deletes a version from storage
     * @param componentId the main component id
     * @param internal true if the component is from internel migration
     * @param versionListId the id of the {@link VersionList}
     * @param versionNumber the number of the version to be deleted
     * @throws MigrationException if an error occurs while deleting version
     */
    @Callable (rights = "Runtime_Rights_Admin_Access", context = "/admin")
    public void deleteVersion(String componentId, boolean internal, String versionListId, String versionNumber) throws MigrationException
    {
        MigrationComponent component = internal ? _migrationDataExtensionPoint.getExtension(componentId) : _migrationExtensionPoint.getExtension(componentId);
        
        VersionList versionList = _getVersionList(component, versionListId);
        
        component.versionStorage().removeVersion(versionList.componentId(), versionNumber, versionList.storageConfiguration());
    }
    
    /**
     * Deletes all versions of a component from storage
     * @param componentId the main component id
     * @param internal true if the component is from internel migration
     * @param versionListId the id of the {@link VersionList}
     * @throws MigrationException if an error occurs while deleting versions
     */
    @Callable (rights = "Runtime_Rights_Admin_Access", context = "/admin")
    public void deleteAllVersions(String componentId, boolean internal, String versionListId) throws MigrationException
    {
        MigrationComponent component = internal ? _migrationDataExtensionPoint.getExtension(componentId) : _migrationExtensionPoint.getExtension(componentId);
        
        VersionList versionList = _getVersionList(component, versionListId);
        
        component.versionStorage().removeAllVersions(versionList.componentId(), versionList.storageConfiguration());
    }
    
    /**
     * (Re-)plays a migration action, optionnaly storing a new version
     * @param componentId the main component id
     * @param internal true if the component is from internel migration
     * @param versionListId the id of the {@link VersionList}
     * @param upgradeNumber the number of the action to be executed
     * @param store if the version should be stored after execution
     * @param comment the comment associated to this operation
     * @return true if the action needs a restart after its execution
     * @throws MigrationException  if an error occurs while executing action
     */
    @Callable (rights = "Runtime_Rights_Admin_Access", context = "/admin")
    public boolean doAction(String componentId, boolean internal, String versionListId, String upgradeNumber, boolean store, String comment) throws MigrationException
    {
        MigrationComponent component = internal ? _migrationDataExtensionPoint.getExtension(componentId) : _migrationExtensionPoint.getExtension(componentId);
        
        ActionConfiguration actionConfiguration = component.upgrades().stream()
                                                                      .filter(ac -> ac.getVersionNumber().equals(upgradeNumber))
                                                                      .findFirst()
                                                                      .orElseThrow(() -> new MigrationException("no such upgrade: " + upgradeNumber));
        
        VersionList versionList = _getVersionList(component, versionListId);
        
        Version version = new Version(component, versionList.componentId(), component.versionStorage(), versionList.storageConfiguration(), null, null, null);
        
        if (versionList.additionalValues() != null)
        {
            for (Entry<String, Object> entry : versionList.additionalValues().entrySet())
            {
                version.addAdditionalValue(entry.getKey(), entry.getValue());
            }
        }
        
        boolean requiresRestart = _migrationEngine.doAction(new ActionData(version, null, actionConfiguration, versionList.id()), _upgradeEP);
        
        if (store)
        {
            // first remove the already stored version
            component.versionStorage().removeVersion(versionList.componentId(), upgradeNumber, versionList.storageConfiguration());
            
            version.setComment(comment);
            version.setExecutionInstant(Instant.now());
            version.setVersionNumber(upgradeNumber);
            
            component.versionStorage().storeVersion(version);
        }
        
        return requiresRestart;
    }
    
    private VersionList _getVersionList(MigrationComponent component, String versionListId) throws MigrationException
    {
        Versions versions = _migrationEngine.getVersions(component);
        VersionList versionList = _migrationEngine.getVersionList(versions, versionListId);
        
        if (versionList == null)
        {
            throw new MigrationException("There's no version list for id: " + versionListId + " in component: " + component);
        }
        
        return versionList;
    }
}
