/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.runtime.plugins.admin.migration;

import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.stream.Collectors;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.migration.MigrationEngine;
import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationEngine.VersionList;
import org.ametys.core.migration.MigrationEngine.Versions;
import org.ametys.core.migration.MigrationEngine.VersionsContainer;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.MigrationExtensionPoint;
import org.ametys.core.migration.action.ActionConfiguration;
import org.ametys.core.migration.version.Version;
import org.ametys.core.ui.Callable;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.I18nUtils;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Component for retrieving info about all past and available automatic upgrades.
 */
public class MigrationsStatus implements Serviceable, Component
{
    private MigrationEngine _migrationEngine;
    private MigrationExtensionPoint _migrationExtensionPoint;
    private MigrationExtensionPoint _migrationInternalExtensionPoint;
    private I18nUtils _i18nUtils;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _migrationEngine = (MigrationEngine) manager.lookup(MigrationEngine.ROLE);
        _migrationExtensionPoint = (MigrationExtensionPoint) manager.lookup(MigrationExtensionPoint.ROLE);
        _migrationInternalExtensionPoint = (MigrationExtensionPoint) manager.lookup(MigrationExtensionPoint.ROLE + "/internal");
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
    }

    /**
     * Retrieves the status of automatic migrations
     * @param clientParameters The client parameters such as the node to refresh
     * @return a JSON view of all existing automatic migrations.
     * @throws ConfigurationException if an error occurred reading the configuration
     * @throws MigrationException if an error occurred retrieving the versions
     */
    @Callable (rights = "Runtime_Rights_Admin_Access", context = "/admin")
    public Map<String, Object> getMigrationsStatus(Map<String, Object> clientParameters) throws ConfigurationException, MigrationException
    {
        Map<String, List<Map<String, Object>>> migrationsByPlugins = new TreeMap<>();
        
        _migrationsByPlugins(true, migrationsByPlugins, _migrationInternalExtensionPoint);
        _migrationsByPlugins(false, migrationsByPlugins, _migrationExtensionPoint);
        
        List<Map<String, Object>> allNodes = migrationsByPlugins.entrySet().stream()
                .map(entry -> _plugin2json(entry.getKey(), entry.getValue()))
                .toList();
        
        List<Map<String, Object>> parentInfos = new ArrayList<>();
        return Map.of(
            "children", _filter(allNodes, StringUtils.removeStart((String) clientParameters.get("path"), "/root"), parentInfos),
            "parentInfos", parentInfos
        );
    }

    @SuppressWarnings("unchecked")
    private List<Map<String, Object>> _filter(List<Map<String, Object>> allNodes, String path, List<Map<String, Object>> parentInfos)
    {
        if (StringUtils.isBlank(path))
        {
            return allNodes;
        }
        else
        {
            int i = path.indexOf('/', 1);
            i = i == -1 ? path.length() : i; 
            String thisNode = path.substring(1, i);
            String subPath = path.substring(i);
            
            for (Map<String, Object> map : allNodes)
            {
                if (thisNode.equals(map.get("component")))
                {
                    parentInfos.add(Map.of(
                        "failed", map.get("failed"),
                        "warning", map.get("warning")
                    ));
                    return _filter((List<Map<String, Object>>) map.get("children"), subPath, parentInfos);
                }
            }
            
            throw new IllegalArgumentException("Cannot find component '" + thisNode + "'");
        }
    }

    private void _migrationsByPlugins(boolean isInternal, Map<String, List<Map<String, Object>>> migrationsByPlugins, MigrationExtensionPoint migrationExtensionPoint)
    {
        for (String extensionId : migrationExtensionPoint.getExtensionsIds())
        {
            MigrationComponent component = migrationExtensionPoint.getExtension(extensionId);
            List<Map<String, Object>> migrations = migrationsByPlugins.computeIfAbsent(component.pluginName(), k -> new ArrayList<>());
            
            try
            {
                // Processing stored versions
                Versions versions = _migrationEngine.getVersions(component);
                
                List<Map<String, Object>> children = _versions2json(isInternal, extensionId, versions, component.upgrades());
                
                migrations.add(_component2json(isInternal, extensionId, extensionId, children, component.versionHandlerType(), component.versionStorage().getId(), true, versions instanceof VersionList vl ? vl.id() : null));
            }
            catch (MigrationException e)
            {
                migrations.add(_componentFailure2json(isInternal, extensionId, component.versionHandlerType(), component.versionStorage().getId(), e));
            }
        }
    }
    
    private List<Map<String, Object>> _versions2json(boolean isInternal, String componentId, Versions versions, List<ActionConfiguration> existingUpgrades)
    {
        if (versions instanceof VersionsContainer versionsContainer)
        {
            return _versionContainer2json(isInternal, componentId, versionsContainer, existingUpgrades);
        }
        else if (versions instanceof VersionList versionList)
        {
            return _versionList2json(isInternal, componentId, versionList, existingUpgrades);
        }
        else 
        {
            throw new IllegalArgumentException("Object " + versions + " is not supported");
        }
    }
    
    private List<Map<String, Object>> _versionContainer2json(boolean isInternal, String componentId, VersionsContainer versionsContainer, List<ActionConfiguration> existingUpgrades)
    {
        List<Map<String, Object>> children = new ArrayList<>();
        
        for (Entry<I18nizableText, Versions> entry : versionsContainer.entrySet())
        {
            Versions versions = entry.getValue();
            List<Map<String, Object>> granChildren = _versions2json(isInternal, componentId, versions, existingUpgrades);
            children.add(_component2json(isInternal, componentId, _i18nUtils.translate(entry.getKey()), granChildren, null, null, false, versions instanceof VersionList vl ? vl.id() : null));
        }
        
        Collections.sort(children, new TreeComparator());
        
        return children;
    }

    private List<Map<String, Object>> _versionList2json(boolean isInternal, String componentId, VersionList versionList, List<ActionConfiguration> existingUpgrades)
    {
        List<Map<String, Object>> children = new ArrayList<>();
        
        Set<String> existingUpgradeNumbers = existingUpgrades.stream().map(ActionConfiguration::getVersionNumber).collect(Collectors.toSet()); 
        
        Set<String> done = new TreeSet<>();
        
        Version latestVersion = _migrationEngine.getLatestVersion(versionList.versions());
        
        for (Version version : versionList.versions())
        {
            done.add(version.getVersionNumber());
            children.add(_version2json(isInternal, componentId, versionList.id(), version, latestVersion == version, "0".equals(version.getVersionNumber()) || existingUpgradeNumbers.contains(version.getVersionNumber())));
        }

        if (_isFailedAction(versionList.id())   // There is a failed action 
            && _failedActionIsTheFutureCurrentVersion(existingUpgrades, done, latestVersion)) // That is the current pending action 
        {
            done.add(StringUtils.defaultIfBlank(_migrationEngine.getFailedAction().targetVersionNumber(), _migrationEngine.getFailedAction().configuration().getVersionNumber()));
            children.add(_failureVersion2json(isInternal, componentId, versionList.id()));
        }
        
        for (ActionConfiguration actionConfiguration : existingUpgrades)
        {
            if (!done.contains(actionConfiguration.getVersionNumber()))
            {
                children.add(_pendingVersion2json(isInternal, componentId, versionList.id(), actionConfiguration, _newest(done), _oldest(done)));
            }
        }
        
        Collections.sort(children, new TreeComparator());
        Collections.reverse(children);
        
        if (children.size() > 0 && !"0".equals(children.get(children.size() - 1).get("component"))
            || children.size() == 0)
        {
            // Add the 0 version in the past if necessary
            children.add(_pendingVersion2json(isInternal, componentId, versionList.id(), null, _newest(done), _oldest(done)));
        }
        
        return children;
    }
    
    private boolean _failedActionIsTheFutureCurrentVersion(List<ActionConfiguration> existingUpgrades, Set<String> done, Version latestVersion)
    {
        Set<String> pendingNumbers = new TreeSet<>();
        for (ActionConfiguration actionConfiguration : existingUpgrades)
        {
            String versionToHandle = actionConfiguration.getVersionNumber();
            if (!done.contains(versionToHandle)
                && (latestVersion == null || latestVersion.getVersionNumber().compareTo(versionToHandle) < 0))
            {
                pendingNumbers.add(versionToHandle);
            }
        }
        
        return _migrationEngine.getFailedAction().currentVersion().getVersionNumber() != null && StringUtils.equals(_oldest(pendingNumbers), _migrationEngine.getFailedAction().configuration().getVersionNumber()) // Upgrade failed
                || _migrationEngine.getFailedAction().currentVersion().getVersionNumber() == null && StringUtils.equals(StringUtils.defaultIfBlank(_newest(pendingNumbers), "0"), StringUtils.defaultIfBlank(_migrationEngine.getFailedAction().targetVersionNumber(), _migrationEngine.getFailedAction().configuration().getVersionNumber())); // Init failed
    }

    private Map<String, Object> _plugin2json(String pluginName, List<Map<String, Object>> children)
    {
        Collections.sort(children, new TreeComparator());
        
        return Map.of(
            "component", pluginName,
            "expanded", true,
            "failed", _any(children, "failed"),
            "warning", _any(children, "warning"),
            "type", "plugin",
            "children", children 
        ); 
    }
    
    private Map<String, Object> _component2json(boolean isInternal, String componentId, String componentLabel, List<Map<String, Object>> children, String versionHandlerType, String versionStorageType, boolean rootComponent, String versionListId)
    {
        return Map.of(
            "component", componentLabel,
            "componentId", componentId, 
            "comment", rootComponent ? ("Type: " + versionHandlerType + (StringUtils.equals(versionHandlerType, versionStorageType) ? "" : "/" + versionStorageType)) : "",
            "versionListId", StringUtils.defaultString(versionListId),
            "failed", _any(children, "failed"), 
            "warning", _any(children, "warning"),
            "type", rootComponent ? "component" : "container",
            "internal", isInternal,
            "children", children
        );
    }
    
    private Map<String, Object> _componentFailure2json(boolean isInternal, String componentId, String versionHandlerType, String versionStorageType, MigrationException ex)
    {
        return Map.of(
            "component", componentId,
            "comment", "Type: " + versionHandlerType + (StringUtils.equals(versionHandlerType, versionStorageType) ? "" : "/" + versionStorageType),
            "errorComment", _migrationExceptionToCommentString(ex), 
            "failed", true, 
            "warning", false,
            "internal", isInternal,
            "type", "component"
        );
    }

    private Map<String, Object> _pendingVersion2json(boolean isInternal, String componentId, String versionListId, ActionConfiguration actionConfiguration, String currentStoredVersion, String oldestStoredVersion)
    {
        String versionNumber = actionConfiguration != null ? actionConfiguration.getVersionNumber() : "0";
        
        boolean past = currentStoredVersion != null && versionNumber.compareTo(currentStoredVersion) < 0;
        boolean beforeInitPast = past && versionNumber.compareTo(oldestStoredVersion) < 0;
        
        return Map.of(
                "component", versionNumber,
                "componentId", componentId,
                "versionListId", versionListId,
                "failed", false,
                "warning", false,
                "comment", actionConfiguration != null ? StringUtils.defaultString(actionConfiguration.getComment()) : "",
                "instant", "",
                "internal", isInternal,
                "type", past ? (beforeInitPast ? "past-before" : "past-notdone") : "pending"
            );
    }
    
    private Map<String, Object> _failureVersion2json(boolean isInternal, String componentId, String versionListId)
    {
        return Map.of(
            "component", StringUtils.defaultIfBlank(_migrationEngine.getFailedAction().targetVersionNumber(), _migrationEngine.getFailedAction().configuration().getVersionNumber()),
            "componentId", componentId,
            "versionListId", versionListId,
            "failed", true,
            "warning", false,
            "comment", StringUtils.defaultString(_migrationEngine.getFailedAction().configuration().getComment()),
            "errorComment", _migrationExceptionToCommentString(_migrationEngine.getFailedException()),
            "instant", DateUtils.zonedDateTimeToString(ZonedDateTime.ofInstant(_migrationEngine.getFailedAction().currentVersion().getExecutionInstant(), ZoneOffset.UTC)),
            "internal", isInternal,
            "type", "error"
        );
    }

    private Map<String, Object> _version2json(boolean isInternal, String componentId, String versionListId, Version version, boolean current, boolean existing)
    {
        return Map.of(
            "component", StringUtils.defaultString(version.getVersionNumber()),
            "componentId", componentId,
            "versionListId", versionListId,
            "comment", StringUtils.defaultString(version.getComment()),
            "errorComment", !existing ? _i18nUtils.translate(new I18nizableText("plugin.admin", "PLUGINS_ADMIN_TOOL_MIGRATIONS_COL_COMMENT_NONEXISTING")) : "", 
            "failed", false,
            "warning", !existing,
            "instant", version.getExecutionInstant() != null ? DateUtils.zonedDateTimeToString(ZonedDateTime.ofInstant(version.getExecutionInstant(), ZoneOffset.UTC)) : "",
            "internal", isInternal,
            "type", current ? "current" : "past-done"
        );
    }
    
    private String _newest(Set<String> s)
    {
        return s.size() > 0 ? s.stream().skip(s.size() - 1).findFirst().orElse(null) : null;
    }
    private String _oldest(Set<String> s)
    {
        return s.stream().findFirst().orElse(null);
    }
    
    private boolean _isFailedAction(String versionListId)
    {
        return _migrationEngine.getFailedAction() != null
            && StringUtils.equals(versionListId, _migrationEngine.getFailedAction().versionListId());
    }
    
    private String _migrationExceptionToCommentString(MigrationException ex)
    {
        return ex != null ? ex.getFailureMessage().replaceAll("<", "&lt;").replaceAll("\n", "<br/>") : "";
    }
    
    @SuppressWarnings({"cast", "unchecked"})
    private boolean _any(List<Map<String, Object>> children, String key)
    {
        for (Map<String, Object> child : children)
        {
            if (child.get(key) == Boolean.TRUE)
            {
                return true;
            }
            if (child.get("children") instanceof List granChildren
                && _any((List<Map<String, Object>>) granChildren, key))
            {
                return true;
            }
        }
        
        return false;
    }
    
    private class TreeComparator implements Comparator<Map> 
    {
        public int compare(Map o1, Map o2)
        {
            return ((String) o1.get("component")).toLowerCase().compareTo(((String) o2.get("component")).toLowerCase());
        }
    }
}
