/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugins.admin.statistics;

import java.lang.management.ManagementFactory;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.authentication.CredentialProvider;
import org.ametys.core.authentication.CredentialProviderFactory;
import org.ametys.core.authentication.CredentialProviderModel;
import org.ametys.core.group.GroupDirectoryDAO;
import org.ametys.core.group.directory.GroupDirectory;
import org.ametys.core.group.directory.GroupDirectoryFactory;
import org.ametys.core.user.directory.UserDirectory;
import org.ametys.core.user.directory.UserDirectoryFactory;
import org.ametys.core.user.population.UserPopulation;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.core.version.VersionsHandler;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.PluginsManager;
import org.ametys.runtime.plugin.component.PluginAware;
import org.ametys.runtime.servlet.AnalyseFileForVirusHelper;

/**
 * All statistics of kernel
 */
public class KernelStatisticsProvider implements StatisticsProvider, Serviceable, PluginAware
{
    private String _id;
    private VersionsHandler _versionsHandler;
    private UserPopulationDAO _userPopulationDAO;
    private UserDirectoryFactory _userDirectoryFactory;
    private CredentialProviderFactory _credentialProviderFactory;
    private GroupDirectoryDAO _groupDirectoryDAO;
    private GroupDirectoryFactory _groupDirectoryFactory;

    public void service(ServiceManager manager) throws ServiceException
    {
        _versionsHandler = (VersionsHandler) manager.lookup(VersionsHandler.ROLE);
        _userPopulationDAO = (UserPopulationDAO) manager.lookup(UserPopulationDAO.ROLE);
        _userDirectoryFactory = (UserDirectoryFactory) manager.lookup(UserDirectoryFactory.ROLE);
        _credentialProviderFactory = (CredentialProviderFactory) manager.lookup(CredentialProviderFactory.ROLE);
        _groupDirectoryDAO = (GroupDirectoryDAO) manager.lookup(GroupDirectoryDAO.ROLE);
        _groupDirectoryFactory = (GroupDirectoryFactory) manager.lookup(GroupDirectoryFactory.ROLE);
    }
    
    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        _id = id;
    }
    
    public Statistics getStatistics()
    {
        return new StatisticsNode(
                _id, 
                new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_LABEL"),
                "ametysicon-trademark-ametys",
                null,
                List.of(
                    _getVersionsStatistics(),
                    _getPluginsStatistics(),
                    _getConfigurationStatistics(),
                    _getPopulationsAndGropDirectoriesStatistics(),
                    _getServerStatistics()
                ),
                true
            );
    }
    
    
    private Statistics _getServerStatistics()
    {
        return new StatisticsNode(
                "server", 
                new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_SERVER_LABEL"),
                "ametysicon-system-server-sync",
                null,
                List.of(
                    new StatisticsValue(
                        "osname", 
                        new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_SERVER_OS_LABEL"), 
                        "ametysicon-system-terminal", 
                        ManagementFactory.getOperatingSystemMXBean().getName() + " (" + ManagementFactory.getOperatingSystemMXBean().getVersion() + ")"
                    ),
                    new StatisticsValue(
                        "cpu", 
                        new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_SERVER_CPU_LABEL"), 
                        "ametysicon-trademark-jackrabbit", 
                        String.valueOf(ManagementFactory.getOperatingSystemMXBean().getAvailableProcessors()) + " x " + ManagementFactory.getOperatingSystemMXBean().getArch()
                    ),
                    new StatisticsValue(
                        "ram", 
                        new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_SERVER_RAM_LABEL"), 
                        "ametysicon-system-sgbd", 
                        ManagementFactory.getMemoryMXBean().getHeapMemoryUsage().getMax()
                    ),
                    new StatisticsValue(
                        "vmvendor", 
                        new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_SERVER_VMVENDOR_LABEL"), 
                        "ametysicon-trademark-jstack", 
                        System.getProperty("java.vendor") + " " + ManagementFactory.getRuntimeMXBean().getVmName()
                    ),
                    new StatisticsValue(
                        "vmversion", 
                        new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_SERVER_VMVERSION_LABEL"), 
                        "ametysicon-trademark-jstack", 
                        System.getProperty("java.version")
                    ),
                    new StatisticsValue(
                        "antivirus", 
                        new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_SERVER_ANTIVIRUS_LABEL"), 
                        "ametysicon-arrow-down-in", 
                        AnalyseFileForVirusHelper.isAntivirusEnabled()
                    )
                ),
                false
                );
    }
    
    private Statistics _getVersionsStatistics()
    {
        return new StatisticsNode(
            "versions", 
            new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_VERSIONS_LABEL"),
            "ametysicon-maths-number-zero-one",
            null,
            _getVersionsChildren(),
            true
        );
    }
    
    private List<Statistics> _getVersionsChildren()
    {
        return _versionsHandler.getVersions().stream()
            .map(v -> new StatisticsValue(
                v.getName(), 
                new I18nizableText(v.getName()), 
                "ametysicon-trademark-ametys", 
                v.getVersion())
            )
            .map(Statistics.class::cast)
            .toList();
    }
    
    private Statistics _getPluginsStatistics()
    {
        List<String> plugins = new ArrayList<>(PluginsManager.getInstance().getBundledPluginsNames());
        Collections.sort(plugins);
        
        return new StatisticsValue(
            "plugins", 
            new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_PLUGINS_LABEL"),
            "ametysicon-puzzle-piece1",
            plugins
        );
    }
    
    private Statistics _getPopulationsAndGropDirectoriesStatistics()
    {
        long populations = 0;
        
        Map<String, Long> userDirectoriesCount = new TreeMap<>();
        Map<String, Long> credentialProvidersCount = new TreeMap<>();
        
        for (UserPopulation userPopulation : _userPopulationDAO.getUserPopulations(false))
        {
            populations++;
            
            for (UserDirectory userDirectory : userPopulation.getUserDirectories())
            {
                Long value = userDirectoriesCount.getOrDefault(userDirectory.getUserDirectoryModelId(), 0L);
                userDirectoriesCount.put(userDirectory.getUserDirectoryModelId(), value + 1);
            }
            
            for (CredentialProvider credentialProvider : userPopulation.getCredentialProviders())
            {
                Long value = credentialProvidersCount.getOrDefault(credentialProvider.getCredentialProviderModelId(), 0L);
                credentialProvidersCount.put(credentialProvider.getCredentialProviderModelId(), value + 1);
            }
        }
        
        List<Statistics> userDirectories = new ArrayList<>();
        for (Entry<String, Long> entry : userDirectoriesCount.entrySet())
        {
            userDirectories.add(new StatisticsValue(
                entry.getKey(),
                _userDirectoryFactory.getExtension(entry.getKey()).getLabel(),
                "ametysicon-body-idcard-white",
                entry.getValue()
            ));
        }
        
        List<Statistics> credentialProviders = new ArrayList<>();
        for (Entry<String, Long> entry : credentialProvidersCount.entrySet())
        {
            CredentialProviderModel cp = _credentialProviderFactory.getExtension(entry.getKey());
            credentialProviders.add(new StatisticsValue(
                entry.getKey(),
                cp.getLabel(),
                StringUtils.defaultIfBlank(cp.getIconGlyph(), "ametysicon-body-idcard-badge"),
                entry.getValue()
            ));
        }
        
        long groups = 0;

        Map<String, Long> groupDirectoriesCount = new TreeMap<>();
        for (GroupDirectory groupDirectory : _groupDirectoryDAO.getGroupDirectories())
        {
            groups++;
            
            Long value = groupDirectoriesCount.getOrDefault(groupDirectory.getGroupDirectoryModelId(), 0L);
            groupDirectoriesCount.put(groupDirectory.getGroupDirectoryModelId(), value + 1);
        }
        
        List<Statistics> groupDirectories = new ArrayList<>();
        for (Entry<String, Long> entry : groupDirectoriesCount.entrySet())
        {
            groupDirectories.add(new StatisticsValue(
                entry.getKey(),
                _groupDirectoryFactory.getExtension(entry.getKey()).getLabel(),
                "ametysicon-body-people",
                entry.getValue()
            ));
        }
        
        return new StatisticsNode(
            "populations-groupdirectories", 
            new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_USERPOPULATIONSANDGROUPDIRECTORIES_LABEL"),
            "ametysicon-body-group",
            populations + groups,
            List.of(
                new StatisticsNode(
                    "populations", 
                    new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_USERPOPULATIONSANDGROUPDIRECTORIES_POPULATIONS_LABEL"),
                    "ametysicon-body-people",
                    populations,
                    List.of(
                        new StatisticsNode(
                            "userdirectories", 
                            new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_USERPOPULATIONSANDGROUPDIRECTORIES_POPULATIONS_USERDIRECTORIES_LABEL"),
                            "ametysicon-body-idcard-white",
                            null,
                            userDirectories,
                            false
                        ),
                        new StatisticsNode(
                            "credentialproviders", 
                            new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_USERPOPULATIONSANDGROUPDIRECTORIES_POPULATIONS_CREDENTIALPROVIDERS_LABEL"),
                            "ametysicon-body-idcard-badge",
                            null,
                            credentialProviders,
                            false
                        )                                
                    ),
                    false
                ),
                new StatisticsNode(
                    "groupdirectories", 
                    new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_USERPOPULATIONSANDGROUPDIRECTORIES_GROUPDIRECTORIES_LABEL"),
                    "ametysicon-body-people",
                    groups,
                    groupDirectories,
                    false
                )
            ),
            false
        );
    }
    
    private StatisticsNode _getConfigurationStatistics()
    { 
        boolean configProd = !Config.getInstance().getValue("runtime.mode.dev", false, true);
        String configCaptchaType = Config.getInstance().getValue("runtime.captcha.type", false, "");
        Long configMaxUpload = Config.getInstance().getValue("runtime.upload.max-size", false, 0L);
        boolean configGravatar = Config.getInstance().getValue("runtime.userprofile.imagesource.gravatar", false, true);
        
        return new StatisticsNode(
            "config",
            new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_CONFIG_LABEL"),
            "ametysicon-gear39",
            null,
            List.of(
                new StatisticsValue(
                    "production",
                    new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_CONFIG_PRODUCTION_LABEL"),
                    "ametysicon-movie16",
                    configProd
                ),
                new StatisticsValue(
                    "captcha",
                    new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_CONFIG_CAPTCHA_LABEL"),
                    "ametysicon-abecedary4",
                    configCaptchaType
                ),
                new StatisticsValue(
                    "upload",
                    new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_CONFIG_UPLOAD_LABEL"),
                    "ametysicon-arrow-up-from",
                    configMaxUpload
                ),
                new StatisticsValue(
                    "gravatar",
                    new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_KERNEL_CONFIG_GRAVATAR_LABEL"),
                    "ametysicon-body-idcard-black",
                    configGravatar
                )
            ),
            false
        );
    }
}
