/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugins.admin.statistics;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.core.ui.Callable;
import org.ametys.core.util.JSONUtils;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentExtensionPoint;
import org.ametys.runtime.util.AmetysHomeHelper;

/**
 * Extension point handling the remove statistics
 */
public class StatisticsProviderExtensionPoint extends AbstractThreadSafeComponentExtensionPoint<StatisticsProvider>
{
    /** The component role */
    public static final String ROLE = StatisticsProviderExtensionPoint.class.getName();
    
    static final String _I18N_PREFIX = "{{i18n ";
    static final String _I18N_SEPARATOR = ":";
    static final String _I18N_SUFFIX = "}}";
    
    private static final String FILE = AmetysHomeHelper.AMETYS_HOME_ADMINISTRATOR_DIR + File.separator + "statistics.json";
    private JSONUtils _jsonUtils;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
    }
    
    /**
     * Generate the statistics
     * @return The non null map of statistics
     * @throws IOException If an error occurred while caching the results
     */
    public Map<String, Object> computeStatistics() throws IOException
    {
        Map<String, Object> jsonStatistics = new HashMap<>();
        List<Map<String, Object>> treeStatistics = new ArrayList<>();
        
        for (String extensionId : getExtensionsIds())
        {
            StatisticsProvider extension = getExtension(extensionId);
            
            Statistics extensionStatistics = extension.getStatistics();
            
            jsonStatistics.putAll(extensionStatistics.toJSONForExport());
            treeStatistics.add(extensionStatistics.toJSONTree());
        }
        
        Path file = AmetysHomeHelper.getAmetysHome().toPath().resolve(FILE);
        Files.createDirectories(file.getParent());

        try (OutputStream os = Files.newOutputStream(file))
        {
            _jsonUtils.convertObjectToJson(os, treeStatistics);
        }
        
        return jsonStatistics;
    }
    
    /**
     * Generate the statistics with tree information for client display
     * @return The non null map of statistics
     * @throws IOException If an error occurred reading the cache file
     */
    @Callable(context = "/admin", rights = "Runtime_Rights_Admin_Access")
    public Object getStatisticsTree() throws IOException
    {
        Path file = AmetysHomeHelper.getAmetysHome().toPath().resolve(FILE);
        if (Files.exists(file))
        {
            String json = Files.readString(file, StandardCharsets.UTF_8);
            List<Object> tree = _jsonUtils.convertJsonToList(json);
            _handleI18n(tree);
            if (tree.size() > 0)
            {
                return tree;
            }
        }
         
        return Map.of(
                "children", List.of(Map.of(
                            "name", "error",
                            "label", new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_TOOL_EMPTY"),
                            "value", new I18nizableText("plugin.admin", "PLUGINS_ADMIN_STATISTICS_TOOL_EMPTY2"),
                            "iconCls", "ametysicon-sign-caution",
                            "leaf", true)
        ));
    }

    @SuppressWarnings("unchecked")
    private void _handleI18n(Object object)
    {
        if (object instanceof List list)
        {
            for (Object node : list)
            {
                _handleI18n(node);
            }
        }
        else if (object instanceof Map map)
        {
            Object label = map.get("label");
            if (label instanceof String string
                && string.startsWith(_I18N_PREFIX) && string.endsWith(_I18N_SUFFIX))
            {
                String[] i18n = string.substring(_I18N_PREFIX.length(), string.length() - _I18N_SUFFIX.length()).trim().split(_I18N_SEPARATOR);
                I18nizableText i18nizableText = new I18nizableText(i18n[0], i18n[1]);
                map.put("label", i18nizableText);
            }
            
            Object children = map.get("children");
            _handleI18n(children);
        }
    }
}
