/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugins.admin.statistics;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;
import org.quartz.JobExecutionContext;

import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.core.impl.schedule.AbstractStaticSchedulable;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.servlet.RuntimeServlet;

/**
 * Compute and optionally sent the anonymous statistics to the central ametys.org server 
 */
public class StatisticsSchedulable extends AbstractStaticSchedulable
{
    private static final String CENTRAL_SERVER_URL = "https://statistics.ametys.org/_update-version/statistics/1.0.0/upload.json";
    private static final String CENTRAL_SERVER_HEADER = "X-Ametys-Statistics";
    
    private JSONUtils _jsonUtils;
    private StatisticsProviderExtensionPoint _statisticsExtensionPoint;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
        _statisticsExtensionPoint = (StatisticsProviderExtensionPoint) manager.lookup(StatisticsProviderExtensionPoint.ROLE);
    }
    
    @Override
    public void execute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception
    {
        getLogger().info("Preparing statistics");
        Map<String, Object> jsonStatistics = _statisticsExtensionPoint.computeStatistics();

        if (Config.getInstance().getValue("runtime.statistics.send-at-night", false, false))
        {
            getLogger().info("Sending remote statistics");
            _sendReport(jsonStatistics);
        }
    }

    private void _sendReport(Map<String, Object> report)
    {
        RequestConfig requestConfig = RequestConfig.custom().setConnectTimeout(30000).setSocketTimeout(30000).build();
        
        try (CloseableHttpClient httpclient = HttpClientBuilder.create().useSystemProperties().setDefaultRequestConfig(requestConfig).build())
        {
            // Prepare a request object
            HttpPost request = new HttpPost(CENTRAL_SERVER_URL);
            request.addHeader(CENTRAL_SERVER_HEADER, RuntimeServlet.getInstanceId());
            request.setEntity(new UrlEncodedFormEntity(Set.of(new BasicNameValuePair("value", _jsonUtils.convertObjectToJson(report))), StandardCharsets.UTF_8));
            
            // Execute the request
            try (CloseableHttpResponse response = httpclient.execute(request))
            {
                if (response.getStatusLine().getStatusCode() != 200)
                {
                    throw new IllegalStateException("Could not join the central ametys.org server at " + CENTRAL_SERVER_URL + ". Error code " + response.getStatusLine().getStatusCode());
                }
                else if (!response.containsHeader(CENTRAL_SERVER_HEADER))
                {
                    throw new IllegalStateException("Could not join the central ametys.org server at " + CENTRAL_SERVER_URL + ". Response code is 200, but there is not " + CENTRAL_SERVER_HEADER + " header");
                }
                
                try
                {
                    String responseAsString = EntityUtils.toString(response.getEntity(), "UTF-8");
                    Map<String, Object> convertJsonToMap = _jsonUtils.convertJsonToMap(responseAsString);
                    
                    if (!(convertJsonToMap.get("success") instanceof Boolean b && b == Boolean.TRUE))
                    {
                        throw new IllegalStateException("Joined the central ametys.org server at " + CENTRAL_SERVER_URL + ". But the operation failed.");
                    }
                }
                catch (IllegalArgumentException e)
                {
                    throw new IllegalStateException("Joined the central ametys.org server at " + CENTRAL_SERVER_URL + ". But cannot parse the response.", e);
                }
            }
        }
        catch (IOException e)
        {
            throw new IllegalStateException("Could not join the central ametys.org server at " + CENTRAL_SERVER_URL, e);
        }
    }
}
