/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugins.admin.system;

import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.runtime.servlet.RuntimeServlet;
import org.ametys.runtime.servlet.RuntimeServlet.ForcedMainteanceInformations;
import org.ametys.runtime.servlet.RuntimeServlet.MaintenanceStatus;
import org.ametys.runtime.servlet.RuntimeServlet.RunMode;

/**
 * The maintenance button
 */
public class MaintenanceClientSideElement extends StaticClientSideElement
{
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        List<Script> scripts = super.getScripts(ignoreRights, contextParameters);
        List<Script> modifiedScripts = new ArrayList<>();
        
        for (Script script : scripts)
        {
            Script modifiedScript = new Script(script);
            modifiedScripts.add(modifiedScript);
            
            switch (RuntimeServlet.getMaintenanceStatus())
            {
                case AUTOMATIC:
                    modifiedScript.getParameters().put("toggle-state", "true");
                    modifiedScript.getParameters().put("disabled", "true");
                    modifiedScript.getParameters().put("description", modifiedScript.getParameters().get("automatic-description"));
                    modifiedScript.getParameters().put("icon-decorator", modifiedScript.getParameters().get("on-icon-decorator"));
                    modifiedScript.getParameters().put("icon-decorator-type", modifiedScript.getParameters().get("on-icon-decorator-type"));
                    break;
                case FORCED: 
                    modifiedScript.getParameters().put("toggle-state", "true");
                    modifiedScript.getParameters().put("disabled", "false");
                    modifiedScript.getParameters().put("description", modifiedScript.getParameters().get("forced-on-description"));
                    modifiedScript.getParameters().put("icon-decorator", modifiedScript.getParameters().get("on-icon-decorator"));
                    modifiedScript.getParameters().put("icon-decorator-type", modifiedScript.getParameters().get("on-icon-decorator-type"));
                    break;
                case NONE:
                default:
                    modifiedScript.getParameters().put("toggle-state", "false");
                    modifiedScript.getParameters().put("disabled", "false");
                    modifiedScript.getParameters().put("description", modifiedScript.getParameters().get("forced-off-description"));
                    modifiedScript.getParameters().put("icon-decorator", modifiedScript.getParameters().get("off-icon-decorator"));
                    modifiedScript.getParameters().put("icon-decorator-type", modifiedScript.getParameters().get("off-icon-decorator-type"));
                    break;
            }
        }
        
        return modifiedScripts;
    }
    
    /**
     * Set maintenance status to normal, so maintenance is not activated anymore
     */
    @Callable(context = "/admin", rights = "Runtime_Rights_Admin_Maintenance_Activate")
    public void goToNormalMode()
    {
        if (RuntimeServlet.getRunMode() == RunMode.MAINTENANCE
            && RuntimeServlet.getMaintenanceStatus() == MaintenanceStatus.FORCED)
        {
            RuntimeServlet.setMaintenanceStatus(MaintenanceStatus.NONE, null);
            RuntimeServlet.setRunMode(RunMode.NORMAL);
        }
    }
    
    /**
     * Set maintenance status to maintenance, so maintenance is active
     * @param comment The maintenance comment
     */
    @Callable(context = "/admin", rights = "Runtime_Rights_Admin_Maintenance_Activate")
    public void goToMaintenanceMode(String comment)
    {
        if (RuntimeServlet.getRunMode() == RunMode.NORMAL)
        {
            RuntimeServlet.setMaintenanceStatus(MaintenanceStatus.FORCED, new ForcedMainteanceInformations(comment, _currentUserProvider.getUser(), ZonedDateTime.now()));
            RuntimeServlet.setRunMode(RunMode.MAINTENANCE);
        }
    }
}
