/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugins.admin.system;

import java.time.ZonedDateTime;

import org.apache.commons.lang3.StringUtils;
import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;

import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.core.impl.schedule.AbstractStaticSchedulable;
import org.ametys.plugins.core.schedule.Scheduler;
import org.ametys.runtime.servlet.RuntimeServlet;
import org.ametys.runtime.servlet.RuntimeServlet.ForcedMainteanceInformations;
import org.ametys.runtime.servlet.RuntimeServlet.MaintenanceStatus;
import org.ametys.runtime.servlet.RuntimeServlet.RunMode;

/**
 * The task to enter or leave the maintenance mode 
 */
public class MaintenanceSchedulable extends AbstractStaticSchedulable
{
    /** The key for mode */
    public static final String JOBDATAMAP_MODE = "mode";
    /** The key for comment */
    public static final String JOBDATAMAP_COMMENT = "comment";
    
    @Override
    public void execute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception
    {
        JobDataMap jobDataMap = context.getJobDetail().getJobDataMap();
        
        boolean mode = jobDataMap.getBoolean(Scheduler.PARAM_VALUES_PREFIX + JOBDATAMAP_MODE);
        String comment = StringUtils.defaultString(jobDataMap.getString(Scheduler.PARAM_VALUES_PREFIX + JOBDATAMAP_COMMENT));
        UserIdentity userIdentity = UserIdentity.stringToUserIdentity(jobDataMap.getString(Scheduler.KEY_RUNNABLE_USERIDENTITY));
        
        if (mode)
        {
            getLogger().info("Entering maintenance mode");
            
            if (RuntimeServlet.getRunMode() != RunMode.NORMAL)
            {
                throw new IllegalStateException("Cannot enter the maintenance mode because the current mode is " + RuntimeServlet.getRunMode());
            }
            
            RuntimeServlet.setMaintenanceStatus(MaintenanceStatus.FORCED, new ForcedMainteanceInformations(comment, userIdentity, ZonedDateTime.now()));
            RuntimeServlet.setRunMode(RunMode.MAINTENANCE);
        }
        else
        {
            getLogger().info("Leaving maintenance mode");
            
            if (RuntimeServlet.getRunMode() != RunMode.MAINTENANCE || RuntimeServlet.getMaintenanceStatus() == MaintenanceStatus.AUTOMATIC)
            {
                throw new IllegalStateException("Cannot leave the maintenance mode because the current mode is " + RuntimeServlet.getRunMode() + " and maintenance status " + RuntimeServlet.getMaintenanceStatus());
            }
            
            RuntimeServlet.setMaintenanceStatus(MaintenanceStatus.NONE, null);
            RuntimeServlet.setRunMode(RunMode.NORMAL);
        }
    }
}
