/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.authentication;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.time.Instant;
import java.time.ZonedDateTime;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Cookie;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.Response;
import org.apache.cocoon.environment.Session;
import org.apache.cocoon.environment.http.HttpCookie;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.core.authentication.AbstractCredentialProvider;
import org.ametys.core.authentication.AuthenticateAction;
import org.ametys.core.authentication.BlockingCredentialProvider;
import org.ametys.core.authentication.LogoutCapable;
import org.ametys.core.authentication.NonBlockingCredentialProvider;
import org.ametys.core.authentication.token.AuthenticationTokenManager;
import org.ametys.core.captcha.CaptchaHelper;
import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.migration.MigrationEngine;
import org.ametys.core.migration.MigrationEngine.ActionData;
import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.MigrationExtensionPoint;
import org.ametys.core.migration.version.Version;
import org.ametys.core.migration.version.storage.SqlVersionStorage.SqlVersionConfiguration;
import org.ametys.core.script.SQLScriptHelper;
import org.ametys.core.trace.ForensicLogger;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.user.directory.NotUniqueUserException;
import org.ametys.core.user.directory.UserDirectory;
import org.ametys.core.user.directory.UserDirectory.CredentialsResult;
import org.ametys.core.user.directory.WeakPasswordException;
import org.ametys.core.user.population.UserPopulation;
import org.ametys.plugins.core.authentication.MultifactorAuthenticationManager;
import org.ametys.plugins.core.authentication.MultifactorAuthenticationManager.MultifactorAuthenticationCode;
import org.ametys.runtime.authentication.AccessDeniedException;
import org.ametys.runtime.plugin.PluginsManager;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * This manager gets the credentials coming from an authentication form. <br>
 * This manager can create a cookie to save credentials
 * <br>
 * Parameters are : - The name of the pool<br>
 *                  - The html field name for user name<br>
 *                  - The html field name for user password<br>
 *                  - The html field name for the check box which allow to create a cookie, must return 'true' when checked<br>
 *                  - A boolean, to activate or not the user info saving by cookie <br>
 *                  - The cookie name, to retrieve info<br>
 *                  - The cookie duration (in seconds), by default set to 1 week<br>
 *                  - A login url (do not start with a "/")<br>
 *                  - A failure login url (do not start with a "/"). The failure Url can receive the login entered by the visitor.<br>
 *                  - A list of URL prefixes that are accessible without authentication. The login and failure URLs are always accessible without authentication.<br><br>
 * 
 * For example :<br>
 *               &lt;username-field&gt;Username&lt;/username-field&gt;<br>
 *               &lt;password-field&gt;Password&lt;/password-field&gt;<br>
 *               &lt;cookie&gt;<br>
 *                   &nbsp;&nbsp;&lt;cookieEnabled&gt;true&lt;/cookieEnabled&gt;<br>
 *                   &nbsp;&nbsp;&lt;cookieLifeTime&gt;604800&lt;/cookieLifeTime&gt;<br>
 *                   &nbsp;&nbsp;&lt;cookieName&gt;AmetysAuthentication&lt;/cookieName&gt;<br>
 *               &lt;/cookie&gt;<br>
 *               &lt;loginUrl internal="true"&gt;login.html&lt;/loginUrl&gt;<br>
 *               &lt;loginFailedUrl provideLoginParameter="true" internal="true"&gt;login_failed.html&lt;/loginFailedUrl&gt;<br>
 *               &lt;unauthenticated&gt;<br>
 *                   &nbsp;&nbsp;&lt;urlPrefix&gt;subscribe.html&lt;/urlPrefix&gt;<br>
 *                   &nbsp;&nbsp;&lt;urlPrefix&gt;lostPassword/&lt;/urlPrefix&gt;<br>
 *               &lt;/unauthenticated&gt;<br>
 * 
 */
public class FormCredentialProvider extends AbstractCredentialProvider implements NonBlockingCredentialProvider, BlockingCredentialProvider, LogoutCapable, Contextualizable, Configurable, Serviceable
{
    /** Password value in case of info retrieved from cookie */
    public static final String AUTHENTICATION_BY_COOKIE = "authentication_by_cookie";
    /** Number of connection attempts allowed */
    public static final Integer NB_CONNECTION_ATTEMPTS = 3;
    /** Default cookie lifetime (15 days in seconds) */
    public static final int COOKIE_LIFETIME = 15 * 24 * 60 * 60;
    
    /** The session attribute containing the password, if the corresponding option has been checked */
    public static final String PASSWORD_SESSION_ATTRIBUTE = "runtime.authentication.password";
    
    /** Name of the Login URL parameter for the current user identity */
    public static final String LOGIN_URL_USER_PARAMETER_NAME = "userIdentity";
    /** Name of the Login URL parameter for the current user identity */
    public static final String LOGIN_URL_INVALID_PASSCODE_PARAMETER_NAME = "invalidMFACode";
    /** Name of the Login URL parameter for the mfa mode to use */
    public static final String LOGIN_URL_MFA_MODE = "MFAMode";
    
    /** Duration in days a connection failure will last */
    protected static final Integer TIME_ALLOWED = 1;
    
    /** Name of the parameter holder the datasource id */
    private static final String __PARAM_DATASOURCE = "runtime.authentication.form.security.storage";
    /** Name of the parameter holding the captcha options */
    private static final String __PARAM_CAPTCHA = "runtime.authentication.form.captcha";
    /** Name of the parameter holding the cookies option */
    private static final String __PARAM_COOKIES = "runtime.authentication.form.cookies";
    /** Name of the parameter holding the allow of login by email */
    private static final String __PARAM_LOGIN_BY_EMAIL = "runtime.authentication.form.login-by-email";
    /** Name of the parameter holding the password in session option */
    private static final String __PARAM_STORE_PASSWORD_IN_SESSION = "runtime.authentication.form.store-password-in-session";
    /** Name of the parameter holding the multifactor authentication option */
    private static final String __PARAM_MULTIFACTOR_AUTHENTICATION = "runtime.authentication.form.multifactor";
    
    /** The session attribute name for storing the identity of the user (during connection process) */
    private static final String __SESSION_CONNECTING_USERIDENTITY = "Runtime:ConnectingUserIdentity";
    
    /** Name of the user name html field */
    protected String _usernameField;

    /** Name of the user password html field */
    protected String _passwordField;
    
    /** Name of the "remember me" html field */
    protected String _rememberMeField;
    
    /** Name of the captcha answer html field */
    protected String _captchaField;
    
    /** Name of the captcha key html field */
    protected String _captchaKeyField;
    
    /** Name of the multifactor authentication code field */
    protected String _mfaCodeField;
    
    /** Name of the multifactor authentication field to force mail code */
    protected String _mfaMode;

    /** Indicates if the user credentials must be saved by a cookie */
    protected boolean _cookieEnabled;

    /** The name of the cookie */
    protected String _cookieName;

    /** Cookie duration in seconds, by default COOKIE_LIFETIME */
    protected long _cookieLifetime;

    /** Set of accepted url prefixes (default : empty). */
    protected Set<String> _acceptedUrlPrefixes;
    
    /** Security level : use captcha after three connection failed */
    protected boolean _useCaptchaOnFailure;
    
    /** Security level : allow storage in cookies */
    protected boolean _allowCookies;
    
    /** Security level : store password in session */
    protected boolean _storePasswordInSession;
    
    /** Allow login by email */
    protected boolean _allowLoginByEmail;
    
    /** use multifactor authentication */
    protected boolean _useMultifactorAuthentication;

    /** Context */
    protected Context _context;
    
    /** The datasource id */
    protected String _datasourceId;
    /** The avalon source resolver */
    protected SourceResolver _sourceResolver;
    /** The token manager */
    protected AuthenticationTokenManager _authenticationTokenManager;
    /** The user manager */
    protected UserManager _userManager;
    /** The multifactor authentication manager */
    protected MultifactorAuthenticationManager _multifactorAuthenticationManager;
    
    /** was lazy initialize done */
    protected boolean _lazyInitialized;
    /** The migration engine */
    protected MigrationEngine _migrationEngine;
    /** The migration extension point */
    protected MigrationExtensionPoint _migrationEP;
    
    /**
     * The mode of MFA to user
     */
    public enum MFAMode
    {
        /** MFA from mail */
        EMAIL,
        /** MFA from application */
        APPLICATION
    }
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _multifactorAuthenticationManager = (MultifactorAuthenticationManager) manager.lookup(MultifactorAuthenticationManager.ROLE);
        _migrationEngine = (MigrationEngine) manager.lookup(MigrationEngine.ROLE);
        _migrationEP = (MigrationExtensionPoint) manager.lookup(MigrationExtensionPoint.ROLE);
        
        try
        {
            _authenticationTokenManager = (AuthenticationTokenManager) manager.lookup(AuthenticationTokenManager.ROLE);
        }
        catch (ServiceException e)
        {
            // nothing... we are in safe mode
        }
        
        // Ensure statics methods will be available during initialize
        manager.lookup(ConnectionHelper.ROLE);
    }
    
    @Override
    public void init(String id, String cpModelId, Map<String, Object> paramValues, String label) throws Exception
    {
        super.init(id, cpModelId, paramValues, label);
        _useCaptchaOnFailure = (Boolean) paramValues.get(__PARAM_CAPTCHA);
        _allowCookies = (Boolean) paramValues.get(__PARAM_COOKIES);
        _allowLoginByEmail = (Boolean) paramValues.get(__PARAM_LOGIN_BY_EMAIL);
        _datasourceId = (String) paramValues.get(__PARAM_DATASOURCE);
        _storePasswordInSession = (Boolean) paramValues.get(__PARAM_STORE_PASSWORD_IN_SESSION);
        _useMultifactorAuthentication = (Boolean) paramValues.get(__PARAM_MULTIFACTOR_AUTHENTICATION);
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _usernameField = configuration.getChild("username-field").getValue("Username");
        _passwordField = configuration.getChild("password-field").getValue("Password");
        _rememberMeField =  configuration.getChild("rememberMe-field").getValue("rememberMe");
        _captchaField =  configuration.getChild("capcha-field").getValue("Captcha");
        _captchaKeyField =  configuration.getChild("captchaKey-field").getValue("CaptchaKey");
        _mfaCodeField =  configuration.getChild("mfaCode-field").getValue("MultifactorAuthenticationCode");
        _mfaMode = configuration.getChild("mfaMode-field").getValue("MultifactorAuthenticationMode");
        _cookieEnabled = configuration.getChild("cookie").getChild("cookieEnabled").getValueAsBoolean(true);
        _cookieLifetime = configuration.getChild("cookie").getChild("cookieLifeTime").getValueAsLong(COOKIE_LIFETIME);
        _cookieName = configuration.getChild("cookie").getChild("cookieName").getValue("AmetysAuthentication");
        _acceptedUrlPrefixes = new HashSet<>();
        for (Configuration prefixConf : configuration.getChild("unauthenticated").getChildren("urlPrefix"))
        {
            String prefix = prefixConf.getValue(null);
            if (prefix != null)
            {
                _acceptedUrlPrefixes.add(prefix);
            }
        }
        
        if (getLogger().isDebugEnabled())
        {
            getLogger().debug(
                    "FormBasedCredentialsProvider values : " + " Name field=" + _usernameField + ", Pwd field="
                            + _passwordField + ", CookieEnabled=" + _cookieEnabled + ", Cookie duration="
                            + _cookieLifetime + ", Cookie name=" + _cookieName
                            + ", accepted prefixes : [" + StringUtils.join(_acceptedUrlPrefixes, ", ") + "]");
        }
    }
    
    /**
     * Get the connection to the database 
     * @return the SQL connection
     */
    protected Connection getSQLConnection()
    {
        Connection connection = ConnectionHelper.getConnection(_datasourceId);
        
        if (!_lazyInitialized)
        {
            MigrationComponent component = _migrationEP.getExtension("credential-provider.form");
            SqlVersionConfiguration sqlVersionConfiguration = new SqlVersionConfiguration(_datasourceId);
            
            try
            {
                boolean tableExists = SQLScriptHelper.tableExists(connection, "Users_FormConnectionFailed");
                if (tableExists && component.versionStorage().getVersions(component.id(), sqlVersionConfiguration, component).size() == 0)
                {
                    throw new IllegalStateException("In datasource " + _datasourceId + " the table Users_FormConnectionFailed is not versionned. Should never happen. Drop the table or restard Ametys.");
                }
                
                if (!tableExists)
                {
                    
                    Version version = component.versionStorage().createVersion(component.id(), component, sqlVersionConfiguration, Map.of("TABLENAME", "Users_FormConnectionFailed"));
                    if (_migrationEngine.initializeVersion(new ActionData(version, _migrationEngine.getHighestUpgradeVersionNumber(component), component.initialization(), null)))
                    {
                        throw new IllegalStateException("Requiring a restart after in a table initialization for a " + this.getClass().getName() + " is not authorized");
                    }
                }
            }
            catch (SQLException | MigrationException e)
            {
                getLogger().error("The tables requires by the " +  this.getClass().getName() + " could not be created. A degraded behavior will occur", e);
            }
            
            _lazyInitialized = true;
        }
        
        return connection;
    }
    
    @Override
    public void logout(Redirector redirector)
    {
        // let's remove cookie from database
        String token = _getCookieValue();
        if (_authenticationTokenManager != null && StringUtils.isNotEmpty(token))
        {
            _authenticationTokenManager.deleteTokenByValue(token, null);
        }
        
        _deleteCookie();
    }

    
    @Override
    public boolean nonBlockingIsStillConnected(UserIdentity userIdentity, Redirector redirector)
    {
        return true;
    }
    
    @Override
    public boolean blockingIsStillConnected(UserIdentity userIdentity, Redirector redirector)
    {
        return true;
    }

    @Override
    public boolean blockingGrantAnonymousRequest()
    {
        Request request = ContextHelper.getRequest(_context);
        
        boolean accept = false;
        
        String login = request.getParameter(_usernameField);
        String password = request.getParameter(_passwordField);
        
        // URL without server context and leading slash.
        String url = (String) request.getAttribute(WorkspaceMatcher.IN_WORKSPACE_URL);
        
        // Accept the other urls only if the user didn't provide credentials
        // (if credentials are provided, the user is trying to connect). 
        if (login == null || password == null)
        {
            if (!accept)
            {
                for (String prefix : _acceptedUrlPrefixes)
                {
                    if (url.startsWith(prefix))
                    {
                        accept = true;
                        break;
                    }
                }
            }
            
            if (!accept)
            {
                for (Pattern pattern : CaptchaHelper.getUsedUrlPatterns())
                {
                    if (pattern.matcher(url).matches())
                    {
                        accept = true;
                        break;
                    }
                }
            }
        }
        
        if (accept && getLogger().isInfoEnabled())
        {
            getLogger().info("URL accepted : " + url);
        }
        
        return accept;
    }
    
    @Override
    public boolean nonBlockingGrantAnonymousRequest()
    {
        return blockingGrantAnonymousRequest();
    }
    
    private void _sendMFACodeByEmail(Request request, UserIdentity userIdentity)
    {
        MultifactorAuthenticationCode mfaCode = _multifactorAuthenticationManager.generateMultifactorAuthenticationCode(userIdentity);
        _multifactorAuthenticationManager.sendMultifactorAuthenticationCodeByMail(request, userIdentity, mfaCode.code());
    }

    @Override
    public UserIdentity blockingGetUserIdentity(Redirector redirector) throws Exception
    {
        Request request = ContextHelper.getRequest(_context);

        try
        {
            UserIdentity userIdentity = getUserIdentityFromRequest(request);
            if (userIdentity != null)
            {
                // Remove multifactor authentication information from session (may be useless - it just won't do anything)
                Session session = request.getSession(false);
                if (session != null)
                {
                    session.removeAttribute(__SESSION_CONNECTING_USERIDENTITY);
                }
            }
            else
            {
                Session session = request.getSession(false);
                if (session != null)
                {
                    userIdentity = (UserIdentity) session.getAttribute(__SESSION_CONNECTING_USERIDENTITY);
                }
            }
            
            if (userIdentity != null)
            {
                // User login/password is correct
                // Let's see MFA
                
                // Potential mfaCode
                String mfaCode = request.getParameter(_mfaCodeField);
                if (mfaCode != null && _multifactorAuthenticationManager.isValidMultifactorAuthenticationCode(userIdentity, mfaCode))
                {
                    // The entered code is valid
                    return userIdentity;
                }
                
                User user = _userManager.getUser(userIdentity);
                if (useMultifactorAuthentication() && StringUtils.isNotEmpty(user.getEmail()))
                {
                    // Put multifactor authentication information in session
                    Session session = request.getSession(true);
                    session.setAttribute(__SESSION_CONNECTING_USERIDENTITY, userIdentity);

                    MFAMode mode; 
                    if (!_multifactorAuthenticationManager.isAuthenticationApplicationActivated(userIdentity)
                        || MFAMode.EMAIL.toString().equals(request.getParameter(_mfaMode)))
                    {
                        mode = MFAMode.EMAIL;
                        if (mfaCode == null)
                        {
                            // Generate and send authentication code to user by email, except if a wrong code has been entered
                            _sendMFACodeByEmail(request, userIdentity);
                        }
                    }
                    else
                    {
                        mode = MFAMode.APPLICATION;
                    }
                    
                    // Redirect to login url with multifactor authentication parameters
                    String url = _getLoginURLWithMFACodeInformation(request, userIdentity, mode, mfaCode != null);
                    redirector.redirect(false, url);
                    return null;
                }
                
                // Multifactor authentication is not activated, or the user do not have email address, or a config parameter is missing
                // So the user does not need to enter a passcode
                return userIdentity;
            }
            else
            {
                // The general login screen will display the form for us
                redirector.redirect(false, (String) request.getAttribute(AuthenticateAction.REQUEST_ATTRIBUTE_LOGIN_URL));
                return null;
            }
        }
        catch (AccessDeniedException | NotUniqueUserException e)
        {
            blockingUserNotAllowed(redirector);
            return null;
        }
    }
    
    /**
     * Retrieves the login URL with parameters for multifactor authentication
     * @param request the request
     * @param userIdentity the identity of the connecting user
     * @param mode The mode to user to connect
     * @param isMFACodeInvalid <code>true</code> if the multifactor authentication is invalid, <code>false</code> otherwise
     * @return the login URL with parameters for multifactor authentication
     */
    protected String _getLoginURLWithMFACodeInformation(Request request, UserIdentity userIdentity, MFAMode mode, boolean isMFACodeInvalid)
    {
        String url = (String) request.getAttribute(AuthenticateAction.REQUEST_ATTRIBUTE_LOGIN_URL);

        StringBuilder parameters = new StringBuilder();

        parameters.append(url.indexOf('?') >= 0 ? "&" : "?");
        parameters.append(LOGIN_URL_USER_PARAMETER_NAME).append("=").append(UserIdentity.userIdentityToString(userIdentity));
        parameters.append("&").append(LOGIN_URL_MFA_MODE).append("=").append(mode.toString());
        parameters.append("&").append(LOGIN_URL_INVALID_PASSCODE_PARAMETER_NAME).append("=").append(isMFACodeInvalid);
        
        return url + parameters;
    }
    
    @Override
    public UserIdentity nonBlockingGetUserIdentity(Redirector redirector) throws Exception
    {
        Request request = ContextHelper.getRequest(_context);
        
        try
        {
            UserIdentity userIdentity = getUserIdentityFromRequest(request);
            if (userIdentity != null)
            {
                return userIdentity;
            }
    
            if (_allowCookies)
            {
                String token = _getCookieValue();
                if (_authenticationTokenManager != null && StringUtils.isNotEmpty(token))
                {
                    UserIdentity userFromToken = _authenticationTokenManager.validateToken(token);
                    if (userFromToken != null)
                    {
                        return userFromToken;
                    }
                    _deleteCookie();
                }
            }
        }
        catch (AccessDeniedException | NotUniqueUserException e)
        {
            nonBlockingUserNotAllowed(redirector);
            return null;
        }

        return null;
    }
    
    /**
     * Get user identity from request
     * @param request the request
     * @return the user identity
     * @throws AccessDeniedException If the user is rejected
     * @throws NotUniqueUserException If many users match this email
     * @throws WeakPasswordException If the user is authenticated with a weak password
     */
    protected UserIdentity getUserIdentityFromRequest(Request request) throws AccessDeniedException, NotUniqueUserException, WeakPasswordException
    {
        String login = request.getParameter(_usernameField);
        String password = request.getParameter(_passwordField);

        if (StringUtils.isNotBlank(login) && password != null)
        {
            UserPopulation userPopulation = _getPopulation(request);
            
            Integer nbConnect = requestNbConnectBDD(login, userPopulation.getId());
            
            if (_useCaptchaOnFailure)
            {
                if (nbConnect >= NB_CONNECTION_ATTEMPTS)
                {
                    String answer = request.getParameter(_captchaField);
                    String captchaKey = request.getParameter(_captchaKeyField);

                    if (!CaptchaHelper.checkAndInvalidate(captchaKey, answer)) 
                    {
                        // Captcha is invalid
                        _logLoginFailed(login, userPopulation, "invalid-captcha");
                        throw new AccessDeniedException("Captcha is invalid for user '" + login + "'");
                    }
                }
            }
            
            // Let's check password
            for (UserDirectory userDirectory : userPopulation.getUserDirectories())
            {
                User user = userDirectory.getUser(login);
                if (user == null && _allowLoginByEmail)
                {
                    // Try to get user by email
                    user = userDirectory.getUserByEmail(login);
                }
                
                if (user != null)
                {
                    CredentialsResult credentialsResult = userDirectory.checkCredentials(user.getIdentity().getLogin(), password);
                    switch (credentialsResult)
                    {
                        case NOT_AUTHENTICATED:
                            if (nbConnect >= NB_CONNECTION_ATTEMPTS)
                            {
                                _logLoginFailed(login, userPopulation, "invalid-credentials");
                            }
                            throw new AccessDeniedException("Password is incorrect for user '" + user.getIdentity().getLogin() + "'");
                        case WEAK_PASSWORD:
                            throw new WeakPasswordException(user.getIdentity());
                        default:
                            return user.getIdentity();
                    }
                }
                
            }
            
            // User does not exists in any directory
            _logLoginFailed(login, userPopulation, "unknown-user");

            throw new AccessDeniedException("Unknown user '" + login + "'");
        }
        
        return null;
    }
    
    private void _logLoginFailed(String login, UserPopulation userPopulation, String errorCause)
    {
        Map<String, Object> args = Map.of("login", login, "population", userPopulation.getId(), "error", errorCause);
        ForensicLogger.info("authentication.form.failed", args, null);
    }

    /**
     * Get user population from request
     * @param request the request
     * @return the user population
     */
    protected UserPopulation _getPopulation(Request request)
    {
        // With forms, the population should always be known!
        @SuppressWarnings("unchecked")
        List<UserPopulation> userPopulations = (List<UserPopulation>) request.getAttribute(AuthenticateAction.REQUEST_ATTRIBUTE_AVAILABLE_USER_POPULATIONS_LIST);

        // If the list has one element only...
        if (userPopulations.size() == 1)
        {
            return userPopulations.get(0);
        }
        
        // In this list a population was maybe chosen?
        final String chosenUserPopulationId = (String) request.getAttribute(AuthenticateAction.REQUEST_ATTRIBUTE_USER_POPULATION_ID);
        if (StringUtils.isNotBlank(chosenUserPopulationId))
        {
            UserPopulation chosenUserPopulation = userPopulations.stream().filter(userPopulation -> StringUtils.equals(userPopulation.getId(), chosenUserPopulationId)).findFirst().get();
            return chosenUserPopulation;
        }
        
        // Can not work here...
        throw new IllegalStateException("The " + this.getClass().getName() + " does not work when population is not known");
    }

    @Override
    public void blockingUserNotAllowed(Redirector redirector) throws Exception
    {
        Request request = ContextHelper.getRequest(_context);

        String url = (String) request.getAttribute(AuthenticateAction.REQUEST_ATTRIBUTE_LOGIN_URL);
        StringBuilder parameters = new StringBuilder();
        parameters.append(url.indexOf('?') >= 0 ? "&" : "?");
        parameters.append("login=");
        parameters.append(request.getParameter(_usernameField));
        parameters.append("&authFailure=true");

        int nbConnect = _setNbConnectBDD(request.getParameter(_usernameField), _getPopulation(request).getId());
        if (_useCaptchaOnFailure)
        {
            String captchaKey = request.getParameter(_captchaKeyField);
            int nbAttempts = NB_CONNECTION_ATTEMPTS - 1;
            
            if (nbConnect == nbAttempts || captchaKey == null && nbConnect > nbAttempts)
            {
                parameters.append("&tooManyAttempts=" + true);
            }
        }
        
        if (StringUtils.isNotEmpty(_getCookieValue()))
        {
            parameters.append("&cookieFailure=" + true);
            _deleteCookie(); 
        }
        
        // The general login screen will display the form for us
        redirector.redirect(false, url + parameters);
    }
    
    @Override
    public void nonBlockingUserNotAllowed(Redirector redirector)
    {
        // Do nothing
    }

    @Override
    public void blockingUserAllowed(UserIdentity userConnected, Redirector redirector)
    {
        if (_allowCookies)
        {
            Request request = ContextHelper.getRequest(_context);
            if ("true".equals(request.getParameter(_rememberMeField)))
            {
                nonBlockingUserAllowed(userConnected, redirector);
            }
        }
        
        if (_storePasswordInSession)
        {
            Request request = ContextHelper.getRequest(_context);
            request.getSession(true).setAttribute(PASSWORD_SESSION_ATTRIBUTE, request.getParameter(_passwordField));
        }
        
        _deleteLoginFailedBDD(userConnected.getLogin(), userConnected.getPopulationId());
    }
    
    @Override
    public void nonBlockingUserAllowed(UserIdentity userConnected, Redirector redirector)
    {
        if (_cookieEnabled && _authenticationTokenManager != null)
        {
            String token = _authenticationTokenManager.generateToken(userConnected, _cookieLifetime, 1, "FormsCredentialProvider-Cookie", null);
            _updateCookie(token);
        }
    }

    @Override
    public boolean requiresNewWindow()
    {
        return false;
    }

    /*************************************************************************************************************
     * Connection failure management
     */
    
    /**
     * Delete all past failed connections
     */
    protected void _deleteAllPastLoginFailedBDD()
    {
        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try
        {
            connection = getSQLConnection();
            
            // Build request for authenticating the user
            String sql = "DELETE FROM Users_FormConnectionFailed WHERE last_connect < ?";
            
            ZonedDateTime dateToday = ZonedDateTime.now();
            ZonedDateTime thresholdDate = dateToday.minusDays(TIME_ALLOWED);
            Timestamp threshold = Timestamp.from(thresholdDate.toInstant());
            
            stmt = connection.prepareStatement(sql);
            stmt.setTimestamp(1, threshold);
            
            // Do the request
            stmt.execute();
        }
        catch (SQLException e)
        {
            getLogger().error("Error during the connection to the database", e);
        }
        finally
        {
            // Close connections
            ConnectionHelper.cleanup(rs);
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(connection);
        }
    }
    
    /**
     * Get the number of failed connections with this login
     * @param login The login to request
     * @param populationId The user's population
     * @return the number of connection failed
     */
    public Integer requestNbConnectBDD(String login, String populationId)
    {
        _deleteAllPastLoginFailedBDD();
        
        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try
        {
            connection = getSQLConnection();
            
            // Build request for authenticating the user
            String sql = "SELECT nb_connect FROM Users_FormConnectionFailed WHERE LOWER(login) = LOWER(?) and population_id = ?";
            
            stmt = connection.prepareStatement(sql);
            stmt.setString(1, login);
            stmt.setString(2, populationId);
    
            // Do the request
            rs = stmt.executeQuery();
    
            Integer nbConnect = 0;
            if (rs.next()) 
            {
                nbConnect = rs.getInt("nb_connect"); 
            }
            return nbConnect;
        }
        catch (SQLException e)
        {
            getLogger().error("Error during the connection to the database", e);
            return 0;
        }
        finally
        {
            // Fermer les ressources de connexion
            ConnectionHelper.cleanup(rs);
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(connection);
        }
    }
    
    /**
     * Get the number of failed connections with this login
     * @param login The login to set
     * @param populationId The population id of the user
     * @return the number of failed connection
     */
    protected Integer _setNbConnectBDD(String login, String populationId)
    {
        Integer nbConnect = requestNbConnectBDD(login, populationId);
        if (nbConnect == 0)
        {
            _insertLoginNbConnectBDD(login, populationId);
        }
        else
        {
            _updateLoginNbConnectBDD(login, populationId, nbConnect);
        }
        
        return nbConnect;
        
    }
    
    /**
     * Insert the login with one failed connection in the BDD
     * @param login The login to insert
     * @param populationId The population id
     */
    protected void _insertLoginNbConnectBDD(String login, String populationId)
    {
        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try
        {
            connection = getSQLConnection();
            
            String sqlUpdate = "INSERT INTO Users_FormConnectionFailed (login, population_id, nb_connect, last_connect) VALUES (?, ?, ?, ?)";
           
            stmt = connection.prepareStatement(sqlUpdate);
            stmt.setString(1, login);
            stmt.setString(2, populationId);
            stmt.setInt(3, 1);
            
            Timestamp date = Timestamp.from(Instant.now());
            
            stmt.setTimestamp(4, date);
            
            stmt.execute();
        }
        catch (SQLException e)
        {
            getLogger().error("Error during the connection to the database", e);
        }
        finally
        {
            // Close connections
            ConnectionHelper.cleanup(rs);
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(connection);
        }
    }
    
    /**
     * Delete the login from the table of the failed connection
     * @param login The login to remove
     * @param populationId The populationId of the user
     */
    protected void _deleteLoginFailedBDD(String login, String populationId)
    {
        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try
        {
            connection = getSQLConnection();
            
            // Build request for authenticating the user
            String sql = "DELETE FROM Users_FormConnectionFailed WHERE LOWER(login) = LOWER(?) and population_id = ?";
            
            stmt = connection.prepareStatement(sql);
            stmt.setString(1, login);
            stmt.setString(2, populationId);
            
            // Do the request
            stmt.execute();
        }
        catch (SQLException e)
        {
            getLogger().error("Error during the connection to the database", e);
        }
        finally
        {
            // Close connections
            ConnectionHelper.cleanup(rs);
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(connection);
        }
    }
    
    /**
     * Update the number of failed connections of the login in the BDD
     * @param login The login to update
     * @param populationId The user's population
     * @param nbConnect The nb of connection to set
     */
    protected void _updateLoginNbConnectBDD(String login, String populationId, Integer nbConnect)
    {
        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try
        {
            connection = getSQLConnection();
            
            String sqlUpdate = "UPDATE Users_FormConnectionFailed SET nb_connect = ? WHERE LOWER(login) = LOWER(?) and population_id = ?";
           
            stmt = connection.prepareStatement(sqlUpdate);
            stmt.setInt(1, nbConnect + 1);
            stmt.setString(2, login);
            stmt.setString(3, populationId);
            
            stmt.execute();
        }
        catch (SQLException e)
        {
            getLogger().error("Error during the connection to the database", e);
        }
        finally
        {
            // Close connections
            ConnectionHelper.cleanup(rs);
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(connection);
        }
    }
    
    /************************************************************************************************
     *    COOKIE EXCHANGES MANAGEMENT
     */
     
     /**
      * Return the cookie value corresponding to the searched name
      * @return the value of the cookie or null if not
      */
    protected String _getCookieValue()
    {
        Cookie[] cookies = ContextHelper.getRequest(_context).getCookies();

        if (cookies != null)
        {
            for (Cookie cookie : cookies)
            {
                if (_cookieName.equals(cookie.getName()))
                {
                    return cookie.getValue();
                }
            }
        }

        return null;
    }

     /**
      * Checks if cookie already exists
      * @return boolean
      */
    protected boolean _isCookieAlreadySet()
    {
        Cookie[] cookies = ContextHelper.getRequest(_context).getCookies();

        if (cookies != null)
        {
            for (Cookie cookie : cookies)
            {
                if (_cookieName.equals(cookie.getName()))
                {
                    return true;
                }
            }
        }

        return false;
    }
     
     /**
      * Update the cookie for client-side purpose
      * @param value the cookie value
      */
    protected void _updateCookie(String value)
    {
        Request request = ContextHelper.getRequest(_context);
        Response response = ContextHelper.getResponse(_context);
        
        Cookie cookie = response.createCookie(_cookieName, value);
        cookie.setSecure(request.isSecure());
        cookie.setPath(ContextHelper.getRequest(_context).getContextPath());
        cookie.setMaxAge((int) _cookieLifetime);
        ((HttpCookie) cookie).getServletCookie().setHttpOnly(true);
        ContextHelper.getResponse(_context).addCookie(cookie);
    }
     
     /**
      * Delete the cookie
      */
    protected void _deleteCookie()
    {
        Cookie cookie = new HttpCookie(_cookieName, "");
        cookie.setPath(ContextHelper.getRequest(_context).getContextPath());
        cookie.setMaxAge(0);
        ContextHelper.getResponse(_context).addCookie(cookie);
    }
    
    /**
     * Checks if multifactor authentication is activated for this credential provider 
     * @return <code>true</code> if multifactor authentication is activated, <code>false</code> otherwise
     */
    public boolean useMultifactorAuthentication()
    {
        return _useMultifactorAuthentication && !PluginsManager.getInstance().isSafeMode();
    }
}
