/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.cache;

import java.lang.reflect.Field;
import java.time.Duration;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.ametys.core.cache.Cache;
import org.ametys.core.cache.CacheException;
import org.ametys.core.cache.CacheStats;
import org.ametys.core.util.SizeUtils;
import org.ametys.runtime.i18n.I18nizableText;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.Weigher;
import com.google.common.util.concurrent.ExecutionError;
import com.google.common.util.concurrent.UncheckedExecutionException;

/**
 * Implementation of AmetysCache with Guava library
 *
 * @param <K> the key
 * @param <V> the value
 */
public class GuavaCache<K, V> implements Cache<K, V>
{

    /** the cache itself, using Guava implementation */
    protected com.google.common.cache.Cache<K, Optional<V>> _cache;

    /** Id of the cache */
    protected String _id;
    
    /** Label of the cache */
    protected I18nizableText _label;

    /** Description of the cache */
    protected I18nizableText _description;

    /** Maximum size of the cache in bytes */
    protected long _size;
    
    /** True if the size of the cache can be computed, false otherwise (e.g., the cache store a component) */
    protected boolean _computableSize;

    /** The length of time after an entry is created that it should be automatically*/
    protected Duration _duration;
    
    /** True if the put or putAll have been called (even without values), false otherwises. become false again when the cache is reseted or if all values have been cleaned.  */
    protected boolean _isInitialized;
    
    /** True if the cache can be transmitted in sub-requests of DispatchGenerator, false otherwises. */
    protected boolean _isDispatchable;
    
   
    /**
     * Implementation of AmetysCache with Guava library
     * @param id id of the cache
     * @param label label of the cache
     * @param description description
     * @param size maximum size of the cache
     * @param computableSize true if the size of the cache can be computed
     * @param duration the length of time after an entry is created that it should be automatically removed
     * @param isDispatchable true if the cache can be transmitted in sub-requests of DispatchGenerator
     */
    public GuavaCache(String id, I18nizableText label, I18nizableText description, long size, boolean computableSize, Duration duration, boolean isDispatchable)
    {
        _label = label;
        _description = description;
        _id = id;
        _size = size;
        _computableSize = computableSize;
        _duration = duration;
        _isInitialized = false;
        _isDispatchable = isDispatchable;
        resetCache();
    }

    public void resetCache()
    {
        CacheBuilder cacheBuilder = com.google.common.cache.CacheBuilder.newBuilder();
        
        if (_duration != null)
        {
            cacheBuilder = cacheBuilder.expireAfterWrite(_duration);
        }
        _cache = cacheBuilder.recordStats()
                .maximumWeight(_size)
                .weigher(new Weigher<K, Optional<V>>()
                {
                    public int weigh(K k, Optional<V> v)
                    {
                        if (_computableSize)
                        {
                            return (int) (SizeUtils.sizeOf(k) + SizeUtils.sizeOf(v.orElse(null)));
                        }
                        else
                        {
                            return 0;
                        }
                    }
                }).build();
        _isInitialized = false;
    }

    public V get(K key, Function<K, V> function) throws CacheException
    {
        if (key instanceof AbstractCacheKey)
        {
            if (((AbstractCacheKey) key).isPartialKey())
            {
                throw new CacheException("Complex key is not valid because it contains null values");
            }
        }
        try
        {
            return _cache.get(key, new Callable<Optional<V>>()
            {
                @Override
                public Optional<V> call()
                {
                    return Optional.ofNullable(function.apply(key));
                }
            }).orElse(null);
        }
        catch (ExecutionException | ExecutionError | UncheckedExecutionException e)
        {
            // assume that the real exception is always e.getCause() as LocalCache wrap the exception into a ExecutionException, ExecutionError or UncheckedExecutionException
            throw new CacheException("An error occurred while computing the new value for key " + key, e.getCause());
        }
    }

    public V get(K key)
    {
        Optional<V> optional = _cache.getIfPresent(key);
        if (optional != null)
        {
            return optional.orElse(null);
        }
        else
        {
            return null;
        }
    }

    public void put(K key, V value)
    {
        if (key instanceof AbstractCacheKey)
        {
            if (((AbstractCacheKey) key).isPartialKey())
            {
                throw new RuntimeException("complex key is not valid because it contains null values");
            }
        }
        _cache.put(key, Optional.ofNullable(value));
        _isInitialized = true;
    }

    public void putAll(Map<K, V> map)
    {
        map.forEach((k, v) -> 
        {
            _cache.put(k, Optional.ofNullable(v));
        }); 
        _isInitialized = true;
    }

    public CacheStats getCacheStats()
    {
        return new GuavaCacheStats(_cache.stats());
    }

    public long getMemorySize() throws CacheException
    {
        if (!_computableSize)
        {
            return -1;
        }
        try
        {
            long sum = 0;
            Field localCacheField = _cache.getClass().getDeclaredField("localCache"); // NoSuchFieldException
            localCacheField.setAccessible(true);
            var localCache = localCacheField.get(_cache); // IllegalAccessException

            Field segmentsField = localCache.getClass().getDeclaredField("segments"); // NoSuchFieldException
            segmentsField.setAccessible(true);
            final Object[] segments = (Object[]) segmentsField.get(localCache); // IllegalAccessException

            for (int i = 0; i < segments.length; i++)
            {
                Field totalWeight = segments[i].getClass().getDeclaredField("totalWeight"); // NoSuchFieldException
                totalWeight.setAccessible(true);
                long totalWeightValue = (long) totalWeight.get(segments[i]); // IllegalAccessException
                sum += totalWeightValue;
            }

            return sum;
            
        }
        catch (Exception e)
        {
            throw new CacheException("Cannot compute object size", e);
        }
    }

    public long getNumberOfElements()
    {
        return _cache.size();
    }

    public void invalidate(K key)
    {
        List<K> keyToRemove = new ArrayList<>();

        if (key instanceof AbstractCacheKey)
        {
            AbstractCacheKey invalidationKey = (AbstractCacheKey) key;
            if (invalidationKey.isPartialKey())
            {
                // Remove null element
                List<Object> keyList = invalidationKey.getFields();

                keyToRemove = _cache.asMap().keySet().stream().filter(k ->
                {
                    AbstractCacheKey abstractK = (AbstractCacheKey) k;
                    boolean isEquals = true;
                    
                    for (int i = 0; i < keyList.size(); i++)
                    {
                        isEquals = isEquals && (keyList.get(i) == null || keyList.get(i).equals(abstractK.getFields().get(i)));
                    }
                    
                    return isEquals;
                    
                }).collect(Collectors.toList());
            }
            else
            {
                keyToRemove.add(key);
            }
        }
        else
        {
            keyToRemove.add(key);
        }
        _cache.invalidateAll(keyToRemove);

    }

    public void invalidateAll()
    {
        _cache.invalidateAll();
        _isInitialized = false;
    }

    public I18nizableText getDescription()
    {
        return _description;
    }

    public I18nizableText getLabel()
    {
        return _label;
    }

    public String getId()
    {
        return _id;
    }

    public long getMaxSize()
    {
        return _size;
    }

    public boolean hasKey(K key)
    {
        if (key instanceof AbstractCacheKey)
        {
            AbstractCacheKey invalidationKey = (AbstractCacheKey) key;
            if (invalidationKey.isPartialKey())
            {
                // we ignore null element
                List<Object> keyList = invalidationKey.getFields();
                
                return _cache.asMap().keySet().stream()
                    .anyMatch(k -> {
                        boolean hasKey = true;
                        AbstractCacheKey abstractKey = (AbstractCacheKey) k;
                        List<Object> fields = abstractKey.getFields();
                        int i = 0;
                        while (hasKey && i < keyList.size())
                        {
                            hasKey = keyList.get(i) == null || keyList.get(i).equals(fields.get(i));
                            i++;
                        }
                        return hasKey;
                    });
            }
            else
            {
                return _cache.asMap().containsKey(key);
            }
        }
        else
        {
            return _cache.asMap().containsKey(key);
        }
    }
    
    public Map<K, V> asMap()
    {
        Map<K, V> map = new HashMap<>();
        
        Map<K, Optional<V>> maps = _cache.asMap();
        maps.forEach((k, v) -> 
        {
            map.put(k, v.orElse(null));
        });
        return map;
    }
    
    public boolean isComputableSize()
    {
        return _computableSize;
    }

    public boolean isInitialized()
    {
        return _isInitialized;
    }
    
    public boolean isDispatchable()
    {
        return _isDispatchable;
    }

    public Map<K, V> getAll(AbstractCacheKey filterKey)
    {
        return this.asMap()
                .entrySet()
                .stream()
                .filter(entry -> _checkKey(entry.getKey(), filterKey.getFields()))
                .collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));
    }
    
    private boolean _checkKey(K key, List<Object> fields)
    {
        if (key instanceof AbstractCacheKey)
        {
            if (((AbstractCacheKey) key).getFields().size() != fields.size())
            {
                throw new RuntimeException("Complex key lengths does not match");
            }
        }
        else
        {
            throw new RuntimeException("Key is not a complex key");
        }
        
        for (int i = 0; i < fields.size(); i++) 
        {
            if (fields.get(i) != null && !fields.get(i).equals(((AbstractCacheKey) key).getFields().get(i)))
            {
                return false;
            }
        }
        
        return true;
    }
}
