/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.captcha.jcaptcha;

import java.awt.Color;
import java.awt.Font;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;

import org.ametys.core.captcha.AbstractCaptcha;
import org.ametys.runtime.i18n.I18nizableText;

import nl.captcha.Captcha;
import nl.captcha.Captcha.Builder;
import nl.captcha.gimpy.DropShadowGimpyRenderer;
import nl.captcha.gimpy.FishEyeGimpyRenderer;
import nl.captcha.gimpy.RippleGimpyRenderer;
import nl.captcha.noise.CurvedLineNoiseProducer;
import nl.captcha.text.producer.DefaultTextProducer;
import nl.captcha.text.renderer.DefaultWordRenderer;

/**
 * Captcha implementation with images
 */
public class JCaptcha extends AbstractCaptcha
{
    /**
     * Prefix for static captchas
     */
    protected static final String STATIC_PREFIX_KEY = "STATIC-";

    /**
     * Prefix for dynamic captchas
     */
    protected static final String DYNAMIC_PREFIX_KEY = "DYNAMIC-";
    
    /**
     * Map of static captchas
     */
    protected static Map<String, List<ValidableCaptcha>> _mapStaticCaptcha = new HashMap<>();
    
    /**
     * Map of dynamic captchas
     */
    protected static Map<String, ValidableCaptcha> _mapDynamicCaptcha = new HashMap<>();
    
    @Override
    public boolean checkAndInvalidateCaptcha(String key, String value)
    {
        if (StringUtils.isEmpty(key) || StringUtils.isEmpty(value))
        {
            return false;
        }
        if (key.startsWith(STATIC_PREFIX_KEY))
        {
            List<ValidableCaptcha> list = _mapStaticCaptcha.get(key);
            if (list != null)
            {
                for (ValidableCaptcha c : list)
                {
                    if (c.isValid())
                    {
                        Captcha captcha = c.getCaptcha();
                        if (captcha.isCorrect(value))
                        {
                            c.invalidate();
                            return true;
                        }
                    }
                }
            }
            
            return false;
        }
        else if (key.startsWith(DYNAMIC_PREFIX_KEY))
        {
            ValidableCaptcha vc = _mapDynamicCaptcha.get(key);
            if (vc == null)
            {
                return false;
            }
            else if (!vc.isValid())
            {
                _mapDynamicCaptcha.remove(key);
                return false;
            }
            else
            {
                _mapDynamicCaptcha.remove(key);

                Captcha c = vc.getCaptcha();
                return c.isCorrect(value);
            }
        }
        else
        {
            throw new IllegalArgumentException("The key '" + key + "' is not a valid captcha key because it does not starts with '" + DYNAMIC_PREFIX_KEY + "' or '" + STATIC_PREFIX_KEY + "'");
        }
    }

    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.core-impl", "PLUGINS_CORE_CAPTCHA_CONFIG_TYPE_JCAPTCHA");
    }

    @Override
    public I18nizableText getLoginFailedBecauseCaptchaFailedLabel()
    {
        return new I18nizableText("plugin.core-impl", "PLUGINS_CORE_UI_LOGIN_SCREEN_FORM_FAILED_WITH_CAPTCHA_jcaptcha");
    }

    @Override
    public I18nizableText getLoginFailedBecauseTooManyAttemptLabel()
    {
        return new I18nizableText("plugin.core-impl", "PLUGINS_CORE_UI_LOGIN_SCREEN_FORM_FAILED_MANY_TIME_jcaptcha");
    }
    
    @Override
    public List<Pattern> getUsedUrlPatterns()
    {
        return Arrays.asList(new Pattern[]{Pattern.compile("^plugins/core-impl/captcha/core-impl/jcaptcha/[^/]+/image.png")});
    }


    /**
     * Generate an image captcha to PNG format. The key has to be unique, if you cannot generate a key use the other form of the method.
     * @param key the wanted key. Can not be null. You can use RandomStringUtils.randomAlphanumeric(10) to generates one
     * @param color The color for font
     * @param addNoise true to add noise to captcha image
     * @param fisheye true to add fish eye background to captcha image
     * @param width The image width
     * @param height The image height
     * @return The corresponding image
     */
    public synchronized BufferedImage generateImageCaptcha (String key, Integer color, boolean addNoise, boolean fisheye, int width, int height)
    {
        Captcha captcha = _generateImageCaptcha(color, addNoise, fisheye, width, height);
        ValidableCaptcha vc = new ValidableCaptcha(captcha);
        
        if (key.startsWith(STATIC_PREFIX_KEY))
        {
            if (!_mapStaticCaptcha.containsKey(key))
            {   
                _mapStaticCaptcha.put(key, new ArrayList<>());
            }
            List<ValidableCaptcha> captchas = _mapStaticCaptcha.get(key);
            captchas.add(new ValidableCaptcha(captcha));
        }
        else if (key.startsWith(DYNAMIC_PREFIX_KEY))
        {
            // If there were a key there, we scrach it! (this is a way to invalidate it - this may happen when the user do 'back' in its browser
            _mapDynamicCaptcha.put(key, vc);
        }
        else
        {
            throw new IllegalArgumentException("The key '" + key + "' is not a valid captcha key because it does not starts with '" + DYNAMIC_PREFIX_KEY + "' or '" + STATIC_PREFIX_KEY + "'");
        }
        
        return vc.getCaptcha().getImage();
    }
    
    private Captcha _generateImageCaptcha(Integer color, boolean addNoise, boolean fisheye, int width, int height)
    {
        List<Color> colors = new ArrayList<>();
        colors.add(new Color(color));
        
        List<Font> fonts = new ArrayList<>();
        fonts.add(new Font("Arial", Font.BOLD, 40));
        fonts.add(new Font("Courier", Font.BOLD, 40));
        
        Builder builder = new Captcha.Builder(width, height)
            .addText(new DefaultTextProducer(6, "abcdefghijklmnopqrstuvwxyz".toCharArray()), new DefaultWordRenderer(colors, fonts))
            .gimp(new RippleGimpyRenderer())
            .addNoise(new CurvedLineNoiseProducer(new Color(color), 3));
        
        if (addNoise)
        {
            builder.addNoise(new CurvedLineNoiseProducer(new Color(color), 3))
                .gimp(new DropShadowGimpyRenderer());
        }
        
        if (fisheye)
        {
            builder.gimp(new FishEyeGimpyRenderer());
        }
        
        return builder.build();
    }

    /**
     * Remove a captcha
     * @param key the key value
     */
    public synchronized void removeCaptcha (String key)
    {
        if (key.startsWith(STATIC_PREFIX_KEY))
        {
            _mapStaticCaptcha.remove(key);
        }
        else if (key.startsWith(DYNAMIC_PREFIX_KEY))
        {
            _mapDynamicCaptcha.remove(key);
        }
    }
    
    /**
     * Clean the outdated captchas
     */
    public synchronized void cleanOldCaptchas()
    {
        _cleanOldStaticCaptchas();
        _cleanOldDynamicCaptchas();
    }
    
    private synchronized void _cleanOldDynamicCaptchas()
    {
        Iterator<String> cIt = _mapDynamicCaptcha.keySet().iterator();
        while (cIt.hasNext())
        {
            String id = cIt.next();
            ValidableCaptcha vc = _mapDynamicCaptcha.get(id);
            if (!vc.isValid())
            {
                cIt.remove();
            }
        }
    }
    
    private synchronized void _cleanOldStaticCaptchas()
    {
        Iterator<String> cIt = _mapStaticCaptcha.keySet().iterator();
        while (cIt.hasNext())
        {
            String id = cIt.next();
            List<ValidableCaptcha> c = _mapStaticCaptcha.get(id);
            
            Iterator<ValidableCaptcha> it = c.iterator();
            while (it.hasNext())
            {
                ValidableCaptcha vc  = it.next();
                if (!vc.isValid())
                {
                    it.remove();
                }
            }
            
            if (c.isEmpty())
            {
                cIt.remove();
            }
        }
    }

    
    /**
     * Bean for a captcha and a validity date 
     */
    protected class ValidableCaptcha
    {
        private Captcha _captcha;
        private Date _date;
        private boolean _valid;
        
        /**
         * Build the captcha wrapper
         * @param c the captcha to wrap
         */
        public ValidableCaptcha(Captcha c)
        {
            _captcha = c;
            _date = new Date();
            _valid = true;
        }

        /**
         * Determine if the validity date has not expired
         * @return true if the captcha is still valid
         */
        public boolean isValid()
        {
            if (!_valid)
            {
                return false;
            }
            
            Calendar validity = new GregorianCalendar();
            validity.add(Calendar.MINUTE, -20);
            
            return _date.after(validity.getTime()); 
        }
        
        /**
         * Get the wrapper captcha
         * @return captcha
         */
        public Captcha getCaptcha()
        {
            return _captcha;
        }
        
        /**
         * Mark the captcha as invalid 
         */
        public void invalidate()
        {
            _valid = false;
        }
    }
}
