/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.captcha.recaptcha;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.ContentType;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;

import org.ametys.core.captcha.AbstractCaptcha;
import org.ametys.core.util.JSONUtils;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Captcha implementation with reCaptcha Enterprise
 */
public class ReCaptchaEnterprise extends AbstractCaptcha implements Serviceable, Initializable
{
    private static final String __RECAPTCHA_API_KEY = "runtime.captcha.recaptcha.enterprise.apikey";

    private static final String __RECAPTCHA_SITE_KEY = "runtime.captcha.recaptcha.enterprise.sitekey";
    
    private static final String __RECAPTCHA_PROJECT_ID = "runtime.captcha.recaptcha.enterprise.projectId";
    
    private static final String __RECAPTCHA_THRESHOLD = "runtime.captcha.recaptcha.enterprise.threshold";
    
    private static final String __RECAPTCHA_URL = "https://recaptchaenterprise.googleapis.com/v1/projects/";
    
    private JSONUtils _jsonUtils;

    private String _projectId;
    private String _apiKey;
    private Object _siteKey;
    private Double _threshold;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
    }
    
    public void initialize() throws Exception
    {
        _initializeConfig();
    }

    private boolean _initializeConfig()
    {
        Config config = Config.getInstance();
        if (config != null)
        {
            _apiKey = config.getValue(__RECAPTCHA_API_KEY);
            _projectId = config.getValue(__RECAPTCHA_PROJECT_ID);
            _siteKey = config.getValue(__RECAPTCHA_SITE_KEY);
            _threshold = config.getValue(__RECAPTCHA_THRESHOLD);
            return true;
        }
        return false;
    }
    
    public boolean requireUserInteraction()
    {
        return false;
    }
    
    @Override
    public synchronized boolean checkAndInvalidateCaptcha(String key, String value)
    {
        if (StringUtils.isEmpty(value))
        {
            return false;
        }
        if (_apiKey != null || _initializeConfig())
        {
            RequestConfig requestConfig = RequestConfig.custom()
                    .setConnectTimeout(2000)
                    .setSocketTimeout(2000)
                    .build();
            
            try (CloseableHttpClient httpclient = HttpClientBuilder.create()
                                                                   .setDefaultRequestConfig(requestConfig)
                                                                   .useSystemProperties()
                                                                   .build())
            {
                HttpPost post = new HttpPost(__RECAPTCHA_URL + _projectId + "/assessments?key=" + _apiKey);
                
                Map<String, Object> eventParams = new HashMap<>(2);
                eventParams.put("token", value);
                eventParams.put("siteKey", _siteKey);
                
                post.setEntity(new StringEntity(_jsonUtils.convertObjectToJson(Map.of("event", eventParams)), ContentType.APPLICATION_JSON));
                
                try (CloseableHttpResponse httpResponse = httpclient.execute(post))
                {
                    if (httpResponse.getStatusLine().getStatusCode() != 200)
                    {
                        return false;
                    }
                    
                    ByteArrayOutputStream bos = new ByteArrayOutputStream();
                    try (InputStream is = httpResponse.getEntity().getContent())
                    {
                        IOUtils.copy(is, bos);
                    }
                    
                    Map<String, Object> jsonObject = _jsonUtils.convertJsonToMap(bos.toString());
                    
                    @SuppressWarnings("unchecked")
                    Map<String, Object> tokenProperties = (Map<String, Object>) jsonObject.get("tokenProperties");
                    if (tokenProperties != null && (Boolean) tokenProperties.getOrDefault("valid", false))
                    {
                        @SuppressWarnings("unchecked")
                        Map<String, Object> riskAnalysis = (Map<String, Object>) jsonObject.get("riskAnalysis");
                        Double score = (Double) riskAnalysis.get("score");
                        @SuppressWarnings("unchecked")
                        List<String> reasons = (List<String>) riskAnalysis.computeIfAbsent("reasons", str -> List.of("CLASSIFICATION_REASON_UNSPECIFIED"));
                        
                        if (score >= _threshold || reasons.contains("LOW_CONFIDENCE_SCORE"))
                        {
                            return true;
                        }
                    }
                    return false;
                }
            }
            catch (IOException e)
            {
                getLogger().error("Unable to contact Google server to validate reCaptcha.", e);
                return false;
            }
        }
        
        return false;
    }

    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.core-impl", "PLUGINS_CORE_CAPTCHA_CONFIG_TYPE_RECAPTCHA_ENTERPRISE");
    }
    
    @Override
    public I18nizableText getLoginFailedBecauseCaptchaFailedLabel()
    {
        return new I18nizableText("plugin.core-impl", "PLUGINS_CORE_UI_LOGIN_SCREEN_FORM_FAILED_WITH_CAPTCHA_RECAPTCHA_ENTERPRISE");
    }

    @Override
    public I18nizableText getLoginFailedBecauseTooManyAttemptLabel()
    {
        return new I18nizableText("plugin.core-ui", "PLUGINS_CORE_UI_LOGIN_SCREEN_FORM_AUTH_FAILURE");
    }

    @Override
    public List<Pattern> getUsedUrlPatterns()
    {
        return List.of();
    }
}
