/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.captcha.recaptcha;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.NameValuePair;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.message.BasicNameValuePair;

import org.ametys.core.captcha.AbstractCaptcha;
import org.ametys.core.util.JSONUtils;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Captcha implementation with reCaptcha v2
 */
public class ReCaptchaV2 extends AbstractCaptcha implements Serviceable
{
    private static String __RECAPTCHA_SECRET_KEY = "runtime.captcha.recaptcha.secretkey";
    
    private static String __RECAPTCHA_URL = "https://www.google.com/recaptcha/api/siteverify";
    
    private JSONUtils _jsonUtils;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
    }
    
    @Override
    public synchronized boolean checkAndInvalidateCaptcha(String key, String value)
    {
        if (StringUtils.isEmpty(value))
        {
            return false;
        }
        if (Config.getInstance() != null)
        {
            String secretKey = Config.getInstance().getValue(__RECAPTCHA_SECRET_KEY);
            
            RequestConfig requestConfig = RequestConfig.custom()
                    .setConnectTimeout(2000)
                    .setSocketTimeout(2000)
                    .build();
            
            try (CloseableHttpClient httpclient = HttpClientBuilder.create()
                                                                   .setDefaultRequestConfig(requestConfig)
                                                                   .useSystemProperties()
                                                                   .build())
            {
                // Prepare a request object
                HttpPost post = new HttpPost(__RECAPTCHA_URL);
                List<NameValuePair> params = new ArrayList<>();
                params.add(new BasicNameValuePair("secret", secretKey));
                params.add(new BasicNameValuePair("response", value));
                post.setEntity(new UrlEncodedFormEntity(params, "UTF-8"));
                
                try (CloseableHttpResponse httpResponse = httpclient.execute(post))
                {
                    if (httpResponse.getStatusLine().getStatusCode() != 200)
                    {
                        return false;
                    }
                    
                    ByteArrayOutputStream bos = new ByteArrayOutputStream();
                    try (InputStream is = httpResponse.getEntity().getContent())
                    {
                        IOUtils.copy(is, bos);
                    }
                    
                    Map<String, Object> jsonObject = _jsonUtils.convertJsonToMap(bos.toString());
                    
                    return (Boolean) jsonObject.getOrDefault("success", false);
                }
            }
            catch (IOException e)
            {
                getLogger().error("Unable to contact Google server to validate reCaptcha.", e);
                return false;
            }
        }
        
        return false;
    }

    @Override
    public I18nizableText getLabel()
    {
        return new I18nizableText("plugin.core-impl", "PLUGINS_CORE_CAPTCHA_CONFIG_TYPE_RECAPTCHA");
    }
    
    @Override
    public I18nizableText getLoginFailedBecauseCaptchaFailedLabel()
    {
        return new I18nizableText("plugin.core-impl", "PLUGINS_CORE_UI_LOGIN_SCREEN_FORM_FAILED_WITH_CAPTCHA_recaptcha");
    }

    @Override
    public I18nizableText getLoginFailedBecauseTooManyAttemptLabel()
    {
        return new I18nizableText("plugin.core-impl", "PLUGINS_CORE_UI_LOGIN_SCREEN_FORM_FAILED_MANY_TIME_recaptcha");
    }

    @Override
    public List<Pattern> getUsedUrlPatterns()
    {
        return List.of();
    }
}
