/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.checker;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.cocoon.util.StringUtils;
import org.apache.commons.lang3.ArrayUtils;

import org.ametys.core.util.HttpUrlUtils;
import org.ametys.runtime.model.checker.ItemChecker;
import org.ametys.runtime.model.checker.ItemCheckerTestFailureException;

/**
 * Checks if the url written is correct and if it allows to establish a connection with the given configuration. 
 */
public class HttpUrlChecker implements ItemChecker, Configurable
{
    /** The user agent */
    protected String _userAgent;
    
    /** The method */
    protected String _method;
    
    /** The timeout */
    protected int _timeout;
    
    /** The acceptable response codes */
    protected String _okCodes;
    
    /** The values the header must have */
    protected Map<String, String> _requestHeaderValues;

    /** The values the header must have */
    protected Map<String, String> _responseHeaderValues;

    /** Something to add to the parameter */
    protected String _additionnalURL;

    public void configure(Configuration configuration) throws ConfigurationException
    {
        Configuration config = configuration.getChild("configuration");
        Configuration requestConfig = config.getChild("request");
        Configuration responseConfig = config.getChild("response");
        
        Configuration timeoutConfig = requestConfig.getChild("timeout", false);
        
        _timeout = timeoutConfig != null ? timeoutConfig.getValueAsInteger() : -1;
        _userAgent = requestConfig.getChild("user-agent").getValue(null);
        _method = requestConfig.getChild("method").getValue(null);
        _additionnalURL = requestConfig.getChild("additionnal-url").getValue("");
        _requestHeaderValues = new HashMap<> ();
        for (Configuration headerChildConfig : requestConfig.getChild("header").getChildren())
        {
            _requestHeaderValues.put(headerChildConfig.getName(), headerChildConfig.getValue());
        }
        
        _okCodes = responseConfig.getChild("code").getValue(null);
        
        _responseHeaderValues = new HashMap<> ();
        for (Configuration headerChildConfig : responseConfig.getChild("header").getChildren())
        {
            _responseHeaderValues.put(headerChildConfig.getName(), headerChildConfig.getValue());
        }
    }
    
    @Override
    public void check(List<String> values) throws ItemCheckerTestFailureException
    {
        String configUrl = values.get(0);
        
        try
        {
            HttpURLConnection httpUrlConnection = HttpUrlUtils.prepareConnection(configUrl + _additionnalURL, _userAgent, _method, _timeout, _timeout, false, _requestHeaderValues);
            
            // Do the connection
            int responseCode = httpUrlConnection.getResponseCode();
            
            _testResponseCode(httpUrlConnection, responseCode);
            _testHeaders(httpUrlConnection);
        }
        catch (IOException e)
        {
            throw new ItemCheckerTestFailureException("Unable to contact '" + configUrl + _additionnalURL + "' (" + e.getMessage() + ")", e);
        }
    }
    
    private void _testHeaders(HttpURLConnection httpUrlConnection) throws ItemCheckerTestFailureException
    {
        Set<String> keys = _responseHeaderValues.keySet();
        for (String key : keys)
        {
            if (httpUrlConnection.getHeaderField(key) == null)
            {
                throw new ItemCheckerTestFailureException("The header field '" + key + "' does not exist");
            }
            
            if (!httpUrlConnection.getHeaderField(key).equals(_responseHeaderValues.get(key)))
            {
                throw new ItemCheckerTestFailureException("The header field '" + key + "' does not have the expected value '" + _responseHeaderValues.get(key) + "', found '" + httpUrlConnection.getHeaderField(key) + "' instead.");
            }
        }
    }

    private void _testResponseCode(HttpURLConnection httpUrlConnection, int responseCode) throws ItemCheckerTestFailureException, IOException
    {
        String[] okCodes = {Integer.toString(HttpURLConnection.HTTP_OK)};
        if (_okCodes != null)
        {
            okCodes = StringUtils.split(_okCodes, ",");
        }
        if (!ArrayUtils.contains(okCodes, Integer.toString(responseCode)))
        {
            throw new ItemCheckerTestFailureException("Error code " + responseCode + ". Message: " + httpUrlConnection.getResponseMessage());
        }
    }
}
