/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.datasource;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.ametys.core.datasource.DataSourceConsumer;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.config.ConfigManager;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.parameter.Validator;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Implementation of {@link DataSourceConsumer} allowing to know whether a data source is used in the configuration or not.
 * It also allows to retrieve the data source ids that are currently in use.
 */
public class ConfigurationDataSourceConsumer extends AbstractLogEnabled implements DataSourceConsumer
{
    @Override
    public TypeOfUse isInUse(String id)
    {
        Map<String, TypeOfUse> usedDataSourceIds = getUsedDataSourceIds();
        return usedDataSourceIds.containsKey(id) ? usedDataSourceIds.get(id) : TypeOfUse.NOT_USED;
    }
    
    @Override
    public Map<String, TypeOfUse> getUsedDataSourceIds()
    {
        Map<String, TypeOfUse> usedDataSourceIds = new HashMap<>();
        Config config = Config.getInstance();
        if (config != null)
        {
            Map<String, Object> values = new HashMap<> ();
            try
            {
                values = config.getValues();
            }
            catch (Exception e)
            {
                getLogger().error("Cannot read the configuration file.", e);
            }
            
            ConfigManager configManager = ConfigManager.getInstance();
            Collection<ElementDefinition> definitions = configManager.getConfigurationParameters();
            for (ElementDefinition definition : definitions)
            {
                if (ModelItemTypeConstants.DATASOURCE_ELEMENT_TYPE_ID.equals(definition.getType().getId()) && !configManager.evaluateDisableConditions(definition.getDisableConditions(), values))
                {
                    // A data source is not considered used when it is disabled
                    // Type of use depends on mandatory property
                    usedDataSourceIds.put(config.getValue(definition.getName()), _isMandatory(definition) ? TypeOfUse.BLOCKING : TypeOfUse.NON_BLOCKING);
                }
            }
        }
        
        return usedDataSourceIds;
    }
    
    private boolean _isMandatory(ElementDefinition def)
    {
        Validator validator = def.getValidator();
        if (validator != null)
        {
            return validator.getConfiguration().containsKey("mandatory") && (Boolean) validator.getConfiguration().get("mandatory");
        }
        
        return false;
    }
}
