/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.enumerator;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.SourceException;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.TraversableSource;

import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.Enumerator;

/**
 * An enumeration that list will list all the file in a source recursively.
 * Source inside a collection will be listed as "path/to/filename"
 */
public abstract class AbstractTraversableSourceEnumerator implements Enumerator<String>, Serviceable
{
    /** The source resolver */
    private SourceResolver _srcResolver;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _srcResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    public Map<String, I18nizableText> getTypedEntries() throws MalformedURLException, IOException
    {
        TraversableSource extractionDir = (TraversableSource) _srcResolver.resolveURI(getLocation());
        
        if (extractionDir.exists())
        {
            return _getCollectionFilename(extractionDir, StringUtils.EMPTY);
        }
        else
        {
            return Map.of();
        }
    }

    /**
     * Provide the location of the Source to extract the content from
     * @return the location of the source
     */
    protected abstract String getLocation();

    private Map<String, I18nizableText> _getCollectionFilename(TraversableSource source, String prefix) throws SourceException
    {
        Map<String, I18nizableText> filenames = new HashMap<>();
        
        for (TraversableSource child : (Collection<TraversableSource>) source.getChildren())
        {
            String childName = child.getName();

            if (child.isCollection())
            {
                filenames.putAll(_getCollectionFilename(child, prefix + childName + File.separator));
            }
            else if (acceptSource(child))
            {
                String filePath = prefix + childName;
                filenames.put(filePath, new I18nizableText(filePath));
            }
        }
        
        return filenames;
    }

    /**
     * Specify if a source should be included in the result.
     * 
     * @param source the Source to test
     * @return true if the source should be included, false otherwise
     */
    protected boolean acceptSource(TraversableSource source)
    {
        return true;
    }

    public I18nizableText getEntry(String value) throws Exception
    {
        return new I18nizableText(value);
    }

}
