/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.group.directory;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.group.Group;
import org.ametys.core.group.GroupIdentity;
import org.ametys.core.group.directory.GroupDirectory;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.directory.UserDirectory;
import org.ametys.core.user.population.PopulationContextHelper;
import org.ametys.core.user.population.UserPopulation;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.SizeUtils.ExcludeFromSizeCalculation;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Implementation of a {@link GroupDirectory} based on user's populations
 *
 */
public class UserPopulationsGroupDirectory extends AbstractLogEnabled implements GroupDirectory, Serviceable
{
    /** The user population DAO */
    protected UserPopulationDAO _userPopulationDAO;
    
    /** The population context helper */
    protected PopulationContextHelper _populationContextHelper;
    
    /** The i18n utils */
    protected I18nUtils _i18nUtils;
    
    private String _id;
    private I18nizableText _label;
    private String _gdModelId;
    private Map<String, Object> _params;

    
    public void service(ServiceManager smanager) throws ServiceException
    {
        _userPopulationDAO = (UserPopulationDAO) smanager.lookup(UserPopulationDAO.ROLE);
        _populationContextHelper = (PopulationContextHelper) smanager.lookup(PopulationContextHelper.ROLE);
        _i18nUtils = (I18nUtils) smanager.lookup(I18nUtils.ROLE);
    }
    
    @Override
    public String getId()
    {
        return _id;
    }

    @Override
    public I18nizableText getLabel()
    {
        return _label;
    }

    @Override
    public void setId(String id)
    {
        _id = id;
    }

    @Override
    public void setLabel(I18nizableText label)
    {
        _label = label;
    }

    @Override
    public String getGroupDirectoryModelId()
    {
        return _gdModelId;
    }

    public Map<String, Object> getParameterValues()
    {
        return _params;
    }

    public void init(String groupDirectoryModelId, Map<String, Object> paramValues) throws Exception
    {
        _gdModelId = groupDirectoryModelId;
        _params = paramValues;    
    }

    public Group getGroup(String groupID)
    {
        UserPopulation up = _userPopulationDAO.getUserPopulation(groupID);
        if (up != null)
        {
            // Check user population is part of available populations
            if (getAvailableUserPopulationsForContext().contains(up))
            {
                return new UserPopulationGroup(up, _id, _getGroupLabel(up), this);
            }
        }
        return null;
    }

    public Set<Group> getGroups()
    {
        List<UserPopulation> userPopulations = getAvailableUserPopulationsForContext();
        
        return userPopulations.stream()
            .map(up -> new UserPopulationGroup(up, _id, _getGroupLabel(up), this))
            .collect(Collectors.toSet());
    }
    
    private String _getGroupLabel(UserPopulation userPopulation)
    {
        Map<String, I18nizableTextParameter> i18nParams = Map.of("populationLabel", userPopulation.getLabel());
        I18nizableText groupLabel = new I18nizableText("plugin.core-impl", "PLUGINS_CORE_GROUPS_USER_POULATIONS_GROUP_LABEL", i18nParams);
        return _i18nUtils.translate(groupLabel);
    }
    
    /**
     * Get the available user populations for the current context
     * @return the available user populations
     */
    protected List<UserPopulation> getAvailableUserPopulationsForContext()
    {
        // By default returns all enabled user populations except admin population
        return _userPopulationDAO.getEnabledUserPopulations(false);
    }

    public Set<String> getUserGroups(UserIdentity userIdentity)
    {
        Group group = getGroup(userIdentity.getPopulationId());
        
        if (group != null)
        {
            return Set.of(group.getIdentity().getId());
        }
        else
        {
            return Set.of();
        }
    }

    public List<Group> getGroups(int count, int offset, Map parameters)
    {
        String pattern = (String) parameters.get("pattern");
        
        return getGroups().stream()
                .filter(group -> _filterMatchingGroup(group, pattern))
                .skip(offset)
                .limit(count < 0 ? Integer.MAX_VALUE : count)
                .collect(Collectors.toList());
    }
    
    private boolean _filterMatchingGroup(Group group, String pattern)
    {
        return StringUtils.isEmpty(pattern)
                || group.getLabel().toLowerCase().indexOf(pattern.toLowerCase()) != -1
                || group.getIdentity() != null && group.getIdentity().getId().toLowerCase().indexOf(pattern.toLowerCase()) != -1;
    }
    
    private static final class UserPopulationGroup implements Group
    {
        private GroupIdentity _identity;
        private String _label;
        
        @ExcludeFromSizeCalculation
        private UserPopulation _userPopulation;

        @ExcludeFromSizeCalculation
        private GroupDirectory _groupDirectory;
        
        UserPopulationGroup(UserPopulation userPopulation, String directoryId, String label, GroupDirectory groupDirectory)
        {
            _identity = new GroupIdentity(userPopulation.getId(), directoryId);
            _label = label;
            _groupDirectory = groupDirectory;
            _userPopulation = userPopulation;
        }
        
        public GroupIdentity getIdentity()
        {
            return _identity;
        }

        public String getLabel()
        {
            return _label;
        }

        public GroupDirectory getGroupDirectory()
        {
            return _groupDirectory;
        }

        public Set<UserIdentity> getUsers()
        {
            Set<User> users = _userPopulation.getUserDirectories().stream()
                        .map(UserDirectory::getUsers)
                        .flatMap(Collection::stream)
                        .collect(Collectors.toSet());
            
            return users.stream()
                        .map(User::getIdentity)
                        .collect(Collectors.toSet());
        }
        
    }

}
