/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.group.directory.jdbc;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.core.ObservationConstants;
import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;
import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.datasource.dbtype.SQLDatabaseTypeExtensionPoint;
import org.ametys.core.group.Group;
import org.ametys.core.group.GroupIdentity;
import org.ametys.core.group.InvalidModificationException;
import org.ametys.core.group.ModifiableGroup;
import org.ametys.core.group.directory.GroupDirectory;
import org.ametys.core.group.directory.GroupDirectoryModel;
import org.ametys.core.group.directory.ModifiableGroupDirectory;
import org.ametys.core.migration.MigrationEngine;
import org.ametys.core.migration.MigrationEngine.ActionData;
import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.MigrationExtensionPoint;
import org.ametys.core.migration.version.Version;
import org.ametys.core.migration.version.storage.SqlVersionStorage.SqlVersionConfiguration;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.script.SQLScriptHelper;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.Cacheable;
import org.ametys.core.util.SizeUtils.ExcludeFromSizeCalculation;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Standard implementation of {@link GroupDirectory} from the core database.
 */
public class JdbcGroupDirectory extends AbstractLogEnabled implements ModifiableGroupDirectory, Serviceable, Cacheable, Disposable
{
    /** Name of the parameter holding the datasource id */
    private static final String __DATASOURCE_PARAM_NAME = "runtime.groups.jdbc.datasource";
    /** Name of the parameter holding the SQL table name for storing the groups */
    private static final String __GROUPS_LIST_TABLE_PARAM_NAME = "runtime.groups.jdbc.list.table";
    /** Name of the parameter holding the SQL table name for storing the composition (i.e. users) of the groups */
    private static final String __GROUPS_COMPOSITION_TABLE_PARAM_NAME = "runtime.groups.jdbc.composition.table";
    
    private static final String __GROUPS_LIST_COLUMN_ID = "Id";
    private static final String __GROUPS_LIST_COLUMN_LABEL = "Label";
    private static final String __GROUPS_COMPOSITION_COLUMN_GROUPID = "Group_Id";
    private static final String __GROUPS_COMPOSITION_COLUMN_LOGIN = "Login";
    private static final String __GROUPS_COMPOSITION_COLUMN_POPULATIONID = "UserPopulation_Id";
    
    private static final String __JDBC_GROUPDIRECTORY_GROUPS_BY_USER_CACHE_NAME_PREFIX = JdbcGroupDirectory.class.getName() + "$groups.by.user$";
    
    /** The observation manager */
    protected ObservationManager _observationManager;
    /** The current user provider */
    protected CurrentUserProvider _currentUserProvider;
    /** The cocoon source resolver */
    protected SourceResolver _sourceResolver;
    /** The cache manager */
    protected AbstractCacheManager _cacheManager;
    
    /** The identifier of data source */
    protected String _dataSourceId;
    /** The name of the SQL table storing the groups */
    protected String _groupsListTableName;
    /** The name of the SQL table storing the composition (i.e. users) of the groups*/
    protected String _groupsCompositionTableName;
    
    /** Avalon manager */
    protected ServiceManager _manager;
    
    /** The id */
    protected String _id;
    /** The label */
    protected I18nizableText _label;
    /** The id of the {@link GroupDirectoryModel} */
    private String _groupDirectoryModelId;
    /** The map of the values of the parameters */
    private Map<String, Object> _paramValues;
    private boolean _lazyInitialized;
    
    // Cannot use _id as two GroupDirectories with same id can co-exist during a short amount of time (during GroupDirectoryDAO#_read)
    private final String _uniqueCacheSuffix = org.ametys.core.util.StringUtils.generateKey();
    
    private MigrationExtensionPoint _migrationEP;
    private MigrationEngine _migrationEngine;
    
    private SQLDatabaseTypeExtensionPoint _sqlDatabaseTypeExtensionPoint;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _observationManager = (ObservationManager) manager.lookup(ObservationManager.ROLE);
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _cacheManager = (AbstractCacheManager) manager.lookup(AbstractCacheManager.ROLE);
        _migrationEP = (MigrationExtensionPoint) manager.lookup(MigrationExtensionPoint.ROLE);
        _migrationEngine = (MigrationEngine) manager.lookup(MigrationEngine.ROLE);
    }
    
    private void _initTable(Connection connection) throws SQLException, MigrationException
    {
        if (!SQLScriptHelper.tableExists(connection, _groupsListTableName))
        {
            MigrationComponent component = _migrationEP.getExtension("group-directory.jdbc");
            
            Version version = component.versionStorage().createVersion("group-directory.jdbc_" + _groupsListTableName, 
                                                                          component, 
                                                                          new SqlVersionConfiguration(_dataSourceId), 
                                                                          Map.of("TABLENAME", _groupsListTableName, 
                                                                                 "TABLENAME_COMPOSITION", _groupsCompositionTableName));
            if (_migrationEngine.initializeVersion(new ActionData(version, _migrationEngine.getHighestUpgradeVersionNumber(component), component.initialization(), null)))
            {
                throw new IllegalStateException("Requiring a restart after in a table initialization for a JdbcGroupDirectory is not authorized");
            }
        }
    }
    
    private SQLDatabaseTypeExtensionPoint getSQLDatabaseTypeExtensionPoint()
    {
        if (_sqlDatabaseTypeExtensionPoint == null)
        {
            try
            {
                _sqlDatabaseTypeExtensionPoint = (SQLDatabaseTypeExtensionPoint) _manager.lookup(SQLDatabaseTypeExtensionPoint.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException(e);
            }
        }
        return _sqlDatabaseTypeExtensionPoint;
    }
    
    @Override
    public String getId()
    {
        return _id;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return _label;
    }
    
    @Override
    public void setId(String id)
    {
        _id = id;
    }
    
    @Override
    public void setLabel(I18nizableText label)
    {
        _label = label;
    }
    
    @Override
    public String getGroupDirectoryModelId ()
    {
        return _groupDirectoryModelId;
    }
    
    @Override
    public Map<String, Object> getParameterValues()
    {
        return _paramValues;
    }
    
    @Override
    public void dispose()
    {
        removeCaches();
    }
    
    @Override
    public Collection<SingleCacheConfiguration> getManagedCaches()
    {
        return Arrays.asList(
                SingleCacheConfiguration.of(
                        __JDBC_GROUPDIRECTORY_GROUPS_BY_USER_CACHE_NAME_PREFIX + _uniqueCacheSuffix, 
                        _buildI18n("PLUGINS_CORE_GROUPS_JDBC_CACHE_GROUPS_BY_USER_LABEL"), 
                        _buildI18n("PLUGINS_CORE_GROUPS_JDBC_CACHE_GROUPS_BY_USER_DESC"))
        );
    }
    
    private I18nizableText _buildI18n(String i18Key)
    {
        String catalogue = "plugin.core-impl";
        I18nizableText groupDirectoryId = new I18nizableText(getId());
        Map<String, I18nizableTextParameter> params = Map.of("id", groupDirectoryId);
        return new I18nizableText(catalogue, i18Key, params);
    }
    
    private Cache<UserIdentity, Set<String>> _getCacheGroupsByUser()
    {
        return getCacheManager().get(__JDBC_GROUPDIRECTORY_GROUPS_BY_USER_CACHE_NAME_PREFIX + _uniqueCacheSuffix);
    }
    
    @Override
    public AbstractCacheManager getCacheManager()
    {
        return _cacheManager;
    }
    
    @Override
    public void init(String groupDirectoryModelId, Map<String, Object> paramValues)
    {
        _groupDirectoryModelId = groupDirectoryModelId;
        _paramValues = paramValues;
        
        _groupsListTableName = (String) paramValues.get(__GROUPS_LIST_TABLE_PARAM_NAME);
        _groupsCompositionTableName = (String) paramValues.get(__GROUPS_COMPOSITION_TABLE_PARAM_NAME);
        _dataSourceId = (String) paramValues.get(__DATASOURCE_PARAM_NAME);
        
        createCaches();
    }
    
    /**
     * Get the connection to the database 
     * @return the SQL connection
     */
    protected Connection getSQLConnection()
    {
        Connection connection = ConnectionHelper.getConnection(_dataSourceId);
        
        if (!_lazyInitialized)
        {
            try
            {
                _initTable(connection);
            }
            catch (Exception e)
            {
                getLogger().error("The tables requires by the " + this.getClass().getName() + " could not be created. A degraded behavior will occur", e);
            }
            
            _lazyInitialized = true;
        }
        
        return connection;
    }
    
    @Override
    public ModifiableGroup getGroup(String groupID)
    {
        JdbcGroup group = null;

        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;

        try
        {
            connection = getSQLConnection();
            String dbType = ConnectionHelper.getDatabaseType(connection);
            
            String sql = "SELECT " + __GROUPS_LIST_COLUMN_LABEL + " FROM " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsListTableName) + " WHERE " + __GROUPS_LIST_COLUMN_ID + " = ?";
            stmt = connection.prepareStatement(sql);
            stmt.setInt(1, Integer.parseInt(groupID));

            if (getLogger().isDebugEnabled())
            {
                getLogger().debug(sql);
            }
            rs = stmt.executeQuery();

            // Iterate over all the groups
            if (rs.next())
            {
                String label = rs.getString(__GROUPS_LIST_COLUMN_LABEL);
                group = new JdbcGroup(new GroupIdentity(groupID, getId()), label, this);

                _fillGroup(group, connection);
            }
        }
        catch (NumberFormatException e)
        {
            getLogger().error("Group ID must be an integer.", e);
            return null;
        }
        catch (SQLException e)
        {
            getLogger().error("Error communication with database", e);
            return null;
        }
        finally
        {
            ConnectionHelper.cleanup(rs);       
            ConnectionHelper.cleanup(stmt);       
            ConnectionHelper.cleanup(connection);       
        }

        // Return the found group or null
        return group;
    }

    @Override
    public Set<Group> getGroups()
    {
        Set<Group> groups = new LinkedHashSet<>();

        Connection connection = null;
        Statement stmt = null;
        ResultSet rs = null;
        
        try
        {
            connection = getSQLConnection();
            String dbType = ConnectionHelper.getDatabaseType(connection);
            
            stmt = connection.createStatement();
            String sql = _createGetGroupsClause(dbType);

            if (getLogger().isDebugEnabled())
            {
                getLogger().debug(sql);
            }
            rs = stmt.executeQuery(sql);

            // Iterate over all the groups
            while (rs.next())
            {
                String groupID = rs.getString(__GROUPS_LIST_COLUMN_ID);
                String label = rs.getString(__GROUPS_LIST_COLUMN_LABEL);
                JdbcGroup group = new JdbcGroup(new GroupIdentity(groupID, getId()), label, this);

                _fillGroup(group, connection);

                // Add current group
                groups.add(group);
            }
        }
        catch (SQLException e)
        {
            getLogger().error("Error communication with database", e);
            return Collections.emptySet();
        }
        finally
        {
            ConnectionHelper.cleanup(rs);       
            ConnectionHelper.cleanup(stmt);       
            ConnectionHelper.cleanup(connection);       
        }

        return groups;
    }
    
    /**
     * Get the sql clause that gets all groups
     * @param dbType The connection type
     * @return A non null sql clause (e.g. "select ... from ... where ...")
     */
    protected String _createGetGroupsClause(String dbType)
    {
        return "SELECT " + __GROUPS_LIST_COLUMN_ID + ", " + __GROUPS_LIST_COLUMN_LABEL + " FROM " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsListTableName) + " ORDER BY " + __GROUPS_LIST_COLUMN_LABEL;
    }
    
    /**
     * Fill users set in a group.
     * 
     * @param group The group to fill.
     * @param connection The SQL connection.
     * @throws SQLException If a problem occurs.
     */
    protected void _fillGroup(JdbcGroup group, Connection connection) throws SQLException
    {
        PreparedStatement stmt = null;
        ResultSet rs = null;

        try
        {
            String dbType = ConnectionHelper.getDatabaseType(connection);
            String sql = "SELECT " + __GROUPS_COMPOSITION_COLUMN_LOGIN + ", " + __GROUPS_COMPOSITION_COLUMN_POPULATIONID + " FROM " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsCompositionTableName) + " WHERE " + __GROUPS_COMPOSITION_COLUMN_GROUPID + " = ?";

            stmt = connection.prepareStatement(sql);

            stmt.setInt(1, Integer.parseInt(group.getIdentity().getId()));

            if (getLogger().isDebugEnabled())
            {
                getLogger().debug(sql);
            }
            rs = stmt.executeQuery();

            // Iterate over all the users from current group 
            while (rs.next())
            {
                UserIdentity identity = new UserIdentity(rs.getString(__GROUPS_COMPOSITION_COLUMN_LOGIN), rs.getString(__GROUPS_COMPOSITION_COLUMN_POPULATIONID));
                group.addUser(identity);
            }
        }
        finally
        {
            ConnectionHelper.cleanup(rs);
            ConnectionHelper.cleanup(stmt);
        }
    }

    @Override
    public Set<String> getUserGroups(UserIdentity userIdentity)
    {
        if (isCachingEnabled() && _getCacheGroupsByUser().hasKey(userIdentity))
        {
            Set<String> userGroups = _getCacheGroupsByUser().get(userIdentity);
            // Cache hit, return the results. 
            return userGroups;
        }
        
        Set<String> userGroups = _executeSqlForUserGroups(userIdentity);
        
        // Cache the results.
        if (isCachingEnabled())
        {
            _getCacheGroupsByUser().put(userIdentity, userGroups);
        }
        
        return userGroups;
    }
    
    private Set<String> _executeSqlForUserGroups(UserIdentity userIdentity)
    {
        String login = userIdentity.getLogin();
        String populationId = userIdentity.getPopulationId();
        
        Set<String> groups = new HashSet<>();
        if (login == null)
        {
            return groups;
        }
        
        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;

        try
        {
            connection = getSQLConnection();
            String dbType = ConnectionHelper.getDatabaseType(connection);
            
            String sql = "SELECT " + __GROUPS_COMPOSITION_COLUMN_GROUPID + " FROM " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsCompositionTableName) + " WHERE " + __GROUPS_COMPOSITION_COLUMN_LOGIN + " = ? AND " + __GROUPS_COMPOSITION_COLUMN_POPULATIONID + " = ?";
            stmt = connection.prepareStatement(sql);
            stmt.setString(1, login);
            stmt.setString(2, populationId);

            if (getLogger().isDebugEnabled())
            {
                getLogger().debug(sql);
            }
            rs = stmt.executeQuery();

            // Iterate over all the groups
            while (rs.next())
            {
                String groupID = rs.getString(__GROUPS_COMPOSITION_COLUMN_GROUPID);

                // Add the current group
                groups.add(groupID);
            }
        }
        catch (SQLException e)
        {
            getLogger().error("Error communication with database", e);
            return Collections.emptySet();
        }
        finally
        {
            ConnectionHelper.cleanup(rs);       
            ConnectionHelper.cleanup(stmt);       
            ConnectionHelper.cleanup(connection);       
        }

        // Return the groups, potentially empty
        return groups;
    }

    @Override
    public List<Group> getGroups(int count, int offset, Map parameters)
    {
        List<Group> groups = new ArrayList<>();
        
        String pattern = (String) parameters.get("pattern");

        Iterator iterator = getGroups().iterator();

        //int totalCount = 0;
        int currentOffset = offset;

        while (currentOffset > 0 && iterator.hasNext())
        {
            Group group = (Group) iterator.next();
            if (StringUtils.isEmpty(pattern)
                || group.getLabel().toLowerCase().indexOf(pattern.toLowerCase()) != -1
                || group.getIdentity() != null && group.getIdentity().getId().toLowerCase().indexOf(pattern.toLowerCase()) != -1)
            {
                currentOffset--;
                //totalCount++;
            }
        }

        int currentCount = count;
        while ((count == -1 || currentCount > 0) && iterator.hasNext())
        {
            Group group = (Group) iterator.next();

            if (StringUtils.isEmpty(pattern)
                || group.getLabel().toLowerCase().indexOf(pattern.toLowerCase()) != -1
                || group.getIdentity() != null && group.getIdentity().getId().toLowerCase().indexOf(pattern.toLowerCase()) != -1)
            {
                groups.add(group);
                currentCount--;
                //totalCount++;
            }
        }

        /*while (iterator.hasNext())
        {
            Group group = (Group) iterator.next();
            
            if (StringUtils.isEmpty(pattern) || group.getLabel().toLowerCase().indexOf(pattern.toLowerCase()) != -1)
            {
                totalCount++;
            }
        }*/
        
        // TODO return totalCount
        return groups;
    }

    @Override
    public ModifiableGroup add(String name) throws InvalidModificationException
    {
        Connection connection = null; 
        PreparedStatement statement = null;
        ResultSet rs = null;

        String id = null;

        try
        {
            connection = getSQLConnection();
            String dbType = ConnectionHelper.getDatabaseType(connection);
            
            if (ConnectionHelper.DATABASE_ORACLE.equals(dbType))
            {
                statement = connection.prepareStatement("SELECT seq_" + _groupsListTableName + ".nextval FROM dual");
                rs = statement.executeQuery();
                if (rs.next())
                {
                    id = rs.getString(1);
                }
                ConnectionHelper.cleanup(rs);
                ConnectionHelper.cleanup(statement);

                statement = connection.prepareStatement("INSERT INTO " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsListTableName) + " (" + __GROUPS_LIST_COLUMN_ID + ", " + __GROUPS_LIST_COLUMN_LABEL + ") VALUES(?, ?)");
                statement.setString(1, id);
                statement.setString(2, name);
            }
            else
            {
                statement = connection.prepareStatement("INSERT INTO " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsListTableName) + " (" + __GROUPS_LIST_COLUMN_LABEL + ") VALUES (?)");
                statement.setString(1, name);
            }

            statement.executeUpdate();

            ConnectionHelper.cleanup(statement);

            //FIXME Write query working with all database
            if (ConnectionHelper.DATABASE_MYSQL.equals(dbType))
            {
                statement = connection.prepareStatement("SELECT " + __GROUPS_LIST_COLUMN_ID + " FROM " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsListTableName) + " WHERE " + __GROUPS_LIST_COLUMN_ID + " = last_insert_id()");    
                rs = statement.executeQuery();
                if (rs.next())
                {
                    id = rs.getString(__GROUPS_LIST_COLUMN_ID);
                }
                else
                {
                    if (connection.getAutoCommit())
                    {
                        throw new InvalidModificationException("Cannot retrieve inserted group. Group was created but listeners not called : base may be inconsistant");
                    }
                    else
                    {
                        connection.rollback();
                        throw new InvalidModificationException("Cannot retrieve inserted group. Rolling back");
                    }
                }
            }
            else if (ConnectionHelper.DATABASE_DERBY.equals(dbType))
            {
                statement = connection.prepareStatement("VALUES IDENTITY_VAL_LOCAL ()");
                rs = statement.executeQuery();
                if (rs.next())
                {
                    id = rs.getString(1);
                }
            }
            else if (ConnectionHelper.DATABASE_HSQLDB.equals(dbType))
            {
                statement = connection.prepareStatement("CALL IDENTITY ()");
                rs = statement.executeQuery();
                if (rs.next())
                {
                    id = rs.getString(1);
                }
            }
            else if (ConnectionHelper.DATABASE_POSTGRES.equals(dbType))
            {
                statement = connection.prepareStatement("SELECT lastval()");
                rs = statement.executeQuery();
                if (rs.next())
                {
                    id = rs.getString(1);
                }
            }

            if (id != null)
            {
                Map<String, Object> eventParams = new HashMap<>();
                eventParams.put(ObservationConstants.ARGS_GROUP, new GroupIdentity(id, getId()));
                _observationManager.notify(new Event(ObservationConstants.EVENT_GROUP_ADDED, _currentUserProvider.getUser(), eventParams));
            }
        }
        catch (SQLException ex)
        {
            throw new RuntimeException(ex);
        }
        finally
        {
            ConnectionHelper.cleanup(rs);       
            ConnectionHelper.cleanup(statement);       
            ConnectionHelper.cleanup(connection);       
        }

        return new JdbcGroup(new GroupIdentity(id, getId()), name, this);
    }

    @Override
    public void update(ModifiableGroup userGroup) throws InvalidModificationException
    {
        Connection connection = null;
        PreparedStatement statement = null;

        if (getLogger().isDebugEnabled())
        {
            getLogger().debug("Updating group " + GroupIdentity.groupIdentityToString(userGroup.getIdentity()) + " with " + userGroup.getUsers().size() + " user(s)");
        }
        
        try
        {
            
            connection = getSQLConnection();
            String dbType = ConnectionHelper.getDatabaseType(connection);
            
            // Start transaction.
            connection.setAutoCommit(false);

            statement = connection.prepareStatement("UPDATE " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsListTableName) + " SET " + __GROUPS_LIST_COLUMN_LABEL + "=? WHERE " + __GROUPS_LIST_COLUMN_ID + " = ?");
            statement.setString(1, userGroup.getLabel());
            statement.setInt(2, Integer.parseInt(userGroup.getIdentity().getId()));

            if (statement.executeUpdate() == 0)
            {
                throw new InvalidModificationException("No group with id '" + userGroup.getIdentity().getId() + "' may be removed");
            }
            ConnectionHelper.cleanup(statement);

            statement = connection.prepareStatement("DELETE FROM " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsCompositionTableName) + " WHERE " + __GROUPS_COMPOSITION_COLUMN_GROUPID + " = ?");
            statement.setInt(1, Integer.parseInt(userGroup.getIdentity().getId()));

            statement.executeUpdate();
            ConnectionHelper.cleanup(statement);

            if (!userGroup.getUsers().isEmpty())
            {
                // Tests if the connection supports batch updates.
                boolean supportsBatch = connection.getMetaData().supportsBatchUpdates();

                statement = connection.prepareStatement("INSERT INTO " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsCompositionTableName) + " (" + __GROUPS_COMPOSITION_COLUMN_GROUPID + ", " + __GROUPS_COMPOSITION_COLUMN_LOGIN + ", " + __GROUPS_COMPOSITION_COLUMN_POPULATIONID + ") VALUES (?, ?, ?)");
                
                for (UserIdentity identity : userGroup.getUsers())
                {
                    String login = identity.getLogin();
                    String populationId = identity.getPopulationId();
                    statement.setInt(1, Integer.parseInt(userGroup.getIdentity().getId()));
                    statement.setString(2, login);
                    statement.setString(3, populationId);
                    
                    // If batch updates are supported, add to the batch, else execute directly.
                    if (supportsBatch)
                    {
                        statement.addBatch();
                    }
                    else
                    {
                        statement.executeUpdate();
                    }
                }
                
                // If the insert queries were queued in a batch, execute it.
                if (supportsBatch)
                {
                    statement.executeBatch();
                }
            }

            ConnectionHelper.cleanup(statement);

            // Commit transaction.
            connection.commit();
            
            // Clear global cache
            _getCacheGroupsByUser().invalidateAll();

            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_GROUP, userGroup.getIdentity());
            _observationManager.notify(new Event(ObservationConstants.EVENT_GROUP_UPDATED, _currentUserProvider.getUser(), eventParams));
        }
        catch (NumberFormatException ex)
        {
            throw new InvalidModificationException("No group with id '" + userGroup.getIdentity().getId() + "' may be removed", ex);
        }
        catch (SQLException ex)
        {
            throw new RuntimeException(ex);
        }
        finally
        {
            ConnectionHelper.cleanup(statement);
            ConnectionHelper.cleanup(connection);
        }
        
        if (getLogger().isDebugEnabled())
        {
            getLogger().debug("Updated group " + GroupIdentity.groupIdentityToString(userGroup.getIdentity()) + " with " + userGroup.getUsers().size() + " user(s)");
        }
    }

    @Override
    public void remove(String groupID) throws InvalidModificationException
    {
        Connection connection = null;
        PreparedStatement statement = null;

        try
        {
            connection = getSQLConnection();
            String dbType = ConnectionHelper.getDatabaseType(connection);
            
            statement = connection.prepareStatement("DELETE FROM " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsListTableName) + " WHERE " + __GROUPS_LIST_COLUMN_ID + " = ?");
            statement.setInt(1, Integer.parseInt(groupID));

            if (statement.executeUpdate() == 0)
            {
                throw new InvalidModificationException("No group with id '" + groupID + "' may be removed");
            }
            ConnectionHelper.cleanup(statement);       

            statement = connection.prepareStatement("DELETE FROM " + getSQLDatabaseTypeExtensionPoint().languageEscapeTableName(dbType, _groupsCompositionTableName) + " WHERE " + __GROUPS_COMPOSITION_COLUMN_GROUPID + " = ?");
            statement.setInt(1, Integer.parseInt(groupID));

            statement.executeUpdate();
            
            // Clear global cache
            _getCacheGroupsByUser().invalidateAll();

            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_GROUP, new GroupIdentity(groupID, getId()));
            _observationManager.notify(new Event(ObservationConstants.EVENT_GROUP_DELETED, _currentUserProvider.getUser(), eventParams));
        }
        catch (NumberFormatException ex)
        {
            throw new InvalidModificationException("No group with id '" + groupID + "' may be removed, the ID must be a number.", ex);
        }
        catch (SQLException ex)
        {
            throw new RuntimeException(ex);
        }
        finally
        {
            ConnectionHelper.cleanup(statement);       
            ConnectionHelper.cleanup(connection);       
        }
    }
    
    private static final class JdbcGroup implements ModifiableGroup
    {
        private Set<UserIdentity> _users;
        private GroupIdentity _identity;
        private String _groupLabel;
        
        @ExcludeFromSizeCalculation
        private GroupDirectory _groupDirectory;
        
        JdbcGroup(GroupIdentity identity, String label, GroupDirectory groupDirectory)
        {
            _identity = identity;
            _groupLabel = label;
            _groupDirectory = groupDirectory;
            _users = new HashSet<>();
        }
        
        @Override
        public GroupIdentity getIdentity()
        {
            return _identity;
        }

        @Override
        public String getLabel()
        {
            return _groupLabel;
        }

        @Override
        public GroupDirectory getGroupDirectory()
        {
            return _groupDirectory;
        }
        
        @Override
        public void setLabel(String label)
        {
            _groupLabel = label;
        }

        @Override
        public void addUser(UserIdentity user)
        {
            _users.add(user);
        }

        @Override
        public void removeUser(UserIdentity user)
        {
            _users.remove(user);
        }
        
        @Override
        public void removeUsers()
        {
            _users.clear();
        }
        
        @Override
        public Set<UserIdentity> getUsers()
        {
            return _users;
        }
        
        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("UserGroup[");
            sb.append(_identity);
            sb.append(" (");
            sb.append(_groupLabel);
            sb.append(") => ");
            sb.append(_users.toString());
            sb.append("]");
            return sb.toString();
        }    
        
        @Override
        public boolean equals(Object another)
        {
            if (another == null || !(another instanceof JdbcGroup))
            {
                return false;
            }
            
            JdbcGroup otherGroup = (JdbcGroup) another;
            
            return _identity != null && _identity.equals(otherGroup.getIdentity());
        }
        
        @Override
        public int hashCode()
        {
            return _identity.hashCode();
        }
    }
}
