/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.hash;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.SecureRandom;
import java.util.Base64;

import org.apache.commons.io.IOUtils;
import org.bouncycastle.crypto.generators.Argon2BytesGenerator;
import org.bouncycastle.crypto.params.Argon2Parameters;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.runtime.config.Config;
import org.ametys.runtime.util.AmetysHomeHelper;

/**
 * Password encoder using Argon2 hashing algorithm.
 */
public class Argon2PasswordEncoder
{
    /** Argon2PasswordEncoder is based on Argon2PasswordEncoder from spring framework crypto library, and add some pepper */
    
    /** Avalon Role */
    public static final String ROLE = Argon2PasswordEncoder.class.getName();

    // Logger for traces
    private static Logger __logger = LoggerFactory.getLogger(Config.class);
    
    private static final String __PEPPER_FILE_PREFIX = "auth/pepper_";
    
    private static final int __SALT_LENGTH = 16;
    private static final int __PEPPER_LENGTH = 16;
    private static final int __HASH_LENGTH = 32;
    private static final int __PARALLELISM = 1;
    private static final int __MEMORY = 1 << 13;
    private static final int __ITERATIONS = 5;

    private static final Base64.Encoder __B64ENCODER = Base64.getEncoder().withoutPadding();
    private static final Base64.Decoder __B64DECODER = Base64.getDecoder();

    private byte[] _pepper;

    /**
     * Constructor for Argon2PasswordEncoder
     * @param pepper the pepper used by the Argon2PasswordEncoder
     */
    public Argon2PasswordEncoder(byte[] pepper)
    {
        _pepper = pepper;
    }
    
    /**
     * Get a pepper for a given id and generate it if it does not already exists
     * @param id the id of the pepper
     * @return the value of the paper
     * @throws IOException if I/O error occurred.
     */
    public static byte[] getOrCreatePepper(String id) throws IOException
    {
        File pepperFile = new File(AmetysHomeHelper.getAmetysHomeData(), __PEPPER_FILE_PREFIX + id);

        if (!pepperFile.exists())
        {
            if (!pepperFile.getParentFile().exists())
            {
                pepperFile.getParentFile().mkdirs();
            }
            
            pepperFile.createNewFile();
            
            byte[] pepperByes = generateBytes(__PEPPER_LENGTH);
            try (FileOutputStream fileOutputStream = new FileOutputStream(pepperFile))
            {
                IOUtils.write(pepperByes, fileOutputStream);
            }
        }
        
        try (FileInputStream inputStream = new FileInputStream(pepperFile)) 
        {
            return IOUtils.toByteArray(inputStream);
        }
    }
    
    /**
     * Encode the raw password with argon2id hash algorithm.
     * @param rawPassword the raw password
     * @return the hash
     */
    public String encode(CharSequence rawPassword) 
    {
        byte[] hash = new byte[__HASH_LENGTH];
        Argon2Parameters params = new Argon2Parameters
                .Builder(Argon2Parameters.ARGON2_id)
                .withSalt(generateBytes(__SALT_LENGTH))
                .withParallelism(__PARALLELISM)
                .withMemoryAsKB(__MEMORY)
                .withIterations(__ITERATIONS)
                .withSecret(this._pepper)
                .build();
        
        Argon2BytesGenerator generator = new Argon2BytesGenerator();
        generator.init(params);
        generator.generateBytes(rawPassword.toString().toCharArray(), hash);
        return encode(hash, params);
    }

    /**
     * Verify the encoded password obtained from storage matches the submitted raw
     * password after it too is encoded. Returns true if the passwords match, false if
     * they do not. The stored password itself is never decoded.
     * @param rawPassword the raw password to encode and match
     * @param encodedPassword the encoded password from storage to compare with
     * @return true if the raw password, after encoding, matches the encoded password from
     * storage
     */
    public boolean matches(CharSequence rawPassword, String encodedPassword) 
    {
        if (encodedPassword == null) 
        {
            __logger.warn("password hash is null");
            return false;
        }
        
        Argon2Hash decoded;
        try 
        {
            decoded = decode(encodedPassword);
        }
        catch (IllegalArgumentException ex) 
        {
            __logger.warn("Malformed password hash", ex);
            return false;
        }
        
        byte[] hashBytes = new byte[decoded.getHash().length];
        Argon2BytesGenerator generator = new Argon2BytesGenerator();
        Argon2Parameters decodedParams = decoded.getParameters();

        Argon2Parameters params = new Argon2Parameters
                .Builder(Argon2Parameters.ARGON2_id)
                .withSalt(decodedParams.getSalt())
                .withParallelism(decodedParams.getLanes())
                .withMemoryAsKB(decodedParams.getMemory())
                .withIterations(decodedParams.getIterations())
                .withSecret(this._pepper)
                .build();
        
        generator.init(params);
        generator.generateBytes(rawPassword.toString().toCharArray(), hashBytes);
        return constantTimeArrayEquals(decoded.getHash(), hashBytes);
    }

    private boolean constantTimeArrayEquals(byte[] expected, byte[] actual) 
    {
        if (expected.length != actual.length) 
        {
            return false;
        }
        
        int result = 0;
        for (int i = 0; i < expected.length; i++) 
        {
            result |= expected[i] ^ actual[i];
        }
        
        return result == 0;
    }
    
    private static byte[] generateBytes(int keyLength)
    {
        byte[] bytes = new byte[keyLength];
        SecureRandom random = new SecureRandom();
        random.nextBytes(bytes);
        return bytes;
    }
    
    /**
     * Encodes a raw Argon2-hash and its parameters into the standard Argon2-hash-string
     * as specified in the reference implementation
     * (https://github.com/P-H-C/phc-winner-argon2/blob/master/src/encoding.c#L244):
     *
     * {@code $argon2<T>[$v=<num>]$m=<num>,t=<num>,p=<num>$<bin>$<bin>}
     *
     * where {@code <T>} is either 'd', 'id', or 'i', {@code <num>} is a decimal integer
     * (positive, fits in an 'unsigned long'), and {@code <bin>} is Base64-encoded data
     * (no '=' padding characters, no newline or whitespace).
     *
     * The last two binary chunks (encoded in Base64) are, in that order, the salt and the
     * output. If no salt has been used, the salt will be omitted.
     * @param hash the raw Argon2 hash in binary format
     * @param parameters the Argon2 parameters that were used to create the hash
     * @return the encoded Argon2-hash-string as described above
     * @throws IllegalArgumentException if the Argon2Parameters are invalid
     */
    private String encode(byte[] hash, Argon2Parameters parameters) throws IllegalArgumentException 
    {
        StringBuilder stringBuilder = new StringBuilder();
        switch (parameters.getType()) 
        {
            case Argon2Parameters.ARGON2_d:
                stringBuilder.append("$argon2d");
                break;
            case Argon2Parameters.ARGON2_i:
                stringBuilder.append("$argon2i");
                break;
            case Argon2Parameters.ARGON2_id:
                stringBuilder.append("$argon2id");
                break;
            default:
                throw new IllegalArgumentException("Invalid algorithm type: " + parameters.getType());
        }
        
        stringBuilder.append("$v=").append(parameters.getVersion()).append("$m=").append(parameters.getMemory())
                .append(",t=").append(parameters.getIterations()).append(",p=").append(parameters.getLanes());
        
        if (parameters.getSalt() != null) 
        {
            stringBuilder.append("$").append(__B64ENCODER.encodeToString(parameters.getSalt()));
        }
        
        stringBuilder.append("$").append(__B64ENCODER.encodeToString(hash));
        return stringBuilder.toString();
    }

    /**
     * Decodes an Argon2 hash string as specified in the reference implementation
     * (https://github.com/P-H-C/phc-winner-argon2/blob/master/src/encoding.c#L244) into
     * the raw hash and the used parameters.
     *
     * The hash has to be formatted as follows:
     * {@code $argon2<T>[$v=<num>]$m=<num>,t=<num>,p=<num>$<bin>$<bin>}
     *
     * where {@code <T>} is either 'd', 'id', or 'i', {@code <num>} is a decimal integer
     * (positive, fits in an 'unsigned long'), and {@code <bin>} is Base64-encoded data
     * (no '=' padding characters, no newline or whitespace).
     *
     * The last two binary chunks (encoded in Base64) are, in that order, the salt and the
     * output. Both are required. The binary salt length and the output length must be in
     * the allowed ranges defined in argon2.h.
     * @param encodedHash the Argon2 hash string as described above
     * @return an {@link Argon2Hash} object containing the raw hash and the
     * {@link Argon2Parameters}.
     * @throws IllegalArgumentException if the encoded hash is malformed
     */
    private Argon2Hash decode(String encodedHash) throws IllegalArgumentException 
    {
        Argon2Parameters.Builder paramsBuilder;
        String[] parts = encodedHash.split("\\$");
        if (parts.length < 4) 
        {
            throw new IllegalArgumentException("Invalid encoded Argon2-hash");
        }
        
        int currentPart = 1;
        switch (parts[currentPart++]) 
        {
            case "argon2d":
                paramsBuilder = new Argon2Parameters.Builder(Argon2Parameters.ARGON2_d);
                break;
            case "argon2i":
                paramsBuilder = new Argon2Parameters.Builder(Argon2Parameters.ARGON2_i);
                break;
            case "argon2id":
                paramsBuilder = new Argon2Parameters.Builder(Argon2Parameters.ARGON2_id);
                break;
            default:
                throw new IllegalArgumentException("Invalid algorithm type: " + parts[0]);
        }
        
        if (parts[currentPart].startsWith("v=")) 
        {
            paramsBuilder.withVersion(Integer.parseInt(parts[currentPart].substring(2)));
            currentPart++;
        }
        
        String[] performanceParams = parts[currentPart++].split(",");
        if (performanceParams.length != 3) 
        {
            throw new IllegalArgumentException("Amount of performance parameters invalid");
        }
        
        if (!performanceParams[0].startsWith("m=")) 
        {
            throw new IllegalArgumentException("Invalid memory parameter");
        }
        
        paramsBuilder.withMemoryAsKB(Integer.parseInt(performanceParams[0].substring(2)));
        
        if (!performanceParams[1].startsWith("t=")) 
        {
            throw new IllegalArgumentException("Invalid iterations parameter");
        }
        
        paramsBuilder.withIterations(Integer.parseInt(performanceParams[1].substring(2)));
        
        if (!performanceParams[2].startsWith("p=")) 
        {
            throw new IllegalArgumentException("Invalid parallelity parameter");
        }
        
        paramsBuilder.withParallelism(Integer.parseInt(performanceParams[2].substring(2)));
        paramsBuilder.withSalt(__B64DECODER.decode(parts[currentPart++]));
        return new Argon2Hash(__B64DECODER.decode(parts[currentPart]), paramsBuilder.build());
    }
    
    /**
     * Object representing Argon hash, with the hash itself and the parameters
     */
    private static class Argon2Hash 
    {
        private byte[] _hash;

        private Argon2Parameters _parameters;

        Argon2Hash(byte[] hash, Argon2Parameters parameters) 
        {
            _hash = hash.clone();
            _parameters = parameters;
        }

        /**
         * get the hash
         * @return the hash
         */
        public byte[] getHash() 
        {
            return _hash.clone();
        }

        /**
         * get the parameters
         * @return the parameters
         */
        public Argon2Parameters getParameters() 
        {
            return _parameters;
        }
    }
}
