/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.right;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.ametys.core.group.GroupIdentity;
import org.ametys.core.right.AccessController;
import org.ametys.core.user.UserIdentity;

/**
 * Abstract {@link AccessController} for a hierarchical type of object. 
 * @param <T> The class of a supported object, from which you can retrieve its parent with {@link #_getParents(Object, Object)}
 * @param <C> The class of the permission context use to retrieve the parents with {@link #_getParents(Object, Object)} 
 */
public abstract class AbstractHierarchicalWithPermissionContextAccessController<T, C> extends AbstractProfileStorageBasedAccessController
{
    /**
     * Gets the parents of the object. Must return null when the object is the "root" (where "root" means here the root of the hierarchy of the access controller)
     * @param object The object
     * @param permissionCtx The permission context
     * @return the parents of the object, or null if the object is the "root" of the hierarchy
     */
    protected abstract Set<T> _getParents(T object, C permissionCtx);
    
    /**
     * Get the permission context. By default the permission context is the initial object context
     * @param initialObject the initial object context
     * @return the permission context. 
     */
    @SuppressWarnings("unchecked")
    protected C _getPermissionContext (T initialObject)
    {
        return (C) initialObject;
    }
    
    /**
     * Determines if the inheritance of permissions is disallowed on the given object
     * @param object The object
     * @return true if the inheritance of permissions is disallowed on the given object
     */
    protected boolean isInheritanceDisallowed(T object)
    {
        return _profileAssignmentStorageEP.isInheritanceDisallowed(object);
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public Map<String, AccessResult> getPermissionByRight(UserIdentity user, Set<GroupIdentity> userGroups, Object object)
    {
        return getPermissionByRight(user, userGroups, (T) object, _getPermissionContext((T) object));
    }
    
    /**
     * Gets the user access has on an object for all rights with the permission context
     * @param user The user. Cannot be null.
     * @param userGroups The groups the user belongs to
     * @param object The context object to check the access
     * @param permissionCtx The permisson context
     * @return the kind of access a user has on an object for all rights
     */
    protected Map<String, AccessResult> getPermissionByRight(UserIdentity user, Set<GroupIdentity> userGroups, T object, C permissionCtx)
    {
        Map<String, AccessResult> permissions = super.getPermissionByRight(user, userGroups, object);
        
        boolean inheritanceDisallowed = isInheritanceDisallowed(object);
        if (!inheritanceDisallowed)
        {
            Set<T> parents = _getParents(object, permissionCtx);
            if (parents != null && !_profileAssignmentStorageEP.isInheritanceDisallowed(object))
            {
                // Determine parents permissions
                Map<String, AccessResult> parentsPermissions = new HashMap<>();
                for (T parent : parents)
                {
                    Map<String, AccessResult> parentPermissions = getPermissionByRight(user, userGroups, parent, permissionCtx);
                    for (String rightId : parentPermissions.keySet())
                    {
                        parentsPermissions.put(rightId, AccessResult.merge(parentPermissions.get(rightId), parentsPermissions.get(rightId)));
                    }
                }
                
                // Apply it to local permission
                for (String rightId : parentsPermissions.keySet())
                {
                    if (!permissions.containsKey(rightId) || permissions.get(rightId) == AccessResult.UNKNOWN)
                    {
                        permissions.put(rightId, parentsPermissions.get(rightId));
                    }
                }
            }
        }
        
        return permissions;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected AccessResult _getPermission(UserIdentity user, Set<GroupIdentity> userGroups, Set<String> profilesIds, Object object, Object convertedObject)
    {
        return _getPermission(user, userGroups, profilesIds, (T) object, convertedObject, _getPermissionContext((T) object));
    }

    /**
     * Get user permission on an object with the permission context
     * @param user The use
     * @param userGroups The groups
     * @param profilesIds The profiles
     * @param object The original context
     * @param convertedObject The converted context
     * @param permissionCtx The permission context
     * @return the computed result
     */
    @SuppressWarnings("unchecked")
    protected AccessResult _getPermission(UserIdentity user, Set<GroupIdentity> userGroups, Set<String> profilesIds, T object, Object convertedObject, C permissionCtx)
    {
        // Let's manage the cache by ourselves, to avoid useless call to #_getParents
        // Try to retrieve in second cache (do not delegate the cache reading to the superclass because we known the unknown result is final
        Map<UserIdentity, AccessResult> cacheResult = (Map<UserIdentity, AccessResult>) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.USER);
        if (cacheResult != null && cacheResult.containsKey(user))
        {
            return cacheResult.get(user);
        }

        AccessResult permission = super._getPermission(user, userGroups, profilesIds, object, convertedObject);
        
        boolean inheritanceDisallowed = isInheritanceDisallowed(object);
        
        if (permission == AccessResult.UNKNOWN && !inheritanceDisallowed)
        {
            Set<T> parents = _getParents(object, permissionCtx);
            if (parents != null)
            {
                for (T parent : parents)
                {
                    Object convertedParentObject = _convertContext(parent);
                    AccessResult parentResult = _getPermission(user, userGroups, profilesIds, parent, convertedParentObject, permissionCtx);
                    permission = AccessResult.merge(permission, parentResult);
                }
                
                if (permission != AccessResult.UNKNOWN)
                {
                    if (cacheResult == null)
                    {
                        cacheResult = (Map<UserIdentity, AccessResult>) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.USER);
                        cacheResult = cacheResult == null ? new HashMap<>() : cacheResult;
                    }
                    cacheResult.put(user, permission);
                    _putInSecondCache(profilesIds, convertedObject, cacheResult, CacheKind.USER);
                }
            }
        }
        
        return permission;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected AccessResult _getPermissionForAnonymous(Set<String> profilesIds, Object object, Object convertedObject)
    {
        return _getPermissionForAnonymous(profilesIds, (T) object, convertedObject, _getPermissionContext((T) object));
    }
    
    /**
     * Get permission of anonymous on an object with the permission context
     * @param profilesIds The profiles ids
     * @param object The context
     * @param convertedObject The converted context
     * @param permissionCtx The permission context
     * @return The access result
     */
    protected AccessResult _getPermissionForAnonymous(Set<String> profilesIds, T object, Object convertedObject, C permissionCtx)
    {
        // Let's manage the cache by ourselves, to avoid useless call to #_getParents
        // Try to retrieve in second cache (do not delegate the cache reading to the superclass because we known the unknown result is final
        AccessResult cacheResult = (AccessResult) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.ANONYMOUS);
        if (cacheResult != null)
        {
            return cacheResult;
        }

        AccessResult permission = super._getPermissionForAnonymous(profilesIds, object, convertedObject);
        
        boolean inheritanceDisallowed = isInheritanceDisallowed(object);
        
        if (permission == AccessResult.UNKNOWN && !inheritanceDisallowed)
        {
            Set<T> parents = _getParents(object, permissionCtx);
            if (parents != null)
            {
                for (T parent : parents)
                {
                    Object convertedParentObject = _convertContext(parent);
                    AccessResult parentResult = _getPermissionForAnonymous(profilesIds, parent, convertedParentObject, permissionCtx);
                    permission = AccessResult.merge(permission, parentResult);
                }
                
                if (permission != AccessResult.UNKNOWN)
                {
                    // LET'S ADD THE ANSWER TO THE CACHE... because #_getParents can be long to execute
                    _putInSecondCache(profilesIds, convertedObject, permission, CacheKind.ANONYMOUS);
                }
            }
        }

        return permission;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected AccessResult _getPermissionForAnyConnectedUser(Set<String> profilesIds, Object object, Object convertedObject)
    {
        return _getPermissionForAnyConnectedUser(profilesIds, (T) object, convertedObject, _getPermissionContext((T) object));
    }
    
    /**
     * Get permission of anyconnected on an object with the permission context
     * @param profilesIds The profiles ids
     * @param object The context
     * @param convertedObject The converted context
     * @param permissionCtx The permission context
     * @return The access result
     */
    protected AccessResult _getPermissionForAnyConnectedUser(Set<String> profilesIds, T object, Object convertedObject, C permissionCtx)
    {
        // Let's manage the cache by ourselves, to avoid useless call to #_getParents
        // Try to retrieve in second cache (do not delegate the cache reading to the superclass because we known the unknown result is final
        AccessResult cacheResult = (AccessResult) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.ANY_CONNECTED_USER);
        if (cacheResult != null)
        {
            return cacheResult;
        }
        
        AccessResult permission = super._getPermissionForAnyConnectedUser(profilesIds, object, convertedObject);
        
        boolean inheritanceDisallowed = isInheritanceDisallowed(object);
        
        if (permission == AccessResult.UNKNOWN && !inheritanceDisallowed)
        {
            Set<T> parents = _getParents(object, permissionCtx);
            if (parents != null)
            {
                for (T parent : parents)
                {
                    Object convertedParentObject = _convertContext(parent);
                    AccessResult parentResult = _getPermissionForAnyConnectedUser(profilesIds, parent, convertedParentObject, permissionCtx);
                    permission = AccessResult.merge(permission, parentResult);
                }
                
                if (permission != AccessResult.UNKNOWN)
                {
                    // LET'S ADD THE ANSWER TO THE CACHE... because #_getParents can be long to execute
                    _putInSecondCache(profilesIds, convertedObject, permission, CacheKind.ANY_CONNECTED_USER);
                }
            }
        }

        return permission;
    }
    
    @SuppressWarnings("unchecked")
    @Override
    protected Map<UserIdentity, AccessResult> _getPermissionByUser(Set<String> profilesIds, Object object, Object convertedObject)
    {
        return _getPermissionByUser(profilesIds, (T) object, convertedObject, _getPermissionContext((T) object));
    }

    /**
     * Get the permission by users on an object with the permission context
     * @param profilesIds The profiles ids
     * @param object The context
     * @param convertedObject The converted context
     * @param permissionCtx The permission context
     * @return The users and their access results
     */
    protected Map<UserIdentity, AccessResult> _getPermissionByUser(Set<String> profilesIds, T object, Object convertedObject, C permissionCtx)
    {
        // Let's manage the cache by ourselves, to avoid useless call to #_getParents
        // Try to retrieve in second cache (do not delegate the cache reading to the superclass because we known the unknown result is final
        @SuppressWarnings("unchecked")
        Map<UserIdentity, AccessResult> cacheResult = (Map<UserIdentity, AccessResult>) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.USERS);
        if (cacheResult != null)
        {
            return cacheResult;
        }

        Map<UserIdentity, AccessResult> permissions = super._getPermissionByUser(profilesIds, object, convertedObject);
        
        boolean inheritanceDisallowed = isInheritanceDisallowed(object);
        if (!inheritanceDisallowed)
        {
            Set<T> parents = _getParents(object, permissionCtx);
            if (parents != null)
            {
                // Determine parents permissions
                Map<UserIdentity, AccessResult> parentsPermissions = new HashMap<>();
                for (T parent : parents)
                {
                    Object convertedParentObject = _convertContext(parent);
                    Map<UserIdentity, AccessResult> parentPermissions = _getPermissionByUser(profilesIds, parent, convertedParentObject, permissionCtx);
                    for (UserIdentity user : parentPermissions.keySet())
                    {
                        parentsPermissions.put(user, AccessResult.merge(parentPermissions.get(user), parentsPermissions.get(user)));
                    }
                }
                
                // Apply it to local permission
                for (UserIdentity user : parentsPermissions.keySet())
                {
                    if (!permissions.containsKey(user) || permissions.get(user) == AccessResult.UNKNOWN)
                    {
                        permissions.put(user, parentsPermissions.get(user));
                    }
                }
            }
        }
        
        _putInSecondCache(profilesIds, convertedObject, permissions, CacheKind.USERS);

        return permissions;
    }

    @SuppressWarnings("unchecked")
    @Override
    protected Map<GroupIdentity, AccessResult> _getPermissionByGroup(Set<String> profilesIds, Object object, Object convertedObject)
    {
        return _getPermissionByGroup(profilesIds, (T) object, convertedObject, _getPermissionContext((T) object));
    }
    
    /**
     * Get the permission by groups on an object with the permission context
     * @param profilesIds The profiles ids
     * @param object The context
     * @param convertedObject The converted context
     * @param permissionCtx The permission context
     * @return The users and their access results
     */
    protected Map<GroupIdentity, AccessResult> _getPermissionByGroup(Set<String> profilesIds, T object, Object convertedObject, C permissionCtx)
    {
        // Let's manage the cache by ourselves, to avoid useless call to #_getParents
        // Try to retrieve in second cache (do not delegate the cache reading to the superclass because we known the unknown result is final
        @SuppressWarnings("unchecked")
        Map<GroupIdentity, AccessResult> cacheResult = (Map<GroupIdentity, AccessResult>) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.GROUPS);
        if (cacheResult != null)
        {
            return cacheResult;
        }

        Map<GroupIdentity, AccessResult> permissions = super._getPermissionByGroup(profilesIds, object, convertedObject);
        
        boolean inheritanceDisallowed = isInheritanceDisallowed(object);
        if (!inheritanceDisallowed)
        {
            Set<T> parents = _getParents(object, permissionCtx);
            if (parents != null)
            {
                // Determine parents permissions
                Map<GroupIdentity, AccessResult> parentsPermissions = new HashMap<>();
                for (T parent : parents)
                {
                    Object convertedParentObject = _convertContext(parent);
                    Map<GroupIdentity, AccessResult> parentPermissions = _getPermissionByGroup(profilesIds, parent, convertedParentObject, permissionCtx);
                    for (GroupIdentity group : parentPermissions.keySet())
                    {
                        parentsPermissions.put(group, AccessResult.merge(parentPermissions.get(group), parentsPermissions.get(group)));
                    }
                }
                
                // Apply it to local permission
                for (GroupIdentity group : parentsPermissions.keySet())
                {
                    if (!permissions.containsKey(group) || permissions.get(group) == AccessResult.UNKNOWN)
                    {
                        permissions.put(group, parentsPermissions.get(group));
                    }
                }
            }
        }
        
        _putInSecondCache(profilesIds, convertedObject, permissions, CacheKind.GROUPS);

        return permissions;
    }
}
