/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.right;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.collections.CollectionUtils;

import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;
import org.ametys.core.group.GroupIdentity;
import org.ametys.core.right.AccessController;
import org.ametys.core.right.ProfileAssignmentStorageExtensionPoint;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightProfilesDAO;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.core.impl.cache.AbstractCacheKey;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * This class delegates all it can to the profile assignment storage extension point
 */
public abstract class AbstractProfileStorageBasedAccessController extends AbstractLogEnabled implements AccessController, Component, Serviceable, Initializable
{
    /**
     * The knd of cache to get/set
     */
    protected enum CacheKind
    {
        /** for anonymous user */
        ANONYMOUS,
        /** for any connected user */
        ANY_CONNECTED_USER,
        /** for users */
        USERS,
        /** for user */
        USER,
        /** for groups */
        GROUPS
    }
    
    /** The instance of ObjectUserIdentity for anonymous */
    protected static final UserIdentity __ANONYMOUS_USER_IDENTITY = null; 
    /** The instance of ObjectUserIdentity for any connected user */
    protected static final UserIdentity __ANY_CONTECTED_USER_IDENTITY = new UserIdentity(null, null);   

    /** The extension point for the profile assignment storages */
    protected ProfileAssignmentStorageExtensionPoint _profileAssignmentStorageEP;
    /** The right profile DAO */
    protected RightProfilesDAO _rightProfileDAO;
    /** Cache Manager */
    protected AbstractCacheManager _cacheManager;
    
    private final String _cache1 = RightManager.ROLE + "$" + this.getClass().getName() + "$Cache-1";
    
    private final String _cache2 = RightManager.ROLE + "$" + this.getClass().getName() + "$Cache-2";

    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _rightProfileDAO = (RightProfilesDAO) manager.lookup(RightProfilesDAO.ROLE);
        _profileAssignmentStorageEP = (ProfileAssignmentStorageExtensionPoint) manager.lookup(ProfileAssignmentStorageExtensionPoint.ROLE);
        _cacheManager = (AbstractCacheManager) manager.lookup(AbstractCacheManager.ROLE);
    }

    public void initialize() throws Exception
    {
        if (!_cacheManager.hasCache(_cache1))
        {
            _cacheManager.createRequestCache(_cache1, 
                    _buildI18n("PLUGINS_CORE_RIGHT_CACHE_1_LABEL"),
                    _buildI18n("PLUGINS_CORE_RIGHT_CACHE_1_DESCRIPTION"),
                    true);
        }

        if (!_cacheManager.hasCache(_cache2))
        {
            _cacheManager.createRequestCache(_cache2, 
                    _buildI18n("PLUGINS_CORE_RIGHT_CACHE_2_LABEL"),
                    _buildI18n("PLUGINS_CORE_RIGHT_CACHE_2_DESCRIPTION"),
                    true);
        }
       
    }

    private I18nizableText _buildI18n(String i18Key)
    {
        String catalogue = "plugin.core";
        I18nizableText className = new I18nizableText(this.getClass().getSimpleName());
        Map<String, I18nizableTextParameter> params = Map.of("id", className);
        return new I18nizableText(catalogue, i18Key, params);
    }
    
    public Map<String, AccessResult> getPermissionByRight(UserIdentity user, Set<GroupIdentity> userGroups, Object object)
    {
        Map<String, AccessResult> rights = new HashMap<>();
        
        Map<String, AccessResult> permissionsByProfile = _profileAssignmentStorageEP.getPermissionsByProfile(user, userGroups, _convertContext(object));
        
        // Convert profile <-> accessresult to right <-> accessresult
        for (String profileId : permissionsByProfile.keySet())
        {
            List<String> rights2 = _rightProfileDAO.getRights(profileId);
            for (String rightId : rights2)
            {
                rights.put(rightId, AccessResult.merge(permissionsByProfile.get(profileId), rights.get(rightId)));
            }
        }

        return rights;
    }
    
    public AccessResult getPermission(UserIdentity user, Set<GroupIdentity> userGroups, String rightId, Object object)
    {
        Set<String> profilesIds = _rightProfileDAO.getProfilesWithRight(rightId);
        if (profilesIds == null || profilesIds.isEmpty())
        {
            // No need for cache...
            return AccessResult.UNKNOWN;
        }
        else
        {
            Object convertedObject = _convertContext(object);
            return _getPermission(user, userGroups, profilesIds, object, convertedObject);
        }
    }
    public AccessResult getReadAccessPermission(UserIdentity user, Set<GroupIdentity> userGroups, Object object)
    {
        Set<String> profilesIds = Collections.singleton(RightManager.READER_PROFILE_ID);
        // Some converted context may change so cache must be done with the convertedObject
        Object convertedObject = _convertContext(object);
        return _getPermission(user, userGroups, profilesIds, object, convertedObject);
    }
    /**
     * Works for getPermission or getReadAccessPermission
     * @param user The use
     * @param userGroups The groups
     * @param profilesIds The profiles
     * @param object The original context
     * @param convertedObject The converted context
     * @return the computed result
     */
    protected AccessResult _getPermission(UserIdentity user, Set<GroupIdentity> userGroups, Set<String> profilesIds, Object object, Object convertedObject)
    {
        @SuppressWarnings("unchecked")
        Map<UserIdentity, AccessResult> cacheResult = (Map<UserIdentity, AccessResult>) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.USER);
        if (cacheResult != null && cacheResult.containsKey(user))
        {
            return cacheResult.get(user);
        }

        Map<String, AccessResult> permissions = _profileAssignmentStorageEP.getPermissions(user, userGroups, profilesIds, convertedObject);
        AccessResult result = AccessResult.merge(permissions.values());
        cacheResult = cacheResult == null ? new HashMap<>() : cacheResult;
        cacheResult.put(user, result);
        _putInSecondCache(profilesIds, convertedObject, cacheResult, CacheKind.USER);
        return result;
    }

    public AccessResult getPermissionForAnonymous(String rightId, Object object)
    {
        Set<String> profilesIds = _rightProfileDAO.getProfilesWithRight(rightId);
        if (profilesIds == null || profilesIds.isEmpty())
        {
            // No need for cache...
            return AccessResult.UNKNOWN;
        }
        else
        {
            // Some converted context may change so cache must be done with the convertedObject
            Object convertedObject = _convertContext(object);
            return _getPermissionForAnonymous(profilesIds, object, convertedObject);
        }
    }
    public AccessResult getReadAccessPermissionForAnonymous(Object object)
    {
        Set<String> profilesIds = Collections.singleton(RightManager.READER_PROFILE_ID);
        // Some converted context may change so cache must be done with the convertedObject
        Object convertedObject = _convertContext(object);
        return _getPermissionForAnonymous(profilesIds, object, convertedObject);
    }
    /**
     * Works for getPermissionForAnonymous and getReadAccessPermissionForAnonymous
     * @param profilesIds The profiles ids
     * @param object The context
     * @param convertedObject The converted context
     * @return The access result
     */
    protected AccessResult _getPermissionForAnonymous(Set<String> profilesIds, Object object, Object convertedObject)
    {
        // Try to retrieve in second cache
        AccessResult cacheResult = (AccessResult) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.ANONYMOUS);
        if (cacheResult != null)
        {
            return cacheResult;
        }
        
        AccessResult result = _profileAssignmentStorageEP.getPermissionForAnonymous(profilesIds, convertedObject);
        _putInSecondCache(profilesIds, convertedObject, result, CacheKind.ANONYMOUS);
        return result;
    }
    
    public AccessResult getPermissionForAnyConnectedUser(String rightId, Object object)
    {
        Set<String> profilesIds = _rightProfileDAO.getProfilesWithRight(rightId);
        if (profilesIds == null || profilesIds.isEmpty())
        {
            // No need for cache...
            return AccessResult.UNKNOWN;
        }
        else
        {
            // Some converted context may change so cache must be done with the convertedObject
            Object convertedObject = _convertContext(object);
            return _getPermissionForAnyConnectedUser(profilesIds, object, convertedObject);
        }
    }
    public AccessResult getReadAccessPermissionForAnyConnectedUser(Object object)
    {
        Set<String> profilesIds = Collections.singleton(RightManager.READER_PROFILE_ID);
        // Some converted context may change so cache must be done with the convertedObject
        Object convertedObject = _convertContext(object);
        return _getPermissionForAnyConnectedUser(profilesIds, object, convertedObject);
    }
    /**
     * Works for getPermissionForAnyConnectedUser and getReadAccessPermissionForAnyConnectedUser
     * @param profilesIds The profiles ids
     * @param object The context
     * @param convertedObject The converted context
     * @return the access result
     */
    protected AccessResult _getPermissionForAnyConnectedUser(Set<String> profilesIds, Object object, Object convertedObject)
    {
        // Try to retrieve in second cache
        AccessResult cacheResult = (AccessResult) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.ANY_CONNECTED_USER);
        if (cacheResult != null)
        {
            return cacheResult;
        }
        
        AccessResult result = _profileAssignmentStorageEP.getPermissionForAnyConnectedUser(profilesIds, convertedObject);
        _putInSecondCache(profilesIds, convertedObject, result, CacheKind.ANY_CONNECTED_USER);
        return result;
    }

    public Map<UserIdentity, AccessResult> getPermissionByUser(String rightId, Object object)
    {
        Set<String> profilesIds = _rightProfileDAO.getProfilesWithRight(rightId);
        if (profilesIds == null || profilesIds.isEmpty())
        {
            // No need for cache...
            return Collections.EMPTY_MAP;
        }
        else
        {
            // Some converted context may change so cache must be done with the convertedObject
            Object convertedObject = _convertContext(object);
            return _getPermissionByUser(profilesIds, object, convertedObject);
        }
    }
    public Map<UserIdentity, AccessResult> getReadAccessPermissionByUser(Object object)
    {
        Set<String> profilesIds = Collections.singleton(RightManager.READER_PROFILE_ID);
        // Some converted context may change so cache must be done with the convertedObject
        Object convertedObject = _convertContext(object);
        return _getPermissionByUser(profilesIds, object, convertedObject);
    }
    /**
     * Works for getPermissionByUser and getReadAccessPermissionByUser
     * @param profilesIds The profiles ids
     * @param object The context
     * @param convertedObject The converted context
     * @return The users and their access results
     */
    protected Map<UserIdentity, AccessResult> _getPermissionByUser(Set<String> profilesIds, Object object, Object convertedObject)
    {
        // Try to retrieve in second cache
        @SuppressWarnings("unchecked")
        Map<UserIdentity, AccessResult> cacheResult = (Map<UserIdentity, AccessResult>) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.USERS);
        if (cacheResult != null)
        {
            return cacheResult;
        }
        
        Map<UserIdentity, AccessResult> result = _profileAssignmentStorageEP.getPermissionsByUser(profilesIds, convertedObject);
        _putInSecondCache(profilesIds, convertedObject, result, CacheKind.USERS);
        return result;
    }
    
    public Map<GroupIdentity, AccessResult> getPermissionByGroup(String rightId, Object object)
    {
        Set<String> profilesIds = _rightProfileDAO.getProfilesWithRight(rightId);
        if (profilesIds == null || profilesIds.isEmpty())
        {
            // No need for cache...
            return Collections.EMPTY_MAP;
        }
        else
        {
            // Some converted context may change so cache must be done with the convertedObject
            Object convertedObject = _convertContext(object);
            return _getPermissionByGroup(profilesIds, object, convertedObject);
        }
    }
    public Map<GroupIdentity, AccessResult> getReadAccessPermissionByGroup(Object object)
    {
        Set<String> profilesIds = Collections.singleton(RightManager.READER_PROFILE_ID);
        // Some converted context may change so cache must be done with the convertedObject
        Object convertedObject = _convertContext(object);
        return _getPermissionByGroup(profilesIds, object, convertedObject);
    }
    /**
     * Works for getPermissionByGroup and getReadAccessPermissionByGroup
     * @param profilesIds The profiles ids
     * @param object The context
     * @param convertedObject The converted context
     * @return The users and their access results
     */
    protected Map<GroupIdentity, AccessResult> _getPermissionByGroup(Set<String> profilesIds, Object object, Object convertedObject)
    {
        // Try to retrieve in second cache
        @SuppressWarnings("unchecked")
        Map<GroupIdentity, AccessResult> cacheResult = (Map<GroupIdentity, AccessResult>) _hasRightResultInSecondCache(convertedObject, profilesIds, CacheKind.GROUPS);
        if (cacheResult != null)
        {
            return cacheResult;
        }
        
        Map<GroupIdentity, AccessResult> result = _profileAssignmentStorageEP.getPermissionsByGroup(profilesIds, convertedObject);
        _putInSecondCache(profilesIds, convertedObject, result, CacheKind.GROUPS);
        return result;
    }

    /**
     * For methods getXXXXPermissionYYY allow to have a modification of the context before transfering it to the profile assignment storage extension point
     * The default implemenation keep the context as it is
     * @param initialContext The right context that is supported
     * @return the context modified
     */
    protected Object _convertContext(Object initialContext)
    {
        return initialContext;
    }
    
    public boolean hasAnonymousAnyReadAccessPermissionOnWorkspace(Set<Object> workspacesContexts)
    {
        Set<String> profilesIds = Collections.singleton(RightManager.READER_PROFILE_ID);
        return _hasAnonymousAnyPermissionOnWorkspace(workspacesContexts, profilesIds);
    }
    public boolean hasAnonymousAnyPermissionOnWorkspace(Set<Object> workspacesContexts, String rightId)
    {
        Set<String> profilesIds = _rightProfileDAO.getProfilesWithRight(rightId);
        return _hasAnonymousAnyPermissionOnWorkspace(workspacesContexts, profilesIds);
    }
    private boolean _hasAnonymousAnyPermissionOnWorkspace(Set<Object> workspacesContexts, Set<String> profilesIds)
    {
        // Try to retrieve in first cache (the one which manages non-null contexts)
        Boolean cacheResult = _hasRightResultInFirstCache(__ANONYMOUS_USER_IDENTITY, profilesIds, workspacesContexts);
        if (cacheResult != null)
        {
            return cacheResult;
        }

        // Otherwise continue
        Set<? extends Object> rootContexts = _convertWorkspaceToRootRightContexts(workspacesContexts);
        
        Set<String> determinedProfiles = profilesIds;
        boolean rightResult = false;
        
        if (CollectionUtils.isNotEmpty(rootContexts))
        {
            Set<String> hasAnonymousAnyPermission = _profileAssignmentStorageEP.hasAnonymousAnyPermission(rootContexts, profilesIds);
            if (!hasAnonymousAnyPermission.isEmpty())
            {
                rightResult = true;
                determinedProfiles = hasAnonymousAnyPermission;
            }
        }
        
        _putInFirstCache(__ANONYMOUS_USER_IDENTITY, determinedProfiles, workspacesContexts, rightResult);
        return rightResult;
    }

    public boolean hasAnyConnectedUserAnyReadAccessPermissionOnWorkspace(Set<Object> workspacesContexts)
    {
        Set<String> profilesIds = Collections.singleton(RightManager.READER_PROFILE_ID);
        return _hasAnyConnectedUserAnyPermissionOnWorkspace(workspacesContexts, profilesIds);
    }
    public boolean hasAnyConnectedUserAnyPermissionOnWorkspace(Set<Object> workspacesContexts, String rightId)
    {
        Set<String> profilesIds = _rightProfileDAO.getProfilesWithRight(rightId);
        return _hasAnyConnectedUserAnyPermissionOnWorkspace(workspacesContexts, profilesIds);
    }
    private boolean _hasAnyConnectedUserAnyPermissionOnWorkspace(Set<Object> workspacesContexts, Set<String> profilesIds)
    {
        // Try to retrieve in first cache (the one which manages non-null contexts)
        Boolean cacheResult = _hasRightResultInFirstCache(__ANY_CONTECTED_USER_IDENTITY, profilesIds, workspacesContexts);
        if (cacheResult != null)
        {
            return cacheResult;
        }

        // Otherwise continue
        Set<? extends Object> rootContexts = _convertWorkspaceToRootRightContexts(workspacesContexts);
        
        Set<String> determinedProfiles = profilesIds;
        boolean rightResult = false;
        
        if (CollectionUtils.isNotEmpty(rootContexts))
        {
            Set<String> hasAnyConnectedUserAnyPermission = _profileAssignmentStorageEP.hasAnyConnectedUserAnyPermission(rootContexts, profilesIds);
            if (!hasAnyConnectedUserAnyPermission.isEmpty())
            {
                rightResult = true;
                determinedProfiles = hasAnyConnectedUserAnyPermission;
            }
        }
        
        _putInFirstCache(__ANY_CONTECTED_USER_IDENTITY, determinedProfiles, workspacesContexts, rightResult);
        return rightResult;        
    }
    
    public boolean hasUserAnyReadAccessPermissionOnWorkspace(Set<Object> workspacesContexts, UserIdentity user, Set<GroupIdentity> userGroups)
    {
        Set<String> profilesIds = Collections.singleton(RightManager.READER_PROFILE_ID);
        return _hasUserAnyPermissionOnWorkspace(workspacesContexts, user, userGroups, profilesIds);
    }
    public boolean hasUserAnyPermissionOnWorkspace(Set<Object> workspacesContexts, UserIdentity user, Set<GroupIdentity> userGroups, String rightId)
    {
        Set<String> profilesIds = _rightProfileDAO.getProfilesWithRight(rightId);
        return _hasUserAnyPermissionOnWorkspace(workspacesContexts, user, userGroups, profilesIds);
    }
    private boolean _hasUserAnyPermissionOnWorkspace(Set<Object> workspacesContexts, UserIdentity user, Set<GroupIdentity> userGroups, Set<String> profilesIds)
    {
        // Try to retrieve in first cache (the one which manages non-null contexts)
        Boolean cacheResult = _hasRightResultInFirstCache(user, profilesIds, workspacesContexts);
        if (cacheResult != null)
        {
            return cacheResult;
        }

        // Otherwise continue
        Set<? extends Object> rootContexts = _convertWorkspaceToRootRightContexts(workspacesContexts);
        
        Set<String> determinedProfiles = profilesIds;
        boolean rightResult = false;
        
        if (CollectionUtils.isNotEmpty(rootContexts))
        {
            Set<String> hasUserAnyPermission = _profileAssignmentStorageEP.hasUserAnyPermission(rootContexts, user, userGroups, profilesIds);
            if (!hasUserAnyPermission.isEmpty())
            {
                rightResult = true;
                determinedProfiles = hasUserAnyPermission;
            }
        }
        
        _putInFirstCache(user, determinedProfiles, workspacesContexts, rightResult);
        return rightResult;         
    }
    
    /**
     * Get the current workspaces contexts and turn it into root contexts in order to allow methods hasXXXAnyPermissionOnWorkspace to work
     * @param workspacesContexts The workspace contexts. Such as '/${WorkspaceName}', '/admin'
     * @return A null or empty set if the current AccessController does not apply to any workspace context, or the root object where ProfileAssignmentStorageExtension should start looking at to find any permission
     */
    protected abstract Set<? extends Object> _convertWorkspaceToRootRightContexts(Set<Object> workspacesContexts);
    
    
    
    

    
    /**
     * Seek in cache
     * @param userIdentity The user identity or AbstractProfileStorageBasedAccessController.__ANONYMOUS_USER_IDENTITY or AbstractProfileStorageBasedAccessController.__ANY_CONTECTED_USER_IDENTITY
     * @param profilesIds The profiles identifiers
     * @param object The context
     * @return true or false if in cache. null otherwise
     */
    protected Boolean _hasRightResultInFirstCache(UserIdentity userIdentity, Set<String> profilesIds, Object object)
    {
        // On the contrary of the other cache, this one is split between profiles, as we check for the first true result (no negativity)
        
        if (profilesIds == null || profilesIds.isEmpty())
        {
            // No need for cache...
            return false;
        }
        
        Cache<Cache1Key, Boolean> mapCache = _cacheManager.get(_cache1);
        if (mapCache != null)
        {
            int negativeProfiles = 0;
            
            for (String profileId: profilesIds)
            {
                Cache1Key key = Cache1Key.of(userIdentity, profileId, object);
                if (mapCache.hasKey(key))
                {
                    if (mapCache.get(key))
                    {
                        getLogger().debug("Find entry in cache for [{}, {}, {}] => true", userIdentity, profilesIds, object);
                        return true;
                    }
                    else
                    {
                        negativeProfiles++;
                    }
                }
            }
            
            if (negativeProfiles == profilesIds.size())
            {
                // All required profiles were negative in cache
                return false;
            }  
        }
        getLogger().debug("Did not find entry in cache for [{}, {}, {}]", userIdentity, profilesIds, object);
        return null;
    }
    
    /**
     * Add to cache
     * @param userIdentity The user identity or AbstractProfileStorageBasedAccessController.__ANONYMOUS_USER_IDENTITY or AbstractProfileStorageBasedAccessController.__ANY_CONTECTED_USER_IDENTITY
     * @param profilesIds The profiles identifiers
     * @param object The context
     * @param rightResult The cache value. true if hasXXX or false otherwise.
     */
    protected void _putInFirstCache(UserIdentity userIdentity, Set<String> profilesIds, Object object, boolean rightResult)
    {
        // On the contrary of the other cache, this one is split between profiles, as we check for the first true result (no negativity)

        Cache<Cache1Key, Boolean> mapCache = _cacheManager.get(_cache1);
        if (mapCache != null)
        {
            for (String profileId: profilesIds)
            {
                mapCache.put(Cache1Key.of(userIdentity, profileId, object), rightResult);
            }   
        }
    }

    /**
     * Seek in cache
     * @param object The context
     * @param profilesIds The set of profile ids to consider
     * @return The cached result per group. null otherwise
     * @param key The kind of cache to use
     */
    protected Object _hasRightResultInSecondCache(Object object, Set<String> profilesIds, CacheKind key)
    {
        Cache<Cache2Key, Object> mapCache = _cacheManager.get(_cache2);

        Object cacheResult = mapCache.get(Cache2Key.of(profilesIds, object, key));
        if (cacheResult != null)
        {
            getLogger().debug("Find entry in cache for [{}, {}, {}] => {}", profilesIds, object, key, cacheResult);
            return cacheResult;
        }

        getLogger().debug("Did not find entry in cache for [{}, {}, {}]", profilesIds, object, key);
        return null;
    }

    /**
     * Add to cache
     * @param profilesIds The profiles ids to consider
     * @param object The context
     * @param result The result
     * @param key The kind of cache to use
     */
    protected void _putInSecondCache(Set<String> profilesIds, Object object, Object result, CacheKind key)
    {
        Cache<Cache2Key, Object> mapCache = _cacheManager.get(_cache2);

        if (mapCache != null)
        {
            mapCache.put(Cache2Key.of(profilesIds, object, key), result);
        }
        
    }

    static class Cache1Key extends AbstractCacheKey
    {
        Cache1Key(UserIdentity userIdentity, String profileId, Object object)
        {
            super(userIdentity, profileId, object);
        }

        static Cache1Key of(UserIdentity userIdentity, String profileId, Object object)
        {
            return new Cache1Key(userIdentity, profileId, object);
        } 
        
    }
    static class Cache2Key extends AbstractCacheKey
    {
        Cache2Key(Set<String> profileIds, Object object, CacheKind cacheKind)
        {
            super(profileIds, object, cacheKind);
        }

        static Cache2Key of(Set<String> profileIds, Object object, CacheKind cacheKind)
        {
            return new Cache2Key(profileIds, object, cacheKind);
        } 
        
    }
   
}
