/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.right;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.ibatis.session.SqlSession;

import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;
import org.ametys.core.datasource.AbstractMyBatisDAO;
import org.ametys.core.group.GroupIdentity;
import org.ametys.core.right.ModifiableProfileAssignmentStorage;
import org.ametys.core.right.ProfileAssignmentStorage;
import org.ametys.core.user.UserIdentity;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Jdbc implementation of {@link ProfileAssignmentStorage} which stores profile assignments in database.
 * This only supports String objects as contexts.
 */
public class JdbcProfileAssignmentStorage extends AbstractMyBatisDAO implements ModifiableProfileAssignmentStorage, Initializable
{   

    private static final String __JDBC_PROFILE_ASSIGNMENT_STORAGE_CACHE = JdbcProfileAssignmentStorage.class.getName() + "$storage";
    
    private static final String __JDBC_PROFILE_ASSIGNMENT_INHERITANCE_CACHE = JdbcProfileAssignmentStorage.class.getName() + "$inheritance";
    
    /** The handled context */
    protected String _supportedContext;
    
    private AbstractCacheManager _cacheManager;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _cacheManager = (AbstractCacheManager) _manager.lookup(AbstractCacheManager.ROLE);
    }
      
    @Override
    public synchronized void initialize() throws Exception
    {
        if (!_cacheManager.hasCache(__JDBC_PROFILE_ASSIGNMENT_STORAGE_CACHE))
        {
            _cacheManager.createMemoryCache(
                    __JDBC_PROFILE_ASSIGNMENT_STORAGE_CACHE,
                    new I18nizableText("plugin.admin", "PLUGINS_ADMIN_RIGHTS_JDBC_PROFILE_ASSIGNMENT_STORAGE_CACHE_LABEL"),
                    new I18nizableText("plugin.admin", "PLUGINS_ADMIN_RIGHTS_JDBC_PROFILE_ASSIGNMENT_STORAGE_CACHE_DESCRIPTION"),
                    true, 
                    null);
        }

        if (!_cacheManager.hasCache(__JDBC_PROFILE_ASSIGNMENT_INHERITANCE_CACHE))
        {
            _cacheManager.createMemoryCache(
                    __JDBC_PROFILE_ASSIGNMENT_INHERITANCE_CACHE,
                    new I18nizableText("plugin.admin", "PLUGINS_ADMIN_RIGHTS_JDBC_PROFILE_ASSIGNMENT_INHERITANCE_CACHE_LABEL"),
                    new I18nizableText("plugin.admin", "PLUGINS_ADMIN_RIGHTS_JDBC_PROFILE_ASSIGNMENT_INHERITANCE_CACHE_DESCRIPTION"),
                    true, 
                    null);
        }
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        super.configure(configuration);
        _supportedContext = configuration.getChild("context").getValue();
    }
    
    /**
     * Dump the SQL database in an cached object
     * @param context The context to seek
     * @return The database
     */
    protected synchronized Database _getFullData(Object context)
    {
        String rootContext = "/" + StringUtils.split((String) context, '/')[0];
     
        return _getStorageCache().get(rootContext, key -> 
        {

            try (SqlSession session = getSession())
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("context", rootContext);
                
                List<Map<String, String>> profiles = session.selectList("ProfilesAssignment.getAllProfiles", parameters);
                Database db = new Database(profiles);
                
                return db;
            }
        });
        
    }
    
    /**
     * Clear the cache of _getFullData
     */
    protected synchronized void _clearCache()
    {
        _getStorageCache().invalidateAll();
    }
    
    /**
     * Clear the cache of _getFullData
     * @param context The context to seek
     */
    protected synchronized void _clearCache(Object context)
    {
        String rootContext = "/" + StringUtils.split((String) context, '/')[0];
    
        if (_getStorageCache().hasKey(rootContext))
        {
            _getStorageCache().invalidate(rootContext);
        }
    }
    
    /**
     * Get the full inheritances from SQL database in an cached object
     * @return The inheritances
     */
    protected synchronized Cache<String, Boolean> _getInheritances()
    {
        Cache<String, Boolean> inheritanceCache = _getInheritanceCache();
        if (!inheritanceCache.isInitialized())
        {
            try (SqlSession session = getSession())
            {
                List<Map<String, Object>> inheritances = session.selectList("ProfilesAssignment.getInheritances", new HashMap<>());
                
                inheritances.stream().forEach(result -> 
                {
                    inheritanceCache.put((String) result.get("context"), (Boolean) result.get("disallow"));
                });
            }
        }
        
        return inheritanceCache;
    }
    
    /**
     * Clear the cache of _getFullData
     */
    protected synchronized void _clearInheritanceCache()
    {
        _getInheritanceCache().invalidateAll();
    }
    
    /* -------------- */
    /* HAS PERMISSION */
    /* -------------- */
    
    /**
     * Get the object context with prefix if necessary
     * @param context The context object
     * @return The prefixed object
     */
    protected Object getObjectWithPrefix (Object context)
    {
        return context;
    }
    
    /**
     * Get the prefix for object context 
     * @return The prefix. Can be null if no prefix is necessary
     */
    protected String getPrefix ()
    {
        return null;
    }

    public Set<String> hasUserAnyAllowedProfile(Set<? extends Object> rootContexts, UserIdentity user, Set<String> profileIds)
    {
        String prefix = getPrefix();
        Set<String> stringPrefixContexts = rootContexts.stream().filter(String.class::isInstance).map(context -> (prefix != null ? prefix : "") + (String) context).collect(Collectors.toSet());

        Set<String> anyAllowedProfiles = new HashSet<>();
        for (String stringPrefixContext : stringPrefixContexts)
        {
            Set<String> profiles = _getFullData(stringPrefixContext)
                .getAlloweUserData()
                .entrySet().stream()
                .filter(entry -> _equalsOrStartsWith(entry.getKey(), stringPrefixContext))
                .map(Map.Entry::getValue)
                .map(m -> m.get(user))
                .filter(Objects::nonNull)
                .flatMap(Set::stream)
                .collect(Collectors.toSet());
            
            anyAllowedProfiles.addAll(profiles);
        }
        
        for (String profileId : profileIds)
        {
            if (anyAllowedProfiles.contains(profileId))
            {
                return anyAllowedProfiles;
            }
        }
        return Set.of();
    }

    public Set<String> hasGroupAnyAllowedProfile(Set<? extends Object> rootContexts, Set<GroupIdentity> groups, Set<String> profileIds)
    {
        String prefix = getPrefix();
        Set<String> stringPrefixContexts = rootContexts.stream().filter(String.class::isInstance).map(context -> (prefix != null ? prefix : "") + (String) context).collect(Collectors.toSet());

        Set<String> anyAllowedProfiles = new HashSet<>();
        for (String stringPrefixContext : stringPrefixContexts)
        {
            _getFullData(stringPrefixContext)
                .getAlloweGroupData()
                .entrySet()
                .stream()
                .filter(entry -> _equalsOrStartsWith(entry.getKey(), stringPrefixContext))
                .map(Map.Entry::getValue)
                .forEach(profilesByGroup -> 
                {
                    for (GroupIdentity group : groups)
                    {
                        Set<String> profiles = profilesByGroup.get(group);
                        if (profiles != null)
                        {
                            anyAllowedProfiles.addAll(profiles);
                        }
                    }
                });
        }
        
        for (String profileId : profileIds)
        {
            if (anyAllowedProfiles.contains(profileId))
            {
                return anyAllowedProfiles;
            }
        }
        
        return Set.of();
    }

    public Set<String> hasAnyConnectedAnyAllowedProfile(Set<? extends Object> rootContexts, Set<String> profileIds)
    {
        String prefix = getPrefix();
        Set<String> stringPrefixContexts = rootContexts.stream().filter(String.class::isInstance).map(context -> (prefix != null ? prefix : "") + (String) context).collect(Collectors.toSet());

        Set<String> anyAllowedProfiles = new HashSet<>();
        for (String stringPrefixContext : stringPrefixContexts)
        {
            Set<String> profiles = _getFullData(stringPrefixContext)
                .getAllowedAnyConnected()
                .entrySet().stream()
                .filter(entry -> _equalsOrStartsWith(entry.getKey(), stringPrefixContext))
                .map(Map.Entry::getValue)
                .flatMap(Set::stream)
                .collect(Collectors.toSet());
            
            anyAllowedProfiles.addAll(profiles);
        }
        
        for (String profileId : profileIds)
        {
            if (anyAllowedProfiles.contains(profileId))
            {
                return anyAllowedProfiles;
            }
        }
        return Set.of();
    }
    
    public Set<String> hasAnonymousAnyAllowedProfile(Set<? extends Object> rootContexts, Set<String> profileIds)
    {
        String prefix = getPrefix();
        Set<String> stringPrefixContexts = rootContexts.stream().filter(String.class::isInstance).map(context -> (prefix != null ? prefix : "") + (String) context).collect(Collectors.toSet());
        
        Set<String> anyAllowedProfiles = new HashSet<>();
        for (String stringPrefixContext : stringPrefixContexts)
        {
            Set<String> profiles = _getFullData(stringPrefixContext)
                .getAllowedAnonymous()
                .entrySet().stream()
                .filter(entry -> _equalsOrStartsWith(entry.getKey(), stringPrefixContext))
                .map(Map.Entry::getValue)
                .flatMap(Set::stream)
                .collect(Collectors.toSet());
            
            anyAllowedProfiles.addAll(profiles);
        }
        
        for (String profileId : profileIds)
        {
            if (anyAllowedProfiles.contains(profileId))
            {
                return anyAllowedProfiles;
            }
        }
        return Set.of();
    }
    
    private static boolean _equalsOrStartsWith(String value, String prefix)
    {
        return value.equals(prefix) || value.startsWith(prefix + "/");
    }
    
    public Map<Object, Map<AnonymousOrAnyConnectedKeys, Set<String>>> getAllProfilesForAnonymousAndAnyConnectedUser()
    {
        Map<Object, Map<AnonymousOrAnyConnectedKeys, Set<String>>> result = new HashMap<>();
        Database databaseValue = _getFullData(_supportedContext);
        databaseValue.getAllowedAnonymous()
            .entrySet().stream()
            .forEach(entry -> {
                Set<String> allowedProfiles = entry.getValue();
                if (!allowedProfiles.isEmpty())
                {
                    Map<AnonymousOrAnyConnectedKeys, Set<String>> contextMap = result.computeIfAbsent(entry.getKey(), k -> new HashMap<>());
                    contextMap.put(AnonymousOrAnyConnectedKeys.ANONYMOUS_ALLOWED, allowedProfiles);
                }
            });
            
        databaseValue.getDeniedAnonymous()
            .entrySet().stream()
            .forEach(entry -> {
                Set<String> deniedProfiles = entry.getValue();
                if (!deniedProfiles.isEmpty())
                {
                    Map<AnonymousOrAnyConnectedKeys, Set<String>> contextMap = result.computeIfAbsent(entry.getKey(), k -> new HashMap<>());
                    contextMap.put(AnonymousOrAnyConnectedKeys.ANONYMOUS_DENIED, deniedProfiles);
                }
            });
        
        databaseValue.getAllowedAnyConnected()
            .entrySet().stream()
            .forEach(entry -> {
                Set<String> allowedProfiles = entry.getValue();
                if (!allowedProfiles.isEmpty())
                {
                    Map<AnonymousOrAnyConnectedKeys, Set<String>> contextMap = result.computeIfAbsent(entry.getKey(), k -> new HashMap<>());
                    contextMap.put(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_ALLOWED, allowedProfiles);
                }
            });
        
        databaseValue.getDeniedAnyConnected()
            .entrySet().stream()
            .forEach(entry -> {
                Set<String> deniedProfiles = entry.getValue();
                if (!deniedProfiles.isEmpty())
                {
                    Map<AnonymousOrAnyConnectedKeys, Set<String>> contextMap = result.computeIfAbsent(entry.getKey(), k -> new HashMap<>());
                    contextMap.put(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_DENIED, deniedProfiles);
                }
            });
        
        return result;
    }
    
    public Map<Object, Map<GroupIdentity, Map<UserOrGroup, Set<String>>>> getAllProfilesForGroups(Set<GroupIdentity> groups)
    {
        Map<Object, Map<GroupIdentity, Map<UserOrGroup, Set<String>>>> result = new HashMap<>();

        Database databaseValue = _getFullData(_supportedContext);
        
        databaseValue.getAlloweGroupData()
            .entrySet().stream()
            .filter(entry -> CollectionUtils.containsAny(entry.getValue().keySet(), groups))
            .forEach(entry -> {
                Map<GroupIdentity, Set<String>> assignments = entry.getValue();
                for (GroupIdentity group : assignments.keySet())
                {
                    if (groups.contains(group))
                    {
                        Set<String> allowed = assignments.get(group);
                        if (allowed != null && !allowed.isEmpty())
                        {
                            Map<GroupIdentity, Map<UserOrGroup, Set<String>>> contextMap = result.computeIfAbsent(entry.getKey(), k -> new HashMap<>());
                            Map<UserOrGroup, Set<String>> groupAssignements = contextMap.computeIfAbsent(group, k -> new HashMap<>());
                            groupAssignements.put(UserOrGroup.ALLOWED, allowed);
                        }
                    }
                }
            });
        
        databaseValue.getDeniedGroupData()
            .entrySet().stream()
            .filter(entry -> CollectionUtils.containsAny(entry.getValue().keySet(), groups))
            .forEach(entry -> {
                Map<GroupIdentity, Set<String>> assignments = entry.getValue();
                for (GroupIdentity group : assignments.keySet())
                {
                    if (groups.contains(group))
                    {
                        Set<String> denied = assignments.get(group);
                        if (denied != null && !denied.isEmpty())
                        {
                            Map<GroupIdentity, Map<UserOrGroup, Set<String>>> contextMap = result.computeIfAbsent(entry.getKey(), k -> new HashMap<>());
                            Map<UserOrGroup, Set<String>> groupAssignements = contextMap.computeIfAbsent(group, k -> new HashMap<>());
                            groupAssignements.put(UserOrGroup.DENIED, denied);
                        }
                        
                    }
                }
            });
        
        return result;
    }
    
    public Map<Object, Map<UserOrGroup, Set<String>>> getAllProfilesForUser(UserIdentity user)
    {
        Map<Object, Map<UserOrGroup, Set<String>>> result = new HashMap<>();
        Database databaseValue = _getFullData(_supportedContext);
        databaseValue
            .getAlloweUserData()
            .entrySet().stream()
            .filter(entry -> entry.getValue().keySet().contains(user))
            .forEach(entry -> {
                Set<String> allowed = entry.getValue().get(user);
                if (allowed != null && !allowed.isEmpty())
                {
                    Map<UserOrGroup, Set<String>> contextMap = result.computeIfAbsent(entry.getKey(), k -> new HashMap<>());
                    contextMap.put(UserOrGroup.ALLOWED, allowed);
                }
            });
        
        databaseValue.getDeniedUserData()
            .entrySet().stream()
            .filter(entry -> entry.getValue().keySet().contains(user))
            .forEach(entry -> {
                Set<String> denied = entry.getValue().get(user);
                if (denied != null && !denied.isEmpty())
                {
                    Map<UserOrGroup, Set<String>> contextMap = result.computeIfAbsent(entry.getKey(), k -> new HashMap<>());
                    contextMap.put(UserOrGroup.DENIED, denied);
                }
            });
        
        return result;
    }
    
    /* --------------------------------------- */
    /* ALLOWED PROFILES FOR ANY CONNECTED USER */
    /* --------------------------------------- */
    
    public Map<UserIdentity, Map<UserOrGroup, Set<String>>> getProfilesForUsers(Object object, UserIdentity user)
    {
        Map<UserIdentity, Map<UserOrGroup, Set<String>>> result = new HashMap<>();
        
        Map<UserIdentity, Set<String>> allowedProfilesByUser = Optional.ofNullable(_getFullData(object).getAlloweUserData().get(object)).orElse(Map.of());
        Map<UserIdentity, Set<String>> deniedProfilesByUser = Optional.ofNullable(_getFullData(object).getDeniedUserData().get(object)).orElse(Map.of());
        
        Set<UserIdentity> userKeys;
        if (user != null)
        {
            userKeys = Set.of(user);
        }
        else
        {
            userKeys = new HashSet<>();
            userKeys.addAll(allowedProfilesByUser.keySet());
            userKeys.addAll(deniedProfilesByUser.keySet());
        }
        
        for (UserIdentity userKey : userKeys)
        {
            result.put(userKey, Map.of(UserOrGroup.ALLOWED, allowedProfilesByUser.containsKey(userKey) ? allowedProfilesByUser.get(userKey) : Set.of(),
                                     UserOrGroup.DENIED, deniedProfilesByUser.containsKey(userKey) ? deniedProfilesByUser.get(userKey) : Set.of()));
        }
        
        return result;
    }
    
    public void addAllowedProfilesForAnyConnectedUser(Object object, Set<String> profileIds)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            Object prefixedObject = getObjectWithPrefix(object);
            for (String profileId : profileIds)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("context", prefixedObject);
                parameters.put("profileIds", Arrays.asList(profileId));
                
                List<Map<String, String>> allowedProfiles = session.selectList("ProfilesAssignment.getAnyConnectedAllowedProfiles", parameters);
                
                if (allowedProfiles.isEmpty())
                {
                    parameters.put("profileId", profileId);
                    session.insert("ProfilesAssignment.addAllowedAnyConnected", parameters);
                }
                else
                {
                    getLogger().debug("Profile {} is already allowed for anyconnected on context {}", profileId, prefixedObject);
                }
                
            }
            
            session.commit();
        }
    }
    
    public void removeAllowedProfilesForAnyConnectedUser(Object object, Set<String> profileIds)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession(true))
        {
            Map<String, Object> parameters = new HashMap<>();
            parameters.put("context", getObjectWithPrefix(object));
            parameters.put("profileIds", profileIds);
            session.delete("ProfilesAssignment.deleteAllowedAnyConnected", parameters);
        }
    }
    
    
    /* -------------------------------------- */
    /* DENIED PROFILES FOR ANY CONNECTED USER */
    /* -------------------------------------- */
    
    public Map<AnonymousOrAnyConnectedKeys, Set<String>> getProfilesForAnonymousAndAnyConnectedUser(Object object)
    {
        return Map.of(AnonymousOrAnyConnectedKeys.ANONYMOUS_ALLOWED, Optional.ofNullable(_getFullData(object).getAllowedAnonymous().get(object)).orElse(Set.of()),
                AnonymousOrAnyConnectedKeys.ANONYMOUS_DENIED, Optional.ofNullable(_getFullData(object).getDeniedAnonymous().get(object)).orElse(Set.of()),
                AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_ALLOWED, Optional.ofNullable(_getFullData(object).getAllowedAnyConnected().get(object)).orElse(Set.of()),
                AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_DENIED, Optional.ofNullable(_getFullData(object).getDeniedAnyConnected().get(object)).orElse(Set.of()));
    }
    
    public void addDeniedProfilesForAnyConnectedUser(Object object, Set<String> profileIds)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            Object prefixedObject = getObjectWithPrefix(object);
            
            for (String profileId : profileIds)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("context", prefixedObject);
                parameters.put("profileIds", Arrays.asList(profileId));
                
                List<Map<String, String>> deniedProfiles = session.selectList("ProfilesAssignment.getAnyConnectedDeniedProfiles", parameters);
                
                if (deniedProfiles.isEmpty())
                {
                    parameters.put("profileId", profileId);
                    session.insert("ProfilesAssignment.addDeniedAnyConnected", parameters);
                }
                else
                {
                    getLogger().debug("Profile {} is already denied for anyconnected on context {}", profileId, prefixedObject);
                }
                
            }
            
            session.commit();
        }
    }
    
    public void removeDeniedProfilesForAnyConnectedUser(Object object, Set<String> profileIds)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession(true))
        {
            Map<String, Object> parameters = new HashMap<>();
            parameters.put("context", getObjectWithPrefix(object));
            parameters.put("profileIds", profileIds);
            session.delete("ProfilesAssignment.deleteDeniedAnyConnected", parameters);
        }
    }
    
    /* ------------------------------ */
    /* ALLOWED PROFILES FOR ANONYMOUS */
    /* ------------------------------ */
    
    public void addAllowedProfilesForAnonymous(Object object, Set<String> profileIds)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            Object prefixedObject = getObjectWithPrefix(object);
            
            for (String profileId : profileIds)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("context", prefixedObject);
                parameters.put("profileIds", Arrays.asList(profileId));
                
                List<Map<String, String>> allowedProfiles = session.selectList("ProfilesAssignment.getAnonymousAllowedProfiles", parameters);
                
                if (allowedProfiles.isEmpty())
                {
                    parameters.put("profileId", profileId);
                    session.insert("ProfilesAssignment.addAllowedAnonymous", parameters);
                }
                else
                {
                    getLogger().debug("Profile {} is already allowed for anonymous on context {}", profileId, prefixedObject);
                }
            }
            
            session.commit();
        }
    }
    
    public void removeAllowedProfilesForAnonymous(Object object, Set<String> profileIds)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession(true))
        {
            Map<String, Object> parameters = new HashMap<>();
            parameters.put("context", getObjectWithPrefix(object));
            parameters.put("profileIds", profileIds);
            session.delete("ProfilesAssignment.deleteAllowedAnonymous", parameters);
        }
    }
    
    
    /* ----------------------------- */
    /* DENIED PROFILES FOR ANONYMOUS */
    /* ----------------------------- */
    
    public void addDeniedProfilesForAnonymous(Object object, Set<String> profileIds)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            Object prefixedObject = getObjectWithPrefix(object);
            for (String profileId : profileIds)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("context", prefixedObject);
                parameters.put("profileIds", Arrays.asList(profileId));
                
                List<Map<String, String>> deniedProfiles = session.selectList("ProfilesAssignment.getAnonymousDeniedProfiles", parameters);
                
                if (deniedProfiles.isEmpty())
                {
                    parameters.put("profileId", profileId);
                    session.insert("ProfilesAssignment.addDeniedAnonymous", parameters);
                }
                else
                {
                    getLogger().debug("Profile {} is already denied for anonymous on context {}", profileId, prefixedObject);
                }
            }
            
            session.commit();
        }
    }
    
    public void removeDeniedProfilesForAnonymous(Object object, Set<String> profileIds)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession(true))
        {
            Map<String, Object> parameters = new HashMap<>();
            parameters.put("context", getObjectWithPrefix(object));
            parameters.put("profileIds", profileIds);
            session.delete("ProfilesAssignment.deleteDeniedAnonymous", parameters);
        }
    }
    
    
    /* --------------------------- */
    /* MANAGEMENT OF ALLOWED USERS */
    /* --------------------------- */
    
    public void addAllowedUsers(Set<UserIdentity> users, Object object, String profileId)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            Object prefixedObject = getObjectWithPrefix(object);
            
            for (UserIdentity userIdentity : users)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("login", userIdentity.getLogin());
                parameters.put("population", userIdentity.getPopulationId());
                parameters.put("context", prefixedObject);
                parameters.put("profileIds", Arrays.asList(profileId));
                
                List<Map<String, String>> allowedProfiles = session.selectList("ProfilesAssignment.getUserAllowedProfiles", parameters);
                
                if (allowedProfiles.isEmpty())
                {
                    parameters.put("profileId", profileId);
                    session.insert("ProfilesAssignment.addAllowedUser", parameters);
                }
                else
                {
                    getLogger().debug("Login {} has already profile {} on context {}", userIdentity, profileId, prefixedObject);
                }
                
            }
            
            session.commit();
        }
    }
    
    public void removeAllowedUsers(Set<UserIdentity> users, Object object, String profileId)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            for (UserIdentity userIdentity : users)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("login", userIdentity.getLogin());
                parameters.put("population", userIdentity.getPopulationId());
                parameters.put("profileIds", Arrays.asList(profileId));
                if (object != null)
                {
                    parameters.put("context", getObjectWithPrefix(object));
                }
                
                session.delete("ProfilesAssignment.deleteAllowedUser", parameters);
            }
            session.commit();
        }
    }
    
    public void removeAllowedUsers(Set<UserIdentity> users, Object object)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            for (UserIdentity userIdentity : users)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("login", userIdentity.getLogin());
                parameters.put("population", userIdentity.getPopulationId());
                if (object != null)
                {
                    parameters.put("context", getObjectWithPrefix(object));
                }
                
                session.delete("ProfilesAssignment.deleteAllowedUser", parameters);
            }
            session.commit();
        }
    }
    
    
    /* -------------------- */
    /* MANAGEMENT OF GROUPS */
    /* -------------------- */
    
    public Map<GroupIdentity, Map<UserOrGroup, Set<String>>> getProfilesForGroups(Object object, Set<GroupIdentity> groups)
    {
        Map<GroupIdentity, Map<UserOrGroup, Set<String>>> result = new HashMap<>();
        
        Map<GroupIdentity, Set<String>> allowedProfilesByGroup = Optional.ofNullable(_getFullData(object).getAlloweGroupData().get(object)).orElse(Map.of());
        Map<GroupIdentity, Set<String>> deniedProfilesByGroup = Optional.ofNullable(_getFullData(object).getDeniedGroupData().get(object)).orElse(Map.of());
        
        Set<GroupIdentity> groupKeys;
        if (groups != null)
        {
            groupKeys = groups;
        }
        else
        {
            groupKeys = new HashSet<>();
            groupKeys.addAll(allowedProfilesByGroup.keySet());
            groupKeys.addAll(deniedProfilesByGroup.keySet());
        }
        
        for (GroupIdentity group : groupKeys)
        {
            result.put(group, Map.of(UserOrGroup.ALLOWED, allowedProfilesByGroup.containsKey(group) ? allowedProfilesByGroup.get(group) : Set.of(),
                                     UserOrGroup.DENIED, deniedProfilesByGroup.containsKey(group) ? deniedProfilesByGroup.get(group) : Set.of()));
        }
        
        return result;
    }
    
    public void addAllowedGroups(Set<GroupIdentity> groups, Object object, String profileId)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            Object prefixedObject = getObjectWithPrefix(object);
            for (GroupIdentity group : groups)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("groupId", group.getId());
                parameters.put("groupDirectory", group.getDirectoryId());
                parameters.put("context", prefixedObject);
                parameters.put("profileIds", Arrays.asList(profileId));
                
                List<Map<String, String>> allowedProfiles = session.selectList("ProfilesAssignment.getGroupAllowedProfiles", parameters);
                
                if (allowedProfiles.isEmpty())
                {
                    parameters.put("profileId", profileId);
                    session.insert("ProfilesAssignment.addAllowedGroup", parameters);
                }
                else
                {
                    getLogger().debug("Group {} is already allowed for profile {} on context {}", group, profileId, prefixedObject);
                }
            }
            
            session.commit();
        }
    }
    
    public void removeAllowedGroups(Set<GroupIdentity> groups, Object object, String profileId)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            for (GroupIdentity group : groups)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("groupId", group.getId());
                parameters.put("groupDirectory", group.getDirectoryId());
                parameters.put("profileIds", Arrays.asList(profileId));
                if (object != null)
                {
                    parameters.put("context", getObjectWithPrefix(object));
                }
                
                session.delete("ProfilesAssignment.deleteAllowedGroup", parameters);
            }
            session.commit();
        }
    }
    
    public void removeAllowedGroups(Set<GroupIdentity> groups, Object object)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            for (GroupIdentity group : groups)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("groupId", group.getId());
                parameters.put("groupDirectory", group.getDirectoryId());
                if (object != null)
                {
                    parameters.put("context", getObjectWithPrefix(object));
                }
                
                session.delete("ProfilesAssignment.deleteAllowedGroup", parameters);
            }
            session.commit();
        }
    }
    
    public void addDeniedUsers(Set<UserIdentity> users, Object object, String profileId)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            Object prefixedObject = getObjectWithPrefix(object);
            for (UserIdentity userIdentity : users)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("login", userIdentity.getLogin());
                parameters.put("population", userIdentity.getPopulationId());
                parameters.put("context", prefixedObject);
                parameters.put("profileIds", Arrays.asList(profileId));
                
                List<Map<String, String>> deniedProfiles = session.selectList("ProfilesAssignment.getUserDeniedProfiles", parameters);
                
                if (deniedProfiles.isEmpty())
                {
                    parameters.put("profileId", profileId);
                    session.insert("ProfilesAssignment.addDeniedUser", parameters);
                }
                else
                {
                    getLogger().debug("Login {} is already denied for profile {} on context {}", userIdentity, profileId, prefixedObject);
                }
                
            }
            
            session.commit();
        }
    }
    
    public void removeDeniedUsers(Set<UserIdentity> users, Object object, String profileId)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            for (UserIdentity userIdentity : users)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("login", userIdentity.getLogin());
                parameters.put("population", userIdentity.getPopulationId());
                parameters.put("profileIds", Arrays.asList(profileId));
                if (object != null)
                {
                    parameters.put("context", getObjectWithPrefix(object));
                }
                
                session.delete("ProfilesAssignment.deleteDeniedUser", parameters);
            }
            session.commit();
        }
    }
    
    public void removeDeniedUsers(Set<UserIdentity> users, Object object)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            for (UserIdentity userIdentity : users)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("login", userIdentity.getLogin());
                parameters.put("population", userIdentity.getPopulationId());
                if (object != null)
                {
                    parameters.put("context", getObjectWithPrefix(object));
                }
                
                session.delete("ProfilesAssignment.deleteDeniedUser", parameters);
            }
            session.commit();
        }
    }
    
    public void addDeniedGroups(Set<GroupIdentity> groups, Object object, String profileId)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            Object prefixedObject = getObjectWithPrefix(object);
            for (GroupIdentity group : groups)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("groupId", group.getId());
                parameters.put("groupDirectory", group.getDirectoryId());
                parameters.put("context", prefixedObject);
                parameters.put("profileIds", Arrays.asList(profileId));
                
                List<Map<String, String>> deniedProfiles = session.selectList("ProfilesAssignment.getGroupDeniedProfiles", parameters);
                
                if (deniedProfiles.isEmpty())
                {
                    parameters.put("profileId", profileId);
                    session.insert("ProfilesAssignment.addDeniedGroup", parameters);
                }
                else
                {
                    getLogger().debug("Group {} is already denied for profile {} on context {}", group, profileId, prefixedObject);
                }
            }
            
            session.commit();
        }
    }
    
    public void removeDeniedGroups(Set<GroupIdentity> groups, Object object, String profileId)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            for (GroupIdentity group : groups)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("groupId", group.getId());
                parameters.put("groupDirectory", group.getDirectoryId());
                parameters.put("profileIds", Arrays.asList(profileId));
                if (object != null)
                {
                    parameters.put("context", getObjectWithPrefix(object));
                }
                
                session.delete("ProfilesAssignment.deleteDeniedGroup", parameters);
            }
            session.commit();
        }
    }
    
    public void removeDeniedGroups(Set<GroupIdentity> groups, Object object)
    {
        _clearCache(object);
        
        try (SqlSession session = getSession())
        {
            for (GroupIdentity group : groups)
            {
                Map<String, Object> parameters = new HashMap<>();
                parameters.put("groupId", group.getId());
                parameters.put("groupDirectory", group.getDirectoryId());
                if (object != null)
                {
                    parameters.put("context", getObjectWithPrefix(object));
                }
                
                session.delete("ProfilesAssignment.deleteDeniedGroup", parameters);
            }
            session.commit();
        }
    }
    
    /* ----------- */
    /* INHERITANCE */
    /* ----------- */
    
    public void disallowInheritance(Object object, boolean disallow)
    {
        try (SqlSession session = getSession())
        {
            Object ctxWithPrefix = getObjectWithPrefix(object);
            
            Map<String, Object> parameters = new HashMap<>();
            
            parameters.put("context", ctxWithPrefix);
            parameters.put("disallow", disallow);
            
            if (_getInheritances().hasKey((String) ctxWithPrefix))
            {
                session.insert("ProfilesAssignment.updateInheritance", parameters);
            }
            else
            {
                session.insert("ProfilesAssignment.insertInheritance", parameters);
            }
            session.commit();
            
            _clearInheritanceCache();
        }
    }
    
    public boolean isInheritanceDisallowed(Object object)
    {
        Object ctxWithPrefix = getObjectWithPrefix(object);
        
        Boolean isDisallowed = _getInheritances().get((String) ctxWithPrefix);
        return isDisallowed != null && Boolean.TRUE.equals(isDisallowed);
    }
    
    
    /* ------ */
    /* REMOVE */
    /* ------ */
    
    public void removeProfile(String profileId)
    {
        _clearCache();
        
        try (SqlSession session = getSession())
        {
            Map<String, Object> parameters = new HashMap<>();
            parameters.put("profileIds", Arrays.asList(profileId));
            
            session.delete("ProfilesAssignment.deleteAllowedUser", parameters);
            session.delete("ProfilesAssignment.deleteDeniedUser", parameters);
            session.delete("ProfilesAssignment.deleteAllowedGroup", parameters);
            session.delete("ProfilesAssignment.deleteDeniedGroup", parameters);
            session.delete("ProfilesAssignment.deleteAllowedAnonymous", parameters);
            session.delete("ProfilesAssignment.deleteDeniedAnonymous", parameters);
            session.delete("ProfilesAssignment.deleteAllowedAnyConnected", parameters);
            session.delete("ProfilesAssignment.deleteDeniedAnyConnected", parameters);
            
            session.commit();
        }
    }
    
    public void removeUser(UserIdentity user)
    {
        _clearCache();
        
        try (SqlSession session = getSession())
        {
            Map<String, String> parameters = new HashMap<>();
            parameters.put("login", user.getLogin());
            parameters.put("population", user.getPopulationId());
            
            session.delete("ProfilesAssignment.deleteAllowedUser", parameters);
            session.delete("ProfilesAssignment.deleteDeniedUser", parameters);
            
            session.commit();
        }
    }
    
    public void removeGroup(GroupIdentity group)
    {
        _clearCache();
        
        try (SqlSession session = getSession())
        {
            Map<String, String> parameters = new HashMap<>();
            parameters.put("groupId", group.getId());
            parameters.put("groupDirectory", group.getDirectoryId());
            
            session.delete("ProfilesAssignment.deleteAllowedGroup", parameters);
            session.delete("ProfilesAssignment.deleteDeniedGroup", parameters);
            
            session.commit();
        }
    }
    
    /* ------------------------------ */
    /* SUPPORT OF OBJECT AND PRIORITY */
    /* ------------------------------ */
    
    public boolean isSupported(Object object)
    {
        if (object instanceof String)
        {
            String context = (String) object;
            return context.equals(_supportedContext) || context.startsWith(_supportedContext + "/");
        }
        return false;
    }
    
    public boolean isRootContextSupported(Object rootContext)
    {
        return isSupported(rootContext);
    }
    
    public int getPriority()
    {
        return ProfileAssignmentStorage.MIN_PRIORITY;
    }
    
    private static class Database
    {
        // Context, ProfileId
        private Map<String, Set<String>> _allowedAnonymousData;
        // Context, ProfileId
        private Map<String, Set<String>> _deniedAnonymousData;
        // Context, ProfileId
        private Map<String, Set<String>> _allowedAnyConnectedData;
        // Context, ProfileId
        private Map<String, Set<String>> _deniedAnyConnectedData;
        // Context, UserIdentity/GroupIdenty ProfileId
        private Map<String, Map<UserIdentity, Set<String>>> _allowedUserData;
        // Context, UserIdentity/GroupIdenty ProfileId
        private Map<String, Map<UserIdentity, Set<String>>> _deniedUserData;
        // Context, UserIdentity/GroupIdenty ProfileId
        private Map<String, Map<GroupIdentity, Set<String>>> _allowedGroupData;
        // Context, UserIdentity/GroupIdenty ProfileId
        private Map<String, Map<GroupIdentity, Set<String>>> _deniedGroupData;

        Database(List<Map<String, String>> data)
        {
            _allowedAnonymousData = new HashMap<>();
            _deniedAnonymousData = new HashMap<>();
            _allowedAnyConnectedData = new HashMap<>();
            _deniedAnyConnectedData = new HashMap<>();
            _allowedUserData = new HashMap<>();
            _deniedUserData = new HashMap<>();
            _allowedGroupData = new HashMap<>();
            _deniedGroupData = new HashMap<>();
            
            for (Map<String, String> map : data)
            {
                String type = map.get("type").trim();
                String profileId = map.get("profileId");
                String context = map.get("context");

                if ("ALLOWED_ANONYMOUS".equals(type) || "DENIED_ANONYMOUS".equals(type) || "ALLOWED_ANYCONNECTED".equals(type) || "DENIED_ANYCONNECTED".equals(type))
                {
                    _buildAnonymousOrAnyConnectedData(type, profileId, context);
                }
                else if ("ALLOWED_USER".equals(type) || "DENIED_USER".equals(type))
                {
                    String targetId = map.get("targetId");
                    String targetGroup = map.get("targetGroup");
                    UserIdentity user = new UserIdentity(targetId, targetGroup);

                    _buildUserData(type, profileId, context, user);
                }
                else // if ("ALLOWED_GROUP".equals(type) || "DENIED_GROUP".equals(type))
                {
                    String targetId = map.get("targetId");
                    String targetGroup = map.get("targetGroup");
                    GroupIdentity group = new GroupIdentity(targetId, targetGroup);
                    
                    _buildGroupData(type, profileId, context, group);
                }
            }
        }

        private void _buildGroupData(String type, String profileId, String context, GroupIdentity group)
        {
            Map<String, Map<GroupIdentity, Set<String>>> wData;
            if ("ALLOWED_GROUP".equals(type))
            {
                wData = _allowedGroupData;
            }
            else // if ("DENIED_GROUP".equals(type))
            {
                wData = _deniedGroupData;
            } 

            if (!wData.containsKey(context))
            {
                wData.put(context, new HashMap<>());
            }
            Map<GroupIdentity, Set<String>> contextMap = wData.get(context);
            
            if (!contextMap.containsKey(group))
            {
                contextMap.put(group, new HashSet<>());
            }
            Set<String> profileSet = contextMap.get(group);
            
            profileSet.add(profileId);
        }

        private void _buildUserData(String type, String profileId, String context, UserIdentity user)
        {
            Map<String, Map<UserIdentity, Set<String>>> wData;
            if ("ALLOWED_USER".equals(type))
            {
                wData = _allowedUserData;
            }
            else // if ("DENIED_USER".equals(type))
            {
                wData = _deniedUserData;
            } 
            
            if (!wData.containsKey(context))
            {
                wData.put(context, new HashMap<>());
            }
            Map<UserIdentity, Set<String>> contextMap = wData.get(context);
            
            if (!contextMap.containsKey(user))
            {
                contextMap.put(user, new HashSet<>());
            }
            Set<String> profileSet = contextMap.get(user);
            
            profileSet.add(profileId);
        }

        private void _buildAnonymousOrAnyConnectedData(String type, String profileId, String context)
        {
            Map<String, Set<String>> wData;
            if ("ALLOWED_ANONYMOUS".equals(type))
            {
                wData = _allowedAnonymousData;
            }
            else if ("DENIED_ANONYMOUS".equals(type))
            {
                wData = _deniedAnonymousData;
            } 
            else if ("ALLOWED_ANYCONNECTED".equals(type))
            {
                wData = _allowedAnyConnectedData;
            } 
            else // if ("DENIED_ANYCONNECTED".equals(type)) 
            {
                wData = _deniedAnyConnectedData;
            } 

            if (!wData.containsKey(context))
            {
                wData.put(context, new HashSet<>());
            }
            Set<String> profileSet = wData.get(context);
            
            profileSet.add(profileId);
        }

        public Map<String, Set<String>> getAllowedAnonymous()
        {
            return _allowedAnonymousData;
        }
        public Map<String, Set<String>> getDeniedAnonymous()
        {
            return _deniedAnonymousData;
        }
        public Map<String, Set<String>> getAllowedAnyConnected()
        {
            return _allowedAnyConnectedData;
        }
        public Map<String, Set<String>> getDeniedAnyConnected()
        {
            return _deniedAnyConnectedData;
        }
        public Map<String, Map<UserIdentity, Set<String>>> getAlloweUserData()
        {
            return _allowedUserData;
        }
        public Map<String, Map<UserIdentity, Set<String>>> getDeniedUserData()
        {
            return _deniedUserData;
        }
        public Map<String, Map<GroupIdentity, Set<String>>> getAlloweGroupData()
        {
            return _allowedGroupData;
        }
        public Map<String, Map<GroupIdentity, Set<String>>> getDeniedGroupData()
        {
            return _deniedGroupData;
        }
    }

    private Cache<String, Database> _getStorageCache()
    {
        return _cacheManager.get(__JDBC_PROFILE_ASSIGNMENT_STORAGE_CACHE);
    }
    
    private Cache<String, Boolean> _getInheritanceCache()
    {
        return _cacheManager.get(__JDBC_PROFILE_ASSIGNMENT_INHERITANCE_CACHE);
    }
}
