/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.right;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Pattern;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.right.RightAssignmentContext;

/**
 * This implementation will manage hierarchical string (parents separated by '/').
 * For example '/node1/node2/node3', will seek for rights on '/node1/node2/node3', if nothing is found will try on '/node1/node2', and finally on '/node1'.
 * Configurable with several &lt;prefix&gt;/myprefix&lt;/prefix&gt;
 * 
 * If this AcessController is limited to a workspace (by the corresponding {@link RightAssignmentContext}), this should be specified 
 * with a configuration &lt;workspace&gt; that is a negativable regexp of a workspace context (such as '/cms' or '!/admin')
 */
public class StringHierarchicalAccessController extends AbstractHierarchicalAccessController<String> implements Configurable
{
    /** The separator character */
    protected static final char SEPARATOR = '/';
    /** The prefixes */
    protected Set<String> _prefixes;
    /** The regexp that should match current workspace */
    protected Pattern _workspaceMatcher;
    /** Consider _workspaceMatcher negatively */
    protected boolean _reverseWorkspaceMather;
    
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _prefixes = new HashSet<>();
        
        for (Configuration prefixConfiguration : configuration.getChildren("prefix"))
        {
            String prefix = prefixConfiguration.getValue();
            _prefixes.add(prefix);
        }
        
        String workspace = configuration.getChild("workspace").getValue("");
        if (StringUtils.isNotBlank(workspace))
        {
            if (workspace.startsWith("!"))
            {
                _reverseWorkspaceMather = true;
                workspace = workspace.substring(1);
            }
            _workspaceMatcher = Pattern.compile(workspace);
        }
    }
    
    /**
     * Get the prefix of the path accepted
     * @return The non-null and non-empty prefixes, starting by '/'
     */
    protected Set<String> getSupportedPrefixes()
    {
        return _prefixes;
    }

    /**
     * Get the prefix of the path accepted
     * @return The non-null and non-empty prefixes, starting by '/'
     */
    protected Set<String> getRootPrefixes()
    {
        return _prefixes;
    }

    @Override
    protected Set<? extends Object> _convertWorkspaceToRootRightContexts(Set<Object> workspacesContexts)
    {
        if (_workspaceMatcher != null)
        {
            for (Object workspaceContext : workspacesContexts)
            {
                if (workspaceContext instanceof String)
                {
                    boolean match = _workspaceMatcher.matcher((String) workspaceContext).matches();
                    if (!match && _reverseWorkspaceMather || match && !_reverseWorkspaceMather)
                    {
                        return getRootPrefixes();
                    }
                }
            }
            return null;
        }
        else
        {
            return getRootPrefixes();
        }
    }
    
    public boolean isSupported(Object object)
    {
        if (!(object instanceof String))
        {
            return false;
        }
        String context = (String) object;
        
        Set<String> prefixes = getSupportedPrefixes();
        if (prefixes != null)
        {
            for (String prefix : prefixes)
            {
                if (StringUtils.isNotBlank(prefix) && StringUtils.equals(context, prefix) || context.startsWith(prefix + SEPARATOR))
                {
                    return true;
                }
            }
        }
        
        return false;
    }

    @Override
    protected Set<String> _getParents(String object)
    {
        if (object == null)
        {
            return null;
        }
        
        int index = object.lastIndexOf(SEPARATOR);
        if (index <= 1)
        {
            return null;
        }
        
        return Collections.singleton(object.substring(0, index));
    }
}
