/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.schedule;

import java.time.LocalTime;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;

import org.ametys.runtime.config.Config;

/**
 * Abstract {@link Runnable} for runnable that will run at the specified hour defined in config throught param 'param-hour'
 */
public abstract class AbstractConfigBasedHourRunnable extends ConfigBasedRunnable
{
    @Override
    protected void configureFireProcess(Configuration configuration, DefaultConfiguration modifiedConfiguration) throws ConfigurationException
    {
        Configuration hourCfg = configuration.getChild("param-hour", false);
        if (hourCfg != null)
        {
            String hourCfgValue = Config.getInstance().getValue(hourCfg.getValue("").trim());
            LocalTime lt = LocalTime.parse(hourCfgValue);
            
            // FireProcess = cron
            DefaultConfiguration fireProcessConfig = new DefaultConfiguration("fire-process");
            fireProcessConfig.setValue(FireProcess.CRON.name());
            modifiedConfiguration.addChild(fireProcessConfig);
            
            String cronExpression = getCronExpression(configuration, lt);
            
            DefaultConfiguration cronConfig = new DefaultConfiguration("cron");
            cronConfig.setValue(cronExpression);
            modifiedConfiguration.addChild(cronConfig);
        }
        else
        {
            throw new ConfigurationException("Mandatory configuration 'param-hour' is missing for config based hour runnable " + this.getClass().getName());
        }
    }
    
    /**
     * Get the CRON expression for this runnable. Use when "param-hour" is used.
     * @param configuration the original configuration
     * @param time The computed time, expressed in the server timezone
     * @return the cron expression 
     * @throws ConfigurationException if configuration is invalid
     */
    protected abstract String getCronExpression(Configuration configuration, LocalTime time) throws ConfigurationException;
}
