/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.schedule;

import java.io.IOException;
import java.time.ZonedDateTime;
import java.util.Locale;
import java.util.Map;

import javax.script.ScriptException;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.slf4j.Logger;

import org.ametys.core.schedule.Schedulable;
import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.core.schedule.Scheduler;
import org.ametys.plugins.core.ui.script.AsyncScriptHandler;
import org.ametys.plugins.core.ui.script.ScriptHandler;
import org.ametys.runtime.servlet.RuntimeServlet;
import org.ametys.runtime.servlet.RuntimeServlet.ForcedMainteanceInformations;
import org.ametys.runtime.servlet.RuntimeServlet.MaintenanceStatus;
import org.ametys.runtime.servlet.RuntimeServlet.RunMode;

import jakarta.mail.MessagingException;

/**
 * A {@link Schedulable} job for executing scripts.
 */
public class ScriptSchedulable extends AbstractStaticSchedulable
{
    /** The key for the script (as string) to execute */
    public static final String SCRIPT_KEY = "script";
    /** The key for the recipient of the report mail */
    public static final String RECIPIENT_KEY = "recipient";
    /** The key for the workspace */
    public static final String WORKSPACE_KEY = "workspace";
    /** The key for the maintenance mode */
    public static final String MAINTENANCE_KEY = "maintenance";
    
    private static final String __JOBDATAMAP_SCRIPT_KEY = Scheduler.PARAM_VALUES_PREFIX + SCRIPT_KEY;
    private static final String __JOBDATAMAP_RECIPIENT_KEY = Scheduler.PARAM_VALUES_PREFIX + RECIPIENT_KEY;
    private static final String __JOBDATAMAP_WORKSPACE_KEY = Scheduler.PARAM_VALUES_PREFIX + WORKSPACE_KEY;
    private static final String __JOBDATAMAP_MAINTENANCE_KEY = Scheduler.PARAM_VALUES_PREFIX + MAINTENANCE_KEY;
    
    /** The script handler */
    protected ScriptHandler _scriptHandler;
    /** The async script handler */
    protected AsyncScriptHandler _asyncScriptHandler;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _scriptHandler = (ScriptHandler) manager.lookup(ScriptHandler.ROLE);
        
        if (manager.hasService(AsyncScriptHandler.COMPONENT_ROLE))
        {
            _asyncScriptHandler = (AsyncScriptHandler) manager.lookup(AsyncScriptHandler.COMPONENT_ROLE);
            _scriptHandler = _asyncScriptHandler;
        }
    }
    
    @Override
    public void execute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception
    {
        JobDataMap jobDataMap = context.getJobDetail().getJobDataMap();
        String script = (String) jobDataMap.get(__JOBDATAMAP_SCRIPT_KEY);
        String recipient = (String) jobDataMap.get(__JOBDATAMAP_RECIPIENT_KEY);
        String workspace = (String) jobDataMap.get(__JOBDATAMAP_WORKSPACE_KEY);
        Boolean maintenance = (Boolean) jobDataMap.get(__JOBDATAMAP_MAINTENANCE_KEY);
        UserIdentity userIdentity = UserIdentity.stringToUserIdentity(jobDataMap.getString(Scheduler.KEY_RUNNABLE_USERIDENTITY));
        
        RunMode oldMode = null;
        if (Boolean.TRUE.equals(maintenance))
        {
            oldMode = RuntimeServlet.getRunMode();
            
            if (oldMode != RunMode.MAINTENANCE)
            {
                RuntimeServlet.setMaintenanceStatus(MaintenanceStatus.FORCED, new ForcedMainteanceInformations("", userIdentity, ZonedDateTime.now()));
                RuntimeServlet.setRunMode(RunMode.MAINTENANCE);
            }
        }
        
        try
        {
            Map<String, Object> scriptResults = _executeScript(script, workspace, progressionTracker);
            _sendMail(scriptResults, recipient, userIdentity);
            
            if (scriptResults.containsKey("stacktrace"))
            {
                throw new JobExecutionException("Script failure. See preceding exception to know more.");
            }
        }
        finally 
        {
            if (Boolean.TRUE.equals(maintenance) && oldMode != RunMode.MAINTENANCE)
            {
                RuntimeServlet.setRunMode(oldMode);
                RuntimeServlet.setMaintenanceStatus(MaintenanceStatus.NONE, null);
            }
        }
    }
    
    private Map<String, Object> _executeScript(String script, String workspace, ContainerProgressionTracker progressionTracker) throws ScriptException
    {
        return _scriptHandler.executeScript("function main() { \n " + script + " \n }", Map.of("progressionTracker", progressionTracker), workspace);
    }
    
    private void _sendMail(Map<String, Object> scriptResults, String recipient, UserIdentity userIdentity) throws MessagingException, IOException
    {
        Locale locale = Locale.getDefault(); // FIXME cannot have the user locale at this time
        Logger logger = getLogger();
        
        if (_asyncScriptHandler != null)
        {
            _asyncScriptHandler.sendReportMail(scriptResults, userIdentity, recipient, locale, logger);
        }
    }
}
