/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.plugins.core.impl.upload;

import java.io.File;
import java.util.Calendar;
import java.util.Collection;
import java.util.GregorianCalendar;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.thread.ThreadSafe;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.AbstractFileFilter;
import org.apache.commons.io.filefilter.TrueFileFilter;
import org.quartz.JobExecutionContext;

import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.upload.UploadManager;
import org.ametys.plugins.core.impl.schedule.AbstractStaticSchedulable;

/**
 * Schedulable that remove empty directories or directories if the first file is one day old
 */
public class CleanFSUploadSchedulable extends AbstractStaticSchedulable implements ThreadSafe
{
    /** The Upload Manager */
    protected UploadManager _uploadManager;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _uploadManager = (UploadManager) manager.lookup(UploadManager.ROLE);
    }
    
    @Override
    public void execute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception
    {
        getLogger().info("Time to clean old uploads");
        
        try
        {
            Calendar calendar = new GregorianCalendar();
            calendar.add(Calendar.DAY_OF_YEAR, -1);
            FSUploadFileFilter filter = new FSUploadFileFilter(calendar.getTimeInMillis());
            
            if (_uploadManager instanceof FSUploadManager fSUploadManager) 
            {
                for (File effectiveLoginDir : fSUploadManager.getUploadsDirectories())
                {
                    // Compute upload directories to remove
                    String[] dirsToRemove = effectiveLoginDir.list(filter);
                    
                    if (dirsToRemove != null)
                    {
                        for (String dirToRemove : dirsToRemove)
                        {
                            File uploadDir = new File(effectiveLoginDir, dirToRemove);
                            
                            getLogger().debug("Removing directory: " + uploadDir);
            
                            FileUtils.deleteDirectory(uploadDir);
                        }
                    }
                }
            }
            else
            {
                getLogger().warn("Current implementation of UploadManager does not support the clean of uploads directory");
            }
        }
        catch (Exception e)
        {
            getLogger().error("Unable to clean old uploads", e);
        }
    }
    
    private static class FSUploadFileFilter extends AbstractFileFilter
    {
        private long _yesterday;
        
        FSUploadFileFilter(long yesterday)
        {
            super();
            this._yesterday = yesterday;
        }
        
        @Override
        public boolean accept(File file)
        {
            if (!file.isDirectory())
            {
                return false;
            }
            
            Collection<File> uploadFiles = FileUtils.listFiles(file, TrueFileFilter.INSTANCE, null);

            if (uploadFiles.isEmpty())
            {
                // Remove empty directory
                return true;
            }
            else
            {
                File firstFile = uploadFiles.iterator().next();

                // Remove directory if the first file is one day old
                return firstFile.lastModified() < _yesterday;
            }
        }
    }
}
