/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.plugins.core.impl.upload;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.NoSuchElementException;
import java.util.UUID;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.LogEnabled;
import org.apache.avalon.framework.logger.Logger;
import org.apache.avalon.framework.thread.ThreadSafe;
import org.apache.cocoon.Constants;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.filefilter.DirectoryFileFilter;
import org.apache.commons.io.filefilter.FalseFileFilter;
import org.apache.commons.io.filefilter.TrueFileFilter;

import org.ametys.core.upload.Upload;
import org.ametys.core.upload.UploadManager;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.FilenameUtils;
import org.ametys.core.util.URIUtils;
import org.ametys.runtime.util.AmetysHomeHelper;

/**
 * {@link UploadManager} which stores uploaded files into the
 * <code>uploads-user</code> directory located in Ametys home
 * <p>
 * Note that this implementation is not cluster safe.
 */
public class FSUploadManager implements UploadManager, ThreadSafe, Initializable, LogEnabled, Contextualizable
{
    /**
     * The path to the global uploads directory relative to ametys home
     */
    public static final String UPLOADS_DIRECTORY = "uploads-user";

    /** Context. */
    protected org.apache.cocoon.environment.Context _cocoonContext;
    /** Global uploads directory. */
    protected File _globalUploadsDir;
    /** Logger available to subclasses. */
    private Logger _logger;

    @Override
    public void contextualize(Context context) throws ContextException
    {
        // Retrieve context-root from context
        _cocoonContext = (org.apache.cocoon.environment.Context) context.get(Constants.CONTEXT_ENVIRONMENT_CONTEXT);
    }
    
    @Override
    public void enableLogging(Logger logger)
    {
        _logger = logger;
    }
        
    @Override
    public void initialize() throws Exception
    {
        _globalUploadsDir = new File(AmetysHomeHelper.getAmetysHomeData(), UPLOADS_DIRECTORY);
    }
    
    @Override
    public Upload storeUpload(UserIdentity user, String filename, InputStream is) throws IOException
    {
        if (!_globalUploadsDir.exists())
        {
            if (!_globalUploadsDir.mkdirs())
            {
                throw new IOException("Unable to create directory: " + _globalUploadsDir);
            }
        }
        
        // Create unique id
        String id = UUID.randomUUID().toString();
        
        File uploadFile = null;

        uploadFile = new File(_getUploadDir(user != null ? UserIdentity.userIdentityToString(user) : null, id), FilenameUtils.encodeName(filename));
        
        if (_logger.isInfoEnabled())
        {
            _logger.info("Using file: " + uploadFile);
        }
        
        if (!uploadFile.getParentFile().mkdirs())
        {
            throw new IOException("Unable to create directory: " + uploadFile.getParent());
        }
        
        try (OutputStream os = new FileOutputStream(uploadFile))
        {
            IOUtils.copy(is, os);
        }
        
        return new FSUpload(_cocoonContext, uploadFile);
    }
    
    @Override
    public Upload getUpload(UserIdentity user, String id) throws NoSuchElementException
    {
        File uploadDir = _getUploadDir(user != null ? UserIdentity.userIdentityToString(user) : null, id);
        
        if (_logger.isDebugEnabled())
        {
            _logger.debug("Using directory: " + uploadDir);
        }
        
        if (!uploadDir.exists() || !uploadDir.isDirectory())
        {
            throw new NoSuchElementException("No directory: " + uploadDir);
        }
        
        Collection<File> files = FileUtils.listFiles(uploadDir, TrueFileFilter.INSTANCE, FalseFileFilter.INSTANCE);
        
        if (_logger.isInfoEnabled())
        {
            _logger.info("Found files: " + files);
        }
        
        if (files.isEmpty())
        {
            throw new NoSuchElementException("No files in directory: " + uploadDir);
        }
        
        // Use first file found
        return new FSUpload(_cocoonContext, files.iterator().next());
    }
    
    /**
     * Retrieves the upload directory for a login and an upload id.
     * @param login the login.
     * @param id the upload id.
     * @return the upload directory.
     */
    protected File _getUploadDir(String login, String id)
    {
        String fileName = login != null ? FilenameUtils.encodeName(login) : "_Anonymous";
        return new File(new File(_globalUploadsDir, fileName), id);
    }
    
    /**
     * {@link Upload} implementation on file system.
     */
    protected static class FSUpload implements Upload
    {
        private org.apache.cocoon.environment.Context _context;
        private File _file;
        
        /**
         * Creates a FSUpload from a file.
         * @param context the context.
         * @param file the file.
         */
        public FSUpload(org.apache.cocoon.environment.Context context, File file)
        {
            _context = context;
            _file = file;
        }

        @Override
        public String getId()
        {
            return _file.getParentFile().getName();
        }
        
        @Override
        public ZonedDateTime getUploadedDate()
        {
            return DateUtils.asZonedDateTime(_file.lastModified());
        }
        
        @Override
        public String getFilename()
        {
            return URIUtils.decode(_file.getName());
        }
        
        @Override
        public String getMimeType()
        {
            String mimeType = _context.getMimeType(getFilename().toLowerCase());
            
            if (mimeType == null)
            {
                mimeType = "application/unknown";
            }
            
            return mimeType;
        }
        
        @Override
        public long getLength()
        {
            return _file.length();
        }
        
        @Override
        public InputStream getInputStream()
        {
            try
            {
                return new FileInputStream(_file);
            }
            catch (FileNotFoundException e)
            {
                throw new RuntimeException("Missing file: " + _file, e);
            }
        }
    }
    
    /**
     * get directories under upload-user directory
     * @return the directories
     */
    public File[] getUploadsDirectories()
    {
        return _globalUploadsDir.listFiles((FilenameFilter) DirectoryFileFilter.INSTANCE);
    }
}
