/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.user.directory;

import java.util.Arrays;
import java.util.Collection;
import java.util.Map;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;
import org.ametys.core.user.User;
import org.ametys.core.user.directory.UserDirectory;
import org.ametys.core.util.Cacheable;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Abstract class providing caching features to {@link UserDirectory}.
 */
public abstract class AbstractCachingUserDirectory extends AbstractLogEnabled implements UserDirectory, Serviceable, Cacheable, Disposable
{
    private String _id;
    private String _label;
    private String _udModelId;
    private Map<String, Object> _paramValues;
    private String _populationId;
    
    // Cannot use _populationId + "#" + _id as two UserDirectories with same id can co-exist during a short amount of time (during UserPopulationDAO#_readPopulations)
    private final String _uniqueCacheSuffix = org.ametys.core.util.StringUtils.generateKey();
    
    private AbstractCacheManager _cacheManager;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _cacheManager = (AbstractCacheManager) manager.lookup(AbstractCacheManager.ROLE);
    }
    
    public String getId()
    {
        return _id;
    }
    
    public String getLabel()
    {
        return _label;
    }
    
    @Override
    public void setPopulationId(String populationId)
    {
        _populationId = populationId;
    }
    
    @Override
    public String getPopulationId()
    {
        return _populationId;
    }
    
    @Override
    public Map<String, Object> getParameterValues()
    {
        return _paramValues;
    }
    
    @Override
    public String getUserDirectoryModelId()
    {
        return _udModelId;
    }
    
    public void init(String id, String udModelId, Map<String, Object> paramValues, String label) throws Exception
    {
        _id = id;
        _udModelId = udModelId;
        _label = label;
        _paramValues = paramValues;        
    }
    
    @Override
    public Collection<SingleCacheConfiguration> getManagedCaches()
    {
        return Arrays.asList(
                SingleCacheConfiguration.of(
                        getClass().getName() + "$by.login$" + _uniqueCacheSuffix, 
                        _buildI18n("PLUGINS_CORE_USERS_CACHE_BY_LOGIN_LABEL"), 
                        _buildI18n("PLUGINS_CORE_USERS_CACHE_BY_LOGIN_DESC")),
                SingleCacheConfiguration.of(
                        getClass().getName() + "$by.mail$" + _uniqueCacheSuffix, 
                        _buildI18n("PLUGINS_CORE_USERS_CACHE_BY_MAIL_LABEL"), 
                        _buildI18n("PLUGINS_CORE_USERS_CACHE_BY_MAIL_DESC"))
        );
    }
    
    private I18nizableText _buildI18n(String i18Key)
    {
        String catalogue = "plugin.core-impl";
        I18nizableText userDirectoryId = new I18nizableText(getPopulationId() + "#" + getId());
        I18nizableText type = new I18nizableText(getCacheTypeLabel());
        Map<String, I18nizableTextParameter> params = Map.of("type", type, "id", userDirectoryId);
        return new I18nizableText(catalogue, i18Key, params);
    }
    
    /**
     * Returns the user's cache by login
     * @return the user's cache by login
     */
    protected Cache<String, User> getCacheByLogin()
    {
        return getCache(getClass().getName() + "$by.login$" + _uniqueCacheSuffix);
    }
    
    /**
     * Returns the user's cache by email
     * @return the user's cache by email
     */
    protected Cache<String, User> getCacheByMail()
    {
        return getCache(getClass().getName() + "$by.mail$" + _uniqueCacheSuffix);
    }
    
    /**
     * Returns a String identifying the type of this directory, used in the cache screen in _admin,
     * such as 'SQL' or 'LDAP'.
     * @return this directory's type
     */
    protected abstract String getCacheTypeLabel();

    @Override
    public AbstractCacheManager getCacheManager()
    {
        return _cacheManager;
    }
    
    @Override
    public void dispose()
    {
        removeCaches();
    }
}
