/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.validator;

import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;

import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.parameter.DefaultValidator;
import org.ametys.runtime.parameter.Errors;

/**
 * This validator validates that the regular expression set is valid
 *
 */
public class RegexpValidator extends DefaultValidator
{
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        Configuration validatorConfig = configuration.getChild("validation").getChild("custom-validator");
        
        _isMandatory = validatorConfig.getChild("mandatory", false) != null;

        String regexp = validatorConfig.getChild("regexp").getValue(null);
        if (regexp != null)
        {
            _regexp = Pattern.compile(regexp);
        }
        
        Configuration textConfig = validatorConfig.getChild("invalidText", false);
        if (textConfig != null)
        {
            _invalidText = I18nizableText.parseI18nizableText(textConfig, "plugin." + _pluginName);
        }
    }
    
    @Override
    protected void validateSingleValue(Object value, Errors errors)
    {
        super.validateSingleValue(value, errors);
        
        _validateRegexp((String) value, errors);
    }
    
    @Override
    protected void validateArrayValues(Object[] values, Errors errors)
    {
        super.validateArrayValues(values, errors);
        
        for (Object value : values)
        {
            _validateRegexp((String) value, errors);
        }
    }
    
    private void _validateRegexp (String regexp, Errors errors)
    {
        if (regexp != null)
        {
            try
            {
                Pattern.compile(regexp);
            }
            catch (PatternSyntaxException e)
            {
                if (getLogger().isDebugEnabled())
                {
                    getLogger().debug("The regular expression '" + regexp + "' is not a valid one.");
                }
                
                errors.addError(new I18nizableText("plugin.core-impl", "PLUGINS_CORE_VALIDATOR_REGEXP_FAILED"));
            }
        }
    }
}
