/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * A mixin to have methods for widgets (such as relative fields...)
 */
Ext.define('Ametys.form.Widget', {
    
    statics: {
        /**
         * Get the value of another item
         * @param {String} relativePath The relative path to another value
         * @param {Ext.form.Field/Object} data An extjs field when in a ConfigurableFormPanel or an object with 'record' and 'dataPath'
         * @return {Object} The value of the field. Can be null. undefined if there is no such field. 
         */
        getRelativeValue(relativePath, data) {
            if (!data)
            {
                return null;
            }
            else if (data.isComponent)
            {
                let form = data.form;
                if (!form)
                {
                    // repeater for example
                    form = data;
                    while (!form.isConfigurableFormPanel)
                    {
                        form = form.ownerCt;
                    }
                }
                
                // Configurable Form Panel
                let referencedField = form.getRelativeField(relativePath, data);
                return referencedField ? referencedField.getValue() : undefined;
            }
            else if (data.record && data.dataPath)
            {
                const SEPARATOR = "/";
                
                let prefix = data.dataPath.substring(0, data.dataPath.lastIndexOf(SEPARATOR));
                
                // Handling '..' in field name.
                Ext.Array.forEach(relativePath.split(SEPARATOR), function(pathPart) {
                    if (pathPart == '..')
                    {
                        prefix = prefix.substring(0, prefix.lastIndexOf(SEPARATOR));
                        relativePath = relativePath.substring(3);
                    }
                });
                
                let valuePath = prefix == '' ? relativePath : (prefix + SEPARATOR + relativePath);
                let cursorData = data.record.data;
                
                // Split on /
                let paths = valuePath.split(SEPARATOR);
                let acumulator = "";
                for (let path of paths)
                {
                    if (acumulator)
                    {
                        acumulator += SEPARATOR;
                    }
                    acumulator += path;
                    
                    if (acumulator.endsWith("]"))
                    {
                        // repeater
                        let i = acumulator.indexOf("[");
                        let index = parseInt(acumulator.substring(i + 1, acumulator.length - 1));
                        cursorData = cursorData[acumulator.substring(0, i) + "_repeater"].entries[index - 1].values;
                        acumulator = "";
                    }
                }
                
                let relativeValue = cursorData[acumulator];
                if (relativeValue === undefined)
                {
                    var message = "{{i18n PLUGINS_CORE_UI_WIDGET_UNKNOWN_FIELD}}" + valuePath;
                    this.getLogger().warn(message);
                }
                return relativeValue;
            }
            else
            {
                throw new Error("getRelativeValue not supported for " + data);
            }
        },
        
        /**
         * Set the value at the dataPath
         * @param {String} valuePath The path to the value to change
         * @param {Ext.data.Model} record The record to change
         * @param {Object} value the new value
         */
        setRecordValue: function(valuePath, record, value)
        {
            const SEPARATOR = "/";
            
            let cursorData = record.data;
            
            // Split on /
            let paths = valuePath.split(SEPARATOR);
            let acumulator = "";
            
            let firstRepeaterAcumulator;
            
            for (let path of paths)
            {
                if (acumulator)
                {
                    acumulator += SEPARATOR;
                }
                acumulator += path;
                
                if (acumulator.endsWith("]"))
                {
                    // repeater
                    let i = acumulator.indexOf("[");
                    let index = parseInt(acumulator.substring(i + 1, acumulator.length - 1));
                    
                    let pointer = acumulator.substring(0, i) + "_repeater";
                    
                    if (!firstRepeaterAcumulator)
                    {
                        firstRepeaterAcumulator = pointer;
                    }
                    
                    cursorData = cursorData[pointer].entries[index - 1].values;
                    acumulator = "";
                }
            }
            
            if (firstRepeaterAcumulator)
            {
                cursorData[acumulator] = value;
                
                record.set(firstRepeaterAcumulator.substring(0, firstRepeaterAcumulator.length - "_repeater".length), record.get(firstRepeaterAcumulator));
            }
            else
            {
                record.set(acumulator, value);
            }
        },

        /**
         * Get the values corresponding the model path. Will seek in repeaters.
         * @param {Ext.data.Model} record The record to analyse
         * @param {String} modelPath The model path to consider
         * @return {Object} An object where id is the dataPath and value the corresponding value
         */
        getRecordValues: function(record, modelPath)
        {
            let values = {};
            
            const SEPARATOR = "/";
                
            // Split on /
            let paths = modelPath.split(SEPARATOR);
            seek(paths, 0, "", "", record.data, record.data);
            
            function seek(paths, index, totalAccumulator, accumulator, cursorData)
            {
                let path = paths[index];
                
                if (accumulator)
                {
                    accumulator += SEPARATOR;
                }
                if (totalAccumulator)
                {
                    totalAccumulator += SEPARATOR;
                }
                accumulator += path;
                totalAccumulator += path;
                
                if (accumulator.endsWith("]"))
                {
                    // repeater entry
                    let i = accumulator.indexOf("[");
                    let index = parseInt(accumulator.substring(i + 1, accumulator.length - 1));
                    
                    seek(paths, index + 1, totalAccumulator, "" , cursorData[accumulator.substring(0, i) + "_repeater"].entries[index - 1].values);
                }
                else if (cursorData[accumulator + "_repeater"])
                {
                    // repeater loop
                    let entries = cursorData[accumulator + "_repeater"].entries;
                    for (let entry of entries)
                    {
                        seek(paths, index + 1, totalAccumulator + "[" + entry.position + "]", "", entry.values)
                    }
                }
                else if (index < paths.length - 1)
                {
                    seek(paths, index+1, totalAccumulator, accumulator, cursorData)
                }
                else
                {
                    values[totalAccumulator] = cursorData[accumulator];
                }
            }
            
            return values;
        },
        
        /**
         * Register a listener when other values changed
         * @param {String/String[]} relativePaths The relative paths to other values
         * @param {Ext.form.Field/Object} data An extjs field when in a ConfigurableFormPanel or an object with 'grid' and 'dataPath'
         * @param {Function} handler The on change handler
         * @param {String} handler.relativePath The relative path to the value that has triggered the on change event.
         * @param {Ext.form.Field/Object} handler.data The data
         * @param {Object} handler.newValue The new value
         * @param {Object} handler.oldValue The old value
         * @param {Object} scope The scope handler. Default to the field or data.grid.
         */
        onRelativeValueChange: function(relativePaths, data, handler, scope) 
        {
            relativePaths = Ext.Array.from(relativePaths);
                
            if (!data)
            {
                throw new Error("Cannot register a reltive value change on a null component");
            }
            else if (data.isComponent)
            {
                let form = data.form;
                if (!form)
                {
                    // repeater for example
                    form = data;
                    while (!form.isConfigurableFormPanel)
                    {
                        form = form.ownerCt;
                    }
                }
                
                for (let relativePath of relativePaths)
                {
                    // Configurable Form Panel
                    function proxyHandler(field, newValue, oldValue)
                    {
                        // do not simple call handler() to keep the 'this' scope
                        handler.call(this, relativePath, data, newValue, oldValue);
                    }
                    form.onRelativeFieldsChange(relativePath, data, proxyHandler, scope);
                }
            }
            else if (data.grid && data.dataPath)
            {
                for (let relativePath of relativePaths)
                {
                    const SEPARATOR = "/";
                    
                    let path = relativePath;
                    
                    let prefix = data.dataPath.substring(0, data.dataPath.lastIndexOf(SEPARATOR));
                    
                    // Handling '..' in field name.
                    Ext.Array.forEach(path.split(SEPARATOR), function(pathPart) {
                        if (pathPart == '..')
                        {
                            prefix = prefix.substring(0, prefix.lastIndexOf(SEPARATOR));
                            path = path.substring(3);
                        }
                    });
                    
                    let valuePath = prefix == '' ? path : (prefix + SEPARATOR + path);
                    
                    data.grid.changeListeners = data.grid.changeListeners || {};
                    data.grid.changeListeners[valuePath] = data.grid.changeListeners[valuePath] || [];
                    data.grid.changeListeners[valuePath].push(function(localData, newValue, oldValue) {
                        handler.call(this, relativePath, {grid: data.grid, dataPath: data.dataPath, record: localData.record}, newValue, oldValue);
                    });
                }
            }
            else
            {
                throw new Error("onRelativeValueChange not supported for " + data);
            }
        },
        
        /**
         * Evaluate the disable conditions of the field now
         * @param {Ext.form.Field/Object} data An extjs field or the object with {disableCondition, dataPath and record}
         * @return {Boolean} Should disable? 
         */
        evaluateDisableCondition: function(data) 
        {
            if (data.disableCondition === 'string')
            {
                data.disableCondition = JSON.parse(data.disableCondition, data);
            }
            
            return Ametys.form.Widget._evaluateDisableCondition(data.disableCondition, data);
        },
        
        /**
         * @private
         * Evaluates the disable condition when a matching field is changing and enables/disables the field accordingly.
         * @param {Object} disableCondition the disable condition.
         * @param {Ext.form.Field/Object} data An extjs field or the object with grid info
         */
        _evaluateDisableCondition: function(disableCondition, data)
        {
            if (!disableCondition || !disableCondition.conditions && !disableCondition.condition)
            {
                return false;
            }
            
            var disable = disableCondition['type'] != "and" ? false : true;
            
            if (disableCondition.conditions)
            {
                var conditionsList = disableCondition.conditions,
                    conditionsListLength = conditionsList.length;
                
                for (var i = 0; i < conditionsListLength; i++)
                {
                    var result = Ametys.form.Widget._evaluateDisableCondition(conditionsList[i], data);
                    disable = disableCondition['type'] != "and" ? disable || result : disable && result;
                }
            }
            
            if (disableCondition.condition)
            {
                var conditionList = disableCondition.condition,
                    conditionListLength = conditionList.length;
                
                for (var i = 0; i < conditionListLength; i++)
                {
                    var id = conditionList[i]['id'],
                        op = conditionList[i]['operator'],
                        val = conditionList[i]['value'];
                        
                    var result = Ametys.form.Widget._evaluateCondition(id, op, val, data);
                    disable = disableCondition['type'] != "and" ? disable || result : disable && result;
                }
            }
            
            return disable;
        },
        
        /**
         * @private
         * Evaluates a single condition.
         * @param {String} id the id of the field.
         * @param {String} operator the operator.
         * @param {String} untypedValue the untyped value the field's value will be compared to.
         * @param {Ext.form.Field/Object} data An extjs field or the object with grid info
         * @return {Boolean} result true if the condition is verified, false otherwise.
         */
        _evaluateCondition: function(id, operator, untypedValue, data)
        {
            let type;
            if (data.isComponent)
            {
                let form = data.form;
                if (!form)
                {
                    // repeater for example
                    form = data;
                    while (!form.isConfigurableFormPanel)
                    {
                        form = form.ownerCt;
                    }
                }
                
                let formField = form.getRelativeField(id, data);
                if (!formField)
                {
                    return false; // no field, not disabled...
                }
                type = formField.type;
            }
            else if (data.record)
            {
                const SEPARATOR = "/";
                
                let localId = id;
                while (localId.startsWith('..' + SEPARATOR))
                {
                    localId = localId.substring(3);
                    data.record = Ametys.plugins.cms.search.SearchGridHelper._getParentRecord(data.record);
                }
                
                let field = data.record.getField(localId);
                if (!field)
                {
                    return false; // no field, not disabled...
                } 
                type = field.ftype || field.type;
            }
            else
            {
                throw new Error("Disable conditions not supported for " + data);
            }
            
            var fieldValue = Ametys.form.Widget.getRelativeValue(id, data);
            fieldValue = Ametys.form.widget.Externalizable.getValueInUse(fieldValue);
            
            var typedValue = Ametys.form.Widget._convertUntypedValue(type, untypedValue);

            if (fieldValue === undefined || fieldValue === "")
            {
                fieldValue = null;
            }
            if (typedValue === undefined || typedValue === "")
            {
                typedValue = null;
            }
                         
            switch (operator)
            {
                case "gt" : 
                    if (!Ext.isArray(fieldValue))
                    {
                        return fieldValue > typedValue;
                    }
                    else
                    {
                        // All entries need to match operator
                        for (var i = 0; i < fieldValue.length; i++)
                        {
                            if (!(fieldValue[i] > typedValue))
                            {
                                // One entry does not match!
                                return false;
                            }
                        }
                        return true;
                    }
                case "geq" : 
                    if (!Ext.isArray(fieldValue))
                    {
                        return fieldValue >= typedValue;
                    }
                    else
                    {
                        // All entries need to match operator
                        for (var i = 0; i < fieldValue.length; i++)
                        {
                            if (!(fieldValue[i] >= typedValue))
                            {
                                // One entry does not match!
                                return false;
                            }
                        }
                        return true;
                    }
                case "lt" : 
                    if (!Ext.isArray(fieldValue))
                    {
                        return fieldValue < typedValue;
                    }
                    else
                    {
                        // All entries need to match operator
                        for (var i = 0; i < fieldValue.length; i++)
                        {
                            if (!(fieldValue[i] < typedValue))
                            {
                                // One entry does not match!
                                return false;
                            }
                        }
                        return true;
                    }
                case "leq" : 
                    if (!Ext.isArray(fieldValue))
                    {
                        return fieldValue <= typedValue;
                    }
                    else
                    {
                        // All entries need to match operator
                        for (var i = 0; i < fieldValue.length; i++)
                        {
                            if (!(fieldValue[i] <= typedValue))
                            {
                                // One entry does not match!
                                return false;
                            }
                        }
                        return true;
                    }
                case "neq" : 
                    if (!Ext.isArray(fieldValue))
                    {
                        return fieldValue  !== typedValue; // null and empty string should be considered the same
                    }
                    else
                    {
                        // All entries need to match operator
                        for (var i = 0; i < fieldValue.length; i++)
                        {
                            if (!(fieldValue[i] !== typedValue))
                            {
                                // One entry does not match!
                                return false;
                            }
                        }
                        return true;
                    }
                case "eq" : 
                    if (!Ext.isArray(fieldValue))
                    {
                        return fieldValue === typedValue;
                    }
                    else
                    {
                        // One entry needs to match operator
                        for (var i = 0; i < fieldValue.length; i++)
                        {
                            if (fieldValue[i] === typedValue)
                            {
                                // One entry does match!
                                return true;
                            }
                        }
                        return false;
                    }
                default :
                    throw "Unknown operator " + operator;
            }
        },
        
        /**
         * @private
         * Converts an untyped value to a typed value according the given type
         * @param {String} fieldType The type to convert into
         * @param {String} untypedValue The untyped value, as a String
         * @return The typed value
         */
        _convertUntypedValue: function (fieldType, untypedValue)
        {
            switch (fieldType) {
                case "boolean":
                    if (Ext.isString(untypedValue))
                    {
                        return untypedValue.toLowerCase() === "true"
                    }
                    else if (untypedValue === undefined || untypedValue === null)
                    {
                        return untypedValue;
                    }
                    return !!untypedValue;
                case "long":
                    return parseInt(untypedValue);
                case "double":
                    return parseFloat(untypedValue);
                case "date":
                    return Ext.Date.parse(untypedValue, Ext.Date.patterns.ISO8601DateTime);
                default:
                    return untypedValue;
            }
        },
    }
});