/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * @private
 * Tool that the display all the profile assignments of a user
 */

Ext.define('Ametys.plugins.coreui.profiles.UserProfilesTool', {
    extend: "Ametys.tool.Tool",

    createPanel: function() {
        // return an empty container.
        // once we know the profile list, we will be able to fill it with a grid
        return Ext.create('Ext.container.Container', {flex:1, layout: 'fit'});
    },
    
    setParams: function(params)
    {
        this.callParent(arguments);
        
        this._login = params.login;
        this._populationId = params.populationId;
        // Call the server to retrieve the user full name and update the tool title
        // (or close the tool if the user doesn't exists)
        this.serverCall('getUser', [this._login, this._populationId], this._getUserCb, {scope: this, ignoreCallbackOnError: false});
        
        // mark the tool out of date to start the process or retrieving the profiles
        this.showOutOfDate();
    },
    
    /**
     * @private
     * The callback for getuser
     * @param {Object} user the user's properties
     */
    _getUserCb: function(user)
    {
        if (user)
        {
            this._fullName = user.fullname || user.login + " / " + user.populationLabel;
            
            this.setTitle(new Ext.Template("{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_LABEL}}").apply([user.fullname]));
            this.setDescription(new Ext.Template("{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_DESCRIPTION}}").apply([user.fullname, user.login, user.populationLabel]))
        }
        else
        {
            this.close();
            Ametys.notify({
                title: "{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_AUTOCLOSE_TITLE}}",
                description: "{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_AUTOCLOSE_DESCRIPTION}}",
                type: "warn"
            });
            this.getLogger().error("Unknown user " + this._login + "#" + this._populationId);
        }
    },
    
    
    refresh: function (manual)
    {
        this.showRefreshing();
        
        // retrieve the profiles to configure the grid panel (every profile is a column)
        Ametys.data.ServerComm.send({
            plugin: "core",
            url: 'rights/profiles.json',
            parameters: {
                limit: null,
                context: this._getProfilesContext()
            },
            priority: Ametys.data.ServerComm.PRIORITY_MAJOR,
            callback: {
                handler: this._getProfilesCb,
                scope: this,
                arguments: []
            },
            errorMessage: true,
            waitMessage: false,
            responseType: 'text'
        });
    },
    
    /**
     * @protected
     * Get the context to display profiles (additionnaly to null context)
     * @returns {String} the context to display or empty
     */
    _getProfilesContext: function()
    {
        return "";
    },
    
    /**
     * @private
     * Callback function after retrieving the profiles
     * @param {Object} response The server response
     */
    _getProfilesCb: function(response)
    {
        // This async call can be done in parallele with the getUser call
        // If no user is found, the getUser will close the tool and we need to ignore the result
        if (this.isNotDestroyed())
        {
            // The server returned a string json, decode it into an array of object
            var stringData = response.firstChild.textContent;
            var data = Ext.decode(stringData, true);
            
            var wrapper = this.getContentPanel();
            wrapper.remove('user-profiles-grid');
            wrapper.add(this._createAssignmentsGrid(data.profiles));
            
            this._profiles = data.profiles;
            
            // Do not send the showRefreshed now. We don't know if the store is loaded yet.
            // Instead, the showRefreshed is provided as a callback to the store load.
            // this.showRefreshed();
        }
    },
    
    /**
     * @private
     * Create a grid displaying all the assignments of a user
     * @param {Object[]} a list of object representing the profiles
     * @returns {Ext.grid.Panel} the assignment grid panel
     */
    _createAssignmentsGrid: function(profiles) {
        this._store = Ext.create('Ext.data.Store', this._getStoreCfg(profiles));
        
        this._assignmentsGrid = Ext.create("Ext.grid.Panel", {
            dockedItems: this._getGridDockedItemsCfg(),
            
            itemId: 'user-profiles-grid',
            store: this._store,
            
            // First the context then all the profiles
            columns: this._getFirstColumn().concat(Ametys.plugins.coreui.profiles.ProfilesGridHelper.getProfilesColumns(profiles, this._renderAssignment, this)),
            
            enableColumnHide: false, // disable column hiding by the user. Column visibility is handled by the right filter
            scrollable: true,
            
            stateful: true,
            stateId: this.self.getName() + "$grid",
            features: [
                {
                    ftype:'grouping',
                    enableGroupingMenu: false,
                    groupHeaderTpl: [
                        '{children:this.formatRightAssignmentContext} ({children.length})',
                        {
                            formatRightAssignmentContext: function(children)
                            {
                                return children[0].get('rightAssignmentContextLabel');
                            }
                        }
                    ],
                    groupers: [{
                        property: 'rightAssignmentContext',
                        groupFn: function(item) {return item.get('rightAssignmentContextId')},
                        sorterFn: function (record1, record2) {
                            var localRecord1 = record1, localRecord2 = record2;
                            if (record1.isGroup)
                            {
                                localRecord1 = record1.items[0];
                                localRecord2 = record2.items[0];
                            }
                            var label1 = localRecord1.data.rightAssignmentContext.label,
                                label2 = localRecord2.data.rightAssignmentContext.label;
                            // make sure that if two group with the same label but not the same id
                            // are correctly sorted. Otherwise, this will create multiple groups…
                            if (label1 == label2)
                            {
                            var id1 = localRecord1.get('rightAssignmentContextId'),
                                id2 = localRecord2.get('rightAssignmentContextId');
                                return id1.localeCompare(id2);
                            }
                            else
                            {
                                return label1.localeCompare(label2);
                            }
                        },
                        direction: 'ASC'
                    }]
                }
            ]
        });
        
        // reset column visibility
        // column are stateful, but the visibility is linked to the right filter and not the user interaction
        Ext.Array.forEach(this._assignmentsGrid.getColumns(), function(column, index) {
            column.setVisible(true);
        }, this);
        
        return this._assignmentsGrid;
    },

    /**
     * @private
     * Get the store configuration for the assignments grid
     * @param {Object[]} a list of object representing all the existing profiles
     * @returns {Object} the store configuration
     */
    _getStoreCfg: function(profiles)
    {
        return {
            model: 'Ametys.plugins.coreui.profiles.UserProfilesTool.Entry',
            proxy: {
                type: 'ametys',
                role: this.getServerRole(),
                serverId : this.getServerId(),
                methodName: 'getUserPermissions',
                extraParams: {
                    login: this._login,
                    population: this._populationId,
                    ctxParams: Ametys.getAppParameters()
                },
                methodArguments: ['login', 'population', 'ctxParams'],
            },
            autoLoad: {
                scope: this,
                callback: function(records, operation, success) {
                    this.showRefreshed();
                }
            },
            groupField: 'rightAssignmentContext',
            sorters: [{property: 'rootContext', direction:'DESC'}, {property: 'contextLabel', direction:'ASC'}]
        }
    },
    
    /**
     * @private
     * Get the definition of the components to dock on the assignments grid
     * @returns {Object/Object[]} the component definition
     */
    _getGridDockedItemsCfg: function()
    {
        return {
            dock: 'top',
            xtype: 'toolbar',
            layout: {
                type: 'hbox',
                align: 'stretch'
            },
            border: false,
            defaultType: 'textfield',
            items: [{
                xtype: 'component',
                html: '{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_FILTERS}}'
            }, {
                xtype: 'edition.right',
                itemId: 'profile-filter',
                name: 'profile-filter',
                cls: 'ametys',
                allowBlank: true,
                multiple: false,
                stacked: "false",
                width: 400,
                emptyText: "{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_RIGHT_FILTER}}",
                listeners: {change: Ext.bind(this._filterByRight, this)}
            }, {
                itemId: 'context-filter',
                name: 'context-filter',
                cls: 'ametys',
                width: 400,
                emptyText: "{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_CONTEXT_FILTER}}",
                listeners: {change: Ext.Function.createBuffered(this._filterByContext, 500, this)}
            }, {
                xtype: 'button',
                // Clear filter
                tooltip: "{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_CLEAR_CONTEXT_FILTER}}",
                handler: Ext.bind (this._clearFilter, this, ['context-filter'], false),
                iconCls: 'a-btn-glyph ametysicon-eraser11 size-16',
                cls: 'a-btn-light'
            }
            ]
        };
    },
    
    _getFirstColumn: function()
    {
        return [
            {
                stateId: 'grid-context-label',
                text: "",
                locked: true,
                lockable: false,
                dataIndex: "contextLabel",
                width: 300,
                minWidth: 300,
                sortable: false,
                hideable: false,
            }
        ];
    },
    
    /**
     * @private
     * Filters the columns by right
     * @param {Ametys.form.widget.Right} field The right field
     */
    _filterByRight: function(field)
    {
        Ext.suspendLayouts();
        var rightId = field.getValue();
        if (Ext.isEmpty(rightId))
        {
            Ext.Array.forEach(this._assignmentsGrid.getColumns(), function(column, index) {
                column.setVisible(true);
            }, this);
            this._store.removeFilter('emptyProfile');
        }
        else
        {
            // Computes the columns to let visible
            var matchingProfiles = [];
            Ext.Array.forEach(this._profiles, function(profile) {
                if (Ext.Array.contains(profile.rights, rightId))
                {
                    matchingProfiles.push(profile.id);
                }
            }, this);
            
            // Hide the others (except the first column)
            Ext.Array.forEach(this._assignmentsGrid.getColumns(), function(column, index) {
                var visible = index == 0 || Ext.Array.contains(matchingProfiles, column.dataIndex);
                column.setVisible(visible);
            }, this);
            
            this._store.addFilter([{
                id: 'emptyProfile',
                filterFn: function(record)
                {
                    for (var profile of matchingProfiles)
                    {
                        var assignment = record.get(profile)
                        if (assignment != undefined && assignment.accessResult != Ametys.plugins.coreui.rights.AccessResult.UNKNOWN)
                        {
                            return true;
                        }
                    }
                    return false;
                }
            }]);
        }
        Ext.resumeLayouts(true);
    },
    
    /**
     * @private
     * Filters the row by the context label
     * @param {Ext.form.field.Text} field The context field
     */
    _filterByContext: function(field)
    {
        Ext.suspendLayouts();
        var query = field.getValue();
        if (Ext.isEmpty(query))
        {
            this._store.removeFilter('context');
        }
        else
        {
            this._store.addFilter([{
                id: 'context',
                value: query,
                filterFn: function(item) {return item.data.context.label != null && item.data.context.label.toLowerCase().includes(this.getValue().toLowerCase())}
            }])
        }
        Ext.resumeLayouts(true);
    },
    
    /**
     * @private
     * Clear the current text filter
     * @param {String} id The input id
     */
    _clearFilter: function(id)
    {
        this.getContentPanel().down("#"+id).reset();
    },
    
    /**
     * @private
     * Renderer for the assignment cells, which draws an icon representing the assignment
     * @param {Object} value The data value for the current cell
     * @param {Object} metaData A collection of metadata about the current cell
     * @param {Ext.data.Model} record The record for the current row
     * @return {String} The HTML string to be rendered
     */
    _renderAssignment: function(value, metaData, record)
    {
        var value = value || {};
        var glyph = this._computeGlyph(value.accessResult);
        
        var tooltip = this._computeTooltip(value.assignments);
        metaData.tdAttr = `data-qtip="${ Ext.String.htmlEncode(tooltip)}"`; // surround with quote to avoid breaking rendering
        
        return '<span class="a-grid-access-result-glyph ' + glyph + '"/>';
    },
    
    /**
     * @private
     * Gets a list of css classes based on an access result.
     * @param accessResult a string representing an access result
     * @return {String} The HTML string to be rendered
     */
    _computeGlyph: function (accessResult)
    {
        switch (accessResult) 
        {
            case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_ALLOWED:
            case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_ALLOWED:
                return "ametysicon-check allowed";
            case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_DENIED:
            case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_DENIED:
                return "ametysicon-cross denied"
            case Ametys.plugins.coreui.rights.AccessResult.USER_ALLOWED:
            case Ametys.plugins.coreui.rights.AccessResult.GROUP_ALLOWED:
                return "ametysicon-check-1 allowed";
            case Ametys.plugins.coreui.rights.AccessResult.USER_DENIED:
            case Ametys.plugins.coreui.rights.AccessResult.GROUP_DENIED:
                return "ametysicon-cross-1 denied";
            case Ametys.plugins.coreui.rights.AccessResult.UNKNOWN:
            default:
                // Undetermined
                return "ametysicon-minus-symbol unknown disabled";
        }
    },
    
    /**
     * @private
     * Compute the tool tip for a record
     * @param {Object[]} assignments A list of access result and their details
     * @return {String} The HTML string to be rendered
     */
    _computeTooltip: function (assignments)
    {
        var tooltip = '';
        
        for (i in assignments)
        {
            var assignment = assignments[i];
            var glyph = this._computeTooltipGlyph(assignment.accessResult);
            var label = this._computeTooltipLabel(assignment);
            tooltip += `<span class="a-grid-access-result-tooltip"><i class="icon ${glyph}"></i>${label}</span>`;
        }
        
        return tooltip;
    },
    
    /**
     * @private
     * Gets a list of css classes based on an access result.
     * @param accessResult a string representing an access result
     * @return {String} The HTML string to be rendered
     */
    _computeTooltipGlyph: function (accessResult)
    {
        switch (accessResult) 
        {
            case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_ALLOWED:
            case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_ALLOWED:
            case Ametys.plugins.coreui.rights.AccessResult.GROUP_ALLOWED:
            case Ametys.plugins.coreui.rights.AccessResult.USER_ALLOWED:
                return "ametysicon-check-1 allowed";
            case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_DENIED:
            case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_DENIED:
            case Ametys.plugins.coreui.rights.AccessResult.GROUP_DENIED:
            case Ametys.plugins.coreui.rights.AccessResult.USER_DENIED:
                return "ametysicon-cross-1 denied";
            case Ametys.plugins.coreui.rights.AccessResult.UNKNOWN:
            default:
                // Undetermined
                return "ametysicon-minus-symbol unknown";
        }
    },
    
    /**
     * @private
     * Compute the tool tip for a record
     * @param {Object} assignment An object representing an access result and its details
     * @return {String} The HTML string to be rendered
     */
    _computeTooltipLabel: function(assignmentInfo)
    {
        switch (assignmentInfo.accessResult)
        {
            case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_ALLOWED:
            case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_DENIED:
                return "{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_ASSIGNMENT_TOOLTIP_ANONYMOUS_TARGET_LABEL}}";
            case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_ALLOWED:
            case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_DENIED:
                return "{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_ASSIGNMENT_TOOLTIP_ANYCONNECTED_TARGET_LABEL}}";
            case Ametys.plugins.coreui.rights.AccessResult.GROUP_ALLOWED:
            case Ametys.plugins.coreui.rights.AccessResult.GROUP_DENIED:
                return Ext.String.format("{{i18n PLUGINS_CORE_UI_TOOL_USER_PROFILES_ASSIGNMENT_TOOLTIP_GROUP_TARGET_LABEL}}", assignmentInfo.target.label);
            case Ametys.plugins.coreui.rights.AccessResult.USER_ALLOWED:
            case Ametys.plugins.coreui.rights.AccessResult.USER_DENIED:
                return assignmentInfo.target.fullname;
            case Ametys.plugins.coreui.rights.AccessResult.UNKNOWN:
            default:
                return "";
        }
    },
    
    getMBSelectionInteraction: function() {
        return Ametys.tool.Tool.MB_TYPE_NOSELECTION;
    }
});

 /**
 * This class is the data model for user profile grid entries.
 * @private
 */
Ext.define('Ametys.plugins.coreui.profiles.UserProfilesTool.Entry', {
    extend: 'Ext.data.Model',
    
    fields: [
        {name: 'rightAssignmentContextId', mapping: 'rightAssignmentContext.id'},
        {name: 'rightAssignmentContextLabel', mapping: 'rightAssignmentContext.label'},
        
        {name: 'contextId', mapping: 'context.id'},
        {name: 'contextLabel', type: 'string', mapping: 'context.label'},
        {name: 'rootContext', type: 'boolean', mapping: 'context.isRoot'}
    ]
});

