/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/**
 * Tool to display the result of all relevant AccessController for the context
 */

Ext.define('Ametys.plugins.coreui.rights.ExplainUserRightTool', {
    extend: "Ametys.tool.SelectionTool",
    
    constructor: function(config)
    {
        this.callParent(arguments);
        
        // Add the selection changed listener because we listen to all bus message
        Ametys.message.MessageBus.on(Ametys.message.Message.SELECTION_CHANGED, this._onSelectionChanged, this);

        // Instantiate all the convertor once and for all
        this._convertor = [];
        for (idx in config.convertors)
        {
            var convertor = Ext.create(config.convertors[idx].className, config.convertors[idx].parameters);
            convertor.setServerId(config.convertors[idx].serverId);
            this._convertor.push(convertor)
        }
        
        // Create the result template
        this._resultTpl = new Ext.XTemplate(
            `
            {[this.setAnyUserMode(values.mode)]}
            <h2>{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_SUMMARY_LABEL}}</h2>
            <p><i class="icon {[this.computeGlyph(values.result)]}"></i>{[this.computeSummaryLabel(values.result)]}</p>
            
            <h2>{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_DETAILS_LABEL}}</h2>
            <tpl if="values.result != 'UNKNOWN'">
                <ul class="explanation-list">
                <tpl for="explanations">
                    <li><i class="icon {[this.computeGlyph(values.result)]}"></i>{[this.computeLabel(values)]}</li>
                </tpl>
                </ul>
            <tpl else>
                <p>-</p>
            </tpl>
            `,
            {
                setAnyUserMode(mode)
                {
                    this._anyUser = mode == "anyUser";
                },
                
                computeGlyph: function(result)
                {
                    switch (result) 
                    {
                        case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_ALLOWED:
                        case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_ALLOWED:
                        case Ametys.plugins.coreui.rights.AccessResult.GROUP_ALLOWED:
                        case Ametys.plugins.coreui.rights.AccessResult.USER_ALLOWED:
                            return "ametysicon-check-1 allowed";
                        case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_DENIED:
                        case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_DENIED:
                        case Ametys.plugins.coreui.rights.AccessResult.GROUP_DENIED:
                        case Ametys.plugins.coreui.rights.AccessResult.USER_DENIED:
                        case Ametys.plugins.coreui.rights.AccessResult.UNKNOWN:
                        default:
                            // Unknown result is interpretted as denied
                            return "ametysicon-cross-1 denied";
                    }
                },
                
                computeSummaryLabel: function(result)
                {
                    switch (result) 
                    {
                        case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_ALLOWED:
                        case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_ALLOWED:
                        case Ametys.plugins.coreui.rights.AccessResult.GROUP_ALLOWED:
                        case Ametys.plugins.coreui.rights.AccessResult.USER_ALLOWED:
                            return this._anyUser ? "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_SUMMARY_ANY_USER_ALLOWED}}" : "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_SUMMARY_CURRENT_USER_ALLOWED}}";
                        case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_DENIED:
                        case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_DENIED:
                        case Ametys.plugins.coreui.rights.AccessResult.GROUP_DENIED:
                        case Ametys.plugins.coreui.rights.AccessResult.USER_DENIED:
                        case Ametys.plugins.coreui.rights.AccessResult.UNKNOWN:
                        default:
                            return this._anyUser ? "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_SUMMARY_ANY_USER_DENIED}}" : "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_SUMMARY_CURRENT_USER_DENIED}}";
                    }
                },
                
                computeLabel: function(explanation)
                {
                    var label;
                    switch (explanation.result) 
                    {
                        case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_ALLOWED:
                            label = "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_ANONYMOUS_ALLOWED}}";
                            break;
                        case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_ALLOWED:
                            label = "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_ANY_CONNECTED_ALLOWED}}";
                            break;
                        case Ametys.plugins.coreui.rights.AccessResult.GROUP_ALLOWED:
                            label = "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_GROUP_ALLOWED}}";
                            break;
                        case Ametys.plugins.coreui.rights.AccessResult.USER_ALLOWED:
                            label = "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_USER_ALLOWED}}";
                            break;
                        case Ametys.plugins.coreui.rights.AccessResult.ANONYMOUS_DENIED:
                            label = "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_ANONYMOUS_DENIED}}";
                            break;
                        case Ametys.plugins.coreui.rights.AccessResult.ANY_CONNECTED_DENIED:
                            label = "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_ANY_CONNECTED_DENIED}}";
                            break;
                        case Ametys.plugins.coreui.rights.AccessResult.GROUP_DENIED:
                            label = "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_GROUP_DENIED}}";
                            break;
                        case Ametys.plugins.coreui.rights.AccessResult.USER_DENIED:
                            label = "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_USER_DENIED}}";
                            break;
                        case Ametys.plugins.coreui.rights.AccessResult.UNKNOWN:
                        default:
                            // should never happens
                            label = "{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RESULT_UNKNOWN}}";
                    }
                    if (explanation.controllerLabel)
                    {
                        label += ` "<i data-qtip="${explanation.controllerId}">${explanation.controllerLabel}</i>"`;
                    }
                    else
                    {
                        label += ` "<i>${explanation.controllerId}</i>"`;
                    }
                    return label;
                }
            }
        );
        
        this._contextTpl = new Ext.XTemplate("...");
    },
    
    createPanel: function()
    {
        var me = this;
        
        // content will be filled with refresh
        this._contextCmp = Ext.create('Ext.Component', {ui: 'tool-hintmessage', anchor:"100%"});
        
        // content will be filled with refresh
        this._resultCmp = Ext.create('Ext.Component', {
            cls: 'results',
            anchor: "100%"
        });
        var currentUser = Ametys.getAppParameter('user');
        
        return Ext.create('Ext.container.Container', {
            layout: {
                type: 'card'
            },
            cls: 'uitool-explain-user-right',
            items: [
                {
                    itemId: 'invalid-selection-card',
                    xtype: 'component',
                    cls: 'a-panel-text-empty',
                    border: false,
                    html: ''
                }, {
                    itemId: 'result-card',
                    layout: "anchor",
                    defaults: {
                        anchor: "100%",
                        minWidth: 300,
                    },
                    scrollable: true,
                    items: [
                        this._contextCmp,
                        Ext.create("Ametys.form.field.RightByProfile", {
                            itemId: 'right',
                            fieldLabel: '{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RIGHT_WIDGET_LABEL}}',
                            labelAlign: 'top',
                            labelSeparator: '',
                            showProfilesIncludingRight: true,
                            value: Ametys.form.widget.Right.READER_RIGHT,
                            ametysDescription: '{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_RIGHT_WIDGET_DESC}}',
                            cls: 'ametys',
                            padding: "5 5 0 5",
                            allowBlank: false,
                            listeners: {change: function(field, newValue) {me._right = newValue; me.refresh(true);}}
                        }), {
                            xtype: 'edition.user',
                            itemId: 'user',
                            fieldLabel: '{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_USER_WIDGET_LABEL}}',
                            labelAlign: 'top',
                            labelSeparator: '',
                            cls: 'ametys',
                            padding: "5 5 0 5",
                            value: currentUser.login + "#" + currentUser.population,
                            multiple: false,
                            emptyText: `{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_USER_WIDGET_EMPTY}}`,
                            ametysDescription: `{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_USER_WIDGET_DESC}}`,
                            listeners: {change: function(field, newValue) {me._user = newValue; me.refresh(true);}}
                        }, 
                        this._resultCmp
                    ]
                }
            ]
        });
    },
    
    setParams: function(params)
    {
        this.callParent(arguments);
        
        this._switchMode(params.mode || "anyUser")
    },
    
    _switchMode: function(mode)
    {
        if(this._mode != mode)
        {
            this._mode = mode;
            var contextLabel;
            if (mode == 'anyUser')
            {
                this.getContentPanel().getComponent('result-card').getComponent('user').setVisible(true);
                
                contextLabel = `{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_CONTEXT_HEADER_ANY_USER}}`
            }
            else
            {
                this.getContentPanel().getComponent('result-card').getComponent('user').setVisible(false);
                
                contextLabel = `{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_CONTEXT_HEADER_CURRENT_USER}}`
            }
            
            contextLabel += `</br><tpl if="contextType">({contextType})</tpl> <strong data-qtip="{contextId}">{contextLabel}</strong>`
            this._contextTpl.set(contextLabel, true);
        }
    },
    
    _getMatchingSelectionTargets: function(message)
    {
        var me = this;
        return message.getTargets(
            function(target)
            {
                for (idx in me._convertor)
                {
                    if (me._convertor[idx].isSupportedTarget(target))
                    {
                        me._matchingConvertorIdx = idx;
                        return true;
                    }
                }
                return false;
            }
        );
    },
    
    // Overrides to allow target to context convertor to provide their own definition of
    // same targets
    areSameTargets: function(target1, target2)
    {
        // to be the same, two target must be handled by the same convertor
        if (this._currentConvertorIdx == this._matchingConvertorIdx)
        {
            return this._convertor[this._currentConvertorIdx].areSameTargets(target1, target2);
        }
        return false;
    },
    
    setNoSelectionMatchState: function (message)
    {
        var mainPanel = this.getContentPanel()
        if (message == null) // no message for empty selection to enable support for general context
        {
            // Do not call parent as it always send an upToDate signal
            // We will handle the _currentSelectionTargets and upToDate ourself
            
            // previous valid selection was not empty or there was an invalid selection
            if (this.getCurrentSelectionTargets().length > 0 || mainPanel.getLayout().getActiveItem().getItemId() != 'result-card')
            {
                // There was a previous selection.
                this._currentSelectionTargets = [];
                this.showOutOfDate(true);
            }
            else
            {
                // There was already no selection, do nothing
                this.showUpToDate(true);
            }
        }
        else
        {
            mainPanel.setActiveItem('invalid-selection-card');
            mainPanel.getComponent('invalid-selection-card').setHtml(message);
            // call parent to ensure normal operation
            this.callParent(arguments);
        }
    },
    
    refresh: function(manual)
    {
        var targets = this.getCurrentSelectionTargets(); // [][0] == undefined
        var target;
        // Store the current convertor
        if (targets.length == 1)
        {
            this._currentConvertorIdx = this._matchingConvertorIdx;
            target = targets[0];
        }
        else if (targets.length == 0 && Ametys.message.MessageBus.getCurrentSelectionMessage().getTargets().length == 0)
        {
            // reset the convertor 
            this._currentConvertorIdx = null;
        }
        else
        {
            // if selection multiple or unsupported, do nothing
            this.showUpToDate(true);
            return;
        }
        
        this.getContentPanel().setActiveItem('result-card');
        
        // Update context header
        var contextId = this._currentConvertorIdx != null ? this._convertor[this._currentConvertorIdx].convert(target) : null;
        var contextInfo = this._currentConvertorIdx != null ? this._convertor[this._currentConvertorIdx].getContextInfo(target) : {label:"{{i18n PLUGINS_CORE_UI_RIGHTS_EXPLAIN_TOOL_CONTEXT_WORKSPACE_CONTEXT_LABEL}}"};
        this._contextCmp.setHtml(this._contextTpl.apply({contextLabel: contextInfo.label, contextId: contextId, contextType: contextInfo.type}));

        // Check if right is set. If so call server. Otherwise stop and hide the result component
        if (this._right != null)
        {
            this._resultCmp.setVisible(true);
            this._updateResultCard(contextId);
        }
        else
        {
            this._resultCmp.setVisible(false);
            
            this.showUpToDate();
        }
    },
    
    _updateResultCard: function(context)
    {
        var serverId = this._currentConvertorIdx != null ? this._convertor[this._currentConvertorIdx].getServerId() : null;
        
        var arguments = {
            context: context,
        };
        
        if (this._mode == "anyUser")
        {
            arguments.user = this._user;
            this.serverCall('explainRight', [this._right, this._user, context, serverId], this._explainRightCb, {arguments: arguments, cancelCode: this.getId() + "$explain"});
        }
        else
        {
            this.serverCall('explainCurrentUserRight',[this._right, context, serverId], this._explainRightCb, {arguments: arguments, cancelCode: this.getId() + "$explain"})
        }
    },
    
    _getFirstSupportingConvertor: function(target)
    {
        for (idx in this._convertor)
        {
            if (this._convertor[idx].isSupportedTarget(target))
            {
                return this._convertor[idx];
            }
        }
        return null;
    },
    
    _explainRightCb: function(response, args)
    {
        this._resultCmp.setHtml(
            this._resultTpl.apply({
                result : response.result,
                explanations: response.explanations,
                user: args.user,
                mode: this._mode
            }));
        this.showUpToDate();
    }
   });

