/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.ui;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

import org.apache.commons.lang.StringUtils;

import org.ametys.core.right.RightAssignmentContext;

/**
 * Mark the annotated method as callable by a client-side component.
 */
@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.METHOD)
public @interface Callable 
{
    /** Constant to use to explicitly do not check right */
    public static final String SKIP_BUILTIN_CHECK = StringUtils.EMPTY;
    
    /** Constant to check read access */
    public static final String READ_ACCESS = "__READ_ACCESS";
    
    /**
     * Indicate how to check multiple rights
     */
    public enum RightMode
    {
        /** All the right must be allowed */
        AND,
        /** At least one right must be allowed */
        OR
    }
    
    /** 
     * The ids of rights to be checked
     * @return The ids or empty if there is no right protection
     */
    public String[] rights() default SKIP_BUILTIN_CHECK;
    
    /**
     * The mode to use to check multiple rights
     * @return The mode to use
     */
    public RightMode rightMode() default RightMode.OR;
    
    /**
     * Determines if the method can be accessed by a anonymous user
     * @return true if anonymous user is allowed
     */
    public boolean allowAnonymous() default false;
    
    /**
     * The context to check the right. Defaults to '/${WorkspaceName}'
     * @return The context
     */
    public String context() default "/${WorkspaceName}";
    
    /**
     * The id of a type of the {@link RightAssignmentContext} to use to convert JS object into into a Java object
     * @return The id of the {@link RightAssignmentContext}
     */
    public String rightContext() default "";
    
    /**
     * The index of method's argument holding the context value to be converted by {@link RightAssignmentContext}
     * @return The parameter's index
     */
    public int paramIndex() default -1;
}
