/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.ui.dispatcher;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.ResourceNotFoundException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.util.location.LocatedException;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.xml.sax.SAXParser;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import org.ametys.core.authentication.AuthenticateAction;
import org.ametys.core.util.IgnoreRootHandler;
import org.ametys.core.util.JSONUtils;
import org.ametys.core.util.URIUtils;
import org.ametys.plugins.core.ui.util.RequestAttributesHelper;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * This generator read the request incoming from the client org.ametys.servercomm.ServerComm component,
 * then dispatch it to given url
 * and aggregate the result 
 */
public class DispatchGenerator extends ServiceableGenerator
{
    /** Request Attributes Helper */
    protected RequestAttributesHelper _requestAttributesHelper;
    /** The source resolver */
    protected SourceResolver _sourceResolver;
    /** The dispatch process extension point */
    protected DispatchProcessExtensionPoint _dispatchProcessExtensionPoint;
    /** The JSON utils */
    protected JSONUtils _jsonUtils;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        
        _sourceResolver = (SourceResolver) smanager.lookup(org.apache.excalibur.source.SourceResolver.ROLE);
        _dispatchProcessExtensionPoint = (DispatchProcessExtensionPoint) manager.lookup(DispatchProcessExtensionPoint.ROLE);
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
        _requestAttributesHelper = (RequestAttributesHelper) manager.lookup(RequestAttributesHelper.ROLE);
    }
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        _setResponseHeader();
        
        String requestBodyAsString = _getRequestBody();
        Map<String, Object> requestBodyAsJson = _jsonUtils.convertJsonToMap(requestBodyAsString);
        
        String contextAsString = _getRequestContext();
        Map<String, Object> contextAsJson = _jsonUtils.convertJsonToMap(contextAsString);

        _dispatching(requestBodyAsJson, contextAsJson);
    }
    
    /**
     * Prepare the response header
     */
    protected void _setResponseHeader()
    {
        // Put the header for the client side ServerComm to ensure the authentication was correctly handled
        ObjectModelHelper.getResponse(objectModel).addHeader("Ametys-Dispatched", "true");
    }

    /** 
     * Get the request body as JSON string
     * @return the request body as JSON string
     */
    protected String _getRequestBody()
    {
        return ObjectModelHelper.getRequest(objectModel).getParameter("content");
    }

    /**
     * Get the request context as JSON string
     * @return the request context as JSON string
     */
    protected String _getRequestContext()
    {
        return ObjectModelHelper.getRequest(objectModel).getParameter("context.parameters");
    }

    /**
     * Dispatch the request
     * @param requestBody the request body as json object
     * @param contextAsMap the request context as json
     * @throws SAXException if failed to sax response
     */
    protected void _dispatching(Map<String, Object> requestBody, Map<String, Object> contextAsMap) throws SAXException
    {
        contentHandler.startDocument();
        XMLUtils.startElement(contentHandler, "responses");
        
        Map<String, Object> attributes = _requestAttributesHelper.saveRequestAttributes();
        contextAsMap.putAll(transmitAttributes(attributes));
        
        Map<String, Long> times = new HashMap<>();
        for (String parameterKey : requestBody.keySet())
        {
            @SuppressWarnings("unchecked")
            Map<String, Object> subRequestParameters = (Map<String, Object>) requestBody.get(parameterKey);
            long time = _dispatchingSubRequest(subRequestParameters, contextAsMap, parameterKey);
            times.put(parameterKey, time);
        }
        
        _requestAttributesHelper.restoreRequestAttributes(attributes);
        
        // Send measured times
        _saxMeasuredTimes(times);
        
        XMLUtils.endElement(contentHandler, "responses");
        contentHandler.endDocument();
    }
    
    /**
     * SAX the measured times for each sub requests
     * @param times the measured times
     * @throws SAXException if failed to sax times
     */
    protected void _saxMeasuredTimes(Map<String, Long> times) throws SAXException
    {
        AttributesImpl attrs = new AttributesImpl();
        attrs.addCDATAAttribute("duration", Long.toString(times.values().stream().mapToLong(Long::longValue).sum()));
        XMLUtils.startElement(contentHandler, "times", attrs);
        for (String parameterKey: times.keySet())
        {
            AttributesImpl internalAttrs = new AttributesImpl();
            internalAttrs.addCDATAAttribute("id", parameterKey);
            internalAttrs.addCDATAAttribute("duration", Long.toString(times.get(parameterKey)));
            XMLUtils.createElement(contentHandler, "time", internalAttrs); 
        }
        XMLUtils.endElement(contentHandler, "times");
    }
    
    /**
     * Dispatch a sub request
     * @param parametersAsMap the sub requets parameters as map
     * @param contextAsMap the request context as json
     * @param parameterKey the parameter key holding the sub request
     * @return the computed time to handle the sub request
     * @throws SAXException if failed to sax response
     */
    protected long _dispatchingSubRequest(Map<String, Object> parametersAsMap, Map<String, Object> contextAsMap, String parameterKey) throws SAXException
    {
        long t0 = System.currentTimeMillis();

        _removeRequestAttributes();
        
        for (String extension : _dispatchProcessExtensionPoint.getExtensionsIds())
        {
            DispatchRequestProcess processor = _dispatchProcessExtensionPoint.getExtension(extension);
            processor.preProcess(ObjectModelHelper.getRequest(objectModel));
        }

        _setContextInRequestAttributes(contextAsMap);

        String pluginOrWorkspace = (String) parametersAsMap.get("pluginOrWorkspace");
        String relativeUrl = (String) parametersAsMap.get("url");
        String responseType = (String) parametersAsMap.get("responseType");
        
        @SuppressWarnings("unchecked")
        Map<String, Object> requestParameters = (Map<String, Object>) parametersAsMap.get("parameters");
        _handleFiles(requestParameters, parameterKey);
        
        Source response = null;

        ResponseHandler responseHandler = null;
        try
        {
            String url = _createUrl(pluginOrWorkspace, relativeUrl, requestParameters != null ? requestParameters : new HashMap<String, Object>());
            
            if (getLogger().isInfoEnabled())
            {
                getLogger().info("Dispatching url '" + url + "'");
            }

            response = _sourceResolver.resolveURI(url, null, requestParameters);

            responseHandler = new ResponseHandler(contentHandler, parameterKey, "200");

            _handleResponse(response, responseType, responseHandler);
        }
        catch (Throwable e)
        {
            String message = String.format("Can not dispatch request '%s' : '%s' '%s' '%s'",  parameterKey , pluginOrWorkspace,  relativeUrl,  requestParameters);
            
            _handleError(responseHandler, parameterKey, message, e);
        }
        finally
        {
            _sourceResolver.release(response);
            
            for (String extension : _dispatchProcessExtensionPoint.getExtensionsIds())
            {
                DispatchRequestProcess processor = _dispatchProcessExtensionPoint.getExtension(extension);
                processor.postProcess(ObjectModelHelper.getRequest(objectModel));
            }
        }
        
        long t1 = System.currentTimeMillis();
        long time = t1 - t0;
        getLogger().debug("Request '" + parameterKey + "' took " + time + "ms");
        
        return time;
    }
    
    /**
     * Remove the request attributes
     */
    protected void _removeRequestAttributes()
    {
        _requestAttributesHelper.removeRequestAttributes();
    }
    
    /**
     * Sax the response in case of success 
     * @param response the response
     * @param responseType the type of response
     * @param responseHandler the handler to sax response
     * @throws IOException if failed to get response's input stream
     * @throws SAXException  if failed to sax response
     * @throws ProcessingException if failed to get sax parser
     */
    protected void _handleResponse(Source response, String responseType, ResponseHandler responseHandler) throws IOException, SAXException, ProcessingException
    {
        try (InputStream is = response.getInputStream())
        {
            if ("xml".equalsIgnoreCase(responseType))
            {
                SAXParser saxParser = null;
                try
                {
                    // DO NOT USE SitemapSource.toSAX in this case
                    saxParser = (SAXParser) manager.lookup(SAXParser.ROLE);
                    saxParser.parse(new InputSource(is), responseHandler);
                }
                catch (ServiceException e)
                {
                    throw new ProcessingException("Unable to get a SAX parser", e);
                }
                finally
                {
                    manager.release(saxParser);
                }
            }
            else 
            {
                responseHandler.startDocument();
                
                String data = IOUtils.toString(is, "UTF-8");
                if ("xml2text".equalsIgnoreCase(responseType))
                {
                    // removing xml prolog and surrounding 'text' tag
                    data = data.substring(data.indexOf(">", data.indexOf("?>") + 2) + 1, data.lastIndexOf("<"));
                }
                XMLUtils.data(responseHandler, data);
                
                responseHandler.endDocument();
            }
        }
    }
    
    /**
     * SAX eroor
     * @param responseHandler the handler to sax error
     * @param parameterKey the key of parameter
     * @param message the error message
     * @param e the throwable
     * @throws SAXException if failed to sax error
     */
    protected void _handleError(ResponseHandler responseHandler, String parameterKey, String message, Throwable e) throws SAXException
    {
        // Ensure SAXException are unrolled the right way
        getLogger().error(message, new LocatedException(message, e));
        
        // Makes the output xml ok 
        if (responseHandler != null)
        {
            responseHandler.exceptionFinish();
        }
        
        Throwable t = _unroll(e);
        
        String code = "500";
        if (t instanceof ResourceNotFoundException || t.toString().startsWith("org.apache.cocoon.ResourceNotFoundException:"))
        {
            code = "404";
        }
        
        AttributesImpl attrs = new AttributesImpl();
        attrs.addCDATAAttribute("id", parameterKey);
        attrs.addCDATAAttribute("code", code);
        
        XMLUtils.startElement(contentHandler, "response", attrs);
        XMLUtils.createElement(contentHandler, "message", _escape(_exceptionToMessageInformation(t)));
        XMLUtils.createElement(contentHandler, "stacktrace", _escape(_exceptionToStackTraceInformation(t)));
        XMLUtils.endElement(contentHandler, "response");
    }
    
    /**
     * Handle uploaded files
     * @param params the request parameters
     * @param parameterKey the parameter key holding the sub request
     */
    protected void _handleFiles(Map<String, Object> params, String parameterKey)
    {
        Map<String, Object> uploadedFiles = _extractFiles(parameterKey);
        if (!uploadedFiles.isEmpty())
        {
            for (Entry<String, Object> entry : uploadedFiles.entrySet())
            {
                Object parent = params;
                
                String[] cursors = StringUtils.split(entry.getKey(), ".");
                for (int i = 0; i < cursors.length; i++)
                {
                    String cursor = cursors[i];
                    
                    if (parent instanceof Map)
                    {
                        @SuppressWarnings("unchecked")
                        Map<String, Object> parentMap = (Map) parent;
                        
                        if (i == cursors.length - 1)
                        {
                            parentMap.put(cursor, entry.getValue());
                        }
                        else
                        {
                            parent = parentMap.get(cursor);
                        }
                    }
                    else if (parent instanceof List)
                    {
                        @SuppressWarnings("unchecked")
                        List<Object> parentList = (List) parent;
                        int cursorIndex = Integer.parseInt(cursor);
                        
                        if (i == cursors.length - 1)
                        {
                            parentList.remove(cursorIndex);
                            parentList.add(cursorIndex, entry.getValue());
                            
                        }
                        else
                        {
                            parent = parentList.get(cursorIndex);
                        }
                    }
                    else
                    {
                        getLogger().warn("Cannot replace files in objects that are not java.util.List nor java.util.Map: " + parent.getClass().getName());
                        break;
                    }
                }
            }
        }
    }
    
    private Map<String, Object> _extractFiles(String index)
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        Map<String, Object> files = new HashMap<>();
        
        // Extract the optional list of uploaded files
        Enumeration paramNames = request.getParameterNames();
        while (paramNames.hasMoreElements())
        {
            String paramName = (String) paramNames.nextElement();
            if (paramName.startsWith("request#" + index + "#"))
            {
                files.put(StringUtils.removeStart(paramName, "request#" + index + "#"), request.get(paramName));
            }
        }
        
        return files;
    }

    
    /**
     * Filters attributes that should be transmitted to the dispatched request
     * @param attributes The full list of attributes
     * @return The attributes filtered
     */
    protected Map<String, Object> transmitAttributes(Map<String, Object> attributes)
    {
        Map<String, Object> contextAsMap = new HashMap<>();
        
        String[] attributesToTransmit = new String[] 
        {
            AuthenticateAction.REQUEST_ATTRIBUTE_AUTHENTICATED,
            WorkspaceMatcher.IN_WORKSPACE_URL,
            WorkspaceMatcher.WORKSPACE_NAME,
            WorkspaceMatcher.WORKSPACE_THEME,
            WorkspaceMatcher.WORKSPACE_THEME_URL,
            WorkspaceMatcher.WORKSPACE_URI
        };
        
        for (String attributeToTransmit : attributesToTransmit)
        {
            contextAsMap.put(attributeToTransmit, attributes.get(attributeToTransmit));
        }

        return contextAsMap;
    }

    private Throwable _unroll(Throwable initial)
    {
        Throwable t = initial;
        while (t.getCause() != null || t instanceof SAXException && ((SAXException) t).getException() != null)
        {
            if (t instanceof SAXException)
            {
                t = ((SAXException) t).getException();
            }
            else
            {
                t = t.getCause();
            }
        }
        
        return t;
    }
    
    /**
     * Set the request attributes from the contextual parameters
     * @param contextAsMap The contextual parameters
     */
    protected void _setContextInRequestAttributes(Map<String, Object> contextAsMap)
    {
        if (contextAsMap != null)
        {
            Request request = ObjectModelHelper.getRequest(objectModel);
    
            contextAsMap.forEach((name, value) -> request.setAttribute(name, value));
        }
    }
    
    /**
     * Convert the exception to a viewable trace
     * @param t The exception
     * @return The visible message
     */
    protected String _exceptionToStackTraceInformation(Throwable t)
    {
        return ExceptionUtils.getStackTrace(t);
    }

    /**
     * Convert the exception to a viewable message
     * @param t The exception
     * @return The visible message
     */
    protected String _exceptionToMessageInformation(Throwable t)
    {
        return StringUtils.defaultString(t.getMessage());
    }

    private String _escape(String value)
    {
        return value.replaceAll("&", "&amp;").replaceAll("<", "&lt;".replaceAll(">", "&gt;"));
    }
    
    /**
     * Create url to call
     * @param pluginOrWorkspace the plugin or workspace name
     * @param relativeUrl the relative url
     * @param requestParameters the request parameters. Can not be null.
     * @return the full url
     */
    @SuppressWarnings("unchecked")
    protected String _createUrl(String pluginOrWorkspace, String relativeUrl, Map<String, Object> requestParameters)
    {
        StringBuilder url = new StringBuilder();
        
        String urlPrefix = _getUrlPrefix(pluginOrWorkspace);
        url.append(urlPrefix);
        
        url.append(_getRelativePath(relativeUrl));
        
        int queryIndex = relativeUrl.indexOf("?");
        
        if (queryIndex == -1 && !requestParameters.isEmpty())
        {
            // no existing parameters in request
            url.append("?");
            
            for (String key : requestParameters.keySet())
            {
                Object value = requestParameters.get(key);
                if (value instanceof List)
                {
                    List<Object> valueAsList = (List<Object>) value;
                    for (Object v : valueAsList)
                    {
                        if (v != null)
                        {
                            url.append(_buildQueryParameter(key, v));
                        }
                    }
                }
                else if (value != null)
                {
                    url.append(_buildQueryParameter(key, value));
                }
            }
        }
        else if (queryIndex > 0)
        {
            url.append("?");
            
            String queryUrl = relativeUrl.substring(queryIndex + 1, relativeUrl.length());
            String[] queryParameters = queryUrl.split("&");
            
            for (String queryParameter : queryParameters)
            {
                if (StringUtils.isNotBlank(queryParameter))
                {
                    String[] part = queryParameter.split("=");
                    String key = part[0];
                    String v = part.length > 1 ? part[1] : "";
                    String value = URIUtils.decode(v);
                    url.append(_buildQueryParameter(key, value));
                    
                    if (!requestParameters.containsKey(key))
                    {
                        requestParameters.put(key, value);
                    }
                }
            }
        }
        
        return url.toString();
    }
    
    private String _getRelativePath(String url)
    {
        int beginIndex = url.length() != 0 && url.charAt(0) == '/' ? 1 : 0;
        int endIndex = url.indexOf("?");
        return endIndex == -1 ? url.substring(beginIndex) : url.substring(beginIndex, endIndex);
    }
    
    private StringBuilder _buildQueryParameter(String key, Object value)
    {
        StringBuilder queryParameter = new StringBuilder();
        queryParameter.append(key);
        queryParameter.append("=");
        queryParameter.append(String.valueOf(value).replaceAll("%", "%25").replaceAll("=", "%3D").replaceAll("&", "%26").replaceAll("\\+", "%2B"));
        queryParameter.append("&");
        
        return queryParameter;
    }
    
    /**
     * Get the url prefix
     * @param pluginOrWorkspace the plugin or workspace name
     * @return the url prefix
     */
    protected String _getUrlPrefix (String pluginOrWorkspace)
    {
        StringBuffer url = new StringBuffer("cocoon://");
        if (pluginOrWorkspace != null && !pluginOrWorkspace.startsWith("_"))
        {
            url.append("_plugins/");
            url.append(pluginOrWorkspace);
            url.append("/");
        }
        else if (pluginOrWorkspace != null)
        {
            url.append(pluginOrWorkspace);
            url.append("/");
        }
        
        return url.toString();
    }
    
    /**
     * Wrap the handler ignore start and end document, but adding a response tag. 
     */
    public static class ResponseHandler extends IgnoreRootHandler
    {
        private final String _parameterKey;
        private final ContentHandler _handler;
        private final String _code;
        
        private final List<String> _startedElements;
        
        /**
         * Create the wrapper
         * @param handler The content handler to wrap
         * @param parameterKey The id of the response
         * @param code The status code of the response
         */
        public ResponseHandler(ContentHandler handler, String parameterKey, String code)
        {
            super(handler);
            _handler = handler;
            _parameterKey = parameterKey;
            _code = code;
            _startedElements = new ArrayList<>();
        }
        
        /**
         * Finish abruptly this handler to obtain a correct XML
         * @throws SAXException if an error occurred
         */
        public void exceptionFinish() throws SAXException
        {
            while (_startedElements.size() > 0)
            {
                XMLUtils.endElement(_handler, _startedElements.get(_startedElements.size() - 1));
                _startedElements.remove(_startedElements.size() - 1);
            }
        }
        
        @Override
        public void startDocument() throws SAXException
        {
            super.startDocument();
            
            AttributesImpl attrs = new AttributesImpl();
            attrs.addCDATAAttribute("id", _parameterKey);
            attrs.addCDATAAttribute("code", _code);
            XMLUtils.startElement(_handler, "response", attrs);

            _startedElements.add("response");
        }
        
        @Override
        public void startElement(String uri, String loc, String raw, Attributes a) throws SAXException
        {
            super.startElement(uri, loc, raw, a);
            _startedElements.add(loc);
        }

        @Override
        public void endElement(String uri, String loc, String raw) throws SAXException
        {
            super.endElement(uri, loc, raw);
            
            if (!StringUtils.equals(_startedElements.get(_startedElements.size() - 1), loc))
            {
                throw new SAXException("Sax events are not consistents. Cannot close <" + loc + "> while it should be <" + _startedElements.get(_startedElements.size() - 1) + ">");
            }
            
            _startedElements.remove(_startedElements.size() - 1);
        }
        
        @Override
        public void endDocument() throws SAXException
        {
            XMLUtils.endElement(_handler, "response");
            
            if (_startedElements.size() != 1)
            {
                throw new SAXException("Sax events are not consistents. Remaining " + _startedElements.size() + " events (should be one).");
            }
            _startedElements.remove(_startedElements.size() - 1);
            super.endDocument();
        }
    }
}
