/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.ui.mail;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.time.ZonedDateTime;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.core.cocoon.ActionResultGenerator;
import org.ametys.core.user.User;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Helper to build standard wrapped HTML body for mail.
 * Components that is supported by this helper are: title, text messages, details, user actions (such as comment), a button link and footer link
 */
public class StandardMailBodyHelper implements Contextualizable, Serviceable, Initializable
{
    /** I18n text for autogenerated email text */
    public static final I18nizableText AUTOGENERATED_MAIL_TEXT = new I18nizableText("plugin.core-ui", "PLUGINS_CORE_UI_MAIL_AUTOGENERATED_TEXT");
    /** I18n text for autogenerated email text with do not answer */
    public static final I18nizableText AUTOGENERATED_DO_NOT_ANSWER_TEXT = new I18nizableText("plugin.core-ui", "PLUGINS_CORE_UI_MAIL_AUTOGENERATED_DO_NOT_ANSWER_TEXT");
   
    /** The avalon context */
    protected static Context _context;
    
    /** The user helper */
    protected static UserHelper _userHelper;
    
    private static SourceResolver _srcResolver;
    private static StandardMailBodyHelper _instance;

    @Override
    public void contextualize(Context context)
    {
        _context = context;
    }
    
    public void initialize() throws Exception
    {
        _instance = this;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _srcResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
    }

    /**
     * Get the URI to resolve wrapped HTML email body
     * @return the uri
     */
    public String getMailBodyUri()
    {
        return "cocoon://_plugins/core-ui/wrapped-mail/standard/body.html";
    }
    
    /**
     * Creates a new mail body builder.
     * @return the newly created {@link MailBodyBuilder}.
     */
    public static MailBodyBuilder newHTMLBody()
    {
        return new MailBodyBuilder(_instance.getMailBodyUri());
    }
    
    /**
     * Implements the builder pattern for creating and sending emails.
     */
    public static class MailBodyBuilder
    {
        private Object _title;
        private List<Object> _messages;
        private I18nizableText _linkText;
        private String _linkUrl;
        private I18nizableText _footerLinkText;
        private String _footerLinkUrl;
        private String _footerImgPluginName;
        private String _footerImgPath;
        private Object _footerBottomText;
        private I18nizableText _userInputHeader;
        private List<UserInput> _userInputs;
        private List<UnauthenticatedUserInput> _unauthenticatedUserInputs;
        private I18nizableText _detailsTitle;
        private Object _detailsText;
        private boolean _technicalDetails;
        private String _lang;
        private String _uri;
        
        MailBodyBuilder(String uri)
        {
            _uri = uri;
        }
        
        /**
         * Set the language
         * @param lang the language
         * @return this builder
         */
        public MailBodyBuilder withLanguage(String lang)
        {
            _lang = lang;
            return this;
        }
        
        /**
         * Set the title
         * @param title the title as {@link I18nizableText}
         * @return this builder
         */
        public MailBodyBuilder withTitle(I18nizableText title)
        {
            _title = title;
            return this;
        }
        
        /**
         * Set the title
         * @param title the title as String
         * @return this builder
         */
        public MailBodyBuilder withTitle(String title)
        {
            _title = title;
            return this;
        }
        
        /**
         * Set the main message
         * @param message the message as String or {@link I18nizableText}
         * @return this builder
         */
        public MailBodyBuilder withMessage(I18nizableText message)
        {
            _messages = List.of(message);
            return this;
        }
        
        /**
         * Set the main message
         * @param message the message as String or {@link I18nizableText}
         * @return this builder
         */
        public MailBodyBuilder withMessage(String message)
        {
            _messages = List.of(_text2html(message));
            return this;
        }
        
        /**
         * Add a new message
         * @param message the message as {@link I18nizableText}
         * @return this builder
         */
        public MailBodyBuilder addMessage(I18nizableText message)
        {
            if (_messages == null)
            {
                _messages = new ArrayList<>();
            }
            
            _messages.add(message);
            return this;
        }
        
        /**
         * Add a new message
         * @param message the message as String
         * @return this builder
         */
        public MailBodyBuilder addMessage(String message)
        {
            if (_messages == null)
            {
                _messages = new ArrayList<>();
            }
            
            _messages.add(_text2html(message));
            return this;
        }
        
        /**
         * Add a main link to be displayed as a button
         * @param linkUrl the link url
         * @param linkText the link text
         * @return this builder
         */
        public MailBodyBuilder withLink(String linkUrl, I18nizableText linkText)
        {
            _linkText = linkText;
            _linkUrl = linkUrl;
            return this;
        }
        
        /**
         * Add a link to be displayed on footer
         * @param linkUrl the link url
         * @param linkText the link text
         * @return this builder
         */
        public MailBodyBuilder withFooterLink(String linkUrl, I18nizableText linkText)
        {
            return withFooterLink(linkUrl, linkText, null, null);
        }
        
        /**
         * Add a link to be displayed on footer
         * @param linkUrl the link url
         * @param linkText the link text
         * @param imgPluginName the plugin name that constains the image
         * @param imgPath the path of the image in the plugin
         * @return this builder
         */
        public MailBodyBuilder withFooterLink(String linkUrl, I18nizableText linkText, String imgPluginName, String imgPath)
        {
            _footerLinkText = linkText;
            _footerLinkUrl = linkUrl;
            _footerImgPluginName = imgPluginName;
            _footerImgPath = imgPath;
            return this;
        }
        
        /**
         * Set the text to display at footer bottom such as "Autogenerated email. Please do not answser"
         * @param footerText The footer text as String
         * @return this builder
         */
        public MailBodyBuilder withFooterBottomText(String footerText)
        {
            _footerBottomText = _text2html(footerText);
            return this;
        }
        
        /**
         * Set the text to display at footer bottom such as "Autogenerated email. Please do not answser"
         * @param footerText The footer text as {@link I18nizableText}
         * @return this builder
         */
        public MailBodyBuilder withFooterBottomText(I18nizableText footerText)
        {
            _footerBottomText = footerText;
            return this;
        }
        
        
        /**
         * Set the user action
         * @param userInputHeader The header text for user input(s)
         * @return this builder
         */
        public MailBodyBuilder withUserInputHeader(I18nizableText userInputHeader)
        {
            _userInputHeader = userInputHeader;
            return this;
        }
        
        /**
         * Set the user action
         * @param userInput the user action such as a comment
         * @return this builder
         */
        public MailBodyBuilder withUserInput(UserInput userInput)
        {
            _userInputs = List.of(userInput);
            return this;
        }
        
        /**
         * Set the user action
         * @param userInput the user action such as a comment
         * @param userInputHeader The header text for user input 
         * @return this builder
         */
        public MailBodyBuilder withUserInput(UserInput userInput, I18nizableText userInputHeader)
        {
            _userInputs = List.of(userInput);
            _userInputHeader = userInputHeader;
            return this;
        }
        
        /**
         * Add an user input 
         * @param userInput the user input such as a comment
         * @return this builder
         */
        public MailBodyBuilder addUserInput(UserInput userInput)
        {
            if (_userInputs == null)
            {
                _userInputs = new ArrayList<>();
            }
            
            _userInputs.add(userInput);
            return this;
        }
        
        /**
         * Add an user input from a non-authenticated user
         * @param unauthenticatedUserInput the input such as a comment
         * @param userInputHeader The header text for user input as String or {@link I18nizableText}
         * @return this builder
         */
        public MailBodyBuilder withUnauthenticatedUserInput(UnauthenticatedUserInput unauthenticatedUserInput, I18nizableText userInputHeader)
        {
            _unauthenticatedUserInputs = List.of(unauthenticatedUserInput);
            _userInputHeader = userInputHeader;
            return this;
        }
        
        /**
         * Add an user input from a non-authenticated user
         * @param unauthenticatedUserInput the input such as a comment
         * @return this builder
         */
        public MailBodyBuilder withUnauthenticatedUserInput(UnauthenticatedUserInput unauthenticatedUserInput)
        {
            _unauthenticatedUserInputs = List.of(unauthenticatedUserInput);
            return this;
        }
        
        /**
         * Add an user input from a non-authenticated user
         * @param unauthenticatedUserInput the input such as a comment
         * @return this builder
         */
        public MailBodyBuilder addUnauthenticatedUserInput(UnauthenticatedUserInput unauthenticatedUserInput)
        {
            if (_unauthenticatedUserInputs == null)
            {
                _unauthenticatedUserInputs = new ArrayList<>();
            }
            
            _unauthenticatedUserInputs.add(unauthenticatedUserInput);
            return this;
        }
        
        /**
         * Set details or additional information to display in the mail
         * @param detailsTitle the details title
         * @param detailsText the details text
         * @param technical true if details are technical details (such as an error stack trace, script result, ...)
         * @return this builder
         */
        public MailBodyBuilder withDetails(I18nizableText detailsTitle, String detailsText, boolean technical)
        {
            _detailsTitle = detailsTitle;
            _detailsText = _text2html(detailsText);
            _technicalDetails = technical;
            return this;
        }
        
        /**
         * Set details or additional information to display in the mail
         * @param detailsTitle the details title
         * @param detailsText the details text
         * @param technical true if details are technical details (such as an error stack trace, script result, ...)
         * @return this builder
         */
        public MailBodyBuilder withDetails(I18nizableText detailsTitle, I18nizableText detailsText, boolean technical)
        {
            _detailsTitle = detailsTitle;
            _detailsText = detailsText;
            _technicalDetails = technical;
            return this;
        }
        
        private String _text2html(String text)
        {
            return text.replaceAll("\n", "<br/>");
        }
        
        /**
         * Build the body of the email as HTML
         * @return the HTML body
         * @throws IOException if an error occurred when building the email
         */
        public String build() throws IOException
        {
            Source src = null;
            Request request = ContextHelper.getRequest(_context);
            
            Object pluginName = request.getAttribute("pluginName");
            Object requestAttr = request.getAttribute(ActionResultGenerator.MAP_REQUEST_ATTR);
            
            try
            {
                request.setAttribute(ActionResultGenerator.MAP_REQUEST_ATTR, toJson());
                
                Map<String, Object> parameters = new HashMap<>();
                
                if (StringUtils.isEmpty(_lang))
                {
                    Map objectModel = ContextHelper.getObjectModel(_context);
                    Locale locale = org.apache.cocoon.i18n.I18nUtils.findLocale(objectModel, "locale", null, Locale.getDefault(), true);
                    
                    parameters.put("lang", locale.toString());
                }
                else
                {
                    parameters.put("lang", _lang);
                }
                
                src = _srcResolver.resolveURI(_uri, null, parameters);
                Reader reader = new InputStreamReader(src.getInputStream(), "UTF-8");
                return IOUtils.toString(reader);
            }
            finally
            {
                // The resolveURI of _plugins/myPlugin set the pluginName to myPlugin, so we need to re-set the current plugin after that 
                request.setAttribute("pluginName", pluginName);
                request.setAttribute(ActionResultGenerator.MAP_REQUEST_ATTR, requestAttr);
                
                _srcResolver.release(src);
            }
        }
        
        /**
         * Get the JSON representation of the email ingredients
         * @return the JSON ingredients
         */
        public Map<String, Object> toJson()
        {
            Map<String, Object> json = new HashMap<>();
            
            json.put("title", _title);
            
            if (_messages != null && !_messages.isEmpty())
            {
                json.put("message", _messages);
            }
            
            if (StringUtils.isNotEmpty(_linkUrl))
            {
                json.put("link", Map.of("url", _linkUrl, "text", _linkText));
            }
            
            if (_userInputHeader != null)
            {
                json.put("userInputHeader", _userInputHeader);
            }
            
            if (_userInputs != null && !_userInputs.isEmpty())
            {
                json.put("userInput", _userInputs.stream()
                        .map(this::_userAction2json)
                        .toList());
            }
            
            if (_unauthenticatedUserInputs != null && !_unauthenticatedUserInputs.isEmpty())
            {
                json.put("unauthenticatedUserInput", _unauthenticatedUserInputs.stream()
                        .map(this::unauthenticatedUserAction2json)
                        .toList());
            }
            
            if (_detailsText != null)
            {
                json.put("details", _details2json());
            }
            
            json.put("footer", _footer2json());
            
            return json;
        }
        
        private Map<String, Object> _footer2json()
        {
            Map<String, Object> footer2json = new HashMap<>();
            
            if (_footerLinkUrl != null && _footerLinkText != null)
            {
                Map<String, Object> link2json = new HashMap<>();
                
                link2json.put("text", _footerLinkText);
                link2json.put("url", _footerLinkUrl);
                if (_footerImgPath != null && _footerImgPluginName != null)
                {
                    link2json.put("imgPluginName", _footerImgPluginName);
                    link2json.put("imgPath", _footerImgPath);
                }
                footer2json.put("link", link2json);
            }
            
            if (_footerBottomText != null)
            {
                footer2json.put("additionalText", _footerBottomText);
            }
            
            
            return footer2json;
        }
        
        private Map<String, Object> _details2json()
        {
            Map<String, Object> details2json = new HashMap<>();
            
            if (_detailsText != null)
            {
                if (_detailsTitle != null)
                {
                    details2json.put("title", _detailsTitle);
                }
                details2json.put("text", _detailsText);
                details2json.put("technical", _technicalDetails);
            }
            return details2json;
            
        }
        
        private Map<String, Object> _userAction2json(UserInput action)
        {
            Map<String, Object> action2json = new HashMap<>();
            action2json.put("author", _userHelper.user2json(action.author()));
            action2json.put("date", DateUtils.zonedDateTimeToString(action.date()));
            action2json.put("description", action.description());
            return action2json;
        }
        
        private Map<String, Object> unauthenticatedUserAction2json(UnauthenticatedUserInput action)
        {
            Map<String, Object> action2json = new HashMap<>();
            action2json.put("authorName", action.authorName());
            action2json.put("date", DateUtils.zonedDateTimeToString(action.date()));
            action2json.put("description", action.description());
            return action2json;
        }
        
        /**
         * Record for an user input
         * @param author The author. Can not be null.
         * @param date the date of the action
         * @param description the description or text of input as String or {@link I18nizableText}
         */
        public record UserInput(User author, ZonedDateTime date, Object description) { /* empty */ }
        
        /**
         * Record for an user input from a non-authenticated user
         * @param authorName The author's name
         * @param date the date of the action
         * @param description the description or text of action as String or {@link I18nizableText}
         */
        public record UnauthenticatedUserInput(String authorName, ZonedDateTime date, Object description) { /* empty */ }
        
    }
}
