/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.ui.ribbonconfiguration;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.xml.sax.SAXException;

/**
 * A ribbon configuration source and related data
 */
public class RibbonConfigurationSource
{
    Source _source;
    String _uri;
    Long _lastModified;
    
    /**
     * Constructor
     * @param uri The source uri
     * @param source The source
     * @param returnedParameters The resolve parameters for additional data
     */
    public RibbonConfigurationSource(String uri, Source source, Map<String, Object> returnedParameters)
    {
        _uri = uri;
        _source = source;
        _lastModified = returnedParameters.containsKey("lastModified") ? (Long) returnedParameters.get("lastModified") : source.getLastModified();
    }
    
    /**
     * Create a new ribbon configuration source
     * @param uri The uri
     * @param resolver The source resolver 
     * @return the ribbon configuration source
     * @throws IOException If an error occurred when resolving the URI
     */
    public static RibbonConfigurationSource createFromUri(String uri, SourceResolver resolver) throws IOException
    {
        Map<String, Object> returnedParameters = new HashMap<>();
        Source resolvedSource = resolver.resolveURI(uri, null, returnedParameters);
        return new RibbonConfigurationSource(uri, resolvedSource, returnedParameters);   
    }

    /**
     * Get the source
     * @return The source
     */
    public Source getSource()
    {
        return _source;
    }
    
    /**
     * Get the ribbon configuration URI
     * @return The uri
     */
    public String getUri()
    {
        return _uri;
    }
    
    /**
     * Get the last modified date
     * @return The last modified date
     */
    public Long getLastModified()
    {
        return _lastModified;
    }
    
    /**
     * build the configuration from the source
     * @return The configuration
     * @throws ConfigurationException If an error occurred
     * @throws SAXException If an error occurred
     * @throws IOException If an error occurred
     */
    public Configuration getConfiguration() throws ConfigurationException, SAXException, IOException
    {
        if (_source.exists())
        {
            try (InputStream is = _source.getInputStream())
            {
                return new DefaultConfigurationBuilder().build(is);
            }
        }

        // default empty configuration
        return new DefaultConfiguration("ribbon");
    }
    
}
