/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.ui.right;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.core.right.AccessController.AccessResult;
import org.ametys.core.right.RightAssignmentContextExtensionPoint;
import org.ametys.core.right.RightManager.RightExplanation;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.ClientSideElementHelper;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.core.user.UserHelper;

/**
 * Client side element for the explain user tool.
 * Notably, the class will provide a list of all the registered {@link TargetToContextConvertor}
 */
public class ExplainUserRightTool extends StaticClientSideElement
{
    /** rights to explain rights for any user */
    public static final String RIGHT_EXPLAIN_ALL_USERS_RIGHTS = "Runtime_Rights_ExplainAllUsersRights";
    /** rights to explain rights for their own user */
    public static final String RIGHT_EXPLAIN_OWN_RIGHTS = "Runtime_Rights_ExplainOwnRights";

    /** {@link TargetToContextConvertorExtensionPoint} */
    protected TargetToContextConvertorExtensionPoint _targetToContextConvertorEP;
    /** {@link UserHelper} */
    protected UserHelper _userHelper;
    /** {@link RightAssignmentContextExtensionPoint} */
    protected RightAssignmentContextExtensionPoint _rightAssignmentContextEP;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _rightAssignmentContextEP = (RightAssignmentContextExtensionPoint) manager.lookup(RightAssignmentContextExtensionPoint.ROLE);
        _targetToContextConvertorEP = (TargetToContextConvertorExtensionPoint) manager.lookup(TargetToContextConvertorExtensionPoint.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
    }
    
    @Override
    public List<Script> getScripts(boolean ignoreRights, Map<String, Object> contextParameters)
    {
        List<Script> scripts = super.getScripts(ignoreRights, contextParameters);
        
        // We need to provide the tool with the list of convertor and load the convertor scripts
        Map<String, Object> targetToContextConvertor = new HashMap<>();
        
        if (scripts.size() > 0)
        {
            Script toolScript = ClientSideElementHelper.cloneScript(scripts.get(0));
            
            for (String id :_targetToContextConvertorEP.getExtensionsIds())
            {
                ClientSideElement convertor = _targetToContextConvertorEP.getExtension(id);
                List<Script> convertorScripts = convertor.getScripts(ignoreRights, contextParameters);
                int index = 0;
                for (Script convertorScript:  convertorScripts)
                {
                    Map<String, Object> classInfo = new HashMap<>();
                    classInfo.put("className", convertorScript.getScriptClassname());
                    classInfo.put("serverId", convertor.getId());
                    classInfo.put("parameters", convertorScript.getParameters());
                    targetToContextConvertor.put(convertor.getId() + "-" + index++, classInfo);
                    
                    toolScript.getCSSFiles().addAll(convertorScript.getCSSFiles());
                    toolScript.getScriptFiles().addAll(convertorScript.getScriptFiles());
                }
            }
            
            toolScript.getParameters().put("convertors", targetToContextConvertor);
            return List.of(toolScript);
        }
        return scripts;
    }
    
    /**
     * Explain why the current user has a given right or not on a context.
     * @param rightId the right to explain
     * @param contextJS the context has provided by the {@link TargetToContextConvertor}
     * @param targetToContextConvertorId the role of the {@link TargetToContextConvertor} that provided the context
     * @return the explanations
     */
    @Callable(rights = RIGHT_EXPLAIN_OWN_RIGHTS)
    public Map<String, Object> explainCurrentUserRight(String rightId, Object contextJS, String targetToContextConvertorId)
    {
        return explainUserRight(rightId, _currentUserProvider.getUser(), contextJS, targetToContextConvertorId);
    }
    
    /**
     * Explain why the user has a given right or not on a context.
     * @param rightId the right to explain
     * @param user the user identity as JSON
     * @param contextJS the context has provided by the {@link TargetToContextConvertor}
     * @param targetToContextConvertorId the role of the {@link TargetToContextConvertor} that provided the context
     * @return the explanations
     */
    @Callable(rights = RIGHT_EXPLAIN_ALL_USERS_RIGHTS)
    public Map<String, Object> explainRight(String rightId, Map<String, Object> user, Object contextJS, String targetToContextConvertorId)
    {
        return explainUserRight(rightId, _userHelper.json2userIdentity(user), contextJS, targetToContextConvertorId);
    }

    private Map<String, Object> explainUserRight(String rightId, UserIdentity userIdentity, Object contextJS, String targetToContextConvertorId)
    {
        Object context;
        if (contextJS == null)
        {
            context = "/${WorkspaceName}";
        }
        else
        {
            context = _targetToContextConvertorEP.getExtension(targetToContextConvertorId).convertJSContext(contextJS);
        }
        
        List<RightExplanation> explanations = _rightManager.explain(userIdentity, rightId.equals("READER") ? null : rightId , context);
        
        explanations.sort(null);
        AccessResult summary;
        if (explanations.isEmpty())
        {
            summary = AccessResult.UNKNOWN;
        }
        else
        {
            summary = explanations.stream().map(RightExplanation::result).collect(Collectors.collectingAndThen(Collectors.toList(), AccessResult::merge));
        }
        return Map.of("result", summary, "explanations", explanations);
    }
}
