/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.ui.right;

import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.excalibur.source.ModifiableSource;
import org.apache.excalibur.source.SourceResolver;
import org.apache.xml.serializer.OutputPropertiesFactory;
import org.xml.sax.SAXException;

import org.ametys.core.right.Profile;
import org.ametys.core.right.Right;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightProfilesDAO;
import org.ametys.core.right.RightsExtensionPoint;
import org.ametys.core.ui.Callable;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Helper to manage rights from rights.xml file
 */
public class ExternalRightsDAO extends AbstractLogEnabled implements Component, Serviceable
{
    /** The rights extension point */ 
    protected RightsExtensionPoint _rightsExtensionPoint;
    
    /** The right manager */
    protected RightManager _rightManager;
    
    /** The profile DAO */
    protected RightProfilesDAO _profileDAO;
    
    /** Avalon SourceResolver */
    protected SourceResolver _resolver;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _rightsExtensionPoint = (RightsExtensionPoint) manager.lookup(RightsExtensionPoint.ROLE);
        _rightManager = (RightManager) manager.lookup(RightManager.ROLE);
        _profileDAO = (RightProfilesDAO) manager.lookup(RightProfilesDAO.ROLE);
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    /**
     * Get all the right categories
     * @return a map of the categories' labels 
     */
    @Callable (rights = "CORE_Rights_Edit_Right")
    public Map<String, Object> getRightsCategories()
    {
        Set<Map<String, Object>> categories2Json = new HashSet<>();
        for (String rightId : _rightsExtensionPoint.getExtensionsIds())
        {
            Right right = _rightsExtensionPoint.getExtension(rightId);
            I18nizableText category = right.getCategory();
            categories2Json.add(Map.of("label", category, "id", category.toString()));
        }
        return Map.of("data", categories2Json);
    }
    
    /**
     * Get the available rights for deletion
     * @return a map of the rights id and label
     */
    @Callable (rights = "CORE_Rights_Edit_Right")
    public Map<String, Object> getRightsToDelete()
    {
        List<Map<String, Object>> rights2Json = new ArrayList<>();
        List<Profile> profiles = _profileDAO.getProfiles();
        for (String rightId : _rightManager.getExternalRightIds())
        {
            if (!_isRightUsed(profiles, rightId))
            {
                Right right = _rightsExtensionPoint.getExtension(rightId);
                rights2Json.add(Map.of("label", right.getLabel(), "id", right.getId()));
            }
        }
        return Map.of("data", rights2Json);
    }

    private boolean _isRightUsed(List<Profile> profiles, String rightId)
    {
        return profiles.stream()
                .map(p -> _profileDAO.getRights(p))
                .filter(rights -> rights.contains(rightId))
                .findAny()
                .isPresent();
    }
    
    /**
     * Add a right to the right EP then write it in rights.xml
     * @param id the id for the new right
     * @param label the label for the new right 
     * @param description the description for the new right 
     * @param category the category for the new right 
     * @return a map with an optionnal error message
     */
    @Callable (rights = "CORE_Rights_Edit_Right")
    public Map<String, Object> addRight(String id, String label, String description, String category)
    {
        Map<String, Object> results = new HashMap<>();
        if (!getRightsIds().contains(id))
        {
            _rightManager.addExternalRight(id, label, description, category);
            
            try
            {
                writeRightsXML();
                results.put("id", id);
            }
            catch (Exception e)
            {
                getLogger().error("An error occured while writing in rights.xml", e);
                results.put("message", "error");
            }
        }
        else
        {
            results.put("message", "duplicate-id");
        }
        return results;
    }
    
    /**
     * Remove the right from the rightEP and rights.xml
     * @param id the right to delete's id
     * @return a map with rightId or error message
     */
    @Callable (rights = "CORE_Rights_Edit_Right")
    public Map<String, Object> removeRight(String id)
    {
        Map<String, Object> results = new HashMap<>();
        if (!_isRightUsed(_profileDAO.getProfiles(), id))
        {
            _rightManager.removeExternalRight(id);

            try
            {
                writeRightsXML();
                results.put("id", id);
            }
            catch (Exception e)
            {
                getLogger().error("An error occured while writing in rights.xml", e);
                results.put("message", "error");
            }
        }
        else 
        {
            results.put("message", "rigth-used");
        }
        return results;
    }

    /**
     * Write the rights.xml file with the list of external rights
     * @throws Exception if an error occured
     */
    protected void writeRightsXML() throws Exception 
    {
        String filepath = "context://WEB-INF/param/rights.xml";
        ModifiableSource source = (ModifiableSource) _resolver.resolveURI(filepath);

        try (OutputStream os = source.getOutputStream())
        {
            // create a transformer for saving sax into a file
            TransformerHandler th = ((SAXTransformerFactory) TransformerFactory.newInstance()).newTransformerHandler();

            StreamResult result = new StreamResult(os);
            th.setResult(result);

            // create the format of result
            Properties format = new Properties();
            format.put(OutputKeys.METHOD, "xml");
            format.put(OutputKeys.INDENT, "yes");
            format.put(OutputKeys.ENCODING, "UTF-8");
            format.put(OutputPropertiesFactory.S_KEY_INDENT_AMOUNT, "2");
            th.getTransformer().setOutputProperties(format);

            // sax the config into the transformer
            _saxRights(th);
        }
        finally
        {
            if (source != null)
            {
                _resolver.release(source);
            }
        }
    }
    
    /**
     * Sax external rights in rights.xml ?
     * @param handler the transformer
     * @throws SAXException exception during SAX
     */
    protected void _saxRights(TransformerHandler handler) throws SAXException
    {
        handler.startDocument();
        AttributesImpl attribute = new AttributesImpl();
        
        XMLUtils.startElement(handler, "rights", attribute);
        List<String> externalRightIds = _rightManager.getExternalRightIds();
        for (String rightId : externalRightIds)
        {
            Right right = _rightsExtensionPoint.getExtension(rightId);
            attribute.addCDATAAttribute("id", right.getId());
            XMLUtils.startElement(handler, "right", attribute);
            
            _saxI18nizableText(handler, right.getLabel(), "label");
            _saxI18nizableText(handler, right.getDescription(), "description");
            _saxI18nizableText(handler, right.getCategory(), "category");
            
            XMLUtils.endElement(handler, "right");
        }
        XMLUtils.endElement(handler, "rights");
        handler.endDocument();
    }
    
    private void _saxI18nizableText(TransformerHandler handler, I18nizableText text, String tagName) throws SAXException
    {
        AttributesImpl attributeI18n = new AttributesImpl();
        attributeI18n.addCDATAAttribute("type", "i18n");
        if (text.isI18n())
        {
            XMLUtils.createElement(handler, tagName, attributeI18n, text.toString());
        }
        else
        {
            XMLUtils.createElement(handler, tagName, text.toString());
        }
    }
    
    /**
     * Get the set of used right ids
     * @return a set of right ids
     */
    @Callable (rights = "CORE_Rights_Edit_Right")
    public Set<String> getRightsIds()
    {
        return _rightsExtensionPoint.getExtensionsIds();
    }
}
