/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.ui.widgets.richtext;

import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.ClientSideElement.ScriptFile;

/**
 * Extension for the {@link RichTextConfigurationExtensionPoint}.
 * Extensions are classified by categories: "" is the default category. 
 */
public interface RichTextConfiguration
{
    /**
     * Get the categories supported by this extension
     * @return The categories supported. Cannot be null, but may be empty and of course contains the "" key for the default category.
     */
    public Set<String> getCategories();
    
    /**
     * Get the authorized tags for the given category
     * @param category The category. "" is the default category.
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return The tags supported (can be null). The key is the html tag name supported and the value is the associated value.
     */
    public Collection<RichTextConfigurationTag> getHandledTags(String category, Map<String, Object> contextualParameters);
    
    /**
     * Get the CSS files required during the edition of the given category.
     * @param category The category. "" is the default category.
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return The CSS files requires. Can be null.
     */
    public List<ScriptFile> getCSSFiles(String category, Map<String, Object> contextualParameters);
    
    /**
     * Get the validators used during an edition of the rich text to ensure the value is correct. Validators can be differents by category.
     * @param category The category. "" is the default category.
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return The validators to use. The class of the ClientSideElement must inherit the front-side validor js interface. Can be null.
     */
    public Set<ClientSideElement> getValidators(String category, Map<String, Object> contextualParameters);
    
    /**
     * Get the convertors used when richtext switch from code to view and from view to code. Convertors can be differents by category.
     * @param category The category. "" is the default category.
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return The convertors to use. The class of the ClientSideElement must inherit the front-side convertor js interface. Can be null.
     */
    public Set<ClientSideElement> getConvertors(String category, Map<String, Object> contextualParameters);

    /**
     * Get the available styles classified by identifiers such as "table", "link"...
     * @param category The category. "" is the default category.
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return The available styles. Can be null.
     */
    public Map<String, Map<RichTextConfigurationStyleGroup, List<RichTextConfigurationStyle>>> getAvailableStyles(String category, Map<String, Object> contextualParameters);
}
