/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.ui.widgets.richtext;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.ametys.core.ui.Callable;
import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.ClientSideElement.Script;
import org.ametys.core.ui.ClientSideElement.ScriptFile;
import org.ametys.core.ui.widgets.richtext.RichTextConfigurationTag.EMPTY_TAG;
import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentExtensionPoint;

/**
 * Extension point for configuring the widget 'richtext' 
 */
public class RichTextConfigurationExtensionPoint extends AbstractThreadSafeComponentExtensionPoint<RichTextConfiguration>
{
    /** The component role */
    public static final String ROLE = RichTextConfigurationExtensionPoint.class.getName();
    
    /**
     * Converts all values to a single JSON object
     * { 
     *     "category1" : 
     *     {
     *         "tags" : {
     *             "p" : {
     *                 empty: "CLOSE|OPEN|REMOVE_EMPTY_CONTENT|PADDING|REMOVE_EMPTY_ATTRIBUTES",
     *                 synonyms: ["div"],
     *                 attributes: {
     *                     "id": {},
     *                     "alt": {
     *                         "default-value": ""
     *                     },
     *                     "class": {
     *                         "default-value": "value1",
     *                         "values": ["value1", "value2"],
     *                         "technical-values": [ "value1" ] // special for 'class' attribute
     *                     }
     *                 }
     *             }
     *         },
     *         
     *         "styles" : {
     *             "paragraph" : {
     *                  "groups": [ {
     *                      "label": "Customized styles",
     *                      "priority": 1,
     *                      "values" : [
     *                          {
     *                              "tagname": "h1",
     *                              "cssclass": "big",
     *                          
     *                              "label": "Header 1 (big)",
     *                              "description": "A big header",
     *                              "buttonCSSClass": "h1-big",
     *                              "buttonSmallIcon": "img/h1_16.png",
     *                              "buttonMediumIcon": "img/h1_32.png",
     *                              "buttonLargeIcon": "img/h1_48.png"
     *                          }
     *                      ]
     *                  } ]
     *             }
     *         },
     *         
     *         "css" : [ 
     *             { language: false, url: "file1", debug: "all|true|false", rtl: "all|true|false" }, 
     *             { language: true, default: "en", url : {"fr": "file1"} 
     *         ],
     *         
     *         "validators": [
     *             "class": {
     *                 "name": "classname",
     *                 "parameters": { ... }
     *             }
     *         ]
     *         
     *         "convertors": [
     *             "class": {
     *                 "name": "classname",
     *                 "parameters": { ... }
     *             }
     *         ]
     *     }
     * }
     * @param contextualParameters Contextuals parameters transmitted by the environment.
     * @return The non-null configuration
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> toJSON(Map<String, Object> contextualParameters)
    {
        Map<String, Object> json = new HashMap<>();

        for (String extensionId : getExtensionsIds())
        {
            _addExtensionToJSON(contextualParameters, json, getExtension(extensionId));
        }
        
        return json;
    }
    
    @SuppressWarnings("unchecked")
    private  void _addExtensionToJSON(Map<String, Object> contextualParameters, Map<String, Object> json, RichTextConfiguration extension)
    {
        Set<String> categories = extension.getCategories();
        for (String category : categories)
        {
            if (!json.containsKey(category))
            {
                Map<String, Object> jsonCategory = new HashMap<>();
                json.put(category, jsonCategory);
                
                jsonCategory.put("tags", new HashMap<>());
                jsonCategory.put("styles", new HashMap<>());
                jsonCategory.put("css", new HashSet<>());
                jsonCategory.put("validators", new ArrayList<>());
                jsonCategory.put("convertors", new ArrayList<>());
            }
            
            Map<String, Object> jsonCategory = (Map<String, Object>) json.get(category);
            
            _addTagsToJSON((Map<String, Object>) jsonCategory.get("tags"), extension.getHandledTags(category, contextualParameters));
            _addStylesToJSON((Map<String, Map<String, List<Map<String, Object>>>>) jsonCategory.get("styles"), extension.getAvailableStyles(category, contextualParameters));
            _addFilesToJSON((Set<Map<String, Object>>) jsonCategory.get("css"), extension.getCSSFiles(category, contextualParameters));
            _addClientSideElementToJSON((List<Object>) jsonCategory.get("validators"), extension.getValidators(category, contextualParameters));
            _addClientSideElementToJSON((List<Object>) jsonCategory.get("convertors"), extension.getConvertors(category, contextualParameters));
        }
        
    }

    @SuppressWarnings("unchecked")
    private void _addTagsToJSON(Map<String, Object> object, Collection<RichTextConfigurationTag> handledTags)
    {
        if (handledTags != null)
        {
            for (RichTextConfigurationTag handledTag : handledTags)
            {
                if (!object.containsKey(handledTag.getTag()))
                {
                    Map<String, Object> jsonTag = new HashMap<>();
                    jsonTag.put("empty", EMPTY_TAG.CLOSE);
                    jsonTag.put("synonyms", new HashSet<>());
                    jsonTag.put("attributes", new HashMap<>()); 
                    object.put(handledTag.getTag(), jsonTag);
                }
    
                Map<String, Object> jsonTag = (Map<String, Object>) object.get(handledTag.getTag());
                
                jsonTag.put("empty", handledTag.onEmptyTag().compareTo((EMPTY_TAG) jsonTag.get("empty")) < 0 ? handledTag.onEmptyTag() : (EMPTY_TAG) jsonTag.get("empty"));
                
                ((Set<String>) jsonTag.get("synonyms")).addAll(handledTag.getSynonyms());
                
                Map<String, Object> attributes = (Map<String, Object>) jsonTag.get("attributes");
                for (RichTextConfigurationAttribute attribute : handledTag.getAttributes())
                {
                    boolean classAttribute = "class".equals(attribute.getName());
                    
                    if (!attributes.containsKey(attribute.getName()))
                    {
                        Map<String, Object> jsonAttribute = new HashMap<>();
                        jsonAttribute.put("default-value", null);
                        jsonAttribute.put("values", new ArrayList<>());
                        if (classAttribute)
                        {
                            jsonAttribute.put("technical-values", new ArrayList<>());
                        }
                        attributes.put(attribute.getName(), jsonAttribute);
                    }
                    
                    Map<String, Object> jsonAttribute = (Map<String, Object>) attributes.get(attribute.getName());
                    jsonAttribute.put("default-value", attribute.getDefaultValue()); // Let's us take the last one
                    ((List<String>) jsonAttribute.get("values")).addAll(attribute.getAuthorizedValues());
                    if (classAttribute)
                    {
                        ((List<String>) jsonAttribute.get("technical-values")).addAll(attribute.getTechnicalValues());
                    }
                }
            }
        }
    }
    
    private void _addStylesToJSON(Map<String, Map<String, List<Map<String, Object>>>> object, Map<String, Map<RichTextConfigurationStyleGroup, List<RichTextConfigurationStyle>>> availableStyles)
    {
        if (availableStyles != null)
        {
            for (Entry<String, Map<RichTextConfigurationStyleGroup, List<RichTextConfigurationStyle>>> entry : availableStyles.entrySet())
            {
                if (!object.containsKey(entry.getKey()))
                {
                    Map<String, List<Map<String, Object>>> jsonStylesGroups = new HashMap<>();
                    jsonStylesGroups.put("groups", new ArrayList<>());
                    object.put(entry.getKey(), jsonStylesGroups);
                }
    
                List<Map<String, Object>> jsonStylesGroups = object.get(entry.getKey()).get("groups");
                
                for (Entry<RichTextConfigurationStyleGroup, List<RichTextConfigurationStyle>> entry2 : entry.getValue().entrySet())
                {
                    List<Map<String, Object>> jsonStyles = new ArrayList<>();
                    for (RichTextConfigurationStyle richTextConfigurationStyle : entry2.getValue())
                    {
                        Map<String, Object> styleAsJSON = new HashMap<>();
                        jsonStyles.add(styleAsJSON);
                        
                        styleAsJSON.put("tagname", richTextConfigurationStyle.getTagName());
                        styleAsJSON.put("cssclass", richTextConfigurationStyle.getClassName());
                        
                        styleAsJSON.put("label", richTextConfigurationStyle.getButtonLabel());
                        styleAsJSON.put("description", richTextConfigurationStyle.getButtonDescription());
                        styleAsJSON.put("buttonCSSClass", richTextConfigurationStyle.getButtonCSSClass());
                        styleAsJSON.put("buttonSmallIcon", richTextConfigurationStyle.getButtonSmallIcon());
                        styleAsJSON.put("buttonMediumIcon", richTextConfigurationStyle.getButtonMediumIcon());
                        styleAsJSON.put("buttonLargeIcon", richTextConfigurationStyle.getButtonLargeIcon());
                    }
                    
                    Map<String, Object> group = new HashMap<>();
                    group.put("label", entry2.getKey().getLabel());
                    group.put("priority", entry2.getKey().getPriority());
                    group.put("values", jsonStyles);
                    jsonStylesGroups.add(group);
                }
            }
        }
    }
    
    private void _addFilesToJSON(Set<Map<String, Object>> object, List<ScriptFile> cssFiles)
    {
        if (cssFiles != null)
        {
            for (ScriptFile file : cssFiles)
            {
                object.add(_scriptFileToJson(file));
            }
        }
    }
    
    private void _addClientSideElementToJSON(List<Object> list, Set<ClientSideElement> clientSideElements)
    {
        if (clientSideElements != null)
        {
            for (ClientSideElement clientSideElementn : clientSideElements)
            {
                List<Script> scripts = clientSideElementn.getScripts(Collections.EMPTY_MAP);
                for (Script script : scripts)
                {
                    Map<String, Object> jsonClientSideElement = new HashMap<>();
                    
                    Map<String, Object> jsonClass = new HashMap<>();
                    jsonClass.put("name", script.getScriptClassname());
                    jsonClass.put("parameters", script.getParameters());
                    jsonClientSideElement.put("class", jsonClass);
                    
                    list.add(jsonClientSideElement);
                }
            }
        }
    }
    
    private Map<String, Object> _scriptFileToJson(ScriptFile scriptFile)
    {
        Map<String, Object> json = new HashMap<>();
        
        if (scriptFile.isLangSpecific())
        {
            json.put("language", true);
            json.put("default", scriptFile.getDefaultLang());
            json.put("url", scriptFile.getLangPaths());
        }
        else
        {
            json.put("language", false);
            json.put("url", scriptFile.getPath());
            json.put("rtl", scriptFile.getRtlMode());
        }

        return json;
    }
}
