/*
 *  Copyright 2010 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui;

import java.io.IOException;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.xml.sax.SAXException;

import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.ClientSideElementDependenciesManager;
import org.ametys.core.ui.MessageTargetFactoriesManager;
import org.ametys.core.ui.RelationsManager;
import org.ametys.core.ui.RibbonConfigurationManager;
import org.ametys.core.ui.RibbonControlsManager;
import org.ametys.core.ui.RibbonImportManager;
import org.ametys.core.ui.RibbonManager;
import org.ametys.core.ui.RibbonManagerCache;
import org.ametys.core.ui.RibbonTabsManager;
import org.ametys.core.ui.SAXClientSideElementHelper;
import org.ametys.core.ui.StaticFileImportsManager;
import org.ametys.core.ui.UIToolsConfigurationManager;
import org.ametys.core.ui.UIToolsFactoriesManager;
import org.ametys.core.ui.ribbonconfiguration.RibbonConfigurationSource;
import org.ametys.core.ui.widgets.ClientSideWidget;
import org.ametys.core.ui.widgets.WidgetsManager;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.runtime.plugin.PluginsManager;
import org.ametys.runtime.servlet.RuntimeServlet;
import org.ametys.runtime.servlet.RuntimeServlet.RunMode;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * Generates the uitools factories definition using the component associated 
 */
public class WorkspaceGenerator extends ServiceableGenerator
{
    /** The ribbon control manager */
    protected RibbonControlsManager _ribbonControlManager;
    /** The ribbon tab manager */
    protected RibbonTabsManager _ribbonTabManager;
    /** The ui tools factories manager */
    protected UIToolsFactoriesManager _uitoolsFactoriesManager;
    /** The widgets manager */
    protected WidgetsManager _widgetsManager;
    /** The sax clientside element helper */
    protected SAXClientSideElementHelper _saxClientSideElementHelper;
    /** The Excalibur source resolver */
    protected SourceResolver _resolver;
    /** The current user provider component */
    protected CurrentUserProvider _currentUserProvider;
    /** The json utils component */
    protected JSONUtils _jsonUtils;
    /** The User Helper */
    protected UserHelper _userHelper;
    /** The ribbon manager cache helper */
    protected RibbonManagerCache _ribbonManagerCache;
    /** The ribbon import manager */
    protected RibbonImportManager _ribbonImportManager;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _uitoolsFactoriesManager = (UIToolsFactoriesManager) smanager.lookup(UIToolsFactoriesManager.ROLE);
        _ribbonTabManager = (RibbonTabsManager) smanager.lookup(RibbonTabsManager.ROLE);
        _ribbonControlManager = (RibbonControlsManager) smanager.lookup(RibbonControlsManager.ROLE);
        _widgetsManager = (WidgetsManager) smanager.lookup(WidgetsManager.ROLE);
        _saxClientSideElementHelper = (SAXClientSideElementHelper) smanager.lookup(SAXClientSideElementHelper.ROLE);
        _resolver = (SourceResolver) smanager.lookup(SourceResolver.ROLE);
        _currentUserProvider = (CurrentUserProvider) smanager.lookup(CurrentUserProvider.ROLE);
        _jsonUtils = (JSONUtils) smanager.lookup(JSONUtils.ROLE);
        _userHelper = (UserHelper) smanager.lookup(UserHelper.ROLE);
        _ribbonManagerCache = (RibbonManagerCache) smanager.lookup(RibbonManagerCache.ROLE);
        _ribbonImportManager = (RibbonImportManager) smanager.lookup(RibbonImportManager.ROLE);
    }
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        doGenerate(getContextualParameters());
    }
    
    /**
     * Get the contextual parameters
     * @return The contextual parameters
     */
    protected Map<String, Object> getContextualParameters()
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        String workspaceName = (String) request.getAttribute(WorkspaceMatcher.WORKSPACE_NAME);
        
        Map<String, Object> contextParameters = new HashMap<>();
        contextParameters.put(WorkspaceMatcher.WORKSPACE_NAME, workspaceName);
        
        return contextParameters;
    }
    
    /**
     * Generates the UI factories definitions, with parameters
     * @param contextParameters context parameters.
     * @throws IOException if an error occurred
     * @throws SAXException if an error occurred
     * @throws ProcessingException if an error occurred
     */
    protected void doGenerate(Map<String, Object> contextParameters) throws IOException, SAXException, ProcessingException
    {
        long startTime = new Date().getTime();
        
        contentHandler.startDocument();
        XMLUtils.startElement(contentHandler, "workspace");
        
        UserIdentity currentUser = _currentUserProvider.getUser();
        if (currentUser != null)
        {
            String login = currentUser.getLogin();
            String userPopulationId = currentUser.getPopulationId();
            if (StringUtils.isNotBlank(login))
            {
                UserIdentity userIdentity = new UserIdentity(login, userPopulationId);
                _userHelper.saxUserIdentity(userIdentity, contentHandler);
            }
        }
        
        ClientSideElementDependenciesManager dependenciesManager = new ClientSideElementDependenciesManager(this.manager);
        
        RibbonConfigurationSource ribbonConfig = getRibbonConfiguration();
        Map<String, List<ClientSideElement>> elementsToSax;
        RibbonManager ribbonManager = null;
        try
        {
            ribbonManager = _ribbonManagerCache.getManager(ribbonConfig.getUri());
            String workspaceName = (String) ObjectModelHelper.getRequest(objectModel).getAttribute("workspaceName");
            RibbonConfigurationManager ribbonConfigurationManager = new RibbonConfigurationManager();
            ribbonConfigurationManager.setup(_ribbonControlManager, _ribbonTabManager, _ribbonImportManager, _saxClientSideElementHelper, _resolver, _ribbonManagerCache);
            ribbonConfigurationManager.configure(ribbonManager, dependenciesManager, ribbonConfig, workspaceName);
            Map<String, Set<String>> effectivelySaxedElements = ribbonConfigurationManager.saxRibbonDefinition(contentHandler, contextParameters);
            elementsToSax = getElementsToSax(dependenciesManager, ribbonConfigurationManager, contextParameters, effectivelySaxedElements);
        }
        catch (Exception e)
        {
            throw new ProcessingException("Unable to get or create a ribbon manager for ribbon specific components", e);
        }
        finally
        {
            _ribbonManagerCache.dispose(ribbonManager);
        }
        
        saxUITools(contextParameters, elementsToSax.get(UIToolsFactoriesManager.ROLE));
        saxMessageTargetFactories(contextParameters, elementsToSax.get(MessageTargetFactoriesManager.ROLE));
        saxRelationsHandlers(contextParameters, elementsToSax.get(RelationsManager.ROLE));
        saxWidgets(contextParameters);
        saxStaticFileImports (contextParameters, elementsToSax.get(StaticFileImportsManager.ROLE));
        saxAdditionnalInfo(contextParameters);
        
        XMLUtils.endElement(contentHandler, "workspace");
        contentHandler.endDocument();

        
        long endTime = new Date().getTime();
        getLogger().debug("Workspace generated in " + (endTime - startTime) + " ms");
    }

    /**
     * Retrieve the list of elements to generate the Workspace
     * @param dependenciesManager The dependencies manager
     * @param ribbonManager The ribbon manager for this workspace
     * @param contextParameters Contextuals parameters transmitted by the environment.
     * @param effectivelySaxedElements Items effectively saxed
     * @return The list of elements, mapped by extension points.
     * @throws SAXException If an error occurs
     */
    protected Map<String, List<ClientSideElement>> getElementsToSax(ClientSideElementDependenciesManager dependenciesManager, RibbonConfigurationManager ribbonManager, Map<String, Object> contextParameters, Map<String, Set<String>> effectivelySaxedElements) throws SAXException
    {
        Set<String> effectivelySaxedControls = effectivelySaxedElements.get("controls");
        
        List<ClientSideElement> ribbonControls = ribbonManager.getControls(contextParameters);
        for (ClientSideElement control : ribbonControls)
        {
            if (effectivelySaxedControls.contains(control.getId()))
            {
                dependenciesManager.register(control);
            }
        }
        List<ClientSideElement> ribbonTabs = ribbonManager.getTabs();
        for (ClientSideElement control : ribbonTabs)
        {
            dependenciesManager.register(control);
        }
        
        for (String extensionId: _widgetsManager.getExtensionsIds())
        {
            ClientSideWidget element = _widgetsManager.getExtension(extensionId);
            dependenciesManager.register(element);
        }
        
        try
        {
            return dependenciesManager.computeDependencies();
        }
        catch (ServiceException e)
        {
            throw new SAXException("Unable to compute dependencies", e);
        }
    }

    /**
     * SAX the UI Tools
     * @param contextParameters the context parameters
     * @param elements The list of elements to sax
     * @throws IOException if an error occurred
     * @throws SAXException if an error occurred
     */
    protected void saxUITools(Map<String, Object> contextParameters, List<ClientSideElement> elements) throws IOException, SAXException
    {
        Source configSource = getUIToolsConfiguration();
        UIToolsConfigurationManager uitoolsManager = new UIToolsConfigurationManager(_uitoolsFactoriesManager, _saxClientSideElementHelper, configSource, ObjectModelHelper.getRequest(objectModel));
        uitoolsManager.saxDefaultState(contentHandler, contextParameters, elements);
    }

    /**
     * Get the ribbon configuration
     * @return the ribbon configuration
     * @throws IOException if an errors occurs getting the ribbon configuration
     */
    protected RibbonConfigurationSource getRibbonConfiguration() throws IOException
    {
        String ribbonFileName = parameters.getParameter("ribbonFileName", "ribbon");
        String mode = parameters.getParameter("mode", null);
        return RibbonConfigurationSource.createFromUri("context://WEB-INF/param/" + ribbonFileName + (mode != null ? "-" + mode : "") + ".xml", _resolver);
    }
    
    /**
     * Get the UI tools configuration
     * @return the UI tools configuration
     * @throws IOException if an errors occurs getting the UI tools configuration
     */
    protected Source getUIToolsConfiguration() throws IOException
    {
        String toolsFileName = parameters.getParameter("toolsFileName", "uitools");
        String mode = parameters.getParameter("mode", null);
        return _resolver.resolveURI("context://WEB-INF/param/" + toolsFileName + (mode != null ? "-" + mode : "") + ".xml");
    }
    
    /**
     * SAX the message target factories
     * @param contextParameters the context parameters
     * @param elements The list of elements for the message target factories 
     * @throws SAXException if an error occurred
     */
    protected void saxMessageTargetFactories(Map<String, Object> contextParameters, List<ClientSideElement> elements) throws SAXException
    {
        contentHandler.startPrefixMapping("i18n", "http://apache.org/cocoon/i18n/2.1");
        XMLUtils.startElement(contentHandler, "messagetarget-factories");

        if (elements != null)
        {
            for (ClientSideElement element : elements)
            {
                _saxClientSideElementHelper.saxDefinition("messagetarget-factory", element, MessageTargetFactoriesManager.ROLE, contentHandler, contextParameters);
            }
        }
        
        XMLUtils.endElement(contentHandler, "messagetarget-factories");
        contentHandler.endPrefixMapping("i18n");
    }
    
    /**
     * SAX the relations handlers
     * @param contextParameters the context parameters
     * @param elements The list of relation handlers
     * @throws SAXException if an error occurred
     */
    protected void saxRelationsHandlers(Map<String, Object> contextParameters, List<ClientSideElement> elements) throws SAXException
    {
        contentHandler.startPrefixMapping("i18n", "http://apache.org/cocoon/i18n/2.1");
        XMLUtils.startElement(contentHandler, "relations-handlers");

        if (elements != null)
        {
            for (ClientSideElement element: elements)
            {
                _saxClientSideElementHelper.saxDefinition("relation-handler", element, RelationsManager.ROLE, contentHandler, contextParameters);
            }
        }
        
        XMLUtils.endElement(contentHandler, "relations-handlers");
        contentHandler.endPrefixMapping("i18n");
    }
    
    /**
     * SAX the widgets
     * @param contextParameters the context parameters
     * @throws SAXException if an error occurred
     */
    protected void saxWidgets(Map<String, Object> contextParameters) throws SAXException
    {
        contentHandler.startPrefixMapping("i18n", "http://apache.org/cocoon/i18n/2.1");
        
        Map<String, Map<String, Map<String, String>>> defaultWidgets = _widgetsManager.getDefaultWidgets();
        AttributesImpl wattrs = new AttributesImpl();
        wattrs.addCDATAAttribute("default-widgets", _jsonUtils.convertObjectToJson(defaultWidgets));
        
        XMLUtils.startElement(contentHandler, "widgets", wattrs);

        for (String extensionId: _widgetsManager.getExtensionsIds())
        {
            ClientSideWidget element = _widgetsManager.getExtension(extensionId);
            
            AttributesImpl attrs = new AttributesImpl();
            attrs.addCDATAAttribute("ftypes", StringUtils.join(element.getFormTypes(contextParameters), ","));
            attrs.addCDATAAttribute("supports-enumerated", Boolean.toString(element.supportsEnumerated(contextParameters)));
            attrs.addCDATAAttribute("supports-non-enumerated", Boolean.toString(element.supportsNonEnumerated(contextParameters)));
            attrs.addCDATAAttribute("supports-multiple", Boolean.toString(element.supportsMultiple(contextParameters)));
            attrs.addCDATAAttribute("supports-non-multiple", Boolean.toString(element.supportsNonMultiple(contextParameters)));
            
            XMLUtils.startElement(contentHandler, "widget-wrapper", attrs);
            _saxClientSideElementHelper.saxDefinition("widget", element, WidgetsManager.ROLE, contentHandler, contextParameters);
            XMLUtils.endElement(contentHandler, "widget-wrapper"); 
        }
        
        XMLUtils.endElement(contentHandler, "widgets");
        contentHandler.endPrefixMapping("i18n");
    }
    
    /**
     * SAX the static file imports
     * @param contextParameters the context parameters
     * @param elements The list of static file imports elements
     * @throws SAXException if an error occurred
     */
    protected void saxStaticFileImports (Map<String, Object> contextParameters, List<ClientSideElement> elements) throws SAXException
    {
        XMLUtils.startElement(contentHandler, "static-imports");
        if (elements != null)
        {
            for (ClientSideElement element : elements)
            {
                _saxClientSideElementHelper.saxDefinition("import", element, StaticFileImportsManager.ROLE, contentHandler, contextParameters);
            }
        }
        XMLUtils.endElement(contentHandler, "static-imports");
    }

    /**
     * Use this method when inheriting the WorkspaceGenerator to sax additional data 
     * @param contextParameters the context parameters
     * @throws SAXException if an error occurred
     */
    protected void saxAdditionnalInfo(Map<String, Object> contextParameters) throws SAXException
    {
        if (PluginsManager.getInstance().isSafeMode())
        {
            XMLUtils.createElement(contentHandler, "safe-mode", PluginsManager.getInstance().getStatus().toString());
        }
        if (RuntimeServlet.getRunMode() == RunMode.MAINTENANCE)
        {
            XMLUtils.createElement(contentHandler, "maintenance");
        }
    }
}
