/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.glyph;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.core.ui.Callable;
import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentExtensionPoint;

/**
 * Manager for glyphs provide by the application
 *
 */
public class ApplicationGlyphManager extends AbstractThreadSafeComponentExtensionPoint<GlyphProvider>
{
    private CssFontHelper _cssFontHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _cssFontHelper = (CssFontHelper) smanager.lookup(CssFontHelper.ROLE);
    }
    
    /**
     * Determines if the back office has available glyphs
     * @return true if the back office has available glyphs
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public boolean hasGlyphs()
    {
        return !getGlyphs().isEmpty();
    }
    
    /**
     * Get the glyphs provide by the application
     * @return The CSS class names for glyphs
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public List<Map<String, String>> getGlyphsStore()
    {
        List<Map<String, String>> glyphs = new ArrayList<>();

        Set<String> glyphClassNames = getGlyphs();
        for (String glyphClassName : glyphClassNames)
        {
            Map<String, String> glyph = new HashMap<>();
            glyph.put("cssClassName", glyphClassName);
            glyphs.add(glyph);
        }
        
        return glyphs;
    }
    
    /**
     * Get the CSS class names for glyphs provide by the application
     * @return The CSS class names for glyphs
     */
    public Set<String> getGlyphs()
    {
        Set<String> glyphs = new LinkedHashSet<>();
        
        for (String id : getExtensionsIds())
        {
            GlyphProvider glyphProvider = getExtension(id);
            
            Map<String, String> cssFiles = glyphProvider.getCSSFiles();
            
            for (Entry<String, String> cssFile : cssFiles.entrySet())
            {
                String cssFileURI = cssFile.getKey();
                String prefix = cssFile.getValue();
                glyphs.addAll(_cssFontHelper.getGlyphClassNames(cssFileURI, prefix));
            }
        }
        
        return glyphs;
    }
}
