/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.glyph;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceNotFoundException;
import org.apache.excalibur.source.SourceResolver;
import org.w3c.css.sac.InputSource;
import org.w3c.dom.css.CSSFontFaceRule;
import org.w3c.dom.css.CSSRule;
import org.w3c.dom.css.CSSRuleList;
import org.w3c.dom.css.CSSStyleDeclaration;
import org.w3c.dom.css.CSSStyleRule;
import org.w3c.dom.css.CSSStyleSheet;

import org.ametys.runtime.plugin.component.AbstractLogEnabled;

import com.steadystate.css.parser.CSSOMParser;
import com.steadystate.css.parser.SACParserCSS3;

/**
 * Helper for parsing of CSS file and extract glyphs
 *
 */
public class CssFontHelper extends AbstractLogEnabled implements Component, Serviceable
{
    /** The Avalon role */
    public static final String ROLE = CssFontHelper.class.getName();
    
    /** Pattern for CSS class */
    private static final Pattern __CSS_CLASS_NAME = Pattern.compile("\\.-?[_a-zA-Z]+[_a-zA-Z0-9-]*\\s*");
    
    private SourceResolver _resolver;
    
    private Map<String, Long> _lastUpdates = new HashMap<>();
    private Map<String, Set<String>> _glyphs = new HashMap<>();
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    /**
     * Get the CSS class names for glyphes
     * @param cssURI The URI of CSS file
     * @param prefix The glyph prefix. Can be null.
     * @return The CSS class names for glyphes
     */
    public Set<String> getGlyphClassNames (String cssURI, String prefix)
    {
        Source cssSource = null;
        try
        {
            cssSource = _resolver.resolveURI(cssURI);
            if (cssSource.exists())
            {
                Long lastUpdate = _lastUpdates.get(cssSource.getURI());
                
                if (lastUpdate == null || lastUpdate < cssSource.getLastModified())
                {
                    Set<String> classNames = _getGlyphClassNames(cssSource, prefix);
                    
                    getLogger().debug("Found {} glyphes into CSS file {}", classNames.size(), cssSource.getURI());
                    
                    _lastUpdates.put(cssSource.getURI(), cssSource.getLastModified());
                    _glyphs.put(cssSource.getURI(), classNames);
                }
                
                return _glyphs.get(cssSource.getURI());
            }
            else
            {
                throw new SourceNotFoundException("The CSS file " + cssURI + " does not exist");
            }
        }
        catch (SourceNotFoundException e)
        {
            getLogger().warn("The CSS file at uri <{}> does not exist. It will be ignored.", cssURI, e);
        }
        catch (IOException e)
        {
            getLogger().warn("Fail to read the CSS file at uri <{}>. It will be ignored.", cssURI, e);
        }
        finally 
        {
            _resolver.release(cssSource);
        }
        
        return Collections.EMPTY_SET;
    }
    
    private Set<String> _getGlyphClassNames(Source cssSource, String prefix) throws IOException
    {
        Set<String> classNames = new LinkedHashSet<>();
        
        CSSStyleSheet styleSheet = _getCssStyleSheet(cssSource);
        
        Set<String> commonClassNames = getFontClassNames(styleSheet);
        
        CSSRuleList rules = styleSheet.getCssRules();
        for (int i = 0; i < rules.getLength(); i++)
        {
            CSSRule rule = rules.item(i);
            
            if (rule instanceof CSSStyleRule)
            {
                CSSStyleDeclaration style = ((CSSStyleRule) rule).getStyle();
                
                // Filter CSS style with 'content' property
                if (StringUtils.isNotEmpty(style.getPropertyValue("content")))
                {
                    String selectorText = ((CSSStyleRule) rule).getSelectorText();
                    Matcher matcher = __CSS_CLASS_NAME.matcher(selectorText);
                    while (matcher.find())
                    {
                        // Remove point at the begining of css class name
                        String className = matcher.group().substring(1);
                        
                        if (commonClassNames.isEmpty()
                            && (StringUtils.isNotBlank(prefix) && className.startsWith(prefix)
                                || StringUtils.isBlank(prefix)))
                        {
                            classNames.add(className);
                        }
                        else if (commonClassNames.contains(className))
                        {
                            // This case means that the class name for glyph already contains the font-family in another rule and so was detected as a "common" class name
                            // but it is not a common class name (see in icomoon.css for example)
                            classNames.add(className);
                        }
                        else
                        {
                            for (String commonClassName : commonClassNames)
                            {
                                String computedClassName = commonClassName + " " + className;
                                if (StringUtils.isNotBlank(prefix) && computedClassName.startsWith(prefix)
                                    || StringUtils.isBlank(prefix))
                                {
                                    classNames.add(computedClassName);
                                }
                            }
                        }
                    }
                }
            }
        }
        
        return classNames;
    }
    
    /**
     * Determines if the CSS file has a 'font-face' rule 
     * @param cssURI The URI of CSS file
     * @return cssURI The URI of CSS file
     */
    public boolean hasFontFaceRule(String cssURI)
    {
        Source cssSource = null;
        try
        {
            cssSource = _resolver.resolveURI(cssURI);
            if (cssSource.exists())
            {
                Long lastUpdate = _lastUpdates.get(cssSource.getURI());
                
                if (lastUpdate == null || lastUpdate < cssSource.getLastModified())
                {
                    CSSStyleSheet cssStyleSheet = _getCssStyleSheet(cssSource);
                    CSSRuleList cssRules = cssStyleSheet.getCssRules();
                    
                    for (int i = 0; i < cssRules.getLength(); i++)
                    {
                        CSSRule cssRule = cssRules.item(i);
                        if (cssRule instanceof CSSFontFaceRule)
                        {
                            return true;
                        }
                    }
                }
                else
                {
                    // if file is in last update cache, the file is necessary a font css
                    return true;
                }
            }
        }
        catch (IOException e)
        {
            getLogger().warn("Fail to read the CSS file at uri <{}>.", cssURI, e);
        }
        finally 
        {
            _resolver.release(cssSource);
        }
        
        return false;
    }
    
    /**
     * Get CSS class name to use for all icons. Can be empty if the icons contains the font-family to use.
     * @param styleSheet The CSS stylesheet to parse
     * @return The name of the common CSS class name for icons (ex: 'fa' for FontAwesome). Can be empty if the icons contains the font-family to use.
     */
    public Set<String> getFontClassNames(CSSStyleSheet styleSheet)
    {
        CSSRuleList rules = styleSheet.getCssRules();

        Set<String> fontClassNames = new HashSet<>();
        Set<String> fontFamilies = _getFontFamily(styleSheet);

        for (int i = 0; i < rules.getLength(); i++)
        {
            CSSRule rule = rules.item(i);
            if (rule instanceof CSSStyleRule)
            {
                // Search if there is a common CSS class to all icons as FontAwesome does :
                // .fa { font: normal normal normal 14px/1 FontAwesome } 
                // .fa-bath { content: "\f2cd"}
                // => We need here to extract "fa" CSS classname
                
                CSSStyleRule styleRule = (CSSStyleRule) rule;
                CSSStyleDeclaration style = styleRule.getStyle();
                
                if (StringUtils.isEmpty(style.getPropertyValue("content")))
                {
                    if (_matchFontFamily(fontFamilies, style))
                    {
                        String selectorText = styleRule.getSelectorText();
                        Matcher matcher = __CSS_CLASS_NAME.matcher(selectorText);
                        while (matcher.find())
                        {
                            fontClassNames.add(matcher.group().substring(1));
                        }
                    }
                }
            }
        }
        
        return fontClassNames;
    }
    
    private boolean _matchFontFamily(Set<String> fontFamilies, CSSStyleDeclaration style)
    {
        String fontValue = style.getPropertyValue("font");
        String fontFamilyValue = style.getPropertyValue("font-family");
        
        if (StringUtils.isNotBlank(fontValue) || StringUtils.isNotBlank(fontFamilyValue))
        {
            for (String fontFamily : fontFamilies)
            {
                if (StringUtils.isNotBlank(fontValue) && fontValue.contains(fontFamily)
                    || StringUtils.isNotBlank(fontFamilyValue) && fontFamilyValue.contains(fontFamily))
                {
                    return true;
                }
            }
        }
        
        
        return false;
    }
    
    private CSSStyleSheet _getCssStyleSheet(Source source) throws MalformedURLException, IOException
    {
        try
        {
            InputSource inputSource = new InputSource();
            inputSource.setByteStream(source.getInputStream());
            inputSource.setEncoding("UTF-8");
            CSSOMParser parser = new CSSOMParser(new SACParserCSS3());
            return parser.parseStyleSheet(inputSource, null, null);
        }
        finally
        {
            _resolver.release(source);
        }
    }
    
    private Set<String> _getFontFamily(CSSStyleSheet styleSheet)
    {
        Set<String> fontsFamily = new HashSet<>();
        
        CSSRuleList cssRules = styleSheet.getCssRules();
        for (int i = 0; i < cssRules.getLength(); i++)
        {
            CSSRule cssRule = cssRules.item(i);
            if (cssRule instanceof CSSFontFaceRule)
            {
                CSSStyleDeclaration style = ((CSSFontFaceRule) cssRule).getStyle();
                String fontFamily = style.getPropertyValue("font-family");
                fontsFamily.add(StringUtils.strip(StringUtils.strip(fontFamily, "\""), "'"));
            }
        }
        return fontsFamily;
    }
}
