/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.help;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.avalon.framework.thread.ThreadSafe;

import org.ametys.runtime.plugin.ExtensionPoint;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;


/**
 * This extension point handle a list of help URLs handled by the plugins or the application.
 */
public class HelpExtensionPoint extends AbstractLogEnabled implements ExtensionPoint<String>, Initializable, ThreadSafe, Component, Serviceable
{
    /** The avalon role */
    public static final String ROLE = HelpExtensionPoint.class.getName();
    
    /** The map of rightId, Right of declared rights */
    protected Map<String, HelpLink> _links;
    /** Set of extensions points added */
    protected Set<String> _points;
    /** The help manager */
    protected HelpManager _helpManager;
    
    public void initialize() throws Exception
    {
        _links = new HashMap<>();
        _points = new HashSet<>();
    }
    public void service(ServiceManager manager) throws ServiceException
    {
        _helpManager = (HelpManager) manager.lookup(HelpManager.ROLE);
    }
    
    public boolean hasExtension(String id)
    {
        return _points.contains(id);
    }

    public void addExtension(String id, String pluginName, String featureName, Configuration configuration) throws ConfigurationException
    {
        _points.add(id);
        getLogger().debug("Adding help URLs from plugin {}/{}", pluginName,  featureName);
        
        Map<String, HelpLink> links = _helpManager.parseConfigFile(configuration, pluginName, featureName);
        
        for (Entry<String, HelpLink> link : links.entrySet())
        {
            if (_links.containsKey(link.getKey()))
            {
                getLogger().warn("There is already a help key for this point{}id : {}", HelpLink.SEPARATOR, link.getKey());
            }
            else
            {
                _links.put(link.getKey(), link.getValue());
            }
        }
    }
    
    /**
     * Return the helpLink available for this point/id
     * @param point extension point
     * @param id id of the extension for this extension point
     * @return a {@link HelpLink}
     */
    public HelpLink getHelpLink(String point, String id)
    {
        String uniqueId = HelpLink.getUniqueId(point, id);
        return _links.get(uniqueId);
    }

    public Set<String> getExtensionsIds()
    {
        return _points;
    }

    public void initializeExtensions() throws Exception
    {
        // empty
    }
    
    public String getExtension(String id)
    {
        if (_points.contains(id))
        {
            return id;
        }
        else
        {
            return null;
        }
    }
}
