/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.plugins.core.ui.help;

import java.util.Collections;
import java.util.Map;
import java.util.Set;

/**
 * Represent a link used for help (point/id and list of urls by lang)
 */
public class HelpLink
{
    /** default language for the links if the requested one is not found */
    public static final String DEFAULT_LANGUAGE = "en";
    /** Separator to create the unique id in getUniqueId */
    public static final String SEPARATOR = "#";
    private String _family;
    private String _id;
    private Map<String, String> _urls;
    
    /**
     * Create a HelpLink for an extension point/id and a list of urls
     * @param family the family for this link (for example an extension point)
     * @param id the id in the family (must be unique in the family)
     * @param urls a map with language/url for the help pages
     */
    public HelpLink(String family, String id, Map<String, String> urls)
    {
        this._family = family;
        this._id = id;
        this._urls = urls;
    }
    
    /**
     * Generate an unique ID to access this link accross a map
     * @param family the family for this link (for example an extension point)
     * @param id the id in the family (must be unique in the family)
     * @return an Unique ID
     */
    public static String getUniqueId(String family, String id)
    {
        return family + SEPARATOR + id;
    }
    
    /**
     * Get an unique ID across the application
     * @return an unique ID
     */
    public String getUniqueId()
    {
        return getUniqueId(this._family, this._id);
    }
    
    /**
     * Get all urls in a map of language/url
     * @return the list of urls available (can be null)
     */
    public Map<String, String> getUrls()
    {
        return _urls;
    }
    
    /**
     * Get the url for a language, or a default one if the requested language is null
     * @param language language requested (null if you want any language available, starting with the default one)
     * @return a help url, or null if nothing is found or map empty
     */
    public String getUrl(String language)
    {
        if (_urls != null)
        {
            return _urls.get(language);
        }
        else
        {
            return null;
        }
    }
    
    /**
     * Get a set of available languages
     * @return a set of available languages (can not be null)
     */
    public Set<String> getLanguages()
    {
        if (_urls != null)
        {
            return _urls.keySet();
        }
        else
        {
            return Collections.emptySet();
        }
    }
    
    /**
     * Get the the family for this link (for example an extension point)
     * @return the extension point
     */
    public String getFamily()
    {
        return _family;
    }
    
    /**
     * Get the extension id in the extension point, for this link
     * @return the id in the family (unique in the family)
     */
    public String getId()
    {
        return _id;
    }
}
