/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.help;

import java.io.InputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;

import org.ametys.core.ui.Callable;
import org.ametys.core.util.filereloader.FileReloader;
import org.ametys.core.util.filereloader.FileReloaderUtils;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Provides the url associated to a family/id
 */
public class HelpManager extends AbstractLogEnabled implements Component, Serviceable, FileReloader, Initializable
{
    /** The avalon role */
    public static final String ROLE = HelpManager.class.getName();
    
    private static final String DEFAULT_HELP_URL = "https://docs.ametys.org/{lang}/index.html";
    private static final String DEFAULT_SEARCH_URL = "https://docs.ametys.org/{lang}/search.html?textfield={query}&submit-form";
    
    /** The map of rightId, Right of declared rights */
    protected Map<String, HelpLink> _links;

    private FileReloaderUtils _fileReloader;
    private HelpExtensionPoint _helpEP;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _fileReloader = (FileReloaderUtils) manager.lookup(FileReloaderUtils.ROLE);
        _helpEP = (HelpExtensionPoint) manager.lookup(HelpExtensionPoint.ROLE);
    }
    public void initialize() throws Exception
    {
        _links = new HashMap<>();
    }
    
    /**
     * Get the url for help for this family/id.
     * The result is a map containing "url", "error" and "error-cause"
     * @param family the family for this link (for example an extension point)
     * @param id id in the family (must be unique in the family)
     * @param lang requested language
     * @return an Map containg the URL for help. If this language is not found, a default language will be searched, then any language. If nothing is found, the url will be null.
     * @throws Exception impossible to read the default help file
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> getHelpUrl (String family, String id, String lang) throws Exception
    {
        Map<String, Object> resultMap = new HashMap<>();
        
        String url = getHelp(family, id, lang);
        resultMap.put("url", url);
        
        if (StringUtils.isEmpty(url))
        {
            resultMap.put("error", "true");
            resultMap.put("error-cause", "not-found");
        }
        
        return resultMap;
    }
    
    /**
     * Get the default help url
     * @return The result (map with an url key, or possible error)
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> getDefaultHelpUrl()
    {
        Map<String, Object> resultMap = new HashMap<>();
        
        String url = null;
        if (Config.getInstance() != null)
        {
            url = Config.getInstance().getValue("online.help.index.url");
        }
        if (StringUtils.isBlank(url))
        {
            url = DEFAULT_HELP_URL;
        }
        resultMap.put("url", url);
        return resultMap;
    }
    /**
     * Get the search url
     * @return The result (map with an url key, or possible error)
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> getSearchUrl ()
    {
        Map<String, Object> resultMap = new HashMap<>();
        String url = null;
        if (Config.getInstance() != null)
        {
            url = Config.getInstance().getValue("online.help.search.url");
        }
        if (StringUtils.isBlank(url))
        {
            url = DEFAULT_SEARCH_URL;
        }
        resultMap.put("url", url);
        
        return resultMap;
    }
    
    /**
     * Get the url for help for this family/id or null if not found
     * @param family the family for this link (for example an extension point)
     * @param id id in the family (must be unique in the family)
     * @param lang requested language
     * @return an URL for help. If this language is not found, a default language will be searched, then any language, and null if really nothing is found
     * @throws Exception impossible to read the default help file
     */
    public String getHelp(String family, String id, String lang) throws Exception
    {
        List<String> languages = Collections.singletonList(lang);
        return getHelp(family, id, languages);
    }
    
    /**
     * Get the url for help for this family/id or null if not found
     * @param family the family for this link (for example an extension point)
     * @param id id in the family (must be unique in the family)
     * @param languages Ordered list of requested languages
     * @return an URL for help. If no language is found, a default language will be searched, then any other language, and null if really nothing is found
     * @throws Exception impossible to read the default help file
     */
    public String getHelp(String family, String id, List<String> languages) throws Exception
    {
        _fileReloader.updateFile("context://WEB-INF/param/help.xml", this);

        String result = null;
        for (String lang : languages)
        {
            result = _getHelp(family, id, lang);
            if (result != null)
            {
                break; //we stop to iterate on the first match
            }
        }
        // Search for the requested lang
        
        // If not found, search default language
        if (result == null)
        {
            result = _getHelp(family, id, HelpLink.DEFAULT_LANGUAGE);
        }
        
        // If still not found, search any language
        if (result == null)
        {
            result = _getAnyLanguageHelp(family, id);
        }
        
        // If blank, return null
        if (StringUtils.isBlank(result))
        {
            return null;
        }
        else
        {
            return result;
        }
    }

    private String _getHelp(String family, String id, String lang)
    {
        
        String url = null;
        
        HelpLink helpLink = _links.get(HelpLink.getUniqueId(family, id));
        if (helpLink != null)
        {
            url = helpLink.getUrl(lang);
        }
        
        if (StringUtils.isNotBlank(url))
        {
            return url;
        }
        else
        {
            helpLink = _helpEP.getHelpLink(family, id);
            if (helpLink != null)
            {
                url = helpLink.getUrl(lang);
            }
        }
        
        return url;
    }
    
    private String _getAnyLanguageHelp(String family, String id)
    {
        String url = null;
        
        HelpLink helpLink = _links.get(HelpLink.getUniqueId(family, id));
        if (helpLink != null)
        {
            url = _getAnyLanguageHelp(helpLink);
        }
        
        if (StringUtils.isNotBlank(url))
        {
            return url;
        }
        else
        {
            helpLink = _helpEP.getHelpLink(family, id);
            if (helpLink != null)
            {
                url = _getAnyLanguageHelp(helpLink);
            }
        }
        
        return url;
    }
    
    private String _getAnyLanguageHelp(HelpLink helpLink)
    {
        if (helpLink != null)
        {
            Map<String, String> urls = helpLink.getUrls();
            if (urls != null && !urls.isEmpty())
            {
                return urls.values().iterator().next();
            }
        }
        return null;
    }
    
    public void updateFile(String sourceUrl, Source source, InputStream is) throws Exception
    {
        if (is != null)
        {
            Configuration cfg = new DefaultConfigurationBuilder().build(is);
            _links = parseConfigFile(cfg, "WEB-INF", "help");
        }
        else
        {
            _links = new HashMap<>();
        }
    }
    public String getId(String sourceUrl)
    {
        return HelpManager.class.getName();
    }

    /**
     * parse a configuration file for a plugin (plugin name and id)
     * @param configuration the configuration
     * @param pluginName Unique identifier for the plugin hosting the extension (for logs)
     * @param featureName Unique feature identifier (unique for a given pluginName) (for logs)
     * @return a map of HelpLink.uniqueId HelpLink representing the help URLs found in the configuration file 
     */
    public Map<String, HelpLink> parseConfigFile(Configuration configuration, String pluginName, String featureName)
    {
        Map<String, HelpLink> links = new HashMap<>();
        Configuration[] linkConfigurations = configuration.getChildren("link");
        for (Configuration linkConfiguration : linkConfigurations)
        {
            HelpLink link = _addLink(linkConfiguration, pluginName, featureName);
            if (link != null)
            {
                String id = link.getUniqueId();
                if (links.containsKey(id))
                {
                    getLogger().warn("There is already a help key for this plugin.feature '{}.{}' on key '{}'", pluginName, featureName, id);
                }
                else
                {
                    links.put(id, link);
                }
            }
        }
        return links;
    }
    
    /**
     * adds a link in the global plugin map
     * @param linkConfiguration link node in the configuration
     * @param pluginName Unique identifier for the plugin hosting the extension (for logs)
     * @param featureName Unique feature identifier (unique for a given pluginName) (for logs)
     * @return the link or null
     */
    private HelpLink _addLink(Configuration linkConfiguration, String pluginName, String featureName)
    {
        String family = null;
        String id = null;
        try
        {
            family = linkConfiguration.getAttribute("family");
            id = linkConfiguration.getAttribute("id");
            
            Configuration[] languages = linkConfiguration.getChildren();
            Map<String, String> urls = new HashMap<>();
            for (Configuration language : languages)
            {
                String lang = language.getName();
                String url = language.getValue(StringUtils.EMPTY); // default to empty String
                urls.put(lang, url);
            }
            return new HelpLink(family, id, urls);
        }
        catch (ConfigurationException e)
        {
            getLogger().warn("The plugin '{}.{}' has an help extension but has an incorrect configuration (for family '{}' and feature '{}'", pluginName, featureName, StringUtils.defaultIfBlank(family, "<unknown>"), StringUtils.defaultIfBlank(id, "<unknown>"), e);
            return null;
        }
    }
    
}
