/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.help;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Locale;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.slf4j.Logger;

import org.ametys.runtime.plugin.component.LogEnabled;

import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;

/**
 * Serializer for {@link HelpLink} objects.
 * Returns the help url for the right language, for JSON value.
 */
public class HelpSerializer extends StdSerializer<HelpLink> implements Component, Serviceable, LogEnabled, Contextualizable
{
    /** The Avalon Role */
    public static final String ROLE = HelpSerializer.class.getName();
    private Logger _logger;
    private HelpManager _helpManager;
    private Context _context;
    
    /**
     * Constructor
     */
    public HelpSerializer()
    {
        super(HelpLink.class);
    }

    @Override
    public void serialize(HelpLink value, JsonGenerator gen, SerializerProvider provider) throws IOException
    {
        if (value != null)
        {
            Request request = ContextHelper.getRequest(_context);
            List<String> languages = _getLanguages(request);
            try
            {
                String msg = _helpManager.getHelp(value.getFamily(), value.getId(), languages);
                if (msg == null)
                {
                    gen.writeNull();
                }
                else
                {
                    gen.writeString(msg);
                }
            }
            catch (Exception e)
            {
                _logger.warn("Impossible to serialize help link for point : '{}' and id : '{}'", value.getFamily(), value.getId(), e);
            }
        }
    }

    public void setLogger(Logger logger)
    {
        _logger = logger;
    }

    public void service(ServiceManager manager) throws ServiceException
    {
        _helpManager = (HelpManager) manager.lookup(HelpManager.ROLE);
    }

    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    private List<String> _getLanguages(Request request)
    {
        List<String> languages = new ArrayList<>();
        if (request != null)
        {
            Enumeration locales = request.getLocales();
            while (locales.hasMoreElements())
            {
                Locale locale = (Locale) locales.nextElement();
                String lang = locale.getLanguage();
                if (!languages.contains(lang))
                {
                    languages.add(lang);
                }
            }
        }
        return languages;
    }
}
