/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.log;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.core.ui.log.parser.LogFileParser;
import org.ametys.runtime.util.AmetysHomeHelper;

/**
 * Client side element to display archived logs.
 */
public class ArchivedLogClientSideElement extends StaticClientSideElement
{
    /** JSON Utils */
    protected JSONUtils _jsonUtils;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _jsonUtils = (JSONUtils) smanager.lookup(JSONUtils.ROLE);
    }
    
    /**
     * Get the log lines of the given log file.
     * @param logFileName The log file name.
     * @param filtersJson The filters (received as a JSON object)
     * @param limit Limit of displayed events (can be 0 or -1 for no limit)
     * @return A {@link Map}, in "lines" there are the log lines, in "error" there is the error message if an error occurs.
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> getLogLines(String logFileName, String filtersJson, Integer limit)
    {
        Map<String, Object> result = new HashMap<>();
        File logFile = new File(_getLogsDirectory(), logFileName);  
        if (logFile.exists())
        {
            try
            {
                result.put("lines", LogFileParser.parseFile(logFile, _convertFilters(filtersJson), limit, getLogger()));
            }
            catch (IOException e)
            {
                String errorMessage = String.format("An error occurs while parsing the log file '%s'.", logFileName);
                result.put("error", errorMessage);
                getLogger().error(errorMessage, e);
            }
        }
        else
        {
            result.put("error", String.format("The file '%s' doesn't exists in the logs directory.", logFileName));
        }
        return result;
    }
    
    /**
     * Get the logs directory.
     * @return The directory where the logs are stored
     */
    private File _getLogsDirectory()
    {
        return new File(AmetysHomeHelper.getAmetysHome(), "logs");
    }
    
    @SuppressWarnings("unchecked")
    private Map<String, Object> _convertFilters(String filtersJson)
    {
        Map<String, Object> filtersAsMap = new HashMap<>();
        
        if (StringUtils.isNotBlank(filtersJson))
        {
            List<Object> filtersAsList = _jsonUtils.convertJsonToList(filtersJson);
            for (Object filterObject : filtersAsList)
            {
                Map<String, Object> filterMap = (Map<String, Object>) filterObject;
                filtersAsMap.put(filterMap.get("property").toString(), filterMap.get("value"));
            }
        }
        
        return filtersAsMap;
    }
}
