/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.log.parser;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.io.input.ReversedLinesFileReader;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;

/**
 * A log file parser.
 * The log file is parsed by the end with a limit of [limit] events.
 */
public final class LogFileParser
{
    private static final DateTimeFormatter __DATE_FORMAT = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss,SSS");
    private static final Pattern __REGEXP = Pattern.compile("^([0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2},[0-9]{3}) ([A-Z ]{5}) \\[([^\\]]+)\\] \\(([^ ]+)\\) (.*)$");

    /** Default constructor */
    private LogFileParser()
    {
        // Nothing to do
    }
    
    /**
     * Parse the log file.
     * @param logFile The log file
     * @param filters The filters on logs
     * @param limit The limit of events
     * @param logger The logger
     * @return A {@link List} of parsed log lines
     * @throws IOException if an error occurs
     */
    public static List<Map<String, Object>> parseFile(File logFile, Map<String, Object> filters, int limit, Logger logger) throws IOException
    {
        List<Map<String, Object>> logLines = new ArrayList<>();
        
        try (ReversedLinesFileReader reader = ReversedLinesFileReader.builder().setFile(logFile).setCharset(StandardCharsets.UTF_8).get())
        {
            StringBuilder stackTrace = new StringBuilder();
            String line = reader.readLine();
            while (line != null && (limit <= 0 || logLines.size() < limit))
            {
                Matcher matcher = __REGEXP.matcher(line);
                if (matcher.matches())
                {
                    Map<String, Object> logLine = _getLogLine(matcher, stackTrace.toString(), logger);
                    
                    // Add if filters are OK
                    if (_filter(logLine, filters))
                    {
                        logLines.add(logLine);
                    }
                    
                    // Reset the detail
                    stackTrace.setLength(0);
                }
                else
                {
                    // Stack trace
                    stackTrace.insert(0, line + "<br/>");
                }
                line = reader.readLine();
            }
        }
        
        return logLines;
    }
    
    private static Map<String, Object> _getLogLine(Matcher matcher, String stackTrace, Logger logger)
    {
        Map<String, Object> logLine = new HashMap<>();
        logLine.put("timestamp", _convertToMs(matcher.group(1), logger));
        logLine.put("level", matcher.group(2).trim());
        logLine.put("category", matcher.group(3));
        logLine.put("thread", matcher.group(4));
        logLine.put("message", matcher.group(5));
        logLine.put("callstack", stackTrace);
        return logLine;
    }
    
    private static long _convertToMs(String dateTime, Logger logger)
    {
        long milliseconds = 0;
        try
        {
            LocalDateTime ldt = LocalDateTime.parse(dateTime, __DATE_FORMAT);
            ZonedDateTime zdt = ldt.atZone(ZoneId.systemDefault());
            milliseconds = zdt.toInstant().toEpochMilli();
        }
        catch (DateTimeParseException e)
        {
            logger.error("Impossible to parse the date {} to milliseconds with the format {}", dateTime, __DATE_FORMAT, e);
        }
        return milliseconds;
    }
    
    @SuppressWarnings("unchecked")
    private static boolean _filter(Map<String, Object> logLine, Map<String, Object> filters)
    {
        if (filters != null && !filters.isEmpty())
        {
            List<String> filterLevels = (List<String>) filters.get("level");
            if (filterLevels != null && !filterLevels.isEmpty() && !filterLevels.contains(logLine.get("level")))
            {
                return false;
            }

            String filterMessage = (String) filters.get("message");
            if (StringUtils.isNotBlank(filterMessage) && !StringUtils.containsIgnoreCase((String) logLine.get("message"), filterMessage))
            {
                return false;
            }
            
            String filterCategory = (String) filters.get("category");
            if (StringUtils.isNotBlank(filterCategory) && !StringUtils.containsIgnoreCase((String) logLine.get("category"), filterCategory))
            {
                return false;
            }
        }
        
        return true;
    }
}
