/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.plugins.core.ui.minimize;

import java.io.IOException;
import java.io.Serializable;
import java.nio.charset.StandardCharsets;
import java.util.List;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.caching.CacheableProcessingComponent;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.Response;
import org.apache.cocoon.reading.ServiceableReader;
import org.apache.commons.io.IOUtils;
import org.apache.excalibur.source.SourceValidity;
import org.apache.excalibur.source.impl.validity.NOPValidity;
import org.xml.sax.SAXException;

import org.ametys.core.DevMode;
import org.ametys.core.DevMode.DEVMODE;
import org.ametys.core.resources.ProxiedContextPathProvider;
import org.ametys.plugins.core.ui.minimize.HashCache.UriData;

/**
 * This generator generates a single file to load all ui items files.
 * Can generates a list of imports of directly intergrates all files.
 */
public abstract class AbstractMinimizeReader extends ServiceableReader implements CacheableProcessingComponent
{
    private HashCache _hashCache;
    private ProxiedContextPathProvider _proxiedContextPathProvider;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _hashCache = (HashCache) smanager.lookup(HashCache.ROLE);
        _proxiedContextPathProvider = (ProxiedContextPathProvider) manager.lookup(ProxiedContextPathProvider.ROLE);
    }

    @Override
    public Serializable getKey()
    {
        return source + "*" + _proxiedContextPathProvider.getContextPath();
    }
    
    @Override
    public SourceValidity getValidity()
    {
        return new NOPValidity();
    }

    @Override
    public long getLastModified()
    {
        return 1000; // the smallest second-precision value
    }
    
    @Override
    public void setup(org.apache.cocoon.environment.SourceResolver res, java.util.Map obj, String src, org.apache.avalon.framework.parameters.Parameters par) throws ProcessingException, SAXException, IOException 
    {
        super.setup(res, obj, src, par);
        Response response = ObjectModelHelper.getResponse(objectModel);
        response.setHeader("Cache-Control", "max-age=31536000, immutable"); // one year from now
    }

    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        List<UriData> files = _hashCache.getFilesForHash(source, true);
        if (files == null) 
        {
            throw new IllegalStateException("There is no list of file to minimize for hashcode '" + source + "'");
        }
        
        Request request = ObjectModelHelper.getRequest(objectModel);
        boolean generateSourceMap = !DEVMODE.PRODUCTION.equals(DevMode.getDeveloperMode(request));
        
        String str = _handleFiles(files, generateSourceMap);

        IOUtils.write(str, out, StandardCharsets.UTF_8);
        out.flush();
    }
    
    /**
     * Minimize files
     * @param files The files
     * @param generateSourceMap true to generate the source map
     * @return The minimized string
     */
    protected abstract String _handleFiles(List<UriData> files, boolean generateSourceMap);
}
