/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.plugins.core.ui.minimize;

import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.transformation.ServiceableTransformer;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import org.ametys.core.DevMode;
import org.ametys.core.DevMode.DEVMODE;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * This transformer will minimize every scripts together
 */
public class MinimizeTransformer extends ServiceableTransformer implements Contextualizable, Configurable
{
    private static final String __DEFAULT_URI_PATTERN = "^/plugins/[^/]+/resources/";
    private static final Set<String> __ALL_MEDIA = Set.of("print", "screen", "speech"); // Keep alphabetical order
    
    /** Is developer mode? */
    protected Boolean _isSuperDevMode;
    /** The avalon context */
    protected Context _context;
    /** The list of patterns available for minimization */
    protected List<Pattern> _patterns;
    /** The current local */
    protected String _locale;
    /** Should css medias be taken in account? */
    protected Boolean _inlineCssMedias;
    /** The current context */
    protected String _currentContextPath;

    /* Current queue being minified */
    private Map<String, Map<String, String>> _filesQueue;
    private String _queueTag;
    private Set<String> _queueMedia;
    private boolean _isCurrentTagQueued;
    private long _startupTime = ManagementFactory.getRuntimeMXBean().getStartTime();
    
    private HashCache _hashCache;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        
        _hashCache = (HashCache) smanager.lookup(HashCache.ROLE);
    }
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        Configuration paternsConfig = configuration.getChild("patterns");
        
        _patterns = new ArrayList<>();
        for (Configuration paternConfig : paternsConfig.getChildren("pattern"))
        {
            Pattern pattern = Pattern.compile(paternConfig.getValue());
            _patterns.add(pattern);
        }
        
        if (_patterns.isEmpty())
        {
            _patterns.add(Pattern.compile(__DEFAULT_URI_PATTERN));
        }
        
        _inlineCssMedias = Boolean.valueOf(configuration.getChild("inline-css-medias").getValue("true"));
    }
    
    /**
     * Get the url of the minimizer
     * @return The minimizer url
     */
    protected String _getMinimizeUrl()
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        return request.getContextPath() + StringUtils.defaultString((String) request.getAttribute(WorkspaceMatcher.WORKSPACE_URI)) + "/plugins/core-ui/resources-minimized/";
    }
    
    @Override
    public void setup(org.apache.cocoon.environment.SourceResolver res, Map obj, String src, Parameters par) throws ProcessingException, SAXException, IOException 
    {
        super.setup(res, obj, src, par);
        
        Request request = ContextHelper.getRequest(_context);
        _locale = request.getLocale().getLanguage();
        _isSuperDevMode = DEVMODE.SUPER_DEVELOPPMENT.equals(DevMode.getDeveloperMode(request));
        _currentContextPath = request.getContextPath();
    }
    
    @Override
    public void startDocument() throws SAXException
    {
        _filesQueue = new LinkedHashMap<>();
        _queueTag = "";
        _isCurrentTagQueued = false;
        _queueMedia = new HashSet<>();
        
        super.startDocument();
    }
    
    @Override
    public void startElement(String uri, String loc, String raw, Attributes a) throws SAXException
    {
        if (!_isSuperDevMode)
        {
            if (_isMinimizable(StringUtils.lowerCase(loc), a))
            {
                _isCurrentTagQueued = true;
                _addToQueue(StringUtils.lowerCase(loc), a);
            }
            else
            {
                _isCurrentTagQueued = false;
                
                if (_filesQueue.size() > 0)
                {
                    _processQueue();
                }
                
                super.startElement(uri, loc, raw, a);
            }
        }
        else
        {
            super.startElement(uri, loc, raw, a);
        }
    }
    
    @Override
    public void endElement(String uri, String loc, String raw) throws SAXException
    {
        if (!_isSuperDevMode)
        { 
            if (_isCurrentTagQueued)
            {
                _isCurrentTagQueued = false;
            }
            else
            {
                if (_filesQueue.size() > 0)
                {
                    _processQueue();
                }

                super.endElement(uri, loc, raw);
            }
        }
        else
        {
            super.endElement(uri, loc, raw);
        }
    }
    
    /**
     * Check if the current tag can be minimized by this transformer
     * @param tagName The tag name
     * @param attrs The attribute
     * @return True if minimizable
     */
    private boolean _isMinimizable(String tagName, Attributes attrs)
    {
        if ("true".equals(attrs.getValue("data-donotminimize")))
        {
            return false;
        }
        
        String uri = null;
        
        if ("script".equals(tagName))
        {
            String type = attrs.getValue("type");
            if (StringUtils.isNotEmpty(type) && !"text/javascript".equals(type) && !"application/javascript".equals(type))
            {
                // unsupported script type
                return false;
            }
            
            uri = attrs.getValue("src");
        }
        
        if ("link".equals(tagName))
        {
            String type = attrs.getValue("type");
            if (StringUtils.isNotEmpty(type) && !"text/css".equals(type))
            {
                // unsupported script type
                return false;
            }
            
            String rel = attrs.getValue("rel");
            if (!"stylesheet".equals(rel))
            {
                // unsupported relation type
                return false;
            }
            
            uri = attrs.getValue("href");
        }
        
        if (StringUtils.isNotEmpty(uri))
        {
            uri = FilenameUtils.normalize(uri, true);
            uri = _removeContextPath(uri);
            
            for (Pattern pattern : _patterns)
            {
                Matcher matcher = pattern.matcher(uri);
                if (matcher.find())
                {
                    return true;
                }
            }
        }
        
        return false;
    }
    
    /**
     * Remove the context path from the uri to be able to analyse correctly the patterns
     * @param uri The incomming uri
     * @return The uri with no context path
     */
    protected String _removeContextPath(String uri)
    {
        return StringUtils.prependIfMissing(StringUtils.removeStart(uri, _currentContextPath + "/"), "/");
    }

    /**
     * Add a new tag to the current queue. If the new tag is not compatible with the current queue, it is processed first and emptied before adding the new type of tag.
     * @param tagName The tag name
     * @param attrs The tag attribute
     * @throws SAXException If an error occurred
     */
    private void _addToQueue(String tagName, Attributes attrs) throws SAXException
    {
        Map<String, String> fileOptions = new HashMap<>();
        
        if (_filesQueue.size() > 0)
        {
            // Check if can be added to current queue. If not, process queue before adding.
            boolean sameAsCurrentQueue = tagName.equals(_queueTag);
            
            if (sameAsCurrentQueue && "link".equals(tagName))
            {
                String media = attrs.getValue("media");
                Set<String> medias = _toMediaList(media);
                
                if (_inlineCssMedias || _areAlwaysInlinableMedias(medias))
                {
                    // allow different medias in queue, but store the value as file option
                    fileOptions.put("media", StringUtils.join(medias, ","));
                }
                else
                {
                    // media must be the same as current queue
                    sameAsCurrentQueue = medias.equals(_queueMedia);
                }
            }
            
            if (!sameAsCurrentQueue)
            {
                _processQueue();
            }
        }
        
        String uri = FilenameUtils.normalize("script".equals(tagName) ? attrs.getValue("src") : attrs.getValue("href"), true);
        
        if (StringUtils.isNotEmpty(uri))
        {
            uri = FilenameUtils.normalize(uri, true);
            uri = _removeContextPath(uri);
            
            fileOptions.put("tag", tagName);
            _filesQueue.put(uri, fileOptions);
            if (_filesQueue.size() == 1)
            {
                // first item in queue, set the queue attributes for further additions compatibility checks.
                _queueTag = tagName;
                
                if (!_inlineCssMedias && "link".equals(tagName))
                {
                    String media = attrs.getValue("media");
                    _queueMedia = _toMediaList(media);
                }
            }
        }
    }
    
    private boolean _areAlwaysInlinableMedias(Set<String> medias)
    {
     // Since IE9, print can be inlined in other media css files
        return Collections.singleton("print").equals(medias) // The new file is print only 
                && (_queueMedia.contains("print") || _queueMedia.isEmpty()); // AND The current file has print included (empty media == all)
    }

    /**
     * Process the current files queue by saxing it as one file, which name is the hash of the sum of files. 
     * This hash is cached, to allow the retrieval of the real files from the hash value.
     * The files parameters, such as the media value of css files, are also stored in the cache.
     * @throws SAXException If an error occurred
     */
    private void _processQueue() throws SAXException
    {
        String hash = _hashCache.createHash(_filesQueue, getHashSalt());
        
        AttributesImpl attrs = new AttributesImpl();
        
        String minimizeUrl = _getMinimizeUrl();
        
        if ("script".equals(_queueTag))
        {
            attrs.addCDATAAttribute("type", "text/javascript");
            attrs.addCDATAAttribute("src", minimizeUrl + hash + ".js");
        }
        
        if ("link".equals(_queueTag))
        {
            attrs.addCDATAAttribute("type", "text/css");
            attrs.addCDATAAttribute("rel", "stylesheet");
            attrs.addCDATAAttribute("href", minimizeUrl + hash + ".css");
            
            if (!_inlineCssMedias && !_queueMedia.isEmpty())  
            {
                String medias = StringUtils.join(_queueMedia, ",");
                attrs.addCDATAAttribute("media", medias);
            }
        }
        
        super.startElement("", _queueTag, _queueTag, attrs);
        super.endElement("", _queueTag, _queueTag);
        
        _queueTag = null;
        _filesQueue.clear();
    }
    
    /**
     * Return the salt used by the hash cache
     * @return The salt string
     */
    protected String getHashSalt()
    {
        // hashes change at every application start
        return _startupTime + "#" + _locale;
    }
    
    private Set<String> _toMediaList(String media)
    {
        Set<String> medias = Arrays.asList(StringUtils.split(StringUtils.defaultString(media, "").toLowerCase(), ",")).stream()
            .map(String::trim)
            .sorted(Comparator.naturalOrder())
            .collect(Collectors.toSet());
        
        if (medias.contains("all") || medias.isEmpty() || __ALL_MEDIA.equals(medias))
        {
            return Collections.emptySet();
        }

        return medias;
    }
}
