/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.minimize;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.nio.charset.StandardCharsets;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.io.IOUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceNotFoundException;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.core.minimize.SourceMapCache;
import org.ametys.core.ui.Callable;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;
import org.ametys.runtime.workspace.WorkspaceMatcher;

import com.google.debugging.sourcemap.SourceMapConsumerFactory;
import com.google.debugging.sourcemap.SourceMapParseException;
import com.google.debugging.sourcemap.SourceMapping;
import com.google.debugging.sourcemap.proto.Mapping.OriginalMapping;

/**
 * Manager for minimized source map
 */
public class UnminimizeHelper extends AbstractLogEnabled implements Component, Serviceable, Contextualizable
{
    private static final String __MINIMIZED_RESOURCES_PREFIX = "/plugins/core-ui/resources-minimized/";
    
    /** The SourceMap cache */
    protected SourceMapCache _sourceMapCache;

    /** Avalon context */
    protected Context _context;
    
    /** SourceResolver */
    protected SourceResolver _resolver;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _sourceMapCache = (SourceMapCache) manager.lookup(SourceMapCache.ROLE);
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    /**
     * Undo the mapping of the stacktrace by providing the original filename, line and column
     * @param stacktraceData The list of stacktrace data, containing the filename, line and column of the minimized file
     * @return The list of data, complemented with the original file data when found
     * @throws SourceNotFoundException If an error occurred
     * @throws IOException If an error occurred
     * @throws SourceMapParseException If an error occurred
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public List<Map<String, String>> unminifyStackTrace(List<Map<String, String>> stacktraceData) throws SourceNotFoundException, IOException, SourceMapParseException
    {
        Map<String, String> resolvedCache = new HashMap<>();
        Request request = ContextHelper.getRequest(_context);
        String workspaceURI = (String) request.getAttribute(WorkspaceMatcher.WORKSPACE_URI);

        for (Map<String, String> data : stacktraceData)
        {
            if (data.containsKey("filename") && data.containsKey("column") && data.containsKey("line"))
            {
                String sourceMapContent = _getSourceMap(data.get("filename"), resolvedCache, workspaceURI);
                _setMappingData(data, sourceMapContent);
            }
        }
        
        return stacktraceData;
    }

    private String _getSourceMap(String filename, Map<String, String> resolvedCache, String workspaceURI) throws MalformedURLException, IOException, SourceNotFoundException
    {
        if (!filename.startsWith(workspaceURI + __MINIMIZED_RESOURCES_PREFIX))
        {
            return null;
        }
        String mapFilename = filename.substring((workspaceURI + __MINIMIZED_RESOURCES_PREFIX).length()) + ".map";
        
        if (!resolvedCache.containsKey(mapFilename))
        {
            Source source = _sourceMapCache.get(mapFilename);
            
            if (source == null)
            {
                // file was not resolved previously (client may have it in cache): Force loading to generate the map
                String uri = "cocoon:/" + __MINIMIZED_RESOURCES_PREFIX + mapFilename;
                source = _resolver.resolveURI(uri);
            }
            
            if (source != null)
            {
                InputStream is = source.getInputStream();
                StringWriter writer = new StringWriter();
                IOUtils.copy(is, writer, StandardCharsets.UTF_8);
                resolvedCache.put(mapFilename, writer.toString());
            }
        }
        
        if (resolvedCache.containsKey(mapFilename))
        {
            return resolvedCache.get(mapFilename);
        }
        
        return null;
    }
    
    private void _setMappingData(Map<String, String> data, String sourceMapContent) throws SourceMapParseException
    {
        if (sourceMapContent != null)
        {
            SourceMapping mapping = SourceMapConsumerFactory.parse(sourceMapContent);
            OriginalMapping mappingForEntry = mapping.getMappingForLine(Integer.parseInt(data.get("line")), Integer.parseInt(data.get("column")));
            if (mappingForEntry != null)
            {
                data.put("realFilename", mappingForEntry.getOriginalFile());
                data.put("realLine", String.valueOf(mappingForEntry.getLineNumber()));
                data.put("realColumn", String.valueOf(mappingForEntry.getColumnPosition()));
            }
        }
    }

}
