/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.parameter.files;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.TraversableSource;

import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.util.DateUtils;

/**
 * Abstract action for getting child files/folders of a local application folder
 */
public abstract class AbstractGetFilesAction extends ServiceableAction implements Contextualizable
{
    /** The cocoon context */
    protected org.apache.cocoon.environment.Context _cocoonContext;
    /** The source resolver */
    protected org.apache.excalibur.source.SourceResolver _srcResolver;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _srcResolver = (org.apache.excalibur.source.SourceResolver) smanager.lookup(org.apache.excalibur.source.SourceResolver.ROLE);
    }
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _cocoonContext = (org.apache.cocoon.environment.Context) context.get(org.apache.cocoon.Constants.CONTEXT_ENVIRONMENT_CONTEXT);
    }
    
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        String path = StringUtils.defaultString(request.getParameter("path"), "");
        
        String rootURI = getRootURI(request);
        TraversableSource rootDir = (TraversableSource) _srcResolver.resolveURI(rootURI);
        TraversableSource currentDir = (TraversableSource) _srcResolver.resolveURI(rootURI + (path.length() > 0 ? "/" + path : ""));
        
        List<Map<String, Object>> nodes = new ArrayList<>();
        
        if (currentDir.exists())
        {
            for (TraversableSource child : (Collection<TraversableSource>) currentDir.getChildren())
            {
                if (!isIgnoredSource(child))
                {
                    if (child.isCollection())
                    {
                        nodes.add(_collection2JsonObject(child, rootDir));
                    }
                    else
                    {
                        nodes.add(_resource2JsonObject(child, rootDir));
                    }
                }
            }
        }
        
        Map<String, Object> result = new HashMap<>();
        result.put("nodes", nodes);
        
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        return EMPTY_MAP;
    }
    
    /**
     * Get the URI of root directory such as 'context://WEB-INF/param'
     * @param request the request
     * @return The root URI
     */
    protected abstract String getRootURI (Request request);
    
    /**
     * Test if the source has to be ignored
     * @param source The source
     * @return true if the source has to be ignored
     */
    protected boolean isIgnoredSource (TraversableSource source)
    {
        return false;
    }
    
    /**
     * Convert collection to JSON object
     * @param folder the folder
     * @param root the root directory
     * @return JSON object
     */
    protected Map<String, Object> _collection2JsonObject (TraversableSource folder, TraversableSource root)
    {
        Map<String, Object> jsonObject = new HashMap<>();
        jsonObject.put("type", "collection");
        jsonObject.put("name", folder.getName());
        jsonObject.put("path", _getRelativePath(root, folder));
        return jsonObject;
    }
    
    /**
     * Convert file to JSON object
     * @param file the file
     * @param root the root directory
     * @return JSON object
     */
    protected Map<String, Object> _resource2JsonObject (TraversableSource file, TraversableSource root)
    {
        Map<String, Object> jsonObject = new HashMap<>();
        
        jsonObject.put("type", "resource");
        jsonObject.put("name", file.getName());
        jsonObject.put("path", _getRelativePath(root, file));
        
        jsonObject.put("size", file.getContentLength());
        jsonObject.put("lastModified", DateUtils.dateToString(new Date(file.getLastModified())));
        
        String mimeType = _cocoonContext.getMimeType(file.getName().toLowerCase());
        jsonObject.put("mimetype", mimeType != null ? mimeType : "application/unknown");
        
        return jsonObject;
    }
    
    /**
     * Get the relative path from root directory
     * @param root The root directory
     * @param file The file
     * @return The relative path
     */
    protected String _getRelativePath (TraversableSource root, TraversableSource file)
    {
        String relPath = file.getURI().substring(root.getURI().length());
        
        if (relPath.endsWith("/"))
        {
            relPath = relPath.substring(0, relPath.length() - 1);
        }
        
        return relPath;
    }

}
