/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.parameter.files;

import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.servlet.multipart.Part;
import org.apache.excalibur.source.impl.FileSource;

import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.file.FileHelper;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Create or update a file
 */
public class AddOrUpdateFile extends AbstractLogEnabled implements Serviceable, Component, Contextualizable
{
    /** Avalon Role */
    public static final String ROLE = AddOrUpdateFile.class.getName();
    
    /** The avalon context */
    protected Context _context;
    private org.apache.excalibur.source.SourceResolver _sResolver;
    private FileHelper _fileHelper;
    private CurrentUserProvider _currentUserProvider;
    private RightManager _rightManager;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _sResolver = (org.apache.excalibur.source.SourceResolver) smanager.lookup(org.apache.excalibur.source.SourceResolver.ROLE);
        _fileHelper = (FileHelper) smanager.lookup(FileHelper.ROLE);
        _currentUserProvider = (CurrentUserProvider) smanager.lookup(CurrentUserProvider.ROLE);
        _rightManager = (RightManager) smanager.lookup(RightManager.ROLE);
    }
    
    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }

    /**
     * Add, add and rename or update a file
     * @param filePath The file path
     * @param mode The mode
     * @param part The part to import
     * @return The results
     * @throws Exception If an error occurs
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> addOrUpdateFile(String filePath, String mode, Part part) throws Exception
    {
        Map<String, Object> result = new HashMap<>();
        
        Request request = ContextHelper.getRequest(_context);
        
        FileSource paramsDir = (FileSource) _sResolver.resolveURI("context://WEB-INF/param");
        FileSource parentDir = (FileSource) _sResolver.resolveURI("context://WEB-INF/param" + (filePath.length() > 0 ? "/" + filePath : ""));
        
        // Check user can access this feature
        _checkUserRight("CORE_Rights_EditParameterFile", "/${WorkspaceName}");
        
        result.putAll(_fileHelper.addOrUpdateFile(part, parentDir, mode, false));
        
        if (result.containsKey("uri"))
        {
            String fileUrl = (String) result.get("uri");
            String path = fileUrl.substring(paramsDir.getURI().length());
            result.put("path", path.endsWith("/") ? path.substring(0, path.length() - 1) : path);
            String parentPath = parentDir.getURI().substring(paramsDir.getURI().length());
            result.put("parentPath", parentPath.endsWith("/") ? parentPath.substring(0, parentPath.length() - 1) : parentPath);
        }
        
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        return result;
    }
    
    /**
     * Check if the user right to access the feature
     * @param rightId The right id
     * @param context The right context
     * @throws IllegalStateException if the user has no right
     */
    protected void _checkUserRight(String rightId, String context) throws IllegalStateException
    {
        UserIdentity user = _currentUserProvider.getUser();
        if (_rightManager.hasRight(user, rightId, context) != RightResult.RIGHT_ALLOW)
        {
            getLogger().error("User '" + user + "' tried to access a privileged feature without convenient right. Should have right '" + rightId + "' on context '" + context + "'");
            throw new IllegalStateException("You have no right to access this feature.");
        }
    }
}
