/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.parameter.files;

import java.io.IOException;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.TraversableSource;
import org.apache.excalibur.source.impl.FileSource;

import org.ametys.core.file.FileHelper;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.user.UserIdentity;

/**
 * Component for operations on parameters files and folders
 */
public class ParametersClientSideElement extends StaticClientSideElement implements Component, Initializable
{
    private static final String _ROOT_PARAMETERS_DIRECTORY_URI = "context://WEB-INF/param";
    
    private FileHelper _fileHelper;
    private SourceResolver _srcResolver;

    private String _baseResolvedURI;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _fileHelper = (FileHelper) serviceManager.lookup(FileHelper.ROLE);
        _srcResolver = (org.apache.excalibur.source.SourceResolver) serviceManager.lookup(org.apache.excalibur.source.SourceResolver.ROLE);        
    }
    
    public void initialize() throws Exception
    {
        _baseResolvedURI = _srcResolver.resolveURI(_ROOT_PARAMETERS_DIRECTORY_URI).getURI();
    }

    /**
     * Saves parameters file.
     * @param relPath the relative file's path from parameters files root directory
     * @param text the UTF-8 file content to save
     * @return The result map
     * @throws IOException If an error occurred while saving 
     */
    @Callable (rights = "CORE_Rights_EditParameterFile")
    public Map<String, Object> saveParameters(String relPath, String text) throws IOException 
    {
        String fileURI = _ROOT_PARAMETERS_DIRECTORY_URI + (relPath.length() > 0 ? "/" + relPath : "");
        return _fileHelper.saveFile(fileURI, text);
    }
    
    /**
     * Add a new folder
     * @param parentRelPath the relative parent file's path from parameters files root directory
     * @param name The name of folder to create
     * @return a map containing the name of the created folder, its path and the path of its parent
     * @throws IOException If an error occurred while creating folder
     */
    @Callable (rights = "CORE_Rights_EditParameterFile")
    public Map<String, Object> addFolder(String parentRelPath, String name) throws IOException 
    {
        FileSource rootDir = (FileSource) _srcResolver.resolveURI(_ROOT_PARAMETERS_DIRECTORY_URI);
        
        String parentURI = _ROOT_PARAMETERS_DIRECTORY_URI + (parentRelPath.length() > 0 ? "/" + parentRelPath : "");
        
        Map<String, Object> result = _fileHelper.addFolder(parentURI, name, true);
        
        if (result.containsKey("uri"))
        {
            String folderUri = (String) result.get("uri");
            String path = folderUri.substring(rootDir.getURI().length());
            result.put("path", path.endsWith("/") ? path.substring(0, path.length() - 1) : path);
            result.put("parentPath", parentRelPath.endsWith("/") ? parentRelPath.substring(0, parentRelPath.length() - 1) : parentRelPath);
        }

        return result;
    }
    
    /**
    * Remove a folder or a file
    * @param relPath the relative file's path from parameters files root directory
    * @return the result map
    * @throws IOException If an error occurs while removing the folder
    */
    @Callable (rights = "CORE_Rights_EditParameterFile")
    public Map<String, Object> deleteFile(String relPath) throws IOException
    {
        String fileUri = _ROOT_PARAMETERS_DIRECTORY_URI + (relPath.length() > 0 ? "/" + relPath : "");
        return _fileHelper.deleteFile(fileUri);
    }
    
    /**
    * Rename a file or a folder 
    * @param relPath the relative file's path from parameters files root directory
    * @param name the new name of the file/folder
    * @return the result map
    * @throws IOException if an error occurs while renaming the file/folder
    */
    @Callable (rights = "CORE_Rights_EditParameterFile")
    public Map<String, Object> renameFile(String relPath, String name) throws IOException 
    {
        FileSource rootDir = (FileSource) _srcResolver.resolveURI(_ROOT_PARAMETERS_DIRECTORY_URI);
        
        String fileUri = _ROOT_PARAMETERS_DIRECTORY_URI + (relPath.length() > 0 ? "/" + relPath : "");
        Map<String, Object> result = _fileHelper.renameFile(fileUri, name);
        
        if (result.containsKey("uri"))
        {
            String newURI = (String) result.get("uri");
            String path = newURI.substring(rootDir.getURI().length());
            result.put("path", path);
        }
        return result;
    }
    
    /**
     * Move a file or folder inside a given directory
     * 
     * @param srcRelPath The relative URI of file/folder to move
     * @param parentRelPath The URI relative of parent target file
     * @return a result map with the name and uri of moved file in case of
     *         success.
     * @throws IOException If an error occurred manipulating the source
     */
    @Callable (rights = "CORE_Rights_EditParameterFile")
    public Map<String, Object> moveParameterFile(String srcRelPath, String parentRelPath) throws IOException
    {
        String srcPath = _ROOT_PARAMETERS_DIRECTORY_URI + (srcRelPath.length() > 0 ? "/" + srcRelPath : "");
        String targetPath = _ROOT_PARAMETERS_DIRECTORY_URI + (parentRelPath.length() > 0 ? "/" + parentRelPath : "");
        
        Map<String, Object> result = _fileHelper.moveSource(srcPath, targetPath);
        
        if (result.containsKey("uri"))
        {
            String newURI = (String) result.get("uri");
            String path = newURI.substring(_ROOT_PARAMETERS_DIRECTORY_URI.length()  + 1);
            result.put("path", path);
        }
        
        return result;
    }
    
    /**
    * Check if a file with same name already exist
    * @param parentRelPath the path where the file will be added
    * @param name the name of the file 
    * @return result the server's response in JSON
    * @throws Exception if an error occurred
    */
    @Callable (rights = "CORE_Rights_EditParameterFile")
    public boolean fileExists(String parentRelPath, String name) throws Exception
    {
        String parentUri = _ROOT_PARAMETERS_DIRECTORY_URI + (parentRelPath.length() > 0 ? "/" + parentRelPath : "");
        return _fileHelper.hasChild(parentUri, name);
    }
    
    /**
     * Check if the user right to access the feature
     * @param rightId The right id
     * @param context The right context
     * @throws IllegalStateException if the user has no right
     */
    protected void _checkUserRight(String rightId, String context) throws IllegalStateException
    {
        UserIdentity user = _currentUserProvider.getUser();
        if (_rightManager.hasRight(user, rightId, context) != RightResult.RIGHT_ALLOW)
        {
            getLogger().error("User '" + user + "' tried to access a privileged feature without convenient right. Should have right '" + rightId + "' on context '" + context + "'");
            throw new IllegalStateException("You have no right to access this feature.");
        }
    }
    
    /**
     * Get the path of all the children that match the provided value.
     * The value will be search in filename and content for text file
     * @param value the value to search
     * @param path the parameter file path to use as root
     * @return the list of path that match
     */
    @Callable (rights = "CORE_Rights_EditParameterFile")
    public List<String> getFilteredPath(String value, String path)
    {
        try
        {
            TraversableSource currentSrc = (TraversableSource) _srcResolver.resolveURI(_ROOT_PARAMETERS_DIRECTORY_URI + (path.length() > 0 ? "/" + path : ""));
            
            return _fileHelper.filterSources(currentSrc, value).stream()
                .map(this::toRelativePath)
                .toList();
        }
        catch (IOException e)
        {
            getLogger().error("Failed to filter source at path '" + path + "'", e);
            return List.of();
        }
    }

    private String toRelativePath(String absoluteURI)
    {
        String path = absoluteURI.substring(_baseResolvedURI.length() - 1); // -1 to keep the head /
        if (path.endsWith("/"))
        {
            path = path.substring(0, path.length() - 1);
        }
        return path;
    }

}
