/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.resources.css;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.io.FilenameUtils;

/**
 * Helper for CSS Files
 */
public final class CSSFileHelper
{
    /** Regex that matches an @import statement in a css file, with optional media */
    public static final Pattern IMPORT_PATTERN = Pattern.compile("(?<=^|\n|;)" // Matches the start of the line or end of previous statement
                                                                + "[ \t]*@import[ \t]*" // Matches the '@import' mention
                                                                + "(?:url)?" // Optional match with 'url'
                                                                + "\\(?[ \t]*[\"']?" // Optional match with parenthesis, whitespace and/or quotes
                                                                + "([^)\"'\n]*)" // Captures a string that does not contains a parenthesis or quote, the import url
                                                                + "[\"']?\\)?[ \t]*" // Optional match with closing parenthesis, whitespace and/or quotes
                                                                + "([^; \t\r\n]*)[ \t]*" // Captures an optional string, which is the "media" part of the @import
                                                                + "(?:;|\n|$)?", // Matches optional end of statement or line 
                                                                Pattern.CASE_INSENSITIVE);
    
    // This regex matches 'src=' declarations
    private static final Pattern CSS_URL_PATTERN_SRC = Pattern.compile("src" // the string 'src' literally
                                                                     + "\\s*=\\s*" // the symbol equal, with any whitespace before or after
                                                                     + "['\"](.*?)['\"]", // capture a string in between quotes, singles or double
                                                                     Pattern.DOTALL | Pattern.CASE_INSENSITIVE);
    
    // This regex matches any 'url()' that is not preceded by an @import directive
    private static final Pattern CSS_URL_PATTERN_NOTIMPORT_URL = Pattern.compile("(?:@import\\s+url)" // The pattern '@import url' that we want to ignore. When matching this, group(1) will be null
                                                                               + "|" // The right side of this "or" is only evaluated when we don't match the previous pattern
                                                                               
                                                                               + "(?:"
                                                                               + "\\burl" // The word 'url'
                                                                               + "\\s*"
                                                                               + "\\(" // The opening parenthesis before the url
                                                                               + "\\s*['\"]?(.*?)['\"]?\\s*" // Capture a string that could be surrounded by whitespace and quotes
                                                                               + "\\)" // The closing parenthesis after the url
                                                                               + ")", 
                                                                               Pattern.DOTALL | Pattern.CASE_INSENSITIVE);
    
    // This regex matches any 'url()' that is preceded by an @import directive
    private static final Pattern CSS_URL_PATTERN_IMPORT_URL = Pattern.compile("@import\\s+url\\s*" // The literal '@import url' with any number of whitespace
                                                                            + "\\(" // The opening parenthesis of the 'url()'
                                                                            + "\\s*['\"]?(.*?)['\"]?\\s*" // Capture a string that could be surrounded by whitespace and quotes
                                                                            + "\\)", // The closing parenthesis of the 'url()'
                                                                            Pattern.DOTALL | Pattern.CASE_INSENSITIVE);
    
    // This regex matches '@import' directives without 'url()'
    private static final Pattern CSS_URL_PATTERN_IMPORT = Pattern.compile("@import\\s+" // The literal '@import' followed by whitespace
                                                                        + "['\"](.*?)['\"]", // Capture a string that could be surrounded by quotes
                                                                        Pattern.DOTALL | Pattern.CASE_INSENSITIVE);
    
    private CSSFileHelper()
    {
    }
    
    /**
     * Replace the relative URI inside a css file with the new context path
     * @param content The file content
     * @param fileUri The file Uri
     * @param jsassResourceURIExtensionPoint The JSASS Resource URI extension point
     * @param internalContextPath The internal context path of the application
     * @param externalContextPath The external context path of the application
     * @return The file content, with URIs replaced
     * @throws URISyntaxException If an exception occurred
     */
    public static String replaceRelativeUri(String content, String fileUri, JSASSResourceURIExtensionPoint jsassResourceURIExtensionPoint, String internalContextPath, String externalContextPath) throws URISyntaxException
    {
        String c1 = _replaceRelativeUri(content, externalContextPath, fileUri, CSS_URL_PATTERN_SRC, jsassResourceURIExtensionPoint);
        String c2 = _replaceRelativeUri(c1, externalContextPath, fileUri, CSS_URL_PATTERN_NOTIMPORT_URL, jsassResourceURIExtensionPoint);
        String c3 = _replaceRelativeUri(c2, internalContextPath, fileUri, CSS_URL_PATTERN_IMPORT_URL, jsassResourceURIExtensionPoint);
        String result = _replaceRelativeUri(c3, internalContextPath, fileUri, CSS_URL_PATTERN_IMPORT, jsassResourceURIExtensionPoint);

        return result;
    }
    
    /**
     * Replace the relative URI inside a css file with the new context path
     * @param content The file content
     * @param fileUri The file Uri, without context path, for example "/plugins/pluginName/resources/style.css" or "/skins/skinName/resources/style.scss"
     * @param jsassResourceURIExtensionPoint The JSASS Resource URI extension point
     * @param externalContextPath The external context path of the application
     * @return The file content, with URIs replaced
     * @throws URISyntaxException If an exception occurred
     */
    public static String replaceRelativeResourcesUri(String content, String fileUri, JSASSResourceURIExtensionPoint jsassResourceURIExtensionPoint, String externalContextPath) throws URISyntaxException
    {
        String c1 = _replaceRelativeUri(content, externalContextPath, fileUri, CSS_URL_PATTERN_SRC, jsassResourceURIExtensionPoint);
        String result = _replaceRelativeUri(c1, externalContextPath, fileUri, CSS_URL_PATTERN_NOTIMPORT_URL, jsassResourceURIExtensionPoint);

        return result;
    }
    
    private static String _replaceRelativeUri(String content, String contextPath, String fileUri, Pattern pattern, JSASSResourceURIExtensionPoint jsassResourceURIExtensionPoint)
    {
        Matcher urlMatcher = pattern.matcher(content);
        StringBuffer sb = new StringBuffer();
        
        while (urlMatcher.find()) 
        {
            String fullMatch = urlMatcher.group();
            String cssUrl = urlMatcher.group(1);
            
            if (cssUrl != null && !cssUrl.startsWith("data:"))
            {
                URI uri;
                try
                {
                    uri = new URI(cssUrl.replaceAll("\\|", "%7C"));
                    
                    if (!uri.isAbsolute() && cssUrl.indexOf("/") != 0  && cssUrl.indexOf("#") != 0)
                    {
                        String fullFileUri = fileUri.indexOf('/') == 0 ? contextPath + fileUri : fileUri;
                        String fullPathUrl = new URI(FilenameUtils.getFullPath(fullFileUri) + cssUrl).normalize().toString();
                        urlMatcher.appendReplacement(sb, Matcher.quoteReplacement(fullMatch.replace(cssUrl, fullPathUrl)));
                    }
                    else
                    {
                        String relativeUri = jsassResourceURIExtensionPoint.resolve(cssUrl);
                        if (relativeUri != null)
                        {
                            urlMatcher.appendReplacement(sb, Matcher.quoteReplacement(fullMatch.replace(cssUrl, relativeUri)));
                        }
                    }
                }
                catch (URISyntaxException e)
                {
                    urlMatcher.appendTail(sb);
                    return sb.toString();
                }
            }
        }
        
        urlMatcher.appendTail(sb);
        
        return sb.toString();
    }

}
