/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.resources.css;

import java.io.IOException;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.core.resources.AbstractResourceHandlerProvider;
import org.ametys.core.resources.CssResourceHandler;
import org.ametys.core.resources.ResourceHandler;
import org.ametys.core.resources.ResourceHandlerProvider;
import org.ametys.plugins.core.ui.resources.css.less.LessResourceHandler;
import org.ametys.plugins.core.ui.resources.css.sass.SassResourceHandler;

/**
 * {@link ResourceHandlerProvider} for compiled CSS files.
 */
public class CompiledCssResourceHandlerProvider extends AbstractResourceHandlerProvider
{
    private static final String[] __LESS_EXTENSION = new String[] {".less"};
    private static final String[] __SASS_EXTENSION = new String[] {".scss", ".sass"};
    
    private SourceResolver _resolver;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    public int getPriority()
    {
        return MIN_PRIORITY + 1000;
    }
    
    @Override
    public ResourceHandler getResourceHandler(String source) throws Exception
    {
        String lcSource = StringUtils.lowerCase(source);
        if (!StringUtils.endsWith(lcSource, ".css"))
        {
            return null;
        }

        // If the requested source exists, fallback to default CSS handler
        Source src = _resolver.resolveURI(source);
        if (src.exists())
        {
            return setup(new CssResourceHandler(src));
        }
        
        _resolver.release(src);
        
        String sourceWithoutExt = StringUtils.substringBeforeLast(source, ".css");
        
        ResourceHandler handler = _getSassResourceHandler(sourceWithoutExt);
        if (handler != null)
        {
            return handler;
        }
        
        handler = _getLessResourceHandler(sourceWithoutExt);
        if (handler != null)
        {
            return handler;
        }
        
        return null;
    }
    
    private ResourceHandler _getLessResourceHandler(String sourceWithoutExt) throws Exception
    {
        for (String ext : __LESS_EXTENSION)
        {
            try
            {
                Source src = _resolver.resolveURI(sourceWithoutExt + ext);
                if (src.exists())
                {
                    return setup(new LessResourceHandler(src));
                }
                else
                {
                    _resolver.release(src);
                }
            }
            catch (IOException e)
            {
                // Nothing
            }
        }
        
        return null;
    }
    
    private ResourceHandler _getSassResourceHandler(String sourceWithoutExt) throws Exception
    {
        for (String ext : __SASS_EXTENSION)
        {
            Source src = null;
            try
            {
                String sassLocation = StringUtils.removeEnd(sourceWithoutExt, ".min") + ext;
                src = _resolver.resolveURI(sassLocation);
                if (src.exists())
                {
                    return setup(new SassResourceHandler(src, sassLocation));
                }
                else
                {
                    _resolver.release(src);
                }
            }
            catch (IOException e)
            {
                // Nothing
            }
        }
        
        return null;
    }
}
