/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.resources.css;

import java.util.Arrays;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.util.LambdaUtils;

/**
 * Default configurable implementation for {@link SupportedContextResourceUri}. Can be configured to map contexts with URIs
 */
public class ConfigurableSupportedContextResourceUri implements SupportedContextResourceUri, Configurable
{
    private Map<String, String> _supportedContext;

    public void configure(Configuration configuration) throws ConfigurationException
    {
        Function<Configuration, String> confToContextAttribute = LambdaUtils.wrap(conf -> "context:/" + conf.getAttribute("context"));
        Function<Configuration, String> confToValue = LambdaUtils.wrap(conf -> conf.getAttribute("url"));
        
        _supportedContext = Arrays.stream(configuration.getChildren("supportedContext"))
                .filter(conf -> StringUtils.isNotEmpty(conf.getAttribute("context", null)))
                .filter(conf -> StringUtils.isNotEmpty(conf.getAttribute("url", null)))
                .collect(Collectors.toMap(confToContextAttribute, confToValue));
    }
    
    @Override
    public String resolve(String uri)
    {
        if (StringUtils.isEmpty(uri))
        {
            return null;
        }
        
        return _supportedContext.entrySet().stream()
                .filter(supported -> uri.startsWith(supported.getKey()))
                .findFirst()
                .map(supported -> uri.replace(supported.getKey(), supported.getValue()))
                .orElse(null);
    }
    
    @Override
    public String resolvePath(String path)
    {
        if (StringUtils.isEmpty(path))
        {
            return null;
        }
        
        return _supportedContext.entrySet().stream()
                .filter(supported -> path.startsWith(supported.getValue()))
                .findFirst()
                .map(supported -> path.replace(supported.getValue(), supported.getKey()))
                .orElse(null);
    }
}
