/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.resources.css;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.core.resources.ProxiedContextPathProvider;
import org.ametys.core.resources.ResourceReader;
import org.ametys.plugins.core.ui.minimize.HashCache.UriData;
import org.ametys.plugins.core.ui.resources.ResourceDependenciesList;
import org.ametys.plugins.core.ui.util.RequestAttributesHelper;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Dependencies list for css files
 */
public class CssDependenciesList extends AbstractLogEnabled implements ResourceDependenciesList, Serviceable
{
    private static final Pattern __EXTERNAL_URL = Pattern.compile("^(http[s]?://[^/]+)(/.*)?$");

    /** RequestAttributesHelper */
    protected RequestAttributesHelper _requestAttributesHelper;

    /* Dependencies cache */
    private Map<String, Pair<List<String>, Long>> _dependenciesCache = new ConcurrentHashMap<>();

    private SourceResolver _sourceResolver;
    
    private ProxiedContextPathProvider _proxiedContextPathProvider;

    public void service(ServiceManager manager) throws ServiceException
    {
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _requestAttributesHelper = (RequestAttributesHelper) manager.lookup(RequestAttributesHelper.ROLE);
        _proxiedContextPathProvider = (ProxiedContextPathProvider) manager.lookup(ProxiedContextPathProvider.ROLE);
    }
    
    @Override
    public boolean isSupported(String uri)
    {
        return StringUtils.endsWith(uri, ".css");
    }

    @Override
    public int getPriority()
    {
        return 100;
    }

    public Set<UriData> getDependenciesList(String uri, Map<String, String> data, boolean onlyFirstLevel)
    {
        return _getFileDependencies(uri, data.getOrDefault("media", null), true, onlyFirstLevel);
    }
    
    private Set<UriData> _getFileDependencies(String cssUri, String media, boolean firstLevel, boolean onlyFirstLevel) throws IllegalArgumentException
    {
        Set<UriData> dependencies = new LinkedHashSet<>();
        
        Source fileSource = null;
        Map<String, Object> resolveParameters = new HashMap<>();
        
        try
        {
            fileSource = _getFileSource(cssUri, resolveParameters);
            
            long fileLastModified = resolveParameters.get(ResourceReader.LAST_MODIFIED) != null ? (long) resolveParameters.get("lastModified") : -1;
             
            UriData fileInfos = new UriData(cssUri, firstLevel);
            fileInfos.setLastModified(fileLastModified);
            fileInfos.setMedia(media);
            dependencies.add(fileInfos);
            
            if (onlyFirstLevel)
            {
                return dependencies;
            }
            
            Long validity = Optional.ofNullable(_dependenciesCache.get(cssUri))
                    .map(Pair::getRight)
                    .orElse(null);
            
            if (validity == null || validity != fileLastModified)
            {
                // cache is outdated
                List<String> dependenciesCache = new ArrayList<>();
                
                for (UriData cssDependency : _getCssFileDependencies(cssUri, fileSource))
                {
                    if (!dependenciesCache.contains(cssDependency.getUri()))
                    {
                        dependenciesCache.add(cssDependency.getUri());
                        dependencies.add(cssDependency);
                    }
                }
                
                _dependenciesCache.put(cssUri, Pair.of(dependenciesCache, fileLastModified));
            }
            else
            {
                // cache is up to date
                Pair<List<String>, Long> dependenciesFromCache = _dependenciesCache.get(cssUri);
                if (dependenciesFromCache != null)
                {
                    for (String dependencyCached : dependenciesFromCache.getLeft())
                    {
                        dependencies.addAll(_getFileDependencies(dependencyCached, null, false, false));
                    }
                }
            }
        }
        finally 
        {
            _sourceResolver.release(fileSource);
        }
        
        return dependencies;
    }

    private Source _getFileSource(String cssUri, Map<String, Object> resolveParameters)
    {
        try
        {
            String uriToResolve = cssUri;
            URI uri = new URI(cssUri);
            if (!uri.isAbsolute())
            {
                uriToResolve = "cocoon:/" + uriToResolve;
            }
            
            _requestAttributesHelper.removeRequestAttributes();
            
            return _sourceResolver.resolveURI(uriToResolve, null, resolveParameters);
        }
        catch (Exception e)
        {
            throw new IllegalArgumentException("Unable to resolve the dependencies of specified uri", e);
        }
    }
    
    private List<UriData> _getCssFileDependencies(String uri, Source cssSource) throws IllegalArgumentException
    {
        List<UriData> cssDependencies = new ArrayList<>();
        
        String fileContent;
        try (InputStream is = cssSource.getInputStream())
        {
            fileContent = IOUtils.toString(is, StandardCharsets.UTF_8);
        }
        catch (IOException e)
        {
            getLogger().error("Unable to retrieve css file dependencies for '" + uri + "'", e);
            return Collections.EMPTY_LIST;
        }
        
        String contextPath = _proxiedContextPathProvider.getContextPath();
        
        Matcher urlMatcher = CSSFileHelper.IMPORT_PATTERN.matcher(fileContent);
        
        while (urlMatcher.find()) 
        {
            String cssUrl = urlMatcher.group(1);
            
            Matcher externalMatcher = __EXTERNAL_URL.matcher(cssUrl);
            
            if (!externalMatcher.find())
            {
                try
                {
                    cssUrl = StringUtils.removeStart(cssUrl, contextPath);
                    URI cssUri = new URI(cssUrl);
                    if (!cssUri.isAbsolute() && !cssUrl.startsWith("/"))
                    {
                        cssUri = new URI(FilenameUtils.getFullPath(uri) + cssUrl);
                    }
                    cssDependencies.addAll(_getFileDependencies(cssUri.normalize().toString(), null, false, false));
                }
                catch (URISyntaxException e)
                {
                    // Invalid URI inside a file, but should not be blocking
                    getLogger().warn("Invalid URI in a file, could not calculate dependancies for file : " + uri , e);
                }
            }
        }
        
        return cssDependencies;
    }
}
