/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.resources.css.sass;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.plugins.core.ui.minimize.HashCache.UriData;
import org.ametys.plugins.core.ui.resources.ResourceDependenciesList;
import org.ametys.plugins.core.ui.resources.css.sass.SassImportHelper.SassImportInfo;
import org.ametys.plugins.core.ui.util.RequestAttributesHelper;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Dependencies list for sass files compiled into css
 * Uncompiled sass files don't depend on anything and are managed by the DefaultResourceDependenciesList
 */
public class SassDependenciesList extends AbstractLogEnabled implements ResourceDependenciesList, Serviceable
{
    /** List of sass extensions */
    protected static final String[] __SASS_EXTENSION = new String[] {".scss", ".sass"};
    
    /** RequestAttributesHelper */
    protected RequestAttributesHelper _requestAttributesHelper;

    /** The source resolver */
    protected SourceResolver _sourceResolver;

    /** The Sass import helper */
    protected SassImportHelper _sassImportHelper;

    /* Dependencies cache */
    private Map<String, Pair<List<String>, Long>> _dependenciesCache = new ConcurrentHashMap<>();

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _requestAttributesHelper = (RequestAttributesHelper) manager.lookup(RequestAttributesHelper.ROLE);
        _sassImportHelper = (SassImportHelper) manager.lookup(SassImportHelper.ROLE);
    }
    
    @Override
    public boolean isSupported(String uri)
    {
        if (StringUtils.endsWith(uri, ".css"))
        {
            return getSassImportInfo(uri) != null;
        }

        return false;
    }

    @Override
    public int getPriority()
    {
        return 200;
    }

    @Override
    public Set<UriData> getDependenciesList(String uri, Map<String, String> data, boolean onlyFirstLevel)
    {
        SassImportInfo fileSource = null;
        
        try
        {
            fileSource = getSassImportInfo(uri);
            
            if (fileSource != null)
            {
                String media = data.getOrDefault("media", null);
                return getDependenciesList(uri, media, fileSource.getSource(), fileSource.getLastModified(), true, onlyFirstLevel);
            }
        }
        finally 
        {
            if (fileSource != null)
            {
                _sourceResolver.release(fileSource.getSource());
            }
        }
        
        return new HashSet<>();
    }
    
    /**
     * Get the Sass import informations from the URI
     * @param uri The URI
     * @return The information
     */
    protected SassImportInfo getSassImportInfo(String uri)
    {
        String uriWithoutExtension = StringUtils.removeEnd(uri, ".css");
        return _sassImportHelper.findExistingImportSource(Arrays.stream(__SASS_EXTENSION).map(ext -> uriWithoutExtension + ext).collect(Collectors.toList()));
    }

    private Set<UriData> getDependenciesList(SassImportInfo sassImportInfo, String media)
    {
        return getDependenciesList(sassImportInfo, media, false);
    }

    private Set<UriData> getDependenciesList(SassImportInfo sassImportInfo, String media, boolean firstLevel)
    {
        return getDependenciesList(sassImportInfo.getUri(), media, sassImportInfo.getSource(), sassImportInfo.getLastModified(), firstLevel, false);
    }
    
    private Set<UriData> getDependenciesList(String uri, String media, Source source, long lastModified, boolean firstLevel, boolean onlyFirstLevel)
    {
        Set<UriData> dependencies = new LinkedHashSet<>();
        UriData fileInfos = new UriData(uri, firstLevel);
        fileInfos.setLastModified(lastModified);
        fileInfos.setMedia(media);
        dependencies.add(fileInfos);
        
        if (onlyFirstLevel)
        {
            return dependencies;
        }
        
        Pair<List<String>, Long> dependencyFromCache = _dependenciesCache.get(uri);

        if (dependencyFromCache == null || dependencyFromCache.getRight() != lastModified)
        {
            // cache is outdated
            List<String> dependenciesCache = new ArrayList<>();
            
            Consumer<UriData> addDependencyToCache = cssDependency -> 
            {
                if (!dependenciesCache.contains(cssDependency.getUri()))
                {
                    dependenciesCache.add(cssDependency.getUri());
                    dependencies.add(cssDependency);
                }
            };
            
            _sassImportHelper.getDependenciesList(source).values()
                    .stream()
                    .map(sassImport -> getDependenciesList(sassImport, media))
                    .flatMap(Set::stream)
                    .forEach(addDependencyToCache);
            
            _dependenciesCache.put(uri, Pair.of(dependenciesCache, lastModified));
        }
        else
        {
            // cache is up to date, lookup dependencies of each entry from the cache to build the full list
            dependencyFromCache.getLeft()
                    .stream()
                    .map(uriFromCache -> _sassImportHelper.getImportSource(uriFromCache))
                    .filter(Objects::nonNull)
                    .map(sassImport -> getDependenciesList(sassImport, media))
                    .forEach(subDependencies -> dependencies.addAll(subDependencies));
        }
        
        return dependencies;
    }
}
