/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.resources.css.sass;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.regex.Matcher;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.TraversableSource;

import org.ametys.core.resources.ResourceReader;
import org.ametys.plugins.core.ui.resources.css.CSSFileHelper;
import org.ametys.plugins.core.ui.resources.css.JSASSResourceURIExtensionPoint;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Helper component for sass contents
 */
public class SassImportHelper extends AbstractLogEnabled implements Serviceable, Component
{
    /** The avalon ROLE */
    public static final String ROLE = SassImportHelper.class.getName();
    
    /** The source resolver */
    protected SourceResolver _resolver;

    /** JsassResourceURIExtensionPoint */
    protected JSASSResourceURIExtensionPoint _jsassResourceURIExtensionPoint;


    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _jsassResourceURIExtensionPoint = (JSASSResourceURIExtensionPoint) manager.lookup(JSASSResourceURIExtensionPoint.ROLE);
    }
    
    
    /**
     * Get the list of direct dependencies of a sass source
     * @param inputSource The sass source
     * @return The list of dependencies, mapped by URI
     */
    public Map<String, SassImportInfo> getDependenciesList(Source inputSource)
    {
        Map<String, SassImportInfo> result = new HashMap<>();
        
        try (InputStream is = inputSource.getInputStream())
        {
            String previous = inputSource.getURI();
            String content = IOUtils.toString(is, StandardCharsets.UTF_8);
            
            Matcher matcher = CSSFileHelper.IMPORT_PATTERN.matcher(content);
            
            while (matcher.find())
            {
                String cssUrl = matcher.group(1);
                
                if (!StringUtils.contains(cssUrl, "http://") && !StringUtils.contains(cssUrl, "https://")) 
                {
                    URI currentUri = new URI(cssUrl);
                    if (!currentUri.isAbsolute())
                    {
                        currentUri = new URI(FilenameUtils.getFullPath(previous.toString()) + cssUrl);
                    }
            
                    SassImportInfo importSource = getImportSource(currentUri.toString());
                    if (importSource != null)
                    {
                        result.put(importSource.getUri(), importSource);
                    }
                }
            }
        }
        catch (IOException | URISyntaxException e)
        {
            getLogger().warn("Invalid " + inputSource.getURI(), e);
        }
        
        return result;
    }
    

    /**
     * Get the Sass source from the current Uri
     * @param currentUri The URI
     * @return The real URI, the Sass source and its last modified
     */
    public SassImportInfo getImportSource(String currentUri)
    {
        List<String> uriMatching = new ArrayList<>();
        uriMatching.add(currentUri);
        
        // extension is optional
        uriMatching.add(currentUri + ".scss");
        uriMatching.add(currentUri + ".sass");
        uriMatching.add(currentUri + ".css");
        
        // add an underscore prefix to the file name for sass partial imports
        String name = FilenameUtils.getName(currentUri);
        String partialUri = currentUri.substring(0, currentUri.length() - name.length()) + "_" + name;
        uriMatching.add(partialUri);
        uriMatching.add(partialUri + ".scss");
        uriMatching.add(partialUri + ".sass");
        uriMatching.add(partialUri + ".css");
        
        return findExistingImportSource(uriMatching);
    }

    /**
     * Find the first existing import source from a list of URIs to try
     * @param uriMatching The list of URIs to try
     * @return the first matching source, or null if no source matches
     */
    public SassImportInfo findExistingImportSource(List<String> uriMatching)
    {
        for (String uri : uriMatching)
        {
            try
            {
                Map<String, Object> resolveParameters = new HashMap<>();
                String absoluteUri = new URI(uri).isAbsolute() ? uri : _jsassResourceURIExtensionPoint.resolvePath(uri);
                Source importSource = _resolver.resolveURI(absoluteUri, null, resolveParameters);
                if (importSource.exists() && !(importSource instanceof TraversableSource && ((TraversableSource) importSource).isCollection()))
                {
                    SassImportInfo sassImportInfo = new SassImportInfo(uri, importSource);
                    sassImportInfo.setLastModified(resolveParameters.containsKey(ResourceReader.LAST_MODIFIED) ? (long) resolveParameters.get("lastModified") : importSource.getLastModified());
                    return sassImportInfo;
                }
            }
            catch (IOException | URISyntaxException e)
            {
                // source does not exists. Do nothing
            }
        }
        
        return null;
    }
    
    /**
     * Informations about sass import, such as real URI and real last modified
     */
    public static class SassImportInfo
    {
        private String _uri;
        private Source _source;
        private Long _lastModified;
        
        /**
         * Default constructor for the sass import info
         * @param uri The import real URI
         * @param source The import source
         */
        SassImportInfo(String uri, Source source)
        {
            _uri = uri;
            _source = source;
        }
        
        /**
         * Get the real uri of the import, with the sass extension
         * @return The real uri
         */
        public String getUri()
        {
            return _uri;
        }
        
        /**
         * Get the source of the import
         * @return The source
         */
        public Source getSource()
        {
            return _source;
        }
        
        /**
         * Set the last modified time of the import
         * @param lastModified The last modified
         */
        public void setLastModified(Long lastModified)
        {
            _lastModified = lastModified;
        }
        
        /**
         * Get the last modified of the import
         * @return The last modified
         */
        public long getLastModified()
        {
            return Optional.ofNullable(_lastModified).orElse(_source.getLastModified());
        }
    }

}
