/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.plugins.core.ui.resources.css.sass;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;

import org.ametys.core.minimize.css.sass.MinimizeSassManager;
import org.ametys.core.resources.ProxiedContextPathProvider;
import org.ametys.plugins.core.ui.resources.AbstractCompiledResourceHandler;
import org.ametys.plugins.core.ui.resources.ResourceDependenciesListExtensionPoint;

import io.bit3.jsass.CompilationException;

/**
 * Reader for SASS files, compile them on the fly into CSS files, and minimize them if location in .min.css
 */
public class SassResourceHandler extends AbstractCompiledResourceHandler
{
    /** ResourceDependenciesListExtensionPoint */
    protected ResourceDependenciesListExtensionPoint _resourceDependenciesListEP;

    /** CSS minimize manager */
    protected MinimizeSassManager _sassMinimizeManager;
    
    /** Sass import helper */
    protected SassImportHelper _sassImportHelper;
    
    /** Proxied context path provider */
    protected ProxiedContextPathProvider _proxiedContextPathProvider;
    
    private boolean _minimize;
    private String _sassLocation;

    /**
     * Constructor with an already resolved {@link Source}.
     * @param source the source
     * @param sassLocation the uri of the sass source
     */
    public SassResourceHandler(Source source, String sassLocation)
    {
        super(source);
        _sassLocation = sassLocation;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        
        _sassMinimizeManager = (MinimizeSassManager) manager.lookup(MinimizeSassManager.ROLE);
        _resourceDependenciesListEP = (ResourceDependenciesListExtensionPoint) manager.lookup(ResourceDependenciesListExtensionPoint.ROLE);
        _sassImportHelper = (SassImportHelper) manager.lookup(SassImportHelper.ROLE);
        _proxiedContextPathProvider = (ProxiedContextPathProvider) manager.lookup(ProxiedContextPathProvider.ROLE);
    }
    
    @Override
    public Source setup(String location, Map objectModel, Parameters par, boolean readForDownload) throws ProcessingException, IOException
    {
        Source source = super.setup(location, objectModel, par, readForDownload);
        
        _minimize = StringUtils.endsWith(location, ".min.css");
        
        return source;
    }

    @Override
    public void generate(OutputStream out) throws IOException, ProcessingException
    {
        try (InputStream is = _source.getInputStream())
        {
            String sassContent = IOUtils.toString(is, StandardCharsets.UTF_8);

            String result = _sassMinimizeManager.compileCss(sassContent, _sassLocation, _requestedLocation, _minimize, _source.getLastModified());
            IOUtils.write(result, out, StandardCharsets.UTF_8);
        }
        catch (CompilationException | URISyntaxException e)
        {
            throw new ProcessingException("Unable to compile the SASS file: " + _source.getURI(), e);
        }
    }
    
    @Override
    protected List<String> getDependenciesList(Source inputSource)
    {
        return new ArrayList<>(_sassImportHelper.getDependenciesList(inputSource).keySet());
    }
    
    @Override
    public Serializable getKey()
    {
        return super.getKey() + "*" + (_minimize ? "min" : "") + "*" + (_proxiedContextPathProvider.getContextPath());
    }

    @Override
    public String getMimeType()
    {
        return "text/css";
    }
}
