/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.resources.css.sass;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.Map;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.components.ContextHelper;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;

import org.ametys.core.minimize.SourceMapCache;
import org.ametys.core.minimize.css.sass.MinimizeSassManager;
import org.ametys.core.resources.ProxiedContextPathProvider;
import org.ametys.core.resources.ResourceHandlerProvider;
import org.ametys.core.resources.SimpleResourceHandler;
import org.ametys.plugins.core.ui.resources.css.JSASSResourceURIExtensionPoint;

import io.bit3.jsass.CompilationException;

/**
 * Resource handler for source map of compiled sass files
 */
public class SassSourceMapResourceHandler extends SimpleResourceHandler implements Contextualizable
{
    /** sassMinimizeManager */
    protected MinimizeSassManager _sassMinimizeManager;
    
    private boolean _minimize;
    private String _sourceMapLocation;

    private SourceMapCache _sourceMapCache;
    private JSASSResourceURIExtensionPoint _jsassResourceURIExtensionPoint;
    private ProxiedContextPathProvider _proxiedContextPathProvider;
    
    private Context _context;
    
    /**
     * Called
     */
    public SassSourceMapResourceHandler()
    {
        super();
    }
    
    /**
     * If the {@link Source} is already resolved by the {@link ResourceHandlerProvider}, 
     * it may provide it through the constructor to avoid resolving it again.
     * @param source the source.
     */
    public SassSourceMapResourceHandler(Source source)
    {
        super(source);
    }

    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);

        _sourceMapCache = (SourceMapCache) manager.lookup(SourceMapCache.ROLE);
        _sassMinimizeManager = (MinimizeSassManager) manager.lookup(MinimizeSassManager.ROLE);
        _jsassResourceURIExtensionPoint = (JSASSResourceURIExtensionPoint) manager.lookup(JSASSResourceURIExtensionPoint.ROLE);
        _proxiedContextPathProvider = (ProxiedContextPathProvider) manager.lookup(ProxiedContextPathProvider.ROLE);
    }
    
    @Override
    public Source setup(String location, Map objectModel, Parameters parameters, boolean readForDownload) throws IOException, ProcessingException
    {
        _sourceMapLocation = location;
        
        // example: plugin:core-ui://resources/example/test.scss or plugin:core-ui://resources/example/test.css
        String originalSassLocation = StringUtils.removeEnd(location, ".map");

        return super.setup(originalSassLocation, objectModel, parameters, readForDownload);
    }
    
    @Override
    public Serializable getKey()
    {
        return super.getKey() + ".map" + (_minimize ? "*min" : "");
    }
    
    @Override
    public void generate(OutputStream out) throws IOException, ProcessingException
    {
        // Check if the source map is in cache
        
        // example: plugin:core-ui://resources/example/test.scss.map or plugin:core-ui://resources/example/test.css.map
        Source cachedSourceMap = null;
        try
        {
            // example: /plugins/core-ui/resources/example/test.scss.map or /plugins/core-ui/resources/example/test.css.map
            String sourceMapCacheKey = _jsassResourceURIExtensionPoint.resolve(_sourceMapLocation);
            cachedSourceMap = _sourceMapCache.get(sourceMapCacheKey);
        }
        catch (URISyntaxException e)
        {
            getLogger().error("Unable to resolve URI location because URI syntax '" + _sourceMapLocation + "' is not supported", e);
        }
        
        if (cachedSourceMap != null)
        {
            try (InputStream is = cachedSourceMap.getInputStream())
            {
                IOUtils.copy(is, out);
                return;
            }
        }
        
        // Generate the source map
        try (InputStream is = _source.getInputStream())
        {
            String sassContent = IOUtils.toString(is, StandardCharsets.UTF_8);
            String extension = FilenameUtils.getExtension(_source.getURI());
            String sourceMap = _sassMinimizeManager.generateCssSourceMap(sassContent, _sourceMapLocation, extension, ContextHelper.getRequest(_context).getContextPath(), _proxiedContextPathProvider.getContextPath(), _minimize, _source.getLastModified());
            IOUtils.write(sourceMap, out, "UTF-8");
        }
        catch (CompilationException | URISyntaxException e)
        {
            throw new ProcessingException("Unable to compile the SASS file: " + _source.getURI(), e);
        }
    }
}
