/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.plugins.core.ui.resources.vuejs;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.Response;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.RegExUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceException;
import org.apache.excalibur.source.TraversableSource;
import org.apache.excalibur.source.impl.FileSource;

import org.ametys.core.resources.ProxiedContextPathProvider;
import org.ametys.plugins.core.ui.resources.AbstractCompiledResourceHandler;

/**
 * Resource handler to compile any VueJS resource on the fly if needed, or serve it
 * The sources have to be located in a directory X/vuejs, while the resources will be thought at X/resources/vuejs 
 */
public class VueJsResourceHandler extends AbstractCompiledResourceHandler implements Contextualizable
{
    private Context _context;
    private LocationParser _lp;
    private ProxiedContextPathProvider _proxiedContextPathProvider;

    /**
     * Constructor with an already resolved {@link Source}.
     * @param source the source
     */
    public VueJsResourceHandler(Source source)
    {
        super(source);
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _proxiedContextPathProvider = (ProxiedContextPathProvider) manager.lookup(ProxiedContextPathProvider.ROLE);
    }

    @Override
    public Source setup(String rawLocation, Map objectModel, Parameters parameters, boolean readForDownload) throws IOException, ProcessingException
    {
        _requestedLocation = rawLocation;
        _objectModel = objectModel;
        _parameters = parameters;
        _readForDownload = readForDownload;
        
        // css files are already minimized files
        String location = rawLocation.replaceAll("\\.min\\.css$", ".css");
        
        // In this implementation, it is not: one source => one target file
        // We cannot know if a given file will exists after compilation without compiling
        // So we pre-compile all the time
        _lp = new LocationParser(location);
        if (!_lp.matches())
        {
            throw new IOException("Path does not match: " + location);
        }
        
        // Special case for source maps: we serve ".vue" files in the source directory
        Source vueSourceFileSource = _handleSouresFiles();
        if (vueSourceFileSource != null)
        {
            _source = vueSourceFileSource;
            return vueSourceFileSource;
        }

        // Now lets do the job
        Source componentSource = _resolver.resolveURI(_lp.getComponentLocation());
        Source binarySource = _resolver.resolveURI(_lp.getBinaryLocation());
        
        synchronized (this)
        {
            // Are compiled files not up-to-date?
            if (!binarySource.exists() || binarySource.getLastModified() < _getLastModified(componentSource))
            {
                Source packageJsonSource = _resolver.resolveURI(_lp.getPackageJSONLocation());
                if (!binarySource.exists() || binarySource.getLastModified() < _getLastModified(packageJsonSource))
                {
                    // we cannot get last modification time of node_modules (too long)
                    // so lets compare if package.json file is newer than last compilation
                    
                    try
                    {
                        _installDependencies();
                    }
                    catch (ProcessingException e)
                    {
                        throw new IOException("Error while retrieving dependencies: " + location, e);
                    }
                }
                
                try
                {
                    _compile(binarySource);
                }
                catch (ProcessingException e)
                {
                    throw new IOException("Compilation error with module: " + location, e);
                }
            }
        }
        
        Source binaryResource = _resolver.resolveURI(_lp.getBinaryLocation() + _lp.getVuejsFile());
        _source = binaryResource;
        return binaryResource;
    }
    
    private Source _handleSouresFiles() throws MalformedURLException, IOException
    {
        if (!StringUtils.equals(_lp.getComponent(), "/_components") && !StringUtils.endsWith(_lp.getVuejsFile(), ".vue") && !StringUtils.endsWith(_lp.getVuejsFile(), "/main.js"))
        {
            return null;
        }
        
        // Fix a chrome issue with sourcemaps file and cache...
        Response response = ContextHelper.getResponse(_context);
        response.setHeader("Cache-Control", "no-cache, no-store, must-revalidate");
        response.setDateHeader("Expires", 0);
        
        // the real path is artificially prefixed by the resource type to bypass browser issues
        String file = _lp.getVuejsFile();
        file = file.substring(file.indexOf('/', 1)); // file is prefixed by /css or /js 
        
        String root = _lp.getSourceLocation();
        if (StringUtils.equals(_lp.getComponent(), "/_components"))
        {
            // for common sources, there's no 'src' folder
            root = StringUtils.removeEnd(root, "/src");
        }
        
        // This is related to sourcemap
        Source originalResource = _resolver.resolveURI(root + file);
        return originalResource;
    }
    
    private void _executeCommandeLine(String[] command, File directory, String successLog, String errorKeyword, String errorLog) throws IOException, ProcessingException
    {
        // We cannot execute directly "vue", because java do not take in the path env var
        String[] cmdPrefix = System.getProperty("os.name").toLowerCase().startsWith("windows") ? new String[] {"cmd", "/c"} : new String[] {};
        
        Process process = Runtime.getRuntime().exec(ArrayUtils.addAll(cmdPrefix, command), null, directory);
        
        List<String> out = new ArrayList<>();
        ReadStream inputStreamReader = new ReadStream(process.getInputStream(), out);
        ReadStream errorStreamReader = new ReadStream(process.getErrorStream(), out);
        Thread inputThread = new Thread(inputStreamReader);
        Thread errorThread = new Thread(errorStreamReader);
        inputThread.start();
        errorThread.start();
        
        try
        {
            int exitCode = process.waitFor();
            
            // We cannot get the process exit code since these executables always returns 0...
            // We cannot check if errorStreamReader.isEmpty() since it is not
            // So let's search the ERROR keyword
            String outAsString = StringUtils.join(out, '\n');
            
            if (exitCode != 0 || outAsString.contains(errorKeyword)) 
            {
                throw new ProcessingException(errorLog + "\n" + outAsString);
            }
            else if (getLogger().isInfoEnabled())
            {
                getLogger().info(successLog + "\n" + outAsString);
            }
        }
        catch (InterruptedException e)
        {
            throw new ProcessingException(e);
        }
    }

    private void _compile(Source binarySource) throws IOException, ProcessingException
    {
        if (getLogger().isInfoEnabled())
        {
            getLogger().info("Compiling " + _lp.getSourceLocation());
        }
        
        File sourceDir = ((FileSource) _resolver.resolveURI(_lp.getSourceLocation())).getFile();
        File outputDir = ((FileSource) binarySource).getFile();
        outputDir.delete();
        _executeCommandeLine(new String[] {"vue", "build", "--target", "lib", "--dest", outputDir.getAbsolutePath()}, 
                            sourceDir, 
                            "Compilation of " + _lp.getSourceLocation() + " ended with", 
                            "ERROR", 
                            "Could not compile " + _lp.getSourceLocation() + " due to:");
    }
    
    private void _installDependencies() throws IOException, ProcessingException
    {
        if (getLogger().isInfoEnabled())
        {
            getLogger().info("Getting dependencies of " + _lp.getComponentLocation());
        }
        
        File sourceDir = ((FileSource) _resolver.resolveURI(_lp.getComponentLocation())).getFile();
        _executeCommandeLine(new String[] {"npm", "install"},
                            sourceDir,
                            "Installing dependencies of " + _lp.getComponentLocation() + " ended with:",
                            "ERR!",
                            "Could not get dependencies " + _lp.getComponentLocation() + " due to:");
    }
    
    @Override
    public void generate(OutputStream out) throws IOException, ProcessingException
    {
        if (_source.getURI().endsWith(".map"))
        {
            // Sourcemap url needs to be adapted
            Request request = ObjectModelHelper.getRequest(_objectModel);
            
            String currentURI = request.getSitemapURI();
            String currentPrefix = StringUtils.substringBefore(currentURI, "/vuejs");
            
            String sourceUri = currentURI.substring(0, currentURI.lastIndexOf('.')); // strip trailing .map
            String ext = sourceUri.substring(sourceUri.lastIndexOf('.') + 1);
            
            String componentPrefix = _proxiedContextPathProvider.getContextPath() + "/" + currentPrefix + "/vuejs" + _lp.getComponent() + "/" + ext;
            String commonPrefix = _proxiedContextPathProvider.getContextPath() + "/" + currentPrefix + "/vuejs/_components/" + ext;
            
            try (InputStream is = _source.getInputStream())
            {
                // all following regexp aims to clean webpack-generated source maps
                String mapContent = IOUtils.toString(is, StandardCharsets.UTF_8);
                String mapContentFixed = RegExUtils.replaceAll(mapContent, "\"webpack://[^.\"]*(/vueserve/[^?\"]*\")", "\"" + componentPrefix + "$1"); // our sources from the component
                
                if (ext.equals("js"))
                {
                    mapContentFixed = RegExUtils.replaceAll(mapContentFixed, "\"webpack://[^\"]*/\\.\\./_components(/[^?\"]*\\.vue\")", "\"" + commonPrefix + "$1"); // our common .vue
                    mapContentFixed = RegExUtils.replaceAll(mapContentFixed, "\"webpack://[^\"]*/_components(/[^?\"]*\\.js\")", "\"" + commonPrefix + "$1"); // our common .js
                    mapContentFixed = RegExUtils.replaceAll(mapContentFixed, "\"webpack://[^\"]*(/main\\.js\")", "\"" + componentPrefix + "$1"); // our main.js
                    mapContentFixed = RegExUtils.replaceAll(mapContentFixed, "\"webpack://main/external \\\\\"[^\"]*\\\\\"\"", "\"unused\""); // external dependencies
                    mapContentFixed = RegExUtils.replaceAll(mapContentFixed, "\"webpack://[^\"]*/node_modules/([^\"]*/)([^\"]*\\.js\")", "\"sources-unavailable://$1source-unavailable-$2"); // internal dependencies (node_modules)
                    mapContentFixed = RegExUtils.replaceAll(mapContentFixed, "\"webpack://[^\"]*\"", "\"unused\""); // other webpack-generated artefacts
                }
                else if (ext.equals("css"))
                {
                    mapContentFixed = RegExUtils.replaceAll(mapContentFixed, "\"webpack://[^\"]*/_components(/[^?\"]*\\.vue\")", "\"" + commonPrefix + "$1"); // our common .vue
                }
                
                IOUtils.write(mapContentFixed, out, StandardCharsets.UTF_8);
            }
        }
        else
        {
            super.generate(out);
        }
    }

    @Override
    public long getLastModified()
    {
        return _getLastModified(_source);
    }

    // Recursively get the most recent last modified date
    private long _getLastModified(Source inputSource)
    {
        long result = inputSource.getLastModified();
        
        if (_isSourceDirectory(inputSource))
        {
            TraversableSource folder = (TraversableSource) inputSource;
            try
            {
                Collection<Source> children = folder.getChildren();
                for (Source child : children)
                {
                    result = Math.max(result, _getLastModified(child));
                }
            }
            catch (SourceException e)
            {
                getLogger().warn("Cannot get dependencies files of source " + inputSource.getURI(), e);
            }
            
        }
        
        return result;
    }
    
    @Override
    protected List<String> getDependenciesList(Source inputSource)
    {
        List<String> dependencies = new ArrayList<>();

        if (_isSourceDirectory(inputSource))
        {
            TraversableSource folder = (TraversableSource) inputSource;
            try
            {
                Collection<Source> children = folder.getChildren();
                for (Source child : children)
                {
                    dependencies.add(child.getURI());
                }
            }
            catch (SourceException e)
            {
                getLogger().warn("Cannot get dependencies files of source " + inputSource.getURI(), e);
            }
        }
        
        return dependencies;
    }

    private boolean _isSourceDirectory(Source inputSource)
    {
        return inputSource instanceof TraversableSource && ((TraversableSource) inputSource).isCollection() && !inputSource.getURI().endsWith("/node_modules/") && !inputSource.getURI().endsWith("/dist/");
    }

    static final class LocationParser
    {
        private static final Pattern __LOCATION = Pattern.compile("^(.*)(/[^/]*)(/vuejs)(/[^/]*)(/.*)$");
        private static final String __OUTPUT_DIR = "/dist";
        private static final String __SOURCE_DIR = "/src";
        private static final String __PACKAGE_JSON_FILE = File.separator + "package.json";
        
        private String _location;
        private String _mainLocation;
        private String _parentFolder;
        private String _vuejsDirectoryName;
        private String _vuejsComponentName;
        private String _vuejsFile;

        LocationParser(String location)
        {
            _location = location;
            
            Matcher matcher = __LOCATION.matcher(location);
            if (matcher.matches())
            {
                // cocoon://plugins/test/resources/vuejs/mycomponent/main.js
                _mainLocation = matcher.group(1);       // cocoon://plugins/test
                _parentFolder = matcher.group(2);       // /resources
                _vuejsDirectoryName = matcher.group(3); // /vuesjs
                _vuejsComponentName = matcher.group(4); // /mycomponent
                _vuejsFile = matcher.group(5);          // /main.js
            }
        }
        
        boolean matches()
        {
            return _mainLocation != null;
        }
        
        String getLocation()
        {
            return _location;
        }
        
        String getParentFolder()
        {
            return _parentFolder;
        }
        
        String getComponent()
        {
            return _vuejsComponentName;
        }
        
        String getComponentLocation()
        {
            return _mainLocation + _vuejsDirectoryName + _vuejsComponentName;
        }
        
        String getPackageJSONLocation()
        {
            return getComponentLocation() + __PACKAGE_JSON_FILE;
        }
        
        String getVuejsFile()
        {
            return _vuejsFile;
        }
        
        String getBinaryLocation()
        {
            return _mainLocation + _vuejsDirectoryName + _vuejsComponentName + __OUTPUT_DIR;
        }
        
        String getSourceLocation()
        {
            return _mainLocation + _vuejsDirectoryName + _vuejsComponentName + __SOURCE_DIR;
        }
    }
    
    class ReadStream implements Runnable 
    {
        private final InputStream _inputStream;
        private final List<String> _sf;
        private boolean _isEmpty;

        ReadStream(InputStream inputStream, List<String> sf) 
        {
            _inputStream = inputStream;
            _sf = sf;
            _isEmpty = true;
        }

        private BufferedReader getBufferedReader(InputStream is) 
        {
            return new BufferedReader(new InputStreamReader(is));
        }

        @Override
        public void run() 
        {
            BufferedReader br = getBufferedReader(_inputStream);
            String line = "";
            try 
            {
                while ((line = br.readLine()) != null) 
                {
                    _sf.add(line);
                    _isEmpty = false;
                }
            } 
            catch (IOException e) 
            {
                e.printStackTrace();
            }
        }
        
        boolean isEmpty()
        {
            return _isEmpty;
        }
    }
}
