/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.right;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.acting.ServiceableAction;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;

import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.group.Group;
import org.ametys.core.group.GroupDirectoryDAO;
import org.ametys.core.group.GroupIdentity;
import org.ametys.core.group.GroupManager;
import org.ametys.core.right.ProfileAssignmentStorage.AnonymousOrAnyConnectedKeys;
import org.ametys.core.right.ProfileAssignmentStorage.UserOrGroup;
import org.ametys.core.right.ProfileAssignmentStorageExtensionPoint;
import org.ametys.core.right.RightAssignmentContext;
import org.ametys.core.right.RightAssignmentContextExtensionPoint;
import org.ametys.core.right.RightProfilesDAO;
import org.ametys.core.ui.right.ProfileAssignmentsToolClientSideElement.AccessType;
import org.ametys.core.ui.right.ProfileAssignmentsToolClientSideElement.TargetType;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.user.population.UserPopulationDAO;

/**
 * Action for generating the grid for profile assignments
 */
public class GetProfileAssignmentsAction extends ServiceableAction
{
    /** The profile assignment storage component */
    protected ProfileAssignmentStorageExtensionPoint _profileAssignmentStorageEP;
    /** The extension point for right assignment contexts */
    protected RightAssignmentContextExtensionPoint _rightAssignmentContextEP;
    /** The profiles DAO */
    protected RightProfilesDAO _profilesDAO;
    /** The DAO for user populations */
    protected UserPopulationDAO _userPopulationDAO;
    /** The user manager */
    protected UserManager _userManager;
    /** The DAO for group directories */
    protected GroupDirectoryDAO _groupDirectoryDAO;
    /** The group manager */
    protected GroupManager _groupManager;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _profileAssignmentStorageEP = (ProfileAssignmentStorageExtensionPoint) smanager.lookup(ProfileAssignmentStorageExtensionPoint.ROLE);
        _rightAssignmentContextEP = (RightAssignmentContextExtensionPoint) smanager.lookup(RightAssignmentContextExtensionPoint.ROLE);
        _userPopulationDAO = (UserPopulationDAO) smanager.lookup(UserPopulationDAO.ROLE);
        _userManager = (UserManager) smanager.lookup(UserManager.ROLE);
        _groupDirectoryDAO = (GroupDirectoryDAO) smanager.lookup(GroupDirectoryDAO.ROLE);
        _groupManager = (GroupManager) smanager.lookup(GroupManager.ROLE);
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        if (_profilesDAO == null)
        {
            _profilesDAO = (RightProfilesDAO) manager.lookup(RightProfilesDAO.ROLE);
        }
        
        Map<String, Object> result = new HashMap<>();
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        Map jsParameters = (Map<String, Object>) objectModel.get(ObjectModelHelper.PARENT_CONTEXT);
        
        String rightAssignmentCtxId = (String) jsParameters.get("rightAssignmentContextId");
        RightAssignmentContext rightCtx = _rightAssignmentContextEP.getExtension(rightAssignmentCtxId);
        
        Object jsContext = jsParameters.get("context");
        Object context = rightCtx.convertJSContext(jsContext);
        
        List<String> profileIds = (List<String>) jsParameters.get("profileIds");
        if (profileIds == null)
        {
            // Get the identifiers of all existing profiles
            profileIds = _profilesDAO.getProfiles().stream().map(profile -> profile.getId()).collect(Collectors.toList());
        }
        
        List<Map<String, Object>> assignments = new ArrayList<>();
        
        assignments.add(_getAssignmentForAnonymous(rightCtx, context, profileIds));
        assignments.add(_getAssignmentForAnyConnectedUser(rightCtx, context, profileIds));
        assignments.addAll(_getAssignmentForUsers(rightCtx, context, context, profileIds).values());
        assignments.addAll(_getAssignmentForGroups(rightCtx, context, context, profileIds).values());
        
        result.put("assignments", assignments);
        
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);
        
        return EMPTY_MAP;
    }
    
    private Map<String, Object> _getAssignmentForAnonymous(RightAssignmentContext rightCtx, Object context, List<String> profileIds)
    {
        Map<String, Object> assignment = new HashMap<>();
        
        assignment.put("targetType", TargetType.ANONYMOUS.toString());
        
        for (String profileId : profileIds)
        {
            _getAssignmentForAnonymous(rightCtx, assignment, context, context, profileId);
        }
        
        return assignment;
    }
    
    private void _getAssignmentForAnonymous(RightAssignmentContext rightCtx, Map<String, Object> assignment, Object initialContext, Object currentContext, String profileId)
    {
        Map<AnonymousOrAnyConnectedKeys, Set<String>> profilesForAnonymousAndAnyConnectedUser = _profileAssignmentStorageEP.getProfilesForAnonymousAndAnyConnectedUser(currentContext);

        Set<String> deniedProfiles = Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANONYMOUS_DENIED)).orElse(Set.of());
        if (deniedProfiles.contains(profileId))
        {
            assignment.put(profileId, currentContext == initialContext ? AccessType.DENY.toString() : AccessType.INHERITED_DENY.toString());
            return;
        }
        
        Set<String> allowedProfiles = Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANONYMOUS_ALLOWED)).orElse(Set.of());
        if (allowedProfiles.contains(profileId))
        {
            assignment.put(profileId, currentContext == initialContext ? AccessType.ALLOW.toString() : AccessType.INHERITED_ALLOW.toString());
            return;
        }
        
        if (!_profileAssignmentStorageEP.isInheritanceDisallowed(currentContext))
        {
            // Can not determine assignment on current context, up to parent context
            Set<Object> parentContexts = rightCtx.getParentContexts(currentContext); 
            if (parentContexts != null)
            {
                for (Object parentContext : parentContexts)
                {
                    Map<String, Object> parentAssignment = new HashMap<>();
                    _getAssignmentForAnonymous(rightCtx, parentAssignment, initialContext, parentContext, profileId);
                    
                    String parentValue = (String) parentAssignment.get(profileId);
                    
                    if (AccessType.INHERITED_DENY.toString().equals(parentValue))
                    {
                        assignment.put(profileId, AccessType.INHERITED_DENY.toString());
                        return;
                    }
                    else if (AccessType.INHERITED_ALLOW.toString().equals(parentValue))
                    {
                        assignment.put(profileId, AccessType.INHERITED_ALLOW.toString());
                    }
                }
            }
        }
    }
    
    private Map<String, Object> _getAssignmentForAnyConnectedUser(RightAssignmentContext rightCtx, Object context, List<String> profileIds)
    {
        Map<String, Object> assignment = new HashMap<>();
        
        assignment.put("targetType", TargetType.ANYCONNECTED_USER.toString());
        
        for (String profileId : profileIds)
        {
            _getAssignmentForAnyConnectedUser(rightCtx, assignment, context, context, profileId);
        }
        
        return assignment;
    }
    
    private void _getAssignmentForAnyConnectedUser(RightAssignmentContext rightCtx, Map<String, Object> assignment, Object initialContext, Object currentContext, String profileId)
    {
        Map<AnonymousOrAnyConnectedKeys, Set<String>> profilesForAnonymousAndAnyConnectedUser = _profileAssignmentStorageEP.getProfilesForAnonymousAndAnyConnectedUser(currentContext);

        Set<String> deniedProfiles = Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_DENIED)).orElse(Set.of());
        if (deniedProfiles.contains(profileId))
        {
            assignment.put(profileId, currentContext == initialContext ? AccessType.DENY.toString() : AccessType.INHERITED_DENY.toString());
            return;
        }
        
        Set<String> allowedProfiles = Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_ALLOWED)).orElse(Set.of());
        if (allowedProfiles.contains(profileId))
        {
            assignment.put(profileId, currentContext == initialContext ? AccessType.ALLOW.toString() : AccessType.INHERITED_ALLOW.toString());
            return;
        }
        
        if (!_profileAssignmentStorageEP.isInheritanceDisallowed(currentContext))
        {   
            // Can not determines assignment on current context, up to parent context
            Set<Object> parentContexts = rightCtx.getParentContexts(currentContext); 
            if (parentContexts != null)
            {
                for (Object parentContext : parentContexts)
                {
                    Map<String, Object> parentAssignment = new HashMap<>();
                    _getAssignmentForAnyConnectedUser(rightCtx, parentAssignment, initialContext, parentContext, profileId);
                    
                    String parentValue = (String) parentAssignment.get(profileId);
                    
                    if (AccessType.INHERITED_DENY.toString().equals(parentValue))
                    {
                        assignment.put(profileId, AccessType.INHERITED_DENY.toString());
                        return;
                    }
                    else if (AccessType.INHERITED_ALLOW.toString().equals(parentValue))
                    {
                        assignment.put(profileId, AccessType.INHERITED_ALLOW.toString());
                    }
                }
            }
        }
    }    
    
    private Map<UserIdentity, Map<String, Object>> _getAssignmentForUsers(RightAssignmentContext rightCtx, Object initialContext, Object context, List<String> profileIds)
    {
        Map<UserIdentity, Map<String, Object>> assignments = new LinkedHashMap<>();
        
        // Get all user with a denied profile on current context
        Map<UserIdentity, Map<UserOrGroup, Set<String>>> profilesForUsers = _profileAssignmentStorageEP.getProfilesForUsers(context, null);
        for (Entry<UserIdentity, Map<UserOrGroup, Set<String>>> entry : profilesForUsers.entrySet())
        {
            _getAssignementForDeniedUser(initialContext, profileIds, assignments, context, Optional.ofNullable(entry.getValue().get(UserOrGroup.DENIED)).orElse(Set.of()), entry.getKey());
        }
        
        // Get all user with a allowed profile on current context
        for (Entry<UserIdentity, Map<UserOrGroup, Set<String>>> entry : profilesForUsers.entrySet())
        {
            _getAssignementForAllowedUser(initialContext, profileIds, assignments, context, Optional.ofNullable(entry.getValue().get(UserOrGroup.ALLOWED)).orElse(Set.of()), entry.getKey());
        }

        if (!_profileAssignmentStorageEP.isInheritanceDisallowed(context))
        {
            // Up to parent context
            Set<Object> parentContexts = rightCtx.getParentContexts(context);
            if (parentContexts != null)
            {        
                Map<UserIdentity, Map<String, Object>> allParentsAssignments = new HashMap<>();
                for (Object parentContext : parentContexts)
                {
                    Map<UserIdentity, Map<String, Object>> parentAssignment = _getAssignmentForUsers(rightCtx, initialContext, parentContext, profileIds);
                    _mergeUser(allParentsAssignments, parentAssignment, false);
                }
                
                _mergeUser(assignments, allParentsAssignments, true);
            }
        }
        
        return assignments;
    }

    private void _getAssignementForAllowedUser(Object context, List<String> profileIds, Map<UserIdentity, Map<String, Object>> assignments, Object currentContext,
            Set<String> allowedProfilesForUser, UserIdentity userIdentity)
    {
        // Check if the user still exits
        User user = _userManager.getUser(userIdentity);
        if (user != null)
        { 
            if (!assignments.containsKey(userIdentity))
            {
                assignments.put(userIdentity, _user2json(user));
            }
            
            Map<String, Object> userAssignment = assignments.get(userIdentity);
            
            for (String profileId : allowedProfilesForUser)
            {
                if (profileIds.contains(profileId) && !userAssignment.containsKey(profileId))
                {
                    userAssignment.put(profileId, currentContext == context ? AccessType.ALLOW.toString() : AccessType.INHERITED_ALLOW.toString());
                }
            }
        }
    }

    private void _getAssignementForDeniedUser(Object context, List<String> profileIds, Map<UserIdentity, Map<String, Object>> assignments, Object currentContext,
            Set<String> deniedProfilesForUser, UserIdentity userIdentity)
    {
        // Check if the user still exits
        User user = _userManager.getUser(userIdentity);
        if (user != null)
        { 
            if (!assignments.containsKey(userIdentity))
            {
                Map<String, Object> user2json = _user2json(user);
                if (user2json != null)
                {
                    assignments.put(userIdentity, user2json);
                }
            }
            
            Map<String, Object> userAssignment = assignments.get(userIdentity);
            
            for (String profileId : deniedProfilesForUser)
            {
                if (profileIds.contains(profileId) && !userAssignment.containsKey(profileId))
                {
                    userAssignment.put(profileId, currentContext == context ? AccessType.DENY.toString() : AccessType.INHERITED_DENY.toString());
                }
            }
        }
    }
    
    private Map<GroupIdentity, Map<String, Object>> _getAssignmentForGroups(RightAssignmentContext rightCtx, Object initialContext, Object context, List<String> profileIds)
    {
        Map<GroupIdentity, Map<String, Object>> assignments = new LinkedHashMap<>();
        
        // Get all user with a denied profile on current context
        Map<GroupIdentity, Map<UserOrGroup, Set<String>>> profilesForGroups = _profileAssignmentStorageEP.getProfilesForGroups(context, null);
        for (Entry<GroupIdentity, Map<UserOrGroup, Set<String>>> entry : profilesForGroups.entrySet())
        {
            _getAssignmentForDeniedGroup(initialContext, profileIds, assignments, context, Optional.ofNullable(entry.getValue().get(UserOrGroup.DENIED)).orElse(Set.of()), entry.getKey());
        }
        
        // Get all user with a allowed profile on current context
        for (Entry<GroupIdentity, Map<UserOrGroup, Set<String>>> entry : profilesForGroups.entrySet())
        {
            _getAssignementForAllowedGroup(initialContext, profileIds, assignments, context, Optional.ofNullable(entry.getValue().get(UserOrGroup.ALLOWED)).orElse(Set.of()), entry.getKey());
        }
        
        if (!_profileAssignmentStorageEP.isInheritanceDisallowed(context))
        {
            // Up to parent context
            Set<Object> parentContexts = rightCtx.getParentContexts(context);
            if (parentContexts != null)
            {        
                Map<GroupIdentity, Map<String, Object>> allParentsAssignments = new HashMap<>();
                for (Object parentContext : parentContexts)
                {
                    Map<GroupIdentity, Map<String, Object>> parentAssignment = _getAssignmentForGroups(rightCtx, initialContext, parentContext, profileIds);
                    _mergeGroup(allParentsAssignments, parentAssignment, false);
                }
                
                _mergeGroup(assignments, allParentsAssignments, true);
            }
        }
        
        return assignments;
    }
    
    private void _mergeUser(Map<UserIdentity, Map<String, Object>> finalAssignments, Map<UserIdentity, Map<String, Object>> parentsAssignments, boolean finalListPrevails)
    {
        for (UserIdentity identity : parentsAssignments.keySet())
        {
            if (!finalAssignments.containsKey(identity))
            {
                // Was not existing... overwrite
                finalAssignments.put(identity, parentsAssignments.get(identity));
            }
            else
            {
                // Was existing... merge
                for (String profileId : parentsAssignments.get(identity).keySet())
                {
                    Object assignmentsValue = finalAssignments.get(identity).get(profileId);
                    Object allParentsAssignmentsValue = parentsAssignments.get(identity).get(profileId);
                    
                    if (assignmentsValue == null 
                            || AccessType.UNKNOWN.toString().equals(assignmentsValue)
                            || !finalListPrevails && AccessType.INHERITED_ALLOW.toString().equals(assignmentsValue) && !AccessType.UNKNOWN.toString().equals(allParentsAssignmentsValue))
                    {
                        finalAssignments.get(identity).put(profileId, allParentsAssignmentsValue);
                    }
                }
            }
        }
    }
    private void _mergeGroup(Map<GroupIdentity, Map<String, Object>> finalAssignments, Map<GroupIdentity, Map<String, Object>> parentsAssignments, boolean finalListPrevails)
    {
        for (GroupIdentity identity : parentsAssignments.keySet())
        {
            if (!finalAssignments.containsKey(identity))
            {
                // Was not existing... overwrite
                finalAssignments.put(identity, parentsAssignments.get(identity));
            }
            else
            {
                // Was existing... merge
                for (String profileId : parentsAssignments.get(identity).keySet())
                {
                    Object assignmentsValue = finalAssignments.get(identity).get(profileId);
                    Object allParentsAssignmentsValue = parentsAssignments.get(identity).get(profileId);
                    
                    if (assignmentsValue == null 
                            || AccessType.UNKNOWN.toString().equals(assignmentsValue)
                            || !finalListPrevails && AccessType.INHERITED_ALLOW.toString().equals(assignmentsValue) && !AccessType.UNKNOWN.toString().equals(allParentsAssignmentsValue))
                    {
                        finalAssignments.get(identity).put(profileId, allParentsAssignmentsValue);
                    }
                }
            }
        }
    }

    private void _getAssignementForAllowedGroup(Object context, List<String> profileIds, Map<GroupIdentity, Map<String, Object>> assignments, Object currentContext,
            Set<String> allowedProfilesForGroup, GroupIdentity gpIdentity)
    {
        Group group = _groupManager.getGroup(gpIdentity.getDirectoryId(), gpIdentity.getId());
        if (group != null)
        {
            if (!assignments.containsKey(gpIdentity))
            {
                assignments.put(gpIdentity, _group2json(group));
            }
            
            Map<String, Object> gpAssignment = assignments.get(gpIdentity);
            
            for (String profileId : allowedProfilesForGroup)
            {
                if (profileIds.contains(profileId) && !gpAssignment.containsKey(profileId))
                {
                    gpAssignment.put(profileId, currentContext == context ? AccessType.ALLOW.toString() : AccessType.INHERITED_ALLOW.toString());
                }
            }
        }
    }

    private void _getAssignmentForDeniedGroup(Object context, List<String> profileIds, Map<GroupIdentity, Map<String, Object>> assignments, Object currentContext,
            Set<String> deniedProfilesForGroup, GroupIdentity gpIdentity)
    {
        Group group = _groupManager.getGroup(gpIdentity.getDirectoryId(), gpIdentity.getId());
        if (group != null)
        {
            if (!assignments.containsKey(gpIdentity))
            {
                assignments.put(gpIdentity, _group2json(group));
            }
            
            Map<String, Object> gpAssignment = assignments.get(gpIdentity);
            
            for (String profileId : deniedProfilesForGroup)
            {
                if (profileIds.contains(profileId) && !gpAssignment.containsKey(profileId))
                {
                    gpAssignment.put(profileId, currentContext == context ? AccessType.DENY.toString() : AccessType.INHERITED_DENY.toString());
                }
            }
        }
    }
    
    private Map<String, Object> _user2json(User user)
    {
        Map<String, Object> assignment = new HashMap<>();
        assignment.put("targetType", TargetType.USER.toString());
        
        String login = user.getIdentity().getLogin();
        String populationId = user.getIdentity().getPopulationId();
        assignment.put("login", login);
        assignment.put("populationId", populationId);
        assignment.put("populationLabel", _userPopulationDAO.getUserPopulation(populationId).getLabel());
        assignment.put("groups", _groupManager.getUserGroups(user.getIdentity()).stream()
                                    .map(this::_userGroup2json)
                                    .collect(Collectors.toList()));
        assignment.put("userSortableName", user.getSortableName());
        
        return assignment;
    }
   
    
    private Map<String, Object> _group2json(Group group)
    {
        Map<String, Object> assignment = new HashMap<>();
        assignment.put("targetType", TargetType.GROUP.toString());
        
        String groupId = group.getIdentity().getId();
        String directoryId = group.getIdentity().getDirectoryId();
        assignment.put("groupId", groupId);
        assignment.put("groupDirectory", directoryId);
        assignment.put("groupDirectoryLabel", _groupDirectoryDAO.getGroupDirectory(directoryId).getLabel());
        assignment.put("groupLabel", group.getLabel());
        
        return assignment;
    }
    
    private Map<String, Object> _userGroup2json(GroupIdentity group)
    {
        Map<String, Object> result = new HashMap<>();
        result.put("groupId", group.getId());
        result.put("groupDirectory", group.getDirectoryId());
        return result;
    }
}
