/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.script;

import java.io.File;
import java.io.IOException;
import java.time.Duration;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.FormatStyle;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.time.DurationFormatUtils;
import org.slf4j.Logger;

import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.mail.SendMailHelper;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.PluginsManager;

import jakarta.mail.MessagingException;

/**
 * Component able to execute scripts asynchronously, either from a schedulable or because the user asked to do so.
 * Script results are sent by email. 
 */
public class AsyncScriptHandler extends ScriptHandler implements Initializable
{
    /** Avalon role. */
    public static final String COMPONENT_ROLE = AsyncScriptHandler.class.getName();
    
    private static final String __ASYNC_EXECUTE_MAIL_SUBJECT_KEY = "PLUGINS_CORE_UI_SCRIPT_ASYNC_EXECUTE_MAIL_SUBJECT";
    private static final String __ASYNC_EXECUTE_MAIL_SUBJECT_ERROR_KEY = "PLUGINS_CORE_UI_SCRIPT_ASYNC_EXECUTE_MAIL_SUBJECT_ERROR";
    private static final String __ASYNC_EXECUTE_MAIL_BODY_KEY = "PLUGINS_CORE_UI_SCRIPT_ASYNC_EXECUTE_MAIL_BODY";
    private static final String __ASYNC_EXECUTE_MAIL_UNDEFINED = "PLUGINS_CORE_UI_SCRIPT_ASYNC_EXECUTE_MAIL_UNDEFINED";
    
    private Function<String, String> _mailSenderProvider;
    
    @Override
    public void initialize() throws Exception
    {
        _initializeMailSenderProvider();
    }
    
    /**
     * Initializes {@link #_mailSenderProvider}. The argument of the provider is the mail of the recipient (as string)
     */
    protected void _initializeMailSenderProvider()
    {
        if (PluginsManager.getInstance().isSafeMode())
        {
            // Safe mode, the sender will be the recipient of the mail (because of the absence of information about a potential sender in this mode)
            _mailSenderProvider = mailRecipient -> mailRecipient;
        }
        else
        {
            String mailSender = Config.getInstance().getValue("smtp.mail.from");
            // Normal mode, the sender will always be the sender of the Configuration
            _mailSenderProvider = mailRecipient -> mailSender;
        }
    }
    
    @Override
    protected ResultProcessor getProcessor()
    {
        return new AsyncResultProcessor();
    }
    
    /**
     * Send a mail report from the result of a script
     * @param scriptResults The result of the script
     * @param user The user
     * @param mailRecipient The recipient of the result mail
     * @param locale The user locale
     * @param logger The logger
     * @throws MessagingException If an error occurred while preparing or sending email
     * @throws IOException If an error occurred with an attachment.
     */
    public void sendReportMail(Map<String, Object> scriptResults, UserIdentity user, String mailRecipient, Locale locale, Logger logger) throws MessagingException, IOException
    {
        if (StringUtils.isNotBlank(mailRecipient)) 
        {
            String output = (String) scriptResults.get("output");
            String error = StringUtils.defaultString((String) scriptResults.get("error"));
            String errorMessage = StringUtils.defaultString((String) scriptResults.get("message"));
            String errorStacktrace = StringUtils.defaultString((String) scriptResults.get("stacktrace"));
            
            @SuppressWarnings("unchecked")
            List<File> attachments = (List<File>) scriptResults.get("attachments");
            
            Object result = scriptResults.getOrDefault("result", "");

            String i18nCatalog = "plugin.core-ui";
            I18nizableText undefinedText = new I18nizableText(i18nCatalog, __ASYNC_EXECUTE_MAIL_UNDEFINED);
            
            // Define start date and duration
            ZonedDateTime startDate = _getZDTFromScriptResults(scriptResults, "start");
            ZonedDateTime endDate = startDate != null ? _getZDTFromScriptResults(scriptResults, "end") : null;
            
            I18nizableText duration = Optional.ofNullable(endDate)
                .map(end -> Duration.between(startDate, endDate))
                .map(Duration::toMillis)
                .map(DurationFormatUtils::formatDurationHMS)
                .map(I18nizableText::new)
                .orElse(undefinedText);
            
            I18nizableText start = Optional.ofNullable(startDate)
                .map(DateTimeFormatter
                    .ofLocalizedDateTime(FormatStyle.LONG)
                    .withLocale(locale)
                    .withZone(ZoneId.systemDefault())
                    ::format
                )
                .map(I18nizableText::new)
                .orElse(undefinedText);
            
            String htmlBody = _i18nUtils.translate(
                new I18nizableText(
                    i18nCatalog, 
                    __ASYNC_EXECUTE_MAIL_BODY_KEY, 
                    Map.of(
                        "out", new I18nizableText(output), 
                        "htmlResult", new I18nizableText(ScriptResultFormatter.htmlFormatResult(result)), 
                        "error", new I18nizableText(error), 
                        "message", new I18nizableText(errorMessage), 
                        "htmlStacktrace", new I18nizableText(ScriptResultFormatter.htmlFormatStacktrace(errorStacktrace)), 
                        "start", start,
                        "duration", duration
                    )
                )
            ); 
            String title = _mailTitle(i18nCatalog, errorStacktrace);
            
            String mailSender = _mailSenderProvider.apply(mailRecipient);
            SendMailHelper.newMail()
                          .withSubject(title)
                          .withHTMLBody(htmlBody)
                          .withAttachments(attachments)
                          .withRecipient(mailRecipient)
                          .withSender(mailSender)
                          .sendMail();
        }
        else
        {
            logger.warn("User {} launched an asynchronous script but there is no email to send the output.", user);
        }
    }
    
    private ZonedDateTime _getZDTFromScriptResults(Map<String, Object> scriptResults, String paramName)
    {
        return Optional.of(paramName)
                .map(scriptResults::get)
                .map(String.class::cast)
                .map(DateUtils::parseZonedDateTime)
                .orElse(null);
    }
    
    private String _mailTitle(String i18nCatalog, String errorStacktrace)
    {
        String i18nKey = StringUtils.isEmpty(errorStacktrace)
                ? __ASYNC_EXECUTE_MAIL_SUBJECT_KEY
                : __ASYNC_EXECUTE_MAIL_SUBJECT_ERROR_KEY;
        return _i18nUtils.translate(new I18nizableText(i18nCatalog, i18nKey));
    }
    
    static class AsyncResultProcessor extends ResultProcessor
    {
        @Override
        protected Object process(Map<String, Object> results, Object scriptResult)
        {
            if (scriptResult instanceof File)
            {
                @SuppressWarnings("unchecked")
                List<File> attachments = (List<File>) results.computeIfAbsent("attachments", __ -> new ArrayList<>());
                attachments.add((File) scriptResult);
                
                return ((File) scriptResult).getName();
            }
            
            return super.process(results, scriptResult);
        }
    }
}
