/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.script;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * A class that hold the documentation of one function or one variable
 */
public class ScriptBindingDocumentation
{
    private I18nizableText _name;
    private I18nizableText _text;
    private Signature _signature;
    private List<Example> _examples;

    /**
     * Creates a documentation object
     * @param name The name of the documented stuff
     * @param text The description
     * @param signature The signature
     * @param examples The examples
     */
    public ScriptBindingDocumentation(I18nizableText name, I18nizableText text, Signature signature, List<Example> examples)
    {
        _name = name;
        _text = text;
        _signature = signature;
        _examples = examples;
    }
    
    /**
     * The documentation as a map
     * @return The documentation as a map
     */
    public Map<String, Object> asMap()
    {
        Map<String, Object> m = new HashMap<>();
        
        m.put("name", getName());
        m.put("description", getText());
        m.put("signature", getSignature().asMap());
        m.put("examples", getExamples() != null ? getExamples().stream().map(Example::asMap).collect(Collectors.toList()) : null);
        
        return m;
    }
    
    /**
     * The documentation name
     * @return The documentation name
     */
    public I18nizableText getName()
    {
        return _name;
    }
    
    /**
     * The description
     * @return The description
     */
    public I18nizableText getText()
    {
        return _text;
    }
    
    /**
     * The signature
     * @return The signature
     */
    public Signature getSignature()
    {
        return _signature;
    }
    
    /**
     * The examples
     * @return The examples
     */
    public List<Example> getExamples()
    {
        return _examples;
    }
    
    /**
     * A signature. 
     */
    public static class Signature
    {
        private String _returnType;
        private List<Argument> _returnSubtype;
        private I18nizableText _returnText;
        private List<Argument> _arguments;

        /**
         * Builds a signature
         * @param returnType The type or return type for a function
         * @param signatureSubtype Description of the type when it is an object. null otherwise
         * @param returnText The text
         * @param arguments The arguments for a function
         */
        public Signature(String returnType, List<Argument> signatureSubtype, I18nizableText returnText, List<Argument> arguments)
        {
            _returnType = returnType;
            _returnSubtype = signatureSubtype;
            _returnText = returnText;
            _arguments = arguments;
        }
        
        /**
         * The signature as map
         * @return The signature as map
         */
        public Map<String, Object> asMap()
        {
            Map<String, Object> m = new HashMap<>();
            
            m.put("type", getType());
            
            List<Argument> subtype = getSubtype();
            if (subtype != null)
            {
                Collection<Object> subm = new ArrayList<>();
                for (Argument argument : getSubtype())
                {
                    subm.add(argument.asMap());
                }
                m.put("subtype", subm);
            }
            
            m.put("text", getText());
            m.put("arguments", getArguments() != null ? getArguments().stream().map(Argument::asMap).collect(Collectors.toList()) : null);
            
            return m;
        }

        /**
         * The type (or return type for a function)
         * @return The type (or return type for a function)
         */
        public String getType()
        {
            return _returnType;
        }
        
        /**
         * The description of the type when it is an object.
         * @return Description or null.
         */
        public List<Argument> getSubtype()
        {
            return _returnSubtype;
        }
        
        /**
         * The description
         * @return The description
         */
        public I18nizableText getText()
        {
            return _returnText;
        }
        
        /**
         * The arguments (when a function)
         * @return The arguments (when a function)
         */
        public List<Argument> getArguments()
        {
            return _arguments;
        }
    }
    
    /**
     * Describes an argument when a function is documented 
     */
    public static class Argument
    {
        private String _argumentName;
        private String _argumentType;
        private List<Argument> _argumentSubtype;
        private I18nizableText _argumentText;
        private String _argumentOptionalValue;

        /**
         * Builds an argument
         * @param argumentName The name
         * @param argumentType The type
         * @param argumentSubtype Description of the type when it is an object. null otherwise 
         * @param argumentText The description
         * @param optionalValue Null if the argument is mandatory, the default value otherwise such as "null" or "\"default\""
         */
        public Argument(String argumentName, String argumentType, List<Argument> argumentSubtype, I18nizableText argumentText, String optionalValue)
        {
            _argumentName = argumentName;
            _argumentType = argumentType;
            _argumentSubtype = argumentSubtype;
            _argumentText = argumentText;
            _argumentOptionalValue = optionalValue;
        }
        
        /**
         * The argument as map
         * @return The argument as map
         */
        public Map<String, Object> asMap()
        {
            Map<String, Object> m = new HashMap<>();
            
            m.put("name", getName());
            m.put("type", getType());

            List<Argument> subtype = getSubtype();
            if (subtype != null)
            {
                Collection<Object> subm = new ArrayList<>();
                for (Argument argument : getSubtype())
                {
                    subm.add(argument.asMap());
                }
                m.put("subtype", subm);
            }
            
            m.put("text", getText());
            m.put("optional", getOptionalValue());
            
            return m;
        }

        /**
         * The name of the argument
         * @return The name of the argument
         */
        public String getName()
        {
            return _argumentName;
        }
        
        /**
         * The type/class name of the argument
         * @return The type/class name of the argument
         */
        public String getType()
        {
            return _argumentType;
        }
        
        /**
         * Description of the type when it is an object.
         * @return Description or null otherwise
         */
        public List<Argument> getSubtype()
        {
            return _argumentSubtype;
        }
        
        /**
         * The description of the argument
         * @return The description of the argument
         */
        public I18nizableText getText()
        {
            return _argumentText;
        }
        
        /**
         * The type/class name of the argument
         * @return The type/class name of the argument
         */
        public String getOptionalValue()
        {
            return _argumentOptionalValue;
        }
    }
    
    /**
     * Example 
     */
    public static class Example 
    {
        private String _code;
        private I18nizableText _exampleText;

        /**
         * Creates an exemple
         * @param code Code
         * @param text Associated text
         */
        public Example(String code, I18nizableText text)
        {
            _code = code;
            _exampleText = text;
        }
        
        /**
         * The example as map
         * @return The example as map
         */
        public Map<String, Object> asMap()
        {
            Map<String, Object> m = new HashMap<>();
            
            m.put("code", getCode());
            m.put("text", getText());
            
            return m;
        }
        
        /**
         * The code
         * @return The code
         */
        public String getCode()
        {
            return _code;
        }
        
        /**
         * Get the text
         * @return the text
         */
        public I18nizableText getText()
        {
            return _exampleText;
        }
    }
}
