/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.script;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceNotFoundException;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.SourceUtil;
import org.xml.sax.SAXException;

import org.ametys.plugins.core.ui.script.ScriptBindingDocumentation.Argument;
import org.ametys.plugins.core.ui.script.ScriptBindingDocumentation.Example;
import org.ametys.plugins.core.ui.script.ScriptBindingDocumentation.Signature;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;
import org.ametys.runtime.plugin.component.PluginAware;

/**
 * Static implementation of the ScriptBinding that can read the functions from an xml file
 */
public class StaticConfigurableScriptBinding extends AbstractLogEnabled implements ScriptBinding, Configurable, Serviceable, PluginAware, Contextualizable
{
    /** Source Resolver */
    protected SourceResolver _sourceResolver;

    /** List of variables per configuration file */
    protected List<String> _variablesScripts;

    /** List of functions per configuration file */
    protected List<String> _functions;
    
    /** List of function descriptions */
    protected Map<String, ScriptBindingDocumentation> _functionsDescriptions;

    /** List of variable descriptions */
    protected Map<String, ScriptBindingDocumentation> _variablesDescriptions;

    /** The plugin name */
    protected String _pluginName;

    /** The pattern corresponding to the script binding */
    protected Pattern _workspacePattern;

    /** The avalon context */
    protected Context _context;

    /** List of tutorials*/
    protected Map<I18nizableText, ScriptBindingDocumentation> _tutorials;
    
    @Override
    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        _pluginName = pluginName;
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        String workspacePattern = configuration.getChild("workspace").getValue(".*");
        _workspacePattern = Pattern.compile(workspacePattern);

        _variablesScripts = new ArrayList<>();
        _variablesDescriptions = new HashMap<>();
        _functions = new ArrayList<>();
        _functionsDescriptions = new HashMap<>();
        _tutorials = new HashMap<>();

        String pluginName = configuration.getChild("file").getAttribute("plugin", _pluginName);
        String filePath = configuration.getChild("file").getValue(null);
        if (StringUtils.isNotEmpty(filePath))
        {
            filePath = _getURI(pluginName, filePath, null);
        }
        
        Configuration fileConfiguration = StringUtils.isNotEmpty(filePath) ? _getFileConfiguration(filePath, configuration) : configuration;
        
        try
        {
            _configureVariables(pluginName, filePath, fileConfiguration);
            _configureFunctions(pluginName, filePath, fileConfiguration);
            _configureTutorials(pluginName, filePath, fileConfiguration);
        }
        catch (IOException e)
        {
            throw new ConfigurationException("An error occurred while configuring the script binding", fileConfiguration, e);
        }
    }

    /**
     * Configure the function part
     * @param pluginName Current plugin
     * @param filePath The conf file path
     * @param configuration the function
     * @throws IOException If an error occurred
     * @throws ConfigurationException If an error occurred 
     */
    protected void _configureFunctions(String pluginName, String filePath, Configuration configuration) throws ConfigurationException, IOException
    {
        
        Configuration functionsConf = configuration.getChild("functions", false);
        if (functionsConf != null)
        {
            for (Configuration functionConf : functionsConf.getChildren("function"))
            {
                for (Configuration scriptConf : functionConf.getChildren("script"))
                {
                    _functions.add(_getContent(pluginName, filePath, scriptConf));
                }
                
                ScriptBindingDocumentation d = _parseDocumentation(pluginName, filePath, functionConf);
                _functionsDescriptions.put(d.getName().getLabel(), d);
            }
        }
    }
    
    /**
     * Configure the variable part
     * @param pluginName Current plugin
     * @param filePath The conf file path
     * @param configuration the variable
     * @throws IOException If an error occurred
     * @throws ConfigurationException If an error occurred 
     */
    protected void _configureVariables(String pluginName, String filePath, Configuration configuration) throws ConfigurationException, IOException
    {
        Configuration variablesConf = configuration.getChild("variables", false);
        if (variablesConf != null)
        {
            for (Configuration variableConf : variablesConf.getChildren("variable"))
            {
                for (Configuration scriptConf : variableConf.getChildren("script"))
                {
                    _variablesScripts.add(_getContent(pluginName, filePath, scriptConf));
                }
                
                ScriptBindingDocumentation d = _parseDocumentation(pluginName, filePath, variableConf);
                _variablesDescriptions.put(d.getName().getLabel(), d);
            }
        }
    }
    
    /**
     * Configure the tutorial part
     * @param pluginName Current plugin
     * @param filePath The conf file path
     * @param configuration the tutorial
     * @throws IOException If an error occurred
     * @throws ConfigurationException If an error occurred 
     */
    protected void _configureTutorials(String pluginName, String filePath, Configuration configuration) throws ConfigurationException, IOException
    {
        Configuration tutorialsConf = configuration.getChild("tutorials", false);
        if (tutorialsConf != null)
        {
            for (Configuration tutorialConf : tutorialsConf.getChildren("tutorial"))
            {
                ScriptBindingDocumentation d = _parseDocumentation(pluginName, filePath, tutorialConf);
                _tutorials.put(d.getName(), d);
            }
        }
    }
    
    private ScriptBindingDocumentation _parseDocumentation(String pluginName, String filePath, Configuration initialConfiguration) throws ConfigurationException, IOException
    {
        I18nizableText name = I18nizableText.parseI18nizableText(initialConfiguration.getChild("name", false), "plugin." + pluginName, "");
        I18nizableText text = I18nizableText.parseI18nizableText(initialConfiguration.getChild("text", false), "plugin." + pluginName, "");
     
        Configuration signatureConfiguration = initialConfiguration.getChild("signature", true);
       
        String signatureType = signatureConfiguration.getChild("type", true).getValue(null); 
        I18nizableText signatureText = I18nizableText.parseI18nizableText(signatureConfiguration.getChild("text", false), "plugin." + pluginName, "");
        List<Argument> signatureSubtype = _configureSubtype(signatureConfiguration, pluginName);
       
        List<ScriptBindingDocumentation.Argument> arguments = null;
        Configuration argumentsConfiguration = signatureConfiguration.getChild("arguments", false);
        if (argumentsConfiguration != null)
        {
            arguments = new ArrayList<>();
            for (Configuration argumentConfiguration : argumentsConfiguration.getChildren("argument"))
            {
                Argument argument = _parseArgument(argumentConfiguration, pluginName);
                arguments.add(argument);
            }
        }
       
        Signature signature = new Signature(signatureType, signatureSubtype, signatureText, arguments);
       
        List<Example> examples = null;
        Configuration examplesConfiguration = initialConfiguration.getChild("examples", false);
        if (examplesConfiguration != null)
        {
            examples = new ArrayList<>();
            for (Configuration exampleConfiguration : examplesConfiguration.getChildren("example"))
            {
                String code = _getContent(pluginName, filePath, exampleConfiguration.getChild("script"));
                I18nizableText exampleText = I18nizableText.parseI18nizableText(exampleConfiguration.getChild("text", false), "plugin." + pluginName, "");
                
                Example example = new Example(code, exampleText);
                examples.add(example);
            }
        }
       
        ScriptBindingDocumentation doc = new ScriptBindingDocumentation(name, text, signature, examples);
        return doc;
    }
   
    private List<Argument> _configureSubtype(Configuration configuration, String pluginName)
    {
        Configuration subtypeConfiguration = configuration.getChild("subtype", false);
        if (subtypeConfiguration == null)
        {
            return null;
        }
        
        List<Argument> subtype = new ArrayList<>();
        
        for (Configuration argumentConfiguration : subtypeConfiguration.getChildren("argument"))
        {
            Argument argument = _parseArgument(argumentConfiguration, pluginName);
            subtype.add(argument);
        }
        
        return subtype;
    }

    private Argument _parseArgument(Configuration argumentConfiguration, String pluginName)
    {
        String argumentName = argumentConfiguration.getChild("name", true).getValue(null); 
        String argumentType = argumentConfiguration.getChild("type", true).getValue(null); 
        I18nizableText argumentText = I18nizableText.parseI18nizableText(argumentConfiguration.getChild("text", false), "plugin." + pluginName, "");
        String argumentOptionalValue = argumentConfiguration.getChild("optional", true).getValue(null);
        List<Argument> argumentSubtype = _configureSubtype(argumentConfiguration, pluginName);
        
        return new Argument(argumentName, argumentType, argumentSubtype, argumentText, argumentOptionalValue);
    }

    private String _getContent(String pluginName, String filePath, Configuration initialConfiguration) throws ConfigurationException, IOException
    {
        if (initialConfiguration != null)
        {
            String fileName = initialConfiguration.getAttribute("file", null);
            String value = StringUtils.isEmpty(fileName) ? initialConfiguration.getValue() : _getRelativeFileContent(pluginName, fileName, filePath);
            
            return value;
        }
        else
        {
            return null;
        }
    }

    private Configuration _getFileConfiguration(String uri, Configuration initialConfiguration)
    {
        try
        {
            Configuration fileConfiguration = _getConfiguration(uri);
            
            return fileConfiguration;
        }
        catch (SourceNotFoundException e)
        {
            getLogger().info("No script binding file was found at {}. Using internally declared config.", uri);
        }
        catch (ConfigurationException | SAXException | IOException exception)
        {
            getLogger().error("The script binding file at {} cannot be read. Using internally declared config.", uri, exception);
        }
        return initialConfiguration;
    }
    
    private String _getURI(String pluginName, String filePath, String relativeTo)
    {
        if (filePath.startsWith("/"))
        {
            return "context:/" + filePath;
        }
        else
        {
            if (StringUtils.isNotEmpty(relativeTo))
            {
                return SourceUtil.absolutize(relativeTo, filePath);
            }
            else
            {
                return "plugin:" + pluginName + "://" + filePath;
            }
        }
    }

    private String _getRelativeFileContent(String pluginName, String filePath, String relativeTo) throws IOException
    {
        Source source = null;
        
        try
        {
            String uri = _getURI(pluginName, filePath, relativeTo);
            source = _sourceResolver.resolveURI(uri);
            
            try (InputStream is = source.getInputStream())
            {
                return IOUtils.toString(is, StandardCharsets.UTF_8);
            }
        }
        finally
        {
            if (source != null)
            {
                _sourceResolver.release(source);
            }
        }
    }

    private Configuration _getConfiguration(String uri) throws ConfigurationException, SAXException, IOException
    {
        Source source = null;
        
        try
        {
            source = _sourceResolver.resolveURI(uri);
            
            try (InputStream is = source.getInputStream())
            {
                return new DefaultConfigurationBuilder(true).build(is, uri);
            }
        }
        finally
        {
            if (source != null)
            {
                _sourceResolver.release(source);
            }
        }
    }
    
    public Pattern getWorkspacePattern()
    {
        return _workspacePattern;
    }

    public String getFunctions()
    {
        return StringUtils.join(_functions, "\n");
    }
    
    public String getVariablesScripts()
    {
        return StringUtils.join(_variablesScripts, "\n");
    }

    public Map<String, ScriptBindingDocumentation> getFunctionsDescriptions()
    {
        return _functionsDescriptions;
    }

    public Map<I18nizableText, ScriptBindingDocumentation> getTutorials()
    {
        return _tutorials;
    }

    public Map<String, Object> getVariables(ScriptExecArguments execArgs)
    {
        // Do nothing
        return null;
    }

    public Map<String, ScriptBindingDocumentation> getVariablesDescriptions()
    {
        return _variablesDescriptions;
    }

    public void cleanVariables(Map<String, Object> variables)
    {
        // Do nothing
    }
}
