/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.system;

import java.io.IOException;
import java.lang.management.ManagementFactory;
import java.time.ZonedDateTime;
import java.util.Locale;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.i18n.I18nUtils;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.DateUtils;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.runtime.plugins.admin.system.SystemHelper;
import org.ametys.runtime.servlet.RuntimeServlet;
import org.ametys.runtime.servlet.RuntimeServlet.ForcedMainteanceInformations;
import org.ametys.runtime.servlet.RuntimeServlet.RunMode;

/**
 * Generates the system announcement
 */
public class SystemAnnouncementGenerator extends ServiceableGenerator
{
    private SystemHelper _systemHelper;
    private UserHelper _userHelper;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _systemHelper = (SystemHelper) smanager.lookup(SystemHelper.ROLE);
        _userHelper = (UserHelper) smanager.lookup(UserHelper.ROLE);
    }
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        contentHandler.startDocument();
        XMLUtils.startElement(contentHandler, "SystemAnnounce");
        
        boolean isAvailable = _systemHelper.isSystemAnnouncementAvailable();
        
        XMLUtils.createElement(contentHandler, "IsAvailable", String.valueOf(isAvailable));
        if (isAvailable)
        {
            Locale locale = I18nUtils.findLocale(objectModel, "locale", null, Locale.getDefault(), true);
            
            XMLUtils.createElement(contentHandler, "LastModification", String.valueOf(_systemHelper.getSystemAnnoucementLastModificationDate())); 
            XMLUtils.createElement(contentHandler, "Message", _systemHelper.getSystemAnnouncement(locale.getLanguage()));
        }
        
        AttributesImpl attrs = new AttributesImpl();
        attrs.addCDATAAttribute("active", RuntimeServlet.getRunMode() == RunMode.MAINTENANCE ? "true" : "false");
        attrs.addCDATAAttribute("mode", RuntimeServlet.getMaintenanceStatus().toString()); 
        
        long maintenanceTimestamp = ManagementFactory.getRuntimeMXBean().getStartTime();
        
        XMLUtils.startElement(contentHandler, "Maintenance", attrs);
        ForcedMainteanceInformations maintenanceStatusForcedInformations = RuntimeServlet.getMaintenanceStatusForcedInformations();
        if (maintenanceStatusForcedInformations != null)
        {
            XMLUtils.createElement(contentHandler, "comment", RuntimeServlet.getMaintenanceStatusForcedInformations().comment());
            
            UserIdentity initiator = RuntimeServlet.getMaintenanceStatusForcedInformations().initiator();
            if (initiator != null)
            {
                String userFullName = StringUtils.defaultString(_userHelper.getUserFullName(initiator));
                XMLUtils.createElement(contentHandler, "initiator", userFullName + " (" + UserIdentity.userIdentityToString(initiator) + ")");
            }
            
            ZonedDateTime since = RuntimeServlet.getMaintenanceStatusForcedInformations().since();
            if (since != null)
            {
                maintenanceTimestamp = since.toInstant().getEpochSecond();
                String sinceAsString = DateUtils.zonedDateTimeToString(since);
                XMLUtils.createElement(contentHandler, "since", sinceAsString);
            }
        }
        
        XMLUtils.createElement(contentHandler, "timestamp", Long.toString(maintenanceTimestamp));
        
        XMLUtils.endElement(contentHandler, "Maintenance");
        
        XMLUtils.endElement(contentHandler, "SystemAnnounce");
        contentHandler.endDocument();
    }
}
