/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.user;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Collection;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.caching.CacheableProcessingComponent;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Response;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.cocoon.reading.ServiceableReader;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.SourceValidity;
import org.apache.excalibur.source.impl.validity.TimeStampValidity;
import org.xml.sax.SAXException;

import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.ImageHelper;
import org.ametys.plugins.core.ui.user.ProfileImageProvider.UserProfileImage;

/**
 * {@link Reader} for generating the user profile image stored in the 'profile-image' user pref.
 * The source of the image can also be forced to retrieve the image of a given source (gravatar, initial, upload etc...)
 */
public class ProfileImageReader extends ServiceableReader implements CacheableProcessingComponent
{
    /** Allowed image format */
    public static final Collection<String> ALLOWED_IMG_FORMATS = Arrays.asList("png", "gif", "jpg", "jpeg");
    
    /** Current user provider */
    protected CurrentUserProvider _currentUserProvider;
    
    /** User profile image provider */
    private ProfileImageProvider _profileImageProvider;

    private UserProfileImage _image;

    private int _size;

    private int _maxSize;

    private UserIdentity _user;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        super.service(serviceManager);
        _currentUserProvider = (CurrentUserProvider) serviceManager.lookup(CurrentUserProvider.ROLE);
        _profileImageProvider = (ProfileImageProvider) manager.lookup(ProfileImageProvider.ROLE);
    }

    @Override
    public void setup(SourceResolver sresolver, Map mObjectModel, String src, Parameters par) throws ProcessingException, SAXException, IOException
    {
        super.setup(sresolver, mObjectModel, src, par);
        
        String imageSourceStr = StringUtils.defaultString(parameters.getParameter("image-source", null));
        _user = _getUser();
        _size = parameters.getParameterAsInteger("size", 0);
        _maxSize = parameters.getParameterAsInteger("maxSize", 0);

        _image = _profileImageProvider.getImage(_user, imageSourceStr, _size, _maxSize);
    }
    
    @Override
    public void generate() throws IOException, ProcessingException
    {
        boolean download = parameters.getParameterAsBoolean("download", false);
        
        // Read the image
        try
        {
            Response response = ObjectModelHelper.getResponse(objectModel);
            _readImage(response, _user, _image, download, _size, _maxSize);
        }
        finally
        {
            IOUtils.closeQuietly(out);
        }
    }
    
    @Override
    public long getLastModified()
    {
        Long lastModified = _image.getLastModified();
        return lastModified != null ? lastModified : 0;
    }
    
    public Serializable getKey()
    {
        return this.getClass().getName() + "$" + UserIdentity.userIdentityToString(_user) + "$" + _size + "$" + _maxSize + "$" + _image.getType();
    }
    
    public SourceValidity getValidity()
    {
        return new TimeStampValidity(getLastModified());
    }
    
    private UserIdentity _getUser()
    {
        String login =  parameters.getParameter("login", StringUtils.EMPTY);
        String populationId =  parameters.getParameter("populationId", StringUtils.EMPTY);
        // Default to current user if login not provided, except for the default source for which a login is not needed.
        if (StringUtils.isEmpty(login) || StringUtils.isEmpty(populationId))
        {
            return _currentUserProvider.getUser();
        }
        else
        {
            return new UserIdentity(login, populationId);
        }
    }

    /**
     * Read the image from an input stream
     * @param response The response
     * @param user The user
     * @param image The user profile image to read
     * @param download To request a download
     * @param size The desired size
     * @param maxSize The max size
     * @throws IOException If an I/O error occurs while manipulating streams
     */
    protected void _readImage(Response response, UserIdentity user, UserProfileImage image, boolean download, int size, int maxSize) throws IOException
    {
        try (InputStream is = image.getInputstream())
        {
            String filename = StringUtils.defaultIfEmpty(image.getFilename(), user.getLogin() + ".png");
            String format = FilenameUtils.getExtension(filename);
            
            format = ALLOWED_IMG_FORMATS.contains(format) ? format : "png";
            
            if (download)
            {
                response.setHeader("Content-Disposition", "attachment; filename=\"" + filename + "\"");
            }
            
            if (size > 0 || maxSize > 0)
            {
                // Profile image are cropped by default
                ImageHelper.generateThumbnail(is, out, format, 0, 0, maxSize, maxSize, size, size);
            }
            else
            {
                Long length = image.getLength();
                if (length != null && length > 0)
                {
                    response.setHeader("Content-Length", Long.toString(length));
                }
                
                IOUtils.copy(is, out);
            }
        }
    }
    
    @Override
    public void recycle()
    {
        super.recycle();
        
        if (_image != null)
        {
            IOUtils.closeQuietly(_image.getInputstream());
            _image = null;
        }
    }
}
