/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.user;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.SourceResolver;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.AmetysXSLTHelper;
import org.ametys.core.util.ImageResolverHelper;
import org.ametys.core.util.URIUtils;
import org.ametys.plugins.core.ui.user.ProfileImageProvider.UserProfileImage;

/**
 * Helper to get the URI for the profile image of a user
 */
public class ProfileImageResolverHelper implements Serviceable, Contextualizable, Initializable
{
    /** The source resolver */
    protected static SourceResolver _srcResolver;
    /** The avalon context */
    protected static Context _context;
    
    private static final Logger __LOGGER = LoggerFactory.getLogger(ProfileImageResolverHelper.class.getName());


    private static ProfileImageResolverHelper _instance;
    private ProfileImageProvider _profileImageProvider;

    public void service(ServiceManager manager) throws ServiceException
    {
        _srcResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _profileImageProvider = (ProfileImageProvider) manager.lookup(ProfileImageProvider.ROLE);
    }
    
    public void initialize() throws Exception
    {
        _instance = this;
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    /**
     * Get the input stream of the profile image
     * @param login the login of the user
     * @param populationId the population of the user
     * @param size the size for the image
     * @param lang the lang to use to resolve image. Can be null.
     * @return the image as input stream
     */
    public InputStream getProfileImage(String login, String populationId, int size, String lang)
    {
        UserIdentity user = new UserIdentity(login, populationId);
        try
        {
            UserProfileImage image = _profileImageProvider.getImage(user, lang, size, size);
            if (image != null)
            {
                return image.getInputstream();
            }
        }
        catch (ProcessingException e)
        {
            __LOGGER.error("Failed to retrieve profile image for user " + user, e);
        }
        
        return null;
    }
    
    /**
     * Get the profile image uri
     * @param login the login of the user
     * @param populationId the population of the user
     * @param size the size for the image
     * @return the uri
     */
    public String getProfileImageUri(String login, String populationId, int size)
    {
        return "/_plugins/core-ui/user/" + populationId + "/" + login + "/image_" + size;
    }
    
    /**
     * Resolve an uri for profile image of an user
     * 
     * This method provide a convenient way to obtain the proxy URL on the base64 representation of the avatar
     * based on the request attribute "forceBase64Encoding".
     * 
     * @param login the login of the user
     * @param populationId the population of the user
     * @param size the size for the image
     * @param lang the lang to use to resolve image uri. Can be null.
     * @return the resolved URI
     */
    public static String resolve(String login, String populationId, int size, String lang)
    {
        return resolve(login, populationId, size, lang, false);
    }
    
    /**
     * Resolve an uri for profile image of an user
     * 
     * This method provide a convenient way to obtain the proxy URL on the base64 representation of the avatar
     * based on the request attribute "forceBase64Encoding".
     * 
     * @param login the login of the user
     * @param populationId the population of the user
     * @param size the size for the avatar
     * @param lang the lang to use to resolve image uri. Can be null.
     * @param absolute true to generate absolute url
     * @return the resolved URI
     */
    public static String resolve(String login, String populationId, int size, String lang, boolean absolute)
    {
        return resolve(login, populationId, size, lang, absolute, false);
    }
    
    /**
     * Resolve an uri for profile image of an user
     * 
     * This method provide a convenient way to obtain the proxy URL on the base64 representation of the avatar
     * based on the request attribute "forceBase64Encoding".
     * 
     * @param login the login of the user
     * @param populationId the population of the user
     * @param size the size for the avatar
     * @param lang the lang to use to resolve image uri. Can be null.
     * @param absolute true to generate absolute url
     * @param internal true to get an internal URI.
     * @return the resolved URI
     */
    public static String resolve(String login, String populationId, int size, String lang, boolean absolute, boolean internal)
    {
        String path = _instance.getProfileImageUri(login, populationId, size);
        
        Map<String, String> params = new HashMap<>();
        if (lang != null)
        {
            params.put("lang", lang);
        }
        
        String resultUri = _instance.getUri(path, absolute, internal);
        
        return internal ? URIUtils.buildURI(resultUri, params) : URIUtils.encodeURI(resultUri, params);
    }
    
    /**
     * Resolve an user image and return it as a base64-encoded string.
     * @param login the login of the user
     * @param populationId the population of the user
     * @param cropSize the size for the avatar. The avatar will be cropped to the given dimension.
     * @param lang the lang to use to resolve image uri. Can be null.
     * @return a base64-encoded string representing the image.
     */
    public static String resolveAsBase64(String login, String populationId, int cropSize, String lang)
    {
        try
        {
            try (InputStream dataIs = _instance.getProfileImage(login, populationId, cropSize, lang))
            {
                if (dataIs != null)
                {
                    return ImageResolverHelper.resolveImageAsBase64(dataIs, "image/png", 0, 0, 0, 0, cropSize, cropSize);
                }
            }
        }
        catch (IOException e)
        {
            __LOGGER.error("Failed to retrieve profile image as Base64 for user " + login + "/" + populationId, e);
        }
        
        return StringUtils.EMPTY;
    }
    
    /**
     * Get the URI prefix
     * @param path the resource path
     * @param absolute true if the url must be absolute
     * @param internal true to get an internal URI.
     * @return the URI prefix
     */
    protected String getUri(String path, boolean absolute, boolean internal)
    {
        if (internal)
        {
            return "cocoon:/" + path;
        }
        else if (absolute)
        {
            return AmetysXSLTHelper.uriPrefix(true) + path;
        }
        else
        {
            return AmetysXSLTHelper.absoluteUriPrefix(true) + path;
        }
    }
}
