/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.ui.util;

import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;

import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;

/**
 * Helper for saving and restoring request attributes
 */
public class RequestAttributesHelper implements Component, Contextualizable
{
    /** Avalon role */
    public static final String ROLE = RequestAttributesHelper.class.getName();
    
    private Context _context;
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    /**
     * Transforms the request attributes into a map and clean the attributes
     * @return A copy of all the request attributes
     */
    public Map<String, Object> saveRequestAttributes()
    {
        Map<String, Object> attrs = new HashMap<>();
        
        Request request = ContextHelper.getRequest(_context);
        
        Enumeration<String> attrNames = request.getAttributeNames();
        while (attrNames.hasMoreElements())
        {
            String attrName = attrNames.nextElement();
            Object value = request.getAttribute(attrName);
            
            attrs.put(attrName, value);
        }

        return attrs;
    }

    /**
     * Clean the requests attributes and add those in the map
     * @param attributes The attributes to restore
     */
    public void restoreRequestAttributes(Map<String, Object> attributes)
    {
        removeRequestAttributes();
        
        Request request = ContextHelper.getRequest(_context);
        
        for (String attrName : attributes.keySet())
        {
            request.setAttribute(attrName, attributes.get(attrName));
        }
    }
    
    /**
     * Remove the request attributes
     */
    @SuppressWarnings("unchecked")
    public void removeRequestAttributes()
    {
        Request request = ContextHelper.getRequest(_context);
        List<String> attrNames = Collections.list(request.getAttributeNames());
        for (String attrName : attrNames)
        {
            // Check if the current attribute is a dispatchable cache, remove it otherwise
            if (!_isADispatchableCache(attrName, request))
            {
                request.removeAttribute(attrName);
            }
        }
        
    }
    private boolean _isADispatchableCache(String attrName, Request request)
    {
        Object attribute = request.getAttribute(attrName);
        return attribute instanceof Cache && attrName.startsWith(AbstractCacheManager.ROLE) && ((Cache) attribute).isDispatchable();
        
    }
    
}
