/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.cache;

/**
 * CacheStats store hit, miss, and evictions of a cache.
 * It can also compute rates, and compute the sum/difference between two CacheStats
 */
public interface CacheStats
{

    /**
     * Returns the number of times cache lookup methods have returned either a cached or uncached value. 
     * This is defined as hitCount + missCount.
     * @return the number of times cache lookup methods have returned either a cached or uncached value.
     */
    public long requestCount();

    /**
     * Returns the number of times cache lookup methods have returned a cached value.
     * @return the number of times cache lookup methods have returned a cached value.
     */
    public long hitCount();

    /**
     * Returns the ratio of cache requests which were hits. This is defined as hitCount/requestCount. If requestCount equals 0, return 1
     * @return the ratio of cache requests which were hits.
     */
    public double hitRate();

    /**
     * Returns the number of times cache lookup methods have returned an uncached (newly loaded) value, or null.
     * @return missCount
     */
    public long missCount();

    /**
     * Returns the ratio of cache requests which were misses. This is defined as
     * missCount/requestCount. If requestCount equals 0, return 0
     * @return the ratio of cache requests which were misses.
     */
    public double missRate();
    
    /**
     * Returns the number of times an entry has been evicted. This count does not include manual invalidations.
     * @return the number of times an entry has been evicted.
     */
    public long evictionCount();
    
    /**
     * Returns the ratio of cache requests that lead to an eviction. This is defined as
     * evictionCount/requestCount. If requestCount equals 0, return 0
     * @return the ratio of cache requests that lead to an eviction.
     */
    public double evictionRate();

    /**
     * Returns a new CacheStats representing the difference between this CacheStats and other.
     * @param cacheStats the other CacheStats
     * @return the CacheStats representing the difference between this CacheStats and other.
     */
    public CacheStats minus(CacheStats cacheStats);
    
    /**
     * Returns a new CacheStats representing the sum between this CacheStats and other
     * @param cacheStats the other CacheStats
     * @return the new CacheStats representing the sum between this CacheStats and other.
     */
    public CacheStats plus(CacheStats cacheStats);
}
