/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.cache;

import java.util.Collections;
import java.util.List;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.quartz.JobDataMap;
import org.quartz.JobExecutionContext;

import org.ametys.core.ObservationConstants;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.schedule.Schedulable;
import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.plugins.core.impl.schedule.AbstractStaticSchedulable;
import org.ametys.plugins.core.schedule.Scheduler;

/**
 * A {@link Schedulable} job to clean caches.
 */
public class CleanCacheSchedulable extends AbstractStaticSchedulable
{
    /** The key for the cache Ids */
    public static final String JOBDATAMAP_CACHE_IDS = "cacheIds";

    /** The key for the all cache flag */
    public static final String JOBDATAMAP_ALL_CACHES = "allCaches";
    
    /** The cache manager */
    protected AbstractCacheManager _cacheManager;

    private ObservationManager _observationManager;

    private CurrentUserProvider _currentUserProvider;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _cacheManager = (AbstractCacheManager) manager.lookup(AbstractCacheManager.ROLE);
        _observationManager = (ObservationManager) manager.lookup(ObservationManager.ROLE);
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
    }
    
    @Override
    public void execute(JobExecutionContext context, ContainerProgressionTracker progressionTracker) throws Exception
    {
        JobDataMap jobDataMap = context.getJobDetail().getJobDataMap();
        List<String> cleanedCaches;
        
        String[] cacheIds = (String[]) jobDataMap.get(Scheduler.PARAM_VALUES_PREFIX + JOBDATAMAP_CACHE_IDS);
        if (cacheIds.length == 0)
        {
            cleanedCaches = _cacheManager.resetAllMemoryCaches();
        }
        else
        {
            cleanedCaches = _cacheManager.resetCaches(List.of(cacheIds));
        }

        for (String cacheId : cleanedCaches)
        {
            _observationManager.notify(new Event(ObservationConstants.EVENT_CACHE_RESET, 
                    _currentUserProvider.getUser(), 
                    Collections.singletonMap(ObservationConstants.ARGS_CACHE_ID, cacheId)));
            
        }
    }
}
