/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.captcha;

import java.util.Collection;
import java.util.List;
import java.util.regex.Pattern;

import org.apache.excalibur.source.SourceException;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * Interface for Captchas
 */
public interface Captcha
{    
    /**
     * Get the Captcha type label
     * @return the label
     */
    public I18nizableText getLabel();
    
    /**
     * Get the Captcha type id
     * @return the id
     */
    public String getId();
    
    /**
     * Get the list of config parameters referenced by this captcha implementation
     * @return list of parameters
     */
    public Collection<String> getConfigParameters();

    /**
     * Check if the Captcha is correct
     * @param key the key
     * @param value the value
     * @return true if correct
     */
    public boolean checkAndInvalidateCaptcha(String key, String value);

    /**
     * Get the helper xsl file
     * @return the source
     * @throws SourceException if an error occurred while reading the xsl file
     */
    public String getXSLHelperURL() throws SourceException;
    
    /**
     * Get the scss file
     * @return the source
     * @throws SourceException if an error occurred while reading the xsl file
     */
    public String getLoginSCSSURL() throws SourceException;

    /**
     * Get the label to display when login has failed with this captcha
     * @return the label
     */
    public I18nizableText getLoginFailedBecauseCaptchaFailedLabel();
    
    /**
     * Get the label to display when too many attempt have been done
     * @return the label
     */
    public I18nizableText getLoginFailedBecauseTooManyAttemptLabel();

    /**
     * Get the minimal url patterns used by the captcha implementation. 
     * This is necessary to grant those url even when the user is not authentified (in particular to set a catpcha on the login page). 
     * Do no put a too wide range (such as ^.*$) since it may introduce security issues.
     * @return The minimal url patterns used by the captcha implementation.
     */
    public List<Pattern> getUsedUrlPatterns();
    
    /**
     * Indicate if the captcha requires user interaction and should be displayed as a form entry
     * @return true if the captcha requires user interaction
     */
    public default boolean requireUserInteraction()
    {
        // Most implementation will require user interaction
        return true;
    }
}
