/*
 *  Copyright 2014 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.captcha;

import java.util.Collection;
import java.util.regex.Pattern;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.runtime.config.Config;

/**
 * Helper for generating a captcha regardless of its implementation
 */
public final class CaptchaHelper implements Serviceable, Initializable, Disposable
{
    /** The avalon role */
    public static final String ROLE = CaptchaHelper.class.getName();
    
    private static final String _CAPTCHA_TYPE_KEY = "runtime.captcha.type";
    private static final String _DEFAULT_CAPTCHA_IMPL = "jcaptcha";
    
    private static CaptchaExtensionPoint _captchaExtensionPoint;

    private static Captcha _captcha;
    
    public void initialize() throws Exception
    {
        _captcha = null;
    }
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _captchaExtensionPoint = (CaptchaExtensionPoint) manager.lookup(CaptchaExtensionPoint.ROLE);
    }

    /**
     * Retrieve the type of captcha used
     * @return The type of captcha.
     */
    public static String getCaptchaType()
    {
        if (_captcha != null)
        {
            return _captcha.getId();
        }
        else if (Config.getInstance() != null && _captchaExtensionPoint != null && _captchaExtensionPoint.hasExtension(Config.getInstance().getValue(_CAPTCHA_TYPE_KEY)))
        {
            return Config.getInstance().getValue(_CAPTCHA_TYPE_KEY);
        }
        else
        {
            return _DEFAULT_CAPTCHA_IMPL;
        }
    }

    /**
     * Retrieve the used captcha
     * @return The captcha.
     */
    public static Captcha getCaptcha()
    {
        if (_captcha != null)
        {
            return _captcha;
        }
        else if (_captchaExtensionPoint != null)
        {
            String key = Config.getInstance() != null ? Config.getInstance().getValue(_CAPTCHA_TYPE_KEY) : null;
            // In safe mod or if configured captcha does not exists, use default impl
            if (key == null || !_captchaExtensionPoint.hasExtension(key))
            {
                key = _DEFAULT_CAPTCHA_IMPL;
            }
            _captcha = _captchaExtensionPoint.getExtension(key);
            return _captcha;
        }
        else
        {
            throw new IllegalStateException("No captcha implementation found");
        }
    }
    
    /**
     * Check a captcha
     * @param key The captcha key. Can be empty or null when using reCaptcha.
     * @param value The value to check
     * @return <code>true</code> if the captcha is valid, false otherwise.
     */
    public static boolean checkAndInvalidate(String key, String value)
    {
        return getCaptcha().checkAndInvalidateCaptcha(key, value);
    }
    
    /**
     * Get the i18n key to use for the message to display when captcha verification has failed
     * @return The i18n key.
     */
    public static String getFailedI18NKey()
    {
        return getCaptcha().getLoginFailedBecauseCaptchaFailedLabel().getKey();
    }
    
    /**
     * Get the catalog to use for the message to display when captcha verification has failed
     * @return The catalog.
     */
    public static String getFailedI18NCatalog()
    {
        return getCaptcha().getLoginFailedBecauseCaptchaFailedLabel().getCatalogue();
    }
    
    /**
     * Get the i18n key to use for the message to display when too many connection attempts have been made
     * @return The key.
     */
    public static String getTooManyAttemptI18NKey()
    {
        return getCaptcha().getLoginFailedBecauseCaptchaFailedLabel().getKey();
    }
    
    /**
     * Get the catalog to use for the message to display when too many connection attempts have been made
     * @return The catalog.
     */
    public static String getTooManyAttemptI18NCatalog()
    {
        return getCaptcha().getLoginFailedBecauseCaptchaFailedLabel().getCatalogue();
    }

    /**
     * Get the minimal url patterns used by the captcha implementation. 
     * This is necessary to grant those url even when the user is not authentified (in particular to set a catpcha on the login page). 
     * Do no put a too wide range (such as ^.*$) since it may introduce security issues.
     * @return The minimal url patterns used by the captcha implementation.
     */
    public static Collection<Pattern> getUsedUrlPatterns()
    {
        return getCaptcha().getUsedUrlPatterns();
    }
    
    /**
     * Determine if the captcha provider requires user interaction
     * If not, no visible rendering should be done
     * @return true if user interaction are needed
     */
    public static boolean requiresUserInteraction()
    {
        return getCaptcha().requireUserInteraction();
    }
    
    @Override
    public void dispose()
    {
        staticDispose();
    }
    
    /**
     * Dispose the CaptchaHelper
     */
    public static void staticDispose()
    {
        _captcha = null;
    }
}
