/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.cocoon;

import java.util.Enumeration;
import java.util.Locale;
import java.util.Map;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.i18n.Bundle;

import org.ametys.core.DevMode;
import org.ametys.core.DevMode.DEVMODE;
import org.ametys.core.ObservationConstants;
import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.observation.Observer;
import org.ametys.core.util.I18nUtils;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This XMLResourceBundleFactory creates a <code>Bundle</code> for XML resources which can be invalidated
 */
public class XMLResourceBundleFactory extends org.apache.cocoon.i18n.XMLResourceBundleFactory implements Observer, Initializable
{
    /** bundle cache id */
    public static final String BUNDLE_CACHE = XMLResourceBundleFactory.class.getName() + "$bundle";

    private ObservationManager _observationManager;
    
    private AbstractCacheManager _cacheManager;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException 
    {
        super.service(smanager);
        
        if (smanager.hasService(ObservationManager.ROLE))
        {
            _observationManager = (ObservationManager) smanager.lookup(ObservationManager.ROLE);
        }
        
        _cacheManager = (AbstractCacheManager) manager.lookup(AbstractCacheManager.ROLE);
    }
    
    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        super.configure(configuration);
        
        // never reload in production, always reload in development
        interval = DevMode.getDeveloperMode() == DEVMODE.PRODUCTION ? -1 : 0;
    }
    
    public void initialize() throws Exception
    {
        if (_observationManager != null)
        {
            _observationManager.registerObserver(this);
        }
        
        _createCache();
    }
    
    @Override
    protected org.apache.cocoon.i18n.XMLResourceBundle create(String sourceURI, Locale locale, org.apache.cocoon.i18n.XMLResourceBundle parent)
    {
        if (getLogger().isDebugEnabled()) 
        {
            getLogger().debug("Creating bundle <" + sourceURI + ">");
        }

        XMLResourceBundle bundle = new XMLResourceBundle(sourceURI, locale, parent);
        bundle.enableLogging(getLogger());
        bundle.reload(manager, resolver, interval);
        return bundle;
    }
    
    /**
     * Invalidate catalogue 
     * @param location catalogue base location (URI)
     * @param name bundle name
     * @param localeName  locale name
     * @throws ComponentException If an error occurred
     */
    public void invalidateCatalogue (String location, String name, String localeName) throws ComponentException
    {
        String cacheKey = "XRB" + getCacheKey(new String[] {location}, 0, name, new Locale(localeName));
        cache.remove(cacheKey);
    }
    
    public int getPriority(Event event)
    {
        return MIN_PRIORITY;
    }
    
    public boolean supports(Event event)
    {
        return event.getId().equals(ObservationConstants.EVENT_CACHE_RESET)
               && I18nUtils.I18N_CACHE.equals(event.getArguments().get(ObservationConstants.ARGS_CACHE_ID));
    }
    
    public void observe(Event event, Map<String, Object> transientVars) throws Exception
    {
        // also clear the bundle cache when clearing the i18n cache 
        Enumeration<String> keys = cache.keys();
        
        while (keys.hasMoreElements())
        {
            String key = keys.nextElement();
            if (key.startsWith("XRB"))
            {
                cache.remove(key);
            }
        }
    }
    
    @Override
    public Bundle select(String[] directories, String name, Locale locale) throws ComponentException 
    {
        if (DevMode.getDeveloperMode() == DEVMODE.PRODUCTION)
        {
            return super.select(directories, name, locale);
        }
        else
        {
            Cache<String, Bundle> requestBundleCache = _getBundleCache();
            
            String cacheKey = getCacheKey(directories, 0, name, locale);
            if (requestBundleCache.hasKey(cacheKey))
            {
                return requestBundleCache.get(cacheKey);
            }
            else
            {
                Bundle select = super.select(directories, name, locale);
                requestBundleCache.put(cacheKey, select);
                return select;
            }
        }
    }
    
    /**
     * get the i18n cache (link language and I18nizable text to a translated value)
     * @return the i18n cache
     */
    protected Cache<String, Bundle> _getBundleCache()
    {
        return _cacheManager.get(BUNDLE_CACHE);
    }
    
    /**
     * Create the i18n cache
     */
    protected void _createCache()
    {
        _cacheManager.createRequestCache(BUNDLE_CACHE, 
                new I18nizableText("plugin.core", "PLUGINS_CORE_I18N_BUNDLECACHE_LABEL"),
                new I18nizableText("plugin.core", "PLUGINS_CORE_I18N_BUNDLECACHE_DESCRIPTION"),
                true);
    }
}
