/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.datasource;

import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Map;
import java.util.stream.Stream;

/**
 * Interface for entities that use data sources
 */
public interface DataSourceConsumer
{
    /**
     * The type of use of a data source
     */
    public enum TypeOfUse
    {
        /** The data source is in use and in a blocking way : the application must start in safe-mode if not available  */
        BLOCKING,
        /** The data source is in use but in a non-blocking way */
        NON_BLOCKING,
        /** The data source is not used */
        NOT_USED;
        
       /**
        * Merge several types of use to keep only the most important
        * @param typesOfUse The types of use to merge. Cannot be null but can be empty.
        * @return The most important type of use
        */
        public static TypeOfUse merge(Collection<TypeOfUse> typesOfUse)
        {
            return _merge(typesOfUse.stream());
        }
        
        /**
         * Merge several types of use to keep only the most important
         * @param typesOfUse The types of use to merge. Cannot be null but can be empty.
         * @return The most important type of use
         */
        public static TypeOfUse merge(TypeOfUse... typesOfUse)
        {
            return _merge(Arrays.stream(typesOfUse));
        }
        
        private static TypeOfUse _merge(Stream<TypeOfUse> typesOfUse)
        {
            return typesOfUse.filter(r -> r != null)
                    .min(Comparator.naturalOrder())
                    .orElse(TypeOfUse.NOT_USED);
        }
    }
    
    /**
     * Determines if a data source is used
     * @param id The id of data source to check
     * @return true if the data source is currently in use
     */
    public TypeOfUse isInUse(String id);
    
    /**
     * Retrieve the ids of the used data sources
     * @return the set of ids of the data source(s) used by this data source customer
     */
    public Map<String, TypeOfUse> getUsedDataSourceIds();
}
