/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.datasource;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.ametys.core.datasource.DataSourceConsumer.TypeOfUse;
import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentExtensionPoint;

/**
 * This class is in charge to load and initialize the various {@link DataSourceConsumer} 
 */
public class DataSourceConsumerExtensionPoint extends AbstractThreadSafeComponentExtensionPoint<DataSourceConsumer>
{
    /** Avalon Role */
    public static final String ROLE = DataSourceConsumerExtensionPoint.class.getName();

    /**
     * Determines if a data source is in use by a {@link DataSourceConsumer}
     * @param id The id of the data source to check
     * @return true if the data source is used by at least one of the {@link DataSourceConsumer}
     */
    public TypeOfUse isInUse(String id)
    {
        Set<TypeOfUse> typesOfUse = new HashSet<>();
        
        for (String extensionId : getExtensionsIds())
        {
            DataSourceConsumer dataSourceClient = getExtension(extensionId);
            typesOfUse.add(dataSourceClient.isInUse(id));
        }
        
        return TypeOfUse.merge(typesOfUse);
    }
    
    /**
     * Retrieve the ids of the used data sources
     * @return the set of ids of data source used by all of the {@link DataSourceConsumer}
     */
    public Map<String, TypeOfUse> getUsedDataSourceIds()
    {
        Set<String> extensionsIds = getExtensionsIds();
        Map<String, TypeOfUse> usedDataSourceIds = new HashMap<>();
        
        for (String extensionId : extensionsIds)
        {
            DataSourceConsumer dataSourceClient = getExtension(extensionId);
            for (Entry<String, TypeOfUse> usedDataSource : dataSourceClient.getUsedDataSourceIds().entrySet())
            {
                if (usedDataSourceIds.containsKey(usedDataSource.getKey()))
                {
                    usedDataSourceIds.put(usedDataSource.getKey(), TypeOfUse.merge(usedDataSourceIds.get(usedDataSource.getKey()), usedDataSource.getValue()));
                }
                else
                {
                    usedDataSourceIds.put(usedDataSource.getKey(), usedDataSource.getValue());
                }
            }
        }
        
        return usedDataSourceIds;
    }
}
