/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.datasource;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.apache.commons.collections.MapUtils;
import org.xml.sax.SAXException;

import org.ametys.plugins.core.impl.checker.LDAPConnectionChecker;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.checker.ItemChecker;
import org.ametys.runtime.model.checker.ItemCheckerTestFailureException;
import org.ametys.runtime.util.AmetysHomeHelper;

/**
 * This component handles SQL data sources. 
 * It is associated with the configuration file $AMETYS_HOME/config/datasources-ldap.xml 
 */
public class LDAPDataSourceManager extends AbstractDataSourceManager
{
    /** Avalon Role */
    public static final String ROLE = LDAPDataSourceManager.class.getName();
    
    /** LDAP parameter's name for base URL */
    public static final String PARAM_BASE_URL = "baseURL";
    /** LDAP parameter's name for base DN */
    public static final String PARAM_BASE_DN = "baseDN";
    /** LDAP parameter's name for use SSL */
    public static final String PARAM_USE_SSL = "useSSL";
    /** LDAP parameter's name for alias dereferencing */
    public static final String PARAM_ALIAS_DEREFERENCING = "aliasDereferencing";
    /** LDAP parameter's name for follow referrals property */
    public static final String PARAM_FOLLOW_REFERRALS = "followReferrals";
    /** LDAP parameter's name for authentication method */
    public static final String PARAM_AUTHENTICATION_METHOD = "authenticationMethod";
    /** LDAP parameter's name for administrator DN */
    public static final String PARAM_ADMIN_DN = "adminDN";
    /** LDAP parameter's name for administration password */
    public static final String PARAM_ADMIN_PASSWORD = "adminPassword";
    /** LDAP parameter's name for enabling server side sorting */
    public static final String PARAM_SERVER_SIDE_SORTING = "serverSideSorting";

    /** The id of the internal DataSource */
    public static final String LDAP_DATASOURCE_PREFIX = "LDAP-";
    
    private static String __filename;
    
    /**
     * Set the config filename. Only use for tests.
     * @param filename Name with path of the config file
     */
    public static void setFilename(String filename)
    {
        __filename = filename;
    }
    
    @Override
    public File getFileConfiguration()
    {
        if (__filename != null)
        {
            return new File(__filename);
        }
        
        return new File(AmetysHomeHelper.getAmetysHomeConfig(), "datasources-ldap.xml");
    }
    
    @Override
    protected String getDataSourcePrefixId()
    {
        return LDAP_DATASOURCE_PREFIX;
    }
    
    @Override
    protected Map<String, DataSourceDefinition> doReadConfiguration(File file)
    {
        Map<String, DataSourceDefinition> definitions = new HashMap<>();
        
        try
        {
            if (file.exists())
            {
                Configuration configuration = new DefaultConfigurationBuilder().buildFromFile(file);
                for (Configuration dsConfig : configuration.getChildren("datasource"))
                {
                    String id = dsConfig.getAttribute("id");
                    
                    I18nizableText name = I18nizableText.parseI18nizableText(dsConfig.getChild("name"), "plugin.core");
                    I18nizableText description = I18nizableText.parseI18nizableText(dsConfig.getChild("description"), "plugin.core", "");
                    
                    boolean isPrivate = dsConfig.getAttributeAsBoolean("private", false);
                    boolean isDefault = dsConfig.getAttributeAsBoolean("default", false);
                    
                    Map<String, Object> parameters = new HashMap<>();
                    
                    Configuration[] paramsConfig = dsConfig.getChild("parameters").getChildren();
                    for (Configuration paramConfig : paramsConfig)
                    {
                        String paramName = paramConfig.getName();
                        Object value;
                        
                        if (paramName.equals(PARAM_USE_SSL) || paramName.equals(PARAM_FOLLOW_REFERRALS) || paramName.equals(PARAM_SERVER_SIDE_SORTING))
                        {
                            value = paramConfig.getValueAsBoolean(false);
                        }
                        else
                        {
                            value = paramConfig.getValue("");
                        }
                        
                        parameters.put(paramConfig.getName(), value);
                    }
                    
                    DataSourceDefinition dataSource = new DataSourceDefinition(id, name, description, parameters, isPrivate, isDefault);
                    definitions.put(id, dataSource);
                }
            }
            
            return definitions;
        }
        catch (IOException | ConfigurationException | SAXException e)
        {
            throw new RuntimeException("Unable to parse datasource configuration file.", e);
        }
    }
    
    @Override
    public void checkParameters(Map<String, Object> parameters) throws ItemCheckerTestFailureException
    {
        // Order the parameters
        List<String> values = new ArrayList<> ();
        values.add(parameters.get(PARAM_BASE_URL).toString());
        values.add(parameters.get(PARAM_AUTHENTICATION_METHOD).toString());
        values.add(parameters.get(PARAM_ADMIN_DN).toString());
        values.add(parameters.get(PARAM_ADMIN_PASSWORD).toString());
        values.add(parameters.get(PARAM_USE_SSL).toString());
        values.add(parameters.get(PARAM_FOLLOW_REFERRALS).toString());
        values.add(parameters.get(PARAM_BASE_DN).toString());
        
        ItemChecker paramChecker = new LDAPConnectionChecker();
        paramChecker.check(values);
    }
    
    @Override
    protected void createDataSource(DataSourceDefinition dataSource)
    {
        // Empty
    }
    
    @Override
    protected void deleteDataSource(DataSourceDefinition dataSource)
    {
        // Empty
    }

    @Override
    protected void editDataSource(DataSourceDefinition dataSource)
    {
        // Empty
    }
    
    @Override
    protected void internalSetDefaultDataSource()
    {
        // Arbitrarily set the LDAP default data source
        if (MapUtils.isNotEmpty(_dataSourcesDef))
        {
            DataSourceDefinition defaultDataSourceDef = _dataSourcesDef.values().iterator().next();
            defaultDataSourceDef.setDefault(true);
            _dataSourcesDef.put(defaultDataSourceDef.getId(), defaultDataSourceDef);
            
            saveConfiguration();
            editDataSource(defaultDataSourceDef);
            
        }
    }
}
