/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.datasource.dbtype;

import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * Interface for the SQL database types
 */
public interface SQLDatabaseType
{
    /**
     * Get the id of the database type.
     * @return the id. Can not be null.
     */
    public String getId();
    
    /**
     * Get the label of the database type.
     * @return the label
     */
    public I18nizableText getLabel();
    
    /**
     * Get the driver of the database type.
     * @return the driver. Can not be null.
     */
    public String getDriver();
    
    /**
     * Get the query to validate a connection
     * @return A SQL query to ensure that a connection is still alive
     */
    public String getValidationQuery();
    
    /**
     * Get the url template to use for this database type
     * @return the url template
     */
    public String getTemplate();
    
    /**
     * Get the error message key when the specified driver was not found
     * @return the error message 
     */
    public I18nizableText getDriverNotFoundMessage();
    
    /**
     * Some db systems requires the table names to be escaped
     * @param tableNameToEscape The non-null table name
     * @return The escaped table name or at least the table name if no escape is required
     */
    public String languageEscapeTableName(String tableNameToEscape);
    
    /**
     * Add a limit/offset element to the given query
     * @param queryToLimit The sql query that will be amended
     * @param limit The max number of the results to return
     * @param offset The initial offset of the results to return
     * @return The initial query amended to handle the limit and offset.
     */
    public String languageLimitQuery(String queryToLimit, String limit, String offset);
    
    /**
     * Get the InputStream to read a blob
     * @param resultSet the result set containing the blob.
     * @param columnName column
     * @return an InputStream to read the blob or null if the column is null
     * @throws SQLException if the columnLabel is not valid; if a database access error occurs or this method is called on a closed result set
     */
    public InputStream getBlob(ResultSet resultSet, String columnName) throws SQLException;
    
    /**
     * Get the InputStream to read a blob
     * @param resultSet the result set containing the blob.
     * @param pos column position
     * @return an InputStream to read the blob
     * @throws SQLException if the columnLabel is not valid; if a database access error occurs or this method is called on a closed result set
     */
    public InputStream getBlob(ResultSet resultSet, int pos) throws SQLException;
    
    /**
     * Set an String into a blob
     * @param statement The satement where the blob will be set
     * @param pos position in the statement
     * @param blob String representing the blob. Can be null
     * @throws SQLException if the columnLabel is not valid; if a database access error occurs or this method is called on a closed result set
     * @throws UnsupportedEncodingException if UTF-8 is not supported
     */
    public void setBlob(PreparedStatement statement, int pos, String blob) throws SQLException, UnsupportedEncodingException;

    /**
     * Set an array of bytes (UTF-8 encoded) into a blob, if you want to pass a String, you can use {@link #setBlob(PreparedStatement, int, String)}
     * @param statement The satement where the blob will be set
     * @param pos position in the statement
     * @param bytes byte[] representing the blob, in UTF-8. Can be null
     * @throws SQLException if the columnLabel is not valid; if a database access error occurs or this method is called on a closed result set
     */
    public void setBlob(PreparedStatement statement, int pos, byte[] bytes) throws SQLException;

    /**
     * Set an array of bytes (UTF-8 encoded) into a blob, if you want to pass a String, you can use {@link #setBlob(PreparedStatement, int, String)}
     * @param statement The satement where the blob will be set
     * @param pos position in the statement
     * @param is inputStream to put in the blob. Can be null
     * @param length length of the stream
     * @throws SQLException if the columnLabel is not valid; if a database access error occurs or this method is called on a closed result set
     */
    public void setBlob(PreparedStatement statement, int pos, InputStream is, long length) throws SQLException;
}

