/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.datasource.dbtype;

import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentExtensionPoint;

/**
 * This class is in charge to load and initialize the {@link SQLDatabaseType} 
 */
public class SQLDatabaseTypeExtensionPoint extends AbstractThreadSafeComponentExtensionPoint<SQLDatabaseType>
{
    /** Avalon Role */
    public static final String ROLE = SQLDatabaseTypeExtensionPoint.class.getName();
    
    @Override
    public void initializeExtensions() throws Exception
    {
        super.initializeExtensions();
        
        for (String databaseTypeId : getExtensionsIds())
        {
            SQLDatabaseType databaseType = getExtension(databaseTypeId);
            
            try
            {
                Class.forName(databaseType.getDriver());
            }
            catch (ClassNotFoundException e)
            {
                getLogger().warn("JDBC Driver cannot be found for extension '" + databaseTypeId + "' with classname '" + databaseType.getDriver() + "'");
            }
        }
    }
    
    /**
     * Get the SQL database types with their label
     * @return the SQL database types with their label
     */
    public Map<String, I18nizableText> getSQLDatabaseTypes()
    {
        Map<String, I18nizableText> db = new HashMap<>();
        
        Set<String> extensionsIds = getExtensionsIds();
        for (String extensionId : extensionsIds)
        {
            SQLDatabaseType dbType = getExtension(extensionId);
            db.put(extensionId, dbType.getLabel());
        }
        
        return db;
    }
    
    /**
     * Some db systems requires the table names to be escaped
     * @param dbType The identifier of the db type used
     * @param tableNameToEscape The non-null table name
     * @return The escaped table name or at least the table name if no escape is required. Or null if the dbtype is unknown.
     */
    public String languageEscapeTableName(String dbType, String tableNameToEscape)
    {
        if (hasExtension(dbType))
        {
            return getExtension(dbType).languageEscapeTableName(tableNameToEscape);
        }
        else
        {
            return null;
        }
    }
    
    /**
     * Add a limit/offset element to the given query
     * @param dbType The identifier of the db type used
     * @param queryToLimit The sql query that will be amended
     * @param limit The max number of the results to return
     * @param offset The initial offset of the results to return
     * @return The initial query amended to handle the limit and offset. Or null if the dbtype is unknown.
     */
    public String languageLimitQuery(String dbType, String queryToLimit, String limit, String offset)
    {
        if (hasExtension(dbType))
        {
            return getExtension(dbType).languageLimitQuery(queryToLimit, limit, offset);
        }
        else
        {
            return null;
        }
    }
    
    /**
     * Get the InputStream to read a blob
     * @param dbType The identifier of the db type used
     * @param resultSet the result set containing the blob.
     * @param columnName column
     * @return an InputStream to read the blob
     * @throws SQLException if the columnLabel is not valid; if a database access error occurs or this method is called on a closed result set
     */
    public InputStream getBlob(String dbType, ResultSet resultSet, String columnName) throws SQLException
    {
        if (hasExtension(dbType))
        {
            return getExtension(dbType).getBlob(resultSet, columnName);
        }
        else
        {
            return null;
        }
    }
    
    /**
     * Get the InputStream to read a blob
     * @param dbType The identifier of the db type used
     * @param resultSet the result set containing the blob.
     * @param pos column position
     * @return an InputStream to read the blob
     * @throws SQLException if the columnLabel is not valid; if a database access error occurs or this method is called on a closed result set
     */
    public InputStream getBlob(String dbType, ResultSet resultSet, int pos) throws SQLException
    {
        if (hasExtension(dbType))
        {
            return getExtension(dbType).getBlob(resultSet, pos);
        }
        else
        {
            return null;
        }
    }
    
    /**
     * Set an String into a blob
     * @param dbType The identifier of the db type used
     * @param statement The satement where the blob will be set
     * @param pos position in the statement
     * @param blob String representing the blob
     * @throws SQLException if the columnLabel is not valid; if a database access error occurs or this method is called on a closed result set
     * @throws UnsupportedEncodingException if UTF-8 is not supported
     * @throws IllegalArgumentException if the dbType is not found
     */
    public void setBlob(String dbType, PreparedStatement statement, int pos, String blob) throws SQLException, UnsupportedEncodingException, IllegalArgumentException
    {
        if (hasExtension(dbType))
        {
            getExtension(dbType).setBlob(statement, pos, blob);
        }
        else
        {
            throw new IllegalArgumentException("Database of type '" + dbType + "' is not supported");
        }
    }
    
    /**
     * Set an array of bytes (UTF-8 encoded) into a blob, if you want to pass a String, you can use {@link #setBlob(String, PreparedStatement, int, String)}
     * @param dbType The identifier of the db type used
     * @param statement The satement where the blob will be set
     * @param pos position in the statement
     * @param bytes byte[] representing the blob, in UTF-8
     * @throws SQLException if the columnLabel is not valid; if a database access error occurs or this method is called on a closed result set
     * @throws IllegalArgumentException if the dbType is not found
     */
    public void setBlob(String dbType, PreparedStatement statement, int pos, byte[] bytes) throws SQLException, IllegalArgumentException
    {
        if (hasExtension(dbType))
        {
            getExtension(dbType).setBlob(statement, pos, bytes);
        }
        else
        {
            throw new IllegalArgumentException("Database of type '" + dbType + "' is not supported");
        }
    }
    

    
    /**
     * Set an array of bytes (UTF-8 encoded) into a blob, if you want to pass a String, you can use {@link #setBlob(String, PreparedStatement, int, String)}
     * @param dbType The identifier of the db type used
     * @param statement The satement where the blob will be set
     * @param pos position in the statement
     * @param is inputStream to put in the blob
     * @param length length of the stream
     * @throws SQLException if the columnLabel is not valid; if a database access error occurs or this method is called on a closed result set
     * @throws IllegalArgumentException if the dbType is not found
     */
    public void setBlob(String dbType, PreparedStatement statement, int pos, InputStream is, long length) throws SQLException, IllegalArgumentException
    {
        if (hasExtension(dbType))
        {
            getExtension(dbType).setBlob(statement, pos, is, length);
        }
        else
        {
            throw new IllegalArgumentException("Database of type '" + dbType + "' is not supported");
        }
    }
}
