/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.engine;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.util.Map;

import org.apache.avalon.framework.logger.Logger;
import org.apache.cocoon.environment.AbstractEnvironment;
import org.apache.cocoon.environment.Context;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.commandline.CommandLineResponse;
import org.apache.cocoon.util.NullOutputStream;

/**
 * A simple implementation of <code>org.apache.cocoon.environment.Environment</code> 
 * for pipeline calls which are not externally triggered.
 */
public class BackgroundEnvironment extends AbstractEnvironment 
{
    /**
     * Constructs the environment.
     * @param logger a logger for traces.
     * @param ctx the application {@link Context}.
     * @throws MalformedURLException if the context is ill-formed
     */
    public BackgroundEnvironment(Logger logger, Context ctx) throws MalformedURLException 
    {
        super("", null, new File(ctx.getRealPath("/")), null);
        enableLogging(logger);

        this.outputStream = new NullOutputStream();

        Request request = new BackgroundRequest(this);
        
        this.objectModel.put(ObjectModelHelper.REQUEST_OBJECT, request);
        this.objectModel.put(ObjectModelHelper.RESPONSE_OBJECT, new CommandLineResponse());
        this.objectModel.put(ObjectModelHelper.CONTEXT_OBJECT, ctx);
    }

    /**
     * Constructs the environment.
     * @param uri the processed uri.
     * @param requestParameters request parameters.
     * @param requestAttributes request attributes.
     * @param requestHeaders request headers.
     * @param contextPath the application context path.
     * @param ctx the application {@link Context}.
     * @param logger a logger for traces.
     * @param os the {@link OutputStream} to send the response to.
     * @throws MalformedURLException if the context is ill-formed
     */
    public BackgroundEnvironment(String uri, Map requestParameters, Map requestAttributes, Map requestHeaders, String contextPath, Context ctx, Logger logger, OutputStream os) throws MalformedURLException
    {
        super(uri, null, new File(ctx.getRealPath("/")), null);
        enableLogging(logger);
        
        Request request = new BackgroundRequest(this, contextPath, requestAttributes, requestParameters, requestHeaders);
        
        objectModel.put(ObjectModelHelper.REQUEST_OBJECT, request);
        objectModel.put(ObjectModelHelper.RESPONSE_OBJECT, new CommandLineResponse());
        objectModel.put(ObjectModelHelper.CONTEXT_OBJECT, ctx);
        
        outputStream = os;
    }
    
    public void redirect(boolean sessionmode, String newURL) throws IOException 
    {
        // nothing to do
    }

    public void setContentType(String mimeType) 
    {
        // nothing to do
    }

    public String getContentType() 
    {
        return null;
    }

    public void setContentLength(int length)
    {
        // nothing to do
    }

    public boolean isExternal()
    {
        return false;
    }
}
