/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.group;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;
import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.group.directory.GroupDirectory;
import org.ametys.core.ui.Callable;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Helper for associating {@link GroupDirectory}(ies) to contexts.
 */
public class GroupDirectoryContextHelper extends AbstractLogEnabled implements Component, Serviceable, Initializable
{
    /** Avalon Role */
    public static final String ROLE = GroupDirectoryContextHelper.class.getName();
    
    /** The "admin" context */
    public static final String ADMIN_CONTEXT = "/admin";
    
    /** The name of the JDBC table for group directories by context */
    private static final String __GROUP_DIRECTORIES_TABLE = "GroupDirectoriesByContext";

    private static final String __GROUP_DIRECTORIES_BY_CONTEXT_CACHE = "groupDirectoriesByContext";
    
    private static final String IS_CACHE_FILLED = "###iscachefilled###";

    
    /** The DAO for {@link GroupDirectory}(ies) */
    private GroupDirectoryDAO _groupDirectoryDAO;

    private AbstractCacheManager _abstractCacheManager;

    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _groupDirectoryDAO = (GroupDirectoryDAO) manager.lookup(GroupDirectoryDAO.ROLE);
        _abstractCacheManager = (AbstractCacheManager) manager.lookup(AbstractCacheManager.ROLE);
    }
    
    @Override
    public void initialize() throws Exception
    {
        _abstractCacheManager.createMemoryCache(__GROUP_DIRECTORIES_BY_CONTEXT_CACHE, 
                new I18nizableText("plugin.core", "PLUGINS_CORE_CACHE_GROUP_DIRECTORY_CONTEXT_LABEL"),
                new I18nizableText("plugin.core", "PLUGINS_CORE_CACHE_GROUP_DIRECTORY_CONTEXT_DESCRIPTION"),
                true,
                null);
    }
    
    /**
     * Get the connection to the database 
     * @return the SQL connection
     */
    protected Connection getSQLConnection ()
    {
        String datasourceId = Config.getInstance().getValue("runtime.assignments.groups");
        return ConnectionHelper.getConnection(datasourceId);
    }
    
    /**
     * Links given group directories to a context.
     * @param context The context
     * @param ids The ids of the group directories to link
     * @return The ids of the linked group directories
     */
    @SuppressWarnings("resource")
    @Callable (rights = "Runtime_Rights_Admin_Access", context = "/admin")
    public Set<String> link(String context, List<String> ids)
    {
        Set<String> result = getGroupDirectoriesOnContext(context);
        
        Connection connection = null;
        PreparedStatement stmt = null;
        try
        {
            connection = getSQLConnection();
            
            // Remove all the ids affected to this context
            String sql = "DELETE FROM " + __GROUP_DIRECTORIES_TABLE + " WHERE Context=?";
            stmt = connection.prepareStatement(sql);
            stmt.setString(1, context);
            stmt.executeUpdate();
            getLogger().info("{}\n[{}]", sql, context);
            
            // Set the new ids to this context
            sql = "INSERT INTO " + __GROUP_DIRECTORIES_TABLE + " (Context, GroupDirectory_Id) VALUES(?, ?)";
            stmt = connection.prepareStatement(sql);
            for (String id : ids)
            {
                if (_groupDirectoryDAO.getGroupDirectory(id) != null)
                {
                    stmt.setString(1, context);
                    stmt.setString(2, id);
                    stmt.executeUpdate();
                    
                    getLogger().info("{}\n[{}, {}]", sql, context, id);
                    if (result.contains(id))
                    {
                        result.remove(id); // the group directory was already linked, so its status didn't changed
                    }
                    else
                    {
                        result.add(id);
                    }
                }
                else
                {
                    getLogger().warn("The GroupDirectory with id '{}' does not exist. It will not be linked.", id);
                }
            }            
        }
        catch (SQLException e)
        {
            getLogger().error("Error in sql query", e);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
            ConnectionHelper.cleanup(stmt);
            _getCache().invalidateAll();
        }
        
        return result;
    }
    
    /**
     * Gets the group directories linked to the given context
     * @param context The context
     * @return The ids of group directories linked to the context
     */
    @Callable (rights = "Runtime_Rights_Admin_Access", context = "/admin")
    public Set<String> getGroupDirectoriesOnContext(String context)
    {
        if (ADMIN_CONTEXT.equals(context))
        {
            // Return all the group directories
            return _groupDirectoryDAO.getGroupDirectoriesIds();
        }
        else
        {
            return _getDirectoriesOnContextFromDatabase(context);
        }
    }
    
    /**
     * Gets the group directories linked to at least one of the given contexts
     * @param contexts The contexts
     * @return The ids of group directories linked to the contexts
     */
    @Callable (rights = "Runtime_Rights_Admin_Access", context = "/admin")
    public Set<String> getGroupDirectoriesOnContexts(Set<String> contexts)
    {
        return contexts.stream()
                .map(this::getGroupDirectoriesOnContext)
                .flatMap(Set::stream)
                .collect(Collectors.toSet());
    }
    
    private Set<String> _getDirectoriesOnContextFromDatabase(String context)
    {
        return _getCacheAsMap().computeIfAbsent(context, str -> new HashSet<>());
    }
    
    /**
     * create a group directory map from the database
     * @return Non null map of group directory content
     */
    private synchronized Map<String, Set<String>> _getAllGroupDirectoryContent()
    {
        Map<String, Set<String>> cache = new HashMap<>();
        
        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try
        {
            connection = getSQLConnection();
            
            String sql = "SELECT Context, GroupDirectory_Id FROM " + __GROUP_DIRECTORIES_TABLE;
            stmt = connection.prepareStatement(sql);
            rs = stmt.executeQuery();
            
            while (rs.next())
            {
                String context = rs.getString(1);
                String groupDirectoryId = rs.getString(2);
                
                if (_groupDirectoryDAO.getGroupDirectory(groupDirectoryId) == null)
                {
                    getLogger().warn("The group directory of id '{}' is linked to a context, but does not exist anymore.", groupDirectoryId);
                }
                else
                {
                    Set<String> cacheForContext = cache.computeIfAbsent(context, c -> new HashSet<>());
                    
                    cacheForContext.add(groupDirectoryId);
                }
            }
        }
        catch (SQLException e)
        {
            throw new IllegalStateException("Cannot get group directories and contexts association because of an error in a sql query", e);
            
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(rs);
        }
        return cache;
    }
    
    /**
     * Get the cache
     * @return Non null cache
     */
    protected synchronized Cache<String, Set<String>> _getCache()
    {
        Cache<String, Set<String>> cache = _abstractCacheManager.get(__GROUP_DIRECTORIES_BY_CONTEXT_CACHE); 
        
        if (!cache.hasKey(IS_CACHE_FILLED))
        {
            Map<String, Set<String>> cacheMap = _getAllGroupDirectoryContent();
            cache.putAll(cacheMap);
            cache.put(IS_CACHE_FILLED, null);
        }
        return cache;
    }
    
    /**
     * Get the cache as map
     * @return Non null cache
     */
    protected synchronized Map<String, Set<String>> _getCacheAsMap()
    {
        Map<String, Set<String>> cacheAsMap = _getCache().asMap();
        cacheAsMap.remove(IS_CACHE_FILLED);
        return cacheAsMap;
    }
}
