/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.mbean;

import java.util.List;

import javax.management.MBeanAttributeInfo;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;

/**
 * MBean providing information about Ametys cache usage
 */
public class CacheUsageMBean extends AbstractAmetysDynamicMBean implements Serviceable
{
    /** the ROLE of the component */
    public static final String ROLE = CacheUsageMBean.class.getName();
    /** The cache manager */
    AbstractCacheManager _cacheManager;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _cacheManager = (AbstractCacheManager) manager.lookup(AbstractCacheManager.ROLE);
    }
    
    /**
     * Get the total size of all Ametys cache
     * @return the cache size
     */
    public long getTotalCacheSize()
    {
        long sum = _cacheManager.getAllCaches()
                .values()
                .stream()
                .flatMap(List::stream)
                .filter(Cache::isComputableSize)
                .mapToLong(Cache::getMemorySize)
                .sum();
        return sum;
    }
    
    @Override
    protected Object _getAttribute(String attributeName)
    {
        // Call the corresponding getter for a recognized attribute_name
        if (attributeName.equals("TotalCacheSize"))
        {
            return FileUtils.byteCountToDisplaySize(getTotalCacheSize());
        }
        if (attributeName.equals("TotalCacheSizeRaw"))
        {
            return getTotalCacheSize();
        }
        return null; 
    }

    @Override
    protected String getMBeanName()
    {
        return StringUtils.substringBefore(StringUtils.substringAfterLast(ROLE, "."), "MBean");
    }

    @Override
    protected String getMBeanDescription()
    {
        return "Cache Memory Usage";
    }

    @Override
    protected MBeanAttributeInfo[] getMBeanAttributes()
    {
        MBeanAttributeInfo[] attributes = new MBeanAttributeInfo[2];
        attributes[0] = new MBeanAttributeInfo("TotalCacheSize", "String", "Total size of the different Ametys cache", true, false, false);
        attributes[1] = new MBeanAttributeInfo("TotalCacheSizeRaw", "float", "Total size of the different Ametys cache in byte", true, false, false);
        return attributes;
    }
}
