/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration;

import java.time.Instant;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.Constants;
import org.apache.cocoon.util.log.SLF4JLoggerAdapter;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;

import org.ametys.core.ObservationConstants;
import org.ametys.core.engine.BackgroundEngineHelper;
import org.ametys.core.migration.action.Action;
import org.ametys.core.migration.action.ActionConfiguration;
import org.ametys.core.migration.action.ActionExtensionPoint;
import org.ametys.core.migration.version.Version;
import org.ametys.core.migration.version.VersionConfiguration;
import org.ametys.core.migration.version.handler.VersionHandler;
import org.ametys.core.migration.version.storage.VersionStorage;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;
import org.ametys.runtime.plugin.component.DeferredServiceable;
import org.ametys.runtime.servlet.RuntimeServlet;
import org.ametys.runtime.servlet.RuntimeServlet.MaintenanceStatus;

/**
 * Main entry point for all automatic migration related tasks.
 */
public class MigrationEngine extends AbstractLogEnabled implements Contextualizable, DeferredServiceable, Component
{
    /** Avalon Role */
    public static final String ROLE = MigrationEngine.class.getName();
    
    private MigrationExtensionPoint _migrationExtensionPoint;
    private MigrationExtensionPoint _migrationDataExtensionPoint;
    
    private ActionExtensionPoint _upgradeEP;
    private ActionExtensionPoint _initializationEP;

    private ObservationManager _observationManager;

    private ServiceManager _manager;

    private org.apache.cocoon.environment.Context _context;
    
    private ActionData _failedAction;
    private MigrationException _failedException;
    
    public void contextualize(Context context) throws ContextException
    {
        _context = (org.apache.cocoon.environment.Context) context.get(Constants.CONTEXT_ENVIRONMENT_CONTEXT);
    }
    
    public void deferredService(ServiceManager smanager) throws ServiceException
    {
        _manager = smanager;
        
        _migrationExtensionPoint = (MigrationExtensionPoint) smanager.lookup(MigrationExtensionPoint.ROLE);
        _migrationDataExtensionPoint = (MigrationExtensionPoint) smanager.lookup(MigrationExtensionPoint.ROLE + "/internal");
        
        _upgradeEP = (ActionExtensionPoint) smanager.lookup(ActionExtensionPoint.ROLE_UPGRADE);
        _initializationEP = (ActionExtensionPoint) smanager.lookup(ActionExtensionPoint.ROLE_INITIALIZATION);
        
        if (smanager.hasService(ObservationManager.ROLE))
        {
            _observationManager = (ObservationManager) smanager.lookup(ObservationManager.ROLE);
        }
    }
    
    /**
     * All migrations are parsed and executed one per one in the right order
     * @return true if the migration requires a server restart ; false otherwise
     */
    public boolean migrate()
    {
        getLogger().info("Automatic migration start");
        try
        {
            boolean hasToRestart = _doMigration(_migrationDataExtensionPoint);
            
            if (hasToRestart)
            {
                return true;
            }
            
            hasToRestart = _doMigration(_migrationExtensionPoint);
            
            if (hasToRestart)
            {
                return true;
            }
            
            _notifyEndOfMigration();
          
            getLogger().info("Automatic migration finished.");
        }
        catch (MigrationException e)
        {
            RuntimeServlet.setMaintenanceStatus(MaintenanceStatus.AUTOMATIC, null);
            getLogger().error("Error during the automatic migration", e);
        }
        
        return false;
    }

    private boolean _doMigration(MigrationExtensionPoint extPoint) throws MigrationException
    {
        boolean versionDeterminationFailed = false;
        List<ActionData> allInitActions = new ArrayList<>(); 
        List<ActionData> allUpgradeActions = new ArrayList<>();
        
        // First gather all versions to init and upgrade from all migration components
        for (String componentId : extPoint.getExtensionsIds())
        {
            MigrationComponent component = extPoint.getExtension(componentId);
            
            try
            {
                Versions versions = getVersions(component);
                
                List<VersionList> versionLists = _getVersionLists(versions);
                
                for (VersionList list : versionLists)
                {
                    Version versionToUpgrade = null;
                    
                    if (list.versions().isEmpty())
                    {
                        // no versions, the associated component has to be initialized
                        Version newVersion = component.versionStorage().createVersion(list.componentId(), component, list.storageConfiguration(), list.additionalValues());

                        if (newVersion.getVersionNumber() == null)
                        {
                            allInitActions.add(new ActionData(newVersion, getHighestUpgradeVersionNumber(component), component.initialization(), list.id()));
                        }
                        else
                        {
                            // non null versionNumer, meaning we are using it as our current version
                            // let's store it and add it to the upgrade flow 
                            _createAndStoreVersion(new ActionData(newVersion, newVersion.getVersionNumber(), null, list.id()), "Initial version");
                            versionToUpgrade = newVersion;
                        }
                    }
                    else
                    {
                        versionToUpgrade = getLatestVersion(list.versions());
                    }
                    
                    if (versionToUpgrade != null)
                    {
                        for (ActionConfiguration configuration : getUpgrades(versionToUpgrade, component.upgrades()))
                        {
                            allUpgradeActions.add(new ActionData(versionToUpgrade, null, configuration, list.id()));
                        }
                    }
                }
            }
            catch (MigrationException e)
            {
                // we're trying to execute as much actions as possible, so we store the issue and throw later
                versionDeterminationFailed = true;
                _failedException = e;
                getLogger().error("Error during version determination for component {}", componentId,  e);
            }
        }
        
        // then execute all needed actions
        if (_executeInitializationActions(allInitActions) || _executeUpgradeActions(allUpgradeActions))
        {
            getLogger().info("Automatic migration is restarting the server to continue migration");
            return true;
        }
        
        if (versionDeterminationFailed)
        {
            // at least one version determination has failed
            throw new MigrationException("Version determination failed. See previous error messages for more details.");
        }
        
        return false;
    }
    
    /**
     * Return all upgrades to execute for the current version
     * @param currentVersion the current version
     * @param availableUpgrades all available upgrades
     * @return all upgrade to execute
     * @throws MigrationException if the upgrades are not compatible
     */
    public List<ActionConfiguration> getUpgrades(Version currentVersion, List<ActionConfiguration> availableUpgrades) throws MigrationException
    {
        List<ActionConfiguration> actualUpgrades = new ArrayList<>();
        
        String versionNumber = currentVersion.getVersionNumber();

        String higherCurrentUpgradeVersionNumber = getHighestUpgradeVersionNumber(currentVersion.getComponent());
        if (higherCurrentUpgradeVersionNumber.compareTo(versionNumber) < 0)
        {
            getLogger().warn("There is a version stored more recent that any version available in conf for version: {}", currentVersion.toString());
        }
        
        for (ActionConfiguration upgrade : availableUpgrades)
        {
            String from = upgrade.getFrom();
            String upgradeVersionNumber = upgrade.getVersionNumber();
            
            if (versionNumber.compareTo(upgradeVersionNumber) < 0 // Accepts all upgrades to newer versions
                    && (from == null || versionNumber.compareTo(from) <= 0)) // and older than the from attribute, if any (allowing to bypass some upgrades)
            {
                actualUpgrades.add(upgrade);
            }
        }

        return _removeDuplicatedActions(actualUpgrades);
    }
    
    /**
     * Retrives all stored versions for the given component
     * @param component the component to check
     * @return all stored versions.
     * @throws MigrationException if something wrong occured
     */
    public Versions getVersions(MigrationComponent component) throws MigrationException
    {
        Map<String, Object> environmentInformation = null;
        try
        {
            environmentInformation = BackgroundEngineHelper.createAndEnterEngineEnvironment(_manager, _context, new SLF4JLoggerAdapter(getLogger()));

            return component.versionHandler.getVersions(component);
        }
        finally
        {
            BackgroundEngineHelper.leaveEngineEnvironment(environmentInformation);
        }                            
    }
    
    private List<VersionList> _getVersionLists(Versions versions)
    {
        if (versions instanceof VersionList versionList)
        {
            return List.of(versionList);
        }
        else
        {
            List<VersionList> versionLists = new ArrayList<>();

            for (Versions subVersions : ((VersionsContainer) versions).values())
            {
                versionLists.addAll(_getVersionLists(subVersions));
            }
            
            return versionLists;
        }
    }
    
    /**
     * Search a given version list among all thos returned by a {@link VersionHandler}
     * @param versions the version tree
     * @param id the wanted id
     * @return the corresponding {@link VersionList}
     */
    public VersionList getVersionList(Versions versions, String id)
    {
        List<VersionList> lists = _getVersionLists(versions);
        return lists.stream()
                    .filter(list -> id.equals(list.id()))
                    .findFirst()
                    .orElse(null);
    }
    
    /**
     * Returns the latest version (according to version numbers) among the given list of versions
     * @param versions a list of {@link Version}
     * @return the latest version, or null if the list is empty
     */
    public Version getLatestVersion(List<Version> versions)
    {
        return versions.stream()
                       .sorted(Comparator.comparing(Version::getVersionNumber).reversed())
                       .findFirst()
                       .orElse(null);
    }

    private void _notifyEndOfMigration()
    {
        if (_observationManager != null)
        {
            _observationManager.notify(new Event(ObservationConstants.EVENT_MIGRATION_ENDED, null, Map.of()));
        }
    }
    
    private boolean _executeInitializationActions(List<ActionData> allInitializationActions) throws MigrationException
    {
        for (ActionData action : allInitializationActions)
        {
            try
            {
                if (initializeVersion(action))
                {
                    // The migration require a restart. Stop the iteration here.
                    return true;
                }
            }
            catch (MigrationException e)
            {
                // store the failing action and rethrow
                _failedAction = e.getFailedAction();
                _failedException = e;
                throw e;
            }
        }
        
        return false;
    }
    
    private boolean _executeUpgradeActions(List<ActionData> allUpgradeActions) throws MigrationException
    {
        allUpgradeActions.sort(Comparator.comparing(action -> action.configuration().getVersionNumber()));
        
        if (allUpgradeActions.isEmpty())
        {
            getLogger().debug("No upgrade to do");
        }
        else
        {
            _logUpgradeActionsBeforeExecution(allUpgradeActions);
            for (ActionData action : allUpgradeActions)
            {
                try
                {
                    if (upgradeVersion(action))
                    {
                        // The migration require a restart. Stop the iteration here.
                        return true;
                    }
                }
                catch (MigrationException e)
                {
                    // store the failing action and rethrow
                    _failedAction = action;
                    _failedException = e;
                    throw e;
                }
            }
        }
        return false;
    }

    /**
     * Execute all needed initializations
     * @param action the version and action configuration to initialize
     * @return true if an action requires a restart
     * @throws MigrationException Something went wrong
     */
    public boolean initializeVersion(ActionData action) throws MigrationException
    {
        ActionConfiguration configuration = action.configuration();
        String versionComment = configuration != null ? configuration.getComment() : null;
        if (StringUtils.isBlank(versionComment))
        {
            versionComment = "Automatic Initialization";
        }
        else
        {
            versionComment = "Automatic Initialization: " + versionComment;
        }
        
        boolean restartRequired = false;
        if (configuration != null)
        {
            restartRequired = doAction(action, _initializationEP);
        }
        
        _createAndStoreVersion(action, versionComment);
        
        return restartRequired;
    }
   
    /**
     * Execute an upgrade
     * @param action the version and action configuration to upgrade
     * @return true if an action requires a restart
     * @throws MigrationException Something went wrong
     */
    public boolean upgradeVersion(ActionData action) throws MigrationException
    {
        String versionComment = action.configuration().getComment();
        if (StringUtils.isBlank(versionComment))
        {
            versionComment = "Automatic Upgrade";
        }
        else
        {
            versionComment = "Automatic Upgrade: " + versionComment;
        }
        
        boolean restartRequired = doAction(action, _upgradeEP);
        _createAndStoreVersion(action, versionComment);
        return restartRequired;
    }
   
    /**
     * Executes a migration action.
     * @param actionData  the version and action configuration to execute
     * @param extensionPoint the related {@link ActionExtensionPoint}
     * @return true if the action needs a restart after its execution
     * @throws MigrationException Something went wrong
     */
    public boolean doAction(ActionData actionData, ActionExtensionPoint extensionPoint) throws MigrationException
    {
        Map<String, Object> environmentInformation = null;
        
        try
        {
            // Create the environment.
            environmentInformation = BackgroundEngineHelper.createAndEnterEngineEnvironment(_manager, _context, new SLF4JLoggerAdapter(getLogger()));
            
            Action action = extensionPoint.getExtension(actionData.configuration().getType());
            getLogger().info("Run action : {}", actionData.configuration());
            
            action.act(actionData);
            
            if (actionData.configuration().requiresRestart())
            {
                return true;
            }
        }
        catch (MigrationException e)
        {
            // rethrow with attached context
            actionData.currentVersion().setExecutionInstant(Instant.now());
            throw new MigrationException(e.getMessage(), e.getFailureMessage(), e.getCause(), actionData);
        }
        finally
        {
            BackgroundEngineHelper.leaveEngineEnvironment(environmentInformation);
        }
        
        return false;
    }

    private void _logUpgradeActionsBeforeExecution(List<ActionData> allActions)
    {
        if (getLogger().isInfoEnabled())
        {
            int count = allActions.size();
            StringBuilder sb = new StringBuilder().append(count).append(count > 1 ? " pending upgrade(s):" : " pending upgrade:");
            for (ActionData action : allActions)
            {
                sb.append(System.lineSeparator());
                sb.append("* ").append(action.toString());
            }
            getLogger().info(sb.toString());
        }
    }

    private Version _createAndStoreVersion(ActionData action, String comment) throws MigrationException
    {
        Version currentVersion = action.currentVersion();
        Version newVersion = currentVersion.getStorage().createVersion(currentVersion.getComponentId(), currentVersion.getComponent(), currentVersion.getStorageConfiguration(), currentVersion.getAdditionalValues());
        newVersion.setExecutionInstant(Instant.now());
        newVersion.setComment(comment);
        newVersion.setVersionNumber(action.targetVersionNumber() != null ? action.targetVersionNumber() : action.configuration() != null ? action.configuration().getVersionNumber() : "0");
        
        getLogger().info("Store new version: {}", newVersion);

        newVersion.getStorage().storeVersion(newVersion);
        
        return newVersion;
    }
    
    /**
     * Returns the highest version number among all components' upgrades
     * @param component the {@link MigrationComponent} to consider
     * @return the highest upgrade version number
     */
    public String getHighestUpgradeVersionNumber(MigrationComponent component)
    {
        return component.upgrades().stream()
                                   .map(ActionConfiguration::getVersionNumber)
                                   .sorted(Comparator.reverseOrder())
                                   .findFirst()
                                   .orElse("0");
    }
    
    private List<ActionConfiguration> _removeDuplicatedActions(List<ActionConfiguration> actions) throws MigrationException
    {
        // First, sort the upgrades by if to have the right order
        actions.sort(Comparator.comparing(ActionConfiguration::getVersionNumber)
                               .thenComparing(ActionConfiguration::getFrom, Comparator.nullsFirst(Comparator.naturalOrder())));
        
        // Create a list of pairs for each upgrade containing a "from" attribute to know from which version to which version we should remove the lines
        List<Pair<String, String>> fromTo = actions.stream()
                                                   .filter(u -> StringUtils.isNotBlank(u.getFrom()))
                                                   .map(u -> Pair.of(u.getFrom(), u.getVersionNumber()))
                                                   .collect(Collectors.toList());

        _checkFromUpgrades(actions, fromTo);
        
        List<ActionConfiguration> result = actions;
        
        // Invert the order so we will start to apply the most recent one, it may override other lines with a From inside
        Collections.reverse(fromTo);
        
        // For each pair, remove the right versions
        for (Pair<String, String> pair : fromTo)
        {
            String from = pair.getLeft();
            String to = pair.getRight();
            
            result = result.stream()
                           .filter(a -> a.getVersionNumber().compareTo(from) <= 0 
                               || to.equals(a.getVersionNumber()) && from.equals(a.getFrom())
                               || a.getVersionNumber().compareTo(to) > 0)
                           .collect(Collectors.toList());
        }
        
        return result;
    }
    
    /**
     * Tests that in the list of fromTo, that each "fromTo" is linked to an action with the same id and without "from" 
     * @param actions list of actions
     * @param fromTo list of actions overriding multiple actions
     * @throws MigrationException there is an overriding version without a "simple" version with the same id
     */
    protected void _checkFromUpgrades(List<ActionConfiguration> actions, List<Pair<String, String>> fromTo) throws MigrationException
    {
        for (Pair<String, String> pair : fromTo)
        {
            String from = pair.getLeft();
            String to = pair.getRight();
            
            boolean anyMatch = actions.stream().anyMatch(a -> a.getFrom() == null && to.equals(a.getVersionNumber()));
            
            if (!anyMatch)
            {
                throw new MigrationException("The action from '" + from + "' to '" + to + "' does not contain a normal upgrade from '" + from + "'");
            }
        }
    }
    
    /**
     * If the upgrade process has failed, returns the corresponding {@link ActionConfiguration}.<br>
     * Returns null if the process went well.
     * @return the failed action data 
     */
    public ActionData getFailedAction()
    {
        return _failedAction;
    }
    
    /**
     * If the upgrade process has failed, returns the corresponding {@link MigrationException}.<br>
     * Returns null if the process went well.
     * @return the exception causing the upgrade failure 
     */
    public MigrationException getFailedException()
    {
        return _failedException;
    }
    
    /**
     * A migration component represents a versioned feature set 
     * @param id the component id
     * @param internal if this component is used for internal migration data
     * @param pluginName the component's plugin
     * @param featureName the component's feature
     * @param versionHandlerType the {@link VersionHandler} type
     * @param versionHandler the {@link VersionHandler}
     * @param versionStorage the {@link VersionStorage}
     * @param versionConfiguration the specific {@link VersionConfiguration} created by the associated {@link VersionHandler}
     * @param initialization the initialization action, if any
     * @param upgrades all actions defined for uprades
     */
    public record MigrationComponent(String id, boolean internal, String pluginName, String featureName, String versionHandlerType, VersionHandler versionHandler, VersionStorage versionStorage, VersionConfiguration versionConfiguration, ActionConfiguration initialization, List<ActionConfiguration> upgrades) { /* empty */ }
    
    /**
     * A version tree, grouped recursively by sub-components.
     */
    public sealed interface Versions permits VersionsContainer, VersionList 
    { 
        // empty interface
    }
    
    /**
     * Contains a list of versions for a sub-component
     */
    public static final class VersionsContainer extends HashMap<I18nizableText, Versions> implements Versions
    { 
        // empty class
    }
    
    /**
     * The leaf of the {@link Versions} tree. Contains the actual stored data and all informations needed to create a new version from storage.
     * @param id the id of this list, unique for a given {@link MigrationComponent}.
     * @param versions the version list
     * @param componentId the component id
     * @param storageConfiguration the storage configuration
     * @param additionalValues an opaque object holding necessary information for the {@link VersionStorage} to create new versions if needed.
     */
    public record VersionList(String id, List<Version> versions, String componentId, VersionConfiguration storageConfiguration, Map<String, Object> additionalValues) implements Versions { /* empty */ }
    
    /**
     * All data needed to actually execute an upgrade action: the current version and the action configuration.
     * @param currentVersion the current version
     * @param targetVersionNumber the target versionNumber. If not present (mainly in case of upgrades), the target version is taken from the configuration.
     * @param configuration the action configuration
     * @param versionListId the encapsulating {@link VersionList}'s id, or null if not known
     */
    public record ActionData(Version currentVersion, String targetVersionNumber, ActionConfiguration configuration, String versionListId) { /* empty */ }
}
