/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration;

import java.util.ArrayList;
import java.util.List;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.action.Action;
import org.ametys.core.migration.action.ActionConfiguration;
import org.ametys.core.migration.action.ActionExtensionPoint;
import org.ametys.core.migration.version.VersionConfiguration;
import org.ametys.core.migration.version.handler.VersionHandler;
import org.ametys.core.migration.version.handler.VersionHandlerExtensionPoint;
import org.ametys.core.migration.version.storage.VersionStorage;
import org.ametys.runtime.plugin.AbstractExtensionPoint;

/**
 * Migration Extension Point that will list all migration needed by the current state of the application
 */
public class MigrationExtensionPoint extends AbstractExtensionPoint<MigrationComponent> implements Serviceable, Configurable
{
    /** Avalon Role */
    public static final String ROLE = MigrationExtensionPoint.class.getName();
    
    private VersionHandlerExtensionPoint _versionHandlerEP;

    private ActionExtensionPoint _upgradeEP;
    private ActionExtensionPoint _initializationEP;
    
    private boolean _internal;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _versionHandlerEP = (VersionHandlerExtensionPoint) manager.lookup(VersionHandlerExtensionPoint.ROLE);
        _upgradeEP = (ActionExtensionPoint) manager.lookup(ActionExtensionPoint.ROLE_UPGRADE);
        _initializationEP = (ActionExtensionPoint) manager.lookup(ActionExtensionPoint.ROLE_INITIALIZATION);
    }
    
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _internal = configuration.getChild("internal").getValueAsBoolean(false);
    }
    
    public void addExtension(String id, String pluginName, String featureName, Configuration configuration) throws ConfigurationException
    {
        // Get VersionHandler, its associated VersionStorage and their VersionConfiguration
        Configuration versionsConfiguration = configuration.getChild("versions");
        String versionHandlerType = configuration.getChild("versions").getAttribute("type");
        VersionHandler handler = _versionHandlerEP.getExtension(versionHandlerType);
        
        VersionConfiguration handlerConfiguration;
        VersionStorage storage;
        try
        {
            handlerConfiguration = handler.createVersionConfiguration(versionsConfiguration);
            storage = handler.getVersionStorage(handlerConfiguration);
        }
        catch (NotMigrableInSafeModeException e)
        {
            getLogger().warn("The migration '{}' cannot be done in safe mode", id, e);
            return;
        }
        catch (MigrationException e)
        {
            getLogger().warn("Unable to find storage for  migration '{}'", id, e);
            return;
        }

        List<ActionConfiguration> upgrades = new ArrayList<>();
        
        Configuration[] upgradeList = configuration.getChild("upgrades").getChildren("upgrade");
        for (Configuration upgradeConf : upgradeList)
        {
            String upgradeVersionNumber = upgradeConf.getAttribute("versionNumber");
            String type = upgradeConf.getAttribute("type");
            String from = upgradeConf.getAttribute("from", null);
            String comment = upgradeConf.getAttribute("comment", null);
            boolean restartRequired = upgradeConf.getAttributeAsBoolean("restart", false);
            
            Action action = _upgradeEP.getExtension(type);
            
            if (action == null)
            {
                throw new ConfigurationException("The type '" + type + "' does not exist.", upgradeConf);
            }
            
            upgrades.add(action.createActionConfiguration(upgradeVersionNumber, type, comment, from, restartRequired, false, upgradeConf));
        }
        
        ActionConfiguration initialization = null;
        
        Configuration initializationConf = configuration.getChild("initialization", false);
        if (initializationConf != null)
        {
            String type = initializationConf.getAttribute("type");
            String comment = initializationConf.getAttribute("comment", null);
            boolean restartRequired = initializationConf.getAttributeAsBoolean("restart", false);
            
            Action action = _initializationEP.getExtension(type);
            
            if (action == null)
            {
                throw new ConfigurationException("The type '" + type + "' does not exist.", initializationConf);
            }
            
            initialization = action.createActionConfiguration(null, type, comment, null, restartRequired, true, initializationConf);
        }
        
        _extensions.put(id, new MigrationComponent(id, _internal, pluginName, featureName, versionHandlerType, handler, storage, handlerConfiguration, initialization, upgrades));
    }

    public void initializeExtensions() throws Exception
    {
        // Nothing here
    }
}
