/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.action;

import org.apache.commons.lang3.StringUtils;

/**
 * Representation of an upgrade line in the extension
 */
public class DefaultActionConfiguration implements ActionConfiguration
{
    private String _versionNumber;
    private String _type;
    private String _comment;
    private String _from;
    private boolean _restartRequired;
    private boolean _initialization;
    
    /**
     * Create the Upgrade based on the upgrade xml line
     * Must contains id and type
     * May contain restartAfter, component and/or file
     * @param versionNumber version number of the action
     * @param type type of action
     * @param comment The comment about this action
     * @param from if this actions is the equivalent of multiple actions, this is the version id just before the 1st action impacted by this action
     * @param restartRequired true if a restart is required after the action
     * @param isInitialization true for initialization action, false for upgrade
     */
    public DefaultActionConfiguration(String versionNumber, String type, String comment, String from, boolean restartRequired, boolean isInitialization)
    {
        _versionNumber = versionNumber;
        _type = type;
        _comment = comment;
        _from = from;
        _restartRequired = restartRequired;
        _initialization = isInitialization;
    }
    
    @Override
    public String toString()
    {
        String log;
        if (_initialization)
        {
            log = "Initialization of type '" + _type + "'. Last version is '" + _versionNumber + "'";
        }
        else
        {
            log = "Upgrade '" + _versionNumber + "' of type '" + _type + "'";
        }
        
        if (StringUtils.isNotBlank(_comment))
        {
            log += " (" + _comment + ")";
        }
        
        return log;
    }
    
    @Override
    public String getVersionNumber()
    {
        return _versionNumber;
    }

    @Override
    public String getType()
    {
        return _type;
    }

    @Override
    public String getComment()
    {
        return _comment;
    }

    @Override
    public String getFrom()
    {
        return _from;
    }
    
    @Override
    public boolean requiresRestart()
    {
        return _restartRequired;
    }
    
    @Override
    public void setRequiresRestart(boolean restart)
    {
        _restartRequired = restart;
    }

}
