/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.action;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.core.migration.MigrationEngine.ActionData;
import org.ametys.core.migration.MigrationException;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * SQL action : A script will be executed
 */
public class JavaAction extends AbstractLogEnabled implements Action, Serviceable
{
    /** service Manager */
    protected ServiceManager _manager;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
    }
    
    public void act(ActionData actionData) throws MigrationException
    {
        getLogger().debug("Start initialization for : {}", actionData.toString());
        
        if (!(actionData.configuration() instanceof JavaActionConfiguration javaActionData))
        {
            throw new MigrationException("Java Upgrade can only be created for a Java upgrade, this is not the case for upgrade : " + actionData.toString());
        }
        
        try
        {
            JavaActionComponent action = (JavaActionComponent) _manager.lookup(javaActionData.getRole());
            action.doAction(actionData.currentVersion(), javaActionData);
        }
        catch (ClassCastException | ServiceException e)
        {
            throw new MigrationException("In component '" + actionData.currentVersion().getComponentId() + "', the role for the version can not be found or is not a JavaActionComponent.", e);
        }
        
        getLogger().debug("End upgrade for : {}", actionData.toString());
    }
    
    public ActionConfiguration createActionConfiguration(String versionNumber, String type, String comment, String from, boolean restartRequired, boolean isInitialization, Configuration configuration) throws ConfigurationException
    {
        return new JavaActionConfiguration(versionNumber, type, comment, from, restartRequired, isInitialization, configuration.getAttribute("role"));
    }
    
    /**
     * {@link ActionConfiguration} with Java component role
     */
    public static class JavaActionConfiguration extends DefaultActionConfiguration
    {
        private String _role;
        
        JavaActionConfiguration(String versionNumber, String type, String comment, String from, boolean restartRequired, boolean isInitialization, String role)
        {
            super(versionNumber, type, comment, from, restartRequired, isInitialization);
            _role = role;
        }
        
        /**
         * Get the role of the component to call
         * @return The role of the component to call
         */
        public String getRole()
        {
            return _role;
        }
        
        @Override
        public String toString()
        {
            return super.toString() + " role : '" + _role + "'";
        }
    }
}
