/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.action;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.util.Map;

import javax.script.ScriptException;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceNotFoundException;
import org.apache.excalibur.source.SourceResolver;
import org.slf4j.LoggerFactory;

import org.ametys.core.migration.MigrationEngine.ActionData;
import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.version.Version;
import org.ametys.plugins.core.ui.script.ScriptHandler;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Script action: A script will be executed
 */
public class ScriptAction extends AbstractLogEnabled implements Action, Serviceable
{
    private SourceResolver _sourceResolver;
    private ScriptHandler _scriptHandler;

    public void service(ServiceManager manager) throws ServiceException
    {
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _scriptHandler = (ScriptHandler) manager.lookup(ScriptHandler.ROLE);
    }
    
    public void act(ActionData actionData) throws MigrationException
    {
        getLogger().debug("Start initialization for : {}", actionData.configuration().toString());
        
        if (!(actionData.configuration() instanceof ScriptActionConfiguration scriptActionConfiguration))
        {
            throw new MigrationException("Script Upgrade can only be created for a Script upgrade, this is not the case for upgrade : " + actionData.configuration().toString());
        }
        
        try
        {
            Version currentVersion = actionData.currentVersion();
            
            String script = _getScript(currentVersion.getComponent(), scriptActionConfiguration);
            
            Map<String, Object> variables = Map.of(
                "version", currentVersion,
                "actionData", actionData.configuration(), // for legacy purposes
                "actionConfiguration", actionData.configuration(),
                "logger", LoggerFactory.getLogger(getLogger().getName() + "." + currentVersion.getComponentId())
            );
            
            // Execute the script in the admin workspace 
            Map<String, Object> scriptResult = _scriptHandler.executeScript(script, variables, "admin");
            if (!StringUtils.isAllBlank((String) scriptResult.get("error"), (String) scriptResult.get("message"), (String) scriptResult.get("stacktrace")))
            {
                throw new MigrationException("Error while executing script to upgrade component '" + currentVersion.getComponentId() + "' to version '" + actionData.configuration().getVersionNumber() + "'. Error : '" + scriptResult.get("error") + "', message : '" + scriptResult.get("message") + "', stacktrace : '" + scriptResult.get("stacktrace") + "'", (String) scriptResult.get("message"));
            }
        }
        catch (ScriptException e)
        {
            throw new MigrationException("Error while executing script to upgrade : '" + actionData.configuration().toString() + "'", e);
        }
        
        getLogger().debug("End upgrade for : {}", actionData.configuration().toString());
    }

    private String _getScript(MigrationComponent component, ScriptActionConfiguration scriptActionConfiguration) throws MigrationException
    {
        String plugin = scriptActionConfiguration.getPlugin() != null ? scriptActionConfiguration.getPlugin() : component.pluginName();
        
        String script;
        if (StringUtils.isNotBlank(scriptActionConfiguration.getScript()))
        {
            script = scriptActionConfiguration.getScript();
        }
        else
        {
            String uri = "plugin:" + plugin + "://" + scriptActionConfiguration.getFile();
            Source source = null;
            try
            {
                source = _sourceResolver.resolveURI(uri);
                
                if (!source.exists())
                {
                    throw new SourceNotFoundException("URI '" + uri + "' does not exist.");
                }
                try (InputStream is = source.getInputStream())
                {
                    script = IOUtils.toString(is, "UTF-8");
                }
            }
            catch (MalformedURLException e)
            {
                throw new MigrationException("Impossible to find the script for the upgrade : " + scriptActionConfiguration.toString(), e);
            }
            catch (IOException e)
            {
                throw new MigrationException("Impossible to run the upgrade : " + scriptActionConfiguration.toString(), e);
            }
            finally
            {
                if (source != null)
                {
                    _sourceResolver.release(source);
                }
            }
        }
        return "function main() { \n " + script + " \n }";
    }
    
    public ActionConfiguration createActionConfiguration(String versionNumber, String type, String comment, String from, boolean restartRequired, boolean isInitialization, Configuration configuration) throws ConfigurationException
    {
        String file = configuration.getAttribute("file", null);
        String plugin = configuration.getAttribute("plugin", null);
        String script = configuration.getValue(null);
        
        if (StringUtils.isAllBlank(script, file))
        {
            throw new ConfigurationException("The configuration must contain at least a script, or a file to fetch the script.", configuration);
        }
        
        return new ScriptActionConfiguration(versionNumber, type, comment, from, restartRequired, isInitialization, file, plugin, script);
    }
    
    /**
     * {@link ActionConfiguration} with script location
     */
    public static class ScriptActionConfiguration extends DefaultActionConfiguration
    {
        private String _file;
        private String _plugin;
        private String _script;
        
        ScriptActionConfiguration(String versionNumber, String type, String comment, String from, boolean restartRequired, boolean isInitialization, String file, String plugin, String script)
        {
            super(versionNumber, type, comment, from, restartRequired, isInitialization);
            _file = file;
            _plugin = plugin;
            _script = script;
        }
        
        /**
         * Get the path for the script to run
         * @return The path for the script to run
         */
        public String getFile()
        {
            return _file;
        }

        /**
         * Get the path for the script to run
         * @return The path for the script to run
         */
        public String getScript()
        {
            return _script;
        }
        
        /**
         * Get the plugin containing the sql script, or null to default to the component's plugin
         * @return The plugin containing the sql script
         */
        public String getPlugin()
        {
            return _plugin;
        }
    }
}
