/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.action;

import java.sql.Connection;
import java.sql.SQLException;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.migration.MigrationEngine.ActionData;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.version.storage.SqlVersionStorage.SqlVersionConfiguration;
import org.ametys.core.script.SQLScriptHelper;


/**
 * Initialization of an SQL table
 * Similar to {@link SqlUpgradeAction} with a test about the existence of a table
 */
public class SqlInitializationAction extends SqlUpgradeAction
{
    @Override
    public void act(ActionData actionData) throws MigrationException
    {
        if (!(actionData.configuration() instanceof SqlInitializationActionConfiguration sqlActionConfiguration))
        {
            throw new MigrationException("SQL Initialization can only be created with a SQL action, this is not the case for action: " + actionData.configuration().toString());
        }
        
        Connection connection = null;
    
        try
        {
            String datasourceId = ((SqlVersionConfiguration) actionData.currentVersion().getStorageConfiguration()).dataSource();
            connection = ConnectionHelper.getConnection(datasourceId);
            // If no table configured, we consider it does not exists (maybe the script works with many tables or test it itself)
            boolean tableExists = StringUtils.isNotBlank(sqlActionConfiguration.getTable()) ? SQLScriptHelper.tableExists(connection, sqlActionConfiguration.getTable()) : false;
            if (tableExists)
            {
                // If the table exists, throw an exception, no initialization for this component, the admin will need to do a manual upgrade
                throw new MigrationException("Impossible to initialize the version as the table '" + sqlActionConfiguration.getTable() + "' already exists. " + actionData.configuration().toString());
            }
            else
            {
                _upgrade(sqlActionConfiguration, actionData.currentVersion());
            }
        }
        catch (SQLException e)
        {
            throw new MigrationException("SQL exception when trying to upgrade: " + actionData.configuration().toString(), e);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
        }
    }
    
    @Override
    public ActionConfiguration createActionConfiguration(String versionNumber, String type, String comment, String from, boolean restartRequired, boolean isInitialization, Configuration configuration) throws ConfigurationException
    {
        String file = configuration.getAttribute("file");
        String plugin = configuration.getAttribute("plugin", null);
        String table = configuration.getAttribute("table", null);

        return new SqlInitializationActionConfiguration(versionNumber, type, comment, from, restartRequired, isInitialization, file, plugin, table);
    }
    
    /**
     * {@link ActionConfiguration} for sql initialisation with table to check
     */
    public static class SqlInitializationActionConfiguration extends SqlUpgradeActionConfiguration
    {
        private String _table;
        
        /**
         * Create an {@link ActionConfiguration} for use with sql initialization actions.
         * @param versionNumber version number of the action
         * @param type type of action
         * @param comment The comment about this action
         * @param from if this actions is the equivalent of multiple actions, this is the version id just before the 1st action impacted by this action
         * @param restartRequired true if a restart is required after the action
         * @param isInitialization true for initialization action, false for upgrade
         * @param file the location of the sql script
         * @param plugin the plugin of the sql script or null to default to the component plugin
         * @param table the table name to check
         */
        public SqlInitializationActionConfiguration(String versionNumber, String type, String comment, String from, boolean restartRequired, boolean isInitialization, String file, String plugin, String table)
        {
            super(versionNumber, type, comment, from, restartRequired, isInitialization, file, plugin);
            _table = table;
        }
        
        /**
         * Get the table that will need to be tested before the initialization is executed
         * @return The table that will need to be tested before the initialization is executed
         */
        public String getTable()
        {
            return _table;
        }
    }
}
