/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.action;

import java.io.IOException;
import java.io.InputStream;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.IOUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.migration.MigrationEngine.ActionData;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.version.Version;
import org.ametys.core.migration.version.storage.SqlVersionStorage.SqlVersionConfiguration;
import org.ametys.core.script.SQLScriptHelper;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * SQL action: A script will be executed
 */
public class SqlUpgradeAction extends AbstractLogEnabled implements Action, Serviceable
{
    /** Source resolver */
    protected SourceResolver _sourceResolver;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    @Override
    public void act(ActionData actionData) throws MigrationException
    {
        getLogger().debug("Start ugprade for: {}", actionData.configuration().toString());
        
        if (!(actionData.configuration() instanceof SqlUpgradeActionConfiguration sqlActionConfiguration))
        {
            throw new MigrationException("SQL Upgrade can only be created for an SQL upgrade, this is not the case for upgrade: " + actionData.configuration().toString());
        }
        
        _upgrade(sqlActionConfiguration, actionData.currentVersion());
        
        getLogger().debug("End upgrade for: {}", actionData.configuration().toString());
    }

    /**
     * Do the upgrade with the right objects
     * @param sqlUpgradeConfiguration upgrade data
     * @param version version
     * @throws MigrationException Something went wrong
     */
    protected void _upgrade(SqlUpgradeActionConfiguration sqlUpgradeConfiguration, Version version) throws MigrationException
    {
        Connection connection = null;
        Source source = null;
        try
        {
            String file = sqlUpgradeConfiguration.getFile();
            
            String datasourceId = ((SqlVersionConfiguration) version.getStorageConfiguration()).dataSource();
            connection = ConnectionHelper.getConnection(datasourceId);

            String plugin = sqlUpgradeConfiguration.getPlugin() != null ? sqlUpgradeConfiguration.getPlugin() : version.getComponent().pluginName();
            
            file = "plugin:" + plugin + "://" + file.replaceAll("%DATABASE_TYPE%", ConnectionHelper.getDatabaseType(connection));
            source = _sourceResolver.resolveURI(file);
            
            try (InputStream is = source.getInputStream())
            {
                String script = IOUtils.toString(is, "UTF-8");
                
                Map<String, Object> additionalScriptValue = version.getAdditionalValues();
                for (Entry<String, Object> keyValue : additionalScriptValue.entrySet())
                {
                    Object value = keyValue.getValue();
                    if (value instanceof String)
                    {
                        script = script.replaceAll("%" + keyValue.getKey() + "%", (String) value);
                    }
                }

                SQLScriptHelper.runScript(connection, script);
            }
        }
        catch (SQLException | IOException e)
        {
            throw new MigrationException("Impossible to run the upgrade: " + sqlUpgradeConfiguration.toString(), e);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
            _sourceResolver.release(source);
        }
    }
    
    public ActionConfiguration createActionConfiguration(String versionNumber, String type, String comment, String from, boolean restartRequired, boolean isInitialization, Configuration configuration) throws ConfigurationException
    {
        String file = configuration.getAttribute("file");
        String plugin = configuration.getAttribute("plugin", null);

        return new SqlUpgradeActionConfiguration(versionNumber, type, comment, from, restartRequired, isInitialization, file, plugin);
    }
    
    /**
     * {@link ActionConfiguration} with sql script location
     */
    public static class SqlUpgradeActionConfiguration extends DefaultActionConfiguration
    {
        private String _file;
        private String _plugin;

        /**
         * Create an {@link ActionConfiguration} for use with sql upgrades actions.
         * @param versionNumber version number of the action
         * @param type type of action
         * @param comment The comment about this action
         * @param from if this actions is the equivalent of multiple actions, this is the version id just before the 1st action impacted by this action
         * @param restartRequired true if a restart is required after the action
         * @param isInitialization true for initialization action, false for upgrade
         * @param file the location of the sql script
         * @param plugin the plugin of the sql script or null to default to the component plugin
         */
        public SqlUpgradeActionConfiguration(String versionNumber, String type, String comment, String from, boolean restartRequired, boolean isInitialization, String file, String plugin)
        {
            super(versionNumber, type, comment, from, restartRequired, isInitialization);
            
            _file = file;
            _plugin = plugin;
        }

        /**
         * Get the path for the script to run
         * @return The path for the script to run
         */
        public String getFile()
        {
            return _file;
        }
        
        /**
         * Get the plugin containing the sql script (either the default plugin of the extension, or the one passed in the config)
         * @return The plugin containing the sql script (either the default plugin of the extension, or the one passed in the config)
         */
        public String getPlugin()
        {
            return _plugin;
        }
    }
}
