/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.version;

import java.time.Instant;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

import org.apache.commons.lang3.StringUtils;

import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.version.storage.VersionStorage;

/**
 * Representation of a version
 * Each upgrade create a new version
 */
public class Version
{
    private MigrationComponent _component;
    private String _componentId;
    private VersionStorage _storage;
    private VersionConfiguration _configuration;
    private String _versionNumber;
    private Instant _executionInstant;
    private String _comment;
    private Map<String, Object> _additionalValues;

    /**
     * Create a version to be stored
     * @param component the context component
     * @param componentId id of the component
     * @param storage the {@link VersionStorage} associated with this Version
     * @param configuration the storage configuration
     * @param versionNumber version number of the upgrade that created this version
     * @param executionInstant time of the application of this version
     * @param comment comment stored about this version
     */
    public Version(MigrationComponent component, String componentId, VersionStorage storage, VersionConfiguration configuration, String versionNumber, Instant executionInstant, String comment)
    {
        _component = component;
        _componentId = componentId;
        _storage = storage;
        _configuration = configuration;
        _versionNumber = versionNumber;
        _executionInstant = executionInstant;
        _comment = comment;
        _additionalValues = new HashMap<>();
    }
    
    /**
     * Build a readable log for this Version
     * @return a readable log
     */
    @Override
    public String toString()
    {
        String log = "Version '" + _versionNumber + "' for component '" + _componentId + "'";
        if (StringUtils.isNotBlank(_comment))
        {
            log += " (" + _comment + ")";
        }
        return log;
    }
    
    @Override
    public boolean equals(Object obj)
    {
        if (obj instanceof Version version)
        {
            return Objects.equals(_component, version._component)
                    && Objects.equals(_componentId, version._componentId)
                    && Objects.equals(_configuration, version._configuration)
                    && Objects.equals(_versionNumber, version._versionNumber)
                    && Objects.equals(_executionInstant, version._executionInstant)
                    && Objects.equals(_comment, version._comment)
                    && Objects.equals(_additionalValues, version._additionalValues);
        }
        
        return false;
    }
    
    @Override
    public int hashCode()
    {
        return Objects.hash(_component, _componentId, _configuration, _versionNumber, _executionInstant, _comment, _additionalValues);
    }

    /**
     * Get the context component
     * @return The context component
     */
    public MigrationComponent getComponent()
    {
        return _component;
    }
    
    /**
     * Get the component id
     * @return The component id
     */
    public String getComponentId()
    {
        return _componentId;
    }
    
    /**
     * Get the storage for this version
     * @return the associated storage
     */
    public VersionStorage getStorage()
    {
        return _storage;
    }
    
    /**
     * Get the configuration for the associated storage
     * @return the configuration
     */
    public VersionConfiguration getStorageConfiguration()
    {
        return _configuration;
    }
    
    /**
     * Get the number of the version (null means that initialization is needed)
     * @return The number of the version
     */
    public String getVersionNumber()
    {
        return _versionNumber;
    }
    
    /**
     * Get the time when the version was created
     * @return The time when the version was created
     */
    public Instant getExecutionInstant()
    {
        return _executionInstant;
    }
    
    /**
     * Get the comment about this version
     * @return The comment about this version
     */
    public String getComment()
    {
        return _comment;
    }

    /**
     * Set the number of the version
     * Should only be used in initialization actions, so the version contains the version to reach, not the version to start (which should be null in the case of initialization)
     * @param versionNumber number to set for this version
     */
    public void setVersionNumber(String versionNumber)
    {
        _versionNumber = versionNumber;
    }

    /**
     * Set the execution date of this version (only to be used when creating a new version after cloning an old version)
     * @param executionInstant instant to set
     */
    public void setExecutionInstant(Instant executionInstant)
    {
        _executionInstant = executionInstant;
    }

    /**
     * Change the comment of the version
     * @param comment comment to set
     */
    public void setComment(String comment)
    {
        _comment = comment;
    }
    
    /**
     * Add a parameter that will be used by upgrade actions
     * @param key key to replace
     * @param value value that will be inserted
     * @return the previous value associated with key, or null if there was no mapping for key.
     */
    public Object addAdditionalValue(String key, Object value)
    {
        return _additionalValues.put(key, value);
    }
    
    /**
     * Get the values that will be used by upgrade actions
     * @return the values
     */
    public Map<String, Object> getAdditionalValues()
    {
        return _additionalValues;
    }
}
