/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.version.handler;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationEngine.VersionsContainer;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.NotMigrableInSafeModeException;
import org.ametys.core.migration.version.VersionConfiguration;
import org.ametys.core.migration.version.storage.VersionStorage;
import org.ametys.core.migration.version.storage.VersionStorageExtensionPoint;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Version Handler which use a java class as version storage
 */
public class JavaVersionHandler extends AbstractLogEnabled implements VersionHandler, Serviceable
{
    private ServiceManager _manager;
    private VersionStorageExtensionPoint _storageExtensionPoint;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _storageExtensionPoint = (VersionStorageExtensionPoint) manager.lookup(VersionStorageExtensionPoint.ROLE);
    }
    
    public VersionStorage getVersionStorage(VersionConfiguration configuration) throws MigrationException
    {
        if (!(configuration instanceof JavaVersionConfiguration javaConfiguration))
        {
            throw new MigrationException("The version configuration object should be an instance of JavaVersionConfiguration");
        }

        return _storageExtensionPoint.getExtension(javaConfiguration.storage());
    }
    
    public VersionsContainer getVersions(MigrationComponent component) throws MigrationException
    {
        VersionConfiguration configuration = component.versionConfiguration();
        if (!(configuration instanceof JavaVersionConfiguration javaConfiguration))
        {
            throw new MigrationException("The version configuration object should be an instance of JavaVersionConfiguration");
        }
        
        JavaVersionHandlerComponent handler;
        try
        {
            handler = (JavaVersionHandlerComponent) _manager.lookup(javaConfiguration.role());
        }
        catch (ClassCastException | ServiceException e)
        {
            throw new MigrationException("In component '" + component.id() + "', the role '" + javaConfiguration.role() + "' for the version can not be found or is not a JavaVersionHandlerComponent.", e);
        }
        
        return handler.getVersions(component);
    }

    public VersionConfiguration createVersionConfiguration(Configuration configuration) throws ConfigurationException, NotMigrableInSafeModeException
    {
        return new JavaVersionConfiguration(configuration.getAttribute("role"), configuration.getAttribute("storage"), configuration);
    }
    
    /**
     * A {@link VersionConfiguration} with a Java component role.
     * @param role the component role
     * @param storage the {@link VersionStorage} type
     * @param configuration the initial configuration, possibly used by scripts
     */
    public record JavaVersionConfiguration(String role, String storage, Configuration configuration)  implements VersionConfiguration { /* empty */ }
}
