/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.version.handler;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.script.ScriptException;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.slf4j.LoggerFactory;

import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationEngine.VersionList;
import org.ametys.core.migration.MigrationEngine.Versions;
import org.ametys.core.migration.MigrationEngine.VersionsContainer;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.NotMigrableInSafeModeException;
import org.ametys.core.migration.version.Version;
import org.ametys.core.migration.version.VersionConfiguration;
import org.ametys.core.migration.version.storage.VersionStorage;
import org.ametys.core.migration.version.storage.VersionStorageExtensionPoint;
import org.ametys.plugins.core.ui.script.ScriptHandler;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Version Handler which use a script as version storage
 */
public class ScriptVersionHandler extends AbstractLogEnabled implements VersionHandler, Serviceable
{
    private SourceResolver _sourceResolver;
    private ScriptHandler _scriptHandler;
    private VersionStorageExtensionPoint _storageExtensionPoint;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _scriptHandler = (ScriptHandler) manager.lookup(ScriptVersionHandlerScriptHandler.COMPONENT_ROLE);
        _storageExtensionPoint = (VersionStorageExtensionPoint) manager.lookup(VersionStorageExtensionPoint.ROLE);
    }
    
    public VersionStorage getVersionStorage(VersionConfiguration configuration) throws MigrationException
    {
        if (!(configuration instanceof ScriptVersionConfiguration scriptConfiguration))
        {
            throw new MigrationException("The version configuration object should be an instance of JavaVersionConfiguration");
        }

        return _storageExtensionPoint.getExtension(scriptConfiguration.storage());
    }
    
    public Versions getVersions(MigrationComponent component) throws MigrationException
    {
        try
        {
            Object resultObject = _executeScript(component);
            return _getVersions(resultObject);
        }
        catch (ScriptException | RuntimeException e)
        {
            throw new MigrationException("Error while getting version by script for component : '" + component.id() + "'", e);
        }
    }
    
    private Versions _getVersions(Object resultObject) throws MigrationException
    {
        if (resultObject instanceof VersionList list)
        {
            return list;
        }
        else if (resultObject instanceof Version version)
        {
            String id = version.getComponentId();
            return new VersionList(id, _filterList(List.of(version)), id, version.getStorageConfiguration(), null);
        }
        else if (resultObject instanceof Version[] versions && versions.length > 0)
        {
            String id = versions[0].getComponentId();
            return new VersionList(id, _filterList(List.of(Arrays.asList(versions))), id, versions[0].getStorageConfiguration(), null);
        }
        else if (resultObject instanceof List versions && versions.size() > 0)
        {
            String id = ((Version) versions.get(0)).getComponentId();
            return new VersionList(id, _filterList(versions), id, ((Version) versions.get(0)).getStorageConfiguration(), null);
        }
        else if (resultObject instanceof Map versions)
        {
            VersionList list = _getVersionListFromMap(versions);

            if (list != null)
            {
                return list;
            }
            else
            {
                VersionsContainer allVersions = new VersionsContainer();
                
                for (Entry version : (Set<Entry<Object, Object>>) versions.entrySet())
                {
                    if (version.getKey() instanceof String subComponentId)
                    {
                        allVersions.put(new I18nizableText(subComponentId), _getVersions(version.getValue()));
                    }
                    else if (version.getKey() instanceof I18nizableText i18n)
                    {
                        allVersions.put(i18n, _getVersions(version.getValue()));
                    }
                    else
                    {
                        throw new MigrationException("A map returned by a version script should be either <I18nizableText, ?> or <String, ?>");
                    }
                }
                
                return allVersions;
            }
        }
        
        throw new MigrationException("The script returned an unexpected object, it must be a Version, Version[], List<Version> or Map<String, List<Version>>: " + resultObject);
    }
    
    @SuppressWarnings("unchecked")
    private VersionList _getVersionListFromMap(Map versions)
    {
        Set keys = versions.keySet();
        if (keys.size() == 5 && versions.get("id") instanceof String id
                             && versions.get("versions") instanceof List list
                             && versions.get("componentId") instanceof String subComponentId 
                             && (versions.get("versionConfiguration") == null || versions.get("versionConfiguration") instanceof VersionConfiguration)
                             && versions.get("additionalValues") instanceof Map values)
        {
            return new VersionList(id, _filterList(list), subComponentId, (VersionConfiguration) versions.get("versionConfiguration"), values);
        }
        
        return null;
    }
    
    private List<Version> _filterList(List<?> versions)
    {
        return versions.stream()
                       .filter(Version.class::isInstance)
                       .map(Version.class::cast)
                       .filter(v -> v.getVersionNumber() != null)
                       .toList();
    }
    
    private Object _executeScript(MigrationComponent component) throws MigrationException, ScriptException
    {
        if (!(component.versionConfiguration() instanceof ScriptVersionConfiguration scriptVersionConfiguration))
        {
            throw new MigrationException("The version configuration object should be an instance of ScriptVersionConfiguration");
        }
        
        String script = scriptVersionConfiguration.script();
        
        Map<String, Object> variables = _getVariables(component, scriptVersionConfiguration);
        
        // Execute the script in the admin workspace 
        Map<String, Object> scriptResult = _scriptHandler.executeScript(script, variables, "admin");
        if (!StringUtils.isAllBlank((String) scriptResult.get("error"), (String) scriptResult.get("message"), (String) scriptResult.get("stacktrace")))
        {
            throw new MigrationException("Error while executing script to get version for component '" + component.id() + "'. Error : '" + scriptResult.get("error") + "', message : '" + scriptResult.get("message") + "', stacktrace : '" + scriptResult.get("stacktrace") + "'");
        }
        
        return scriptResult.get("result");
    }
    
    private Map<String, Object> _getVariables(MigrationComponent component, ScriptVersionConfiguration scriptVersionConfiguration)
    {
        Map<String, Object> variables = new HashMap<>();
        variables.put("component", component);
        variables.put("componentId", component.id());
        variables.put("versionConfiguration", scriptVersionConfiguration.configuration());
        variables.put("logger", LoggerFactory.getLogger(getLogger().getName() + "." + component.id()));
        return variables;
    }

    private String _getScript(Configuration versionConfiguration) throws ConfigurationException
    {
        String script = versionConfiguration.getValue(null);
        if (StringUtils.isBlank(script))
        {
            String uri = versionConfiguration.getAttribute("uri", ""); // Should be plugin://path/to/script
            if (StringUtils.isBlank(uri))
            {
                throw new ConfigurationException("The script should be directly written in the version, or in a linked file.", versionConfiguration);
            }
            
            Source source = null;
            try
            {
                source = _sourceResolver.resolveURI(uri);
                
                if (!source.exists())
                {
                    throw new ConfigurationException("URI '" + uri + "' does not exist.", versionConfiguration);
                }
                try (InputStream is = source.getInputStream())
                {
                    script = IOUtils.toString(is, "UTF-8");
                }
            }
            catch (IOException e)
            {
                throw new ConfigurationException("Impossible to run the script at uri '" + uri + "'.", versionConfiguration, e);
            }
            finally
            {
                if (source != null)
                {
                    _sourceResolver.release(source);
                }
            }
        }

        return "function main() { \n " + script + " \n }";
    }

    public VersionConfiguration createVersionConfiguration(Configuration configuration) throws ConfigurationException, NotMigrableInSafeModeException
    {
        String uri = configuration.getAttribute("uri", null);
        String script = _getScript(configuration);
        String storage = configuration.getAttribute("storage");
        
        return new ScriptVersionConfiguration(uri, script, storage, configuration);
    }
    
    /**
     * A {@link VersionConfiguration} with script location. Either uri or script should be provided.
     * @param uri the script uri, if any
     * @param script the actual script
     * @param storage the {@link VersionStorage} type
     * @param configuration the initial configuration, possibly used by scripts
     */
    public record ScriptVersionConfiguration(String uri, String script, String storage, Configuration configuration)  implements VersionConfiguration { /* empty */ }
}
