/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.version.handler;

import java.util.List;

import org.apache.avalon.framework.configuration.Configurable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationEngine.VersionList;
import org.ametys.core.migration.MigrationEngine.Versions;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.NotMigrableInSafeModeException;
import org.ametys.core.migration.version.Version;
import org.ametys.core.migration.version.VersionConfiguration;
import org.ametys.core.migration.version.storage.VersionStorage;
import org.ametys.core.migration.version.storage.VersionStorageExtensionPoint;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Class for {@link VersionHandler} delegating to a single {@link VersionStorage}.
 */
public class StorageBasedVersionHandler extends AbstractLogEnabled implements VersionHandler, Serviceable, Configurable
{
    private VersionStorageExtensionPoint _versionStorageEP;
    private VersionStorage _versionStorage;
    private String _versionStorageId;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _versionStorageEP = (VersionStorageExtensionPoint) manager.lookup(VersionStorageExtensionPoint.ROLE);
    }

    @Override
    public void configure(Configuration configuration) throws ConfigurationException
    {
        _versionStorageId = configuration.getChild("version-storage").getValue(null);
        
        if (StringUtils.isBlank(_versionStorageId))
        {
            throw new ConfigurationException("<version-storage> is mandatory and should not be blank", configuration);
        }
        
        _versionStorage = _versionStorageEP.getExtension(_versionStorageId);
    }
    
    public VersionStorage getVersionStorage(VersionConfiguration configuration) throws MigrationException
    {
        return _versionStorage;
    }

    public Versions getVersions(MigrationComponent component) throws MigrationException
    {
        getLogger().debug("Get all versions for component id: " + component.id());
        List<Version> versions = _versionStorage.getVersions(component.id(), component.versionConfiguration(), component);
        
        return new VersionList(component.id(), versions, component.id(), component.versionConfiguration(), null);
    }

    @Override
    public VersionConfiguration createVersionConfiguration(Configuration versionConfiguration) throws ConfigurationException, NotMigrableInSafeModeException
    {
        return _versionStorage.createVersionConfiguration(versionConfiguration);
    }
}
