/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.version.storage;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.migration.MigrationEngine;
import org.ametys.core.migration.MigrationEngine.ActionData;
import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.NotMigrableInSafeModeException;
import org.ametys.core.migration.action.ActionConfiguration;
import org.ametys.core.migration.action.SqlInitializationAction.SqlInitializationActionConfiguration;
import org.ametys.core.migration.action.SqlUpgradeAction.SqlUpgradeActionConfiguration;
import org.ametys.core.migration.version.Version;
import org.ametys.core.migration.version.VersionConfiguration;
import org.ametys.core.script.SQLScriptHelper;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;
import org.ametys.runtime.plugin.component.PluginAware;

/**
 * SQL implementation of {@link VersionStorage}, to manage the list of versions in a database
 */
public class SqlVersionStorage extends AbstractLogEnabled implements VersionStorage, Serviceable, PluginAware, Initializable
{
    private static final String __VERSION_TABLE_NAME = "Ametys_Versions";

    private static final String __COLUMN_COMPONENT = "component_identifier";
    private static final String __COLUMN_VERSION = "ametys_version_number";
    private static final String __COLUMN_DATE = "execution_date";
    private static final String __COLUMN_COMMENT = "execution_comment";
    
    private static final String __AMETYS_VERSIONS_COMPONENT_ID = "ametys.versions.table";
    
    private static final TreeMap<String, String> __AMETYS_VERSIONS_UDPATES = new TreeMap<>();
    static
    {
        __AMETYS_VERSIONS_UDPATES.put("20210401T0000", "Modify column 'component_identifier' to enlarge it to 250 characters.");
    }
    
    private MigrationEngine _migrationEngine;
    
    private String _plugin;
    private String _feature;
    private String _id;
    
    private ActionConfiguration _internalInitialization;
    private List<ActionConfiguration> _internalUpgrades = new ArrayList<>();
    
    private List<String> _checkedDataSources = new ArrayList<>();

    public void setPluginInfo(String pluginName, String featureName, String id)
    {
        _id = id;
        _plugin = pluginName;
        _feature = featureName;
    }
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _migrationEngine = (MigrationEngine) manager.lookup(MigrationEngine.ROLE);
    }
    
    public void initialize() throws Exception
    {
        for (String versionNumber : __AMETYS_VERSIONS_UDPATES.keySet())
        {
            String comment = __AMETYS_VERSIONS_UDPATES.get(versionNumber);
            _internalUpgrades.add(new SqlUpgradeActionConfiguration(versionNumber, "sql", comment, null, false, false, String.format("scripts/%%DATABASE_TYPE%%/ametys_versions/v%s.sql", versionNumber), "core"));
        }
        
        String highestVersionNumber = __AMETYS_VERSIONS_UDPATES.lastKey();

        _internalInitialization = new SqlInitializationActionConfiguration(highestVersionNumber, "sql", null, null, false, true, "scripts/%DATABASE_TYPE%/ametys_versions.sql", "core", null);
    }
    
    public String getId()
    {
        return _id;
    }
    
    public List<Version> getVersions(String componentId, VersionConfiguration configuration, MigrationComponent component) throws MigrationException
    {
        if (!(configuration instanceof SqlVersionConfiguration sqlConfiguration))
        {
            throw new MigrationException("The version configuration object should be an instance of SqlVersionConfiguration");
        }

        String dataSourceId = sqlConfiguration.dataSource();
        
        // First migrate the storage itself if needed
        _internalUpgrade(dataSourceId);

        Connection connection = null;
        
        try
        {
            connection = ConnectionHelper.getConnection(dataSourceId);
            return _getVersionsForComponent(connection, componentId, component, sqlConfiguration);
        }
        catch (SQLException e)
        {
            String message = String.format("Error while getting the versions for: %s (dataSource: %s)", componentId, dataSourceId);
            throw new MigrationException(message, e);
        }
        finally
        {
            // Close connection
            ConnectionHelper.cleanup(connection);
        }
    }
    
    private List<Version> _getVersionsForComponent(Connection connection, String componentId, MigrationComponent component, VersionConfiguration configuration) throws SQLException
    {
        PreparedStatement stmt = null;
        ResultSet rs = null;
        
        try
        {
            String sql = "SELECT * FROM " + __VERSION_TABLE_NAME + " WHERE " + __COLUMN_COMPONENT + " = ? ORDER BY " + __COLUMN_VERSION + " DESC";
            
            stmt = connection.prepareStatement(sql);
            stmt.setString(1, componentId);
            
            rs = stmt.executeQuery();
            
            List<Version> versions = new ArrayList<>();
            
            while (rs.next())
            {
                String versionNumber = rs.getString(__COLUMN_VERSION);
                Timestamp timestamp = rs.getTimestamp(__COLUMN_DATE);
                String comment = rs.getString(__COLUMN_COMMENT);
                
                Version version = new Version(component, componentId, this, configuration, versionNumber, timestamp.toInstant(), comment);
                versions.add(version);
            }
            
            return versions;
        }
        finally
        {
            ConnectionHelper.cleanup(rs);
            ConnectionHelper.cleanup(stmt);
        }
    }
    
    public Version createVersion(String componentId, MigrationComponent component, VersionConfiguration storageConfiguration, Map<String, Object> additionalValues) throws MigrationException
    {
        Version version = VersionStorage.super.createVersion(componentId, component, storageConfiguration, additionalValues);
        
        // if there's no initialization, let's assume we are in the default "0" version, so that all upgrades may be executed 
        if (component.initialization() == null)
        {
            version.setVersionNumber("0");
        }
        
        return version;
    }

    @Override
    public void storeVersion(Version version) throws MigrationException
    {
        getLogger().debug("Add version for: {}", version.toString());
        
        if (!(version.getStorageConfiguration() instanceof SqlVersionConfiguration configuration))
        {
            throw new MigrationException("Impossible to create a SQL version in a non-sql version: " + version.toString());
        }
        
        String dataSourceId = configuration.dataSource();
    
        _internalUpgrade(dataSourceId);

        Connection connection = null;
        
        try
        {
            connection = ConnectionHelper.getConnection(dataSourceId);
            _addVersion(connection, version);
        }
        catch (SQLException e)
        {
            throw new MigrationException("Error during SQL query to set the new version for: " + version.toString() + " (dataSource: " + dataSourceId + ")", e);
        }
        finally
        {
            // Close connections
            ConnectionHelper.cleanup(connection);
        }

        getLogger().debug("End add version for: {}", version.toString());
    }
    
    private void _addVersion(Connection connection, Version version) throws SQLException
    {
        PreparedStatement stmt = null;
        
        try
        {
            String sql = "INSERT INTO " + __VERSION_TABLE_NAME + " (" + __COLUMN_COMPONENT + ", " + __COLUMN_VERSION + ", " + __COLUMN_DATE + ", " + __COLUMN_COMMENT + ") VALUES (?, ?, ?, ?)";
            
            stmt = connection.prepareStatement(sql);
            stmt.setString(1, version.getComponentId());
            stmt.setString(2, version.getVersionNumber());
            stmt.setTimestamp(3, Timestamp.from(version.getExecutionInstant()));
            stmt.setString(4, version.getComment());
            
            stmt.executeUpdate();
        }
        finally
        {
            // Close connections
            ConnectionHelper.cleanup(stmt);
        }
    }
    
    public void removeVersion(String componentId, String versionNumber, VersionConfiguration configuration) throws MigrationException
    {
        getLogger().debug("Start remove version {} for component: {}", versionNumber, componentId);

        if (!(configuration instanceof SqlVersionConfiguration sqlConfiguration))
        {
            throw new MigrationException("The version configuration object should be an instance of SqlVersionConfiguration");
        }
        
        String dataSourceId = sqlConfiguration.dataSource();
        Connection connection = null;
        PreparedStatement stmt = null;
        
        try
        {
            _internalUpgrade(dataSourceId);
            
            connection = ConnectionHelper.getConnection(dataSourceId);
            String sql = "DELETE FROM " + __VERSION_TABLE_NAME + " WHERE " + __COLUMN_COMPONENT + " = ? AND " + __COLUMN_VERSION + " = ?";
            
            stmt = connection.prepareStatement(sql);
            stmt.setString(1, componentId);
            stmt.setString(2, versionNumber);
            
            stmt.executeUpdate();
        }
        catch (SQLException e)
        {
            String message = String.format("Error while getting the versions for %s (datasource: %s)", componentId, dataSourceId);
            throw new MigrationException(message, e);
        }
        finally
        {
            // Close connections
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(connection);
        }
        
        getLogger().debug("End remove version {} for component: {}", versionNumber, componentId);
    }

    @Override
    public void removeAllVersions(String componentIdentifier, VersionConfiguration versionConfiguration) throws MigrationException
    {
        getLogger().debug("Start remove all version for component: {}", componentIdentifier);

        if (!(versionConfiguration instanceof SqlVersionConfiguration))
        {
            throw new MigrationException("The version configuration object should be an instance of SqlVersionConfiguration");
        }
        
        String dataSourceId = ((SqlVersionConfiguration) versionConfiguration).dataSource();
        Connection connection = null;
        PreparedStatement stmt = null;
        
        try
        {
            _internalUpgrade(dataSourceId);
            
            connection = ConnectionHelper.getConnection(dataSourceId);
            String sql = "DELETE FROM " + __VERSION_TABLE_NAME + " WHERE " + __COLUMN_COMPONENT + " = ?";
            
            stmt = connection.prepareStatement(sql);
            stmt.setString(1, componentIdentifier);
            
            stmt.executeUpdate();
        }
        catch (SQLException e)
        {
            String message = String.format("Error while getting the versions for %s (datasource: %s)", componentIdentifier, dataSourceId);
            throw new MigrationException(message, e);
        }
        finally
        {
            // Close connections
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(connection);
        }
        
        getLogger().debug("End remove all version for component: {}", componentIdentifier);
    }
    
    private void _internalUpgrade(String dataSource) throws MigrationException
    {
        if (_checkedDataSources.contains(dataSource))
        {
            return;
        }
        
        getLogger().info("Check internal versions table for datasource {}", dataSource);

        // avoid to check the internal structure on each method call
        _checkedDataSources.add(dataSource);
        
        MigrationComponent component = _createVersionsTableMigrationComponent(dataSource);
        
        Connection connection = null;
        try
        {
            connection = ConnectionHelper.getConnection(dataSource);
            
            // Use a fake MigrationComponent to initiate the migration system itself
            if (!SQLScriptHelper.tableExists(connection, __VERSION_TABLE_NAME))
            {
                Version version = createVersion(component.id(), component, new SqlVersionConfiguration(dataSource), null);
                if (_migrationEngine.initializeVersion(new ActionData(version, null, _internalInitialization, null)))
                {
                    throw new UnsupportedOperationException("Restarting after internal intialization is not supported");
                }
            }
            else
            {
                List<Version> versions = _getVersionsForComponent(connection, component.id(), component, new SqlVersionConfiguration(dataSource));
                Version currentVersion = versions.stream().sorted(Comparator.comparing(Version::getVersionNumber).reversed())
                                                          .findFirst()
                                                          .orElse(null);
                
                if (currentVersion == null)
                {
                    // If no version, all upgrades should be applied
                    currentVersion = createVersion(component.id(), component, new SqlVersionConfiguration(dataSource), null);
                    currentVersion.setVersionNumber("0");
                }
                
                List<ActionConfiguration> upgrades = _migrationEngine.getUpgrades(currentVersion, _internalUpgrades);
                for (ActionConfiguration action : upgrades)
                {
                    if (_migrationEngine.upgradeVersion(new ActionData(currentVersion, null, action, null)))
                    {
                        throw new UnsupportedOperationException("Restarting after internal upgrade is not supported");
                    }
                }
            }
        }
        catch (SQLException e)
        {
            String errorMsg = String.format("Error during SQL tables initialization for data source id: '%s'.", StringUtils.defaultString(dataSource));
            throw new MigrationException(errorMsg, e);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
        }
    }
    
    private MigrationComponent _createVersionsTableMigrationComponent(String datasource)
    {
        // fake MigrationComponent with no version handler for internal purposes
        return new MigrationComponent(__AMETYS_VERSIONS_COMPONENT_ID, false, _plugin, _feature, "sql", null, this, new SqlVersionConfiguration(datasource), _internalInitialization, _internalUpgrades);
    }
    
    public VersionConfiguration createVersionConfiguration(Configuration versionConfiguration) throws ConfigurationException, NotMigrableInSafeModeException
    {
        String dataSourceId = null;
        Configuration dataSourceConf = versionConfiguration.getChild("datasource");
        String dataSourceConfParam = dataSourceConf.getValue(null);
        if (StringUtils.isBlank(dataSourceConfParam))
        {
            throw new ConfigurationException("The 'datasource' configuration node must be defined and contain non empty value.", dataSourceConf);
        }
        
        String dataSourceConfType = dataSourceConf.getAttribute("type", null);
        
        if (StringUtils.equals(dataSourceConfType, "config"))
        {
            if (Config.getInstance() != null)
            {
                dataSourceId = Config.getInstance().getValue(dataSourceConfParam);
            }
            else
            {
                throw new NotMigrableInSafeModeException("Cannot get datasource for config parameter '" + dataSourceConfParam + "' in safe mode");
            }
        }
        else // expecting type="id"
        {
            dataSourceId = dataSourceConfParam;
        }
        
        return new SqlVersionConfiguration(dataSourceId);
    }
    
    /**
     * A {@link VersionConfiguration} with a SQL datasource id
     * @param dataSource the datasource id
     */
    public record SqlVersionConfiguration(String dataSource)  implements VersionConfiguration { /* empty */ }
}
