/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.migration.version.storage;

import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;

import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.NotMigrableInSafeModeException;
import org.ametys.core.migration.version.Version;
import org.ametys.core.migration.version.VersionConfiguration;

/**
 * Store the versions of a migration component and acts as a factory to create specific configuration objects.
 */
public interface VersionStorage
{
    /**
     * Get the unique identifier
     * @return The identifier
     */
    String getId();
    
    /**
     * Get the history of all versions for a component identifier
     * @param componentId the component id
     * @param component the context component
     * @param storageConfiguration configuration for this storage
     * @return list of all versions for this component
     * @throws MigrationException Something went wrong
     */
    List<Version> getVersions(String componentId, VersionConfiguration storageConfiguration, MigrationComponent component) throws MigrationException;
    
    /**
     * Add a version
     * @param version Version to store (contains the component identifier)
     * @throws MigrationException Something went wrong
     */
    void storeVersion(Version version) throws MigrationException;
    
    /**
     * Removes a stored version
     * @param componentId the component
     * @param versionNumber the version to remove
     * @param configuration the associated configuration
     * @throws MigrationException Something went wrong
     */
    public void removeVersion(String componentId, String versionNumber, VersionConfiguration configuration) throws MigrationException;
    
    /**
     * Remove all versions for a component identifier
     * @param componentIdentifier component to clear
     * @param configuration configuration to use
     * @throws MigrationException Something went wrong
     */
    void removeAllVersions(String componentIdentifier, VersionConfiguration configuration) throws MigrationException;  
    
    /**
     * Create a specific object containing useful data for the handler/storage
     * @param versionConfiguration configuration from the plugin.xml
     * @return the confuguration object
     * @throws ConfigurationException something wrong in the configuration
     * @throws NotMigrableInSafeModeException it it's impossible to configure the versions in safe mode
     */
    VersionConfiguration createVersionConfiguration(Configuration versionConfiguration) throws ConfigurationException, NotMigrableInSafeModeException;
    
    /**
     * Creates a new version object, used to initialize the migration process. This version is actually NOT stored.<br>
     * Depending of the context, the created {@link Version} could have the "0" number (ie. all actions should be executed), 
     * or <pre>null</pre> number (ie. only initialization action should be executed and the actual version number set to the highest available).  
     * @param componentId the component id
     * @param component the context component
     * @param storageConfiguration configuration for this storage
     * @param additionalValues all specific values used by upgrade actions
     * @return the newly created Version
     * @throws MigrationException Impossible to create the version
     */
    default Version createVersion(String componentId, MigrationComponent component, VersionConfiguration storageConfiguration, Map<String, Object> additionalValues) throws MigrationException
    {
        Version version =  new Version(component, componentId, this, storageConfiguration, null, null, null);
        
        if (additionalValues != null)
        {
            for (Entry<String, Object> entry : additionalValues.entrySet())
            {
                version.addAdditionalValue(entry.getKey(), entry.getValue());
            }
        }
        
        return version;
    }
}
