/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.nio.charset.StandardCharsets;
import java.util.Map;

import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceFactory;
import org.apache.excalibur.source.SourceNotFoundException;
import org.apache.excalibur.source.SourceValidity;
import org.apache.excalibur.source.impl.validity.TimeStampValidity;

/**
 * Fake factory for SourceMapCache
 */
public class SourceMapSourceFactory implements SourceFactory
{
    /** Parameter last modified */
    public static final String LAST_MODIFIED = "lastmodified";
    /** Parameter content*/
    public static final String CONTENT = "content";
    /** Declared protocol */
    public static final String SCHEME = "sourcemap";
    
    public Source getSource(String location, Map parameters) throws IOException, MalformedURLException
    {
        return new SourceMapSource(location, (String) parameters.get(CONTENT), (Long) parameters.get(LAST_MODIFIED));
    }

    public void release(Source source)
    {
        // Nothing to do
    }

    /**
     * The Source of a source map
     */
    public static final class SourceMapSource implements Source
    {
        private String _content;
        private Long _lastModified;
        private String _location;
        private TimeStampValidity _timeStampValidity;
        
        /**
         * Default constructor for a memory source
         * @param location The location of the source
         * @param content The content of the source
         * @param lastModified The last modified date
         */
        public SourceMapSource(String location, String content, Long lastModified)
        {
            _location = location;
            _content = content;
            _lastModified = lastModified;
        }
        
        public boolean exists()
        {
            return true;
        }

        public InputStream getInputStream() throws IOException, SourceNotFoundException
        {
            return new ByteArrayInputStream(_content.getBytes(StandardCharsets.UTF_8));
        }

        public String getURI()
        {
            return _location;
        }

        public String getScheme()
        {
            return SCHEME;
        }

        public SourceValidity getValidity()
        {
            if (_timeStampValidity == null)
            {
                _timeStampValidity = new TimeStampValidity(_lastModified);
            }
            return _timeStampValidity;
        }

        public void refresh()
        {
            // do nothing
        }

        public String getMimeType()
        {
            return "application/json";
        }

        public long getContentLength()
        {
            return _content.length();
        }

        public long getLastModified()
        {
            return _lastModified;
        }
    }
}
