/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize.css;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;

import org.ametys.core.minimize.AbstractSourceMapResourceHandler;
import org.ametys.plugins.core.ui.resources.css.JSASSResourceURIExtensionPoint;

/**
 * Resource handler for source map files
 */
public class CSSSourceMapResourceHandler extends AbstractSourceMapResourceHandler
{
    /** CSS minimize manager */
    protected MinimizeCSSManager _cssMinimizeManager;
    
    private JSASSResourceURIExtensionPoint _jsassResourceURIExtensionPoint;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        
        _cssMinimizeManager = (MinimizeCSSManager) manager.lookup(MinimizeCSSManager.ROLE);
        _jsassResourceURIExtensionPoint = (JSASSResourceURIExtensionPoint) manager.lookup(JSASSResourceURIExtensionPoint.ROLE);
    }

    @Override
    protected Source _getAlternateSource(String location, Map objectModel) throws IOException
    {
        Source source = null;
        
        // 1. check if the source map is in cache
        try
        {
            String resolvedLocation = _jsassResourceURIExtensionPoint.resolve(location);
            source = _sourceMapCache.get(resolvedLocation != null ? resolvedLocation : location);
        }
        catch (URISyntaxException e)
        {
            getLogger().error("Unable to resolve the uri because the syntax '" + location + "' is not supported ", e);
        }

        if (source == null)
        {
            // 2. generate the source map and store it in cache
            source = _generateCssSourceMap(objectModel, location);
        }
        
        return source;
    }

    private Source _generateCssSourceMap(Map objectModel, String location) throws IOException
    {
        String originalUri = "/" + StringUtils.removeEnd(ObjectModelHelper.getRequest(objectModel).getSitemapURI(), ".map");
        
        Source sourceResolved = _resolver.resolveURI("cocoon:" + originalUri);
        if (sourceResolved != null && sourceResolved.exists())
        {
            try (InputStream inputStream = sourceResolved.getInputStream())
            {
                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                IOUtils.copy(inputStream, baos);
                
                // Stores the source of the sourceMap in the cache
                String sourceMapKey = _jsassResourceURIExtensionPoint.resolve(location);
                _cssMinimizeManager.minimizeCss(baos.toString(StandardCharsets.UTF_8.toString()), originalUri, sourceMapKey, sourceResolved.getLastModified());
                
                return _sourceMapCache.get(sourceMapKey);
            }
            catch (ProcessingException | URISyntaxException e)
            {
                getLogger().error("Unable to compile CSS file", e);
            }
        }
        else
        {
            _resolver.release(sourceResolved);
        }
        
        return null;
    }
}
