/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize.css;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Arrays;
import java.util.function.Predicate;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Helper for JSASS source maps
 */
public final class JSASSFixHelper
{

    // Regex that matches a JSASS injected import in the sources of a source map
    private static final Pattern SOURCEMAP_SOURCES_PATTERN = Pattern.compile("^([^\"]*\")" // Captures the start of the line until the first double quote
                                                                        + "([^\"]*)" // Captures any character that is the path of the JSASS injected import
                                                                        + "(\",?)$", // Captures the end of the line, which is the closing double quote, and an optional comma
                                                                        Pattern.MULTILINE);
    
    // List of JSASS files injected automatically in the source map
    private static final String[] JSASS_INJECTED_FILES = new String[] {"JSASS_CUSTOM.scss", "JSASS_PRE_IMPORT.scss", "JSASS_POST_IMPORT.scss"};
    
    private JSASSFixHelper()
    {
        // static helper
    }
    
    /**
     * Fix the common problems with JSASS generated source maps
     * @param sourceMap The source map
     * @param externalContextPath The context path 
     * @return The source map fixed
     */
    public static String fixJsassSourceMap(String sourceMap, String externalContextPath)
    {
        // Replace relative path with absolute
        Matcher matcher = SOURCEMAP_SOURCES_PATTERN.matcher(sourceMap);
        StringBuffer sb = new StringBuffer();
        while (matcher.find())
        {
            String source = matcher.group(2);
            Predicate<String> isJsassInjectedFile = file -> source.endsWith(file);
            if (Arrays.stream(JSASS_INJECTED_FILES).anyMatch(isJsassInjectedFile))
            {
                // JSASS injected file, replace with placeholder
                matcher.appendReplacement(sb, "$1" + externalContextPath + "/plugins/core/resources/css/jsass_injected_file_placeholder.css$3");
            }
            else
            {
                try
                {
                    if (!source.startsWith("/") && !(new URI(source).isAbsolute()))
                    {
                        // To prevent JSASS from resolving URIs, URIs are transformed to relative URIs before going through JSASS,
                        // and must be converted back to absolute (by adding the initial / at the start) afterward.
                        matcher.appendReplacement(sb, "$1/$2$3");
                    }
                }
                catch (URISyntaxException e)
                {
                    // invalid uri in sources of the sourceMap. 
                }
            }
        }
        
        matcher.appendTail(sb);
        
        return sb.toString();
    }
    
}
