/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize.css;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.regex.Matcher;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.input.BOMInputStream;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;

import org.ametys.core.minimize.AbstractMinimizeManager;
import org.ametys.plugins.core.ui.resources.css.CSSFileHelper;
import org.ametys.plugins.core.ui.resources.css.JSASSResourceURIExtensionPoint;

import com.google.debugging.sourcemap.SourceMapGeneratorV3;
import com.google.debugging.sourcemap.SourceMapParseException;

import io.bit3.jsass.CompilationException;
import io.bit3.jsass.Compiler;
import io.bit3.jsass.Options;
import io.bit3.jsass.Output;
import io.bit3.jsass.OutputStyle;
import io.bit3.jsass.importer.Importer;

/**
 * Minimize manager for CSS files
 */
public class MinimizeCSSManager extends AbstractMinimizeManager implements Component, Initializable
{
    /** The avalon ROLE */
    public static final String ROLE = MinimizeCSSManager.class.getName();
    
    private Compiler _jsassCompiler;

    private JSASSResourceURIExtensionPoint _jsassResourceURIExtensionPoint;

    public void initialize() throws Exception
    {
        _jsassCompiler = new Compiler();
    }
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _jsassResourceURIExtensionPoint = (JSASSResourceURIExtensionPoint) smanager.lookup(JSASSResourceURIExtensionPoint.ROLE);
    }
    
    /**
     * Minimize a CSS string
     * @param code The CSS code
     * @param location The CSS location, can be null if no location was determined
     * @return The minimized CSS
     * @throws ProcessingException If an error occurred during minimization
     * @throws IOException If an error occurred while retrieving the source map
     */
    public String minimizeCss(String code, String location)  throws ProcessingException, IOException
    {
        return minimizeCss(code, location, null, null);
    }

    /**
     * Minimize a CSS string
     * @param code The CSS code
     * @param location The CSS location, can be null if no location was determined, in which case no source map will be generated
     * @param sourceMapKey The key to store the source map in cache. Can be null to prevent source map generation
     * @param lastModified The last modified date of the code, used to assert validity of the source map cache. Can be null
     * @return The minimized CSS
     * @throws ProcessingException If an error occurred during minimization
     * @throws IOException If an error occurred while retrieving the source map
     */
    public String minimizeCss(String code, String location, String sourceMapKey, Long lastModified) throws IOException, ProcessingException 
    {
        try
        {
            boolean withSourceMap = location != null && sourceMapKey != null && lastModified != null;
            Output output = _minimizeCss(location, "", code, withSourceMap);
            
            String sourceResult = output.getCss();
            
            if (StringUtils.isBlank(sourceResult))
            {
                return "";
            }
            
            if (withSourceMap)
            {
                // Source map contains "skins/demo/resources/css/ametys.css" and we need to convert it to "/cms/preview/skins/demo/resources/css/ametys.css"
                String unfixedSourceMap = super.convertSourceMapURIs(output.getSourceMap(), "/");
                String sourceMap = JSASSFixHelper.fixJsassSourceMap(unfixedSourceMap, _proxiedContextPathProvider.getContextPath());
                _sourceMapCache.put(sourceMapKey, sourceMap, lastModified);
            }
            
            sourceResult = sourceResult.trim();
            if (location != null)
            {
                String fileName = location.substring(location.lastIndexOf("/") + 1);
                sourceResult += "\n" + formatSourceMappingURL(fileName + ".map");
            }
            
            return sourceResult;
        }
        catch (URISyntaxException e)
        {
            throw new ProcessingException("An error occurred while converting source map of " + location, e);
        }
    }

    @Override
    protected String getMinimizedContent(String fileUri, String nestedParentFilesName)
    {
        Source cssSource = null;
        
        try
        {
            // example: test.min.css
            String minimizeFileUri = StringUtils.endsWith(fileUri, ".min.css") ? fileUri : StringUtils.removeEnd(fileUri, ".css") + ".min.css";
            URI minimizeUri = new URI(minimizeFileUri);
            
            String uriToResolve = minimizeUri.isAbsolute() ? minimizeFileUri : "cocoon:/" + org.apache.cocoon.util.NetUtils.normalize(minimizeFileUri);
            cssSource = _resolver.resolveURI(uriToResolve);
            
            String originalContent;
            try (InputStream is = cssSource.getInputStream();
                 BOMInputStream bomIs = BOMInputStream.builder().setInputStream(is).get())
            {
                originalContent = IOUtils.toString(bomIs, "UTF-8");
            }
            
            originalContent = CSSFileHelper.replaceRelativeUri(originalContent, minimizeFileUri, _jsassResourceURIExtensionPoint,  _proxiedContextPathProvider.getContextPath(),  _proxiedContextPathProvider.getContextPath());
            
            StringBuffer result = new StringBuffer();
            result.append("/*! File : ");
            result.append(nestedParentFilesName);
            result.append(fileUri);
            result.append(" */\n");
            result.append(originalContent);
            
            return result.toString();
        }
        catch (Exception e)
        {
            getLogger().error("Cannot open CSS for aggregation " + fileUri, e);
            return "/** ERROR " + e.getMessage() + "*/\n";
        }
        finally
        {
            _resolver.release(cssSource);
        }
    }
    
    @Override
    protected String applyMediaToContent(String content, String media)
    {
        return "@media " + media + " {\n" + (StringUtils.endsWith(content, "\n") ? (content.substring(0, content.length() - 1) + "}\n") : (content + "}"));
    }

    private Output _minimizeCss(String cssUri, String nestedParentFilesName, String originalContent, boolean generateSourceMap) throws ProcessingException
    {
        try
        {
            String content = originalContent;
            Options options = new Options();
            options.setOutputStyle(OutputStyle.COMPRESSED);
            if (cssUri != null && generateSourceMap)
            {
                options.setSourceMapFile(new URI("unused")); // Parameter required to generate the source map, but value is unused
                options.setOmitSourceMapUrl(true);
            }
            Importer cssImporter = new AmetysCssImporter();
            options.getImporters().add(cssImporter);

            URI inputPath;
            URI outputPath;
            if (cssUri != null)
            {
                URI uri = new URI(cssUri);
                String contextualizedUri = uri.isAbsolute() ? _jsassResourceURIExtensionPoint.resolve(uri.toString()) : uri.toString();
                content = CSSFileHelper.replaceRelativeUri(content, contextualizedUri, _jsassResourceURIExtensionPoint, _proxiedContextPathProvider.getContextPath(), _proxiedContextPathProvider.getContextPath());
    
                content = _resolveImportUrl(content, nestedParentFilesName + cssUri + " > ");
                
                inputPath = new URI("." + cssUri);
                outputPath = _computeMinimizedUri(FilenameUtils.getName(cssUri));
            }
            else
            {
                // Minimize the css file as best as possible, due to an unknown location. No source map will be generated
                options.setSourceMapFile(null);
                inputPath = new URI("unknown-location");
                outputPath = new URI("unknown-location");
            }
            return _jsassCompiler.compileString(content, inputPath, outputPath, options);
        }
        catch (CompilationException | URISyntaxException e)
        {
            throw new ProcessingException("Unable to Minimize the css file: " + cssUri, e);
        }
    }
    
    
    private URI _computeMinimizedUri(String sourceMapUri) throws URISyntaxException
    {
        return new URI(StringUtils.removeEnd(sourceMapUri, ".css") + ".min.css");
    }

    private String _resolveImportUrl(String content, String nestedParentFilesName) throws URISyntaxException
    {
        StringBuffer sb = new StringBuffer();
        
        Matcher importMatcher = CSSFileHelper.IMPORT_PATTERN.matcher(content);

        while (importMatcher.find()) 
        {
            String cssUrl = importMatcher.group(1);
            String media = importMatcher.group(2);
            
            if (cssUrl != null && !cssUrl.startsWith("http://") && !cssUrl.startsWith("https://") && !cssUrl.startsWith("//"))
            {
                URI uri = new URI(cssUrl);
                String uriToCompile = uri.isAbsolute() ? cssUrl : StringUtils.removeStart(cssUrl,  _proxiedContextPathProvider.getContextPath());
                String importedContent = getMinimizedContent(uriToCompile, nestedParentFilesName);
                importedContent += " /*! File end : " + nestedParentFilesName + uriToCompile + " */";
                if (StringUtils.isNotEmpty(media))
                {
                    importedContent = applyMediaToContent(importedContent, media);
                }
                importMatcher.appendReplacement(sb, Matcher.quoteReplacement(importedContent));
            }
        }
        
        importMatcher.appendTail(sb);
        
        return sb.toString();
    }
    
    @Override
    protected void addSourceMap(SourceMapGeneratorV3 sourceMapGenerator, int lineCount, String fileContent, String fileUri, String sourceMapUri)
    {
        try
        {
            String sourceMapContent = null;
            
            // 1. Get source map from cache
            String sourceMapKey = fileUri.indexOf('/') > -1 ? new URI(StringUtils.substringBeforeLast(fileUri, "/") + "/" + sourceMapUri).normalize().toString() : sourceMapUri;
            sourceMapKey = StringUtils.removeStart(sourceMapKey, _proxiedContextPathProvider.getContextPath());
            
            Source mapSource = _sourceMapCache.get(sourceMapKey);
            if (mapSource != null)
            {
                try (InputStream is = mapSource.getInputStream())
                {
                    sourceMapContent = IOUtils.toString(is, StandardCharsets.UTF_8);
                }
            }
            else
            {
                // 2. If not found, get source map content from pipeline .css.map
                String uriToResolve = new URI(sourceMapKey).isAbsolute() ? sourceMapKey : "cocoon:/" + sourceMapKey;
                
                try
                {
                    mapSource = _resolver.resolveURI(uriToResolve);
                    if (mapSource.exists())
                    {
                        try (InputStream is = mapSource.getInputStream())
                        {
                            sourceMapContent = IOUtils.toString(is, "UTF-8");
                        }
                    }
                }
                catch (IOException e)
                {
                    // Nothing
                }
            }
            
            if (sourceMapContent != null)
            {
                // 3. aggregate content
                sourceMapContent = JSASSFixHelper.fixJsassSourceMap(sourceMapContent, _proxiedContextPathProvider.getContextPath());
        
                // Add 1 to the line count to jump after the comment "/** File : */"
                sourceMapGenerator.mergeMapSection(lineCount + 1, 0, sourceMapContent);
            }
            else
            {
                getLogger().warn("Unable to retrieve source map when aggregating hash file, for sourceMappingURL '" + sourceMapUri + "' of file '" + fileUri + "'");
            }
        }
        catch (SourceMapParseException | IOException | URISyntaxException | IllegalStateException e)
        {
            // IllegalStateException occurs in "sourceMapGenerator.mergeMapSection" when a previous invalid source map offsets the line count too much
            getLogger().error("Unable to merge source map of '" + fileUri + "' to the CSS minimized source map", e);
        }
    }

    @Override
    protected boolean isSourceMappingURLLine(String line)
    {
        return line != null && line.startsWith("/*# sourceMappingURL=") && line.endsWith(" */");
    }

    @Override
    protected String getSourceMappingURL(String line)
    {
        return line.substring("/*# sourceMappingURL=".length(), line.length() - " */".length()).trim();
    }

    @Override
    protected String removeSourceMappingURLLine(String content)
    {
        return content.substring(0, content.lastIndexOf("/*# sourceMappingURL="));
    }

    @Override
    protected String formatSourceMappingURL(String sourceMapName)
    {
        return "/*# sourceMappingURL=" + sourceMapName + " */\n";
    }


}
