/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize.css;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceException;

import org.ametys.core.DevMode;
import org.ametys.core.DevMode.DEVMODE;
import org.ametys.core.resources.SimpleResourceHandler;
import org.ametys.plugins.core.ui.resources.css.JSASSResourceURIExtensionPoint;

/**
 * Resource handler for minimized css
 */
public class MinimizedCSSResourceHandler extends SimpleResourceHandler
{
    /** JS minimize manager */
    protected MinimizeCSSManager _cssMinimizeManager;
    private JSASSResourceURIExtensionPoint _jsassResourceURIExtensionPoint;

    private String _minimizedUri;
    private String _originalUri;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        
        _cssMinimizeManager = (MinimizeCSSManager) manager.lookup(MinimizeCSSManager.ROLE);
        _jsassResourceURIExtensionPoint = (JSASSResourceURIExtensionPoint) manager.lookup(JSASSResourceURIExtensionPoint.ROLE);
    }
    
    @Override
    public Source setup(String location, Map objectModel, Parameters par, boolean readForDownload) throws IOException, ProcessingException
    {
        // The .min.css location
        _minimizedUri = location;

        Source source = null;
        try 
        {
            source = _resolver.resolveURI(location);
        } 
        catch (SourceException e) 
        {
            // Nothing
        }
        
        if (source == null || !source.exists())
        {
            // If the file does not exists at the original location, but it is a .min.css file, retrieve the .css file instead to process it
            _resolver.release(source);
            
            String sitemapURI = "/" + StringUtils.removeEnd(ObjectModelHelper.getRequest(objectModel).getSitemapURI(), ".min.css") + ".css";
            
            source = super.setup("cocoon:" + sitemapURI, objectModel, par, readForDownload);
            
            // The .css location
            _originalUri = sitemapURI;
        }
        
        _source = source;
        return _source;
    }
    
    @Override
    public Serializable getKey()
    {
        Serializable key = super.getKey();
        return key + "*" + _minimizedUri;
    }
    
    @Override
    public void generate(OutputStream out) throws IOException, ProcessingException
    {
        if (_originalUri == null)
        {
            _cssMinimizeManager.validateAndOutputMinimizedFile(_source, out, _minimizedUri);
            return;
        }
        
        // Retrieve the .css file and minimize it to serve a .min.css generated on the fly
        try (InputStream is = _source.getInputStream(); ByteArrayOutputStream bos = new ByteArrayOutputStream())
        {
            IOUtils.copy(is, bos);
            String code = bos.toString("UTF-8");
            
            // when lazy loading the source map, wait for the .map to be directly called to generate it
            boolean lazyLoadSourceMap = DEVMODE.PRODUCTION.equals(DevMode.getDeveloperMode(ObjectModelHelper.getRequest(_objectModel)));
            
            String resolvedOriginalUri = null;
            try
            {
                resolvedOriginalUri = _jsassResourceURIExtensionPoint.resolve(_originalUri);
            }
            catch (URISyntaxException e)
            {
                getLogger().error("Unable to resolve URI location because URI syntax '" + _originalUri + "' is not supported", e);
            }
            
            String result;
            if (resolvedOriginalUri == null || lazyLoadSourceMap)
            {
                result = _cssMinimizeManager.minimizeCss(code, resolvedOriginalUri);
            }
            else
            {
                result = _cssMinimizeManager.minimizeCss(code, _originalUri, resolvedOriginalUri + ".map", _source.getLastModified());
            }

            out.write(result.getBytes(StandardCharsets.UTF_8));
        }
    }
}
