/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize.css.sass;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceNotFoundException;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.TraversableSource;

import org.ametys.core.cocoon.source.ResourceSource;
import org.ametys.plugins.core.ui.resources.css.CSSFileHelper;
import org.ametys.plugins.core.ui.resources.css.JSASSResourceURIExtensionPoint;

import io.bit3.jsass.importer.Import;
import io.bit3.jsass.importer.Importer;

/**
 * Sass Importer which can resolve Ametys resources
 */
public class AmetysScssImporter implements Importer 
{
    private SourceResolver _resolver;
    private JSASSResourceURIExtensionPoint _jsassResourceURIExtensionPoint;
    private String _internalContextPath;
    private String _externalContextPath;

    /**
     * Default constructor for the Ametys SassImporter.
     * @param internalContextPath The internal context path
     * @param externalContextPath The external context path
     * @param resolver The source resolver
     * @param jsassResourceURIExtensionPoint The JSASS resourceUri extension point 
     */
    public AmetysScssImporter(String internalContextPath, String externalContextPath, SourceResolver resolver, JSASSResourceURIExtensionPoint jsassResourceURIExtensionPoint)
    {
        _internalContextPath = internalContextPath;
        _externalContextPath = externalContextPath;
        _resolver = resolver;
        _jsassResourceURIExtensionPoint = jsassResourceURIExtensionPoint;
    }
    
    public Collection<Import> apply(String url, Import previous)
    {
        String previousUri = Optional.of(previous)
                                     .map(Import::getAbsoluteUri)
                                     .map(URI::toString)
                                     .filter(uri -> !uri.endsWith(".css"))
                                     .orElse(null);
        
        if (previousUri == null || url.endsWith(".css") || url.startsWith("http://") || url.startsWith("https://") || url.startsWith("//"))
        {
            // returning null keeps the original @import mention
            return null;
        }
        
        List<Import> list = new LinkedList<>();
        
        try
        {
            boolean isUrlAbsolute = new URI(url).isAbsolute();
            String currentUri = isUrlAbsolute || url.startsWith("/") ? url : StringUtils.removeStart("/" + new URI(FilenameUtils.getFullPath(previousUri) + url).normalize(), _externalContextPath);
            
            Source importSource = _getImportSource(currentUri, _resolver);
            String importSourcePath = StringUtils.substringBefore(importSource.getURI(), "?");

            if (importSourcePath.endsWith(".scss") 
                    || importSourcePath.endsWith(".sass") 
                    || importSourcePath.endsWith(".css") && !currentUri.endsWith(".css")) // An url ending at the origin by .css should stay an import at the end (while a .css added automatically should be resolved)
            {
                String importText = IOUtils.toString(importSource.getInputStream(), StandardCharsets.UTF_8);
                
                if (importSourcePath.endsWith(".css"))
                {
                    String contextualizedUri = isUrlAbsolute ? _jsassResourceURIExtensionPoint.resolve(url) : url;
                    importText = CSSFileHelper.replaceRelativeUri(importText, contextualizedUri, _jsassResourceURIExtensionPoint, _internalContextPath, _externalContextPath);
                    currentUri = StringUtils.appendIfMissing(currentUri, ".css");
                }
                
                String uri = StringUtils.removeStart(StringUtils.prependIfMissing(_jsassResourceURIExtensionPoint.resolve(currentUri), _externalContextPath), "/");
                list.add(new Import(uri, StringUtils.appendIfMissing(uri, "." + FilenameUtils.getExtension(importSourcePath)), importText));
            }
            else
            {
                // returning null keeps the original @import mention
                return null;
            }
        }
        catch (URISyntaxException | IOException e) 
        {
            throw new RuntimeException(e);
        }

        return list;
    }
    
    private Source _getImportSource(String currentPath, SourceResolver resolver) throws URISyntaxException, IOException
    {
        List<String> pathMatching = new ArrayList<>();
        pathMatching.add(currentPath);
        
        // extension is optional
        pathMatching.add(currentPath + ".scss");
        pathMatching.add(currentPath + ".sass");
        pathMatching.add(currentPath + ".css");
        
        // add an underscore prefix to the file name for sass partial imports
        String name = FilenameUtils.getName(currentPath);
        String partialPath = currentPath.substring(0, currentPath.length() - name.length()) + "_" + name;
        pathMatching.add(partialPath);
        pathMatching.add(partialPath + ".scss");
        pathMatching.add(partialPath + ".sass");
        pathMatching.add(partialPath + ".css");
        
        for (String uri : pathMatching)
        {
            try
            {
                Source importSource = resolver.resolveURI(_jsassResourceURIExtensionPoint.resolvePath(uri));
                if (_isSourceValid(importSource))
                {
                    return importSource;
                }
            }
            catch (SourceNotFoundException e)
            {
                // source does not exists. Do nothing
            }
        }
        
        throw new URISyntaxException(currentPath, "Unable to resolve SASS import, no matching source found");
    }
    
    private boolean _isSourceValid(Source source)
    {
        if (!source.exists())
        {
            return false;
        }
        
        // ignore folders
        if (source instanceof TraversableSource && ((TraversableSource) source).isCollection())
        {
            return false;
        }
        
        // Folders inside a zip entry (such as a jar file) are not traversable
        if (source instanceof ResourceSource)
        {
            URL location = ((ResourceSource) source).getLocation();
            if (location != null && StringUtils.endsWith(location.toString(), "/"))
            {
                return false;
            }
        }
        
        return true;
    }
}

