/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize.css.sass;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.caching.CacheableProcessingComponent;
import org.apache.cocoon.reading.ServiceableReader;
import org.apache.commons.io.IOUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.SourceValidity;
import org.apache.excalibur.source.impl.validity.NOPValidity;
import org.xml.sax.SAXException;

import io.bit3.jsass.CompilationException;

/**
 * Reader that compile a sass file into css
 */
public class CompileSassReader extends ServiceableReader implements CacheableProcessingComponent
{
    
    private MinimizeSassManager _minimizeSassManager;
    private SourceResolver _resolver;

    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _minimizeSassManager = (MinimizeSassManager) smanager.lookup(MinimizeSassManager.ROLE);
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        try
        {
            Source src = _resolver.resolveURI(this.source);
            if (!src.exists())
            {
                throw new ProcessingException("Unable to resolve sass file '" + source + "' for compilation, source does not exist");
            }
            
            try (ByteArrayOutputStream bos = new ByteArrayOutputStream())
            {
                try (InputStream is = src.getInputStream())
                {
                    IOUtils.copy(is, bos);
                }
                
                String compileCss = _minimizeSassManager.compileCss(bos.toString(StandardCharsets.UTF_8.toString()), source, source, false, 0);
                IOUtils.write(compileCss, out, StandardCharsets.UTF_8);
                IOUtils.closeQuietly(out);
            }
            catch (URISyntaxException | CompilationException e)
            {
                throw new ProcessingException("Unable to compile sass file '" + source + "'", e);
            }
            
        }
        catch (IOException e)
        {
            throw new ProcessingException("Unable to resolve sass file '" + source + "' for compilation", e);
        }
    }

    @Override
    public Serializable getKey()
    {
        return this.source;
    }

    @Override
    public SourceValidity getValidity()
    {
        return new NOPValidity();
    }

    @Override
    public String getMimeType()
    {
        return "text/css";
    }
    
}
