/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize.css.sass;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.List;
import java.util.stream.Collectors;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.SourceResolver;

import org.ametys.core.minimize.SourceMapCache;
import org.ametys.core.minimize.css.JSASSFixHelper;
import org.ametys.core.resources.ProxiedContextPathProvider;
import org.ametys.plugins.core.ui.resources.css.CSSFileHelper;
import org.ametys.plugins.core.ui.resources.css.JSASSResourceURIExtensionPoint;
import org.ametys.plugins.core.ui.resources.css.sass.SassFunctionsProvider;
import org.ametys.plugins.core.ui.resources.css.sass.SassFunctionsProviderExtensionPoint;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

import io.bit3.jsass.CompilationException;
import io.bit3.jsass.Compiler;
import io.bit3.jsass.Options;
import io.bit3.jsass.Output;
import io.bit3.jsass.OutputStyle;

/**
 * Minimize manager for CSS files
 */
public class MinimizeSassManager extends AbstractLogEnabled implements Component, Serviceable, Initializable, Contextualizable
{
    /** The avalon ROLE */
    public static final String ROLE = MinimizeSassManager.class.getName();
    private Compiler _jsassCompiler;

    private JSASSResourceURIExtensionPoint _jsassResourceURIExtensionPoint;

    private SourceResolver _resolver;

    private List<SassFunctionsProvider> _sassFunctionsProviders;

    private SourceMapCache _sourceMapCache;
    
    private ProxiedContextPathProvider _proxiedContextPathProvider;
    
    private Context _context;
    
    public void initialize() throws Exception
    {
        _jsassCompiler = new Compiler();
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _jsassResourceURIExtensionPoint = (JSASSResourceURIExtensionPoint) smanager.lookup(JSASSResourceURIExtensionPoint.ROLE);
        _resolver = (SourceResolver) smanager.lookup(SourceResolver.ROLE);
        SassFunctionsProviderExtensionPoint sassFunctionsProviderEP = (SassFunctionsProviderExtensionPoint) smanager.lookup(SassFunctionsProviderExtensionPoint.ROLE);
        _sassFunctionsProviders = sassFunctionsProviderEP.getExtensionsIds().stream().map(id -> sassFunctionsProviderEP.getExtension(id)).collect(Collectors.toList());
        _sourceMapCache = (SourceMapCache) smanager.lookup(SourceMapCache.ROLE);
        _proxiedContextPathProvider = (ProxiedContextPathProvider) smanager.lookup(ProxiedContextPathProvider.ROLE);
    }
    
    
    /**
     * Compile a sass file. 
     * The source map of the file is generated and stored in the source map cache
     * @param content The content of the sass file
     * @param sassLocation The location of the sass file, e.g. plugin:core-ui:/resources/example/test.scss
     * @param location The location of the compiled css, e.g. plugin:core-ui:/resources/example/test.css or plugin:core-ui:/resources/example/test.min.css
     * @param minimize Minimize the output
     * @param lastModified The date of last modification
     * @return The content of the compiled, and optionally minimized, sass file
     * @throws URISyntaxException If an exception occurred
     * @throws CompilationException If an exception occurred
     * @throws IOException If an exception occurred
     */
    public String compileCss(String content, String sassLocation, String location, boolean minimize, long lastModified) throws URISyntaxException, CompilationException, IOException
    {
        String externalContextPath = _proxiedContextPathProvider.getContextPath();
        
        String contextualizedUri = null;
        try
        {
            // example: /plugins/core-ui/resources/example/test.scss
            contextualizedUri = _jsassResourceURIExtensionPoint.resolve(sassLocation);
        }
        catch (URISyntaxException e)
        {
            getLogger().error("Unable to resolve URI location because URI syntax '" + location + "' is not supported", e);
        }
        
        if (contextualizedUri != null && contextualizedUri.contains("://"))
        {
            throw new URISyntaxException(contextualizedUri, "Unable to compile Sass file, unsupported URI starts with unknown protocol");
        }
        
        // example: /EXTERNAL_CONTEXT/plugins/core-ui/resources/example/test.scss
        URI uri = contextualizedUri != null ? new URI(externalContextPath + contextualizedUri) : null;

        // example: test.css or test.min.css
        String locationFilename = FilenameUtils.getName(location);
        Output compiledString = _compileCss(content, locationFilename, uri, minimize, FilenameUtils.getExtension(sassLocation));
        String compileResult = compiledString.getCss();
        
        if (uri != null)
        {
            String result = CSSFileHelper.replaceRelativeResourcesUri(compileResult, contextualizedUri, _jsassResourceURIExtensionPoint, externalContextPath);
            if (!result.equals(compileResult))
            {
                getLogger().warn("Relative path found in '{}'. Relative paths inside SASS files need to be avoided, as it will offset the source mapping. Please use functions ('pluginUrl()' or similar helpers) instead.", uri.toString());
            }
            
            String sourceMap = compiledString.getSourceMap();
            // example: /EXTERNAL_CONTEXT/plugins/core-ui/resources/example/test.scss.map or /EXTERNAL_CONTEXT/plugins/core-ui/resources/example/test.css.map
            String sourceMapName = !minimize ? uri.toString() + ".map" : StringUtils.removeEnd(uri.toString(), ".scss") + ".css.map";
            
            sourceMap = JSASSFixHelper.fixJsassSourceMap(sourceMap, externalContextPath);
    
            // example: /plugins/core-ui/resources/example/test.scss.map or /plugins/core-ui/resources/example/test.css.map
            String cacheKey = StringUtils.removeStart(sourceMapName, externalContextPath);
            _sourceMapCache.put(cacheKey, sourceMap, lastModified);
            
            result += "/*# sourceMappingURL=" + FilenameUtils.getName(sourceMapName) + " */\n";
            return result;
        }
        
        return compileResult;
    }

    
    /**
     * Generate the source map of compiled sass file. 
     * The source map of the file is generated and stored in the source map cache
     * @param content The content of the file
     * @param location The location of the source map file, e.g. plugin:core-ui:/resources/example/test.scss.map or plugin:core-ui:/resources/example/test.css.map 
     * @param extension The extension of the sass file, e.g. scss or sass
     * @param internalContextPath The internal context path
     * @param externalContextPath The external context path
     * @param minimize Minimize the output
     * @param lastModified The date of last modification
     * @return The content of the source map, or null if no source map was generated
     * @throws URISyntaxException If an exception occurred
     * @throws CompilationException If an exception occurred
     * @throws IOException If an exception occurred
     */
    public String generateCssSourceMap(String content, String location, String extension, String internalContextPath, String externalContextPath, boolean minimize, long lastModified) throws URISyntaxException, CompilationException, IOException
    {
        String contextualizedUri = null;
        try
        {
            // example: /plugins/core-ui/resources/example/test.scss or /plugins/core-ui/resources/example/test.css
            contextualizedUri = _jsassResourceURIExtensionPoint.resolve(StringUtils.removeEnd(location, ".map"));
        }
        catch (URISyntaxException e)
        {
            getLogger().error("Unable to resolve URI location because URI syntax '" + location + "' is not supported", e);
            return null; // No source map generated
        }
        
        if (contextualizedUri.contains("://"))
        {
            throw new URISyntaxException(contextualizedUri, "Unable to compile Sass file, unsupported URI '" + location + "' contains unknown protocol");
        }

        // example: /EXTERNAL_CONTEXT/plugins/core-ui/resources/example/test.scss
        URI sassUri = new URI(externalContextPath + (!minimize ? contextualizedUri : StringUtils.removeEnd(contextualizedUri, ".css") + "." + extension));

        // example: test.css
        String locationFilename = FilenameUtils.getName(StringUtils.removeEnd(sassUri.toString(), "." + extension) + (!minimize ? ".css" : ".min.css"));
        Output compiledString = _compileCss(content, locationFilename, sassUri, minimize, extension);
        
        String sourceMap = compiledString.getSourceMap();
        // example: /EXTERNAL_CONTEXT/plugins/core-ui/resources/example/test.scss or /EXTERNAL_CONTEXT/plugins/core-ui/resources/example/test.css
        URI uri = new URI(externalContextPath + contextualizedUri);        
        // example: /EXTERNAL_CONTEXT/plugins/core-ui/resources/example/test.scss.map or /EXTERNAL_CONTEXT/plugins/core-ui/resources/example/test.css.map
        String sourceMapName = !minimize ? uri.toString() + ".map" : StringUtils.removeEnd(uri.toString(), "." + extension) + ".css.map";
        
        sourceMap = JSASSFixHelper.fixJsassSourceMap(sourceMap, externalContextPath);
        
        // example: /plugins/core-ui/resources/example/test.scss.map or /plugins/core-ui/resources/example/test.css.map
        String cacheKey = StringUtils.removeStart(sourceMapName, externalContextPath);
        _sourceMapCache.put(cacheKey, sourceMap, lastModified);
        
        return sourceMap;
    }
    
    /**
     * Compile a SCSS and output the JSASS result containing the content and source map
     * @param content The content of the file
     * @param fileName The name of the compiled sass file, e.g. test.css or test.min.css
     * @param uri The contextualized uri of the sass file, e.g. /EXTERNAL_CONTEXT/plugins/core-ui/resources/example/test.scss
     * @param minimize Minimize the output
     * @param extension The sass extension such as scss or sass
     * @return The output of the compilation
     * @throws URISyntaxException if an error occurred with the uri syntax
     * @throws CompilationException if an error occured when compiling the sass
     */
    private Output _compileCss(String content, String fileName, URI uri, boolean minimize, String extension) throws URISyntaxException, CompilationException
    {
        String externalContextPath = _proxiedContextPathProvider.getContextPath();

        Options options = new Options();
        AmetysScssImporter importer = new AmetysScssImporter(ContextHelper.getRequest(_context).getContextPath(), externalContextPath, _resolver, _jsassResourceURIExtensionPoint);
        options.getImporters().add(importer);
        options.getFunctionProviders().addAll(_sassFunctionsProviders);
        
        if (minimize)
        {
            options.setOutputStyle(OutputStyle.COMPRESSED);
        }
        
        URI inputPath;
        URI outputPath;
        if (fileName != null && uri != null)
        {
            options.setSourceMapFile(new URI("unused")); // Parameter required to generate the source map, but value is unused
            options.setOmitSourceMapUrl(true);
            
            // Transform the absolute path in relative path "/contextPath/plugins" becomes "./contextPath/plugins" to prevent jsass messing with it
            // The relative path is transformed back into absolute in post-processing of the source map
            inputPath = new URI("." + uri.toString());
            outputPath = new URI(fileName);
        }
        else
        {
            // Compile the sass file as best as possible, due to an unknown location. No source map will be generated
            options.setSourceMapFile(null);
            inputPath = new URI("unknown-location");
            outputPath = new URI("unknown-location");
        }
        
        return _jsassCompiler.compileString(content, inputPath, outputPath, options);
    }
}
