/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize.js;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.Locale;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;

import org.ametys.core.minimize.AbstractSourceMapResourceHandler;

/**
 * Resource handler for source map files
 */
public class JSSourceMapResourceHandler extends AbstractSourceMapResourceHandler implements Contextualizable
{
    private static final Pattern __LOCALE_PATTERN = Pattern.compile("^(.*resources/.*)\\.([^/.]+)\\.js$");
    
    /** JS minimize manager */
    protected MinimizeJSManager _jSMinimizeManager;
    
    private Context _context;
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _jSMinimizeManager = (MinimizeJSManager) manager.lookup(MinimizeJSManager.ROLE);
    }
    
    @Override
    protected Source _getAlternateSource(String location, Map objectModel) throws ProcessingException, IOException
    {
        Source source = null;
        // 1. Check if the source map exist with and without the .min
        if (location.endsWith(".min.js.map"))
        {
            source = _getExistingSource(location.substring(0, location.length() - ".min.js.map".length()) + ".js.map");
        }
        else
        {
            source = _getExistingSource(location.substring(0, location.length() - ".js.map".length()) + ".min.js.map");
        }
        
        if (source == null)
        {
            // 2. check if the source map is in cache
            String localizedLocation = _getLocalizedLocation(location);
            source = _sourceMapCache.get(localizedLocation);

            if (source == null)
            {
                // 3. generate the source map and store it in cache
                source = _generateJSSourceMap(objectModel, localizedLocation);
            }
        }
        
        return source;
    }
    
    private String _getLocalizedLocation(String location)
    {
        Matcher matcher = __LOCALE_PATTERN.matcher(location);
        String locale = matcher.matches() ? matcher.group(2) : null;
        
        if (locale == null)
        {
            // Default locale
            Map objectModel = ContextHelper.getObjectModel(_context);
            locale = org.apache.cocoon.i18n.I18nUtils.findLocale(objectModel, "locale", null, Locale.getDefault(), true).getLanguage();
        }
        
        return location + (locale != null ? "#" + locale : "");
    }
    

    private Source _generateJSSourceMap(Map objectModel, String localizedLocation) throws ProcessingException, IOException
    {
        String originalUri = "cocoon:/" + StringUtils.removeEnd(ObjectModelHelper.getRequest(objectModel).getSitemapURI(), ".map");
        
        Source sourceResolved = _resolver.resolveURI(originalUri);
        if (sourceResolved != null && sourceResolved.exists())
        {
            try (InputStream inputStream = sourceResolved.getInputStream())
            {
                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                IOUtils.copy(inputStream, baos);
                
                // Stores the source of the sourceMap in the cache
                _jSMinimizeManager.generateJSSourceMap(baos.toString(StandardCharsets.UTF_8.toString()), originalUri, localizedLocation, sourceResolved.getLastModified());
                
                return _sourceMapCache.get(localizedLocation);
            }
        }
        else
        {
            _resolver.release(sourceResolved);
        }
        
        return null;
    }
}
