/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize.js;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.Collections;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.avalon.framework.component.Component;
import org.apache.cocoon.ProcessingException;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.exception.ExceptionUtils;
import org.apache.commons.text.StringEscapeUtils;
import org.apache.excalibur.source.Source;

import org.ametys.core.minimize.AbstractMinimizeManager;

import com.google.debugging.sourcemap.SourceMapGeneratorV3;
import com.google.debugging.sourcemap.SourceMapParseException;
import com.google.javascript.jscomp.CompilationLevel;
import com.google.javascript.jscomp.Compiler;
import com.google.javascript.jscomp.CompilerOptions;
import com.google.javascript.jscomp.CompilerOptions.TracerMode;
import com.google.javascript.jscomp.JSError;
import com.google.javascript.jscomp.Result;
import com.google.javascript.jscomp.SourceFile;
import com.google.javascript.jscomp.WarningLevel;

/**
 * Manager for JS minimization and source map generation
 */
public class MinimizeJSManager extends AbstractMinimizeManager implements Component
{
    /** The avalon ROLE */
    public static final String ROLE = MinimizeJSManager.class.getName();

    // regex that matches the "file" property of the json sourcemap
    private static final Pattern __SOURCEMAP_FILE_NAME = Pattern.compile(".*\"file\"\\s*:" // Matches the literal '"file" :' with any whitespace
                                                                       + "\\s*"
                                                                       + "\"([^\"]+)\".*", // Captures a string in between quotes 
                                                                       Pattern.MULTILINE | Pattern.DOTALL);
    
    private static final Pattern __SOURCEMAP_LINE = Pattern.compile("^(.*)(\n//# sourceMappingURL=.*)\r?\n?$", Pattern.DOTALL);
    
    /**
     * Minimize a JS string
     * @param code The JS code
     * @param location The JS location
     * @return The minimized JS
     * @throws ProcessingException If an error occurred during minimization
     * @throws IOException If an error occurred while retrieving the source map
     */
    public String minimizeJS(String code, String location) throws ProcessingException, IOException
    {
        return minimizeJS(code, location, null, null);
    }
    
    /**
     * Minimize a JS string
     * @param code The JS code
     * @param location The JS location
     * @param sourceMapKey The key to store the source map in cache. Can be null to prevent source map generation
     * @param lastModified The last modified date of the code, used to assert validity of the source map cache. Can be null
     * @return The minimized JS
     * @throws ProcessingException If an error occurred during minimization
     * @throws IOException If an error occurred while retrieving the source map
     */
    public String minimizeJS(String code, String location, String sourceMapKey, Long lastModified) throws IOException, ProcessingException
    {
        String fileName = StringUtils.substringAfterLast(location, "/");
        boolean withSourceMap = sourceMapKey != null && lastModified != null;
        Compiler compiler = _minimizeJS(code, fileName, withSourceMap);
        
        String sourceResult = compiler.toSource();
        if (withSourceMap)
        {
            StringBuilder sb = new StringBuilder();
            
            compiler.getSourceMap().appendTo(sb, fileName);
            _sourceMapCache.put(sourceMapKey, sb.toString(), lastModified);
            
            sourceResult += "\n" + formatSourceMappingURL(fileName + ".map");
        }
        
        return sourceResult;
    }
    
    /**
     * Minimize a JS string and return the source map
     * @param code The JS code
     * @param location The JS location
     * @param sourceMapKey The key to store the source map in cache
     * @param lastModified The last modified date of the code, used to assert validity of the source map cache
     * @return The minimized JS
     * @throws ProcessingException If an error occurred during minimization
     * @throws IOException If an error occurred while retrieving the source map
     */
    public String generateJSSourceMap(String code, String location, String sourceMapKey, Long lastModified) throws ProcessingException, IOException
    {
        String fileName = location.substring(location.lastIndexOf("/") + 1);
        Compiler compiler = _minimizeJS(code, fileName, true);

        StringBuilder sb = new StringBuilder();
        compiler.toSource(); // required to generate the JS, and the source Map
        compiler.getSourceMap().appendTo(sb, fileName);
        String sourceMap = sb.toString();
        _sourceMapCache.put(sourceMapKey, sourceMap, lastModified);
        
        return sourceMap;
    }
    
    private Compiler _minimizeJS(String code, String fileName, boolean withSourceMap) throws ProcessingException
    {
        Compiler compiler = new Compiler();
        CompilerOptions options = new CompilerOptions();
        
        options.setTracerMode(TracerMode.OFF);
        WarningLevel.QUIET.setOptionsForWarningLevel(options);
        if (withSourceMap)
        {
            options.setSourceMapOutputPath("unused");
        }
        options.setEmitUseStrict(false);
        options.setOutputCharset(StandardCharsets.UTF_8);
        
        CompilationLevel level = CompilationLevel.WHITESPACE_ONLY;
        level.setOptionsForCompilationLevel(options);
        
        SourceFile file = SourceFile.fromCode(fileName, code);
        Result result = compiler.compile(Collections.emptyList(), Arrays.asList(file), options);
        if (result.errors.size() > 0)
        {
            StringBuilder errorsString = new StringBuilder();
            errorsString.append("Unable to minimize " + fileName);
            for (JSError error : result.errors)
            {
                errorsString.append("\n" + error.toString());
            }
            throw new ProcessingException(errorsString.toString());
        }
        
        return compiler;
    }

    @Override
    protected void addSourceMap(SourceMapGeneratorV3 sourceMapGenerator, int lineCount, String fileContent, String fileUri, String sourceMapUri)
    {
        try
        {
            String sourceMapCacheKey = fileUri.indexOf('/') > -1 ? fileUri.substring(0, fileUri.lastIndexOf("/") + 1) + sourceMapUri : sourceMapUri;
            
            Source mapSource = _sourceMapCache.get(sourceMapCacheKey);
            
            if (mapSource == null)
            {
                // If not found, get source map content from pipeline .css.map, e.g. if the js file was already minified
                String uriToResolve = new URI(sourceMapCacheKey).isAbsolute() ? sourceMapCacheKey : "cocoon:/" + sourceMapCacheKey;
                
                try
                {
                    mapSource = _resolver.resolveURI(uriToResolve);
                }
                catch (IOException e)
                {
                    // Nothing
                }
            }
            
            if (mapSource != null && mapSource.exists())
            {
                String sourceMapContent;
                try (InputStream is = mapSource.getInputStream())
                {
                    sourceMapContent = IOUtils.toString(is, StandardCharsets.UTF_8);
                }
                
                Matcher fileMatcher = __SOURCEMAP_FILE_NAME.matcher(sourceMapContent);
                if (fileMatcher.matches())
                {
                    sourceMapContent = sourceMapContent.replaceFirst("\"file\":\"" + fileMatcher.group(1) + "\"", "\"file\":\"" + _proxiedContextPathProvider.getContextPath() + fileUri + "\"");
                }
                
                sourceMapContent = super.convertSourceMapURIs(sourceMapContent, fileUri);
                
                // Add 1 to the line count to jump after the comment "/** File : */ 
                // and 1 to the line count for document_currentScript"
                sourceMapGenerator.mergeMapSection(lineCount + 2, 0, sourceMapContent);
            }
        }
        catch (SourceMapParseException | IOException | URISyntaxException | IllegalStateException e)
        {
            // IllegalStateException occurs in "sourceMapGenerator.mergeMapSection" when a previous invalid source map offsets the line count too much
            getLogger().error("Unable to retrieve source map when aggregating hash file, for sourceMappingURL '" + sourceMapUri + "' of file '" + fileUri + "'", e);
        }
    }


    @Override
    protected String getMinimizedContent(String uri, String nestedParentFilesName)
    {
        StringBuffer sb = new StringBuffer();
        
        sb.append("/** File : " + uri + " */\n");
        sb.append("document_currentScript = (function() { a = document.createElement('script'); a.src='" + _proxiedContextPathProvider.getContextPath() + uri + "'; return a; })(); try {\n");
        
        Source jssource = null;
        String suffix = null;
        try
        {
            String minimizeFileUri = StringUtils.endsWith(uri, ".min.js") ? uri : StringUtils.removeEnd(uri, ".js") + ".min.js";
            URI minimizeUri = new URI(minimizeFileUri);
            
            String uriToResolve = minimizeUri.isAbsolute() ? minimizeFileUri : "cocoon:/" + org.apache.cocoon.util.NetUtils.normalize(minimizeFileUri);
            jssource = _resolver.resolveURI(uriToResolve);
            
            String s;
            try (InputStream is = jssource.getInputStream())
            {
                s = IOUtils.toString(is, StandardCharsets.UTF_8);
            }
            
            s = s.replaceAll("document\\.currentScript", "document_currentScript");

            // We extract the sourcemap comment, to put it after the "catch"
            Matcher matcher = __SOURCEMAP_LINE.matcher(s);
            if (matcher.matches())
            {
                s = matcher.group(1);
                suffix = matcher.group(2) + "\n"; // \n is optional in the regexp. we force it to be present this way. 
            }
            
            sb.append(s);
        }
        catch (Exception e)
        {
            getLogger().error("Cannot minimize JS for file '" + uri + "'.", e);
            sb.append("throw new Error('" + StringEscapeUtils.escapeEcmaScript(ExceptionUtils.getRootCause(e).getMessage()) + "');");
        }
        finally
        {
            _resolver.release(jssource);
            
            sb.append("} catch (e) { console.error(\"" + uri + "\\n\", e) }");
            if (StringUtils.isNotBlank(suffix))
            {
                sb.append(suffix);
            }
        }

        sb.append("\n");
        return sb.toString();
    }

    @Override
    protected boolean isSourceMappingURLLine(String line)
    {
        return line != null && line.startsWith("//# sourceMappingURL=");
    }

    @Override
    protected String getSourceMappingURL(String line)
    {
        return line.substring("//# sourceMappingURL=".length()).trim();
    }

    @Override
    protected String removeSourceMappingURLLine(String content)
    {
        return content.substring(0, content.lastIndexOf("//# sourceMappingURL="));
    }

    @Override
    protected String formatSourceMappingURL(String sourceMapName)
    {
        return "//# sourceMappingURL=" + sourceMapName + "\n";
    }

}
