/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.minimize.js;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.nio.charset.StandardCharsets;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceException;

import org.ametys.core.DevMode;
import org.ametys.core.DevMode.DEVMODE;
import org.ametys.core.resources.JSResourceHandler;

/**
 * Resource handler for minimized javascript file
 */
public class MinimizedJSResourceHandler extends JSResourceHandler
{
    /** JS minimize manager */
    protected MinimizeJSManager _jSMinimizeManager;

    private String _realFileUri;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _jSMinimizeManager = (MinimizeJSManager) manager.lookup(MinimizeJSManager.ROLE);
    }
    
    @Override
    public Source setup(String location, Map objectModel, Parameters par, boolean readForDownload) throws IOException, ProcessingException
    {
        _requestedLocation = location;
        _objectModel = objectModel;
        
        // Cached
        _setExpiresToNoCache = false;

        Source source = null;
        try 
        {
            source = _resolver.resolveURI(location);
        } 
        catch (SourceException e) 
        {
            // Nothing
        }
        
        if (source == null || !source.exists())
        {
            // If the file does not exists at the original location, but it is a .min.js file, retrieve the .js file instead to process it
            _resolver.release(source);
            
            String realLocation = location.substring(0, location.length() - ".min.js".length()) + ".js";
            source = super.setup(realLocation, objectModel, par, readForDownload);
            
            // The .js location
            _realFileUri = realLocation;
        }
        
        _source = source;
        return _source;
    }
    
    @Override
    public Serializable getKey()
    {
        Serializable key = super.getKey();
        return key + "*" + _requestedLocation;
    }
    
    @Override
    public void generate(OutputStream out) throws IOException, ProcessingException
    {
        if (_realFileUri == null)
        {
            _jSMinimizeManager.validateAndOutputMinimizedFile(_source, out, _requestedLocation);
            return;
        }
        
        // Retrieve the .js file and minimize it to serve a .min.js generated on the fly
        try (ByteArrayOutputStream bos = new ByteArrayOutputStream())
        {
            super.generate(bos);
            
            String code = bos.toString("UTF-8");
            String locale = getLocale();
            
            // when lazy loading the source map, wait for the .map to be directly called to generate it
            boolean lazyLoadSourceMap = DEVMODE.PRODUCTION.equals(DevMode.getDeveloperMode(ObjectModelHelper.getRequest(_objectModel)));
            String result;
            if (lazyLoadSourceMap)
            {
                result = _jSMinimizeManager.minimizeJS(code, _realFileUri);
            }
            else
            {
                result = _jSMinimizeManager.minimizeJS(code, _realFileUri, _realFileUri + ".map" + (locale != null ? "#" + locale : ""), _source.getLastModified());
            }

            out.write(result.getBytes(StandardCharsets.UTF_8));
        }
    }
}
