/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.runtime.model.AbstractModelItem;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.disableconditions.DisableCondition;
import org.ametys.runtime.model.disableconditions.DisableConditions;

/**
 * Helper for model items
 */
public class ModelItemHelper implements Component, Serviceable
{
    /** Avalon ROLE. */
    public static final String ROLE = ModelItemHelper.class.getName();
    
    public void service(ServiceManager manager) throws ServiceException
    {
        AbstractModelItem.setServiceManager(manager);
    }
    
    /**
     * Converts the model item's disable conditions in a JSON map
     * @param modelItem The model item containing the disable conditions
     * @return The model item's disable conditions as a JSON map
     */
    public Map<String, Object> disableConditionsToJSON(ModelItem modelItem)
    {
        return _disableConditionsToJSON(modelItem.getDisableConditions());
    }
    
    /**
     * Converts the given disable conditions in a JSON map
     * @param disableConditions the disable conditions to convert
     * @return The disable conditions as a JSON map
     */
    private Map<String, Object> _disableConditionsToJSON(DisableConditions disableConditions)
    {
        Map<String, Object> map = new HashMap<>();
        
        // Handle simple conditions
        List<Map<String, String>> disableConditionList = new ArrayList<>();
        map.put("condition", disableConditionList);
        for (DisableCondition disableCondition : disableConditions.getConditions())
        {
            Map<String, String> disableConditionAsMap = disableConditionToJSON(disableCondition);
            disableConditionList.add(disableConditionAsMap);
        }

        // Handle nested conditions
        List<Map<String, Object>> disableConditionsList = new ArrayList<>();
        map.put("conditions", disableConditionsList);
        for (DisableConditions subDisableConditions : disableConditions.getSubConditions())
        {
            Map<String, Object> disableConditionsAsMap = _disableConditionsToJSON(subDisableConditions);
            disableConditionsList.add(disableConditionsAsMap);
        }
        
        // Handle type
        map.put("type", disableConditions.getAssociationType().toString().toLowerCase());
        
        return map;
    }

    /**
     * Converts the given disable condition in a JSON map
     * @param disableCondition the disable condition to convert
     * @return The disable condition as a JSON map
     */
    public Map<String, String> disableConditionToJSON(DisableCondition disableCondition)
    {
        Map<String, String> map = new HashMap<>();
        map.put("id", disableCondition.getId());
        map.put("operator", disableCondition.getOperator().toString().toLowerCase());
        map.put("value", disableCondition.getValue());
        return map;
    }
}
