/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.model.type;

import java.time.Instant;
import java.time.ZonedDateTime;
import java.time.format.DateTimeParseException;
import java.util.Date;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Triple;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.core.util.DateUtils;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;

/**
 * Abstract class for date time element types
 */
public abstract class AbstractDateTimeElementType extends AbstractElementType<ZonedDateTime>
{
    /**
     * An ISO date-time pattern that parses a date-time.
     * The fraction of second and/or the offset can also be provided
     */
    private static final String __ISO_DATETIME_PATTERN_WITH_OPTIONAL_SECOND_FRACTION_AND_OFFSET = "uuuu-MM-dd'T'HH:mm:ss[.SSS]XXX";
    
    @Override
    public ZonedDateTime convertValue(Object value) throws BadItemTypeException
    {
        if (value instanceof String stringValue)
        {
            if (StringUtils.isEmpty(stringValue))
            {
                return null;
            }
            
            try
            {
                // For legacy purpose, use the date pattern that can parse date with or without time
                return DateUtils.parseZonedDateTime(stringValue, __ISO_DATETIME_PATTERN_WITH_OPTIONAL_SECOND_FRACTION_AND_OFFSET);
            }
            catch (DateTimeParseException e)
            {
                throw new BadItemTypeException("Unable to cast '" + value + "' to a date", e);
            }
        }
        else if (value instanceof Date dateValue)
        {
            return DateUtils.asZonedDateTime(dateValue);
        }
        else if (value instanceof Instant instantValue)
        {
            return DateUtils.asZonedDateTime(instantValue);
        }
        
        return super.convertValue(value);
    }
    
    @Override
    public String toString(ZonedDateTime value)
    {
        if (value == null)
        {
            return null;
        }
        return DateUtils.zonedDateTimeToString(value);
    }
    
    @Override
    protected Object _singleTypedValueToJSON(ZonedDateTime value, DataContext context)
    {
        return toString(value);
    }
    
    @Override
    protected void _singleTypedNotEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, ZonedDateTime value, DataContext context, AttributesImpl attributes) throws SAXException
    {
        String valueAsString = DateUtils.zonedDateTimeToString(value);
        XMLUtils.createElement(contentHandler, tagName, attributes, valueAsString);
    }
    
    @Override
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareSingleValues(ZonedDateTime value1, ZonedDateTime value2)
    {
        return Stream.of(ModelItemTypeHelper.compareSingleDates(value1, value2, StringUtils.EMPTY))
                     .filter(Optional::isPresent)
                     .map(Optional::get);
    }
    
    public boolean isSimple()
    {
        return true;
    }
}
