/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.model.type;

import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.Locale;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Triple;

import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;
import org.ametys.runtime.model.exception.BadItemTypeException;

/**
 * Abstract class for double element types
 */
public abstract class AbstractDoubleElementType extends AbstractElementType<Double>
{
    private static final ThreadLocal<DecimalFormat> __FORMATTER = new ThreadLocal<>();

    @Override
    public Double convertValue(Object value) throws BadItemTypeException
    {
        if (value instanceof Number)
        {
            return ((Number) value).doubleValue();
        }
        else if (value instanceof String)
        {
            if (StringUtils.isEmpty((String) value))
            {
                return null;
            }
            
            try
            {
                return Double.valueOf((String) value);
            }
            catch (NumberFormatException e)
            {
                throw new BadItemTypeException("Unable to cast '" + value + "' to a double", e);
            }
        }
        
        return super.convertValue(value);
    }
    
    @Override
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareSingleValues(Double value1, Double value2)
    {
        return Stream.of(ModelItemTypeHelper.compareSingleNumbers(value1, value2, StringUtils.EMPTY))
                     .filter(Optional::isPresent)
                     .map(Optional::get);
    }
    
    @Override
    public boolean isCompatible(Object value)
    {
        return super.isCompatible(value) || value instanceof double[] || value instanceof Float || value instanceof Float[] || value instanceof float[];
    }
    
    public boolean isSimple()
    {
        return true;
    }
    
    @Override
    public String toString(Double value)
    {
        if (value == null)
        {
            return null;
        }
        DecimalFormat df = __FORMATTER.get(); 
        if (df == null)
        {
            df = new DecimalFormat("0.0", DecimalFormatSymbols.getInstance(Locale.ENGLISH)); // English to have . as fraction separator
            df.setMaximumFractionDigits(340); // 340 is the max supporter by the implementation (DecimalFormat.DOUBLE_FRACTION_DIGITS)
            
            __FORMATTER.set(df);
        }
        return df.format(value);
    }    
}
