/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.model.type;

import java.lang.reflect.Array;
import java.lang.reflect.ParameterizedType;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.BiFunction;
import java.util.stream.Stream;

import javax.xml.transform.TransformerException;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.ConfigurationUtil;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.ImmutableTriple;
import org.apache.commons.lang3.tuple.Triple;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.core.util.JSONUtils;
import org.ametys.core.util.dom.DOMUtils;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.Enumerator;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;
import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ElementType;

/**
 * Abstract class for element types
 * @param <T> Type of the element value
 */
@SuppressWarnings("unchecked")
public abstract class AbstractElementType<T> extends AbstractModelItemType implements ElementType<T>, Component, Serviceable
{
    /** The JSON conversion utilities. */
    protected JSONUtils _jsonUtils;

    private Class<T> _type;
    private Class<T[]> _typeArray;
    
    {
        // Get the <T> class bu going up in the implementation untill the superclass is this abstract class (AbstractElementType)
        Class c = this.getClass();
        while (!c.getSuperclass().equals(AbstractElementType.class))
        {
            c = c.getSuperclass();
        }

        // The super class is now AbstractElementType, which is Parameterized, so we can get the type arguments
        _type = (Class<T>) ((ParameterizedType) c.getGenericSuperclass()).getActualTypeArguments()[0];
        try
        {
            // We have the class, so we can create a <T>[] by using the syntax [L.....; in Class.forName
            _typeArray = (Class<T[]>) Class.forName("[L" + _type.getName() + ";");
        }
        catch (ClassNotFoundException e)
        {
            throw new IllegalStateException("Impossible to find class '" + _type.getName() + "' in an implementation of AbstractElementType", e);
        }
    }
    
    public void service(ServiceManager manager) throws ServiceException
    {
        if (manager.hasService(JSONUtils.ROLE))
        {
            _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
        }
    }
    
    public String toString(T value)
    {
        if (value != null)
        {
            return value.toString();
        }
        return null;
    }
    
    public Object fromJSONForClient(Object json, DataContext context) throws BadItemTypeException
    {
        if (json == null)
        {
            return null;
        }
        else if (json instanceof List jsonList)
        {
            Object arrayValue = Array.newInstance(getManagedClass(), jsonList.size());
            for (int i = 0; i < jsonList.size(); i++)
            {
                Object singleJSONValue = jsonList.get(i);
                T singleValue = (T) fromJSONForClient(singleJSONValue, context);
                Array.set(arrayValue, i, singleValue);
            }
            
            return arrayValue;
        }
        else if (_isEnumeratedJSONObject(json))
        {
            Object value = ((Map) json).get("value");
            return _singleValueFromJSON(value, context);
        }
        else
        {
            return _singleValueFromJSON(json, context);
        }
    }
    
    /**
     * Check if the given json object represents a value from an enumeration 
     * @param json the json object to check
     * @return <code>true</code> if the given json object represents a value from an enumeration, <code>false</code> otherwise
     */
    protected boolean _isEnumeratedJSONObject(Object json)
    {
        return json instanceof Map jsonMap && jsonMap.size() == 2 && jsonMap.containsKey("value") && jsonMap.containsKey("label");
    }
    
    /**
     * Convert the given client side JSON object to the typed single value 
     * @param json the JSON object to convert
     * @param context the context of the data to retrieve
     * @return the typed single value corresponding to the JSON object
     * @throws BadItemTypeException if the given json can not be converted into the typed single value
     */
    protected T _singleValueFromJSON(Object json, DataContext context) throws BadItemTypeException
    {
        return castValue(json);
    }
    
    public Object valueToJSONForClient(Object value, DataContext context) throws BadItemTypeException
    {
        return _valueToJSON(value, context, this::_singleValueToJSON);
    }
    
    public Object valueToJSONForEdition(Object value, DataContext context) throws BadItemTypeException
    {
        return _valueToJSON(value, context, this::_singleValueToJSONForEdition);
    }
    
    /**
     * Convert the value into a JSON object using the given function
     * @param value the value to convert
     * @param context The context of the data to convert
     * @param singleValueToJSONFunction the function to apply to each single value to convert it into a JSON object
     * @return The value as JSON
     * @throws BadItemTypeException if the given value is not compatible is the current type
     */
    protected Object _valueToJSON(Object value, DataContext context, BiFunction<Object, DataContext, Object> singleValueToJSONFunction) throws BadItemTypeException
    {
        if (value == null)
        {
            return null;
        }
        else if (value.getClass().isArray())
        {
            return Stream.of((Object[]) value)
                .filter(Objects::nonNull)
                .map(v -> _singleValueToJSON(v, context, singleValueToJSONFunction))
                .filter(Objects::nonNull)
                .toList();
        }
        else
        {
            return _singleValueToJSON(value, context, singleValueToJSONFunction);
        }
    }
    
    /**
     * Convert the single value into a JSON object using the given function
     * @param value the value to convert
     * @param context The context of the data to convert
     * @param singleValueToJSONFunction the function to apply to the given value to convert it into a JSON object
     * @return The single value as JSON
     * @throws BadItemTypeException if the given value is not compatible is the current type
     */
    protected Object _singleValueToJSON(Object value, DataContext context, BiFunction<Object, DataContext, Object> singleValueToJSONFunction) throws BadItemTypeException
    {
        Object jsonValue = singleValueToJSONFunction.apply(value, context);
        
        Optional<Enumerator> optionalEnumerator = context.getModelItem()
                                                         .filter(ElementDefinition.class::isInstance)
                                                         .map(ElementDefinition.class::cast)
                                                         .map(ElementDefinition::getEnumerator);
        
        if (optionalEnumerator.isPresent())
        {
            Enumerator enumerator = optionalEnumerator.get();
            I18nizableText i18n = null;
            
            try
            {
                i18n = enumerator.getEntry(value);
            }
            catch (Exception e)
            {
                getLogger().warn("JSON conversion for data '" + context.getFullDataPath() + "' requires a label for enumerated value", e);
            }
            
            if (i18n != null)
            {
                return Map.of("value", jsonValue,
                              "label", i18n);
            }
            else
            {
                return jsonValue;
            }
        }
        else
        {
            return jsonValue;
        }
    }
    
    /**
     * Convert the single value into a JSON object
     * @param value the value to convert
     * @param context The context of the data to convert
     * @return The value as JSON
     * @throws BadItemTypeException if the given value is not compatible is the current type
     */
    protected Object _singleValueToJSON(Object value, DataContext context) throws BadItemTypeException
    {
        if (getManagedClass().isInstance(value))
        {
            return _singleTypedValueToJSON(getManagedClass().cast(value), context);
        }
        else
        {
            throw new BadItemTypeException("Try to convert the non " + getId() + " JSON object '" + value + "' into a " + getManagedClass().getName());
        }
    }
    
    /**
     * Convert the single value into a JSON object
     * @param value the value to convert
     * @param context The context of the data to convert
     * @return The value as JSON
     */
    protected Object _singleTypedValueToJSON(T value, DataContext context)
    {
        return value;
    }
    
    /**
     * Convert the single value into a JSON object to use in edition mode
     * @param value the value to convert
     * @param context The context of the data to convert
     * @return The value as JSON
     * @throws BadItemTypeException if the given value is not compatible is the current type
     */
    protected Object _singleValueToJSONForEdition(Object value, DataContext context) throws BadItemTypeException
    {
        if (getManagedClass().isInstance(value))
        {
            return _singleTypedValueToJSONForEdition(getManagedClass().cast(value), context);
        }
        else
        {
            throw new BadItemTypeException("Try to convert the non " + getId() + " JSON object '" + value + "' into a " + getManagedClass().getName());
        }
    }
    
    /**
     * Convert the single value into a JSON object to use in edition mode
     * @param value the value to convert
     * @param context The context of the data to convert
     * @return The value as JSON
     */
    protected Object _singleTypedValueToJSONForEdition(T value, DataContext context)
    {
        return _singleTypedValueToJSON(value, context);
    }
    
    public T parseConfiguration(Configuration configuration) throws ConfigurationException
    {
        Element element = ConfigurationUtil.toElement(configuration);
        try
        {
            return _singleValueFromXML(element, Optional.empty());
        }
        catch (Exception e)
        {
            throw new ConfigurationException("An error occurs while parsin configuration of an I/O data", e);
        }
    }
    
    public Object valueFromXML(Element parent, String name, Optional<Object> additionalData) 
    {
        if (parent != null)
        {
            List<Element> children = DOMUtils.getChildElementsByTagName(parent, name);
            List<T> values = new ArrayList<>();
            boolean isMultiple = false;

            for (Element element : children)
            {
                try
                {
                    T singleValue = _singleValueFromXML(element, additionalData);
                    if (singleValue != null)
                    {
                        values.add(singleValue);
                    }
                    
                    isMultiple = _isValueFromXMLMultiple(element);
                }
                catch (TransformerException e)
                {
                    throw new IllegalArgumentException("Unable to parse DOM element to get the value", e);
                }
            }

            if (values.isEmpty())
            {
                return null;
            }
            else if (isMultiple)
            {
                Object[] array = (Object[]) Array.newInstance(getManagedClass(), values.size());
                return values.toArray(array);
            }
            else
            {
                return values.get(0);
            }
        }
        else
        {
            return null;
        }
    }
    
    /**
     * Checks if the value in the given node is marked as multiple
     * @param element the DOM element
     * @return <code>true</code> if the value is marked as multiple, <code>false</code> otherwise
     * @throws TransformerException if an error occurs while parsing the DOM node
     */
    protected boolean _isValueFromXMLMultiple(Element element) throws TransformerException
    {
        return Boolean.valueOf(element.getAttribute("multiple"));
    }
    
    /**
     * Retrieves a single typed value from the given DOM node
     * @param element the DOM element containing the single value
     * @param additionalData additional data needed to get the typed value
     * @return the value
     */
    protected T _singleValueFromXML(Element element, Optional<Object> additionalData)
    {
        // Check in "value" attribute for i18n values
        String value = Optional.ofNullable(element.getAttribute("value"))
                               .filter(StringUtils::isNotEmpty)
                               .orElseGet(element::getTextContent);
        
        return StringUtils.isNotEmpty(value) ? castValue(value) : null;
    }
    
    public void valueToSAXForEdition(ContentHandler contentHandler, String tagName, Object value, DataContext context) throws SAXException, BadItemTypeException
    {
        _valueToSAX(contentHandler, tagName, value, context, true);
    }
    
    public void valueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context) throws SAXException, BadItemTypeException
    {
        _valueToSAX(contentHandler, tagName, value, context, false);
    }
    
    /**
     * Generates SAX events for the given value
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param value the value to SAX
     * @param context The context of the data to SAX
     * @param isEdition <code>true</code> if SAX events are generated in edition mode, <code>false</code> otherwise
     * @throws SAXException if an error occurs during the SAX events generation
     * @throws BadItemTypeException if the given value is not compatible with the current type
     */
    protected void _valueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context, boolean isEdition) throws SAXException, BadItemTypeException
    {
        AttributesImpl attributes = _getValueAttributes(value, context, isEdition);
        if (value == null)
        {
            XMLUtils.createElement(contentHandler, tagName, attributes);
        }
        else if (isEdition)
        {
            _valueToSAXForEdition(contentHandler, tagName, value, context, attributes);
        }
        else
        {
            _valueToSAX(contentHandler, tagName, value, context, attributes);
        }
    }

    /**
     * Generates SAX events for the given value in edition mode.
     * This default implementation generates SAX event using the {@link #valueToJSONForClient(Object, DataContext)} method
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param value the value to SAX
     * @param context The context of the data to SAX
     * @param attributes the attributes for the SAX events to generate
     * @throws SAXException if an error occurs during the SAX events generation
     * @throws BadItemTypeException if the given value is not compatible with the current type
     */
    protected void _valueToSAXForEdition(ContentHandler contentHandler, String tagName, Object value, DataContext context, AttributesImpl attributes) throws SAXException, BadItemTypeException
    {
        if (_useJSONForEdition())
        {
            if (_jsonUtils != null)
            {
                Object valueAsJSON = valueToJSONForClient(value, context);
                String jsonString = _jsonUtils.convertObjectToJson(valueAsJSON);
                XMLUtils.createElement(contentHandler, tagName, attributes, jsonString);
            }
            else
            {
                throw new IllegalStateException("Unable to generate SAX events in tag name '" + tagName + "'. The component '" + JSONUtils.ROLE + "' has not been initialized");
            }
        }
        else
        {
            _valueToSAX(contentHandler, tagName, value, context, attributes);
        }
    }
    
    /**
     * Generates SAX events for the given value
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param value the value to SAX
     * @param context The context of the data to SAX
     * @param attributes the attributes for the SAX events to generate
     * @throws SAXException if an error occurs during the SAX events generation
     * @throws BadItemTypeException if the given value is not compatible with the current type
     */
    protected void _valueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context, AttributesImpl attributes) throws SAXException, BadItemTypeException
    {
        if (value.getClass().isArray())
        {
            if (Array.getLength(value) <= 0)
            {
                XMLUtils.createElement(contentHandler, tagName, attributes);
            }
            else
            {
                for (Object singleValue : (Object[]) value)
                {
                    if (singleValue != null)
                    {
                        _singleValueToSAX(contentHandler, tagName, singleValue, context, attributes);
                    }
                }
            }
        }
        else
        {
            _singleValueToSAX(contentHandler, tagName, value, context, attributes);
        }
    }
    
    /**
     * Generates SAX events for the given single value
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param value the single value to SAX
     * @param attributes the attributes for the SAX event to generate
     * @param context The context of the data to SAX
     * @throws SAXException if an error occurs during the SAX events generation
     * @throws BadItemTypeException if the given value is not compatible with the current type
     */
    protected void _singleValueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context, AttributesImpl attributes) throws SAXException, BadItemTypeException
    {
        AttributesImpl localAttributes = new AttributesImpl(attributes);
        
        Optional<Enumerator> optionalEnumerator = context.getModelItem()
                                                         .filter(ElementDefinition.class::isInstance)
                                                         .map(ElementDefinition.class::cast)
                                                         .map(ElementDefinition::getEnumerator);
        
        if (optionalEnumerator.isPresent())
        {
            Enumerator enumerator = optionalEnumerator.get();
            I18nizableText i18n = null;
            
            try
            {
                i18n = enumerator.getEntry(value);
            }
            catch (Exception e)
            {
                getLogger().warn("SAX events generated at tag '" + tagName + "' requires a label for enumerated value", e);
            }
            
            if (i18n != null)
            {
                _singleEnumeratedValueToSAX(contentHandler, tagName, value, i18n, context, localAttributes);
            }
            else
            {
                _singleNotEnumeratedValueToSAX(contentHandler, tagName, value, context, localAttributes);
            }
        }
        else
        {
            _singleNotEnumeratedValueToSAX(contentHandler, tagName, value, context, localAttributes);
        }
    }
    
    /**
     * Generates SAX events for the given single not enumerated value
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param value the single value to SAX
     * @param attributes the attributes for the SAX event to generate
     * @param context The context of the data to SAX
     * @throws SAXException if an error occurs during the SAX events generation
     * @throws BadItemTypeException if the given value is not compatible with the current type
     */
    protected void _singleNotEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, Object value, DataContext context, AttributesImpl attributes) throws SAXException, BadItemTypeException
    {
        if (getManagedClass().isInstance(value))
        {
            _singleTypedNotEnumeratedValueToSAX(contentHandler, tagName, getManagedClass().cast(value), context, attributes);
        }
        else
        {
            throw new BadItemTypeException("Try to sax the non compatible value '" + value + "' in tag name '" + tagName + "' with the type '" + getId() + "'");
        }
    }
    
    /**
     * Generates SAX events for the given single typed not enumerated value
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param value the single value to SAX
     * @param attributes the attributes for the SAX event to generate
     * @param context The context of the data to SAX
     * @throws SAXException if an error occurs during the SAX events generation
     */
    protected void _singleTypedNotEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, T value, DataContext context, AttributesImpl attributes) throws SAXException
    {
        XMLUtils.createElement(contentHandler, tagName, attributes, toString(value));
    }
    
    /**
     * Generates SAX events for the given single enumerated value
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param value the single value to SAX
     * @param i18n label of the enumerated value
     * @param attributes the attributes for the SAX event to generate
     * @param context The context of the data to SAX
     * @throws SAXException if an error occurs during the SAX events generation
     * @throws BadItemTypeException if the given value is not compatible with the current type
     */
    protected void _singleEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, Object value, I18nizableText i18n, DataContext context, AttributesImpl attributes) throws SAXException, BadItemTypeException
    {
        if (getManagedClass().isInstance(value))
        {
            _singleTypedEnumeratedValueToSAX(contentHandler, tagName, getManagedClass().cast(value), i18n, context, attributes);
        }
        else
        {
            throw new BadItemTypeException("Try to sax the non compatible value '" + value + "' in tag name '" + tagName + "' with the type '" + getId() + "'");
        }
    }
    
    /**
     * Generates SAX events for the given single typed enumerated value
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param tagName the tag name of the SAX event to generate.
     * @param value the single value to SAX
     * @param i18n label of the enumerated value
     * @param attributes the attributes for the SAX event to generate
     * @param context The context of the data to SAX
     * @throws SAXException if an error occurs during the SAX events generation
     */
    protected void _singleTypedEnumeratedValueToSAX(ContentHandler contentHandler, String tagName, T value, I18nizableText i18n, DataContext context, AttributesImpl attributes) throws SAXException
    {
        AttributesImpl localAttributes = new AttributesImpl(attributes);
        localAttributes.addCDATAAttribute("value", toString(value));
        
        XMLUtils.startElement(contentHandler, tagName, localAttributes);
        i18n.toSAX(contentHandler);
        XMLUtils.endElement(contentHandler, tagName);
    }
    
    /**
     * Retrieves the attributes of the single value to SAX
     * @param value the single value
     * @param context The context of the data to SAX
     * @param isEdition <code>true</code> if SAX events are generated in edition mode, <code>false</code> otherwise
     * @return the attributes
     */
    protected AttributesImpl _getValueAttributes(Object value, DataContext context, boolean isEdition)
    {
        AttributesImpl attributes = ModelItemTypeHelper.getXMLAttributesFromDataContext(context);
        attributes.addCDATAAttribute("typeId", getId());
        
        boolean isMultiple = value != null && value.getClass().isArray();
        attributes.addCDATAAttribute("multiple", String.valueOf(isMultiple));

        if (isEdition && _useJSONForEdition())
        {
            attributes.addCDATAAttribute("json", "true");
        }
        
        return attributes;
    }
   
    /**
     * Determine if the JSON format has to be used in edition mode for this element type 
     * @return <code>true</code> if the JSON format has to be used, <code>false</code> otherwise
     */
    protected boolean _useJSONForEdition()
    {
        return false;
    }
    
    public Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> compareValues(Object value1, Object value2)
    {
        if (_isMultiple(value1) && _isMultiple(value2))
        {
            // Both values are multiples
            T[] value1Array = _castMultipleValue(value1);
            T[] value2Array = _castMultipleValue(value2);
            return _compareMultipleValues(value1Array, value2Array);
        }
        else if (_isSingle(value1) && _isSingle(value2))
        {
            // Both values are single (or null)
            T typedValue1 = castValue(value1);
            T typedValue2 = castValue(value2);
            return _compareSingleValues(typedValue1, typedValue2);
        }
        else
        {
            // One value is single and the other one is multiple
            return Stream.of(new ImmutableTriple<>(DataChangeType.MODIFIED, DataChangeTypeDetail.NONE, StringUtils.EMPTY));
        }
    }
    
    /**
     * Checks if the given value is considered as multiple for this element type
     * @param value the value to check
     * @return <code>true</code> if the value is multiple, <code>false</code> otherwise
     */
    protected boolean _isMultiple(Object value)
    {
        return getManagedClassArray().isInstance(value);
    }
    
    /**
     * Checks if the given value is considered as single for this element type
     * @param value the value to check
     * @return <code>true</code> if the value is single, <code>false</code> otherwise
     */
    protected boolean _isSingle(Object value)
    {
        return value == null || getManagedClass().isInstance(value);
    }
    
    private T[] _castMultipleValue(Object value) throws BadItemTypeException
    {
        List<T> valuesAsList = new ArrayList<>();
        
        for (int i = 0; i < Array.getLength(value); i++)
        {
            T singleValue = castValue(Array.get(value, i));
            valuesAsList.add(singleValue);
        }
        
        return valuesAsList.toArray((T[]) Array.newInstance(getManagedClass(), valuesAsList.size()));
    }
    
    /**
     * Compares the given multiple values and retrieves the changes as a stream of {@link Triple}s. The {@link Triple}s contains:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>some details about this change if possible (after or before for a date, more or less for a number, ...) as a {@link DataChangeTypeDetail}</li>
     * <li>The data concerned by this change if not the element itself (or an empty String)</li>
     * </ul>
     * @param value1 the 1st multiple value
     * @param value2 the 2nd multiple value
     * @return the changes between the two given multiple values as a stream of {@link Triple}. Retrieves an empty stream if there is no change
     */
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareMultipleValues(T[] value1, T[] value2)
    {
        if (!Arrays.deepEquals(value1, value2))
        {
            DataChangeType type = DataChangeType.MODIFIED;
            DataChangeTypeDetail detail = DataChangeTypeDetail.NONE;
            if (value1.length == value2.length)
            {
                T[] diff = _removeAllOnce(value1, value2);
                if (diff.length == 0)
                {
                    detail = DataChangeTypeDetail.ORDER;
                }
            }
            else
            {
                if (_containsAll(value2, value1))
                {
                    detail = DataChangeTypeDetail.MORE;
                }
                else if (_containsAll(value1, value2))
                {
                    detail = DataChangeTypeDetail.LESS;
                }
            }
            
            return Stream.of(new ImmutableTriple<>(type, detail, StringUtils.EMPTY));
        }
        else
        {
            return Stream.empty();
        }
    }
    
    /**
     * Compares the given single values and retrieves the changes as a stream of {@link Triple}s. The {@link Triple} contains:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>some details about this change if possible (after or before for a date, more or less for a number, ...) as a {@link DataChangeTypeDetail}</li>
     * <li>The data concerned by this change if not the element itself (or an empty String)</li>
     * </ul>
     * @param value1 the 1st single value
     * @param value2 the 2nd single value
     * @return the changes between the two given single values as a stream of {@link Triple}. Retrieves an empty stream if there is no change
     */
    protected Stream<Triple<DataChangeType, DataChangeTypeDetail, String>> _compareSingleValues(T value1, T value2)
    {
        return Stream.of(ModelItemTypeHelper.compareSingleObjects(value1, value2, StringUtils.EMPTY))
                     .filter(Optional::isPresent)
                     .map(Optional::get);
    }
    
    /**
     * Remove once all elements from list2 in list1
     * [A, A, B, C], [A, B] will result [A, C]
     * @param list1 origin list
     * @param list2 list of objects to remove
     * @return a new list with removed objects
     */
    private T[] _removeAllOnce(T[] list1, T[] list2)
    {
        T[] result = list1.clone();
        for (T element : list2)
        {
            result = ArrayUtils.removeElement(result, element);
        }
        return result;
    }
    
    /**
     * Check if list1 contains all elements of list2
     * [A, B, C], [A, A, B, C] will return false
     * [A, A, B, C], [A, A, A, B, C] will return false
     * [A, A, B, C], [A, B, C] will return true
     * [A, B, C, D], [A, B, C] will return true
     * @param list1 complete list
     * @param list2 elements to check
     * @return true if all elements in list2 are once in list1
     */
    private boolean _containsAll(T[] list1, T[] list2)
    {
        T[] listClone = list1.clone();
        for (T element : list2)
        {
            if (!ArrayUtils.contains(listClone, element))
            {
                return false;
            }
            else
            {
                listClone = ArrayUtils.removeElement(listClone, element);
            }
        }
        return true;
    }
    
    public boolean isCompatible(Object value)
    {
        return value != null ? getManagedClass().isInstance(value) || getManagedClassArray().isInstance(value) : false;
    }
    
    public T castValue(Object value) throws BadItemTypeException
    {
        if (value == null)
        {
            return null;
        }
        else if (getManagedClass().isInstance(value))
        {
            return getManagedClass().cast(value);
        }
        
        return convertValue(value);
    }
    
    /**
     * Try to convert the given value, that is not of the managed type
     * @param value the value to convert
     * @return the converted value or null if it's not possible.
     * @throws BadItemTypeException if the given value is not compatible with the current type
     */
    protected T convertValue(Object value) throws BadItemTypeException
    {
        throw new BadItemTypeException("Unable to convert the given value '" + value + "'. This value is not compatible with the type '" + getId() + "'");
    }
    
    public Class<T> getManagedClass()
    {
        return _type;
    }
    
    public Class<T[]> getManagedClassArray()
    {
        return _typeArray;
    }
}
