/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.model.type;

import java.time.Instant;
import java.time.LocalDate;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Optional;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.ImmutableTriple;
import org.apache.commons.lang3.tuple.Triple;

import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.compare.DataChangeType;
import org.ametys.runtime.model.compare.DataChangeTypeDetail;
import org.ametys.runtime.model.type.DataContext;

/**
 * Helper model item types
 */
public final class ModelItemTypeHelper
{
    private ModelItemTypeHelper()
    {
        // Empty constructor
    }
    
    /**
     * Compare the given single strings and retrieves the change as an {@link Optional} of {@link Triple}. The {@link Triple} contains:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>detail to precise if the changes between the two given strings,</li>
     * <li>the data name</li>
     * </ul>
     * @param string1 the 1st single string
     * @param string2 the 2nd single string
     * @param dataName the name of the data
     * @return the change between the two given single strings
     */
    public static Optional<Triple<DataChangeType, DataChangeTypeDetail, String>> compareSingleStrings(String string1, String string2, String dataName)
    {
        if (StringUtils.isNotEmpty(string1) && StringUtils.isNotEmpty(string2))
        {
            if (!string1.equals(string2))
            {
                DataChangeTypeDetail detail = DataChangeTypeDetail.NONE;
                
                if (string1.startsWith(string2))
                {
                    detail = DataChangeTypeDetail.LESS_CONTENT_END;
                }
                else if (string2.startsWith(string1))
                {
                    detail = DataChangeTypeDetail.MORE_CONTENT_END;
                }
                else if (string1.endsWith(string2))
                {
                    detail = DataChangeTypeDetail.LESS_CONTENT_START;
                }
                else if (string2.endsWith(string1))
                {
                    detail = DataChangeTypeDetail.MORE_CONTENT_START;
                }
                else if (string1.contains(string2))
                {
                    detail = DataChangeTypeDetail.LESS_CONTENT;
                }
                else if (string2.contains(string1))
                {
                    detail = DataChangeTypeDetail.MORE_CONTENT;
                }
                
                return Optional.of(new ImmutableTriple<>(DataChangeType.MODIFIED, detail, dataName));
            }
            else
            {
                return Optional.empty();
            }
        }
        else if (StringUtils.isNotEmpty(string1) && StringUtils.isEmpty(string2))
        {
            return Optional.of(new ImmutableTriple<>(DataChangeType.REMOVED, DataChangeTypeDetail.NONE, dataName));
        }
        else if (StringUtils.isEmpty(string1) && StringUtils.isNotEmpty(string2))
        {
            return Optional.of(new ImmutableTriple<>(DataChangeType.ADDED, DataChangeTypeDetail.NONE, dataName));
        }
        else
        {
            // Both values are null
            return Optional.empty();
        }
    }
    
    /**
     * Compare the given single objects and retrieves the change as an {@link Optional} of {@link Triple}. The {@link Triple} contains:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>detail about this change if possible (after or before for a date, more or less for a number, ...) as a {@link DataChangeTypeDetail},</li>
     * <li>the data name</li>
     * </ul>
     * @param object1 the 1st single object
     * @param object2 the 2nd single object
     * @param dataName the name of the data
     * @return the change between the two given single objects
     */
    public static Optional<Triple<DataChangeType, DataChangeTypeDetail, String>> compareSingleObjects(Object object1, Object object2, String dataName)
    {
        if (object1 != null && object2 == null)
        {
            return Optional.of(new ImmutableTriple<>(DataChangeType.REMOVED, DataChangeTypeDetail.NONE, dataName));
        }
        else if (object1 == null && object2 != null)
        {
            return Optional.of(new ImmutableTriple<>(DataChangeType.ADDED, DataChangeTypeDetail.NONE, dataName));
        }
        else if (object1 == null || object1.equals(object2))
        {
            // values are equals or both are null
            return Optional.empty();
        }
        else
        {
            // Both values are not null and different
            return Optional.of(new ImmutableTriple<>(DataChangeType.MODIFIED, DataChangeTypeDetail.NONE, dataName));
        }
    }
    
    /**
     * Compare the given single numbers and retrieves the change as an {@link Optional} of {@link Triple}. The {@link Triple} contains:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>detail to precise if the second number is lower or greater than the first one,</li>
     * <li>the data name</li>
     * </ul>
     * @param number1 the 1st single number
     * @param number2 the 2nd single number
     * @param dataName the name of the data
     * @return the change between the two given single numbers
     */
    public static Optional<Triple<DataChangeType, DataChangeTypeDetail, String>> compareSingleNumbers(Number number1, Number number2, String dataName)
    {
        if (areSingleObjectsBothNotNullAndDifferents(number1, number2))
        {
            DataChangeTypeDetail detail = DataChangeTypeDetail.LESS;
            if (number2.doubleValue() > number1.doubleValue())
            {
                detail =  DataChangeTypeDetail.MORE;
            }
            
            return Optional.of(new ImmutableTriple<>(DataChangeType.MODIFIED, detail, dataName));
        }
        else
        {
            return compareSingleObjects(number1, number2, dataName);
        }
    }
    
    /**
     * Compare the given single {@link LocalDate} and retrieves the change as an {@link Optional} of {@link Triple}. The {@link Triple} contains:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>detail to precise if the second date is before or after the first one,</li>
     * <li>the data name</li>
     * </ul>
     * @param date1 the 1st single date
     * @param date2 the 2nd single date
     * @param dataName the name of the data
     * @return the change between the two given single dates
     */
    public static Optional<Triple<DataChangeType, DataChangeTypeDetail, String>> compareSingleDates(LocalDate date1, LocalDate date2, String dataName)
    {
        ZonedDateTime zdt1 = date1 != null ? date1.atStartOfDay(ZoneId.systemDefault()) : null;
        ZonedDateTime zdt2 = date2 != null ? date2.atStartOfDay(ZoneId.systemDefault()) : null;
        return compareSingleDates(zdt1, zdt2, dataName);
    }
    
    /**
     * Compare the given single {@link ZonedDateTime} and retrieves the change as an {@link Optional} of {@link Triple}. The {@link Triple} contains:
     * <ul>
     * <li>the general type of the change (added, modified or removed) as a {@link DataChangeType},</li>
     * <li>detail to precise if the second date is before or after the first one,</li>
     * <li>the data name</li>
     * </ul>
     * @param date1 the 1st single date
     * @param date2 the 2nd single date
     * @param dataName the name of the data
     * @return the change between the two given single dates
     */
    public static Optional<Triple<DataChangeType, DataChangeTypeDetail, String>> compareSingleDates(ZonedDateTime date1, ZonedDateTime date2, String dataName)
    {
        // We only want to compare instants of dates, not their zones
        Instant instant1 = date1 != null ? date1.toInstant() : null;
        Instant instant2 = date2 != null ? date2.toInstant() : null;
        
        if (areSingleObjectsBothNotNullAndDifferents(instant1, instant2))
        {
            DataChangeTypeDetail detail = DataChangeTypeDetail.AFTER;
            if (instant2 != null && instant2.isBefore(instant1)) // instant2 can not be null, it has already been checked by the previous test
            {
                detail =  DataChangeTypeDetail.BEFORE;
            }
            
            return Optional.of(new ImmutableTriple<>(DataChangeType.MODIFIED, detail, dataName));
        }
        else
        {
            return compareSingleObjects(instant1, instant2, dataName);
        }
    }
    
    /**
     * Checks if the given are both not <code>null</code> and if they are different
     * @param object1 the first object
     * @param object2 the second object
     * @return <code>true</code> if both object are not <code>null</code> and if they are different, <code>false</code> otherwise
     */
    public static boolean areSingleObjectsBothNotNullAndDifferents(Object object1, Object object2)
    {
        return object1 != null && object2 != null && !object1.equals(object2);
    }
    
    /**
     * Retrieves the XML attributes from the given context
     * @param context The context
     * @return the XML attributes
     */
    public static AttributesImpl getXMLAttributesFromDataContext(DataContext context)
    {
        return getXMLAttributesFromDataContext(context, true);
    }
    
    /**
     * Retrieves the XML attributes from the given context
     * @param context The context
     * @param renderModelInformation set to <code>true</code> to render model information, <code>false</code> otherwise
     * @return the XML attributes
     */
    public static AttributesImpl getXMLAttributesFromDataContext(DataContext context, boolean renderModelInformation)
    {
        AttributesImpl attributes = new AttributesImpl();
        
        // Get the data path including the parent context
        String dataPath = context.getDataPath();
        attributes.addCDATAAttribute("dataPath", dataPath);
        
        String fullDataPath = context.getFullDataPath();
        attributes.addCDATAAttribute("fullDataPath", fullDataPath);
        
        // Transform the data path to model path
        if (renderModelInformation)
        {
            attributes.addCDATAAttribute("modelPath", ModelHelper.getDefinitionPathFromDataPath(dataPath));
            attributes.addCDATAAttribute("fullModelPath", ModelHelper.getDefinitionPathFromDataPath(fullDataPath));
        }
        
        return attributes;
    }
}
