/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.resources;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.net.URISyntaxException;
import java.nio.charset.StandardCharsets;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.excalibur.source.Source;

import org.ametys.core.DevMode;
import org.ametys.core.DevMode.DEVMODE;
import org.ametys.plugins.core.ui.resources.css.CSSFileHelper;
import org.ametys.plugins.core.ui.resources.css.JSASSResourceURIExtensionPoint;

/**
 * Resource handler for CSS files
 */
public class CssResourceHandler extends ExpiresResourceHandler
{
    private ProxiedContextPathProvider _proxiedContextPathProvider;
    private JSASSResourceURIExtensionPoint _jsassResourceURIExtensionPoint;
    
    /**
     * Default constructor
     */
    public CssResourceHandler()
    {
        super();
    }
    
    /**
     * Constructor with an already resolved {@link Source}.
     * @param source the source
     */
    public CssResourceHandler(Source source)
    {
        super(source);
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        
        _proxiedContextPathProvider = (ProxiedContextPathProvider) manager.lookup(ProxiedContextPathProvider.ROLE);
        _jsassResourceURIExtensionPoint = (JSASSResourceURIExtensionPoint) manager.lookup(JSASSResourceURIExtensionPoint.ROLE);
    }
    
    @Override
    public Serializable getKey()
    {
        // files are cached by source and by context path
        return super.getKey() + "*" + _proxiedContextPathProvider.getContextPath();
    }
    
    @Override
    public void generate(OutputStream out) throws IOException, ProcessingException
    {
        Request request = ObjectModelHelper.getRequest(_objectModel);
        if (DEVMODE.SUPER_DEVELOPPMENT.equals(DevMode.getDeveloperMode(request)))
        {
            // Files are not aggregated, thus don't need to be modified
            super.generate(out);
            return;
        }
        
        // Replace relative URIs with absolute values, required before aggregating files served through a new location
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        super.generate(baos);
        String content = baos.toString(StandardCharsets.UTF_8.toString());
        
        try
        {
            String locationPath = _jsassResourceURIExtensionPoint.resolve(_requestedLocation);
            content = CSSFileHelper.replaceRelativeUri(content, locationPath, _jsassResourceURIExtensionPoint, _proxiedContextPathProvider.getContextPath(), _proxiedContextPathProvider.getContextPath());
        }
        catch (URISyntaxException e)
        {
            throw new ProcessingException("Unable to replace relative URIs inside of css file '" + _requestedLocation + "'", e);
        }
        
        out.write(content.getBytes(StandardCharsets.UTF_8));
    }
}
