/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.resources;

import java.io.IOException;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.cocoon.ProcessingException;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceValidity;

/**
 * Interface used to handle resources
 */
public interface ResourceHandler
{
    /**
     * Initialize the resource handler with a resource.
     * @param source The source uri
     * @param objectModel The object model
     * @param par The parameters
     * @param readForDownload if the resource is to be downloaded and not rendered.
     * @return the resolved source
     * @throws IOException If an error occurs
     * @throws ProcessingException If an error occurs
     */
    public Source setup(String source, Map objectModel, Parameters par, boolean readForDownload) throws IOException, ProcessingException;
    
    /**
     * Returns true if this {@link ResourceHandler} handles "Range" requests. false by default.
     * @return true if this {@link ResourceHandler} handles "Range" requests.
     */
    public default boolean acceptRanges()
    {
        return false;
    }
    
    /**
     * Generate the resource configured during setup, and output it
     * @param out The output stream to write to
     * @throws IOException If an error occurs
     * @throws ProcessingException If an error occurs
     */
    public void generate(OutputStream out) throws IOException, ProcessingException;
    
    /**
     * Generate the resource configured during setup in the context of a range request.
     * @param out The output stream to write to
     * @param offset the first byte to send
     * @param length the stream length to send
     * @throws IOException If an error occurs
     * @throws ProcessingException If an error occurs
     */
    public default void generate(OutputStream out, long offset, long length) throws IOException, ProcessingException
    {
        throw new UnsupportedOperationException("Ranges requests are not supported");
    }
    
    /**
     * Get the unique key for this resource, for cache purpose.
     * @return The cache key.
     */
    public Serializable getKey();
    
    /**
     * Get the resource validity, for cache purpose.
     * @return The resource validity.
     */
    public SourceValidity getValidity();
    
    /**
     * Get the resource size, if available. -1 if unknown.
     * @return The resource size.
     */
    public default long getLength()
    {
        return -1;
    }
    
    /**
     * Get the resource last modified time
     * @return The last modified
     */
    public long getLastModified();
    
    /**
     * Return the mime type of the configured resource.
     * @return The mime type.
     */
    public String getMimeType();

    /**
     * Returns true if the response Content-Length header should be set according to the source's length 
     * (ie. the source is not transformed during processing).
     * @return true if the source's length can be relied on
     */
    public default boolean shouldUseSourceContentLength()
    {
        return false;
    }
}
