/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.resources;

import java.io.IOException;
import java.io.Serializable;
import java.util.Map;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.Constants;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.caching.CacheableProcessingComponent;
import org.apache.cocoon.components.LifecycleHelper;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.Response;
import org.apache.cocoon.environment.http.HttpResponse;
import org.apache.cocoon.reading.AbstractReader;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.SourceValidity;
import org.xml.sax.SAXException;

import org.ametys.core.cocoon.source.NamedSource;
import org.ametys.core.util.URIUtils;

/**
 * Default resource reader, that handle different resources type using the ResourcesExtensionPoint.
 */
public class ResourceReader extends AbstractReader implements CacheableProcessingComponent, Serviceable, Contextualizable
{
    /** last modified parameter name for resources parameters */
    public static final String LAST_MODIFIED = "lastModified";
    
    private SourceResolver _resolver;
    private org.apache.cocoon.environment.Context _cocoonContext;

    private ResourceHandlerProviderExtensionPoint _resourcesHandlerEP;
    private ResourceHandler _resourceHandler;

    private Source _source;
    private boolean _readForDownload;
    
    private boolean _processRange;
    private long _rangeStart;
    private long _rangeEnd;
    
    public void contextualize(Context context) throws ContextException
    {
        _cocoonContext = (org.apache.cocoon.environment.Context) context.get(Constants.CONTEXT_ENVIRONMENT_CONTEXT);
    }
    
    public void service(ServiceManager sManager) throws ServiceException
    {
        _resourcesHandlerEP = (ResourceHandlerProviderExtensionPoint) sManager.lookup(ResourceHandlerProviderExtensionPoint.ROLE);
        _resolver = (SourceResolver) sManager.lookup(SourceResolver.ROLE);
    }
    
    @Override
    public void setup(org.apache.cocoon.environment.SourceResolver res, Map objModel, String src, Parameters setupParameters) throws ProcessingException, SAXException, IOException
    {
        super.setup(res, objModel, src, setupParameters);
        
        try
        {
            _resourceHandler = _resourcesHandlerEP.getResourceHandler(src);
        }
        catch (Exception e)
        {
            throw new ProcessingException("Exception while retrieving resource handler for resource '" + src + "'", e);
        }
        
        _readForDownload = setupParameters.getParameterAsBoolean("download", false);
        _source = _resourceHandler.setup(src, objModel, setupParameters, _readForDownload);
        
        // Minimizer does not receive the real lastModified through sitemap source 
        @SuppressWarnings("unchecked")
        Map<String, Object> params = (Map<String, Object>) objModel.get(ObjectModelHelper.PARENT_CONTEXT);
        if (params != null)
        {
            params.put(LAST_MODIFIED, getLastModified());
        }
        
        Response response = ObjectModelHelper.getResponse(objModel);

        if (_resourceHandler.acceptRanges())
        {
            response.setHeader("Accept-Ranges", "bytes");
            
            Request request = ObjectModelHelper.getRequest(objModel);
            String range = request.getHeader("Range");
            if (range != null)
            {
                try
                {
                    _parseRange(range.trim());
                    _processRange = !(_rangeStart == 0 && _rangeEnd == Long.MAX_VALUE); // only process range if it's not the whole resource
                }
                catch (IllegalArgumentException e)
                {
                    ((HttpResponse) response).setStatus(416); // Range not satisfiable
                    getLogger().error("Illegal range request: " + e.getMessage());
                }
            }
        }
        else
        {
            response.setHeader("Accept-Ranges", "none");
        }
        
        if (_readForDownload)
        {
            String name = _source instanceof NamedSource ? URIUtils.encodeHeader(((NamedSource) _source).getName()) : null;
            response.setHeader("Content-Disposition", "attachment" + (name != null ? ";filename=\"" + name + "\";filename*=UTF-8''" + name : ""));
        }
    }

    @Override
    public void generate() throws IOException, ProcessingException
    {
        Response response = ObjectModelHelper.getResponse(objectModel);
        
        long contentLength = _resourceHandler.getLength();
        
        if (!_processRange || contentLength == -1)
        {
            if (contentLength != -1) 
            {
                response.setHeader("Content-Length", Long.toString(contentLength));
            }
            
            _resourceHandler.generate(out);
        }
        else
        {
            ((HttpResponse) response).setStatus(206); // partial content
            
            long actualEnd = _rangeEnd == Long.MAX_VALUE ? contentLength - 1 : _rangeEnd;
            long length = actualEnd - _rangeStart + 1;
            
            response.setHeader("Content-Range", "bytes " + _rangeStart + "-" + actualEnd + "/" + contentLength);
            response.setHeader("Content-Length", Long.toString(length));
            
            _resourceHandler.generate(out, _rangeStart, length);
        }
        
        out.flush();
    }

    @Override
    public Serializable getKey()
    {
        if (_processRange)
        {
            return null;
        }
        
        return _resourceHandler.getKey();
    }
    
    @Override
    public SourceValidity getValidity()
    {
        return _resourceHandler.getValidity();
    }

    @Override
    public void recycle()
    {
        super.recycle();
        
        _resolver.release(_source);
        LifecycleHelper.dispose(_resourceHandler);
        
        _processRange = false;
        _rangeStart = 0;
        _rangeEnd = 0;
    }
    
    @Override
    public String getMimeType()
    {
        String sourceMimeType = _source.getMimeType();
        
        if (sourceMimeType != null)
        {
            return sourceMimeType;
        }
        
        if (_cocoonContext != null) 
        {
            final String mimeType = _cocoonContext.getMimeType(_source.getURI());
            
            if (mimeType != null) 
            {
                return mimeType;
            }
        }
        
        return null;
    }
    
    @Override
    public long getLastModified()
    {
        return _resourceHandler.getLastModified();
    }
    
    private void _parseRange(String requestedRange)
    {
        if (!requestedRange.startsWith("bytes="))
        {
            throw new IllegalArgumentException("Ranges are only accepted for bytes: " + requestedRange);
        }
        
        String range = requestedRange.substring("bytes=".length());

        if (range.contains(","))
        {
            throw new IllegalArgumentException("Only single-values ranges are allowed: " + range);
        }
        
        int i = range.indexOf('-');
        if (i < 0)
        {
            throw new IllegalArgumentException("Wrong format for range: " + range);
        }
        
        if (i == 0)
        {
            throw new IllegalArgumentException("Suffix ranges not supported: " + range);
        }
        
        _rangeStart = Long.parseLong(range.substring(0, i));
        
        int length = range.length();
        if (i < length - 1) 
        {
            _rangeEnd = Integer.parseInt(range.substring(i + 1, length));
        } 
        else 
        {
            _rangeEnd = Long.MAX_VALUE;
        }
        
        if (_rangeStart > _rangeEnd) 
        {
            throw new IllegalArgumentException("Start value is greater than end value in range: " + range);
        }
    }
}
