/*
 *  Copyright 2021 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.resources;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.ResourceNotFoundException;
import org.apache.commons.io.IOUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.SourceValidity;

import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Simple common ancestor for all {@link ResourceHandler} implementations, delegating all methods to the underlying {@link Source}.
 */
public class SimpleResourceHandler extends AbstractLogEnabled implements ResourceHandler, Serviceable
{
    /** The source resolver */
    protected SourceResolver _resolver;
    
    /** The resolved source */
    protected Source _source;
    
    /** The sitemap parameters */
    protected Parameters _parameters;
    
    /** The object model */
    protected Map _objectModel;
    
    /** if the resource has been request for download */
    protected boolean _readForDownload;
    
    /** The initially requested resource location */
    protected String _requestedLocation;
    
    /**
     * Default constructor.
     */
    public SimpleResourceHandler()
    {
        // empty constructor
    }
    
    /**
     * If the {@link Source} is already resolved by the {@link ResourceHandlerProvider}, 
     * it may provide it through the constructor to avoid resolving it again.
     * @param source the source.
     */
    public SimpleResourceHandler(Source source)
    {
        _source = source;
    }

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    @Override
    public Source setup(String location, Map objectModel, Parameters parameters, boolean readForDownload) throws IOException, ProcessingException
    {
        _requestedLocation = location;
        _objectModel = objectModel;
        _parameters = parameters;
        _readForDownload = readForDownload;
        
        // If the source has already been given through constructor, no need to resolve again
        if (_source == null)
        {
            _source = _resolver.resolveURI(location);

            if (!_source.exists())
            {
                throw new ResourceNotFoundException("Resource not found for URI : " + _source.getURI());
            }
        }
        
        return _source;
    }
    
    @Override
    public void generate(OutputStream out) throws IOException, ProcessingException
    {
        try (InputStream is = _source.getInputStream())
        {
            IOUtils.copy(is, out);
        }
    }
    
    @Override
    public Serializable getKey()
    {
        return _source.getURI();
    }

    @Override
    public SourceValidity getValidity()
    {
        return _source.getValidity();
    }

    @Override
    public long getLastModified()
    {
        return _source.getLastModified();
    }
    
    public String getMimeType()
    {
        return _source.getMimeType();
    }
}
