/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.right;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.tuple.Pair;

import org.ametys.core.group.GroupIdentity;
import org.ametys.core.right.AccessController.AccessResult;
import org.ametys.core.right.ProfileAssignmentStorage.AnonymousOrAnyConnectedKeys;
import org.ametys.core.right.ProfileAssignmentStorage.UserOrGroup;
import org.ametys.core.ui.right.ProfileAssignmentsToolClientSideElement.TargetType;
import org.ametys.core.user.UserIdentity;
import org.ametys.runtime.plugin.ExtensionPoint;
import org.ametys.runtime.plugin.component.AbstractThreadSafeComponentPrioritizableExtensionPoint;

/**
 * {@link ExtensionPoint} handling {@link ProfileAssignmentStorage}s.
 */
public class ProfileAssignmentStorageExtensionPoint extends AbstractThreadSafeComponentPrioritizableExtensionPoint<ProfileAssignmentStorage>
{
    /** Avalon Role */
    public static final String ROLE = ProfileAssignmentStorageExtensionPoint.class.getName();
    
    /* ---------- */
    /* PUBLIC API */
    /* ---------- */
    
    /**
     * Gets the permissions a user has, given some groups and profiles, on an object.
     * @param user The user
     * @param userGroups The groups
     * @param profileIds The ids of the profiles
     * @param object The object
     * @return the permissions a user has, given some groups and profiles on an object.
     */
    public Map<String, AccessResult> getPermissions(UserIdentity user, Set<GroupIdentity> userGroups, Set<String> profileIds, Object object)
    {
        getLogger().debug("Try to determine permissions for user '{}' and groups {} on context {} for profiles [{}]", user, userGroups, object, profileIds);
        
        Map<String, AccessResult> results = new HashMap<>();
        
        // 1) First initialize the access results with the allowed profiles for Anonymous
        Map<AnonymousOrAnyConnectedKeys, Set<String>> profilesForAnonymousAndAnyConnectedUser = getProfilesForAnonymousAndAnyConnectedUser(object);
        
        Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANONYMOUS_ALLOWED)).orElse(Set.of())
            .stream()
            .filter(profileIds::contains)
            .forEach(p -> results.putIfAbsent(p, AccessResult.ANONYMOUS_ALLOWED));
        
        if (results.size() == profileIds.size())
        {
            // Stop process here if access were determined for all profiles
            _logResult(user, userGroups, object, results);
            return results;
        }
        
        // 2) Add to access results the denied profiles for user (among the still undetermined profiles)
        Map<UserIdentity, Map<UserOrGroup, Set<String>>> profilesForUsers = getProfilesForUsers(object, user);
        Optional.ofNullable(profilesForUsers.get(user)).map(a -> a.get(UserOrGroup.DENIED)).orElse(Set.of())
            .stream()
            .filter(profileIds::contains)
            .forEach(p -> results.putIfAbsent(p, AccessResult.USER_DENIED));

        if (results.size() == profileIds.size())
        {
            // Stop process here if access were determined for all profiles
            _logResult(user, userGroups, object, results);
            return results;
        }
        
        // 3) Add to access results the allowed profiles for user (among the still undetermined profiles)
        Optional.ofNullable(profilesForUsers.get(user)).map(a -> a.get(UserOrGroup.ALLOWED)).orElse(Set.of())
            .stream()
            .filter(profileIds::contains)
            .forEach(p -> results.putIfAbsent(p, AccessResult.USER_ALLOWED));

        if (results.size() == profileIds.size())
        {
            // Stop process here if access were determined for all profiles
            _logResult(user, userGroups, object, results);
            return results;
        }
        
        // 4) Add to access results the denied profiles for group (among the still undetermined profiles)
        Map<GroupIdentity, Map<UserOrGroup, Set<String>>> profilesForGroups = getProfilesForGroups(object, userGroups);
        for (GroupIdentity userGroup : userGroups)
        {
            Optional.ofNullable(profilesForGroups.get(userGroup)).map(a -> a.get(UserOrGroup.DENIED)).orElse(Set.of())
                .stream()
                .filter(profileIds::contains)
                .forEach(p -> results.putIfAbsent(p, AccessResult.GROUP_DENIED));
        }

        if (results.size() == profileIds.size())
        {
            // Stop process here if access were determined for all profiles
            _logResult(user, userGroups, object, results);
            return results;
        }
        
        // 5) Add to access results the allowed profiles for group (among the still undetermined profiles)
        for (GroupIdentity userGroup : userGroups)
        {
            Optional.ofNullable(profilesForGroups.get(userGroup)).map(a -> a.get(UserOrGroup.ALLOWED)).orElse(Set.of())
                .stream()
                .filter(profileIds::contains)
                .forEach(p -> results.putIfAbsent(p, AccessResult.GROUP_ALLOWED));
        }
        
        if (results.size() == profileIds.size())
        {
            // Stop process here if access were determined for all profiles
            _logResult(user, userGroups, object, results);
            return results;
        }
        
        // 6) Add to access results the denied profiles for any connected user (among the still undetermined profiles)
        Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_DENIED)).orElse(Set.of())
            .stream()
            .filter(profileIds::contains)
            .forEach(p -> results.putIfAbsent(p, AccessResult.ANY_CONNECTED_DENIED));

        if (results.size() == profileIds.size())
        {
            // Stop process here if access were determined for all profiles
            _logResult(user, userGroups, object, results);
            return results;
        }
        
        // 7) Add to access results the allowed profiles for any connected user (among the still undetermined profiles)
        Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_ALLOWED)).orElse(Set.of())
            .stream()
            .filter(profileIds::contains)
            .forEach(p -> results.putIfAbsent(p, AccessResult.ANY_CONNECTED_ALLOWED));

        if (results.size() == profileIds.size())
        {
            // Stop process here if access were determined for all profiles
            _logResult(user, userGroups, object, results);
            return results;
        }
        
        // 8) Add to access results the denied profiles for any connected user (among the still undetermined profiles)
        Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANONYMOUS_DENIED)).orElse(Set.of())
            .stream()
            .filter(profileIds::contains)
            .forEach(p -> results.putIfAbsent(p, AccessResult.ANONYMOUS_DENIED));

        if (results.size() == profileIds.size())
        {
            // Stop process here if access were determined for all profiles
            _logResult(user, userGroups, object, results);
            return results;
        }
        
        // 9) Finally, add to access results the profiles for which no access could be determined
        for (String profileId : profileIds)
        {
            results.putIfAbsent(profileId, AccessResult.UNKNOWN);
        }
        
        _logResult(user, userGroups, object, results);
        return results;
    }
    
    /**
     * Get all permissions for a given user, including permission given by anonymous, any connected user, user's groups and user itself.
     * I.e assignment targeting anonymous, any connected user, one of the provided group, or the user themselves
     * @param user the user
     * @param userGroups the groups of the user
     * @return a map associating each context with stored permission to the profile and permission assigned
     */
    public Map<Object, Map<AccessResultInfo, Set<String>>> getAllPermissions(UserIdentity user, Set<GroupIdentity> userGroups)
    {
        getLogger().debug("Try to determine all stored permissions for user '{}' and its groups {}", user, userGroups);
        
        List<ProfileAssignmentStorage> sortedPas = getExtensionsIds().stream()
                .map(this::getExtension)
                .collect(Collectors.toList());
        
        Map<Object, Map<AccessResultInfo, Set<String>>> userPermissions = new HashMap<>();
        
        for (ProfileAssignmentStorage profileAssignmentStorage : sortedPas)
        {
            Map<Object, Map<AccessResultInfo, Set<String>>> pasStoredPermissions = _getUserPermission(profileAssignmentStorage, user, userGroups);
            // Only stored new context. Context already present were handled by a storage with higher priority
            // The existing result is therefore the valid one
            pasStoredPermissions.keySet().stream()
                .forEach(context -> userPermissions.computeIfAbsent(context, k -> pasStoredPermissions.get(k)));
        }
        getLogger().debug("Found {} context with stored permissions for user '{}' and groups {} with profiles {}", userPermissions.keySet().size(), user, userGroups);
        return userPermissions;
    }
    
    /**
     * Get all the user permission recorded by this storage
     * @param profileAssignmentStorage the storage
     * @param user the user
     * @param userGroups the groups of the user
     * @return permissions on all supported context
     */
    private Map<Object, Map<AccessResultInfo, Set<String>>> _getUserPermission(ProfileAssignmentStorage profileAssignmentStorage, UserIdentity user, Set<GroupIdentity> userGroups)
    {
        Map<Object, Map<AccessResultInfo, Set<String>>> result = new HashMap<>();
        
        // 1) Search profiles in "allowed-anonymous-profiles" and "denied-anonymous-profiles"
        Map<Object, Map<AnonymousOrAnyConnectedKeys, Set<String>>> anonymousProfiles = profileAssignmentStorage.getAllProfilesForAnonymousAndAnyConnectedUser();
        
        // 2) Search profiles in "allowed-profiles" and "denied-profiles" for user
        Map<Object, Map<UserOrGroup, Set<String>>> userProfiles = profileAssignmentStorage.getAllProfilesForUser(user);
        
        // 3) Search at least one profile in "allowed-profiles" for groups
        Map<Object, Map<GroupIdentity, Map<UserOrGroup, Set<String>>>> groupsProfiles = profileAssignmentStorage.getAllProfilesForGroups(userGroups);
        
        // Now, iterate on all the retrieved context to merge the result
        // TODO Est-ce qu'il y a pas mieux à faire (perf ? lisibilité ?)
        Stream.of(anonymousProfiles.keySet().stream(), userProfiles.keySet().stream(), groupsProfiles.keySet().stream())
            .flatMap(i -> i) // Concatenate the 3 streams on contexts
            .distinct() // remove duplicate context having assignment for different "levels"
            .forEach(context -> {
                Map<AccessResultInfo, Set<String>> contextResult = new HashMap<>();
                Map<AnonymousOrAnyConnectedKeys, Set<String>> contextAnonymousProfiles = anonymousProfiles.get(context);
                if (contextAnonymousProfiles != null)
                {
                    for (Entry<AnonymousOrAnyConnectedKeys, Set<String>> profiles : contextAnonymousProfiles.entrySet())
                    {
                        AccessResult accessResult = profiles.getKey().toAccessResult();
                        AccessResultInfo resultInfo = new AccessResultInfo(accessResult, null);
                        contextResult.put(resultInfo, profiles.getValue());
                    }
                }
                
                Map<UserOrGroup, Set<String>> contextUserProfiles = userProfiles.get(context);
                if (contextUserProfiles != null)
                {
                    for (Entry<UserOrGroup, Set<String>> profiles : contextUserProfiles.entrySet())
                    {
                        AccessResult accessResult = profiles.getKey().toAccessResult(TargetType.USER);
                        AccessResultInfo resultInfo = new AccessResultInfo(accessResult, user);
                        contextResult.put(resultInfo, profiles.getValue());
                    }
                }
                
                Map<GroupIdentity, Map<UserOrGroup, Set<String>>> contextGroupsProfiles = groupsProfiles.get(context);
                if (contextGroupsProfiles != null)
                {
                    for (Entry<GroupIdentity, Map<UserOrGroup, Set<String>>> groupProfiles : contextGroupsProfiles.entrySet())
                    {
                        for (Entry<UserOrGroup, Set<String>> profiles : groupProfiles.getValue().entrySet())
                        {
                            AccessResult accessResult = profiles.getKey().toAccessResult(TargetType.GROUP);
                            AccessResultInfo resultInfo = new AccessResultInfo(accessResult, groupProfiles.getKey());
                            contextResult.put(resultInfo, profiles.getValue());
                        }
                    }
                }
                
                result.put(context, contextResult);
            });
        
        return result;
    }
    
    private void _logResult(UserIdentity user, Set<GroupIdentity> userGroups, Object object, Map<String, AccessResult> accessResultsByProfile)
    {
        getLogger().debug("Access result found for {} and groups {} on context '{}' are {}", user, userGroups, object, accessResultsByProfile);
    }    
    
    /**
     * Returns some profiles that are matching if the user has a permission on at least one object, given some groups and profiles
     * @param rootContexts The root contexts object where to seek
     * @param user The user
     * @param userGroups The groups
     * @param profileIds The ids of the profiles
     * @return If the Set is empty, it means any connected user has no matching profile.<br>
     *         If the Set is non empty, it contains at least one of the given profile BUT it may not contains all the matching profiles for anyconnected user AND it can contains some other profiles that were not in the given profiles
     */
    public Set<String> hasUserAnyPermission(Set<? extends Object> rootContexts, UserIdentity user, Set<GroupIdentity> userGroups, Set<String> profileIds)
    {
        getLogger().debug("Try to determine permissions on any context for user '{}' and groups {} with profiles {}", user, userGroups, profileIds);
        
        List<ProfileAssignmentStorage> sortedPas = getExtensionsIds().stream()
                .map(this::getExtension)
                .collect(Collectors.toList());
        
        Set<? extends Object> remainingRootsToTest = new HashSet<>(rootContexts);
        for (ProfileAssignmentStorage profileAssignmentStorage : sortedPas)
        {
            if (!remainingRootsToTest.isEmpty())
            {
                Set<? extends Object> filteredContexts = remainingRootsToTest.stream().filter(profileAssignmentStorage::isRootContextSupported).collect(Collectors.toSet());
                if (!filteredContexts.isEmpty())
                {
                    Set<String> hasUserAnyPermission = _hasUserAnyPermission(profileAssignmentStorage, filteredContexts, user, userGroups, profileIds);
                    if (!hasUserAnyPermission.isEmpty())
                    {
                        getLogger().debug("Find permission on any context for user '{}' and groups {} with profiles {}", user, userGroups, profileIds);
                        return hasUserAnyPermission;
                    }
                }
                // Remove the already supported contexts
                remainingRootsToTest.removeAll(filteredContexts);
            }
        }
        
        getLogger().debug("Find no permission on any context for user '{}' and groups {} with profiles {}", user, userGroups, profileIds);
        return Set.of();
    }
    
    
    
    private Set<String> _hasUserAnyPermission(ProfileAssignmentStorage profileAssignmentStorage, Set<? extends Object> rootContexts, UserIdentity user, Set<GroupIdentity> userGroups, Set<String> profileIds)
    {
        if (profileIds.isEmpty() || rootContexts.isEmpty())
        {
            return Set.of();
        }
        
        
        // 1) Search at least one profile in "allowed-anonymous-profiles"
        Set<String> hasAnonymousAnyAllowedProfile = profileAssignmentStorage.hasAnonymousAnyAllowedProfile(rootContexts, profileIds);
        if (!hasAnonymousAnyAllowedProfile.isEmpty())
        {
            return hasAnonymousAnyAllowedProfile;
        }
        
        // 2) Search at least one profile in "allowed-profiles" for user
        Set<String> hasUserAnyAllowedProfile = profileAssignmentStorage.hasUserAnyAllowedProfile(rootContexts, user, profileIds);
        if (!hasUserAnyAllowedProfile.isEmpty())
        {
            return hasUserAnyAllowedProfile;
        }
        
        // 3) Search at least one profile in "allowed-profiles" for groups
        Set<String> hasGroupAnyAllowedProfile = profileAssignmentStorage.hasGroupAnyAllowedProfile(rootContexts, userGroups, profileIds);
        if (!hasGroupAnyAllowedProfile.isEmpty())
        {
            return hasGroupAnyAllowedProfile;
        }
            
        // 4) Search at least one profile in "allowed-any-connected-profiles"
        Set<String> hasAnyConnectedAnyAllowedProfile = profileAssignmentStorage.hasAnyConnectedAnyAllowedProfile(rootContexts, profileIds);
        if (!hasAnyConnectedAnyAllowedProfile.isEmpty())
        {
            return hasAnyConnectedAnyAllowedProfile;
        }
        
        // 5) Not found, return nothing
        return Set.of();
    }

    /**
     * Returns some profiles that are matching if anybody has a permission on at least one object, given some profiles
     * @param rootContexts The root contexts object where to seek
     * @param profileIds The ids of the profiles
     * @return If the Set is empty, it means anonymous has no matching profile.<br>
     *         If the Set is non empty, it contains at least one of the given profile BUT it may not contains all the matching profiles for anonymous AND it can contains some other profiles that were not in the given profiles
     */
    public Set<String> hasAnonymousAnyPermission(Set<? extends Object> rootContexts, Set<String> profileIds)
    {
        getLogger().debug("Try to determine permissions on any context for anonymous with profiles {}", profileIds);
        
        List<ProfileAssignmentStorage> sortedPas = getExtensionsIds().stream()
                .map(this::getExtension)
                .collect(Collectors.toList());
        
        Set<? extends Object> remainingRootsToTest = new HashSet<>(rootContexts);
        for (ProfileAssignmentStorage profileAssignmentStorage : sortedPas)
        {
            if (!remainingRootsToTest.isEmpty())
            {
                Set<? extends Object> filteredContexts = remainingRootsToTest.stream().filter(profileAssignmentStorage::isRootContextSupported).collect(Collectors.toSet());
                if (!filteredContexts.isEmpty())
                {
                    Set<String> hasAnonymousAnyPermission = _hasAnonymousAnyPermission(profileAssignmentStorage, filteredContexts, profileIds);
                    if (!hasAnonymousAnyPermission.isEmpty())
                    {
                        getLogger().debug("Find permission on any context for anonymous with profiles {}", profileIds);
                        return hasAnonymousAnyPermission;
                    }
                }
                // Remove the already supported contexts
                remainingRootsToTest.removeAll(filteredContexts);
            }
        }
        
        getLogger().debug("Find no permission on any context for anonymous with profiles {}", profileIds);
        return Set.of();
    }
    
    private Set<String> _hasAnonymousAnyPermission(ProfileAssignmentStorage profileAssignmentStorage, Set< ? extends Object> rootContexts, Set<String> profileIds)
    {
        if (profileIds.isEmpty() || rootContexts.isEmpty())
        {
            return Set.of();
        }
        
        // Search at least one profile in "allowed-anonymous-profiles", if found return true
        return profileAssignmentStorage.hasAnonymousAnyAllowedProfile(rootContexts, profileIds);
    }

    /**
     * Returns some profiles that are matching if any connected user has a permission on at least one object, given some profiles
     * @param rootContexts The root contexts object where to seek
     * @param profileIds The ids of the profiles
     * @return If the Set is empty, it means the user has no matching profile.<br>
     *         If the Set is non empty, it contains at least one of the given profile BUT it may not contains all the matching profiles for the user AND it can contains some other profiles that were not in the given profiles
     */
    public Set<String> hasAnyConnectedUserAnyPermission(Set<? extends Object> rootContexts, Set<String> profileIds)
    {
        getLogger().debug("Try to determine permissions on any context for any connected user with profiles {}", profileIds);
        
        List<ProfileAssignmentStorage> sortedPas = getExtensionsIds().stream()
                .map(this::getExtension)
                .collect(Collectors.toList());
        
        Set<? extends Object> remainingRootsToTest = new HashSet<>(rootContexts);
        for (ProfileAssignmentStorage profileAssignmentStorage : sortedPas)
        {
            if (!remainingRootsToTest.isEmpty())
            {
                Set<? extends Object> filteredContexts = remainingRootsToTest.stream().filter(profileAssignmentStorage::isRootContextSupported).collect(Collectors.toSet());
                if (!filteredContexts.isEmpty())
                {
                    Set<String> hasAnyConnectedUserAnyPermission = _hasAnyConnectedUserAnyPermission(profileAssignmentStorage, filteredContexts, profileIds);
                    if (!hasAnyConnectedUserAnyPermission.isEmpty())
                    {
                        getLogger().debug("Find permission on any context for any connected user with profiles {}", profileIds);
                        return hasAnyConnectedUserAnyPermission;
                    }
                }
                // Remove the already supported contexts
                remainingRootsToTest.removeAll(filteredContexts);
            }
        }
        
        getLogger().debug("Find no permission on any context for any connected user with profiles {}", profileIds);
        return Set.of();
    }
    
    private Set<String> _hasAnyConnectedUserAnyPermission(ProfileAssignmentStorage profileAssignmentStorage, Set< ? extends Object> rootContexts, Set<String> profileIds)
    {
        if (profileIds.isEmpty() || rootContexts.isEmpty())
        {
            return Set.of();
        }
        
        
        // 1) Search at least one profile in "allowed-anonymous-profiles", if found return true
        Set<String> hasAnonymousAnyAllowedProfile = profileAssignmentStorage.hasAnonymousAnyAllowedProfile(rootContexts, profileIds);
        if (!hasAnonymousAnyAllowedProfile.isEmpty())
        {
            return hasAnonymousAnyAllowedProfile;
        }
            
        // 2) Search at least one profile in "allowed-any-connected-profiles", if found return true
        Set<String> hasAnyConnectedAnyAllowedProfile = profileAssignmentStorage.hasAnyConnectedAnyAllowedProfile(rootContexts, profileIds);
        if (!hasAnyConnectedAnyAllowedProfile.isEmpty())
        {
            return hasAnyConnectedAnyAllowedProfile;
        }
        
        // 3) Not found, return false
        return Set.of();
    }

    /**
     * Gets the permissions a user has on an object, for every profile in the application.
     * @param user The user
     * @param userGroups The groups
     * @param object The object
     * @return the permissions a user has on an object, for every profile in the application.
     */
    public Map<String, AccessResult> getPermissionsByProfile(UserIdentity user, Set<GroupIdentity> userGroups, Object object)
    {
        getLogger().debug("Try to determine permissions for each profile on context {} for user '{}' and groups {}", object, user, userGroups);
        
        Map<String, AccessResult> result = new HashMap<>();
        
        // Allowed profiles for anonymous
        Map<AnonymousOrAnyConnectedKeys, Set<String>> profilesForAnonymousAndAnyConnectedUser = getProfilesForAnonymousAndAnyConnectedUser(object);
        Set<String> anonymousAllowed = Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANONYMOUS_ALLOWED)).orElse(Set.of());
        Set<String> anonymousDenied = Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANONYMOUS_DENIED)).orElse(Set.of());
        ((Collection<String>) CollectionUtils.removeAll(anonymousAllowed, anonymousDenied))
            .forEach(profile -> result.putIfAbsent(profile, AccessResult.ANONYMOUS_ALLOWED));
        
        // Denied profiles for user
        Map<UserIdentity, Map<UserOrGroup, Set<String>>> profilesForUsers = getProfilesForUsers(object, user);
        Optional.ofNullable(profilesForUsers.get(user)).map(a -> a.get(UserOrGroup.DENIED)).orElse(Set.of())
            .forEach(profile -> result.putIfAbsent(profile, AccessResult.USER_DENIED));
        
        // Allowed profiles for user
        Optional.ofNullable(profilesForUsers.get(user)).map(a -> a.get(UserOrGroup.ALLOWED)).orElse(Set.of())
            .forEach(profile -> result.putIfAbsent(profile, AccessResult.USER_ALLOWED));
        
        // Denied profiles for groups
        Map<GroupIdentity, Map<UserOrGroup, Set<String>>> profilesForGroups = getProfilesForGroups(object, userGroups);
        for (GroupIdentity userGroup : userGroups)
        {
            Optional.ofNullable(profilesForGroups.get(userGroup)).map(a -> a.get(UserOrGroup.DENIED)).orElse(Set.of())
                .forEach(profile -> result.putIfAbsent(profile, AccessResult.GROUP_DENIED));
        }
        
        // Allowed profiles for groups
        for (GroupIdentity userGroup : userGroups)
        {
            Optional.ofNullable(profilesForGroups.get(userGroup)).map(a -> a.get(UserOrGroup.ALLOWED)).orElse(Set.of())
                .forEach(profile -> result.putIfAbsent(profile, AccessResult.GROUP_ALLOWED));
        }
        
        // Denied profiles for any connected user
        Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_DENIED)).orElse(Set.of())
            .forEach(profile -> result.putIfAbsent(profile, AccessResult.ANY_CONNECTED_DENIED));
        
        // Allowed profiles for any connected user
        Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_ALLOWED)).orElse(Set.of())
            .forEach(profile -> result.putIfAbsent(profile, AccessResult.ANY_CONNECTED_ALLOWED));
        
        // Denied profiles for anonymous
        anonymousDenied
            .forEach(profile -> result.putIfAbsent(profile, AccessResult.ANONYMOUS_DENIED));
        
        getLogger().debug("The permissions by profile on context {} for user '{}' and groups {} are : {}", object, user, userGroups, result);
        return result;
    }
    
    /**
     * Gets the permissions for Anonymous for the given profiles
     * @param profileIds The profiles to get permissions on
     * @param object The object
     * @return the access result for each profile
     */
    public AccessResult getPermissionForAnonymous (Set<String> profileIds, Object object)
    {
        getLogger().debug("Try to determine permission for Anonymous on context {} and profiles {}", object, profileIds);
     
        Map<AnonymousOrAnyConnectedKeys, Set<String>> profilesForAnonymousAndAnyConnectedUser = getProfilesForAnonymousAndAnyConnectedUser(object);
        
        Set<String> deniedProfiles = Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANONYMOUS_DENIED)).orElse(Set.of());
        Set<String> allowedProfiles = Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANONYMOUS_ALLOWED)).orElse(Set.of());
        
        AccessResult result = AccessResult.UNKNOWN;
        
        for (String profileId : profileIds)
        {
            if (deniedProfiles.contains(profileId))
            {
                return AccessResult.ANONYMOUS_DENIED;
            }
            else if (allowedProfiles.contains(profileId))
            {
                result = AccessResult.ANONYMOUS_ALLOWED;
            }
        }
        
        return result;
    }
    
    /**
     * Gets the permissions for Anonymous for the given profiles
     * @param profileIds The profiles to get permissions on
     * @param object The object
     * @return the access result for each profile
     */
    public AccessResult getPermissionForAnyConnectedUser (Set<String> profileIds, Object object)
    {
        getLogger().debug("Try to determine permission for AnyConnectedUser on context {} and profiles {}", object, profileIds);
     
        Map<AnonymousOrAnyConnectedKeys, Set<String>> profilesForAnonymousAndAnyConnectedUser = getProfilesForAnonymousAndAnyConnectedUser(object);
        
        Set<String> deniedProfiles = Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_DENIED)).orElse(Set.of());
        Set<String> allowedProfiles = Optional.ofNullable(profilesForAnonymousAndAnyConnectedUser.get(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_ALLOWED)).orElse(Set.of());
        
        AccessResult result = AccessResult.UNKNOWN;
        
        for (String profileId : profileIds)
        {
            if (deniedProfiles.contains(profileId))
            {
                return AccessResult.ANY_CONNECTED_DENIED;
            }
            else if (allowedProfiles.contains(profileId))
            {
                result = AccessResult.ANY_CONNECTED_ALLOWED;
            }
        }
        
        return result;
    }
    
    private AccessResult _getPermissionsByUser(Map<UserOrGroup, Set<String>> userProfiles, Set<String> profileIds)
    {
        if (userProfiles != null)
        {
            Set<String> deniedProfiles = userProfiles.get(UserOrGroup.DENIED);
            if (deniedProfiles != null)
            {
                if (deniedProfiles.stream().anyMatch(p -> profileIds.contains(p)))
                {
                    return AccessResult.USER_DENIED;
                }
            }
    
            Set<String> allowedProfiles = userProfiles.get(UserOrGroup.ALLOWED);
            if (allowedProfiles != null)
            {
                if (allowedProfiles.stream().anyMatch(p -> profileIds.contains(p)))
                {
                    return AccessResult.USER_ALLOWED;
                }
            }
        }

        return null;
    }
    
    /**
     * Gets the permission by user only on an object, according to the given profiles. It does not take account of the groups of the user, etc.
     * @param profileIds The ids of the profiles
     * @param object The object
     * @return the permission by user only on an object, according to the given profiles
     */
    public Map<UserIdentity, AccessResult> getPermissionsByUser(Set<String> profileIds, Object object)
    {
        getLogger().debug("Try to determine permissions by users on context {} and profiles {}", object, profileIds);
        
        Map<UserIdentity, AccessResult> result = getProfilesForUsers(object, null)
            .entrySet()
            .stream()
            .map(entry -> Pair.of(entry.getKey(), _getPermissionsByUser(entry.getValue(), profileIds)))
            .filter(pair -> pair.getRight() != null)
            .collect(Collectors.toMap(Pair::getLeft, Pair::getRight));
        
        getLogger().debug("The permissions by users on context {} and profiles {} are: {}", object, profileIds, result);
        return result;
    }
    
    private AccessResult _getPermissionsByGroup(Map<UserOrGroup, Set<String>> groupProfiles, Set<String> profileIds)
    {
        if (groupProfiles != null)
        {
            Set<String> deniedProfiles = groupProfiles.get(UserOrGroup.DENIED);
            if (deniedProfiles != null)
            {
                if (deniedProfiles.stream().anyMatch(p -> profileIds.contains(p)))
                {
                    return AccessResult.GROUP_DENIED;
                }
            }
    
            Set<String> allowedProfiles = groupProfiles.get(UserOrGroup.ALLOWED);
            if (allowedProfiles != null)
            {
                if (allowedProfiles.stream().anyMatch(p -> profileIds.contains(p)))
                {
                    return AccessResult.GROUP_ALLOWED;
                }
            }
        }

        return null;
    }
    
    /**
     * Gets the permission by group only on an object, according to the given profiles.
     * @param profileIds The ids of the profiles
     * @param object The object
     * @return the permission by group only on an object, according to the given profiles
     */
    public Map<GroupIdentity, AccessResult> getPermissionsByGroup(Set<String> profileIds, Object object)
    {
        getLogger().debug("Try to determine permissions by groups on context {} and profiles {}", object, profileIds);
        
        Map<GroupIdentity, AccessResult> result = getProfilesForGroups(object, null)
            .entrySet()
            .stream()
            .map(entry -> Pair.of(entry.getKey(), _getPermissionsByGroup(entry.getValue(), profileIds)))
            .filter(pair -> pair.getRight() != null)
            .collect(Collectors.toMap(Pair::getLeft, Pair::getRight));
        
        getLogger().debug("The permissions by groups on context {} and profiles {} are: ", object, profileIds, result);
        return result;
    }
    
    /* ------- */
    /* STORAGE */
    /* ------- */

    /**
     * Gets the allowed profiles any connected user has on the given object
     * @param context The object
     * @return a map containing allowed/denied profiles that anonymous and any connected user has on the given object
     */
    public Map<AnonymousOrAnyConnectedKeys, Set<String>> getProfilesForAnonymousAndAnyConnectedUser(Object context)
    {
        return _getFirstProfileAssignmentStorage(context)
                .map(pas -> pas.getProfilesForAnonymousAndAnyConnectedUser(context))
                .orElse(Map.of());
    }
    
    /**
     * Gets the users that have allowed profiles assigned on the given object
     * @param context The object to test 
     * @param user The user to get profiles for. Can be null to get profiles for all users that have rights
     * @return The map of allowed users with their assigned allowed/denied profiles
     */
    public Map<UserIdentity, Map<UserOrGroup, Set<String>>> getProfilesForUsers(Object context, UserIdentity user)
    {
        return _getFirstProfileAssignmentStorage(context)
                .map(pas -> pas.getProfilesForUsers(context, user))
                .orElse(Map.of());
    }
    
    /**
     * Gets the groups that have allowed profiles assigned on the given object
     * @param context The object to test 
     * @param groups The group to get profiles for. Can be null to get profiles for all groups that have rights
     * @return The map of allowed/denied groups with their assigned profiles
     */
    public Map<GroupIdentity, Map<UserOrGroup, Set<String>>> getProfilesForGroups(Object context, Set<GroupIdentity> groups)
    {
        return _getFirstProfileAssignmentStorage(context)
                .map(pas -> pas.getProfilesForGroups(context, groups))
                .orElse(Map.of());
    }
    
    /* ------------- */
    /* ANY CONNECTED */
    /* ------------- */
    
    /**
     * Adds allowed profile any connected user has on the given object
     * @param context The object context
     * @param profileId The profile to add
     */
    public void allowProfileToAnyConnectedUser(String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.addAllowedProfilesForAnyConnectedUser(context, Collections.singleton(profileId)));
    }
    
    /**
     * Adds denied profile any connected user has on the given object
     * @param profileId The profile to add
     * @param context The object context
     */
    public void denyProfileToAnyConnectedUser(String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.addDeniedProfilesForAnyConnectedUser(context, Collections.singleton(profileId)));
    }
    
    /**
     * Removes allowed profile any connected user has on the given object
     * @param profileId The profile to remove
     * @param context The object context
     */
    public void removeAllowedProfileFromAnyConnectedUser(String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.removeAllowedProfilesForAnyConnectedUser(context, Collections.singleton(profileId)));
    }
    
    /**
     * Removes denied profile any connected user has on the given object
     * @param context The object context
     * @param profileId The profile to remove
     */
    public void removeDeniedProfileFromAnyConnectedUser(String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.removeDeniedProfilesForAnyConnectedUser(context, Collections.singleton(profileId)));
    }
    
    /* --------- */
    /* ANONYMOUS */
    /* --------- */
    
    /**
     * Adds allowed profile an anonymous user has on the given object
     * @param profileId The profile to add
     * @param context The object context
     */
    public void allowProfileToAnonymous(String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.addAllowedProfilesForAnonymous(context, Collections.singleton(profileId)));
    }
    
    /**
     * Adds denied profile an anonymous user has on the given object
     * @param profileId The profile to add
     * @param context The object context
     */
    public void denyProfileToAnonymous(String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.addDeniedProfilesForAnonymous(context, Collections.singleton(profileId)));
    }
    
    /**
     * Removes allowed profile an anonymous user has on the given object
     * @param profileId The profile to remove
     * @param context The object context
     */
    public void removeAllowedProfileFromAnonymous(String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.removeAllowedProfilesForAnonymous(context, Collections.singleton(profileId)));
    }
    
    /**
     * Removes denied profile an anonymous user has on the given object
     * @param context The object context
     * @param profileId The profile to remove
     */
    public void removeDeniedProfileFromAnonymous(String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.removeDeniedProfilesForAnonymous(context, Collections.singleton(profileId)));
    }
    
    /* ----- */
    /* USERS */
    /* ----- */
    
    /**
     * Allows a user to a profile on a given object
     * @param user The user to add
     * @param profileId The id of the profile
     * @param context The object context
     */
    public void allowProfileToUser(UserIdentity user, String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.addAllowedUsers(Collections.singleton(user), context, profileId));
    }
    
    /**
     * Denies a user to a profile on a given object
     * @param user The user to add
     * @param profileId The id of the profile
     * @param context The object context
     */
    public void denyProfileToUser(UserIdentity user, String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.addDeniedUsers(Collections.singleton(user), context, profileId));
    }
    
    /**
     * Removes the association between a user and an allowed profile on a given object
     * @param user The user to remove
     * @param context The object context
     * @param profileId The id of the profile
     */
    public void removeAllowedProfileFromUser(UserIdentity user, String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.removeAllowedUsers(Collections.singleton(user), context, profileId));
    }
    
    /**
     * Removes the association between a user and a denied profile on a given object
     * @param user The user to remove
     * @param profileId The id of the profile
     * @param context The object context
     */
    public void removeDeniedProfileFromUser(UserIdentity user, String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.removeDeniedUsers(Collections.singleton(user), context, profileId));
    }
    
    
    /* ------ */
    /* GROUPS */
    /* ------ */
    
    /**
     * Allows a group to a profile on a given object
     * @param group The group to add
     * @param profileId The id of the profile
     * @param context The object context
     */
    public void allowProfileToGroup(GroupIdentity group, String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.addAllowedGroups(Collections.singleton(group), context, profileId));
    }
    
    /**
     * Denies a group to a profile on a given object
     * @param group The group to add
     * @param profileId The id of the profile
     * @param context The object context
     */
    public void denyProfileToGroup(GroupIdentity group, String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.addDeniedGroups(Collections.singleton(group), context, profileId));
    }
    
    /**
     * Removes the association between a group and an allowed profile on a given object
     * @param group The group to remove
     * @param profileId The id of the profile
     * @param context The object context
     */
    public void removeAllowedProfileFromGroup(GroupIdentity group, String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.removeAllowedGroups(Collections.singleton(group), context, profileId));
    }
    
    /**
     * Removes the association between a group and a denied profile on a given object
     * @param group The group to remove
     * @param profileId The id of the profile
     * @param context The object context
     */
    public void removeDeniedProfileFromGroup(GroupIdentity group, String profileId, Object context)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
                .ifPresent(pas -> pas.removeDeniedGroups(Collections.singleton(group), context, profileId));
    }
    
    /* ----------- */
    /* INHERITANCE */
    /* ----------- */
    /**
     * Determines if the inheritance of permissions is disallowed on a given context
     * @param context The object context
     * @return true if the inheritance is disallowed
     */
    public boolean isInheritanceDisallowed(Object context)
    {
        return _getFirstProfileAssignmentStorage(context)
            .map(pas -> pas.isInheritanceDisallowed(context))
            .orElse(Boolean.FALSE);
    }
    
    /**
     * Allow or disallow the inheritance of permissions on a given context
     * @param context The object context
     * @param disallow true to disallow the inheritance
     */
    public void disallowInheritance(Object context, boolean disallow)
    {
        _getFirstModifiableProfileAssignmentStorage(context)
            .ifPresent(pas -> pas.disallowInheritance(context, disallow));
    }
    
    /* -------------------------- */
    /* PRIVATE CONVENIENT METHODS */
    /* -------------------------- */
    
    private Optional<ProfileAssignmentStorage> _getFirstProfileAssignmentStorage(Object object)
    {
        return getExtensionsIds().stream()
                .map(this::getExtension)
                .filter(pas -> pas.isSupported(object))
                .findFirst();
    }
    
    private Optional<ModifiableProfileAssignmentStorage> _getFirstModifiableProfileAssignmentStorage(Object object)
    {
        return getExtensionsIds().stream()
                .map(this::getExtension)
                .filter(pas -> pas.isSupported(object) && pas instanceof ModifiableProfileAssignmentStorage)
                .map(ModifiableProfileAssignmentStorage.class::cast)
                .findFirst();
    }
    
    /**
     * Store an access result associated to a object target (user or group)
     * @param accessResult an access result
     * @param target the actual target of the assignment (user or group). Can be null for anonymous or any connected user assignment
     */
    public record AccessResultInfo(AccessResult accessResult, Object target) implements Comparable<AccessResultInfo> {
        // Keep the natural order provided by the access result
        public int compareTo(AccessResultInfo o)
        {
            return this.accessResult.compareTo(o.accessResult());
        }
    }
}
