/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.right;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ametys.core.ui.ClientSideElement;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * This interface represents a context of right assignment
 */
public interface RightAssignmentContext extends ClientSideElement
{
    /** Use this separator for path of context */
    public static final String CONTEXT_LABEL_PATH_SEPARATOR = " > ";
    
    /**
     * Determines if this rights assignment context is a private context
     * @return true if is private 
     */
    public boolean isPrivate();
    
    /**
     * Determines if this right assignment context should be visible in the user profiles tool
     * @return true if the right assignment should be visible
     */
    public boolean showInUserProfilesTool();
    
    /**
     * From a JavaScript context object, converts it into a Java object
     * @param jsContext the JS context object
     * @return the Java object
     */
    public Object convertJSContext(Object jsContext);
    
    /**
     * Gets the identifier of the object context
     * @param context The object context
     * @return the identifier of the object context
     */
    public String getContextIdentifier(Object context);
    
    /**
     * Get the parent contexts of the object context
     * @param context The object context
     * @return The parents if exist or <code>null</code> otherwise
     */
    public Set<Object> getParentContexts(Object context);
    
    /**
     * Get the list of root contexts if the current workspace matches the allowed workspace
     * @param contextParameters The contextual parameters
     * @return The root context object in a list
     */
    public List<Object> getRootContexts(Map<String, Object> contextParameters);
    
    /**
     * Determines if the right assignment context is available in the current application context
     * @param contextualParameters The contextual parameters 
     * @return true if the current workspace matches
     */
    public default boolean matchWorkspace(Map<String, Object> contextualParameters)
    {
        return true;
    }
    
    /**
     * Indicate for a given context, if this RightAssignmentContext is able to handle it.
     * This method should only be call if the result of {@link #matchWorkspace} for the given 
     * contextual parameters is true. Otherwise the behavior of the method is not defined.
     * Note that the result of this method should be in coherence with the list of item displayed
     * by in the assignment tool.
     * @param context a context
     * @param contextualParameters the application context
     * @return true if the context is supported
     */
    public boolean isSupportedContext(Object context, Map<String, Object> contextualParameters);
    
    /**
     * Get info about the context and its right assignment context.
     * 
     * Note : this method returns info about the context and the right assignment context to 
     * allow implementations to emulate different right assignment contexts based on the context
     * @param context a context object. The context object must be supported (see {@link #isSupportedContext}).
     * @return Info about the context and the right assignment context
     */
    public ContextInfo getContextInfo(Object context);
    
    /**
     * Record describing a context and its associated right assignment context
     * @param rightAssignmentContextId the right assignment context id
     * @param rightAssignmentContextLabel the right assignment context label
     * @param contextId the context id
     * @param contextLabel the context label
     * @param isRootContext if the context is a root context
     */
    public record ContextInfo(
            String rightAssignmentContextId,
            I18nizableText rightAssignmentContextLabel,
            String contextId,
            I18nizableText contextLabel,
            boolean isRootContext) {
        /**
         * Serialize the context info as a JSON
         * @return a JSON map.
         */
        public Map<String, Object> toJSON()
        {
            Map<String, Object> json = new HashMap<>();
            json.put("rightAssignmentContext", Map.of(
                "id", this.rightAssignmentContextId,
                "label", this.rightAssignmentContextLabel
                    ));
            
            json.put("context", Map.of(
                    "id", this.contextId,
                    "label", this.contextLabel,
                    "isRoot", this.isRootContext
                    ));
            
            return json;
        }
    }
}
