/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.right;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.apache.ibatis.session.SqlSession;

import org.ametys.core.ObservationConstants;
import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;
import org.ametys.core.datasource.AbstractMyBatisDAO;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Manages registration of profiles
 */
public class RightProfilesDAO extends AbstractMyBatisDAO implements Initializable
{
    /** The component role. */
    public static final String ROLE = RightProfilesDAO.class.getName();

    /** profile cache id */
    private static final String PROFILES_CACHE = ROLE + "$profiles";

    /** profile rights cache id */
    private static final String PROFILE_RIGHTS_CACHE = ROLE + "$profileRigths";
    
    private ServiceManager _smanager;

    private ObservationManager _observationManager;

    private CurrentUserProvider _currentUserProvider;

    private AbstractCacheManager _cacheManager;

    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _smanager = smanager;
        super.service(smanager);
        _cacheManager = (AbstractCacheManager) smanager.lookup(AbstractCacheManager.ROLE);
    }


    public void initialize() throws Exception
    {
        _cacheManager.createMemoryCache(PROFILES_CACHE, 
                new I18nizableText("plugin.core", "PLUGINS_CORE_RIGHT_PROFILES_CACHE_LABEL"),
                new I18nizableText("plugin.core", "PLUGINS_CORE_RIGHT_PROFILES_CACHE_DESCRIPTION"),
                true,
                null);
        _cacheManager.createMemoryCache(PROFILE_RIGHTS_CACHE, 
                new I18nizableText("plugin.core", "PLUGINS_CORE_RIGHT_PROFILE_RIGHTS_CACHE_LABEL"),
                new I18nizableText("plugin.core", "PLUGINS_CORE_RIGHT_PROFILE_RIGHTS_CACHE_DESCRIPTION"),
                true,
                null);
    }
    
    /**
     * Clear the profiles cache
     */
    public synchronized void clearCache()
    {
        _getProfilesCache().invalidateAll();
        _getProfileRightsCache().invalidateAll();
    }
    
    private synchronized Cache<String, Profile> _getProfilesCache()
    {
        Cache<String, Profile> cache = this._cacheManager.get(PROFILES_CACHE);

        if (!cache.isInitialized())
        {
            try (SqlSession session = getSession())
            {
                Map<String, Profile> profilesMap = new HashMap<>();
                List<Profile> profiles = session.selectList("Profiles.getProfiles");
                profilesMap = profiles.stream().collect(Collectors.toMap(Profile::getId, p -> p));
                cache.putAll(profilesMap);
            }
        }
        
        return cache;
    }

    private synchronized Cache<String, List<String>> _getProfileRightsCache()
    {

        Cache<String, List<String>> cache = this._cacheManager.get(PROFILE_RIGHTS_CACHE);

        //As the cache can be cleared from admin tool, we add an entry to check is filled or not.
        if (!cache.isInitialized())
        {
            try (SqlSession session = getSession())
            {

                Map<String, List<String>> profileRightsMap = new HashMap<>();
                List<Map<String, String>> rightAssociations = session.selectList("Profiles.getProfileRights");
                
                profileRightsMap = rightAssociations.stream()
                        .collect(Collectors.groupingBy(
                            rightAssociation -> rightAssociation.get("profileId"),
                            Collectors.mapping(
                                rightAssociation -> rightAssociation.get("rightId"), 
                                Collectors.toList())));
                cache.putAll(profileRightsMap);
            }
        }
        
        return cache;
    }
    
    /**
     * Get all existing profiles
     * @return The list for profiles
     */
    public List<Profile> getProfiles()
    {
        Collection<Profile> profiles = _getProfilesCache().asMap().values();
        return new ArrayList<>(profiles);
    }
    
    /**
     * Get the profiles on a given context
     * @param context The context. Can be null. If null, the profiles with no context are returned.
     * @return The list for profiles for this context
     */
    public List<Profile> getProfiles(String context)
    {
        return _getProfilesCache().asMap().values().stream()
                                      .filter(p -> StringUtils.equals(p.getContext(), context))
                                      .collect(Collectors.toList());
    }
    
    /**
     * Get the profile with given identifier
     * @param id The id of profile to retrieve
     * @return The profile
     */
    public Profile getProfile(String id)
    {
        return _getProfilesCache().get(id);
    }
    
    /**
     * Get all profiles containing the right with given id
     * @param rightId The id of right
     * @return The id of profiles with this right
     */
    public Set<String> getProfilesWithRight (String rightId)
    {
        return _getProfileRightsCache().asMap().entrySet().stream()
                .filter(e -> e.getValue().contains(rightId))
                .map(e -> e.getKey())
                .collect(Collectors.toSet());
    }
    
    /**
     * Creates a new profile with null context. The identifier of the profile will be automatically generated from label.
     * @param label The label of profile
     * @return The create profile
     */
    public Profile addProfile (String label)
    {
        return addProfile(label, null);
    }
    
    /**
     * Creates a new profile. The identifier of the profile will be automatically generated from label.
     * @param label The label of profile
     * @param context The context. Can be null
     * @return The create profile
     */
    public Profile addProfile (String label, String context)
    {
        String id = _generateUniqueId(label);
        Profile profile = new Profile(id, label, context);
        addProfile(profile);
        return profile;
    }
    
    private String _generateUniqueId(String label)
    {
        // Id generated from name lowercased, trimmed, and spaces and underscores replaced by dashes
        String value = label.toLowerCase().trim().replaceAll("[\\W_]", "-").replaceAll("-+", "-").replaceAll("^-", "");
        int i = 2;
        String suffixedValue = value;
        while (getProfile(suffixedValue) != null)
        {
            suffixedValue = value + i;
            i++;
        }
        
        return suffixedValue;
    }
    
    /**
     * Creates a new profile
     * @param id The unique identifier of profile
     * @param label The label of profile
     * @param context The context. Can be null
     * @return The create profile
     */
    public Profile addProfile (String id, String label, String context)
    {
        Profile profile = new Profile(id, label, context);
        addProfile(profile);
        return profile;
    }
    
    /**
     * Add a new profile
     * @param profile The profile to add
     * @param silent Set to true to not notify observer of this update
     */
    public void addProfile (Profile profile, boolean silent)
    {
        try (SqlSession session = getSession(true))
        {
            session.insert("Profiles.addProfile", profile);
            clearCache();
            if (!silent)
            {
                _notifyEvent(profile, ObservationConstants.EVENT_PROFILE_ADDED);
            }
        }
    }
    
    /**
     * Add a new profile
     * @param profile The profile to add
     */
    public void addProfile (Profile profile)
    {
        addProfile(profile, false);
    }
    
    /**
     * Rename a profile
     * @param profile The profile to rename
     * @param newLabel The updated label
     */
    public void renameProfile (Profile profile, String newLabel)
    {
        renameProfile(profile, newLabel, false);
    }
    
    /**
     * Rename a profile
     * @param profile The profile to rename
     * @param newLabel The updated label
     * @param silent Set to true to not notify observer of this update
     */
    public void renameProfile (Profile profile, String newLabel, boolean silent)
    {
        try (SqlSession session = getSession(true))
        {
            Map<String, Object> params = new HashMap<>();
            params.put("id", profile.getId());
            params.put("label", newLabel);
            session.update("Profiles.renameProfile", params);
            
            clearCache();
            
            if (!silent)
            {
                _notifyEvent(profile, ObservationConstants.EVENT_PROFILE_UPDATED);
            }
        }
    }

    
    /**
     * Get the rights of a profile
     * @param profileId The profile id
     * @return The rights
     */
    public List<String> getRights (String profileId)
    {
        if (!StringUtils.isEmpty(profileId))
        {
            List<String> list = _getProfileRightsCache().get(profileId);
            if (list != null)
            {
                return list;
            }
        }
        return Collections.EMPTY_LIST;
    }

    /**
     * Get the rights of a profile
     * @param profile The profile
     * @return The rights
     */
    public List<String> getRights (Profile profile)
    {
        if (profile == null)
        {
            return Collections.EMPTY_LIST;
        }
        else
        {
            return getRights(profile.getId());
        }
    }
    
    /**
     * Add a right to a profile
     * @param profile The profile
     * @param rightId The id of right to add
     */
    public void addRight (Profile profile, String rightId)
    {
        try (SqlSession session = getSession(true))
        {
            _addRight (session, profile, rightId);
        }
        
        clearCache();
    }
    
    /**
     * Add a right to a profile
     * @param profile The profile
     * @param rightIds The id of rights to add
     */
    public void addRights (Profile profile, List<String> rightIds)
    {
        try (SqlSession session = getSession())
        {
            for (String rightId : rightIds)
            {
                _addRight (session, profile, rightId);
            }
            
            session.commit();
        }
        
        clearCache();
    }
    
    /**
     * Update the rights of a profile
     * @param profile The profile
     * @param rights The rights of the profile
     */
    public void updateRights (Profile profile, List<String> rights)
    {
        updateRights(profile, rights, false);
    }
    
    /**
     * Update the rights of a profile
     * @param profile The profile
     * @param rights The rights of the profile
     * @param silent Set to true to not notify observer of this update
     */
    public void updateRights (Profile profile, List<String> rights, boolean silent)
    {
        try (SqlSession session = getSession())
        {
            session.delete("Profiles.deleteProfileRights", profile.getId());
            
            if (rights != null)
            {
                for (String rightId : rights)
                {
                    _addRight (session, profile, rightId);
                }
            }
            
            session.commit();
            
            clearCache();
            
            if (!silent)
            {
                _notifyEvent(profile, ObservationConstants.EVENT_PROFILE_UPDATED);
            }
        }
    }
    
    private void _addRight (SqlSession session, Profile profile, String rightId)
    {
        Map<String, Object> params = new HashMap<>();
        params.put("profileId", profile.getId());
        params.put("rightId", rightId);
        
        session.insert("Profiles.addRight", params);
    }
    
    /**
     * Add a right to a profile
     * @param profile The profile
     */
    public void removeRights (Profile profile)
    {
        removeRights(profile, false);
    }
    
    /**
     * Add a right to a profile
     * @param profile The profile
     * @param silent Set to true to not notify observer of this update
     */
    public void removeRights (Profile profile, boolean silent)
    {
        try (SqlSession session = getSession(true))
        {
            session.delete("Profiles.deleteProfileRights", profile.getId());
            
            clearCache();
            
            if (!silent)
            {
                _notifyEvent(profile, ObservationConstants.EVENT_PROFILE_UPDATED);
            }
        }
    }
    
    /**
     * Remove a right from a profile
     * @param profile The profile
     * @param rightId The id of right to add
     */
    public void removeRight (Profile profile, String rightId)
    {
        removeRight(profile, rightId, false);
    }

    /**
     * Remove a right from a profile
     * @param profile The profile
     * @param rightId The id of right to add
     * @param silent Set to true to not notify observer of this update
     */
    public void removeRight (Profile profile, String rightId, boolean silent)
    {
        try (SqlSession session = getSession(true))
        {
            Map<String, Object> params = new HashMap<>();
            params.put("profileId", profile.getId());
            params.put("rightId", rightId);
            session.delete("Profiles.deleteProfileRight", params);
            
            clearCache();
            
            if (!silent)
            {
                _notifyEvent(profile, ObservationConstants.EVENT_PROFILE_UPDATED);
            }
        }
        
        clearCache();
    }
    
    /**
     * Delete a profile
     * @param profile The profile to delete
     */
    public void deleteProfile (Profile profile)
    {
        deleteProfile(profile, false);
    }
    
    /**
     * Delete a profile
     * @param profile The profile to delete
     * @param silent Set to true to not notify observer of this update
     */
    public void deleteProfile (Profile profile, boolean silent)
    {
        try (SqlSession session = getSession())
        {
            session.delete("Profiles.deleteProfile", profile.getId());
            session.delete("Profiles.deleteProfileRights", profile.getId());
            
            session.commit();
            
            clearCache();
            
            if (!silent)
            {
                _notifyEvent(profile, ObservationConstants.EVENT_PROFILE_DELETED);
            }
        }
    }
    
    private void _notifyEvent (Profile profile, String eventId)
    {
        try
        {
            if (_observationManager == null)
            {
                _observationManager = (ObservationManager) _smanager.lookup(ObservationManager.ROLE);
            }
            if (_currentUserProvider == null)
            {
                _currentUserProvider = (CurrentUserProvider) _smanager.lookup(CurrentUserProvider.ROLE);
            }
            
            Map<String, Object> eventParams = new HashMap<>();
            eventParams.put(ObservationConstants.ARGS_PROFILE, profile);
            _observationManager.notify(new Event(eventId, _currentUserProvider.getUser(), eventParams));
        }
        catch (ServiceException e)
        {
            getLogger().error("Fail to notify observers for event '" + eventId + "'", e);
        }
    }
}
