/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.right;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;

import org.ametys.core.group.GroupIdentity;
import org.ametys.core.group.GroupManager;
import org.ametys.core.right.AccessController.AccessResult;
import org.ametys.core.right.ProfileAssignmentStorageExtensionPoint.AccessResultInfo;
import org.ametys.core.right.RightAssignmentContext.ContextInfo;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.plugins.core.group.GroupHelper;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * Client side element of the tool displaying all the profile assignments stored for a given user
 */
public class UserProfilesToolClientSideElement extends StaticClientSideElement implements Contextualizable
{

    private GroupManager _groupManager;
    private RightAssignmentContextExtensionPoint _rightAssignmentContextEP;
    private ProfileAssignmentStorageExtensionPoint _storageEP;
    private Context _context;
    private GroupHelper _groupHelper;
    private UserHelper _userHelper;
    private UserManager _userManager;

    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        super.service(manager);
        _groupHelper = (GroupHelper) manager.lookup(GroupHelper.ROLE);
        _groupManager = (GroupManager) manager.lookup(GroupManager.ROLE);
        _rightAssignmentContextEP = (RightAssignmentContextExtensionPoint) manager.lookup(RightAssignmentContextExtensionPoint.ROLE);
        _storageEP = (ProfileAssignmentStorageExtensionPoint) manager.lookup(ProfileAssignmentStorageExtensionPoint.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    /**
     * Provided a user identity, this method will retrieve all stored assignments concerning this user
     * ie assignment for anonymous, any connected user, user's groups, and the user itself
     * @param login the user login
     * @param populationId the user population
     * @param jsParameters context parameters provided by the client
     * @return Result is a JSON list like:
     * [
     *      {
     *          rightAssignmentContext: {"id": &lt;assignmentContextId&gt;, "label": &lt;assignmentContextLabel&gt;},
     *          context: {"label": &lt;contextLabel&gt;},
     *          profile1: {
     *              accessResult: accessResult,
     *              assignments: [accessResultInfo,…]
     *          }
     *      }
     * ]
     */
    @Callable(rights = "Runtime_Rights_SeeUserProfiles")
    public List<Map<String, Object>> getUserPermissions(String login, String populationId, Map<String, Object> jsParameters)
    {
        List<Map<String, Object>> result = new ArrayList<>();
        
        UserIdentity user = new UserIdentity(login, populationId);
        Map<String, Object> contextualParameters = getContextualParameters(jsParameters);
        
        Map<Object, Map<AccessResultInfo, Set<String>>> storedAssignments = _storageEP.getAllPermissions(user, _groupManager.getUserGroups(user));
        if (storedAssignments.isEmpty())
        {
            return List.of();
        }
        
        List<RightAssignmentContext> filteredAssignments = _rightAssignmentContextEP.getExtensionsIds()
                .stream()
                .map(id -> _rightAssignmentContextEP.getExtension(id))
                .filter(rac -> rac.matchWorkspace(contextualParameters))
                .filter(rac -> rac.showInUserProfilesTool())
                .toList();
        
        for (Object context: storedAssignments.keySet())
        {
            Map<String, Object> jsonAssignments = null;
            for (RightAssignmentContext assignmentContext: filteredAssignments)
            {
                if (assignmentContext.isSupportedContext(context, contextualParameters))
                {
                    ContextInfo contextInfo = assignmentContext.getContextInfo(context);
                    
                    if (jsonAssignments == null)
                    {
                        // Compute the assignment only if at least on right assignment context support the current context
                        jsonAssignments = _getAssignmentFromStorage(storedAssignments.get(context));
                    }
                    
                    Map<String, Object> contextResult = contextInfo.toJSON();
                    contextResult.putAll(jsonAssignments);
                    result.add(contextResult);
                }
            }
        }
        
        return result;
    }
    
    /**
     * Take a map associating AccessResultInfo to a set of profile
     * and serialize it into a JSON object organized by profile,
     * storing the result of the access result targeting this profile
     * and the list of each access result with the target sorted by weight
     * {
     *      profile1 : {
     *          accessResult: &lt;the merge result of all the AccessResult targeting the profile&gt;
     *          assignments: [
     *              {
     *                  accessResult: …
     *                  target: …
     *              },
     *              …
     *          ]
     *      },…
     * @param storedAssignments a map of assignments provided by the storage
     * @return a JSON object representing those assignments
     */
    private Map<String, Object> _getAssignmentFromStorage(Map<AccessResultInfo, Set<String>> storedAssignments)
    {
        Map<String, List<AccessResultInfo>> assignmentByProfiles = storedAssignments.entrySet().stream()
            .collect(
                // Supply a result object
                HashMap<String, List<AccessResultInfo>>::new,
                // Accumulate stream element in a result object
                (json, entry) -> {
                    for (String profile: entry.getValue())
                    {
                        json.compute(profile, (k, v) -> {
                            List<AccessResultInfo> infos;
                            if (v == null)
                            {
                                infos = new ArrayList<>();
                            }
                            else
                            {
                                infos = v;
                            }
                            infos.add(entry.getKey());
                            return infos;
                        });
                    }
                },
                // Combine result objects
                (json1, json2) -> {
                    for (Entry<String, List<AccessResultInfo>> entry: json2.entrySet())
                    {
                        List<AccessResultInfo> list = json1.get(entry.getKey());
                        if (list == null)
                        {
                            json1.put(entry.getKey(), entry.getValue());
                        }
                        else
                        {
                            list.addAll(entry.getValue());
                        }
                    }
                }
            );
        
        // assignments are now organized the way we want for the result
        // we now need to JSONify it
        Map<String, Object> result = new HashMap<>();
        for (Entry<String, List<AccessResultInfo>> entry : assignmentByProfiles.entrySet())
        {
            List<AccessResultInfo> assignments = entry.getValue();
            if (!assignments.isEmpty())
            {
                List<Map<String, Object>> assignmentsJSON = new ArrayList<>();
                List<AccessResult> accessResults = new ArrayList<>();
                // sort to have user > group > any connected > anonymous
                assignments.sort(Comparator.naturalOrder());
                
                for (AccessResultInfo assignment : assignments)
                {
                    AccessResult accessResult = assignment.accessResult();
                    // Store a list of the accessResult to merge later
                    accessResults.add(accessResult);
                    
                    Map<String, Object> assignmentJSON = new HashMap<>();
                    assignmentJSON.put("accessResult", accessResult.toString());
                    Object target = assignment.target();
                    if (target instanceof UserIdentity userIdentity)
                    {
                        assignmentJSON.put("target", _userHelper.user2json(userIdentity));
                    }
                    else if (target instanceof GroupIdentity groupIdentity)
                    {
                        assignmentJSON.put("target", _groupHelper.group2JSON(groupIdentity, false));
                    }
                    assignmentsJSON.add(assignmentJSON);
                }
                
                result.put(entry.getKey(), Map.of(
                        "accessResult", AccessResult.merge(accessResults),
                        "assignments", assignmentsJSON
                        ));
            }
        }
        return result;
    }

    /**
     * Get the contextual parameters used to determines the root contexts
     * @param jsParameters the client-side parameters
     * @return the contextual parameters
     */
    protected Map<String, Object> getContextualParameters(Map<String, Object> jsParameters)
    {
        Map<String, Object> contextParameters = new HashMap<>(jsParameters);
        
        // Add the current workspace to the contextualParameters
        Request request = ContextHelper.getRequest(_context);
        String workspaceName = (String) request.getAttribute(WorkspaceMatcher.WORKSPACE_NAME);
        contextParameters.put(WorkspaceMatcher.WORKSPACE_NAME, workspaceName);
        
        return contextParameters;
    }
    
    /**
     * Get user's information
     * @param login The user's login
     * @param populationId The id of the population
     * @return The user's information
     */
    @Callable(rights = "Runtime_Rights_SeeUserProfiles")
    public Map<String, Object> getUser (String login, String populationId)
    {
        return _userHelper.user2json(_userManager.getUser(populationId, login), true);
    }

}
