/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.schedule.progression;

import java.util.Map;

import org.slf4j.Logger;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * The abstract implementation of a {@link ProgressionTracker}
 */
public abstract sealed class AbstractProgressionTracker implements ProgressionTracker permits SimpleProgressionTracker, ContainerProgressionTracker
{
    /** The factory */
    protected ProgressionTrackerFactory _factory;
    /** The label */
    protected I18nizableText _label;
    /** The parent of the tracker */
    protected ContainerProgressionTracker _parent;
    
    private Logger _logger;
    private String _id;

    /**
     * A constructor with a {@link I18nizableText} argument
     * @param id The identifier
     * @param label The label
     * @param factory The factory
     * @param logger The logger
     * @param parent The parent of the tracker, null if it does not have one
     */
    protected AbstractProgressionTracker(String id, I18nizableText label, ProgressionTrackerFactory factory, Logger logger, ContainerProgressionTracker parent)
    {
        _id = id != null ? id : (parent != null ? parent.getId() + "-" : "") + label.toString();
        _label = label;
        _factory = factory;
        _logger = logger;
        _parent = parent;
    }
    
    public ContainerProgressionTracker getParent()
    {
        return _parent;
    }
    
    @Override
    public I18nizableText getLabel()
    {
        return _label;
    }
    
    @Override
    public Map<String, Object> toJSON()
    {
        double percentage = asPercentage();
        return Map.of(
            "id", getId(),
            "label", getLabel(),
            "index", getIndex(),
            "size", getSize(),
            "progression", percentage,
            "running", _parent == null ? true : _parent.getCurrentStep() == this
        );
    }
    
    /**
     * Get the log prefix
     * @return The log prefix
     */
    protected String _getLogPrefix()
    {
        return "[Step '" + _getHierarchicalLabel() + "']";
    }
    
    /**
     * Get the hierarchical label
     * @return The hierarchical label, or just label if the parent is null
     */
    protected String _getHierarchicalLabel()
    {
        if (_parent != null)
        {
            return _parent._getHierarchicalLabel() + " / " + _factory.getI18nUtils().translate(_label, "en");
        }
        else
        {
            return _factory.getI18nUtils().translate(_label, "en");
        }
    }
    
    /**
     * Get the logger
     * @return The logger
     */
    protected Logger _getLogger() 
    {
        return _logger;
    }

    public String getId()
    {
        return _id;
    }
}
