/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.core.schedule.progression;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * Container implementation that tracks a progression with steps
 */
public final class ContainerProgressionTracker extends AbstractProgressionTracker
{
    private static final String __UNIQUE_STEP_ID = "unique-step"; 
    private List<ProgressionTrackerWithWeight> _stepsWithWeight = new ArrayList<>();
    private double _totalStepsWeight;
    private int _currentStepIndex; // 0 based
    
    /**
     * A constructor
     * @param id The identifier
     * @param label The label
     * @param factory The factory
     * @param logger The logger
     * @param parent The parent
     */
    ContainerProgressionTracker(String id, I18nizableText label, ProgressionTrackerFactory factory, Logger logger, ContainerProgressionTracker parent)
    {
        super(id, label, factory, logger, parent);
    }
    
    @Override
    public long getIndex()
    {
        return _currentStepIndex + 1;
    }

    @Override
    public long getSize()
    {
        return _stepsWithWeight.size();
    }
    
    @Override
    public double asPercentage()
    {
        double progression = 0;
        
        if (_totalStepsWeight != 0)
        {
            for (ProgressionTrackerWithWeight progressionTrackerWithWeight : _stepsWithWeight)
            {
                progression += progressionTrackerWithWeight.progressionTracker().asPercentage() * progressionTrackerWithWeight.weight();
            }
            progression /= _totalStepsWeight;
        }
        
        return progression;
    }
    
    /**
     * Get a step by its index
     * @param index The index between 0 and size-1
     * @return The {@link ProgressionTracker} found
     * @throws IllegalArgumentException If no progression tracker was found for id
     */
    public ProgressionTracker getStep(int index) throws IllegalArgumentException
    {
        if (index >= 0 && index < getSize())
        {
            return _stepsWithWeight.get(index).progressionTracker();
        }
        
        throw new IllegalArgumentException("No progression tracker was found at index " + index + "' for parent of id '" + getId() + "'");
    }
    
    /**
     * Get the current step
     * @param <T> The type of tracker
     * @return The {@link ProgressionTracker} found or null if no step exists
     * @throws IllegalArgumentException If no step exists
     */
    @SuppressWarnings("unchecked")
    public <T extends ProgressionTracker> T getCurrentStep() throws IllegalArgumentException
    {
        if (_currentStepIndex < _stepsWithWeight.size())
        {
            return (T) _stepsWithWeight.get(_currentStepIndex).progressionTracker();
        }
        else
        {
            throw new IllegalArgumentException("There is no steps under parent of id '" + getId() + "'");        
        }
    }
    
    
    /**
     * Ensure the given progression tracker is the current one
     * @param progressionTracker The tracker
     */
    void ensureCurrentStep(ProgressionTracker progressionTracker)
    {
        int index = 0;
        for (ProgressionTrackerWithWeight stepWithWeight : _stepsWithWeight)
        {
            if (progressionTracker == stepWithWeight.progressionTracker())
            {
                _currentStepIndex = index;
                break;
            }
            index++;
        }
        
        _ensureCurrentStep();
    }
    
    private void _ensureCurrentStep()
    {
        if (_parent != null)
        {
            _parent.ensureCurrentStep(this);
        }
    }
    
    /**
     * Go to the next step
     */
    void nextStep()
    {
        _ensureCurrentStep();
        
        if (_parent != null && _currentStepIndex == _stepsWithWeight.size() - 1)
        {
            if (_getLogger().isInfoEnabled())
            {
                _getLogger().info(_getLogPrefix() + " Finished");
            }
            
            _parent.nextStep();
        }
        else
        {
            if (_currentStepIndex == _stepsWithWeight.size() - 1)
            {
                if (_getLogger().isInfoEnabled())
                {
                    _getLogger().info(_getLogPrefix() + " Finished");
                }
            }
            else
            {
                _currentStepIndex++;
                
                if (_getLogger().isInfoEnabled())
                {
                    _getLogger().info(_getLogPrefix() + " Starting step " + (_currentStepIndex + 1) + ": " + _factory.getI18nUtils().translate(getCurrentStep().getLabel(), "en"));
                }
            }
        }
    }
    
    /**
     * Add a step
     * @param label The label of the step
     * @param id The unique id of the step
     * @return A new {@link SimpleProgressionTracker} of weight 1
     */
    public SimpleProgressionTracker addSimpleStep(String id, I18nizableText label)
    {
        return addSimpleStep(id, label, 1d);
    }
    
    /**
     * Add a step
     * @param id The unique id of the step
     * @param label The label of the step
     * @param weight The weight wanted
     * @return A new {@link SimpleProgressionTracker}
     */
    public SimpleProgressionTracker addSimpleStep(String id, I18nizableText label, double weight)
    {
        SimpleProgressionTracker step = ProgressionTrackerFactory.createSimpleProgressionTracker(id, label, _getLogger(), this);
        
        _addStep(step, weight);
        
        return step;
    }
    
    /**
     * Add a step
     * @param id The unique id of the step
     * @param label The label of the step
     * @return A new {@link ContainerProgressionTracker} of weight 1
     */
    public ContainerProgressionTracker addContainerStep(String id, I18nizableText label)
    {
        return addContainerStep(id, label, 1d);
    }
    
    /**
     * Add a step
     * @param id The unique id of the step
     * @param label The label of the step
     * @param weight The weight wanted
     * @return A new {@link ContainerProgressionTracker}
     */
    public ContainerProgressionTracker addContainerStep(String id, I18nizableText label, double weight)
    {
        ContainerProgressionTracker step = ProgressionTrackerFactory.createContainerProgressionTracker(id, label, _getLogger(), this);
        
        _addStep(step, weight);
        
        return step;
    }
    
    /**
     * Add a step
     * @param step The step wanted
     * @param weight The weight wanted
     */
    private void _addStep(ProgressionTracker step, double weight)
    {
        if (_stepsWithWeight.size() == 0)
        {
            if (_getLogger().isInfoEnabled())
            {
                _getLogger().info(_getLogPrefix() + " Starting step 1: " + _factory.getI18nUtils().translate(step.getLabel(), "en"));
            }
        }
        
        _totalStepsWeight += weight;
        _stepsWithWeight.add(new ProgressionTrackerWithWeight(step, weight));
    }

    @Override
    public Map<String, Object> toJSON() 
    {
        Map<String, Object> result = new HashMap<>();
        
        result.putAll(super.toJSON());
        
        List<Map<String, Object>> stepsToJSON = _stepsWithWeight.stream()
            .map(ProgressionTrackerWithWeight::progressionTracker)
            .map(ProgressionTracker::toJSON)
            .toList();
        result.put("steps", stepsToJSON);
        
        return result;
    }
    
    /**
     * Get a step by its id in the direct children of this container
     * @param <T> The return type
     * @param id The id
     * @return The progression tracker found
     * @throws IllegalArgumentException If no progression tracker was found for the id wanted
     */
    @SuppressWarnings("unchecked")
    public <T extends ProgressionTracker> T getStep(String id) throws IllegalArgumentException
    {
        for (ProgressionTrackerWithWeight stepWithWeight : _stepsWithWeight)
        {
            ProgressionTracker progressionTracker =  stepWithWeight.progressionTracker();
            if (progressionTracker.getId().equals(id))
            {
                return (T) progressionTracker;
            }
        }

        throw new IllegalArgumentException("No progression tracker was found for id '" + id + "' of parent of id '" + getId() + "'");
    }
    
    /**
     * Get a step by its id, it searches for the step in all the tree of progression tracker
     * @param <T> The return type
     * @param id The id
     * @return The progression tracker found
     * @throws IllegalArgumentException If no progression tracker was found for the id wanted
     */
    public <T extends ProgressionTracker> T findStep(String id) throws IllegalArgumentException
    {
        try
        {
            return getStep(id);
        }
        catch (IllegalArgumentException e)
        {
            // Do Nothing
        }

        for (ProgressionTrackerWithWeight stepWithWeight : _stepsWithWeight)
        {
            ProgressionTracker progressionTracker = stepWithWeight.progressionTracker();
            if (progressionTracker instanceof ContainerProgressionTracker stepContainer)
            {
                try
                {
                    return stepContainer.findStep(id);
                }
                catch (IllegalArgumentException e)
                {
                    // Do Nothing
                }
            }
        }

        throw new IllegalArgumentException("No progression tracker was found for id '" + id + "' of parent of id '" + getId() + "'");
    }
    
    /**
     * Set the steps from a JSON
     * @param steps The steps as JSON
     */
    @SuppressWarnings("unchecked")
    public void setSteps(List<Map<String, Object>> steps)
    {
        for (Map<String, Object> stepMap : steps)
        {
            ProgressionTracker step;
            Double weight = ((Number) stepMap.getOrDefault("weight", 1)).doubleValue();
            
            Object label = stepMap.get("label");
            I18nizableText i18nlabel;
            if (label instanceof I18nizableText ilabel)
            {
                i18nlabel = ilabel;
            }
            else
            {
                i18nlabel = new I18nizableText(label.toString());
            }
            
            String id = (String) stepMap.get("id");
                
            
            if (stepMap.containsKey("children"))
            {
                step = addContainerStep(id, i18nlabel, weight);
            }
            else
            {
                step = addSimpleStep(id, i18nlabel, weight);
            }
            
            if (step instanceof ContainerProgressionTracker containerStep)
            {
                containerStep.setSteps((List<Map<String, Object>>) stepMap.get("children"));
            }
        }
    }
    
    /**
     * Initialize the progression with a unique sub step and set its size 
     * Use only when steps are not initialized yet 
     * @param size The size to set
     * @throws IllegalArgumentException If no progression tracker was found for id or if the {@link ProgressionTracker} of id wanted is not of type {@link SimpleProgressionTracker}
     */
    public void setSize(long size) throws IllegalArgumentException
    {
        setSize(size, "Unique step");
    }

    /**
     * Initialize the progression with a unique sub step and set its size 
     * Use only when steps are not initialized yet 
     * @param size The size to set
     * @param label The unique step label
     * @throws IllegalArgumentException If no progression tracker was found for id or if the {@link ProgressionTracker} of id wanted is not of type {@link SimpleProgressionTracker}
     */
    public void setSize(long size, String label) throws IllegalArgumentException
    {
        setSize(size, new I18nizableText(label));
    }
    
    /**
     * Initialize the progression with a unique sub step and set its size 
     * Use only when steps are not initialized yet 
     * @param size The size to set
     * @param label The unique step label
     * @throws IllegalArgumentException If no progression tracker was found for id or if the {@link ProgressionTracker} of id wanted is not of type {@link SimpleProgressionTracker}
     */
    public void setSize(long size, I18nizableText label) throws IllegalArgumentException
    {
        if (_stepsWithWeight.size() > 0)
        {
            throw new IllegalArgumentException("Cannot call setSize(long size) when the steps are already initialized");
        }
        
        SimpleProgressionTracker uniqueStep = addSimpleStep(__UNIQUE_STEP_ID, label);
        uniqueStep.setSize(size);
    }


    /**
     * Set the size of a {@link ProgressionTracker} by its id
     * @param id The id of the {@link ProgressionTracker} wanted
     * @param size The size to set
     * @throws IllegalArgumentException If no progression tracker was found for id or if the {@link ProgressionTracker} of id wanted is not of type {@link SimpleProgressionTracker}
     */
    public void setSize(String id, long size) throws IllegalArgumentException
    {
        ProgressionTracker step = findStep(id);
        if (step instanceof SimpleProgressionTracker simpleStep)
        {
            simpleStep.setSize(size);
        }
        else
        {
            throw new IllegalArgumentException("The progression tracker of id '" + id + "' is not a SimpleProgressionTracker");
        }
    }
    
    /**
     * Increment the unique-step of the {@link ProgressionTracker} of 1 increment
     * @throws IllegalArgumentException If the {@link ProgressionTracker} of id wanted is not of type {@link SimpleProgressionTracker} or if no progression tracker was found for id
     */
    public void increment() throws IllegalArgumentException
    {
        increment(__UNIQUE_STEP_ID, 1);
    }
    
    /**
     * Increment the unique-step of the {@link ProgressionTracker}
     * @param increment The increment wanted
     * @throws IllegalArgumentException If the {@link ProgressionTracker} of id wanted is not of type {@link SimpleProgressionTracker} or if no progression tracker was found for id
     */
    public void increment(long increment) throws IllegalArgumentException
    {
        increment(__UNIQUE_STEP_ID, increment);
    }
        
    /**
     * Increment the {@link ProgressionTracker} of id wanted of 1 increment
     * @param id The id
     * @throws IllegalArgumentException If the {@link ProgressionTracker} of id wanted is not of type {@link SimpleProgressionTracker} or if no progression tracker was found for id
     */
    public void increment(String id) throws IllegalArgumentException
    {
        increment(id, 1);
    }
    
    /**
     * Increment the {@link ProgressionTracker} of id wanted of the requested increment
     * @param id The id
     * @param increment The increment wanted
     * @throws IllegalArgumentException If the {@link ProgressionTracker} of id wanted is not of type {@link SimpleProgressionTracker} or if no progression tracker was found for id
     */
    public void increment(String id, long increment) throws IllegalArgumentException
    {
        ProgressionTracker step = findStep(id);
        if (step instanceof SimpleProgressionTracker simpleStep)
        {
            simpleStep.increment(increment);
        }
        else
        {
            throw new IllegalArgumentException("The progression tracker of id '" + id + "' is not a SimpleProgressionTracker or does not exist");
        }
    }
    
    /**
     * The record that contains a sub step {@link ProgressionTracker} and its weight
     * @param progressionTracker The sub step {@link ProgressionTracker}
     * @param weight The weight of the sub step {@link ProgressionTracker}
     */
    private record ProgressionTrackerWithWeight(ProgressionTracker progressionTracker, double weight) { }
}
