/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.core.schedule.progression;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.slf4j.Logger;

import org.ametys.core.util.I18nUtils;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * A {@link ProgressionTracker} factory
 */
public class ProgressionTrackerFactory implements Component, Serviceable
{
    /** Avalon Role */
    public static final String ROLE = ProgressionTrackerFactory.class.getName();

    private static I18nUtils _i18nUtils;
    private static ProgressionTrackerFactory _instance;
    
    /**
     * Constructor
     */
    public ProgressionTrackerFactory()
    {
        _instance = this;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
    }
    
    /**
     * Create a {@link SimpleProgressionTracker}
     * @param label The label as a String
     * @param logger The logger
     * @return the new fake {@link SimpleProgressionTracker}
     */
    public static SimpleProgressionTracker createSimpleProgressionTracker(String label, Logger logger)
    {
        return createSimpleProgressionTracker(null, new I18nizableText(label), logger, null);
    }
    
    /**
     * Create a {@link SimpleProgressionTracker}
     * @param label The label
     * @param logger The logger
     * @return the new {@link SimpleProgressionTracker}
     */
    public static SimpleProgressionTracker createSimpleProgressionTracker(I18nizableText label, Logger logger)
    {
        return createSimpleProgressionTracker(null, label, logger, null);
    }
    
    /**
     * Create a {@link SimpleProgressionTracker}
     * @param parent The parent of the tracker, null if it does not have one
     * @param id The identifier
     * @param label The label
     * @param logger The logger
     * @return the new {@link SimpleProgressionTracker}
     */
    static SimpleProgressionTracker createSimpleProgressionTracker(String id, I18nizableText label, Logger logger, ContainerProgressionTracker parent)
    {
        return new SimpleProgressionTracker(id, label, _instance, logger, parent);
    }
    
    /**
     * Create a fake {@link ContainerProgressionTracker}
     * @param label The label as a String
     * @param logger The logger
     * @return the new fake {@link ContainerProgressionTracker}
     */
    public static ContainerProgressionTracker createContainerProgressionTracker(String label, Logger logger)
    {
        return createContainerProgressionTracker(null, new I18nizableText(label), logger, null);
    }
    
    /**
     * Create a {@link ContainerProgressionTracker}
     * @param label The label
     * @param logger The logger
     * @return the new {@link ContainerProgressionTracker}
     */
    public static ContainerProgressionTracker createContainerProgressionTracker(I18nizableText label, Logger logger)
    {
        return createContainerProgressionTracker(null, label, logger, null);
    }
    
    /**
     * Create a {@link ContainerProgressionTracker}
     * @param id The identifier
     * @param parent The parent of the tracker, null if it does not have one
     * @param label The label
     * @param logger The logger
     * @return the new {@link ContainerProgressionTracker}
     */
    static ContainerProgressionTracker createContainerProgressionTracker(String id, I18nizableText label, Logger logger, ContainerProgressionTracker parent)
    {
        return new ContainerProgressionTracker(id, label, _instance, logger, parent);
    }

    /**
     * Get the i18n utils
     * @return The i18nUtils
     */
    I18nUtils getI18nUtils()
    {
        return _i18nUtils;
    }
}
