/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.core.schedule.progression;

import org.slf4j.Logger;

import org.ametys.runtime.i18n.I18nizableText;

/**
 * Simple implementation that tracks a progression with graduations
 */
public final class SimpleProgressionTracker extends AbstractProgressionTracker
{
    private long _index;
    private long _size = 1;
    private int _loggerGranularity = 500;

    /**
     * A constructor
     * @param id The identifier
     * @param label The label
     * @param factory The factory
     * @param logger The logger
     * @param parent The parent
     */
    SimpleProgressionTracker(String id, I18nizableText label, ProgressionTrackerFactory factory, Logger logger, ContainerProgressionTracker parent)
    {
        super(id, label, factory, logger, parent);
    }
    
    /**
     * Add a granularity to the tracker
     * @param loggerGranularity The granularity
     * @return The tracker
     */
    public SimpleProgressionTracker withLoggerGranularity(int loggerGranularity)
    {
        _loggerGranularity = loggerGranularity;
        
        return this;
    }
    
    /**
     * Complete the step
     */
    protected void complete()
    {
        if (_parent != null)
        {
            _parent.nextStep();
        }
    }
    
    @Override
    public long getIndex()
    {
        return _index;
    }

    @Override
    public long getSize()
    {
        return _size;
    }
    
    /**
     * Set the size
     * @param size The size wanted
     */
    public void setSize(long size)
    {
        this._size = size;

        if (size == 0)
        {
            complete();
        }
    }
    
    /**
     * Increment the progression of 1 increment
     */
    public void increment()
    {
        increment(1);
    }
    
    /**
     * Increment the progression of 'increment' graduations
     * @param increment The number of increment to add
     */
    public void increment(long increment)
    {
        long newCurrentGraduation = _index + increment;
        
        _ensureCurrentStep();
        
        if (newCurrentGraduation < _size)
        {
            if (_getLogger().isInfoEnabled())
            {
                if (((newCurrentGraduation - 1) / _loggerGranularity) != ((_index - 1) / _loggerGranularity))
                {
                    _getLogger().info(_getLogPrefix() + " The incremental progression is now " + _index + " out of " + _size + " graduations.");
                }
            }
            
            _index = newCurrentGraduation;
        }
        else 
        {
            _index = _size;
            complete();
            
            if (newCurrentGraduation > _size)
            {
                if (_getLogger().isWarnEnabled())
                {
                    _getLogger().warn("[Step of id '" + getId() + "'] The incrementation of the progression has tried to exceed the maximum.");
                }
            }
        }
    }
    
    private void _ensureCurrentStep()
    {
        if (_parent != null)
        {
            _parent.ensureCurrentStep(this);
        }
    }

    @Override
    public double asPercentage()
    {
        if (_size == 0)
        {
            return 100d;
        }

        if (_index < 0)
        {
            return 0d;
        }
        
        return (((double) _index) / _size) * 100;
    }
}
