/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.servletwrapper.servlet;

import java.io.IOException;
import java.util.Map;

import javax.servlet.Servlet;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.http.HttpEnvironment;
/**
 * This class is a generic servlet wrapping an actual {@link Servlet}. 
 * {@link ServletWrapper} are used to handle servlets from inside the application, instead of inside the servlet engine.
 */
public class ServletWrapper
{
    private Servlet _servlet;

    /**
     * Constructor
     * @param servlet A servlet to apply
     */
    public ServletWrapper(Servlet servlet)
    {
        _servlet = servlet;
    }
    
    /**
     * Initialize the servlet in the wrapper
     * @param servletName Name
     * @param servletContext Context
     * @param parameters Parameters
     * @throws ServletException Exception if servlet init fails
     */
    public void init(String servletName, ServletContext servletContext, Map<String, String> parameters) throws ServletException
    {
        try
        {
            _servlet.init(new ServletWrapperConfig(servletName, servletContext, parameters));
        }
        catch (Exception e)
        {
            throw new ServletException("Impossible to initialize the filter.", e);
        }
    }
    
    /**
     * Calls service in the servlet, wrapping the request and the response
     * @param objectModel object model
     * @param redirect redirector
     * @throws ServletException exception thrown by the servlet
     * @throws IOException exception thrown by the servlet
     */
    public void service(Map objectModel, Redirector redirect) throws ServletException, IOException
    {
        HttpServletRequest request = (HttpServletRequest) objectModel.get(HttpEnvironment.HTTP_REQUEST_OBJECT);
        HttpServletResponse response = (HttpServletResponse) objectModel.get(HttpEnvironment.HTTP_RESPONSE_OBJECT);
        this.service(request, response, objectModel, redirect);
    }
    
    /**
     * Calls service in the servlet
     * @param request request
     * @param response response
     * @param objectModel object model
     * @param redirect redirector
     * @throws ServletException exception thrown by the servlet
     * @throws IOException exception thrown by the servlet
     */
    public void service(HttpServletRequest request, HttpServletResponse response, Map objectModel, Redirector redirect) throws ServletException, IOException
    {
        ServletWrapperResponse filteredResponse = new ServletWrapperResponse(response, redirect);
        
        _servlet.service(request, filteredResponse);
    }
}
