/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.trace;

import java.util.Collection;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Spliterator;
import java.util.Spliterators;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.collections4.MapUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.core.observation.Event;
import org.ametys.core.observation.Observer;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.util.StringUtils;

/**
 * This class is an observer and a helper to trace all forensic events
 */
public class ForensicLogger implements Observer, Contextualizable
{
    // [127.0.0.1] [login#population] event.id (arg1::value1|arg2::value2|...) (header1::value1|header2::value2|...)
    private static final String __DEFAULT_LOG_FORMAT = "[{}] [{}] {} ({}) ({})";
    
    private static final String __DEFAULT_ARGS_SEPARATOR = "|";
    
    private static final List<String> __REQUEST_HEADER_NAME = List.of("origin", "referer", "user-agent", "host");
    
    private static final String __NO_REQUEST = "No Request";
    private static final String __UNKNOWN_ISSUER = "Unknown Issuer";
    private static final String __UNKNOWN_IP = "Unknown IP";
    
    private static Map<String, Logger> _loggers = new HashMap<>();
    private static Context _context;
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    /**
     * Get the logger of a given event ID for tracing.
     * @param eventId The event identifier.
     * @return The logger corresponding to the event
     */
    public static synchronized Logger getLogger(String eventId)
    {
        return _loggers.computeIfAbsent(eventId, key -> LoggerFactory.getLogger(ForensicLogger.class.getName() + "." + key));
    }
    
    public boolean supports(Event event)
    {
        return getLogger(event.getId()).isInfoEnabled();
    }

    public int getPriority(Event event)
    {
        return Integer.MAX_VALUE;
    }

    public void observe(Event event, Map<String, Object> transientVars) throws Exception
    {
        info(event.getId(), event.getArguments(), event.getIssuer());
    }
    
    /**
     * Trace event in INFO level to following format: [IP::127.0.0.1] [Issuer::login#population] event.id (arg1::value1|arg2::value2|...) (header1::value1|header2::value2|...)
     * @param eventId the event id
     * @param eventArgs the event arguments
     * @param issuer the issuer. Can be null.
     */
    public static void info(String eventId, Map<String, Object> eventArgs, UserIdentity issuer)
    {
        Logger logger = getLogger(eventId);
        if (logger.isInfoEnabled())
        {
            logger.info(
                    __DEFAULT_LOG_FORMAT,
                    getIP(),
                    eventIssuer(issuer),
                    eventId,
                    eventArguments(eventArgs),
                    requestHeaders()
            );
        }
    }
    
    
    
    /**
     * Trace event in WARN level to following format: [IP::127.0.0.1] [Issuer::login#population] event.id (arg1::value1|arg2::value2|...) (header1::value1|header2::value2|...)
     * @param eventId the event id
     * @param eventArgs the event arguments
     * @param issuer the issuer. Can be null.
     */
    public static void warn(String eventId, Map<String, Object> eventArgs, UserIdentity issuer)
    {
        Logger logger = getLogger(eventId);
        if (logger.isWarnEnabled())
        {
            logger.warn(
                    __DEFAULT_LOG_FORMAT,
                    getIP(),
                    eventIssuer(issuer),
                    eventId,
                    eventArguments(eventArgs),
                    requestHeaders()
            );
        }
    }
    
    /**
     * Get IP from request
     * @return the IP or [unknown] if there is no request
     */
    public static String getIP()
    {
        try
        {
            Request request = ContextHelper.getRequest(_context);
            if (request != null)
            {
                return Optional.of("X-Forwarded-For")
                        .map(request::getHeader)
                        .map(StringUtils::stringToCollection)
                        .map(Collection::stream)
                        .orElseGet(Stream::of)
                        .findFirst()
                        .orElseGet(request::getRemoteAddr);
            }
            
        }
        catch (Exception e)
        {
            // Ignore if there is no request (not a client call)
        }
        
        return __UNKNOWN_IP;
    }
    
    /**
     * Get the request headers to log
     * @return the request headers
     */
    public static String requestHeaders()
    {
        try
        {
            Request request = ContextHelper.getRequest(_context);
            if (request != null)
            {
                Enumeration<String> headerNames = request.getHeaderNames();
                Stream<String> headerNamesAsStream = StreamSupport.stream(
                    Spliterators.spliteratorUnknownSize(headerNames.asIterator(), Spliterator.ORDERED),
                    false
                );

                return headerNamesAsStream
                    .filter(name -> __REQUEST_HEADER_NAME.contains(name))
                    .map(name -> name + "::" + request.getHeader(name))
                    .collect(Collectors.joining(__DEFAULT_ARGS_SEPARATOR));
            }
            
        }
        catch (Exception e)
        {
            // Ignore if there is no request (not a client call)
        }
        
        return __NO_REQUEST;
    }
    
    
    /**
     * Get readable value of issuer
     * @param user the user. Can be null
     * @return the issuer to display in logs
     */
    public static String eventIssuer(UserIdentity user)
    {
        return Optional.ofNullable(user)
            .map(UserIdentity::userIdentityToString)
            .orElse(__UNKNOWN_ISSUER);
    }
    
    /**
     * Build message for event arguments
     * @param arguments the arguments
     * @return the message to trace for arguments
     */
    public static String eventArguments(Map<String, Object> arguments)
    {
        if (MapUtils.isNotEmpty(arguments))
        {
            return arguments.entrySet()
                .stream()
                .map(e -> e.getKey() + "::" + e.getValue())
                .collect(Collectors.joining(__DEFAULT_ARGS_SEPARATOR));
        }

        return org.apache.commons.lang3.StringUtils.EMPTY;
    }
}
