/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.user.population;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;

import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;

import org.ametys.core.ObservationConstants;
import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;
import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.runtime.authentication.AccessDeniedException;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Helper for associating {@link UserPopulation}s to contexts.
 */
public class PopulationContextHelper extends AbstractLogEnabled implements Component, Serviceable, Initializable
{
    /** Avalon Role */
    public static final String ROLE = PopulationContextHelper.class.getName();
 
    /** The name of request attribute holding the current population contexts */
    public static final String POPULATION_CONTEXTS_REQUEST_ATTR = "populationContexts";
    
    /** The "admin" context */
    public static final String ADMIN_CONTEXT = "/admin";
    
    /** The name of the JDBC table for user populations by context */
    protected static final String __USER_POPULATIONS_TABLE = "UserPopulationsByContext";
    
    /** The id of right for administration access */
    protected static final String __ADMIN_RIGHT_ACCESS = "Runtime_Rights_Admin_Access";

    private static final Comparator<Pair<String, Integer>> _COMPARATOR = Comparator.comparingInt(Pair::getRight);
    
    private static final String IS_CACHE_FILLED = "###iscachefilled###";
    
    /** The service manager */
    protected ServiceManager _manager;
    
    /** The DAO for {@link UserPopulation}s */
    private UserPopulationDAO _userPopulationDAO;

    private ObservationManager _observationManager;

    private CurrentUserProvider _currentUserProvider;

    private RightManager _rightManager;
    
    private AbstractCacheManager _abstractCacheManager;
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _abstractCacheManager = (AbstractCacheManager) _manager.lookup(AbstractCacheManager.ROLE);
    }

    @Override
    public void initialize() throws Exception
    {
        _abstractCacheManager.createMemoryCache(ROLE, 
                new I18nizableText("plugin.core", "PLUGINS_CORE_CACHE_POPULATION_CONTEXT_LABEL"),
                new I18nizableText("plugin.core", "PLUGINS_CORE_CACHE_POPULATION_CONTEXT_DESCRIPTION"),
                true,
                null);
    }
    
    /**
     * Get the observation manager
     * @return the 
     */
    protected ObservationManager getObservationManager()
    {
        if (_observationManager == null)
        {
            try
            {
                _observationManager = (ObservationManager) _manager.lookup(ObservationManager.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException(e);
            }
        }
        return _observationManager;
    }
    
    /**
     * Get the DAO for user populations
     * @return the DAO for user populations
     */
    protected UserPopulationDAO getUserPopulationDAO()
    {
        if (_userPopulationDAO == null)
        {
            try
            {
                _userPopulationDAO = (UserPopulationDAO) _manager.lookup(UserPopulationDAO.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException(e);
            }
        }
        return _userPopulationDAO;
    }
    
    /**
     * Get the current user provider
     * @return the current user provider
     */
    protected CurrentUserProvider getCurrentUserProvider()
    {
        if (_currentUserProvider == null)
        {
            try
            {
                _currentUserProvider = (CurrentUserProvider) _manager.lookup(CurrentUserProvider.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException(e);
            }
        }
        return _currentUserProvider;
    }
    
    /**
     * Get the right manager
     * @return the right manager
     */
    protected RightManager getRightManager()
    {
        if (_rightManager == null)
        {
            try
            {
                _rightManager = (RightManager) _manager.lookup(RightManager.ROLE);
            }
            catch (ServiceException e)
            {
                throw new RuntimeException(e);
            }
        }
        return _rightManager;
    }
    
    /**
     * Get the connection to the database 
     * @return the SQL connection
     */
    protected Connection getSQLConnection ()
    {
        String datasourceId = Config.getInstance().getValue("runtime.assignments.userpopulations");
        return ConnectionHelper.getConnection(datasourceId);
    }
    
    /**
     * Links given populations to a context.
     * @param context The context
     * @param ids The ids of the populations to link
     * @return The ids of the changed user populations (the ones unlinked and the ones linked)
     */
    @SuppressWarnings("resource")
    @Callable (rights = "Runtime_Rights_Admin_Access", context = "/admin")
    public Set<String> link(String context, Collection<String> ids)
    {
        Set<String> result = getUserPopulationsOnContext(context, true); // Get the already linked ids
        
        Connection connection = null;
        PreparedStatement stmt = null;
        try
        {
            connection = getSQLConnection();
            
            // Remove all the ids affected to this context
            String sql = "DELETE FROM " + __USER_POPULATIONS_TABLE + " WHERE Context=?";
            stmt = connection.prepareStatement(sql);
            stmt.setString(1, context);
            stmt.executeUpdate();
            getLogger().info("{}\n[{}]", sql, context);
            
            // Set the new ids to this context
            sql = "INSERT INTO " + __USER_POPULATIONS_TABLE + " (Context, Ordering, UserPopulation_Id) VALUES(?, ?, ?)";
            stmt = connection.prepareStatement(sql);
            
            int index = 0;
            for (String id : ids)
            {
                if (getUserPopulationDAO().getUserPopulation(id) != null)
                {
                    stmt.setString(1, context);
                    stmt.setInt(2, index);
                    stmt.setString(3, id);
                    stmt.executeUpdate();
                    
                    getLogger().info("{}\n[{}, {}, {}]", sql, context, index, id);
                    if (result.contains(id))
                    {
                        result.remove(id); // the population was already linked, so its status didn't changed
                    }
                    else
                    {
                        result.add(id);
                    }
                }
                else
                {
                    getLogger().warn("The user population with id '{}' does not exist. It will not be linked.", id);
                }
                
                index++;
            }
        }
        catch (SQLException e)
        {
            throw new IllegalStateException("SQL error while linking user populations to a context", e);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
            ConnectionHelper.cleanup(stmt);
            _getCache().invalidateAll();
        }
        
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_USERPOPULATION_IDS, ids);
        eventParams.put(ObservationConstants.ARGS_USERPOPULATION_CONTEXT, context);
        getObservationManager().notify(new Event(ObservationConstants.EVENT_USERPOPULATIONS_ASSIGNMENT, getCurrentUserProvider().getUser(), eventParams));
        
        return result;
    }
    
    /**
     * Gets the populations linked to the given context (need the population to be enabled)
     * @param context The context
     * @param withDisabled True to also return disabled populations
     * @return The ids of populations linked to the context
     */
    public Set<String> getUserPopulationsOnContext(String context, boolean withDisabled)
    {
        if (ADMIN_CONTEXT.equals(context))
        {
            // Return all the enabled populations
            List<UserPopulation> populations = withDisabled ? getUserPopulationDAO().getUserPopulations(true) : getUserPopulationDAO().getEnabledUserPopulations(true);
            return populations.stream().map(UserPopulation::getId).collect(Collectors.toSet());
        }
        else
        {
            return _getPopulationsOnContextFromDatabase(context, withDisabled);
        }
    }
    
    /**
     * Gets the populations linked to at least one of the given contexts (need the population to be enabled)
     * @param contexts The contexts
     * @param withDisabled True to also return disabled populations
     * @param checkRights True to check that current user belongs to one of populations or is an administrator user
     * @return The ids of populations linked to the contexts
     */
    public Set<String> getUserPopulationsOnContexts(Collection<String> contexts, boolean withDisabled, boolean checkRights)
    {
        UserIdentity currentUser = getCurrentUserProvider().getUser();
        if (checkRights && contexts.contains(ADMIN_CONTEXT) && getRightManager().currentUserHasRight(__ADMIN_RIGHT_ACCESS, ADMIN_CONTEXT) != RightResult.RIGHT_ALLOW)
        {
            throw new AccessDeniedException("User " + getCurrentUserProvider().getUser() +  " tried to access the list of all user populations without convenient rights");
        }
        
        if (!checkRights)
        {
            return contexts.stream()
                    .map(context -> getUserPopulationsOnContext(context, withDisabled))
                    .flatMap(Set::stream)
                    .collect(Collectors.toSet());
        }
        
        if (currentUser == null)
        {
            throw new AccessDeniedException("Anonymous user tried to access the list of user populations on contexts '" + StringUtils.join(contexts, ",") + "'.");
        }
        
        boolean isAdministrator = getRightManager().currentUserHasRight(__ADMIN_RIGHT_ACCESS, ADMIN_CONTEXT) == RightResult.RIGHT_ALLOW;
        Set<String> populations = new HashSet<>();
        
        // Check current user belongs to a population on each context
        for (String context : contexts)
        {
            Set<String> ctxPopulations = getUserPopulationsOnContext(context, withDisabled);
            
            if (!ctxPopulations.contains(currentUser.getPopulationId()) && !isAdministrator)
            {
                throw new AccessDeniedException("User " + currentUser +  " tried to access the list of user populations on context '" + context + "', but he does not belong to any populations on this context.");
            }
            
            populations.addAll(ctxPopulations);
        }
        
        return populations;
    }
    
    /**
     * Gets the populations linked to at least one of the given contexts (need the population to be enabled)
     * @param contexts The contexts
     * @param withDisabled True to also return disabled populations
     * @return The ids of populations linked to the contexts
     */
    @Callable (rights = "Runtime_Rights_Admin_Access", context = "/admin")
    public Set<String> getUserPopulationsOnContexts(Collection<String> contexts, boolean withDisabled)
    {
        return getUserPopulationsOnContexts(contexts, withDisabled, true);
    }
    
    private Set<String> _getPopulationsOnContextFromDatabase(String context, boolean withDisabled)
    {
        Set<String> result = new LinkedHashSet<>();
        
        for (Pair<String, Integer> pair : _getCacheAsMap().getOrDefault(context, Collections.emptySet()))
        {
            String userPopulationId = pair.getLeft();
            
            if (withDisabled || getUserPopulationDAO().getUserPopulation(userPopulationId).isEnabled())
            {
                result.add(userPopulationId);
            }
            else
            {
                getLogger().warn("The population of id '{}' is linked to a context but disabled. It will not be returned.", userPopulationId);
            }
        }
        
        return result;
    }

    /**
     * create a population content map from the database
     * @return Non null map of population content
     */
    private synchronized Map<String, Set<Pair<String, Integer>>> _getAllPopulationContent()
    {
        Map<String, Set<Pair<String, Integer>>> cache = new HashMap<>();
        
        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        try
        {
            connection = getSQLConnection();
            
            String sql = "SELECT Context, Ordering, UserPopulation_Id FROM " + __USER_POPULATIONS_TABLE;
            stmt = connection.prepareStatement(sql);
            rs = stmt.executeQuery();
            
            while (rs.next())
            {
                String context = rs.getString(1);
                Integer ordering = rs.getInt(2);
                String userPopulationId = rs.getString(3);
                
                if (getUserPopulationDAO().getUserPopulation(userPopulationId) == null)
                {
                    getLogger().warn("The population of id '{}' is linked to a context, but does not exist anymore.", userPopulationId);
                }
                else
                {
                    Set<Pair<String, Integer>> cacheForContext = cache.computeIfAbsent(context, c -> new TreeSet<>(_COMPARATOR)); // Big number mean low priority
                    
                    cacheForContext.add(Pair.of(userPopulationId, ordering));
                }
            }
            
        }
        catch (SQLException e)
        {
            throw new IllegalStateException("Cannot get populations and contexts association because of an error in a sql query", e);
            
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(rs);
        }
        return cache;
    }

    /**
     * Get the cache
     * @return Non null cache
     */
    protected synchronized Cache<String, Set<Pair<String, Integer>>> _getCache()
    {
        Cache<String, Set<Pair<String, Integer>>> cache = _abstractCacheManager.get(ROLE); 
        
        if (!cache.hasKey(IS_CACHE_FILLED))
        {
            Map<String, Set<Pair<String, Integer>>> cacheMap = _getAllPopulationContent();
            cache.putAll(cacheMap);
            cache.put(IS_CACHE_FILLED, null);
        }
        return cache;
    }
    
    /**
     * Get the cache as map
     * @return Non null cache
     */
    protected synchronized Map<String, Set<Pair<String, Integer>>> _getCacheAsMap()
    {
        Map<String, Set<Pair<String, Integer>>> cacheAsMap = _getCache().asMap();
        cacheAsMap.remove(IS_CACHE_FILLED);
        return cacheAsMap;
    }
    
    /**
     * Returns true if the user population is linked to a context.
     * @param upId The id of the user population
     * @return True if the user population is currently linked to a context.
     */
    public boolean isLinked(String upId)
    {
        return _getCacheAsMap().values().stream()
                                    .flatMap(s -> s.stream())
                                    .map(Pair::getLeft)
                                    .anyMatch(u -> StringUtils.equals(u, upId));
    }

}
