/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.core.util;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.xml.dom.DOMBuilder;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.text.StringEscapeUtils;
import org.apache.commons.text.translate.LookupTranslator;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import org.ametys.core.DevMode;
import org.ametys.core.DevMode.DEVMODE;
import org.ametys.core.group.Group;
import org.ametys.core.group.GroupIdentity;
import org.ametys.core.group.GroupManager;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.user.directory.NotUniqueUserException;
import org.ametys.core.util.dom.AmetysNodeList;
import org.ametys.core.util.dom.MapElement;
import org.ametys.core.util.dom.StringElement;
import org.ametys.core.version.Version;
import org.ametys.core.version.VersionsHandler;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.PluginsManager;
import org.ametys.runtime.servlet.RuntimeConfig;
import org.ametys.runtime.workspace.WorkspaceManager;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * Helper component to be used from XSL stylesheets.
 */
public class AmetysXSLTHelper implements Contextualizable, Serviceable
{
    /** The logger */
    protected static final Logger _LOGGER = LoggerFactory.getLogger(AmetysXSLTHelper.class.getName());
    
    /** The i18n utils instance */
    protected static I18nUtils _i18nUtils;
    
    /** The versions handler */
    protected static VersionsHandler _versionHandler;

    /** The current user provider */
    protected static CurrentUserProvider _currentUserProvider;
    /** The groups manager */
    protected static GroupManager _groupManager;
    /** The user helper */
    protected static UserHelper _userHelper;
    /** The json utils */
    protected static JSONUtils _jsonUtils;
    /** The right manager */
    protected static RightManager _rightManager;
    /** The user manager */
    protected static UserManager _userManager;
    /** The source resolver */
    protected static SourceResolver _sourceResolver;
    
    private static Context _context;

    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
        _versionHandler = (VersionsHandler) manager.lookup(VersionsHandler.ROLE);
        
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
        _groupManager = (GroupManager) manager.lookup(GroupManager.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
        _rightManager = (RightManager) manager.lookup(RightManager.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    /**
     * Returns the current URI prefix.
     * @return the current URI prefix.
     */
    public static String uriPrefix()
    {
        return uriPrefix(true);
    }
    
    /**
     * Returns the current URI prefix.
     * @param withWorkspaceURI true to add the workspace URI (recommended)
     * @return the current URI prefix.
     */
    public static String uriPrefix(boolean withWorkspaceURI)
    {
        return getUriPrefix(withWorkspaceURI);
    }
    
    /**
     * Returns the absolute URI prefix.
     * @return the absolute URI prefix.
     */
    public static String absoluteUriPrefix()
    {
        return absoluteUriPrefix(true);
    }
    
    /**
     * Returns the absolute URI prefix.
     * @param withWorkspaceURI true to add the workspace URI (recommended)
     * @return the absolute URI prefix.
     */
    public static String absoluteUriPrefix(boolean withWorkspaceURI)
    {
        return getAbsoluteUriPrefix(withWorkspaceURI);
    }
    
    /**
     * Return the current workspace name
     * @return The workspace name. Cannot be empty.
     */
    public static String workspaceName()
    {
        return getWorkspaceName();
    }
    
    /**
     * Return the current workspace URI
     * @return The workspace name. Can be empty.
     */
    public static String workspacePrefix()
    {
        return getWorkspacePrefix();
    }
    
    /**
     * Return the current workspace theme name
     * @return The name
     */
    public static String workspaceTheme()
    {
        Request request = ContextHelper.getRequest(_context);
        return (String) request.getAttribute(WorkspaceMatcher.WORKSPACE_THEME);
    }

    /**
     * Return the current workspace theme url
     * @return The url without any prefix
     */
    public static String workspaceThemeURL()
    {
        Request request = ContextHelper.getRequest(_context);
        
        String workspaceThemeUrl = (String) request.getAttribute(WorkspaceMatcher.WORKSPACE_THEME_URL);
        if (workspaceThemeUrl == null)
        {
            // fallback to the default workspace
            String workspaceName = RuntimeConfig.getInstance().getDefaultWorkspace();
            WorkspaceManager wm = WorkspaceManager.getInstance();
            if (wm.getWorkspaceNames().contains(workspaceName))
            {
                workspaceThemeUrl = wm.getWorkspaces().get(workspaceName).getThemeURL();
            }
        }
        
        return workspaceThemeUrl;
    }
    
    /**
     * Get the application context path. Can be empty if the application
     * resides in the root context. Use it to create a link beginning with
     * the application root.
     * @param withWorkspaceURI true to add the workspace URI (recommended)
     * @return The application context path with workspace URI
     * @see Request#getContextPath()
     */
    protected static String getUriPrefix(boolean withWorkspaceURI)
    {
        Request request = ContextHelper.getRequest(_context);
        String workspaceURI = withWorkspaceURI ? getWorkspacePrefix() : "";
        
        return request.getContextPath() + workspaceURI;
    }
    
    /**
     * Get the absolutized version of the context path. Use it to create an absolute
     * link beginning with the application root, for instance when sending a mail
     * linking to the application.
     * @param withWorkspaceURI true to add the workspace URI (recommended)
     * @return The absolute context path.
     */
    protected static String getAbsoluteUriPrefix(boolean withWorkspaceURI)
    {
        Request request = ContextHelper.getRequest(_context);
        
        String uriPrefix = getUriPrefix(withWorkspaceURI);
        
        if (!uriPrefix.startsWith("http"))
        {
            uriPrefix = request.getScheme() + "://" + request.getServerName() + (request.getServerPort() != 80 ? ":" + request.getServerPort() : "") + uriPrefix;
        }
        
        return uriPrefix;
    }  
    
    /**
     * Return the current workspace name
     * @return The workspace name. Cannot be empty.
     */
    protected static String getWorkspaceName()
    {
        Request request = ContextHelper.getRequest(_context);
        return (String) request.getAttribute(WorkspaceMatcher.WORKSPACE_NAME);
    }
    
    /**
     * Return the current workspace URI
     * @return The workspace name. Can be empty for the default workspace.
     */
    protected static String getWorkspacePrefix()
    {
        Request request = ContextHelper.getRequest(_context);
        return (String) request.getAttribute(WorkspaceMatcher.WORKSPACE_URI);
    }

    /**
     * Returns the configuration value associated with the given parameter.
     * @param id the configuration parameter.
     * @return the configuration value associated with the given parameter.
     */
    public static Object config(String id)
    {
        if (Config.getInstance() != null)
        {
            return Config.getInstance().getValue(id);
        }
        else
        {
            return null;
        }
    }
    
    /**
     * Return the value of a request parameter.
     * @param parameter the parameter name.
     * @return the request parameter.
     */
    public static String requestParameter(String parameter)
    {
        Request request = ContextHelper.getRequest(_context);
        return request.getParameter(parameter);
    }
    
    /**
     * Translate an i18n key using current user language.
     * @param key The key to translate. Specify the catalog this way: "catalogue:KEY"
     * @return The translation or null.
     */
    public static String translate(String key)
    {
        return translate(key, null, null);
    }
    
    /**
     * Translate an i18n key
     * @param key The key to translate. Specify the catalog this way: "catalogue:KEY"
     * @param lang The language. Can be null to use current user language.
     * @return The translation or null.
     */
    public static String translate(String key, String lang)
    {
        return translate(key, lang, null);
    }
    
    /**
     * Translate an i18n key
     * @param key The key to translate. Specify the catalog this way: "catalogue:KEY"
     * @param lang The language. Can be null to use current user language.
     * @param parameters The key parameters. Can be empty.
     * @return The translation or null.
     */
    public static String translate(String key, String lang, NodeList parameters)
    {
        List<String> i18nparams = new ArrayList<>();
        if (parameters != null && parameters.getLength() == 1)
        {
            NodeList childNodes = parameters.item(0).getChildNodes();
            for (int i = 0; i < childNodes.getLength(); i++) 
            {  
                i18nparams.add(childNodes.item(i).getTextContent());
            }
        }
        
        I18nizableText i18nKey = new I18nizableText(null, key, i18nparams);
        return _i18nUtils.translate(i18nKey, lang);
    }
    
    /**
     * Escape the given string to be used as JS variable.
     * @param str the string to escape.
     * @return the escaped String.
     */
    public static String escapeJS(String str)
    {
        return StringEscapeUtils.escapeEcmaScript(str);
    }
    
    /**
     * Escape the given URL, meant to be used in eg. a background-image CSS style
     * @param url the URL to escape
     * @return the escaped URL
     */
    public static String escapeURLforCSS(String url)
    {
        Map<CharSequence, CharSequence> escapeMap = Map.of("'", "\\'",
                                                           "\"", "\\\"",
                                                           "(", "\\(",
                                                           ")", "\\)");
        
        return new LookupTranslator(escapeMap).translate(url);
    }
    
    /**
     * Split the text.
     * @param textToSplit the text to split.
     * @param tokenizers the tokenizer characters.
     * @param startIndex the minimum number of characters of the result string
     * @return the split text.
     */
    public static String splitText(String textToSplit, String tokenizers, int startIndex)
    {
        String tokenizableText = textToSplit.substring(startIndex != 0 ? startIndex - 1 : 0, textToSplit.length());
        
        int tokenPlace = StringUtils.indexOfAny(tokenizableText, tokenizers);
        
        if (tokenPlace == -1)
        {
            return textToSplit;
        }
        else
        {
            return textToSplit.substring(0, startIndex - 1 + tokenPlace);
        }
    }
    
    /**
     * Split the text.
     * @param textToSplit the text to split.
     * @param tokenizers the tokenizer characters.
     * @param maxCharacters the maximum number of characters of the result string
     * @param currentCharactersNumber the current character number.
     * @return the split text.
     */
    @Deprecated
    public static String splitText(String textToSplit, String tokenizers, int maxCharacters, int currentCharactersNumber)
    {
        int tokenStartIndex = maxCharacters - currentCharactersNumber - 1;
        String tokenizableText = textToSplit.substring(tokenStartIndex, textToSplit.length());
        
        int tokenPlace = StringUtils.indexOfAny(tokenizableText, tokenizers);
        
        if (tokenPlace == -1)
        {
            return textToSplit;
        }
        else
        {
            return textToSplit.substring(0, tokenStartIndex + tokenPlace);
        }
    }
    
    /**
     * Get the versions of the application.
     * Default VersionsHandler impl will return Ametys and Application versions
     * @return The versions &lt;Version&gt;&lt;Version&gt;&lt;Name&gt;X&lt;/Name&gt;&lt;Version&gt;X&lt;/Version&gt;&lt;Date&gt;X&lt;/Date&gt;&lt;/Version&gt;&lt;/Versions&gt; (empty tags are removed)
     */
    public static Node versions()
    {
        Map<String, Object> versionsMap = new HashMap<>();

        List<Object> versionList = new ArrayList<>();

        for (Version version : _versionHandler.getVersions())
        {
            Map<String, Object> versionMap = new HashMap<>();

            String componentName = version.getName();
            String componentVersion = version.getVersion();
            String componentDate =  DateUtils.dateToString(version.getDate());
            
            if (StringUtils.isNotEmpty(componentName))
            {
                versionMap.put("Name", componentName);
            }
            if (StringUtils.isNotEmpty(componentVersion))
            {
                versionMap.put("Version", componentVersion);
            }
            if (StringUtils.isNotEmpty(componentDate))
            {
                versionMap.put("Date", componentDate);
            }
            
            versionList.add(versionMap);
        }
        
        versionsMap.put("Component", versionList);

        return new MapElement("Versions", versionsMap);
    }
    
    /**
     * Get the current mode of the application for the current user.
     * @return True if the application is in developer mode, false if in production mode.
     */
    public static boolean isDeveloperMode()
    {
        Request request = ContextHelper.getRequest(_context);
        
        DEVMODE developerMode = DevMode.getDeveloperMode(request);
        return developerMode == DEVMODE.DEVELOPMENT 
                || developerMode == DEVMODE.SUPER_DEVELOPPMENT;
    }
    
    /**
     * Return the user
     * @return The current connected user object or null
     * @throws SAXException if a problem occured while getting the user
     */
    public static Node user() throws SAXException
    {
        UserIdentity userIdentity = _currentUserProvider.getUser();
        if (userIdentity != null)
        {
            return user(userIdentity.getLogin(), userIdentity.getPopulationId());
        }
        
        return null;
    }
    
    /**
     * Return the given user
     * @param userIdentity the concerned user's login + population
     * @return The informations about the given user
     * @throws SAXException If an error occurred while saxing the user
     */
    public static Node user(String userIdentity) throws SAXException
    {
        UserIdentity userIdentityObject = UserIdentity.stringToUserIdentity(userIdentity);
        if (userIdentityObject == null)
        {
            return null;
        }
        else
        {
            return user(userIdentityObject.getLogin(), userIdentityObject.getPopulationId());
        }
    }
    
    /**
     * Return the given user
     * @param login the concerned user's login
     * @param populationId the concerned user's population id
     * @return The informations about the given user
     * @throws SAXException If an error occurred while saxing the user
     */
    public static Node user(String login, String populationId) throws SAXException
    {
        DOMBuilder domBuilder = new DOMBuilder();
        
        UserIdentity userIdentity = new UserIdentity(login, populationId);
        _userHelper.saxUserIdentity(userIdentity, domBuilder);
        
        return domBuilder.getDocument();
    }
    
    /**
     * Return the given user
     * @param email the concerned user's email
     * @param populationId the concerned user's population id
     * @return The informations about the given user
     * @throws SAXException If an error occurred while saxing the user
     */
    public static Node userByMail(String email, String populationId) throws SAXException
    {
        try
        {
            User user = _userManager.getUserByEmail(populationId, email);
            if (user != null)
            {
                return user(UserIdentity.userIdentityToString(user.getIdentity()));
            }
        }
        catch (NotUniqueUserException e)
        {
            return null;
        }
        return null;
    }
    
    /**
     * Returns the list of the current user's groups.
     * @return the list of the current user's groups. Can be null if there is no connected user.
     */
    public static NodeList groups()
    {
        UserIdentity userIdentity = _currentUserProvider.getUser();
        return userIdentity != null ? groups(userIdentity.getLogin(), userIdentity.getPopulationId()) : null;
    }
    
    /**
     * Returns the of the given user's group.
     * @param userIdentity the concerned user's login + population
     * @return the of the given user's group.
     */
    public static NodeList groups(String userIdentity)
    {
        UserIdentity userIdentityObject = UserIdentity.stringToUserIdentity(userIdentity);
        if (userIdentityObject == null)
        {
            return null;
        }
        else
        {
            return groups(userIdentityObject.getLogin(), userIdentityObject.getPopulationId());
        }
    }
    
    /**
     * Returns the of the given user's group.
     * @param login the concerned user's login.
     * @param populationId the concerned user's population.
     * @return the of the given user's group.
     */
    public static NodeList groups(String login, String populationId)
    {
        ArrayList<Node> groups = new ArrayList<>();
        
        Set<GroupIdentity> userGroups = _groupManager.getUserGroups(new UserIdentity(login, populationId));
        for (GroupIdentity groupId : userGroups)
        {
            Group group = _groupManager.getGroup(groupId);
            if (group != null)
            {
                Map<String, String> attributes = new HashMap<>();
                attributes.put("name", groupId.getId());
                attributes.put("directory", groupId.getDirectoryId());
                groups.add(new StringElement("group", attributes, group.getLabel()));
            }
        }
        
        return new AmetysNodeList(groups);
    }

    /**
     * Parse a JSON string as a Map and return the value with the desired key
     * @param jsonString the JSON representation of the object.
     * @param key name of the value to return
     * @return the value as a String, or empty string if an error occurred or the key was not found.
     */
    public static String getValueFromJsonObject(String jsonString, String key)
    {
        try
        {
            Map<String, Object> jsonMap = _jsonUtils.convertJsonToMap(jsonString);
            if (jsonMap.containsKey(key))
            {
                Object value = jsonMap.get(key);
                if (value instanceof Map || value instanceof Collection)
                {
                    _LOGGER.warn("Unable to get string value for key '{}' from json object {}: the value can not be a map nor collection", key, jsonString);
                }
                else if (value instanceof Date)
                {
                    return DateUtils.dateToString((Date) value);
                }
                else
                {
                    return value.toString();
                }
            }
        }
        catch (Exception e)
        {
            _LOGGER.warn("Unable to parse json object {}", jsonString, e);
        }
        
        return "";
    }
    
    /**
     * Determines if the current logged user has right on a String context
     * @param rightId The id of right
     * @param objectCtx the context. Can be null to search on any context.
     * @return true if the current user is allowed, false otherwise
     */
    public static boolean hasRight(String rightId, String objectCtx)
    {
        return _rightManager.currentUserHasRight(rightId, objectCtx) == RightResult.RIGHT_ALLOW;
    }

    /**
     * Check if a given plugin is enabled
     * @param plugin the plugin's name.
     * @return the plugin is enabled
     */
    public static boolean isPluginEnabled(String plugin)
    {
        return PluginsManager.getInstance().getPluginNames().contains(plugin);
    }
    
    /**
     * Computes the base 64 representation of the image at the specified path in the given plugin.
     * @param plugin the plugin's name.
     * @param path the resource path.
     * @return the base 64 encoding for the given resource.
     * @throws IOException if an error occurs when trying to get the file
     * @throws MalformedURLException if the url is invalid
     */
    public static String pluginImageBase64(String plugin, String path) throws MalformedURLException, IOException
    {
        Source source = null;
        try
        {
            source = _sourceResolver.resolveURI("plugin:" + plugin + "://resources/" + path);
            return _getResourceBase64(source);
        }
        finally
        {
            if (source != null)
            {
                _sourceResolver.release(source);
            }
        }
    }
    
    /**
     * Get the base 64 encoding for the given source
     * @param source the source 
     * @return the base 64 encoding of the source
     */
    protected static String _getResourceBase64(Source source)
    {
        if (source.exists())
        {
            
            try (InputStream dataIs = source.getInputStream())
            {
                return ImageResolverHelper.resolveImageAsBase64(dataIs, source.getMimeType(), 0, 0, 0, 0);
            }
            catch (Exception e)
            {
                throw new IllegalStateException(e);
            }
        }

        return "";
    }
}
