/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.util.Collection;

import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.AbstractCacheManager.CacheType;
import org.ametys.core.cache.Cache;
import org.ametys.runtime.i18n.I18nizableText;

/**
 * Trait for easily cache some elements, in multiple Caches if necessary.
 * <br>This is the preferred way to manage the caching of elements (over {@link CachingComponent}), as it will use
 * the {@link AbstractCacheManager CacheManager mechanism}, and as it is an interface, it will not interfere in the class hierarchy
 * of the implementing component.
 * <br>
 * <br>The methods to implement are:
 * <ul>
 * <li>{@link #getManagedCaches}: for describing the managed caches;</li>
 * <li>{@link #getCacheManager}: for binding the reference of {@link AbstractCacheManager};</li>
 * <li>{@link #isCachingEnabled}: for implementing the potential business logic to determine if the caching is enabled.</li>
 * </ul>
 * <br>
 * <br>The methods to call are:
 * <ul>
 * <li>{@link #createCaches}: for creating the managed caches;</li>
 * <li>{@link #removeCaches}: for removing the managed caches;</li>
 * <li>{@link #getCache}: for getting a reference to a managed cache.</li>
 * </ul>
 */
public interface Cacheable
{
    /**
     * A configuration for describing to {@link Cacheable} interface the managed {@link Cache caches} to {@link AbstractCacheManager#createMemoryCache create}.
     * <br>It simply composes a cache id, a label and a description.
     */
    public static final class SingleCacheConfiguration
    {
        private String _id;
        private I18nizableText _label;
        private I18nizableText _description;
        
        private SingleCacheConfiguration(String id, I18nizableText label, I18nizableText description)
        {
            _id = id;
            _label = label;
            _description = description;
        }
        
        /**
         * Builds a {@link SingleCacheConfiguration}
         * @param id The id of the cache
         * @param label The label of the cache
         * @param description The description of the cache
         * @return A {@link SingleCacheConfiguration}
         */
        public static SingleCacheConfiguration of(String id, I18nizableText label, I18nizableText description)
        {
            return new SingleCacheConfiguration(id, label, description);
        }
    }
    
    /**
     * Gets the managed caches.
     * <br>This <b>is</b> meant to be implemented in order to describe the managed caches and automatically
     * create and remove the corresponding caches in {@link #createCaches} and {@link #removeCaches} default methods.
     * <br>This <b>is not</b> meant to be called manually.
     * @return A collection of {@link SingleCacheConfiguration}s to manage
     */
    Collection<SingleCacheConfiguration> getManagedCaches();
    
    /**
     * Returns <code>true</code> if the cache is enabled
     * <br>The default implementation returns <code>true</code>.
     * <br>You <b>can</b> override this method to modify the behavior.
     * @return <code>true</code> if the cache is enabled
     */
    default boolean isCachingEnabled()
    {
        return true;
    }
    
    /**
     * Returns the instance of the implementation of {@link AbstractCacheManager} to use.
     * <br>This <b>is not</b> meant to be called manually.
     * @return The {@link AbstractCacheManager} to bind
     */
    AbstractCacheManager getCacheManager();
    
    /**
     * Gets the managed cache with the given id.
     * <br>It is advised to create its own business methods which call this method in order to retrieve the caches.
     * <br>For instance, if the implementing class {@link #getManagedCaches declares} two caches,
     * you can create the two methods:
     * <ul>
     * <li><code>Cache&lt;String, Double&gt; getCachePriceByProductId()</code></li>
     * <li>and <code>Cache&lt;Integer, Boolean&gt; getCacheAvailabilityByProductIndex()</code></li>
     * </ul>
     * which simply call this method, so as to have methods to retrieve the caches with more semantic.
     * @param <K> The type of the keys in the cache
     * @param <V> The type of the values in the cache
     * @param cacheId The cache id
     * @return The managed {@link Cache}
     */
    default <K, V> Cache<K, V> getCache(String cacheId)
    {
        return getCacheManager().get(cacheId);
    }
    
    /**
     * Creates the managed caches.
     * <br>This <b>is</b> meant to be called manually, at the beginning of the lifecycle of the implementing class,
     * in order to create and initialize the corresponding caches.
     * <br>This <b>is not</b> meant to be overridden.
     */
    default void createCaches()
    {
        if (isCachingEnabled())
        {
            for (SingleCacheConfiguration cacheConfig : getManagedCaches())
            {
                _createCache(cacheConfig._id, cacheConfig._label, cacheConfig._description);
            }
        }
    }
    
    private void _createCache(String id, I18nizableText label, I18nizableText description)
    {
        getCacheManager().createMemoryCache(
                id, 
                label, 
                description, 
                hasComputableSize(),
                null);
    }
    
    /**
     * Removes the managed caches.
     * <br>This <b>is</b> meant to be called manually, at the end of the lifecycle of the implementing class,
     * in order to remove the corresponding caches.
     * <br>This <b>is not</b> meant to be overridden.
     */
    default void removeCaches()
    {
        if (isCachingEnabled())
        {
            for (SingleCacheConfiguration cacheConfig : getManagedCaches())
            {
                getCacheManager().removeCache(cacheConfig._id, CacheType.MEMORY);
            }
        }
    }
    
    /**
     * Determines if the cache has a computable size. This operation can be very slow.
     * @return <code>true</code> if we can compute the size on the cache.
     */
    default boolean hasComputableSize()
    {
        return true;
    }
}
