/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.time.Duration;
import java.util.concurrent.ExecutionException;
import java.util.function.Supplier;

import com.google.common.base.Ticker;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;

/**
 * A cached value. It can be constructed with {@link #withInitial} with a {@link Supplier} to pass in order to compute the value to cache on the first call of {@link #get()}.
 * <br>It can also be constructed with {@link #withExpiryDuration} with the {@link Supplier} and an expiry {@link Duration} to automatically force the reload of the value.
 * <br>The supplier is ensured to be called once (until {@link #uncache} is called, or after the expiry duration has elapsed since the last computing), even with multiple callers at the same time.
 * <br>The returned value by the supplier <b>must not be null</b>
 * @param <T> The type of the cached value
 */
public class CachedValue<T>
{
    private static final Object __ALL_CACHES_ONLY_KEY = new Object();
    
    private LoadingCache<Object, T> _loadingCache;
    
    CachedValue(Supplier<T> supplier, Duration expiryDuration, Ticker ticker)
    {
        LoadingCache<Object, T> loadingCache = _cacheBuilder(expiryDuration, ticker)
                .build(_cacheLoader(supplier));
        _loadingCache = loadingCache;
    }
    
    /**
     * Creates a cached value. The value will be computed by invoking
     * the given supplier when the {@link #get()} method will be
     * called the first time.
     * @param <T> The type of the cached value
     * @param supplier The supplier which will be invoked only once in order to compute the value to cache. The value <b>must not be null</b>
     * @return The {@link CachedValue}
     */
    public static <T> CachedValue<T> withInitial(Supplier<T> supplier)
    {
        return new CachedValue<>(supplier, null, null);
    }
    
    /**
     * Creates a cached value. The value will be computed by invoking
     * the given supplier when the {@link #get()} method will be
     * called the very first time, and every first time after the given {@link Duration} is expired.
     * @param <T> The type of the cached value
     * @param supplier The supplier which will be invoked only once in order to compute the value to cache. The value <b>must not be null</b>
     * @param expiryDuration The expiry duration
     * @return The {@link CachedValue}
     */
    public static <T> CachedValue<T> withExpiryDuration(Supplier<T> supplier, Duration expiryDuration)
    {
        return new CachedValue<>(supplier, expiryDuration, null);
    }
    
    CacheBuilder<Object, Object> _cacheBuilder(Duration expiryDuration, Ticker ticker)
    {
        CacheBuilder<Object, Object> cacheBuilder = CacheBuilder.newBuilder()
                .initialCapacity(1)
                // key will always be __ALL_CACHES_ONLY_KEY
                .weakKeys();
        
        if (expiryDuration != null)
        {
            cacheBuilder.expireAfterWrite(expiryDuration);
        }
        if (ticker != null)
        {
            cacheBuilder.ticker(ticker);
        }
        
        return cacheBuilder;
    }
    
    private static <T> CacheLoader<Object, T> _cacheLoader(Supplier<T> supplier)
    {
        return new CacheLoader<>()
        {
            @Override
            public T load(Object key) throws Exception
            {
                // key should always be __ALL_CACHES_ONLY_KEY, and is useless
                T value = supplier.get();
                if (value == null)
                {
                    throw new IllegalArgumentException("The given supplier cannot return a null value.");
                }
                return value;
            }
        };
    }
    
    /**
     * Returns the value. If it is the first time it is called, then it is computed
     * by invoking the supplier function. Otherwise, the cached value is returned.
     * @return the value
     */
    public T get()
    {
        try
        {
            return _loadingCache.get(__ALL_CACHES_ONLY_KEY);
        }
        catch (ExecutionException e)
        {
            throw new IllegalStateException("An unexpected error occured when loading the value from the supplier function.", e);
        }
    }
    
    /**
     * Uncache the value. Mostly to force to re-compute the value. 
     */
    public void uncache()
    {
        _loadingCache.invalidate(__ALL_CACHES_ONLY_KEY);
    }
}
