/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.util.HashMap;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.activity.Initializable;

import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Simple memory cache with a thread clearing the cache every day.
 * @param <K> the type of keys for caching objects by this component.
 * @param <V> the type of objects cached by this component.
 */
public class CachingComponent<K, V> extends AbstractLogEnabled implements Initializable, Disposable
{
    private Map<K, V> _objects = new HashMap<>();
    private Timer _timer;
    
    public void initialize() throws Exception
    {
        if (isCacheEnabled())
        {
            _timer = new Timer("CachingComponent", true);
            
            long period = 1000 * 60 * 60 * 24; // one day
            _timer.scheduleAtFixedRate(new TimerTask()
            {
                @Override
                public void run()
                {
                    clearCache();
                }
            }, period, period);
        }
    }
    
    public void dispose()
    {
        if (_timer != null)
        {
            _timer.cancel();
        }
    }
    
    /**
     * Returns an object from the cache, correspondong to the specified key, or null if none.
     * @param key the object's key.
     * @return the object from cache.
     */
    protected V getObjectFromCache(K key)
    {
        V object = _objects.get(key);
        
        getLogger().debug("Getting object {} from cache for key {}", object, key);
        
        return object;
    }
    
    /**
     * Adds a key/object pair in the cache.
     * @param key the object's key.
     * @param object the object to be cached.
     */
    protected void addObjectInCache(K key, V object)
    {
        getLogger().debug("Adding object {} in cache for key {}", object, key);
        
        _objects.put(key, object);
    }
    
    /**
     * Removes a key/object pair in the cache.
     * @param key the object's key.
     */
    protected void removeObjectFromCache(K key)
    {
        getLogger().debug("Removing object in cache for key {}", key);
        
        _objects.remove(key);
    }
    
    /**
     * Removes all entries from the cache.
     */
    protected void clearCache()
    {
        getLogger().debug("Clearing cache");
        
        _objects.clear();
    }
    
    /**
     * Returns true if the cache is enabled.
     * @return true if the cache is enabled.
     */
    protected boolean isCacheEnabled()
    {
        return true;
    }
}
