/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.nio.charset.StandardCharsets;
import java.util.function.Predicate;

/**
 * Utility class for encoding file names and paths.<br>
 * These methods are used to provide safe names to be stored on disk.<br>
 * The encode* methods %-encode all reserved chars, plus the '%' char itself.<br>
 * The filter* methods replace reserved chars with the '_' char.<br><br>
 * Reserved chars are <pre>\ / ? : * | " &lt; &gt;</pre>
 */
public final class FilenameUtils
{
    /**
     * Character used to replace reserved chars in the filter* methods.
     */
    public static final char REPLACEMENT_CHAR = '_';
    
    private static Predicate<Character> _isReserved = c -> ';' == c || '\\' == c || '/' == c || '?' == c || ':' == c || '*' == c || '|' == c || '"' == c || '<' == c || '>' == c;
    
    private FilenameUtils()
    {
        // empty
    }
    
    /**
     * %-encode a file path to be stored on disk.
     * @param path the file path to encode
     * @return the encoded file path.
     */
    public static String encodePath(String path)
    {
        return _encodePath(path, Predicate.not(_isReserved.or(c -> '%' == c)).or(c -> '/' == c), true);
    }
    
    /**
     * %-encode a file name to be stored on disk.
     * @param name the file name to encode
     * @return the encoded file name.
     */
    public static String encodeName(String name)
    {
        return _encodePath(name, Predicate.not(_isReserved.or(c -> '%' == c)), true);
    }
    
    /**
     * Replace reserved characters from a file path with a _.
     * @param path the file path to transform
     * @return the filtered file path
     */
    public static String filterPath(String path)
    {
        return _encodePath(path, Predicate.not(_isReserved).or(c -> '/' == c), false);
    }
    
    /**
     * Replace reserved characters from a file name with a _.
     * @param name the file name to transform
     * @return the filtered file name
     */
    public static String filterName(String name)
    {
        return _encodePath(name, Predicate.not(_isReserved), false);
    }
    
    @SuppressWarnings("all")
    // very similar to URIUtils#encode but this one allows UTF-8 chars and not only ASCII bytes
    private static String _encodePath(String path, Predicate<Character> charactersToKeep, boolean encode)
    {
        StringBuilder result = new StringBuilder();
        
        char[] chars = path.toCharArray();
        
        for (char c : chars)
        {
            if (charactersToKeep.test(c)) 
            {
                result.append(c);
            }
            else if (encode)
            {
                byte[] bytes = String.valueOf(c).getBytes(StandardCharsets.UTF_8);
                for (byte b : bytes) 
                {
                    if (b < 0) 
                    {
                        b += 256;
                    }
                    
                    result.append('%');
                    char hex1 = Character.toUpperCase(Character.forDigit((b >> 4) & 0xF, 16));
                    char hex2 = Character.toUpperCase(Character.forDigit(b & 0xF, 16));
                    result.append(hex1);
                    result.append(hex2);
                }
            }
            else
            {
                result.append(REPLACEMENT_CHAR);
            }
        }
        
        return result.toString();
    }
    
    /**
     * Decodes an URI-encoded String.
     * @param source the String to decode.
     * @return the decoded String.
     */
    @SuppressWarnings("all")
    public static String decode(String source)
    {
        return URIUtils.decode(source);
    }
}
