/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.SocketTimeoutException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.regex.Pattern;

import javax.net.ssl.SSLHandshakeException;

import org.apache.avalon.framework.component.Component;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.core.ui.Callable;

/**
 * Utility class for HTTP urls.
 */
public class HttpUrlUtils implements Component
{
    /** The Avalon role */
    public static final String ROLE = HttpUrlUtils.class.getName();
    
    /** Regexp for HTTP url */
    public static final Pattern HTTP_URL_VALIDATOR = Pattern.compile("^(https?:\\/\\/.+)?$");
    
    /** The status of HTTP check */
    public static enum HttpCheck
    {
        /** All right */
        SUCCESS,
        /** Server error. */
        SERVER_ERROR,
        /** URL not found.*/
        NOT_FOUND,
        /** Unauthorized. */
        UNAUTHORIZED,
        /** Timeout (too long) */
        TIMEOUT,
        /** A redirect occurs */
        REDIRECT,
        /** Security level error */
        SECURITY_LEVEL_ERROR,
        /** No HTTP url */
        NOT_HTTP
        
    }
    
    /**
     * HTTP Status-Code 307: Temporay Redirect
     */
    public static final int HTTP_REDIRECT_TEMP = 307;
    
    /**
     * HTTP Status-Code 308: Temporay Redirect
     */
    public static final int HTTP_REDIRECT_PERM = 308;
    
    
    private static Logger __logger = LoggerFactory.getLogger(HttpUrlUtils.class);
    
    /**
     * Prepare the connection to the remote url
     * @param url The url 
     * @param userAgent The user agent. Can be null.
     * @param method The method for teh URL request. Can be null.
     * @param timeout The connection timeout in milliseconds. Set to -1 to not set a timeout.
     * @param readTimeOut The read timeout in milliseconds. Set to -1 to not set a timeout.
     * @param followRedirects Sets to true to follow HTTP redirects 
     * @return The URL connection
     * @throws MalformedURLException If the given url is a malformed URL
     * @throws IOException if an I/O exception occurs
     */
    public static HttpURLConnection prepareConnection(String url, String userAgent, String method, int timeout, int readTimeOut, boolean followRedirects) throws MalformedURLException, IOException
    {
        return prepareConnection(url, userAgent, method, timeout, readTimeOut, followRedirects, Collections.EMPTY_MAP);
    }
    
    /**
     * Prepare the connection to the remote url
     * @param httpUrl The HTTP url 
     * @param userAgent The user agent. Can be null.
     * @param method The method for teh URL request. Can be null.
     * @param timeout The connection timeout in milliseconds. Set to -1 to not set a timeout.
     * @param readTimeOut The read timeout in milliseconds. Set to -1 to not set a timeout.
     * @param followRedirects Sets to true to follow HTTP redirects 
     * @param requestHeaders The request headers.
     * @return The URL connection
     * @throws MalformedURLException If the given url is a malformed URL
     * @throws IOException if an I/O exception occurs
     */
    public static HttpURLConnection prepareConnection(String httpUrl, String userAgent, String method, int timeout, int readTimeOut, boolean followRedirects, Map<String, String> requestHeaders) throws MalformedURLException, IOException
    {
        return prepareConnection(new URL(httpUrl), userAgent, method, timeout, readTimeOut, followRedirects, requestHeaders);
    }
    
    /**
     * Prepare the connection to the remote url
     * @param url The url
     * @param userAgent The user agent. Can be null.
     * @param method The method for teh URL request. Can be null.
     * @param timeout The connection timeout in milliseconds. Set to -1 to not set a timeout.
     * @param readTimeOut The read timeout in milliseconds. Set to -1 to not set a timeout.
     * @param followRedirects Sets to true to follow HTTP redirects 
     * @param requestHeaders The request headers.
     * @return The URL connection
     * @throws MalformedURLException If the given url is a malformed URL
     * @throws IOException if an I/O exception occurs
     */
    public static HttpURLConnection prepareConnection(URL url, String userAgent, String method, int timeout, int readTimeOut, boolean followRedirects, Map<String, String> requestHeaders) throws MalformedURLException, IOException
    {
        HttpURLConnection httpUrlConnection = (HttpURLConnection) url.openConnection();
        
        if (userAgent != null)
        {
            httpUrlConnection.setRequestProperty("User-Agent", userAgent);
        }
        if (method != null)
        {
            httpUrlConnection.setRequestMethod(method);
        }
        if (timeout != -1)
        {
            httpUrlConnection.setConnectTimeout(timeout);
        }
        
        if (readTimeOut != -1)
        {
            httpUrlConnection.setReadTimeout(readTimeOut);
        }
        
        if (followRedirects)
        {
            httpUrlConnection.setInstanceFollowRedirects(true);
        }
        
        for (Entry<String, String> headers : requestHeaders.entrySet())
        {
            httpUrlConnection.setRequestProperty(headers.getKey(), headers.getValue());
        }
        
        return httpUrlConnection;
    }
    
    /**
     * Check the HTTP url
     * @param httpUrl The HTTP url to test 
     * @param userAgent The user agent. Can be null.
     * @param method The method for teh URL request. Can be null.
     * @param timeout The connection timeout in milliseconds. Set to -1 to not set a timeout.
     * @param readTimeOut The read timeout in milliseconds. Set to -1 to not set a timeout.
     * @param followRedirects Sets to true to follow HTTP redirects 
     * @return The URL connection
     */
    public static HttpCheckReport checkHttpUrl(String httpUrl, String userAgent, String method, int timeout, int readTimeOut, boolean followRedirects)
    {
        return checkHttpUrl(httpUrl, userAgent, method, timeout, readTimeOut, followRedirects, Collections.EMPTY_MAP);
    }
    
    /**
     * Check the HTTP url
     * @param httpUrl The url to test 
     * @param userAgent The user agent. Can be null.
     * @param method The method for teh URL request. Can be null.
     * @param timeout The connection timeout in milliseconds. Set to -1 to not set a timeout.
     * @param readTimeOut The read timeout in milliseconds. Set to -1 to not set a timeout.
     * @param followRedirects Sets to true to follow HTTP redirects 
     * @param requestHeaders The request headers.
     * @return The URL connection
     */
    public static HttpCheckReport checkHttpUrl(String httpUrl, String userAgent, String method, int timeout, int readTimeOut, boolean followRedirects, Map<String, String> requestHeaders)
    {
        if (!HTTP_URL_VALIDATOR.matcher(StringUtils.defaultIfEmpty(httpUrl, "")).matches())
        {
            __logger.debug("Url '{}' is not a valid HTTP url", httpUrl);
            return new HttpCheckReport(HttpCheck.NOT_HTTP);
        }
        
        try
        {
            return checkHttpUrl(new URL(httpUrl), userAgent, method, timeout, readTimeOut, followRedirects, requestHeaders);
        }
        catch (MalformedURLException e) 
        {
            __logger.debug("Unable to parse '{}' as a HTTP url", httpUrl, e);
            return new HttpCheckReport(HttpCheck.NOT_HTTP);
        }
    }
    
    /**
     * Check the HTTP url
     * @param url The url to test 
     * @param userAgent The user agent. Can be null.
     * @param method The method for teh URL request. Can be null.
     * @param timeout The connection timeout in milliseconds. Set to -1 to not set a timeout.
     * @param readTimeOut The read timeout in milliseconds. Set to -1 to not set a timeout.
     * @param followRedirects Sets to true to follow HTTP redirects 
     * @param requestHeaders The request headers.
     * @return The URL connection
     */
    public static HttpCheckReport checkHttpUrl(URL url, String userAgent, String method, int timeout, int readTimeOut, boolean followRedirects, Map<String, String> requestHeaders)
    {
        return _checkHttpUrl(url, userAgent, method, timeout, readTimeOut, followRedirects, requestHeaders, new ArrayList<>(), 5);
    }
    
    private static HttpCheckReport _checkHttpUrl(URL url, String userAgent, String method, int timeout, int readTimeOut, boolean followRedirects, Map<String, String> requestHeaders, List<String> visitedUrl, int maxRedirects)
    {
        try
        {
            HttpURLConnection connection = prepareConnection(url, userAgent, method, timeout, readTimeOut, followRedirects, requestHeaders);
                     
            if (connection.getResponseCode() == HttpURLConnection.HTTP_OK)
            {
                __logger.debug("Check of URL '{}' successed", url);
                return new HttpCheckReport(HttpCheck.SUCCESS);
            }
            else if (connection.getResponseCode() == HttpURLConnection.HTTP_MOVED_TEMP /* 302 */ 
                    || connection.getResponseCode() == HttpURLConnection.HTTP_MOVED_PERM /* 301 */ 
                    || connection.getResponseCode() == HttpURLConnection.HTTP_SEE_OTHER /* 303 */
                    || connection.getResponseCode() == HTTP_REDIRECT_TEMP /* 307 */
                    || connection.getResponseCode() == HTTP_REDIRECT_PERM /* 308 */)
            {
                if (followRedirects)
                {
                    // Redirection between HTTP and HTTPS URLs is not followed (do it manually)
                    if (maxRedirects < 0 || visitedUrl.contains(url.toExternalForm()))
                    {
                        // The maximum number of redirects has been reached or the url was already been visited when following redirects
                        throw new IOException("Exceed max number of redirects");
                    }
                    
                    visitedUrl.add(url.toExternalForm());
                    
                    URL base = connection.getURL();
                    String location = connection.getHeaderField("Location");
                    URL redirect = new URL(base, location);  // Deal with relative URLs
                    
                    return _checkHttpUrl(redirect, userAgent, method, timeout, readTimeOut, true, requestHeaders, visitedUrl, maxRedirects - 1);
                }
                else
                {
                    return new HttpCheckReport(HttpCheck.REDIRECT);
                }
            }
            else
            {
                __logger.debug("Check of URL '{}' returns the status code {}", url, connection.getResponseCode());
                
                int responseCode = connection.getResponseCode();
                
                switch (responseCode)
                {
                    case HttpURLConnection.HTTP_NOT_FOUND:
                        return new HttpCheckReport(HttpCheck.NOT_FOUND);
                    case HttpURLConnection.HTTP_FORBIDDEN:
                    case HttpURLConnection.HTTP_UNAUTHORIZED:
                        return new HttpCheckReport(HttpCheck.UNAUTHORIZED);
                    case HttpURLConnection.HTTP_INTERNAL_ERROR:
                    default:
                        return new HttpCheckReport(HttpCheck.SERVER_ERROR, Optional.ofNullable(connection.getResponseMessage()));
                }
            }
        }
        catch (SSLHandshakeException e)
        {
            __logger.debug("Certificate error for URL '{}'", url, e);
            return new HttpCheckReport(HttpCheck.SECURITY_LEVEL_ERROR, Optional.ofNullable(e.getMessage()));
        }
        catch (SocketTimeoutException e)
        {
            __logger.debug("Aborting test for URL '{}' because too long", url, e);
            return new HttpCheckReport(HttpCheck.TIMEOUT);
        }
        catch (UnknownHostException e)
        {
            __logger.debug("Unknown host for URL '{}'", url, e);
            return new HttpCheckReport(HttpCheck.NOT_FOUND);
        }
        catch (IOException e)
        {
            __logger.debug("Cannot test URL '{}'", url, e);
            return new HttpCheckReport(HttpCheck.SERVER_ERROR, Optional.ofNullable(e.getMessage()));
        }
    }
    
    /**
     * Method to check a HTTP url from client side. 
     * The HTTP redirects will be followed.
     * @param httpUrl the http url to check
     * @return the result of the check
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> checkHttpUrl(String httpUrl)
    {
        Map<String, Object> result = new HashMap<>();
        
        HttpCheckReport report = checkHttpUrl(httpUrl, null, null, 2000, 2000, true);
        result.put("success", HttpCheck.SUCCESS.equals(report.status()));
        if (report.message().isPresent())
        {
            result.put("message", report.message().get());
        }
        result.put("checkResult", report.status().name());
        
        return result;
    }
    
    /**
     * Represent the result of a check and a potential message
     * @param status the result status
     * @param message the message if needed
     */
    public record HttpCheckReport(HttpCheck status, Optional<String> message)
    {
        /**
         * Create a report with the provided status and no message.
         * @param result the result status
         */
        public HttpCheckReport(HttpCheck result)
        {
            this(result, Optional.empty());
        }
    }
}
