/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.io.IOException;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.LogEnabled;
import org.apache.avalon.framework.logger.Logger;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;

import org.ametys.runtime.i18n.I18nizable;
import org.ametys.runtime.i18n.I18nizableText;

import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;

/**
 * Serializer for {@link I18nizable} objects.
 * Returns the translated message for JSON value.
 */
public class I18nizableSerializer extends StdSerializer<I18nizable> implements Component, Serviceable, LogEnabled, Contextualizable
{
    /** Request attribut for locale */
    public static final String REQUEST_ATTR_LOCALE = "locale";
    
    /** The Avalon Role */
    public static final String ROLE = I18nizableSerializer.class.getName();
    
    private I18nUtils _i18nUtils;
    private Logger _logger;

    private Context _context;
    
    /**
     * Constructor
     */
    public I18nizableSerializer()
    {
        super(I18nizable.class);
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
    }

    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void enableLogging(Logger logger)
    {
        _logger = logger;
    }
    
    @Override
    public void serialize(I18nizable i18nizable, JsonGenerator jgen, SerializerProvider provider) throws IOException, JsonGenerationException
    {
        if (i18nizable instanceof I18nizableText)
        {
            I18nizableText value = (I18nizableText) i18nizable;
            if (value.isI18n())
            {
                String msg = _i18nUtils.translate(value, _getLangCode());
                if (msg == null)
                {
                    if (_logger.isWarnEnabled())
                    {
                        _logger.warn("Translation not found for key " + value.getKey() + " in catalogue " + value.getCatalogue());
                    }
                    
                    jgen.writeString(value.getCatalogue() + ':' + value.getKey());
                }
                else
                {
                    jgen.writeString(msg);
                }
            }
            else
            {
                jgen.writeString(value.getLabel());
            }
        }
        else
        {
            jgen.writeString(_i18nUtils.translate(i18nizable, _getLangCode()));
        }
    }
    
    private String _getLangCode()
    {
        Request request = ContextHelper.getRequest(_context);
        return request != null ? (String) request.getAttribute(REQUEST_ATTR_LOCALE) : null;
    }
}
