/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang.StringUtils;

/**
 * Class providing helper methods to resolve images.
 */
public final class ImageResolverHelper
{
    
    private ImageResolverHelper()
    {
        // Hides the default constructor.
    }

    /**
     * Get an image's bytes encoded as base64, optionally resized. 
     * @param dataIs an input stream on the image bytes.
     * @param mimeType the image mime type.
     * @param height the specified height. Ignored if negative.
     * @param width the specified width. Ignored if negative.
     * @param maxHeight the maximum image height. Ignored if height or width is specified.
     * @param maxWidth the maximum image width. Ignored if height or width is specified.
     * @return the image bytes encoded as base64.
     * @throws IOException if an error occurs.
     */
    
    public static final String resolveImageAsBase64(InputStream dataIs, String mimeType, int height, int width, int maxHeight, int maxWidth) throws IOException
    {
        return resolveImageAsBase64(dataIs, mimeType, height, width, maxHeight, maxWidth, 0, 0);
    }
    
    /**
     * Get an image's bytes encoded as base64, optionally resized. 
     * @param dataIs an input stream on the image bytes.
     * @param mimeType the image mime type.
     * @param height the specified height. Ignored if 0.
     * @param width the specified width. Ignored if 0.
     * @param maxHeight the maximum image height. Ignored if height or width is specified.
     * @param maxWidth the maximum image width. Ignored if height or width is specified.
     * @param cropHeight the cropping height. Ignored if 0.
     * @param cropWidth the cropping width. Ignored if 0.
     * @return the image bytes encoded as base64.
     * @throws IOException if an error occurs.
     */
    public static final String resolveImageAsBase64(InputStream dataIs, String mimeType, int height, int width, int maxHeight, int maxWidth, int cropHeight, int cropWidth) throws IOException
    {
        if (!StringUtils.startsWith(mimeType, "image/"))
        {
            return "";
        }
        
        ByteArrayOutputStream buffer = new ByteArrayOutputStream();
        
        if (height == 0 && width == 0 && maxHeight == 0 && maxWidth == 0 && cropHeight == 0 && cropWidth == 0)
        {
            IOUtils.copy(dataIs, buffer);
        }
        else
        {
            String format = StringUtils.substringAfter(mimeType, "/");
            ImageHelper.generateThumbnail(dataIs, buffer, format, height, width, maxHeight, maxWidth, cropHeight, cropWidth);
        }
        
        String data = Base64.encodeBase64String(buffer.toByteArray());
        
        return "data:" + mimeType + ";base64," + data;
    }
}
