/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.io.File;
import java.io.IOException;
import java.net.JarURLConnection;
import java.net.URI;
import java.net.URL;
import java.net.URLConnection;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.nio.file.NoSuchFileException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.core.util.LambdaUtils.LambdaException;

/**
 * This component centralize the access to zip files
 * Beware: All opened zip files are kept opened in memory 
 */
public final class JarFSManager
{
    private static JarFSManager _instance;
    
    private Map<String, FileSystem> _openedFS = new HashMap<>();
    
    private Logger _logger = LoggerFactory.getLogger(getClass());
    
    private JarFSManager()
    {
        // private
    }

    /**
     * Get the unique instance
     * @return the unique instance
     */
    public static JarFSManager getInstance()
    {
        if (_instance == null)
        {
            _instance = new JarFSManager();
        }
        
        return _instance;
    }
    
    /**
     * Return the jar filesystem associated with the given resource
     * @param resourcePath The resource path to seek. Such as 'org/apache/ivy/logo.png'.
     * @return The JAR filesystem or null if the resource cannot be found
     * @throws IOException If an error occurred while getting the JAR file of an existing resource
     */
    public FileSystem getFileSystemByResource(String resourcePath) throws IOException
    {
        String uri = getJARFileURI(resourcePath);
        return getFileSystemByURI(uri);
    }
    
    /**
     * Return the jar filesystem associated with the given jar file
     * @param jarFile the jar file to open
     * @return The JAR filesystem or null if the file has a problem
     * @throws IOException If an error occurred while opening the JAR file
     */
    public FileSystem getFileSystemByFile(File jarFile) throws IOException
    {
        String uri = "jar:" + jarFile.getCanonicalFile().toURI().toString();
        return getFileSystemByURI(uri);
    }
    
    /**
     * Determine the JAR file holding the given resource
     * @param resourcePath The resource path to seek. Such as 'org/apache/ivy/logo.png'.
     * @return The JAR file location. Such as '/path/to/jar/ivy.jar' or null if the resource cannot be found
     * @throws IOException If an error occurred while getting the JAR file of an existing resource
     */
    public String getJARFileURI(String resourcePath) throws IOException
    {
        URL resource = getClass().getClassLoader().getResource(resourcePath);
        if (resource == null)
        {
            return null;
        }
        
        URLConnection connection = resource.openConnection();
        URL jarFileURL = ((JarURLConnection) connection).getJarFileURL();
        return "jar:" + jarFileURL.toString();
    }
    
    /**
     * Return the jar filesystem associated with the given jar file denoted by its URI
     * @param jarURI Such as "jar:file:/path/to/file.jar"
     * @return The filesystem or null if the jarURI is null
     * @throws IOException If the JAR file cannot be opened correctly
     */
    public synchronized FileSystem getFileSystemByURI(String jarURI) throws IOException
    {
        if (jarURI == null)
        {
            return null;
        }
        
        try
        {
            _logger.debug("Opening the jar at uri: {}", jarURI);
            return _openedFS.computeIfAbsent(jarURI, LambdaUtils.wrap(u -> {
                return FileSystems.newFileSystem(new URI(u), Collections.emptyMap());
            }));
        }
        catch (LambdaException e)
        {
            if (e.getCause() instanceof NoSuchFileException)
            {
                _logger.error("Could not find the jar at uri: {}", jarURI);
            }
            throw (IOException) e.getCause();
        }
    }
}
