/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.BinaryOperator;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Utility methods for {@link Map}s
 */
public final class MapUtils
{
    private static BinaryOperator<?> __DEFAULT_MERGE_FUNCTION = (value1, value2) -> value1;
    
    private MapUtils()
    {
        // Nothing
    }
    
    @SuppressWarnings("unchecked")
    private static <V> BinaryOperator<V> _defaultMergeFunction()
    {
        return (BinaryOperator<V>) __DEFAULT_MERGE_FUNCTION;
    }
    
    /**
     * Merges the entries of two maps into one map. Entries of the first one have the priority (i.e. kept if same keys) over ones of the second one.
     * @param <K> The type of the keys
     * @param <V> The type of the values
     * @param map1 The first map (not modified)
     * @param map2 The second map (not modified)
     * @return The merged map
     */
    public static <K, V> Map<K, V> merge(
            Map<? extends K, ? extends V> map1, 
            Map<? extends K, ? extends V> map2)
    {
        Stream<Map<? extends K, ? extends V>> sequentialStreamOfMaps = Stream.of(map1, map2);
        return _merge(_defaultMergeFunction(), sequentialStreamOfMaps);
    }
    
    
    /**
     * Merges the entries of several maps into one map. Entries of any input map one have the priority (i.e. kept if same keys) over ones of any following map.
     * @param <K> The type of the keys
     * @param <V> The type of the values
     * @param map1 The first map (not modified)
     * @param map2 The second map (not modified)
     * @param others The other maps (not modified)
     * @return The merged map
     */
    @SafeVarargs // Creating a stream from an array is safe
    public static <K, V> Map<K, V> merge(
            Map<? extends K, ? extends V> map1, 
            Map<? extends K, ? extends V> map2, 
            Map<? extends K, ? extends V>... others)
    {
        return merge(_defaultMergeFunction(), map1, map2, others);
    }
    
    /**
     * Merges the entries of two maps into one map.
     * @param <K> The type of the keys
     * @param <V> The type of the values
     * @param mergeFunction The merging function, to determine which value of the entry to keep when same keys are encountered.
     * @param map1 The first map (not modified)
     * @param map2 The second map (not modified)
     * @return The merged map
     */
    public static <K, V> Map<K, V> merge(
            BinaryOperator<V> mergeFunction, 
            Map<? extends K, ? extends V> map1, 
            Map<? extends K, ? extends V> map2)
    {
        Stream<Map<? extends K, ? extends V>> sequentialStreamOfMaps = Stream.of(map1, map2);
        return _merge(mergeFunction, sequentialStreamOfMaps);
    }
    
    /**
     * Merges the entries of several maps into one map. Entries of any input map one have the priority (i.e. kept if same keys) over ones of any following map.
     * @param <K> The type of the keys
     * @param <V> The type of the values
     * @param mergeFunction The merging function, to determine which value of the entry to keep when same keys are encountered.
     * @param map1 The first map (not modified)
     * @param map2 The second map (not modified)
     * @param others The other maps (not modified)
     * @return The merged map
     */
    @SafeVarargs // Creating a stream from an array is safe
    public static <K, V> Map<K, V> merge(
            BinaryOperator<V> mergeFunction, 
            Map<? extends K, ? extends V> map1, 
            Map<? extends K, ? extends V> map2, 
            Map<? extends K, ? extends V>... others)
    {
        Stream<Map<? extends K, ? extends V>> sequentialStreamOfMaps = Stream.of(
                Stream.of(map1, map2),
                Stream.of(others)
            )
            .flatMap(stream -> stream)
            .sequential();
        
        return _merge(mergeFunction, sequentialStreamOfMaps);
    }
    
    
    private static <K, V> Map<K, V> _merge(
            BinaryOperator<V> mergeFunction,
            Stream<Map<? extends K, ? extends V>> sequentialStreamOfMaps)
    {
        return sequentialStreamOfMaps
                .map(Map::entrySet)
                .map(ArrayList::new) // Convert Set into List in order to be sure entries of a given map all come before any entry of a following map
                .flatMap(List::stream)
                .collect(Collectors.toMap(
                        Map.Entry::getKey, 
                        Map.Entry::getValue,
                        mergeFunction));
    }

    /**
     * Function to merge deeply two maps. If the two values are maps, they are merged, but if not, value1 is returned.
     * This operation is done recursively.
     * @param value1 First value to merge
     * @param value2 Second value to merge
     * @return Both values merged if they are maps.
     */
    @SuppressWarnings("unchecked")
    public static Object deepMerge(Object value1, Object value2)
    {
        if (value1 instanceof Map && value2 instanceof Map)
        {
            Map<Object, Object> firstMap = new HashMap<>((Map<Object, Object>) value1);
            Map<Object, Object> secondMap = new HashMap<>((Map<Object, Object>) value2);
            for (Object key : secondMap.keySet())
            {
                Object value = secondMap.get(key);
                if (value != null)
                {
                    firstMap.merge(key, value, MapUtils::deepMerge);
                }
                // null value are not supported by remapping function, add it only if the value is not in value1
                else if (!firstMap.containsKey(key))
                {
                    firstMap.put(key, value);
                }
            }
            
            return firstMap;
        }
        return value1;
    }
}
