/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;

import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.DefinitionAndValue;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelHelper;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.model.type.xml.XMLElementType;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Helper to read XML data
 */
public class ReadXMLDataHelper extends AbstractLogEnabled implements Component
{
    /** The avalon role */
    public static final String ROLE = ReadXMLDataHelper.class.getName();
    
    /**
     * Read data from the given {@link Configuration} and validate the values
     * @param configuration the configuration
     * @param prefix the definition's prefix to use
     * @param definitions the definitions of the data to retrieve
     * @param allErrors a map to store all validation errors on read data
     * @return the values, indexed by name
     * @throws ConfigurationException if an error occurs while reading a value or if the value is not valid
     */
    public Map<String, Object> readAndValidateXMLData(Configuration configuration, String prefix, Map<String, ? extends ElementDefinition> definitions, Map<String, List<I18nizableText>> allErrors) throws ConfigurationException
    {
        Map<String, DefinitionAndValue> definitionAndValues = new LinkedHashMap<>();
        for (String dataName : definitions.keySet())
        {
            ElementDefinition definition = definitions.get(dataName);
            ElementType type = definition.getType();
            if (type instanceof XMLElementType xmlElementType)
            {
                Object value = xmlElementType.read(configuration, dataName);
                definitionAndValues.put(prefix + dataName, new DefinitionAndValue<>(null, definition,  value));
            }
        }
        
        Map<String, Object> xmlValues = new LinkedHashMap<>();
        for (String dataName : definitions.keySet())
        {
            ElementDefinition definition = definitions.get(dataName);
            boolean isDisabled = ModelHelper.evaluateDisableConditions(definition.getDisableConditions(), definitionAndValues, getLogger());
            
            Object value = definitionAndValues.get(prefix + dataName).getValue();
            if (!isDisabled)
            {
                List<I18nizableText> errors = ModelHelper.validateValue(definition, value);
                if (errors.isEmpty())
                {
                    xmlValues.put(dataName, value);
                }
                else
                {
                    allErrors.put(dataName, errors);
                }
            }
            else
            {
                xmlValues.put(dataName, value);
            }
        }
        
        return xmlValues;
    }
}
