/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collections;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.lang3.tuple.Pair;
import org.openjdk.jol.info.ClassData;
import org.openjdk.jol.info.FieldData;
import org.openjdk.jol.layouters.CurrentLayouter;
import org.openjdk.jol.util.ObjectUtils;
import org.openjdk.jol.vm.VM;
import org.slf4j.LoggerFactory;

/**
 * Helper for manipulating objects heap size.
 */
public final class SizeUtils
{
    private static Map<Class<?>, Pair<Long, List<Field>>> __cache = new ConcurrentHashMap<>();
    
    static
    {
        System.setProperty("jol.magicFieldOffset", "true");

        // add opens to JOL for all java.base packages, to avoid illegal access warnings
        Object.class.getModule().getPackages().forEach(p -> 
        {
            if (Object.class.getModule().isOpen(p, ObjectUtils.class.getModule()))
            {
                Object.class.getModule().addOpens(p, ObjectUtils.class.getModule());
            }
        });
    }
    
    private SizeUtils()
    {
        // empty
    }
    
    /**
     * Returns true if size computation is enabled for this VM
     * @return true if size computation is enabled
     */
    public static boolean enabled()
    {
        try
        {
            VM.current();
            return true;
        }
        catch (IllegalStateException e)
        {
            // only Hotspot/OpenJDK VM are supported
            return false;
        }
    }
    
    /**
     * Calculates the heap size of the given object.
     * @param object the object
     * @return the size of the object.
     */
    public static long sizeOf(Object object)
    {
        if (object == null)
        {
            return 0;
        }
        
        try
        {
            return _sizeOf(object, Collections.newSetFromMap(new IdentityHashMap<>()));
        }
        catch (Error e)
        {
            // Special case for intercepting eg. StackOverflowError and log the concerned object
            LoggerFactory.getLogger(SizeUtils.class.getName()).error("Error computing size of {}", object, e);
            throw e;
        }
    }
    
    private static long _sizeOf(Object object, Set<Object> visited)
    {
        if (!visited.add(object))
        {
            return 0;
        }
        
        // allow to bypass the internal computation
        if (object instanceof Sized sized)
        {
            return sized.getSize();
        }
        
        Class<?> klass = object.getClass();
        
        if (klass.isArray())
        {
            long size = new CurrentLayouter().layout(ClassData.parseInstance(object)).instanceSize();

            if (klass.getComponentType().isPrimitive()) 
            {
                return size;
            }
            
            Object[] array = (Object[]) object;
            
            for (Object item : array)
            {
                if (_isComputable(item))
                {
                    size += _sizeOf(item, visited);
                }
            }
            
            return size;
        }
        
        Pair<Long, List<Field>> data = _shallowSizeOf(klass);
        
        // the actual size of a non-array object is the sum of its shallow size and the size of each of its fields
        long size = data.getLeft();
        
        for (Field field : data.getRight())
        {
            Object value = ObjectUtils.value(object, field);
            
            if (_isComputable(value))
            {
                size += _sizeOf(value, visited);
            }
        }
        
        return size;
    }
    
    private static Pair<Long, List<Field>> _shallowSizeOf(Class<?> klass)
    {
        // we cache the shallow class size and the list of class' fields
        return __cache.computeIfAbsent(klass, c -> 
        {
            ClassData classData = ClassData.parseClass(c);

            List<Field> fields = new ArrayList<>();
            for (FieldData fieldData : classData.fields())
            {
                Field field = fieldData.refField();
                if (!field.getType().isPrimitive() 
                        && !field.isAnnotationPresent(ExcludeFromSizeCalculation.class))
                {
                    fields.add(field);
                }
            }
            
            return Pair.of(new CurrentLayouter().layout(classData).instanceSize(), fields);
        });
    }
    
    /**
     * Calculates the shallow size of the given object (ie the size of its internal structure).
     * @param object the object
     * @return the shallow size of the object.
     */
    public static long shallowSizeOf(Object object)
    {
        Class<?> klass = object.getClass();
        Pair<Long, List<Field>> data = _shallowSizeOf(klass);
        return data.getKey();
    }

    /**
     * Helper to compute the size of many objects at once
     * @param objects the objects
     * @return the sum of the sizes
     */
    public static long totalSizeOf(Object... objects)
    {
        long size = 0;
        
        if (objects != null)
        {
            for (Object object : objects)
            {
                size += sizeOf(object);
            }
        }
        
        return size;
    }
    
    private static boolean _isComputable(Object object)
    {
        return object != null && !(object instanceof Class);
    }
    
    /**
     * Fields annotated with {@link ExcludeFromSizeCalculation} are excluded for size calculation
     */
    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.FIELD)
    public static @interface ExcludeFromSizeCalculation
    {
        // marker annotation
    }
    
    /**
     * Implemented by objects knowing how to compute their size.
     */
    public static interface Sized
    {
        /**
         * Returns the size of this object
         * @return the size of this object
         */
        long getSize();
    }
}
