/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.io.ByteArrayOutputStream;
import java.nio.charset.StandardCharsets;
import java.util.Map;
import java.util.function.Predicate;

/**
 * Utility class for encoding and decoding URL, following the RFC 3986 
 * @see <a href="https://tools.ietf.org/html/rfc3986">https://tools.ietf.org/html/rfc3986</a>
 */
public final class URIUtils
{
    private static final String __NAME_VALUE_SEPARATOR = "=";
    private static final String __PARAMETER_SEPARATOR = "&";
    
    private static Predicate<Byte> _isAlpha = c -> c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z';
    private static Predicate<Byte> _isDigit = c -> c >= '0' && c <= '9';
    private static Predicate<Byte> _isSubDelimiter = c -> '!' == c || '$' == c || '&' == c || '\'' == c || '(' == c || ')' == c 
                                                       || '*' == c || '+' == c || ',' == c || ';' == c || '=' == c;
    private static Predicate<Byte> _isUnreserved = _isAlpha.or(_isDigit).or(c -> '-' == c || '.' == c || '_' == c || '~' == c);
    private static Predicate<Byte> _isPchar = _isUnreserved.or(_isSubDelimiter).or(c -> ':' == c || '@' == c);
    
    private URIUtils()
    {
        // empty
    }
    
    /**
     * Encode a parameter's value using UTF-8 encoding
     * @param value the value or the name of the request's parameter to encode
     * @return the encoded value.
     */
    public static String encodeParameter(String value)
    {
        return _encodeUriComponent(value, _isPchar.or(c -> '/' == c || '?' == c)
                                                  .and(Predicate.not(c -> '=' == c || '+' == c || '&' == c)));
    }
    
    /**
     * Encode a fragment's value using UTF-8 encoding
     * @param value the value or the name of the request's fragment to encode
     * @return the encoded value.
     */
    public static String encodeFragment(String value)
    {
        return _encodeUriComponent(value, _isPchar.or(c -> '/' == c || '?' == c));
    }
    
    /**
     * Encode an URL path 
     * @param path the path to encode (before question-mark character)
     * @return the encoded path.
     */
    public static String encodePath(String path)
    {
        return _encodeUriComponent(path, _isPchar.or(c -> '/' == c)
                                                 .and(Predicate.not(c -> ';' == c)));
    }
    
    /**
     * Encode an URL path segment
     * @param pathSegment the path segment to encode
     * @return the encoded path segment.
     */
    public static String encodePathSegment(String pathSegment)
    {
        return _encodeUriComponent(pathSegment, _isPchar.and(Predicate.not(c -> ';' == c)));
    }
    
    /**
     * Encode a request header value.
     * @param header the value to encode
     * @return the encoded value.
     */
    public static String encodeHeader(String header)
    {
        return _encodeUriComponent(header, _isUnreserved);
    }

    // Implementation taken from Spring's UriUtils#encode
    @SuppressWarnings("all")
    private static String _encodeUriComponent(String uriComponent, Predicate<Byte> charactersToKeep)
    {
        byte[] bytes = uriComponent.getBytes(StandardCharsets.UTF_8);
        ByteArrayOutputStream bos = new ByteArrayOutputStream(bytes.length);
        
        for (byte b : bytes) 
        {
            if (b < 0) 
            {
                b += 256;
            }
            if (charactersToKeep.test(b))
            {
                bos.write(b);
            }
            else 
            {
                bos.write('%');
                char hex1 = Character.toUpperCase(Character.forDigit((b >> 4) & 0xF, 16));
                char hex2 = Character.toUpperCase(Character.forDigit(b & 0xF, 16));
                bos.write(hex1);
                bos.write(hex2);
            }
        }

        return new String(bos.toByteArray(), StandardCharsets.US_ASCII);
    }
    
    /**
     * Encode an URL path
     * @param path the path to encode (before question-mark character)
     * @param parameters the parameters. Can be null.
     * @return the encoded path
     */
    public static String encodeURI(String path, Map<String, String> parameters)
    {
        return _buildURI(path, parameters, true);
    }
    
    /**
     * Build an URL.
     * @param path the URL path.
     * @param parameters the URL parameters. Can be null.
     * @return the computed URL.
     */
    public static String buildURI(String path, Map<String, String> parameters)
    {
        return _buildURI(path, parameters, false);
    }
    
    private static String _buildURI(String path, Map<String, String> parameters, boolean encode)
    {
        StringBuilder sb = new StringBuilder();
        sb.append(encode ? encodePath(path) : path);
        
        if (parameters != null && !parameters.isEmpty())
        {
            StringBuilder query = new StringBuilder();
            
            for (String paramName : parameters.keySet())
            {
                String encodedName = encode ? encodeParameter(paramName) : paramName;
                String encodedValue = encode ? encodeParameter(parameters.get(paramName)) : parameters.get(paramName);
                
                if (query.length() > 0) 
                {
                    query.append(__PARAMETER_SEPARATOR);
                }
                
                query.append(encodedName)
                    .append(__NAME_VALUE_SEPARATOR)
                    .append(encodedValue);
            }
            
            sb.append("?").append(query.toString());
        }
        
        return sb.toString();
    }
    
    /**
     * Decodes an URI-encoded String.
     * @param source the String to decode.
     * @return the decoded String.
     */
    // Implementation taken from Spring's UriUtils#decode
    @SuppressWarnings("all")
    public static String decode(String source)
    {
        int length = source.length();
        ByteArrayOutputStream bos = new ByteArrayOutputStream(length);
        boolean changed = false;
        for (int i = 0; i < length; i++) 
        {
            char ch = source.charAt(i);
            
            if (ch == '%') 
            {
                if ((i + 2) < length) 
                {
                    char hex1 = source.charAt(i + 1);
                    char hex2 = source.charAt(i + 2);
                    int u = Character.digit(hex1, 16);
                    int l = Character.digit(hex2, 16);
                    
                    if (u == -1 || l == -1) 
                    {
                        throw new IllegalArgumentException("Invalid encoded sequence \"" + source.substring(i) + "\"");
                    }
                    
                    bos.write((byte) ((u << 4) + l));
                    
                    i += 2;
                    changed = true;
                }
                else 
                {
                    throw new IllegalArgumentException("Invalid encoded sequence \"" + source.substring(i) + "\"");
                }
            }
            else if (ch > 0 && ch < 128)
            {
                // to avoid the overhead of decoding/recoding an ASCII char
                bos.write(ch);
            }
            else if (Character.isHighSurrogate(ch))
            {
                if ((i + 1) < length && Character.isLowSurrogate(source.charAt(i + 1))) 
                {
                    for (byte b : String.valueOf(new char[]{ch, source.charAt(i + 1)}).getBytes(StandardCharsets.UTF_8))
                    {
                        bos.write(b);
                    }
                    
                    i += 1;
                }
                else 
                {
                    throw new IllegalArgumentException("Invalid encoded sequence \"" + source.substring(i) + "\"");
                }
            }
            else 
            {
                for (byte b : String.valueOf(ch).getBytes(StandardCharsets.UTF_8))
                {
                    bos.write(b);
                }
            }
        }
        
        return changed ? new String(bos.toByteArray(), StandardCharsets.UTF_8) : source;
    }
}
