/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.io.IOException;
import java.io.InputStream;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.excalibur.xml.sax.SAXParser;
import org.slf4j.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import org.ametys.runtime.i18n.I18nizable;
import org.ametys.runtime.plugin.component.LogEnabled;

/**
 * Component with XML utils methods
 */
public class XMLUtils extends org.apache.cocoon.xml.XMLUtils implements Component, Serviceable, LogEnabled
{
    /** The avalon role */
    public static final String ROLE = XMLUtils.class.getName();

    /** Avalon service manager */
    protected ServiceManager _manager;
    
    private Logger _logger;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
    }
    
    /**
     * Returns the {@link Logger}.
     * @return the {@link Logger}.
     */
    protected Logger getLogger()
    {
        return _logger;
    }

    public void setLogger(Logger logger)
    {
        _logger = logger;
    }
    
    /**
     * Create a start and endElement with a empty Namespace and without Attributes with the given {@link I18nizable}
     * If the {@link I18nizable} is <code>null</code>, the element is not created
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param localName The local name (without prefix)
     * @param i18n the {@link I18nizable} to SAX
     * @throws SAXException if an error occurs during the SAX events generation
     */
    public static void createI18nElementIfNotNull(ContentHandler contentHandler, String localName, I18nizable i18n) throws SAXException
    {
        if (i18n != null)
        {
            i18n.toSAX(contentHandler, localName);
        }
    }
    
    /**
     * Create a start and endElement with a empty Namespace and without Attributes
     * If the string value is <code>null</code>, the element is not created
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param localName The local name (without prefix)
     * @param stringValue The content of the Element
     * @throws SAXException if an error occurs during the SAX events generation
     */
    public static void createElementIfNotNull(ContentHandler contentHandler, String localName, String stringValue) throws SAXException
    {
        if (stringValue != null)
        {
            createElement(contentHandler, localName, stringValue);
        }
    }
    
    /**
     * Create a start and endElement with a empty Namespace and without Attributes
     * If the string value is <code>null</code>, the element is not created
     * @param contentHandler the {@link ContentHandler} that will receive the SAX events
     * @param localName The local name (without prefix)
     * @param stringValue The content of the Element
     * @param attributes The attributes attached to the element.  If there are no attributes, it shall be an empty Attributes object.
     * @throws SAXException if an error occurs during the SAX events generation
     */
    public static void createElementIfNotNull(ContentHandler contentHandler, String localName, String stringValue, Attributes attributes) throws SAXException
    {
        if (stringValue != null)
        {
            createElement(contentHandler, localName, attributes, stringValue);
        }
    }

    /**
     * Get a XML as a string and extract the text only
     * @param is The inputstream of XML
     * @return The text or null if the XML is not well formed
     */
    public String toString(InputStream is)
    {
        SAXParser saxParser = null;
        try
        {
            TxtHandler txtHandler = new TxtHandler();
            saxParser = (SAXParser) _manager.lookup(SAXParser.ROLE);
            saxParser.parse(new InputSource(is), txtHandler);
            return txtHandler.getValue();
        }
        catch (ServiceException e)
        {
            getLogger().error("Unable to get a SAX parser", e);
            return null;
        }
        catch (IOException | SAXException e)
        {
            getLogger().error("Cannot parse inputstream", e);
            return null;
        }
        finally
        {
            _manager.release(saxParser);
        }
    }
    
    final class TxtHandler extends DefaultHandler
    {
        private StringBuilder _internal = new StringBuilder();
        
        @Override
        public void characters(char[] ch, int start, int length) throws SAXException
        {
            _internal.append(new String(ch, start, length));
        }
        
        public String getValue()
        {
            return _internal.toString();
        }
    }
}
