/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util.date;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.temporal.TemporalUnit;
import java.util.Objects;

/**
 * This represents an advanced {@link LocalDateTime}, allowing to be a static one, 
 * or a relative one to the moment the {@link #resolveDateTime} method is called.
 * <br>A single instance of {@link AdaptableDate} can be {@link #resolveDateTime() resolved} mutliple times, without being a problem at all, it just can lead to different results.
 * <br>
 * <br>You can get an {@link AdaptableDate} with:
 * <ul>
 * <li>{@link #fromDateTime} to have a static one, always resolved as the same given {@link LocalDateTime}</li>
 * <li>{@link #fromDate} to have a static one, always resolved as the same given {@link LocalDate} (at the start of the day)</li>
 * <li> {@link #now} to have one resolved as the current {@link LocalDateTime} when {@link #resolveDateTime} is called</li>
 * <li> {@link #future} to have one resolved with the given amount of given {@link TemporalUnit} added to the current {@link LocalDateTime} when {@link #resolveDateTime} is called</li>
 * <li> {@link #past} to have one resolved with the given amount of given {@link TemporalUnit} subtracted to the current {@link LocalDateTime} when {@link #resolveDateTime} is called</li>
 * </ul>
 */
public final class AdaptableDate
{
    /**
     * Ways available to resolve dates.
     */
    public enum AdaptableDateType
    {
        /** Absolute */
        STATIC,
        /** Relative to the current instant */
        RELATIVE_TO_NOW,
        /** Relative to the beginning of the current day */
        RELATIVE_TO_TODAY
    }
    
    private enum OffsetType
    {
        NOW,
        PAST,
        FUTURE,
    }
    
    private static final AdaptableDate __TODAY = new AdaptableDate(AdaptableDateType.RELATIVE_TO_TODAY, OffsetType.NOW, 0, null);
    private static final AdaptableDate __NOW = new AdaptableDate(AdaptableDateType.RELATIVE_TO_NOW, OffsetType.NOW, 0, null);
    
    private final AdaptableDateType _type;
    private final LocalDate _staticDate;
    private final LocalDateTime _staticDateTime;
    private final TemporalUnit _relativeUnit;
    private final long _relativeAmount;
    private final OffsetType _relativeOffsetType;
    
    private AdaptableDate(LocalDate date, LocalDateTime dateTime)
    {
        _type = AdaptableDateType.STATIC;
        _staticDate = date;
        _staticDateTime = dateTime;
        _relativeUnit = null;
        _relativeAmount = 0;
        _relativeOffsetType = null;
    }
    
    private AdaptableDate(AdaptableDateType type, OffsetType offsetType, long amount, TemporalUnit unit)
    {
        if (type == AdaptableDateType.STATIC)
        {
            throw new IllegalArgumentException("There can't be any offset for a static AdaptableDate");
        }
        
        _type = type;
        _staticDate = null;
        _staticDateTime = null;
        _relativeUnit = unit;
        _relativeAmount = amount;
        _relativeOffsetType = offsetType;
    }
    
    /**
     * Gets a static {@link AdaptableDate}, always resolved as the given {@link LocalDateTime}
     * @param dateTime The {@link LocalDateTime}
     * @return The {@link AdaptableDate}
     */
    public static AdaptableDate fromDateTime(LocalDateTime dateTime)
    {
        return new AdaptableDate(null, dateTime);
    }
    
    /**
     * Gets a static {@link AdaptableDate}, always resolved as the given {@link LocalDate}
     * @param date The {@link LocalDate}
     * @return The {@link AdaptableDate}
     */
    public static AdaptableDate fromDate(LocalDate date)
    {
        return new AdaptableDate(date, null);
    }
    
    /**
     * Gets an {@link AdaptableDate}, resolved as the current {@link LocalDateTime} at the moment of the {@link #resolveDateTime() call}
     * @return The {@link AdaptableDate}
     */
    public static AdaptableDate now()
    {
        return __NOW;
    }
    
    /**
     * Gets an {@link AdaptableDate}, resolved at the beggining of the current day at the moment of the {@link #resolveDateTime() call}
     * @return The {@link AdaptableDate}
     */
    public static AdaptableDate today()
    {
        return __TODAY;
    }
    
    /**
     * Gets an {@link AdaptableDate}, resolved as the current {@link LocalDateTime} at the moment of the {@link #resolveDateTime() call}, 
     * minus the given amount of the given {@link TemporalUnit}, so as to get a {@link LocalDateTime} in the past
     * @param type the type of the offset to consider (relative to now or the beginning of the current day)
     * @param amount The amount to subtract
     * @param unit The {@link TemporalUnit} of the amount to subtract
     * @return The {@link AdaptableDate}
     */
    public static AdaptableDate past(AdaptableDateType type, long amount, TemporalUnit unit)
    {
        return new AdaptableDate(type, OffsetType.PAST, amount, unit);
    }
    
    /**
     * Gets an {@link AdaptableDate}, resolved as the current {@link LocalDateTime} at the moment of the {@link #resolveDateTime() call}, 
     * plus the given amount of the given {@link TemporalUnit}, so as to get a {@link LocalDateTime} in the future
     * @param type the type of the offset to consider (relative to now or the beginning of the current day)
     * @param amount The amount to add
     * @param unit The {@link TemporalUnit} of the amount to add
     * @return The {@link AdaptableDate}
     */
    public static AdaptableDate future(AdaptableDateType type, long amount, TemporalUnit unit)
    {
        return new AdaptableDate(type, OffsetType.FUTURE, amount, unit);
    }
    
    /**
     * Resolves this {@link AdaptableDate}
     * @return The resolved {@link LocalDateTime}
     */
    public DateTimeInterval resolveDateTime()
    {
        if (_type == AdaptableDateType.STATIC)
        {
            if (_staticDateTime != null)
            {
                return new DateTimeInterval(_staticDateTime, _staticDateTime);
            }
            
            LocalDateTime min = _staticDate.atStartOfDay();
            LocalDateTime max = _staticDate.atTime(LocalTime.MAX);

            return new DateTimeInterval(min, max);
        }
        
        if (_type == AdaptableDateType.RELATIVE_TO_NOW)
        {
            LocalDateTime now = LocalDateTime.now();
            
            switch (_relativeOffsetType)
            {
                case NOW:
                    return new DateTimeInterval(now, now);
                case PAST:
                    LocalDateTime past = now.minus(_relativeAmount, _relativeUnit);
                    return new DateTimeInterval(past, past);
                case FUTURE:
                default:
                    LocalDateTime future = now.plus(_relativeAmount, _relativeUnit);
                    return new DateTimeInterval(future, future);
            }
        }
        else
        {
            LocalDate today = LocalDate.now();

            switch (_relativeOffsetType)
            {
                case NOW:
                    LocalDateTime min = today.atStartOfDay();
                    LocalDateTime max = today.atTime(LocalTime.MAX);
                    
                    return new DateTimeInterval(min, max);
                case PAST:
                    LocalDate past = today.minus(_relativeAmount, _relativeUnit);
                    LocalDateTime minPast = past.atStartOfDay();
                    LocalDateTime maxPast = past.atTime(LocalTime.MAX);
                    
                    return new DateTimeInterval(minPast, maxPast);
                case FUTURE:
                default:
                    LocalDate future = today.plus(_relativeAmount, _relativeUnit);
                    LocalDateTime minFuture = future.atStartOfDay();
                    LocalDateTime maxFuture = future.atTime(LocalTime.MAX);
                    
                    return new DateTimeInterval(minFuture, maxFuture);
            }
        }
    }
    
    /**
     * Resolves this {@link AdaptableDate}
     * @return The resolved {@link LocalDate}
     */
    public LocalDate resolveDate()
    {
        if (_type == AdaptableDateType.STATIC)
        {
            if (_staticDate != null)
            {
                return _staticDate;
            }
            else
            {
                return _staticDateTime.toLocalDate();
            }
        }
        
        LocalDate now = LocalDate.now();
        
        switch (_relativeOffsetType)
        {
            case NOW:
                return now;
            case PAST:
                return now.minus(_relativeAmount, _relativeUnit);
            case FUTURE:
            default:
                return now.plus(_relativeAmount, _relativeUnit);
        }
    }
    
    @Override
    public String toString()
    {
        switch (_type)
        {
            case STATIC:
                return _staticDate != null ? _staticDate.toString() : _staticDateTime.toString();
            case RELATIVE_TO_TODAY:
                switch (_relativeOffsetType)
                {
                    case NOW:
                        return "Today";
                    case PAST:
                        return _relativeAmount + " " + _relativeUnit + " ago today";
                    case FUTURE:
                    default:
                        return "Today in " + _relativeAmount + " " + _relativeUnit;
                }
            case RELATIVE_TO_NOW:
            default:
                switch (_relativeOffsetType)
                {
                    case NOW:
                        return "Now";
                    case PAST:
                        return _relativeAmount + " " + _relativeUnit + " ago";
                    case FUTURE:
                    default:
                        return "In " + _relativeAmount + " " + _relativeUnit;
                }
        }
    }
    
    @Override
    public int hashCode()
    {
        return Objects.hash(_type, _staticDate, _staticDateTime, _relativeAmount, _relativeOffsetType, _relativeUnit);
    }
    
    @Override
    public boolean equals(Object obj)
    {
        if (!(obj instanceof AdaptableDate other))
        {
            return false;
        }
        
        return Objects.equals(_type, other._type)
            && Objects.equals(_staticDate, other._staticDate)
            && Objects.equals(_staticDateTime, other._staticDateTime)
            && Objects.equals(_relativeAmount, other._relativeAmount)
            && Objects.equals(_relativeOffsetType, other._relativeOffsetType)
            && Objects.equals(_relativeUnit, other._relativeUnit);
    }
    
    /**
     * Interval of {@link LocalDateTime}
     * @param min the minimum LocalDateTime
     * @param max the maximum LocalDateTime
     */
    public record DateTimeInterval(LocalDateTime min, LocalDateTime max) { }
}

