/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util.dom;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.StringUtils;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Some useful utility methods for {@link Element} use 
 */
public final class DOMUtils
{
    private DOMUtils()
    {
        // Empty constructor
    }
    
    /**
     * Retrieves all child elements of the given DOM element.
     * @param element The DOM element to analyze
     * @return a {@link List} of child elements instances
     */
    public static List<Element> getChildElements(Element element)
    {
        if (element == null)
        {
            throw new IllegalArgumentException("The element must not be null");
        }
        
        List<Element> elements = new ArrayList<>();
        
        NodeList nodes = element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++)
        {
            Node node = nodes.item(i);
            if (node.getNodeType() == Node.ELEMENT_NODE)
            {
                elements.add((Element) node);
            }
        }
        
        return elements;
    }
    
    /**
     * Get the first child element of the given DOM element.
     * @param element The DOM element to analyze
     * @return the first child element or null if there is no child element
     */
    public static Element getFirstChildElement(Element element)
    {
        if (element == null)
        {
            throw new IllegalArgumentException("The element must not be null");
        }
        
        NodeList nodes = element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++)
        {
            Node node = nodes.item(i);
            if (node.getNodeType() == Node.ELEMENT_NODE)
            {
                return (Element) node;
            }
        }
        
        return null;
    }
    
    /**
     * Retrieves some child elements of the given DOM element
     * If some elements have the same tag name, only the first of these elements is retrieved
     * @param element the DOM element to analyze
     * @return a {@link List} of child elements instances
     */
    public static List<Element> getUniqueChildElements(Element element)
    {
        if (element == null)
        {
            throw new IllegalArgumentException("The element must not be null");
        }
        
        List<Element> elements = new ArrayList<>();
        Set<String> tagNames = new HashSet<>();
        
        NodeList nodes = element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++)
        {
            Node node = nodes.item(i);
            String dataName = node.getNodeName();
            if (!tagNames.contains(dataName) && node.getNodeType() == Node.ELEMENT_NODE)
            {
                elements.add((Element) node);
                tagNames.add(dataName);
            }
        }
        
        return elements;
    }
    
    /**
     * Retrieves all child elements of the given DOM element that match the given element name.
     * Only look at the direct child level of the given element.
     * Do not go into further depth (in contrast to the DOM API's getElementsByTagName method).
     * @param element The DOM element to analyze
     * @param name The child element name to look for
     * @return a {@link List} of child elements instances
     */
    public static List<Element> getChildElementsByTagName(Element element, String name)
    {
        if (element == null)
        {
            throw new IllegalArgumentException("The element must not be null");
        }
        
        if (StringUtils.isEmpty(name))
        {
            throw new IllegalArgumentException("The name must not be empty");
        }
        
        List<Element> elements = new ArrayList<>();
        
        NodeList nodes = element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++)
        {
            Node node = nodes.item(i);
            if (node.getNodeType() == Node.ELEMENT_NODE && name.equals(node.getNodeName()))
            {
                elements.add((Element) node);
            }
        }
        
        return elements;
    }
    
    /**
     * Retrieves the first child element of the given DOM element that matches the given element name.
     * Only look at the direct child level of the given element. Do not go into further depth.
     * @param element The DOM element to analyze
     * @param name The child element name to look for
     * @return the first child element instance
     */
    public static Element getChildElementByTagName(Element element, String name)
    {
        if (element == null)
        {
            throw new IllegalArgumentException("The element must not be null");
        }
        
        if (StringUtils.isEmpty(name))
        {
            throw new IllegalArgumentException("The name must not be empty");
        }
        
        NodeList nodes = element.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++)
        {
            Node node = nodes.item(i);
            if (node.getNodeType() == Node.ELEMENT_NODE && name.equals(node.getNodeName()))
            {
                return (Element) node;
            }
        }
        
        return null;
    }
    
    /**
     * Checks if the given DOM element has a child element that matches the given element name.
     * @param element The DOM element to analyze
     * @param name The child element name to look for
     * @return <code>true</code> if the given DOM element has a child element that matches the given element name, <code>false</code> otherwise
     */
    public static boolean hasChildElement(Element element, String name)
    {
        return getChildElementByTagName(element, name) != null;
    }
}
