/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.core.util.dom;

import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceException;
import org.apache.excalibur.source.TraversableSource;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * DOM layer on top if a {@link File} hierarchy.
 */
public class SourceElement extends AbstractWrappingAmetysElement<Source>
{
    /**
     * Constructor.
     * @param source the underlying {@link Source}.
     */
    public SourceElement(Source source)
    {
        this(source, null);
    }

    /**
     * Constructor.
     * @param source the underlying {@link Source}.
     * @param parent the parent {@link Element}.
     */
    public SourceElement(Source source, SourceElement parent)
    {
        super(source, parent);
    }
    
    @Override
    public String getTagName()
    {
        return _object instanceof TraversableSource && ((TraversableSource) _object).isCollection() ? "collection" : "resource";
    }
    
    @Override
    protected Map<String, AmetysAttribute> _lookupAttributes()
    {
        Map<String, AmetysAttribute> result = new HashMap<>();
        
        result.put("name", new AmetysAttribute("name", "name", null, StringUtils.substringAfterLast(_object.getURI(), "/"), this));
        
        return result;
    }
    
    @Override
    public boolean hasChildNodes()
    {
        try
        {
            return _object instanceof TraversableSource
                    && ((TraversableSource) _object).isCollection()
                    && CollectionUtils.isNotEmpty(((TraversableSource) _object).getChildren());
        }
        catch (SourceException e)
        {
            throw new RuntimeException("Cannot list child elements of " + _object.getURI(), e);
        }
    }
    
    @Override
    public Node getFirstChild()
    {
        try
        {
            if (_object instanceof TraversableSource
                && ((TraversableSource) _object).isCollection())
            {
                Collection<Source> children = ((TraversableSource) _object).getChildren();
                if (children.size() > 0)
                {
                    return new SourceElement(children.iterator().next(), this);
                }
            }
        }
        catch (SourceException e)
        {
            throw new RuntimeException("Cannot list child elements of " + _object.getURI(), e);
        }
        
        return null;
    }
    
    @Override
    public Node getNextSibling()
    {
        if (_parent == null)
        {
            return null;
        }
        
        TraversableSource parent = (TraversableSource) ((AbstractWrappingAmetysElement) _parent).getWrappedObject();
        
        Iterator<Source> iterator;
        try
        {
            iterator = parent.getChildren().iterator();
        }
        catch (SourceException e)
        {
            throw new RuntimeException("Cannot list child elements of " + parent.getURI(), e);
        }
        
        boolean isNext = false;
        Source nextSibling = null;
        
        while (nextSibling == null && iterator.hasNext())
        {
            Source child = iterator.next();
            
            if (isNext)
            {
                nextSibling = child;
            }
            else if (_object.getURI().equals(child.getURI()))
            {
                isNext = true;
            }
        }
        
        return nextSibling == null ? null : new SourceElement(nextSibling, (SourceElement) _parent);
    }
}
