/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util.mail;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Files;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.util.mail.SendMailHelper.NamedStream;

import jakarta.activation.DataSource;

/**
 * The input stream data source
 */
public final class InputStreamDataSource implements DataSource
{
    private String _name;
    private String _type;
    private File _tmpFile;
    
    /**
     * The constructor of the input stream data source
     * @param attachment the attachment as stream
     * @throws IOException if an error occurred
     */
    public InputStreamDataSource(NamedStream attachment) throws IOException
    {
        _name = attachment.name();
        _type = attachment.mimeType();

        String prefix = StringUtils.substringBeforeLast(_name, ".");
        String suffix = StringUtils.substringAfterLast(_name, ".");
        
        _tmpFile = Files.createTempFile(prefix, "." + suffix).toFile();
        
        IOUtils.copy(attachment.inputStream(), new FileOutputStream(_tmpFile));
    }
    
    public InputStream getInputStream() throws IOException
    {
        return new FileInputStream(_tmpFile);
    }

    public OutputStream getOutputStream() throws IOException
    {
        throw new UnsupportedOperationException("Method getOutputStream can't be called for InputStreamDataSource");
    }

    public String getContentType()
    {
        return _type;
    }

    public String getName()
    {
        return _name;
    }
    
    /**
     * Dispose the input stream data source deleting the tmp file
     * @throws IOException if an error occurred
     */
    public void dispose() throws IOException
    {
        FileUtils.forceDeleteOnExit(_tmpFile);
    }
}
