/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util.path;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;

import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.SourceValidity;

import org.ametys.core.util.JarFSManager;

/**
 * A validation object for time-stamps.
 */
public class PathTimeStampValidity implements SourceValidity
{
    /** The timestamp */
    protected long _timeStamp;
    private transient Path _file;
    private String _absolutePath;

    /**
     * Creates from a name
     * @param filename File name
     */
    public PathTimeStampValidity(final String filename)
    {
        this(Path.of(filename));
    }

    /**
     * Creates from a file
     * @param file file wrapped
     */
    public PathTimeStampValidity(final Path file)
    {
        this(file, _getTimeStamp(file));
    }

    /**
     * Creates from a file
     * @param file file wrapped
     * @param timeStamp file timestamp
     */
    public PathTimeStampValidity(final Path file, final long timeStamp)
    {
        _file = file;
        try
        {
            _absolutePath = (new URI(file.toUri().getRawSchemeSpecificPart())).getPath();
        }
        catch (URISyntaxException e)
        {
            throw new RuntimeException("Cannot convert " + file.toString() + " to URI", e);
        }
        _timeStamp = timeStamp;
    }
    
    /**
     * Empty constructor for sub classes only.
     */
    protected PathTimeStampValidity()
    {
        // Nothing
    }

    /**
     * Get the timestamp from a file
     * @param file the file
     * @return the file's timestamp
     */
    protected static long _getTimeStamp(Path file)
    {
        try
        {
            return Files.getLastModifiedTime(file).toMillis();
        }
        catch (IOException e)
        {
            return -1;
        }
    }
    
    public int isValid()
    {
        return _getTimeStamp(getFile()) == _timeStamp ? 1 : -1;
    }

    public int isValid(final SourceValidity newValidity)
    {
        if (newValidity instanceof PathTimeStampValidity)
        {
            final long timeStamp = ((PathTimeStampValidity) newValidity).getTimeStamp();
            return _timeStamp == timeStamp ? 1 : -1;
        }
        return -1;
    }

    /**
     * Get the file
     * @return the file
     */
    public Path getFile()
    {
        if (_file == null)
        {
            _file = _getFile(_absolutePath);
        }
        
        return this._file;
    }
    
    /**
     * Get file from an absolute path
     * @param absolutePath the absolute path
     * @return the file
     */
    protected static Path _getFile(String absolutePath)
    {
        String[] path = StringUtils.split(absolutePath, "!");
        if (path.length == 2)
        {
            try
            {
                return JarFSManager.getInstance().getFileSystemByFile(new java.io.File(path[0])).getPath(path[1]);
            }
            catch (IOException e)
            {
                throw new RuntimeException("Cannot unserialize the path " + absolutePath, e);
            }
        }
        else
        {
            return new java.io.File(path[0]).toPath();
        }
    }

    /**
     * Get the timestamp
     * @return the timestamp or -1 if it cannot be computed
     */
    public long getTimeStamp()
    {
        return this._timeStamp;
    }

    @Override
    public String toString()
    {
        return "FileTimeStampValidity: " + getFile().toString() + ": " + this._timeStamp;
    }

}
