/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.authentication;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.generation.ServiceableGenerator;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.commons.lang3.StringUtils;
import org.xml.sax.SAXException;

import org.ametys.core.authentication.BlockingCredentialProvider;
import org.ametys.core.authentication.CredentialProvider;
import org.ametys.core.authentication.CredentialProviderFactory;
import org.ametys.core.authentication.CredentialProviderModel;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.user.population.UserPopulation;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.core.impl.authentication.FormCredentialProvider;
import org.ametys.runtime.config.Config;

/**
 * SAX configuration of the login screen
 */
public class LoginScreenGenerator extends ServiceableGenerator
{
    /** Name of the input parameters html field */
    protected CredentialProviderFactory _credentialProviderFactory;
    
    /** The DAO for user populations */
    protected UserPopulationDAO _userPopulationDAO;

    /** The JSON helper */
    protected JSONUtils _jsonUtils;

    /** The I18N utils */
    protected I18nUtils _i18nUtils;
    
    /** The user manager */
    protected UserManager _userManager;
   
    /** The multifactor authentication manager */
    protected MultifactorAuthenticationManager _multifactorAuthenticationManager;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        super.service(smanager);
        _credentialProviderFactory = (CredentialProviderFactory) smanager.lookup(CredentialProviderFactory.ROLE);
        _userPopulationDAO = (UserPopulationDAO) smanager.lookup(UserPopulationDAO.ROLE);
        _jsonUtils = (JSONUtils) smanager.lookup(JSONUtils.ROLE);
        _i18nUtils = (I18nUtils) smanager.lookup(I18nUtils.ROLE);
        _userManager = (UserManager) smanager.lookup(UserManager.ROLE);
        _multifactorAuthenticationManager = (MultifactorAuthenticationManager) smanager.lookup(MultifactorAuthenticationManager.ROLE);
    }
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        contentHandler.startDocument();
        XMLUtils.startElement(contentHandler, "LoginScreen");
        
        // LoginScreenGenerator can be used from the frontoffice AuthenticateAction: it can not use the requests attributes
        boolean isAmetysPublic = Config.getInstance() != null ? Config.getInstance().getValue("runtime.ametys.public") : false/* in safe mode, we only have one population */;
        
        boolean invalidPopulationIds = "true".equals(request.getParameter("invalidPopulationIds"));
        
        boolean shouldDisplayUserPopulationsList = "true".equals(request.getParameter("shouldDisplayUserPopulationsList"));
        
        List<UserPopulation> usersPopulations = null;
        String usersPopulationsIdsAsString = request.getParameter("usersPopulations");
        if (usersPopulationsIdsAsString != null)
        {
            usersPopulations = Arrays.stream(usersPopulationsIdsAsString.split(",")).map(_userPopulationDAO::getUserPopulation).collect(Collectors.toList());
        }
        
        String chosenPopulationId = request.getParameter("chosenPopulationId");
        
        if (shouldDisplayUserPopulationsList)
        {
            _generatePopulations(usersPopulations, isAmetysPublic, invalidPopulationIds, chosenPopulationId);
        }
        
        boolean availableCredentialProviders = "true".equals(request.getParameter("availableCredentialProviders"));
        List<CredentialProvider> credentialProviders = null;
        if (availableCredentialProviders && usersPopulations != null && !usersPopulations.isEmpty())
        {
            if (StringUtils.isNotBlank(chosenPopulationId))
            {
                credentialProviders = usersPopulations.stream().filter(userPop -> chosenPopulationId.equals(userPop.getId())).findAny().get().getCredentialProviders();
            }
            else
            {
                credentialProviders = usersPopulations.get(0).getCredentialProviders();
            }
        }
        int credentialProviderIndex = Integer.parseInt(request.getParameter("credentialProviderIndex"));

        _generateCredentialProviders(credentialProviders, credentialProviderIndex);
        
        _generateLoginForm(request, credentialProviders, credentialProviderIndex, invalidPopulationIds, chosenPopulationId != null || usersPopulations == null ? chosenPopulationId : usersPopulations.get(0).getId());
        
        String contextsAsString = request.getParameter("contexts");
        XMLUtils.createElement(contentHandler, "contexts", contextsAsString);
        
        XMLUtils.endElement(contentHandler, "LoginScreen");
        contentHandler.endDocument();
    }
    
    private void _generatePopulations(List<UserPopulation> usersPopulations, boolean isAmetysPublic, boolean triedAnInvalidPopulationId, String chosenPopulationId) throws SAXException
    {
        if (usersPopulations != null)
        {
            AttributesImpl attrs = new AttributesImpl();
            attrs.addCDATAAttribute("invalid", triedAnInvalidPopulationId ? "true" : "false");
            attrs.addCDATAAttribute("public", isAmetysPublic ? "true" : "false");
            attrs.addCDATAAttribute("size", Integer.toString(usersPopulations.size()));
            if (chosenPopulationId != null)
            {
                attrs.addCDATAAttribute("currentValue", chosenPopulationId);
            }
            XMLUtils.startElement(contentHandler, "UserPopulations", attrs);
            
            if (!isAmetysPublic)
            {
                for (UserPopulation up : usersPopulations)
                {
                    AttributesImpl attrs2 = new AttributesImpl();
                    attrs2.addCDATAAttribute("id", up.getId());
                    XMLUtils.startElement(contentHandler, "UserPopulation", attrs2);
                    up.getLabel().toSAX(contentHandler, "label");
                    XMLUtils.endElement(contentHandler, "UserPopulation");
                }
            }
        
            XMLUtils.endElement(contentHandler, "UserPopulations");
        }
    }
    
    private void _generateCredentialProviders(List<CredentialProvider> credentialProviders, int currentCredentialProvider) throws SAXException
    {
        if (credentialProviders == null)
        {
            return;
        }
        
        XMLUtils.startElement(contentHandler, "CredentialProviders");
        for (int index = 0; index < credentialProviders.size(); index++)
        {
            CredentialProvider cp = credentialProviders.get(index);
            if (cp instanceof BlockingCredentialProvider blockingCP)
            {
                CredentialProviderModel cpModel = _credentialProviderFactory.getExtension(cp.getCredentialProviderModelId());
                
                AttributesImpl attrs = new AttributesImpl();
                attrs.addCDATAAttribute("index", String.valueOf(index));
                attrs.addCDATAAttribute("selected", index == currentCredentialProvider ? "true" : "false");
                attrs.addCDATAAttribute("isForm", cp instanceof FormCredentialProvider ? "true" : "false");
                attrs.addCDATAAttribute("isNewWindowRequired", blockingCP.requiresNewWindow() ? "true" : "false");
                XMLUtils.startElement(contentHandler, "CredentialProvider", attrs);
                XMLUtils.createElement(contentHandler, "label", _i18nUtils.translate(cpModel.getConnectionLabel()));
                if (StringUtils.isNotBlank(cp.getLabel()))
                {
                    XMLUtils.createElement(contentHandler, "additionalLabel", cp.getLabel());
                }
                if (StringUtils.isNotEmpty(cpModel.getIconGlyph()))
                {
                    XMLUtils.createElement(contentHandler, "iconGlyph", cpModel.getIconGlyph());
                    XMLUtils.createElement(contentHandler, "iconDecorator", cpModel.getIconDecorator());
                }
                else if (StringUtils.isNotEmpty(cpModel.getIconMedium()))
                {
                    XMLUtils.createElement(contentHandler, "iconMedium", cpModel.getIconMedium());
                }
                XMLUtils.createElement(contentHandler, "color", cpModel.getColor());
                XMLUtils.endElement(contentHandler, "CredentialProvider");
            }
        }
        XMLUtils.endElement(contentHandler, "CredentialProviders");
    }
    
    private void _generateLoginForm(Request request, List<CredentialProvider> credentialProviders, int currentCredentialProvider, boolean invalidPopulationIds, String chosenPopulationId) throws SAXException
    {
        if (credentialProviders == null)
        {
            return;
        }
        
        FormCredentialProvider formBasedCP = null;
        if (currentCredentialProvider != -1 && credentialProviders.get(currentCredentialProvider) instanceof FormCredentialProvider)
        {
            formBasedCP = (FormCredentialProvider) credentialProviders.get(currentCredentialProvider);
        }
        else
        {
            Optional<CredentialProvider> foundAnyFormCredentialProvider = credentialProviders.stream().filter(cp -> cp instanceof FormCredentialProvider).findAny();
            if (foundAnyFormCredentialProvider.isPresent())
            {
                formBasedCP = (FormCredentialProvider) foundAnyFormCredentialProvider.get();
            }
            else
            {
                // We found no form based
                return;
            }
        }

        boolean useCaptcha = (Boolean) formBasedCP.getParameterValues().get("runtime.authentication.form.captcha");
        
        boolean autoComplete = true; // FIXME not well implemented
        boolean rememberMe = (Boolean) formBasedCP.getParameterValues().get("runtime.authentication.form.cookies");
        boolean allowLoginByEmail = (Boolean) formBasedCP.getParameterValues().get("runtime.authentication.form.login-by-email");
        boolean useMultifactorAuthentication = formBasedCP.useMultifactorAuthentication();
        
        boolean captcha = false;
        
        if (useCaptcha)
        {
            String login = request.getParameter("login");
            if (StringUtils.isNotBlank(login) && StringUtils.isNotBlank(chosenPopulationId))
            {
                int nbConnect = formBasedCP.requestNbConnectBDD(login, chosenPopulationId);
                captcha = nbConnect >= FormCredentialProvider.NB_CONNECTION_ATTEMPTS;
            }
        }
        
        boolean showErrors = !invalidPopulationIds;
        
        XMLUtils.startElement(contentHandler, "LoginForm");
        
        if (StringUtils.isNotBlank(formBasedCP.getLabel()))
        {
            XMLUtils.createElement(contentHandler, "additionalLabel", formBasedCP.getLabel());
        }
        XMLUtils.createElement(contentHandler, "autocomplete", String.valueOf(autoComplete));
        XMLUtils.createElement(contentHandler, "rememberMe", String.valueOf(rememberMe));
        XMLUtils.createElement(contentHandler, "useCaptcha", String.valueOf(captcha));
        XMLUtils.createElement(contentHandler, "allowLoginByEmail", String.valueOf(allowLoginByEmail));
        XMLUtils.createElement(contentHandler, "useMultifactorAuthentication", String.valueOf(useMultifactorAuthentication));
        XMLUtils.createElement(contentHandler, "showErrors", String.valueOf(showErrors));
        
        _generateMultifactorAuthentication(request);

        XMLUtils.endElement(contentHandler, "LoginForm");
    }
    
    private void _generateMultifactorAuthentication(Request request) throws SAXException
    {
        String currentUserAsString = request.getParameter(FormCredentialProvider.LOGIN_URL_USER_PARAMETER_NAME);
        UserIdentity currentUser = UserIdentity.stringToUserIdentity(currentUserAsString);
        if (currentUser == null)
        {
            return;
        }
        
        Optional<String> email = _getUserEmail(currentUser);
        if (email.isEmpty())
        {
            // Do not display this part of the form if the current user does not have email address
            return;
        }
        
        String mfaMode = request.getParameter(FormCredentialProvider.LOGIN_URL_MFA_MODE);
        boolean invalidMFACode = "true".equals(request.getParameter(FormCredentialProvider.LOGIN_URL_INVALID_PASSCODE_PARAMETER_NAME));
        
        AttributesImpl attrs = new AttributesImpl();
        attrs.addCDATAAttribute("invalid", Boolean.toString(invalidMFACode));
        attrs.addCDATAAttribute("email", email.get());
        attrs.addCDATAAttribute("duration", Integer.toString(_multifactorAuthenticationManager.getEmailCodeDuration()));
        attrs.addCDATAAttribute("mode", mfaMode);
        XMLUtils.createElement(contentHandler, "MultifactorAuthentication", attrs);
    }
    
    private Optional<String> _getUserEmail(UserIdentity currentUserIdentity)
    {
        User currentUser = _userManager.getUser(currentUserIdentity);
        String email = currentUser.getEmail();
        
        if (StringUtils.isNotEmpty(email))
        {
            int at = email.indexOf("@");
            String escapedEmail = at > 2 ? email.substring(0, 2) + email.substring(2, at).replaceAll(".", "*") + email.substring(at) : email; 
            return Optional.of(escapedEmail);
        }
        else
        {
            return Optional.empty();
        }
    }
}
