/*
 *  Copyright 2022 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.authentication;

import java.io.IOException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.ZonedDateTime;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.LifecycleHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.util.log.SLF4JLoggerAdapter;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.datasource.dbtype.SQLDatabaseTypeExtensionPoint;
import org.ametys.core.ui.Callable;
import org.ametys.core.ui.mail.StandardMailBodyHelper;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.user.population.UserPopulation;
import org.ametys.core.util.CryptoHelper;
import org.ametys.core.util.I18nUtils;
import org.ametys.core.util.mail.SendMailHelper;
import org.ametys.plugins.core.impl.authentication.FormCredentialProvider;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

import dev.samstevens.totp.code.CodeGenerator;
import dev.samstevens.totp.code.DefaultCodeGenerator;
import dev.samstevens.totp.code.DefaultCodeVerifier;
import dev.samstevens.totp.exceptions.CodeGenerationException;
import dev.samstevens.totp.qr.QrData;
import dev.samstevens.totp.secret.DefaultSecretGenerator;
import dev.samstevens.totp.secret.SecretGenerator;
import dev.samstevens.totp.time.SystemTimeProvider;
import dev.samstevens.totp.time.TimeProvider;
import jakarta.mail.MessagingException;

/**
 * The component to handle Multifactor authentication
 * Provides methods to generate, send and check MFA codes
 */
public class MultifactorAuthenticationManager extends AbstractLogEnabled implements Component, Serviceable, Initializable, Contextualizable, Disposable
{
    /** Avalon Role */
    public static final String ROLE = MultifactorAuthenticationManager.class.getName();
    
    private static final String __SQL_TABLE_NAME = "Authentication_MFA";
    
    private static final String __QRCODE_ISSUER_KEY = "APPLICATION_PRODUCT_LABEL";
    private static final String __QRCODE_ISSUER_CATALOG = "application";
    
    private static final String __SEND_CODE_MAIL_SUBJECT_DEFAULT_PREFIX_KEY = "PLUGINS_CORE_MULTIFACTOR_AUTHENTICATION_MAIL_SUBJECT_DEFAULT_PREFIX";
    private static final String __SEND_CODE_MAIL_SUBJECT_KEY = "PLUGINS_CORE_MULTIFACTOR_AUTHENTICATION_MAIL_SUBJECT";
    private static final String __SEND_CODE_MAIL_BODY_TITLE_KEY = "PLUGINS_CORE_MULTIFACTOR_AUTHENTICATION_MAIL_BODY_TITLE";
    private static final String __SEND_CODE_MAIL_BODY_KEY = "PLUGINS_CORE_MULTIFACTOR_AUTHENTICATION_MAIL_BODY";
    
    private static final int __BUCKET_PERIOD_IN_SECONDS = 30;
    private static final int __ALLOWED_EMAIL_TIME_PERIOD_DISCREPANCY = 10;
    
    /** The service manager */
    protected ServiceManager _serviceManager;
    /** The coccon context */
    protected Context _context;
    
    /** I18n utils */
    protected I18nUtils _i18nUtils;
    /** The user manager */
    protected UserManager _userManager;
    /** The extension point for SQL database types */
    protected SQLDatabaseTypeExtensionPoint _sqlDatabaseTypeExtensionPoint;
    /** The current user provider */
    protected CurrentUserProvider _currentUserProvider;
    /** The crypto helpers by user population */
    protected Map<String, MultifactorAuthenticationCryptoHelper> _cryptoHelpers = new HashMap<>();
    
    /** The id of the data source storing the user's secrets */ 
    protected String _datasourceId;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _serviceManager = manager;
        
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _sqlDatabaseTypeExtensionPoint = (SQLDatabaseTypeExtensionPoint) manager.lookup(SQLDatabaseTypeExtensionPoint.ROLE);
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
    }
    
    public void initialize() throws Exception
    {
        _datasourceId = Config.getInstance() != null ? Config.getInstance().getValue("runtime.assignments.multifactorauthentication") : "";
    }
    
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    public void dispose()
    {
        _cryptoHelpers.clear();
    }
    
    /**
     * Email code validity duration
     * @return The number of seconds
     */
    public int getEmailCodeDuration() 
    {
        return __ALLOWED_EMAIL_TIME_PERIOD_DISCREPANCY * __BUCKET_PERIOD_IN_SECONDS;
    }
    
    /**
     * Initialize the {@link MultifactorAuthenticationCryptoHelper} for the given user population
     * A key file will be generated for secret encrypting/decrypting
     * @param userPopulation the user population
     * @throws IllegalStateException If an error occurs during helper setup
     */
    public void initializeMFACryptoComponent(UserPopulation userPopulation) throws IllegalStateException
    {
        if (!_cryptoHelpers.containsKey(userPopulation.getId()))
        {
            try
            {
                // If the user population has a credential provider with multifactor authentication,
                // Create an instance of CryptoHelper and generate a key file to encrypt/decrypt user's secret
                if (userPopulation.getCredentialProviders().stream()
                                       .filter(FormCredentialProvider.class::isInstance)
                                       .map(FormCredentialProvider.class::cast)
                                       .anyMatch(FormCredentialProvider::useMultifactorAuthentication))
                {
                    Configuration helperConfig = _getCryptoHelperConfiguration(userPopulation);
                    MultifactorAuthenticationCryptoHelper cryptoHelper = new MultifactorAuthenticationCryptoHelper();
                    LifecycleHelper.setupComponent(cryptoHelper, new SLF4JLoggerAdapter(getLogger()), _context, _serviceManager, helperConfig);
                    _cryptoHelpers.put(userPopulation.getId(), cryptoHelper);
                }
            }
            catch (Exception e)
            {
                throw new IllegalStateException("An error occured during the initialization of the multifactor authentication crypto component for UserPopulation '" + userPopulation.getId() + "'.", e);
            }
        }
    }
    
    private Configuration _getCryptoHelperConfiguration(UserPopulation userPopulation)
    {
        DefaultConfiguration helperConfig = new DefaultConfiguration("mfaCryptoHelper");
        
        DefaultConfiguration userPopulationConfig = new DefaultConfiguration("userPopulationId");
        userPopulationConfig.setValue(userPopulation.getId());
        helperConfig.addChild(userPopulationConfig);
        
        return helperConfig;
    }
    
    /**
     * Generate the multifactor authentication code
     * @param userIdentity the user trying to connect
     * @return The generated the multifactor authentication code, or <code>null</code> if the code couldn't have bee generated
     * @throws RuntimeException if an error occurs while getting or generating the secret or generating the MFA code
     */
    public MultifactorAuthenticationCode  generateMultifactorAuthenticationCode(UserIdentity userIdentity) throws RuntimeException
    {
        // Check the datasource existence. The datasource id can be empty if the config parameter is not valued yet 
        if (StringUtils.isEmpty(_datasourceId))
        {
            return null;
        }
        
        String secret = _getUserSecret(userIdentity);
        try
        {
            ZonedDateTime now = ZonedDateTime.now();
            long currentBucket = Math.floorDiv(now.toEpochSecond(), __BUCKET_PERIOD_IN_SECONDS);
            
            CodeGenerator codeGenerator = new DefaultCodeGenerator();
            String code = codeGenerator.generate(secret, currentBucket);
            
            ZonedDateTime expirationDate = now.plusSeconds(getEmailCodeDuration());
            return new MultifactorAuthenticationCode(code, expirationDate);
        }
        catch (CodeGenerationException e)
        {
            throw new RuntimeException("Unable to generate the multifactor authentication code for user " + UserIdentity.userIdentityToString(userIdentity) + ".", e);
        }
    }
    
    /**
     * Check if the current user has activated multifactor authentication by application and gets its secret
     * @return A map with key "active" <code>true</code> if the current user has activated multifactor authentication by application, <code>false</code> otherwise, and key "secret" with its secret
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> getUserSecretForCurrentUser()
    {
        UserIdentity userIdentity = _currentUserProvider.getUser();
        return Map.of("active", isAuthenticationApplicationActivated(userIdentity), "secret", _getUserSecret(userIdentity));

    }
    
    /**
     * Creates a new secret for the current user
     * @return The new secret
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public String renewSecretForCurrentUser()
    {
        UserIdentity userIdentity = _currentUserProvider.getUser();
        return renewSecret(userIdentity);
    }
    
    /**
     * Check if the given user has activated multifactor authentication by application
     * @param userIdentity the user identity to check
     * @return <code>true</code> if the given user has activated multifactor authentication by application, <code>false</code> otherwise
     */
    public boolean isAuthenticationApplicationActivated(UserIdentity userIdentity)
    {
        Connection connection = null;
        try
        {
            connection = ConnectionHelper.getConnection(_datasourceId);
            
            try (PreparedStatement statement = _getSelectUseApplicationStatement(connection, userIdentity);
                 ResultSet resultSet = statement.executeQuery();)
            {
                if (resultSet.next())
                {
                    return resultSet.getBoolean("use_application");
                }
            }
            
            return false;
        }
        catch (SQLException e)
        {
            throw new RuntimeException("Communication error with the database. Unable to retrieve the multifactor authentication state for user " + UserIdentity.userIdentityToString(userIdentity) + ".", e);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
        }
    }
    
    private PreparedStatement _getSelectUseApplicationStatement(Connection connection, UserIdentity user) throws SQLException
    {
        String dbType = ConnectionHelper.getDatabaseType(connection);
        String sqlRequest = "SELECT use_application FROM " + _sqlDatabaseTypeExtensionPoint.languageEscapeTableName(dbType, __SQL_TABLE_NAME) + " WHERE login=? AND population_id=?";

        PreparedStatement statement = connection.prepareStatement(sqlRequest);
        statement.setString(1, user.getLogin());
        statement.setString(2, user.getPopulationId());

        return statement;
    }
    
    /**
     * Activate/deactivate the multifactor authentication by application for the current user.
     * @param value True to activate. False to deactivate
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public void authenticationApplicationForCurrentUser(boolean value)
    {
        UserIdentity userIdentity = _currentUserProvider.getUser();
        _storeAuthenticationActivationInDataSource(userIdentity, value);
    }
    
    private void _storeAuthenticationActivationInDataSource(UserIdentity userIdentity, boolean useApplication) throws RuntimeException
    {
        Connection connection = null;
        try
        {
            connection = ConnectionHelper.getConnection(_datasourceId);
            String dbType = ConnectionHelper.getDatabaseType(connection);
            PreparedStatement statement = connection.prepareStatement("UPDATE " + _sqlDatabaseTypeExtensionPoint.languageEscapeTableName(dbType, __SQL_TABLE_NAME) + " SET use_application=? WHERE login=? AND population_id=?");

            statement.setBoolean(1, useApplication);
            statement.setString(2, userIdentity.getLogin());
            statement.setString(3, userIdentity.getPopulationId());
    
            statement.executeUpdate();
        }
        catch (SQLException e)
        {
            throw new RuntimeException("Communication error with the database. Unable to store the generated secret for multifactor authentication for user " + UserIdentity.userIdentityToString(userIdentity) + ".", e);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
        }
    }
    
    /**
     * Retrieves the QrData containing the current user secret
     * @return the QrData for the current user
     */
    public QrData getCurrentUserQrData()
    {
        UserIdentity userIdentity = _currentUserProvider.getUser();
        String secret = _getUserSecret(userIdentity);
        I18nizableText i18nIssuer = new I18nizableText(__QRCODE_ISSUER_CATALOG, __QRCODE_ISSUER_KEY);
        String issuer = _i18nUtils.translate(i18nIssuer);
        return new QrData.Builder()
                .secret(secret)
                .label(userIdentity.getLogin())
                .issuer(issuer)
                .period(__BUCKET_PERIOD_IN_SECONDS)
                .build();
    }
    
    private String _getUserSecret(UserIdentity userIdentity)
    {
        return _getSecretFromDatasourceIfExists(userIdentity)
                .orElseGet(() -> _generateSecret(userIdentity));
    }
    
    /**
     * Check if the given code is valid
     * @param userIdentity the user trying to connect
     * @param multifactorAuthenticationCode the code submitted by the user 
     * @return <code>true</code> if the given code is valid, <code>false</code> otherwise
     * @throws RuntimeException if an error occurs while getting the secret from the datasource
     */
    public boolean isValidMultifactorAuthenticationCode(UserIdentity userIdentity, String multifactorAuthenticationCode) throws RuntimeException
    {
        Optional<String> optionalSecret = _getSecretFromDatasourceIfExists(userIdentity);
        if (optionalSecret.isPresent())
        {
            TimeProvider timeProvider = new SystemTimeProvider();
            CodeGenerator codeGenerator = new DefaultCodeGenerator();
            DefaultCodeVerifier verifier = new DefaultCodeVerifier(codeGenerator, timeProvider);
            verifier.setTimePeriod(__BUCKET_PERIOD_IN_SECONDS);
            verifier.setAllowedTimePeriodDiscrepancy(__ALLOWED_EMAIL_TIME_PERIOD_DISCREPANCY);

            String secret = optionalSecret.get();
            return verifier.isValidCode(secret, multifactorAuthenticationCode);
        }
        else
        {
            // Should never happen. There is no way to be here without having generated a secret for the given user
            return false;
        }
    }
    
    private Optional<String> _getSecretFromDatasourceIfExists(UserIdentity userIdentity) throws RuntimeException
    {
        Connection connection = null;
        try
        {
            connection = ConnectionHelper.getConnection(_datasourceId);
            
            try (PreparedStatement statement = _getSelectSecretStatement(connection, userIdentity);
                 ResultSet resultSet = statement.executeQuery();)
            {
                if (resultSet.next())
                {
                    String cryptedSecret = resultSet.getString("secret");
                    MultifactorAuthenticationCryptoHelper cryptoHelper = _cryptoHelpers.get(userIdentity.getPopulationId());
                    String secret = cryptoHelper.decrypt(cryptedSecret);
                    return Optional.of(secret);
                }
            }
            
            return Optional.empty();
        }
        catch (CryptoHelper.WrongKeyException e)
        {
            throw new RuntimeException("Secret of user " + UserIdentity.userIdentityToString(userIdentity) + " cannot be decrypted. Unable to retrieve the multifactor authentication secret.", e);
        }
        catch (SQLException e)
        {
            throw new RuntimeException("Communication error with the database. Unable to retrieve the multifactor authentication secret for user " + UserIdentity.userIdentityToString(userIdentity) + ".", e);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
        }
    }
    
    /**
     * Creates a new secret for the user
     * @param userIdentity The user to change
     * @return The new secret
     */
    public String renewSecret(UserIdentity userIdentity)
    {
        Connection connection = null;
        try
        {
            connection = ConnectionHelper.getConnection(_datasourceId);
            
            SecretGenerator secretGenerator = new DefaultSecretGenerator();
            String newSecret = secretGenerator.generate();
            
            try (PreparedStatement statement = _getUpdateSecretStatement(connection, userIdentity, newSecret))
            {
                if (statement.executeUpdate() == 1)
                {
                    return newSecret;
                }
            }
            
            throw new RuntimeException("Secret of user " + UserIdentity.userIdentityToString(userIdentity) + " cannot be updated"); 
        }
        catch (SQLException e)
        {
            throw new RuntimeException("Communication error with the database. Unable to retrieve the multifactor authentication secret for user " + UserIdentity.userIdentityToString(userIdentity) + ".", e);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
        }   
    }    
    
    private PreparedStatement _getUpdateSecretStatement(Connection connection, UserIdentity user, String newSecret) throws SQLException
    {
        String dbType = ConnectionHelper.getDatabaseType(connection);
        String sqlRequest = "UPDATE " + _sqlDatabaseTypeExtensionPoint.languageEscapeTableName(dbType, __SQL_TABLE_NAME) + " SET secret=? WHERE login=? AND population_id=?"; 

        MultifactorAuthenticationCryptoHelper cryptoHelper = _cryptoHelpers.get(user.getPopulationId());
        String cryptedSecret = cryptoHelper.encrypt(newSecret);
        
        PreparedStatement statement = connection.prepareStatement(sqlRequest);
        statement.setString(1, cryptedSecret);
        statement.setString(2, user.getLogin());
        statement.setString(3, user.getPopulationId());

        return statement;
    }
    
    
    private PreparedStatement _getSelectSecretStatement(Connection connection, UserIdentity user) throws SQLException
    {
        String dbType = ConnectionHelper.getDatabaseType(connection);
        String sqlRequest = "SELECT secret FROM " + _sqlDatabaseTypeExtensionPoint.languageEscapeTableName(dbType, __SQL_TABLE_NAME) + " WHERE login=? AND population_id=?";

        PreparedStatement statement = connection.prepareStatement(sqlRequest);
        statement.setString(1, user.getLogin());
        statement.setString(2, user.getPopulationId());

        return statement;
    }
    
    private String _generateSecret(UserIdentity userIdentity) throws RuntimeException
    {
        SecretGenerator secretGenerator = new DefaultSecretGenerator();
        String secret = secretGenerator.generate();
        
        _storeSecretInDataSource(userIdentity, secret);
        
        return secret;
    }
    
    private void _storeSecretInDataSource(UserIdentity userIdentity, String secret) throws RuntimeException
    {
        Connection connection = null;
        try
        {
            connection = ConnectionHelper.getConnection(_datasourceId);
            String dbType = ConnectionHelper.getDatabaseType(connection);

            PreparedStatement statement = connection.prepareStatement(
                    "INSERT INTO " + _sqlDatabaseTypeExtensionPoint.languageEscapeTableName(dbType, __SQL_TABLE_NAME) + " (login, population_id, secret) VALUES (?, ?, ?)");
            
            statement.setString(1, userIdentity.getLogin());
            statement.setString(2, userIdentity.getPopulationId());

            MultifactorAuthenticationCryptoHelper cryptoHelper = _cryptoHelpers.get(userIdentity.getPopulationId());
            String cryptedSecret = cryptoHelper.encrypt(secret);
            statement.setString(3, cryptedSecret);
    
            statement.executeUpdate();
        }
        catch (SQLException e)
        {
            throw new RuntimeException("Communication error with the database. Unable to store the generated secret for multifactor authentication for user " + UserIdentity.userIdentityToString(userIdentity) + ".", e);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
        }
    }
    
    /**
     * Record to get MFA code and its expiration date
     * @param code the multifactor authentication code
     * @param expirationDate the expiration date of the code
     *
     */
    public record MultifactorAuthenticationCode(String code, ZonedDateTime expirationDate) { /* empty */ }
    
    /**
     * Generate a MFA code and send it by mail to the given user 
     * @param request the request
     * @param userIdentity the user trying to connect
     * @param multifactorAuthenticationCode the multifactor authentication code to send
     * @throws RuntimeException if an error occurs while sending the mail with the code
     */
    public void sendMultifactorAuthenticationCodeByMail(Request request, UserIdentity userIdentity, String multifactorAuthenticationCode) throws RuntimeException
    {
        try
        {
            String i18nCatalog = "plugin.core-impl";
            
            String allowedTimePeriodInMinutes = String.valueOf(getEmailCodeDuration() / 60);
            String subjectPrefix = _i18nUtils.translate(_getMailSubjectPrefix(request));
            
            List<String> params = List.of(subjectPrefix, multifactorAuthenticationCode, allowedTimePeriodInMinutes);
            
            I18nizableText subjectWithParams = new I18nizableText(i18nCatalog, __SEND_CODE_MAIL_SUBJECT_KEY, params);
            String subject = _i18nUtils.translate(subjectWithParams);
            
            String body = StandardMailBodyHelper.newHTMLBody()
                .withTitle(new I18nizableText(i18nCatalog, __SEND_CODE_MAIL_BODY_TITLE_KEY))
                .withMessage(new I18nizableText(i18nCatalog, __SEND_CODE_MAIL_BODY_KEY, params))
                .withFooterBottomText(StandardMailBodyHelper.AUTOGENERATED_DO_NOT_ANSWER_TEXT)
                .build();
            
            User user = _userManager.getUser(userIdentity);
            String recipient = user.getEmail();
            
            if (StringUtils.isEmpty(recipient))
            {
                // This should never happen. This method should not be called with a user without an address email
                throw new RuntimeException("Unable to send the multifactor authentication code to user " + UserIdentity.userIdentityToString(userIdentity) + ". This user has no email address");
            }
            
            SendMailHelper.newMail()
                .withSubject(subject)
                .withHTMLBody(body)
                .withRecipient(recipient)
                .sendMail();
        }
        catch (MessagingException | IOException e)
        {
            throw new RuntimeException("Unable to send the multifactor authentication code to user " + UserIdentity.userIdentityToString(userIdentity) + ".", e);
        }
    }
    
    /**
     * Retrieves the prefix of mail's subject 
     * @param request the request
     * @return the prefix of mail's subject
     */
    protected I18nizableText _getMailSubjectPrefix(Request request)
    {
        return new I18nizableText("plugin.core-impl", __SEND_CODE_MAIL_SUBJECT_DEFAULT_PREFIX_KEY);
    }
}
