/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.group;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.core.group.Group;
import org.ametys.core.group.GroupIdentity;
import org.ametys.core.group.GroupManager;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Helper class for groups, for common operations on {@link Group}
 */
public class GroupHelper extends AbstractLogEnabled implements Component, Serviceable
{
    /** The Avalon role */
    public static final String ROLE = GroupHelper.class.getName();
    private GroupManager _groupManager;
    private UserHelper _userHelper;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _groupManager = (GroupManager) manager.lookup(GroupManager.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
    }
    
    /**
     * Sax a Group identity
     * @param groupIdentity the group identity to sax
     * @param handler the content handler
     * @throws SAXException If a SAX error occurs
     */
    public void saxGroupIdentity(GroupIdentity groupIdentity, ContentHandler handler) throws SAXException
    {
        saxGroupIdentity(groupIdentity, handler, "group", false);
    }
    
    /**
     * Sax a Group identity
     * @param groupIdentity the group identity to sax
     * @param handler the content handler
     * @param tagName the XML tag for saxed group
     * @throws SAXException If a SAX error occurs
     */
    public void saxGroupIdentity(GroupIdentity groupIdentity, ContentHandler handler, String tagName) throws SAXException
    {
        saxGroupIdentity(groupIdentity, handler, tagName, false);
    }
    
    /**
     * Sax a Group identity
     * @param groupIdentity the group identity to sax
     * @param handler the content handler
     * @param tagName the XML tag for saxed group
     * @param withUsers if true, users of the group will be saxed with the group
     * @throws SAXException If a SAX error occurs
     */
    public void saxGroupIdentity(GroupIdentity groupIdentity, ContentHandler handler, String tagName, boolean withUsers) throws SAXException
    {
        
        Group group = _groupManager.getGroup(groupIdentity);
        if (group != null)
        {
            saxGroup(group, handler, tagName, withUsers);
        }
        else
        {
            getLogger().warn("Unable to sax unknown group with identity {}", groupIdentity);
        }
    }
    
    /**
     * Sax a Group
     * @param group the group to sax
     * @param handler the content handler
     * @throws SAXException If a SAX error occurs
     */
    public void saxGroup(Group group, ContentHandler handler) throws SAXException
    {
        saxGroup(group, handler, "group", false);
    }
    
    /**
     * Sax a Group
     * @param group the group to sax
     * @param handler the content handler
     * @param tagName the XML tag for saxed group
     * @throws SAXException If a SAX error occurs
     */
    public void saxGroup(Group group, ContentHandler handler, String tagName) throws SAXException
    {
        saxGroup(group, handler, tagName, false);
    }
    
    /**
     * Sax a Group
     * @param group the group to sax
     * @param handler the content handler
     * @param tagName the XML tag for saxed group
     * @param withUsers if true, users of the group will be saxed with the group
     * @throws SAXException If a SAX error occurs
     */
    public void saxGroup(Group group, ContentHandler handler, String tagName, boolean withUsers) throws SAXException
    {
        if (group != null)
        {
            AttributesImpl attr = new AttributesImpl();
            attr.addCDATAAttribute("id", group.getIdentity().getId());
            attr.addCDATAAttribute("groupDirectory", group.getIdentity().getDirectoryId());
            
            XMLUtils.startElement(handler, tagName, attr);
            
            XMLUtils.createElement(handler, "label", group.getLabel());
            group.getGroupDirectory().getLabel().toSAX(handler, "groupDirectoryLabel");
            
            if (withUsers)
            {
                XMLUtils.startElement(handler, "users");
                for (UserIdentity user: group.getUsers())
                {
                    _userHelper.saxUserIdentity(user, handler);
                }
                XMLUtils.endElement(handler, "users");
            }
    
            XMLUtils.endElement(handler, tagName);
        }
        else if (getLogger().isWarnEnabled())
        {
            getLogger().warn("Unable to sax null group", new Exception());
        }
    }
    
    /**
     * Get group as JSON object
     * @param groupIdentity the group identity
     * @param users true to get users' group
     * @return the group as JSON object
     */
    public Map<String, Object> group2JSON(GroupIdentity groupIdentity, boolean users)
    {
        return group2JSON(_groupManager.getGroup(groupIdentity), users);
    }
    
    /**
     * Get group as JSON object
     * @param group the group
     * @param users true to get users' group
     * @return the group as JSON object
     */
    public Map<String, Object> group2JSON(Group group, boolean users)
    {
        if (group == null)
        {
            return null;
        }
        
        Map<String, Object> group2json = new HashMap<>();
        group2json.put("id", group.getIdentity().getId());
        group2json.put("groupDirectory", group.getIdentity().getDirectoryId());
        group2json.put("groupDirectoryLabel", group.getGroupDirectory().getLabel());
        group2json.put("label", group.getLabel());
        if (users)
        {
            group2json.put("users", group.getUsers());
        }
        return group2json;
    }
    

    /**
     * Get groups as JSON object
     * @param groups the groups
     * @param users true to get users' group
     * @return the group as JSON object
     */
    public List<Map<String, Object>> groups2JSON(List<Group> groups, boolean users)
    {
        return groups.stream()
            .map(group -> this.group2JSON(group, users))
            .collect(Collectors.toList());
    }
}
