/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.schedule;

import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Properties;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.activity.Disposable;
import org.apache.avalon.framework.activity.Initializable;
import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.SourceResolver;
import org.quartz.CronScheduleBuilder;
import org.quartz.CronTrigger;
import org.quartz.JobBuilder;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobExecutionContext;
import org.quartz.JobKey;
import org.quartz.SchedulerException;
import org.quartz.Trigger;
import org.quartz.Trigger.TriggerState;
import org.quartz.TriggerBuilder;
import org.quartz.TriggerKey;
import org.quartz.impl.StdSchedulerFactory;
import org.quartz.impl.matchers.GroupMatcher;
import org.quartz.utils.PoolingConnectionProvider;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.datasource.SQLDataSourceManager;
import org.ametys.core.datasource.dbtype.SQLDatabaseType;
import org.ametys.core.datasource.dbtype.SQLDatabaseTypeExtensionPoint;
import org.ametys.core.right.RightManager;
import org.ametys.core.schedule.AmetysJob;
import org.ametys.core.schedule.Runnable;
import org.ametys.core.schedule.Runnable.FireProcess;
import org.ametys.core.schedule.RunnableExtensionPoint;
import org.ametys.core.schedule.Schedulable;
import org.ametys.core.schedule.SchedulableExtensionPoint;
import org.ametys.core.schedule.progression.ProgressionTracker;
import org.ametys.core.script.SQLScriptHelper;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.util.DateUtils;
import org.ametys.core.util.LambdaUtils;
import org.ametys.plugins.core.impl.schedule.DefaultRunnable;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * The scheduler component
 */
public class Scheduler extends AbstractLogEnabled implements Component, Initializable, Disposable, Serviceable, Contextualizable
{
    /** The Avalon Role */
    public static final String ROLE = Scheduler.class.getName();
    
    /** The group name for jobs */
    public static final String JOB_GROUP = "runtime.job";
    /** The group name for triggers */
    public static final String TRIGGER_GROUP = "runtime.trigger";
    /** The key for the id of the schedulable to execute */
    public static final String KEY_SCHEDULABLE_ID = "schedulableId";
    /** The key for the runnable id */
    public static final String KEY_RUNNABLE_ID = "id";
    /** The key for the runnable label */
    public static final String KEY_RUNNABLE_LABEL = "label";
    /** The key for the runnable description */
    public static final String KEY_RUNNABLE_DESCRIPTION = "description";
    /** The key for the runnable fire process property */
    public static final String KEY_RUNNABLE_FIRE_PROCESS = "fireProcess";
    /** The key for 'run at startup' jobs indicating if the job has already been executed and is now completed */
    public static final String KEY_RUNNABLE_STARTUP_COMPLETED = "runAtStartupCompleted";
    /** The key for the runnable cron expression */
    public static final String KEY_RUNNABLE_CRON = "cron";
    /** The key for the runnable removable property */
    public static final String KEY_RUNNABLE_REMOVABLE = "removable";
    /** The key for the runnable modifiable property */
    public static final String KEY_RUNNABLE_MODIFIABLE = "modifiable";
    /** The key for the runnable deactivatable property */
    public static final String KEY_RUNNABLE_DEACTIVATABLE = "deactivatable";
    /** The key for the runnable volatile property */
    public static final String KEY_RUNNABLE_VOLATILE = "volatile";
    /** The key to retrieve the {@link UserIdentity} */
    public static final String KEY_RUNNABLE_USERIDENTITY = "userIdentity";
    /** The prefix for the parameter values of the runnable job in the job data map */
    public static final String PARAM_VALUES_PREFIX = "parameterValues#";
    /** Name of the parameter holding the datasource id for Quartz */
    public static final String DATASOURCE_CONFIG_NAME = "runtime.scheduler.datasource";
    
    /** The name of the configuration file for Quartz */
    private static final String __QUARTZ_CONFIG_FILE_NAME = "quartz.properties";
    /** The id of the right to execute actions on tasks */
    private static final String __RIGHT_SCHEDULER = "CORE_Rights_TaskScheduler";
    
    /** The service manager */
    protected ServiceManager _manager;
    /** The context */
    protected Context _context;
    /** The extension point for {@link Runnable}s */
    protected RunnableExtensionPoint _runnableEP;
    /** The extension point for {@link Schedulable}s */
    protected SchedulableExtensionPoint _schedulableEP;
    /** The Quartz scheduler */
    protected org.quartz.Scheduler _scheduler;
    /** The manager for SQL datasources */
    protected SQLDataSourceManager _sqlDataSourceManager;
    /** The source resolver */
    protected SourceResolver _sourceResolver;
    /** The rights manager */
    protected RightManager _rightManager;
    /** The provider of current user */
    protected CurrentUserProvider _currentUserProvider;
    /** The sql database type ep */
    protected SQLDatabaseTypeExtensionPoint _sqlDatabaseTypeEP;
    /** The user manager */
    protected UserManager _userManager;
    /** The user helper */
    protected UserHelper _userHelper;

    @Override
    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    @Override
    public void service(ServiceManager manager) throws ServiceException
    {
        _manager = manager;
        _runnableEP = (RunnableExtensionPoint) manager.lookup(RunnableExtensionPoint.ROLE);
        _schedulableEP = (SchedulableExtensionPoint) manager.lookup(SchedulableExtensionPoint.ROLE);
        _sqlDataSourceManager = (SQLDataSourceManager) manager.lookup(SQLDataSourceManager.ROLE);
        _sqlDatabaseTypeEP = (SQLDatabaseTypeExtensionPoint) manager.lookup(SQLDatabaseTypeExtensionPoint.ROLE);
        _sourceResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _rightManager = (RightManager) manager.lookup(RightManager.ROLE);
        _currentUserProvider = (CurrentUserProvider) manager.lookup(CurrentUserProvider.ROLE);
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
        
        // Ensure statics methods will be available during initialize
        manager.lookup(ConnectionHelper.ROLE);
    }
    
    @Override
    public void initialize() throws Exception
    {
        // Be sure the tables are created
        String dsId = Config.getInstance().getValue(DATASOURCE_CONFIG_NAME);
        _checkAndCreateTables(dsId);
        
        // Get info about datasource
        Map<String, Object> dsParameters = _sqlDataSourceManager.getDataSourceDefinition(dsId).getParameters();
        String url = (String) dsParameters.get("url");
        String user = (String) dsParameters.get("user");
        String password = (String) dsParameters.get("password");
        String dbtype = (String) dsParameters.get("dbtype");
        SQLDatabaseType sqlDbType = _sqlDatabaseTypeEP.getExtension(dbtype);
        
        String dbType = ConnectionHelper.getDatabaseType(url);
        String dsName = "quartzDb";
        
        // Set the properties for Quartz
        Properties props = new Properties();
        props.load(Scheduler.class.getResourceAsStream(__QUARTZ_CONFIG_FILE_NAME));
        props.setProperty("org.quartz.jobStore.dataSource", dsName);
        props.setProperty(StdSchedulerFactory.PROP_DATASOURCE_PREFIX + "." + dsName + "." + PoolingConnectionProvider.DB_DRIVER, sqlDbType.getDriver());
        props.setProperty(StdSchedulerFactory.PROP_DATASOURCE_PREFIX + "." + dsName + "." + PoolingConnectionProvider.DB_URL, url);
        props.setProperty(StdSchedulerFactory.PROP_DATASOURCE_PREFIX + "." + dsName + "." + PoolingConnectionProvider.DB_USER, user);
        props.setProperty(StdSchedulerFactory.PROP_DATASOURCE_PREFIX + "." + dsName + "." + PoolingConnectionProvider.DB_PASSWORD, password);
        props.setProperty("org.quartz.jobStore.driverDelegateClass", _getDriverDelegateClass(dbType));
        StdSchedulerFactory factory = new StdSchedulerFactory(props);
        _scheduler = factory.getScheduler();
        
        // Initialize AmetysJob class
        AmetysJob.initialize(_manager, _context);
    }
    
    /* copy/paste of SqlTablesInit#init() because SqlTablesInit comes too late */
    private void _checkAndCreateTables(String dataSourceId)
    {
        // Test and create tables
        try
        {
            SQLScriptHelper.createTableIfNotExists(dataSourceId, "QRTZ_JOB_DETAILS", "plugin:core://scripts/%s/quartz.sql", _sourceResolver);
        }
        catch (Exception e)
        {
            getLogger().error("Error during SQL tables initialization for data source id: '{}'.", StringUtils.defaultString(dataSourceId), e);
        }
    }
    
    private String _getDriverDelegateClass(String dbType)
    {
        switch (dbType)
        {
            case ConnectionHelper.DATABASE_HSQLDB:
                return "org.quartz.impl.jdbcjobstore.HSQLDBDelegate";
            case ConnectionHelper.DATABASE_POSTGRES:
                return "org.quartz.impl.jdbcjobstore.PostgreSQLDelegate";
            default:
                return "org.quartz.impl.jdbcjobstore.StdJDBCDelegate";
        }
    }
    
    /**
     * Starts the Quartz scheduler. Only call this method once !
     * @throws SchedulerException if an error occured
     */
    public void start() throws SchedulerException
    {
        _removeVolatileJobs();
        _scheduleConfigurableJobs();
        _triggerRunAtStartupJobs();
        
        getLogger().info("Start up the scheduler");
        _scheduler.start();
    }
    
    /**
     * Get the Quartz scheduler
     * @return the scheduler
     */
    public org.quartz.Scheduler getScheduler()
    {
        return _scheduler;
    }
    
    private void _removeVolatileJobs() throws SchedulerException
    {
        for (JobKey jobKey : getJobs())
        {
            JobDataMap jobDataMap = _scheduler.getJobDetail(jobKey).getJobDataMap();
            if (jobDataMap.getBoolean(KEY_RUNNABLE_VOLATILE))
            {
                _scheduler.deleteJob(jobKey);
            }
        }
    }
    
    private void _scheduleConfigurableJobs()
    {
        try
        {
            for (String runnableId : _runnableEP.getExtensionsIds())
            {
                JobKey jobKey = new JobKey(runnableId, JOB_GROUP);
                if (_scheduler.checkExists(jobKey))
                {
                    // Check if exists, but it should never do since the configurable are told volatile and were removed before
                    _scheduler.deleteJob(jobKey);
                }
                Runnable runnable = _runnableEP.getExtension(runnableId);
                
                scheduleJob(runnable);
            }
        }
        catch (SchedulerException e)
        {
            getLogger().error("An exception occured during the scheduling of configurable runnables", e);
        }
    }
    
    private void _triggerRunAtStartupJobs() throws SchedulerException
    {
        for (JobKey jobKey : getJobs())
        {
            JobDataMap jobDataMap = _scheduler.getJobDetail(jobKey).getJobDataMap();
            if (FireProcess.STARTUP.toString().equals(jobDataMap.getString(KEY_RUNNABLE_FIRE_PROCESS))
                    && !jobDataMap.getBoolean(KEY_RUNNABLE_STARTUP_COMPLETED))
            {
                _scheduler.triggerJob(jobKey);
            }
        }
    }
    
    /**
     * Schedules a job
     * @param runnable The runnable job to schedule
     * @throws SchedulerException if the Job or Trigger cannot be added to the Scheduler, or there is an internal Scheduler error.
     */
    public void scheduleJob(Runnable runnable) throws SchedulerException
    {
        String runnableId = runnable.getId();
        String schedulableId = runnable.getSchedulableId();
        
        // Schedule the job
        JobDetail jobDetail = null;
        JobBuilder jobBuilder = JobBuilder.newJob(AmetysJob.class)
                .withIdentity(runnableId, JOB_GROUP)
                .usingJobData(KEY_SCHEDULABLE_ID, schedulableId)
                .usingJobData(_runnableToJobDataMap(runnable));
        
        Date firstTime = null;
        switch (runnable.getFireProcess())
        {
            case NEVER:
                // will never fire
                jobDetail = jobBuilder.storeDurably().build();
                
                _scheduler.addJob(jobDetail, true);
                getLogger().info("{} has been scheduled to never run", jobDetail.getKey());
                break;
            case STARTUP:
                // will fire at next startup
                jobDetail = jobBuilder.storeDurably().build();
                
                _scheduler.addJob(jobDetail, true);
                getLogger().info("{} has been scheduled to run at next startup of the application", jobDetail.getKey());
                break;
            case NOW:
                // will be triggered now
                jobDetail = jobBuilder.storeDurably().build();
                
                Trigger simpleTrigger = TriggerBuilder.newTrigger()
                        .withIdentity(runnableId, TRIGGER_GROUP)
                        .startNow()
                        .build();
                
                firstTime = _scheduler.scheduleJob(jobDetail, simpleTrigger);
                getLogger().info("{} has been scheduled to run as soon as possible", jobDetail.getKey());
                break;
            case CRON:
            default:
                // based on a cron trigger
                jobDetail = jobBuilder.storeDurably().build();
                
                CronScheduleBuilder schedBuilder = CronScheduleBuilder.cronSchedule(runnable.getCronExpression());
                
                switch (runnable.getMisfirePolicy())
                {
                    case IGNORE:
                        schedBuilder.withMisfireHandlingInstructionIgnoreMisfires();
                        break;
                    case FIRE_ONCE:
                        schedBuilder.withMisfireHandlingInstructionFireAndProceed();
                        break;
                    case DO_NOTHING:
                    default:
                        schedBuilder.withMisfireHandlingInstructionDoNothing();
                        break;
                }
                CronTrigger cronTrigger = TriggerBuilder.newTrigger()
                        .withIdentity(runnableId, TRIGGER_GROUP)
                        .startNow()
                        .withSchedule(schedBuilder)
                        .build();
                
                firstTime = _scheduler.scheduleJob(jobDetail, cronTrigger);
                getLogger().info("{} has been scheduled to run at: {} and repeat based on expression: {}", jobDetail.getKey(), firstTime, cronTrigger.getCronExpression());
                break;
        }
        
    }
    
    private JobDataMap _runnableToJobDataMap(Runnable runnable)
    {
        Map<String, Object> result = new HashMap<>();
        result.put(KEY_RUNNABLE_ID, runnable.getId());
        result.put(KEY_RUNNABLE_LABEL, I18nizableText.i18nizableTextToString(runnable.getLabel()));
        result.put(KEY_RUNNABLE_DESCRIPTION, I18nizableText.i18nizableTextToString(runnable.getDescription()));
        result.put(KEY_RUNNABLE_FIRE_PROCESS, runnable.getFireProcess().toString());
        result.put(KEY_RUNNABLE_CRON, runnable.getCronExpression());
        result.put(KEY_RUNNABLE_REMOVABLE, runnable.isRemovable());
        result.put(KEY_RUNNABLE_MODIFIABLE, runnable.isModifiable());
        result.put(KEY_RUNNABLE_DEACTIVATABLE, runnable.isDeactivatable());
        result.put(KEY_RUNNABLE_VOLATILE, runnable.isVolatile());
        result.put(KEY_RUNNABLE_USERIDENTITY, UserIdentity.userIdentityToString(runnable.getUserIdentity()));
        
        // parameter values
        for (String paramId : runnable.getParameterValues().keySet())
        {
            result.put(PARAM_VALUES_PREFIX + paramId, runnable.getParameterValues().get(paramId));
        }
        
        return new JobDataMap(result);
    }
    
    /**
     * Gets the jobs of the Quartz scheduler
     * @return the jobs
     * @throws SchedulerException if an error occured
     */
    public Set<JobKey> getJobs() throws SchedulerException
    {
        return _scheduler.getJobKeys(GroupMatcher.jobGroupEquals(JOB_GROUP));
    }
    
    /**
     * Gets tasks information
     * @param taskIds The ids of the tasks
     * @return The tasks information
     * @throws SchedulerException if an error occurred
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public List<Map<String, Object>> getTasksInformation(List<String> taskIds) throws SchedulerException
    {
        List<Map<String, Object>> result = new ArrayList<>();
        
        for (JobKey jobKey : getJobs())
        {
            String id = jobKey.getName();
            if (taskIds.contains(id))
            {
                Map<String, Object> task = new HashMap<>();
                JobDataMap jobDataMap = _scheduler.getJobDetail(jobKey).getJobDataMap();
                task.put("id", id);
                task.put("schedulable", jobDataMap.getString(KEY_SCHEDULABLE_ID));
                task.put("modifiable", jobDataMap.getBoolean(KEY_RUNNABLE_MODIFIABLE));
                task.put("removable", jobDataMap.getBoolean(KEY_RUNNABLE_REMOVABLE));
                task.put("deactivatable", jobDataMap.getBoolean(KEY_RUNNABLE_DEACTIVATABLE));
                
                // Get information to compute job state
                boolean successed = jobDataMap.getBoolean(AmetysJob.KEY_SUCCESS);
                boolean isRunning = _isRunningJob(jobKey, _getExecutingJobs());
                TriggerKey triggerKey = new TriggerKey(jobKey.getName(), TRIGGER_GROUP);
                boolean triggerExists = _scheduler.checkExists(triggerKey);
                boolean completed = _isCompleted(triggerExists, jobDataMap);
                boolean enabled = _isEnabled(triggerKey, triggerExists, jobDataMap.getString(KEY_RUNNABLE_FIRE_PROCESS));
                task.put("state", _getJobState(completed, successed, isRunning, enabled));
                
                result.add(task);
            }
        }
        
        return result;
    }
    
    /**
     * Gets all the scheduled tasks
     * @return the scheduled tasks
     * @throws Exception if an error occured
     */
    public List<Map<String, Object>> getTasksAsJson() throws Exception
    {
        Map<JobKey, JobExecutionContext> executingJobs = _getExecutingJobs();
        
        return getJobs().stream()
                .map(LambdaUtils.wrap(key-> _jobToJson(key, executingJobs)))
                .filter(Objects::nonNull)
                .collect(Collectors.toList());
    }
    
    private Map<JobKey, JobExecutionContext> _getExecutingJobs() throws SchedulerException
    {
        return _scheduler.getCurrentlyExecutingJobs().stream()
                .collect(Collectors.toMap(
                            e -> e.getJobDetail().getKey(), 
                            e -> e));
    }
    
    private Map<String, Object> _jobToJson(JobKey jobKey, Map<JobKey, JobExecutionContext> executingJobs) throws Exception
    {
        Map<String, Object> result = new HashMap<>();
        
        JobDataMap jobDataMap = _scheduler.getJobDetail(jobKey).getJobDataMap();
        
        String id = jobDataMap.getString(KEY_RUNNABLE_ID);
        result.put("id", id);
        
        String schedulableId = jobDataMap.getString(KEY_SCHEDULABLE_ID);
        Schedulable schedulable;
        if (_schedulableEP.hasExtension(schedulableId))
        {
            schedulable = _schedulableEP.getExtension(schedulableId);
        }
        else
        {
            // Abort trying to JSONify this job, as its associated schedulable does not exist anymore
            // It will not be displayed in the UI tool
            getLogger().warn("The Runnable '{}' is associated to a Schedulable that does not exist anymore ({}).", id, schedulableId);
            return null;
        }
        
        result.put("label", I18nizableText.stringToI18nizableText((String) jobDataMap.get(KEY_RUNNABLE_LABEL)));
        result.put("description", I18nizableText.stringToI18nizableText((String) jobDataMap.get(KEY_RUNNABLE_DESCRIPTION)));
        String fireProcess = jobDataMap.getString(KEY_RUNNABLE_FIRE_PROCESS);
        result.put("fireProcess", fireProcess);
        result.put("launchUser", Optional.ofNullable(jobDataMap.getString(KEY_RUNNABLE_USERIDENTITY)) // Test if KEY_RUNNABLE_USERIDENTITY is null
                                         .map(UserIdentity::stringToUserIdentity) // Get the UserIdentity
                                         .map(_userManager::getUser) // Get the User from the UserIdentity
                                         .map(u -> _userHelper.user2json(u, true)) // Transform the User in JSON
                                         .orElse(Collections.EMPTY_MAP)); // Return an empty map if we get an Optionnal.EMPTY
        result.put("removable", jobDataMap.getBoolean(KEY_RUNNABLE_REMOVABLE));
        result.put("modifiable", jobDataMap.getBoolean(KEY_RUNNABLE_MODIFIABLE));
        result.put("deactivatable", jobDataMap.getBoolean(KEY_RUNNABLE_DEACTIVATABLE));
        Object lastDuration = jobDataMap.get(AmetysJob.KEY_LAST_DURATION);
        result.put("lastDuration", lastDuration != null ? lastDuration : StringUtils.EMPTY);
        
        boolean successed = jobDataMap.getBoolean(AmetysJob.KEY_SUCCESS);
        result.put("success", successed);
        boolean isRunning = _isRunningJob(jobKey, executingJobs);
        result.put("running", isRunning);
        
        if (isRunning) 
        {
            result.putAll(_addProgressionToJson(schedulable, executingJobs.get(jobKey)));
        }
        
        result.put("schedulable", _schedulableToJson(schedulable, false));
        
        Long previousFireTimeMillis = null;
        String cronExpression = null;
        TriggerKey triggerKey = new TriggerKey(jobKey.getName(), TRIGGER_GROUP);
        boolean triggerExists = _scheduler.checkExists(triggerKey);
        if (triggerExists)
        {
            Trigger trigger = _scheduler.getTrigger(triggerKey);
            result.put("nextFireTime", DateUtils.dateToString(trigger.getNextFireTime()));
            if (trigger instanceof CronTrigger)
            {
                cronExpression = ((CronTrigger) trigger).getCronExpression();
            }
        }
        
        boolean enabled = _isEnabled(triggerKey, triggerExists, fireProcess);
        result.put("enabled", enabled);
        
        boolean completed = _isCompleted(triggerExists, jobDataMap);
        result.put("completed", completed);
        
        if (cronExpression == null && Runnable.FireProcess.CRON.toString().equals(fireProcess))
        {
            // CRON based job, but no trigger (because no next execution), so get the cron expr in the job data map
            cronExpression = jobDataMap.getString(KEY_RUNNABLE_CRON);
        }
        result.put("cronExpression", cronExpression);
        
        if (jobDataMap.containsKey(AmetysJob.KEY_PREVIOUS_FIRE_TIME))
        {
            previousFireTimeMillis = (Long) jobDataMap.get(AmetysJob.KEY_PREVIOUS_FIRE_TIME);
        } 
        
        result.put("previousFireTime", previousFireTimeMillis != null ? DateUtils.dateToString(new Date(previousFireTimeMillis)) : StringUtils.EMPTY);
        result.put("state", _getJobState(completed, successed, isRunning, enabled));
        
        return result;
    }
    
    private Map<String, Object> _addProgressionToJson(Schedulable schedulable, JobExecutionContext jobExecutionContext)
    {
        Map<String, Object> result = new HashMap<>();
        
        Date startDate = jobExecutionContext.getFireTime();
        
        result.put("fireTime",  DateUtils.dateToString(startDate));
        result.put("duration", Duration.between(startDate.toInstant(), Instant.now()).toMillis());
        
        ProgressionTracker progressionTracker = schedulable.getProgressionTracker(jobExecutionContext);
        
        result.put("progressionTracker", progressionTracker.toJSON());
        
        return result;
    }

    private boolean _isRunningJob(JobKey jobKey, Map<JobKey, JobExecutionContext> executingJobs)
    {
        return executingJobs.containsKey(jobKey);
    }
    
    private boolean _isCompleted(boolean triggerExists, JobDataMap jobDataMap)
    {
        // Detect incomplete first:
        //  - trigger still exists (job is programmed or currently running) - triggerExists = true
        //  - or the job programmation is "never" (no trigger but not complete) - fireProcess = NEVER
        //  - or the job programmation is "startup" without trigger and completed status (next startup) - fireProcess = STARTUP and runAtStartupCompleted = false
        // Then invert the result
        
        String fireProcess = jobDataMap.getString(KEY_RUNNABLE_FIRE_PROCESS);
        return !(
                triggerExists
                ||
                Runnable.FireProcess.NEVER.toString().equals(fireProcess)
                ||
                    Runnable.FireProcess.STARTUP.toString().equals(fireProcess)
                    &&
                    !jobDataMap.getBoolean(KEY_RUNNABLE_STARTUP_COMPLETED)
            );
    }
    
    private boolean _isEnabled(TriggerKey triggerKey, boolean triggerExists, String fireProcess) throws SchedulerException
    {
        return triggerExists
            ? !TriggerState.PAUSED.equals(_scheduler.getTriggerState(triggerKey))
            : !Runnable.FireProcess.NEVER.toString().equals(fireProcess);
    }
    
    private String _getJobState(boolean completed, boolean successed, boolean isRunning, boolean enabled)
    {
        if (completed)
        {
            return successed ? "success" : "failure";
        }
        else if (isRunning)
        {
            return "running";
        }
        else if (enabled)
        {
            return "waiting";
        }
        else
        {
            return "disabled";
        }
    }
    
    private Map<String, Object> _schedulableToJson(Schedulable schedulable, boolean isEdition) throws Exception
    {
        Map<String, Object> result = new HashMap<>();
        result.put("id", schedulable.getId());
        result.put("label", schedulable.getLabel());
        result.put("description", schedulable.getDescription());
        result.put("iconGlyph", schedulable.getIconGlyph());
        result.put("iconSmall", schedulable.getIconSmall());
        result.put("iconMedium", schedulable.getIconMedium());
        result.put("iconLarge", schedulable.getIconLarge());
        result.put("private", schedulable.isPrivate());
        
        Map<String, Object> params = new LinkedHashMap<>();
        for (String paramId : schedulable.getParameters().keySet())
        {
            params.put(schedulable.getId() + "$" + paramId, schedulable.getParameters().get(paramId).toJSON(DefinitionContext.newInstance().withEdition(isEdition)));
        }
        result.put("parameters", params);
        
        return result;
    }
    
    /**
     * Gets the configuration for creating/editing a runnable (so returns all the schedulables and their parameters)
     * @return A map containing information about what is needed to create/edit a runnable
     * @throws Exception If an error occurs.
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public Map<String, Object> getEditionConfiguration() throws Exception
    {
        Map<String, Object> result = new HashMap<>();
        
        List<Map<String, Object>> fireProcesses = new ArrayList<>();
        for (FireProcess fireProcessVal : FireProcess.values())
        {
            if (fireProcessVal.equals(FireProcess.NEVER))
            {
                // do not display the 'NEVER' option
                continue;
            }
            Map<String, Object> fireProcess = new LinkedHashMap<>();
            fireProcess.put("value", fireProcessVal.toString());
            String i18nKey = "PLUGINS_CORE_UI_TASKS_DIALOG_FIRE_PROCESS_OPTION_" + fireProcessVal.toString() + "_LABEL";
            fireProcess.put("label", new I18nizableText("plugin.core-ui", i18nKey));
            fireProcesses.add(fireProcess);
        }
        result.put("fireProcesses", fireProcesses);
        
        List<Object> schedulables = new ArrayList<>();
        for (String schedulableId : _schedulableEP.getExtensionsIds())
        {
            Schedulable schedulable = _schedulableEP.getExtension(schedulableId);
            if (!schedulable.isPrivate())
            {
                schedulables.add(_schedulableToJson(schedulable, true));
            }
        }
        result.put("schedulables", schedulables);
        
        return result;
    }
    
    /**
     * Gets the parameters of a given schedulable
     * @param schedulableId The id of the schedulable
     * @return A map containing parameters (prefixed) of the given schedulable
     * @throws Exception If an error occurs.
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> getParameters(String schedulableId) throws Exception
    {
        Schedulable schedulable = _schedulableEP.getExtension(schedulableId);
        Map<String, Object> params = new LinkedHashMap<>();
        for (String paramId : schedulable.getParameters().keySet())
        {
            params.put(schedulable.getId() + "$" + paramId, schedulable.getParameters().get(paramId).toJSON(DefinitionContext.newInstance().withEdition(true)));
        }
        return params;
    }
    
    /**
     * Gets the values of the parameters of the given task
     * @param id The id of the task
     * @return The values of the parameters
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public Map<String, Object> getParameterValues(String id)
    {
        Map<String, Object> result = new HashMap<>();
        try
        {
            JobDetail jobDetail = _scheduler.getJobDetail(new JobKey(id, JOB_GROUP));
            JobDataMap jobDataMap = jobDetail.getJobDataMap();
            result.put("id", id);
            result.put("label", I18nizableText.stringToI18nizableText((String) jobDataMap.get(KEY_RUNNABLE_LABEL)));
            result.put("description", I18nizableText.stringToI18nizableText((String) jobDataMap.get(KEY_RUNNABLE_DESCRIPTION)));
            result.put("fireProcess", jobDataMap.getString(KEY_RUNNABLE_FIRE_PROCESS));
            result.put("cron", jobDataMap.getString(KEY_RUNNABLE_CRON));
            String schedulableId = jobDataMap.getString(KEY_SCHEDULABLE_ID);
            result.put("schedulableId", schedulableId);
            result.put("launchUser", jobDataMap.getString(KEY_RUNNABLE_USERIDENTITY));
            Map<String, Object> params = new HashMap<>();
            result.put("params", params);
            for (String param : jobDataMap.keySet())
            {
                if (param.startsWith(PARAM_VALUES_PREFIX))
                {
                    params.put(schedulableId + "$" + param.substring(PARAM_VALUES_PREFIX.length()), jobDataMap.get(param));
                }
            }
            return result;
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to retrieve the parameter values of the task {}", id, e);
            result.put("error", "scheduler-error");
            return result;
        }
    }
    
    /**
     * Returns true if the given task is modifiable.
     * @param id The id of the task
     * @return true if the given task is modifiable.
     * @throws SchedulerException if an error occured 
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public Map<String, Object> isModifiable(String id) throws SchedulerException
    {
        Map<String, Object> result = new HashMap<>();
        
        JobDetail jobDetail = _scheduler.getJobDetail(new JobKey(id, JOB_GROUP));
        if (jobDetail == null)
        {
            // Does not exist anymore
            result.put("error", "not-found");
            return result;
        }
        JobDataMap jobDataMap = jobDetail.getJobDataMap();
        result.put("modifiable", jobDataMap.getBoolean(KEY_RUNNABLE_MODIFIABLE));
        return result;
    }

    /**
     * Adds a new task.
     * @param label The label
     * @param description The description
     * @param fireProcess the fire process
     * @param cron The cron expression
     * @param schedulableId The id of the schedulable model
     * @param params The values of the parameters
     * @return A result map
     * @throws SchedulerException if an error occured
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> add(String label, String description, String fireProcess, String cron, String schedulableId, Map<String, Object> params) throws SchedulerException
    {
        Request request = ContextHelper.getRequest(_context);
        String workspaceName = (String) request.getAttribute(WorkspaceMatcher.WORKSPACE_NAME);
        UserIdentity taskUser = _currentUserProvider.getUser();
        if (taskUser == null && !workspaceName.equals("admin"))
        {
            Map<String, Object> result = new HashMap<>();
            result.put("error", "invalid-schedulable");
            return result;
        }
        return add(label, description, fireProcess, cron, schedulableId, _userHelper.user2json(taskUser), params);
    }
    
    /**
     * Adds a new task.
     * @param label The label
     * @param description The description
     * @param fireProcess the fire process
     * @param cron The cron expression
     * @param schedulableId The id of the schedulable model
     * @param launchUser The user to launch the task
     * @param params The values of the parameters
     * @return A result map
     * @throws SchedulerException if an error occured
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public Map<String, Object> add(String label, String description, String fireProcess, String cron, String schedulableId, Map<String, Object> launchUser, Map<String, Object> params) throws SchedulerException
    {
        Map<String, Object> result = new HashMap<>();
        
        if (_schedulableEP.getExtension(schedulableId) == null)
        {
            result.put("error", "invalid-schedulable");
            return result;
        }
        else if (_schedulableEP.getExtension(schedulableId).isPrivate())
        {
            result.put("error", "private");
            return result;
        }
        
        String id = _generateUniqueId(label);
        return _edit(id, label, description, FireProcess.valueOf(fireProcess.toUpperCase()), cron, schedulableId, false, launchUser, params);
    }
    
    private String _generateUniqueId(String label) throws SchedulerException
    {
        String value = label.toLowerCase().trim().replaceAll("[\\W_]", "-").replaceAll("-+", "-").replaceAll("^-", "");
        int i = 2;
        String suffixedValue = value;
        while (_scheduler.checkExists(new JobKey(suffixedValue, JOB_GROUP)))
        {
            suffixedValue = value + i;
            i++;
        }
        
        return suffixedValue;
    }
    
    /**
     * Edits the given task.
     * @param id The id of the task
     * @param label The label
     * @param description The description
     * @param fireProcess the fire process
     * @param cron The cron expression
     * @param launchUser The user to launch the task
     * @param params The values of the parameters
     * @return A result map
     * @throws SchedulerException if an error occured
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public Map<String, Object> edit(String id, String label, String description, String fireProcess, String cron, Map<String, Object> launchUser, Map<String, Object> params) throws SchedulerException
    {
        Map<String, Object> result = new HashMap<>();
        
        // Check if exist
        JobDetail jobDetail = null;
        JobKey jobKey = new JobKey(id, JOB_GROUP);
        jobDetail = _scheduler.getJobDetail(jobKey);
        if (jobDetail == null)
        {
            result.put("error", "not-found");
            return result;
        }
        
        // Check if modifiable
        JobDataMap jobDataMap = jobDetail.getJobDataMap();
        boolean isModifiable = jobDataMap.getBoolean(KEY_RUNNABLE_MODIFIABLE);
        if (!isModifiable)
        {
            result.put("error", "no-modifiable");
            return result;
        }
        
        // Remove the associated job as we will create another one
        if (!_scheduler.deleteJob(jobKey))
        {
            // Deletion did not succeed
            result.put("error", "not-found");
            return result;
        }
        
        String schedulableId = jobDataMap.getString(KEY_SCHEDULABLE_ID);
        boolean isVolatile = jobDataMap.getBoolean(KEY_RUNNABLE_VOLATILE);
        return _edit(id, label, description, FireProcess.valueOf(fireProcess.toUpperCase()), cron, schedulableId, isVolatile, launchUser, params);
    }
    
    private Map<String, Object> _edit(String id, String label, String description, FireProcess fireProcess, String cron, String schedulableId, boolean isVolatile, Map<String, Object> launchUser, Map<String, Object> params)
    {
        Map<String, Object> result = new HashMap<>();
        
        Map<String, Object> typedParams = _getTypedParams(params, schedulableId);
        
        boolean deactivatable = !FireProcess.STARTUP.equals(fireProcess); // cannot disable a startup job as we do not attach any trigger to it
        
        // If the user is empty, we get the current user
        UserIdentity userIdentity = _userHelper.json2userIdentity(launchUser);
        if (userIdentity == null)
        {
            userIdentity = _currentUserProvider.getUser();
        }
        Runnable runnable = new DefaultRunnable(id, new I18nizableText(label), new I18nizableText(description), fireProcess, cron, schedulableId, true, true, deactivatable, null, isVolatile, userIdentity, typedParams);
        
        try
        {
            scheduleJob(runnable);
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to add/edit the task {}", id, e);
            result.put("error", "scheduler-error");
            return result;
        }
        
        result.put("id", id);
        return result;
    }
    
    private Map<String, Object> _getTypedParams(Map<String, Object> params, String schedulableId)
    {
        Map<String, Object> result = new HashMap<>();
        
        Map<String, ElementDefinition> declaredParams = _schedulableEP.getExtension(schedulableId).getParameters();
        for (String nameWithPrefix : params.keySet())
        {
            String[] splitStr = nameWithPrefix.split("\\$", 2);
            String prefix = splitStr[0];
            String paramName = splitStr[1];
            if (prefix.equals(schedulableId) && declaredParams.containsKey(paramName))
            {
                Object originalValue = params.get(nameWithPrefix);
                
                ElementDefinition definition = declaredParams.get(paramName);
                ElementType type = definition.getType();
                Object typedValue = type.fromJSONForClient(originalValue, DataContext.newInstance());
                result.put(paramName, typedValue);
            }
            else if (prefix.equals(schedulableId))
            {
                getLogger().warn("The parameter {} is not declared in schedulable {}. It will be ignored", paramName, schedulableId);
            }
        }
        
        return result;
    }
    
    /**
     * Removes the given task.
     * @param id The id of the task
     * @return A result map
     * @throws SchedulerException if an error occured
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public Map<String, Object> remove(String id) throws SchedulerException
    {
        Map<String, Object> result = new HashMap<>();
        JobKey jobKey = new JobKey(id, JOB_GROUP);
        JobDetail jobDetail = null;
        try
        {
            jobDetail = _scheduler.getJobDetail(jobKey);
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to remove the task {}", id, e);
            result.put("error", "scheduler-error");
            return result;
        }
        
        if (jobDetail == null)
        {
            result.put("error", "not-found");
            return result;
        }
        else if (_isRunningJob(jobKey, _getExecutingJobs()))
        {
            result.put("error", "is-running");
            return result;
        }
        
        JobDataMap jobDataMap = jobDetail.getJobDataMap();
        boolean isRemovable = jobDataMap.getBoolean(KEY_RUNNABLE_REMOVABLE);
        if (!isRemovable)
        {
            result.put("error", "no-removable");
            return result;
        }
        
        try
        {
            if (_scheduler.deleteJob(jobKey))
            {
                result.put("id", id);
                return result;
            }
            else
            {
                // Deletion did not succeed
                result.put("error", "no-delete");
                return result;
            }
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to remove the task {}", id, e);
            result.put("error", "scheduler-error");
            return result;
        }
    }
    
    /**
     * Immediately trigger the given task.
     * @param id The id of the task
     * @return A result map
     * @throws SchedulerException if an error occurred
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public Map<String, Object> trigger(String id) throws SchedulerException
    {
        Map<String, Object> result = new HashMap<>();
        JobKey jobKey = new JobKey(id, JOB_GROUP);
        JobDetail jobDetail = null;
        try
        {
            jobDetail = _scheduler.getJobDetail(jobKey);
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to trigger the task " + id, e);
            result.put("error", "scheduler-error");
            return result;
        }
        
        if (jobDetail == null)
        {
            result.put("error", "not-found");
            return result;
        }
        else if (_isRunningJob(jobKey, _getExecutingJobs()))
        {
            result.put("error", "is-running");
            return result;
        }
        
        try
        {
            _scheduler.triggerJob(jobKey);

            result.put("id", id);
            return result;
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to remove the task " + id, e);
            result.put("error", "scheduler-error");
            return result;
        }
    }
    
    /**
     * Enables/disables the given task.
     * @param id The id of the task
     * @param enabled true to enable the task, false to disable it.
     * @return A result map
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public Map<String, Object> enable(String id, boolean enabled)
    {
        Map<String, Object> result = new HashMap<>();
        
        JobKey jobKey = new JobKey(id, JOB_GROUP);
        JobDetail jobDetail = null;
        try
        {
            jobDetail = _scheduler.getJobDetail(jobKey);
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to enable/disable the task {}", id, e);
            result.put("error", "scheduler-error");
            return result;
        }
        if (jobDetail == null)
        {
            result.put("error", "not-found");
            return result;
        }
        JobDataMap jobDataMap = jobDetail.getJobDataMap();
        boolean isDeactivatable = jobDataMap.getBoolean(KEY_RUNNABLE_DEACTIVATABLE);
        if (!isDeactivatable)
        {
            result.put("error", "no-deactivatable");
            return result;
        }
        
        try
        {
            if (enabled)
            {
                _scheduler.resumeJob(jobKey);
            }
            else
            {
                _scheduler.pauseJob(jobKey);
            }
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to enable/disable the task {}", id, e);
            result.put("error", "scheduler-error");
            return result;
        }
        
        result.put("id", id);
        return result;
    }
    
    /**
     * Removes the completed tasks
     * @return The information of deleted tasks
     * @throws SchedulerException if an error occured
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public List<Map<String, Object>> removeCompletedTasks() throws SchedulerException
    {
        List<Map<String, Object>> targets = new ArrayList<>();
        
        List<JobKey> jobsToRemove = new ArrayList<>();
        for (JobKey jobKey : getJobs())
        {
            JobDetail jobDetail = _scheduler.getJobDetail(jobKey);
            JobDataMap jobDataMap = jobDetail.getJobDataMap();
            if (_isCompleted(_scheduler.checkExists(new TriggerKey(jobKey.getName(), TRIGGER_GROUP)), jobDataMap))
            {
                jobsToRemove.add(jobKey);
            }
        }
        
        targets.addAll(getTasksInformation(jobsToRemove.stream().map(JobKey::getName).collect(Collectors.toList())));
        for (JobKey jobKey : jobsToRemove)
        {
            _scheduler.deleteJob(jobKey);
        }
        
        return targets;
    }
    
    /**
     * Returns the enabled state of the task.
     * @param id The id of the task
     * @return A result map
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public Map<String, Object> isEnabled(String id)
    {
        Map<String, Object> result = new HashMap<>();
        
        try
        {
            JobDetail jobDetail = _scheduler.getJobDetail(new JobKey(id, JOB_GROUP));
            if (jobDetail == null)
            {
                // Does not exist anymore
                result.put("error", "not-found");
                return result;
            }
            TriggerKey triggerKey = new TriggerKey(id, TRIGGER_GROUP);
            boolean triggerExists = _scheduler.checkExists(triggerKey);
            String fireProcess = jobDetail.getJobDataMap().getString(KEY_RUNNABLE_FIRE_PROCESS);
            result.put("enabled", _isEnabled(triggerKey, triggerExists, fireProcess));
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to retrieve the enable state of the task {}", id, e);
            result.put("error", "scheduler-error");
            return result;
        }
        return result;
    }
    
    /**
     * Returns the running state of the task.
     * @param id The id of the task
     * @return A result map
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public Map<String, Object> isRunning(String id)
    {
        Map<String, Object> result = new HashMap<>();
        
        try
        {
            result.put("running", _isRunningJob(new JobKey(id, JOB_GROUP), _getExecutingJobs()));
        }
        catch (SchedulerException e)
        {
            getLogger().error("An error occured when trying to retrieve the running state of the task {}", id, e);
            result.put("error", "scheduler-error");
        }
        return result;
    }
    
    /**
     * Gets all the schedulables (private or not)
     * @return the schedulables
     */
    @Callable(rights = __RIGHT_SCHEDULER, context = "/admin")
    public List<Map<String, Object>> getSchedulables()
    {
        List<Map<String, Object>> result = new ArrayList<>();
        
        for (String schedulableId : _schedulableEP.getExtensionsIds())
        {
            Schedulable schedulable = _schedulableEP.getExtension(schedulableId);
            Map<String, Object> map = new HashMap<>();
            map.put("id", schedulable.getId());
            map.put("text", schedulable.getLabel());
            result.add(map);
        }
        
        return result;
    }

    @Override
    public void dispose()
    {
        try
        {
            _scheduler.shutdown();
            AmetysJob.dispose();
        }
        catch (SchedulerException e)
        {
            getLogger().error("Fail to shutdown scheduler", e);
        }
    }
}
