/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.search;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;

import org.ametys.core.group.Group;
import org.ametys.core.group.GroupDirectoryContextHelper;
import org.ametys.core.group.GroupDirectoryDAO;
import org.ametys.core.group.directory.GroupDirectory;
import org.ametys.core.user.User;
import org.ametys.core.user.UserManager;
import org.ametys.core.user.population.PopulationContextHelper;
import org.ametys.core.user.population.UserPopulation;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.plugins.core.group.GroupHelper;
import org.ametys.plugins.core.user.UserHelper;

/**
 * Common helper for user and group searches
 */
public class UserAndGroupSearchManager extends AbstractLogEnabled implements Component, Serviceable 
{
    /** The avalon role */
    public static final String ROLE = UserAndGroupSearchManager.class.getName();
    
    private static final int _DEFAULT_OFFSET_VALUE = 0;
    private static final int _DEFAULT_COUNT_VALUE = 20;
    
    private UserManager _userManager;
    private UserHelper _userHelper;
    private GroupDirectoryDAO _groupDirectoryDAO;
    private GroupDirectoryContextHelper _directoryContextHelper;
    private PopulationContextHelper _populationContextHelper;
    private UserPopulationDAO _userPopulationDAO;
    private GroupHelper _groupHelper;
    
    public void service(ServiceManager manager) throws ServiceException
    {
        _userManager = (UserManager) manager.lookup(UserManager.ROLE);
        _userHelper = (UserHelper) manager.lookup(UserHelper.ROLE);
        _userPopulationDAO = (UserPopulationDAO) manager.lookup(UserPopulationDAO.ROLE);
        _groupDirectoryDAO = (GroupDirectoryDAO) manager.lookup(GroupDirectoryDAO.ROLE);
        _directoryContextHelper = (GroupDirectoryContextHelper) manager.lookup(GroupDirectoryContextHelper.ROLE);
        _populationContextHelper = (PopulationContextHelper) manager.lookup(PopulationContextHelper.ROLE);
        _groupHelper = (GroupHelper) manager.lookup(GroupHelper.ROLE);
    }

    /**
     * Mixed users and groups sorted search, with multiple offsets
     * @param contexts The context of the search
     * @param limit The maximum number of results per group directory and population
     * @param searchData The search data to fetch the next batch of the search. Must be null on the first search.
     * @param params The parameters to pass to the group directories and user populations when retrieving their contents
     * @param sort true to sort users
     * @return An object with the list of results, sorted, and the search data required to fetch the following batch
     */
    public Map<String, Object> searchUsersAndGroupByContext(Set<String> contexts, int limit, Map<String, Object> searchData, Map<String, Object> params, boolean sort)
    {
        Set<String> groupDirectories = _directoryContextHelper.getGroupDirectoriesOnContexts(contexts);
        Set<String> userPopulations = _populationContextHelper.getUserPopulationsOnContexts(contexts, false, true);
        return searchUsersAndGroup(userPopulations, groupDirectories, limit, searchData, params, sort);
    }

    /**
     * Mixed users and groups sorted search, with multiple offsets
     * @param availablePopulations The populations of users available for search. On first search, all populations will be used
     * @param availableDirectories The directories of groups available for search. On first search, all directories will be used
     * @param limit The maximum number of results per group directory and population
     * @param searchData The search data to fetch the next batch of the search. Must be null on the first search.
     * @param params The parameters to pass to the group directories and user populations when retrieving their contents
     * @param sort true to sort users
     * @return An object with the list of results, sorted, and the search data required to fetch the following batch
     */
    public Map<String, Object> searchUsersAndGroup(Set<String> availablePopulations, Set<String> availableDirectories, int limit, Map<String, Object> searchData, Map<String, Object> params, boolean sort)
    {
        Map<String, Integer> usersFromSearchData = _getFromSearchData("users", searchData);
        Set<String> userPopulations = usersFromSearchData == null ? availablePopulations : _filterSetUnion(availablePopulations, usersFromSearchData.keySet());
        Map<String, Integer> groupsFromSearchData = _getFromSearchData("groups", searchData);
        Set<String> groupDirectories = groupsFromSearchData == null ? availableDirectories : _filterSetUnion(availableDirectories, groupsFromSearchData.keySet());
        
        Function<UserPopulation, Integer> getPopulationOffset = population -> usersFromSearchData != null 
                ? usersFromSearchData.getOrDefault(population.getId(), _DEFAULT_OFFSET_VALUE) 
                : _DEFAULT_OFFSET_VALUE;
        Function<GroupDirectory, Integer> getDirectoryOffset = directory -> groupsFromSearchData != null 
                ? groupsFromSearchData.getOrDefault(directory.getId(), _DEFAULT_OFFSET_VALUE) 
                : _DEFAULT_OFFSET_VALUE;
        
        int count = limit <= 0 ? _DEFAULT_COUNT_VALUE : limit;

        List<SearchDataResult> searchDataResults = Stream.concat(
            userPopulations.stream()
                    .map(_userPopulationDAO::getUserPopulation)
                    .filter(Objects::nonNull)
                    .map(population -> _searchUsers(population, count, getPopulationOffset.apply(population), params, sort)),
            groupDirectories.stream()
                    .map(_groupDirectoryDAO::getGroupDirectory)
                    .filter(Objects::nonNull)
                    .map(directory -> _searchGroups(directory, count, getDirectoryOffset.apply(directory), params)))
            .collect(Collectors.toList());
        
        if (searchDataResults.size() == 0)
        {
            return _formatSearchResults(Collections.emptyList(), null);
        }
        
        if (searchDataResults.size() == 1)
        {
            SearchDataResult searchDataResult = searchDataResults.get(0);
            return _formatSearchResults(searchDataResult.results2Json(), searchDataResult.isFinished() ? null : searchDataResults);
        }
        
        if (searchDataResults.stream().allMatch(SearchDataResult::isFinished))
        {
            return _formatSearchResults(_mergeSearchResults(searchDataResults), null);
        }
        
        searchDataResults = searchDataResults.stream()
                .filter(searchDataResult -> searchDataResult.getLastSortableName() != null)
                .sorted(Comparator.comparing(SearchDataResult::getLastSortableName))
                .collect(Collectors.toList());
        
        String entryDelimitor = null;
        for (SearchDataResult searchDataResult : searchDataResults)
        {
            if (entryDelimitor != null)
            {
                searchDataResult.truncateAfter(entryDelimitor);
            }
            else if (!searchDataResult.isFinished())
            {
                entryDelimitor = searchDataResult.getLastSortableName();
            }
        }
        
        List<SearchDataResult> unfinishedSearchDataResults = searchDataResults.stream()
                .filter(searchDataResult -> !searchDataResult.isFinished())
                .collect(Collectors.toList());
        
        return _formatSearchResults(_mergeSearchResults(searchDataResults), unfinishedSearchDataResults);
    }
    
    private Map<String, Integer> _getFromSearchData(String property, Map<String, Object> searchData)
    {
        if (searchData == null)
        {
            return null;
        }
        
        return Optional.ofNullable(searchData.getOrDefault(property, null))
                .filter(Map.class::isInstance)
                .map(Map.class::cast)
                .orElse(Collections.emptyMap());
    }
    
    private Set<String> _filterSetUnion(Set<String> values, Set<String> referenceSet)
    {
        return values.stream()
                .filter(referenceSet::contains)
                .collect(Collectors.toSet());
    }
    
    private SearchDataResult _searchUsers(UserPopulation population, int limit, int offset, Map<String, Object> params, boolean sort)
    {
        List<User> users = new ArrayList<>(_userManager.getUsers(population, limit, offset, params, sort));
        
        return new PopulationSearchDataResult(_userHelper, population.getId(), users.size() < limit, users, offset);
    }
    
    private SearchDataResult _searchGroups(GroupDirectory directory, int limit, int offset, Map<String, Object> params)
    {
        List<Group> groups = directory.getGroups(limit, offset, params);
        
        return new DirectorySearchDataResult(_groupHelper, directory.getId(), groups.size() < limit, groups, offset);
    }
    
    private Map<String, Object> _formatSearchResults(List<Map<String, Object>> results, List<SearchDataResult> searchDataResults)
    {
        Map<String, Object> jsonResult = new HashMap<>();
        jsonResult.put("results", results);
        if (searchDataResults == null)
        {
            jsonResult.put("finished", true);
        }
        else
        {
            Map<String, Map<String, Integer>> searchDataJson = new HashMap<>();
            for (SearchDataResult searchDataResult: searchDataResults)
            {
                String type = searchDataResult.getType();
                if (!searchDataJson.containsKey(type))
                {
                    searchDataJson.put(type, new HashMap<>());
                }
                searchDataJson.get(type).put(searchDataResult.getSourceId(), searchDataResult.getOffset());
            }
            
            if (searchDataJson.size() > 0)
            {
                jsonResult.put("searchData", searchDataJson);
            }
            else
            {
                jsonResult.put("finished", true);
            }
        }
        
        return jsonResult;
    }
    
    private List<Map<String, Object>> _mergeSearchResults(List<SearchDataResult> searchDataResults)
    {
        return searchDataResults.stream()
                .map(SearchDataResult::results2Json)
                .flatMap(List::stream)
                .sorted((o1, o2) -> _getSortableName(o1).compareTo(_getSortableName(o2)))
                .collect(Collectors.toList());
    }
    
    /**
     * Get the sortable name for the json object
     * @param o The object
     * @return The sortable name, or an empty string if no name was found
     */
    protected static String _getSortableName(Map<String, Object> o)
    {
        if (o.containsKey("sortablename"))
        {
            return ((String) o.get("sortablename")).toLowerCase();
        }
        
        if (o.containsKey("label"))
        {
            return ((String) o.get("label")).toLowerCase();
        }
        
        return "";
    }

    private static interface SearchDataResult
    {
        public String getType();
        public String getSourceId();
        public boolean isFinished();
        public void truncateAfter(String entry);
        public int getOffset();
        public List<Map<String, Object>> results2Json();
        public String getLastSortableName();
    }
    
    private static class PopulationSearchDataResult implements SearchDataResult
    {
        private boolean _isFinished;
        private List<User> _users;
        private int _offset;
        private UserHelper _helper;
        private String _sourceId;
        
        public PopulationSearchDataResult(UserHelper userHelper, String sourceId, boolean isFinished, List<User> users, int initialOffset)
        {
            _helper = userHelper;
            _sourceId = sourceId;
            _isFinished = isFinished;
            _users = users;
            _offset = initialOffset;
        }

        public String getType()
        {
            return "users";
        }
        
        public String getSourceId()
        {
            return _sourceId;
        }

        public boolean isFinished()
        {
            return _isFinished;
        }

        public void truncateAfter(String entry)
        {
            if (_users.removeIf(user -> user.getSortableName().toLowerCase().compareTo(entry) > 0))
            {
                _isFinished = false;
            }
        }

        public int getOffset()
        {
            return _offset + _users.size();
        }

        public List<Map<String, Object>> results2Json()
        {
            return _helper.users2json(_users, true);
        }
        
        public String getLastSortableName()
        {
            return _users.size() > 0 ? _users.get(_users.size() - 1).getSortableName().toLowerCase() : null;
        }
        
        @Override
        public String toString()
        {
            return SearchDataResult.class.getName() + "@" + Integer.toHexString(hashCode()) 
                + "[" + _users.size() + " user(s) from population " + _sourceId + " at offset " + _offset + "]" 
                + (_isFinished ? "[Finished]" : "");
        }
    }

    private static class DirectorySearchDataResult implements SearchDataResult
    {
        private String _sourceId;
        private boolean _isFinished;
        private List<Group> _groups;
        private int _offset;
        private GroupHelper _helper;

        public DirectorySearchDataResult(GroupHelper helper, String sourceId, boolean isFinished, List<Group> groups, int initialOffset)
        {
            _helper = helper;
            _sourceId = sourceId;
            _isFinished = isFinished;
            _groups = groups;
            _offset = initialOffset;
        }
        
        public String getType()
        {
            return "groups";
        }

        public String getSourceId()
        {
            return _sourceId;
        }
        
        public boolean isFinished()
        {
            return _isFinished;
        }

        public void truncateAfter(String entry)
        {
            if (_groups.removeIf(group -> group.getLabel().toLowerCase().compareTo(entry) > 0))
            {
                _isFinished = false;
            }
        }

        public int getOffset()
        {
            return _offset + _groups.size();
        }

        public List<Map<String, Object>> results2Json()
        {
            return _helper.groups2JSON(_groups, false);
        }

        public String getLastSortableName()
        {
            return _groups.size() > 0 ? _groups.get(_groups.size() - 1).getLabel().toLowerCase() : null;
        }

        @Override
        public String toString()
        {
            return SearchDataResult.class.getName() + "@" + Integer.toHexString(hashCode()) 
                + "[" + _groups.size() + " group(s) from directory " + _sourceId + " at offset " + _offset + "]" 
                + (_isFinished ? "[Finished]" : "");
        }

    }
}
