/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.upload;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.cocoon.acting.Action;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.apache.cocoon.servlet.multipart.Part;
import org.apache.cocoon.servlet.multipart.PartOnDisk;
import org.apache.cocoon.servlet.multipart.RejectedPart;
import org.apache.commons.lang3.exception.ExceptionUtils;

import org.ametys.core.cocoon.JSonReader;
import org.ametys.core.upload.Upload;
import org.ametys.core.upload.UploadManager;
import org.ametys.core.util.cocoon.AbstractCurrentUserProviderServiceableAction;

/**
 * {@link Action} for uploading a file and store it using the {@link UploadManager}.
 */
public class UploadAction extends AbstractCurrentUserProviderServiceableAction
{
    /** The Upload Manager */
    protected UploadManager _uploadManager;
    
    @Override
    public Map act(Redirector redirector, SourceResolver resolver, Map objectModel, String source, Parameters parameters) throws Exception
    {
        Request request = ObjectModelHelper.getRequest(objectModel);
        
        _lookupComponents();
        
        Map<String, Object> result = new LinkedHashMap<>();
        _doUpload(request, parameters, result);
        request.setAttribute(JSonReader.OBJECT_TO_READ, result);

        return EMPTY_MAP;
    }
    
    /**
     * Lookup components (lazy initialize)
     * @throws ServiceException If some components cannot be looked up
     */
    protected void _lookupComponents() throws ServiceException
    {
        // Lazy initialize the upload manager because it cannot be found if the config is not complete  
        if (_uploadManager == null)
        {
            _uploadManager = (UploadManager) manager.lookup(UploadManager.ROLE);
        }
    }
    
    /**
     * Do upload the file and fill the result map.
     * @param request The request
     * @param parameters The parameters
     * @param result The result map to fill
     * @throws Exception if an error occurs.
     */
    protected void _doUpload(Request request, Parameters parameters, Map<String, Object> result) throws Exception
    {
        Part partUploaded = _getPart(request);
        
        if (!(partUploaded instanceof PartOnDisk))
        {
            result.put("success", false);
            if (partUploaded instanceof RejectedPart rejectedPart && rejectedPart.getMaxContentLength() == 0)
            {
                result.put("error", "infected");
            }
            else // if (partUploaded == null || partUploaded instanceof RejectedPart)
            {
                result.put("error", "rejected");
            }
        }
        else
        {
            try (InputStream is = partUploaded.getInputStream())
            {
                _storeUpload(is, partUploaded.getFileName(), result);
            }
            catch (IOException e)
            {
                _handleStoreUploadException(e, partUploaded, result);
            }
        }
    }
    
    /**
     * Gets the {@link Part} object which is trying to be uploaded
     * @param request The request
     * @return The {@link Part} object which is trying to be uploaded
     * @throws Exception if an error occurs.
     */
    protected Part _getPart(Request request) throws Exception
    {
        Part partUploaded = (Part) request.get("file");
        
        if (partUploaded == null)
        {
            throw new Exception("Missing request parameter file");
        }
        
        return partUploaded;
    }
    
    /**
     * Stores the given upload data (as an {@link InputStream}) into the {@link UploadManager} system and fill the result map.
     * @param is The upload data
     * @param filename The file name
     * @param result The result map to fill
     * @throws IOException if an I/O error occurs
     */
    protected void _storeUpload(InputStream is, String filename, Map<String, Object> result) throws IOException
    {
        Upload upload = _uploadManager.storeUpload(_getCurrentUser(), filename, is);
        _fillSuccess(upload, result);
    }
    
    /**
     * Fill the result map.
     * @param upload The upload
     * @param result The result map to fill
     */
    protected void _fillSuccess(Upload upload, Map<String, Object> result)
    {
        result.put("success", true);
        result.put("id", upload.getId());
        result.put("filename", upload.getFilename());
        result.put("size", upload.getLength());
        result.put("viewHref", _getUrlForView(upload));
        result.put("downloadHref", _getUrlForDownload(upload));
    }
    
    /**
     * Handles a store upload exception by filling the result map with information about the exception.
     * @param e The exception
     * @param obj The object representing the upload data (for logging purpose)
     * @param result The result map to fill
     */
    protected void _handleStoreUploadException(IOException e, Object obj, Map<String, Object> result)
    {
        getLogger().error("Unable to store uploaded file: " + obj, e);
        
        result.put("success", false);
        
        Map<String, String> ex = new HashMap<>();
        ex.put("message", e.getMessage());
        ex.put("stacktrace", ExceptionUtils.getStackTrace(e));
        
        result.put("error", ex);
    }
    
    /**
     * Get the url for view the uploaded file
     * @param upload The file uploaded
     * @return The url for view
     */
    protected String _getUrlForView(Upload upload)
    {
        return "/plugins/core/upload/file?id=" + upload.getId();
    }
    
    /**
     * Get the url for download the uploaded file
     * @param upload The file uploaded
     * @return The url for view
     */
    protected String _getUrlForDownload(Upload upload)
    {
        return "/plugins/core/upload/file?id=" + upload.getId() + "&download=true";
    }
}
