/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.upload;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.Map;
import java.util.NoSuchElementException;

import org.apache.avalon.framework.parameters.Parameters;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.ResourceNotFoundException;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Request;
import org.apache.cocoon.environment.SourceResolver;
import org.xml.sax.SAXException;

import org.ametys.core.upload.Upload;
import org.ametys.core.upload.UploadManager;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.util.cocoon.AbstractResourceReader;

/**
 * {@link Reader} for generating binary output of previously uploaded file.
 */
public class UploadReader extends AbstractResourceReader implements Serviceable
{
    private CurrentUserProvider _currentUserProvider;
    private UploadManager _uploadManager;
    private Upload _upload;
    
    @Override
    public void service(ServiceManager serviceManager) throws ServiceException
    {
        _currentUserProvider = (CurrentUserProvider) serviceManager.lookup(CurrentUserProvider.ROLE);
        _uploadManager = (UploadManager) serviceManager.lookup(UploadManager.ROLE);
    }
    
    @Override
    protected void doSetup(SourceResolver res, Map objModel, String src, Parameters par)
    {
        Request request = ObjectModelHelper.getRequest(objModel);
        String uploadId = request.getParameter("id");

        try
        {
            _upload = _uploadManager.getUpload(_currentUserProvider.getUser(), uploadId);
        }
        catch (NoSuchElementException e)
        {
            // Invalid upload id
            getLogger().warn("Cannot find the temporary uploaded file with id " + (uploadId != null ? "'" + uploadId + "'" : "<null>"));
        }
    }
    
    @Override
    protected InputStream getInputStream()
    {
        return _upload.getInputStream();
    }
    
    @Override
    protected String getFilename()
    {
        return _upload.getFilename();
    }
    
    @Override
    protected String getEncodedFilename()
    {
        return null;
    }
    
    @Override
    protected long getLength()
    {
        return _upload.getLength();
    }
    
    @Override
    public long getLastModified()
    {
        if (_upload != null)
        {
            return _upload.getUploadedDate().toInstant().toEpochMilli();
        }
        
        return super.getLastModified();
    }
    
    @Override
    public String getMimeType()
    {
        if (_upload != null)
        {
            return _upload.getMimeType();
        }

        return super.getMimeType();
    }
    
    
    @Override
    public void generate() throws IOException, SAXException, ProcessingException
    {
        if (_upload == null)
        {
            throw new ResourceNotFoundException("No upload for source: " + source);
        }
        
        super.generate();
    }
    
    @Override
    public void recycle()
    {
        super.recycle();
        _upload = null;
    }
}
