/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.upload.image;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

import javax.imageio.ImageIO;

import org.apache.avalon.framework.parameters.ParameterException;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.cocoon.environment.Request;
import org.apache.commons.io.FilenameUtils;

import org.ametys.core.upload.Upload;
import org.ametys.core.upload.UploadManager;
import org.ametys.core.util.ImageHelper;
import org.ametys.plugins.core.upload.UploadAction;

/**
 * Generates a new image by cropping an already uploaded image and store it.
 */
public class CropImageAction extends UploadAction
{
    @Override
    protected void _doUpload(Request request, Parameters parameters, Map<String, Object> result) throws Exception
    {
        Upload originalImage = _getUploadObject(request, parameters);
        Cropping cropping = new Cropping(parameters);
        _doCrop(originalImage, cropping, result);
    }
    
    /**
     * Gets the already uploaded image
     * @param request The request
     * @param parameters The parameters
     * @return The previously uploaded image
     * @throws ParameterException if the <code>imageId</code> parameter is not specified
     */
    protected Upload _getUploadObject(Request request, Parameters parameters) throws ParameterException
    {
        String imageId = parameters.getParameter("imageId");
        return _uploadManager.getUpload(_getCurrentUser(), imageId);
    }
    
    /**
     * Do crop the original image and fill the result map.
     * @param originalImage The original image to crop
     * @param cropping The cropping to apply
     * @param result The result map to fill
     * @throws IOException if an I/O error occurs
     */
    protected void _doCrop(Upload originalImage, Cropping cropping, Map<String, Object> result) throws IOException
    {
        if (cropping.isOriginal())
        {
            // Avoid an unnecessary cropping and the creation of a new image which will be the same as the original.
            _fillSuccess(originalImage, result);
            return;
        }
        
        try (InputStream is = originalImage.getInputStream())
        {
            BufferedImage src = _uploadAsBufferedImage(originalImage);
            int originalWidth = src.getWidth();
            int originalHeight = src.getHeight();
            
            int x = cropping.getX(originalWidth);
            int y = cropping.getY(originalHeight);
            
            int width = cropping.getWidth(originalWidth);
            width = _checkValidity(width, x, originalWidth);
            int height = cropping.getHeight(originalHeight);
            height = _checkValidity(height, y, originalHeight);
            
            BufferedImage croppedImage = ImageHelper.generateCroppedImage(src, x, y, height, width);
            _storeCroppedImage(croppedImage, originalImage.getFilename(), result);
        }
    }
    
    private int _checkValidity(int widthOrHeight, int xOrY, int originalWidthOrHeight)
    {
        if (xOrY + widthOrHeight > originalWidthOrHeight)
        {
            // It somehow overflows the original size, do return the maximum valid width/height
            return originalWidthOrHeight - xOrY;
        }
        else
        {
            return widthOrHeight;
        }
    }
    
    /**
     * Converts an uploaded image as {@link Upload} into a {@link BufferedImage}
     * @param upload The uploaded image
     * @return The upload image as a {@link BufferedImage}
     * @throws IOException if an I/O error occurs
     */
    protected BufferedImage _uploadAsBufferedImage(Upload upload) throws IOException
    {
        try (InputStream is = upload.getInputStream())
        {
            return ImageHelper.read(is);
        }
    }
    
    /**
     * Stores the cropped image into the {@link UploadManager} system and fill the result map.
     * @param croppedImage The cropped image
     * @param originalName The original image file name
     * @param result The result map to fill
     * @throws IOException if an I/O error occurs
     */
    protected void _storeCroppedImage(BufferedImage croppedImage, String originalName, Map<String, Object> result) throws IOException
    {
        try (ByteArrayOutputStream baos = new ByteArrayOutputStream())
        {
            ImageIO.write(croppedImage, _getFormatName(originalName), baos);
            byte[] buffer = baos.toByteArray();
            try (InputStream is = new ByteArrayInputStream(buffer))
            {
                _storeUpload(is, _getCroppedImageFilename(originalName), result);
            }
            catch (IOException e)
            {
                _handleStoreUploadException(e, croppedImage, result);
            }
        }
    }
    
    /**
     * Gets the String containing the informal name of the format of the image
     * @param iamgeFilename The file name 
     * @return the String containing the informal name of the format of the image
     */
    protected String _getFormatName(String iamgeFilename)
    {
        String format = FilenameUtils.getExtension(iamgeFilename);
        return format.isEmpty() ? "png" : format;
    }
    
    /**
     * Gets the file name of the cropped image
     * @param originalName The file name of the original image
     * @return the file name of the cropped image
     */
    protected String _getCroppedImageFilename(String originalName)
    {
        // By default keep name of the original image file
        return originalName;
    }
    
    /**
     * Represents a rectangular cropping of an image.
     * <br>It stores only relative float values (which are between 0 and 1) as ratio of the width and height of the original image, which are unknown.
     */
    protected static class Cropping
    {
        private float _relativeX1;
        private float _relativeY1;
        private float _relativeWidth;
        private float _relativeHeight;
        
        /**
         * Builds a new Cropping object from {@link Parameters}
         * @param parameters The parameters
         */
        protected Cropping(Parameters parameters)
        {
            _relativeX1 = parameters.getParameterAsFloat("x1", 0.0f);
            _relativeY1 = parameters.getParameterAsFloat("y1", 0.0f);
            _relativeWidth = parameters.getParameterAsFloat("width", 1.0f);
            _relativeHeight = parameters.getParameterAsFloat("height", 1.0f);
            
            assert _relativeX1 >= 0;
            assert _relativeX1 <= 1;
            
            assert _relativeY1 >= 0;
            assert _relativeY1 <= 1;
            
            assert _relativeWidth >= 0;
            assert _relativeWidth <= 1;
            
            assert _relativeHeight >= 0;
            assert _relativeHeight <= 1;
        }
        
        /**
         * Returns <code>true</code> if this cropping actually represents the original image
         * @return <code>true</code> if this cropping actually represents the original image
         */
        protected boolean isOriginal()
        {
            return _relativeX1 <= 0.0f && _relativeY1 <= 0.0f
                && _relativeWidth >= 1.0f && _relativeHeight >= 1.0f;
        }
        
        /**
         * Gets the X coordinate of the upper-left corner of the cropping rectangular region
         * @param originalWidth The width of the original image
         * @return the X coordinate of the upper-left corner of the cropping rectangular region
         */
        protected int getX(int originalWidth)
        {
            return Math.round(originalWidth * _relativeX1);
        }
        
        /**
         * Gets the Y coordinate of the upper-left corner of the cropping rectangular region
         * @param originalHeight The height of the original image
         * @return the Y coordinate of the upper-left corner of the cropping rectangular region
         */
        protected int getY(int originalHeight)
        {
            return Math.round(originalHeight * _relativeY1);
        }
        
        /**
         * Gets the width of the cropping rectangular region
         * @param originalWidth The width of the original image
         * @return the width of the cropping rectangular region
         */
        protected int getWidth(int originalWidth)
        {
            return Math.round(originalWidth * _relativeWidth);
        }
        
        /**
         * Gets the height of the cropping rectangular region
         * @param originalHeight The height of the original image
         * @return the height of the cropping rectangular region
         */
        protected int getHeight(int originalHeight)
        {
            return Math.round(originalHeight * _relativeHeight);
        }
    }
}
