/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.user;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.servlet.multipart.Part;
import org.apache.cocoon.servlet.multipart.PartOnDisk;
import org.apache.cocoon.servlet.multipart.RejectedPart;
import org.apache.commons.collections4.ListUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.input.BOMInputStream;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.ObservationConstants;
import org.ametys.core.observation.Event;
import org.ametys.core.observation.ObservationManager;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.InvalidModificationException;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.util.I18nUtils;
import org.ametys.runtime.authentication.AccessDeniedException;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.parameter.Errors;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Import users from a CSV or text file.
 */
public class ImportUsers extends AbstractLogEnabled implements Component, Serviceable
{
    /** Avalon Role */
    public static final String ROLE = ImportUsers.class.getName();
    
    private static final String[] _ALLOWED_EXTENSIONS = new String[] {"txt", "csv"};
    private static final String[] _COLUMNS = new String[] {"login", "password", "firstname", "lastname", "email"};
    private static final String _COLUMN_LOGIN = "login";
    private static final String _COLUMN_PASSWORD = "password";
    private static final String _COLUMN_FIRSTNAME = "firstname";
    private static final String _COLUMN_LASTNAME = "lastname";
    private static final String _COLUMN_EMAIL = "email";

    private static final String _DEFAULT_HASHED_PASSWORD = "ImportedUserWithoutPassword";

    /** The subscribers DAO. */
    protected UserDAO _usersDao;
    /** User Manager. */
    protected UserManager _userManager;
    /** I18n utils */
    protected I18nUtils _i18nUtils;
    /** Observation Manager */
    protected ObservationManager _observationManager;
    /** Current User Provider */
    protected CurrentUserProvider _currentUserProvider;
    /** The right manager */
    protected RightManager _rightManager;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _usersDao = (UserDAO) smanager.lookup(UserDAO.ROLE);
        _userManager = (UserManager) smanager.lookup(UserManager.ROLE);
        _i18nUtils = (I18nUtils) smanager.lookup(I18nUtils.ROLE);
        _currentUserProvider = (CurrentUserProvider) smanager.lookup(CurrentUserProvider.ROLE);
        _rightManager = (RightManager) smanager.lookup(RightManager.ROLE);
        if (smanager.hasService(ObservationManager.ROLE))
        {
            _observationManager = (ObservationManager) smanager.lookup(ObservationManager.ROLE);
        }
    }
    
    /**
     * Read a file from disk and add users inside in the requested population/directory
     * @param filePart file
     * @param populationId population Id
     * @param userDirectoryId user directory
     * @param cleanDirectory true to clear every user that was not in the import file
     * @return a map containing success, addedCount, existingCount, errorCount, deletedCount, message
     * success : true if success
     * addedCount, existingCount, errorCount : number of imported user added, modified, an error occured (see server log)
     * message : error message if something wrong happened (usually malformed file)
     * @throws FileNotFoundException error reading the file
     * @throws InvalidModificationException error modifying the directory
     */
    public Map<String, Object> importUsers(Part filePart, String populationId, String userDirectoryId, boolean cleanDirectory) throws FileNotFoundException, InvalidModificationException
    {
        Map<String, Object> result = new HashMap<>();

        if (filePart instanceof RejectedPart)
        {
            result.put("success", false);
            result.put("error", "rejected-file");
            result.put("message", new I18nizableText("plugin.core", "PLUGINS_CORE_USERS_IMPORT_ERROR_REJECTED_FILE"));
            return result;
        }

        PartOnDisk uploadedFilePart = (PartOnDisk) filePart;
        File uploadedFile = (uploadedFilePart != null) ? uploadedFilePart.getFile() : null;
        String filename = (uploadedFilePart != null) ? uploadedFilePart.getFileName().toLowerCase() : null;

        if (!FilenameUtils.isExtension(filename, _ALLOWED_EXTENSIONS))
        {
            result.put("success", false);
            result.put("error", "invalid-extension");
            result.put("message", new I18nizableText("plugin.core", "PLUGINS_CORE_USERS_IMPORT_ERROR_INVALID_EXTENSION"));
            return result;
        }
        
        UserIdentity currentUser = _currentUserProvider.getUser();
        
        boolean hasRight = _rightManager.hasRight(currentUser, "Runtime_Rights_User_Handle", "/${WorkspaceName}") == RightResult.RIGHT_ALLOW
                  || _rightManager.hasRight(currentUser, "Runtime_Rights_User_OwnPopulation_Handle", "/${WorkspaceName}") == RightResult.RIGHT_ALLOW && currentUser.getPopulationId().equals(populationId);
        
        if (!hasRight)
        {
            throw new AccessDeniedException("The user " + currentUser + " tried to import users without sufficient rights");
        }

        try (FileInputStream fileIS =  new FileInputStream(uploadedFile); 
             BOMInputStream bomIS = BOMInputStream.builder().setInputStream(fileIS).get())
        {
            Map<String, List<UserIdentity>> parsedUsers = parseFile(bomIS, populationId, userDirectoryId);
            
            List<UserIdentity> addedUsers = parsedUsers.get("added");
            List<UserIdentity> editedUsers = parsedUsers.get("edited");
            
            List<Map<String, String>> removedUsers;
            if (cleanDirectory)
            {
                List<UserIdentity> usersToSave = ListUtils.union(addedUsers, editedUsers);
                removedUsers = clearUserDirectory(populationId, userDirectoryId, usersToSave);
            }
            else
            {
                removedUsers = new ArrayList<>();
            }
    
            result.put("success", true);
            result.put("addedCount", addedUsers.size());
            result.put("existingCount", editedUsers.size());
            result.put("errorCount", parsedUsers.get("error").size());
            result.put("deletedCount", removedUsers.size());
        }
        catch (IOException ioe)
        {
            result.put("success", false);
            result.put("error", IOException.class.getName());
            result.put("message", ioe.getLocalizedMessage());
        }
        catch (ImportUserActionException iuae)
        {
            result.put("success", false);
            result.put("error", iuae.getErrorName());
            result.put("message", iuae.getI18nizableText());
        }

        return result;
    }
    
    /**
     * Remove all users in a userDirectory, avoiding those in usersToSave list
     * @param populationId population Id
     * @param userDirectoryId directory Id
     * @param usersToSave list of users that need to be saved
     * @return list of deleted users (containing login and populationId)
     * @throws InvalidModificationException If modification is not possible
     */
    protected List<Map<String, String>> clearUserDirectory(String populationId, String userDirectoryId, List<UserIdentity> usersToSave) throws InvalidModificationException
    {
        Map<String, Object> parameters =  new HashMap<>();
        Collection<User> existingUsersInDirectory = _userManager.getUsersByDirectory(populationId, userDirectoryId, Integer.MAX_VALUE, 0, parameters);
        List<Map<String, String>> usersToDelete = new ArrayList<>();
        UserIdentity currentUser = _usersDao._getCurrentUser();
        for (User user : existingUsersInDirectory)
        {
            UserIdentity ui = user.getIdentity();
            // If the user is not in the "save" list, and is not the current user, add it to the delete list
            if (!usersToSave.contains(ui) && (currentUser == null || !currentUser.equals(ui)))
            {
                Map<String, String> userTodelete = new HashMap<>();
                userTodelete.put("login", ui.getLogin());
                userTodelete.put("populationId", ui.getPopulationId());
                usersToDelete.add(userTodelete);
            }
        }
        _usersDao.deleteUsers(usersToDelete);
        return usersToDelete;
    }
    
    /**
     * Parse the file to add new users in the directory
     * @param bomIS input stream
     * @param populationId population Id
     * @param userDirectoryId directory Id
     * @return a map, containing added/edited/error, each containing the logins in each list
     * @throws IOException if an error occurs
     * @throws ImportUserActionException file parsing failed
     */
    @SuppressWarnings("unchecked")
    protected Map<String, List<UserIdentity>> parseFile(BOMInputStream bomIS, String populationId, String userDirectoryId) throws IOException, ImportUserActionException
    {
        Map<String, Integer> order = new HashMap<>();

        List<UserIdentity> addedUsers = new ArrayList<>();
        List<UserIdentity> editedUsers = new ArrayList<>();
        List<UserIdentity> errors = new ArrayList<>();

        boolean headerRead = false;
        
        for (String line : IOUtils.readLines(bomIS, StandardCharsets.UTF_8))
        {
            String[] columns = Arrays.stream(line.split(";")).map(StringUtils::normalizeSpace).toArray(String[]::new);
            if (!headerRead)
            {
                _checkCsvHeader(columns);
                for (int i = 0; i < columns.length; i++)
                {
                    String string = columns[i];
                    order.put(string, i);
                }
                headerRead = true;
            }
            else
            {
                Map<String, String> untypedValues = new HashMap<>();
                String login = columns[order.get(_COLUMN_LOGIN)];
                UserIdentity userIdentity = new UserIdentity(login, populationId);
                untypedValues.put("login", login);
                untypedValues.put("firstname", columns[order.get(_COLUMN_FIRSTNAME)]);
                untypedValues.put("lastname", columns[order.get(_COLUMN_LASTNAME)]);
                untypedValues.put("email", columns[order.get(_COLUMN_EMAIL)]);
                
                if (order.containsKey(_COLUMN_PASSWORD))
                {
                    untypedValues.put("password", columns[order.get(_COLUMN_PASSWORD)]);
                    untypedValues.put("clearText", "true");
                }
                
                try
                {
                    User user = _userManager.getUser(populationId, login);
                    if (user == null)
                    {
                        if (!order.containsKey(_COLUMN_PASSWORD))
                        {
                            // If there is no password column, the user is created with an invalid password
                            // To connect, the user will have to reset it's password
                            untypedValues.put("password", _DEFAULT_HASHED_PASSWORD);
                            untypedValues.put("clearText", "false");
                        }
                        
                        Map<String, Object> addedUser = _usersDao.addUser(populationId, userDirectoryId, untypedValues);
                        if (addedUser.containsKey("errors"))
                        {
                            Map<String, Errors> fieldErrors = (Map<String, Errors>) addedUser.get("errors");
                            errors.add(userIdentity);
                            getLogger().error("Failed to import user from inputs [{}] for population '{}' and user directory '{}' with the following error(s):\n{}", line, populationId, userDirectoryId, _parseErrors(fieldErrors));
                        }
                        else
                        {
                            addedUsers.add(userIdentity);
                        }
                    }
                    else
                    {
                        Map<String, Object> editedUser = _usersDao.editUser(populationId, untypedValues);
                        if (editedUser.containsKey("errors"))
                        {
                            Map<String, Errors> fieldErrors = (Map<String, Errors>) editedUser.get("errors");
                            errors.add(userIdentity);
                            getLogger().error("Failed to edit user from inputs [{}] for population '{}' and user directory '{}' with the following error(s):\n{}", line, populationId, userDirectoryId, _parseErrors(fieldErrors));
                        }
                        else
                        {
                            editedUsers.add(userIdentity);
                        }
                    }
                }
                catch (InvalidModificationException e)
                {
                    errors.add(userIdentity);
                    getLogger().error("Unable to import user from inputs [{}] for population '{}' and user directory '{}'", line, populationId, userDirectoryId, e);
                }
            }
        }
        
        Map<String, Object> eventParams = new HashMap<>();
        eventParams.put(ObservationConstants.ARGS_USERS_ADDED, addedUsers);
        eventParams.put(ObservationConstants.ARGS_USERS_UPDATED, editedUsers);
        
        if (_observationManager != null)
        {
            _observationManager.notify(new Event(ObservationConstants.EVENT_USER_IMPORTED, _currentUserProvider.getUser(), eventParams));
        }
        
        Map<String, List<UserIdentity>> result = new HashMap<>();
        result.put("added", addedUsers);
        result.put("edited", editedUsers);
        result.put("error", errors);
        return result;
    }
    
    private String _parseErrors(Map<String, Errors> errors)
    {
        StringBuilder sb = new StringBuilder();
        
        for (Entry<String, Errors> entry : errors.entrySet())
        {
            sb.append("[" + entry.getKey() + "] ");
            List<I18nizableText> errorLabels = entry.getValue().getErrors();
            for (I18nizableText errorLabel : errorLabels)
            {
                sb.append(_i18nUtils.translate(errorLabel, Locale.ENGLISH.getLanguage())).append("; ");
            }
        }
        return sb.toString();
    }
    
    /**
     * Check a line (the header) and throw an exception if not correct
     * @param columns list of fields
     * @throws ImportUserActionException thrown if error
     */
    private void _checkCsvHeader(String[] columns) throws ImportUserActionException
    {
        for (String expectedColumn : _COLUMNS)
        {
            // If column not found (unless this is the password)
            if (!ArrayUtils.contains(columns, expectedColumn) && !expectedColumn.equals(_COLUMN_PASSWORD))
            {
                List<String> params = new ArrayList<>(1);
                params.add(expectedColumn);
                throw new ImportUserActionException("PLUGINS_CORE_USERS_IMPORT_ERROR_COLUMN_MISSING", "column-missing", params);
            }
            // If column is present, but multiple times
            if (ArrayUtils.contains(columns, expectedColumn) && ArrayUtils.indexOf(columns, expectedColumn) != ArrayUtils.lastIndexOf(columns, expectedColumn))
            {
                List<String> params = new ArrayList<>(1);
                params.add(expectedColumn);
                throw new ImportUserActionException("PLUGINS_CORE_USERS_IMPORT_ERROR_COLUMN_DUPLICATE", "column-duplicate", params);
            }
        }
    }

    /**
     * Internal exception to throw with a i18n message
     */
    private static class ImportUserActionException extends Exception
    {
        private I18nizableText _i18nizableText;
        private String _errorName;

        /**
         * Create an exception with a parametrized message
         * @param i18nkey i18n key
         * @param errorName technical error name
         * @param params parameters
         */
        public ImportUserActionException(String i18nkey, String errorName, List<String> params)
        {
            this._i18nizableText = new I18nizableText("plugin.core", i18nkey, params);
            this._errorName = errorName;
        }
        
        /**
         * Get the I18nizableText
         * @return I18nizableText representing the error message
         */
        public I18nizableText getI18nizableText()
        {
            return this._i18nizableText;
        }
        /**
         * Get the technical error name
         * @return technical error name
         */
        public String getErrorName()
        {
            return this._errorName;
        }
    }
}
