/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.user;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.ContextException;
import org.apache.avalon.framework.context.Contextualizable;
import org.apache.avalon.framework.logger.AbstractLogEnabled;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.components.ContextHelper;
import org.apache.cocoon.environment.Redirector;
import org.apache.cocoon.environment.Request;
import org.apache.commons.lang3.StringUtils;

import org.ametys.core.authentication.AbstractCredentialProvider;
import org.ametys.core.authentication.AuthenticateAction;
import org.ametys.core.authentication.BlockingCredentialProvider;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.trace.ForensicLogger;
import org.ametys.core.ui.Callable;
import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.InvalidModificationException;
import org.ametys.core.user.User;
import org.ametys.core.user.User.UserCreationOrigin;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.user.directory.ModifiableUserDirectory;
import org.ametys.core.user.directory.UserDirectory;
import org.ametys.core.user.population.UserPopulation;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.plugins.core.impl.user.directory.JdbcUserDirectory;
import org.ametys.plugins.core.impl.user.directory.JdbcUserDirectory.StrongPasswordRequirements;
import org.ametys.runtime.authentication.AccessDeniedException;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.model.DefinitionContext;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewElement;
import org.ametys.runtime.parameter.Errors;

/**
 * DAO for manipulating {@link User}
 *
 */
public class UserDAO extends AbstractLogEnabled implements Component, Contextualizable, Serviceable 
{
    /** The avalon role */
    public static final String ROLE = UserDAO.class.getName();
    
    /** The service manager */
    protected ServiceManager _smanager;
    /** The user manager */
    protected UserManager _userManager;
    /** The user population DAO */
    protected UserPopulationDAO _userPopulationDAO;
    /** The current user provider. */
    protected CurrentUserProvider _currentUserProvider;
    /** The Avalon context */
    protected Context _context;
    /** The user helper */
    protected UserHelper _userHelper;
    /** The right manager */
    protected RightManager _rightManager;

    public void contextualize(Context context) throws ContextException
    {
        _context = context;
    }
    
    public void service(ServiceManager smanager) throws ServiceException
    {
        _smanager = smanager;
        _userManager = (UserManager) smanager.lookup(UserManager.ROLE);
        _userPopulationDAO = (UserPopulationDAO) smanager.lookup(UserPopulationDAO.ROLE);
        _userHelper = (UserHelper) smanager.lookup(UserHelper.ROLE);
        _rightManager = (RightManager) smanager.lookup(RightManager.ROLE);
    }
    
    /**
     * Get user's information
     * @param login The user's login
     * @param populationId The id of the population
     * @return The user's information
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> getUser (String login, String populationId)
    {
        if (!_canHandle(populationId))
        {
            throw new AccessDeniedException("The user " + _getCurrentUser() + " tried to get user without sufficient rights");
        }
        
        return _userHelper.user2json(_userManager.getUser(populationId, login), true);
    }
    
    /**
     * Checks if the user is modifiable and if current user is allowed to edit it
     * @param login The users's login
     * @param populationId The id of the population of the user
     * @return A map with the "isModifiable" at true if the user is modifiable
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> isModifiable (String login, String populationId)
    {
        Map<String, Object> result = new HashMap<>();
        
        boolean isModifiable = _userManager.getUserDirectory(populationId, login) instanceof ModifiableUserDirectory;
        result.put("isModifiable", isModifiable && _canHandle(populationId));
        result.put("additionalDescription", new I18nizableText("plugin.core-ui", "PLUGINS_CORE_UI_USERS_EDIT_NO_MODIFIABLE_DESCRIPTION"));
        
        return result;
    }
    
    /**
     * Checks if the user is removable and if current user is allowed to remove it
     * @param login The users's login
     * @param populationId The id of the population of the user
     * @return A map with the "isRemovable" at true if the user is removable
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> isRemovable (String login, String populationId)
    {
        Map<String, Object> result = new HashMap<>();
        
        UserDirectory userDirectory = _userManager.getUserDirectory(populationId, login);
        if (userDirectory != null && populationId.equals(UserPopulationDAO.ADMIN_POPULATION_ID) && userDirectory.getUsers().size() == 1)
        {
            // Impossible to delete the last user of the admin population
            result.put("isRemovable", false);
            result.put("additionalDescription", new I18nizableText("plugin.core-ui", "PLUGINS_CORE_UI_USERS_DELETE_LAST_ADMIN_DESCRIPTION"));
        }
        else
        {
            boolean isRemovable = userDirectory instanceof ModifiableUserDirectory;
            result.put("isRemovable", isRemovable && _canHandle(populationId));
            result.put("additionalDescription", new I18nizableText("plugin.core-ui", "PLUGINS_CORE_UI_USERS_DELETE_NO_MODIFIABLE_DESCRIPTION"));
        }
        return result;
    }
    
    /**
     * Creates a User
     * @param populationId The id of the user population
     * @param userDirectoryId The id of the user directory
     * @param untypedValues The untyped user's parameters
     * @return The created user as JSON object
     * @throws InvalidModificationException If modification is not possible 
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> addUser(String populationId, String userDirectoryId, Map<String, String> untypedValues) throws InvalidModificationException
    {
        Map<String, Object> result = new HashMap<>();
        
        UserPopulation userPopulation = _userPopulationDAO.getUserPopulation(populationId);
        UserDirectory userDirectory = userPopulation.getUserDirectory(userDirectoryId);

        if (userDirectory == null)
        {
            throw new InvalidModificationException("User directory with ID '" + userDirectoryId + "' in population '" + populationId + "' has not been found.");
        }
        
        if (!(userDirectory instanceof ModifiableUserDirectory))
        {
            throw new InvalidModificationException("User directory is not modifiable!");
        }
        
        if (!_canHandle(populationId))
        {
            throw new AccessDeniedException("The user " + _getCurrentUser() + " tried to add user without sufficient rights");
        }

        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) userDirectory;
        
        String login = untypedValues.get("login");
        
        try
        {
            if (getLogger().isInfoEnabled())
            {
                getLogger().info(String.format("User %s is adding a new user '%s'", _getCurrentUser(), login));
            }

            modifiableUserDirectory.add(untypedValues, UserCreationOrigin.ADMIN);
            return _userHelper.user2json(modifiableUserDirectory.getUser(login), true);
        }
        catch (InvalidModificationException e)
        {
            Map<String, Errors> fieldErrors = e.getFieldErrors();
            
            if (fieldErrors != null && fieldErrors.size() > 0)
            {
                result.put("errors", fieldErrors);
            }
            else
            {
                throw e;
            }
        }

        if (getLogger().isDebugEnabled())
        {
            getLogger().debug("Ending user's edition");
        }
        
        return result;
    }
    
    /**
     * Check current user can handle the population of users
     * @param populationId the id of population
     * @return true if user is allowed to handle this population
     */
    protected boolean _canHandle(String populationId)
    {
        UserIdentity currentUser = _getCurrentUser();
        
        return  _rightManager.hasRight(currentUser, "Runtime_Rights_User_Handle", "/${WorkspaceName}") == RightResult.RIGHT_ALLOW
                  || _rightManager.hasRight(currentUser, "Runtime_Rights_User_OwnPopulation_Handle", "/${WorkspaceName}") == RightResult.RIGHT_ALLOW && currentUser.getPopulationId().equals(populationId);
    }
    
    /**
     * Edits a User
     * @param populationId The id of the population of the user to edit
     * @param untypedValues The untyped user's parameters
     * @return The update user as JSON object
     * @throws InvalidModificationException If modification is not possible 
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> editUser (String populationId, Map<String, String> untypedValues) throws InvalidModificationException
    {
        Map<String, Object> result = new HashMap<>();
        
        String login = untypedValues.get("login");
        UserDirectory userDirectory = _userManager.getUserDirectory(populationId, login);

        if (userDirectory == null)
        {
            throw new InvalidModificationException("User directory for user '" + login + "' with population '" + populationId + "' has not been found, maybe the user does not exists anymore.");
        }
        
        if (!(userDirectory instanceof ModifiableUserDirectory))
        {
            throw new InvalidModificationException("Users are not modifiable!");
        }
        
        if (!_canHandle(populationId))
        {
            throw new AccessDeniedException("The user " + _getCurrentUser() + " tried to edit user without sufficient rights");
        }
        
        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) userDirectory;
        
        try
        {
            if (getLogger().isInfoEnabled())
            {
                getLogger().info(String.format("User %s is updating information about user '%s'", _getCurrentUser(), login));
            }

            modifiableUserDirectory.update(untypedValues);
            return _userHelper.user2json(modifiableUserDirectory.getUser(login), true);
        }
        catch (InvalidModificationException e)
        {
            Map<String, Errors> fieldErrors = e.getFieldErrors();
            
            if (fieldErrors != null && fieldErrors.size() > 0)
            {
                result.put("errors", fieldErrors);
            }
            else
            {
                throw e;
            }
        }

        if (getLogger().isDebugEnabled())
        {
            getLogger().debug("Ending user's edition");
        }
        
        return result;
    }
    
    /**
     * Deletes users
     * @param users The users to delete
     * @return the result with deleted users and undeleted users
     * @throws InvalidModificationException If modification is not possible 
     */
    @Callable (rights = Callable.SKIP_BUILTIN_CHECK)
    public Map<String, Object> deleteUsers (List<Map<String, String>> users) throws InvalidModificationException
    {
        List<Map<String, String>> deletedUsers = new ArrayList<>();
        List<Map<String, String>> unauthorizedUsers = new ArrayList<>();
        
        for (Map<String, String> user : users)
        {
            String login = user.get("login");
            String populationId = user.get("populationId");
            
            try
            {
                _deleteUser(login, populationId);
                deletedUsers.add(user);
            }
            catch (AccessDeniedException e)
            {
                unauthorizedUsers.add(user);
            }
            
        }
       
        if (getLogger().isDebugEnabled())
        {
            getLogger().debug("Ending user's removal");
        }
        
        return Map.of("deletedUsers", deletedUsers, "unauthorizedUsers", unauthorizedUsers);
    }
    
    private void _deleteUser(String login, String populationId) throws InvalidModificationException
    {
        UserDirectory userDirectory = _userManager.getUserDirectory(populationId, login);

        if (userDirectory == null)
        {
            throw new InvalidModificationException("User directory for user '" + login + "' with population '" + populationId + "' has not been found, maybe the user does not exists anymore.");
        }
        
        if (!(userDirectory instanceof ModifiableUserDirectory))
        {
            throw new InvalidModificationException("User directory is not modifiable!");
        }
        
        if (populationId.equals(UserPopulationDAO.ADMIN_POPULATION_ID) && userDirectory.getUsers().size() == 1)
        {
            // Impossible to delete the last user of the admin population
            throw new InvalidModificationException("You cannot delete the last user of the 'admin' population!");
        }
        
        if (!_canHandle(populationId))
        {
            throw new AccessDeniedException("The user " + _getCurrentUser() + " tried to delete user without sufficient rights");
        }
        
        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) userDirectory;
        
        if (getLogger().isInfoEnabled())
        {
            getLogger().info(String.format("User %s is removing user '%s'", _getCurrentUser(), login));
        }
        
        modifiableUserDirectory.remove(login);
    }
    
    /**
     * Get the users edition model 
     * @param login The user's login to edit
     * @param populationId The id of the population of the user to edit
     * @throws InvalidModificationException If modification is not possible 
     * @throws ProcessingException If there is another exception
     * @return The edition model as an object
     */
    @Callable (rights = {"Runtime_Rights_User_Handle", "Runtime_Rights_User_OwnPopulation_Handle"})
    public Map<String, Object> getEditionModelForUSer(String login, String populationId) throws InvalidModificationException, ProcessingException
    {
        UserDirectory userDirectory = _userManager.getUserDirectory(populationId, login);

        if (userDirectory == null)
        {
            throw new InvalidModificationException("User directory for user '" + login + "' with population '" + populationId + "' has not been found, maybe the user does not exists anymore.");
        }
        
        if (!_canHandle(populationId))
        {
            throw new AccessDeniedException("The user " + _getCurrentUser() + " tried to access user handle without sufficient rights");
        }
        
        return _getEditionModel(userDirectory);
    }
    
    /**
     * Get the users edition model 
     * @param populationId The id of the population where to add a user
     * @param userDirectoryId The id of the user directory where to add a user
     * @throws InvalidModificationException If modification is not possible 
     * @throws ProcessingException If there is another exception
     * @return The edition model as an object
     */
    @Callable (rights = {"Runtime_Rights_User_Handle", "Runtime_Rights_User_OwnPopulation_Handle"})
    public Map<String, Object> getEditionModelForDirectory(String populationId, String userDirectoryId) throws InvalidModificationException, ProcessingException
    {
        UserDirectory userDirectory = _userPopulationDAO.getUserPopulation(populationId).getUserDirectory(userDirectoryId);

        if (userDirectory == null)
        {
            throw new InvalidModificationException("User directory with ID '" + userDirectoryId + "' in population '" + populationId + "' has not been found.");
        }
        
        if (!_canHandle(populationId))
        {
            throw new AccessDeniedException("The user " + _getCurrentUser() + " tried to handle user without sufficient rights");
        }
        
        return _getEditionModel(userDirectory);
    }
    
    private Map<String, Object> _getEditionModel (UserDirectory userDirectory) throws InvalidModificationException, ProcessingException
    {
        if (!(userDirectory instanceof ModifiableUserDirectory))
        {
            throw new InvalidModificationException("User directory is not modifiable!");
        }
        
        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) userDirectory;
        View view = modifiableUserDirectory.getView();
        
        if (modifiableUserDirectory instanceof JdbcUserDirectory jdbcUserDirectory)
        {
            if (jdbcUserDirectory.useStrongPassword())
            {
                StrongPasswordRequirements passwordRequirements = jdbcUserDirectory.getStrongPasswordRequirements();
                Map<String, I18nizableText> widgetParameters = passwordRequirements.getWidgetParameters();
                
                ViewElement modelViewItem = (ViewElement) view.getModelViewItem("password");
                modelViewItem.getDefinition().setWidgetParameters(widgetParameters);
            }
        }
        
        return view.toJSON(DefinitionContext.newInstance().withEdition(true));
    }
    
    /**
     * Impersonate the selected user
     * @param login the login of the user to impersonate
     * @param populationId The id of the population
     * @return a map of information on the user
     * @throws AccessDeniedException If the currently connected user has no right to do so
     */
    @Callable (rights = "Runtime_Rights_User_Handle", context = "/admin")
    public Map<String, String> impersonate(String login, String populationId) throws AccessDeniedException
    {
        UserIdentity currentUser = _getCurrentUser();
        
        if (StringUtils.isEmpty(login))
        {
            throw new IllegalArgumentException("'login' parameter is null or empty");
        }
        
        Map<String, String> result = new HashMap<>();
        
        User user = _userManager.getUser(populationId, login);
        if (user == null)
        {
            result.put("error", "unknown-user");   
        }
        else
        {
            try
            {
                _currentUserProvider.logout(null);
            }
            catch (ProcessingException e)
            {
                getLogger().error("An error occurred while logging out current user " + currentUser);
            }
            
            Request request = ContextHelper.getRequest(_context);
            
            AuthenticateAction.setUserIdentityInSession(request, user.getIdentity(), new ImpersonateCredentialProvider(), true);
            
            result.put("login", login);
            result.put("populationId", populationId);
            result.put("name", user.getFullName());
            
            Map<String, Object> impersonateArgs = Map.of("login", login, "population", populationId);
            ForensicLogger.info("authentication.impersonation", impersonateArgs, currentUser);
        }
        
        return result;
    }
    
    /**
     * Provides the login of the current user.
     * @return the login which cannot be <code>null</code>.
     */
    protected UserIdentity _getCurrentUser()
    {
        if (_currentUserProvider == null)
        {
            try
            {
                _currentUserProvider = (CurrentUserProvider) _smanager.lookup(CurrentUserProvider.ROLE);
            }
            catch (ServiceException e)
            {
                throw new IllegalStateException(e);
            }
        }
        
        return _currentUserProvider.getUser();
    }
    
    /**
     * A fake credential provider used by the impersonate process
     */
    public static class ImpersonateCredentialProvider extends AbstractCredentialProvider implements BlockingCredentialProvider
    {
        public boolean blockingGrantAnonymousRequest()
        {
            return false;
        }

        public boolean blockingIsStillConnected(UserIdentity userIdentity, Redirector redirector)
        {
            return true;
        }

        public UserIdentity blockingGetUserIdentity(Redirector redirector)
        {
            return null;
        }

        public void blockingUserNotAllowed(Redirector redirector)
        {
            // nothing
        }

        public void blockingUserAllowed(UserIdentity userIdentity, Redirector redirector)
        {
            // nothing
        }

        public boolean requiresNewWindow()
        {
            return false;
        }
    }
}
