/*
 *  Copyright 2015 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.user;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import javax.xml.transform.TransformerException;

import org.apache.avalon.framework.component.Component;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.ConfigurationUtil;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.apache.xpath.XPathAPI;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;

import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.UserManager;
import org.ametys.core.user.directory.NotUniqueUserException;
import org.ametys.core.user.directory.UserDirectoryFactory;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.runtime.plugin.component.AbstractLogEnabled;

/**
 * Simple user helper, for common function working on {@link User}
 */
public class UserHelper extends AbstractLogEnabled implements Component, Serviceable
{
    /** The Avalon role */
    public static final String ROLE = UserHelper.class.getName();
    
    /** The user population DAO */
    private UserPopulationDAO _userPopulationDAO;

    /** The user directory factory */
    private UserDirectoryFactory _userDirectoryFactory;

    private UserManager _userManager;
    
    @Override
    public void service(ServiceManager smanager) throws ServiceException
    {
        _userManager = (UserManager) smanager.lookup(UserManager.ROLE);
        _userPopulationDAO = (UserPopulationDAO) smanager.lookup(UserPopulationDAO.ROLE);
        _userDirectoryFactory = (UserDirectoryFactory) smanager.lookup(UserDirectoryFactory.ROLE);
    }
    
    /**
     * Get the user's full name handling request cache
     * @param userIdentity The user identity
     * @return user's full name or null if user does not exist
     */
    public String getUserFullName (UserIdentity userIdentity)
    {
        User user = _userManager.getUser(userIdentity);
        return user != null ? user.getFullName() : null;
    }
    
    /**
     * Get the user's sortable name handling request cache
     * @param userIdentity The user identity
     * @return user's full name or null if user does not exist
     */
    public String getUserSortableName (UserIdentity userIdentity)
    {
        User user = _userManager.getUser(userIdentity);
        return user != null ? user.getSortableName() : null;
    }
    
    /**
     * Populate a list of map, where each map representing an user.
     * @param users The list of users
     * @return The list of map.
     */
    public List<Map<String, Object>> userIdentities2json(Collection<UserIdentity> users)
    {
        return users.stream().map(identity -> _userManager.getUser(identity)).filter(Objects::nonNull).map(user -> user2json(user)).collect(Collectors.toList());
    }
    
    
    /**
     * Populate a list of map, where each map representing an user.
     * @param users The list of users
     * @return The list of map.
     */
    public List<Map<String, Object>> users2json(Collection<User> users)
    {
        return users2json(users, false);
    }
    
    /**
     * Populate a list of map, where each map representing an user.
     * @param users The list of users
     * @param full Set to <code>true</code> to get full information on user
     * @return The list of map.
     */
    public List<Map<String, Object>> users2json(Collection<User> users, boolean full)
    {
        List<Map<String, Object>> userList = new ArrayList<>();
        
        Set<User> distinctUsers = new LinkedHashSet<>(users);
        for (User user : distinctUsers)
        {
            userList.add(user2json(user, full));
        }
        
        return userList;
    }
        
    /**
     * Get the JSON object representing a user
     * @param userIdentity The user identity
     * @return The user as JSON object
     */
    public Map<String, Object> user2json (UserIdentity userIdentity)
    {
        return user2json(userIdentity, false);
    }
    
    /**
     * Get the JSON object representing a user
     * @param userIdentity The user identity
     * @param full Set to <code>true</code> to get full information on user
     * @return The user as JSON object
     */
    public Map<String, Object> user2json (UserIdentity userIdentity, boolean full)
    {
        User user = _userManager.getUser(userIdentity);
        if (user != null)
        {
            return user2json(user, full);
        }
        return Collections.EMPTY_MAP;
    }
    
    /**
     * Get the JSON object representing a user
     * @param user The user
     * @return The user as JSON object
     */
    public Map<String, Object> user2json(User user)
    {
        return _user2json(user, false);
    }
    
    /**
     * Get the JSON object representing a user
     * @param user The user
     * @param full Set to <code>true</code> to get full information on user
     * @return The user as JSON object or null, if the given user is null
     */
    public Map<String, Object> user2json(User user, boolean full)
    {
        return _user2json(user, full);
    }
    
    private Map<String, Object> _user2json(User user, boolean full)
    {
        if (user == null)
        {
            return null;
        }
        
        Map<String, Object> userInfos = new HashMap<>();
        
        userInfos.put("login", user.getIdentity().getLogin());
        
        String populationId = user.getIdentity().getPopulationId();
        userInfos.put("populationId", populationId);
        
        userInfos.put("fullname", user.getFullName());
        userInfos.put("sortablename", user.getSortableName());
        
        if (full)
        {
            String udModelId = user.getUserDirectory() != null ? user.getUserDirectory().getUserDirectoryModelId() : "";
            
            userInfos.put("populationLabel", _userPopulationDAO.getUserPopulation(populationId).getLabel());
            userInfos.put("directory", _userDirectoryFactory.hasExtension(udModelId) ? _userDirectoryFactory.getExtension(udModelId).getLabel() : "");
            
            userInfos.put("lastname", user.getLastName());
            userInfos.put("firstname", user.getFirstName());
            userInfos.put("email", user.getEmail());
            
            userInfos.put("creationDate", user.getCreationDate());
            userInfos.put("creationOrigin", user.getCreationOrigin());
        }
        
        return userInfos;
    }
    
    /**
     * Get the user identity from a JSON object
     * @param json the JSON object representing the user
     * @return the user identity
     */
    public UserIdentity json2userIdentity(Map<String, ? extends Object> json)
    {
        // If the user is empty, return null
        return Optional.ofNullable(json)
                .filter(u -> !u.isEmpty() && u.get("login") != null && u.get("populationId") != null)
                .map(u -> new UserIdentity((String) u.get("login"), (String) u.get("populationId")))
                .orElse(null);
    }
    
    /**
     * Get the user from its identity
     * @param userIdentity The user identity
     * @return The user or null if not found
     * @deprecated use {@link UserManager#getUser(UserIdentity)} instead
     */
    @Deprecated
    public User getUser (UserIdentity userIdentity)
    {
        if (userIdentity == null)
        {
            return null;
        }
         
        return _userManager.getUser(userIdentity.getPopulationId(), userIdentity.getLogin());
    }
    
    /**
     * Get the user from its email
     * @param populationId The population id
     * @param email The email
     * @return The user or null if not found or if multiple user were found with this email
     * @deprecated use {@link UserManager#getUserByEmail(String, String)} instead
     */
    @Deprecated
    public User getUserByEmail (String populationId, String email)
    {
        if (email == null)
        {
            return null;
        }

        try
        {
            return _userManager.getUserByEmail(populationId, email);
        }
        catch (NotUniqueUserException e)
        {
            getLogger().error("Cannot find user in population '" + populationId + "' by email '" + email + "' because 2 or more users matches", e);
            return null;
        }
    }

    /**
     * SAX an user identity
     * @param userIdentity The user identity to SAX
     * @param handler The content handler
     * @throws SAXException If a SAX error occurs
     */
    public void saxUserIdentity(UserIdentity userIdentity, ContentHandler handler) throws SAXException
    {
        saxUserIdentity(userIdentity, handler, "user");
    }
    
    /**
     * SAX an user identity
     * @param userIdentity The user identity to SAX
     * @param handler The content handler
     * @param tagName The XML tag for saxed user
     * @throws SAXException If a SAX error occurs
     */
    public void saxUserIdentity(UserIdentity userIdentity, ContentHandler handler, String tagName) throws SAXException
    {
        User user = _userManager.getUser(userIdentity);
        if (user != null)
        {
            saxUser(user, handler, tagName);
        }
        else
        {
            getLogger().warn("Unable to sax unknown user with identity {}", userIdentity);
        }
           
    }
    
    /**
     * SAX an user
     * @param user The user. If null no sax events will be generated.
     * @param handler The content handler
     * @throws SAXException If a SAX error occurs
     */
    public void saxUser(User user, ContentHandler handler) throws SAXException
    {
        saxUser(user, handler, "user");
    }
    
    /**
     * SAX an user
     * @param user The user. If null no sax events will be generated.
     * @param handler The content handler
     * @param tagName The XML tag for saxed user
     * @throws SAXException If a SAX error occurs
     */
    public void saxUser(User user, ContentHandler handler, String tagName) throws SAXException
    {
        if (user != null)
        {
            AttributesImpl attr = new AttributesImpl();
            attr.addCDATAAttribute("login", user.getIdentity().getLogin());
            attr.addCDATAAttribute("population", user.getIdentity().getPopulationId());
            
            XMLUtils.startElement(handler, tagName, attr);
            
            XMLUtils.createElement(handler, "lastname", user.getLastName());
            XMLUtils.createElement(handler, "firstname", user.getFirstName());
            XMLUtils.createElement(handler, "email", user.getEmail());
            
            XMLUtils.createElement(handler, "fullname", user.getFullName());
            XMLUtils.createElement(handler, "sortablename", user.getSortableName());
    
            _userPopulationDAO.getUserPopulation(user.getIdentity().getPopulationId()).getLabel().toSAX(handler, "populationLabel");
    
            XMLUtils.endElement(handler, tagName);
        }
        else if (getLogger().isWarnEnabled())
        {
            getLogger().warn("Unable to sax null user", new Exception());
        }
    }
    
    /**
     * Get the user identity from an XML configuration
     * @param userConfiguration The configuration
     * @return The user identity
     * @throws ConfigurationException if an error occurs during value retrieving
     */
    public UserIdentity xml2userIdentity(Configuration userConfiguration) throws ConfigurationException
    {
        Element userElement = ConfigurationUtil.toElement(userConfiguration);
        try
        {
            return xml2userIdentity(userElement);
        }
        catch (TransformerException e)
        {
            throw new ConfigurationException("An error occurred while parsing the configuration of an I/O data", e);
        }
    }
    
    /**
     * Get the user identity from a DOM Node object
     * @param userNode The node
     * @return The user identity
     * @throws TransformerException if an error occurs during value retrieving
     */
    public UserIdentity xml2userIdentity(Node userNode) throws TransformerException
    {
        String login = Objects.requireNonNull(XPathAPI.eval(userNode, "@login").str());
        String population = Objects.requireNonNull(XPathAPI.eval(userNode, "@population").str());
        return new UserIdentity(login, population);
    }
}
