/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import org.junit.After;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Rule;
import org.junit.rules.TestName;
import org.junit.runner.RunWith;
import org.junit.runner.Runner;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Ametys Test Cases must <b>always</b> extend this class.
 * <br>It <ul>
 * <li>indicates to run with our extended {@link Runner}</li>
 * <li>provides a static and a non-static {@link Logger}</li>
 * <li>logs before/after the Test Case and before/after each Test</li>
 * </ul>
 */
@RunWith(AmetysJUnitClassRunner.class)
public abstract class AbstractBaseTestCase
{
    /** The static logger to use in static methods */
    protected static Logger _staticLogger = LoggerFactory.getLogger(AbstractBaseTestCase.class);
    
    /** The logger */
    protected Logger _logger = LoggerFactory.getLogger(getClass());
    
    private TestName _name = new TestName();
    
    /**
     * Returns the test name.
     * @return the test name
     */
    @Rule
    public final TestName getTestName()
    {
        return _name;
    }
    
    /**
     * Executed before the Test Case. For printing log purposes.
     */
    @BeforeClass
    public static final void baseBeforeClass()
    {
        _logStartTestCase();
    }
    
    private static void _logStartTestCase()
    {
        _staticLogger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
        _staticLogger.info("Begin test case");
        _staticLogger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
    }
    
    /**
     * Executed after the Test Case. For printing log purposes.
     */
//    @AfterClass
    // Not annotated AfterClass to avoid to log before doing our custom job of BeforeAfterClassTestCase#afterClass (as we want these logs to be the last ones)
    public static final void baseAfterClass()
    {
        _logEndTestCase();
    }
    
    private static void _logEndTestCase()
    {
        _staticLogger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
        _staticLogger.info("End test case");
        _staticLogger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
    }
    
    /**
     * Executed before each Test. For printing log purposes.
     */
    @Before
    public final void baseSetup()
    {
        _logStartTest();
    }
    
    private void _logStartTest()
    {
        _logger.info("========================================");
        _logger.info("Begin test " + getTestName().getMethodName());
        _logger.info("========================================");
    }
    
    /**
     * Executed after each Test. For printing log purposes.
     */
    @After
    public final void baseTeardown()
    {
        _logEndTest();
    }
    
    private void _logEndTest()
    {
        _logger.info("========================================");
        _logger.info("End test " + getTestName().getMethodName());
        _logger.info("========================================");
    }
}
