/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.core.user.CurrentUserProvider;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.runtime.test.SimpleUserProvider;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Abstract test case providing a population and users.
 */
@NeedFeatures({
    "core/runtime.users.manager", 
    "core-impl/runtime.user.directory.impl.static", 
    "core-impl/runtime.authentication.impl.defined",
    
    "test/test.user.provider",
})
public abstract class AbstractPopulationAwareTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    /** The id of the test population. It is already created and removed. */
    protected static final String POPULATION_TEST_ID = "static";
    /** The current user */
    protected static final UserIdentity CURRENT_USER_TEST_IDENTITY = new UserIdentity("current-user", POPULATION_TEST_ID);
    /** The test user 'anonymous' */
    protected static final UserIdentity ANONYMOUS_TEST_IDENTITY = new UserIdentity("anonymous", POPULATION_TEST_ID);
    /** The test user 'user1' */
    protected static final UserIdentity USER1_TEST_IDENTITY = new UserIdentity("user1", POPULATION_TEST_ID);
    /** The test user 'user2' */
    protected static final UserIdentity USER2_TEST_IDENTITY = new UserIdentity("user2", POPULATION_TEST_ID);
    /** The test user 'user3' */
    protected static final UserIdentity USER3_TEST_IDENTITY = new UserIdentity("user3", POPULATION_TEST_ID);

    /** The current user provider */
    protected static SimpleUserProvider _userProvider;
    /** The user population DAO */
    protected static UserPopulationDAO _userPopulationDAO;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _userProvider = (SimpleUserProvider) manager.lookup(CurrentUserProvider.ROLE);
        _userPopulationDAO = (UserPopulationDAO) manager.lookup(UserPopulationDAO.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        _setCurrentUser();
        _createPopulation();
    }
    
    private static void _setCurrentUser()
    {
        _userProvider.setUser(CURRENT_USER_TEST_IDENTITY);
    }
    
    private static void _createPopulation()
    {
        Map<String, Object> userDirectory = new LinkedHashMap<>();
        String udModelId = "org.ametys.plugins.core.user.directory.Static";
        userDirectory.put("udModelId", udModelId);
        String usersAsString =
                CURRENT_USER_TEST_IDENTITY.getLogin() + ":CurrentUser:User\n"
                + ANONYMOUS_TEST_IDENTITY.getLogin() + ":Anonymous:User\n"
                + USER1_TEST_IDENTITY.getLogin() + ":One:User\n"
                + USER2_TEST_IDENTITY.getLogin() + ":Two:User\n"
                + USER3_TEST_IDENTITY.getLogin() + ":Three:User";
        userDirectory.put(udModelId + "$" + "runtime.users.static.users", usersAsString);
        
        Map<String, Object> credentialProvider = new LinkedHashMap<>();
        String cpModelId = "org.ametys.core.authentication.Defined";
        credentialProvider.put("cpModelId", cpModelId);
        credentialProvider.put(cpModelId + "$" + "runtime.authentication.defined.user", "anonymous");
        
        _userPopulationDAO.add(POPULATION_TEST_ID, "Static Population", Collections.singletonList(userDirectory), Collections.singletonList(credentialProvider));
    }
    
    @Override
    protected void essentialsAfter() throws Exception
    {
        _removeUserPopulation();
        super.essentialsAfter();
    }
    
    private static void _removeUserPopulation()
    {
        boolean hasTestPopulation = false;
        try
        {
            hasTestPopulation = _userPopulationDAO != null && _userPopulationDAO.getUserPopulation(POPULATION_TEST_ID) != null;
        }
        catch (Exception e)
        {
            // Silently ignore, Cocoon must have been not well loaded for instance, etc.
        }
        
        if (hasTestPopulation)
        {
            _userPopulationDAO.remove(POPULATION_TEST_ID);
        }
    }
}
