/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collection;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.environment.Environment;
import org.apache.cocoon.environment.ObjectModelHelper;
import org.apache.cocoon.environment.Response;
import org.apache.cocoon.environment.commandline.CommandLineRequest;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.io.FileUtils;
import org.junit.After;
import org.junit.Before;

import org.ametys.runtime.data.AmetysHomeLock;
import org.ametys.runtime.plugin.IncludePolicyFeatureActivator.IncludedFeature;
import org.ametys.runtime.plugin.PluginsManager;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

import com.sun.net.httpserver.HttpServer;

/**
 * Test Cases which needs to be {@link ScriptTestCase} or {@link CocoonTestCase} must extend this class.
 * <br>
 * <br>By default, processings in {@link #essentialsBefore()} and {@link #essentialsAfter()} (such as running a script, start a Cocoon application...)
 * are done once before any test and once after all tests. Annotate your Test Case class with {@link Unoptimized}
 * if you need to do those processings before and after each test.
 * <br>
 * <br>If the class implements {@link CocoonTestCase}, indicates the features the test needs by annotating the class or/and the method with {@link NeedFeatures}.
 */
public abstract class AbstractRuntimeTestCase extends AbstractBaseTestCase implements BeforeAndAfterClassTestCase, WithRuntimeFeaturesTestCase
{
    /** The path for Ametys Home which will be copied for starting Cocoon */
    protected static final Path DATA_HOME_PATH = Paths.get("test", "environments", "webapp", "WEB-INF", "data");
    /** The tmp path */
    protected static final Path TMP_PATH = Paths.get("tmp");
    
    /** The lock on Ametys home */
    protected static AmetysHomeLock _ametysHomeLock;
    private static String _webappPath;
    
    private static DatabaseProperties _dbProperties;
    private static VariableHandler _variableHandler;
    private static CocoonWrapper _cocoon;
    private static TestConfigSet _resolvedConfigSet;
    private static Map<String, Object> _environment;
    // by default, is optimized, but if annotation is present, then it is not
    private final boolean _optimized = !getClass().isAnnotationPresent(Unoptimized.class);
    
    @Override
    public final void beforeClass() throws Exception
    {
        if (_optimized)
        {
            _beforeCatchingException();
        }
    }
    
    /**
     * Executed before each test
     * @throws Exception If an error occurs
     */
    @Before
    public final void beforeTest() throws Exception
    {
        if (!_optimized)
        {
            _beforeCatchingException();
        }
    }
    
    private void _beforeCatchingException() throws Exception
    {
        try
        {
            essentialsBefore();
        }
        catch (Throwable t)
        {
            try
            {
                essentialsAfter();
            }
            catch (Exception e)
            {
                _logger.error("An unexpected exception occured during tearDown after an error occured during setUp", e);
                _logger.error("The initial error was: ", t);
            }
            throw new Exception("Exception during setUp", t);
        }
    }
    
    /**
     * Essential processings (such as executing script, start Cocoon...)
     * done before each test (if class is annotated {@link Unoptimized})
     * or done once before any test (if class is not annotated {@link Unoptimized}).
     * @throws Exception If an error occurs
     */
    protected void essentialsBefore() throws Exception
    {
        _deleteDirectories();
        _dbProperties = new DatabaseProperties();
        _variableHandler = new VariableHandler(this, _dbProperties);
        _dbProperties.executeStartScripts(this, _variableHandler.variableValues());
        _startCocoon();
    }
    
    private static void _deleteDirectories()
    {
        long start = System.currentTimeMillis();
        for (Path path : new Path[] {DATA_HOME_PATH, TMP_PATH})
        {
            try
            {
                FileUtils.deleteDirectory(path.toFile());
                _staticLogger.info("Directory '{}' was successfully removed.", path);
            }
            catch (Exception e)
            {
                _staticLogger.error("Cannot remove directory '{}'. It can lead to unstable test running.", path, e);
            }
        }
        _staticLogger.info("Deletion of directories ended in {}ms", System.currentTimeMillis() - start);
    }
    
    private void _startCocoon() throws Exception
    {
        if (this instanceof CocoonTestCase)
        {
            long start = System.currentTimeMillis();
            
            CocoonTestCase cocoonTestInstance = (CocoonTestCase) this;
            
            TestConfigSet configSet = cocoonTestInstance.getConfigSet();
            Map<String, String> variableValues = _variableHandler.variableValues();
            _logger.info("Using configSet {}\n with values {}", configSet, _variableHandler.variableValuesWithHiddenPasswords());
            _resolvedConfigSet = configSet.resolve(variableValues);
            _webappPath = _resolvedConfigSet.getWebappPath();
            _logger.info("Using resolved configSet {}", _resolvedConfigSet);
            
            String contextPath = configSet.getContextPath();
            
            startCocoon(_resolvedConfigSet);
            if (cocoonTestInstance.withInitialEnvironment())
            {
                _environment = enterEnvironment(contextPath);
            }
            
            ServiceManager serviceManager = _cocoon.getServiceManager();
            if (isPluginsManagerOk() && serviceManager != null)
            {
                cocoonTestInstance.lookupComponents(serviceManager);
                cocoonTestInstance.prepareComponents();
            }
            else
            {
                _logger.error("No ServiceManager available. It may be due to PluginsManager in safe or error mode. Please refer to previous logs.");
            }
            
            _logger.info("Start of Cocoon ended in {}ms", System.currentTimeMillis() - start);
        }
    }
    
    /**
     * <b>Use with care, if you know what you do</b>
     * <br>Starts Cocoon manually with all features. You often do not need to call this method, and implements {@link CocoonTestCase} instead.
     * @param configSet The ConfigSet 
     * @throws Exception if an error occurs
     */
    protected void startCocoon(TestConfigSet configSet) throws Exception
    {
        _startCocoonApplication(
                configSet, 
                this instanceof CocoonTestCase 
                    ? new NeededFeatures(this, _optimized).compute(Optional.ofNullable(((CocoonTestCase) this).getDBType()))
                    : null
        );
    }
    
    /**
     * <b>Use with care, if you know what you do</b>
     * <br>Starts Cocoon manually. You often do not need to call this method, and implements {@link CocoonTestCase} instead.
     * @param configSet The ConfigSet 
     * @param useIncludePolicyWithFeatures If not null, Cocoon will initialize with only the given features and their dependencies.
     * @throws Exception if an error occurs
     */
    protected void startCocoon(TestConfigSet configSet, Collection<IncludedFeature> useIncludePolicyWithFeatures) throws Exception
    {
        _startCocoonApplication(
                configSet, 
                useIncludePolicyWithFeatures == null
                    ? null
                    : CollectionUtils.union(
                          useIncludePolicyWithFeatures, 
                          new NeededFeatures(this, _optimized).compute(Optional.empty())
                      )
        );
    }
    
    private void _startCocoonApplication(TestConfigSet configSet, Collection<IncludedFeature> useIncludePolicyWithFeatures) throws Exception
    {
        Objects.requireNonNull(configSet);
        Objects.requireNonNull(configSet.getRuntime());
        Objects.requireNonNull(configSet.getConfigFile());
        Objects.requireNonNull(configSet.getWebappPath());
        AmetysHomeLockWrapper lockWrapper = new AmetysHomeLockWrapper();
        try
        {
            _cocoon = CocoonHelper.startApplication(configSet, _logger, lockWrapper, useIncludePolicyWithFeatures);
        }
        finally
        {
            _ametysHomeLock = lockWrapper.getLock().orElse(null);
        }
    }
    
    /**
     * Gets the Coocon Wrapper instance (if exists)
     * <br><b>Use with care, if you know what you do</b>
     * @return the Coocon Wrapper instance
     */
    protected final CocoonWrapper getCocoonWrapper()
    {
        return _cocoon;
    }
    
    /**
     * Gets the webapp path
     * <br><b>Use with care, if you know what you do</b>
     * @return the webapp path
     */
    protected final String getWebappPath()
    {
        return _webappPath;
    }
    
    @Override
    public final void afterClass() throws Exception
    {
        if (_optimized)
        {
            essentialsAfter();
        }
    }
    
    /**
     * Executed after each test
     * @throws Exception If an error occurs
     */
    @After
    public final void afterTest() throws Exception
    {
        if (!_optimized)
        {
            essentialsAfter();
        }
    }
    
    /**
     * Essential processings (such as stopping Cocoon...)
     * done after each test (if class is annotated {@link Unoptimized})
     * or done once after all tests (if class is not annotated {@link Unoptimized}).
     * @throws Exception If an error occurs
     */
    protected void essentialsAfter() throws Exception
    {
        if (_dbProperties != null && _cocoon != null)
        {
            _dbProperties.retrieveDbsToShutdown(_cocoon.getServiceManager());
            _stopCocoon();
            _dbProperties.executeStopScripts(this, _resolvedConfigSet);
        }
        _dbProperties = null;
        _deleteTmpFiles();
        _deleteDirectories();
    }
    
    private void _stopCocoon()
    {
        if (this instanceof CocoonTestCase)
        {
            long start = System.currentTimeMillis();
            leaveEnvironment();
            stopCocoon();
            _logger.info("End of Cocoon ended in {}ms", System.currentTimeMillis() - start);
        }
    }
    
    /**
     * <b>Use with care, if you know what you do</b>
     * <br>Stops Cocoon manually. You often do not need to call this method, and implements {@link CocoonTestCase} instead.
     */
    protected void stopCocoon()
    {
        CocoonHelper.stopApplication(_cocoon, _ametysHomeLock, _logger);
        _cocoon = null;
        _ametysHomeLock = null;
    }
    
    private static void _deleteTmpFiles()
    {
        if (_resolvedConfigSet != null)
        {
            _resolvedConfigSet.removeTmpFiles(_staticLogger);
            _resolvedConfigSet = null;
        }
    }
    
    /**
     * Creates and enters a cocoon environment specific to the engine, mount at the given contextPath.
     * <br>Use only if TestCase implements {@link CocoonTestCase}
     * @param contextPath the underlying app contextPath.
     * @return a Map with the environment information.
     */
    protected Map<String, Object> enterEnvironment(String contextPath)
    {
        leaveEnvironment();
        
        if (_cocoon != null)
        {
            _environment = _cocoon.enterEnvironment(contextPath);
            return _environment;
        }
        
        return null;
    }
    
    /**
     * Leaves the cocoon environment.
     * <br>Use only if TestCase implements {@link CocoonTestCase}
     */
    protected void leaveEnvironment()
    {
        if (_cocoon != null && _environment != null)
        {
            try
            {
                _cocoon.leaveEnvironment(_environment);
            }
            catch (Exception e)
            {
                // silently ignore
            }
            _environment = null;
        }
    }
    
    /**
     * Creates and mock a {@link HttpServer}.
     * <br>The {@link HttpServer#start()} method is already called. You <b>MUST</b> call {@link HttpServer#stop(int)} when your test has finished to run. 
     * @param port The port number
     * @return the mocked {@link HttpServer}
     * @throws IOException if an I/O exception occured
     */
    protected static final HttpServer mockServer(int port) throws IOException
    {
        return ServerMocker.mockServer(port, _cocoon);
    }
    
    /**
     * Returns true if it is a {@link CocoonTestCase} 
     * and the Ametys runtime {@link PluginsManager} started with {@link org.ametys.runtime.plugin.PluginsManager.Status#OK OK} status
     * @return <code>true</code> if the PluginsManager started with OK status
     */
    protected static final boolean isPluginsManagerOk()
    {
        boolean isOkStatus = Optional.ofNullable(PluginsManager.getInstance())
                .map(PluginsManager::getStatus)
                .map(PluginsManager.Status.OK::equals)
                .orElse(false);
        return isOkStatus;
    }
    
    /**
     * Get the request from the environment.
     * @return The request.
     */
    protected final CommandLineRequest getRequest()
    {
        Objects.requireNonNull(_environment, "Error. The environment is not available to get the request.");
        Environment env = (Environment) _environment.get("environment");
        return (CommandLineRequest) ObjectModelHelper.getRequest(env.getObjectModel());
    }
    
    /**
     * Get the response from the environment.
     * @return The response.
     */
    protected final Response getResponse()
    {
        Objects.requireNonNull(_environment, "Error. The environment is not available to get the response.");
        Environment env = (Environment) _environment.get("environment");
        return ObjectModelHelper.getResponse(env.getObjectModel());
    }
}
