/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.net.URL;
import java.util.Collection;

import javax.xml.XMLConstants;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.xml.sax.XMLReader;

import org.ametys.core.datasource.LDAPDataSourceManager;
import org.ametys.core.datasource.SQLDataSourceManager;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.data.AmetysHomeLock;
import org.ametys.runtime.data.AmetysHomeLockException;
import org.ametys.runtime.plugin.IncludePolicyFeatureActivator.IncludedFeature;
import org.ametys.runtime.servlet.RuntimeConfig;

/**
 * Utility methods relative to {@link CocoonTestCase}s.
 */
public final class CocoonHelper
{
    /** The LDAP datasource file for testing real LDAP server */
    public static final String DEFAULT_LDAP_DATASOURCE_FILE = "test/environments/datasources/datasource-ldap.xml";
    
    private static final String __AMETYS_HOME_DIR = "/WEB-INF/data";
    
    private CocoonHelper()
    {
        // nothing
    }
    
    /**
     * Starts the application with the provided config files in {@link TestConfigSet} object.
     * @param configSet The configset to use
     * @param logger The logger
     * @param ametysHomeLockWrapper The wrapper of {@link AmetysHomeLock}
     * @param useIncludePolicyWithFeatures If not null, Cocoon will initialize with only the given features and their dependencies.
     * @return the CocoonWrapper used to process URIs
     * @throws Exception if an error occurred
     */
    public static CocoonWrapper startApplication(TestConfigSet configSet, Logger logger, AmetysHomeLockWrapper ametysHomeLockWrapper, Collection<IncludedFeature> useIncludePolicyWithFeatures) throws Exception
    {
        CocoonWrapper cocoon = null;
        try
        {
            configureRuntime(configSet, logger, ametysHomeLockWrapper);
            
            String configFile = configSet.getConfigFile();
            Config.setFilename(configFile);
            
            SQLDataSourceManager.setFilename(configSet.getSqlFile());
            LDAPDataSourceManager.setFilename(configSet.getLdapFile());
            
            cocoon = _startCocoon(configSet.getWebappPath(), useIncludePolicyWithFeatures);
            return cocoon;
        }
        catch (Exception e)
        {
            stopApplication(cocoon, ametysHomeLockWrapper.getLock().orElse(null), logger);
            throw e;
        }
    }
    
    /**
     * Stops the application and releases associated resources.
     * @param cocoon The cocoon wrapper
     * @param ametysHomeLock  The Ametys Home lock
     * @param logger The logger
     */
    public static void stopApplication(CocoonWrapper cocoon, AmetysHomeLock ametysHomeLock, Logger logger)
    {
        if (cocoon != null)
        {
            cocoon.dispose();
        }
        
        if (ametysHomeLock != null)
        {
            logger.info("Releasing lock on Ametys home");
            
            ametysHomeLock.release();
        }
    }
    
    /**
     * Configures the RuntimeConfig with the given file
     * @param configSet The configset to use
     * @param logger The logger
     * @param ametysHomeLockWrapper The wrapper of {@link AmetysHomeLock} 
     * @throws Exception if the Runtime config cannot be loaded
     */
    public static void configureRuntime(TestConfigSet configSet, Logger logger, AmetysHomeLockWrapper ametysHomeLockWrapper) throws Exception
    {
        String webappPath = configSet.getWebappPath();
        AmetysRuntime runtime = configSet.getRuntime();
        
        File ametysHome = _initAmetysHome(webappPath, logger, ametysHomeLockWrapper);
        
        Configuration runtimeConf = null;
        try
        {
            // XML Schema validation
            SAXParserFactory factory = SAXParserFactory.newInstance();
            factory.setNamespaceAware(true);
            
            URL schemaURL = CocoonHelper.class.getResource("/org/ametys/runtime/servlet/runtime-4.1.xsd");
            Schema schema = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI).newSchema(schemaURL);
            factory.setSchema(schema);
            
            XMLReader reader = factory.newSAXParser().getXMLReader();
            DefaultConfigurationBuilder runtimeConfBuilder = new DefaultConfigurationBuilder(reader);
            
            try (InputStream runtimeIs = runtime.getInputStream())
            {
                runtimeConf = runtimeConfBuilder.build(runtimeIs, runtime.getSystemId());
            }
        }
        catch (Exception e)
        {
            logger.error("Unable to load runtime at '" + runtime.getSystemId() + "'. PluginsManager will enter in safe mode.", e);
        }
        
        Configuration externalConf = null;
        DefaultConfigurationBuilder externalConfBuilder = new DefaultConfigurationBuilder();

        File externalConfigFile = new File("test/environments/external-locations.xml");
        if (externalConfigFile.exists())
        {
            try (InputStream external = new FileInputStream(externalConfigFile))
            {
                externalConf = externalConfBuilder.build(external, externalConfigFile.getAbsolutePath());
            }
        }
        else
        {
            externalConf = _generateExternalLocationFile(logger);
        }
        
        RuntimeConfig.configure(runtimeConf, externalConf, ametysHome, webappPath);
    }
    
    private static File _initAmetysHome(String contextPath, Logger logger, AmetysHomeLockWrapper ametysHomeLockWrapper) throws AmetysHomeLockException
    {
        String ametysHomePath = contextPath + __AMETYS_HOME_DIR;
        
        logger.info("Acquiring lock on {}", ametysHomePath);
        
        // Acquire the lock on Ametys home
        File ametysHome = new File(ametysHomePath);
        ametysHome.mkdirs();
        
        AmetysHomeLock ametysHomeLock = new AmetysHomeLock(ametysHome);
        ametysHomeLockWrapper.setLock(ametysHomeLock);
        ametysHomeLock.acquire();
        
        return ametysHome;
    }
    
    private static Configuration _generateExternalLocationFile(Logger logger)
    {
        DefaultConfiguration externalConfiguration = new DefaultConfiguration("external");
        
        DefaultConfiguration pluginsConfiguration = new DefaultConfiguration("plugins");
        DefaultConfiguration workspacesConfiguration = new DefaultConfiguration("workspaces");
        
        File main = new File("main");
        if (main.exists())
        {
            for (File file : main.listFiles())
            {
                if (file.isDirectory())
                {
                    if (file.getName().startsWith("plugin-"))
                    {
                        DefaultConfiguration pluginConfiguration = new DefaultConfiguration("plugin");
                        pluginConfiguration.setAttribute("name", StringUtils.substringAfter(file.getName(), "plugin-"));
                        pluginConfiguration.setValue(file.getAbsolutePath());
                        pluginsConfiguration.addChild(pluginConfiguration);
                    }
                    else if (file.getName().startsWith("workspace-") && new File(file, "workspace.xml").exists())
                    {
                        DefaultConfiguration workspaceConfiguration = new DefaultConfiguration("workspace");
                        workspaceConfiguration.setAttribute("name", StringUtils.substringAfter(file.getName(), "workspace-"));
                        workspaceConfiguration.setValue(file.getAbsolutePath());
                        workspacesConfiguration.addChild(workspaceConfiguration);
                    }
                    else if (file.getName().equals("kernel"))
                    {
                        DefaultConfiguration kernelConfiguration = new DefaultConfiguration("kernel");
                        kernelConfiguration.setValue(file.getAbsolutePath());
                        externalConfiguration.addChild(kernelConfiguration);
                    }
                    else
                    {
                        logger.warn("File " + file + " is not added to the generated external-locations.xml");
                    }
                }
            }
        }
        
        externalConfiguration.addChild(pluginsConfiguration);
        externalConfiguration.addChild(workspacesConfiguration);
        
        return externalConfiguration;
    }
    
    static CocoonWrapper _startCocoon(String applicationPath, Collection<IncludedFeature> useIncludePolicyWithFeatures) throws Exception
    {
        // Set this property in order to avoid a System.err.println (CatalogManager.java)
        if (System.getProperty("xml.catalog.ignoreMissing") == null)
        {
            System.setProperty("xml.catalog.ignoreMissing", "true");
        }
        
        CocoonWrapper cocoon = new CocoonWrapper(applicationPath, "tmp/work");
        cocoon.initialize(useIncludePolicyWithFeatures);
        return cocoon;
    }
}
