/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import static java.lang.annotation.ElementType.METHOD;
import static java.lang.annotation.ElementType.TYPE_USE;
import static java.lang.annotation.RetentionPolicy.RUNTIME;

import java.lang.annotation.Inherited;
import java.lang.annotation.Retention;
import java.lang.annotation.Target;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.environment.Environment;
import org.apache.commons.lang3.RandomStringUtils;

import org.ametys.runtime.plugin.ExcludePolicyFeatureActivator;
import org.ametys.runtime.plugin.IncludePolicyFeatureActivator.IncludedFeature;
import org.ametys.runtime.plugin.PluginsManager;

/**
 * Implements this interface if your Test Case needs to start a Cocoon application before the Test Case (by default) or before each test (by annotating with {@link Unoptimized}).
 * <br>Lookup the components you need with {@link #lookupComponents(ServiceManager)} and use them in {@link #prepareComponents()}.
 * <br>Indicates the features the test needs by annotating the class or/and the method with {@link NeedFeatures}.
 */
public interface CocoonTestCase extends ScriptTestCase
{
    /**
     * Overrides this method to configure the files the Cocoon application will start with.
     * <br>By default,<ul>
     * <li>The {@link AmetysRuntime runtime} used is the default one ({@link TestConfigSet#DEFAULT_RUNTIME})
     * <li>The configuration file is <code>test/environments/configs/config.xml</code></li>
     * <li>The SQL data source used is the one returned by {@link #getDataSourceFile()}</li>
     * <li>No LDAP data source is used</li>
     * <li>The webapp path is <code>test/environments/webapp</code></li>
     * <li>The context path is an empty string</li>
     * </ul>
     * The provided files can contain some variable with the following syntax:
     * <pre>
     * &lt;foo&gt;http://server.name:${serverPort}/index&lt;/foo&gt;
     * </pre>
     * <br>See also {@link #variableValues()} to give a value to those variables.
     * @return The configset to run the test with
     */
    default TestConfigSet getConfigSet()
    {
        return new TestConfigSet()
                .withRuntime(TestConfigSet.DEFAULT_RUNTIME)
                .withConfigFile("test/environments/configs/config.xml")
                .withSqlFile(getDataSourceFile())
                .withLdapFile(null)
                .withWebappPath("test/environments/webapp")
                .withContextPath("");
    }
    
    /**
     * If the provided files by {@link #getConfigSet()} contain
     * some variables, you must assign them a value by overriding this method.
     * <br>The syntax is the following:
     * <pre>${variableName}</pre>
     * <br>See also {@link #getConfigSet()}.
     * @return The variable values
     */
    default Map<String, String> variableValues()
    {
        String randomString = RandomStringUtils.randomAlphanumeric(20);
        return Map.of(
                "tmpFolder", String.format("tmp/tmp-%s", randomString)
        );
    }
    
    /**
     * If you do not need an initial {@link Environment}, override it and set to <code>false</code>.
     * <br>Default to <code>true</code>
     * @return <code>true</code> if initial environment must be created, <code>false</code> otherwise.
     */
    default boolean withInitialEnvironment()
    {
        return true;
    }
    
    /**
     * Lookup components needed for the Test Case.
     * <br>You <b>must</b> store your components into <b>static</b> fields as the test can be optimized and this method be called once before all tests.
     * <br>This method is called before each test (if class is annotated {@link Unoptimized})
     * or once before any test (if class is not annotated {@link Unoptimized}).
     * @param manager The service manager
     * @throws ServiceException if an error occurs
     */
    void lookupComponents(ServiceManager manager) throws ServiceException;
    
    /**
     * Prepares components just after the start of Cocoon (if it started with OK status).
     * <br>This method is called before each test (if class is annotated {@link Unoptimized})
     * or once before any test (if class is not annotated {@link Unoptimized}).
     * @throws Exception if an error occurs
     */
    default void prepareComponents() throws Exception
    {
        // Nothing by default
    }
    
    /**
     * Annotate your class or/and your method with this annotation so as to tell the feature names you need to run your test(s).
     */
    @Retention(RUNTIME)
    @Target({TYPE_USE, METHOD})
    public static @interface NeedFeatures
    {
        /**
         * The feature names you need to run your test(s).
         * @return The feature names you need to run your test(s).
         */
        String[] value();
    }
    
    /**
     * Helper class to transform a {@link NeedFeatures} annotation to a list of {@link IncludedFeature}s
     */
    public static final class NeedFeaturesHelper
    {
        /**
         * Transforms a {@link NeedFeatures} annotation to a list of {@link IncludedFeature}s
         * @param annotation The annotation
         * @param cause The inclusion cause
         * @return The features to include
         */
        public static List<IncludedFeature> toIncludedFeatures(NeedFeatures annotation, String cause)
        {
            return Stream.of(annotation.value())
                    .flatMap(Stream::of)
                    .map(f -> IncludedFeature.of(f, cause))
                    .collect(Collectors.toList());
        }
        
    }
    
    /**
     * Annotate your class with this annotation so as to run your test(s) with all features (i.e. {@link PluginsManager} will load features with {@link ExcludePolicyFeatureActivator}).
     * <br><b>Use with caution</b>, as it is a bad practice to run a test with all features, and not use {@link NeedFeatures} annotation.
     */
    @Retention(RUNTIME)
    @Target(TYPE_USE)
    @Inherited
    public static @interface AllFeatures
    {
        // Nothing
    }
}
