/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import java.io.File;
import java.io.OutputStream;
import java.lang.reflect.Field;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.apache.avalon.excalibur.logger.LogKitLoggerManager;
import org.apache.avalon.framework.CascadingRuntimeException;
import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.container.ContainerUtil;
import org.apache.avalon.framework.context.Context;
import org.apache.avalon.framework.context.DefaultContext;
import org.apache.avalon.framework.logger.LogKitLogger;
import org.apache.avalon.framework.logger.Logger;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.WrapperServiceManager;
import org.apache.cocoon.Cocoon;
import org.apache.cocoon.Constants;
import org.apache.cocoon.components.CocoonComponentManager;
import org.apache.cocoon.components.pipeline.ProcessingPipeline;
import org.apache.cocoon.environment.Environment;
import org.apache.cocoon.environment.commandline.AbstractCommandLineEnvironment;
import org.apache.cocoon.environment.commandline.CommandLineContext;
import org.apache.cocoon.environment.commandline.CommandLineSession;
import org.apache.cocoon.util.NullOutputStream;
import org.apache.cocoon.xml.ContentHandlerWrapper;
import org.apache.cocoon.xml.XMLConsumer;
import org.apache.commons.text.CharacterPredicates;
import org.apache.commons.text.RandomStringGenerator;
import org.apache.log.Hierarchy;
import org.apache.log.Priority;
import org.xml.sax.ContentHandler;

import org.ametys.core.engine.BackgroundEnvironment;
import org.ametys.core.observation.ObservationManager;
import org.ametys.runtime.plugin.IncludePolicyFeatureActivator.IncludedFeature;
import org.ametys.runtime.plugin.PluginsManager;
import org.ametys.runtime.plugin.component.PluginsComponentManager;
import org.ametys.runtime.servlet.RuntimeServlet;

/**
 * The Cocoon Wrapper simplifies usage of the Cocoon object. Allows to create, configure Cocoon instance and process single requests.
 * 
 * @author <a href="mailto:stefano@apache.org">Stefano Mazzocchi</a>
 * @author <a href="mailto:nicolaken@apache.org">Nicola Ken Barozzi</a>
 * @author <a href="mailto:vgritsenko@apache.org">Vadim Gritsenko</a>
 * @author <a href="mailto:uv@upaya.co.uk">Upayavira</a>
 */
public class CocoonWrapper
{
    private static RandomStringGenerator _randomStringGenerator = new RandomStringGenerator.Builder()
            .withinRange('0', 'z')
            .filteredBy(CharacterPredicates.LETTERS, CharacterPredicates.DIGITS)
            .build();

    // User Supplied Parameters
    private String _contextDir;

    private String _workDir;

    private String _logLevel = "ERROR";

    // Internal Objects
    private CommandLineContext _cliContext;

    private Cocoon _cocoon;
    
    private ServiceManager _serviceManager;
    
    private Context _context;

    private Logger _logger;

    /**
     * Creates a new Cocoon wrapper
     * @param contextDir The context directory
     * @param workDir The work directory
     */
    public CocoonWrapper(String contextDir, String workDir)
    {
        _contextDir = contextDir;
        _workDir = workDir;
    }

    /**
     * Initializes Cocoon
     * @param useIncludePolicyWithFeatures If not null, Cocoon will initialize with only the given features and their dependencies.
     * @throws Exception if an error occurs
     */
    void initialize(Collection<IncludedFeature> useIncludePolicyWithFeatures) throws Exception
    {
        Hierarchy hierarchy = Hierarchy.getDefaultHierarchy();

        Priority priority = Priority.getPriorityForName(_logLevel);
        hierarchy.setDefaultPriority(priority);

        // Install a temporary logger so that getDir() can log if needed
        _logger = new LogKitLogger(hierarchy.getLoggerFor(""));

        try
        {
            DefaultContext appContext = new DefaultContext();
            appContext.put(Constants.CONTEXT_WORK_DIR, new File(_workDir));

            LogKitLoggerManager logManager = new LogKitLoggerManager(hierarchy);
            logManager.enableLogging(_logger);

            URL configFile = getClass().getResource("/org/ametys/runtime/cocoon/cocoon.xconf");

            _cliContext = new CommandLineContext(_contextDir);
            _cliContext.enableLogging(_logger);
            
            File cacheDir = new File(_workDir, "cache-dir");
            cacheDir.mkdirs();

            appContext.put(Constants.CONTEXT_ENVIRONMENT_CONTEXT, _cliContext);
            appContext.put(Constants.CONTEXT_CLASS_LOADER, CocoonWrapper.class.getClassLoader());
            appContext.put(Constants.CONTEXT_CLASSPATH, "unknown");
            appContext.put(Constants.CONTEXT_UPLOAD_DIR, _contextDir + "upload-dir");
            appContext.put(Constants.CONTEXT_CACHE_DIR, cacheDir);
            appContext.put(Constants.CONTEXT_CONFIG_URL, configFile);
            appContext.put(Constants.CONTEXT_DEFAULT_ENCODING, "ISO-8859-1");
            if (useIncludePolicyWithFeatures != null)
            {
                appContext.put(PluginsManager.INCLUDED_FEATURES_CONTEXT_ATTRIBUTE, useIncludePolicyWithFeatures);
            }

            _cocoon = new Cocoon();
            ContainerUtil.enableLogging(_cocoon, _logger);
            ContainerUtil.contextualize(_cocoon, appContext);
            _cocoon.setLoggerManager(logManager);
            ContainerUtil.initialize(_cocoon);
            
            PluginsComponentManager pluginCM = (PluginsComponentManager) _cliContext.getAttribute("PluginsComponentManager");
            
            if (!PluginsManager.getInstance().isSafeMode())
            {
                if (RuntimeServlet.doMigrationAndInit(pluginCM))
                {
                    throw new UnsupportedOperationException("Tests does not support migrations that requires a restart");
                }
                
                // Expose service manager
                _serviceManager = new WrapperServiceManager(pluginCM);
                
                // Expose context
                _exposeContext();
            }
            
            CommandLineSession.invalidateSession();
        }
        catch (Exception e)
        {
            _logger.fatalError("Exception caught initializing Cocoon", e);
            throw e;
        }
    }
    
    private void _exposeContext()
    {
        try
        {
            // Ok using reflection is bad, but Cocoon class should have made it at least protected to make my life easier !
            Field contextField = _cocoon.getClass().getDeclaredField("context");
            contextField.setAccessible(true);
            _context = (Context) contextField.get(_cocoon);
        }
        catch (NoSuchFieldException | SecurityException | IllegalArgumentException | IllegalAccessException e)
        {
            _logger.error("Error trying to expose Context", e);
        }
    }
    
    /**
     * Returns the service manager
     * <br><b>Use with care, if you know what you do</b>
     * @return the service manager
     */
    public ServiceManager getServiceManager()
    {
        return _serviceManager;
    }
    
    /**
     * Returns the context
     * <br><b>Use with care, if you know what you do</b>
     * @return the context
     */
    public Context getContext()
    {
        return _context;
    }

    /**
     * Process single URI into given content handler, skipping final serializer
     * @param uri to process
     * @param handler to write generated contents into
     * @param requestParameters the request parameters
     * @param requestAttributes the request attributes
     * @param requestHeaders the request headers
     * @throws Exception if an error occurs
     */
    public void processURI(String uri, ContentHandler handler, Map<String, String> requestParameters, Map<String, Object> requestAttributes, Map<String, String> requestHeaders) throws Exception
    {
        _logger.info("Processing URI: " + uri);

        // Get parameters, headers, deparameterized URI and path from URI
        Map<String, String> headers = requestHeaders;
        if (headers == null)
        {
            headers = new HashMap<>();
        }
        
        headers.put("user-agent", "Ametys Runtime");
        headers.put("accept", "text/html, */*");

        Environment env = new BackgroundEnvironment(uri, requestParameters, requestAttributes, headers, "", _cliContext, _logger, null);

        XMLConsumer consumer = new ContentHandlerWrapper(handler);
        CocoonComponentManager.enterEnvironment(env, _cocoon.getComponentManager(), _cocoon);
        ProcessingPipeline pipeline = _cocoon.buildPipeline(env);
        try
        {
            pipeline.prepareInternal(env);
            pipeline.process(env, consumer);
        }
        finally
        {
            CocoonComponentManager.leaveEnvironment();
        }
    }
    
    /**
     * Process single URI into given output stream
     * @param uri to process
     * @param outputStream to write into
     * @param requestParameters the request parameters
     * @param requestAttributes the request attributes
     * @param requestHeaders the request headers
     * @throws Exception if an error occurs
     */
    public void processURI(String uri, OutputStream outputStream, Map<String, String> requestParameters, Map<String, Object> requestAttributes, Map<String, String> requestHeaders) throws Exception
    {
        _logger.info("Processing URI: " + uri);
        
        // Get parameters, headers, deparameterized URI and path from URI
        Map<String, String> headers = requestHeaders;
        if (headers == null)
        {
            headers = new HashMap<>();
        }
        
        headers.put("user-agent", "Ametys Runtime");
        headers.put("accept", "text/html, */*");
        
        Environment env = new BackgroundEnvironment(uri, requestParameters, requestAttributes, headers, "", _cliContext, _logger, outputStream);
        
        CocoonComponentManager.enterEnvironment(env, _cocoon.getComponentManager(), _cocoon);
        ProcessingPipeline pipeline = _cocoon.buildPipeline(env);
        try
        {
            pipeline.process(env);
        }
        finally
        {
            CocoonComponentManager.leaveEnvironment();
        }
    }
    
    /**
     * Create and enter a cocoon environment specific to the engine.
     * @param contextPath the underlying app contextPath.
     * @return a Map with the environment information.
     */
    public Map<String, Object> enterEnvironment(String contextPath)
    {
        try
        {
            Environment env = new BackgroundEnvironment("", new HashMap<>(), new HashMap<>(), Collections.EMPTY_MAP, contextPath, _cliContext, _logger, new NullOutputStream());
            
            // Random request ID: used by AbstractCachingPipeline to implement pipeline locking.
            String requestId = _randomStringGenerator.generate(8);
            env.getObjectModel().put(AbstractCommandLineEnvironment.CLI_REQUEST_ID, requestId);
            
            Object processingKey = CocoonComponentManager.startProcessing(env);
            int environmentDepth = CocoonComponentManager.markEnvironment();
            
            CocoonComponentManager.enterEnvironment(env, _cocoon.getComponentManager(), _cocoon);
            
            Map<String, Object> result = new HashMap<>();
            
            result.put("environment", env);
            result.put("processingKey", processingKey);
            result.put("environmentDepth", Integer.valueOf(environmentDepth));
            
            return result;
        }
        catch (MalformedURLException e)
        {
            throw new RuntimeException(e);
        }
    }
    
    /**
     * Leave the cocoon environment.
     * @param environmentInformation the environment information.
     */
    public void leaveEnvironment(Map<String, Object> environmentInformation)
    {
        _waitAsyncObservers();
        
        Environment environment = (Environment) environmentInformation.get("environment");
        Object processingKey = environmentInformation.get("processingKey");
        int environmentDepth = ((Integer) environmentInformation.get("environmentDepth")).intValue();

        CocoonComponentManager.leaveEnvironment();
        CocoonComponentManager.endProcessing(environment, processingKey);
        
        try
        {
            CocoonComponentManager.checkEnvironment(environmentDepth, _logger);
        }
        catch (Exception e)
        {
            throw new CascadingRuntimeException("Error checking the environment", e);
        }
    }
    
    private void _waitAsyncObservers()
    {
        // Wait for Ametys async Observers to finish in order to avoid some issues when Cocoon is disposed but some Observers still work
        PluginsComponentManager pluginCM = (PluginsComponentManager) _cliContext.getAttribute("PluginsComponentManager");
        if (pluginCM != null && !PluginsManager.getInstance().isSafeMode())
        {
            try
            {
                ObservationManager observationManager = (ObservationManager) pluginCM.lookup(ObservationManager.ROLE);
                List<Future> futures = observationManager.getFuturesForRequest();
                futures.forEach(f ->
                {
                    try
                    {
                        f.get();
                    }
                    catch (InterruptedException | ExecutionException e)
                    {
                        _logger.error("An exception occured when calling Future#get()", e);
                    }
                });
            }
            catch (ComponentException e)
            {
                _logger.error("An exception occured when getting the ObservationManager component", e);
            }
        }
        
    }
    
    /**
     * Dispose current cocoon instance.
     */
    void dispose()
    {
        // Dispose the plugins component manager.
        PluginsComponentManager pluginCM = (PluginsComponentManager) _cliContext.getAttribute("PluginsComponentManager");
        if (pluginCM != null)
        {
            pluginCM.dispose();
        }
        
        // Dispose the cocoon instance.
        _cocoon.dispose();
    }
}
