/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Properties;
import java.util.UUID;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.text.RandomStringGenerator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.core.datasource.AbstractDataSourceManager.DataSourceDefinition;
import org.ametys.core.datasource.SQLDataSourceManager;
import org.ametys.runtime.test.framework.ScriptHelper.ShutdownStatus;

class DatabaseProperties
{
    private static Properties __DB_PROPERTIES;
    private static final Path __DB_PROPERTIES_LOCATION = Paths.get(
            System.getProperty("user.home"), 
            ".ametys", 
            "ametys_db.properties");
    
    private static final Logger __LOGGER = LoggerFactory.getLogger(AbstractBaseTestCase.class);
    
    private static RandomStringGenerator __ORACLE_USER_GENERATOR;
    static
    {
        if (__ORACLE_USER_GENERATOR == null)
        {
            __ORACLE_USER_GENERATOR = new RandomStringGenerator.Builder()
                    .withinRange(new char[] {'0', '9'}, new char[] {'a', 'z'})
                    .build();
        }
    }
    
    private Optional<String> _name = Optional.empty();
    private String _type;
    private String _urlForCreateAndDrop;
    private String _baseUrl;
    private String _user;
    private String _password;
    private List<DataSourceDefinition> _datasourcesToShutdown = new ArrayList<>();
    
    private static void _loadDbProperties()
    {
        if (__DB_PROPERTIES != null)
        {
            return;
        }
        
        __DB_PROPERTIES = new Properties();
        try
        {
            __DB_PROPERTIES.load(Files.newBufferedReader(__DB_PROPERTIES_LOCATION));
        }
        catch (IOException e)
        {
            throw new IllegalStateException("DB properties cannot be loaded. Maybe the .properties file is not present.", e);
        }
    }
    
    void addVariables(Map<String, String> variableValues)
    {
        if (_name.isPresent())
        {
            variableValues.put("dbName", _name.get());
            variableValues.put("dbUrl", _baseUrl);
            variableValues.put("dbUser", _user);
            variableValues.put("dbPassword", _password);
        }
    }
    
    void executeStartScripts(AbstractRuntimeTestCase test, Map<String, String> variableValues) throws Exception
    {
        if (test instanceof ScriptTestCase)
        {
            long start = System.currentTimeMillis();
            
            ScriptTestCase scriptTestInstance = (ScriptTestCase) test;
            String dbType = scriptTestInstance.getDBType();
            
            // Create unique datasource if needed
            _initFieldsAndCreateDatabase(dbType);
            
            // Add variables
            addVariables(variableValues);
            
            // Execute start scripts
            TestConfigSet configSet = new TestConfigSet()
                    .withSqlFile(scriptTestInstance.getDataSourceFile())
                    .resolve(variableValues);
            try
            {
                String dataSourceFilePath = configSet.getSqlFile();
                ScriptHelper.executeStartScripts(
                        Arrays.asList(scriptTestInstance.getStartScriptFiles()), 
                        Arrays.asList(scriptTestInstance.getStartScripts()), 
                        dataSourceFilePath, 
                        dbType);
            }
            finally
            {
                configSet.removeTmpFiles(__LOGGER);
            }
            
            __LOGGER.info("Execution of start scripts ended in {}ms", System.currentTimeMillis() - start);
        }
    }
    
    private void _initFieldsAndCreateDatabase(String dbType) throws Exception
    {
        switch (dbType)
        {
            case "derby":
            case "hsqldb":
                // local database, do nothing
                return;
            default:
                // continue, create a unique db
        }
        
        _loadDbProperties();
        
        String name;
        String roleForOracle = null;
        switch (dbType)
        {
            case "oracle":
                // Cannot be longer than 30 characters
                int length = 30;
                name = "T_";
                name += __ORACLE_USER_GENERATOR.generate(length - name.length());
                roleForOracle = __DB_PROPERTIES.getProperty("oracle.role");
                _baseUrl = __DB_PROPERTIES.getProperty("oracle.url");
                _urlForCreateAndDrop = _baseUrl;
                _user = __DB_PROPERTIES.getProperty("oracle.user");
                _password = __DB_PROPERTIES.getProperty("oracle.password");
                break;
            case "mysql":
                name = "Ametys_Test_" + UUID.randomUUID().toString().replaceAll("\\-", "_");
                _baseUrl = __DB_PROPERTIES.getProperty("mysql.url");
                _urlForCreateAndDrop = _baseUrl;
                _user = __DB_PROPERTIES.getProperty("mysql.user");
                _password = __DB_PROPERTIES.getProperty("mysql.password");
                break;
            case "postgresql":
                name = "ametys_test_" + UUID.randomUUID().toString().replaceAll("\\-", "_");
                _baseUrl = __DB_PROPERTIES.getProperty("postgresql.url");
                _urlForCreateAndDrop = _baseUrl + "/postgres";
                _user = __DB_PROPERTIES.getProperty("postgresql.user");
                _password = __DB_PROPERTIES.getProperty("postgresql.password");
                break;
            default:
                throw new IllegalArgumentException("Unsupported dbType '" + dbType + "'");
        }
        _name = Optional.of(name);
        _type = dbType;
        
        ScriptHelper.createDatabase(_urlForCreateAndDrop, _user, _password, name, roleForOracle, dbType);
    }
    
    void retrieveDbsToShutdown(ServiceManager serviceManager) throws Exception
    {
        if (serviceManager != null && serviceManager.hasService(SQLDataSourceManager.ROLE))
        {
            SQLDataSourceManager sqlDataSourceManager = (SQLDataSourceManager) serviceManager.lookup(SQLDataSourceManager.ROLE);
            _datasourcesToShutdown = sqlDataSourceManager.getDataSourceDefinitions(true, true, true)
                    .values()
                    .stream()
                    .distinct()
                    .filter(def ->
                    {
                        String type = (String) def.getParameters().get(SQLDataSourceManager.PARAM_DATABASE_TYPE);
                        return "hsqldb".equals(type) || "derby".equals(type);
                    })
                    .collect(Collectors.toList());
        }
    }
    
    void executeStopScripts(AbstractRuntimeTestCase test, TestConfigSet resolvedConfigSet) throws Exception
    {
        if (test instanceof ScriptTestCase)
        {
            if (_name.isPresent())
            {
                ScriptHelper.dropDatabase(_urlForCreateAndDrop, _user, _password, _name.get(), _type);
            }
            
            for (DataSourceDefinition datasource : _datasourcesToShutdown)
            {
                String sqlFile = resolvedConfigSet.getSqlFile();
                String datasourceId =  datasource.getId();
                String dbType = (String) datasource.getParameters().get(SQLDataSourceManager.PARAM_DATABASE_TYPE);
                try
                {
                    ShutdownStatus status = ScriptHelper.shutdownDatabase(sqlFile, datasourceId, dbType);
                    __LOGGER.info("{} was shutdown with status {}.", datasourceId, status);
                }
                catch (Exception e)
                {
                    __LOGGER.error("Cannot shutdown {} database. It can lead to unstable test running.", datasourceId, e);
                }
            }
            _datasourcesToShutdown.clear();
        }
    }
}
