/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.commons.collections4.ListUtils;
import org.junit.Ignore;
import org.junit.Test;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.runtime.plugin.IncludePolicyFeatureActivator.IncludedFeature;
import org.ametys.runtime.test.framework.CocoonTestCase.AllFeatures;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeaturesHelper;
import org.ametys.runtime.util.Annotations;

class NeededFeatures
{
    private AbstractRuntimeTestCase _test;
    private boolean _optimized;
    
    NeededFeatures(AbstractRuntimeTestCase test, boolean optimized)
    {
        _test = test;
        _optimized = optimized;
    }
    
    Collection<IncludedFeature> compute(Optional<String> dbType) throws NoSuchMethodException
    {
        if (_test.getClass().isAnnotationPresent(AllFeatures.class))
        {
            // Force ExcludePolicyFeatureActivator
            _test._logger.info("'CocoonTestCase.AllFeatures' annotation is present, ExcludePolicyFeatureActivator mode is thus forced.");
            return null;
        }
        
        List<IncludedFeature> result = ListUtils.union(
                _featuresForMethods(),
                _featuresForClassAndSuper()
        );
        
        // Assert there is at least one feature
        if (result.isEmpty())
        {
            _test._logger.warn("No needed feature was specified. At least one feature should be declared with the 'CocoonTestCase.NeedFeatures' annotation on the TestCase Class or on the Test method");
        }
        
        _featureForDbType(dbType).ifPresent(result::add);
        
        if (_test._logger.isInfoEnabled())
        {
            _test._logger.info("Cocoon will start with features: \n[\n{}\n]\n", 
                    result.stream()
                        .map(f -> "    " + f.toString())
                        .collect(Collectors.joining("\n")));
        }
        return result;
    }
    
    private List<IncludedFeature> _featuresForMethods() throws NoSuchMethodException
    {
        if (_optimized)
        {
            // The test is optimized, so even though we are running @Test testA(), 
            // it should include the features declared on @Test testB() 
            // so as to have the same test result when running all methods (testA AND testB) AND running just testA
            return _getMethods()
                    .stream()
                    .filter(method -> method.getAnnotation(Test.class) != null)
                    .filter(method -> method.getAnnotation(Ignore.class) == null)
                    .map(this::_featuresForMethod)
                    .flatMap(Collection::stream)
                    .collect(Collectors.toList());
        }
        else
        {
            // The test is unoptimized, just take the features declared by the current method @Test
            String currentTestMethodName = _test.getTestName().getMethodName();
            Method currentTestMethod = Objects.requireNonNull(_test.getClass().getMethod(currentTestMethodName));
            return _featuresForMethod(currentTestMethod);
        }
    }
    
    private List<Method> _getMethods()
    {
        List<Method> methods = new ArrayList<>();
        for (Class< ? > clazz = _test.getClass(); clazz != null; clazz = clazz.getSuperclass())
        {
            methods.addAll(Arrays.asList(clazz.getDeclaredMethods()));
        }
        return methods;
    }
    
    private List<IncludedFeature> _featuresForMethod(Method method)
    {
        return Optional.ofNullable(method.getAnnotation(NeedFeatures.class))
                .map(methodAnnotation -> NeedFeaturesHelper.toIncludedFeatures(methodAnnotation, _toCause(method)))
                .orElseGet(Collections::emptyList);
    }
    
    private String _toCause(Method method)
    {
        return _test.getClass().getSimpleName() + "::" + method.getName();
    }
    
    private List<IncludedFeature> _featuresForClassAndSuper()
    {
        return Annotations.getAnnotationAttributeValues(
            _test.getClass(), 
            NeedFeatures.class, 
            (annotation, annotated) -> NeedFeaturesHelper.toIncludedFeatures(annotation, annotated.getSimpleName())
        )
                .flatMap(List::stream)
                .collect(Collectors.toList());
                
    }
    
    private static Optional<IncludedFeature> _featureForDbType(Optional<String> optionalDbType)
    {
        if (optionalDbType.isEmpty())
        {
            return Optional.empty();
        }
        
        String featureId;
        switch (optionalDbType.get())
        {
            case ConnectionHelper.DATABASE_DERBY:
                featureId = "core-impl/runtime.core.datasources.sql.derby";
                break;
            case ConnectionHelper.DATABASE_HSQLDB:
                featureId = "core-impl/runtime.core.datasources.sql.hsqldb";
                break;
            case ConnectionHelper.DATABASE_MYSQL:
                featureId = "core-impl/runtime.core.datasources.sql.mysql";
                break;
            case ConnectionHelper.DATABASE_ORACLE:
                featureId = "core-impl/runtime.core.datasources.sql.oracle";
                break;
            case ConnectionHelper.DATABASE_POSTGRES:
                featureId = "core-impl/runtime.core.datasources.sql.postgresql";
                break;
            default:
                featureId = null;
                break;
        }
        
        return Optional.of(IncludedFeature.of(featureId, "DATABASE TYPE"));
    }
}
