/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import java.io.File;
import java.io.FileInputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.time.Duration;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.DefaultConfigurationBuilder;
import org.apache.commons.dbcp2.ConnectionFactory;
import org.apache.commons.dbcp2.DriverManagerConnectionFactory;
import org.apache.commons.dbcp2.PoolableConnection;
import org.apache.commons.dbcp2.PoolableConnectionFactory;
import org.apache.commons.dbcp2.PoolingDataSource;
import org.apache.commons.pool2.impl.GenericObjectPool;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.datasource.SQLDataSourceManager;
import org.ametys.core.script.SQLScriptHelper;

/**
 * Utility methods relative to {@link ScriptTestCase}s.
 */
public final class ScriptHelper
{
    private ScriptHelper()
    {
        // nothing
    }
    
    /**
     * Gets the datasource file associated to the given database type
     * @param dbType the database type
     * @return the datasource file
     */
    public static String getDataSourceFile(String dbType)
    {
        return "test/environments/datasources/datasource-" + dbType + ".xml";
    }
    
    /**
     * Gets the start script files
     * @param dbType the database type
     * @return the start script files (empty array for the current implementation)
     */
    static File[] getStartScriptFiles(String dbType)
    {
        return new File[] {
        };
    }
    
    /**
     * Executes the given script files.
     * @param scriptFiles The script files to execute.
     * @param scripts The script (as strings) to execute
     * @param dataSourceFileName The file name of datasources.
     * @param dbType The database type
     * @throws Exception if an exception occured
     */
    public static void executeStartScripts(List<File> scriptFiles, List<String> scripts, String dataSourceFileName, String dbType) throws Exception
    {
        // Execute the start scripts, before starting Ametys, so without using any component
        if (dataSourceFileName == null)
        {
            return;
        }
        
        File dataSourceFile = new File(dataSourceFileName);
        GenericObjectPool<PoolableConnection> connectionPool = _getConnectionPool(dbType, dataSourceFile, "SQL-test");
        if (connectionPool == null)
        {
            // SQL-test was not found
            return;
        }
        Connection connection = null;
        
        try (connectionPool;
                PoolingDataSource<PoolableConnection> dataSource = new PoolingDataSource<>(connectionPool))
        {
            connection = dataSource.getConnection();
            
            for (File scriptFile : scriptFiles)
            {
                SQLScriptHelper.runScript(connection, new FileInputStream(scriptFile));
            }
            for (String script : scripts)
            {
                SQLScriptHelper.runScript(connection, script);
            }
        }
        finally 
        {
            ConnectionHelper.cleanup(connection);
        }
    }
    
    private static GenericObjectPool<PoolableConnection> _getConnectionPool(String dbType, File file, String datasourceId) throws Exception
    {
        Map<String, Object> parameters = _getSqlDatasourceParameters(file, datasourceId);
        if (parameters != null)
        {
            String url = (String) parameters.get(SQLDataSourceManager.PARAM_DATABASE_URL);
            String user = (String) parameters.get(SQLDataSourceManager.PARAM_DATABASE_USER);
            String password = (String) parameters.get(SQLDataSourceManager.PARAM_DATABASE_PASSWORD);
            
            return _getConnectionPool(url, user, password, dbType);
        }
        
        return null;
    }
    
    private static Map<String, Object> _getSqlDatasourceParameters(File file, String datasourceId) throws Exception
    {
        // Read datasources
        if (file.exists())
        {
            Configuration configuration = new DefaultConfigurationBuilder().buildFromFile(file);
            for (Configuration dsConfig : configuration.getChildren("datasource"))
            {
                String id = dsConfig.getAttribute("id");
                if (!id.equals(datasourceId))
                {
                    // this is not the wanted datasource
                    break;
                }
                
                Map<String, Object> parameters = new HashMap<>();
                
                Configuration[] paramsConfig = dsConfig.getChild("parameters").getChildren();
                for (Configuration paramConfig : paramsConfig)
                {
                    String value = paramConfig.getValue("");
                    parameters.put(paramConfig.getName(), value);
                }
                
                return parameters;
            }
        }
        
        return null;
    }
    
    private static GenericObjectPool<PoolableConnection> _getConnectionPool(String url, String user, String password, String dbType)
    {
        ConnectionFactory connectionFactory = new DriverManagerConnectionFactory(url, user, password);
        PoolableConnectionFactory poolableConnectionFactory = new PoolableConnectionFactory(connectionFactory, null);
        
        GenericObjectPool<PoolableConnection> connectionPool = new GenericObjectPool<>(poolableConnectionFactory);
        connectionPool.setMaxTotal(-1);
        connectionPool.setMaxIdle(10);
        connectionPool.setMinIdle(2);
        connectionPool.setTestOnBorrow(true);
        connectionPool.setTestOnReturn(false);
        connectionPool.setTestWhileIdle(true);
        connectionPool.setTimeBetweenEvictionRuns(Duration.ofMillis(1000 * 60 * 30));
        
        poolableConnectionFactory.setPool(connectionPool);
        poolableConnectionFactory.setValidationQuery(_getValidationQuery(dbType));
        poolableConnectionFactory.setDefaultAutoCommit(true);
        poolableConnectionFactory.setDefaultReadOnly(false);
        
        return connectionPool;
    }
    
    private static String _getValidationQuery(String dbType)
    {
        // Dirty way to do it, but we have not access to SQLDataSourcemanager at this time...
        switch (dbType)
        {
            case "derby":
                return "SELECT 1 FROM SYS.SYSTABLES";
            case "hsqldb":
                return "SELECT 1 FROM INFORMATION_SCHEMA.SYSTEM_USERS";
            case "oracle":
                return "SELECT 1 FROM DUAL";
            case "mysql":
            case "postgresql":
            default:
                return "SELECT 1";
        }
    }
    
    static void createDatabase(String url, String user, String password, String name, String roleForOracle, String dbType) throws Exception
    {
        _createOrDropDatabase(url, user, password, name, roleForOracle, dbType, true);
    }
    
    static void dropDatabase(String url, String user, String password, String name, String dbType) throws Exception
    {
        _createOrDropDatabase(url, user, password, name, null, dbType, false);
    }
    
    private static void _createOrDropDatabase(String url, String user, String password, String name, String roleForOracle, String dbType, boolean isCreate) throws Exception
    {
        Connection connection = null;
        try (GenericObjectPool<PoolableConnection> connectionPool = _getConnectionPool(url, user, password, dbType);
                PoolingDataSource<PoolableConnection> dataSource = new PoolingDataSource<>(connectionPool);)
        {
            connection = dataSource.getConnection();
            String script = isCreate ? _createQuery(dbType, name, roleForOracle) : _dropQuery(dbType, name);
            SQLScriptHelper.runScript(connection, script);
        }
        finally
        {
            ConnectionHelper.cleanup(connection);
        }
    }
    
    private static String _createQuery(String dbType, String dbName, String roleForOracle)
    {
        switch (dbType)
        {
            case "oracle":
                return "CREATE USER " + dbName + " IDENTIFIED BY " + dbName + " DEFAULT TABLESPACE USERS QUOTA UNLIMITED ON USERS;"
                        + "\nGRANT " + roleForOracle + " TO " + dbName + ";";
            case "mysql":
            case "postgresql":
                return "CREATE DATABASE " + dbName + ";";
            default:
                return null;
        }
    }
    
    private static String _dropQuery(String dbType, String dbName)
    {
        switch (dbType)
        {
            case "oracle":
                return "DROP USER " + dbName + " CASCADE;";
            case "mysql":
            case "postgresql":
                return "DROP DATABASE IF EXISTS " + dbName + ";";
            default:
                return null;
        }
    }
    
    static enum ShutdownStatus
    {
        CORRECTLY_SHUTDOWN,
        NOTHING_DONE,
        DATABASE_NOT_FOUND,
    }
    
    static ShutdownStatus shutdownDatabase(String sqlFile, String datasourceId, String dbType) throws Exception
    {
        switch (dbType)
        {
            case "hsqldb":
                String sqlStatement = "SHUTDOWN;";
                ScriptHelper.executeStartScripts(Collections.EMPTY_LIST, Collections.singletonList(sqlStatement), sqlFile, dbType);
                return ShutdownStatus.CORRECTLY_SHUTDOWN;
            case "derby":
                return _shutdownDerbyDb(sqlFile, datasourceId);
            default:
                return ShutdownStatus.NOTHING_DONE;
        }
    }
    
    private static ShutdownStatus _shutdownDerbyDb(String sqlFile, String datasourceId) throws Exception
    {
        Map<String, Object> parameters;
        if (SQLDataSourceManager.AMETYS_INTERNAL_DATASOURCE_ID.equals(datasourceId))
        {
            parameters = SQLDataSourceManager.getInternalDataSourceDefinition().getParameters();
        }
        else if (sqlFile == null)
        {
            return ShutdownStatus.DATABASE_NOT_FOUND;
        }
        else
        {
            parameters = _getSqlDatasourceParameters(new File(sqlFile), datasourceId);
        }
        
        if (parameters == null)
        {
            return ShutdownStatus.NOTHING_DONE;
        }
        
        String url = (String) parameters.get(SQLDataSourceManager.PARAM_DATABASE_URL);
        
        // https://db.apache.org/derby/papers/DerbyTut/embedded_intro.html#shutdown
        String shutdownUrl = url + ";shutdown=true";
        try
        {
            DriverManager.getConnection(shutdownUrl);
            return ShutdownStatus.CORRECTLY_SHUTDOWN;
        }
        catch (SQLException e)
        {
            return _handleDerbyExceptionOnShutdown(e);
        }
    }
    
    private static ShutdownStatus _handleDerbyExceptionOnShutdown(SQLException e) throws SQLException
    {
        // http://db.apache.org/derby/docs/10.8/ref/rrefexcept71493.html
        switch (e.getSQLState())
        {
            case "08006":
                // it is the expected SQLSTATE
                return ShutdownStatus.CORRECTLY_SHUTDOWN;
            case "XJ004":
                // database not found because was probably not created
                return ShutdownStatus.DATABASE_NOT_FOUND;
            default:
                throw e;
        }
    }
}
