/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import java.io.File;

import org.ametys.core.datasource.ConnectionHelper;

/**
 * Implements this interface if your Test Case needs to execute scripts before the Test Case (by default) or before each test (by annotating with {@link Unoptimized}). 
 */
public interface ScriptTestCase
{
    /**
     * Get the test dbtype.
     * <br>By default return {@link ConnectionHelper#DATABASE_DERBY Derby} DB.
     * @return One of '{@link ConnectionHelper#DATABASE_MYSQL mysql}', '{@link ConnectionHelper#DATABASE_DERBY derby}', 
     * '{@link ConnectionHelper#DATABASE_HSQLDB hsqldb}', '{@link ConnectionHelper#DATABASE_ORACLE oracle}', '{@link ConnectionHelper#DATABASE_POSTGRES postgresql}' 
     */
    default String getDBType()
    {
        return ConnectionHelper.DATABASE_DERBY;
    }
    
    /**
     * Gets the datasource configuration file to use.
     * <br>The start scripts from {@link #getStartScriptFiles()} will be executed on the "SQL-test" datasource if found.
     * <br>If you override this method, you probably have to also override {@link #getStartScriptFiles()} method.
     * @return the datasource file to use.
     */
    default String getDataSourceFile()
    {
        return ScriptHelper.getDataSourceFile(getDBType());
    }
    
    /**
     * Provides the SQL scripts to run before any test / before each test invocation.
     * <br>Those scripts are executed before starting Ametys, usually for dropping tables.
     * <br>If you override this method, you probably have to also override {@link #getDataSourceFile()} method.
     * @return the SQL script files to run.
     */
    default File[] getStartScriptFiles()
    {
        return ScriptHelper.getStartScriptFiles(getDBType());
    }
    
    /**
     * Provides the SQL scripts to run before any test / before each test invocation.
     * <br>Those scripts are executed before starting Ametys, usually for dropping tables.
     * <br>If you override this method, you probably have to also override {@link #getDataSourceFile()} method.
     * @return the SQL scripts to run.
     * @throws Exception If an error occurred building the script
     */
    default String[] getStartScripts() throws Exception
    {
        return new String[0];
    }
}
