/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.InetSocketAddress;
import java.net.URI;
import java.net.URLDecoder;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sun.net.httpserver.Filter;
import com.sun.net.httpserver.Headers;
import com.sun.net.httpserver.HttpContext;
import com.sun.net.httpserver.HttpExchange;
import com.sun.net.httpserver.HttpServer;

/**
 * Utility methods for mocking a http server.
 */
public final class ServerMocker
{
    private ServerMocker()
    {
        // nothing
    }
    
    /**
     * Creates and mock a {@link HttpServer}.
     * <br>The {@link HttpServer#start()} method is already called. You <b>MUST</b> call {@link HttpServer#stop(int)} when your test has finished to run. 
     * @param port The port number
     * @param cocoon The cocoon wrapper
     * @return the mocked {@link HttpServer}
     * @throws IOException if an I/O exception occured
     */
    public static final HttpServer mockServer(int port, CocoonWrapper cocoon) throws IOException
    {
        Logger logger = LoggerFactory.getLogger(ServerMocker.class);
        HttpServer httpServer = HttpServer.create(new InetSocketAddress(port), 0);
        HttpContext context = httpServer.createContext("/", exchange ->
        {
            // Parameters
            @SuppressWarnings("unchecked")
            Map<String, String> requestParameters = (Map<String, String>) exchange.getAttribute("parameters");
            
            // Headers
            Headers headers = exchange.getRequestHeaders();
            Map<String, String> requestHeaders = new HashMap<>();
            for (String headerName : headers.keySet())
            {
                requestHeaders.put(headerName, headers.getFirst(headerName));
            }
            
            // Cocoon process URI
            String uri = exchange.getRequestURI().toString();
            String contextPath = "/cms/";
            uri = uri.startsWith(contextPath) ? uri.substring(contextPath.length()) : uri;
            logger.info("Cocoon will process request '{}' (params: {})", uri, requestParameters);
                
            try
            {
                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                cocoon.processURI(uri, baos, requestParameters, new HashMap<>(), requestHeaders);
                
                // Write response
                byte[] bytes = baos.toByteArray();
                exchange.sendResponseHeaders(HttpURLConnection.HTTP_OK, bytes.length);
                OutputStream out = exchange.getResponseBody();
                out.write(bytes);
                out.flush();
            }
            catch (Exception e)
            {
                logger.error("An error occured when trying to process request '{}' (params: {})", uri, requestParameters, e);
                throw new IOException(e);
            }
            finally
            {
                exchange.close();
            }
        });
        context.getFilters().add(new ParameterFilter());
        httpServer.start();
        return httpServer;
    }
    
    static class ParameterFilter extends Filter
    {
        @Override
        public String description()
        {
            return "Parses the requested URI for parameters";
        }

        @Override
        public void doFilter(HttpExchange exchange, Chain chain) throws IOException
        {
            _parseGetParameters(exchange);
            _parsePostParameters(exchange);
            chain.doFilter(exchange);
        }
        
        private void _parseGetParameters(HttpExchange exchange) throws UnsupportedEncodingException
        {
            Map<String, Object> parameters = new HashMap<>();
            URI requestedUri = exchange.getRequestURI();
            String query = requestedUri.getRawQuery();
            _parseQuery(query, parameters);
            exchange.setAttribute("parameters", parameters);
        }

        private void _parsePostParameters(HttpExchange exchange) throws IOException
        {
            if ("post".equalsIgnoreCase(exchange.getRequestMethod()))
            {
                @SuppressWarnings("unchecked")
                Map<String, Object> parameters = (Map<String, Object>) exchange.getAttribute("parameters");
                InputStreamReader isr = new InputStreamReader(exchange.getRequestBody(), StandardCharsets.UTF_8);
                BufferedReader br = new BufferedReader(isr);
                String query = br.readLine();
                _parseQuery(query, parameters);
            }
        }

        @SuppressWarnings("unchecked")
        private void _parseQuery(String query, Map<String, Object> parameters) throws UnsupportedEncodingException
        {
            if (query != null)
            {
                String[] pairs = query.split("[&]");
                for (String pair : pairs) 
                {
                    String[] param = pair.split("[=]");
                    String key = null;
                    String value = null;
                    if (param.length > 0)
                    {
                        key = URLDecoder.decode(param[0], System.getProperty("file.encoding"));
                    }
                    if (param.length > 1)
                    {
                        value = URLDecoder.decode(param[1], System.getProperty("file.encoding"));
                    }
                    if (parameters.containsKey(key))
                    {
                        Object obj = parameters.get(key);
                        if (obj instanceof List<?>) 
                        {
                            List<String> values = (List<String>) obj;
                            values.add(value);
                        }
                        else if (obj instanceof String)
                        {
                            List<String> values = new ArrayList<>();
                            values.add((String) obj);
                            values.add(value);
                            parameters.put(key, values);
                        }
                    }
                    else
                    {
                        parameters.put(key, value);
                    }
                }
            }
        }
    }
}
