/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.framework;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.apache.commons.io.FileUtils;
import org.slf4j.Logger;

/**
 * Represents a set of configuration items (mostly file names) for running a Cocoon application.
 */
public class TestConfigSet
{
    /** The default runtime */
    public static final AmetysRuntime DEFAULT_RUNTIME = AmetysRuntime.fromContent(
              "<runtime xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"" + System.lineSeparator()
            + "         xsi:noNamespaceSchemaLocation=\"http://www.ametys.org/schema/runtime/runtime-4.1.xsd\">" + System.lineSeparator()
            + "    <application><version>2.0.1</version></application>" + System.lineSeparator()
            + "    <plugins/>" + System.lineSeparator()
            + "    <components/>" + System.lineSeparator()
            + "</runtime>", 
            TestConfigSet.class.getSimpleName() + "::DEFAULT_RUNTIME");
    
    private AmetysRuntime _runtime;
    private String _configFile;
    private String _sqlFile;
    private String _ldapFile;
    private String _contextPath = "";
    private String _webappPath;
    private List<Path> _tmpFiles = new ArrayList<>();
    
    /**
     * Creates a new configset.
     */
    public TestConfigSet()
    {
        // nothing
    }
    
    AmetysRuntime getRuntime()
    {
        return _runtime;
    }
    
    String getConfigFile()
    {
        return _configFile;
    }
    
    String getSqlFile()
    {
        return _sqlFile;
    }
    
    String getLdapFile()
    {
        return _ldapFile;
    }
    
    String getWebappPath()
    {
        return _webappPath;
    }
    
    String getContextPath()
    {
        return _contextPath;
    }
    
    /**
     * Use the given runtime file
     * @param runtimeFile The runtime file to use
     * @return The same configset
     */
    public TestConfigSet withRuntimeFile(String runtimeFile)
    {
        _runtime = AmetysRuntime.fromFile(runtimeFile);
        return this;
    }
    
    /**
     * Use the given runtime content
     * @param runtimeContent The runtime content to use
     * @param source The source of the runtime content
     * @return The same configset
     */
    public TestConfigSet withRuntimeContent(String runtimeContent, String source)
    {
        _runtime = AmetysRuntime.fromContent(runtimeContent, source);
        return this;
    }
    
    /**
     * Use the given runtime object
     * @param runtime The runtime object to use
     * @return The same configset
     */
    public TestConfigSet withRuntime(AmetysRuntime runtime)
    {
        _runtime = runtime;
        return this;
    }
    
    /**
     * Use the given configuration file
     * @param configFile The configuration file to use
     * @return The same configset
     */
    public TestConfigSet withConfigFile(String configFile)
    {
        _configFile = configFile;
        return this;
    }
    
    /**
     * Use the given SQL datasource file
     * @param sqlFile The SQL datasource file to use
     * @return The same configset
     */
    public TestConfigSet withSqlFile(String sqlFile)
    {
        _sqlFile = sqlFile;
        return this;
    }
    
    /**
     * Use the given LDAP datasource file
     * @param ldapFile The LDAP datasource file to use
     * @return The same configset
     */
    public TestConfigSet withLdapFile(String ldapFile)
    {
        _ldapFile = ldapFile;
        return this;
    }
    
    /**
     * Use the given path for localizing webapp.
     * @param webappPath The path to use.
     * @return The same configset
     */
    public TestConfigSet withWebappPath(String webappPath)
    {
        _webappPath = webappPath;
        return this;
    }
    
    /**
     * Use the given context path for requests.
     * @param contextPath The context path to use.
     * @return The same configset
     */
    public TestConfigSet withContextPath(String contextPath)
    {
        _contextPath = contextPath;
        return this;
    }
    
    TestConfigSet resolve(Map<String, String> variableValues) throws IOException
    {
        List<Path> tmpFiles = new ArrayList<>();
        AmetysRuntime runtime;
        String configFile;
        String sqlFile;
        String ldapFile;
        
        if (variableValues.isEmpty())
        {
            runtime = _runtime;
            configFile = _configFile;
            sqlFile = _sqlFile;
            ldapFile = _ldapFile;
        }
        else
        {
            runtime = _runtime != null && _runtime.isFromFile()
                    ? AmetysRuntime.fromFile(_getResolvedFile(_runtime.getRuntimeFile().getPath(), variableValues, tmpFiles))
                    : _runtime /* we do not handle yet variables in an AmetysRuntime from string content */;
            configFile = _getResolvedFile(_configFile, variableValues, tmpFiles);
            sqlFile = _getResolvedFile(_sqlFile, variableValues, tmpFiles);
            ldapFile = _getResolvedFile(_ldapFile, variableValues, tmpFiles);
        }
        
        String webappPath = _getResolvedWebappPath(_webappPath, tmpFiles);
        
        TestConfigSet resolvedConfigSet = new ResolvedTestConfigSet()
                .withRuntime(runtime)
                .withConfigFile(configFile)
                .withSqlFile(sqlFile)
                .withLdapFile(ldapFile)
                .withWebappPath(webappPath)
                .withContextPath(_contextPath);
        resolvedConfigSet._tmpFiles = tmpFiles;
        
        return resolvedConfigSet;
    }
    
    void removeTmpFiles(Logger logger)
    {
        for (Path tmpFile : _tmpFiles)
        {
            try
            {
                _removeTmpFile(tmpFile, logger);
            }
            catch (IOException e)
            {
                logger.error("Cannot delete file or directory '{}'", tmpFile, e);
            }
        }
    }
    
    private void _removeTmpFile(Path tmpFile, Logger logger) throws IOException
    {
        if (Files.isDirectory(tmpFile))
        {
            FileUtils.deleteDirectory(tmpFile.toFile());
        }
        else
        {
            Files.delete(tmpFile);
        }
        logger.info("'{}' was successfully removed.", tmpFile);
    }
    
    private static String _getResolvedFile(String origFile, Map<String, String> variableValues, List<Path> tmpFiles) throws IOException
    {
        if (origFile == null)
        {
            return null;
        }
        
        File source = new File(origFile);
        if (!source.exists() || !source.isFile())
        {
            return origFile;
        }
        
        Path dir = Paths.get("tmp");
        String suffix = source.getName() + "-";
        Files.createDirectories(dir);
        Path target = Files.createTempFile(dir, suffix, ".tmp");
        tmpFiles.add(target);
        
        try (BufferedWriter writer = Files.newBufferedWriter(target, StandardCharsets.UTF_8);)
        {
            _writeLines(source, variableValues, writer);
            writer.flush();
        }
        
        return target.toString();
    }
    
    private static void _writeLines(File source, Map<String, String> variableValues, BufferedWriter writer) throws IOException
    {
        List<String> allLines = Files.readAllLines(source.toPath());
        for (String line : allLines)
        {
            String replacedLine = _replace(line, variableValues);
            writer.write(replacedLine);
            writer.write(System.lineSeparator());
        }
    }
    
    private static String _replace(String line, Map<String, String> variableValues)
    {
        String replacedLine = line;
        for (String variableName : variableValues.keySet())
        {
            replacedLine = replacedLine.replaceAll("\\$\\{" + variableName + "\\}", variableValues.get(variableName));
        }
        return replacedLine;
    }
    
    private static String _getResolvedWebappPath(String origWebappPath, List<Path> tmpFiles) throws IOException
    {
        if (origWebappPath == null)
        {
            return null;
        }
        
        File sourceDir = new File(origWebappPath);
        if (!sourceDir.exists() || !sourceDir.isDirectory())
        {
            return origWebappPath;
        }
        
        Path dir = Paths.get("tmp");
        String prefix = sourceDir.getName() + "-";
        Files.createDirectories(dir);
        Path target = Files.createTempDirectory(dir, prefix);
        tmpFiles.add(target);
        
        FileUtils.copyDirectory(sourceDir, target.toFile());
        
        return target.toString();
    }
    
    @Override
    public String toString()
    {
        return "TestConfigSet {"
                + "\n\truntime=" + _runtime + ", "
                + "\n\tconfigFile=" + _configFile + ", "
                + "\n\tsqlFile=" + _sqlFile + ", "
                + "\n\tldapFile=" + _ldapFile + ", "
                + "\n\twebappPath=" + _webappPath + ", "
                + "\n\tcontextPath=" + _contextPath
                + "\n}";
    }
    
    static final class ResolvedTestConfigSet extends TestConfigSet
    {
        @Override
        TestConfigSet resolve(Map<String, String> variableValues) throws IOException
        {
            throw new IllegalStateException("The TestConfigSet is already resolved, it cannot be resolved again.");
        }
    }
}
