/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.rights.access.controller;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.group.GroupIdentity;
import org.ametys.core.right.AccessController;
import org.ametys.core.right.AccessController.AccessResult;
import org.ametys.core.user.UserIdentity;
import org.ametys.runtime.test.framework.Unoptimized;

/**
 * Common test class for testing a string based {@link AccessController}
 */
@Unoptimized
public abstract class AbstractStringBasedAccessControllerTestCase extends AbstractAccessControllerTestCase
{
    /**
     * The first test right context
     * @return The context to test
     */
    protected Object _getTest1()
    {
        return "/test1";
    }

    /**
     * The second test right context
     * @return The context to test
     */
    protected Object _getTest2()
    {
        return "/test2";
    }
    
    /**
     * The third test right context to test inheritance
     * @return The context to test
     */
    protected Object _getTest3()
    {
        return "/test3";
    }
    
    /**
     * A sub context of test 3
     * @return The context to test
     */
    protected Object _getTest3SubContext()
    {
        return "/test3/foo";
    }
    
    /**
     * Empty the cache of RightManager
     * To use when testing #getPermission and edit the allow/deny profile
     */
    protected void emptyRightManagerCache()
    {
        _rightManager.clearCaches();
    }
    
    /**
     * Tests responding UNKNOWN when no assignments set
     */
    @Test
    public void testEmpty()
    {
        Object test = _getTest1();
        
        UserIdentity user1 = new UserIdentity("user1", "population1");
        
        GroupIdentity group1 = new GroupIdentity("group1", "directory1");
        GroupIdentity group2 = new GroupIdentity("group2", "directory2");
        Set<GroupIdentity> bothGroups = Stream.of(group1, group2).collect(Collectors.toSet());
        
        // initially nothing is set
        Assert.assertEquals(AccessResult.UNKNOWN, _accessController.getPermission(user1, Collections.EMPTY_SET, "right1", null));
        Assert.assertEquals(AccessResult.UNKNOWN, _accessController.getPermission(user1, Collections.EMPTY_SET, "right1", test));

        Assert.assertEquals(AccessResult.UNKNOWN, _accessController.getPermission(user1, bothGroups, "right1", test));
    }
    
    /**
     * Tests getPermissions method
     */
    @Test
    public void testGetPermissions()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        
        initProfiles();
        String profile1 = _profileIds.get(0);
        String profile2 = _profileIds.get(1);
        String profile3 = _profileIds.get(2);
        String profile4 = _profileIds.get(3);
        String profile5 = _profileIds.get(4);
        String profile6 = _profileIds.get(5);
        String profile7 = _profileIds.get(6);
        
        UserIdentity user1 = new UserIdentity("user1", "population1");
        UserIdentity user2 = new UserIdentity("user2", "population1");
        UserIdentity user3 = new UserIdentity("user3", "population1");
        UserIdentity user4 = new UserIdentity("user4", "population1");
        
        GroupIdentity group1 = new GroupIdentity("group1", "directory1");
        GroupIdentity group2 = new GroupIdentity("group2", "directory23");
        GroupIdentity group3 = new GroupIdentity("group3", "directory23");
        
        _profileAssignmentStorageEP.allowProfileToAnyConnectedUser(profile1, test1);
        _profileAssignmentStorageEP.allowProfileToUser(user1, profile2, test1);
        _profileAssignmentStorageEP.denyProfileToUser(user1, profile3, test1);
        
        _profileAssignmentStorageEP.allowProfileToGroup(group1, profile4, test1);
        _profileAssignmentStorageEP.denyProfileToGroup(group1, profile5, test1);
        
        _profileAssignmentStorageEP.allowProfileToGroup(group2, profile6, test1);
        _profileAssignmentStorageEP.denyProfileToGroup(group2, profile7, test1);
        
        _profileAssignmentStorageEP.allowProfileToGroup(group3, profile5, test1);
        _profileAssignmentStorageEP.denyProfileToGroup(group3, profile4, test1);
        
        // test
        Assert.assertEquals(AccessResult.ANY_CONNECTED_ALLOWED, _accessController.getPermission(user1, Collections.EMPTY_SET, "right1", test1));
        Assert.assertEquals(AccessResult.USER_DENIED, _accessController.getPermission(user1, Collections.EMPTY_SET, "rightall", test1));
        Assert.assertEquals(AccessResult.UNKNOWN, _accessController.getPermission(user1, Collections.EMPTY_SET, "rightall", test2));
        Assert.assertEquals(AccessResult.UNKNOWN, _accessController.getPermission(user1, Collections.EMPTY_SET, "right4", test1));
        
        Assert.assertEquals(AccessResult.USER_ALLOWED, _accessController.getPermission(user1, Collections.EMPTY_SET, "right2", test1));
        Assert.assertEquals(AccessResult.USER_DENIED, _accessController.getPermission(user1, Collections.EMPTY_SET, "right3", test1));
        
        Assert.assertEquals(AccessResult.GROUP_ALLOWED, _accessController.getPermission(user2, Collections.singleton(group1), "right4", test1));
        Assert.assertEquals(AccessResult.GROUP_DENIED, _accessController.getPermission(user2, Collections.singleton(group1), "right5", test1));
        Assert.assertEquals(AccessResult.GROUP_ALLOWED, _accessController.getPermission(user3, Collections.singleton(group2), "right6", test1));
        Assert.assertEquals(AccessResult.GROUP_DENIED, _accessController.getPermission(user3, Collections.singleton(group2), "right7", test1));
        Assert.assertEquals(AccessResult.GROUP_DENIED, _accessController.getPermission(user3, Collections.singleton(group2), "right67", test1));
        Assert.assertEquals(AccessResult.GROUP_ALLOWED, _accessController.getPermission(user4, Collections.singleton(group3), "right5", test1));
        Assert.assertEquals(AccessResult.GROUP_DENIED, _accessController.getPermission(user4, Collections.singleton(group3), "right4", test1));
    }
    
    /**
     * Tests getPermission method with hierarchical contexts
     */
    @Test
    public void testInheritance()
    {
        Object ctx = _getTest3();
        Object subCtx = _getTest3SubContext();
        
        initProfiles();
        
        String profile1 = _profileIds.get(0);
        String profile2 = _profileIds.get(1);
        
        UserIdentity user1 = new UserIdentity("user1", "population1");
        UserIdentity user2 = new UserIdentity("user2", "population1");
        
        _profileAssignmentStorageEP.allowProfileToUser(user1, profile1, ctx);
        _profileAssignmentStorageEP.denyProfileToUser(user2, profile1, ctx);
        
        // Test permission by inheritance
        Assert.assertEquals(AccessResult.USER_ALLOWED, _accessController.getPermission(user1, Collections.EMPTY_SET, "right1", ctx));
        Assert.assertEquals(AccessResult.USER_ALLOWED, _accessController.getPermission(user1, Collections.EMPTY_SET, "right1", subCtx));
        Assert.assertEquals(AccessResult.USER_DENIED, _accessController.getPermission(user2, Collections.EMPTY_SET, "right1", ctx));
        Assert.assertEquals(AccessResult.USER_DENIED, _accessController.getPermission(user2, Collections.EMPTY_SET, "right1", subCtx));
        
        // Test getPermissionByRight by inheritance
        Map<String, AccessResult> expectedMap1 = new HashMap<>();
        expectedMap1.put("right1", AccessResult.USER_ALLOWED);
        expectedMap1.put("right12", AccessResult.USER_ALLOWED);
        expectedMap1.put("rightall", AccessResult.USER_ALLOWED);
        Assert.assertEquals(expectedMap1, _accessController.getPermissionByRight(user1, Collections.EMPTY_SET, ctx));
        Assert.assertEquals(expectedMap1, _accessController.getPermissionByRight(user1, Collections.EMPTY_SET, subCtx));
        
        _profileAssignmentStorageEP.allowProfileToUser(user1, profile2, subCtx);
        Map<String, AccessResult> expectedMap2 = new HashMap<>();
        expectedMap2.put("right1", AccessResult.USER_ALLOWED);
        expectedMap2.put("right12", AccessResult.USER_ALLOWED);
        expectedMap2.put("rightall", AccessResult.USER_ALLOWED);
        expectedMap2.put("right2", AccessResult.USER_ALLOWED);
        Assert.assertEquals(expectedMap2, _accessController.getPermissionByRight(user1, Collections.EMPTY_SET, subCtx));
        
        // Disallow inheritance
        _profileAssignmentStorageEP.disallowInheritance(subCtx, true);
        
        // Test getPermissionByRight without inheritance
        Map<String, AccessResult> expectedMap3 = new HashMap<>();
        expectedMap3.put("right12", AccessResult.USER_ALLOWED);
        expectedMap3.put("rightall", AccessResult.USER_ALLOWED);
        expectedMap3.put("right2", AccessResult.USER_ALLOWED);
        Assert.assertEquals(expectedMap3, _accessController.getPermissionByRight(user1, Collections.EMPTY_SET, subCtx));
        
        emptyRightManagerCache();
        
        // Test permission without inheritance
        Assert.assertEquals(AccessResult.UNKNOWN, _accessController.getPermission(user1, Collections.EMPTY_SET, "right1", subCtx));
        Assert.assertEquals(AccessResult.UNKNOWN, _accessController.getPermission(user2, Collections.EMPTY_SET, "right1", subCtx));
        
        emptyRightManagerCache();
        _profileAssignmentStorageEP.denyProfileToUser(user1, profile1, subCtx);
        _profileAssignmentStorageEP.allowProfileToUser(user2, profile1, subCtx);
        Assert.assertEquals(AccessResult.USER_DENIED, _accessController.getPermission(user1, Collections.EMPTY_SET, "right1", subCtx));
        Assert.assertEquals(AccessResult.USER_ALLOWED, _accessController.getPermission(user2, Collections.EMPTY_SET, "right1", subCtx));
    }
    
    /**
     * Tests getPermission with groups with hierarchical contexts
     */
    @Test
    public void testInheritanceWithGroups()
    {
        Object ctx = _getTest3();
        Object subCtx = _getTest3SubContext();
        
        initProfiles();
        
        String profile1 = _profileIds.get(0);
        
        UserIdentity user1 = new UserIdentity("user1", "population1");
        UserIdentity user2 = new UserIdentity("user2", "population1");
        
        GroupIdentity group1 = new GroupIdentity("group1", "directory1");
        GroupIdentity group2 = new GroupIdentity("group2", "directory1");
        
        Set<GroupIdentity> user1Groups = Stream.of(group1, group2).collect(Collectors.toSet()); // Let's assume for this test that user1 belongs to group1 and group2
        Set<GroupIdentity> user2Groups = Stream.of(group1).collect(Collectors.toSet()); // Let's assume for this test that user2 belongs only to group1
        
        _profileAssignmentStorageEP.allowProfileToGroup(group1, profile1, ctx);
        _profileAssignmentStorageEP.allowProfileToGroup(group2, profile1, subCtx);
        
        Assert.assertEquals(AccessResult.GROUP_ALLOWED, _accessController.getPermission(user1, user1Groups, "rightall", ctx));
        Assert.assertEquals(AccessResult.GROUP_ALLOWED, _accessController.getPermission(user2, user2Groups, "rightall", ctx));
        Assert.assertEquals(AccessResult.GROUP_ALLOWED, _accessController.getPermission(user1, user1Groups, "rightall", subCtx));
        Assert.assertEquals(AccessResult.GROUP_ALLOWED, _accessController.getPermission(user2, user2Groups, "rightall", subCtx));
        
        emptyRightManagerCache();
        
        // Disallow inheritance
        _profileAssignmentStorageEP.disallowInheritance(subCtx, true);
        
        Assert.assertEquals(AccessResult.GROUP_ALLOWED, _accessController.getPermission(user1, user1Groups, "rightall", ctx));
        Assert.assertEquals(AccessResult.GROUP_ALLOWED, _accessController.getPermission(user2, user2Groups, "rightall", ctx));
        Assert.assertEquals(AccessResult.GROUP_ALLOWED, _accessController.getPermission(user1, user1Groups, "rightall", subCtx));
        Assert.assertEquals(AccessResult.UNKNOWN, _accessController.getPermission(user2, user2Groups, "rightall", subCtx)); // user2 is no more authorized on subCtx when inheritance is disallowed
    }
    
    /**
     * Tests getPermissionsByProfile method
     */
    @Test
    public void testGetPermissionsByProfile()
    {
        Object test1 = _getTest1();
        
        initProfiles();
        String profile1 = _profileIds.get(0);
        String profile2 = _profileIds.get(1);
        String profile3 = _profileIds.get(2);
        String profile4 = _profileIds.get(3);
        // String profile5 = _profileIds.get(4);
        // String profile6 = _profileIds.get(5);
        // String profile7 = _profileIds.get(6);

        UserIdentity user1 = new UserIdentity("user1", "population1");
        
        GroupIdentity group1 = new GroupIdentity("group1", "directory1");
        
        // test initially empty
        Assert.assertEquals(Collections.EMPTY_MAP, _accessController.getPermissionByRight(user1, Collections.singleton(group1), test1));
        // test 1
        _profileAssignmentStorageEP.allowProfileToAnonymous(profile1, test1);
        _profileAssignmentStorageEP.allowProfileToAnyConnectedUser(profile2, test1);
        _profileAssignmentStorageEP.allowProfileToGroup(group1, profile3, test1);
        _profileAssignmentStorageEP.allowProfileToUser(user1, profile4, test1);
        
        Map<String, AccessResult> expectedMap1 = new HashMap<>();
        expectedMap1.put("right12", AccessResult.ANONYMOUS_ALLOWED);
        expectedMap1.put("right1", AccessResult.ANONYMOUS_ALLOWED);
        expectedMap1.put("rightall", AccessResult.ANONYMOUS_ALLOWED);
        expectedMap1.put("right2", AccessResult.ANY_CONNECTED_ALLOWED);
        expectedMap1.put("right3", AccessResult.GROUP_ALLOWED);
        expectedMap1.put("right4", AccessResult.USER_ALLOWED);
        Assert.assertEquals(expectedMap1, _accessController.getPermissionByRight(user1, Collections.singleton(group1), test1));
        
        // test 2
        _profileAssignmentStorageEP.denyProfileToAnonymous(profile1, test1);

        Map<String, AccessResult> expectedMap2 = new HashMap<>();
        expectedMap2.put("right12", AccessResult.ANY_CONNECTED_ALLOWED);
        expectedMap2.put("right1", AccessResult.ANONYMOUS_DENIED);
        expectedMap2.put("rightall", AccessResult.USER_ALLOWED);
        expectedMap2.put("right2", AccessResult.ANY_CONNECTED_ALLOWED);
        expectedMap2.put("right3", AccessResult.GROUP_ALLOWED);
        expectedMap2.put("right4", AccessResult.USER_ALLOWED);
        Assert.assertEquals(expectedMap2, _accessController.getPermissionByRight(user1, Collections.singleton(group1), test1));
        
        // test 3
        _profileAssignmentStorageEP.denyProfileToAnyConnectedUser(profile2, test1);
        
        Map<String, AccessResult> expectedMap3 = new HashMap<>();
        expectedMap3.put("right12", AccessResult.ANY_CONNECTED_DENIED);
        expectedMap3.put("right1", AccessResult.ANONYMOUS_DENIED);
        expectedMap3.put("rightall", AccessResult.USER_ALLOWED);
        expectedMap3.put("right2", AccessResult.ANY_CONNECTED_DENIED);
        expectedMap3.put("right3", AccessResult.GROUP_ALLOWED);
        expectedMap3.put("right4", AccessResult.USER_ALLOWED);
        Assert.assertEquals(expectedMap3, _accessController.getPermissionByRight(user1, Collections.singleton(group1), test1));
        
        // test 4
        _profileAssignmentStorageEP.denyProfileToGroup(group1, profile3, test1);
        
        Map<String, AccessResult> expectedMap4 = new HashMap<>();
        expectedMap4.put("right12", AccessResult.ANY_CONNECTED_DENIED);
        expectedMap4.put("right1", AccessResult.ANONYMOUS_DENIED);
        expectedMap4.put("rightall", AccessResult.USER_ALLOWED);
        expectedMap4.put("right2", AccessResult.ANY_CONNECTED_DENIED);
        expectedMap4.put("right3", AccessResult.GROUP_DENIED);
        expectedMap4.put("right4", AccessResult.USER_ALLOWED);
        Assert.assertEquals(expectedMap4, _accessController.getPermissionByRight(user1, Collections.singleton(group1), test1));
        
        // test 5
        _profileAssignmentStorageEP.denyProfileToUser(user1, profile4, test1);

        Map<String, AccessResult> expectedMap5 = new HashMap<>();
        expectedMap5.put("right12", AccessResult.ANY_CONNECTED_DENIED);
        expectedMap5.put("right1", AccessResult.ANONYMOUS_DENIED);
        expectedMap5.put("rightall", AccessResult.USER_DENIED);
        expectedMap5.put("right2", AccessResult.ANY_CONNECTED_DENIED);
        expectedMap5.put("right3", AccessResult.GROUP_DENIED);
        expectedMap5.put("right4", AccessResult.USER_DENIED);
        Assert.assertEquals(expectedMap5, _accessController.getPermissionByRight(user1, Collections.singleton(group1), test1));
        
        // Let's restart from scratch
        initProfiles();
        profile1 = _profileIds.get(0);
        // profile2 = _profileIds.get(1);
        // profile3 = _profileIds.get(2);
        // profile4 = _profileIds.get(3);
        // profile5 = _profileIds.get(4);
        // profile6 = _profileIds.get(5);
        // profile7 = _profileIds.get(6);
        
        // test 1
        _profileAssignmentStorageEP.denyProfileToAnonymous(profile1, test1);
        
        Map<String, AccessResult> expectedMap6 = new HashMap<>();
        expectedMap6.put("rightall", AccessResult.ANONYMOUS_DENIED);
        expectedMap6.put("right1", AccessResult.ANONYMOUS_DENIED);
        expectedMap6.put("right12", AccessResult.ANONYMOUS_DENIED);
        Assert.assertEquals(expectedMap6, _accessController.getPermissionByRight(user1, Collections.singleton(group1), test1));
    }
    
    /**
     * Tests hasUserAnyPermission method and its cache
     */
    @Test
    public void testHasUserAnyPermission()
    {
        Object test1 = _getTest1();
        
        initProfiles();
        String profile2 = _profileIds.get(1);
        
        UserIdentity user1 = new UserIdentity("user1", "population1");
        
        GroupIdentity group1 = new GroupIdentity("group1", "directory1");

        _profileAssignmentStorageEP.allowProfileToGroup(group1, profile2, test1);
        Assert.assertEquals(false, _accessController.hasUserAnyPermissionOnWorkspace(Collections.singleton(test1), user1, Collections.singleton(group1), "right1"));
        Assert.assertEquals(true, _accessController.hasUserAnyPermissionOnWorkspace(Collections.singleton(test1), user1, Collections.singleton(group1), "right2"));
        Assert.assertEquals(true, _accessController.hasUserAnyPermissionOnWorkspace(Collections.singleton(test1), user1, Collections.singleton(group1), "right12"));
    }
}
