/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.rights.storage.assignments;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.tuple.Pair;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.group.GroupIdentity;
import org.ametys.core.right.ModifiableProfileAssignmentStorage;
import org.ametys.core.right.ProfileAssignmentStorage;
import org.ametys.core.right.ProfileAssignmentStorage.AnonymousOrAnyConnectedKeys;
import org.ametys.core.right.ProfileAssignmentStorage.UserOrGroup;
import org.ametys.core.right.ProfileAssignmentStorageExtensionPoint;
import org.ametys.core.user.UserIdentity;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.Unoptimized;

/**
 * Common test class for testing the {@link ProfileAssignmentStorage} 
 */
@Unoptimized
public abstract class AbstractProfileAssignmentStorageTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    /** The storage for profile assignments */
    protected static ProfileAssignmentStorage _profileAssignmentStorage;
    private static ProfileAssignmentStorageExtensionPoint _profileAssignmentStorageEP;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _profileAssignmentStorageEP = (ProfileAssignmentStorageExtensionPoint) manager.lookup(ProfileAssignmentStorageExtensionPoint.ROLE);
        _profileAssignmentStorage = _profileAssignmentStorageEP.getExtension(_getExtensionId());
        Assert.assertNotNull("The #_getExtensionId method should point to a ProfileAssignmentStorage extension", _profileAssignmentStorage);
    }
    
    /**
     * Returns the id of the extension for {@link ProfileAssignmentStorage} to use.
     * @return the id of the extension to use.
     */
    protected abstract String _getExtensionId();
    
    /**
     * Gets the typed object 1 for tests
     * @return the typed object 1 for tests
     */
    protected abstract Object _getTest1();
    
    /**
     * Gets the typed object 2 for tests
     * @return the typed object 2 for tests
     */
    protected abstract Object _getTest2();
    
    /**
     * Tests the {@link ProfileAssignmentStorageExtensionPoint#hasUserAnyPermission(Set, UserIdentity, Set, Set)} method
     */
    @Test
    public void testHasPermission()
    {
        Object test1 = _getTest1();
        
        Set<Object> workspacesContexts = new  LinkedHashSet<>();
        workspacesContexts.add(test1);
        
        String profile1 = "1";
        String profile2 = "2";
        String profile3 = "3";
        
        Set<String> userProfiles = Stream.of(profile1, profile2).collect(Collectors.toSet()); // Let's assume for this test that user1 has the profiles profile1 and profile2
        
        UserIdentity user1 = new UserIdentity("user1", "foo");
        UserIdentity user2 = new UserIdentity("user2", "bar");
        
        GroupIdentity group1 = new GroupIdentity("group1", "foofoo");
        GroupIdentity group2 = new GroupIdentity("group2", "barbar");
        GroupIdentity group3 = new GroupIdentity("group3", "barbar");
        Set<GroupIdentity> userGroups = Stream.of(group1, group2).collect(Collectors.toSet()); // Let's assume for this test that user1 belongs to group1 and group2
        
        // test initially empty
        Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).isEmpty());
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
         
            // ### USER TEST ###
            // Add another profile as allowed to test1
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user1), test1, profile3);
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).isEmpty());
            
            // Add profile1 as allowed to test1, but on another user
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user2), test1, profile1);
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).isEmpty());
            
            // Add profile1 as allowed to test1
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user1), test1, profile1);
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).contains("1"));
            
            // Add profile1 as denied to test1
            mProfileAssignmentStorage.removeAllowedUsers(Collections.singleton(user1), test1, profile1);
            mProfileAssignmentStorage.addDeniedUsers(Collections.singleton(user1), test1, profile1);
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).isEmpty());
            
            // Add profile2 as allowed to test1
            mProfileAssignmentStorage.removeAllowedUsers(Collections.singleton(user1), test1, profile2);
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user1), test1, profile2);
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).contains("2"));
            
            // ### GROUPS TEST ###
            // Clean for testing with assignments on groups
            mProfileAssignmentStorage.removeProfile(profile1);
            mProfileAssignmentStorage.removeProfile(profile2);
            
            // Add another profile as allowed to group1
            mProfileAssignmentStorage.addAllowedGroups(Collections.singleton(group1), test1, profile3);
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).isEmpty());
            
            // Add profile1 as allowed to test1, but on another group
            mProfileAssignmentStorage.addAllowedGroups(Collections.singleton(group3), test1, profile1);
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).isEmpty());
            
            // Add profile1 as allowed to test1
            mProfileAssignmentStorage.addAllowedGroups(Collections.singleton(group1), test1, profile1);
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).contains("1"));
            
            // Add profile1 as denied to test1
            mProfileAssignmentStorage.addDeniedGroups(Collections.singleton(group2), test1, profile1);
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).contains("1"));
            
            // Add profile2 as denied to test1
            mProfileAssignmentStorage.addDeniedGroups(Collections.singleton(group1), test1, profile2);
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).contains("1"));
            
            // ### ANYCONNECTED TEST ###
            // Clean for testing with assignments on anyconnected
            mProfileAssignmentStorage.removeProfile(profile1);
            mProfileAssignmentStorage.removeProfile(profile2);
            
            // Add another profile as allowed
            mProfileAssignmentStorage.addAllowedProfilesForAnyConnectedUser(test1, Collections.singleton(profile3));
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).isEmpty());
            
            // Add profile1 as allowed to test1
            mProfileAssignmentStorage.addAllowedProfilesForAnyConnectedUser(test1, Collections.singleton(profile1));
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).contains("1"));
            
            // Add profile1 as denied to test1
            mProfileAssignmentStorage.removeAllowedProfilesForAnyConnectedUser(test1, Collections.singleton(profile1));
            mProfileAssignmentStorage.addDeniedProfilesForAnyConnectedUser(test1, Collections.singleton(profile1));
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).isEmpty());
            
            // Add profile2 as allowed to test1
            mProfileAssignmentStorage.addAllowedProfilesForAnyConnectedUser(test1, Collections.singleton(profile2));
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).contains("2"));
            
            // ### ANONYMOUS TEST ###
            // Clean for testing with assignments on anonymous
            mProfileAssignmentStorage.removeProfile(profile1);
            mProfileAssignmentStorage.removeProfile(profile2);
            
            // Add another profile as allowed
            mProfileAssignmentStorage.addAllowedProfilesForAnonymous(test1, Collections.singleton(profile3));
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).isEmpty());
            
            // Add profile1 as allowed to test1
            mProfileAssignmentStorage.addAllowedProfilesForAnonymous(test1, Collections.singleton(profile1));
            Assert.assertTrue(_profileAssignmentStorageEP.hasUserAnyPermission(workspacesContexts, user1, userGroups, userProfiles).contains("1"));
            
            // Add profile1 as denied to test1
            mProfileAssignmentStorage.addDeniedProfilesForAnonymous(test1, Collections.singleton(profile1));
            
            // Add profile2 as allowed to test1
            mProfileAssignmentStorage.addAllowedProfilesForAnonymous(test1, Collections.singleton(profile2));
        }
        
    }
    
    /**
     * Tests the allowed profiles for any connected user 
     */
    @Test
    public void testAllowedProfilesForAnyConnectedUser()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        String profile1 = "1";
        String profile2 = "2";
        String profile3 = "3";
        String profile4 = "4";
        String profile5 = "5";
        
        Set<String> profilesTest1 = Stream.of(profile1, profile2).collect(Collectors.toSet());
        Set<String> profilesTest2 = Stream.of(profile1, profile3, profile4).collect(Collectors.toSet());
        
        // test initially empty
        Assert.assertTrue(_getAllowedProfilesForAnyConnectedUser(test1).isEmpty());
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
            
            // test add
            mProfileAssignmentStorage.addAllowedProfilesForAnyConnectedUser(test1, profilesTest1);
            
            mProfileAssignmentStorage.addAllowedProfilesForAnyConnectedUser(test2, profilesTest2);
            
            Assert.assertEquals(profilesTest1, _getAllowedProfilesForAnyConnectedUser(test1));
            Assert.assertEquals(profilesTest2, _getAllowedProfilesForAnyConnectedUser(test2));
            
            // test remove
            Set<String> profilesToRemove = Stream.of(profile1, profile4, profile5).collect(Collectors.toSet());
            Set<String> remainingProfiles = Collections.singleton(profile3);
            mProfileAssignmentStorage.removeAllowedProfilesForAnyConnectedUser(test2, profilesToRemove);
            Assert.assertEquals(profilesTest1, _getAllowedProfilesForAnyConnectedUser(test1));
            Assert.assertEquals(remainingProfiles, _getAllowedProfilesForAnyConnectedUser(test2));
            
            // test add
            Set<String> expectedProfiles = Stream.of(profile1, profile2, profile4).collect(Collectors.toSet());
            mProfileAssignmentStorage.addAllowedProfilesForAnyConnectedUser(test1, Collections.singleton(profile4));
            Assert.assertEquals(expectedProfiles, _getAllowedProfilesForAnyConnectedUser(test1));
        }
    }
    
    /**
     * Tests the denied profiles for any connected user 
     */
    @Test
    public void testDeniedProfilesForAnyConnectedUser()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        String profile1 = "1";
        String profile2 = "2";
        String profile3 = "3";
        String profile4 = "4";
        String profile5 = "5";
        
        Set<String> profilesTest1 = Stream.of(profile1, profile2).collect(Collectors.toSet());
        Set<String> profilesTest2 = Stream.of(profile1, profile3, profile4).collect(Collectors.toSet());
        
        // test initially empty
        Assert.assertTrue(_getDeniedProfilesForAnyConnectedUser(test1).isEmpty());
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
            
            // test add
            mProfileAssignmentStorage.addDeniedProfilesForAnyConnectedUser(test1, profilesTest1);
            
            mProfileAssignmentStorage.addDeniedProfilesForAnyConnectedUser(test2, profilesTest2);
            
            Assert.assertEquals(profilesTest1, _getDeniedProfilesForAnyConnectedUser(test1));
            Assert.assertEquals(profilesTest2, _getDeniedProfilesForAnyConnectedUser(test2));
            
            // test remove
            Set<String> profilesToRemove = Stream.of(profile1, profile4, profile5).collect(Collectors.toSet());
            Set<String> remainingProfiles = Collections.singleton(profile3);
            mProfileAssignmentStorage.removeDeniedProfilesForAnyConnectedUser(test2, profilesToRemove);
            Assert.assertEquals(profilesTest1, _getDeniedProfilesForAnyConnectedUser(test1));
            Assert.assertEquals(remainingProfiles, _getDeniedProfilesForAnyConnectedUser(test2));
            
            // test add
            Set<String> expectedProfiles = Stream.of(profile1, profile2, profile4).collect(Collectors.toSet());
            mProfileAssignmentStorage.addDeniedProfilesForAnyConnectedUser(test1, Collections.singleton(profile4));
            Assert.assertEquals(expectedProfiles, _getDeniedProfilesForAnyConnectedUser(test1));
        }
    }
    
    /**
     * Tests the allowed profiles for an anonymous user
     */
    @Test
    public void testAllowedProfilesForAnonymous()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        String profile1 = "1";
        String profile2 = "2";
        String profile3 = "3";
        String profile4 = "4";
        String profile5 = "5";
        
        Set<String> profilesTest1 = Stream.of(profile1, profile2).collect(Collectors.toSet());
        Set<String> profilesTest2 = Stream.of(profile1, profile3, profile4).collect(Collectors.toSet());
        
        // test initially empty
        Assert.assertTrue(_getAllowedProfilesForAnonymous(test1).isEmpty());
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
            
            // test add
            mProfileAssignmentStorage.addAllowedProfilesForAnonymous(test1, profilesTest1);
            
            mProfileAssignmentStorage.addAllowedProfilesForAnonymous(test2, profilesTest2);
            
            Assert.assertEquals(profilesTest1, _getAllowedProfilesForAnonymous(test1));
            Assert.assertEquals(profilesTest2, _getAllowedProfilesForAnonymous(test2));
            
            // test remove
            Set<String> profilesToRemove = Stream.of(profile1, profile4, profile5).collect(Collectors.toSet());
            Set<String> remainingProfiles = Collections.singleton(profile3);
            mProfileAssignmentStorage.removeAllowedProfilesForAnonymous(test2, profilesToRemove);
            Assert.assertEquals(profilesTest1, _getAllowedProfilesForAnonymous(test1));
            Assert.assertEquals(remainingProfiles, _getAllowedProfilesForAnonymous(test2));
            
            // test add
            Set<String> expectedProfiles = Stream.of(profile1, profile2, profile4).collect(Collectors.toSet());
            mProfileAssignmentStorage.addAllowedProfilesForAnonymous(test1, Collections.singleton(profile4));
            Assert.assertEquals(expectedProfiles, _getAllowedProfilesForAnonymous(test1));
        }
    }
    
    /**
     * Tests the denied profiles for an anonymous user
     */
    @Test
    public void testDeniedProfilesForAnonymous()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        String profile1 = "1";
        String profile2 = "2";
        String profile3 = "3";
        String profile4 = "4";
        String profile5 = "5";
        
        Set<String> profilesTest1 = Stream.of(profile1, profile2).collect(Collectors.toSet());
        Set<String> profilesTest2 = Stream.of(profile1, profile3, profile4).collect(Collectors.toSet());
        
        // test initially empty
        Assert.assertTrue(_getDeniedProfilesForAnonymous(test1).isEmpty());
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
            
            // test add
            mProfileAssignmentStorage.addDeniedProfilesForAnonymous(test1, profilesTest1);
            
            mProfileAssignmentStorage.addDeniedProfilesForAnonymous(test2, profilesTest2);
            
            Assert.assertEquals(profilesTest1, _getDeniedProfilesForAnonymous(test1));
            Assert.assertEquals(profilesTest2, _getDeniedProfilesForAnonymous(test2));
            
            // test remove
            Set<String> profilesToRemove = Stream.of(profile1, profile4, profile5).collect(Collectors.toSet());
            Set<String> remainingProfiles = Collections.singleton(profile3);
            mProfileAssignmentStorage.removeDeniedProfilesForAnonymous(test2, profilesToRemove);
            Assert.assertEquals(profilesTest1, _getDeniedProfilesForAnonymous(test1));
            Assert.assertEquals(remainingProfiles, _getDeniedProfilesForAnonymous(test2));
            
            // test add
            Set<String> expectedProfiles = Stream.of(profile1, profile2, profile4).collect(Collectors.toSet());
            mProfileAssignmentStorage.addDeniedProfilesForAnonymous(test1, Collections.singleton(profile4));
            Assert.assertEquals(expectedProfiles, _getDeniedProfilesForAnonymous(test1));
        }
    }
    
    /**
     * Tests the storage of allowed users
     */
    @Test
    public void testAllowedUsers()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        String profile1 = "1";
        String profile2 = "2";
        
        Set<String> bothProfiles = Stream.of(profile1, profile2).collect(Collectors.toSet());
        
        UserIdentity user1 = new UserIdentity("user1", "foo");
        UserIdentity user2 = new UserIdentity("user2", "bar");
        UserIdentity user3 = new UserIdentity("user3", "bar");
        
        Set<UserIdentity> bothUsers = Stream.of(user1, user2).collect(Collectors.toSet());
        Set<UserIdentity> threeUsers = Stream.of(user1, user2, user3).collect(Collectors.toSet());
        
        Map<UserIdentity, Set<String>> expectedUsers = new HashMap<>();
        
        // test initially empty
        Assert.assertTrue(_getAllowedProfilesForUsers(test1).isEmpty());
        Assert.assertTrue(_getAllowedUsers(test1, profile1).isEmpty());
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
            
            // test add
            mProfileAssignmentStorage.addAllowedUsers(bothUsers, test1, profile1);
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user1), test1, profile2);
            expectedUsers.put(user1, bothProfiles);
            expectedUsers.put(user2, Collections.singleton(profile1));
            Assert.assertEquals(expectedUsers, _getAllowedProfilesForUsers(test1));
            Assert.assertEquals(Collections.EMPTY_MAP, _getAllowedProfilesForUsers(test2));
            Assert.assertEquals(bothUsers, _getAllowedUsers(test1, profile1));
            Assert.assertEquals(Collections.singleton(user1), _getAllowedUsers(test1, profile2));
            Assert.assertEquals(Collections.EMPTY_SET, _getAllowedUsers(test2, profile1));
            
            // test removing non-existing associations
            mProfileAssignmentStorage.removeAllowedUsers(Collections.singleton(user3), test1, profile1);
            Assert.assertEquals(bothUsers, _getAllowedUsers(test1, profile1));
            mProfileAssignmentStorage.removeAllowedUsers(Collections.singleton(user1), test2, profile1);
            Assert.assertEquals(Collections.EMPTY_MAP, _getAllowedProfilesForUsers(test2));
            mProfileAssignmentStorage.removeAllowedUsers(Collections.singleton(user2), test1, profile2);
            Assert.assertEquals(Collections.singleton(user1), _getAllowedUsers(test1, profile2));
            
            // test remove profile
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user1), test2, profile1); //for testing the deletion is not made on another object
            mProfileAssignmentStorage.removeAllowedUsers(Collections.singleton(user1), test1, profile1);
            Assert.assertEquals(Collections.singleton(user2), _getAllowedUsers(test1, profile1));
            Assert.assertEquals(Collections.singleton(user1), _getAllowedUsers(test1, profile2)); //but still have the profile2
            Assert.assertEquals(Collections.singleton(user1), _getAllowedUsers(test2, profile1)); //but still it on another object
            mProfileAssignmentStorage.removeAllowedUsers(Collections.singleton(user2), test1, profile1);
            Assert.assertEquals(Collections.EMPTY_SET, _getAllowedUsers(test1, profile1));
            
            // test remove all profiles
            mProfileAssignmentStorage.addAllowedUsers(threeUsers, test1, profile1); // populate
            mProfileAssignmentStorage.addAllowedUsers(bothUsers, test1, profile2); // populate
            mProfileAssignmentStorage.removeAllowedUsers(bothUsers, test1); // this is what we want to test
            expectedUsers.clear();
            expectedUsers.put(user3, Collections.singleton(profile1));
            Assert.assertEquals(expectedUsers, _getAllowedProfilesForUsers(test1));
            expectedUsers.clear();
            expectedUsers.put(user1, Collections.singleton(profile1));
            Assert.assertEquals(expectedUsers, _getAllowedProfilesForUsers(test2));
        }
    }
    
    /**
     * Get anonymous allowed profiles on context
     * @param context The context
     * @return The profiles
     */
    protected Set<String> _getAllowedProfilesForAnonymous(Object context)
    {
        return _profileAssignmentStorage.getProfilesForAnonymousAndAnyConnectedUser(context)
                    .get(AnonymousOrAnyConnectedKeys.ANONYMOUS_ALLOWED);
    }

    /**
     * Get anonymous denied profiles on context
     * @param context The context
     * @return The profiles
     */
    protected Set<String> _getDeniedProfilesForAnonymous(Object context)
    {
        return _profileAssignmentStorage.getProfilesForAnonymousAndAnyConnectedUser(context)
                    .get(AnonymousOrAnyConnectedKeys.ANONYMOUS_DENIED);
    }

    /**
     * Get anyconnecteduser allowed profiles on context
     * @param context The context
     * @return The profiles
     */
    protected Set<String> _getAllowedProfilesForAnyConnectedUser(Object context)
    {
        return _profileAssignmentStorage.getProfilesForAnonymousAndAnyConnectedUser(context)
                    .get(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_ALLOWED);
    }
    
    /**
     * Get anyconnecteduser denied profiles on context
     * @param context The context
     * @return The profiles
     */
    protected Set<String> _getDeniedProfilesForAnyConnectedUser(Object context)
    {
        return _profileAssignmentStorage.getProfilesForAnonymousAndAnyConnectedUser(context)
                    .get(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_DENIED);
    }
    
    /**
     * Get allowed profiles for all users on context
     * @param context The context
     * @return The associated user and profiles
     */
    protected Map<UserIdentity, Set<String>> _getAllowedProfilesForUsers(Object context)
    {
        return _profileAssignmentStorage.getProfilesForUsers(context, null)
                                .entrySet()
                                .stream()
                                .map(entry -> Pair.of(entry.getKey(), entry.getValue().get(UserOrGroup.ALLOWED)))
                                .collect(Collectors.toMap(Pair::getLeft, Pair::getRight));
    }

    /**
     * Get allowed users on context with the profile
     * @param context The context
     * @param profile the profile
     * @return The users
     */
    protected Set<UserIdentity> _getAllowedUsers(Object context, String profile)
    {
        return _profileAssignmentStorage.getProfilesForUsers(context, null)
                                .entrySet()
                                .stream()
                                .filter(entry -> entry.getValue().get(UserOrGroup.ALLOWED).contains(profile))
                                .map(Map.Entry::getKey)
                                .collect(Collectors.toSet());
    }

    /**
     * Get denied profiles for all users on context
     * @param context The context
     * @return The associated user and profiles
     */
    protected Map<UserIdentity, Set<String>> _getDeniedProfilesForUsers(Object context)
    {
        return _profileAssignmentStorage.getProfilesForUsers(context, null)
                                .entrySet()
                                .stream()
                                .map(entry -> Pair.of(entry.getKey(), entry.getValue().get(UserOrGroup.DENIED)))
                                .collect(Collectors.toMap(Pair::getLeft, Pair::getRight));
    }

    /**
     * Get denied users on context with the profile
     * @param context The context
     * @param profile the profile
     * @return The users
     */
    protected Set<UserIdentity> _getDeniedUsers(Object context, String profile)
    {
        return _profileAssignmentStorage.getProfilesForUsers(context, null)
                                .entrySet()
                                .stream()
                                .filter(entry -> entry.getValue().get(UserOrGroup.DENIED).contains(profile))
                                .map(Map.Entry::getKey)
                                .collect(Collectors.toSet());
    }

    /**
     * Get allowed groups on context
     * @param context The context
     * @return The associated groups and profiles
     */
    protected Map<GroupIdentity, Set<String>> _getAllowedProfilesForGroups(Object context)
    {
        return _profileAssignmentStorage.getProfilesForGroups(context, null)
                                .entrySet()
                                .stream()
                                .map(entry -> Pair.of(entry.getKey(), entry.getValue().get(UserOrGroup.ALLOWED)))
                                .collect(Collectors.toMap(Pair::getLeft, Pair::getRight));
    }

    /**
     * Get allowed groups on context with the profile
     * @param context The context
     * @param profile the profile
     * @return The groups
     */
    protected Set<GroupIdentity> _getAllowedGroups(Object context, String profile)
    {
        return _profileAssignmentStorage.getProfilesForGroups(context, null)
                                .entrySet()
                                .stream()
                                .filter(entry -> entry.getValue().get(UserOrGroup.ALLOWED).contains(profile))
                                .map(Map.Entry::getKey)
                                .collect(Collectors.toSet());
    }

    /**
     * Get denied groups on context
     * @param context The context
     * @return The associated groups and profiles
     */
    protected Map<GroupIdentity, Set<String>> _getDeniedProfilesForGroups(Object context)
    {
        return _profileAssignmentStorage.getProfilesForGroups(context, null)
                                .entrySet()
                                .stream()
                                .map(entry -> Pair.of(entry.getKey(), entry.getValue().get(UserOrGroup.DENIED)))
                                .collect(Collectors.toMap(Pair::getLeft, Pair::getRight));
    }

    /**
     * Get denied groups on context with the profile
     * @param context The context
     * @param profile the profile
     * @return The groups
     */
    protected Set<GroupIdentity> _getDeniedGroups(Object context, String profile)
    {
        return _profileAssignmentStorage.getProfilesForGroups(context, null)
                                .entrySet()
                                .stream()
                                .filter(entry -> entry.getValue().get(UserOrGroup.DENIED).contains(profile))
                                .map(Map.Entry::getKey)
                                .collect(Collectors.toSet());
    }

    /**
     * Tests the storage of denied users
     */
    @Test
    public void testDeniedUsers()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        String profile1 = "1";
        String profile2 = "2";
        
        Set<String> bothProfiles = Stream.of(profile1, profile2).collect(Collectors.toSet());
        
        UserIdentity user1 = new UserIdentity("user1", "foo");
        UserIdentity user2 = new UserIdentity("user2", "bar");
        UserIdentity user3 = new UserIdentity("user3", "bar");
        
        Set<UserIdentity> bothUsers = Stream.of(user1, user2).collect(Collectors.toSet());
        Set<UserIdentity> threeUsers = Stream.of(user1, user2, user3).collect(Collectors.toSet());
        
        Map<UserIdentity, Set<String>> expectedUsers = new HashMap<>();
        
        // test initially empty
        Assert.assertTrue(_getDeniedProfilesForUsers(test1).isEmpty());
        Assert.assertTrue(_getDeniedUsers(test1, profile1).isEmpty());
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
            
            // test add
            mProfileAssignmentStorage.addDeniedUsers(bothUsers, test1, profile1);
            mProfileAssignmentStorage.addDeniedUsers(Collections.singleton(user1), test1, profile2);
            expectedUsers.put(user1, bothProfiles);
            expectedUsers.put(user2, Collections.singleton(profile1));
            Assert.assertEquals(expectedUsers, _getDeniedProfilesForUsers(test1));
            Assert.assertEquals(Collections.EMPTY_MAP, _getDeniedProfilesForUsers(test2));
            Assert.assertEquals(bothUsers, _getDeniedUsers(test1, profile1));
            Assert.assertEquals(Collections.singleton(user1), _getDeniedUsers(test1, profile2));
            Assert.assertEquals(Collections.EMPTY_SET, _getDeniedUsers(test2, profile1));
            
            // test removing non-existing associations
            mProfileAssignmentStorage.removeDeniedUsers(Collections.singleton(user3), test1, profile1);
            Assert.assertEquals(bothUsers, _getDeniedUsers(test1, profile1));
            mProfileAssignmentStorage.removeDeniedUsers(Collections.singleton(user1), test2, profile1);
            Assert.assertEquals(Collections.EMPTY_MAP, _getDeniedProfilesForUsers(test2));
            mProfileAssignmentStorage.removeDeniedUsers(Collections.singleton(user2), test1, profile2);
            Assert.assertEquals(Collections.singleton(user1), _getDeniedUsers(test1, profile2));
            
            // test remove profile
            mProfileAssignmentStorage.addDeniedUsers(Collections.singleton(user1), test2, profile1); //for testing the deletion is not made on another object
            mProfileAssignmentStorage.removeDeniedUsers(Collections.singleton(user1), test1, profile1);
            Assert.assertEquals(Collections.singleton(user2), _getDeniedUsers(test1, profile1));
            Assert.assertEquals(Collections.singleton(user1), _getDeniedUsers(test1, profile2)); //but still have the profile2
            Assert.assertEquals(Collections.singleton(user1), _getDeniedUsers(test2, profile1)); //but still it on another object
            mProfileAssignmentStorage.removeDeniedUsers(Collections.singleton(user2), test1, profile1);
            Assert.assertEquals(Collections.EMPTY_SET, _getDeniedUsers(test1, profile1));
            
            // test remove all profiles
            mProfileAssignmentStorage.addDeniedUsers(threeUsers, test1, profile1); // populate
            mProfileAssignmentStorage.addDeniedUsers(bothUsers, test1, profile2); // populate
            mProfileAssignmentStorage.removeDeniedUsers(bothUsers, test1); // this is what we want to test
            expectedUsers.clear();
            expectedUsers.put(user3, Collections.singleton(profile1));
            Assert.assertEquals(expectedUsers, _getDeniedProfilesForUsers(test1));
            expectedUsers.clear();
            expectedUsers.put(user1, Collections.singleton(profile1));
            Assert.assertEquals(expectedUsers, _getDeniedProfilesForUsers(test2));
        }
    }
    
    /**
     * Tests the storage of allowed groups
     */
    @Test
    public void testAllowedGroups()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        String profile1 = "1";
        String profile2 = "2";
        
        Set<String> bothProfiles = Stream.of(profile1, profile2).collect(Collectors.toSet());
        
        GroupIdentity group1 = new GroupIdentity("group1", "foo");
        GroupIdentity group2 = new GroupIdentity("group2", "bar");
        GroupIdentity group3 = new GroupIdentity("group3", "bar");
        
        Set<GroupIdentity> bothGroups = Stream.of(group1, group2).collect(Collectors.toSet());
        Set<GroupIdentity> threeGroups = Stream.of(group1, group2, group3).collect(Collectors.toSet());
        
        // test initially empty
        Assert.assertTrue(_getAllowedProfilesForGroups(test1).isEmpty());
        Assert.assertTrue(_getAllowedGroups(test1, profile1).isEmpty());
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
            
            // test add
            mProfileAssignmentStorage.addAllowedGroups(bothGroups, test1, profile1);
            mProfileAssignmentStorage.addAllowedGroups(Collections.singleton(group1), test1, profile2);
            Map<GroupIdentity, Set<String>> expectedGroups = new HashMap<>();
            expectedGroups.put(group1, bothProfiles);
            expectedGroups.put(group2, Collections.singleton(profile1));
            Assert.assertEquals(expectedGroups, _getAllowedProfilesForGroups(test1));
            Assert.assertEquals(Collections.EMPTY_MAP, _getAllowedProfilesForGroups(test2));
            Assert.assertEquals(bothGroups, _getAllowedGroups(test1, profile1));
            Assert.assertEquals(Collections.singleton(group1), _getAllowedGroups(test1, profile2));
            Assert.assertEquals(Collections.EMPTY_SET, _getAllowedGroups(test2, profile1));
            
            // test removing non-existing associations
            mProfileAssignmentStorage.removeAllowedGroups(Collections.singleton(group3), test1, profile1);
            Assert.assertEquals(bothGroups, _getAllowedGroups(test1, profile1));
            mProfileAssignmentStorage.removeAllowedGroups(Collections.singleton(group1), test2, profile1);
            Assert.assertEquals(Collections.EMPTY_MAP, _getAllowedProfilesForGroups(test2));
            mProfileAssignmentStorage.removeAllowedGroups(Collections.singleton(group2), test1, profile2);
            Assert.assertEquals(Collections.singleton(group1), _getAllowedGroups(test1, profile2));
            
            // test remove profile
            mProfileAssignmentStorage.addAllowedGroups(Collections.singleton(group1), test2, profile1); //for testing the deletion is not made on another object
            mProfileAssignmentStorage.removeAllowedGroups(Collections.singleton(group1), test1, profile1);
            Assert.assertEquals(Collections.singleton(group2), _getAllowedGroups(test1, profile1));
            Assert.assertEquals(Collections.singleton(group1), _getAllowedGroups(test1, profile2)); //but still have the profile2
            Assert.assertEquals(Collections.singleton(group1), _getAllowedGroups(test2, profile1)); //but still it on another object
            mProfileAssignmentStorage.removeAllowedGroups(Collections.singleton(group2), test1, profile1);
            Assert.assertEquals(Collections.EMPTY_SET, _getAllowedGroups(test1, profile1));
            
            // test remove all profiles
            mProfileAssignmentStorage.addAllowedGroups(threeGroups, test1, profile1); // populate
            mProfileAssignmentStorage.addAllowedGroups(bothGroups, test1, profile2); // populate
            mProfileAssignmentStorage.removeAllowedGroups(bothGroups, test1); // this is what we want to test
            expectedGroups.clear();
            expectedGroups.put(group3, Collections.singleton(profile1));
            Assert.assertEquals(expectedGroups, _getAllowedProfilesForGroups(test1));
            expectedGroups.clear();
            expectedGroups.put(group1, Collections.singleton(profile1));
            Assert.assertEquals(expectedGroups, _getAllowedProfilesForGroups(test2));
        }
    }
    
    /**
     * Tests the storage of denied groups
     */
    @Test
    public void testDeniedGroups()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        String profile1 = "1";
        String profile2 = "2";
        
        Set<String> bothProfiles = Stream.of(profile1, profile2).collect(Collectors.toSet());
        
        GroupIdentity group1 = new GroupIdentity("group1", "foo");
        GroupIdentity group2 = new GroupIdentity("group2", "bar");
        GroupIdentity group3 = new GroupIdentity("group3", "bar");
        
        Set<GroupIdentity> bothGroups = Stream.of(group1, group2).collect(Collectors.toSet());
        Set<GroupIdentity> threeGroups = Stream.of(group1, group2, group3).collect(Collectors.toSet());
        
        // test initially empty
        Assert.assertTrue(_getDeniedProfilesForGroups(test1).isEmpty());
        Assert.assertTrue(_getDeniedGroups(test1, profile1).isEmpty());
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
          
            // test add
            mProfileAssignmentStorage.addDeniedGroups(bothGroups, test1, profile1);
            mProfileAssignmentStorage.addDeniedGroups(Collections.singleton(group1), test1, profile2);
            Map<GroupIdentity, Set<String>> expectedGroups = new HashMap<>();
            expectedGroups.put(group1, bothProfiles);
            expectedGroups.put(group2, Collections.singleton(profile1));
            Assert.assertEquals(expectedGroups, _getDeniedProfilesForGroups(test1));
            Assert.assertEquals(Collections.EMPTY_MAP, _getDeniedProfilesForGroups(test2));
            Assert.assertEquals(bothGroups, _getDeniedGroups(test1, profile1));
            Assert.assertEquals(Collections.singleton(group1), _getDeniedGroups(test1, profile2));
            Assert.assertEquals(Collections.EMPTY_SET, _getDeniedGroups(test2, profile1));
            
            // test removing non-existing associations
            mProfileAssignmentStorage.removeDeniedGroups(Collections.singleton(group3), test1, profile1);
            Assert.assertEquals(bothGroups, _getDeniedGroups(test1, profile1));
            mProfileAssignmentStorage.removeDeniedGroups(Collections.singleton(group1), test2, profile1);
            Assert.assertEquals(Collections.EMPTY_MAP, _getDeniedProfilesForGroups(test2));
            mProfileAssignmentStorage.removeDeniedGroups(Collections.singleton(group2), test1, profile2);
            Assert.assertEquals(Collections.singleton(group1), _getDeniedGroups(test1, profile2));
            
            // test remove profile
            mProfileAssignmentStorage.addDeniedGroups(Collections.singleton(group1), test2, profile1); //for testing the deletion is not made on another object
            mProfileAssignmentStorage.removeDeniedGroups(Collections.singleton(group1), test1, profile1);
            Assert.assertEquals(Collections.singleton(group2), _getDeniedGroups(test1, profile1));
            Assert.assertEquals(Collections.singleton(group1), _getDeniedGroups(test1, profile2)); //but still have the profile2
            Assert.assertEquals(Collections.singleton(group1), _getDeniedGroups(test2, profile1)); //but still it on another object
            mProfileAssignmentStorage.removeDeniedGroups(Collections.singleton(group2), test1, profile1);
            Assert.assertEquals(Collections.EMPTY_SET, _getDeniedGroups(test1, profile1));
            
            // test remove all profiles
            mProfileAssignmentStorage.addDeniedGroups(threeGroups, test1, profile1); // populate
            mProfileAssignmentStorage.addDeniedGroups(bothGroups, test1, profile2); // populate
            mProfileAssignmentStorage.removeDeniedGroups(bothGroups, test1); // this is what we want to test
            expectedGroups.clear();
            expectedGroups.put(group3, Collections.singleton(profile1));
            Assert.assertEquals(expectedGroups, _getDeniedProfilesForGroups(test1));
            expectedGroups.clear();
            expectedGroups.put(group1, Collections.singleton(profile1));
            Assert.assertEquals(expectedGroups, _getDeniedProfilesForGroups(test2));
        }
    }
    
    /**
     * Tests the removing of a profile
     */
    @Test
    public void testRemoveProfile()
    {
        Object test1 = _getTest1();
        String profile1 = "1";
        String profile2 = "2";
        Set<String> bothProfiles = Stream.of(profile1, profile2).collect(Collectors.toSet());
        
        UserIdentity user1 = new UserIdentity("user1", "foo");
        
        GroupIdentity group1 = new GroupIdentity("group1", "foo");
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
          
            // populate
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user1), test1, profile1);
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user1), test1, profile2);
            
            mProfileAssignmentStorage.addDeniedUsers(Collections.singleton(user1), test1, profile1);
            mProfileAssignmentStorage.addDeniedUsers(Collections.singleton(user1), test1, profile2);
            
            mProfileAssignmentStorage.addAllowedGroups(Collections.singleton(group1), test1, profile1);
            mProfileAssignmentStorage.addAllowedGroups(Collections.singleton(group1), test1, profile2);
            
            mProfileAssignmentStorage.addDeniedGroups(Collections.singleton(group1), test1, profile1);
            mProfileAssignmentStorage.addDeniedGroups(Collections.singleton(group1), test1, profile2);
            
            mProfileAssignmentStorage.addAllowedProfilesForAnyConnectedUser(test1, bothProfiles);
            mProfileAssignmentStorage.addDeniedProfilesForAnyConnectedUser(test1, bothProfiles);
            
            mProfileAssignmentStorage.addAllowedProfilesForAnonymous(test1, bothProfiles);
            mProfileAssignmentStorage.addDeniedProfilesForAnonymous(test1, bothProfiles);
            
            // test removing
            mProfileAssignmentStorage.removeProfile(profile1);
            
            Map<UserIdentity, Set<String>> expectedUsers = new HashMap<>();
            Map<GroupIdentity, Set<String>> expectedGroups = new HashMap<>();
            
            expectedUsers.put(user1, Collections.singleton(profile2));
            Assert.assertEquals(expectedUsers, _getAllowedProfilesForUsers(test1));
            Assert.assertEquals(expectedUsers, _getDeniedProfilesForUsers(test1));
            
            expectedGroups.put(group1, Collections.singleton(profile2));
            Assert.assertEquals(expectedGroups, _getAllowedProfilesForGroups(test1));
            Assert.assertEquals(expectedGroups, _getDeniedProfilesForGroups(test1));
            
            Assert.assertEquals(Collections.singleton(profile2), _getAllowedProfilesForAnyConnectedUser(test1));
            Assert.assertEquals(Collections.singleton(profile2), _getDeniedProfilesForAnyConnectedUser(test1));
            Assert.assertEquals(Collections.singleton(profile2), _getAllowedProfilesForAnonymous(test1));
            Assert.assertEquals(Collections.singleton(profile2), _getDeniedProfilesForAnonymous(test1));
        }
    }
    
    /**
     * Tests the removing of a user
     */
    @Test
    public void testRemoveUser()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        String profile1 = "1";
        String profile2 = "2";
        
        UserIdentity user1 = new UserIdentity("user1", "foo");
        UserIdentity user2 = new UserIdentity("user2", "bar");
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
          
            // populate
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user1), test1, profile1);
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user2), test1, profile2);
            mProfileAssignmentStorage.addAllowedUsers(Collections.singleton(user1), test2, profile1);
            
            mProfileAssignmentStorage.addDeniedUsers(Collections.singleton(user2), test1, profile1);
            mProfileAssignmentStorage.addDeniedUsers(Collections.singleton(user1), test1, profile2);
            mProfileAssignmentStorage.addDeniedUsers(Collections.singleton(user2), test2, profile1);
            
            // test removing
            mProfileAssignmentStorage.removeUser(user1);
            
            Assert.assertEquals(Collections.EMPTY_SET, _getAllowedUsers(test1, profile1));
            Assert.assertEquals(Collections.singleton(user2), _getAllowedUsers(test1, profile2));
            Assert.assertEquals(Collections.EMPTY_SET, _getAllowedUsers(test2, profile1));
            
            Assert.assertEquals(Collections.singleton(user2), _getDeniedUsers(test1, profile1));
            Assert.assertEquals(Collections.EMPTY_SET, _getDeniedUsers(test1, profile2));
            Assert.assertEquals(Collections.singleton(user2), _getDeniedUsers(test2, profile1));
        }
    }
    
    /**
     * Tests the removing of a group
     */
    @Test
    public void testRemoveGroup()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        String profile1 = "1";
        String profile2 = "2";
        
        GroupIdentity group1 = new GroupIdentity("group1", "foo");
        GroupIdentity group2 = new GroupIdentity("group2", "bar");
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ModifiableProfileAssignmentStorage mProfileAssignmentStorage = (ModifiableProfileAssignmentStorage) _profileAssignmentStorage;
          
            // populate
            mProfileAssignmentStorage.addAllowedGroups(Collections.singleton(group1), test1, profile1);
            mProfileAssignmentStorage.addAllowedGroups(Collections.singleton(group2), test1, profile2);
            mProfileAssignmentStorage.addAllowedGroups(Collections.singleton(group1), test2, profile1);
            
            mProfileAssignmentStorage.addDeniedGroups(Collections.singleton(group2), test1, profile1);
            mProfileAssignmentStorage.addDeniedGroups(Collections.singleton(group1), test1, profile2);
            mProfileAssignmentStorage.addDeniedGroups(Collections.singleton(group2), test2, profile1);
            
            // test removing
            mProfileAssignmentStorage.removeGroup(group1);
            
            Assert.assertEquals(Collections.EMPTY_SET, _getAllowedGroups(test1, profile1));
            Assert.assertEquals(Collections.singleton(group2), _getAllowedGroups(test1, profile2));
            Assert.assertEquals(Collections.EMPTY_SET, _getAllowedGroups(test2, profile1));
            
            Assert.assertEquals(Collections.singleton(group2), _getDeniedGroups(test1, profile1));
            Assert.assertEquals(Collections.EMPTY_SET, _getDeniedGroups(test1, profile2));
            Assert.assertEquals(Collections.singleton(group2), _getDeniedGroups(test2, profile1));
        }
    }
    
    /**
     * Tests the inheritance
     */
    @Test
    public void testInheritance()
    {
        Object test1 = _getTest1();
        Object test2 = _getTest2();
        
        Assert.assertFalse(_profileAssignmentStorage.isInheritanceDisallowed(test1));
        Assert.assertFalse(_profileAssignmentStorage.isInheritanceDisallowed(test2));
        
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage)
        {
            ((ModifiableProfileAssignmentStorage) _profileAssignmentStorage).disallowInheritance(test1, true);
            ((ModifiableProfileAssignmentStorage) _profileAssignmentStorage).disallowInheritance(test2, false);
            
            Assert.assertTrue(_profileAssignmentStorage.isInheritanceDisallowed(test1));
            Assert.assertFalse(_profileAssignmentStorage.isInheritanceDisallowed(test2));
            
            // Test cache
            ((ModifiableProfileAssignmentStorage) _profileAssignmentStorage).disallowInheritance(test2, true);
            Assert.assertTrue(_profileAssignmentStorage.isInheritanceDisallowed(test2));
            
        }
    }
    
    /**
     * Test the getAllProfilesForXXX methods
     */
    @Test
    public void testGetAllProfiles()
    {
        if (_profileAssignmentStorage instanceof ModifiableProfileAssignmentStorage mPAS)
        {
            Object context1 = _getTest1();
            Object context2 = _getTest2();
            
            String profile1 = "1";
            String profile2 = "2";
            
            UserIdentity user1 = new UserIdentity("user1", "foo");
            UserIdentity user2 = new UserIdentity("user2", "bar");
            
            GroupIdentity group1 = new GroupIdentity("group1", "foo");
            GroupIdentity group2 = new GroupIdentity("group2", "bar");
            
            Map<Object, Map<AnonymousOrAnyConnectedKeys, Set<String>>> anonymousOrAnyExpectedResult = new HashMap<>();
            Map<AnonymousOrAnyConnectedKeys, Set<String>> context1Result = new HashMap<>();
            anonymousOrAnyExpectedResult.put(context1, context1Result);
            Map<AnonymousOrAnyConnectedKeys, Set<String>> context2Result = new HashMap<>();
            anonymousOrAnyExpectedResult.put(context2, context2Result);
            
            // anonymous
            mPAS.addAllowedProfilesForAnonymous(context1, Set.of(profile1));
            context1Result.put(AnonymousOrAnyConnectedKeys.ANONYMOUS_ALLOWED, Set.of(profile1));
            mPAS.addDeniedProfilesForAnonymous(context1, Set.of(profile2));
            context1Result.put(AnonymousOrAnyConnectedKeys.ANONYMOUS_DENIED, Set.of(profile2));
            mPAS.addAllowedProfilesForAnonymous(context2, Set.of(profile2));
            context2Result.put(AnonymousOrAnyConnectedKeys.ANONYMOUS_ALLOWED, Set.of(profile2));
            mPAS.addDeniedProfilesForAnonymous(context2, Set.of(profile1));
            context2Result.put(AnonymousOrAnyConnectedKeys.ANONYMOUS_DENIED, Set.of(profile1));
            
            
            // any connected
            mPAS.addDeniedProfilesForAnyConnectedUser(context1, Set.of(profile1));
            context1Result.put(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_DENIED, Set.of(profile1));
            mPAS.addAllowedProfilesForAnyConnectedUser(context1, Set.of(profile2));
            context1Result.put(AnonymousOrAnyConnectedKeys.ANYCONNECTEDUSER_ALLOWED, Set.of(profile2));
            
            Map<Object, Map<AnonymousOrAnyConnectedKeys, Set<String>>> profilesForAnonymousAndAnyConnectedUser = mPAS.getAllProfilesForAnonymousAndAnyConnectedUser();
            Assert.assertEquals(profilesForAnonymousAndAnyConnectedUser, anonymousOrAnyExpectedResult);
            
            // groups
            // We add to two groups but will request only for one to check that the result for group 2 are ignored
            Map<Object, Map<GroupIdentity, Map<UserOrGroup, Set<String>>>> groupExpectedResult = new HashMap<>();
            Map<UserOrGroup, Set<String>> context1Group1Result = new HashMap<>();
            groupExpectedResult.put(context1, Map.of(group1, context1Group1Result));
            mPAS.addAllowedGroups(Set.of(group1, group2), context1, profile1);
            context1Group1Result.put(UserOrGroup.ALLOWED, Set.of(profile1));
            
            Map<UserOrGroup, Set<String>> context2Group1Result = new HashMap<>();
            groupExpectedResult.put(context2, Map.of(group1, context2Group1Result));
            mPAS.addDeniedGroups(Set.of(group1, group2), context2, profile1);
            mPAS.addDeniedGroups(Set.of(group1, group2), context2, profile2);
            context2Group1Result.put(UserOrGroup.DENIED, Set.of(profile1, profile2));
            
            Map<Object, Map<GroupIdentity, Map<UserOrGroup, Set<String>>>> profilesForGroups = mPAS.getAllProfilesForGroups(Set.of(group1));
            Assert.assertEquals(profilesForGroups, groupExpectedResult);
            
            // users
            Map<Object, Map<UserOrGroup, Set<String>>> userExpectedResult = new HashMap<>();
            Map<UserOrGroup, Set<String>> context1UserResult = new HashMap<>();
            userExpectedResult.put(context1, context1UserResult);
            Map<UserOrGroup, Set<String>> context2UserResult = new HashMap<>();
            userExpectedResult.put(context2, context2UserResult);
            
            mPAS.addAllowedUsers(Set.of(user1, user2), context1, profile1);
            mPAS.addAllowedUsers(Set.of(user1, user2), context1, profile2);
            context1UserResult.put(UserOrGroup.ALLOWED, Set.of(profile2, profile1));
            
            mPAS.addDeniedUsers(Set.of(user1), context2, profile1);
            context2UserResult.put(UserOrGroup.DENIED, Set.of(profile1));
            mPAS.addAllowedUsers(Set.of(user1), context2, profile2);
            context2UserResult.put(UserOrGroup.ALLOWED, Set.of(profile2));
            
            Map<Object, Map<UserOrGroup, Set<String>>> profilesForUser = mPAS.getAllProfilesForUser(user1);
            Assert.assertEquals(profilesForUser, userExpectedResult);
        }
    }
}
