/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.time.Duration;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicLong;

import org.junit.Test;

import org.ametys.runtime.test.framework.AbstractBaseTestCase;

import com.google.common.base.Ticker;

/**
 * Test of {@link CachedValue}
 */
public class CachedValueTestCase extends AbstractBaseTestCase
{
    private TestTicker _ticker = new TestTicker();
    
    /**
     * Tests {@link CachedValue#get}
     */
    @Test
    public void testGet()
    {
        CachedValue<String> cachedFoo = CachedValue.withInitial(() -> "foo");
        CachedValue<String> cachedBar = CachedValue.withInitial(() -> "bar");
        assertEquals("foo", cachedFoo.get());
        assertEquals("foo", cachedFoo.get());
        assertEquals("bar", cachedBar.get());
        assertEquals("foo", cachedFoo.get());
        assertEquals("bar", cachedBar.get());
        assertEquals("bar", cachedBar.get());
        
        // Test it is cached
        boolean[] calledOnce = new boolean[] {false};
        CachedValue<Integer> cachedInt = CachedValue.withInitial(() ->
        {
            if (calledOnce[0])
            {
                return -1;
            }
            calledOnce[0] = true;
            return 42;
        });
        
        assertEquals(42, cachedInt.get().intValue());
        String errorMsg = "Getting -1 because it was called twice! The given supplier should be called only once !";
        assertEquals(errorMsg, 42, cachedInt.get().intValue());
        assertEquals(errorMsg, 42, cachedInt.get().intValue());
        assertEquals(errorMsg, 42, cachedInt.get().intValue());
    }
    
    /**
     * Tests {@link CachedValue#get} will fail with clear explanation when the supplier returns null, which is invalid
     */
    @Test
    public void testNullFailsProperly()
    {
        CachedValue<String> cachedNull = CachedValue.withInitial(() -> null);
        try
        {
            cachedNull.get();
            fail("Calling CachedValue#get() while the supplier returns null should fail.");
        }
        catch (Exception e)
        {
            _logger.info("As expected, cachedNull.get() fails. The thrown exception is: ", e);
            assertTrue(e.getMessage().contains("cannot return a null value"));
        }
    }
    
    /**
     * Tests the expiry duration works
     */
    @Test
    public void testExpiryDuration()
    {
        Duration duration = Duration.ofHours(8); // for instance, a value cached for 8 hours
        int[] callCounter = new int[] {0};
        CachedValue<String> cachedFoo = new CachedValue<>(() ->
        {
            String result;
            switch (callCounter[0])
            {
                case 0:
                    result = "foo";
                    break;
                case 1:
                    result = "fooagain";
                    break;
                case 2:
                    result = "fooagainagain";
                    break;
                default:
                    result = "ERROR";
                    break;
            }
            callCounter[0] = callCounter[0] + 1;
            return result;
        }, duration, _ticker);
        
        assertEquals("foo", cachedFoo.get());
        assertEquals("foo", cachedFoo.get());
        _ticker.advance(4, TimeUnit.HOURS);
        assertEquals("foo", cachedFoo.get());
        _ticker.advance(3, TimeUnit.HOURS);
        assertEquals("foo", cachedFoo.get());
        
        // ...
        _ticker.advance(1, TimeUnit.HOURS);
        _ticker.advance(1, TimeUnit.NANOSECONDS);
        // Tick ! a total of 8 hours elapsed
        
        assertEquals("fooagain", cachedFoo.get());
        assertEquals("fooagain", cachedFoo.get());
        _ticker.advance(2, TimeUnit.HOURS);
        assertEquals("fooagain", cachedFoo.get());
        _ticker.advance(5, TimeUnit.HOURS);
        _ticker.advance(55, TimeUnit.MINUTES); // 7h55 elapsed
        assertEquals("fooagain", cachedFoo.get());
        
        // ...
        _ticker.advance(5, TimeUnit.MINUTES);
        _ticker.advance(1, TimeUnit.NANOSECONDS);
        // Tick ! a total of 8 hours elapsed
        
        assertEquals("fooagainagain", cachedFoo.get());
        assertEquals("fooagainagain", cachedFoo.get());
        _ticker.advance(7, TimeUnit.HOURS);
        assertEquals("fooagainagain", cachedFoo.get());
    }
    
    private static class TestTicker extends Ticker
    {
        private final AtomicLong _nanos = new AtomicLong();
        
        void advance(long time, TimeUnit timeUnit)
        {
            _nanos.addAndGet(timeUnit.toNanos(time));
        }
        
        @Override
        public long read()
        {
            return _nanos.getAcquire();
        }
    }
    
    /**
     * Tests {@link CachedValue#uncache}
     */
    @Test
    public void testUncache()
    {
        int[] callCounter = new int[] {0};
        CachedValue<String> cachedFoo = CachedValue.withInitial(() ->
        {
            String result;
            switch (callCounter[0])
            {
                case 0:
                    result = "foo";
                    break;
                case 1:
                    result = "fooagain";
                    break;
                default:
                    result = "ERROR";
                    break;
            }
            callCounter[0] = callCounter[0] + 1;
            return result;
        });
        
        assertEquals("foo", cachedFoo.get());
        assertEquals("foo", cachedFoo.get());
        assertEquals("foo", cachedFoo.get());
        cachedFoo.uncache();
        assertEquals("fooagain", cachedFoo.get());
        assertEquals("fooagain", cachedFoo.get());
        assertEquals("fooagain", cachedFoo.get());
    }
}
