/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import java.time.LocalDate;
import java.time.Month;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Optional;

import org.junit.Assert;
import org.junit.Test;

import org.ametys.runtime.test.framework.AbstractBaseTestCase;

/**
 * Test cases for {@link DateUtils} helper
 */
public class DateUtilsTestCase  extends AbstractBaseTestCase
{
    /**
     * Tests the conversion from a {@link Calendar} to a {@link LocalDate} or a {@link ZonedDateTime} with a date before 1910 in the 'Europe/Paris' timezone.<br>
     * There is a gap of 9 minutes and 21 seconds on date before 1910, so this test checks that this gap does not interfere in the conversions
     */
    @Test
    public void testCalendarToDateConversionsWithDateBefore1910()
    {
        // Set a date before 1910. Choose a time near the end of the day to check that the 9 minutes gap does not interfere in the conversion 
        ZonedDateTime dateBefore1910 = ZonedDateTime.of(1890, 02, 25, 0, 5, 18, 0, ZoneId.of("Europe/Paris"));
        
        // tests the conversion between Calendar and ZonedDateTime. Internally, instants are converted to UTC. 
        Calendar calendar = DateUtils.asCalendar(dateBefore1910);
        ZonedDateTime zonedDateTime = DateUtils.asZonedDateTime(calendar);
        zonedDateTime = zonedDateTime.withZoneSameInstant(ZoneId.of("Europe/Paris"));
        Assert.assertEquals(1890, zonedDateTime.getYear());
        Assert.assertEquals(Month.FEBRUARY, zonedDateTime.getMonth());
        Assert.assertEquals(25, zonedDateTime.getDayOfMonth());
        Assert.assertEquals(0, zonedDateTime.getHour());
        Assert.assertEquals(5, zonedDateTime.getMinute());
        Assert.assertEquals(18, zonedDateTime.getSecond());
        
        // tests the conversion to a date. time components are ignored.
        GregorianCalendar cal = GregorianCalendar.from(dateBefore1910);
        LocalDate localDate = DateUtils.asLocalDate(cal);
        Assert.assertEquals(1890, localDate.getYear());
        Assert.assertEquals(Month.FEBRUARY, localDate.getMonth());
        Assert.assertEquals(25, localDate.getDayOfMonth());
    }
    
    /**
     * Tests the conversion from a {@link String} to a {@link LocalDate} with a {@link DateTimeFormatter} using time
     */
    @Test
    public void testParseLocalDateWithDatetimeFormatter()
    {
        // Create date as string
        String dateTimeAsString = "2018-05-18T14:15:31.000+02:00";
        
        // Try to parse as LocalDate
        DateTimeFormatter formatter = DateUtils.getISODateTimeFormatter();
        LocalDate localDate = DateUtils.parseLocalDate(dateTimeAsString, Optional.of(formatter));
        
        // Check the converted local date
        Assert.assertEquals(18, localDate.getDayOfMonth());
        Assert.assertEquals(Month.MAY, localDate.getMonth());
        Assert.assertEquals(2018, localDate.getYear());
    }
}
