/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util;

import static org.junit.Assert.assertEquals;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.function.BinaryOperator;

import org.junit.Test;

import org.ametys.runtime.test.framework.AbstractBaseTestCase;

/**
 * Test of {@link MapUtils}
 */
public class MapUtilsTestCase extends AbstractBaseTestCase
{
    /**
     * Tests {@link MapUtils#merge(Map, Map)}
     */
    @Test
    public void testMergeTwoMaps()
    {
        assertEquals(
                Collections.EMPTY_MAP, 
                MapUtils.merge(Map.of(), Map.of()));
        assertEquals(
                Map.of("a", "b"), 
                MapUtils.merge(Map.of("a", "b"), Map.of()));
        assertEquals(
                Map.of("a", "b"), 
                MapUtils.merge(Map.of(), Map.of("a", "b")));
        assertEquals(
                Map.of("a", "b", "c", "d"), 
                MapUtils.merge(Map.of("a", "b"), Map.of("c", "d")));
        assertEquals(
                Map.of("a", "b"), 
                MapUtils.merge(Map.of("a", "b"), Map.of("a", "b")));
        assertEquals(
                Map.of("a", "b"), 
                MapUtils.merge(Map.of("a", "b"), Map.of("a", "b2")));
        assertEquals(
                Map.of("a", "b", "c", "d"), 
                MapUtils.merge(Map.of("a", "b"), Map.of("a", "b2", "c", "d")));
    }
    
    /**
     * Tests {@link MapUtils#merge(BinaryOperator, Map, Map)}
     */
    @Test
    public void testMergeTwoMapsWithMergeFunction()
    {
        BinaryOperator<String> takeSecondValue = (s1, s2) -> s2;
        assertEquals(
                Map.of("a", "b2", "c", "d"), 
                MapUtils.merge(takeSecondValue, Map.of("a", "b"), Map.of("a", "b2", "c", "d")));
    }
    
    /**
     * Tests {@link MapUtils#merge(Map, Map, Map...)}
     */
    @Test
    public void testMergeSeveralMaps()
    {
        assertEquals(
                Map.of("a", "b1", "c", "d2", "e", "f3"), 
                MapUtils.merge(Map.of("a", "b1"), Map.of("a", "b2", "c", "d2"), Map.of("c", "d3", "e", "f3", "a", "b3")));
    }
    
    /**
     * Tests {@link MapUtils#merge(BinaryOperator, Map, Map, Map...)}
     */
    @Test
    public void testMergeSeveralMapsWithMergeFunction()
    {
        BinaryOperator<String> takeSecondValue = (s1, s2) -> s2;
        assertEquals(
                Map.of("a", "b3", "c", "d3", "e", "f3"), 
                MapUtils.merge(takeSecondValue, Map.of("a", "b1"), Map.of("a", "b2", "c", "d2"), Map.of("c", "d3", "e", "f3", "a", "b3")));
    }

    /**
     * Tests {@link MapUtils#deepMerge(Object, Object)}
     */
    @Test
    public void testDeepMergeFunction()
    {
        Map<String, Object> map1 = new HashMap<>();
        map1.put("1", "map1.v1");
        map1.put("2", null);
        map1.put("3", "map1.v3");
        map1.put("5", "map1.v5");
        Map<String, Object> subMap1_4 = new HashMap<>();
        subMap1_4.put("4.1", "map1.v4.1");
        subMap1_4.put("4.2", "map1.v4.2");
        subMap1_4.put("4.3", null);
        Map<String, Object> subMap1_4_4 = new HashMap<>();
        subMap1_4_4.put("4.4.1", "map1.v4.4.1");
        subMap1_4_4.put("4.4.2", null);
        subMap1_4_4.put("4.4.3", null);
        subMap1_4.put("4.4", subMap1_4_4);
        map1.put("4", subMap1_4);
        map1.put("6", Map.of("6.1", "map1.v6.1"));
        map1.put("7", "map1.v7");
        map1.put("10", null);

        Map<String, Object> map2 = new HashMap<>();
        map2.put("1", "map2.v1");
        map2.put("2", "map2.v2");
        map2.put("3", null);
        Map<String, Object> subMap2_4 = new HashMap<>();
        subMap2_4.put("4.1", "map2.v4.1");
        Map<String, Object> subMap2_4_4 = new HashMap<>();
        subMap2_4_4.put("4.4.2", "map2.v4.4.2");
        subMap2_4.put("4.4", subMap2_4_4);
        subMap2_4.put("4.5", "map2.v4.5");
        map2.put("4", subMap2_4);
        map2.put("5", Map.of("5.1", "map2.v5.1"));
        map2.put("6", "map2.v6.1");
        map2.put("8", Map.of("8.1", "map2.v8.1"));
        map2.put("9", null);
        
        @SuppressWarnings("unchecked")
        Map<String, Object> mergedMap = (Map<String, Object>) MapUtils.deepMerge(map1, map2);
        
        Map<String, Object> expectedMap = new HashMap<>();
        expectedMap.put("1", "map1.v1");
        expectedMap.put("2", "map2.v2");
        expectedMap.put("3", "map1.v3");
        Map<String, Object> subExpectedMap_4 = new HashMap<>();
        subExpectedMap_4.put("4.1", "map1.v4.1");
        subExpectedMap_4.put("4.2", "map1.v4.2");
        subExpectedMap_4.put("4.3", null);
        Map<String, Object> subExpectedMap_4_4 = new HashMap<>();
        subExpectedMap_4_4.put("4.4.1", "map1.v4.4.1");
        subExpectedMap_4_4.put("4.4.2", "map2.v4.4.2");
        subExpectedMap_4_4.put("4.4.3", null);
        subExpectedMap_4.put("4.4", subExpectedMap_4_4);
        subExpectedMap_4.put("4.5", "map2.v4.5");
        expectedMap.put("4", subExpectedMap_4);
        expectedMap.put("5", "map1.v5");
        expectedMap.put("6", Map.of("6.1", "map1.v6.1"));
        expectedMap.put("7", "map1.v7");
        expectedMap.put("8", Map.of("8.1", "map2.v8.1"));
        expectedMap.put("9", null);
        expectedMap.put("10", null);
        
        assertEquals(expectedMap, mergedMap);
    }
}
