/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.core.util.dom;

import java.util.List;
import java.util.Map;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.cocoon.xml.AttributesImpl;
import org.apache.cocoon.xml.XMLUtils;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.ametys.runtime.test.framework.AbstractBaseTestCase;

/**
 * Test of {@link DOMUtils}
 */
public class DOMUtilsTestCase extends AbstractBaseTestCase
{
    private static Element _element;
    
    /**
     * Prepares all needed stuffs to the {@link DOMUtils} test cases
     * @throws Exception if an error occurs
     */
    @BeforeClass
    public static void prepareTestCase() throws Exception
    {
        _element = _createDOMElement();
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElements(Element)} method
     */
    @Test
    public void testGetChildElements()
    {
        List<Element> childElements = DOMUtils.getChildElements(_element);
        
        Assert.assertEquals(5, childElements.size());
        Assert.assertTrue("The child elements should contain an element named childNode1",
                _containsElement(childElements, "childNode1"));
        Assert.assertTrue("The child elements should contain an element named childNode2 with attribute position '1'",
                _containsElement(childElements, "childNode2", Map.of("position", "1")));
        Assert.assertTrue("The child elements should contain an element named childNode2 with attribute position '2'",
                _containsElement(childElements, "childNode2", Map.of("position", "2")));
        Assert.assertTrue("The child elements should contain an element named childNode2 with attribute position '3'",
                _containsElement(childElements, "childNode2", Map.of("position", "3")));
        Assert.assertTrue("The child elements should contain an element named childNode3",
                _containsElement(childElements, "childNode3"));
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElements(Element)} method on an element without children
     */
    @Test
    public void testGetChildElementsOnElementWithoutChildren()
    {
        NodeList nodesWithoutChildren = _element.getElementsByTagName("childNode3");
        Element elementWithoutChildren = (Element) nodesWithoutChildren.item(0);
        
        List<Element> childElements = DOMUtils.getChildElements(elementWithoutChildren);
        
        Assert.assertTrue("The element 'childNode3' shouldn't have children", childElements.isEmpty());
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElements(Element)} method on a <code>null</code> element
     */
    @Test
    public void testGetChildElementsOnNullElement()
    {
        try
        {
            DOMUtils.getChildElements(null);
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#getFirstChildElement(Element)} method
     */
    @Test
    public void testGetFirstChildElement()
    {
        Element firstChildElement = DOMUtils.getFirstChildElement(_element);
        
        Assert.assertEquals("childNode1", firstChildElement.getNodeName());
    }
    
    /**
     * Tests the {@link DOMUtils#getFirstChildElement(Element)} method on an element without children
     */
    @Test
    public void testGetFirstChildElementOnElementWithoutChildren()
    {
        NodeList nodesWithoutChildren = _element.getElementsByTagName("childNode3");
        Element elementWithoutChildren = (Element) nodesWithoutChildren.item(0);
        
        Assert.assertNull("The element 'childNode3' shouldn't have children", DOMUtils.getFirstChildElement(elementWithoutChildren));
    }
    
    /**
     * Tests the {@link DOMUtils#getFirstChildElement(Element)} method on a <code>null</code> element
     */
    @Test
    public void testGetFirstChildElementOnNullElement()
    {
        try
        {
            DOMUtils.getFirstChildElement(null);
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#getUniqueChildElements(Element)} method
     */
    @Test
    public void testGetUniqueChildElements()
    {
        List<Element> childElements = DOMUtils.getUniqueChildElements(_element);
        
        Assert.assertEquals(3, childElements.size());
        Assert.assertTrue("The child elements should contain an element named childNode1",
                _containsElement(childElements, "childNode1"));
        Assert.assertTrue("The child elements should contain an element named childNode2 with attribute position '1'",
                _containsElement(childElements, "childNode2", Map.of("position", "1")));
        Assert.assertTrue("The child elements should contain an element named childNode3",
                _containsElement(childElements, "childNode3"));
    }
    
    /**
     * Tests the {@link DOMUtils#getUniqueChildElements(Element)} method on an element with only different children
     */
    @Test
    public void testGetUniqueChildElementsOnElementWithOnlyDifferentChildren()
    {
        NodeList nodesWithOnlyDifferentChildren = _element.getElementsByTagName("childNode1");
        Element elementWithOnlyDifferentChildren = (Element) nodesWithOnlyDifferentChildren.item(0);
        List<Element> childElements = DOMUtils.getUniqueChildElements(elementWithOnlyDifferentChildren);
        
        Assert.assertEquals(3, childElements.size());
        Assert.assertTrue("The child elements should contain an element named subChildNode1",
                _containsElement(childElements, "subChildNode1"));
        Assert.assertTrue("The child elements should contain an element named subChildNode2",
                _containsElement(childElements, "subChildNode2"));
        Assert.assertTrue("The child elements should contain an element named subChildNode3",
                _containsElement(childElements, "subChildNode3"));
    }
    
    /**
     * Tests the {@link DOMUtils#getUniqueChildElements(Element)} method on an element without children
     */
    @Test
    public void testGetUniqueChildElementsOnElementWithoutChildren()
    {
        NodeList nodesWithoutChildren = _element.getElementsByTagName("childNode3");
        Element elementWithoutChildren = (Element) nodesWithoutChildren.item(0);
        
        List<Element> childElements = DOMUtils.getUniqueChildElements(elementWithoutChildren);
        
        Assert.assertTrue("The element 'childNode3' shouldn't have children", childElements.isEmpty());
    }
    
    /**
     * Tests the {@link DOMUtils#getUniqueChildElements(Element)} method on a <code>null</code> element
     */
    @Test
    public void testGetUniqueChildElementsOnNullElement()
    {
        try
        {
            DOMUtils.getUniqueChildElements(null);
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElementsByTagName(Element, String)} method
     */
    @Test
    public void testGetChildElementsByTagName()
    {
        List<Element> childElements = DOMUtils.getChildElementsByTagName(_element, "childNode2");
        
        Assert.assertEquals(3, childElements.size());
        Assert.assertTrue("The child elements should contain an element named childNode2 with attribute position '1'",
                _containsElement(childElements, "childNode2", Map.of("position", "1")));
        Assert.assertTrue("The child elements should contain an element named childNode2 with attribute position '2'",
                _containsElement(childElements, "childNode2", Map.of("position", "2")));
        Assert.assertTrue("The child elements should contain an element named childNode2 with attribute position '3'",
                _containsElement(childElements, "childNode2", Map.of("position", "3")));
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElementsByTagName(Element, String)} method with an unexisting tag name
     */
    @Test
    public void testGetChildElementsByTagNameWithUnexistingTagName()
    {
        List<Element> childElements = DOMUtils.getChildElementsByTagName(_element, "childNode4");
        
        Assert.assertTrue("The element shouldn't have children named 'childNode4", childElements.isEmpty());
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElementsByTagName(Element, String)} method on a <code>null</code> element
     */
    @Test
    public void testGetChildElementsByTagNameOnNullElement()
    {
        try
        {
            DOMUtils.getChildElementsByTagName(null, "childNode1");
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElementsByTagName(Element, String)} method with a <code>null</code> tag name
     */
    @Test
    public void testGetChildElementsByTagNameWithNullTagName()
    {
        try
        {
            DOMUtils.getChildElementsByTagName(_element, null);
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElementsByTagName(Element, String)} method with an empty tag name
     */
    @Test
    public void testGetChildElementsByTagNameWithEmptyTagName()
    {
        try
        {
            DOMUtils.getChildElementsByTagName(_element, "");
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElementByTagName(Element, String)} method
     */
    @Test
    public void testGetChildElementByTagName()
    {
        Element childElement = DOMUtils.getChildElementByTagName(_element, "childNode2");
        
        Assert.assertEquals("childNode2", childElement.getNodeName());
        Assert.assertEquals("1", childElement.getAttribute("position"));
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElementByTagName(Element, String)} method with an unexisting tag name
     */
    @Test
    public void testGetChildElementByTagNameWithUnexistingTagName()
    {
        Assert.assertNull("The element shouldn't have a child named 'childNode4", DOMUtils.getChildElementByTagName(_element, "childNode4"));
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElementByTagName(Element, String)} method on a <code>null</code> element
     */
    @Test
    public void testGetChildElementByTagNameOnNullElement()
    {
        try
        {
            DOMUtils.getChildElementByTagName(null, "childNode1");
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElementByTagName(Element, String)} method with a <code>null</code> tag name
     */
    @Test
    public void testGetChildElementByTagNameWithNullTagName()
    {
        try
        {
            DOMUtils.getChildElementByTagName(_element, null);
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#getChildElementByTagName(Element, String)} method with an empty tag name
     */
    @Test
    public void testGetChildElementByTagNameWithEmptyTagName()
    {
        try
        {
            DOMUtils.getChildElementByTagName(_element, "");
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#hasChildElement(Element, String)} method
     */
    @Test
    public void testHasChildElement()
    {
        Assert.assertTrue("The element should have a child named 'childNode1'", DOMUtils.hasChildElement(_element, "childNode1"));
    }
    
    /**
     * Tests the {@link DOMUtils#hasChildElement(Element, String)} method with an unexisting tag name
     */
    @Test
    public void testHasChildElementWithUnexistingTagName()
    {
        Assert.assertFalse("The element shouldn't have a child named 'childNode4'", DOMUtils.hasChildElement(_element, "childNode4"));
    }
    
    /**
     * Tests the {@link DOMUtils#hasChildElement(Element, String)} method on a <code>null</code> element
     */
    @Test
    public void testHasChildElementOnNullElement()
    {
        try
        {
            DOMUtils.hasChildElement(null, "childNode1");
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#hasChildElement(Element, String)} method with a <code>null</code> tag name
     */
    @Test
    public void testHasChildElementWithNullTagName()
    {
        try
        {
            DOMUtils.hasChildElement(_element, null);
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    /**
     * Tests the {@link DOMUtils#hasChildElement(Element, String)} method with an empty tag name
     */
    @Test
    public void testHasChildElementWithEmptyTagName()
    {
        try
        {
            DOMUtils.hasChildElement(_element, "");
            Assert.fail();
        }
        catch (IllegalArgumentException e)
        {
            // Expected
        }
    }
    
    private boolean _containsElement(List<Element> elements, String tagName)
    {
        return _containsElement(elements, tagName, Map.of());
    }
    
    private boolean _containsElement(List<Element> elements, String tagName, Map<String, String> expectedAttributes)
    {
        for (Element element : elements)
        {
            if (tagName.equals(element.getNodeName()))
            {
                if (expectedAttributes.isEmpty())
                {
                    return true;
                }
                else
                {
                    for (Map.Entry<String, String> expectedAttribute : expectedAttributes.entrySet())
                    {
                        String attributeName = expectedAttribute.getKey();
                        String attributeValue = expectedAttribute.getValue();
                        if (element.hasAttribute(attributeName))
                        {
                            if (attributeValue.equals(element.getAttribute(attributeName)))
                            {
                                return true;
                            }
                        }
                    }
                }
            }
        }
        return false;
    }
    
    /**
     * Creates the DOM element used to test the {@link DOMUtils} methods 
     * @throws Exception if an error occurs
     */
    private static Element _createDOMElement() throws Exception
    {
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        
        AttributesImpl elementAttributes = new AttributesImpl();
        elementAttributes.addCDATAAttribute("attribute", "value");
        XMLUtils.startElement(handler, "element", elementAttributes);
        
        XMLUtils.startElement(handler, "childNode1");
        XMLUtils.createElement(handler, "subChildNode1");
        XMLUtils.createElement(handler, "subChildNode2");
        XMLUtils.createElement(handler, "subChildNode3");
        XMLUtils.endElement(handler, "childNode1");
        
        AttributesImpl position1Attributes = new AttributesImpl();
        position1Attributes.addCDATAAttribute("position", "1");
        XMLUtils.createElement(handler, "childNode2", position1Attributes);
        
        AttributesImpl position2Attributes = new AttributesImpl();
        position2Attributes.addCDATAAttribute("position", "2");
        XMLUtils.createElement(handler, "childNode2", position2Attributes);
        
        AttributesImpl position3Attributes = new AttributesImpl();
        position3Attributes.addCDATAAttribute("position", "3");
        XMLUtils.createElement(handler, "childNode2", position3Attributes);
        
        XMLUtils.createElement(handler, "childNode3");
        
        XMLUtils.endElement(handler, "element");
        handler.endDocument();
        
        Document document = (Document) result.getNode();
        
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        return (Element) document.getElementsByTagName("element").item(0);
    }
}
