/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.plugins.core.impl.model.type.xml;

import java.time.LocalDate;
import java.time.Month;
import java.util.Optional;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationUtil;
import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.xml.XMLUtils;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.ContentHandler;

import org.ametys.runtime.model.type.DataContext;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.model.type.ModelItemTypeExtensionPoint;
import org.ametys.runtime.model.type.xml.XMLElementType;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * {@link DateXMLElementType} implementation test case
 */
@NeedFeatures({
    "test/test-data-types",
})
public class DateXMLElementTypeTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static ModelItemTypeExtensionPoint _testDataTypeExtensionPoint;
    private static XMLElementType<LocalDate> _dateType;

    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp-item-types-tests");
    }
    
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _testDataTypeExtensionPoint = (ModelItemTypeExtensionPoint) manager.lookup(ModelItemTypeExtensionPoint.class.getName() + ".Test");
    }
    
    @SuppressWarnings("unchecked")
    @Override
    public void prepareComponents() throws Exception
    {
        _dateType = (XMLElementType<LocalDate>) _testDataTypeExtensionPoint.getExtension(ModelItemTypeConstants.DATE_TYPE_ID);
    }
    
    /**
     * Test the reciprocity between the methods {@link DateXMLElementType#valueToSAX(ContentHandler, String, Object, DataContext)} and {@link DateXMLElementType#valueFromXML(Element, String, Optional)}
     * @throws Exception if an error occurs
     */
    @Test
    public void testReciprocityBetweenToAndFromXML() throws Exception
    {
        LocalDate localDate = LocalDate.of(2020, Month.JUNE, 26);
        
        // Generate SAX events for the date
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        _dateType.valueToSAX(handler, "localDate", localDate, DataContext.newInstance());

        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        Document document = (Document) result.getNode();
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        LocalDate resultDate = (LocalDate) _dateType.valueFromXML(document.getDocumentElement(), "localDate", Optional.empty());
        
        Assert.assertTrue(resultDate.isEqual(localDate));
    }
    
    /**
     * Test the reciprocity between the methods {@link DateXMLElementType#toString(LocalDate)} and {@link DateXMLElementType#castValue(Object)}
     */
    @Test
    public void testReciprocityBetweenToStringAndCast()
    {
        LocalDate localDate = LocalDate.of(2020, Month.JUNE, 26);
        Assert.assertTrue(localDate.isEqual(_dateType.castValue(_dateType.toString(localDate))));
    }
    
    
    /**
     * Test the reciprocity between the methods {@link DateXMLElementType#write(ContentHandler, String, Object)} and {@link DateXMLElementType#read(Configuration, String)}
     * @throws Exception if an error occurs
     */
    @Test
    public void testReciprocityBetweenWriteAndRead() throws Exception
    {
        LocalDate localDate = LocalDate.of(2020, Month.JUNE, 26);
        
        // Generate SAX events for the date
        DOMResult result = new DOMResult();
        TransformerHandler handler = (TransformerHandler) TransformerFactory.newInstance().newTransformer();
        handler.setResult(result);
        
        handler.startDocument();
        XMLUtils.startElement(handler, "result");
        
        _dateType.write(handler, "localDate", localDate);

        XMLUtils.endElement(handler, "result");
        handler.endDocument();
        
        Document document = (Document) result.getNode();
        System.out.println("***************");
        System.out.println(XMLUtils.serializeNode(document));
        System.out.println("***************");
        
        Element element = document.getDocumentElement();
        Configuration configuration = ConfigurationUtil.toConfiguration(element);
        LocalDate resultDate = (LocalDate) _dateType.read(configuration, "localDate");
        
        Assert.assertTrue(resultDate.isEqual(localDate));
    }

    /**
     * Test the method {@link DateXMLElementType#castValue(Object)} with a date without time info
     */
    @Test
    public void testCastValueWithStrinDate()
    {
        String value = "2018-05-18";
        LocalDate date = _dateType.castValue(value);
        Assert.assertEquals(18, date.getDayOfMonth());
        Assert.assertEquals(Month.MAY, date.getMonth());
        Assert.assertEquals(2018, date.getYear());
    }
    
    /**
     * Test the method {@link DateXMLElementType#castValue(Object)} with a date with time info
     */
    @Test
    public void testCastValueWithStrinDateTime()
    {
        String value = "2018-05-18T14:15:31Z";
        LocalDate date = _dateType.castValue(value);
        Assert.assertEquals(18, date.getDayOfMonth());
        Assert.assertEquals(Month.MAY, date.getMonth());
        Assert.assertEquals(2018, date.getYear());
    }
    
    /**
     * Test the method {@link DateXMLElementType#castValue(Object)} with a date with time info, including fraction of second
     */
    @Test
    public void testCastValueWithStrinDateTimeAndFractionOfSecond()
    {
        String value = "2018-05-18T14:15:31.000+02:00";
        LocalDate date = _dateType.castValue(value);
        Assert.assertEquals(18, date.getDayOfMonth());
        Assert.assertEquals(Month.MAY, date.getMonth());
        Assert.assertEquals(2018, date.getYear());
    }
}
