/*
 *  Copyright 2024 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.config;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.disableconditions.DisableConditionsEvaluator;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * {@link ConfigDisableConditionsEvaluator}'s test cases
 */
@NeedFeatures({
    "core/runtime.captcha",
    "core-impl/runtime.captcha.jcaptcha",
    "core-impl/runtime.captcha.recaptcha"
})
public class ConfigDisableConditionsEvaluatorTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static ConfigDisableConditionsEvaluator _evaluator;
    private static ConfigManager _configManager;
    private static Config _config;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _evaluator = (ConfigDisableConditionsEvaluator) manager.lookup(ConfigDisableConditionsEvaluator.ROLE);
    }
    
    public void prepareComponents() throws Exception
    {
        CocoonTestCase.super.prepareComponents();
        
        _configManager = ConfigManager.getInstance();
        _config = Config.getInstance();
    }
    
    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * No value in map, the stored value is jcaptcha -> the clean period is not disabled
     */
    @Test
    public void testDisabledFromStoredValues()
    {
        ModelItem modelItem = _configManager.getModelItem("runtime.captcha.jcaptcha.clean.period");
        boolean disable = _evaluator.evaluateDisableConditions(modelItem, "runtime.captcha.jcaptcha.clean.period", Optional.empty(), Map.of(), _config,  new HashMap<>());
        
        Assert.assertFalse(disable);
    }
    
    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * No value in map, the stored value is jcaptcha -> the public key is disabled
     */
    @Test
    public void testEnabledFromStoredValues()
    {
        ModelItem modelItem = _configManager.getModelItem("runtime.captcha.recaptcha.publickey");
        boolean disable = _evaluator.evaluateDisableConditions(modelItem, "runtime.captcha.recaptcha.publickey", Optional.empty(), Map.of(), _config,  new HashMap<>());
        
        Assert.assertTrue(disable);
    }
    
    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The value in map is recaptcha -> the clean period is disabled
     */
    @Test
    public void testEnabledFromGivenValues()
    {
        ModelItem modelItem = _configManager.getModelItem("runtime.captcha.jcaptcha.clean.period");
        boolean disable = _evaluator.evaluateDisableConditions(modelItem, "runtime.captcha.jcaptcha.clean.period", Optional.empty(), Map.of("runtime.captcha.type", "recaptcha"), _config,  new HashMap<>());
        
        Assert.assertTrue(disable);
    }
    
    /**
     * Tests the {@link DisableConditionsEvaluator#evaluateDisableConditions(ModelItem, String, Optional, Map, Object, Map)} method
     * The value in map is recaptcha -> the public key is not disabled
     */
    @Test
    public void testDisabledFromGivenValues()
    {
        ModelItem modelItem = _configManager.getModelItem("runtime.captcha.recaptcha.publickey");
        boolean disable = _evaluator.evaluateDisableConditions(modelItem, "runtime.captcha.recaptcha.publickey", Optional.empty(), Map.of("runtime.captcha.type", "recaptcha"), _config,  new HashMap<>());
        
        Assert.assertFalse(disable);
    }
}
