/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.config;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.After;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.util.I18nUtils;
import org.ametys.runtime.model.ElementDefinition;
import org.ametys.runtime.model.ModelItem;
import org.ametys.runtime.model.ModelItemContainer;
import org.ametys.runtime.model.ModelItemGroup;
import org.ametys.runtime.model.ModelViewItem;
import org.ametys.runtime.model.ModelViewItemGroup;
import org.ametys.runtime.model.View;
import org.ametys.runtime.model.ViewItem;
import org.ametys.runtime.model.ViewItemContainer;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.plugin.PluginsManager;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.ScriptHelper;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * Tests the ConfigManager
 */
public class ConfigManagerTestCase extends AbstractRuntimeTestCase
{
    /**
     * Called after each test. Stops Cocoon.
     */
    @After
    public void afterConfigManager()
    {
        leaveEnvironment();
        stopCocoon();
    }
    
    /**
     * Test the behaviour when the config.xml file is not present
     * @throws Exception if an error occurs
     */
    @Test
    public void testConfigNotPresent() throws Exception
    {
        startCocoon(new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime01.xml")
                .withConfigFile("test/environments/configs/config0.xml")
                .withWebappPath("test/environments/webapp1"));
        
        Assert.assertTrue(PluginsManager.getInstance().isSafeMode());
        Assert.assertFalse(ConfigManager.getInstance().isComplete());
    }
    
    /**
     * Test that if the datasource is activated, the config parameters are necessary.
     * @throws Exception if an error occurs
     */
    @Test
    public void testMissingParameters() throws Exception
    {
        startCocoon(new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime01.xml")
                .withConfigFile("test/environments/configs/config2.xml")
                .withWebappPath("test/environments/webapp1"));
        
        Assert.assertTrue(PluginsManager.getInstance().isSafeMode());
        Assert.assertFalse(ConfigManager.getInstance().isComplete());
    }
    
    /**
     * Test the behavior when there are 2 switchers defined for the same group.
     */
    @Test
    public void testWith2SwitchersInAGroup()
    {
        try
        {
            startCocoon(new TestConfigSet()
                    .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime04.xml")
                    .withConfigFile("test/environments/configs/config5.xml")
                    .withWebappPath("test/environments/webapp1"));
            Assert.fail("start the application with 2 switchers in the same group should have lead to an error");
        }
        catch (Exception e)
        {
            Assert.assertTrue(e instanceof RuntimeException);
            Assert.assertTrue(e.getMessage().contains("Test Group with 2 switchers"));
            Assert.assertTrue(e.getMessage().contains("switchParam1"));
            Assert.assertTrue(e.getMessage().contains("switchParam2"));
        }
    }
    
    /**
     * Test the behavior with a non boolean switcher defined for a group.
     */
    @Test
    public void testWithANonBooleanSwitcher()
    {
        try
        {
            startCocoon(new TestConfigSet()
                    .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime05.xml")
                    .withConfigFile("test/environments/configs/config6.xml")
                    .withWebappPath("test/environments/webapp1"));
            Assert.fail("start the application with non boolean switcher should have lead to an error");
        }
        catch (Exception e)
        {
            Assert.assertTrue(e instanceof RuntimeException);
            Assert.assertTrue(e.getMessage().contains("Test Group with a non boolean switcher"));
            Assert.assertTrue(e.getMessage().contains("longSwitchParam"));
        }
    }
    
    /**
     * Test the behavior with a parameter value with the wrong type (string value for a long parameter).
     * @throws Exception if an error occurs
     */
    @Test
    public void testWithWrongValueType() throws Exception
    {
        startCocoon(new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime02.xml")
                .withConfigFile("test/environments/configs/config7.xml")
                .withWebappPath("test/environments/webapp1"));
        
        Assert.assertTrue(PluginsManager.getInstance().isSafeMode());
        Assert.assertFalse(ConfigManager.getInstance().isComplete());
    }
    
    /**
     * Test that if the datasource is activated, the config parameters are necessary.
     * @throws Exception if an error occurs
     */
    @Test
    public void testMissingLongParameter() throws Exception
    {
        startCocoon(new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime02.xml")
                .withConfigFile("test/environments/configs/config8.xml")
                .withWebappPath("test/environments/webapp1"));
        
        Assert.assertTrue(PluginsManager.getInstance().isSafeMode());
        Assert.assertFalse(ConfigManager.getInstance().isComplete());
    }

    /**
     * Test that if the datasource is unactivated, the config parameters are not necessary.
     * @throws Exception if an error occurs
     */
    @Test
    public void testUnactivation() throws Exception
    {
        startCocoon(new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime03.xml")
                .withConfigFile("test/environments/configs/config2.xml")
                .withWebappPath("test/environments/webapp1"));
        
        Assert.assertTrue(ConfigManager.getInstance().isComplete());
    }

    /**
     * Test parameters from plugins and features
     * @throws Exception if an error occurs
     */
    @Test
    public void testConfiguration() throws Exception
    {
        startCocoon(new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime02.xml")
                .withConfigFile("test/environments/configs/config1.xml")
                .withSqlFile(ScriptHelper.getDataSourceFile(ConnectionHelper.DATABASE_HSQLDB))
                .withWebappPath("test/environments/webapp1"));
        ServiceManager serviceManager = getCocoonWrapper().getServiceManager();
        enterEnvironment("");
        
        // Tests on configuration API
        Assert.assertTrue(ConfigManager.getInstance().isComplete());
        _checkModelItems();
        _checkView(serviceManager);
        _checkFlatDefinitions();
        _checkValues();
    }

    private void _checkModelItems()
    {
        ConfigManager configManager = ConfigManager.getInstance();
        
        // Category
        ModelItemGroup category = _getModelItemGroup(configManager, "Test Category");
        Assert.assertNotNull(category);
        Assert.assertEquals(3, category.getModelItems().size());
        
        // Simple Group
        ModelItemGroup simpleGroup = _getModelItemGroup(category, "Test Group");
        Assert.assertNotNull(simpleGroup);
        List<ModelItem> simpleGroupDefinitions = simpleGroup.getChildren();
        Assert.assertEquals(4, simpleGroupDefinitions.size());
        
        // Simple Group Switcher
        ModelItem booleanItem = simpleGroup.getModelItem("switchParam");
        ElementDefinition<Boolean> switcherDefinition = simpleGroup.getSwitcher();
        Assert.assertEquals(booleanItem, switcherDefinition);
        Assert.assertEquals(ModelItemTypeConstants.BOOLEAN_TYPE_ID, switcherDefinition.getType().getId());
        Assert.assertFalse(switcherDefinition.getDefaultValue());
        
        // Simple Group String parameter
        @SuppressWarnings("unchecked")
        ElementDefinition<String> stringDefinition = (ElementDefinition<String>) simpleGroup.getModelItem("stringParam");
        Assert.assertEquals(ModelItemTypeConstants.STRING_TYPE_ID, stringDefinition.getType().getId());
        Assert.assertEquals("test2", stringDefinition.getPluginName());
        Assert.assertEquals("String parameter", stringDefinition.getLabel().toString());
        Assert.assertEquals("This parameter is a string and is referenced by an used feature", stringDefinition.getDescription().toString());
        Assert.assertEquals("default value", stringDefinition.getDefaultValue());
        Assert.assertEquals("stringParam", stringDefinition.getPath());
        Assert.assertEquals(configManager, stringDefinition.getModel());
        Assert.assertEquals(simpleGroup, stringDefinition.getParent());
        
        // Simple Group long parameter
        @SuppressWarnings("unchecked")
        ElementDefinition<Long> longDefinition = (ElementDefinition<Long>) simpleGroup.getModelItem("longParam");
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, longDefinition.getType().getId());
        
        // Simple Group empty long parameter
        @SuppressWarnings("unchecked")
        ElementDefinition<Long> emptyLongDefinition = (ElementDefinition<Long>) simpleGroup.getModelItem("emptyLongParam");
        Assert.assertEquals(ModelItemTypeConstants.LONG_TYPE_ID, emptyLongDefinition.getType().getId());
        
        // Ordered Group
        ModelItemGroup orderedGroup = _getModelItemGroup(category, "Ordered Test Group");
        Assert.assertNotNull(orderedGroup);
        Assert.assertEquals(5, orderedGroup.getChildren().size());
        Assert.assertTrue(orderedGroup.hasModelItem("firstParam"));
        Assert.assertTrue(orderedGroup.hasModelItem("secondParam"));
        Assert.assertTrue(orderedGroup.hasModelItem("thirdParam"));
        Assert.assertTrue(orderedGroup.hasModelItem("yyFourthParam"));
        Assert.assertTrue(orderedGroup.hasModelItem("zzFifthParam"));
        
        // Group with a switcher without default value
        ModelItemGroup groupWithSwitcher = _getModelItemGroup(category, "Test Group with switcher without default value");
        Assert.assertNotNull(groupWithSwitcher);
        ModelItem switcherDefinitionFromChildrenList = groupWithSwitcher.getModelItem("switchParamWithoutDefaultValue");
        ElementDefinition<Boolean> switcherWithoutDefaultValueDefinition = groupWithSwitcher.getSwitcher();
        Assert.assertEquals(switcherDefinitionFromChildrenList, switcherWithoutDefaultValueDefinition);
        Assert.assertEquals(ModelItemTypeConstants.BOOLEAN_TYPE_ID, switcherWithoutDefaultValueDefinition.getType().getId());
        Assert.assertFalse(switcherWithoutDefaultValueDefinition.getDefaultValue());
    }

    private ModelItemGroup _getModelItemGroup(ModelItemContainer container, String label)
    {
        for (ModelItem item : container.getModelItems())
        {
            if (label.equals(item.getLabel().toString()) && item instanceof ModelItemGroup)
            {
                return (ModelItemGroup) item;
            }
        }
        
        return null;
    }

    private void _checkView(ServiceManager serviceManager)
    {
        ConfigManager configManager = ConfigManager.getInstance();
        View view = null;
        
        try
        {
            I18nUtils i18nUtils = (I18nUtils) serviceManager.lookup(I18nUtils.ROLE);
            view = configManager.getView(i18nUtils);
        }
        catch (ServiceException e)
        {
            Assert.fail("Lookup to I18nUtils failed, unable to check the configuration's view");
        }
        
        // Category
        ModelViewItemGroup category = _getViewItemGroup(view, "Test Category");
        Assert.assertNotNull(category);
        List<ViewItem> groups = category.getViewItems();
        Assert.assertEquals(3, groups.size());
        
        // Ordered Group
        ModelViewItemGroup orderedGroup = (ModelViewItemGroup) groups.get(0);
        Assert.assertEquals("Ordered Test Group", orderedGroup.getDefinition().getLabel().toString());
        List<ViewItem> orderedGroupViewItems = orderedGroup.getViewItems();
        Assert.assertEquals(5, orderedGroupViewItems.size());
        Assert.assertEquals("firstParam", ((ModelViewItem) orderedGroupViewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("secondParam", ((ModelViewItem) orderedGroupViewItems.get(1)).getDefinition().getName());
        Assert.assertEquals("thirdParam", ((ModelViewItem) orderedGroupViewItems.get(2)).getDefinition().getName());
        Assert.assertEquals("yyFourthParam", ((ModelViewItem) orderedGroupViewItems.get(3)).getDefinition().getName());
        Assert.assertEquals("zzFifthParam", ((ModelViewItem) orderedGroupViewItems.get(4)).getDefinition().getName());
        
        // Simple Group
        ModelViewItemGroup simpleGroup = (ModelViewItemGroup) groups.get(1);
        Assert.assertEquals("Test Group", simpleGroup.getDefinition().getLabel().toString());
        List<ViewItem> simpleGroupViewItems = simpleGroup.getViewItems();
        Assert.assertEquals(4, simpleGroupViewItems.size());
        Assert.assertEquals("switchParam", ((ModelViewItem) simpleGroupViewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("stringParam", ((ModelViewItem) simpleGroupViewItems.get(1)).getDefinition().getName());
        Assert.assertEquals("longParam", ((ModelViewItem) simpleGroupViewItems.get(2)).getDefinition().getName());
        Assert.assertEquals("emptyLongParam", ((ModelViewItem) simpleGroupViewItems.get(3)).getDefinition().getName());
        
        // Group with a switcher without default value
        ModelViewItemGroup groupWithSwitcher = (ModelViewItemGroup) groups.get(2);
        Assert.assertEquals("Test Group with switcher without default value", groupWithSwitcher.getDefinition().getLabel().toString());
        List<ViewItem> groupWithSwitcherViewItems = groupWithSwitcher.getViewItems();
        Assert.assertEquals(1, groupWithSwitcherViewItems.size());
        Assert.assertEquals("switchParamWithoutDefaultValue", ((ModelViewItem) groupWithSwitcherViewItems.get(0)).getDefinition().getName());
    }

    private ModelViewItemGroup _getViewItemGroup(ViewItemContainer container, String label)
    {
        for (ViewItem item : container.getViewItems())
        {
            if (item instanceof ModelViewItemGroup && label.equals(((ModelViewItemGroup) item).getDefinition().getLabel().toString()))
            {
                return (ModelViewItemGroup) item;
            }
        }
        
        return null;
    }
    
    private void _checkFlatDefinitions()
    {
        ConfigManager configManager = ConfigManager.getInstance();
        Collection<ElementDefinition> flatDefinitions = configManager.getConfigurationParameters();
        
        ElementDefinition<Boolean> switchParam = _getElementDefinition(flatDefinitions, "switchParam");
        Assert.assertNotNull(switchParam);
        Assert.assertEquals(switchParam, configManager.getModelItem("switchParam"));
        
        Assert.assertNotNull(_getElementDefinition(flatDefinitions, "stringParam"));
        Assert.assertNotNull(_getElementDefinition(flatDefinitions, "longParam"));
        Assert.assertNotNull(_getElementDefinition(flatDefinitions, "firstParam"));
        Assert.assertNotNull(_getElementDefinition(flatDefinitions, "secondParam"));
        Assert.assertNotNull(_getElementDefinition(flatDefinitions, "thirdParam"));
        Assert.assertNotNull(_getElementDefinition(flatDefinitions, "yyFourthParam"));
        Assert.assertNotNull(_getElementDefinition(flatDefinitions, "zzFifthParam"));

        Assert.assertNull(_getElementDefinition(flatDefinitions, "unusedParam"));
        Assert.assertNull(_getElementDefinition(flatDefinitions, "paramInUnusedFeature"));
    }

    @SuppressWarnings("unchecked")
    private <T> ElementDefinition<T> _getElementDefinition(Collection<ElementDefinition> definitions, String name)
    {
        for (ElementDefinition definition : definitions)
        {
            if (name.equals(definition.getName()))
            {
                return definition;
            }
        }
        
        return null;
    }

    private void _checkValues()
    {
        Config config = Config.getInstance();
        Map<String, Object> values = config.getValues();
        
        // Switcher param value
        Boolean switchParamValue = config.getValue("switchParam");
        Assert.assertNull(switchParamValue);
        Assert.assertEquals(switchParamValue, values.get("switchParam"));
        switchParamValue = config.getValue("switchParam", true, true);
        Assert.assertFalse(switchParamValue);
        
        // String param value
        String stringParamValue = config.getValue("stringParam");
        Assert.assertEquals("string param value", stringParamValue);
        Assert.assertEquals(stringParamValue, values.get("stringParam"));
        stringParamValue = config.getValue("stringParam", true, "other default value");
        Assert.assertEquals("string param value", stringParamValue);
        
        // Long param value
        Long longParamValue = config.getValue("longParam");
        Assert.assertEquals(Long.valueOf(4), longParamValue);
        Assert.assertEquals(longParamValue, values.get("longParam"));
        Assert.assertEquals(Long.valueOf(4), config.getValue("longParam", true, (long) 8));
        Assert.assertEquals(Long.valueOf(4), config.getValue("longParam", false, (long) 8));
        
        // Empty long param value
        longParamValue = config.getValue("emptyLongParam");
        Assert.assertNull(longParamValue);
        Assert.assertEquals(Long.valueOf(1), config.getValue("emptyLongParam", true, (long) 3));
        Assert.assertEquals(Long.valueOf(3), config.getValue("emptyLongParam", false, (long) 3));
        
        // First param value
        String firstParamValue = config.getValue("firstParam");
        Assert.assertEquals("first param value", firstParamValue);
        Assert.assertEquals(firstParamValue, values.get("firstParam"));
        
        // Second param value
        String secondParamValue = config.getValue("secondParam");
        Assert.assertEquals("", secondParamValue);
        Assert.assertEquals(secondParamValue, values.get("secondParam"));
        secondParamValue = config.getValue("secondParam", true, "other default value");
        Assert.assertEquals("", secondParamValue);
        secondParamValue = config.getValue("secondParam", false, "other default value");
        Assert.assertEquals("", secondParamValue);
        
        // Third param value
        String thridParamValue = config.getValue("thridParam");
        Assert.assertNull(thridParamValue);
        Assert.assertEquals(thridParamValue, values.get("thridParam"));
        thridParamValue = config.getValue("thridParam", true, "other default value");
        Assert.assertEquals("other default value", thridParamValue);
        
        // Fourth param value
        String fourthParamValue = config.getValue("yyFourthParam");
        Assert.assertEquals("fourth param value", fourthParamValue);
        Assert.assertEquals(fourthParamValue, values.get("yyFourthParam"));
        
        // Fifth param value
        String fifthParamValue = config.getValue("zzFifthParam");
        Assert.assertEquals("fifth param value", fifthParamValue);
        Assert.assertEquals(fifthParamValue, values.get("zzFifthParam"));
        
        // Unavaible parameters
        Assert.assertNull(config.getValue("unusedParam"));
        Assert.assertEquals("default value", config.getValue("unusedParam", true, "default value"));
        Assert.assertNull(config.getValue("paramInUnusedFeature"));
        Assert.assertEquals("default value", config.getValue("paramInUnusedFeature", true, "default value"));
    }
}
