/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.Collection;

import org.apache.avalon.framework.service.ServiceException;

import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.model.type.ElementType;
import org.ametys.runtime.model.type.ModelItemType;
import org.ametys.runtime.plugin.ExtensionPoint;

/**
 * {@link ElementDefinition} that is also a {@link ModelItemAccessor}.
 * This class is for tests purposes as this case does not exist in the runtime plugin
 * To create an instance of {@link ElementDefinitionAccessorForTests}, use the {@link #of(String, boolean, String, String, Model)} method.
 * This accessor will be able to access to the model items of the iven {@link Model}
 *
 * @param <T> Type of the element value
 */
public class ElementDefinitionAccessorForTests<T> extends DefaultElementDefinition<T> implements ModelItemAccessor
{
    private Model _model;
    
    /**
     * Constructor used to create simple models and items 
     * @param name the name of the definition
     * @param isMultiple the element multiple status
     * @param type the type of the definition
     * @param model the model with the accessed items
     */
    public ElementDefinitionAccessorForTests(String name, boolean isMultiple, ElementType<T> type, Model model)
    {
        super(name, isMultiple, type);
        _model = model;
    }
    
    public Collection< ? extends ModelItem> getModelItems()
    {
        return _model.getModelItems();
    }

    /**
     * Creates an {@link ElementDefinitionAccessorForTests}
     * @param name the definition's name
     * @param isMultiple the definition's cardinality
     * @param typeId the definition's type identifier
     * @param availableTypesExtensionPoint the role of the extension point containing all available types for this {@link ElementDefinition}
     * @param model the model with the accessed items
     * @return the created {@link ElementDefinition}
     * @throws UnknownTypeException if the given type identifier is not available in the extension point
     * @throws BadItemTypeException if the given type identifier can not be used for an {@link ElementDefinition}
     * @throws ServiceException if an error occurs while getting the extension point of available types
     */
    @SuppressWarnings("unchecked")
    public static ElementDefinitionAccessorForTests of(String name, boolean isMultiple, String typeId, String availableTypesExtensionPoint, Model model) throws ServiceException
    {
        ExtensionPoint<ModelItemType> availableTypes = (ExtensionPoint<ModelItemType>) __serviceManager.lookup(availableTypesExtensionPoint);
        if (!availableTypes.hasExtension(typeId))
        {
            throw new UnknownTypeException("The type '" + typeId + "' (used for data '" + name + "') is not available for the given extension point.");
        }
        else
        {
            ModelItemType type = availableTypes.getExtension(typeId);
            if (!(type instanceof ElementType))
            {
                throw new BadItemTypeException("The type '" + typeId + "' (used for data '" + name + "') can not be used for an element definition.");
            }
            else
            {
                return new ElementDefinitionAccessorForTests(name, isMultiple, (ElementType) type, model);
            }
        }
    }

}
