/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.model.type.ModelItemTypeExtensionPoint;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * {@link ModelHelper#getAllModelItemsInPath(String, Collection)} test cases
 */
@NeedFeatures({
    "core/model.helper"
})
public class ModelHelperGetAllModelItemsInPathTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static final String __MODELS_FAMILY_ID = "test_model_for_helper_test_cases";
    
    private static Collection<Model> _models;

    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _models = new ArrayList<>();
        _models.add(_createModel1("model1", __MODELS_FAMILY_ID));
        _models.add(_createModel2("model2", __MODELS_FAMILY_ID));
    }
    
    /**
     * Tests the {@link ModelHelper#getAllModelItemsInPath(String, Collection)} method
     */
    @Test
    public void testGetModelItemPathNominalCases()
    {
        List<ModelItem> modelItemPath = ModelHelper.getAllModelItemsInPath("dateData", _models);
        Assert.assertEquals(1, modelItemPath.size());
        Assert.assertEquals("dateData", modelItemPath.get(0).getName());

        modelItemPath = ModelHelper.getAllModelItemsInPath("otherStringData", _models);
        Assert.assertEquals(1, modelItemPath.size());
        Assert.assertEquals("otherStringData", modelItemPath.get(0).getName());

        modelItemPath = ModelHelper.getAllModelItemsInPath("group", _models);
        Assert.assertEquals(1, modelItemPath.size());
        Assert.assertEquals("group", modelItemPath.get(0).getName());

        modelItemPath = ModelHelper.getAllModelItemsInPath("group/longData", _models);
        Assert.assertEquals(2, modelItemPath.size());
        Assert.assertEquals("group", modelItemPath.get(0).getName());
        Assert.assertEquals("longData", modelItemPath.get(1).getName());

        modelItemPath = ModelHelper.getAllModelItemsInPath("group/subGroup/booleanData", _models);
        Assert.assertEquals(3, modelItemPath.size());
        Assert.assertEquals("group", modelItemPath.get(0).getName());
        Assert.assertEquals("subGroup", modelItemPath.get(1).getName());
        Assert.assertEquals("booleanData", modelItemPath.get(2).getName());
    }
    
    /**
     * Tests the {@link ModelHelper#getAllModelItemsInPath(String, Collection)} method with a path that begins from a group
     */
    @Test
    public void testGetModelItemPathFromGroup()
    {
        ModelItemGroup group = (ModelItemGroup) ModelHelper.getModelItem("group", _models);
        List<ModelItem> modelItemPath = ModelHelper.getAllModelItemsInPath("subGroup/booleanData", Collections.singleton(group));
        Assert.assertEquals(2, modelItemPath.size());
        Assert.assertEquals("subGroup", modelItemPath.get(0).getName());
        Assert.assertEquals("booleanData", modelItemPath.get(1).getName());
    }
    
    /**
     * Tests the {@link ModelHelper#getAllModelItemsInPath(String, Collection)} method with a <code>null</code> or empty path
     * The method should throw an {@link IllegalArgumentException}
     */
    @Test
    public void testGetModelItemPathWithNullOrEmptyPath()
    {
        try
        {
            ModelHelper.getAllModelItemsInPath(null, _models);
            Assert.fail("a null path should have lead to an IllegalArgumentException");
        }
        catch (IllegalArgumentException e)
        {
            // Nothing to do
        }
        
        try
        {
            ModelHelper.getAllModelItemsInPath(StringUtils.EMPTY, _models);
            Assert.fail("an empty path should have lead to an IllegalArgumentException");
        }
        catch (IllegalArgumentException e)
        {
            // Nothing to do
        }
    }
    
    /**
     * Tests the {@link ModelHelper#getAllModelItemsInPath(String, Collection)} method with a path of a data that does not exist
     * The method should throw an {@link UndefinedItemPathException}
     */
    @Test
    public void testGetModelItemPathWithInexistingPath()
    {
        try
        {
            ModelHelper.getAllModelItemsInPath("unexistingData", _models);
            Assert.fail("a path of an unexisting data should have lead to an UndefinedItemPathException");
        }
        catch (UndefinedItemPathException e)
        {
            // Nothing to do
        }
        
        try
        {
            ModelHelper.getAllModelItemsInPath("group/unexistingData", _models);
            Assert.fail("a path of an unexisting data should have lead to an UndefinedItemPathException");
        }
        catch (UndefinedItemPathException e)
        {
            // Nothing to do
        }
    }

    private Model _createModel1(String modelId, String familyId) throws UnknownTypeException, BadItemTypeException, ServiceException
    {
        String role = ModelItemTypeExtensionPoint.ROLE_CONFIG;
        return Model.of(modelId, familyId,
                DefaultElementDefinition.of("stringData", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("longData", false, ModelItemTypeConstants.LONG_TYPE_ID, role),
                DefaultElementDefinition.of("booleanData", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role),
                DefaultElementDefinition.of("dateData", false, ModelItemTypeConstants.DATE_TYPE_ID, role),
                ModelItemGroup.of("group",
                        DefaultElementDefinition.of("longData", false, ModelItemTypeConstants.LONG_TYPE_ID, role),
                        ModelItemGroup.of("subGroup",
                                DefaultElementDefinition.of("booleanData", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role),
                                DefaultElementDefinition.of("stringData", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                                DefaultElementDefinition.of("dateData", false, ModelItemTypeConstants.DATE_TYPE_ID, role))));
    }
    
    private Model _createModel2(String modelId, String familyId) throws UnknownTypeException, BadItemTypeException, ServiceException
    {
        String role = ModelItemTypeExtensionPoint.ROLE_CONFIG;
        return Model.of(modelId, familyId,
                DefaultElementDefinition.of("otherStringData", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("otherBooleanData", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role));
    }
}
