/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.List;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.model.type.ModelItemTypeExtensionPoint;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * {@link ModelViewItemGroup} test case
 */
@NeedFeatures({"core/model.helper"})
public class ModelViewItemGroupTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static final String __MODELS_FAMILY_ID = "test_model_for_view_constructors";
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        // Nothing to do
    }
    
    /**
     * Tests the {@link ModelViewItemGroup#of(ModelItemGroup)} method. Nominal case.
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewConstructionFromModel() throws Exception
    {
        Model model = _createModelForViewConstructionsTests("constructor_from_model");
        ModelViewItemGroup viewItemGroup = ModelViewItemGroup.of((ModelItemGroup) model.getModelItem("group"));
        List<ViewItem> viewItems = viewItemGroup.getViewItems();
        
        Assert.assertEquals(2, viewItems.size());
        Assert.assertEquals("longData", ((ViewElement) viewItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup subGroup = (ModelViewItemGroup) viewItems.get(1);
        List<ViewItem> subGroupViewItems = subGroup.getViewItems();
        Assert.assertEquals(3, subGroupViewItems.size());
        Assert.assertEquals("booleanData", ((ViewElement) subGroupViewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("stringData", ((ViewElement) subGroupViewItems.get(1)).getDefinition().getName());
        Assert.assertEquals("dateData", ((ViewElement) subGroupViewItems.get(2)).getDefinition().getName());
    }
    
    /**
     * Tests the {@link View#of(Model, String...)} method. Nominal Case.
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewConstructionFromItemsList() throws Exception
    {
        Model model = _createModelForViewConstructionsTests("constructor_from_items");
        ModelViewItemGroup viewItemGroup = ModelViewItemGroup.of((ModelItemGroup) model.getModelItem("group"), "longData", "subGroup/stringData");
        List<ViewItem> viewItems = viewItemGroup.getViewItems();
        
        Assert.assertEquals(2, viewItems.size());
        Assert.assertEquals("longData", ((ViewElement) viewItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup subGroup = (ModelViewItemGroup) viewItems.get(1);
        List<ViewItem> subGroupViewItems = subGroup.getViewItems();
        Assert.assertEquals(1, subGroupViewItems.size());
        Assert.assertEquals("stringData", ((ViewElement) subGroupViewItems.get(0)).getDefinition().getName());
    }
    
    private Model _createModelForViewConstructionsTests(String modelId) throws UnknownTypeException, BadItemTypeException, ServiceException
    {
        String role = ModelItemTypeExtensionPoint.ROLE_CONFIG;
        return Model.of(modelId, __MODELS_FAMILY_ID,
                ModelItemGroup.of("group",
                        DefaultElementDefinition.of("longData", false, ModelItemTypeConstants.LONG_TYPE_ID, role),
                        ModelItemGroup.of("subGroup",
                                DefaultElementDefinition.of("booleanData", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role),
                                DefaultElementDefinition.of("stringData", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                                DefaultElementDefinition.of("dateData", false, ModelItemTypeConstants.DATE_TYPE_ID, role))));
    }
}
