/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import static org.junit.Assert.assertEquals;

import java.util.List;
import java.util.Set;
import java.util.function.BiFunction;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UndefinedItemPathException;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.model.type.ModelItemTypeExtensionPoint;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * {@link ViewHelper} test cases
 */
@NeedFeatures({
    "core/model.helper"
})
public class ViewHelperTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static final String __MODELS_FAMILY_ID = "test_model_for_view_helper_test_cases";
    
    private static Model _model;
    private static Model _modelForAccessor;

    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _model = _createModel("model", __MODELS_FAMILY_ID);
        _modelForAccessor = _createModelForAccessors("model_for_accessors", __MODELS_FAMILY_ID);
    }
    
    private Model _createModel(String modelId, String familyId) throws UnknownTypeException, BadItemTypeException, ServiceException
    {
        String role = ModelItemTypeExtensionPoint.ROLE_CONFIG;
        return Model.of(modelId, familyId,
                DefaultElementDefinition.of("stringData", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                ModelItemGroup.of("group",
                        DefaultElementDefinition.of("longData", false, ModelItemTypeConstants.LONG_TYPE_ID, role),
                        DefaultElementDefinition.of("doubleData", false, ModelItemTypeConstants.DOUBLE_TYPE_ID, role),
                        ModelItemGroup.of("subGroup",
                                DefaultElementDefinition.of("dateData", false, ModelItemTypeConstants.DATE_TYPE_ID, role))));
    }
    
    private Model _createModelForAccessors(String modelId, String familyId) throws UnknownTypeException, BadItemTypeException, ServiceException
    {
        String role = ModelItemTypeExtensionPoint.ROLE_CONFIG;
        return Model.of(modelId, familyId,
                DefaultElementDefinition.of("stringData", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("longData", false, ModelItemTypeConstants.LONG_TYPE_ID, role));
    }
    
    /**
     * Tests the {@link ViewHelper#getModelItemsPathsFromView(View)} method
     */
    @Test
    public void testGetViewPaths()
    {
        View view = _createView(_model);
        
        Set<String> viewItemsPaths = ViewHelper.getModelItemsPathsFromView(view);
        
        Assert.assertEquals(5, viewItemsPaths.size());
        Assert.assertTrue(viewItemsPaths.contains("stringData"));
        Assert.assertTrue(viewItemsPaths.contains("group"));
        Assert.assertTrue(viewItemsPaths.contains("group/longData"));
        Assert.assertTrue(viewItemsPaths.contains("group/subGroup"));
        Assert.assertTrue(viewItemsPaths.contains("group/subGroup/dateData"));
    }
    
    /**
     * Tests 
     * <ul>
     * <li>{@link ViewHelper#getModelViewItem}</li>
     * <li>{@link ViewHelper#getViewElement}</li>
     * <li>{@link ViewHelper#getModelViewItemGroup}</li>
     * </ul>
     */
    @Test
    public void testGetViewItem()
    {
        View view = _createView(_model);
        
        // stringData
        final ModelViewItem item1 = ViewHelper.getModelViewItem(view, "stringData");
        assertEquals("stringData", item1.getName());
        final ViewElement modelViewItem1 = ViewHelper.getViewElement(view, "stringData");
        assertEquals("stringData", modelViewItem1.getName());
        assertEquals(item1, modelViewItem1);
        _assertFail(ViewHelper::getModelViewItemGroup, view, "stringData", BadItemTypeException.class);
        
        
        // group
        final ModelViewItem item2 = ViewHelper.getModelViewItem(view, "group");
        assertEquals("group", item2.getName());
        _assertFail(ViewHelper::getViewElement, view, "group", BadItemTypeException.class);
        ModelViewItemGroup itemContainer2 = ViewHelper.getModelViewItemGroup(view, "group");
        assertEquals(item2, itemContainer2);
        
        
        // group/longData
        final ModelViewItem item3 = ViewHelper.getModelViewItem(view, "group/longData");
        assertEquals("longData", item3.getName());
        final ViewElement modelViewItem3 = ViewHelper.getViewElement(view, "group/longData");
        assertEquals("longData", modelViewItem3.getName());
        assertEquals(item3, modelViewItem3);
        _assertFail(ViewHelper::getModelViewItemGroup, view, "group/longData", BadItemTypeException.class);
        
        
        // group/subGroup
        final ModelViewItem item4 = ViewHelper.getModelViewItem(view, "group/subGroup");
        assertEquals("subGroup", item4.getName());
        _assertFail(ViewHelper::getViewElement, view, "group/subGroup", BadItemTypeException.class);
        ModelViewItemGroup itemContainer4 = ViewHelper.getModelViewItemGroup(view, "group/subGroup");
        assertEquals(item4, itemContainer4);
        
        
        // group/subGroup/dateData
        final ModelViewItem item5 = ViewHelper.getModelViewItem(view, "group/subGroup/dateData");
        assertEquals("dateData", item5.getName());
        final ViewElement modelViewItem5 = ViewHelper.getViewElement(view, "group/subGroup/dateData");
        assertEquals("dateData", modelViewItem5.getName());
        assertEquals(item5, modelViewItem5);
        _assertFail(ViewHelper::getModelViewItemGroup, view, "group/subGroup/dateData", BadItemTypeException.class);
        
        
        // Fails
        _assertFail(ViewHelper::getModelViewItem, view, "unexisting", UndefinedItemPathException.class);
        _assertFail(ViewHelper::getViewElement, view, "unexisting", UndefinedItemPathException.class);
        _assertFail(ViewHelper::getModelViewItemGroup, view, "unexisting", UndefinedItemPathException.class);
        _assertFail(ViewHelper::getModelViewItem, view, "group/unexisting", UndefinedItemPathException.class);
        _assertFail(ViewHelper::getViewElement, view, "group/unexisting", UndefinedItemPathException.class);
        _assertFail(ViewHelper::getModelViewItemGroup, view, "group/unexisting", UndefinedItemPathException.class);
    }
    
    private <R> void _assertFail(BiFunction<View, String, R> methodToTest, View view, String path, Class<? extends Exception> expected)
    {
        try
        {
            methodToTest.apply(view, path);
            Assert.fail("it should fail for '" + path + "'");
        }
        catch (Exception e)
        {
            assertEquals(expected, e.getClass());
        }
    }
    
    /**
     * Tests the {@link ViewHelper#areItemsPresentsOnlyOnce(View)} method
     * All items appear only once, method return <code>true</code>
     */
    @Test
    public void testAreItemsPresentsOnlyOnceAllItemsOnlyOnce()
    {
        View view = View.of(_model,
                "stringData",
                "group/longData");
        
        Assert.assertTrue(ViewHelper.areItemsPresentsOnlyOnce(view));
    }
    
    /**
     * Tests the {@link ViewHelper#areItemsPresentsOnlyOnce(View)} method
     * An item appears twice, method return <code>false</code>
     */
    @Test
    public void testAreItemsPresentsOnlyOnceAnItemTwice()
    {
        View view = View.of(_model,
                "stringData",
                "stringData",
                "group/longData");
        
        Assert.assertFalse(ViewHelper.areItemsPresentsOnlyOnce(view));
    }
    
    /**
     * Tests the {@link ViewHelper#areItemsPresentsOnlyOnce(View)} method
     * An item in group appears twice, method return <code>false</code>
     */
    @Test
    public void testAreItemsPresentsOnlyOnceAnItemTwiceInAGroup()
    {
        View view = View.of(_model,
                "stringData",
                "group/longData",
                "group/longData");
        
        Assert.assertFalse(ViewHelper.areItemsPresentsOnlyOnce(view));
    }
    
    /**
     * Tests the {@link ViewHelper#areItemsPresentsOnlyOnce(View)} method
     * A group appears twice, and even if the sub items are not the same, method return <code>false</code>
     */
    @Test
    public void testAreItemsPresentsOnlyOnceAGroupTwiceButDifferentSubItems()
    {
        View view = _createViewWithAGroupTwiceButDifferentSubItems(_model);
        Assert.assertFalse(ViewHelper.areItemsPresentsOnlyOnce(view));
    }
    
    /**
     * Create a view from the given model, with some duplicated items:
     * stringData
     * group (model)
     *   longData
     * group (model)
     *   subgroup (model)
     *     dateData
     *     
     * @param model the model from which the view is created
     * @return the created view
     */
    private View _createViewWithAGroupTwiceButDifferentSubItems(Model model)
    {
        View view = new View();
        
        ViewElement stringData = new ViewElement();
        stringData.setDefinition((ElementDefinition) model.getModelItem("stringData"));
        view.addViewItem(stringData);
        
        ModelItemGroup groupDef = (ModelItemGroup) model.getModelItem("group");
        ModelViewItemGroup<ModelItemGroup> group1 = new ModelViewItemGroup<>();
        group1.setDefinition(groupDef);
        view.addViewItem(group1);
        
        ViewElement longData = new ViewElement();
        longData.setDefinition((ElementDefinition) model.getModelItem("group/longData"));
        group1.addViewItem(longData);
        
        ModelViewItemGroup<ModelItemGroup> group2 = new ModelViewItemGroup<>();
        group2.setDefinition(groupDef);
        view.addViewItem(group2);
        
        ModelViewItemGroup<ModelItemGroup> subGroup = new ModelViewItemGroup<>();
        subGroup.setDefinition((ModelItemGroup) model.getModelItem("group/subGroup"));
        group2.addViewItem(subGroup);
        
        ViewElement dateData = new ViewElement();
        dateData.setDefinition((ElementDefinition) model.getModelItem("group/subGroup/dateData"));
        subGroup.addViewItem(dateData);
        
        return view;
    }
    
    /**
     * Tests the {@link ViewHelper#areItemsPresentsOnlyOnce(View)} method
     * A group appears twice with the same sub item, method return <code>false</code>
     */
    @Test
    public void testAreItemsPresentsOnlyOnceAGroupTwiceWithSameSubItem()
    {
        View view = new View();
        
        ViewElement stringData = new ViewElement();
        stringData.setDefinition((ElementDefinition) _model.getModelItem("stringData"));
        view.addViewItem(stringData);
        
        ModelItemGroup groupDef = (ModelItemGroup) _model.getModelItem("group");
        ModelViewItemGroup<ModelItemGroup> group1 = new ModelViewItemGroup<>();
        group1.setDefinition(groupDef);
        view.addViewItem(group1);
        
        ElementDefinition longDataDef = (ElementDefinition) _model.getModelItem("group/longData");
        ViewElement longData1 = new ViewElement();
        longData1.setDefinition(longDataDef);
        group1.addViewItem(longData1);
        
        ModelViewItemGroup<ModelItemGroup> group2 = new ModelViewItemGroup<>();
        group2.setDefinition(groupDef);
        view.addViewItem(group2);
        
        ViewElement longData2 = new ViewElement();
        longData2.setDefinition(longDataDef);
        group2.addViewItem(longData2);
        
        ModelViewItemGroup<ModelItemGroup> subGroup = new ModelViewItemGroup<>();
        subGroup.setDefinition((ModelItemGroup) _model.getModelItem("group/subGroup"));
        group2.addViewItem(subGroup);
        
        ViewElement dateData = new ViewElement();
        dateData.setDefinition((ElementDefinition) _model.getModelItem("group/subGroup/dateData"));
        subGroup.addViewItem(dateData);
        
        Assert.assertFalse(ViewHelper.areItemsPresentsOnlyOnce(view));
    }
    
    /**
     * Tests the {@link ViewHelper#getTruncatedView(View)} method
     * @throws Exception if an error occurs
     */
    @Test
    public void testGetTruncatedView() throws Exception
    {
        View viewWithAccessors = _createViewWithAccessors(_model, _modelForAccessor);
        
        View truncatedView = ViewHelper.getTruncatedView(viewWithAccessors);
        
        List<ViewItem> viewItems = truncatedView.getViewItems();
        
        Assert.assertEquals(2, viewItems.size());
        
        SimpleViewItemGroup tab1 = (SimpleViewItemGroup) viewItems.get(0);
        Assert.assertEquals("tab1", tab1.getName());
        List<ViewItem> tab1ViewItems = tab1.getViewItems();
        Assert.assertEquals(1, tab1ViewItems.size());
        Assert.assertEquals("stringData", ((ViewElement) tab1ViewItems.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup tab2 = (SimpleViewItemGroup) viewItems.get(1);
        Assert.assertEquals("tab2", tab2.getName());
        List<ViewItem> tab2ViewItems = tab2.getViewItems();
        Assert.assertEquals(1, tab2ViewItems.size());
        
        ModelViewItemGroup group = (ModelViewItemGroup) tab2ViewItems.get(0);
        List<ViewItem> groupViewItems = group.getViewItems();
        Assert.assertEquals(2, groupViewItems.size());
        
        SimpleViewItemGroup fieldset1 = (SimpleViewItemGroup) groupViewItems.get(0);
        Assert.assertEquals("fieldset1", fieldset1.getName());
        List<ViewItem> fieldset1ViewItems = fieldset1.getViewItems();
        Assert.assertEquals(1, fieldset1ViewItems.size());
        Assert.assertEquals("longData", ((ViewElement) fieldset1ViewItems.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup fieldset2 = (SimpleViewItemGroup) groupViewItems.get(1);
        Assert.assertEquals("fieldset2", fieldset2.getName());
        List<ViewItem> fieldset2ViewItems = fieldset2.getViewItems();
        Assert.assertEquals(1, fieldset2ViewItems.size());
        
        ModelViewItemGroup subGroup = (ModelViewItemGroup) fieldset2ViewItems.get(0);
        List<ViewItem> subGroupViewItems = subGroup.getViewItems();
        Assert.assertEquals(1, subGroupViewItems.size());
        Assert.assertEquals("dateData", ((ViewElement) subGroupViewItems.get(0)).getDefinition().getName());
    }
    
    /**
     * Create a view from the given model, with an accessor:
     * tab1 (view)
     *   stringData
     * tab2 (view)
     *   group (model)
     *     fieldset1 (view)
     *       longData
     *         stringData
     *         longData
     *     fieldset2 (view)
     *       subgroup (model)
     *         dateData
     *     
     * @param model the model from which the view is created
     * @param modelForAccessor the model of the item accessor (in order to add linked data in the created view)
     * @return the created view
     */
    private View _createViewWithAccessors(Model model, Model modelForAccessor) throws UnknownTypeException, BadItemTypeException
    {
        View view = _createView(model);
        SimpleViewItemGroup fieldset1 = (SimpleViewItemGroup) ((ModelViewItemGroup) ((SimpleViewItemGroup) view.getViewItems().get(1)).getViewItems().get(0)).getViewItems().get(0);
        fieldset1.clear();
        
        ViewElementAccessor longDataViewElement = new ViewElementAccessor();
        longDataViewElement.setDefinition((ElementDefinition) model.getModelItem("group/longData"));
        fieldset1.addViewItem(longDataViewElement);
        
        ViewElement stringDataViewElementForAccessors = new ViewElement();
        stringDataViewElementForAccessors.setDefinition((ElementDefinition) modelForAccessor.getModelItem("stringData"));
        longDataViewElement.addViewItem(stringDataViewElementForAccessors);
        ViewElement longDataViewElementForAccessors = new ViewElement();
        longDataViewElementForAccessors.setDefinition((ElementDefinition) modelForAccessor.getModelItem("longData"));
        longDataViewElement.addViewItem(longDataViewElementForAccessors);
        
        return view;
    }
    
    /**
     * Tests the {@link ViewHelper#mergeDuplicatedItems(View)} method
     * Merged view should look like:
     * tab1 (view)
     *   stringData
     * tab2 (view)
     *   group (model)
     *     fieldset1 (view)
     *       longData
     *         stringData
     *         longData
     *       doubleData
     *         stringData
     *     fieldset2 (view)
     *       subgroup (model)
     *         dateData
     * tab3 (view)
     * 
     * @throws Exception if an error occurs
     */
    @Test
    public void testMergeDuplicatedItems() throws Exception
    {
        View viewWithDuplicatedItems = _createViewWithDuplicatedItems(_model, _modelForAccessor);
        
        View mergedView = ViewHelper.mergeDuplicatedItems(viewWithDuplicatedItems);
        
        List<ViewItem> viewItems = mergedView.getViewItems();
        
        Assert.assertEquals(3, viewItems.size());
        
        SimpleViewItemGroup tab1 = (SimpleViewItemGroup) viewItems.get(0);
        Assert.assertEquals("tab1", tab1.getName());
        List<ViewItem> tab1ViewItems = tab1.getViewItems();
        Assert.assertEquals(1, tab1ViewItems.size());
        Assert.assertEquals("stringData", ((ViewElement) tab1ViewItems.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup tab2 = (SimpleViewItemGroup) viewItems.get(1);
        Assert.assertEquals("tab2", tab2.getName());
        List<ViewItem> tab2ViewItems = tab2.getViewItems();
        Assert.assertEquals(1, tab2ViewItems.size());
        
        ModelViewItemGroup group = (ModelViewItemGroup) tab2ViewItems.get(0);
        List<ViewItem> groupViewItems = group.getViewItems();
        Assert.assertEquals(2, groupViewItems.size());
        
        SimpleViewItemGroup fieldset1 = (SimpleViewItemGroup) groupViewItems.get(0);
        Assert.assertEquals("fieldset1", fieldset1.getName());
        List<ViewItem> fieldset1ViewItems = fieldset1.getViewItems();
        Assert.assertEquals(2, fieldset1ViewItems.size());
        
        ViewElementAccessor longDataAccesor = (ViewElementAccessor) fieldset1ViewItems.get(0);
        Assert.assertEquals("longData", longDataAccesor.getDefinition().getName());
        List<ViewItem> longDataAccessorViewItems = longDataAccesor.getViewItems();
        Assert.assertEquals(2, longDataAccessorViewItems.size());
        Assert.assertEquals("stringData", ((ViewElement) longDataAccessorViewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("longData", ((ViewElement) longDataAccessorViewItems.get(1)).getDefinition().getName());
        
        ViewElementAccessor doubleDataAccesor = (ViewElementAccessor) fieldset1ViewItems.get(1);
        Assert.assertEquals("doubleData", doubleDataAccesor.getDefinition().getName());
        List<ViewItem> doubleDataAccessorViewItems = doubleDataAccesor.getViewItems();
        Assert.assertEquals(1, doubleDataAccessorViewItems.size());
        Assert.assertEquals("stringData", ((ViewElement) doubleDataAccessorViewItems.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup fieldset2 = (SimpleViewItemGroup) groupViewItems.get(1);
        Assert.assertEquals("fieldset2", fieldset2.getName());
        List<ViewItem> fieldset2ViewItems = fieldset2.getViewItems();
        Assert.assertEquals(1, fieldset2ViewItems.size());
        
        ModelViewItemGroup subGroup = (ModelViewItemGroup) fieldset2ViewItems.get(0);
        List<ViewItem> subGroupViewItems = subGroup.getViewItems();
        Assert.assertEquals(1, subGroupViewItems.size());
        Assert.assertEquals("dateData", ((ViewElement) subGroupViewItems.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup tab3 = (SimpleViewItemGroup) viewItems.get(2);
        Assert.assertEquals("tab3", tab3.getName());
        Assert.assertTrue(tab3.getViewItems().isEmpty());
    }
    
    /**
     * Create a view from the given model, with some duplicated items:
     * tab1 (view)
     *   stringData
     *   stringData
     * tab2 (view)
     *   group (model)
     *     fieldset1 (view)
     *       longData
     *         stringData
     *         stringData
     *       doubleData
     *         stringData
     *         stringData
     * tab3 (view)
     *   stringData
     *   group (model)
     *     longData
     *       longData
     *     fieldset2 (view)
     *       subgroup (model)
     *         dateData
     *     
     * @param model the model from which the view is created
     * @param modelForAccessor the model of the item accessor (in order to add linked data in the created view)
     * @return the created view
     */
    private View _createViewWithDuplicatedItems(Model model, Model modelForAccessor)
    {
        View view = new View();
        
        SimpleViewItemGroup tab1 = new SimpleViewItemGroup();
        tab1.setName("tab1");
        tab1.setRole(ViewItemGroup.TAB_ROLE);
        view.addViewItem(tab1);
        
        ViewElement stringDataViewElement = new ViewElement();
        stringDataViewElement.setDefinition((ElementDefinition) model.getModelItem("stringData"));
        tab1.addViewItem(stringDataViewElement);
        
        ViewElement duplicatedStringDataViewElement1 = new ViewElement();
        duplicatedStringDataViewElement1.setDefinition((ElementDefinition) model.getModelItem("stringData"));
        tab1.addViewItem(duplicatedStringDataViewElement1);
        
        SimpleViewItemGroup tab2 = new SimpleViewItemGroup();
        tab2.setName("tab2");
        tab2.setRole(ViewItemGroup.TAB_ROLE);
        view.addViewItem(tab2);
        
        ModelViewItemGroup<ModelItemGroup> groupViewItem = new ModelViewItemGroup<>();
        groupViewItem.setDefinition((ModelItemGroup) model.getModelItem("group"));
        tab2.addViewItem(groupViewItem);
        
        SimpleViewItemGroup fieldset1 = new SimpleViewItemGroup();
        fieldset1.setName("fieldset1");
        fieldset1.setRole(ViewItemGroup.FIELDSET_ROLE);
        groupViewItem.addViewItem(fieldset1);
        
        ViewElementAccessor longDataViewElement = new ViewElementAccessor();
        longDataViewElement.setDefinition((ElementDefinition) model.getModelItem("group/longData"));
        fieldset1.addViewItem(longDataViewElement);
        
        ViewElement stringDataViewElementForAccessors = new ViewElement();
        stringDataViewElementForAccessors.setDefinition((ElementDefinition) modelForAccessor.getModelItem("stringData"));
        longDataViewElement.addViewItem(stringDataViewElementForAccessors);
        
        ViewElementAccessor doubleDataViewElement = new ViewElementAccessor();
        doubleDataViewElement.setDefinition((ElementDefinition) model.getModelItem("group/doubleData"));
        fieldset1.addViewItem(doubleDataViewElement);
        
        ViewElement stringDataViewElementForAccessorsInDoubleData = new ViewElement();
        stringDataViewElementForAccessorsInDoubleData.setDefinition((ElementDefinition) modelForAccessor.getModelItem("stringData"));
        doubleDataViewElement.addViewItem(stringDataViewElementForAccessorsInDoubleData);
        
        ViewElement duplicatedStringDataViewElementForAccessorsInDoubleData = new ViewElement();
        duplicatedStringDataViewElementForAccessorsInDoubleData.setDefinition((ElementDefinition) modelForAccessor.getModelItem("stringData"));
        doubleDataViewElement.addViewItem(duplicatedStringDataViewElementForAccessorsInDoubleData);
        
        SimpleViewItemGroup tab3 = new SimpleViewItemGroup();
        tab3.setName("tab3");
        tab3.setRole(ViewItemGroup.TAB_ROLE);
        view.addViewItem(tab3);
        
        ViewElement duplicatedStringDataViewElement2 = new ViewElement();
        duplicatedStringDataViewElement2.setDefinition((ElementDefinition) model.getModelItem("stringData"));
        tab3.addViewItem(duplicatedStringDataViewElement2);
        
        ModelViewItemGroup<ModelItemGroup> duplicatedGroupViewItem = new ModelViewItemGroup<>();
        duplicatedGroupViewItem.setDefinition((ModelItemGroup) model.getModelItem("group"));
        tab3.addViewItem(duplicatedGroupViewItem);
        
        ViewElementAccessor duplicatedLongDataViewElement = new ViewElementAccessor();
        duplicatedLongDataViewElement.setDefinition((ElementDefinition) model.getModelItem("group/longData"));
        duplicatedGroupViewItem.addViewItem(duplicatedLongDataViewElement);
        
        ViewElement longDataViewElementForAccessors = new ViewElement();
        longDataViewElementForAccessors.setDefinition((ElementDefinition) modelForAccessor.getModelItem("longData"));
        duplicatedLongDataViewElement.addViewItem(longDataViewElementForAccessors);
        
        SimpleViewItemGroup fieldset2 = new SimpleViewItemGroup();
        fieldset2.setName("fieldset2");
        fieldset2.setRole(ViewItemGroup.FIELDSET_ROLE);
        duplicatedGroupViewItem.addViewItem(fieldset2);
        
        ModelViewItemGroup<ModelItemGroup> subGroupViewItem = new ModelViewItemGroup<>();
        subGroupViewItem.setDefinition((ModelItemGroup) model.getModelItem("group/subGroup"));
        fieldset2.addViewItem(subGroupViewItem);
        
        ViewElement dateDataViewElement = new ViewElement();
        dateDataViewElement.setDefinition((ElementDefinition) model.getModelItem("group/subGroup/dateData"));
        subGroupViewItem.addViewItem(dateDataViewElement);
        
        return view;
    }
    
    /**
     * Create a view from the given model:
     * tab1 (view)
     *   stringData
     * tab2 (view)
     *   group (model)
     *     fieldset1 (view)
     *       longData
     *     fieldset2 (view)
     *       subgroup (model)
     *         dateData
     *     
     * @param model the model from which the view is created
     * @return the created view
     */
    private View _createView(Model model)
    {
        View view = new View();
        
        SimpleViewItemGroup tab1 = new SimpleViewItemGroup();
        tab1.setName("tab1");
        tab1.setRole(ViewItemGroup.TAB_ROLE);
        view.addViewItem(tab1);
        
        ViewElement stringDataViewElement = new ViewElement();
        stringDataViewElement.setDefinition((ElementDefinition) model.getModelItem("stringData"));
        tab1.addViewItem(stringDataViewElement);
        
        SimpleViewItemGroup tab2 = new SimpleViewItemGroup();
        tab2.setName("tab2");
        tab2.setRole(ViewItemGroup.TAB_ROLE);
        view.addViewItem(tab2);
        
        ModelViewItemGroup<ModelItemGroup> groupViewItem = new ModelViewItemGroup<>();
        groupViewItem.setDefinition((ModelItemGroup) model.getModelItem("group"));
        tab2.addViewItem(groupViewItem);
        
        SimpleViewItemGroup fieldset1 = new SimpleViewItemGroup();
        fieldset1.setName("fieldset1");
        fieldset1.setRole(ViewItemGroup.FIELDSET_ROLE);
        groupViewItem.addViewItem(fieldset1);
        
        ViewElement longDataViewElement = new ViewElement();
        longDataViewElement.setDefinition((ElementDefinition) model.getModelItem("group/longData"));
        fieldset1.addViewItem(longDataViewElement);
        
        SimpleViewItemGroup fieldset2 = new SimpleViewItemGroup();
        fieldset2.setName("fieldset2");
        fieldset2.setRole(ViewItemGroup.FIELDSET_ROLE);
        groupViewItem.addViewItem(fieldset2);
        
        ModelViewItemGroup<ModelItemGroup> subGroupViewItem = new ModelViewItemGroup<>();
        subGroupViewItem.setDefinition((ModelItemGroup) model.getModelItem("group/subGroup"));
        fieldset2.addViewItem(subGroupViewItem);
        
        ViewElement dateDataViewElement = new ViewElement();
        dateDataViewElement.setDefinition((ElementDefinition) model.getModelItem("group/subGroup/dateData"));
        subGroupViewItem.addViewItem(dateDataViewElement);
        
        return view;
    }
}
