/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.model;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.runtime.model.exception.BadItemTypeException;
import org.ametys.runtime.model.exception.UnknownTypeException;
import org.ametys.runtime.model.type.ModelItemTypeConstants;
import org.ametys.runtime.model.type.ModelItemTypeExtensionPoint;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * {@link View} test case
 */
@NeedFeatures({"core/model.helper"})
public class ViewTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static final String __MODELS_FAMILY_ID = "test_model_for_view_constructors";
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        // Nothing to do
    }
    
    /**
     * Tests the {@link View#toJSON(DefinitionContext)} method
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testToJSON() throws Exception
    {
        View view = _createViewForJSONTest();
        Map<String, Object> json = view.toJSON(DefinitionContext.newInstance());
        
        Assert.assertEquals("view", json.get("name"));
        
        // Items
        Map<String, Object> items = (Map<String, Object>) json.get("elements");
        
        // element 1
        Map<String, Object> definition1 = (Map<String, Object>) items.get("one");
        Assert.assertEquals("one", definition1.get("name"));
        Assert.assertEquals("string", definition1.get("type"));
        
        // Tab1
        Map<String, Object> tab1 = (Map<String, Object>) items.get("tab1");
        Assert.assertEquals("tab1", tab1.get("name"));
        Assert.assertEquals(ViewItemGroup.TAB_ROLE, tab1.get("role"));
        Map<String, Object> tab1Elements = (Map<String, Object>) tab1.get("elements");
        Map<String, Object> definition2 = (Map<String, Object>) tab1Elements.get("two");
        Assert.assertEquals("two", definition2.get("name"));
        Assert.assertEquals("long", definition2.get("type"));
        Map<String, Object> definition3 = (Map<String, Object>) tab1Elements.get("three");
        Assert.assertEquals("three", definition3.get("name"));
        Assert.assertEquals("boolean", definition3.get("type"));
        Map<String, Object> group = (Map<String, Object>) tab1Elements.get("group");
        Assert.assertEquals("group", group.get("name"));
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, group.get("role"));
        Map<String, Object> groupElements = (Map<String, Object>) group.get("elements");
        Map<String, Object> definition4 = (Map<String, Object>) groupElements.get("four");
        Assert.assertEquals("four", definition4.get("name"));
        Assert.assertEquals("date", definition4.get("type"));
        Map<String, Object> definition5 = (Map<String, Object>) groupElements.get("five");
        Assert.assertEquals("five", definition5.get("name"));
        Assert.assertEquals("string", definition5.get("type"));
        
        // Element 6
        Map<String, Object> definition6 = (Map<String, Object>) items.get("six");
        Assert.assertEquals("six", definition6.get("name"));
        Assert.assertEquals("long", definition6.get("type"));
        
        // Tab2
        Map<String, Object> tab2 = (Map<String, Object>) items.get("tab2");
        Assert.assertEquals("tab2", tab2.get("name"));
        Assert.assertEquals(ViewItemGroup.TAB_ROLE, tab2.get("role"));
        Map<String, Object> tab2Elements = (Map<String, Object>) tab2.get("elements");
        Map<String, Object> definition7 = (Map<String, Object>) tab2Elements.get("seven");
        Assert.assertEquals("seven", definition7.get("name"));
        Assert.assertEquals("boolean", definition7.get("type"));
        Map<String, Object> fieldset2 = (Map<String, Object>) tab2Elements.get("fieldset2");
        Assert.assertEquals("fieldset2", fieldset2.get("name"));
        Assert.assertEquals(ViewItemGroup.FIELDSET_ROLE, fieldset2.get("role"));
        Map<String, Object> fieldset2Elements = (Map<String, Object>) fieldset2.get("elements");
        Map<String, Object> definition8 = (Map<String, Object>) fieldset2Elements.get("eight");
        Assert.assertEquals("eight", definition8.get("name"));
        Assert.assertEquals("date", definition8.get("type"));
        Map<String, Object> definition9 = (Map<String, Object>) fieldset2Elements.get("nine");
        Assert.assertEquals("nine", definition9.get("name"));
        Assert.assertEquals("string", definition9.get("type"));
    }

    private View _createViewForJSONTest() throws Exception
    {
        String role = ModelItemTypeExtensionPoint.ROLE_CONFIG;
        Model model = Model.of("json", __MODELS_FAMILY_ID,
                DefaultElementDefinition.of("one", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("two", false, ModelItemTypeConstants.LONG_TYPE_ID, role),
                DefaultElementDefinition.of("three", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role),
                ModelItemGroup.of("group",
                        DefaultElementDefinition.of("four", false, ModelItemTypeConstants.DATE_TYPE_ID, role),
                        DefaultElementDefinition.of("five", false, ModelItemTypeConstants.STRING_TYPE_ID, role)),
                DefaultElementDefinition.of("six", false, ModelItemTypeConstants.LONG_TYPE_ID, role),
                DefaultElementDefinition.of("seven", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role),
                DefaultElementDefinition.of("eight", false, ModelItemTypeConstants.DATE_TYPE_ID, role),
                DefaultElementDefinition.of("nine", false, ModelItemTypeConstants.STRING_TYPE_ID, role));

        // Create the view
        View view = new View();
        view.setName("view");
        
        ViewElement element1 = new ViewElement();
        element1.setDefinition((ElementDefinition) model.getModelItem("one"));
        view.addViewItem(element1);
        
        SimpleViewItemGroup tab1 = new SimpleViewItemGroup();
        tab1.setName("tab1");
        tab1.setRole(ViewItemGroup.TAB_ROLE);
        view.addViewItem(tab1);
        
        ViewElement element2 = new ViewElement();
        element2.setDefinition((ElementDefinition) model.getModelItem("two"));
        tab1.addViewItem(element2);
        
        ViewElement element3 = new ViewElement();
        element3.setDefinition((ElementDefinition) model.getModelItem("three"));
        tab1.addViewItem(element3);
        
        ModelViewItemGroup<ModelItemGroup> group = new ModelViewItemGroup<>();
        group.setRole(ViewItemGroup.FIELDSET_ROLE);
        group.setDefinition((ModelItemGroup) model.getModelItem("group"));
        tab1.addViewItem(group);
        
        ViewElement element4 = new ViewElement();
        element4.setDefinition((ElementDefinition) model.getModelItem("group/four"));
        group.addViewItem(element4);
        
        ViewElement element5 = new ViewElement();
        element5.setDefinition((ElementDefinition) model.getModelItem("group/five"));
        group.addViewItem(element5);
        
        ViewElement element6 = new ViewElement();
        element6.setDefinition((ElementDefinition) model.getModelItem("six"));
        view.addViewItem(element6);
        
        SimpleViewItemGroup tab2 = new SimpleViewItemGroup();
        tab2.setName("tab2");
        tab2.setRole(ViewItemGroup.TAB_ROLE);
        view.addViewItem(tab2);
        
        ViewElement element7 = new ViewElement();
        element7.setDefinition((ElementDefinition) model.getModelItem("seven"));
        tab2.addViewItem(element7);
        
        SimpleViewItemGroup fieldset2 = new SimpleViewItemGroup();
        fieldset2.setName("fieldset2");
        fieldset2.setRole(ViewItemGroup.FIELDSET_ROLE);
        tab2.addViewItem(fieldset2);
        
        ViewElement element8 = new ViewElement();
        element8.setDefinition((ElementDefinition) model.getModelItem("eight"));
        fieldset2.addViewItem(element8);
        
        ViewElement element9 = new ViewElement();
        element9.setDefinition((ElementDefinition) model.getModelItem("nine"));
        fieldset2.addViewItem(element9);
        
        return view;
    }
    
    /**
     * Tests the {@link View#includeView(View)} method. The included view contains only elements that are not present in the view to test
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncludeViewWithOnlyNewItems() throws Exception
    {
        Model model = _createSimpleModelWithFourElements("include_only_new_items");
        View view = View.of(model, "one", "two");
        View viewToInclude = View.of(model, "three", "four"); // The view to include contains only elements that are not present in the view to test
        
        // Include the second view in the first one
        view.includeView(viewToInclude);
        
        // Test that the view contains all four elements
        List<ViewItem> items = view.getViewItems();
        Assert.assertEquals(4, items.size());
        Assert.assertEquals("one", ((ViewElement) items.get(0)).getDefinition().getName());
        Assert.assertEquals("two", ((ViewElement) items.get(1)).getDefinition().getName());
        Assert.assertEquals("three", ((ViewElement) items.get(2)).getDefinition().getName());
        Assert.assertEquals("four", ((ViewElement) items.get(3)).getDefinition().getName());
    }
    
    /**
     * Tests the {@link View#includeView(View)} method. The included view contains an element that is already present in the view to test
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncludeViewWithACommonItem() throws Exception
    {
        Model model = _createSimpleModelWithFourElements("include_common_item");
        View view = View.of(model, "one", "two");
        View viewToInclude = View.of(model, "two", "three"); // The view to include the element "two", that is already present in the view to test
        
        // Include the second view in the first one
        view.includeView(viewToInclude);
        
        // Test that the element "two" appears only once
        List<ViewItem> items = view.getViewItems();
        Assert.assertEquals(3, items.size());
        Assert.assertEquals("one", ((ViewElement) items.get(0)).getDefinition().getName());
        Assert.assertEquals("two", ((ViewElement) items.get(1)).getDefinition().getName());
        Assert.assertEquals("three", ((ViewElement) items.get(2)).getDefinition().getName());
    }
    
    /**
     * Tests the {@link View#includeView(View)} method. The included view contains an element that is already present in the view to test
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncludeViewWithAnItemSeveralTimesInIncludedView() throws Exception
    {
        Model model = _createSimpleModelWithFourElements("include_an_item_several_times");
        View view = View.of(model, "one");
        View viewToInclude = View.of(model, "two", "two", "three"); // The view to include twice the element "two"
        
        // Include the second view in the first one
        view.includeView(viewToInclude);
        
        // Test that the element "two" appears twice
        List<ViewItem> items = view.getViewItems();
        Assert.assertEquals(4, items.size());
        Assert.assertEquals("one", ((ViewElement) items.get(0)).getDefinition().getName());
        Assert.assertEquals("two", ((ViewElement) items.get(1)).getDefinition().getName());
        Assert.assertEquals("two", ((ViewElement) items.get(2)).getDefinition().getName());
        Assert.assertEquals("three", ((ViewElement) items.get(3)).getDefinition().getName());
    }
    
    /**
     * Tests the {@link View#includeView(View)} method. The included view contains a group with elements that are already present in the view to test
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncludeViewWithACommonItemInAGroup() throws Exception
    {
        Model model = _createSimpleModelWithFourElements("include_common_item_in_a_group");
        View view = View.of(model, "one", "two");
        
        // The view to include contains in a group the element "two", that is already present in the view to test
        View viewToInclude = new View();
        
        SimpleViewItemGroup group = new SimpleViewItemGroup();
        group.setName("group");
        viewToInclude.addViewItem(group);

        ViewElement element2 = new ViewElement();
        element2.setDefinition((ElementDefinition) model.getModelItem("two"));
        group.addViewItem(element2);
        
        ViewElement element3 = new ViewElement();
        element3.setDefinition((ElementDefinition) model.getModelItem("three"));
        group.addViewItem(element3);
        
        // Include the second view in the first one
        view.includeView(viewToInclude);
        
        // Test that the element "two" appears only once
        List<ViewItem> items = view.getViewItems();
        Assert.assertEquals(3, items.size());
        Assert.assertEquals("one", ((ViewElement) items.get(0)).getDefinition().getName());
        Assert.assertEquals("two", ((ViewElement) items.get(1)).getDefinition().getName());
        
        Assert.assertEquals("group", ((SimpleViewItemGroup) items.get(2)).getName());
        List<ViewItem> groupChildren = ((SimpleViewItemGroup) items.get(2)).getViewItems();
        Assert.assertEquals(1, groupChildren.size());
        Assert.assertEquals("three", ((ViewElement) groupChildren.get(0)).getDefinition().getName());
    }
    

    
    /**
     * Tests the {@link View#includeView(View)} method. The included view contains a group with elements that are already present in the view to test
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncludeViewWithAnItemSeveralTimeInAGroupInIncludedView() throws Exception
    {
        Model model = _createSimpleModelWithFourElements("include_an_item_several_times_in_a_group");
        View view = View.of(model, "one");
        
        // The view to include contains twice the element "two", once in a group
        View viewToInclude = new View();

        ViewElement element2 = new ViewElement();
        element2.setDefinition((ElementDefinition) model.getModelItem("two"));
        viewToInclude.addViewItem(element2);
        
        SimpleViewItemGroup group = new SimpleViewItemGroup();
        group.setName("group");
        viewToInclude.addViewItem(group);

        ViewElement element22 = new ViewElement();
        element22.setDefinition((ElementDefinition) model.getModelItem("two"));
        group.addViewItem(element22);
        
        ViewElement element3 = new ViewElement();
        element3.setDefinition((ElementDefinition) model.getModelItem("three"));
        group.addViewItem(element3);
        
        // Include the second view in the first one
        view.includeView(viewToInclude);
        
        // Test that the element "two" appears only once
        List<ViewItem> items = view.getViewItems();
        Assert.assertEquals(3, items.size());
        Assert.assertEquals("one", ((ViewElement) items.get(0)).getDefinition().getName());
        Assert.assertEquals("two", ((ViewElement) items.get(1)).getDefinition().getName());
        
        Assert.assertEquals("group", ((SimpleViewItemGroup) items.get(2)).getName());
        List<ViewItem> groupChildren = ((SimpleViewItemGroup) items.get(2)).getViewItems();
        Assert.assertEquals(2, groupChildren.size());
        Assert.assertEquals("two", ((ViewElement) groupChildren.get(0)).getDefinition().getName());
        Assert.assertEquals("three", ((ViewElement) groupChildren.get(1)).getDefinition().getName());
    }
    
    private Model _createSimpleModelWithFourElements(String modelId) throws UnknownTypeException, BadItemTypeException, ServiceException
    {
        String role = ModelItemTypeExtensionPoint.ROLE_CONFIG;
        return Model.of(modelId, __MODELS_FAMILY_ID,
                DefaultElementDefinition.of("one", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("two", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("three", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("four", false, ModelItemTypeConstants.STRING_TYPE_ID, role));
    }
    
    /**
     * Tests the {@link View#includeView(View)} method. The views contain some groups (simple or with model). Some of the items are commons, others aren't.
     * the result view should be like:
     * title
     * abstract
     * illustration
     *   image
     *   alt-text
     * fieldset-1
     *   string
     *   long
     *   boolean
     *   fieldset-2
     *     date1
     *     date2
     * fieldset-3
     *   group
     *     subGroup
     *       string
     *       long
     * fieldset-4
     * double
     * illustration
     * fieldset-1
     *   stringmult
     *   fieldset-2
     *     datemult
     * fieldset-3
     *   group
     *     subgroup
     *       boolean
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncludeViewWithSeveralGroups() throws Exception
    {
        Model model = _createModelForIncludingTest();
        View view = _createViewForIncludingTest(model);
        View viewToInclude = _createViewToIncludeForIncludingTest(model); // The view to include contains only elements that are not present in the view to test
        
        // Include the second view in the first one
        view.includeView(viewToInclude);
        
        // Test that the view contains all four elements
        List<ViewItem> items = view.getViewItems();
        Assert.assertEquals(10, items.size());
        Assert.assertEquals("title", ((ViewElement) items.get(0)).getDefinition().getName());
        Assert.assertEquals("abstract", ((ViewElement) items.get(1)).getDefinition().getName());
        
        ModelViewItemGroup illustration  = (ModelViewItemGroup) items.get(2);
        Assert.assertEquals("illustration", illustration.getName());
        List<ViewItem> illustrationChildren = illustration.getViewItems(); 
        Assert.assertEquals(2, illustrationChildren.size());
        Assert.assertEquals("image", ((ViewElement) illustrationChildren.get(0)).getDefinition().getName());
        Assert.assertEquals("alt-text", ((ViewElement) illustrationChildren.get(1)).getDefinition().getName());
        
        SimpleViewItemGroup fieldset11 = (SimpleViewItemGroup) items.get(3);
        Assert.assertEquals("fieldset-1", fieldset11.getName());
        List<ViewItem> fieldset11Children = fieldset11.getViewItems();
        Assert.assertEquals(4, fieldset11Children.size());
        Assert.assertEquals("string", ((ViewElement) fieldset11Children.get(0)).getDefinition().getName());
        Assert.assertEquals("long", ((ViewElement) fieldset11Children.get(1)).getDefinition().getName());
        Assert.assertEquals("boolean", ((ViewElement) fieldset11Children.get(2)).getDefinition().getName());
        
        SimpleViewItemGroup fieldset21 = (SimpleViewItemGroup) fieldset11Children.get(3);
        Assert.assertEquals("fieldset-2", fieldset21.getName());
        List<ViewItem> fieldset21Children = fieldset21.getViewItems();
        Assert.assertEquals(2, fieldset21Children.size());
        Assert.assertEquals("date1", ((ViewElement) fieldset21Children.get(0)).getDefinition().getName());
        Assert.assertEquals("date2", ((ViewElement) fieldset21Children.get(1)).getDefinition().getName());
        
        SimpleViewItemGroup fieldset31 = (SimpleViewItemGroup) items.get(4);
        Assert.assertEquals("fieldset-3", fieldset31.getName());
        List<ViewItem> fieldset3Children = fieldset31.getViewItems();
        Assert.assertEquals(1, fieldset3Children.size());
        
        ModelViewItemGroup group1 = (ModelViewItemGroup) fieldset3Children.get(0);
        Assert.assertEquals("group", group1.getName());
        List<ViewItem> group1Children = group1.getViewItems(); 
        Assert.assertEquals(1, group1Children.size());
        
        ModelViewItemGroup subGroup1 = (ModelViewItemGroup) group1Children.get(0);
        Assert.assertEquals("subGroup", subGroup1.getName());
        List<ViewItem> subGroup1Children = subGroup1.getViewItems(); 
        Assert.assertEquals(2, subGroup1Children.size());
        Assert.assertEquals("string", ((ViewElement) subGroup1Children.get(0)).getDefinition().getName());
        Assert.assertEquals("long", ((ViewElement) subGroup1Children.get(1)).getDefinition().getName());
        
        SimpleViewItemGroup fieldset4 = (SimpleViewItemGroup) items.get(5);
        Assert.assertEquals("fieldset-4", fieldset4.getName());
        List<ViewItem> fieldset4Children = fieldset4.getViewItems();
        Assert.assertEquals(0, fieldset4Children.size());
        
        Assert.assertEquals("double", ((ViewElement) items.get(6)).getDefinition().getName());
        
        ModelViewItemGroup illustration2  = (ModelViewItemGroup) items.get(7);
        Assert.assertEquals("illustration", illustration2.getName());
        List<ViewItem> illustration2Children = illustration2.getViewItems(); 
        Assert.assertEquals(0, illustration2Children.size());
        
        SimpleViewItemGroup fieldset12 = (SimpleViewItemGroup) items.get(8);
        Assert.assertEquals("fieldset-1", fieldset12.getName());
        List<ViewItem> fieldset12Children = fieldset12.getViewItems();
        Assert.assertEquals(2, fieldset12Children.size());
        Assert.assertEquals("stringmult", ((ViewElement) fieldset12Children.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup fieldset22 = (SimpleViewItemGroup) fieldset12Children.get(1);
        Assert.assertEquals("fieldset-2", fieldset22.getName());
        List<ViewItem> fieldset22Children = fieldset22.getViewItems();
        Assert.assertEquals(1, fieldset22Children.size());
        Assert.assertEquals("datemult", ((ViewElement) fieldset22Children.get(0)).getDefinition().getName());
        
        SimpleViewItemGroup fieldset32 = (SimpleViewItemGroup) items.get(9);
        Assert.assertEquals("fieldset-3", fieldset32.getName());
        List<ViewItem> fieldset32Children = fieldset32.getViewItems();
        Assert.assertEquals(1, fieldset32Children.size());

        ModelViewItemGroup group2 = (ModelViewItemGroup) fieldset32Children.get(0);
        Assert.assertEquals("group", group2.getName());
        List<ViewItem> group2Children = group2.getViewItems(); 
        Assert.assertEquals(1, group2Children.size());
        
        ModelViewItemGroup subGroup2 = (ModelViewItemGroup) group2Children.get(0);
        Assert.assertEquals("subGroup", subGroup2.getName());
        List<ViewItem> subGroup2Children = subGroup2.getViewItems(); 
        Assert.assertEquals(1, subGroup2Children.size());
        Assert.assertEquals("boolean", ((ViewElement) subGroup2Children.get(0)).getDefinition().getName());
    }
    
    private Model _createModelForIncludingTest() throws UnknownTypeException, BadItemTypeException, ServiceException
    {
        String role = ModelItemTypeExtensionPoint.ROLE_CONFIG;
        return Model.of("include_several_groups", __MODELS_FAMILY_ID,
                DefaultElementDefinition.of("title", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("abstract", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                ModelItemGroup.of("illustration",
                        DefaultElementDefinition.of("image", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                        DefaultElementDefinition.of("alt-text", false, ModelItemTypeConstants.STRING_TYPE_ID, role)),
                DefaultElementDefinition.of("string", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("stringmult", true, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("long", false, ModelItemTypeConstants.LONG_TYPE_ID, role),
                DefaultElementDefinition.of("double", false, ModelItemTypeConstants.DOUBLE_TYPE_ID, role),
                DefaultElementDefinition.of("boolean", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role),
                DefaultElementDefinition.of("date1", false, ModelItemTypeConstants.DATE_TYPE_ID, role),
                DefaultElementDefinition.of("date2", false, ModelItemTypeConstants.DATE_TYPE_ID, role),
                DefaultElementDefinition.of("datemult", true, ModelItemTypeConstants.DATE_TYPE_ID, role),
                ModelItemGroup.of("group",
                        ModelItemGroup.of("subGroup",
                                DefaultElementDefinition.of("string", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                                DefaultElementDefinition.of("boolean", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role),
                                DefaultElementDefinition.of("long", false, ModelItemTypeConstants.LONG_TYPE_ID, role))));
    }
    
    // The created view is the following :
    // - title
    // - abstract
    // - illustration
    //      - image
    //      - alt-text
    // - fieldset-1
    //      - string
    //      - long
    //      - boolean
    //      - fieldset-2
    //          - date1
    //          - date2
    //  - fieldset-3
    //      - group
    //          - subGroup
    //              - string
    //              - long
    // - fieldset-4
    private View _createViewForIncludingTest(Model model)
    {
        View view = new View();
        view.setName("view");
        
        ViewElement title = new ViewElement();
        title.setDefinition((ElementDefinition) model.getModelItem("title"));
        view.addViewItem(title);

        ViewElement abstractElmt = new ViewElement();
        abstractElmt.setDefinition((ElementDefinition) model.getModelItem("abstract"));
        view.addViewItem(abstractElmt);
        
        ModelViewItemGroup<ModelItemGroup> illustration = new ModelViewItemGroup<>();
        illustration.setDefinition((ModelItemGroup) model.getModelItem("illustration"));
        view.addViewItem(illustration);
        
        ViewElement image = new ViewElement();
        image.setDefinition((ElementDefinition) model.getModelItem("illustration/image"));
        illustration.addViewItem(image);
        
        ViewElement altText = new ViewElement();
        altText.setDefinition((ElementDefinition) model.getModelItem("illustration/alt-text"));
        illustration.addViewItem(altText);
        
        SimpleViewItemGroup fieldset1 = new SimpleViewItemGroup();
        fieldset1.setName("fieldset-1");
        fieldset1.setRole(ViewItemGroup.FIELDSET_ROLE);
        view.addViewItem(fieldset1);
        
        ViewElement string = new ViewElement();
        string.setDefinition((ElementDefinition) model.getModelItem("string"));
        fieldset1.addViewItem(string);
        
        ViewElement longElmt = new ViewElement();
        longElmt.setDefinition((ElementDefinition) model.getModelItem("long"));
        fieldset1.addViewItem(longElmt);
        
        ViewElement booleanElmt = new ViewElement();
        booleanElmt.setDefinition((ElementDefinition) model.getModelItem("boolean"));
        fieldset1.addViewItem(booleanElmt);
        
        SimpleViewItemGroup fieldset2 = new SimpleViewItemGroup();
        fieldset2.setName("fieldset-2");
        fieldset2.setRole(ViewItemGroup.FIELDSET_ROLE);
        fieldset1.addViewItem(fieldset2);
        
        ViewElement date1 = new ViewElement();
        date1.setDefinition((ElementDefinition) model.getModelItem("date1"));
        fieldset2.addViewItem(date1);
        
        ViewElement date2 = new ViewElement();
        date2.setDefinition((ElementDefinition) model.getModelItem("date2"));
        fieldset2.addViewItem(date2);
        
        SimpleViewItemGroup fieldset3 = new SimpleViewItemGroup();
        fieldset3.setName("fieldset-3");
        fieldset3.setRole(ViewItemGroup.FIELDSET_ROLE);
        view.addViewItem(fieldset3);
        
        ModelViewItemGroup<ModelItemGroup> group = new ModelViewItemGroup<>();
        group.setDefinition((ModelItemGroup) model.getModelItem("group"));
        fieldset3.addViewItem(group);
        
        ModelViewItemGroup<ModelItemGroup> subGroup = new ModelViewItemGroup<>();
        subGroup.setDefinition((ModelItemGroup) model.getModelItem("group/subGroup"));
        group.addViewItem(subGroup);
        
        ViewElement compStr = new ViewElement();
        compStr.setDefinition((ElementDefinition) model.getModelItem("group/subGroup/string"));
        subGroup.addViewItem(compStr);
        
        ViewElement compLong = new ViewElement();
        compLong.setDefinition((ElementDefinition) model.getModelItem("group/subGroup/long"));
        subGroup.addViewItem(compLong);
        
        SimpleViewItemGroup fieldset4 = new SimpleViewItemGroup();
        fieldset4.setName("fieldset-4");
        fieldset4.setRole(ViewItemGroup.FIELDSET_ROLE);
        view.addViewItem(fieldset4);
        
        return view;
    }
    
    // The created view is the following :
    // - title
    // - double
    // - illustration
    //      - alt-text
    // - fieldset-1
    //      - string
    //      - stringmult
    //      - fieldset-2
    //          - date1
    //          - datemult
    //  - fieldset-3
    //      - group
    //          - subGroup
    //              - boolean
    private View _createViewToIncludeForIncludingTest(Model model)
    {
        View view = new View();
        view.setName("view to include");
        
        ViewElement title = new ViewElement();
        title.setDefinition((ElementDefinition) model.getModelItem("title"));
        view.addViewItem(title);

        ViewElement doubleElmt = new ViewElement();
        doubleElmt.setDefinition((ElementDefinition) model.getModelItem("double"));
        view.addViewItem(doubleElmt);
        
        ModelViewItemGroup<ModelItemGroup> illustration = new ModelViewItemGroup<>();
        illustration.setDefinition((ModelItemGroup) model.getModelItem("illustration"));
        view.addViewItem(illustration);
        
        ViewElement altText = new ViewElement();
        altText.setDefinition((ElementDefinition) model.getModelItem("illustration/alt-text"));
        illustration.addViewItem(altText);
        
        SimpleViewItemGroup fieldset1 = new SimpleViewItemGroup();
        fieldset1.setName("fieldset-1");
        fieldset1.setRole(ViewItemGroup.FIELDSET_ROLE);
        view.addViewItem(fieldset1);
        
        ViewElement string = new ViewElement();
        string.setDefinition((ElementDefinition) model.getModelItem("string"));
        fieldset1.addViewItem(string);
        
        ViewElement stringmult = new ViewElement();
        stringmult.setDefinition((ElementDefinition) model.getModelItem("stringmult"));
        fieldset1.addViewItem(stringmult);
        
        SimpleViewItemGroup fieldset2 = new SimpleViewItemGroup();
        fieldset2.setName("fieldset-2");
        fieldset2.setRole(ViewItemGroup.FIELDSET_ROLE);
        fieldset1.addViewItem(fieldset2);
        
        ViewElement date = new ViewElement();
        date.setDefinition((ElementDefinition) model.getModelItem("date1"));
        fieldset2.addViewItem(date);
        
        ViewElement datemult = new ViewElement();
        datemult.setDefinition((ElementDefinition) model.getModelItem("datemult"));
        fieldset2.addViewItem(datemult);
        
        SimpleViewItemGroup fieldset3 = new SimpleViewItemGroup();
        fieldset3.setName("fieldset-3");
        fieldset3.setRole(ViewItemGroup.FIELDSET_ROLE);
        view.addViewItem(fieldset3);
        
        ModelViewItemGroup<ModelItemGroup> group = new ModelViewItemGroup<>();
        group.setDefinition((ModelItemGroup) model.getModelItem("group"));
        fieldset3.addViewItem(group);
        
        ModelViewItemGroup<ModelItemGroup> subGroup = new ModelViewItemGroup<>();
        subGroup.setDefinition((ModelItemGroup) model.getModelItem("group/subGroup"));
        group.addViewItem(subGroup);
        
        ViewElement compBoolean = new ViewElement();
        compBoolean.setDefinition((ElementDefinition) model.getModelItem("group/subGroup/boolean"));
        subGroup.addViewItem(compBoolean);
        
        return view;
    }
    
    /**
     * Tests the {@link View#of(Model)} method. Nominal case.
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewConstructionFromModel() throws Exception
    {
        Model model = _createModelForViewConstructionsTests("constructor_from_model");
        View view = View.of(model);
        List<ViewItem> viewItems = view.getViewItems();
        
        Assert.assertEquals(6, viewItems.size());
        Assert.assertEquals("stringData", ((ViewElement) viewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("longData", ((ViewElement) viewItems.get(1)).getDefinition().getName());
        Assert.assertEquals("booleanData", ((ViewElement) viewItems.get(2)).getDefinition().getName());
        Assert.assertEquals("dateData", ((ViewElement) viewItems.get(3)).getDefinition().getName());
        
        ViewItemAccessor doubleDataAccessor = (ViewItemAccessor) viewItems.get(4);
        Assert.assertEquals("doubleDataAccessor", ((ViewElement) doubleDataAccessor).getDefinition().getName());
        Assert.assertTrue(doubleDataAccessor.getViewItems().isEmpty());
        
        ModelViewItemGroup group = (ModelViewItemGroup) viewItems.get(5);
        List<ViewItem> groupViewItems = group.getViewItems();
        Assert.assertEquals(2, groupViewItems.size());
        Assert.assertEquals("longData", ((ViewElement) groupViewItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup subGroup = (ModelViewItemGroup) groupViewItems.get(1);
        List<ViewItem> subGroupViewItems = subGroup.getViewItems();
        Assert.assertEquals(3, subGroupViewItems.size());
        Assert.assertEquals("booleanData", ((ViewElement) subGroupViewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("stringData", ((ViewElement) subGroupViewItems.get(1)).getDefinition().getName());
        Assert.assertEquals("dateData", ((ViewElement) subGroupViewItems.get(2)).getDefinition().getName());
    }
    
    /**
     * Tests the {@link View#of(Collection)} method. Nominal case.
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewConstructionFromModels() throws Exception
    {
        Model model1 = _createModelForViewConstructionsTests("constructor_from_models_1");
        Model model2 = _createOtherModelWithNoCommonData("constructor_from_models_2");
        View view = View.of(Arrays.asList(model1, model2));
        List<ViewItem> viewItems = view.getViewItems();
        
        Assert.assertEquals(8, viewItems.size());
        Assert.assertEquals("stringData", ((ViewElement) viewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("longData", ((ViewElement) viewItems.get(1)).getDefinition().getName());
        Assert.assertEquals("booleanData", ((ViewElement) viewItems.get(2)).getDefinition().getName());
        Assert.assertEquals("dateData", ((ViewElement) viewItems.get(3)).getDefinition().getName());
        
        ViewItemAccessor doubleDataAccessor = (ViewItemAccessor) viewItems.get(4);
        Assert.assertEquals("doubleDataAccessor", ((ViewElement) doubleDataAccessor).getDefinition().getName());
        Assert.assertTrue(doubleDataAccessor.getViewItems().isEmpty());
        
        ModelViewItemGroup group = (ModelViewItemGroup) viewItems.get(5);
        List<ViewItem> groupViewItems = group.getViewItems();
        Assert.assertEquals(2, groupViewItems.size());
        Assert.assertEquals("longData", ((ViewElement) groupViewItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup subGroup = (ModelViewItemGroup) groupViewItems.get(1);
        List<ViewItem> subGroupViewItems = subGroup.getViewItems();
        Assert.assertEquals(3, subGroupViewItems.size());
        Assert.assertEquals("booleanData", ((ViewElement) subGroupViewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("stringData", ((ViewElement) subGroupViewItems.get(1)).getDefinition().getName());
        Assert.assertEquals("dateData", ((ViewElement) subGroupViewItems.get(2)).getDefinition().getName());

        Assert.assertEquals("otherStringData", ((ViewElement) viewItems.get(6)).getDefinition().getName());
        Assert.assertEquals("otherBooleanData", ((ViewElement) viewItems.get(7)).getDefinition().getName());
    }
    
    /**
     * Tests the {@link View#of(Model, String...)} method. Nominal Case.
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewConstructionFromItemsList() throws Exception
    {
        Model model = _createModelForViewConstructionsTests("constructor_from_items");
        View view = View.of(model, "stringData", "longData", "doubleDataAccessor/stringData", "group/subGroup/booleanData", "group/subGroup/dateData");
        List<ViewItem> viewItems = view.getViewItems();
        
        Assert.assertEquals(4, viewItems.size());
        Assert.assertEquals("stringData", ((ViewElement) viewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("longData", ((ViewElement) viewItems.get(1)).getDefinition().getName());
        
        ViewItemAccessor doubleDataAccessor = (ViewItemAccessor) viewItems.get(2);
        Assert.assertEquals("doubleDataAccessor", ((ViewElement) doubleDataAccessor).getDefinition().getName());
        List<ViewItem> accessedItems = doubleDataAccessor.getViewItems();
        Assert.assertEquals(1, accessedItems.size());
        Assert.assertEquals("stringData", ((ViewElement) accessedItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup group = (ModelViewItemGroup) viewItems.get(3);
        List<ViewItem> groupViewItems = group.getViewItems();
        Assert.assertEquals(1, groupViewItems.size());
        
        ModelViewItemGroup subGroup = (ModelViewItemGroup) groupViewItems.get(0);
        List<ViewItem> subGroupViewItems = subGroup.getViewItems();
        Assert.assertEquals(2, subGroupViewItems.size());
        Assert.assertEquals("booleanData", ((ViewElement) subGroupViewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("dateData", ((ViewElement) subGroupViewItems.get(1)).getDefinition().getName());
    }

    /**
     * Tests the {@link View#of(Model, String...)} method. One item is a group, all its children should be added.
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewConstructionFromItemsListWithAGroup() throws Exception
    {
        Model model = _createModelForViewConstructionsTests("constructor_from_items_with_group");
        View view = View.of(model, "longData", "group/subGroup");
        List<ViewItem> viewItems = view.getViewItems();
        
        Assert.assertEquals(2, viewItems.size());
        Assert.assertEquals("longData", ((ViewElement) viewItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup group = (ModelViewItemGroup) viewItems.get(1);
        Assert.assertEquals("group", group.getDefinition().getName());
        List<ViewItem> groupViewItems = group.getViewItems();
        Assert.assertEquals(1, groupViewItems.size());
        
        ModelViewItemGroup subGroup = (ModelViewItemGroup) groupViewItems.get(0);
        Assert.assertEquals("subGroup", subGroup.getDefinition().getName());
        List<ViewItem> subGroupViewItems = subGroup.getViewItems();
        Assert.assertEquals(3, subGroupViewItems.size());
        Assert.assertEquals("booleanData", ((ViewElement) subGroupViewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("stringData", ((ViewElement) subGroupViewItems.get(1)).getDefinition().getName());
        Assert.assertEquals("dateData", ((ViewElement) subGroupViewItems.get(2)).getDefinition().getName());
    }
    
    /**
     * Tests the {@link View#of(Model, String...)} method. One of the given path does not exist in the model.
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewConstructionFromItemsListWithInexistingData() throws Exception
    {
        Model model = _createModelForViewConstructionsTests("constructor_from_items_with_inexisting_data");
        
        try
        {
            View.of(model, "stringData", "group/inexistingData", "group/subGroup/booleanData");
            Assert.fail("One of the given items does not existing, the view construction should have fail");
        }
        catch (IllegalArgumentException e)
        {
            Assert.assertTrue(e.getMessage().contains("inexistingData"));
        }
    }
    
    /**
     * Tests the {@link View#of(Collection, String...)} method. Nominal case.
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewConstructionFromModelsAndItemsList() throws Exception
    {
        Model model1 = _createModelForViewConstructionsTests("constructor_from_models_1");
        Model model2 = _createOtherModelWithNoCommonData("constructor_from_models_2");
        View view = View.of(Arrays.asList(model1, model2), "longData", "otherStringData", "doubleDataAccessor/stringData", "group/subGroup/booleanData");
        List<ViewItem> viewItems = view.getViewItems();
        
        Assert.assertEquals(4, viewItems.size());
        Assert.assertEquals("longData", ((ViewElement) viewItems.get(0)).getDefinition().getName());
        Assert.assertEquals("otherStringData", ((ViewElement) viewItems.get(1)).getDefinition().getName());
        
        ViewItemAccessor doubleDataAccessor = (ViewItemAccessor) viewItems.get(2);
        Assert.assertEquals("doubleDataAccessor", ((ViewElement) doubleDataAccessor).getDefinition().getName());
        List<ViewItem> accessedItems = doubleDataAccessor.getViewItems();
        Assert.assertEquals(1, accessedItems.size());
        Assert.assertEquals("stringData", ((ViewElement) accessedItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup group = (ModelViewItemGroup) viewItems.get(3);
        List<ViewItem> groupViewItems = group.getViewItems();
        Assert.assertEquals(1, groupViewItems.size());
        
        ModelViewItemGroup subGroup = (ModelViewItemGroup) groupViewItems.get(0);
        List<ViewItem> subGroupViewItems = subGroup.getViewItems();
        Assert.assertEquals(1, subGroupViewItems.size());
        Assert.assertEquals("booleanData", ((ViewElement) subGroupViewItems.get(0)).getDefinition().getName());
    }
    
    /**
     * Tests the {@link View#of(ModelItem...)} methods
     * @throws Exception if an error occurs
     */
    @Test
    public void testViewConstructionFromModelItems() throws Exception
    {
        Model model = _createModelForViewConstructionsTests("constructor_from_model_items");
        View view = View.of(model.getModelItem("stringData"),
                model.getModelItem("group/longData"),
                model.getModelItem("group/subGroup/stringData"));
        
        List<ViewItem> viewItems = view.getViewItems();
        
        Assert.assertEquals(2, viewItems.size());
        Assert.assertEquals("stringData", ((ViewElement) viewItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup group = (ModelViewItemGroup) viewItems.get(1);
        List<ViewItem> groupViewItems = group.getViewItems();
        Assert.assertEquals(2, groupViewItems.size());
        Assert.assertEquals("longData", ((ViewElement) groupViewItems.get(0)).getDefinition().getName());
        
        ModelViewItemGroup subGroup = (ModelViewItemGroup) groupViewItems.get(1);
        List<ViewItem> subGroupViewItems = subGroup.getViewItems();
        Assert.assertEquals(1, subGroupViewItems.size());
        Assert.assertEquals("stringData", ((ViewElement) subGroupViewItems.get(0)).getDefinition().getName());
        
    }
    
    private Model _createModelForViewConstructionsTests(String modelId) throws UnknownTypeException, BadItemTypeException, ServiceException
    {
        String role = ModelItemTypeExtensionPoint.ROLE_CONFIG;
        return Model.of(modelId, __MODELS_FAMILY_ID,
                DefaultElementDefinition.of("stringData", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("longData", false, ModelItemTypeConstants.LONG_TYPE_ID, role),
                DefaultElementDefinition.of("booleanData", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role),
                DefaultElementDefinition.of("dateData", false, ModelItemTypeConstants.DATE_TYPE_ID, role),
                ElementDefinitionAccessorForTests.of("doubleDataAccessor", false,
                        ModelItemTypeConstants.DOUBLE_TYPE_ID, role, Model.of(modelId + "_accessed", __MODELS_FAMILY_ID,
                                DefaultElementDefinition.of("stringData", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                                DefaultElementDefinition.of("longData", false, ModelItemTypeConstants.LONG_TYPE_ID, role))),
                ModelItemGroup.of("group",
                        DefaultElementDefinition.of("longData", false, ModelItemTypeConstants.LONG_TYPE_ID, role),
                        ModelItemGroup.of("subGroup",
                                DefaultElementDefinition.of("booleanData", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role),
                                DefaultElementDefinition.of("stringData", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                                DefaultElementDefinition.of("dateData", false, ModelItemTypeConstants.DATE_TYPE_ID, role))));
    }
    
    private Model _createOtherModelWithNoCommonData(String modelId) throws UnknownTypeException, BadItemTypeException, ServiceException
    {
        String role = ModelItemTypeExtensionPoint.ROLE_CONFIG;
        return Model.of(modelId, __MODELS_FAMILY_ID,
                DefaultElementDefinition.of("otherStringData", false, ModelItemTypeConstants.STRING_TYPE_ID, role),
                DefaultElementDefinition.of("otherBooleanData", false, ModelItemTypeConstants.BOOLEAN_TYPE_ID, role));
    }
}
