/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugin;

import static org.junit.Assert.assertEquals;

import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.configuration.DefaultConfiguration;
import org.junit.Test;

import org.ametys.runtime.plugin.FeatureActivator.PluginsInformation;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;

/**
 * Tests an implementation of {@link AbstractFeatureActivator}.
 */
public abstract class AbstractFeatureActivatorTestCase extends AbstractRuntimeTestCase
{
    static final String __PLUGIN_NAME = "test-feature-activator";
    
    /*
     * A->B->C->D->E
     * C->C-DEP
     * 
     * D-d1 and D-d2 deactivate D
     * D-o1 and D-o2 override D
     * 
     * D-d1->D-d1-DEP
     * D-o1->D-o1-DEP
     * D-d21->D-d2->D-d2-DEP
     * D-o21->D-o2->D-o2-DEP
     */
    
    void _test(Configuration pluginConfiguration, Set<String> expectedActivatedFeatures, Set<String> expectedInactiveFeatures)
    {
        PluginsInformation info = _computeFeatures(pluginConfiguration);
        
        Set<String> activeFeatures = info.getFeatures().keySet();
        Set<String> inactiveFeatures = info.getInactiveFeatures().keySet();
        
        assertEquals(
                expectedActivatedFeatures.stream()
                    .map("/"::concat)
                    .map(__PLUGIN_NAME::concat)
                    .collect(Collectors.toSet()),
                activeFeatures
        );
        assertEquals(
                expectedInactiveFeatures.stream()
                    .map("/"::concat)
                    .map(__PLUGIN_NAME::concat)
                    .collect(Collectors.toSet()), 
                inactiveFeatures 
        );
    }
    
    abstract FeatureActivator _buildFeatureActivator(Configuration pluginConfiguration);
    
    PluginsInformation _computeFeatures(Configuration pluginConfiguration)
    {
        FeatureActivator featureActivator = _buildFeatureActivator(pluginConfiguration);
        PluginsInformation info = featureActivator.computeActiveFeatures(Collections.EMPTY_MAP, false);
        
        Set<String> activeFeatures = info.getFeatures().keySet();
        Set<String> inactiveFeatures = info.getInactiveFeatures().keySet();
        
        _logger.info("Active features: \n{}", activeFeatures);
        _logger.info("Inactive features: \n{}", inactiveFeatures);
        
        return info;
    }
    
    Map<String, Plugin> _buildPlugin(Configuration pluginConfiguration)
    {
        try
        {
            Plugin plugin = new Plugin(__PLUGIN_NAME);
            plugin.configure(pluginConfiguration);
            return Map.of(__PLUGIN_NAME, plugin);
        }
        catch (ConfigurationException e)
        {
            throw new RuntimeException(e);
        }
    }
    
    Configuration _addFeatureConfigToRoot(String name, String depends, String deactivates, String overrides, Boolean passive, DefaultConfiguration rootConfig)
    {
        DefaultConfiguration feature = new DefaultConfiguration("feature");
        rootConfig.addChild(feature);
        feature.setAttribute("name", name);
        if (depends != null)
        {
            feature.setAttribute("depends", depends);
        }
        if (deactivates != null)
        {
            feature.setAttribute("deactivates", deactivates);
        }
        if (overrides != null)
        {
            feature.setAttribute("overrides", overrides);
        }
        if (passive != null && passive)
        {
            feature.setAttribute("passive", "true");
        }
        return feature;
    }
    
    /**
     * Tests a simple dependency chain
     * <br>A->B->C->D->E
     * <br>C->C-DEP
     * @throws Exception if an exception occurs
     */
    @Test
    public void testChain() throws Exception
    {
        DefaultConfiguration rootConfig = new DefaultConfiguration("plugin");
        _addFeatureConfigToRoot("A", "B", null, null, false, rootConfig);
        _addFeatureConfigToRoot("B", "C", null, null, false, rootConfig);
        _addFeatureConfigToRoot("C", "D, C-DEP", null, null, false, rootConfig);
        _addFeatureConfigToRoot("D", "E", null, null, false, rootConfig);
        _addFeatureConfigToRoot("E", null, null, null, true, rootConfig); // passive
        _addFeatureConfigToRoot("C-DEP", null, null, null, false, rootConfig);
        
        _testChain(rootConfig);
    }
    
    abstract void _testChain(Configuration rootConfig) throws Exception;
    
    /**
     * Tests deactivation
     * <br>A->B->C->D->E
     * <br>C->C-DEP
     * <br>D-d1->D-d1-DEP
     * <br>D-d1 deactivates D
     * @throws Exception if an exception occurs
     */
    @Test
    public void testOneDeactivate() throws Exception
    {
        DefaultConfiguration rootConfig = new DefaultConfiguration("plugin");
        _addFeatureConfigToRoot("A", "B", null, null, false, rootConfig);
        _addFeatureConfigToRoot("B", "C", null, null, false, rootConfig);
        _addFeatureConfigToRoot("C", "D, C-DEP", null, null, false, rootConfig);
        _addFeatureConfigToRoot("D", "E", null, null, false, rootConfig);
        _addFeatureConfigToRoot("E", null, null, null, true, rootConfig); // passive
        _addFeatureConfigToRoot("C-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-d1", "D-d1-DEP", "D", null, false, rootConfig);
        _addFeatureConfigToRoot("D-d1-DEP", null, null, null, false, rootConfig);
        
        _testOneDeactivate(rootConfig);
    }
    
    abstract void _testOneDeactivate(Configuration rootConfig) throws Exception;
    
    /**
     * Tests override
     * <br>A->B->C->D->E
     * <br>C->C-DEP
     * <br>D-o1->D-o1-DEP
     * <br>D-o1 overrides D
     * @throws Exception if an exception occurs
     */
    @Test
    public void testOneOverride() throws Exception
    {
        DefaultConfiguration rootConfig = new DefaultConfiguration("plugin");
        _addFeatureConfigToRoot("A", "B", null, null, false, rootConfig);
        _addFeatureConfigToRoot("B", "C", null, null, false, rootConfig);
        _addFeatureConfigToRoot("C", "D, C-DEP", null, null, false, rootConfig);
        _addFeatureConfigToRoot("D", "E", null, null, false, rootConfig);
        _addFeatureConfigToRoot("E", null, null, null, true, rootConfig); // passive
        _addFeatureConfigToRoot("C-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-o1", "D-o1-DEP", null, "D", false, rootConfig);
        _addFeatureConfigToRoot("D-o1-DEP", null, null, null, false, rootConfig);
        
        _testOneOverride(rootConfig);
    }
    
    abstract void _testOneOverride(Configuration rootConfig) throws Exception;
    
    /**
     * Tests deactivation and override at the same time
     * <br>A->B->C->D->E
     * <br>C->C-DEP
     * <br>D-d1->D-d1-DEP
     * <br>D-d1 deactivates D
     * <br>D-o1->D-o1-DEP
     * <br>D-o1 overrides D
     * @throws Exception if an exception occurs
     */
    @Test
    public void testOneDeactivateAndOneOverride() throws Exception
    {
        DefaultConfiguration rootConfig = new DefaultConfiguration("plugin");
        _addFeatureConfigToRoot("A", "B", null, null, false, rootConfig);
        _addFeatureConfigToRoot("B", "C", null, null, false, rootConfig);
        _addFeatureConfigToRoot("C", "D, C-DEP", null, null, false, rootConfig);
        _addFeatureConfigToRoot("D", "E", null, null, false, rootConfig);
        _addFeatureConfigToRoot("E", null, null, null, true, rootConfig); // passive
        _addFeatureConfigToRoot("C-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-d1", "D-d1-DEP", "D", null, false, rootConfig);
        _addFeatureConfigToRoot("D-o1", "D-o1-DEP", null, "D", false, rootConfig);
        _addFeatureConfigToRoot("D-d1-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-o1-DEP", null, null, null, false, rootConfig);
        
        _testOneDeactivateAndOneOverride(rootConfig);
    }
    
    abstract void _testOneDeactivateAndOneOverride(Configuration rootConfig) throws Exception;
    
    /**
     * Tests two deactivations
     * <br>A->B->C->D->E
     * <br>C->C-DEP
     * <br>D-d1->D-d1-DEP
     * <br>D-d1 deactivates D
     * <br>D-d2->D-d2-DEP
     * <br>D-d2 deactivates D
     * @throws Exception if an exception occurs
     */
    @Test
    public void testTwoDeactivates() throws Exception
    {
        DefaultConfiguration rootConfig = new DefaultConfiguration("plugin");
        _addFeatureConfigToRoot("A", "B", null, null, false, rootConfig);
        _addFeatureConfigToRoot("B", "C", null, null, false, rootConfig);
        _addFeatureConfigToRoot("C", "D, C-DEP", null, null, false, rootConfig);
        _addFeatureConfigToRoot("D", "E", null, null, false, rootConfig);
        _addFeatureConfigToRoot("E", null, null, null, true, rootConfig); // passive
        _addFeatureConfigToRoot("C-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-d1", "D-d1-DEP", "D", null, false, rootConfig);
        _addFeatureConfigToRoot("D-d2", "D-d2-DEP", "D", null, false, rootConfig);
        _addFeatureConfigToRoot("D-d1-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-d2-DEP", null, null, null, false, rootConfig);
        
        _testTwoDeactivates(rootConfig);
    }
    
    abstract void _testTwoDeactivates(Configuration rootConfig) throws Exception;
    
    /**
     * Tests two overrides
     * <br>A->B->C->D
     * <br>D-o1 overrides D
     * <br>D-o2 overrides D
     * @throws Exception if an exception occurs
     */
    @Test
    public void testTwoOverrides() throws Exception
    {
        DefaultConfiguration rootConfig = new DefaultConfiguration("plugin");
        _addFeatureConfigToRoot("A", "B", null, null, false, rootConfig);
        _addFeatureConfigToRoot("B", "C", null, null, false, rootConfig);
        _addFeatureConfigToRoot("C", "D", null, null, false, rootConfig);
        _addFeatureConfigToRoot("D", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-o1", null, null, "D", false, rootConfig);
        _addFeatureConfigToRoot("D-o2", null, null, "D", false, rootConfig);
        
        _testTwoOverrides(rootConfig);
    }
    
    abstract void _testTwoOverrides(Configuration rootConfig) throws Exception;
    
    /**
     * Tests two deactivations with one through a dependency
     * <br>A->B->C->D->E
     * <br>C->C-DEP
     * <br>D-d1->D-d1-DEP
     * <br>D-d1 deactivates D
     * <br>D-d21->D-d2->D-d2-DEP
     * <br>D-d2 deactivates D
     * @throws Exception if an exception occurs
     */
    @Test
    public void testTwoDeactivatesThroughDepend() throws Exception
    {
        DefaultConfiguration rootConfig = new DefaultConfiguration("plugin");
        _addFeatureConfigToRoot("A", "B", null, null, false, rootConfig);
        _addFeatureConfigToRoot("B", "C", null, null, false, rootConfig);
        _addFeatureConfigToRoot("C", "D, C-DEP", null, null, false, rootConfig);
        _addFeatureConfigToRoot("D", "E", null, null, false, rootConfig);
        _addFeatureConfigToRoot("E", null, null, null, true, rootConfig); // passive
        _addFeatureConfigToRoot("C-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-d1", "D-d1-DEP", "D", null, false, rootConfig);
        _addFeatureConfigToRoot("D-d2", "D-d2-DEP", "D", null, false, rootConfig);
        _addFeatureConfigToRoot("D-d21", "D-d2", null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-d1-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-d2-DEP", null, null, null, false, rootConfig);
        
        _testTwoDeactivatesThroughDepend(rootConfig);
    }
    
    abstract void _testTwoDeactivatesThroughDepend(Configuration rootConfig) throws Exception;
    
    /**
     * Tests two overrides with one through a dependency
     * <br>A->B->C->D
     * <br>D-o21->D-o2
     * <br>D-o1 overrides D
     * <br>D-o2 overrides D
     * @throws Exception if an exception occurs
     */
    @Test
    public void testTwoOverridesThroughDepend() throws Exception
    {
        DefaultConfiguration rootConfig = new DefaultConfiguration("plugin");
        _addFeatureConfigToRoot("A", "B", null, null, false, rootConfig);
        _addFeatureConfigToRoot("B", "C", null, null, false, rootConfig);
        _addFeatureConfigToRoot("C", "D", null, null, false, rootConfig);
        _addFeatureConfigToRoot("D", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-o1", null, null, "D", false, rootConfig);
        _addFeatureConfigToRoot("D-o2", null, null, "D", false, rootConfig);
        _addFeatureConfigToRoot("D-o21", "D-o2", null, null, false, rootConfig);
        
        _testTwoOverridesThroughDepend(rootConfig);
    }
    
    abstract void _testTwoOverridesThroughDepend(Configuration rootConfig) throws Exception;
    
    /**
     * Tests overrides in cascade (override of an override of an override...)
     * <br>A->B->C->D->E
     * <br>C->C-DEP
     * <br>D-o1->D-o1-DEP
     * <br>D-o1 overrides D
     * <br>D-o2->D-o2-DEP
     * <br>D-o2 overrides D-o1
     * <br>D-o3->D-o3-DEP
     * <br>D-o3 overrides D-o2
     * @throws Exception if an exception occurs
     */
    @Test
    public void testOverrideInCascade() throws Exception
    {
        DefaultConfiguration rootConfig = new DefaultConfiguration("plugin");
        _addFeatureConfigToRoot("A", "B", null, null, false, rootConfig);
        _addFeatureConfigToRoot("B", "C", null, null, false, rootConfig);
        _addFeatureConfigToRoot("C", "D, C-DEP", null, null, false, rootConfig);
        _addFeatureConfigToRoot("D", "E", null, null, false, rootConfig);
        _addFeatureConfigToRoot("E", null, null, null, true, rootConfig); // passive
        _addFeatureConfigToRoot("C-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-o1", "D-o1-DEP", null, "D", false, rootConfig);
        _addFeatureConfigToRoot("D-o1-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-o2", "D-o2-DEP", null, "D-o1", false, rootConfig);
        _addFeatureConfigToRoot("D-o2-DEP", null, null, null, false, rootConfig);
        _addFeatureConfigToRoot("D-o3", "D-o3-DEP", null, "D-o2", false, rootConfig);
        _addFeatureConfigToRoot("D-o3-DEP", null, null, null, false, rootConfig);
        
        _testOverrideInCascade(rootConfig);
    }
    
    abstract void _testOverrideInCascade(Configuration rootConfig) throws Exception;
}
