/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.plugin;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.Collection;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import org.apache.avalon.framework.configuration.Configuration;

import org.ametys.runtime.plugin.IncludePolicyFeatureActivator.IncludedFeature;

/**
 * Tests {@link IncludePolicyFeatureActivator}
 */
public class IncludePolicyFeatureActivatorTestCase extends AbstractFeatureActivatorTestCase
{
    private Collection<IncludedFeature> _includedFeatures;
    
    @Override
    FeatureActivator _buildFeatureActivator(Configuration pluginConfiguration)
    {
        Map<String, Plugin> allPlugins = _buildPlugin(pluginConfiguration);
        Objects.requireNonNull(_includedFeatures, "The @Test forgot to initialize '_includedFeatures'.");
        FeatureActivator featureActivator = new IncludePolicyFeatureActivator(allPlugins, _includedFeatures);
        return featureActivator;
    }
    
    @Override
    void _testChain(Configuration rootConfig) throws Exception
    {
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/B", "B feature")
        );
        _test(rootConfig, 
                Set.of("B", "C", "D", "E", "C-DEP"), 
                Set.of("A"));
    }
    
    @Override
    void _testOneDeactivate(Configuration rootConfig) throws Exception
    {
        // only D
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/D-d1", "because we want to test mechanism of 'deactivates='")
        );
        _test(rootConfig, 
                Set.of("D-d1", "D-d1-DEP"),
                Set.of("A", "B", "C", "D", "E", "C-DEP"));
        
        
        // D and D-d1
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/D", "D feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d1", "because we want to test mechanism of 'deactivates='")
        );
        _test(rootConfig, 
                Set.of("D-d1", "D-d1-DEP"),
                Set.of("A", "B", "C", "D", "E", "C-DEP"));
        
        
        // B and D-d1
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/B", "B feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d1", "because we want to test mechanism of 'deactivates='")
        );
        try
        {
            _computeFeatures(rootConfig);
            fail("It works while it should not.");
        }
        catch (Exception e)
        {
            _logger.info("Exception is: \n", e);
            assertTrue(e instanceof IllegalStateException);
            assertTrue(e.getMessage().contains("Problem of consistency with included features"));
        }
    }
    
    @Override
    void _testOneOverride(Configuration rootConfig) throws Exception
    {
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/B", "B feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-o1", "because we want to test mechanism of 'overrides='")
        );
        _test(rootConfig, 
                Set.of("B", "C", "D-o1", "C-DEP", "D-o1-DEP"),
                Set.of("A", "D", "E"));
    }
    
    @Override
    void _testOneDeactivateAndOneOverride(Configuration rootConfig) throws Exception
    {
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/B", "B feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d1", "because we want to test mechanism of 'deactivates='"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-o1", "because we want to test mechanism of 'overrides='")
        );
        _test(rootConfig, 
                Set.of("B", "C", "D-d1", "D-o1", "C-DEP", "D-d1-DEP", "D-o1-DEP"),
                Set.of("A", "D", "E"));
    }
    
    @Override
    void _testTwoDeactivates(Configuration rootConfig) throws Exception
    {
        // D-d1 and D-d2
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/D-d1", "first deactivation"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d2", "second deactivation")
        );
        _test(rootConfig, 
                Set.of("D-d1", "D-d2", "D-d1-DEP", "D-d2-DEP"),
                Set.of("A", "B", "C", "D", "E", "C-DEP"));
        
        
        // D and D-d1 and D-d2
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/D", "D feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d1", "first deactivation"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d2", "second deactivation")
        );
        _test(rootConfig, 
                Set.of("D-d1", "D-d2", "D-d1-DEP", "D-d2-DEP"),
                Set.of("A", "B", "C", "D", "E", "C-DEP"));
        
        // B and D-d1 and D-d2
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/B", "B feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d1", "first deactivation"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d2", "second deactivation")
        );
        try
        {
            _computeFeatures(rootConfig);
            fail("It works while it should not.");
        }
        catch (Exception e)
        {
            _logger.info("Exception is: \n", e);
            assertTrue(e instanceof IllegalStateException);
            assertTrue(e.getMessage().contains("Problem of consistency with included features"));
        }
    }
    
    @Override
    void _testTwoOverrides(Configuration rootConfig) throws Exception
    {
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/B", "B feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-o1", "first override"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-o2", "second override")
        );
        try
        {
            _computeFeatures(rootConfig);
            fail("It works while it should not.");
        }
        catch (Exception e)
        {
            _logger.info("Exception is: \n", e);
            assertTrue(e instanceof IllegalStateException);
            assertTrue(e.getMessage().contains("It cannot be overridden by 2 features"));
        }
    }
    
    @Override
    void _testTwoDeactivatesThroughDepend(Configuration rootConfig) throws Exception
    {
        // D-d1 and D-d21
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/D-d1", "first deactivation"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d21", "second deactivation with D-d21->D-d2")
        );
        _test(rootConfig, 
                Set.of("D-d1", "D-d21", "D-d2", "D-d1-DEP", "D-d2-DEP"),
                Set.of("A", "B", "C", "D", "E", "C-DEP"));
        
        
        // D and D-d1 and D-d21
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/D", "D feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d1", "first deactivation"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d21", "second deactivation with D-d21->D-d2")
        );
        _test(rootConfig, 
                Set.of("D-d1", "D-d21", "D-d2", "D-d1-DEP", "D-d2-DEP"),
                Set.of("A", "B", "C", "D", "E", "C-DEP"));
        
        
        // B and D-d1 and D-d21
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/B", "B feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d1", "first deactivation"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-d21", "second deactivation with D-d21->D-d2")
        );
        try
        {
            _computeFeatures(rootConfig);
            fail("It works while it should not.");
        }
        catch (Exception e)
        {
            _logger.info("Exception is: \n", e);
            assertTrue(e instanceof IllegalStateException);
            assertTrue(e.getMessage().contains("Problem of consistency with included features"));
        }
    }
    
    @Override
    void _testTwoOverridesThroughDepend(Configuration rootConfig) throws Exception
    {
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/B", "B feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-o1", "first override"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-o21", "second override with D-o21->D-o2")
        );
        try
        {
            _computeFeatures(rootConfig);
            fail("It works while it should not.");
        }
        catch (Exception e)
        {
            _logger.info("Exception is: \n", e);
            assertTrue(e instanceof IllegalStateException);
            assertTrue(e.getMessage().contains("It cannot be overridden by 2 features"));
        }
    }
    
    @Override
    void _testOverrideInCascade(Configuration rootConfig) throws Exception
    {
        _includedFeatures = Set.of(
                IncludedFeature.of(__PLUGIN_NAME + "/B", "B feature"),
                IncludedFeature.of(__PLUGIN_NAME + "/D-o2", "because we want to test mechanism of 'overrides='")
        );
        _test(rootConfig, 
                Set.of("B", "C", "D-o3", "C-DEP", "D-o3-DEP"),
                Set.of("A", "D", "E", "D-o1", "D-o1-DEP", "D-o2", "D-o2-DEP"));
    }
}
