/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test;

import java.io.File;
import java.io.FileInputStream;
import java.sql.Connection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.group.GroupDirectoryDAO;
import org.ametys.core.group.directory.GroupDirectory;
import org.ametys.core.script.SQLScriptHelper;
import org.ametys.core.user.directory.UserDirectory;
import org.ametys.core.user.population.UserPopulation;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.plugins.core.group.GroupHelper;
import org.ametys.plugins.core.impl.user.directory.JdbcUserDirectory;
import org.ametys.plugins.core.impl.user.directory.JdbcUserDirectory.StrongPasswordRequirements;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Abstract test case for jdbc Runtime test cases.
 */
@NeedFeatures({
    "core/runtime.groups.manager", "core/core.groups.helper", 
    "core-impl/runtime.user.directory.impl.jdbc", 
    "core-impl/runtime.group.directory.impl.jdbc"
})
public abstract class AbstractJDBCTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    /** The group directory DAO */
    protected static GroupDirectoryDAO _groupDirectoryDAO;
    /** The user directory */
    protected static UserDirectory _userDirectory;
    /** The group directory */
    protected static GroupDirectory _groupDirectory;
    /** The group helper */
    protected static GroupHelper _groupHelper;
    /** The User Population DAO */
    protected static UserPopulationDAO _userPopulationDAO;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _userPopulationDAO = (UserPopulationDAO) manager.lookup(UserPopulationDAO.ROLE);
        _groupDirectoryDAO = (GroupDirectoryDAO) manager.lookup(GroupDirectoryDAO.ROLE);
        _groupHelper = (GroupHelper) manager.lookup(GroupHelper.ROLE);
    }
    
    /**
     * Reset the database and insert the scripts. Needs the Init class.
     * @param scripts A list of script to play on db. Scripts should reset tables.
     * @throws Exception if an error occurs
     */
    protected void _setDatabase(List<File> scripts) throws Exception
    {
        Connection connection = null;
        
        try
        {
            connection = ConnectionHelper.getConnection("SQL-test");
            
            for (File script : scripts)
            {
                SQLScriptHelper.runScript(connection, new FileInputStream(script));
            }
        }
        finally 
        {
            ConnectionHelper.cleanup(connection);
        }
    }
    
    /**
     * Creates the JDBC User Directory
     * @return The created JDBC User Directory
     * @throws Exception if an error occurs
     */
    protected UserDirectory _createUserDirectory() throws Exception
    {
        String udModelId = "org.ametys.plugins.core.user.directory.Jdbc";
        Map<String, Object> userDirectoryParameters = new LinkedHashMap<>();
        userDirectoryParameters.put("udModelId", udModelId);
        userDirectoryParameters.put(udModelId + "$" + JdbcUserDirectory.DATASOURCE_PARAM_NAME, "SQL-test");
        userDirectoryParameters.put(udModelId + "$" + JdbcUserDirectory.USERS_TABLE_PARAM_NAME, "Users");
        userDirectoryParameters.put(udModelId + "$" + JdbcUserDirectory.USE_STRONG_PASSWORD_PARAM_NAME, true);
        userDirectoryParameters.put(udModelId + "$" + JdbcUserDirectory.USE_STRONG_PASSWORD_MIN_LENGTH_PARAM_NAME, 8);
        userDirectoryParameters.put(udModelId + "$" + JdbcUserDirectory.USE_STRONG_PASSWORD_MIN_NUMBERS_PARAM_NAME, 1);
        userDirectoryParameters.put(udModelId + "$" + JdbcUserDirectory.USE_STRONG_PASSWORD_MIN_LOWERCASE_PARAM_NAME, 1);
        userDirectoryParameters.put(udModelId + "$" + JdbcUserDirectory.USE_STRONG_PASSWORD_MIN_UPPERCASE_PARAM_NAME, 1);
        userDirectoryParameters.put(udModelId + "$" + JdbcUserDirectory.USE_STRONG_PASSWORD_MIN_SPECIALS_PARAM_NAME, 1);
        userDirectoryParameters.put(udModelId + "$" + JdbcUserDirectory.USE_STRONG_PASSWORD_SPECIAL_CHARACTERS_PARAM_NAME, StrongPasswordRequirements.DEFAULT_SPECIAL_CHARACTERS);
        _userPopulationDAO.add("foo", "Foo Population", Collections.singletonList(userDirectoryParameters), Collections.emptyList());
        UserPopulation population = _userPopulationDAO.getUserPopulation("foo");
        return population.getUserDirectories().stream().findFirst().get();
    }
    
    /**
     * Creates the JDBC Group Directory
     * @return The created JDBC Group Directory
     * @throws Exception if an error occurs
     */
    protected GroupDirectory _createGroupDirectory() throws Exception
    {
        String modelId = "org.ametys.plugins.core.group.directory.Jdbc";
        
        Map<String, String> parameters = new LinkedHashMap<>();
        parameters.put(modelId + "$" + "runtime.groups.jdbc.datasource", "SQL-test");
        parameters.put(modelId + "$" + "runtime.groups.jdbc.list.table", "Groups");
        parameters.put(modelId + "$" + "runtime.groups.jdbc.composition.table", "Groups_Users");
        
        String groupDirectoryId = "jdbc-groups";
        _groupDirectoryDAO.add(groupDirectoryId, "JDBC", modelId, parameters);
        
        return _groupDirectoryDAO.getGroupDirectory(groupDirectoryId);
    }
    
    /**
     * Deletes, if it exists, the Group Directory
     * @throws Exception if an error occurs
     */
    protected void _deleteGroupDirectory() throws Exception
    {
        if (_groupDirectory != null)
        {
            _groupDirectoryDAO.remove(_groupDirectory.getId());
        }
    }
}
