/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import org.apache.avalon.framework.service.ServiceManager;
import org.apache.excalibur.xml.xpath.XPathProcessor;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.util.AmetysXSLTHelper;
import org.ametys.runtime.data.AmetysHomeLock;
import org.ametys.runtime.servlet.RuntimeConfig;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.AmetysHomeLockWrapper;
import org.ametys.runtime.test.framework.CocoonHelper;
import org.ametys.runtime.test.framework.ScriptHelper;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * Tests the RuntimeConfig
 */
public class RuntimeConfigTestCase extends AbstractRuntimeTestCase
{
    /**
     * RuntimeConfig test
     * @throws Exception if an error occurs
     */
    @Test
    public void testRuntimeConfigFail() throws Exception
    {
        // The file runtime00_syntax-error.xml is not valid against the schema
        TestConfigSet configSet = new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime00.xml")
                .withWebappPath("test/environments/webapp1");
        AmetysHomeLockWrapper ametysHomeLockWrapper = new AmetysHomeLockWrapper();
        try
        {
            CocoonHelper.configureRuntime(configSet, _staticLogger, ametysHomeLockWrapper);
            Assert.assertTrue("Must not have been validated", RuntimeConfig.getInstance().isInvalid());
        }
        finally
        {
            ametysHomeLockWrapper.getLock().ifPresent(AmetysHomeLock::release);
        }
    }
    
    /**
     * RuntimeConfig test
     * @throws Exception if an error occurs
     */
    @Test
    public void testRuntimeConfigOk() throws Exception
    {
        TestConfigSet configSet = new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime01.xml")
                .withWebappPath("test/environments/webapp1");
        AmetysHomeLockWrapper ametysHomeLockWrapper = new AmetysHomeLockWrapper();
        try
        {
            CocoonHelper.configureRuntime(configSet, _staticLogger, ametysHomeLockWrapper);
            
            RuntimeConfig config = RuntimeConfig.getInstance();
            
            Assert.assertEquals(config.getDefaultWorkspace(), "myworkspace");
            
            Assert.assertEquals(config.getInitClassName(), "org.ametys.runtime.test.NoOpInit");
            
            Assert.assertEquals(2, config.getPluginsLocations().size());
            Assert.assertTrue(config.getPluginsLocations().contains("plugins/"));
            Assert.assertTrue(config.getPluginsLocations().contains("test"));
            
            Assert.assertEquals(1, config.getExcludedFeatures().size());
            Assert.assertTrue(config.getExcludedFeatures().contains("test"));
            
            Assert.assertEquals(1, config.getExcludedWorkspaces().size());
            Assert.assertTrue(config.getExcludedWorkspaces().contains("test"));
            
            Assert.assertEquals(2, config.getComponents().size());
            
            Assert.assertEquals(config.getApplicationVersion(), "2.0.1");
            Date date = config.getApplicationBuildDate();
            Calendar calendar = new GregorianCalendar();
            calendar.setTime(date);
            
            Assert.assertEquals(calendar.get(Calendar.MINUTE), 15);
            Assert.assertEquals(calendar.get(Calendar.MONTH), 0);
        }
        finally
        {
            ametysHomeLockWrapper.getLock().ifPresent(AmetysHomeLock::release);
        }
    }
    
    /**
     * Tests the VersionGenerator and VersionHandler process
     * @throws Exception if an error occurs
     */
    @Test
    public void testVersion() throws Exception
    {
        TestConfigSet configSet = new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime01.xml")
                .withConfigFile("test/environments/configs/config1.xml")
                .withSqlFile(ScriptHelper.getDataSourceFile(ConnectionHelper.DATABASE_HSQLDB))
                .withWebappPath("test/environments/webapp1");
        try
        {
            startCocoon(configSet);
            ServiceManager serviceManager = getCocoonWrapper().getServiceManager();
            
            XPathProcessor xpath = (XPathProcessor) serviceManager.lookup(XPathProcessor.ROLE);
            Assert.assertEquals("2.0.1", xpath.evaluateAsString(AmetysXSLTHelper.versions(), "Component[Name='Application']/Version"));
        }
        finally
        {
            stopCocoon();
        }
    }
}
