/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.authentication.token;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.authentication.token.AuthenticationTokenManager;
import org.ametys.core.authentication.token.AuthenticationTokenManager.Token;
import org.ametys.core.user.InvalidModificationException;
import org.ametys.core.user.User;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.User.UserCreationOrigin;
import org.ametys.core.user.directory.ModifiableUserDirectory;
import org.ametys.plugins.core.impl.user.directory.JdbcUserDirectory;
import org.ametys.runtime.test.AbstractJDBCTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;
import org.ametys.runtime.test.framework.Unoptimized;

/**
 * Abstract test case for token test cases.
 */
@Unoptimized
@NeedFeatures("core/runtime.core.sql.authenticationtokens.init")
public abstract class AbstractTokenTestCase extends AbstractJDBCTestCase
{
    /** The type of the token **/
    protected static final String TOKEN_TYPE = "test-token";
    
    /** The comment of the token **/
    protected static final String TOKEN_COMMENT = "A comment for a test token";

    private static final int DEFAULT_SLEEP_DURATION = 4000; // milisecond (should always be > 1000 to avoid to crash one particular test that sets a token for 1sec)
    private static final int DEFAULT_TOKEN_DURATION = 6; // second
    
    private static AuthenticationTokenManager _authenticateTokenManager;
    
    /** Current User */
    private User _user;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _authenticateTokenManager = (AuthenticationTokenManager) manager.lookup(AuthenticationTokenManager.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        _userDirectory = _createUserDirectory();
        _createUser();
    }
    
    /**
     * Test the getting of users on mysql
     * @throws Exception if an error occurs
     */
    @Test
    public void testType() throws Exception
    {
        // JDBC IMPL
        Assert.assertTrue(_userDirectory instanceof JdbcUserDirectory);

        // MODIFIABLE
        Assert.assertTrue(_userDirectory instanceof ModifiableUserDirectory);
    }
    
    private void _createUser() throws InvalidModificationException
    {
        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) _userDirectory;
        
        // Correct additions
        Map<String, String> userInformation;
        
        userInformation = new HashMap<>();
        userInformation.put("login", "test");
        userInformation.put("firstname", "Test");
        userInformation.put("lastname", "TEST");
        userInformation.put("email", "");
        userInformation.put("password", "test@Password123");
        modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);
        _user = _userDirectory.getUser("test");
    }
    
    private void _deleteToken(String token, String context)
    {
        _authenticateTokenManager.deleteTokenByValue(token, context);
        UserIdentity tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertNull("Token is still existing, it should be deleted", tokenUserIdentity);
    }
    
    /**
     * Test the creation and use of a simple token (global, unlimited)
     * @throws Exception something went wrong
     */
    @Test
    public void testSimpleToken() throws Exception
    {
        UserIdentity identity = _user.getIdentity();
        String token = _authenticateTokenManager.generateToken(identity, 0, TOKEN_TYPE, null);
        UserIdentity tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        
        tokenUserIdentity = _authenticateTokenManager.validateToken(token, "context-1");
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        
        _deleteToken(token, null);
    }
    
    /**
     * Test the creation and use of a time limited + auto-renew token
     * @throws Exception something went wrong
     */
    @Test
    public void testRenewableToken() throws Exception
    {
        UserIdentity identity = _user.getIdentity();
        String token = _authenticateTokenManager.generateToken(identity, DEFAULT_TOKEN_DURATION, true, null, null, TOKEN_TYPE, TOKEN_COMMENT);
        UserIdentity tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        
        Thread.sleep(DEFAULT_SLEEP_DURATION);
        //should still be OK
        tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        
        Thread.sleep(DEFAULT_SLEEP_DURATION);
        //should still be OK
        tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        
        Thread.sleep(DEFAULT_SLEEP_DURATION * 2);
        //should not be OK
        tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertNull("Token should not be available anymore", tokenUserIdentity);
    }
    
    /**
     * Test the creation and use of a time limited without auto-renew token
     * @throws Exception something went wrong
     */
    @Test
    public void testNotRenewableToken() throws Exception
    {
        UserIdentity identity = _user.getIdentity();
        String token = _authenticateTokenManager.generateToken(identity, DEFAULT_TOKEN_DURATION, false, null, null, TOKEN_TYPE, TOKEN_COMMENT);
        UserIdentity tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        
        Thread.sleep(DEFAULT_SLEEP_DURATION);
        //should still be OK
        tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        
        Thread.sleep(DEFAULT_SLEEP_DURATION);
        //should not be OK
        tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertNull("Token should not be available anymore", tokenUserIdentity);
    }
    
    /**
     * Test the creation and use of a limited by context
     * @throws Exception something went wrong
     */
    @Test
    public void testContextToken() throws Exception
    {
        UserIdentity identity = _user.getIdentity();
        String token = _authenticateTokenManager.generateToken(identity, 0, false, null, Collections.singleton("context-2"), TOKEN_TYPE, TOKEN_COMMENT);
        
        UserIdentity tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertNull("Token should not be available without context", tokenUserIdentity);
        
        tokenUserIdentity = _authenticateTokenManager.validateToken(token, "context-2");
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);

        tokenUserIdentity = _authenticateTokenManager.validateToken(token, "context-3");
        Assert.assertNull("Token should not be available with another context", tokenUserIdentity);
        
        _deleteToken(token, "context-2");
        
        // Tests on multiple contexts
        Set<String> contexts = new HashSet<>();
        contexts.add("context-3");
        contexts.add("context-4");
        contexts.add("context-5");
        String tokenMultiContext = _authenticateTokenManager.generateToken(identity, 0, false, null, contexts, TOKEN_TYPE, TOKEN_COMMENT);
        tokenUserIdentity = _authenticateTokenManager.validateToken(tokenMultiContext, "context-3");
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        tokenUserIdentity = _authenticateTokenManager.validateToken(tokenMultiContext, "context-4");
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        tokenUserIdentity = _authenticateTokenManager.validateToken(tokenMultiContext, "context-5");
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        tokenUserIdentity = _authenticateTokenManager.validateToken(token, "context-6");
        Assert.assertNull("Token should not be available with another context", tokenUserIdentity);
        
        _deleteToken(tokenMultiContext, "context-3");
    }
    
    /**
     * Test the creation and use of a limited without auto-renew token
     * @throws Exception something went wrong
     */
    @Test
    public void testNbUsesToken() throws Exception
    {
        UserIdentity identity = _user.getIdentity();
        String token = _authenticateTokenManager.generateToken(identity, 0, false, 3, null, TOKEN_TYPE, TOKEN_COMMENT);
        
        UserIdentity tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        
        tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        
        tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);

        tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertNull("Token should not be available anymore", tokenUserIdentity);
    }
    
    /**
     * Test the creation and use of a limited without auto-renew token
     * @throws Exception something went wrong
     */
    @Test
    public void testComplexToken() throws Exception
    {
        UserIdentity identity = _user.getIdentity();
        
        // Usable 3 times, for 3 secondes (can be renewed), on a context
        String token = _authenticateTokenManager.generateToken(identity, DEFAULT_TOKEN_DURATION, true, 3, Collections.singleton("context-4"), TOKEN_TYPE, TOKEN_COMMENT);

        UserIdentity tokenUserIdentity = _authenticateTokenManager.validateToken(token);
        Assert.assertNull("Token should not be available without context", tokenUserIdentity);
        
        tokenUserIdentity = _authenticateTokenManager.validateToken(token, "context-5");
        Assert.assertNull("Token should not be available on this context", tokenUserIdentity);

        tokenUserIdentity = _authenticateTokenManager.validateToken(token, "context-4");
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        Thread.sleep(DEFAULT_SLEEP_DURATION);

        tokenUserIdentity = _authenticateTokenManager.validateToken(token, "context-5");
        Assert.assertNull("Token should not be available on this context", tokenUserIdentity);
        
        tokenUserIdentity = _authenticateTokenManager.validateToken(token, "context-4");
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        Thread.sleep(DEFAULT_SLEEP_DURATION);
        tokenUserIdentity = _authenticateTokenManager.validateToken(token, "context-4");
        Assert.assertEquals("User connected and user found in the token are not the same", identity, tokenUserIdentity);
        Thread.sleep(DEFAULT_SLEEP_DURATION);
        tokenUserIdentity = _authenticateTokenManager.validateToken(token, "context-4");
        Assert.assertNull("Token should not be available anymore", tokenUserIdentity);
    }
    
    /**
     * Test the recuperation of multiple tokens for a user
     * @throws Exception something went wrong
     */
    @Test
    public void testGetTokens() throws Exception
    {
        UserIdentity identity = _user.getIdentity();
        
        // Usable 3 times, for 3 secondes (can be renewed), on a context
        String token1 = _authenticateTokenManager.generateToken(identity, DEFAULT_TOKEN_DURATION, true, 3, Collections.singleton("context-6"), TOKEN_TYPE, TOKEN_COMMENT);
        String token2 = _authenticateTokenManager.generateToken(identity, 0, TOKEN_TYPE, TOKEN_COMMENT);
        String token3 = _authenticateTokenManager.generateToken(identity, 1, TOKEN_TYPE, TOKEN_COMMENT); // minimum duration
        String token4 = _authenticateTokenManager.generateToken(identity, 0, "other-type", TOKEN_COMMENT);
        
        UserIdentity falseId = new UserIdentity("someLogin", "somePopulation");
        String token5 = _authenticateTokenManager.generateToken(falseId, 1, TOKEN_TYPE, TOKEN_COMMENT);
        
        Thread.sleep(DEFAULT_SLEEP_DURATION);

        List<Token> tokens = _authenticateTokenManager.getTokens(identity, TOKEN_TYPE);
        
        Assert.assertEquals("There should be 2 tokens only", 2, tokens.size());

        Assert.assertEquals("Comment should be equal", TOKEN_COMMENT, tokens.get(0).getComment());
        
        _deleteToken(token1, "context-6");
        _deleteToken(token2, null);
        _deleteToken(token3, null);
        _deleteToken(token4, null);
        _deleteToken(token5, null);
        
        tokens = _authenticateTokenManager.getTokens(identity, TOKEN_TYPE);
        
        Assert.assertEquals("There should be 0 tokens left", 0, tokens.size());
    }
}
