/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.runtime.test.cache;

import java.util.HashMap;
import java.util.Map;

import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.cache.Cache;
import org.ametys.core.cache.CacheException;
import org.ametys.plugins.core.impl.cache.AbstractCacheKey;
import org.ametys.plugins.core.impl.cache.GuavaCache;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.test.framework.AbstractBaseTestCase;

import junit.framework.TestCase;

/**
 * {@link TestCase} of {@link AbstractCacheManager} helper class.
 */
public class CacheTestCase extends AbstractBaseTestCase
{
    static class MyStringKey extends AbstractCacheKey
    {
        MyStringKey(String layer1, String layer2, String layer3)
        {
            super(layer1, layer2, layer3);
        }

        static MyStringKey of(String layer1, String layer2, String layer3)
        {
            return new MyStringKey(layer1, layer2, layer3);
        }

        static MyStringKey of(String layer1, String layer2)
        {
            return new MyStringKey(layer1, layer2, null);
        }

        static MyStringKey of(String layer1)
        {
            return new MyStringKey(layer1, null, null);
        }
    }

    static class MyIntegerThreeKey extends AbstractCacheKey
    {
        MyIntegerThreeKey(Integer layer1, Integer layer2, Integer layer3)
        {
            super(layer1, layer2, layer3);
        }

        static MyIntegerThreeKey of(Integer layer1, Integer layer2, Integer layer3)
        {
            return new MyIntegerThreeKey(layer1, layer2, layer3);
        }

        static MyIntegerThreeKey of(Integer layer1, Integer layer2)
        {
            return new MyIntegerThreeKey(layer1, layer2, null);
        }

        static MyIntegerThreeKey of(Integer layer1)
        {
            return new MyIntegerThreeKey(layer1, null, null);
        }
    }

    static class MyIntegerSixKey extends AbstractCacheKey
    {
        MyIntegerSixKey(Integer layer1, Integer layer2, Integer layer3, Integer layer4, Integer layer5, Integer layer6)
        {
            super(layer1, layer2, layer3, layer4, layer5, layer6);
        }

        static MyIntegerSixKey of(Integer layer1, Integer layer2, Integer layer3, Integer layer4, Integer layer5, Integer layer6)
        {
            return new MyIntegerSixKey(layer1, layer2, layer3, layer4, layer5, layer6);
        }

        static MyIntegerSixKey of(Integer layer1, Integer layer2, Integer layer3, Integer layer4)
        {
            return new MyIntegerSixKey(layer1, layer2, layer3, layer4, null, null);
        }

        static MyIntegerSixKey of(Integer layer1, Integer layer2, Integer layer3)
        {
            return new MyIntegerSixKey(layer1, layer2, layer3, null, null, null);
        }

        static MyIntegerSixKey of(Integer layer1, Integer layer2)
        {
            return new MyIntegerSixKey(layer1, layer2, null, null, null, null);
        }

        static MyIntegerSixKey of(Integer layer1)
        {
            return new MyIntegerSixKey(layer1, null, null, null, null, null);
        }
    }

    static class MyIntegerTwoKey extends AbstractCacheKey
    {
        MyIntegerTwoKey(Integer layer1, Integer layer2)
        {
            super(layer1, layer2);
        }

        static MyIntegerTwoKey of(Integer layer1, Integer layer2)
        {
            return new MyIntegerTwoKey(layer1, layer2);
        }

        static MyIntegerTwoKey of(Integer layer1)
        {
            return new MyIntegerTwoKey(layer1, null);
        }
    }

    private static Map<MyIntegerThreeKey, String> _heavyMap;

    private static Map<MyIntegerSixKey, Integer> _pageElementCacheMap;

    private static int _numberOfElementsByLevel = 10;

    private static Cache<MyIntegerThreeKey, String> _cache = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);
    
    private static Cache<MyIntegerSixKey, Integer> _pageElementCache = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);

    private static Cache<MyIntegerSixKey, Integer> _pageElementCache1 = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);

    private static Cache<MyIntegerSixKey, Integer> _pageElementCache2 = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);

    private static Cache<MyIntegerSixKey, Integer> _pageElementCache3 = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);

    private static Cache<MyIntegerSixKey, Integer> _pageElementCache4 = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);

    private static Cache<MyIntegerSixKey, Integer> _pageElementCache5 = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);
    
    /**
     * Initialize some caches for tests
     */
    @BeforeClass
    public static void initBeforeTest()
    {
        _pageElementCacheMap = new HashMap<>();
        for (int i = 0; i < 2; i++)
        {
            for (int j = 0; j < 10; j++)
            {
                for (int k = 0; k < 10; k++)
                {
                    for (int l = 0; l < 1000; l++)
                    {
                        for (int m = 0; m < 1; m++)
                        {
                            for (int n = 0; n < 3; n++)
                            {
                                MyIntegerSixKey u = MyIntegerSixKey.of(i, j, k, l, m, n);
                                _pageElementCacheMap.put(u, 5);
                            }
                        }
                    }
                }
            }
        }

        // _pageElementCacheMap.put(MyIntegerSixKey.of(1, 1, 1, 1, 1, 1), 5);
        _heavyMap = new HashMap<>();
        for (int i = 0; i < _numberOfElementsByLevel; i++)
        {
            for (int j = 0; j < _numberOfElementsByLevel; j++)
            {
                for (int k = 0; k < _numberOfElementsByLevel; k++)
                {
                    _heavyMap.put(MyIntegerThreeKey.of(1 + i, 1 + j, 1 + k), Integer.toString(i) + Integer.toString(j) + Integer.toString(k));
                }
            }
        }

        _cache.putAll(_heavyMap);
        _pageElementCache.putAll(_pageElementCacheMap);
        _pageElementCache1.putAll(_pageElementCacheMap);
        _pageElementCache2.putAll(_pageElementCacheMap);
        _pageElementCache3.putAll(_pageElementCacheMap);
        _pageElementCache4.putAll(_pageElementCacheMap);
        _pageElementCache5.putAll(_pageElementCacheMap);
    }

    /**
     * test put and get on cache
     */
    @Test
    public void testPut()
    {
        Cache<String, String> cache = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);
        cache.put("key", "value");
        Assert.assertEquals("value", cache.get("key"));
    }

    /**
     * test that #get(K key, Function function) compute successfully a function
     * when key is not present
     * 
     * @throws CacheException throw CacheException if function is not computed
     *             successfully
     */
    @Test
    public void testGet() throws CacheException
    {
        Cache<String, String> cache = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);
        Assert.assertFalse(cache.hasKey("key"));
        Assert.assertEquals("value", cache.get("key", k -> "value"));
    }

    /**
     * Test that #get(K key) return empty optional when key is missing
     */
    @Test
    public void testGetEmpty()
    {
        GuavaCache<String, String> cache = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);
        Assert.assertNull(cache.get("key"));
    }

    /**
     * test that key-value are deleted from cache after #invalidate(K key)
     */
    @Test
    public void testInvalidate()
    {
        Cache<String, String> cache = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);
        cache.put("key", "value");
        Assert.assertNotNull(cache.get("key"));
        cache.invalidate("key");
        Assert.assertNull(cache.get("key"));
    }

    /**
     * test put and get with complex key
     */
    @Test
    public void testComplexKey()
    {
        Cache<MyStringKey, String> cache = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);
        cache.put(MyStringKey.of("site1", "page1", "element1"), "111");
        Assert.assertNotNull(cache.get(MyStringKey.of("site1", "page1", "element1")));
    }

    /**
     * test multiple invalidation with complex key
     */
    @Test
    public void testInvalidateNode()
    {
        Cache<MyStringKey, String> cache = new GuavaCache<>(null, null, null, Long.MAX_VALUE, false, null, false);
        cache.put(MyStringKey.of("site1", "page1", "element1"), "111");
        Assert.assertEquals(1, cache.getNumberOfElements());
        cache.put(MyStringKey.of("site1", "page1", "element2"), "112");
        cache.put(MyStringKey.of("site1", "page1", "element3"), "113");
        cache.put(MyStringKey.of("site1", "page1", "element4"), "114");
        cache.put(MyStringKey.of("site1", "page2", "element1"), "121");
        cache.put(MyStringKey.of("site1", "page2", "element2"), "122");
        Assert.assertEquals(6, cache.getNumberOfElements());
        cache.invalidate(MyStringKey.of("site1", "page1"));
        Assert.assertEquals(2, cache.getNumberOfElements());
    }

    /**
     * Test that cache is filled with correct number of elements
     */
    @Test
    public void pageElementCacheNumber()
    {
        Assert.assertEquals(600000, _pageElementCache.getNumberOfElements());
    }

    /**
     * Test that cache is filled with correct number of elements after an
     * eviction at level 1
     */
    @Test
    public void pageElementCacheRemoveNodeLevel1()
    {
        Assert.assertEquals(600000, _pageElementCache1.getNumberOfElements());
        _pageElementCache1.invalidate(MyIntegerSixKey.of(0));
        Assert.assertEquals(600000 - 300000, _pageElementCache1.getNumberOfElements());
    }

    /**
     * Test that cache is filled with correct number of elements after an
     * eviction at level 2
     */
    @Test
    public void pageElementCacheRemoveNodeLevel2()
    {
        Assert.assertEquals(600000, _pageElementCache2.getNumberOfElements());
        _pageElementCache2.invalidate(MyIntegerSixKey.of(0, 0));
        Assert.assertEquals(600000 - 30000, _pageElementCache2.getNumberOfElements());
    }

    /**
     * Test that cache is filled with correct number of elements after an
     * eviction at level 3
     */
    @Test
    public void pageElementCacheRemoveNodeLevel3()
    {
        Assert.assertEquals(600000, _pageElementCache3.getNumberOfElements());
        _pageElementCache3.invalidate(MyIntegerSixKey.of(0, 0, 0));
        Assert.assertEquals(600000 - 3000, _pageElementCache3.getNumberOfElements());
    }

    /**
     * Test that cache is filled with correct number of elements after an
     * eviction at level 4
     */
    @Test
    public void pageElementCacheRemoveNodeLevel4()
    {
        Assert.assertEquals(600000, _pageElementCache4.getNumberOfElements());
        _pageElementCache4.invalidate(MyIntegerSixKey.of(0, 0, 0, 0));
        Assert.assertEquals(600000 - 3, _pageElementCache4.getNumberOfElements());
    }

    /**
     * Complete test : create new cache, add some elements, invalidate some
     * elements, try to get present and invalidated element
     */
    @Test
    public void completeTest()
    {
        Cache<MyStringKey, String> cache = new GuavaCache<>("id", new I18nizableText("label"), new I18nizableText("description"), Long.MAX_VALUE, false, null, false);
        cache.put(MyStringKey.of("site1", "page1", "element1"), "111");
        Assert.assertEquals(1, cache.getNumberOfElements());
        cache.put(MyStringKey.of("site1", "page1", "element2"), "112");
        cache.put(MyStringKey.of("site1", "page1", "element3"), "113");
        cache.put(MyStringKey.of("site1", "page1", "element4"), "114");
        cache.put(MyStringKey.of("site1", "page2", "element1"), "121");
        cache.put(MyStringKey.of("site1", "page2", "element2"), "122");
        Assert.assertEquals(6, cache.getNumberOfElements());
        cache.invalidate(MyStringKey.of("site1", "page1"));
        Assert.assertEquals(2, cache.getNumberOfElements());
        String presentElement = cache.get(MyStringKey.of("site1", "page2", "element1"));
        Assert.assertNotNull(presentElement);
        String invalidatedElement = cache.get(MyStringKey.of("site1", "page1", "element1"));
        Assert.assertNull(invalidatedElement);
    }

}
