/*
 *  Copyright 2012 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.cocoon;

import java.io.IOException;
import java.io.InputStream;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathFactory;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.environment.Request;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.apache.excalibur.source.SourceUtil;
import org.apache.excalibur.xml.xslt.XSLTProcessor;
import org.junit.Assert;
import org.junit.Test;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

import org.ametys.runtime.cocoon.ThreadSafeTraxProcessor;
import org.ametys.runtime.cocoon.ThreadSafeTraxProcessor.CacheValidity;
import org.ametys.runtime.cocoon.ThreadSafeTraxProcessor.ExtendedTransformerHandlerAndValidity;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * Tests the XSLT Processor and its cache.
 */
@NeedFeatures({
    "core/runtime.core", 
    "test/test",
})
public class TraxProcessorTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static ThreadSafeTraxProcessor _xsltProcessor;
    private static SourceResolver _resolver;
    private static XPath _xPath;
    private static Source _xml;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp2");
    }
    
    @Override
    public boolean withInitialEnvironment()
    {
        return false;
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _xsltProcessor = (ThreadSafeTraxProcessor) manager.lookup(XSLTProcessor.ROLE + "/xalan");
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        _xPath = XPathFactory.newInstance().newXPath();
        _xml = _resolver.resolveURI("resource://org/ametys/runtime/test/cocoon/trax-xml-input.xml");
    }
    
    /**
     * Tests {@link ThreadSafeTraxProcessor} with regular files.
     * @throws Exception if an error occurs.
     */
    @Test
    public void testRegular() throws Exception
    {
        Source xslt = _resolver.resolveURI("resource://org/ametys/runtime/test/cocoon/trax-test1-root.xsl");

        // First request
        enterEnvironment("");
        try
        {
            _copy(_resolver, "resource://org/ametys/runtime/test/cocoon/trax-level2-1.xsl", "context://plugins/test/trax/trax-level2.xsl");
            
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            // initially not cached
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.NON_CACHED, transformerHandlerAndValidity.getValidity());
            
            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // second request, unchanged, getting XSLT from cache
        enterEnvironment("");
        try
        {
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertTrue(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.CACHED, transformerHandlerAndValidity.getValidity());
            
            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // third request
        enterEnvironment("");
        try
        {
            // Wait 2s due to Linux File.lastModified() second precision
            Thread.sleep(2000);
            // value is changed on level-2 XSLT
            _copy(_resolver, "resource://org/ametys/runtime/test/cocoon/trax-level2-2.xsl", "context://plugins/test/trax/trax-level2.xsl");
            
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.OUT_OF_DATE, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value2", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }
    }
    
    /**
     * Tests {@link ThreadSafeTraxProcessor} with regular files, modifying root.
     * @throws Exception if an error occurs.
     */
    @Test
    public void testRegularRoot() throws Exception
    {
        _copy(_resolver, "resource://org/ametys/runtime/test/cocoon/trax-test1-root.xsl", "context://plugins/test/trax/trax-test4-root.xsl");
        _copy(_resolver, "resource://org/ametys/runtime/test/cocoon/trax-level2-1.xsl", "context://plugins/test/trax/trax-level2.xsl");
        
        Source xslt = _resolver.resolveURI("context://plugins/test/trax/trax-test4-root.xsl");

        // First request
        enterEnvironment("");
        try
        {
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.NON_CACHED, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // second request, unchanged, getting XSLT from cache
        enterEnvironment("");
        try
        {
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertTrue(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.CACHED, transformerHandlerAndValidity.getValidity());
            
            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // third request
        enterEnvironment("");
        try
        {
            // Wait 2s due to Linux File.lastModified() second precision
            Thread.sleep(2000);
            // value is changed on root XSLT
            _copy(_resolver, "resource://org/ametys/runtime/test/cocoon/trax-test1-root.xsl", "context://plugins/test/trax/trax-test4-root.xsl");
            
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.OUT_OF_DATE, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }
    }
    
    /**
     * Tests {@link ThreadSafeTraxProcessor} with regular files.
     * @throws Exception if an error occurs.
     */
    @Test
    public void testRegularChangingUrl() throws Exception
    {
        Source xslt = _resolver.resolveURI("resource://org/ametys/runtime/test/cocoon/trax-test5-root.xsl");

        // First request
        enterEnvironment("");
        try
        {
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.NON_CACHED, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // second request, unchanged, getting XSLT from cache
        enterEnvironment("");
        try
        {
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertTrue(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.CACHED, transformerHandlerAndValidity.getValidity());
            
            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // third request
        enterEnvironment("");
        try
        {
            Request request = getRequest();
            request.setAttribute("test", "test");
            
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.NON_CACHED, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value2", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }
        
        // fourth request, getting XSLT from cache
        enterEnvironment("");
        try
        {
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertTrue(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.CACHED, transformerHandlerAndValidity.getValidity());
            
            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // fifth request
        enterEnvironment("");
        try
        {
            Request request = getRequest();
            request.setAttribute("test", "test");
            
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertTrue(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.CACHED, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value2", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }
    }
    
    /**
     * Tests {@link ThreadSafeTraxProcessor} with regular files.
     * @throws Exception if an error occurs.
     */
    @Test
    public void testRegularChangingRoot() throws Exception
    {
        // First request
        enterEnvironment("");
        try
        {
            Source xslt = _resolver.resolveURI("test://trax-test6-root.xsl");
            
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.NON_CACHED, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // second request, unchanged, getting XSLT from cache
        enterEnvironment("");
        try
        {
            Source xslt = _resolver.resolveURI("test://trax-test6-root.xsl");
            
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertTrue(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.CACHED, transformerHandlerAndValidity.getValidity());
            
            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // third request
        enterEnvironment("");
        try
        {
            Request request = getRequest();
            request.setAttribute("test", "test");

            Source xslt = _resolver.resolveURI("test://trax-test6-root.xsl");

            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.NON_CACHED, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value2", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }
        
        // fourth request, getting XSLT from cache
        enterEnvironment("");
        try
        {
            Source xslt = _resolver.resolveURI("test://trax-test6-root.xsl");

            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertTrue(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.CACHED, transformerHandlerAndValidity.getValidity());
            
            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // fifth request
        enterEnvironment("");
        try
        {
            Request request = getRequest();
            request.setAttribute("test", "test");
            
            Source xslt = _resolver.resolveURI("test://trax-test6-root.xsl");

            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertTrue(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.CACHED, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value2", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }
    }
    
    /**
     * Tests {@link ThreadSafeTraxProcessor} with "cocoon://" inclusions.
     * @throws Exception if an error occurs.
     */
    @Test
    public void testWithCocoonIncludes() throws Exception
    {
        Source xslt = _resolver.resolveURI("resource://org/ametys/runtime/test/cocoon/trax-test2-root.xsl");

        // First request
        enterEnvironment("");
        try
        {
            _copy(_resolver, "resource://org/ametys/runtime/test/cocoon/trax-level2-1.xsl", "context://plugins/test/trax/trax-level2.xsl");
            
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.NON_CACHED, transformerHandlerAndValidity.getValidity());
            
            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // second request, unchanged
        enterEnvironment("");
        try
        {
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.OUT_OF_DATE, transformerHandlerAndValidity.getValidity());
            
            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // third request
        enterEnvironment("");
        try
        {
            // Wait 2s due to Linux File.lastModified() second precision
            Thread.sleep(2000);
            // value is changed on level-2 XSLT
            _copy(_resolver, "resource://org/ametys/runtime/test/cocoon/trax-level2-2.xsl", "context://plugins/test/trax/trax-level2.xsl");

            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.OUT_OF_DATE, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value2", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }
    }
    
    /**
     * Tests {@link ThreadSafeTraxProcessor} with "cocoon://" root xslt.
     * @throws Exception if an error occurs.
     */
    @Test
    public void testWithCocoonRoot() throws Exception
    {
        _copy(_resolver, "resource://org/ametys/runtime/test/cocoon/trax-test1-root.xsl", "context://plugins/test/trax/trax-test3-root.xsl");

        // First request
        enterEnvironment("");
        try
        {
            Source xslt = _resolver.resolveURI("cocoon://_plugins/test/trax-test3-root.xsl");

            _copy(_resolver, "resource://org/ametys/runtime/test/cocoon/trax-level2-1.xsl", "context://plugins/test/trax/trax-level2.xsl");
            
            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.NON_CACHED, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // second request, unchanged
        enterEnvironment("");
        try
        {
            Source xslt = _resolver.resolveURI("cocoon://_plugins/test/trax-test3-root.xsl");

            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.NON_CACHED, transformerHandlerAndValidity.getValidity());
            
            Assert.assertEquals("value1", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }  
        
        // third request
        enterEnvironment("");
        try
        {
            Source xslt = _resolver.resolveURI("cocoon://_plugins/test/trax-test3-root.xsl");

            // Wait 2s due to Linux File.lastModified() second precision
            Thread.sleep(2000);
            // value is changed on level-2 XSLT
            _copy(_resolver, "resource://org/ametys/runtime/test/cocoon/trax-level2-2.xsl", "context://plugins/test/trax/trax-level2.xsl");

            ExtendedTransformerHandlerAndValidity transformerHandlerAndValidity = (ExtendedTransformerHandlerAndValidity) _xsltProcessor.getTransformerHandlerAndValidity(xslt);
            Assert.assertFalse(transformerHandlerAndValidity.isFromCache());
            Assert.assertEquals(CacheValidity.NON_CACHED, transformerHandlerAndValidity.getValidity());

            Assert.assertEquals("value2", _getValue(transformerHandlerAndValidity.getTransfomerHandler()));
        }
        finally
        {
            leaveEnvironment();
        }
    }
    
    private void _copy(SourceResolver resolver, String src, String dest) throws IOException
    {
        Source source = resolver.resolveURI(src);
        Source destination = resolver.resolveURI(dest);
        
        SourceUtil.copy(source, destination);
    }
    
    private String _getValue(TransformerHandler transformerHandler) throws Exception
    {
        DOMResult result = new DOMResult();
        transformerHandler.setResult(result);

        XMLReader reader = SAXParserFactory.newInstance().newSAXParser().getXMLReader();
        reader.setContentHandler(transformerHandler);
        
        try (InputStream is = _xml.getInputStream())
        {
            reader.parse(new InputSource(is));
        }
        
        Node node = result.getNode();
        
        return _xPath.evaluate("/test/value", node);
    }
}
