/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.groups.jdbc;

import java.io.File;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.core.datasource.dbtype.SQLDatabaseType;
import org.ametys.core.group.Group;
import org.ametys.core.group.GroupIdentity;
import org.ametys.core.group.InvalidModificationException;
import org.ametys.core.group.ModifiableGroup;
import org.ametys.core.group.directory.GroupDirectory;
import org.ametys.core.group.directory.ModifiableGroupDirectory;
import org.ametys.core.user.UserIdentity;
import org.ametys.plugins.core.impl.group.directory.jdbc.JdbcGroupDirectory;
import org.ametys.runtime.test.AbstractJDBCTestCase;
import org.ametys.runtime.test.framework.TestConfigSet;
import org.ametys.runtime.test.framework.Unoptimized;

/**
 * Tests the JdbcGroupDirectory
 */
@Unoptimized
public abstract class AbstractJdbcGroupsTestCase extends AbstractJDBCTestCase
{
    @Override
    public TestConfigSet getConfigSet()
    {
        return super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        _userDirectory = _createUserDirectory();
        _groupDirectory = _createGroupDirectory();
    }
    
    /**
     * Provide the scripts to run to populate the database.
     * @return the scripts to run.
     */
    protected File[] getPopulateScripts()
    {
        return new File[] {new File("test/environments/scripts/jdbc/" + getDBType() + "/fillJDBCUsersAndGroups.sql")};
    }
    
    @Override
    protected void essentialsAfter() throws Exception
    {
        _deleteGroupDirectory();
        super.essentialsAfter();
    }
    
    /**
     * Check that the jdbc group manager has the correct type
     * @throws Exception if an error occurs
     */
    @Test
    public void testType() throws Exception
    {
        // DEFAULT IMPL
        Assert.assertTrue(_groupDirectory instanceof JdbcGroupDirectory);

        // MODIFIABLE
        Assert.assertTrue(_groupDirectory instanceof ModifiableGroupDirectory);
    }
    
    /**
     * Check when the db is empty
     * @throws Exception if an error occurs
     */
    @Test
    public void testEmpty() throws Exception
    {
        // GET GROUP
        Assert.assertNull(_groupDirectory.getGroup("foo"));
        
        // GET GROUPS
        Assert.assertEquals(0, _groupDirectory.getGroups().size());
        
        // GET USER GROUPS
        Assert.assertEquals(0, _groupDirectory.getUserGroups(new UserIdentity("foo", "foo")).size());
    }

    /**
     * Check when the db is filled
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    @Test
    public void testFilled() throws Exception
    {
        Set<UserIdentity> users;
        Group group;
        
        UserIdentity test = new UserIdentity("test", "population");
        UserIdentity test2 = new UserIdentity("test2", "population");
        
        // Be sure it is empty before filling and tables are created
        Assert.assertEquals(0, _userDirectory.getUsers().size());
        Assert.assertEquals(0, _groupDirectory.getGroups().size());
        // Fill DB
        _setDatabase(Arrays.asList(getPopulateScripts()));
        
        List<String> groupsIds = _getGroupsIds();
        
        // GET GROUP
        Assert.assertNull(_groupDirectory.getGroup("foo"));
        
        group = _groupDirectory.getGroup(groupsIds.get(0));
        Assert.assertNotNull(group);
        Assert.assertEquals(groupsIds.get(0), group.getIdentity().getId());
        Assert.assertEquals("Group 1", group.getLabel());
        users = group.getUsers();
        Assert.assertEquals(1, users.size());
        Assert.assertEquals(test, users.iterator().next());
        
        group = _groupDirectory.getGroup(groupsIds.get(1));
        Assert.assertNotNull(group);
        Assert.assertEquals(groupsIds.get(1), group.getIdentity().getId());
        Assert.assertEquals("Group 2", group.getLabel());
        users = group.getUsers();
        Assert.assertEquals(2, users.size());
        Assert.assertTrue(users.contains(test));
        Assert.assertTrue(users.contains(test2));
        
        group = _groupDirectory.getGroup(groupsIds.get(2));
        Assert.assertNotNull(group);
        Assert.assertEquals(groupsIds.get(2), group.getIdentity().getId());
        Assert.assertEquals("Group 3", group.getLabel());
        users = group.getUsers();
        Assert.assertEquals(1, users.size());
        Assert.assertEquals(test2, users.iterator().next());
        
        group = _groupDirectory.getGroup(groupsIds.get(3));
        Assert.assertNotNull(group);
        Assert.assertEquals(groupsIds.get(3), group.getIdentity().getId());
        Assert.assertEquals("Group 4", group.getLabel());
        users = group.getUsers();
        Assert.assertEquals(0, users.size());
        
        // GET GROUPS
        Set<GroupIdentity> groups = _groupDirectory.getGroups().stream()
                                                               .map(Group::getIdentity)
                                                               .collect(Collectors.toSet());
        
        Assert.assertNotNull(groups);
        Assert.assertEquals(4, groups.size());        
        GroupIdentity groupIdentity0 = new GroupIdentity(groupsIds.get(0), "jdbc-groups");
        Assert.assertTrue(groups.contains(groupIdentity0));
        GroupIdentity groupIdentity1 = new GroupIdentity(groupsIds.get(1), "jdbc-groups");
        Assert.assertTrue(groups.contains(groupIdentity1));
        GroupIdentity groupIdentity2 = new GroupIdentity(groupsIds.get(2), "jdbc-groups");
        Assert.assertTrue(groups.contains(groupIdentity2));
        GroupIdentity groupIdentity3 = new GroupIdentity(groupsIds.get(3), "jdbc-groups");
        Assert.assertTrue(groups.contains(groupIdentity3));
        
        // GET USER GROUPS
        Set<String> groupsName;
        groupsName = _groupDirectory.getUserGroups(new UserIdentity("test", "population"));
        Assert.assertEquals(2, groupsName.size());
        Assert.assertTrue(groupsName.contains(groupsIds.get(0)));
        Assert.assertTrue(groupsName.contains(groupsIds.get(1)));
        
        groupsName = _groupDirectory.getUserGroups(new UserIdentity("test2", "population"));
        Assert.assertEquals(2, groupsName.size());
        Assert.assertTrue(groupsName.contains(groupsIds.get(1)));
        Assert.assertTrue(groupsName.contains(groupsIds.get(2)));
        
        // GROUPS to JSON
        List<Map<String, Object>> jsonGroups = _groupHelper.groups2JSON(_groupDirectory.getGroups(-1, 0, new HashMap<>()), true);

        Assert.assertEquals(4, jsonGroups.size());
        
        List<Map<String, Object>> filteredJsonGroups;
        Set<UserIdentity> jsonUsers;
        UserIdentity jsonUser;
        filteredJsonGroups = _filterJsonGroupsById(jsonGroups, groupsIds.get(0));
        Assert.assertEquals(1, filteredJsonGroups.size());
        Assert.assertEquals("Group 1", filteredJsonGroups.iterator().next().get("label"));
        jsonUsers = (Set) filteredJsonGroups.iterator().next().get("users");
        Assert.assertEquals(1, jsonUsers.size());
        jsonUser = jsonUsers.iterator().next();
        Assert.assertEquals("test", jsonUser.getLogin());
        Assert.assertEquals("population", jsonUser.getPopulationId());
        
        filteredJsonGroups = _filterJsonGroupsById(jsonGroups, groupsIds.get(1));
        Assert.assertEquals(1, filteredJsonGroups.size());
        Assert.assertEquals("Group 2", filteredJsonGroups.iterator().next().get("label"));
        jsonUsers = (Set) filteredJsonGroups.iterator().next().get("users");
        Assert.assertEquals(2, jsonUsers.size());
        Assert.assertTrue(jsonUsers.contains(test));
        Assert.assertTrue(jsonUsers.contains(test2));
        
        filteredJsonGroups = _filterJsonGroupsById(jsonGroups, groupsIds.get(2));
        Assert.assertEquals(1, filteredJsonGroups.size());
        Assert.assertEquals("Group 3", filteredJsonGroups.iterator().next().get("label"));
        jsonUsers = (Set) filteredJsonGroups.iterator().next().get("users");
        Assert.assertEquals(1, jsonUsers.size());
        jsonUser = jsonUsers.iterator().next();
        Assert.assertEquals("test2", jsonUser.getLogin());
        Assert.assertEquals("population", jsonUser.getPopulationId());
        
        filteredJsonGroups = _filterJsonGroupsById(jsonGroups, groupsIds.get(3));
        Assert.assertEquals(1, filteredJsonGroups.size());
        Assert.assertEquals("Group 4", filteredJsonGroups.iterator().next().get("label"));
        jsonUsers = (Set) filteredJsonGroups.iterator().next().get("users");
        Assert.assertEquals(0, jsonUsers.size());
    }
    
    private List<Map<String, Object>> _filterJsonGroupsById(List<Map<String, Object>> jsonGroups, String id)
    {
        List<Map<String, Object>> results = new ArrayList<>();
        
        for (Map<String, Object> jsonGroup : jsonGroups)
        {
            if (id.equals(jsonGroup.get("id")))
            {
                results.add(jsonGroup);
            }
        }
        
        return results;
    }
    
    /**
     * Equivalent to {@link SQLDatabaseType#languageEscapeTableName(String)}
     * @param tableName The table name to escape
     * @return The escaped table table
     */
    protected String _escapeTableName(String tableName)
    {
        return "\"" + tableName + "\"";
    }
 
    private List<String> _getGroupsIds() throws Exception
    {
        List<String> result = new ArrayList<>();

        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        
        try
        {
            connection = ConnectionHelper.getConnection("SQL-test");
            stmt = connection.prepareStatement("SELECT Id FROM " + _escapeTableName("Groups") + " order by Id");
            
            rs = stmt.executeQuery();
            while (rs.next())
            {
                result.add(rs.getString("Id"));
            }
        }
        finally
        {
            ConnectionHelper.cleanup(rs);
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(connection);
        }
        
        return result;
    }

    /**
     * Test the addition of a new user
     * @throws Exception if an error occurs
     */
    @Test
    public void testCorrectAdd() throws Exception
    {
        ModifiableGroupDirectory groupDirectory = (ModifiableGroupDirectory) _groupDirectory;
        
        // Correct additions
        Group group1 = groupDirectory.add("Group 1");
        Group group2 = groupDirectory.add("Group 2");
        Group group3 = groupDirectory.add("Group 3");

        Assert.assertNotNull(groupDirectory.getGroup(group1.getIdentity().getId()));
        Assert.assertNotNull(groupDirectory.getGroup(group2.getIdentity().getId()));
        Assert.assertNotNull(groupDirectory.getGroup(group3.getIdentity().getId()));
        Assert.assertNotSame(group1.getIdentity().getId(), group2.getIdentity().getId());
    }
    
    /**
     * Test the update of a user incorrectly
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncorrectUpdate() throws Exception
    {
        ModifiableGroupDirectory groupDirectory = (ModifiableGroupDirectory) _groupDirectory;
        
        // Incorrect modification
        try
        {
            ModifiableGroup fooGroup = new TestModifiableGroup(new GroupIdentity("foo", "groupDirectory"), "Foo Group");
            fooGroup.addUser(new UserIdentity("user1", "population"));
            
            groupDirectory.update(fooGroup);
            Assert.fail("Update should have failed");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior since login does not exist
        }
    }
    
    /**
     * Test the update of a user
     * @throws Exception if an error occurs
     */
    @Test
    public void testCorrectUpdate() throws Exception
    {
        ModifiableGroupDirectory groupDirectory = (ModifiableGroupDirectory) _groupDirectory;
        
        ModifiableGroup group1 = groupDirectory.add("Group 1");
        
        group1.addUser(new UserIdentity("test", "population"));
        groupDirectory.update(group1);
        Assert.assertEquals(1, groupDirectory.getGroup(group1.getIdentity().getId()).getUsers().size());

        group1.addUser(new UserIdentity("test2", "population"));
        groupDirectory.update(group1);
        Assert.assertEquals(2, groupDirectory.getGroup(group1.getIdentity().getId()).getUsers().size());
        
        group1.addUser(new UserIdentity("test", "population"));
        groupDirectory.update(group1);
        Assert.assertEquals(2, groupDirectory.getGroup(group1.getIdentity().getId()).getUsers().size());
        
        group1.getUsers().remove(new UserIdentity("test", "population"));
        groupDirectory.update(group1);
        Assert.assertEquals(1, groupDirectory.getGroup(group1.getIdentity().getId()).getUsers().size());
    }
    
    /**
     * Test incorrects remove
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncorrectRemove() throws Exception
    {
        ModifiableGroupDirectory groupDirectory = (ModifiableGroupDirectory) _groupDirectory;

        try
        {
            groupDirectory.remove("foo");
            Assert.fail("Remove should fail");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior
        }

        Group group1 = groupDirectory.add("Group 1");

        try
        {
            groupDirectory.remove("foo");
            Assert.fail("Remove should fail");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior
            Group group = groupDirectory.getGroup(group1.getIdentity().getId());
            Assert.assertNotNull(group);
        }
    }
    
    /**
     * Test corrects remove
     * @throws Exception if an error occurs
     */
    @Test
    public void testCorrectRemove() throws Exception
    {
        ModifiableGroupDirectory groupDirectory = (ModifiableGroupDirectory) _groupDirectory;

        ModifiableGroup group1 = groupDirectory.add("Group 1");
        group1.addUser(new UserIdentity("test", "population"));
        groupDirectory.update(group1);

        groupDirectory.remove(group1.getIdentity().getId());
        
        Assert.assertNull(groupDirectory.getGroup(group1.getIdentity().getId()));
    }
    
    private static final class TestModifiableGroup implements ModifiableGroup
    {
        private Set<UserIdentity> _users;
        private GroupIdentity _identity;
        private String _groupLabel;
        
        TestModifiableGroup(GroupIdentity identity, String label)
        {
            _identity = identity;
            _groupLabel = label;
            _users = new HashSet<>();
        }
        
        @Override
        public GroupIdentity getIdentity()
        {
            return _identity;
        }

        @Override
        public String getLabel()
        {
            return _groupLabel;
        }

        @Override
        public GroupDirectory getGroupDirectory()
        {
            return null;
        }
        
        @Override
        public void setLabel(String label)
        {
            _groupLabel = label;
        }

        @Override
        public void addUser(UserIdentity user)
        {
            _users.add(user);
        }

        @Override
        public void removeUser(UserIdentity user)
        {
            _users.remove(user);
        }
        
        @Override
        public void removeUsers()
        {
            _users.clear();
        }
        
        @Override
        public Set<UserIdentity> getUsers()
        {
            return _users;
        }
        
        @Override
        public String toString()
        {
            StringBuffer sb = new StringBuffer("UserGroup[");
            sb.append(_identity);
            sb.append(" (");
            sb.append(_groupLabel);
            sb.append(") => ");
            sb.append(_users.toString());
            sb.append("]");
            return sb.toString();
        }    
        
        @Override
        public boolean equals(Object another)
        {
            if (another == null || !(another instanceof TestModifiableGroup))
            {
                return false;
            }
            
            TestModifiableGroup otherGroup = (TestModifiableGroup) another;
            
            return _identity != null && _identity.equals(otherGroup.getIdentity());
        }
        
        @Override
        public int hashCode()
        {
            return _identity.hashCode();
        }
    }
}
