/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.groups.ldap;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;

import org.ametys.core.datasource.AbstractDataSourceManager.DataSourceDefinition;
import org.ametys.core.datasource.LDAPDataSourceManager;
import org.ametys.core.group.Group;
import org.ametys.core.group.GroupIdentity;
import org.ametys.core.group.directory.GroupDirectory;
import org.ametys.core.group.directory.GroupDirectoryFactory;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.plugins.core.group.GroupHelper;
import org.ametys.plugins.core.impl.user.LdapUserIdentity;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * Ldap groups' tests 
 */
@NeedFeatures({"core/runtime.core.datasource.ldap.manager", "core/runtime.users.manager", "core/core.groups.helper"})
public abstract class AbstractLdapGroupsTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    /** The group directory */
    protected static GroupDirectory _groupDirectory;
    /** The LDAP Datasource Manager */
    protected static LDAPDataSourceManager _ldapDataSourceManager;
    /** The User Population DAO */
    protected static UserPopulationDAO _userPopulationDAO;
    /** The Group Directory Factory */
    protected static GroupDirectoryFactory _groupDirectoryFactory;
    /** The group helper */
    protected static GroupHelper _groupHelper;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config3.xml")
                .withLdapFile("test/environments/datasources/datasource-ldap.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ldapDataSourceManager = (LDAPDataSourceManager) manager.lookup(LDAPDataSourceManager.ROLE);
        _userPopulationDAO = (UserPopulationDAO) manager.lookup(UserPopulationDAO.ROLE);
        _groupDirectoryFactory = (GroupDirectoryFactory) manager.lookup(GroupDirectoryFactory.ROLE);
        _groupHelper = (GroupHelper) manager.lookup(GroupHelper.ROLE);
    }
    
    /**
     * To avoid alltest failure
     * @param attribute user attribute for login
     * @throws Exception if an error occurs
     */
    @SuppressWarnings("unchecked")
    protected void _testFilled(String attribute) throws Exception
    {
        User u1 = new User("user1", "user1@ametys.org");
        User u2 = new User("user2", "user2@ametys.org");
        User u3 = new User("user3", "user3@ametys.org");
        User u4 = new User("user4", "user4@ametys.org");
        User u5 = new User("user5", "user5@ametys.org");
        User u6 = new User("user6", "user6@ametys.org");
        User u7 = new User("user7", "user7@ametys.org");
        User u8 = new User("user8", "user8@ametys.org");
        User u10 = new User("user10", "user10@ametys.org");
        
        Group group;
        Set<UserIdentity> users;
        
        // IDENTITIES
        String populationId = "ldap_population";
        UserIdentity user1 = new UserIdentity(u1.get(attribute), populationId);
        UserIdentity user2 = new UserIdentity(u2.get(attribute), populationId);
        UserIdentity user3 = new UserIdentity(u3.get(attribute), populationId);
        UserIdentity user4 = new UserIdentity(u4.get(attribute), populationId);
        UserIdentity user5 = new UserIdentity(u5.get(attribute), populationId);
        UserIdentity user6 = new UserIdentity(u6.get(attribute), populationId);
        UserIdentity user7 = new UserIdentity(u7.get(attribute), populationId);
        UserIdentity user8 = new UserIdentity(u8.get(attribute), populationId);
        UserIdentity user10 = new UserIdentity(u10.get(attribute), populationId);
        
        // GET GROUP
        Assert.assertNull(_groupDirectory.getGroup("foo"));
        
        group = _groupDirectory.getGroup("group1");
        Assert.assertNotNull(group);
        Assert.assertEquals("group1", group.getIdentity().getId());
        Assert.assertEquals("Group 1", group.getLabel());
        users = group.getUsers();
        Assert.assertEquals(1, users.size());
        Assert.assertEquals(user1, users.iterator().next());
        
        group = _groupDirectory.getGroup("group2");
        Assert.assertNotNull(group);
        Assert.assertEquals("group2", group.getIdentity().getId());
        Assert.assertEquals("Group 2", group.getLabel());
        users = group.getUsers();
        Assert.assertEquals(4, users.size());
        Assert.assertTrue(users.contains(user1));
        Assert.assertTrue(users.contains(user2));
        Assert.assertTrue(users.contains(user3));
        Assert.assertTrue(users.contains(user4));
        
        group = _groupDirectory.getGroup("group3");
        Assert.assertNotNull(group);
        Assert.assertEquals("group3", group.getIdentity().getId());
        Assert.assertEquals("Group 3", group.getLabel());
        users = group.getUsers();
        Assert.assertEquals(4, users.size());
        Assert.assertTrue(users.contains(user5));
        Assert.assertTrue(users.contains(user6));
        Assert.assertTrue(users.contains(user7));
        Assert.assertTrue(users.contains(user8));
        
        // GET GROUPS
        Set<GroupIdentity> groups = _groupDirectory.getGroups().stream()
                                                        .map(Group::getIdentity)
                                                        .collect(Collectors.toSet());
        
        GroupIdentity group1 = new GroupIdentity("group1", "directory");
        GroupIdentity group2 = new GroupIdentity("group2", "directory");
        GroupIdentity group3 = new GroupIdentity("group3", "directory");
        
        Assert.assertNotNull(groups);
        Assert.assertEquals(3, groups.size());        
        Assert.assertTrue("group1 not found", groups.contains(group1));
        Assert.assertTrue("group2 not found", groups.contains(group2));
        Assert.assertTrue("group3 not found", groups.contains(group3));
        
        // GET USER GROUPS
        Set<String> groupsIds;
        groupsIds = _groupDirectory.getUserGroups(user1);
        Assert.assertEquals(2, groupsIds.size());
        Assert.assertTrue(groupsIds.contains("group1"));
        Assert.assertTrue(groupsIds.contains("group2"));
        
        groupsIds = _groupDirectory.getUserGroups(user2);
        Assert.assertEquals(1, groupsIds.size());
        Assert.assertTrue(groupsIds.contains("group2"));
        
        // Test with LDAPUserIdentity for avoiding a search and do a lookup instead (if memberof attribute on users only)
        String ldapBaseDn = _getLdapBaseDn();
        groupsIds = _groupDirectory.getUserGroups(new LdapUserIdentity(u3.get(attribute), populationId, "uid=user3,ou=people," + ldapBaseDn));
        Assert.assertEquals(1, groupsIds.size());
        Assert.assertTrue(groupsIds.contains("group2"));

        groupsIds = _groupDirectory.getUserGroups(user10);
        Assert.assertEquals(0, groupsIds.size());

        // GROUPS to JSON
        List<Map<String, Object>> jsonGroups = _groupHelper.groups2JSON(_groupDirectory.getGroups(-1, 0, new HashMap<>()), true);
        
        Assert.assertEquals(3, jsonGroups.size());
        
        List<Map<String, Object>> filteredJsonGroups;
        Set<UserIdentity> jsonUsers;
        UserIdentity jsonUser;
        filteredJsonGroups = _filterJsonGroupsById(jsonGroups, "group1");
        Assert.assertEquals(1, filteredJsonGroups.size());
        Assert.assertEquals("Group 1", filteredJsonGroups.iterator().next().get("label"));
        jsonUsers = (Set) filteredJsonGroups.iterator().next().get("users");
        Assert.assertEquals(1, jsonUsers.size());
        jsonUser = jsonUsers.iterator().next();
        Assert.assertEquals(u1.get(attribute), jsonUser.getLogin());
        Assert.assertEquals(populationId, jsonUser.getPopulationId());

        filteredJsonGroups = _filterJsonGroupsById(jsonGroups, "group2");
        Assert.assertEquals(1, filteredJsonGroups.size());
        Assert.assertEquals("Group 2", filteredJsonGroups.iterator().next().get("label"));
        jsonUsers = (Set) filteredJsonGroups.iterator().next().get("users");
        Assert.assertEquals(4, jsonUsers.size());
        Assert.assertTrue(jsonUsers.contains(user1));
        Assert.assertTrue(jsonUsers.contains(user2));
        Assert.assertTrue(jsonUsers.contains(user3));
        Assert.assertTrue(jsonUsers.contains(user4));
        
        filteredJsonGroups = _filterJsonGroupsById(jsonGroups, "group3");
        Assert.assertEquals(1, filteredJsonGroups.size());
        Assert.assertEquals("Group 3", filteredJsonGroups.iterator().next().get("label"));
        jsonUsers = (Set) filteredJsonGroups.iterator().next().get("users");
        Assert.assertEquals(4, jsonUsers.size());
        Assert.assertTrue(jsonUsers.contains(user5));
        Assert.assertTrue(jsonUsers.contains(user6));
        Assert.assertTrue(jsonUsers.contains(user7));
        Assert.assertTrue(jsonUsers.contains(user8));
    }
    
    private List<Map<String, Object>> _filterJsonGroupsById(List<Map<String, Object>> jsonGroups, String id)
    {
        List<Map<String, Object>> results = new ArrayList<>();
        
        for (Map<String, Object> jsonGroup : jsonGroups)
        {
            if (id.equals(jsonGroup.get("id")))
            {
                results.add(jsonGroup);
            }
        }
        
        return results;
    }
    
    private String _getLdapBaseDn()
    {
        DataSourceDefinition ldapDefinition = _ldapDataSourceManager.getDataSourceDefinition(_getLdapDatasourceId());
        if (ldapDefinition != null)
        {
            Map<String, Object> ldapParameters = ldapDefinition.getParameters();
            return (String) ldapParameters.get(LDAPDataSourceManager.PARAM_BASE_DN);
        }
        
        return null;
    }
    
    /**
     * Gets the id of the LDAP datasource
     * @return the id of the LDAP datasource
     */
    protected abstract String _getLdapDatasourceId();
    
    class User 
    {
        private String _uid;
        private String _email;
        
        User(String login, String email)
        {
            _uid = login;
            _email = email;
        }
        
        String get(String attribute)
        {
            if ("uid".equals(attribute))
            {
                return _uid;
            }
            else if ("email".equals(attribute))
            {
                return _email;
            }

            throw new IllegalArgumentException("only 'uid' and 'email' attribute supported");
        }
    }
}
