/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.groups.ldap;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

import org.junit.Test;

import org.ametys.core.group.directory.GroupDirectory;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.test.framework.TestConfigSet;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Tests the LdapGroupsManager (UserDriven, i.e. with memberof parameter set (corresponding to the DN of the groups), but not member)
 */
@NeedFeatures({"core-impl/runtime.user.directory.impl.ldap", "core-impl/runtime.authentication.impl.defined", "core-impl/runtime.group.directory.impl.ldap"})
public class LdapGroupsWithMemberOfTestCase extends AbstractLdapGroupsTestCase
{
    @Override
    public TestConfigSet getConfigSet()
    {
        return super.getConfigSet()
                .withConfigFile("test/environments/configs/config4.xml");
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        _groupDirectory = _createUserDrivenLdapGroupDirectory();
    }
    
    /**
     * Tests groups.
     * @throws Exception if an error occurs
     */
    @Test
    public void testFilled() throws Exception
    {
        _testFilled("uid");
    }
    
    private GroupDirectory _createUserDrivenLdapGroupDirectory()
    {
        // We need a LDAP User Directory where to retrieve the users
        String populationId = _createPopulation();
        
        // Get the id of this LDAP user directory
        String userDirectoryId = _userPopulationDAO.getUserPopulation(populationId).getUserDirectories().get(0).getId();
        
        // Create the Group Directory
        String modelId = "org.ametys.plugins.core.group.directory.Ldap";
        Map<String, Object> parameters = new LinkedHashMap<>();
        parameters.put("runtime.groups.ldap.datasource", _getLdapDatasourceId());
        parameters.put("runtime.groups.ldap.groupDN", "ou=groups");
        parameters.put("runtime.groups.ldap.filter", "(objectclass=organizationalUnit)");
        parameters.put("runtime.groups.ldap.scope", "one");
        parameters.put("runtime.groups.ldap.id", "ou");
        parameters.put("runtime.groups.ldap.description", "description");
        parameters.put("runtime.groups.ldap.userdirectory", populationId + "#" + userDirectoryId);
        parameters.put("runtime.users.ldap.uidAttr", "uid");
        parameters.put("runtime.groups.ldap.memberof", "memberOf");
        return _groupDirectoryFactory.createGroupDirectory("directory", new I18nizableText("User Driven LDAP"), modelId, parameters);
    }
    
    private String _createPopulation()
    {
        Map<String, Object> userDirectoryParameters = new LinkedHashMap<>();
        String udModelId = "org.ametys.plugins.core.user.directory.Ldap";
        userDirectoryParameters.put("udModelId", udModelId);
        userDirectoryParameters.put(udModelId + "$" + "runtime.users.ldap.datasource", _getLdapDatasourceId());
        userDirectoryParameters.put(udModelId + "$" + "runtime.users.ldap.peopleDN", "ou=people");
        userDirectoryParameters.put(udModelId + "$" + "runtime.users.ldap.baseFilter", "(objectclass=inetOrgPerson)");
        userDirectoryParameters.put(udModelId + "$" + "runtime.users.ldap.scope", "sub");
        userDirectoryParameters.put(udModelId + "$" + "runtime.users.ldap.loginAttr", "uid");
        userDirectoryParameters.put(udModelId + "$" + "runtime.users.ldap.firstnameAttr", "givenName");
        userDirectoryParameters.put(udModelId + "$" + "runtime.users.ldap.lastnameAttr", "sn");
        userDirectoryParameters.put(udModelId + "$" + "runtime.users.ldap.emailAttr", "mail");
        userDirectoryParameters.put(udModelId + "$" + "runtime.users.ldap.emailMandatory", "false");
        userDirectoryParameters.put(udModelId + "$" + "runtime.users.ldap.serverSideSorting", "true");
        
        Map<String, Object> credentialProviderParameters = new LinkedHashMap<>();
        String cpModelId = "org.ametys.core.authentication.Defined";
        credentialProviderParameters.put("cpModelId", cpModelId);
        credentialProviderParameters.put(cpModelId + "$" + "runtime.authentication.defined.user", "anonymous");
        
        Map<String, Object> resultMap = _userPopulationDAO.add("ldap_population", "LDAP Population", Collections.singletonList(userDirectoryParameters), Collections.singletonList(credentialProviderParameters));
        return (String) resultMap.get("id");
    }
    
    @Override
    protected String _getLdapDatasourceId()
    {
        return "LDAP-test-users";
    }
}
