/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.migration;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.util.Arrays;
import java.util.List;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.collections4.CollectionUtils;
import org.junit.Test;

import org.ametys.core.migration.MigrationEngine;
import org.ametys.core.migration.MigrationEngine.MigrationComponent;
import org.ametys.core.migration.MigrationEngine.VersionList;
import org.ametys.core.migration.MigrationEngine.Versions;
import org.ametys.core.migration.MigrationEngine.VersionsContainer;
import org.ametys.core.migration.MigrationException;
import org.ametys.core.migration.MigrationExtensionPoint;
import org.ametys.core.migration.version.Version;
import org.ametys.core.migration.version.handler.JavaVersionHandler.JavaVersionConfiguration;
import org.ametys.core.migration.version.handler.ScriptVersionHandler.ScriptVersionConfiguration;
import org.ametys.core.migration.version.storage.SqlVersionStorage.SqlVersionConfiguration;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Tests for reading migration components and getting versions.
 */
@NeedFeatures({
    "test/test.migrations.script",
    "test/test.migrations.sql",
    "test/test.migrations.scriptHandler",
    "test/test.migrations.javaHandler"
})
public class MigrationVersionsTestCase extends AbstractMigrationTestCase
{
    private static MigrationExtensionPoint _migrationExtensionPoint;
    private static MigrationEngine _migrationEngine;
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _migrationExtensionPoint = (MigrationExtensionPoint) manager.lookup(MigrationExtensionPoint.ROLE);
        _migrationEngine = (MigrationEngine) manager.lookup(MigrationEngine.ROLE);
    }
    
    /**
     * Tests the configuration of all migration components.
     */
    @Test
    public void testReadMigrationComponents()
    {
        Set<String> extensionsIds = _migrationExtensionPoint.getExtensionsIds();
        assertEquals(8, extensionsIds.size());
        
        MigrationComponent test1 = _migrationExtensionPoint.getExtension("test.1");
        assertEquals("sql", test1.versionHandlerType());
        assertTrue(test1.versionConfiguration() instanceof SqlVersionConfiguration);
        assertEquals("SQL-test", ((SqlVersionConfiguration) test1.versionConfiguration()).dataSource());
        assertNotNull(test1.initialization());
        assertEquals(null, test1.initialization().getVersionNumber());
        assertEquals(2, test1.upgrades().size());
        assertEquals("20200201T1200", test1.upgrades().get(0).getVersionNumber());
        
        MigrationComponent test7 = _migrationExtensionPoint.getExtension("test.7");
        assertEquals("script", test7.versionHandlerType());
        assertTrue(test7.versionConfiguration() instanceof ScriptVersionConfiguration);
        assertEquals("sql", ((ScriptVersionConfiguration) test7.versionConfiguration()).storage());
        assertTrue(((ScriptVersionConfiguration) test7.versionConfiguration()).script().startsWith("function main() { \n var results"));
        
        MigrationComponent test8 = _migrationExtensionPoint.getExtension("test.8");
        assertEquals("java", test8.versionHandlerType());
        assertEquals("sql", ((JavaVersionConfiguration) test8.versionConfiguration()).storage());
        assertEquals("org.ametys.runtime.test.migration.component.Test8", ((JavaVersionConfiguration) test8.versionConfiguration()).role());
        assertNotNull(test8.initialization());
        assertEquals(1, test8.upgrades().size());
    }

    /**
     * Tests VersionHandler.getVersions for Java VersionHandler
     * @throws MigrationException if an error occurs while getting versions
     */
    @Test
    public void testGetJavaVersions() throws MigrationException
    {
        MigrationComponent test8 = _migrationExtensionPoint.getExtension("test.8");
        VersionsContainer versions = (VersionsContainer) _migrationEngine.getVersions(test8);
        
        assertEquals(2, versions.size());
        
        Versions noVersions = versions.get(new I18nizableText("test.migrations.javaHandler.noversion"));
        assertEquals(1, ((VersionList) noVersions).versions().size());
        Version v1 = ((VersionList) noVersions).versions().get(0);
        assertEquals("1", v1.getVersionNumber());
        assertEquals("test.migrations.javaHandler.noversion", v1.getComponentId());

        Versions version0 = versions.get(new I18nizableText("test.migrations.javaHandler.version0exists"));
        assertEquals(2, ((VersionList) version0).versions().size());
        List<String> versionExists = ((VersionList) version0).versions().stream().map(Version::getVersionNumber).toList();
        assertTrue(CollectionUtils.isEqualCollection(versionExists, Arrays.asList("0", "1")));
    }
    
    /**
     * Tests VersionHandler.getVersions for Script VersionHandler
     * @throws MigrationException if an error occurs while getting versions
     */
    @Test
    public void testGetScriptVersions() throws MigrationException
    {
        MigrationComponent test7 = _migrationExtensionPoint.getExtension("test.7");
        VersionsContainer versions = (VersionsContainer) _migrationEngine.getVersions(test7);
        
        assertEquals(2, versions.size());
        
        Versions noVersions = versions.get(new I18nizableText("test.migrations.scriptHandler.noversion"));
        assertEquals(1, ((VersionList) noVersions).versions().size());
        Version v1 = ((VersionList) noVersions).versions().get(0);
        assertEquals("1", v1.getVersionNumber());
        assertEquals("test.migrations.scriptHandler.noversion", v1.getComponentId());

        Versions version0 = versions.get(new I18nizableText("test.migrations.scriptHandler.version0exists"));
        assertEquals(2, ((VersionList) version0).versions().size());
        List<String> versionExists = ((VersionList) version0).versions().stream().map(Version::getVersionNumber).toList();
        assertTrue(CollectionUtils.isEqualCollection(versionExists, Arrays.asList("0", "1")));
    }
    
    /**
     * Tests VersionHandler.getVersions for Script VersionHandler
     * @throws MigrationException if an error occurs while getting versions
     */
    @Test
    public void testGetSqlVersions() throws MigrationException
    {
        MigrationComponent test1 = _migrationExtensionPoint.getExtension("test.1");
        VersionList versions = (VersionList) _migrationEngine.getVersions(test1);
        
        assertEquals(1, versions.versions().size());
        Version v0 = versions.versions().get(0);
        assertEquals("20200202T1200", v0.getVersionNumber());
        assertEquals("test.1", v0.getComponentId());

        MigrationComponent test3 = _migrationExtensionPoint.getExtension("test.2");
        versions = (VersionList) _migrationEngine.getVersions(test3);

        assertEquals(1, versions.versions().size());
        v0 = versions.versions().get(0);
        assertEquals("0", v0.getVersionNumber());
        assertEquals("test.2", v0.getComponentId());
    }
}
