/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.migration.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.migration.AbstractMigrationTestCase;

/**
 * Test migration with SQL version storage/handler
 */
@NeedFeatures({
    "test/test.migrations.sql",
    "test/test.migrations.scriptHandler",
    "test/test.migrations.javaHandler"
})
public abstract class AbstractSqlMigrationTestCase extends AbstractMigrationTestCase
{
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        // empty
    }
    
    /**
     * Init needed, no upgrade should be done
     * @throws SQLException queries to test failed
     */
    @Test
    public void testCase1() throws SQLException
    {
        Assert.assertEquals("The version for test.1 should be 20200202T1200", "20200202T1200", _getCurrentVersion("test.1"));
        Assert.assertEquals("There should be only 1 line in test_1 about initialization", 1, _countLines("SELECT comment_column FROM test_1 WHERE comment_column = 'Created by initialization'"));
        Assert.assertEquals("There should be only 0 line in test_1 about upgrade", 0, _countLines("SELECT comment_column FROM test_1 WHERE comment_column LIKE 'Upgrade%'"));
        Assert.assertEquals("There should be 1 version in test_1 with a comment", 1, _countLines("SELECT ametys_version_number FROM Ametys_Versions WHERE component_identifier = 'test.1' AND execution_comment = 'Automatic Initialization: Comment set in XML'"));
    }

    /**
     * No init needed, no upgrade should be done
     * @throws SQLException queries to test failed
     */
    @Test
    public void testCase2() throws SQLException
    {
        Assert.assertEquals("The version for test.2 should be 0", "0", _getCurrentVersion("test.2"));
        Assert.assertEquals("There should be 1 version in test_2 without a comment", 1, _countLines("SELECT ametys_version_number FROM Ametys_Versions WHERE component_identifier = 'test.2' AND execution_comment = 'Initial version'"));
    }

    /**
     * Init already done, upgrades with FROM needed
     * @throws SQLException queries to test failed
     */
    @Test
    public void testCase3() throws SQLException
    {
        Assert.assertEquals("The version for test.3 should be 20200210T1200", "20200210T1200", _getCurrentVersion("test.3"));
        Assert.assertEquals("There should be 4 versions", 4, _countVersions("test.3"));
        Assert.assertEquals("There should be only 0 line in test_3 about initialization", 0, _countLines("SELECT comment_column FROM test_3 WHERE comment_column LIKE 'NOK : %'"));
        Assert.assertEquals("There should be only 3 line in test_3 about upgrade", 3, _countLines("SELECT comment_column FROM test_3 WHERE comment_column LIKE 'OK : %'"));
        Assert.assertEquals("There should be 2 version in test_3 with a comment", 2, _countLines("SELECT ametys_version_number FROM Ametys_Versions WHERE component_identifier = 'test.3' AND execution_comment LIKE 'Automatic Upgrade: %'"));
        Assert.assertEquals("There should be 1 version in test_3 without a comment", 1, _countLines("SELECT ametys_version_number FROM Ametys_Versions WHERE component_identifier = 'test.3' AND execution_comment = 'Automatic Upgrade'"));
    }

    /**
     * Test of multiple upgrades with no initialization because the version 0 already exists
     * @throws SQLException queries to test failed
     */
    @Test
    public void testCase4() throws SQLException
    {
        Assert.assertEquals("The version for test.4 should be 20200202T1200", "20200202T1200", _getCurrentVersion("test.4"));
        Assert.assertEquals("There should be only 2 lines in test_4 about upgrade", 2, _countLines("SELECT comment_column FROM test_4 WHERE comment_column LIKE 'Upgrade %'"));
        Assert.assertEquals("There should be 0 version in test_4 with a comment", 0, _countLines("SELECT ametys_version_number FROM Ametys_Versions WHERE component_identifier = 'test.4' AND execution_comment = 'Automatic Initialization: double init'"));
    }

    /**
     * Init needed, no upgrades available, version should be 0
     * @throws SQLException queries to test failed
     */
    @Test
    public void testCase5() throws SQLException
    {
        Assert.assertEquals("The version for test.5 should be 0", "0", _getCurrentVersion("test.5"));
    }
    


    /**
     * Script handler that returns 2 versions
     * @throws SQLException queries to test failed
     */
    @Test
    public void testCase7() throws SQLException
    {
        Assert.assertEquals("There should be only 2 line in test_script", 2, _countLines("SELECT comment_column FROM test_script"));
        Assert.assertEquals("There should be only 1 line in test_script about initialization", 1, _countLines("SELECT comment_column FROM test_script WHERE comment_column = 'init versionNull'"));
        Assert.assertEquals("There should be only 1 line in test_script about update", 1, _countLines("SELECT comment_column FROM test_script WHERE comment_column = 'update versionNotNull'"));
    }

    /**
     * Java handler that returns 2 versions
     * @throws SQLException queries to test failed
     */
    @Test
    public void testCase8() throws SQLException
    {
        Assert.assertEquals("There should be only 2 line in test_java", 2, _countLines("SELECT comment_column FROM test_java"));
        Assert.assertEquals("There should be only 1 line in test_java about initialization", 1, _countLines("SELECT comment_column FROM test_java WHERE comment_column = 'init versionNull'"));
        Assert.assertEquals("There should be only 1 line in test_java about update", 1, _countLines("SELECT comment_column FROM test_java WHERE comment_column = 'update versionNotNull'"));
    }
    
    private String _getCurrentVersion(String componentIdentifier) throws SQLException
    {
        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        
        String version = null;
        
        try
        {
            connection = ConnectionHelper.getConnection("SQL-test");
            stmt = connection.prepareStatement("SELECT ametys_version_number FROM Ametys_Versions WHERE component_identifier = ? ORDER BY ametys_version_number DESC");
            stmt.setString(1, componentIdentifier);
            
            rs = stmt.executeQuery();
            
            if (rs.next())
            {
                version = rs.getString("ametys_version_number");
            }
        }
        finally
        {
            ConnectionHelper.cleanup(rs);
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(connection);
        }
        
        return version;
    }
    
    private int _countVersions(String componentIdentifier) throws SQLException
    {
        String query = "SELECT ametys_version_number FROM Ametys_Versions WHERE component_identifier = '" + componentIdentifier + "'";
        return _countLines(query);
    }
    
    private int _countLines(String query) throws SQLException
    {
        Connection connection = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        
        int count = 0;
        
        try
        {
            connection = ConnectionHelper.getConnection("SQL-test");
            stmt = connection.prepareStatement(query);
            
            rs = stmt.executeQuery();
            
            while (rs.next())
            {
                count++;
            }
        }
        finally
        {
            ConnectionHelper.cleanup(rs);
            ConnectionHelper.cleanup(stmt);
            ConnectionHelper.cleanup(connection);
        }
        
        return count;
    }
}
