/*
 *  Copyright 2016 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.minimize;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.nio.file.attribute.FileTime;
import java.time.Instant;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceNotFoundException;
import org.apache.excalibur.source.SourceResolver;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.DevMode;
import org.ametys.core.DevMode.DEVMODE;
import org.ametys.plugins.core.ui.minimize.HashCache;
import org.ametys.plugins.core.ui.minimize.HashCache.UriData;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * Test case for the minimize transformer
 */
@NeedFeatures({
    "core/runtime.core", "core-ui/minimizer-cache", 
    "test/test", "core/runtime.core.resources-handler",
    "core/xhtmlserializer-empty", "core/xhtmlserializer-xhtml", "core/xhtmlserializer-xml", "core/xhtmlserializer-xlink", "core/xhtmlserializer-svg", "core/xhtmlserializer-mathml",
    "core-ui/plugin.core.ui.resources.dependencies", "core-ui/plugin.core.ui.resources.sass.dependencies",
    "core-ui/runtime.core.ui.resources-handler",
    "core-ui/plugin.core.ui.minimize.resources.uri",
})
public class MinimizeTransformerTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static SourceResolver _resolver;
    private static HashCache _hashCache;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp2")
                .withContextPath(getContextPath());
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _hashCache = (HashCache) manager.lookup(HashCache.ROLE);
    }
    
    /**
     * Get the context path of the requests. Can be overriden
     * @return The context path
     */
    protected String getContextPath()
    {
        return "";
    }
    
    /**
     * This test runs all the possible situations that the minimize transformer may encounter, though a single html file and all the minified results.
     * @throws Exception If an error occurs
     */
    @Test
    public void testMinimizeTransformerFull() throws Exception
    {
        assertTransformation("complete-test.html", "complete-test-result.html", false);
    }
    
    /**
     * Asserts that the minimized transformer works
     * @param inputFilename The input file name
     * @param outputFilename  The result file name
     * @param contextPath Version with a context path
     * @throws Exception If an error occurred
     */
    protected void assertTransformation(String inputFilename, String outputFilename, boolean contextPath) throws Exception
    {
        Source expectedSource = _resolver.resolveURI("plugin:test://pages/minimize/" + outputFilename);
        Source source = _resolver.resolveURI("cocoon://_plugins/test/minimize/" + inputFilename);
        
        String result = IOUtils.toString(source.getInputStream(), "UTF-8");
        String filteredResult = result.replaceAll("/plugins/core-ui/resources-minimized/[^/.]+\\.(js|css)", "/plugins/core-ui/resources-minimized/__HASH__.$1");
        String expected = IOUtils.toString(expectedSource.getInputStream(), "UTF-8");
        Assert.assertEquals("XML output differs", expected.replace("\r", ""), filteredResult.replace("\r", ""));

        Matcher matcher = _getMinimizedUrlPattern().matcher(result);
        
        _assertFilesMinimized(matcher, "css", "complete-test-css-1.css");
        _assertFilesMinimized(matcher, "js", "complete-test-js-1" + (contextPath ? "-contextpath" : "") + ".js"); // File a.js does not exist, but hashed file should still generate
        _assertFilesMinimized(matcher, "css", "complete-test-css-2.css");
        _assertFilesMinimized(matcher, "css", "complete-test-css-3.css");
        _assertFilesMinimized(matcher, "css", "complete-test-css-4.css");
        _assertFilesMinimized(matcher, "js", "complete-test-js-2" + (contextPath ? "-contextpath" : "") + ".js");
        _assertFilesMinimized(matcher, "js", "complete-test-js-3" + (contextPath ? "-contextpath" : "") + ".js");
    }

    private void _assertFilesMinimized(Matcher matcher, String type, String expectedFile) throws Exception
    {
        Source expectedSource = _resolver.resolveURI("plugin:test://resources/" + type + "/minimize-result/" + expectedFile);
        String expected = IOUtils.toString(expectedSource.getInputStream(), "UTF-8");
        
        Assert.assertTrue("Url not minimized", matcher.find());
        String file = matcher.group(1);
        String fileExtension = matcher.group(2);
        Assert.assertEquals("Wrong minimized type", type, fileExtension);
        
        Source source = _resolver.resolveURI("cocoon://_plugins/test/resources-minimized/" + file + "." + fileExtension);
        String actual = IOUtils.toString(source.getInputStream(), "UTF-8");
        Assert.assertEquals("Minimized file content did not match expected value", 
                expected.replace("\r", "") + ("js".equals(type) ? file + ".js.map\n" : file + ".css.map */\n"), 
                actual.replace("\r", "").replaceAll("(file|resource):[^ ]+:[0-9]+:[0-9]+", "file:@CANCHANGE@"));
    }
    
    /**
     * This test runs all the possible situations that the minimize transformer may encounter, though a single html file and all the minified results.
     * @throws Exception If an error occurs
     */
    @Test
    public void testMinimizeTransformerLastModify() throws Exception
    {
        _assertLastModify("last-modify-test.html", "lastModify-create.js", "lastModify-edit.js", "js");
    }

    private void _assertLastModify(String htmlFile, String fileToCreate, String fileToModify, String extension) throws Exception
    {
        // Delete temporary file to create if exists
        File file = new File(getWebappPath() + "/plugins/test/resources/js/minimize/" + fileToCreate);
        if (file.exists())
        {
            Assert.assertTrue("Unable to delete file for last modify tests", file.delete());
        }
        
        
        // Generate a hash with a non existing file
        String hashNotFound = _getHashFromSourceFile(htmlFile);
        List<UriData> filesForHash = _hashCache.getFilesForHash(hashNotFound, false);
        Assert.assertTrue("The hash cache is invalid", filesForHash.size() == 2 
                && filesForHash.get(0).getUri().equals("/plugins/test/resources/js/minimize/" + fileToCreate)
                && filesForHash.get(1).getUri().equals("/plugins/test/resources/js/minimize/" + fileToModify));
        
        _assertMinimizedEquals(hashNotFound + "." + extension, "/** File : /plugins/test/resources/js/minimize/lastModify-create.js */\n"
                + "document_currentScript = (function() { a = document.createElement('script'); a.src='" + getContextPath() + "/plugins/test/resources/js/minimize/lastModify-create.js'; return a; })(); try {\n" 
                + "throw new Error('Resource not found for URI : \\'plugin:test:\\/\\/resources\\/js\\/minimize\\/lastModify-create.js\\'.\\n\\tat <map:read type=\\\"resource\\\"> - file:@CANCHANGE@');} catch (e) { console.error(\"/plugins/test/resources/js/minimize/lastModify-create.js\\n\", e) }\n"
                + "/** File : /plugins/test/resources/js/minimize/" + fileToModify + " */\n" 
                + "document_currentScript = (function() { a = document.createElement('script'); a.src='" + getContextPath() + "/plugins/test/resources/js/minimize/" + fileToModify + "'; return a; })(); try {\n" 
                + "console.log(\"test\");} catch (e) { console.error(\"/plugins/test/resources/js/minimize/lastModify-edit.js\\n\", e) }\n"
                + "\n"
                + "//# sourceMappingURL=" + hashNotFound + ".js.map\n");
        
        
        // Create the file and regenerate the hash
        Assert.assertTrue("Unable to create file to test last modify", file.createNewFile());
        String hashFound = _getHashFromSourceFile(htmlFile);
        Assert.assertFalse("Same hash after creating the file", hashNotFound.equals(hashFound));

        _assertMinimizedEquals(hashFound + "." + extension, "/** File : /plugins/test/resources/js/minimize/" + fileToCreate + " */\n"
                + "document_currentScript = (function() { a = document.createElement('script'); a.src='" + getContextPath() + "/plugins/test/resources/js/minimize/lastModify-create.js'; return a; })(); try {\n"
                + "} catch (e) { console.error(\"/plugins/test/resources/js/minimize/lastModify-create.js\\n\", e) }\n"
                + "/** File : /plugins/test/resources/js/minimize/" + fileToModify + " */\n"
                + "document_currentScript = (function() { a = document.createElement('script'); a.src='" + getContextPath() + "/plugins/test/resources/js/minimize/lastModify-edit.js'; return a; })(); try {\n"
                + "console.log(\"test\");} catch (e) { console.error(\"/plugins/test/resources/js/minimize/lastModify-edit.js\\n\", e) }\n"
                + "\n"
                + "//# sourceMappingURL=" + hashFound + ".js.map\n");
        
        // Edit the file to generate a new hash
        Files.write(Paths.get(getWebappPath() + "/plugins/test/resources/js/minimize/" + fileToModify), Arrays.asList("/* Temporary test file */", "console.log('test')"), Charset.forName("UTF-8"), StandardOpenOption.TRUNCATE_EXISTING);
        
        String hashModified = _getHashFromSourceFile(htmlFile);
        Assert.assertFalse("Hash must change when file is modified", hashNotFound.equals(hashModified) || hashFound.equals(hashModified));

        _assertMinimizedEquals(hashModified + "." + extension, "/** File : /plugins/test/resources/js/minimize/" + fileToCreate + " */\n"
                + "document_currentScript = (function() { a = document.createElement('script'); a.src='" + getContextPath() + "/plugins/test/resources/js/minimize/lastModify-create.js'; return a; })(); try {\n"
                + "} catch (e) { console.error(\"/plugins/test/resources/js/minimize/lastModify-create.js\\n\", e) }\n"
                + "/** File : /plugins/test/resources/js/minimize/" + fileToModify + " */\n"
                + "document_currentScript = (function() { a = document.createElement('script'); a.src='" + getContextPath() + "/plugins/test/resources/js/minimize/lastModify-edit.js'; return a; })(); try {\n"
                + "console.log(\"test\");} catch (e) { console.error(\"/plugins/test/resources/js/minimize/lastModify-edit.js\\n\", e) }\n"
                + "\n"
                + "//# sourceMappingURL=" + hashModified + ".js.map\n");

        String hashNoChange = _getHashFromSourceFile(htmlFile);
        Assert.assertEquals("Hash must not change when there is no modifications", hashModified, hashNoChange);
        
        // Delete temporary file created at the end
        if (file.exists())
        {
            Assert.assertTrue("Unable to delete file after last modify tests", file.delete());
        }
    }

    private void _assertMinimizedEquals(String hashedFile, String expected) throws MalformedURLException, IOException, SourceNotFoundException
    {
        Source source = _resolver.resolveURI("cocoon://_plugins/test/resources-minimized/" + hashedFile);
        String contentFound = IOUtils.toString(source.getInputStream(), "UTF-8");
        contentFound = contentFound.replaceAll("(file|resource):[^ ]+:[0-9]+:[0-9]+", "file:@CANCHANGE@");

        Assert.assertEquals("Minimized file content is not correct", expected, contentFound);
    }

    private String _getHashFromSourceFile(String fileSource) throws MalformedURLException, IOException, SourceNotFoundException
    {
        Source source = _resolver.resolveURI("cocoon://_plugins/test/minimize/" + fileSource);
        String sourceContent = IOUtils.toString(source.getInputStream(), "UTF-8");
        Matcher matcher = _getMinimizedUrlPattern().matcher(sourceContent);
        Assert.assertTrue("File was not successfully minimized", matcher.find());
        return matcher.group(1);
    }
    
    /**
     * This test runs all the possible situations that the minimize transformer may encounter, though a single html file and all the minified results.
     * @throws Exception If an error occurs
     */
    @Test
    public void testMinimizeTransformerFileDependencies() throws Exception
    {
        DEVMODE mode = DevMode.getDeveloperMode(getRequest());
        
        getRequest().getSession(true).setAttribute(DevMode.SESSION_ATTRIBUTE_DEVMODE, DEVMODE.PRODUCTION.toString());
        
        assertCssFileDependencies("file-dependencies-test.html", true);
        
        getRequest().getSession(true).setAttribute(DevMode.SESSION_ATTRIBUTE_DEVMODE, DEVMODE.DEVELOPMENT.toString());
        
        assertCssFileDependencies("file-dependencies-test.html", false);
        
        getRequest().getSession(true).setAttribute(DevMode.SESSION_ATTRIBUTE_DEVMODE, mode.toString());
    }
    
    /**
     * Check if the file dependencies of hash are correctly calculated
     * @param inputFilename The input file name
     * @param productionMode if in production mode
     * @throws Exception If an error occurred
     */
    protected void assertCssFileDependencies(String inputFilename, boolean productionMode) throws Exception
    {
        String hash = _getHashFromSourceFile(inputFilename);
        
        List<UriData> filesForHash = _hashCache.getFilesForHash(hash, false);
        
        Assert.assertNotNull("No hash found in cache for hash generated by MinimizeTransformer", filesForHash);
        Assert.assertTrue("Invalid list of files dependencies for hash. Expected 5 results, but received : " + filesForHash.toString(), filesForHash.size() == 5);
        
        List<UriData> invalidLastModifiedFiles = filesForHash.stream()
                .filter(file -> file.getLastModified() <= 0)
                .collect(Collectors.toList());
        
        Assert.assertTrue("Some files dependencies have an invalid last modified date : " + invalidLastModifiedFiles.toString(), invalidLastModifiedFiles.isEmpty());
        
        // touch a direct dependency
        Files.setLastModifiedTime(Paths.get(getWebappPath() + "/plugins/test/resources/css/minimize/file-dependencies/direct.css"), FileTime.from(Instant.now()));
        String hashDirectModified = _getHashFromSourceFile(inputFilename);
        Assert.assertNotEquals("Hash must change when file is modified", hash, hashDirectModified);
        
        // touch an indirect dependency
        Files.setLastModifiedTime(Paths.get(getWebappPath() + "/plugins/test/resources/css/minimize/file-dependencies/referenced.css"), FileTime.from(Instant.now()));
        
        String hashIndirectModified = _getHashFromSourceFile(inputFilename);
        
        if (productionMode)
        {
            Assert.assertEquals("Hash must not change in production mode when indirect dependency is modified", hashDirectModified, hashIndirectModified);
        }
        else
        {
            Assert.assertNotEquals("Hash must change in development mode when indirect dependency is modified", hashDirectModified, hashIndirectModified);
        }
    }

    private Pattern _getMinimizedUrlPattern()
    {
        return Pattern.compile("['\"]" + getContextPath() + "/plugins/core-ui/resources-minimized/([^/.'\"]+)\\.([^/.'\"]+)['\"]");
    }
}
