/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.plugins;

import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.datasource.ConnectionHelper;
import org.ametys.runtime.config.Config;
import org.ametys.runtime.plugin.ExtensionPoint;
import org.ametys.runtime.plugin.Feature;
import org.ametys.runtime.plugin.IncludePolicyFeatureActivator.IncludedFeature;
import org.ametys.runtime.plugin.PluginsManager;
import org.ametys.runtime.plugin.PluginsManager.InactivityCause;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeaturesHelper;
import org.ametys.runtime.test.framework.ScriptHelper;
import org.ametys.runtime.test.framework.TestConfigSet;
import org.ametys.runtime.test.framework.WithRuntimeFeaturesTestCase;

/**
 * Tests the PluginsManager
 */
public class PluginsTestCase extends AbstractRuntimeTestCase
{
    private static final List<IncludedFeature> __KERNEL_FEATURES = 
            NeedFeaturesHelper.toIncludedFeatures(
                    WithRuntimeFeaturesTestCase.class.getAnnotation(NeedFeatures.class), 
                    WithRuntimeFeaturesTestCase.class.getSimpleName()
            );
        
    /**
     * Tests the concept of plugin location
     * @throws Exception if an error occurs
     */
    @Test
    public void testLocation() throws Exception
    {
        TestConfigSet configSet = new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime02.xml")
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp1");
        try
        {
            startCocoon(configSet);
            Assert.assertTrue(PluginsManager.getInstance().getPluginNames().contains("test"));
            Assert.assertEquals("location1", PluginsManager.getInstance().getPluginLocation("test").getParentFile().getName());
        }
        finally
        {
            stopCocoon();
        }
    }
    
    /**
     * Tests the declaration of extension points
     * @throws Exception if an error occurs
     */
    @Test
    public void testExtensionPoint() throws Exception
    {
        TestConfigSet configSet = new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime11.xml")
                .withConfigFile("test/environments/configs/config1.xml")
                .withSqlFile(ScriptHelper.getDataSourceFile(ConnectionHelper.DATABASE_HSQLDB))
                .withWebappPath("test/environments/webapp1");
        try
        {
            startCocoon(configSet);
            ServiceManager serviceManager = getCocoonWrapper().getServiceManager();
            Assert.assertTrue(PluginsManager.getInstance().getPluginNames().contains("test"));
    
            Assert.assertNotNull(Config.getInstance());
            
            Assert.assertTrue(PluginsManager.getInstance().getComponents().contains("sep-test"));
                    
            Assert.assertTrue(serviceManager.lookup("sep-test") instanceof TestSingleExtensionPoint);
            
            Assert.assertTrue(PluginsManager.getInstance().getExtensionPoints().containsKey("ep-test"));
            
            ExtensionPoint testEP = (ExtensionPoint) serviceManager.lookup("ep-test");
            Assert.assertEquals(1, testEP.getExtensionsIds().size());
            TestExtension ext = (TestExtension) testEP.getExtension("org.ametys.runtime.test.EP");
            Assert.assertNotNull(ext);
            Assert.assertNull(ext.getContext());
            
            Assert.assertTrue(PluginsManager.getInstance().getExtensionPoints().containsKey("ep-component-test"));
            
            ExtensionPoint testCEP = (ExtensionPoint) serviceManager.lookup("ep-component-test");
            Assert.assertEquals(1, testCEP.getExtensionsIds().size());
            TestExtension ext2 = (TestExtension) testCEP.getExtension("org.ametys.runtime.test.EP");
            Assert.assertNotNull(ext2);
            Assert.assertNotNull(ext2.getContext());
        }
        finally
        {
            stopCocoon();
        }
    }
    
    /**
     * Tests active and inactive feature
     * @throws Exception if an error occurs
     */
    @Test
    public void testFeatures() throws Exception
    {
        TestConfigSet configSet = new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/exclude-policy/runtime12.xml")
                .withConfigFile("test/environments/configs/config1.xml")
                .withSqlFile(ScriptHelper.getDataSourceFile(ConnectionHelper.DATABASE_HSQLDB))
                .withWebappPath("test/environments/webapp1");
        try
        {
            startCocoon(configSet);
            Map<String, Feature> activeFeatures = PluginsManager.getInstance().getFeatures();
            Map<String, InactivityCause> inactiveFeatures = PluginsManager.getInstance().getInactiveFeatures();
    
            Assert.assertTrue(activeFeatures.containsKey("test/test-feature2"));
            Assert.assertTrue(activeFeatures.containsKey("test/test-feature3"));
            
            Assert.assertFalse(activeFeatures.containsKey("test/test-feature"));
            Assert.assertTrue(inactiveFeatures.containsKey("test/test-feature"));
            InactivityCause cause = inactiveFeatures.get("test/test-feature");
            Assert.assertEquals(cause, InactivityCause.DEPENDENCY);
            
            Assert.assertFalse(activeFeatures.containsKey("test/test-feature4"));
            Assert.assertTrue(inactiveFeatures.containsKey("test/test-feature4"));
            cause = inactiveFeatures.get("test/test-feature4");
            Assert.assertEquals(cause, InactivityCause.COMPONENT);
            
            Assert.assertFalse(activeFeatures.containsKey("test/test-feature5"));
            Assert.assertTrue(inactiveFeatures.containsKey("test/test-feature5"));
            cause = inactiveFeatures.get("test/test-feature5");
            Assert.assertEquals(cause, InactivityCause.EXCLUDED);
        }
        finally
        {
            stopCocoon();
        }
    }
    
    /**
     * Tests include policy for PluginsManager
     * @throws Exception if an error occurs
     */
    @Test
    public void testSimpleIncludePolicy() throws Exception
    {
        String includedFeatureCause = getTestName().getMethodName();
        List<IncludedFeature> featuresToInclude = Stream.concat(Stream.of(
                IncludedFeature.of("dummy-features/a", includedFeatureCause), 
                IncludedFeature.of("dummy-features/g", includedFeatureCause)
            ), __KERNEL_FEATURES.stream()
        ).collect(Collectors.toList());
        
        TestConfigSet configSet = new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/runtime-include.xml")
                .withConfigFile("test/environments/configs/config-include.xml")
                .withWebappPath("test/environments/webapp-test-include");
        try
        {
            startCocoon(configSet, featuresToInclude);
            Map<String, Feature> activeFeatures = PluginsManager.getInstance().getFeatures();
            Map<String, InactivityCause> inactiveFeatures = PluginsManager.getInstance().getInactiveFeatures();
            
            _assertActiveFeatures(activeFeatures, inactiveFeatures);
            
            Assert.assertTrue(activeFeatures.containsKey("dummy-features/g"));
            Assert.assertTrue(activeFeatures.containsKey("dummy-features/h"));
            
            Assert.assertFalse(activeFeatures.containsKey("dummy-features/g_alt"));
            Assert.assertTrue(inactiveFeatures.containsKey("dummy-features/g_alt"));
            InactivityCause cause = inactiveFeatures.get("dummy-features/g_alt");
            Assert.assertEquals(InactivityCause.UNUSED, cause);
            
            Assert.assertFalse(activeFeatures.containsKey("dummy-features/g_alt_dep"));
            Assert.assertTrue(inactiveFeatures.containsKey("dummy-features/g_alt_dep"));
            cause = inactiveFeatures.get("dummy-features/g_alt_dep");
            Assert.assertEquals(InactivityCause.UNUSED, cause);
        }
        finally
        {
            stopCocoon();
        }
    }
    
    /**
     * Tests include policy for PluginsManager
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncludePolicyWithOverrideByDependency() throws Exception
    {
        // 'dummy-features/feature_which_needs_g_alt' depends on 'dummy-features/g_alt' which overrides 'dummy-features/g', even though it is declared as an included feature
        String includedFeatureCause = getTestName().getMethodName();
        List<IncludedFeature> featuresToInclude = Stream.concat(Stream.of(
                IncludedFeature.of("dummy-features/a", includedFeatureCause), 
                IncludedFeature.of("dummy-features/g", includedFeatureCause), 
                IncludedFeature.of("dummy-features/feature_which_needs_g_alt", includedFeatureCause)
            ), __KERNEL_FEATURES.stream()
        ).collect(Collectors.toList());
        
        TestConfigSet configSet = new TestConfigSet()
                .withRuntimeFile("test/environments/runtimes/runtime-include.xml")
                .withConfigFile("test/environments/configs/config-include.xml")
                .withWebappPath("test/environments/webapp-test-include");
        try
        {
            startCocoon(configSet, featuresToInclude);
            Map<String, Feature> activeFeatures = PluginsManager.getInstance().getFeatures();
            Map<String, InactivityCause> inactiveFeatures = PluginsManager.getInstance().getInactiveFeatures();
            
            _assertActiveFeatures(activeFeatures, inactiveFeatures);
            
            Assert.assertFalse(activeFeatures.containsKey("dummy-features/g"));
            Assert.assertTrue(inactiveFeatures.containsKey("dummy-features/g"));
            InactivityCause cause = inactiveFeatures.get("dummy-features/g");
            Assert.assertEquals(InactivityCause.OVERRIDDEN, cause); // overridden by 'dummy-features/g_alt'
            
            Assert.assertFalse(activeFeatures.containsKey("dummy-features/h"));
            Assert.assertTrue(inactiveFeatures.containsKey("dummy-features/h"));
            cause = inactiveFeatures.get("dummy-features/h");
            Assert.assertEquals(InactivityCause.UNUSED, cause);
            
            Assert.assertTrue(activeFeatures.containsKey("dummy-features/feature_which_needs_g_alt"));
            Assert.assertTrue(activeFeatures.containsKey("dummy-features/g_alt"));
            Assert.assertTrue(activeFeatures.containsKey("dummy-features/g_alt_dep"));
        }
        finally
        {
            stopCocoon();
        }
    }
    
    private void _assertActiveFeatures(Map<String, Feature> activeFeatures, Map<String, InactivityCause> inactiveFeatures)
    {
        Assert.assertTrue(activeFeatures.containsKey("dummy-features/a"));
        Assert.assertTrue(activeFeatures.containsKey("dummy-features/b"));
        Assert.assertTrue(activeFeatures.containsKey("dummy-features/c"));
        Assert.assertTrue(activeFeatures.containsKey("dummy-features/d"));
        Assert.assertTrue(activeFeatures.containsKey("dummy-features/e"));
        Assert.assertTrue(activeFeatures.containsKey("dummy-features/f"));
        
        Assert.assertFalse(activeFeatures.containsKey("dummy-features/i"));
        Assert.assertTrue(inactiveFeatures.containsKey("dummy-features/i"));
        InactivityCause cause = inactiveFeatures.get("dummy-features/i");
        Assert.assertEquals(InactivityCause.UNUSED, cause);
        
        Assert.assertFalse(activeFeatures.containsKey("dummy-features/j"));
        Assert.assertTrue(inactiveFeatures.containsKey("dummy-features/j"));
        cause = inactiveFeatures.get("dummy-features/j");
        Assert.assertEquals(InactivityCause.UNUSED, cause);
        
        Assert.assertFalse(activeFeatures.containsKey("dummy-features/k"));
        Assert.assertTrue(inactiveFeatures.containsKey("dummy-features/k"));
        cause = inactiveFeatures.get("dummy-features/k");
        Assert.assertEquals(InactivityCause.UNUSED, cause);
        
        Assert.assertFalse(activeFeatures.containsKey("dummy-features/l"));
        Assert.assertTrue(inactiveFeatures.containsKey("dummy-features/l"));
        cause = inactiveFeatures.get("dummy-features/l");
        Assert.assertEquals(InactivityCause.UNUSED, cause);
    }
}
