/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package org.ametys.runtime.test.resources;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.util.function.Predicate;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.minimize.css.sass.MinimizeSassManager;
import org.ametys.plugins.core.ui.resources.css.JSASSResourceURIExtensionPoint;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

import io.bit3.jsass.CompilationException;

/**
 * Test case for the {@link JSASSResourceURIExtensionPoint}
 */
@NeedFeatures({
    "core-ui/plugin.core.ui.minimize.resources.uri",
    "test/test.jsass-resource-uri.supported-context-resource-uri",
    "core/runtime.core.resources-handler",
})
public class JsassResourceUriTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static JSASSResourceURIExtensionPoint _jsassResourceURIEP;
    private static MinimizeSassManager _minimizeSassManager;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp2");
    }
    
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _jsassResourceURIEP = (JSASSResourceURIExtensionPoint) manager.lookup(JSASSResourceURIExtensionPoint.ROLE);
        _minimizeSassManager = (MinimizeSassManager) manager.lookup(MinimizeSassManager.ROLE);
    }
    
    /**
     * Test the extension point for replacing protocols to path
     * @throws URISyntaxException If an invalid URI is used
     */
    @Test
    public void testReplaceResourceUriProtocols() throws URISyntaxException
    {
        Assert.assertEquals("An empty URI should not be modified", 
                "", 
                _jsassResourceURIEP.resolve(""));
        
        Assert.assertEquals("An URI without protocol should not be modified", 
                "/test/without/protocol", 
                _jsassResourceURIEP.resolve("/test/without/protocol"));
        
        Assert.assertEquals("An URI with an unknown protocol should not be modified", 
                "unknown-protocol://test", 
                _jsassResourceURIEP.resolve("unknown-protocol://test"));
        
        Assert.assertEquals("An URI with a plugin protocol should be transformed into a plugin path", 
                "/plugins/core-ui/resources/path", 
                _jsassResourceURIEP.resolve("plugin://resources/path"));
        
        Assert.assertEquals("An URI with a specific plugin protocol should be transformed into the plugin specific path", 
                "/plugins/specific/resources/path", 
                _jsassResourceURIEP.resolve("plugin:specific://resources/path"));
        
        Assert.assertEquals("An URI with a workspace protocol should be transformed into a workspace path", 
                "/_test/path", 
                _jsassResourceURIEP.resolve("workspace:test://path"));
        
        try
        {
            _jsassResourceURIEP.resolve("context://unknown-context/path");
            Assert.fail("An URI with an unknown context:// should throw an URISyntaxException");
        }
        catch (URISyntaxException e)
        {
            // As expected
        }
        
        Assert.assertEquals("An URI with a known context protocol should be replaced with the corresponding path", 
                "/plugins/test/path-with-context/path", 
                _jsassResourceURIEP.resolve("context://other-location/test-context/path"));
    }
    
    /**
     * Test the extension point for replacing path with protocols
     * @throws URISyntaxException If an invalid URI is used
     */
    @Test
    public void testReplaceResourcePath() throws URISyntaxException
    {
        Assert.assertEquals("An URI with no known path to protocol mapping should be mapped to cocoon:/", 
                "cocoon://path", 
                _jsassResourceURIEP.resolvePath("/path"));

        Assert.assertEquals("An URI with a protocol should stay untouched", 
                "protocol://path", 
                _jsassResourceURIEP.resolvePath("protocol://path"));

        Assert.assertEquals("An URI with a plugin path should be resolved using plugin:<plugin>:// protocol", 
                "plugin:test://resources/path", 
                _jsassResourceURIEP.resolvePath("/plugins/test/resources/path"));
        
        Assert.assertEquals("An URI with a workspace path should be resolved using workspace:<workspace>:// protocol", 
                "workspace:test://path", 
                _jsassResourceURIEP.resolvePath("/_test/path"));
    }
    
    /**
     * Test the {@link JSASSResourceURIExtensionPoint} when compiling a sass file
     * @throws URISyntaxException If an error occurred
     * @throws CompilationException If an error occurred
     * @throws IOException If an error occurred
     * @throws MalformedURLException  If an error occurred
     */
    @Test
    public void testCompileSassFile() throws URISyntaxException, CompilationException, MalformedURLException, IOException
    {
        String sassContent = ".test { .subclass { color: red; } }";
        String expectedOutput = ".test .subclass {" + System.lineSeparator() + "  color: red; }";
        
        Predicate<String> hasSourceMappingUrl = output -> output.contains("/*# sourceMappingURL=");
        
        // Location ok : css + source map
        String compiledCssWithValidLocation = _minimizeSassManager.compileCss(sassContent, "plugin:test://resources/path/file.css", "plugin:test://resources/path/file.scss", false, 0);
        
        Assert.assertTrue("The sass content with a valid location should be generated with a source map",
                isExpectedOutput(expectedOutput).and(hasSourceMappingUrl).test(compiledCssWithValidLocation));
        
        
        // location with unknown protocol : throw
        try
        {
            _minimizeSassManager.compileCss(sassContent, "unknown-protocol://path/file.css", "unknown-protocol://path/file.scss", false, 0);
            Assert.fail("Compiling a SASS file with an unknown protocol in the location should throw an URISyntaxException");
        }
        catch (URISyntaxException e)
        {
            // As expected
        }
        
        // Location with unknown context : css without source map
        String compileCssWithUnknownContext = _minimizeSassManager.compileCss(sassContent, "context://unknown-context/path/file.css", "context://unknown-context/path/file.scss", false, 0);

        Assert.assertTrue("The sass content with an unknown context in the location should be generated without a source map",
                isExpectedOutput(expectedOutput).and(hasSourceMappingUrl.negate()).test(compileCssWithUnknownContext));
        
        // Location ok with context and indirect import : css with source map
        String compiledCssWithValidContext = _minimizeSassManager.compileCss(
                "@import 'context-imported';",
                "cocoon://plugins/test/path-with-context/context.scss", 
                "context://other-location/test-context/context.css", 
                false, 0);
        
        expectedOutput = ".indirect-context-imported {" + System.lineSeparator()
                + "  color: blue;" + System.lineSeparator()
                + "  background: url(\"/plugins/test/path-with-context/test-relative-path/image2.png\"); }" + System.lineSeparator()
                + System.lineSeparator()
                + ".context-imported {" + System.lineSeparator()
                + "  color: white;" + System.lineSeparator()
                + "  background: url(\"/plugins/test/path-with-context/test-relative-path/image1.png\"); }";
        
        Assert.assertTrue("The sass content with a valid context and import should be generated with a source map, and relative path should be replaced",
                isExpectedOutput(expectedOutput).and(hasSourceMappingUrl).test(compiledCssWithValidContext));
    }
    
    private Predicate<String> isExpectedOutput(String expectedOutput)
    {
        return output -> StringUtils.isNotEmpty(output) && output.startsWith(expectedOutput);
    }
}
