/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.resources;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.minimize.SourceMapCache;
import org.ametys.core.minimize.css.MinimizeCSSManager;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * Test the validity of source maps generated with a context path
 */
@NeedFeatures({
    "core/runtime.core",
    "core/runtime.core.resources-handler",
    "core/runtime.core.resources.contextpath"
})
public class SourceMapContextPathTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{

    private static SourceResolver _resolver;
    private static MinimizeCSSManager _minimizeCssManager;
    private SourceMapCache _sourceMapCache;

    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp2")
                .withContextPath("/contextPath");
    }
    
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _minimizeCssManager = (MinimizeCSSManager) manager.lookup(MinimizeCSSManager.ROLE);
        _sourceMapCache = (SourceMapCache) manager.lookup(SourceMapCache.ROLE);
    }

    /**
     * Test the source map's sources with imports
     * @throws Exception If an error occurred
     */
    @Test
    public void testSourceMapSources() throws Exception
    {
        Source inputSource = _resolver.resolveURI("plugin:test://resources/css/minimize/b.css");
        String cssContent = IOUtils.toString(inputSource.getInputStream(), "UTF-8");
        
        String minimizedCss = _minimizeCssManager.minimizeCss(cssContent, "/plugins/test/resources/css/minimize/b.css", "testSourceMapSources.css", 0L);
        Assert.assertTrue("Failed to minimize css", StringUtils.isNotEmpty(minimizedCss));
        
        Source sourceMapSource = _sourceMapCache.get("testSourceMapSources.css");
        String sourceMap = IOUtils.toString(sourceMapSource.getInputStream(), "UTF-8");
        Assert.assertTrue("The source map could not be generated", StringUtils.isNotEmpty(sourceMap));
        
        Assert.assertTrue("Expected source '/contextPath/plugins/test/resources/css/minimize/b.css' not found in source map :\n"
                + sourceMap,
                sourceMap.contains("\"/contextPath/plugins/test/resources/css/minimize/b.css\""));
    }
}
