/*
 *  Copyright 2018 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.resources;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.function.BiPredicate;
import java.util.function.Predicate;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceNotFoundException;
import org.apache.excalibur.source.SourceResolver;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.minimize.SourceMapCache;
import org.ametys.core.minimize.css.MinimizeCSSManager;
import org.ametys.core.minimize.css.sass.MinimizeSassManager;
import org.ametys.core.util.JSONUtils;
import org.ametys.plugins.core.ui.minimize.HashCache.UriData;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

import com.google.common.base.Predicates;
import com.google.common.collect.ImmutableList;

/**
 * Test the validity of source maps generated
 */
@NeedFeatures({
    "core/runtime.core",
    "core/runtime.core.resources-handler",
    "core/runtime.core.resources.contextpath",
    "core/util.jsonutils",
    "core-ui/plugin.core.ui.minimize.resources.uri",
    "core-ui/runtime.core.ui.resources-handler",
})
public class SourceMapTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{

    private static SourceResolver _resolver;
    private static MinimizeSassManager _minimizeSassManager;
    private static JSONUtils _jsonUtils;
    private static MinimizeCSSManager _minimizeCssManager;
    private static SourceMapCache _sourceMapCache;

    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp2");
    }
    
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _resolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
        _minimizeSassManager = (MinimizeSassManager) manager.lookup(MinimizeSassManager.ROLE);
        _jsonUtils = (JSONUtils) manager.lookup(JSONUtils.ROLE);
        _minimizeCssManager = (MinimizeCSSManager) manager.lookup(MinimizeCSSManager.ROLE);
        _sourceMapCache = (SourceMapCache) manager.lookup(SourceMapCache.ROLE);
    }

    /**
     * Check the structure of generated source map files
     * @throws Exception If an error occurred
     */
    @Test
    public void testSourceMapStructure() throws Exception
    {
        String defaultInput = ".test { .subclass { color: red; } }";
        String sourceMap = _minimizeSassManager.generateCssSourceMap(defaultInput, "plugin:core-ui://resources/example/test.scss.map", 
                "scss", "/internalContextPath", "/externalContextPath", false, 0);

        Map<String, Object> sourceMapJson = _jsonUtils.convertJsonToMap(sourceMap);

        BiPredicate<String, Predicate<Object>> checkSourceMapFieldValidity = (field, test) -> sourceMapJson.containsKey(field) && test.test(sourceMapJson.get(field));
        
        Assert.assertTrue("Only source map version 3 are supported", 
                checkSourceMapFieldValidity.test("version", val -> (int) val == 3));
        
        Assert.assertTrue("Source maps must reference the target file", 
                checkSourceMapFieldValidity.test("file", "test.css"::equals));
        
        Assert.assertTrue("Source maps must reference the original sources", 
                checkSourceMapFieldValidity.test("sources", Predicates.and(
                    List.class::isInstance, 
                    list -> ((List) list).contains("/externalContextPath/plugins/core-ui/resources/example/test.scss"))
                ));
        
        Assert.assertTrue("Source maps must contain the mapping to the original file", 
                checkSourceMapFieldValidity.test("mappings", "AAAA,AAAQ,KAAH,CAAG,SAAS,CAAC;EAAE,KAAK,EAAE,GAAG,GAAI"::equals));
    }
    
    /**
     * Test the source map's sources with imports
     * @throws Exception If an error occurred
     */
    @Test
    public void testSourceMapSources() throws Exception
    {
        Source inputSource = _resolver.resolveURI("plugin:test://resources/css/sass-relative-path/main.scss");
        String sassContent = IOUtils.toString(inputSource.getInputStream(), "UTF-8");
        
        String sourceMap = _minimizeSassManager.generateCssSourceMap(sassContent, "plugin:test://resources/css/sass-relative-path/main.scss.map", 
                "scss", "", "", false, 0);
        
        Assert.assertTrue("The source map could not be generated", StringUtils.isNotEmpty(sourceMap));
        
        Map<String, Object> sourceMapJson = _jsonUtils.convertJsonToMap(sourceMap);
        
        Assert.assertTrue("The source map does not contains a sources list", sourceMapJson != null && sourceMapJson.containsKey("sources") && sourceMapJson.get("sources") instanceof List);
        
        String[] expectedSources = new String[] {
            "/plugins/test/resources/css/sass-relative-path/main.scss",
            "/plugins/test/resources/css/sass-relative-path/scss_same_folder.scss",
            "/plugins/test/resources/css/sass-relative-path/scss_same_folder_prefix_underscore.scss",
            "/plugins/test/resources/css/sass-relative-path/folder1/scss_folder1.scss",
            "/plugins/test/resources/css/sass-relative-path/folder1/scss_folder1_2.scss",
            "/plugins/test/resources/css/sass-relative-path/folder2/imported_by_plugin_name.css",
            "/plugins/test/resources/css/sass-relative-path/folder1/subfolder1/import-this-css.css",
        };
        
        @SuppressWarnings("unchecked")
        List<String> sources = (List<String>) sourceMapJson.get("sources");
        while (sources.remove("/plugins/core/resources/css/jsass_injected_file_placeholder.css")) 
        {
            // remove all "jsass_
        }
        
        for (String expectedSource : expectedSources)
        {
            if (!(sources.contains(expectedSource) && sources.remove(expectedSource)))
            {
                Assert.fail("Expected source '" + expectedSource + "' not found in source map");
            }
        }
        
        if (sources.size() > 0)
        {
            Assert.fail("Unexpected sources found in source map file '" + sources.toString() + "'");
        }
    }
    
    /**
     * Test the mapping of an aggregated source map
     * @throws SourceNotFoundException If an error occurred
     * @throws IOException If an error occurred
     */
    @Test
    public void testSourceMapAggregatedMapping() throws SourceNotFoundException, IOException
    {
        ImmutableList<UriData> uriDataList = ImmutableList.of(
            new UriData("/plugins/test/resources/css/minimize/h.css", true),
            new UriData("/plugins/test/resources/css/sass-relative-path/main.css", true)
        );
        
        // Aggregates the minimized file, and stores the source map in the _sourceMapCache component
        _minimizeCssManager.minimizeAndAggregateURIs(uriDataList, "testSourceMapAggregatedMapping.css", true);
        
        Source sourceMap = _sourceMapCache.get("testSourceMapAggregatedMapping.css.map");
        String sourceMapContent = IOUtils.toString(sourceMap.getInputStream(), "UTF-8");
        
        Map<String, Object> sourceMapJson = _jsonUtils.convertJsonToMap(sourceMapContent);
        Assert.assertTrue("Source map must contain a mapping", sourceMapJson.containsKey("mappings"));
        
        Assert.assertEquals("Source map mapping is not valid", 
                "A;" // first line first character
                + "AAAA,EAAG,CACC,gBAAgB,CAAE,yDAAyD,CAC3E,KAAK,CAAE,GAAG,CACb,C;" // Mapping of h.css
                + ";" // Empty comment line /* File : */
                + "ACDD,OAAO,CAAP,oBAAO," // main.scss (line 3 not imported)
                
                + "CCFP,OAAO,CAAP,iGAAO," // folder2/imported_by_plugin_name.css (line 1 and 2)
                + "CACP," // correct path in folder2/imported_by_plugin_name.css (line 1 and 2)
                + "OAAO,CAAP,qGAAO," // folder2/imported_by_plugin_name.css (line 1 and 2)
                
                + "CCDP,OAAO,CAAP,2CAAO," // folder1/subfolder1/import-this-css.css (line 1) with correct mapping to 'relative-import.css' and after
                + "CCAP,iBAAkB,CACd,KAAK,CAAE,IAAI," // scss_same_folder.scss
                + "CCDf,mCAAoC,CAChC,KAAK,CAAE,IAAI," // scss_same_folder_prefix_underscore.scss
                + "CCDf,aAAc,CACV,gBAAgB,CAAE,sBAAsB," // folder2/imported_by_plugin_name.css with updated resource path (line 4 and 5)
                + "CJE5C,wBAAyB,CACrB,gBAAgB,CAAE,4GAA4G," // folder1/scss_folder1.scss
                + "CKFlI,eAAgB,CACZ,KAAK,CAAE,IAAI," // folder1/scss_folder1_2.scss (line 3 and 4)
                + "CJDf,gBAAiB,CACb,KAAK,CAAE,IAAI,CACd;" // folder1/subfolder1/import-this-css.css (line 3 and 4)
            , sourceMapJson.get("mappings"));
    }
}
