/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.rights.manager;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.environment.Request;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.group.Group;
import org.ametys.core.right.Profile;
import org.ametys.core.right.ProfileAssignmentStorageExtensionPoint;
import org.ametys.core.right.RightManager;
import org.ametys.core.right.RightManager.RightResult;
import org.ametys.core.right.RightProfilesDAO;
import org.ametys.core.user.UserIdentity;
import org.ametys.runtime.test.AbstractJDBCTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;
import org.ametys.runtime.test.framework.Unoptimized;
import org.ametys.runtime.workspace.WorkspaceMatcher;

/**
 * Tests the {@link RightManager}
 */
@Unoptimized
@NeedFeatures({"core/runtime.rights.manager", "core/runtime.rights.profiles.dao", "rights-test/profile.assignment.test.jdbc", "rights-test/rights.access.controller.test.string", "core/reader.profile.init"})
public abstract class AbstractRightManagerTestCase extends AbstractJDBCTestCase
{
    private static RightManager _rightManager;
    private static ProfileAssignmentStorageExtensionPoint _profileAssignmentStorageEP;
    private static RightProfilesDAO _profilesDAO;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp4");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _rightManager = (RightManager) manager.lookup(RightManager.ROLE);
        _profileAssignmentStorageEP = (ProfileAssignmentStorageExtensionPoint) manager.lookup(ProfileAssignmentStorageExtensionPoint.ROLE);
        _profilesDAO = (RightProfilesDAO) manager.lookup(RightProfilesDAO.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        _userDirectory = _createUserDirectory();
        _groupDirectory = _createGroupDirectory();
    }
    
    /**
     * Provide the scripts to run to populate the database.
     * @return the scripts to run.
     */
    protected File[] getPopulateScripts()
    {
        return new File[] {new File("test/environments/scripts/jdbc/" + getDBType() + "/fillProfileRights.sql")};
    }
    
    @Override
    protected void essentialsAfter() throws Exception
    {
        _deleteGroupDirectory();
        super.essentialsAfter();
    }
    
    /**
     * Test an empty db
     * @throws Exception if an error occurs
     */
    @Test
    public void testEmpty() throws Exception
    {
        getRequest().setAttribute("populationContexts", new ArrayList<>());
        
        Set<UserIdentity> users = _rightManager.getAllowedUsers("Runtime_Rights_User_Handle", "foo").resolveAllowedUsers(true);
        Assert.assertEquals(0, users.size());
        
        Assert.assertTrue(_rightManager.getUserRights(new UserIdentity("foo", "population"), "/foo").isEmpty());
        
        List<Profile> profiles = _profilesDAO.getProfiles();
        Assert.assertEquals(1, profiles.size()); // There is the READER profile only
        
        _profilesDAO.clearCache();
    }
    
    /**
     * Test a filled db
     * @throws Exception if an error occurs
     */
    @Test
    public void testFilled() throws Exception
    {
        Request request = getRequest();
        request.setAttribute("populationContexts", new ArrayList<>());
        
        Set<UserIdentity> users;
        UserIdentity test = new UserIdentity("test", "population");
        UserIdentity test2 = new UserIdentity("test2", "population");
        
        // Be sure it is empty before filling and tables are created
        Assert.assertEquals(0, _userDirectory.getUsers().size());
        Assert.assertEquals(0, _groupDirectory.getGroups().size());
        // Fill DB
        _setDatabase(Arrays.asList(getPopulateScripts()));
        
        users = _rightManager.getAllowedUsers("right1", "/test1").resolveAllowedUsers(true);
        Assert.assertEquals(1, users.size());
        Assert.assertTrue(users.contains(test));
        users = _rightManager.getAllowedUsers("right1", "/test2").resolveAllowedUsers(true);
        Assert.assertEquals(0, users.size());
        users = _rightManager.getAllowedUsers("right3", "/test1").resolveAllowedUsers(true);
        Assert.assertEquals(0, users.size());
        
        Set<String> rights;
        rights = _rightManager.getUserRights(test, "/test1");
        Assert.assertEquals(3, rights.size());
        rights = _rightManager.getUserRights(test, "/test2");
        Assert.assertEquals(0, rights.size());
        rights = _rightManager.getUserRights(test2, "/test1");
        Assert.assertEquals(0, rights.size());
        rights = _rightManager.getUserRights(test2, "/test2");
        Assert.assertEquals(0, rights.size());
        
        request.setAttribute(WorkspaceMatcher.WORKSPACE_NAME, "/test");
        
        RightResult hasRight = _rightManager.hasRight(test, "right4", null);
        Assert.assertEquals(RightResult.RIGHT_ALLOW, hasRight);
        hasRight = _rightManager.hasRight(test, "right3", null);
        Assert.assertEquals(RightResult.RIGHT_UNKNOWN, hasRight);
        
        hasRight = _rightManager.hasRight(test, "right1", "/test1");
        Assert.assertEquals(RightResult.RIGHT_ALLOW, hasRight);
        hasRight = _rightManager.hasRight(test, "right1", "/testfoo");
        Assert.assertEquals(RightResult.RIGHT_UNKNOWN, hasRight);
        
        _profilesDAO.clearCache();
    }
    
    /**
     * Test profiles methods
     * @throws Exception if an error occurs
     */
    @Test
    public void testProfiles() throws Exception
    {
        List<Profile> profiles;
        Profile profile;
        List<String> rights;
        
        profile = _profilesDAO.addProfile("Profile 1");
        profile = _profilesDAO.getProfile(profile.getId());
        Assert.assertEquals("Profile 1", profile.getLabel());
        
        profiles = _profilesDAO.getProfiles();
        Assert.assertEquals(2, profiles.size()); // The added profile "Profile 1" and READER
        for (Profile currentProfile : profiles)
        {
            if (!RightManager.READER_PROFILE_ID.equals(currentProfile.getId()))
            {
                Assert.assertEquals("Profile 1", currentProfile.getLabel());
            }
        }
        
        _profilesDAO.renameProfile(profile, "Profile 1 renamed");
        _profilesDAO.addRight(profile, "right1");
        _profilesDAO.addRight(profile, "right2");
        
        profile = _profilesDAO.getProfile(profile.getId());
        Assert.assertEquals("Profile 1 renamed", profile.getLabel());
        rights = _profilesDAO.getRights(profile);
        Assert.assertEquals(2, rights.size());
        Assert.assertTrue(rights.contains("right1"));
        Assert.assertTrue(rights.contains("right2"));
        
        _profilesDAO.removeRights(profile);
        profile = _profilesDAO.getProfile(profile.getId());
        rights = _profilesDAO.getRights(profile);
        Assert.assertEquals(0, rights.size());
        
        _profilesDAO.deleteProfile(profile);
        profile = _profilesDAO.getProfile(profile.getId());
        Assert.assertNull(profile);
        
        profiles = _profilesDAO.getProfiles();
        Assert.assertEquals(1, profiles.size()); // Only READER profile
        
        _profilesDAO.clearCache();
        
    }
    
    /**
     * Test rights modifications
     * @throws Exception if an error occurs
     */
    @Test
    public void testRight() throws Exception
    {
        Request request = getRequest();
        request.setAttribute(WorkspaceMatcher.WORKSPACE_NAME, "/test");
        
        // Be sure it is empty before filling and tables are created
        Assert.assertEquals(0, _userDirectory.getUsers().size());
        Assert.assertEquals(0, _groupDirectory.getGroups().size());
        // Fill DB
        _setDatabase(Arrays.asList(getPopulateScripts()));
        _profilesDAO.clearCache();
        
        Set<String> rights;
        RightResult rightResult;
        
        UserIdentity test = new UserIdentity("test", "population");
        
        for (Profile profile : _profilesDAO.getProfiles())
        {
            if (!RightManager.READER_PROFILE_ID.equals(profile.getId()))
            {
                _profilesDAO.deleteProfile(profile);
            }
        }
        
        Profile profile1 = _profilesDAO.addProfile("MyProfil1");
        _profilesDAO.addRight(profile1, "right1");
        _profilesDAO.addRight(profile1, "right2");
        Profile profile2 = _profilesDAO.addProfile("MyProfil2");
        _profilesDAO.addRight(profile2, "right3");

        // USER
        _profileAssignmentStorageEP.allowProfileToUser(test, profile1.getId(), "/test1");
        rights = _rightManager.getUserRights(test, "/test1");
        Assert.assertEquals(2, rights.size());
        rightResult = _rightManager.hasRight(test, "right1", "/test1");
        Assert.assertEquals(RightResult.RIGHT_ALLOW, rightResult);
        rightResult = _rightManager.hasRight(test, "right1", null);
        Assert.assertEquals(RightResult.RIGHT_ALLOW, rightResult);
        rightResult = _rightManager.hasRight(test, "right3", "/test1");
        Assert.assertEquals(RightResult.RIGHT_UNKNOWN, rightResult);
        rightResult = _rightManager.hasRight(test, "right3", null);
        Assert.assertEquals(RightResult.RIGHT_UNKNOWN, rightResult);
        
        // Need to clear cache when modifying affectation 
        _rightManager.clearCaches();
        
        _profileAssignmentStorageEP.allowProfileToUser(test, profile2.getId(), "/test1");
        rights = _rightManager.getUserRights(test, "/test1");
        Assert.assertEquals(3, rights.size());
        rightResult = _rightManager.hasRight(test, "right3", "/test1");
        Assert.assertEquals(RightResult.RIGHT_ALLOW, rightResult);
        
        // Need to clear cache when modifying affectation 
        _rightManager.clearCaches();
        
        _profileAssignmentStorageEP.removeAllowedProfileFromUser(test, profile2.getId(), "/test1");
        rights = _rightManager.getUserRights(test, "/test1");
        Assert.assertEquals(2, rights.size());
        rightResult = _rightManager.hasRight(test, "right3", "/test1");
        Assert.assertEquals(RightResult.RIGHT_UNKNOWN, rightResult);
        
        // Need to clear cache when modifying affectation 
        _rightManager.clearCaches();
        
        _profileAssignmentStorageEP.allowProfileToUser(test, profile2.getId(), "/test2");
        rights = _rightManager.getUserRights(test, "/test1");
        Assert.assertEquals(2, rights.size());
        rights = _rightManager.getUserRights(test, "/test2");
        Assert.assertEquals(1, rights.size());
        
        // Need to clear cache when modifying affectation 
        _rightManager.clearCaches();

        _profileAssignmentStorageEP.removeAllowedProfileFromUser(test, profile1.getId(), "/test1");
        _profileAssignmentStorageEP.removeAllowedProfileFromUser(test, profile2.getId(), "/test1");
        rights = _rightManager.getUserRights(test, "/test1");
        Assert.assertEquals(0, rights.size());
        rights = _rightManager.getUserRights(test, "/test2");
        Assert.assertEquals(1, rights.size());
        
        // Need to clear cache when modifying affectation 
        _rightManager.clearCaches();

        _profileAssignmentStorageEP.removeAllowedProfileFromUser(test, profile1.getId(), "/test2");
        _profileAssignmentStorageEP.removeAllowedProfileFromUser(test, profile2.getId(), "/test2");
        rights = _rightManager.getUserRights(test, "/test1");
        Assert.assertEquals(0, rights.size());
        rights = _rightManager.getUserRights(test, "/test2");
        Assert.assertEquals(0, rights.size());

        
        // GROUP
        Group group = _groupDirectory.getGroups().iterator().next(); 
        UserIdentity userIdentity = group.getUsers().iterator().next();

        _profileAssignmentStorageEP.allowProfileToGroup(group.getIdentity(), profile1.getId(), "/test1");
        rights = _rightManager.getUserRights(userIdentity, "/test1");
        Assert.assertEquals(2, rights.size());
        
        _profileAssignmentStorageEP.allowProfileToGroup(group.getIdentity(), profile2.getId(), "/test1");
        rights = _rightManager.getUserRights(userIdentity, "/test1");
        Assert.assertEquals(3, rights.size());
        
        _profileAssignmentStorageEP.removeAllowedProfileFromGroup(group.getIdentity(), profile2.getId(), "/test1");
        rights = _rightManager.getUserRights(userIdentity, "/test1");
        Assert.assertEquals(2, rights.size());
        
        _profileAssignmentStorageEP.allowProfileToGroup(group.getIdentity(), profile2.getId(), "/test2");
        rights = _rightManager.getUserRights(userIdentity, "/test1");
        Assert.assertEquals(2, rights.size());
        rights = _rightManager.getUserRights(userIdentity, "/test2");
        Assert.assertEquals(1, rights.size());

        _profileAssignmentStorageEP.removeAllowedProfileFromGroup(group.getIdentity(), profile1.getId(), "/test1");
        _profileAssignmentStorageEP.removeAllowedProfileFromGroup(group.getIdentity(), profile2.getId(), "/test1");
        rights = _rightManager.getUserRights(userIdentity, "/test1");
        Assert.assertEquals(0, rights.size());
        rights = _rightManager.getUserRights(userIdentity, "/test2");
        Assert.assertEquals(1, rights.size());

        _profileAssignmentStorageEP.removeAllowedProfileFromGroup(group.getIdentity(), profile1.getId(), "/test2");
        _profileAssignmentStorageEP.removeAllowedProfileFromGroup(group.getIdentity(), profile2.getId(), "/test2");
        rights = _rightManager.getUserRights(userIdentity, "/test1");
        Assert.assertEquals(0, rights.size());
        rights = _rightManager.getUserRights(userIdentity, "/test2");
        Assert.assertEquals(0, rights.size());
        
        _profilesDAO.clearCache();
    }
}
