/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.rights.storage.assignments;

import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.group.GroupIdentity;
import org.ametys.core.right.AccessController.AccessResult;
import org.ametys.core.right.ModifiableProfileAssignmentStorage;
import org.ametys.core.right.ProfileAssignmentStorageExtensionPoint;
import org.ametys.core.right.ProfileAssignmentStorageExtensionPoint.AccessResultInfo;
import org.ametys.core.user.UserIdentity;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;


/**
 * Test the {@link ProfileAssignmentStorageExtensionPoint}
 */
@NeedFeatures("rights-test/profile.assignment.test.jdbc")
public class ProfileAssignmentStorageExtensionPointTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    /** Profile assignment storage for context '/test1' */
    protected static ModifiableProfileAssignmentStorage _profileAssignmentStorage1;
    /** Profile assignment storage for context '/test2' */
    protected static ModifiableProfileAssignmentStorage _profileAssignmentStorage2;
    /** Profile assignment storage extension point */
    protected static ProfileAssignmentStorageExtensionPoint _profileAssignmentStorageEP;
    
    private static final GroupIdentity _GROUP2 = new GroupIdentity("group2", "bar");
    private static final GroupIdentity _GROUP1 = new GroupIdentity("group1", "bar");
    private static final UserIdentity _USER1 = new UserIdentity("user1", "foo");
    
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp4");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _profileAssignmentStorageEP = (ProfileAssignmentStorageExtensionPoint) manager.lookup(ProfileAssignmentStorageExtensionPoint.ROLE);
        _profileAssignmentStorage1 = (ModifiableProfileAssignmentStorage) _profileAssignmentStorageEP.getExtension("profile.assignment.test.JdbcProfileAssignmentStorage1");
        _profileAssignmentStorage2 = (ModifiableProfileAssignmentStorage) _profileAssignmentStorageEP.getExtension("profile.assignment.test.JdbcProfileAssignmentStorage2");
    }
    
    /**
     * Test the {@link ProfileAssignmentStorageExtensionPoint#getAllPermissions} method
     * The test validate the result of the method for :
     * - different access result on the same context
     * - multiple profiles assigned to the same context and target
     */
    @Test
    public void testGetAllPermissions()
    {
        String profile1 = "foo";
        String profile2 = "bar";
        String profile3 = "baz";
        Set<String> profileSet1 = Set.of(profile1, profile2);
        Set<String> profileSet2 = Set.of(profile1, profile3);
        
        /**
         * Single assignment on context
         */
        String anonymousAllowedCtx = "/test1/ctx1";
        _profileAssignmentStorage1.addAllowedProfilesForAnonymous(anonymousAllowedCtx, profileSet1);
        
        /**
         * Multiple assignments on context with different profile set
         */
        String anonymousDeniedAnyAllowedCtx = "/test1/ctx2";
        _profileAssignmentStorage1.addDeniedProfilesForAnonymous(anonymousDeniedAnyAllowedCtx, profileSet1);
        _profileAssignmentStorage1.addAllowedProfilesForAnyConnectedUser(anonymousDeniedAnyAllowedCtx, profileSet2);
        
        String anonymousAllowedUserDeniedCtx = "/test1/ctx3";
        Set<UserIdentity> user1 = Set.of(_USER1);
        _profileAssignmentStorage1.addAllowedProfilesForAnonymous(anonymousAllowedUserDeniedCtx, profileSet1);
        _profileAssignmentStorage1.addDeniedUsers(user1, anonymousAllowedUserDeniedCtx, profile1);
        _profileAssignmentStorage1.addDeniedUsers(user1, anonymousAllowedUserDeniedCtx, profile3);
        
        String groupDeniedUserAllowedCtx = "/test1/ctx4";
        Set<GroupIdentity> group1 = Set.of(_GROUP1);
        _profileAssignmentStorage1.addDeniedGroups(group1, groupDeniedUserAllowedCtx, profile1);
        _profileAssignmentStorage1.addDeniedGroups(group1, groupDeniedUserAllowedCtx, profile2);
        _profileAssignmentStorage1.addAllowedUsers(user1, groupDeniedUserAllowedCtx, profile1);
        _profileAssignmentStorage1.addAllowedUsers(user1, groupDeniedUserAllowedCtx, profile3);
        
        String groupAllowedGroupDeniedCtx = "/test1/ctx5";
        Set<GroupIdentity> group2 = Set.of(_GROUP2);
        _profileAssignmentStorage1.addDeniedGroups(group1, groupAllowedGroupDeniedCtx, profile1);
        _profileAssignmentStorage1.addDeniedGroups(group1, groupAllowedGroupDeniedCtx, profile2);
        _profileAssignmentStorage1.addAllowedGroups(group2, groupAllowedGroupDeniedCtx, profile1);
        _profileAssignmentStorage1.addAllowedGroups(group2, groupAllowedGroupDeniedCtx, profile3);
        
        /**
         * Read from all storage
         */
        String storage2Ctx = "/test2/ctx1";
        _profileAssignmentStorage2.addAllowedProfilesForAnonymous(storage2Ctx, profileSet1);
        
        // The test does not cover the priority management
        // i.e. different assignment stored on the same context
        // by two supporting storage with different priority
        // Only the assignment for the highest priority storage
        // Should be returned
        
        Map<Object, Map<AccessResultInfo, Set<String>>> allPermissions = _profileAssignmentStorageEP.getAllPermissions(_USER1, Set.of(_GROUP1, _GROUP2));
        // Remove the different contexts one by one and assert that no context are left after
        Map<AccessResultInfo, Set<String>> anonymousAllowedPermissions = allPermissions.remove(anonymousAllowedCtx);
        Assert.assertEquals("Expected 1 access result for context " + anonymousAllowedCtx, 1, anonymousAllowedPermissions.size());
        Assert.assertEquals(profileSet1, anonymousAllowedPermissions.get(new AccessResultInfo(AccessResult.ANONYMOUS_ALLOWED, null)));
        
        Map<AccessResultInfo, Set<String>> anonymousDeniedAnyAllowedPermissions = allPermissions.remove(anonymousDeniedAnyAllowedCtx);
        Assert.assertEquals("Expected 2 access results for context " + anonymousDeniedAnyAllowedCtx, 2, anonymousDeniedAnyAllowedPermissions.size());
        Assert.assertEquals(profileSet1, anonymousDeniedAnyAllowedPermissions.get(new AccessResultInfo(AccessResult.ANONYMOUS_DENIED, null)));
        Assert.assertEquals(profileSet2, anonymousDeniedAnyAllowedPermissions.get(new AccessResultInfo(AccessResult.ANY_CONNECTED_ALLOWED, null)));
        
        Map<AccessResultInfo, Set<String>> anonymousAllowedUserDeniedPermissions = allPermissions.remove(anonymousAllowedUserDeniedCtx);
        Assert.assertEquals("Expected 2 access results for context " + anonymousAllowedUserDeniedCtx, 2, anonymousAllowedUserDeniedPermissions.size());
        Assert.assertEquals(profileSet1, anonymousAllowedUserDeniedPermissions.get(new AccessResultInfo(AccessResult.ANONYMOUS_ALLOWED, null)));
        Assert.assertEquals(profileSet2, anonymousAllowedUserDeniedPermissions.get(new AccessResultInfo(AccessResult.USER_DENIED, _USER1)));
        
        Map<AccessResultInfo, Set<String>> groupDeniedUserAllowedPermissions = allPermissions.remove(groupDeniedUserAllowedCtx);
        Assert.assertEquals("Expected 2 access results for context " + groupDeniedUserAllowedCtx, 2, groupDeniedUserAllowedPermissions.size());
        Assert.assertEquals(profileSet1, groupDeniedUserAllowedPermissions.get(new AccessResultInfo(AccessResult.GROUP_DENIED, _GROUP1)));
        Assert.assertEquals(profileSet2, groupDeniedUserAllowedPermissions.get(new AccessResultInfo(AccessResult.USER_ALLOWED, _USER1)));
        
        Map<AccessResultInfo, Set<String>> groupAllowedGroupDeniedPermissions = allPermissions.remove(groupAllowedGroupDeniedCtx);
        Assert.assertEquals("Expected 2 access results for context " + groupAllowedGroupDeniedCtx, 2, groupAllowedGroupDeniedPermissions.size());
        Assert.assertEquals(profileSet1, groupAllowedGroupDeniedPermissions.get(new AccessResultInfo(AccessResult.GROUP_DENIED, _GROUP1)));
        Assert.assertEquals(profileSet2, groupAllowedGroupDeniedPermissions.get(new AccessResultInfo(AccessResult.GROUP_ALLOWED, _GROUP2)));
        
        Map<AccessResultInfo, Set<String>> storage2Permissions = allPermissions.remove(storage2Ctx);
        Assert.assertEquals("Expected 1 access results for context " + storage2Ctx, 1, storage2Permissions.size());
        Assert.assertEquals(profileSet1, storage2Permissions.get(new AccessResultInfo(AccessResult.ANONYMOUS_ALLOWED, null)));
        
        Assert.assertEquals("All permissions should be removed at this point.", 0, allPermissions.size());
    }
}
