/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.scheduler;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;
import org.quartz.JobDataMap;
import org.quartz.JobKey;
import org.quartz.SchedulerException;

import org.ametys.core.schedule.AmetysJob;
import org.ametys.core.schedule.Runnable.FireProcess;
import org.ametys.core.user.UserIdentity;
import org.ametys.core.user.population.UserPopulationDAO;
import org.ametys.plugins.core.impl.schedule.DefaultRunnable;
import org.ametys.plugins.core.schedule.Scheduler;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * Test case for Schedulable.
 */
@NeedFeatures({
    "core/runtime.core.scheduler",
    "test/test.schedulable",
    "core-impl/runtime.user.directory.impl.static", 
    "core-impl/runtime.authentication.impl.defined",
})
public class SchedulableTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static Scheduler _scheduler;
    private static UserPopulationDAO _userPopulationDAO;

    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                                   .withConfigFile("test/environments/configs/config1.xml")
                                   .withWebappPath("test/environments/webapp5");
    }
    
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _scheduler = (Scheduler) manager.lookup(Scheduler.ROLE);
        _userPopulationDAO = (UserPopulationDAO) manager.lookup(UserPopulationDAO.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        Map<String, Object> userDirectory = new LinkedHashMap<>();
        String udModelId = "org.ametys.plugins.core.user.directory.Static";
        userDirectory.put("udModelId", udModelId);
        String usersAsString = "user1:One:User\n"
                             + "user2:Two:User";
        userDirectory.put(udModelId + "$" + "runtime.users.static.users", usersAsString);
        
        Map<String, Object> credentialProvider = new LinkedHashMap<>();
        String cpModelId = "org.ametys.core.authentication.Defined";
        credentialProvider.put("cpModelId", cpModelId);
        credentialProvider.put(cpModelId + "$" + "runtime.authentication.defined.user", "anonymous");
        
        _userPopulationDAO.add("static", "Static Population", Collections.singletonList(userDirectory), Collections.singletonList(credentialProvider));
    }
    
    /**
     * Tests the case where two schedulables are laucnhed concurrently with different users.
     * @throws SchedulerException if the scheduler fail.
     * @throws InterruptedException if something wrong occurs waiting for schedulables to complete.
     */
    @Test
    public void testConcurrentSchedulables() throws SchedulerException, InterruptedException
    {
        org.ametys.core.schedule.Runnable job = new DefaultRunnable(
                "test-schedulable",
                new I18nizableText("label"),
                new I18nizableText("desc"),
                FireProcess.NOW,
                null,
                "org.ametys.runtime.test.scheduler.TestSchedulable",
                true,
                false,
                false,
                null,
                true,
                new UserIdentity("user1", "static"),
                Collections.EMPTY_MAP);
        
        _scheduler.scheduleJob(job);
        
        JobDataMap dataMap = _scheduler.getScheduler().getJobDetail(new JobKey(job.getId(), Scheduler.JOB_GROUP)).getJobDataMap();
        while (dataMap.get(AmetysJob.KEY_SUCCESS) == null)
        {
            Thread.sleep(100);
            dataMap = _scheduler.getScheduler().getJobDetail(new JobKey(job.getId(), Scheduler.JOB_GROUP)).getJobDataMap();
        }
        
        boolean success = dataMap.getBoolean(AmetysJob.KEY_SUCCESS);
        Assert.assertTrue("The test schedulable failed to execute", success);
    }
}
