/*
 *  Copyright 2023 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.scheduler.progression;

import java.util.List;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import org.ametys.core.schedule.progression.ContainerProgressionTracker;
import org.ametys.core.schedule.progression.ProgressionTrackerFactory;
import org.ametys.core.schedule.progression.SimpleProgressionTracker;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;

/**
 * {@link ContainerProgressionTracker} implementation test case
 */
public class ContainerProgressionTrackerTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static final Logger _LOGGER = LoggerFactory.getLogger(ContainerProgressionTrackerTestCase.class);
    private static final I18nizableText SAME_LABEL = new I18nizableText("LABEL");

    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        // Do Nothing
    }
    
    /**
     * Test the {@link ContainerProgressionTracker} with two sub steps with the same coefficient
     */
    @Test
    public void testTwoStepsOfSameCoef()
    {
        // Create a progression tracker with one step
        ContainerProgressionTracker progressionTrackerWithSubSteps = ProgressionTrackerFactory.createContainerProgressionTracker(SAME_LABEL, _LOGGER);
        
        // Add two sub steps
        progressionTrackerWithSubSteps.addSimpleStep("test1", SAME_LABEL, 1);
        progressionTrackerWithSubSteps.addSimpleStep("test2", SAME_LABEL, 1);

        SimpleProgressionTracker subStep1 = progressionTrackerWithSubSteps.getStep("test1");
        
        // Check the parameters with no progression made
        Assert.assertEquals(0d, subStep1.asPercentage(), 0d);
        Assert.assertEquals(1, progressionTrackerWithSubSteps.getIndex());
        Assert.assertEquals(0d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        // Increment the current progression (= increment the progression of the first sub step)
        subStep1.increment();
        
        // Check the parameters with a progression of 1 graduation on the first sub step
        // The first sub step's progression is at index/graduations * 100 = 1/1 * 100 = 100%
        Assert.assertEquals(100d, subStep1.asPercentage(), 0d);
        Assert.assertEquals(2, progressionTrackerWithSubSteps.getIndex());
        
        SimpleProgressionTracker subStep2 = progressionTrackerWithSubSteps.getStep("test2");

        // Check the parameters with a progression of 1 graduation on the first sub step and 0 on the second substep
        // The second sub step's progression is at index/graduations * 100 = 0/1 * 100 = 0%
        Assert.assertEquals(0d, subStep2.asPercentage(), 0d);
        // The progression is at SUM(subStepsProgressions * coef) = 100 * 0.5 + 0 * 0.5 = 50%
        Assert.assertEquals(50d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        // Increment the current progression (= increment the progression of the second sub step)
        subStep2.increment();
        
        // Check the parameters with a progression of 1 graduation on the first sub step and on the second substep
        // The second sub step's progression is at index/graduations * 100 = 1/1 * 100 = 0%
        Assert.assertEquals(100d, subStep2.asPercentage(), 0d);
        // The progression is at SUM(subStepsProgressions * coef) = 100 * 0.5 + 100 * 0.5 = 100%
        Assert.assertEquals(100d, progressionTrackerWithSubSteps.asPercentage(), 0d);
    }
    
    /**
     * Test the {@link ContainerProgressionTracker} with three sub steps with different coefficients
     */
    @Test
    public void testWithThreeStepsWithDifferentCoefs()
    {
        // Create a progression tracker with one step
        ContainerProgressionTracker progressionTrackerWithSubSteps = ProgressionTrackerFactory.createContainerProgressionTracker(SAME_LABEL, _LOGGER);
        
        // Add the substeps to the progression tracker with weight 1, 8 and 1 so coefs : 0.1, 0.8 and 0.1 (10%, 80% and 10%)
        progressionTrackerWithSubSteps.addSimpleStep("test1", SAME_LABEL, 1);
        progressionTrackerWithSubSteps.addSimpleStep("test2", SAME_LABEL, 8);
        progressionTrackerWithSubSteps.addSimpleStep("test3", SAME_LABEL, 1);

        // Check the parameters with no progression made
        Assert.assertEquals(1, progressionTrackerWithSubSteps.getIndex());
        Assert.assertEquals(0d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        SimpleProgressionTracker subStep1 = progressionTrackerWithSubSteps.getStep("test1");
        Assert.assertEquals(0d, subStep1.asPercentage(), 0d);
        // Increment the current progression (= increment the progression of the first sub step)
        subStep1.increment(1);
      
        // Check the parameters with a progression of 1 graduation on the first sub step
        // The first sub step's progression is at index/graduations * 100 = 1/1 * 100 = 100%
        Assert.assertEquals(100d, subStep1.asPercentage(), 0d);
        Assert.assertEquals(2, progressionTrackerWithSubSteps.getIndex());
        // The progression is at SUM(subStepsProgressions * coef) = 100 * 0.1 = 10%
        Assert.assertEquals(10d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        // Check the parameters with a progression of 1 graduation on the first sub step and 0 on the second substep
        // The second sub step's progression is at index/graduations * 100 = 0/1 * 100 = 0%
        Assert.assertEquals(2, progressionTrackerWithSubSteps.getIndex());
        // The progression is at SUM(subStepsProgressions * coef) = 100 * 0.1 + 0 * 0.8 = 10%
        Assert.assertEquals(10d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        SimpleProgressionTracker subStep2 = progressionTrackerWithSubSteps.getStep("test2");
        Assert.assertEquals(0d, subStep2.asPercentage(), 0d);
        
        // Increment the current progression (= increment the progression of the second sub step)
        subStep2.increment(1);
        
        // Check the parameters with a progression of 1 graduation on the first sub step and on the second substep
        // The second sub step's progression is at index/graduations * 100 = 1/1 * 100 = 100%
        Assert.assertEquals(100d, subStep2.asPercentage(), 0d);
        Assert.assertEquals(3, progressionTrackerWithSubSteps.getIndex());
        // The progression is at SUM(subStepsProgressions * coef) = 100 * 0.1 + 100 * 0.8 = 90%
        Assert.assertEquals(90d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        // Check the parameters with a progression of 1 graduation on the first sub step and on the second substep and 0 on the third
        // The second sub step's progression is at index/graduations * 100 = 0/1 * 100 = 0%
        Assert.assertEquals(3, progressionTrackerWithSubSteps.getIndex());
        // The progression is at SUM(subStepsProgressions * coef) = 100 * 0.1 + 100 * 0.8 + 0 * 0.1 = 90%
        Assert.assertEquals(90d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        SimpleProgressionTracker subStep3 = progressionTrackerWithSubSteps.getStep("test3");
        Assert.assertEquals(0d, subStep3.asPercentage(), 0d);
        // Increment the current progression (= increment the progression of the third sub step)
        subStep3.increment(1);
        
        // Check the parameters with a progression of 1 graduation on the first sub step, on the second substep and on the third
        // The second sub step's progression is at index/graduations * 100 = 1/1 * 100 = 100%
        Assert.assertEquals(100d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        // The progression is at SUM(subStepsProgressions * coef) = 100 * 0.1 + 100 * 0.8 + 100 * 0.1 = 100%
        Assert.assertEquals(100d, subStep2.asPercentage(), 0d);
    }
    
    /**
     * Test the {@link ContainerProgressionTracker} with three levels of sub steps with different coefficients
     */
    @Test
    public void testWithThreeLevelOfStepsWithDifferentCoefs()
    {
        // Create a progression tracker with one step
        ContainerProgressionTracker progressionTrackerWithSubSteps = ProgressionTrackerFactory.createContainerProgressionTracker(SAME_LABEL, _LOGGER);

        // Add the substeps to the progression with weight 6 and 4 (coefs = 60% and 40%)
        progressionTrackerWithSubSteps.addContainerStep("test1", SAME_LABEL, 6);
        progressionTrackerWithSubSteps.addSimpleStep("test2", SAME_LABEL, 4);

        ContainerProgressionTracker subStep1 = progressionTrackerWithSubSteps.getStep("test1");

        // Add the substeps to the substep1 with weight 8 and 2 (coefs = 80% and 20%)
        subStep1.addSimpleStep("test1.1", SAME_LABEL, 8);
        subStep1.addSimpleStep("test1.2", SAME_LABEL, 2);
        
        SimpleProgressionTracker subSubStep1 = subStep1.getStep("test1.1");
        subSubStep1.setSize(10);

        // Increment the progression sub sub step 1
        subSubStep1.increment(5);
        
        // Check the parameters with a progression of 5 graduations on the first sub sub step
        // The first sub sub step's progression is at index/graduations * 100 = 5/10 * 100 = 50%
        Assert.assertEquals(50d, subSubStep1.asPercentage(), 0d);
        // The first sub step's progression is at SUM(subStepsProgressions) = 50% * 0.8 = 40%
        Assert.assertEquals(40d, subStep1.asPercentage(), 0d);
        // The progression is at SUM(subStepsProgressions) = 40% * 0.6 = 24%
        Assert.assertEquals(24d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        // Increment the progression sub sub step 1
        subSubStep1.increment(5);
        
        // Check the parameters with a progression of 10 graduations on the first sub sub step
        // The first sub sub step's progression is at 100% because it is completed
        Assert.assertEquals(100d, subSubStep1.asPercentage(), 0d);
        // The first sub step's progression is at SUM(subStepsProgressions) = 100% * 0.8 = 80%
        Assert.assertEquals(80d, subStep1.asPercentage(), 0d);
        // The progression is at SUM(subStepsProgressions) = 80% * 0.6 = 48%
        Assert.assertEquals(48d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        SimpleProgressionTracker subSubStep2 = subStep1.getStep("test1.2");
        subSubStep2.setSize(8);

        // Increment the progression sub sub step 2
        subSubStep2.increment(4);
        
        // Check the parameters with a progression of 4 graduations on the second sub sub step
        // The second sub sub step's progression is at index/graduations * 100 = 4/8 * 100 = 50%
        Assert.assertEquals(50d, subSubStep2.asPercentage(), 0d);
        // The second sub step's progression is at SUM(subStepsProgressions) = 100% * 0.8 + 50 * 0.2 = 90%
        Assert.assertEquals(90d, subStep1.asPercentage(), 0d);
        // The progression is at SUM(subStepsProgressions) = 90% * 0.6 = 54%
        Assert.assertEquals(54d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        // Increment the progression sub sub step 2
        subSubStep2.increment(4);
        
        // Check the parameters with a progression of 4 graduations on the second sub sub step
        // The second sub sub step's progression is at index/graduations * 100 = 8/8 * 100 = 100%
        Assert.assertEquals(100d, subSubStep2.asPercentage(), 0d);
        // The second sub step's progression is at SUM(subStepsProgressions) = 100% * 0.8 + 100 * 0.2 = 100%
        Assert.assertEquals(100d, subStep1.asPercentage(), 0d);
        // The progression is at SUM(subStepsProgressions) = 100% * 0.6 = 60%
        Assert.assertEquals(60d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        SimpleProgressionTracker subStep2 = progressionTrackerWithSubSteps.getStep("test2");
        subStep2.setSize(2);

        // Increment the progression sub step 2
        subStep2.increment(1);
        
        // Check the parameters with a progression of 1 graduations on the second sub step
        // The second sub sub step's progression is at index/graduations * 100 = 1/2 * 100 = 50%
        Assert.assertEquals(50d, subStep2.asPercentage(), 0d);
        // The progression is at SUM(subStepsProgressions) = 100% * 0.6 + 50 * 0.4 = 80%
        Assert.assertEquals(80d, progressionTrackerWithSubSteps.asPercentage(), 0d);
        
        // Increment the progression sub step 2
        subStep2.increment(1);
        
        // Check the parameters with a progression of 2 graduations on the second sub step
        // The second sub sub step's progression is at 100% because it is completed
        Assert.assertEquals(100d, subStep2.asPercentage(), 0d);
        // The progression is at SUM(subStepsProgressions) = 100% * 0.6 + 100 * 0.4 = 100%
        Assert.assertEquals(100d, progressionTrackerWithSubSteps.asPercentage(), 0d);
    }
    
    /**
     * Test the {@link ContainerProgressionTracker} with three levels of sub steps with different coefficients
     */
    @Test
    public void testMethodsForScripts()
    {
        ContainerProgressionTracker progressionTracker = ProgressionTrackerFactory.createContainerProgressionTracker(SAME_LABEL, _LOGGER);
        
        progressionTracker.setSteps(List.of(
                                Map.of("id", "step1",
                                       "label", "Etape 1",
                                       "weight", 6d,
                                       "children", List.of(
                                               Map.of("id", "step11", 
                                                       "label", "Etape 1 - 1",
                                                       "weight", 8d), 
                                               Map.of("id", "step12",
                                                       "label", "Etape 1 - 2",
                                                       "weight", 2d,
                                                       "children", List.of(
                                                               Map.of("id", "step121", 
                                                                       "label", "Etape 1 - 2 - 1"),
                                                               Map.of("id", "step122", 
                                                                       "label", "Etape 1 - 2 - 2")))
                                               )
                                       ),
                                Map.of("id", "step2", 
                                        "label", "Etape 2",
                                        "weight", 4d)));
        
        Assert.assertEquals(progressionTracker.getCurrentStep(), progressionTracker.getStep("step1"));
        Assert.assertEquals(0d, progressionTracker.getCurrentStep().asPercentage(), 0d);
        Assert.assertEquals(0d, progressionTracker.asPercentage(), 0d);

        progressionTracker.setSize("step11", 10);
        progressionTracker.increment("step11", 5);
        
        Assert.assertEquals(50d, progressionTracker.findStep("step11").asPercentage(), 0d);
        Assert.assertEquals(40d, progressionTracker.getStep("step1").asPercentage(), 0d);
        Assert.assertEquals(24d, progressionTracker.asPercentage(), 0d);
        
        progressionTracker.increment("step11", 5);

        Assert.assertEquals(100d, progressionTracker.findStep("step11").asPercentage(), 0d);
        Assert.assertEquals(80d, progressionTracker.getStep("step1").asPercentage(), 0d);
        Assert.assertEquals(48d, progressionTracker.asPercentage(), 0d);
        
        progressionTracker.setSize("step121", 8);
        progressionTracker.increment("step121", 4);

        Assert.assertEquals(50d, progressionTracker.findStep("step121").asPercentage(), 0d);
        Assert.assertEquals(25d, progressionTracker.findStep("step12").asPercentage(), 0d);
        Assert.assertEquals(85d, progressionTracker.getStep("step1").asPercentage(), 0d);
        Assert.assertEquals(51d, progressionTracker.asPercentage(), 0d);
        
        progressionTracker.increment("step121", 4);

        Assert.assertEquals(100d, progressionTracker.findStep("step121").asPercentage(), 0d);
        Assert.assertEquals(50d, progressionTracker.findStep("step12").asPercentage(), 0d);
        Assert.assertEquals(90d, progressionTracker.getStep("step1").asPercentage(), 0d);
        Assert.assertEquals(54d, progressionTracker.asPercentage(), 0d);
        
        progressionTracker.setSize("step122", 0);
        Assert.assertEquals(100d, progressionTracker.findStep("step122").asPercentage(), 0d);
        Assert.assertEquals(100d, progressionTracker.findStep("step12").asPercentage(), 0d);
        Assert.assertEquals(100d, progressionTracker.getStep("step1").asPercentage(), 0d);
        Assert.assertEquals(60d, progressionTracker.asPercentage(), 0d);
        
        progressionTracker.increment("step2", 1);
        
        Assert.assertEquals(100d, progressionTracker.asPercentage(), 0d);
    }
}
