/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.search;

import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.plugins.core.search.UserAndGroupSearchManager;
import org.ametys.runtime.test.framework.AbstractGroupDirectoriesAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Ordering;

/**
 * Test the {$link {@link UserAndGroupSearchManager} component
 */
@NeedFeatures({
    "core/runtime.core.search.userAndGroup"
})
public class UserAndGroupSearchTestCase extends AbstractGroupDirectoriesAwareTestCase
{
    private static UserAndGroupSearchManager _userAndGroupSearchCommon;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp2");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        _userAndGroupSearchCommon = (UserAndGroupSearchManager) manager.lookup(UserAndGroupSearchManager.ROLE);
    }
    
    /**
     * Test the search on a single user population
     */
    @Test
    public void testSearchOnSingleSource()
    {
        _testSearch(ImmutableSet.of(POPULATION_TEST_ID), ImmutableSet.of(), 3, null, true);
    }
    
    /**
     * Test the search on a mix of user populations and group directories
     */
    @Test
    public void testSearchOnMixedSources()
    {
        _testSearch(ImmutableSet.of(POPULATION_TEST_ID), ImmutableSet.of(DIRECTORY_TEST_ID), 3, null, true);
    }
    
    /**
     * Test the search on a mix of user populations and group directories, when a criteria is provided
     */
    @Test
    public void testSearchWithCriteria()
    {
        _testSearch(ImmutableSet.of(POPULATION_TEST_ID), ImmutableSet.of(DIRECTORY_TEST_ID), 3, "user", true);
    }
    
    @SuppressWarnings("unchecked")
    private void _testSearch(Set<String> populations, Set<String> directories, int limit, String pattern, boolean sorted)
    {
        Map<String, Object> params = pattern != null ? ImmutableMap.of("sort", sorted, "pattern", pattern) : ImmutableMap.of("sort", sorted);
        Map<String, Object> searchData = null;
        String lastResult = null;
        
        do
        {
            Map<String, Object> searchResults = _userAndGroupSearchCommon.searchUsersAndGroup(populations, directories, limit, searchData, params, true);
            
            Assert.assertNotNull(searchResults);
            Assert.assertTrue("Invalid results from a search: " + searchResults,
                    validateResults(searchResults));
            
            boolean isFinished = searchResults.containsKey("finished");
            List<Map<String, Object>> results = (List<Map<String, Object>>) searchResults.get("results");
            
            Assert.assertTrue("Results are not ordered: " + results.stream().map(this::_getSortableName).collect(Collectors.toList()),
                    Ordering.from(Comparator.comparing(this::_getSortableName, String.CASE_INSENSITIVE_ORDER)).isOrdered(results));
            Assert.assertTrue(isFinished || results.size() > 0);
            
            searchData = (Map<String, Object>) searchResults.getOrDefault("searchData", null);
            
            if (lastResult != null && results.size() > 0)
            {
                String firstResultSortableName = _getSortableName(results.get(0));
                Assert.assertTrue("New result '" + firstResultSortableName + "' should be sorted before the previous last result '" + lastResult + "'", 
                        lastResult.toLowerCase().compareTo(firstResultSortableName.toLowerCase()) <= 0);
            }
            lastResult = results.size() > 0 ? _getSortableName(results.get(results.size() - 1)) : null;
            
            if (pattern != null && StringUtils.isNotEmpty(pattern))
            {
                Assert.assertTrue("All results: " + results.stream().map(this::_getSortableName).collect(Collectors.toList()) + " did not match the pattern: " + pattern, 
                        results.stream().allMatch(result -> StringUtils.contains(_getSortableName(result).toLowerCase(), pattern.toLowerCase())));
            }
        } while (searchData != null);
    }
    
    private boolean validateResults(Map<String, Object> results)
    {
        if (results == null 
            || !results.containsKey("results") 
            || !(results.get("results") instanceof List))
        {
            return false;
        }
        
        return results.containsKey("finished") 
                ? (Boolean) results.getOrDefault("finished", false)
                : results.containsKey("searchData") && results.get("searchData") instanceof Map;
    }
    
    private String _getSortableName(Map<String, Object> json)
    {
        return json.containsKey("sortablename") ? (String) json.get("sortablename") : (String) json.getOrDefault("label", "");
    }
}
