/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.ui;

import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.ui.ClientSideElement;
import org.ametys.core.ui.ClientSideElement.Script;
import org.ametys.core.ui.RibbonControlsManager;
import org.ametys.core.ui.StaticClientSideElement;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.TestConfigSet;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;

/**
 * Test the static ui item factory
 */
@NeedFeatures({
    "staticuiitemfactorytest/staticuiitemfactorytest",
})
public class StaticUIItemFactoryTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    /** The runtime ui item maanger */
    protected static RibbonControlsManager _ribbonManager;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp2");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _ribbonManager = (RibbonControlsManager) manager.lookup("ribbonManagerTest");
    }
    
    /**
     * Check that unexisting factory goes well
     * @throws Exception if an error occurs
     */
    @Test
    public void testUnexisting() throws Exception
    {
        ClientSideElement itemFactory0 = _ribbonManager.getExtension("unexisting");
        Assert.assertNull(itemFactory0);
    }
    
    /**
     * Test the first factory
     * @throws Exception if an error occurs
     */
    @Test
    public void testFactory1() throws Exception
    {
        /**
         * 1st FACTORY 
         */
        ClientSideElement itemFactory1 = _ribbonManager.getExtension("staticuiitemfactorytest.1");
        Script script = itemFactory1.getScripts(null).get(0);
        Map<String, Object> parameters = script.getParameters();
        
        Assert.assertNotNull(itemFactory1);
        Assert.assertTrue(itemFactory1 instanceof StaticClientSideElement);
        // action
        Assert.assertNotNull(script.getScriptClassname());
        Assert.assertNotNull(parameters);
        Assert.assertEquals(5, parameters.size());
        Assert.assertEquals("JavascriptClass", script.getScriptClassname());
        Assert.assertNotNull(script.getScriptFiles());
        Assert.assertEquals(1, script.getScriptFiles().size());
        Assert.assertEquals("/plugins/staticuiitemfactorytest/resources/js/script.js", script.getScriptFiles().iterator().next().getPath());
        // label
        Assert.assertNotNull(parameters.get("label"));
        Assert.assertTrue(parameters.get("label") instanceof I18nizableText);
        Assert.assertTrue(((I18nizableText) parameters.get("label")).isI18n());
        Assert.assertEquals("plugin.staticuiitemfactorytest", ((I18nizableText) parameters.get("label")).getCatalogue());
        Assert.assertEquals("label", ((I18nizableText) parameters.get("label")).getKey());
        Assert.assertNull(((I18nizableText) parameters.get("label")).getParameters());
        // description
        Assert.assertNotNull(parameters.get("default-description"));
        Assert.assertTrue(parameters.get("default-description") instanceof I18nizableText);
        Assert.assertTrue(((I18nizableText) parameters.get("default-description")).isI18n());
        Assert.assertEquals("plugin.staticuiitemfactorytest", ((I18nizableText) parameters.get("default-description")).getCatalogue());
        Assert.assertEquals("description", ((I18nizableText) parameters.get("default-description")).getKey());
        Assert.assertNull(((I18nizableText) parameters.get("default-description")).getParameters());
        // iconset
        Assert.assertNotNull(parameters.get("icon-small"));
        Assert.assertEquals("/plugins/staticuiitemfactorytest/resources/img/icon_small.gif", parameters.get("icon-small").toString());
        Assert.assertEquals("/plugins/staticuiitemfactorytest/resources/img/icon_medium.gif", parameters.get("icon-medium").toString());
        Assert.assertEquals("/plugins/staticuiitemfactorytest/resources/img/icon_large.gif", parameters.get("icon-large").toString());
    }
    
    /**
     * Test the third factory
     * @throws Exception if an error occurs
     */
    @Test
    public void testFactory3() throws Exception
    {
        /**
         * 3rd FACTORY 
         */
        ClientSideElement itemFactory3 = _ribbonManager.getExtension("staticuiitemfactorytest.3");
        Script script = itemFactory3.getScripts(null).get(0);
        Map<String, Object> parameters = script.getParameters();
        
        Assert.assertNotNull(itemFactory3);
        Assert.assertTrue(itemFactory3 instanceof StaticClientSideElement);
        // action
        Assert.assertNotNull(script.getScriptClassname());
        Assert.assertNotNull(parameters);
        Assert.assertEquals(7, parameters.size());
        Assert.assertEquals("myurl.html", parameters.get("Link").toString());
        Assert.assertNotNull(script.getScriptFiles());
        Assert.assertEquals(0, script.getScriptFiles().size());
        // label
        Assert.assertNotNull(parameters.get("label"));
        Assert.assertTrue(parameters.get("label") instanceof I18nizableText);
        Assert.assertTrue(((I18nizableText) parameters.get("label")).isI18n());
        Assert.assertEquals("plugin.staticuiitemfactorytest", ((I18nizableText) parameters.get("label")).getCatalogue());
        Assert.assertEquals("label", ((I18nizableText) parameters.get("label")).getKey());
        Assert.assertNull(((I18nizableText) parameters.get("label")).getParameters());
        // description
        Assert.assertNotNull(parameters.get("default-description"));
        Assert.assertTrue(parameters.get("default-description") instanceof I18nizableText);
        Assert.assertTrue(((I18nizableText) parameters.get("default-description")).isI18n());
        Assert.assertEquals("plugin.staticuiitemfactorytest", ((I18nizableText) parameters.get("default-description")).getCatalogue());
        Assert.assertEquals("description", ((I18nizableText) parameters.get("default-description")).getKey());
        Assert.assertNull(((I18nizableText) parameters.get("default-description")).getParameters());
        // iconset
        Assert.assertEquals("/plugins/staticuiitemfactorytest/resources/img/icon_small.gif", parameters.get("icon-small").toString());
        Assert.assertEquals("/plugins/staticuiitemfactorytest/resources/img/icon_medium.gif", parameters.get("icon-medium").toString());
        Assert.assertEquals("/plugins/staticuiitemfactorytest/resources/img/icon_large.gif", parameters.get("icon-large").toString());
    }
    
    /**
     * Test the forth factory
     * @throws Exception if an error occurs
     */
    @Test
    public void testFactory4() throws Exception
    {
        /**
         * 4th FACTORY 
         */
        ClientSideElement itemFactory4 = _ribbonManager.getExtension("staticuiitemfactorytest.4");
        Script script = itemFactory4.getScripts(null).get(0);
        Map<String, Object> parameters = script.getParameters();
        
        Assert.assertNotNull(itemFactory4);
        Assert.assertTrue(itemFactory4 instanceof StaticClientSideElement);
        // action
        Assert.assertNotNull(script.getScriptClassname());
        Assert.assertNotNull(parameters);
        Assert.assertEquals(5, parameters.size());
        Assert.assertEquals("OtherJavascriptClass", script.getScriptClassname());
        Assert.assertNotNull(script.getScriptFiles());
        Assert.assertEquals(1, script.getScriptFiles().size());
        Assert.assertEquals("/plugins/core/resources/js/script.js", script.getScriptFiles().iterator().next().getPath());
        Assert.assertEquals("staticuiitemfactorytest", itemFactory4.getPluginName());
        // label
        Assert.assertNotNull(parameters.get("label"));
        Assert.assertTrue(parameters.get("label") instanceof I18nizableText);
        Assert.assertTrue(((I18nizableText) parameters.get("label")).isI18n());
        Assert.assertEquals("othercatalogue", ((I18nizableText) parameters.get("label")).getCatalogue());
        Assert.assertEquals("label", ((I18nizableText) parameters.get("label")).getKey());
        Assert.assertNull(((I18nizableText) parameters.get("label")).getParameters());
        // description
        Assert.assertNotNull(parameters.get("default-description"));
        Assert.assertTrue(((I18nizableText) parameters.get("default-description")).isI18n());
        Assert.assertEquals("plugin.otherplugin", ((I18nizableText) parameters.get("default-description")).getCatalogue());
        Assert.assertEquals("description", ((I18nizableText) parameters.get("default-description")).getKey());
        Assert.assertNull(((I18nizableText) parameters.get("default-description")).getParameters());
        // iconset
        Assert.assertEquals("/plugins/core/resources/img/icon_small.gif", parameters.get("icon-small").toString());
        Assert.assertEquals("/plugins/core/resources/img/icon_medium.gif", parameters.get("icon-medium").toString());
        Assert.assertEquals("/plugins/core/resources/img/icon_large.gif", parameters.get("icon-large").toString());
    }  
}
