/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.users.jdbc;

import java.util.HashMap;
import java.util.Map;

import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.user.InvalidModificationException;
import org.ametys.core.user.User;
import org.ametys.core.user.User.UserCreationOrigin;
import org.ametys.core.user.directory.ModifiableUserDirectory;
import org.ametys.core.user.directory.UserDirectory.CredentialsResult;
import org.ametys.plugins.core.impl.user.directory.JdbcUserDirectory;
import org.ametys.runtime.test.AbstractJDBCTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;
import org.ametys.runtime.test.framework.Unoptimized;

/**
 * Tests JDBC user directory
 */
@Unoptimized
@NeedFeatures({
    "core-impl/runtime.user.directory.impl.jdbc"})
public abstract class AbstractJdbcUsersTestCase extends AbstractJDBCTestCase
{
    @Override
    public TestConfigSet getConfigSet()
    {
        return super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        _userDirectory = _createUserDirectory();
    }
    
    /**
     * Test the getting of users on mysql
     * @throws Exception if an error occurs
     */
    @Test
    public void testType() throws Exception
    {
        // JDBC IMPL
        Assert.assertTrue(_userDirectory instanceof JdbcUserDirectory);

        // MODIFIABLE
        Assert.assertTrue(_userDirectory instanceof ModifiableUserDirectory);
    }
    
    /**
     * Test the addition of a new user that should failed
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncorrectAdd() throws Exception
    {
        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) _userDirectory;
        
        // Incorrect additions
        Map<String, String> userInformation;
        
        try
        {
            userInformation = new HashMap<>();
            modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);
            Assert.fail("An empty addition should fail");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior
        }

        try
        {
            userInformation = new HashMap<>();
            userInformation.put("login", "test");
            userInformation.put("firstname", "test");
            modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);
            Assert.fail("A non complete addition should fail");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior
        }

        try
        {
            userInformation = new HashMap<>();
            userInformation.put("login", "test");
            userInformation.put("firstname", "test");
            userInformation.put("lastname", "test");
            userInformation.put("email", "");
            modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);
            Assert.fail("A non complete addition should fail");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior
        }

        try
        {
            userInformation = new HashMap<>();
            userInformation.put("login", "test");
            userInformation.put("firstname", "test");
            userInformation.put("lastname", "test");
            userInformation.put("password", "test@Password123");
            userInformation.put("email", "testthatisnotacorrectemail");
            modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);
            Assert.fail("An incorrect addition should fail");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior
        }
        
        try
        {
            userInformation = new HashMap<>();
            userInformation.put("login", "test");
            userInformation.put("firstname", "test");
            userInformation.put("lastname", "test");
            userInformation.put("password", "test");
            userInformation.put("email", "test@email.com");
            modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);
            Assert.fail("Addition with weak password should fail");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior
        }
    }
    
    /**
     * Test the addition of a new user
     * @throws Exception if an error occurs
     */
    @Test
    public void testCorrectAdd() throws Exception
    {
        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) _userDirectory;
        User user;
        
        // Correct additions
        Map<String, String> userInformation;
        
        userInformation = new HashMap<>();
        userInformation.put("login", "test");
        userInformation.put("firstname", "Test");
        userInformation.put("lastname", "TEST");
        userInformation.put("email", "");
        userInformation.put("password", "test@Password123");
        modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);
        Assert.assertEquals(CredentialsResult.AUTHENTICATED, _userDirectory.checkCredentials("test", "test@Password123"));
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, _userDirectory.checkCredentials("test", "wrongpassword"));
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, _userDirectory.checkCredentials("test2", "test@Password123"));
        
        userInformation = new HashMap<>();
        userInformation.put("login", "test2");
        userInformation.put("firstname", "Test2");
        userInformation.put("lastname", "TEST2");
        userInformation.put("email", "test2@test.te");
        userInformation.put("password", "test@Password123");
        modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);
        Assert.assertEquals(CredentialsResult.AUTHENTICATED, _userDirectory.checkCredentials("test", "test@Password123"));
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, _userDirectory.checkCredentials("test", "wrongpassword"));
        Assert.assertEquals(CredentialsResult.AUTHENTICATED, _userDirectory.checkCredentials("test2", "test@Password123"));
        
        user = _userDirectory.getUser("test");
        Assert.assertNotNull(user);
        Assert.assertEquals(user.getIdentity().getLogin(), "test");
        Assert.assertEquals(user.getLastName(), "TEST"); 
        Assert.assertEquals(user.getFirstName(), "Test"); 
        Assert.assertEquals(user.getFullName(), "Test TEST");
        Assert.assertEquals(user.getSortableName(), "TEST Test");
        Assert.assertEquals(user.getEmail(), "");

        user = _userDirectory.getUser("test2");
        Assert.assertNotNull(user);
        Assert.assertEquals(user.getIdentity().getLogin(), "test2");
        Assert.assertEquals(user.getLastName(), "TEST2"); 
        Assert.assertEquals(user.getFirstName(), "Test2"); 
        Assert.assertEquals(user.getFullName(), "Test2 TEST2");
        Assert.assertEquals(user.getSortableName(), "TEST2 Test2");
        Assert.assertEquals(user.getEmail(), "test2@test.te");

        try
        {
            userInformation = new HashMap<>();
            userInformation.put("login", "test");
            userInformation.put("firstname", "Test");
            userInformation.put("lastname", "TEST");
            userInformation.put("email", "");
            userInformation.put("password", "test@Password123");
            modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);
            Assert.fail("Add should have failed");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior since login already exists
        }
    }
    
    /**
     * Test the update of a user incorrectly
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncorrectUpdate() throws Exception
    {
        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) _userDirectory;
        
        // Incorrect modification
        Map<String, String> userInformation;
        
        try
        {
            userInformation = new HashMap<>();
            userInformation.put("login", "test");
            userInformation.put("firstname", "Test");
            userInformation.put("lastname", "TEST");
            userInformation.put("email", "");
            userInformation.put("password", "test@Password123");
            modifiableUserDirectory.update(userInformation);
            Assert.fail("Update should have failed");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior since login does not exist
        }
        
        userInformation = new HashMap<>();
        userInformation.put("login", "test");
        userInformation.put("firstname", "Test");
        userInformation.put("lastname", "TEST");
        userInformation.put("email", "");
        userInformation.put("password", "test@Password123");
        modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);
        
        try
        {
            userInformation = new HashMap<>();
            userInformation.put("firstname", "Test");
            userInformation.put("lastname", "TEST");
            userInformation.put("email", "");
            userInformation.put("password", "test@Password123");
            modifiableUserDirectory.update(userInformation);
            Assert.fail("Update should have failed");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior since no login is given
        }

        try
        {
            userInformation = new HashMap<>();
            userInformation.put("login", "test");
            userInformation.put("firstname", "Test");
            userInformation.put("lastname", "TEST");
            userInformation.put("email", "incorrectemail");
            userInformation.put("password", "test@Password123");
            modifiableUserDirectory.update(userInformation);
            Assert.fail("Update should have failed");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior since email is incorrect
        }
        
        try
        {
            userInformation = new HashMap<>();
            userInformation.put("login", "test");
            userInformation.put("firstname", "Test");
            userInformation.put("lastname", "TEST");
            userInformation.put("password", "");
            modifiableUserDirectory.update(userInformation);
            Assert.fail("Update should have failed");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior since password is empty
        }
        
        try
        {
            userInformation = new HashMap<>();
            userInformation.put("login", "test");
            userInformation.put("firstname", "Test");
            userInformation.put("lastname", "TEST");
            userInformation.put("password", "testpassword");
            modifiableUserDirectory.update(userInformation);
            Assert.fail("Update should have failed");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior since password is weak
        }
        
        Assert.assertEquals(CredentialsResult.AUTHENTICATED, _userDirectory.checkCredentials("test", "test@Password123"));
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, _userDirectory.checkCredentials("test", "wrongpassword"));
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, _userDirectory.checkCredentials("test2", "test@Password123"));
    }
    
    /**
     * Test the update of a user
     * @throws Exception if an error occurs
     */
    @Test
    public void testCorrectUpdate() throws Exception
    {
        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) _userDirectory;
        User user;
        
        Map<String, String> userInformation;
        userInformation = new HashMap<>();
        userInformation.put("login", "test");
        userInformation.put("firstname", "Test");
        userInformation.put("lastname", "TEST");
        userInformation.put("email", "");
        userInformation.put("password", "test@Password123");
        modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);

        // Correct modification
        userInformation = new HashMap<>();
        userInformation.put("login", "test");
        userInformation.put("firstname", "Testmodified");
        userInformation.put("lastname", "TESTMODIFIED");
        userInformation.put("email", "testModified@test.te");
        userInformation.put("password", "testPassword456!");
        modifiableUserDirectory.update(userInformation);
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, _userDirectory.checkCredentials("test", "test@Password123"));
        Assert.assertEquals(CredentialsResult.AUTHENTICATED, _userDirectory.checkCredentials("test", "testPassword456!"));
        
        user = _userDirectory.getUser("test");
        Assert.assertNotNull(user);
        Assert.assertEquals(user.getIdentity().getLogin(), "test");
        Assert.assertEquals(user.getLastName(), "TESTMODIFIED"); 
        Assert.assertEquals(user.getFirstName(), "Testmodified"); 
        Assert.assertEquals(user.getFullName(), "Testmodified TESTMODIFIED");
        Assert.assertEquals(user.getSortableName(), "TESTMODIFIED Testmodified");
        Assert.assertEquals(user.getEmail(), "testModified@test.te");

        // partial modification (no password change)
        userInformation = new HashMap<>();
        userInformation.put("login", "test");
        userInformation.put("firstname", "Testmodifiedtwice");
        modifiableUserDirectory.update(userInformation);
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, _userDirectory.checkCredentials("test", "test@Password123"));
        Assert.assertEquals(CredentialsResult.AUTHENTICATED, _userDirectory.checkCredentials("test", "testPassword456!"));
        
        user = _userDirectory.getUser("test");
        Assert.assertNotNull(user);
        Assert.assertEquals(user.getIdentity().getLogin(), "test");
        Assert.assertEquals(user.getLastName(), "TESTMODIFIED"); 
        Assert.assertEquals(user.getFirstName(), "Testmodifiedtwice"); 
        Assert.assertEquals(user.getFullName(), "Testmodifiedtwice TESTMODIFIED");
        Assert.assertEquals(user.getSortableName(), "TESTMODIFIED Testmodifiedtwice");
        Assert.assertEquals(user.getEmail(), "testModified@test.te");
    }
    
    /**
     * Test incorrects remove
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncorrectRemove() throws Exception
    {
        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) _userDirectory;
        
        try
        {
            modifiableUserDirectory.remove("foo");
            Assert.fail("Remove should fail");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior
        }

        Map<String, String> userInformation;
        userInformation = new HashMap<>();
        userInformation.put("login", "test");
        userInformation.put("firstname", "Test");
        userInformation.put("lastname", "TEST");
        userInformation.put("password", "test@Password123");
        userInformation.put("email", "");
        modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);

        try
        {
            modifiableUserDirectory.remove("foo");
            Assert.fail("Remove should fail");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior
            User user = _userDirectory.getUser("test");
            Assert.assertNotNull(user);
        }

        userInformation = new HashMap<>();
        userInformation.put("login", "test");
        userInformation.put("firstname", "Testmodified");
        userInformation.put("lastname", "TESTMODIFIED");
        userInformation.put("email", "testModified@test.te");
        userInformation.put("password", "testPassword456!");
        modifiableUserDirectory.update(userInformation);

        try
        {
            modifiableUserDirectory.remove("foo");
            Assert.fail("Remove should fail");
        }
        catch (InvalidModificationException e)
        {
            // normal behavior
            User user = _userDirectory.getUser("test");
            Assert.assertNotNull(user);
        }

        Assert.assertEquals(CredentialsResult.AUTHENTICATED, _userDirectory.checkCredentials("test", "testPassword456!"));
    }
    
    /**
     * Test corrects remove
     * @throws Exception if an error occurs
     */
    @Test
    public void testCorrectRemove() throws Exception
    {
        ModifiableUserDirectory modifiableUserDirectory = (ModifiableUserDirectory) _userDirectory;
        
        Map<String, String> userInformation;
        userInformation = new HashMap<>();
        userInformation.put("login", "test");
        userInformation.put("firstname", "Test");
        userInformation.put("lastname", "TEST");
        userInformation.put("email", "");
        userInformation.put("password", "test@Password123");
        modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);

        userInformation = new HashMap<>();
        userInformation.put("login", "test2");
        userInformation.put("firstname", "Test2");
        userInformation.put("lastname", "TEST2");
        userInformation.put("email", "email@email.ma");
        userInformation.put("password", "testPassword456!");
        modifiableUserDirectory.add(userInformation, UserCreationOrigin.ADMIN);

        modifiableUserDirectory.remove("test");
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, _userDirectory.checkCredentials("test", "test@Password123"));
        Assert.assertEquals(CredentialsResult.AUTHENTICATED, _userDirectory.checkCredentials("test2", "testPassword456!"));
        
        User user = _userDirectory.getUser("test2");
        Assert.assertNotNull(user);
        Assert.assertEquals(user.getIdentity().getLogin(), "test2");
        Assert.assertEquals(user.getLastName(), "TEST2"); 
        Assert.assertEquals(user.getFirstName(), "Test2"); 
        Assert.assertEquals(user.getFullName(), "Test2 TEST2");
        Assert.assertEquals(user.getSortableName(), "TEST2 Test2");
        Assert.assertEquals(user.getEmail(), "email@email.ma");
    }
}
