/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.users.ldap;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.user.User;
import org.ametys.core.user.directory.ModifiableUserDirectory;
import org.ametys.core.user.directory.UserDirectory;
import org.ametys.core.user.directory.UserDirectory.CredentialsResult;
import org.ametys.core.user.directory.UserDirectoryFactory;
import org.ametys.plugins.core.impl.user.directory.LdapUserDirectory;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * Tests the LdapUserDirectory
 */
@NeedFeatures({
    "core-impl/runtime.user.directory.impl.ldap", 
    "core-impl/runtime.user.directory.impl.ldap",
})
public class LdapUsersTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    /** The User Directory Factory */
    protected static UserDirectoryFactory _userDirectoryFactory;
    /** the user directory */
    protected static UserDirectory _userDirectory;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config4.xml")
                .withLdapFile("test/environments/datasources/datasource-ldap.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _userDirectoryFactory = (UserDirectoryFactory) manager.lookup(UserDirectoryFactory.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        _userDirectory = _createLdapUserDirectory();
    }
    
    /**
     * Test the getting of users on mysql
     * @throws Exception if an error occurs
     */
    @Test
    public void testType() throws Exception
    {
        // LDAP IMPL
        Assert.assertTrue(_userDirectory instanceof LdapUserDirectory);

        // NOT MODIFIABLE
        Assert.assertFalse(_userDirectory instanceof ModifiableUserDirectory);
    }
    
    /**
     * Test a filled db
     * @throws Exception if an error occurs
     */
    @Test
    public void testFilled() throws Exception
    {
        User user = null;
        Collection<User> users = null;

        // Get unexisting user
        user = _userDirectory.getUser("foo");
        Assert.assertNull(user);

        // Get existing user
        User user1 = _userDirectory.getUser("user1");
        Assert.assertNotNull(user1);
        Assert.assertEquals(user1.getIdentity().getLogin(), "user1");
        Assert.assertEquals(user1.getLastName(), "USER1");
        Assert.assertEquals(user1.getFirstName(), "User1");
        Assert.assertEquals(user1.getFullName(), "User1 USER1");
        Assert.assertEquals(user1.getSortableName(), "USER1 User1");
        Assert.assertEquals(user1.getEmail(), "user1@ametys.org");
        
        User user10 = _userDirectory.getUser("user10");
        Assert.assertNotNull(user10);
        Assert.assertEquals(user10.getIdentity().getLogin(), "user10");
        Assert.assertEquals(user10.getLastName(), "USER10");
        Assert.assertEquals(user10.getFirstName(), "User10");
        Assert.assertEquals(user10.getFullName(), "User10 USER10");
        Assert.assertEquals(user10.getSortableName(), "USER10 User10");
        Assert.assertEquals(user10.getEmail(), "user10@ametys.org");

        // Get users
        users = _userDirectory.getUsers();
        Assert.assertEquals(10, users.size());
        Assert.assertTrue(users.contains(user1));
        Assert.assertTrue(users.contains(user10));
        
        // Get users with parameters
        users = _userDirectory.getUsers(-1, 0, new HashMap<>());
        Assert.assertEquals(10, users.size());
        Assert.assertTrue(users.contains(user1));
        Assert.assertTrue(users.contains(user10));
        
        // Get users with pattern "user1"
        Map<String, Object> parameters = new HashMap<>();
        parameters.put("pattern", "user1");
        users = _userDirectory.getUsers(-1, 0, parameters);
        Assert.assertEquals(2, users.size());
        Assert.assertTrue(users.contains(user1));
        Assert.assertTrue(users.contains(user10));
        
        // Get users with parameters by part
        Set<String> results = new HashSet<>();
        
        users = _userDirectory.getUsers(4, 0, new HashMap<>());
        Assert.assertEquals(4, users.size());
        for (User theUser : users)
        {
            results.add(theUser.getIdentity().getLogin());
        }
        
        users = _userDirectory.getUsers(4, 4, new HashMap<>());
        Assert.assertEquals(4, users.size());
        for (User theUser : users)
        {
            results.add(theUser.getIdentity().getLogin());
        }
        
        users = _userDirectory.getUsers(4, 8, new HashMap<>());
        Assert.assertEquals(2, users.size());
        for (User theUser : users)
        {
            results.add(theUser.getIdentity().getLogin());
        }
        
        Assert.assertEquals(10, results.size());
        Assert.assertTrue(results.contains("user1"));
        Assert.assertTrue(results.contains("user2"));
        Assert.assertTrue(results.contains("user3"));
        Assert.assertTrue(results.contains("user4"));
        Assert.assertTrue(results.contains("user5"));
        Assert.assertTrue(results.contains("user6"));
        Assert.assertTrue(results.contains("user7"));
        Assert.assertTrue(results.contains("user8"));
        Assert.assertTrue(results.contains("user9"));
        Assert.assertTrue(results.contains("user10"));
    }
    
    /**
     * Test that LDAP can authentify incorrectly
     * @throws Exception if an error occurs
     */
    @Test
    public void testIncorrectAuthentication() throws Exception
    {
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, _userDirectory.checkCredentials("foo", "foo"));
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, _userDirectory.checkCredentials("user1", "wrongpassword"));
    }

    /**
     * Test that LDAP can authentify correctly
     * @throws Exception if an error occurs
     */
    @Test
    public void testCorrectAuthentication() throws Exception
    {
        Assert.assertEquals(CredentialsResult.AUTHENTICATED, _userDirectory.checkCredentials("user1", "user1"));
    }
    
    /**
     * Create the user directory
     * @return The user directory
     */
    protected UserDirectory _createLdapUserDirectory()
    {
        String modelId = "org.ametys.plugins.core.user.directory.Ldap";
        
        Map<String, Object> parameters = new LinkedHashMap<>();
        parameters.put("runtime.users.ldap.datasource", "LDAP-test-users");
        
        parameters.put("runtime.users.ldap.peopleDN", "ou=people");
        parameters.put("runtime.users.ldap.baseFilter", "(objectclass=inetOrgPerson)");
        parameters.put("runtime.users.ldap.scope", "sub");
        parameters.put("runtime.users.ldap.loginAttr", "uid");
        parameters.put("runtime.users.ldap.firstnameAttr", "givenName");
        parameters.put("runtime.users.ldap.lastnameAttr", "sn");
        parameters.put("runtime.users.ldap.emailAttr", "mail");
        parameters.put("runtime.users.ldap.emailMandatory", false);
        parameters.put("runtime.users.ldap.serverSideSorting", true);
        
        return _userDirectoryFactory.createUserDirectory("fooldap", modelId, parameters, "foo", null);
    }
}
