/*
 *  Copyright 2017 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.users.modifiable;

import static org.junit.Assert.assertNotNull;

import java.io.File;
import java.io.FileInputStream;
import java.nio.file.CopyOption;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Optional;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.cocoon.servlet.multipart.Part;
import org.apache.cocoon.servlet.multipart.PartOnDisk;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.IOUtils;
import org.junit.After;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.cache.AbstractCacheManager;
import org.ametys.core.right.Profile;
import org.ametys.core.right.ProfileAssignmentStorageExtensionPoint;
import org.ametys.core.right.RightProfilesDAO;
import org.ametys.core.user.InvalidModificationException;
import org.ametys.core.user.User;
import org.ametys.core.user.directory.UserDirectory.CredentialsResult;
import org.ametys.core.user.population.UserPopulation;
import org.ametys.plugins.core.impl.hash.Argon2PasswordEncoder;
import org.ametys.plugins.core.impl.user.directory.JdbcUserDirectory;
import org.ametys.plugins.core.user.ImportUsers;
import org.ametys.plugins.core.user.UserHelper;
import org.ametys.runtime.test.framework.AbstractPopulationAwareTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * Test ImportUsers
 */
@NeedFeatures({
    "core/runtime.users.import", 
    "core/runtime.users.manager", 
    "core-impl/runtime.user.directory.impl.jdbc", 
    "core-impl/runtime.authentication.impl.defined",
    
    "core/runtime.rights.profiles.dao",
    "core-impl/runtime.core.impl.rights.accesscontroller",
    "core-impl/runtime.core.impl.rights.convertor",
    "admin/right.assignment.context.admin"
})
public class ImportUsersTestCase extends AbstractPopulationAwareTestCase
{
    /** import user component */
    protected static ImportUsers _importUser;

    /** The cache manager */
    protected static  AbstractCacheManager _abstractCacheManager;

    /** The argon password encoder */
    protected static  Argon2PasswordEncoder _argon2PasswordEncoder;

    /** The profile assignment storage */
    protected static ProfileAssignmentStorageExtensionPoint _profileAssignmentStorageEP;
    /** The profile dao */
    protected static RightProfilesDAO _profilesDAO;

    private static final String __POPULATION_ID = "test-import-pop";
    private static final String __JDBC_USER_DIRECTORY_ID = "test-import-jdbc-user-directory";
    private static final String __USER_CACHE_ID = UserHelper.class.getName() + "$userCache";

    
    @Override
    public TestConfigSet getConfigSet()
    {
        return super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp2");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        super.lookupComponents(manager);
        
        _importUser = (ImportUsers) manager.lookup(ImportUsers.ROLE);
        _abstractCacheManager = (AbstractCacheManager) manager.lookup(AbstractCacheManager.ROLE);
        _profilesDAO = (RightProfilesDAO) manager.lookup(RightProfilesDAO.ROLE);
        _profileAssignmentStorageEP = (ProfileAssignmentStorageExtensionPoint) manager.lookup(ProfileAssignmentStorageExtensionPoint.ROLE);
    }
    
    @Override
    public void prepareComponents() throws Exception
    {
        super.prepareComponents();
        _createUserPopulation();

        File pepperFile = new File("test/resources/users-import/pepper.txt");
        try (FileInputStream inputStream = new FileInputStream(pepperFile)) 
        {
            _argon2PasswordEncoder = new Argon2PasswordEncoder(IOUtils.toByteArray(inputStream));
        }
        
        Profile profile = _profilesDAO.addProfile("myProfile");
        _profilesDAO.addRight(profile, "Runtime_Rights_User_Handle");
        _profileAssignmentStorageEP.allowProfileToAnyConnectedUser(profile.getId(), "/admin");
    }
    
    /**
     * Delete all remaining users after each test
     * @throws Exception if an error occurs.
     */
    @After
    public void deleteUsersAfterTest() throws Exception
    {
        _deleteUsers();
    }
    
    /**
     * Test various incorrect files
     */
    @Test
    public void testImportBadFiles()
    {
        // Invalid file extension
        Assert.assertEquals("non empty user directory", 0, _retrieveUserDirectory().getUsers().size());
        Map<String, Object> importFile = _importFile("test/resources/users-import/import-error-2.xls", false, true);
        Assert.assertNotNull(importFile);
        Assert.assertTrue("There should be an error in this import", importFile.containsKey("error"));
        Assert.assertEquals("the error should be \"invalid-extension\"", "invalid-extension", importFile.get("error"));

        // Non existing file
        importFile = _importFile("fileNotFound.csv", false, true);
        Assert.assertNotNull(importFile);
        Assert.assertTrue("There should be an error in this import", importFile.containsKey("error"));
        Assert.assertEquals("the error should be java.nio.file.NoSuchFileException", "java.nio.file.NoSuchFileException", importFile.get("error"));

        // Missing column
        importFile = _importFile("test/resources/users-import/import-error-1.csv", false, true);
        Assert.assertNotNull(importFile);
        Assert.assertTrue("There should be an error in this import", importFile.containsKey("error"));
        Assert.assertEquals("the error should be \"column-names\"", "column-missing", importFile.get("error"));

        // Missing column
        importFile = _importFile("test/resources/users-import/import-error-3.txt", false, true);
        Assert.assertNotNull(importFile);
        Assert.assertTrue("There should be an error in this import", importFile.containsKey("error"));
        Assert.assertEquals("the error should be \"column-missing\"", "column-missing", importFile.get("error"));

        // Duplicated column
        importFile = _importFile("test/resources/users-import/import-error-4.txt", false, true);
        Assert.assertNotNull(importFile);
        Assert.assertTrue("There should be an error in this import", importFile.containsKey("error"));
        Assert.assertEquals("the error should be \"column-duplicate\"", "column-duplicate", importFile.get("error"));
    }

    /**
     * Test files with password not hashed
     * @throws InvalidModificationException exception in deleteUsers
     */
    @Test
    public void testImportClearText() throws InvalidModificationException
    {
        _checkImportResult(4, 0, 0, 0, "test/resources/users-import/import-clear-1.csv", false, false);
        _checkUserCount(4);
        Assert.assertEquals("Error in password", CredentialsResult.AUTHENTICATED, _retrieveUserDirectory().checkCredentials("login1", "password1"));

        _checkImportResult(0, 4, 0, 0, "test/resources/users-import/import-clear-2.txt", false, false);
        _checkUserCount(4);
        Assert.assertEquals("Error in password", CredentialsResult.AUTHENTICATED, _retrieveUserDirectory().checkCredentials("login1", "password1"));

        _checkImportResult(1, 1, 0, 1, "test/resources/users-import/import-clear-3.csv", false, false);
        _checkUserCount(5);
        Assert.assertEquals("Error in password", CredentialsResult.AUTHENTICATED, _retrieveUserDirectory().checkCredentials("login5", "password5"));

        _checkImportResult(0, 2, 3, 1, "test/resources/users-import/import-clear-3.csv", true, false);
        _checkUserCount(2);
        Assert.assertEquals("Error in password", CredentialsResult.AUTHENTICATED, _retrieveUserDirectory().checkCredentials("login5", "password5"));
    }
    
    private void _checkImportResult(Integer nbAdded, Integer nbEdited, Integer nbDeleted, Integer nbError, String filePath, boolean cleanDirectory, boolean hashedPasswords)
    {
        _invalidateCache();
        Map<String, Object> importResult = _importFile(filePath, cleanDirectory, hashedPasswords);
        Assert.assertNotNull("No object returned by import", importResult);
        Boolean success = importResult.containsKey("success") && (Boolean) importResult.get("success");
        Integer addedCount = importResult.containsKey("addedCount") ? (Integer) importResult.get("addedCount") : -1;
        Integer existingCount = importResult.containsKey("existingCount") ? (Integer) importResult.get("existingCount") : -1;
        Integer deletedCount = importResult.containsKey("deletedCount") ? (Integer) importResult.get("deletedCount") : -1;
        Integer errorCount = importResult.containsKey("errorCount") ? (Integer) importResult.get("errorCount") : -1;
        Assert.assertTrue("Error in import", success);
        Assert.assertEquals("Nb added problem", nbAdded, addedCount);
        Assert.assertEquals("Nb deleted problem", nbDeleted, deletedCount);
        Assert.assertEquals("Nb existing problem", nbEdited, existingCount);
        Assert.assertEquals("Nb error problem", nbError, errorCount);
    }

    private void _checkUserCount(Integer expected)
    {
        Integer nbUsers = _retrieveUserDirectory().getUsers().size();
        Assert.assertEquals("Number of users in repository error", expected, nbUsers);
    }

    private Map<String, Object> _importFile(String filePath, boolean cleanDirectory, boolean returnException)
    {
        Path from = Paths.get(filePath);
        String extension = FilenameUtils.getExtension(filePath);
        Path destination = Paths.get("test/resources/users-import/temporaty-file." + extension);
        CopyOption[] options = new CopyOption[]{
            StandardCopyOption.REPLACE_EXISTING,
            StandardCopyOption.COPY_ATTRIBUTES
        };
        try
        {
            Files.copy(from, destination, options);
            
            File file = destination.toFile();
            Part part = new PartOnDisk(null, file);
            return _importUser.importUsers(part, __POPULATION_ID, __JDBC_USER_DIRECTORY_ID, cleanDirectory);
        }
        catch (Exception e)
        {
            if (returnException)
            {
                return Collections.singletonMap("error", e.getClass().getName());
            }
            else
            {
                e.printStackTrace();
                Assert.fail(e.getMessage());
            }
        }
        return Collections.singletonMap("error", "empty error"); //should never happen due to try/catch
    }

    private void _deleteUsers() throws InvalidModificationException
    {
        JdbcUserDirectory userDirectory = _retrieveUserDirectory();
        for (User user : _retrieveUserDirectory().getUsers())
        {
            userDirectory.remove(user.getIdentity().getLogin());
        }
    }

    private UserPopulation _createUserPopulation()
    {
        Map<String, Object> udParams = new LinkedHashMap<>();
        String udModelId = "org.ametys.plugins.core.user.directory.Jdbc";
        udParams.put("udModelId", udModelId);
        udParams.put("id", __JDBC_USER_DIRECTORY_ID);
        udParams.put(udModelId + "$" + "runtime.users.jdbc.datasource", "SQL-test");
        udParams.put(udModelId + "$" + "runtime.users.jdbc.table", "Users");
        udParams.put(udModelId + "$runtime.users.jdbc.strong.password", false);
        udParams.put(udModelId + "$runtime.users.jdbc.strong.password.min-length", 8);

        Map<String, Object> cpParams = new LinkedHashMap<>();
        String cpModelId = "org.ametys.core.authentication.Defined";
        cpParams.put("cpModelId", cpModelId);
        cpParams.put(cpModelId + "$" + "runtime.authentication.defined.user", "anonymous");

        _userPopulationDAO.remove(__POPULATION_ID);
        Map<String, Object> populationMap = _userPopulationDAO.add(__POPULATION_ID, __POPULATION_ID, Collections.singletonList(udParams), Collections.singletonList(cpParams));
        assertNotNull(populationMap);
        return _userPopulationDAO.getUserPopulation(__POPULATION_ID);
    }
    
    private JdbcUserDirectory _retrieveUserDirectory()
    {
        return Optional.of(_userPopulationDAO)
                .map(populationDao -> populationDao.getUserPopulation(__POPULATION_ID))
                .map(population -> population.getUserDirectory(__JDBC_USER_DIRECTORY_ID))
                .map(JdbcUserDirectory.class::cast)
                .orElseThrow();
    }

    private void _invalidateCache()
    {
        _abstractCacheManager.get(__USER_CACHE_ID).invalidateAll();
    }
}
