/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.users.others;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.user.User;
import org.ametys.core.user.directory.ModifiableUserDirectory;
import org.ametys.core.user.directory.UserDirectory;
import org.ametys.core.user.directory.UserDirectory.CredentialsResult;
import org.ametys.core.user.directory.UserDirectoryFactory;
import org.ametys.plugins.core.impl.user.directory.StaticUserDirectory;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

/**
 * Tests the DefinedUsersTestCase
 */
@NeedFeatures("core-impl/runtime.user.directory.impl.static")
public class StaticUsersTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static UserDirectoryFactory _userDirectoryFactory;

    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp1");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _userDirectoryFactory = (UserDirectoryFactory) manager.lookup(UserDirectoryFactory.ROLE);
    }
    
    /**
     * Tests that the <code>StaticUsersTestCase</code> is the default <code>UserDirectory</code><br/>
     * Tests the values returned by it
     * @throws Exception if an error occurs
     */
    @Test
    public void testStaticUsers() throws Exception
    {
        UserDirectory userDirectory = _createStaticUserDirectory();
        
        // DEFAULT IMPL
        Assert.assertTrue(userDirectory instanceof StaticUserDirectory);
        
        // NOT MODIFIABLE
        Assert.assertFalse(userDirectory instanceof ModifiableUserDirectory);
        
        // ONE USER
        User user;
        
        user = userDirectory.getUser("foo");
        Assert.assertNull(user);
        
        user = userDirectory.getUser("anonymous");
        Assert.assertNotNull(user);
        Assert.assertEquals(user.getIdentity().getLogin(), "anonymous");
        Assert.assertEquals(user.getLastName(), "Anonymous");
        Assert.assertEquals(user.getFirstName(), "user");
        Assert.assertEquals(user.getFullName(), "user Anonymous");
        Assert.assertEquals(user.getSortableName(), "Anonymous user");
        Assert.assertEquals(user.getEmail(), "user@ametys.org");
        
        // ALL USERS
        Collection<User> users = userDirectory.getUsers();
        Assert.assertEquals(users.size(), 1);
        Assert.assertEquals(users.iterator().next(), user);
        
        // LOGIN
        Assert.assertEquals(CredentialsResult.NOT_AUTHENTICATED, userDirectory.checkCredentials("foo", null));

        Assert.assertEquals(CredentialsResult.AUTHENTICATED, userDirectory.checkCredentials("anonymous", null));
    }
    
    private UserDirectory _createStaticUserDirectory()
    {
        String modelId = "org.ametys.plugins.core.user.directory.Static";
        Map<String, Object> parameters = new LinkedHashMap<>();
        parameters.put("runtime.users.static.users", "anonymous:Anonymous:user:user@ametys.org");
        return _userDirectoryFactory.createUserDirectory("foostatic", modelId, parameters, "foo", null);
    }
}
