/*
 *  Copyright 2019 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.util;

import static java.lang.annotation.ElementType.TYPE_USE;
import static java.lang.annotation.RetentionPolicy.RUNTIME;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.lang.annotation.Retention;
import java.lang.annotation.Target;
import java.util.Collections;
import java.util.Set;
import java.util.function.BiFunction;
import java.util.stream.Collectors;

import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.junit.BeforeClass;
import org.junit.Test;

import org.ametys.runtime.test.framework.AbstractBaseTestCase;
import org.ametys.runtime.util.Annotations;

import junit.framework.TestCase;

/**
 * {@link TestCase} of {@link Annotations} helper class.
 */
public class AnnotationsTestCase extends AbstractBaseTestCase
{
    /*@Inherited*//* meta-annotation @Inherited has no influence in this test as we ignore the class it comes from */
    @Retention(RUNTIME) @Target(TYPE_USE) @interface AnnotationToTest
    {
        String value();
    }
    
    private static final BiFunction<AnnotationToTest, Class<?>, String> __ANNOTATION_ATTRIBUTE_GETTER = 
        (annotation, __) -> annotation.value(); // ignore annotated class
    
    /*
     * IA2 -> IA1
     * IB3 -> IB2 -> @ IB1
     * 
     * 
     * C -> B -> A
     *     ^ ^
     *     |  \
     *    IA2 IB3
     * 
     * G2 -> F -> @ E -> D
     * ^                ^
     * |                |
     * @ IG            IA2
     * 
     * G1 -> F
     */
    
    interface IA1
    { /* nothing */ }
    interface IA2 extends IA1
    { /* nothing */ }
    @AnnotationToTest("-ib1-") interface IB1
    { /* nothing */ }
    interface IB2 extends IB1
    { /* nothing */ }
    interface IB3 extends IB2
    { /* nothing */ }
    @AnnotationToTest("-ig2-") interface IG2
    { /* nothing */ }
    
    class A
    { /* nothing */ }
    class B extends A implements IA2, IB3
    { /* nothing */ }
    class C extends B
    { /* nothing */ }
    class D implements IA2
    { /* nothing */ }
    @AnnotationToTest("-e-") class E extends D
    { /* nothing */ }
    class F extends E
    { /* nothing */ }
    class G1 extends F
    { /* nothing */ }
    class G2 extends F implements IG2
    { /* nothing */ }
    
    /**
     * Sets debug level on catgory of {@link Annotations}
     */
    @BeforeClass
    public static void enableDebugLevel()
    {
        LogManager.getLogger(Annotations.class).setLevel(Level.DEBUG);
    }
    
    /**
     * Tests {@link Annotations#isAnnotationPresent} returns <code>false</code> when it should.
     */
    @Test
    public void testIsNotPresent()
    {
        A a = new A();
        assertFalse(Annotations.isAnnotationPresent(a.getClass(), AnnotationToTest.class));
        
        D d = new D();
        assertFalse(Annotations.isAnnotationPresent(d.getClass(), AnnotationToTest.class));
    }
    
    /**
     * Tests {@link Annotations#isAnnotationPresent} returns <code>true</code> when the annotation is present for a super interface of one of the implemented interfaces.
     * <br>Here because B implements IB3 which extends IB1 which has the annotation
     */
    @Test
    public void testIsPresentFromSuperInterfaceOfImplemented()
    {
        A b = new B();
        assertTrue(Annotations.isAnnotationPresent(b.getClass(), AnnotationToTest.class));
    }
    
    /**
     * Tests {@link Annotations#isAnnotationPresent} returns <code>true</code> when the annotation is present for a super interface of one of the implemented interfaces of the superclass.
     * <br>Here because C extends B which implements IB3 which extends IB1 which has the annotation
     */
    @Test
    public void testIsPresentFromSuperInterfaceOfImplementedOfSuperclass()
    {
        A c = new C();
        assertTrue(Annotations.isAnnotationPresent(c.getClass(), AnnotationToTest.class));
    }
    
    /**
     * Tests {@link Annotations#isAnnotationPresent} returns <code>true</code> when the annotation is present on the class itself.
     * <br>Here because E has the annotation
     */
    @Test
    public void testIsPresentFromItself()
    {
        D e = new E();
        assertTrue(Annotations.isAnnotationPresent(e.getClass(), AnnotationToTest.class));
    }
    
    /**
     * Tests {@link Annotations#isAnnotationPresent} returns <code>true</code> when the annotation is present on the superclass.
     * <br>Here because F extends E which has the annotation
     */
    @Test
    public void testIsPresentFromSuperclass()
    {
        D f = new F();
        assertTrue(Annotations.isAnnotationPresent(f.getClass(), AnnotationToTest.class));
    }
    
    /**
     * Tests {@link Annotations#isAnnotationPresent} returns <code>true</code> when the annotation is present on one of the ancestor classes.
     * <br>Here because G1 extends which F extends E which has the annotation
     */
    @Test
    public void testIsPresentFromAncestor()
    {
        D g1 = new G1();
        assertTrue(Annotations.isAnnotationPresent(g1.getClass(), AnnotationToTest.class));
    }
    
    /**
     * Tests {@link Annotations#isAnnotationPresent} returns <code>true</code> when the annotation is present on one of the ancestor classes AND from one of the implemented interfaces.
     * <br>Here because G2 extends which F extends E which has the annotation AND G2 implements IG2 which has the annotation
     */
    @Test
    public void testIsPresentFromAncestorAndImplemented()
    {
        D g2 = new G2();
        assertTrue(Annotations.isAnnotationPresent(g2.getClass(), AnnotationToTest.class));
    }
    
    
    
    /**
     * Tests {@link Annotations#getAnnotationAttributeValues} when the annotation is not present.
     */
    @Test
    public void testGetEmpty()
    {
        A a = new A();
        assertEquals(
                Collections.EMPTY_SET, 
                Annotations.getAnnotationAttributeValues(a.getClass(), AnnotationToTest.class, __ANNOTATION_ATTRIBUTE_GETTER)
                        .collect(Collectors.toSet())
        );
        
        D d = new D();
        assertEquals(
                Collections.EMPTY_SET, 
                Annotations.getAnnotationAttributeValues(d.getClass(), AnnotationToTest.class, __ANNOTATION_ATTRIBUTE_GETTER)
                        .collect(Collectors.toSet())
        );
    }
    
    /**
     * Tests {@link Annotations#getAnnotationAttributeValues} when the annotation is present for a super interface of one of the implemented interfaces.
     * <br>Here because B implements IB3 which extends IB1 which has the annotation
     */
    @Test
    public void testGetFromSuperInterfaceOfImplemented()
    {
        A b = new B();
        assertEquals(
                Set.of("-ib1-"), 
                Annotations.getAnnotationAttributeValues(b.getClass(), AnnotationToTest.class, __ANNOTATION_ATTRIBUTE_GETTER)
                        .collect(Collectors.toSet())
        );
    }
    
    /**
     * Tests {@link Annotations#getAnnotationAttributeValues} when the annotation is present for a super interface of one of the implemented interfaces of the superclass.
     * <br>Here because C extends B which implements IB3 which extends IB1 which has the annotation
     */
    @Test
    public void testGetFromSuperInterfaceOfImplementedOfSuperclass()
    {
        A c = new C();
        assertEquals(
                Set.of("-ib1-"), 
                Annotations.getAnnotationAttributeValues(c.getClass(), AnnotationToTest.class, __ANNOTATION_ATTRIBUTE_GETTER)
                        .collect(Collectors.toSet())
        );
    }
    
    /**
     * Tests {@link Annotations#getAnnotationAttributeValues} when the annotation is present on the class itself.
     * <br>Here because E has the annotation
     */
    @Test
    public void testGetFromItself()
    {
        D e = new E();
        assertEquals(
                Set.of("-e-"), 
                Annotations.getAnnotationAttributeValues(e.getClass(), AnnotationToTest.class, __ANNOTATION_ATTRIBUTE_GETTER)
                        .collect(Collectors.toSet())
        );
    }
    
    /**
     * Tests {@link Annotations#getAnnotationAttributeValues} when the annotation is present on the superclass.
     * <br>Here because F extends E which has the annotation
     */
    @Test
    public void testGetFromSuperclass()
    {
        D f = new F();
        assertEquals(
                Set.of("-e-"), 
                Annotations.getAnnotationAttributeValues(f.getClass(), AnnotationToTest.class, __ANNOTATION_ATTRIBUTE_GETTER)
                        .collect(Collectors.toSet())
        );
    }
    
    /**
     * Tests {@link Annotations#getAnnotationAttributeValues} when the annotation is present on one of the ancestor classes.
     * <br>Here because G1 extends which F extends E which has the annotation
     */
    @Test
    public void testGetFromAncestor()
    {
        D g1 = new G1();
        assertEquals(
                Set.of("-e-"), 
                Annotations.getAnnotationAttributeValues(g1.getClass(), AnnotationToTest.class, __ANNOTATION_ATTRIBUTE_GETTER)
                        .collect(Collectors.toSet())
        );
    }
    
    /**
     * Tests {@link Annotations#getAnnotationAttributeValues} when the annotation is present on one of the ancestor classes AND from one of the implemented interfaces.
     * <br>Here because G2 extends which F extends E which has the annotation AND G2 implements IG2 which has the annotation
     */
    @Test
    public void testGetFromAncestorAndImplemented()
    {
        D g2 = new G2();
        assertEquals(
                Set.of("-ig2-", "-e-"), 
                Annotations.getAnnotationAttributeValues(g2.getClass(), AnnotationToTest.class, __ANNOTATION_ATTRIBUTE_GETTER)
                .collect(Collectors.toSet())
        );
    }
}

