/*
 *  Copyright 2009 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.util;

import java.time.LocalDate;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.time.temporal.ChronoUnit;
import java.util.Date;

import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.util.DateUtils;
import org.ametys.runtime.test.framework.AbstractBaseTestCase;

import junit.framework.TestCase;

/**
 * {@link TestCase} of {@link DateUtils} for formatting and parsing {@link Date}.
 */
public class DateConversionTestCase extends AbstractBaseTestCase
{
    /**
     * Test date conversion
     */
    @Test   
    public void testDateConversion()
    {
        ZonedDateTime zdt = ZonedDateTime.of(2012, 3, 2, 18, 30, 22, 0, ZoneId.systemDefault());
        Date expected = Date.from(zdt.toInstant());
        
        String dateAsStr = zdt.withZoneSameInstant(ZoneOffset.UTC).format(DateUtils.getISODateTimeFormatter());
        Date actual =  DateUtils.parse(dateAsStr);
        
        Assert.assertEquals(expected, actual);
        
        Assert.assertEquals(DateUtils.dateToString(expected), dateAsStr);
        
        ZonedDateTime atParis = zdt.withZoneSameInstant(ZoneId.of("Europe/Paris"));
        ZonedDateTime atTokyo = zdt.withZoneSameInstant(ZoneId.of("Asia/Tokyo"));
        
        Date dateAtParis = Date.from(atParis.toInstant());
        Date dateAtTokyo = Date.from(atTokyo.toInstant());
        
        Assert.assertEquals(DateUtils.dateToString(dateAtParis), DateUtils.dateToString(dateAtTokyo));
    }
    
    /**
     * Test date to string methods as localDateToString and zonedDateTimeToString
     */
    @Test
    public void testDateToStringMethod()
    {
        LocalDate localDate = LocalDate.of(2012, 3, 2);
        Assert.assertEquals("2012-03-02", DateUtils.localDateToString(localDate));
        
        ZonedDateTime zonedDateTime = ZonedDateTime.of(2012, 3, 2, 18, 30, 22, 153_111_111, ZoneId.of("Europe/Paris"));
        
        Assert.assertEquals("2012-03-02T18:30:22.153+01:00", DateUtils.zonedDateTimeToString(zonedDateTime));
        Assert.assertEquals("2012-03-02T17:30:22.153Z", DateUtils.zonedDateTimeToString(zonedDateTime, ZoneId.of("Z"), DateUtils.ISO_OFFSET_DATE_TIME_PATTERN));
        Assert.assertEquals("2012-03-02T12:30:22.153-05:00", DateUtils.zonedDateTimeToString(zonedDateTime, ZoneId.of("-05:00"), DateUtils.ISO_OFFSET_DATE_TIME_PATTERN));
        
        ZonedDateTime zonedDateTimeWithNoMillis = ZonedDateTime.of(2012, 3, 2, 18, 30, 22, 0, ZoneId.of("Europe/Paris"));
        
        Assert.assertEquals("2012-03-02T18:30:22.000+01:00", DateUtils.zonedDateTimeToString(zonedDateTimeWithNoMillis));
        Assert.assertEquals("2012-03-02T17:30:22.000Z", DateUtils.zonedDateTimeToString(zonedDateTimeWithNoMillis, ZoneId.of("Z"), DateUtils.ISO_OFFSET_DATE_TIME_PATTERN));
        Assert.assertEquals("2012-03-02T12:30:22.000-05:00", DateUtils.zonedDateTimeToString(zonedDateTimeWithNoMillis, ZoneId.of("-05:00"), DateUtils.ISO_OFFSET_DATE_TIME_PATTERN));
    }
    
    /**
     * Test date to string methods as localDateToString and zonedDateTimeToString
     */
    @Test
    public void testParseDateMethod()
    {
        LocalDate ld = LocalDate.of(2012, 3, 2);
        
        ZonedDateTime zdt = ZonedDateTime.of(2012, 3, 2, 18, 30, 22, 153_000_000, ZoneId.of("Europe/Paris"));
        
        LocalDate localDate = DateUtils.parseLocalDate("2012-03-02");
        ZonedDateTime zonedDateTime = DateUtils.parseZonedDateTime("2012-03-02T18:30:22.153+01:00");
        
        Assert.assertTrue(ld.isEqual(localDate));
        Assert.assertTrue(zdt.isEqual(zonedDateTime));
    }
    
    /**
     * Test if after parseDate(dateToString) is the identity
     */
    @Test
    public void testReciprocityBetweenDateMethod()
    {
        LocalDate ld = LocalDate.now();
        ZonedDateTime zdt = ZonedDateTime.now(ZoneId.of("Europe/Paris"));
       
        Assert.assertTrue(
                ld
                .isEqual(DateUtils.parseLocalDate(DateUtils.localDateToString(ld))));
        
        Assert.assertTrue(
                zdt.truncatedTo(ChronoUnit.MILLIS) // We ignore after the milisecond
                .isEqual(DateUtils.parseZonedDateTime(DateUtils.zonedDateTimeToString(zdt))));
        
        Assert.assertTrue(
                "2012-03-02"
                .equals(DateUtils.localDateToString(DateUtils.parseLocalDate("2012-03-02"))));
        
        Assert.assertTrue(
                "2012-03-02T18:30:22.153+01:00"
                .equals(DateUtils.zonedDateTimeToString(DateUtils.parseZonedDateTime("2012-03-02T18:30:22.153+01:00"))));
    }
}
