/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.util;

import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.util.FilenameUtils;
import org.ametys.runtime.test.framework.AbstractBaseTestCase;

import junit.framework.TestCase;

/**
 * {@link TestCase} for encoding or decodong file names
 */
public class FilenameUtilsTestCase extends AbstractBaseTestCase
{
    private String[] _prepareDataForTest()
    {
        return new String[] {
            "value 1.pdf", // 0
            "value@!$2.jpg", // 1
            "value%3:;🙂.doc", // 2
            "path/to/value/4.xslx", // 3
            "pàth/to/àé=çù.txt" // 4
        };
    }

    /**
     * Test encode of path
     */
    @Test
    public void testEncodePath()
    {
        String[] data = _prepareDataForTest();
        
        Assert.assertEquals("value 1.pdf", FilenameUtils.encodePath(data[0]));
        Assert.assertEquals("value@!$2.jpg", FilenameUtils.encodePath(data[1]));
        Assert.assertEquals("value%253%3A%3B🙂.doc", FilenameUtils.encodePath(data[2]));
        Assert.assertEquals("path/to/value/4.xslx", FilenameUtils.encodePath(data[3]));
        Assert.assertEquals("pàth/to/àé=çù.txt", FilenameUtils.encodePath(data[4]));
    }
    
    /**
     * Test encode of path segment
     */
    @Test
    public void testEncodeName()
    {
        String[] data = _prepareDataForTest();
        
        Assert.assertEquals("value 1.pdf", FilenameUtils.encodeName(data[0]));
        Assert.assertEquals("value@!$2.jpg", FilenameUtils.encodeName(data[1]));
        Assert.assertEquals("value%253%3A%3B🙂.doc", FilenameUtils.encodeName(data[2]));
        Assert.assertEquals("path%2Fto%2Fvalue%2F4.xslx", FilenameUtils.encodeName(data[3]));
        Assert.assertEquals("pàth%2Fto%2Fàé=çù.txt", FilenameUtils.encodeName(data[4]));
    }
    
    /**
     * Test filter of path
     */
    @Test
    public void testFilterPath()
    {
        String[] data = _prepareDataForTest();
        
        Assert.assertEquals("value 1.pdf", FilenameUtils.filterPath(data[0]));
        Assert.assertEquals("value@!$2.jpg", FilenameUtils.filterPath(data[1]));
        Assert.assertEquals("value%3__🙂.doc", FilenameUtils.filterPath(data[2]));
        Assert.assertEquals("path/to/value/4.xslx", FilenameUtils.filterPath(data[3]));
        Assert.assertEquals("pàth/to/àé=çù.txt", FilenameUtils.filterPath(data[4]));
    }
    
    /**
     * Test filter of path segment
     */
    @Test
    public void testFilterName()
    {
        String[] data = _prepareDataForTest();
        
        Assert.assertEquals("value 1.pdf", FilenameUtils.filterName(data[0]));
        Assert.assertEquals("value@!$2.jpg", FilenameUtils.filterName(data[1]));
        Assert.assertEquals("value%3__🙂.doc", FilenameUtils.filterName(data[2]));
        Assert.assertEquals("path_to_value_4.xslx", FilenameUtils.filterName(data[3]));
        Assert.assertEquals("pàth_to_àé=çù.txt", FilenameUtils.filterName(data[4]));
    }
    
    /**
     * Test decode
     */
    @Test
    public void testDecode()
    {
        String[] data = _prepareDataForTest();
        
        Assert.assertEquals(data[0], FilenameUtils.decode("value 1.pdf"));
        Assert.assertEquals(data[1], FilenameUtils.decode("value@!$2.jpg"));
        Assert.assertEquals(data[2], FilenameUtils.decode("value%253%3A%3B🙂.doc"));
        Assert.assertEquals(data[3], FilenameUtils.decode("path/to/value/4.xslx"));
        Assert.assertEquals(data[4], FilenameUtils.decode("pàth%2Fto%2Fàé=çù.txt"));
    }
}
