/*
 *  Copyright 2013 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.util;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.time.LocalDate;
import java.time.LocalTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.time.ZonedDateTime;
import java.time.format.FormatStyle;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.commons.io.IOUtils;
import org.apache.excalibur.source.Source;
import org.apache.excalibur.source.SourceResolver;
import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.util.I18nUtils;
import org.ametys.runtime.i18n.I18nizableDate;
import org.ametys.runtime.i18n.I18nizableDateTime;
import org.ametys.runtime.i18n.I18nizableText;
import org.ametys.runtime.i18n.I18nizableTextParameter;
import org.ametys.runtime.test.framework.AbstractRuntimeTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase;
import org.ametys.runtime.test.framework.CocoonTestCase.NeedFeatures;
import org.ametys.runtime.test.framework.TestConfigSet;

import junit.framework.TestCase;

/**
 * {@link TestCase} for programmatically translating i18n keys.
 */
@NeedFeatures({
    "core/runtime.core", 
    "core/util.i18nutils", 
    "test/test", 
    "core/runtime.core.resources-handler", 
    "core/util.i18n.bundlefactory",
})
public class I18nTestCase extends AbstractRuntimeTestCase implements CocoonTestCase
{
    private static I18nUtils _i18nUtils;
    private static SourceResolver _srcResolver;
    
    @Override
    public TestConfigSet getConfigSet()
    {
        return CocoonTestCase.super.getConfigSet()
                .withConfigFile("test/environments/configs/config1.xml")
                .withWebappPath("test/environments/webapp2");
    }
    
    @Override
    public void lookupComponents(ServiceManager manager) throws ServiceException
    {
        _i18nUtils = (I18nUtils) manager.lookup(I18nUtils.ROLE);
        _srcResolver = (SourceResolver) manager.lookup(SourceResolver.ROLE);
    }
    
    /**
     * Tests simple i18n keys, without params.
     */
    @Test
    public void testI18n()
    {
        String translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY"), "en");
        Assert.assertEquals("english text", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY"), "fr");
        Assert.assertEquals("texte français", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY"), "fr_FR");
        Assert.assertEquals("texte français", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY"), "zh");
        Assert.assertEquals("english text", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "LABEL"), "en");
        Assert.assertEquals("overriden label", translatedText);
    }
    
    /**
     * Tests i18n keys with params.
     */
    @Test
    public void testI18nWithParams()
    {
        ArrayList<String> params = new ArrayList<>(1);
        params.add("2");
        
        String translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY_PARAMS", params), "en");
        Assert.assertEquals("english text 2", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY_PARAMS", params), "fr");
        Assert.assertEquals("texte français 2", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY_PARAMS", params), "fr_FR");
        Assert.assertEquals("texte français 2", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY_PARAMS", params), "zh");
        Assert.assertEquals("english text 2", translatedText);
    }
    
    /**
     * Tests i18n keys with i18n params.
     */
    @Test
    public void testI18nWithI18nParams()
    {
        Map<String, I18nizableTextParameter> params = new HashMap<>(1);
        params.put("language", new I18nizableText("plugin.test", "TEST"));
        
        String translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY_I18N_PARAMS", params), "en");
        Assert.assertEquals("english text", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY_I18N_PARAMS", params), "fr");
        Assert.assertEquals("texte français", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY_I18N_PARAMS", params), "fr_FR");
        Assert.assertEquals("texte français", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY_I18N_PARAMS", params), "zh");
        Assert.assertEquals("chinese text", translatedText);
    }
    
    /**
     * Tests i18n keys with date params.
     */
    @Test
    public void testI18nWithDateParams()
    {
        Map<String, I18nizableTextParameter> params = new HashMap<>(1);
        params.put("date", new I18nizableDate(LocalDate.of(2021, 01, 15), FormatStyle.SHORT));
        params.put("date-time", new I18nizableDateTime(ZonedDateTime.of(LocalDate.of(2021, 01, 16), LocalTime.of(16, 12), ZoneOffset.UTC), ZoneId.of("+02:00"), FormatStyle.SHORT));
        
        String translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY_DATE_PARAMS", params), "en");
        Assert.assertEquals("from 1/15/21 to 1/16/21, 6:12 PM", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY_DATE_PARAMS", params), "fr");
        Assert.assertEquals("du 15/01/2021 au 16/01/2021 18:12", translatedText);
    }
    
    /**
     * Test to get raw values from I18nUtils.translate
     */
    @Test
    public void testI18nRawValues()
    {
        String translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY"), "en", true);
        Assert.assertEquals("english text", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "LABEL"), "en", true);
        Assert.assertEquals("overriden label", translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY"), "fr_FR", true);
        Assert.assertEquals(null, translatedText);
        
        translatedText = _i18nUtils.translate(new I18nizableText("plugin.test", "TEST_KEY"), "zh", true);
        Assert.assertEquals(null, translatedText);
    }
    
    /**
     * Test the i18n text reader
     */
    @Test
    public void testI18nTextReader()
    {
        Source pluginSource = null;
        Source expectation = null;
        
        try 
        {
            pluginSource = _srcResolver.resolveURI("cocoon://plugins/test/resources/js/i18ntextreader/input.en.js");
            expectation = _srcResolver.resolveURI("plugin:test://resources/js/i18ntextreader/expectation.js");
            
            try (InputStream is1 = pluginSource.getInputStream();
                 InputStream is2 = expectation.getInputStream();)
            {
                Assert.assertEquals(IOUtils.toString(is2, StandardCharsets.UTF_8), IOUtils.toString(is1, StandardCharsets.UTF_8));
            }
        }
        catch (IOException e)
        {
            throw new RuntimeException("I/O exception", e);
        }
        finally
        {
            _srcResolver.release(expectation);
            _srcResolver.release(pluginSource);
        }
    }
    
    /**
     * Test the i18n text reader at the limits 
     * limit1: file is ending with an escape sequence
     * limit2: file is ending with a '}' within an unfinished i18n declaration
     * limit3: file is ending with a '{' within an unfinished i18n declaration 
     */
    @Test
    public void testI18nTextReaderLimits()
    {
        Source pluginSource1 = null;
        Source pluginSource2 = null;
        Source pluginSource3 = null;
        
        try 
        {
            pluginSource1 = _srcResolver.resolveURI("cocoon://plugins/test/resources/js/i18ntextreader/limit1.js");
            pluginSource2 = _srcResolver.resolveURI("cocoon://plugins/test/resources/js/i18ntextreader/limit2.js");
            pluginSource3 = _srcResolver.resolveURI("cocoon://plugins/test/resources/js/i18ntextreader/limit3.js");
            
            try (InputStream is1 = pluginSource1.getInputStream();
                 InputStream is2 = pluginSource2.getInputStream();
                 InputStream is3 = pluginSource3.getInputStream();)
            {
                // Nothing
            }
        }
        catch (IOException e)
        {
            Assert.fail("Sould not throw an exception: " + e);
        }
        finally
        {
            _srcResolver.release(pluginSource1);
            _srcResolver.release(pluginSource2);
            _srcResolver.release(pluginSource3);
        }
    }
}
