/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.util;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.Assert;
import org.junit.Test;
import org.openjdk.jol.info.GraphLayout;

import org.ametys.core.util.SizeUtils;
import org.ametys.core.util.SizeUtils.ExcludeFromSizeCalculation;
import org.ametys.runtime.test.framework.AbstractBaseTestCase;

/**
 * Tests {@link SizeUtils}.
 */
public class SizeUtilsTestCase extends AbstractBaseTestCase
{
    /**
     * Tests {@link SizeUtils} by comparing to JOL results.
     */
    @Test
    public void testSizeUtils()
    {
        Assert.assertEquals("sizeOf(null) should return 0", 0, SizeUtils.sizeOf(null));
        
        TestObject object = new TestObject();
        long ametysSize = SizeUtils.sizeOf(object);
        long jolSize = GraphLayout.parseInstance(object).totalSize();
        Assert.assertEquals("Wrong size for TestObject", jolSize, ametysSize);
        
        OtherTestObject other = new OtherTestObject(object);
        ametysSize = SizeUtils.sizeOf(other);
        jolSize = GraphLayout.parseInstance(other).totalSize();
        Assert.assertEquals("Wrong size for OtherTestObject", jolSize, ametysSize);
        
        object._other = other;
        ametysSize = SizeUtils.sizeOf(object);
        jolSize = GraphLayout.parseInstance(object).totalSize();
        Assert.assertEquals("Wrong size for recursive TestObject", jolSize, ametysSize);

        TestObjectWithExcludedField excluded = new TestObjectWithExcludedField();
        ametysSize = SizeUtils.sizeOf(excluded);
        jolSize = GraphLayout.parseInstance(excluded).totalSize();
        Assert.assertNotEquals("Same size for TestObjectWithExcludedField", jolSize, ametysSize);

        TestObjectWithExcludedArray excludedArray = new TestObjectWithExcludedArray();
        ametysSize = SizeUtils.sizeOf(excludedArray);
        jolSize = GraphLayout.parseInstance(excludedArray).totalSize();
        Assert.assertNotEquals("Same size for TestObjectWithExcludedArray", jolSize, ametysSize);

        TestObjectWithClassField classClass = new TestObjectWithClassField();
        ametysSize = SizeUtils.sizeOf(classClass);
        jolSize = GraphLayout.parseInstance(classClass).totalSize();
        Assert.assertNotEquals("Same size for TestObjectWithClassField", jolSize, ametysSize);
        
        TestObjectWithClassNull classClassNull = new TestObjectWithClassNull();
        long jolSizeNull = GraphLayout.parseInstance(classClassNull).totalSize();
        Assert.assertEquals("Wrong size for TestObjectWithClassNull", jolSizeNull, ametysSize);

        TestObjectWithClassArray classArray = new TestObjectWithClassArray();
        ametysSize = SizeUtils.sizeOf(classArray);
        jolSize = GraphLayout.parseInstance(classArray).totalSize();
        Assert.assertNotEquals("Same size for TestObjectWithClassArray", jolSize, ametysSize);
        
        TestObjectWithClassArrayNull classArrayNull = new TestObjectWithClassArrayNull();
        jolSizeNull = GraphLayout.parseInstance(classArrayNull).totalSize();
        Assert.assertEquals("Wrong size for TestObjectWithClassArrayNull", jolSizeNull, ametysSize);
    }

    
    @SuppressWarnings("unused")
    private static class TestObject
    {
        private static Object _a = "myStaticString";
        
        private String _b = "ab";
        private String _c = "ba";
        private Map<String, List<String>> _map = new HashMap<>();
        
        private OtherTestObject _other;
        
        public TestObject()
        {
            _map.put("a", List.of("a", "b", "c"));
            _map.put("b", List.of("1", "2"));
            _map.put("c", null);
        }
    }
    
    @SuppressWarnings("unused")
    private static class OtherTestObject
    {
        private TestObject _object;
        
        public OtherTestObject(TestObject object)
        {
            _object = object;
        }
    }

    @SuppressWarnings("unused")
    private static class TestObjectWithExcludedField
    {
        @ExcludeFromSizeCalculation
        private String _a = "ab";
        
        private String _b = "ba";
    }

    @SuppressWarnings("unused")
    private static class TestObjectWithExcludedArray
    {
        @ExcludeFromSizeCalculation
        private String[] _a = {"ab", "cd"};
        
        private String _b = "ba";
    }
    
    @SuppressWarnings("unused")
    private static class TestObjectWithClassField
    {
        private Class _a = TestObject.class;

        private String _b = "ba";
    }
    
    @SuppressWarnings("unused")
    private static class TestObjectWithClassNull
    {
        private Class _a;

        private String _b = "ba";
    }

    @SuppressWarnings("unused")
    private static class TestObjectWithClassArray
    {
        private Class[] _a = {Class.class};
        
        private String _b = "ba";
    }

    @SuppressWarnings("unused")
    private static class TestObjectWithClassArrayNull
    {
        private Class[] _a = {null};
        
        private String _b = "ba";
    }
}
