/*
 *  Copyright 2020 Anyware Services
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.ametys.runtime.test.util;

import org.junit.Assert;
import org.junit.Test;

import org.ametys.core.util.URIUtils;
import org.ametys.runtime.test.framework.AbstractBaseTestCase;

import junit.framework.TestCase;

/**
 * {@link TestCase} for encoding or decodong URIs
 */
public class URIUtilsTestCase extends AbstractBaseTestCase
{
    private String[] _prepareDataForTest()
    {
        return new String[] {
            "value 1.pdf", // 0
            "value@!$2", // 1
            "value%3", // 2
            "value / 4", // 3
            "value=A&value=B", // 4
            "value : A", // 5
            "foo://example.com:8042/test", // 6
            "file:foo/bar", // 7
            "foo (foo)+foo%foo;foo:foo#foo?foo", // 8
            "àéçù" // 9
        };
    }
    
    /**
     * Test encode of header
     */
    @Test
    public void testEncodeHeader()
    {
        String[] data = _prepareDataForTest();
        
        Assert.assertEquals("value%201.pdf", URIUtils.encodeHeader(data[0]));
        Assert.assertEquals("value%40%21%242", URIUtils.encodeHeader(data[1]));
        Assert.assertEquals("value%253", URIUtils.encodeHeader(data[2]));
        Assert.assertEquals("value%20%2F%204", URIUtils.encodeHeader(data[3]));
        Assert.assertEquals("value%3DA%26value%3DB", URIUtils.encodeHeader(data[4]));
        Assert.assertEquals("value%20%3A%20A", URIUtils.encodeHeader(data[5]));
        Assert.assertEquals("foo%3A%2F%2Fexample.com%3A8042%2Ftest", URIUtils.encodeHeader(data[6]));
        Assert.assertEquals("file%3Afoo%2Fbar", URIUtils.encodeHeader(data[7]));
        Assert.assertEquals("foo%20%28foo%29%2Bfoo%25foo%3Bfoo%3Afoo%23foo%3Ffoo", URIUtils.encodeHeader(data[8]));
        Assert.assertEquals("%C3%A0%C3%A9%C3%A7%C3%B9", URIUtils.encodeHeader(data[9]));
    }
    
    /**
     * Test encode of parameter
     */
    @Test
    public void testEncodeParameter()
    {
        String[] data = _prepareDataForTest();
        
        Assert.assertEquals("value%201.pdf", URIUtils.encodeParameter(data[0]));
        Assert.assertEquals("value@!$2", URIUtils.encodeParameter(data[1]));
        Assert.assertEquals("value%253", URIUtils.encodeParameter(data[2]));
        Assert.assertEquals("value%20/%204", URIUtils.encodeParameter(data[3]));
        Assert.assertEquals("value%3DA%26value%3DB", URIUtils.encodeParameter(data[4]));
        Assert.assertEquals("value%20:%20A", URIUtils.encodeParameter(data[5]));
        Assert.assertEquals("foo://example.com:8042/test", URIUtils.encodeParameter(data[6]));
        Assert.assertEquals("file:foo/bar", URIUtils.encodeParameter(data[7]));
        Assert.assertEquals("foo%20(foo)%2Bfoo%25foo;foo:foo%23foo?foo", URIUtils.encodeParameter(data[8]));
        Assert.assertEquals("%C3%A0%C3%A9%C3%A7%C3%B9", URIUtils.encodeParameter(data[9]));
    }
    
    /**
     * Test encode of path
     */
    @Test
    public void testEncodePath()
    {
        String[] data = _prepareDataForTest();
        
        Assert.assertEquals("value%201.pdf", URIUtils.encodePath(data[0]));
        Assert.assertEquals("value@!$2", URIUtils.encodePath(data[1]));
        Assert.assertEquals("value%253", URIUtils.encodePath(data[2]));
        Assert.assertEquals("value%20/%204", URIUtils.encodePath(data[3]));
        Assert.assertEquals("value=A&value=B", URIUtils.encodePath(data[4]));
        Assert.assertEquals("value%20:%20A", URIUtils.encodePath(data[5]));
        Assert.assertEquals("foo://example.com:8042/test", URIUtils.encodePath(data[6]));
        Assert.assertEquals("file:foo/bar", URIUtils.encodePath(data[7]));
        Assert.assertEquals("foo%20(foo)+foo%25foo%3Bfoo:foo%23foo%3Ffoo", URIUtils.encodePath(data[8]));
        Assert.assertEquals("%C3%A0%C3%A9%C3%A7%C3%B9", URIUtils.encodePath(data[9]));
    }
    
    /**
     * Test encode of path segment
     */
    @Test
    public void testEncodePathSegment()
    {
        String[] data = _prepareDataForTest();
        
        Assert.assertEquals("value%201.pdf", URIUtils.encodePathSegment(data[0]));
        Assert.assertEquals("value@!$2", URIUtils.encodePathSegment(data[1]));
        Assert.assertEquals("value%253", URIUtils.encodePathSegment(data[2]));
        Assert.assertEquals("value%20%2F%204", URIUtils.encodePathSegment(data[3]));
        Assert.assertEquals("value=A&value=B", URIUtils.encodePathSegment(data[4]));
        Assert.assertEquals("value%20:%20A", URIUtils.encodePathSegment(data[5]));
        Assert.assertEquals("foo:%2F%2Fexample.com:8042%2Ftest", URIUtils.encodePathSegment(data[6]));
        Assert.assertEquals("file:foo%2Fbar", URIUtils.encodePathSegment(data[7]));
        Assert.assertEquals("foo%20(foo)+foo%25foo%3Bfoo:foo%23foo%3Ffoo", URIUtils.encodePathSegment(data[8]));
        Assert.assertEquals("%C3%A0%C3%A9%C3%A7%C3%B9", URIUtils.encodePathSegment(data[9]));
    }
    
    /**
     * Test decode
     */
    @Test
    public void testDecode()
    {
        String[] data = _prepareDataForTest();
        
        Assert.assertEquals(data[0], URIUtils.decode("value%201.pdf"));
        Assert.assertEquals(data[1], URIUtils.decode("value@!$2"));
        Assert.assertEquals(data[2], URIUtils.decode("value%253"));
        Assert.assertEquals(data[3], URIUtils.decode("value%20%2F%204"));
        Assert.assertEquals(data[4], URIUtils.decode("value=A&value=B"));
        Assert.assertEquals(data[5], URIUtils.decode("value%20:%20A"));
        Assert.assertEquals(data[6], URIUtils.decode("foo:%2F%2Fexample.com:8042%2Ftest"));
        Assert.assertEquals(data[7], URIUtils.decode("file:foo%2Fbar"));
        Assert.assertEquals(data[8], URIUtils.decode("foo%20(foo)+foo%25foo;foo:foo%23foo%3Ffoo"));
        Assert.assertEquals(data[9], URIUtils.decode("%C3%A0%C3%A9%C3%A7%C3%B9"));
    }
}
